#-------------------------------------------------------------------------------
# CMake helper module for find_package.
#
# Copyright (C) 2017  Frank Tackmann
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0.
#-------------------------------------------------------------------------------
#
# Defines the function
#
# find_package_helper (
#    <package>
#    [HEADERS <header> [<header2> ...] [INCLUDE_SUFFIXES <include_dirs...>]]
#    [VERSION_REGEX <regex> [VERSION_FILES <files...>]]
#    [LIBRARY <lib> [<lib2> ...]]
#    [COMPONENTS <libs...>]
#    [OPTIONAL_COMPONENTS <libs...>]
#    [LIB_SUFFIXES <lib_dirs...>]
#    [LANGUAGES <languages...>]
#    [EXECUTABLE <exec> [<exec2> ...] [BIN_SUFFIXES <bin_dirs...>]
#       [EXECUTABLE_REQUIRED] [EXECUTABLE_TARGET <exec_target>]]
#    [DATA_FILES <files...> [SHARE_SUFFIXES <data_dirs...>] [DATA_REQUIRED]]
#    [HELPMSG <errmsg>]
# )
#
# which can be used to write custom Findpackage.cmake scripts for find_package.
# In most cases it should be sufficient to just call find_package_helper in the
# Findpackage.cmake script. The helper uses the <package>_FIND_... variables set
# by find_package (see CMake documentation).
#
# The HEADERS if given are used with find_path to find the include dir. The
# optional INCLUDE_SUFFIXES are used as relative suffixes to the search or root
# path. If not specified it defaults to 'include'. Specifying INCLUDE_SUFFIXES
# without HEADERS is diagnosed as an error.
#
# If specified, the optional VERSION_REGEX is used to search the headers in the
# specified order line by line to extract the package version. The first match
# is used to set <package>_VERSION and checked against <package>_FIND_VERSION.
# If separate VERSION_FILES are specified, they are searched instead of the
# headers (if they exist). They are still interpreted relative to the found
# include dir of the headers.
#
# LIBRARY if given specifies the name of the main package library to search with
# find_library. Names of additional required and optional component libraries to
# search for are specified by COMPONENTS and OPTIONAL_COMPONENTS. These are
# merged with any user-specified components in <package>_FIND_COMPONENTS. The
# optional LIB_SUFFIXES are used as relative suffixes to the search or root
# path. If not specified, it defaults to 'lib'. Specifying LIB_SUFFIXES without
# any libraries to search for is diagnosed as an error.
#
# EXECUTABLE if given specifies the name of the main package executable and is
# searched for with find_program. The optional BIN_SUFFIXES are used as relative
# suffixes. If not specified, it defaults to 'bin'. The executable is treated
# as required if EXECUTABLE_REQUIRED is specified and otherwise as optional.
# If EXECUTABLE_TARGET is specified an imported target <package>::<exec_target>
# for the executable is defined.
#
# The DATA_FILES if given are used with find_path to search for the data
# directory. The optional SHARE_SUFFIXES are used as relative suffixes. If not
# specified, it defaults to 'share'. The first data file found determines the
# <package>_DATA_DIR. The data files are treated as required if DATA_REQUIRED is
# specified and otherwise as optional.
#
# HELPMSG can be used to specify an alternative help message which is displayed
# when some parts of the package are not found.
#
# The helper provides various diagnostic messages for each searched component.
# If VERSION_REGEX is not specified no version check is performed. In this case,
# one can use other means to identify the version of the found installation and
# check it using the function _find_package_check_version(<package>).
#
# If additional dependencies are not satisfied the function
# _find_package_error (<package> <errormsg> [<search info>])
# can be used to display appropriate error messages and clear <package>_FOUND.
#
# The following variables can be set by the user:
#  <package>_ROOT            : The root directory of the package installation (cached)
#  <package>_SEARCH_PATH     : Additional prefix paths to search (cached)
#  <package>_NO_DEFAULT_PATH : If set, skips searching the CMake default paths
#
# The following variables are set:
#  <package>_FOUND         : TRUE if all required parts are found
#
#  <package>_VERSION       : The found version
#  <package>_VERSION_MAJOR : The major version, if any
#  <package>_VERSION_MINOR : The minor version, if any
#  <package>_VERSION_PATCH : The patch version, if any
#  <package>_VERSION_TWEAK : The tweak version, if any
#
# The following imported targets are defined:
#  <package>::<package>     : The primary package target
#  <package>::<c>           : Target for the component library <c>
#  <package>::<exec_target> : Target for the package executable (if specified)
#
# The targets are only defined if <package>_FOUND is TRUE. Targets for optional
# components are only defined if they were found. The targets have their
# IMPORTED_LOCATION property set to the respective found library file and their
# INTERFACE_INCLUDE_DIRECTORIES property set to the found include dir (if any).
# If HEADERS without LIBRARY are specified, the main package target is defined
# as interface target. No INTERFACE_LINK_LIBRARIES for any of the targets are
# defined. Setting them appropriately to define possible dependencies between
# components is left to the calling Findpackage.cmake script. For convenience,
# if LANGUAGES is set such as "C" or "CXX", its value is used to set the
# IMPORTED_LINK_INTERFACE_LANGUAGES of all library targets.
#
# The results are stored in the following cache variables (as applicable):
#  <package>_LIBRARY     : The found library
#  <package>_<c>_LIBRARY : The found component library <c>
#  <package>_INCLUDE_DIR : The found include directory
#  <package>_EXECUTABLE  : The found executable
#  <package>_DATA_DIR    : The found data directory
#
# Note that since we provide imported targets, the traditional variables
# <package>_LIBRARIES and <package>_INCLUDE_DIRS are not defined any longer. If
# desired or needed for backward compatibility, the calling Findpackage.cmake
# script must define them based on the search results.
#
# There are two search modes, depending on whether <package>_ROOT is set or not.
#
# If <package>_ROOT is specified, all package components are searched starting
# from this prefix with appropriate suffixes. In this case, no other prefix
# paths are searched. Whenever <package>_ROOT changes, previous search results
# are discarded (including when <package>_ROOT is unset). This allows the user
# to explicitly force using a specific installation prefix.
#
# Otherwise, the search proceeds as follows:
# 1. No search is performed if <package>_LIBRARY, <package>_INCLUDE_DIR, etc.
#    are present from a previous search (or have been manually set by the user,
#    but see below).
# 2. The default system paths according to CMake are searched (see the CMake
#    documentation) unless <package>_NO_DEFAULT_PATH is set.
# 3. The additional prefix paths in <package>_SEARCH_PATH are searched.
#
# Only if all required parts are found and the deduced version is acceptable,
# the search results are stored and not searched again. If some required parts
# are not found, any partial search results are cleared and not stored.
#
# To completely bypass any searching or override the found location, the user
# can directly set the <package>_LIBRARY, <package>_INCLUDE_DIR, etc. cache
# variables. This should only be used as a last resort in case of nonstandard
# or conflicting installations where specifying <package>_SEARCH_PATH or
# <package>_ROOT is insufficient. In this case the headers are still checked for
# a compatible version but otherwise it is up to the user to ensure that the
# library and headers exist and are consistent with each other.
#-------------------------------------------------------------------------------

include (CMakeParseArguments)

function (find_package_helper package)
   # parse arguments
   set (options EXECUTABLE_REQUIRED DATA_REQUIRED)
   set (singleValueArgs LIBRARY VERSION_REGEX EXECUTABLE_TARGET HELPMSG)
   set (multiValueArgs
      HEADERS INCLUDE_SUFFIXES VERSION_FILES
      COMPONENTS OPTIONAL_COMPONENTS LIB_SUFFIXES
      EXECUTABLE BIN_SUFFIXES
      DATA_FILES SHARE_SUFFIXES
      LANGUAGES
   )
   cmake_parse_arguments (package "${options}" "${singleValueArgs}" "${multiValueArgs}" ${ARGN})

   if (NOT package_HEADERS AND package_INCLUDE_SUFFIXES)
      message (FATAL_ERROR "Specified INCLUDE_SUFFIXES but no HEADERS for package ${package}")
   endif ()
   if (NOT package_INCLUDE_SUFFIXES)
      set (package_INCLUDE_SUFFIXES "include")
   endif ()

   if (NOT package_VERSION_FILES)
      set (package_VERSION_FILES ${package_HEADERS})
   endif ()

   list (APPEND ${package}_FIND_COMPONENTS ${package_COMPONENTS} ${package_OPTIONAL_COMPONENTS})
   list (REMOVE_DUPLICATES ${package}_FIND_COMPONENTS)
   foreach (lib IN LISTS package_COMPONENTS)
      set (${package}_FIND_REQUIRED_${lib} TRUE)
   endforeach ()

   if (NOT package_LIBRARY AND NOT ${package}_FIND_COMPONENTS AND package_LIB_SUFFIXES)
      message (FATAL_ERROR "Specified LIB_SUFFIXES but there are no libraries to search for package ${package}")
   endif ()
   if (NOT package_LIB_SUFFIXES)
      set (package_LIB_SUFFIXES "lib")
   endif ()

   if (NOT package_EXECUTABLE)
       if (package_BIN_SUFFIXES)
         message (FATAL_ERROR "Specified BIN_SUFFIXES but no EXECUTABLE for package ${package}")
      endif ()
      if (package_EXECUTABLE_REQUIRED)
         message (FATAL_ERROR "Specified EXECUTABLE_REQUIRED but no EXECUTABLE for package ${package}")
      endif ()
      if (package_EXECUTABLE_TARGET)
         message (FATAL_ERROR "Specified EXECUTABLE_TARGET but no EXECUTABLE for package ${package}")
      endif ()
   endif ()
   if (NOT package_BIN_SUFFIXES)
      set (package_BIN_SUFFIXES "bin")
   endif ()

   if (NOT package_DATA_FILES)
       if (package_SHARE_SUFFIXES)
         message (FATAL_ERROR "Specified SHARE_SUFFIXES but no DATA_FILES for package ${package}")
      endif ()
       if (package_DATA_REQUIRED)
         message (FATAL_ERROR "Specified DATA_REQUIRED but no DATA_FILES for package ${package}")
      endif ()
   endif ()
   if (NOT package_SHARE_SUFFIXES)
      set (package_SHARE_SUFFIXES "share")
   endif ()

   if (NOT package_HELPMSG)
      set (package_HELPMSG ">> Please specify the ${package} installation path (e.g. by passing -D${package}_ROOT=path/to/${package} to cmake).")
   endif ()

   # setup search paths
   set (${package}_SEARCH_PATH CACHE PATH "List of additional prefix paths to search for the ${package} installation.")

   set (${package}_ROOT CACHE PATH "Root (prefix) directory for the ${package} installation.")
   if (NOT "${${package}_ROOT}" STREQUAL "${${package}_ROOT_LAST}")
      set (${package}_ROOT_LAST "${${package}_ROOT}" CACHE INTERNAL "Last used value of ${package}_ROOT.")
      unset (${package}_LIBRARY CACHE)
      foreach (lib ${${package}_FIND_COMPONENTS})
         unset (${package}_${lib}_LIBRARY CACHE)
      endforeach ()
      unset (${package}_INCLUDE_DIR CACHE)
      unset (${package}_EXECUTABLE CACHE)
      unset (${package}_DATA_DIR CACHE)
   endif ()

   if (${package}_ROOT)
      set (lib_search PATHS "${${package}_ROOT}" PATH_SUFFIXES ${package_LIB_SUFFIXES} NO_DEFAULT_PATH)
      set (include_search PATHS "${${package}_ROOT}" PATH_SUFFIXES ${package_INCLUDE_SUFFIXES} NO_DEFAULT_PATH)
      set (bin_search PATHS "${${package}_ROOT}" PATH_SUFFIXES ${package_BIN_SUFFIXES} NO_DEFAULT_PATH)
      set (data_search PATHS "${${package}_ROOT}" PATH_SUFFIXES ${package_SHARE_SUFFIXES} NO_DEFAULT_PATH)
   else ()
      set (lib_search PATHS ${${package}_SEARCH_PATH} PATH_SUFFIXES ${package_LIB_SUFFIXES})
      set (include_search PATHS ${${package}_SEARCH_PATH} PATH_SUFFIXES ${package_INCLUDE_SUFFIXES})
      set (bin_search PATHS ${${package}_SEARCH_PATH} PATH_SUFFIXES ${package_BIN_SUFFIXES})
      set (data_search PATHS ${${package}_SEARCH_PATH} PATH_SUFFIXES ${package_SHARE_SUFFIXES})
      if (${package}_NO_DEFAULT_PATH)
         list (APPEND lib_search NO_DEFAULT_PATH)
         list (APPEND include_search NO_DEFAULT_PATH)
         list (APPEND bin_search NO_DEFAULT_PATH)
         list (APPEND data_search NO_DEFAULT_PATH)
      endif ()
   endif ()

   # start with the assumption that everything was found
   set (${package}_FOUND TRUE)

   # search the main library
   if (package_LIBRARY)
      find_library (${package}_LIBRARY NAMES ${package_LIBRARY} ${lib_search} DOC "${package} library.")
      mark_as_advanced (${package}_LIBRARY)
      if (${package}_LIBRARY)
         if (NOT ${package}_FIND_QUIETLY)
            message (STATUS  "Found ${package} library: ${${package}_LIBRARY}")
         endif ()
      else ()
         _find_package_error (${package} "Could NOT find ${package} library." ${lib_search})
      endif ()
   endif ()

   # search for all additional components
   foreach (lib IN LISTS ${package}_FIND_COMPONENTS)
      find_library (${package}_${lib}_LIBRARY NAMES ${lib} ${lib_search} DOC "${lib} component library of ${package}.")
      mark_as_advanced (${package}_${lib}_LIBRARY)
      if (${package}_${lib}_LIBRARY)
         list (APPEND libs ${lib})
         if (NOT ${package}_FIND_QUIETLY)
            message (STATUS "Found ${package} component library ${lib}: ${${package}_${lib}_LIBRARY}")
         endif ()
      else ()
         if (${package}_FIND_REQUIRED_${lib})
            _find_package_error (${package} "Could NOT find requested ${package} component library ${lib}." ${lib_search})
         else ()
            if (NOT ${package}_FIND_QUIETLY)
               message (STATUS "Could NOT find optional ${package} component library ${lib} -- proceeding without.")
            endif ()
         endif ()
      endif ()
   endforeach ()

   # search the headers
   if (package_HEADERS)
      find_path (${package}_INCLUDE_DIR NAMES ${package_HEADERS} ${include_search} DOC "${package} include directory.")
      mark_as_advanced (${package}_INCLUDE_DIR)
      if (${package}_INCLUDE_DIR)
         if (NOT ${package}_FIND_QUIETLY)
            message (STATUS  "Found ${package} headers: ${${package}_INCLUDE_DIR}")
         endif ()

         # try to extract version info if VERSION_REGEX was given
         if (package_VERSION_REGEX)
            set (${package}_VERSION "(unknown)")
            foreach (file ${package_VERSION_FILES})
               if (EXISTS "${${package}_INCLUDE_DIR}/${file}")
                  file (STRINGS "${${package}_INCLUDE_DIR}/${file}" version_string REGEX "${package_VERSION_REGEX}")
                  if ("${version_string}" MATCHES "${package_VERSION_REGEX}")
                     set (${package}_VERSION "${CMAKE_MATCH_1}")
                     break ()
                  endif ()
               endif ()
            endforeach ()
            string (REPLACE "." ";" version_list "${${package}_VERSION}...")
            list (GET version_list 0 ${package}_VERSION_MAJOR)
            list (GET version_list 1 ${package}_VERSION_MINOR)
            list (GET version_list 2 ${package}_VERSION_PATCH)
            list (GET version_list 3 ${package}_VERSION_TWEAK)
            _find_package_check_version (${package})
         endif ()
      else ()
         _find_package_error (${package} "Could NOT find ${package} headers." ${include_search})
      endif ()
   endif ()

   # search for executable
   if (package_EXECUTABLE)
      find_program (${package}_EXECUTABLE NAMES ${package_EXECUTABLE} ${bin_search} DOC "${package} executable.")
      mark_as_advanced (${package}_EXECUTABLE)
      if (${package}_EXECUTABLE)
         if (NOT ${package}_FIND_QUIETLY)
            message (STATUS  "Found ${package} executable: ${${package}_EXECUTABLE}")
         endif ()
      else ()
         if (package_EXECUTABLE_REQUIRED)
            _find_package_error (${package} "Could NOT find requested ${package} executable ${package_EXECUTABLE}." ${bin_search})
         else ()
            if (NOT ${package}_FIND_QUIETLY)
               message (STATUS "Could NOT find optional ${package} executable ${package_EXECUTABLE} -- proceeding without.")
            endif ()
         endif ()
      endif ()
   endif ()

   # search for data files
   if (package_DATA_FILES)
      find_path (${package}_DATA_DIR NAMES ${package_DATA_FILES} ${data_search} DOC "${package} data directory.")
      mark_as_advanced (${package}_DATA_DIR)
      if (${package}_DATA_DIR)
         if (NOT ${package}_FIND_QUIETLY)
            message (STATUS  "Found ${package} data files: ${${package}_DATA_DIR}")
         endif ()
      else ()
         if (package_DATA_REQUIRED)
            _find_package_error (${package} "Could NOT find requested ${package} data files." ${data_search})
         else ()
            if (NOT ${package}_FIND_QUIETLY)
               message (STATUS "Could NOT find optional ${package} data files -- proceeding without.")
            endif ()
         endif ()
      endif ()
   endif ()

   # finalize
   if (${package}_FOUND)
      # export variables to parent scope
      set (${package}_FOUND TRUE PARENT_SCOPE)
      set (${package}_VERSION ${${package}_VERSION} PARENT_SCOPE)
      set (${package}_VERSION_MAJOR ${${package}_VERSION_MAJOR} PARENT_SCOPE)
      set (${package}_VERSION_MINOR ${${package}_VERSION_MINOR} PARENT_SCOPE)
      set (${package}_VERSION_PATCH ${${package}_VERSION_PATCH} PARENT_SCOPE)
      set (${package}_VERSION_TWEAK ${${package}_VERSION_TWEAK} PARENT_SCOPE)

      # define imported targets
      if (NOT TARGET ${package}::${package})
         if (${package}_LIBRARY)
            add_library (${package}::${package} UNKNOWN IMPORTED)
            set_target_properties (${package}::${package} PROPERTIES IMPORTED_LOCATION "${${package}_LIBRARY}")
            list (APPEND lib_targets ${package}::${package})
         elseif (${package}_INCLUDE_DIR)
            add_library (${package}::${package} INTERFACE IMPORTED)
            set_target_properties (${package}::${package} PROPERTIES INTERFACE_INCLUDE_DIRECTORIES "${${package}_INCLUDE_DIR}")
         endif ()
      endif ()
      foreach (lib IN LISTS libs)
         if (NOT TARGET ${package}::${lib})
            add_library (${package}::${lib} UNKNOWN IMPORTED)
            set_target_properties (${package}::${lib} PROPERTIES IMPORTED_LOCATION "${${package}_${lib}_LIBRARY}")
            list (APPEND lib_targets ${package}::${lib})
         endif ()
      endforeach ()
      if (${package}_INCLUDE_DIR)
         set_target_properties (${lib_targets} PROPERTIES INTERFACE_INCLUDE_DIRECTORIES "${${package}_INCLUDE_DIR}")
      endif ()
      if (package_LANGUAGES)
         set_target_properties (${lib_targets} PROPERTIES IMPORTED_LINK_INTERFACE_LANGUAGES "${package_LANGUAGES}")
      endif ()

      if (${package}_EXECUTABLE AND package_EXECUTABLE_TARGET)
         if (NOT TARGET ${package}::${package_EXECUTABLE_TARGET})
            add_executable (${package}::${package_EXECUTABLE_TARGET} IMPORTED)
            set_target_properties (${package}::${package_EXECUTABLE_TARGET} PROPERTIES IMPORTED_LOCATION "${${package}_EXECUTABLE}")
         endif ()
      endif ()

      if (NOT ${package}_FIND_QUIETLY)
         if (${package}_FIND_VERSION)
            if (${package}_FIND_VERSION_EXACT)
               set (required "exact")
            else ()
               set (required "minimum")
            endif ()
            message (STATUS "Found suitable ${package} version: ${${package}_VERSION} (requested was ${required} version ${${package}_FIND_VERSION})")
         elseif (${package}_VERSION)
            message (STATUS "Found ${package} version: ${${package}_VERSION}")
         endif ()
      endif ()
   else ()
      # clear partial search results to help ensure consistency
      unset (${package}_LIBRARY CACHE)
      foreach (lib ${${package}_FIND_COMPONENTS})
         unset (${package}_${lib}_LIBRARY CACHE)
      endforeach ()
      unset (${package}_INCLUDE_DIR CACHE)
      unset (${package}_EXECUTABLE CACHE)
      unset (${package}_DATA_DIR CACHE)
      if (NOT ${package}_FIND_QUIETLY)
         message (${package_HELPMSG})
      endif ()
      if (${package}_FIND_REQUIRED)
         message (FATAL_ERROR "Could NOT find required ${package}.")
      endif ()
   endif ()
endfunction ()


#-------------------------------------------------------------------------------
# Private helper functions
#-------------------------------------------------------------------------------

function (_find_package_error package errmsg)
   set (${package}_FOUND FALSE PARENT_SCOPE)
   if (NOT ${package}_FIND_QUIETLY)
      if (${package}_FIND_REQUIRED)
         set (tag "EE")
      else ()
         set (tag "ww")
      endif ()
      message ("${tag} ${errmsg}")
      if (${ARGC} GREATER 3)
         string (REPLACE ";" " " out "${ARGN}")
         message ("${tag} Searched with '${out}'.")
      endif ()
   endif ()
endfunction ()

function (_find_package_check_version package)
   if (${package}_FOUND)
      if (${package}_VERSION VERSION_LESS ${package}_FIND_VERSION)
         set (age "old")
      elseif (${package}_FIND_VERSION_EXACT AND NOT ${package}_VERSION VERSION_EQUAL ${package}_FIND_VERSION)
         set (age "new")
      endif ()
      if (age)
         if (${package}_FIND_VERSION_EXACT)
            set (required "exact")
         else ()
            set (required "minimum")
         endif ()
         _find_package_error (${package} "Detected ${package} version ${${package}_VERSION} is too ${age} (requested was ${required} version ${${package}_FIND_VERSION})")
         set (${package}_FOUND FALSE PARENT_SCOPE)
      endif ()
   endif ()
endfunction ()
