#-------------------------------------------------------------------------------
# CMake module to find the gsl library and headers.
#
# Copyright (C) 2017  Frank Tackmann
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0.
#-------------------------------------------------------------------------------
#
# The following variables can be set by the user:
#  gsl_ROOT            : The root directory of the gsl installation (cache)
#  gsl_SEARCH_PATH     : Additional prefix paths to search (cache)
#  gsl_NO_DEFAULT_PATH : If set, skips searching the CMake default paths
#
# The following imported targets are defined:
#  gsl::gsl      : The gsl library
#  gsl::gslcblas : The gslcblas library (if gslcblas was given as COMPONENT)
#
# The following variables are set:
#  gsl_FOUND        : TRUE if library, headers, and all required components are found
#  gsl_VERSION      : The found version
#
#  gsl_LIBRARY          : Cache variable storing the found library
#  gsl_gslcblas_LIBRARY : Cache variable storing the found gslcblas library
#                         (if gslcblas was given as COMPONENT in find_package)
#  gsl_INCLUDE_DIR      : Cache variable storing the found include directory
#
# There are two search modes, depending on whether gsl_ROOT is set or not.
#
# If gsl_ROOT is specified, the library and headers are searched in
# gsl_ROOT/lib and gsl_ROOT/include. No other paths are searched.
# Whenever gsl_ROOT changes, previous search results are discarded
# (including when gsl_ROOT is unset). This allows the user to explicitly
# force using a specific installation.
#
# Otherwise, the search proceeds as follows:
# 1. No search is performed if gsl_LIBRARY or gsl_INCLUDE_DIR are
#    present from a previous search (or manually set by the user, but see below)
# 2. The default system paths according to CMake are searched (see the CMake
#    documentation) unless gsl_NO_DEFAULT_PATH is set.
# 3. The additional prefix paths in gsl_SEARCH_PATH are searched.
#
# Once both library and headers with an acceptable version are found, their
# location is stored and not searched again.
#
# To completely bypass any searching or override the found location, the user
# can directly set the gsl_LIBRARY and gsl_INCLUDE_DIR cache
# variables. This should only be used as a last resort in case of nonstandard
# or conflicting installations where specifying gsl_SEARCH_PATH or
# gsl_ROOT is insufficient. In this case the headers are still checked for
# a compatible version but otherwise it is up to the user to ensure that the
# library and headers exist and are consistent with each other.
#-------------------------------------------------------------------------------

include (FindPackageHelper)

find_package_helper (gsl
   LIBRARY gsl
   HEADERS "gsl/gsl_version.h"
   VERSION_REGEX ".*#define GSL_VERSION \"([0-9\\.]+)\"$"
   LANGUAGES "C"
)

if (TARGET gsl::gslcblas)
   set_target_properties (gsl::gsl PROPERTIES INTERFACE_LINK_LIBRARIES gsl::gslcblas)
endif ()
