#-------------------------------------------------------------------------------
# CMake module to find the lhapdf library and headers.
#
# Copyright (C) 2017  Frank Tackmann
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0.
#-------------------------------------------------------------------------------
#
# The following variables can be set by the user:
#  lhapdf_ROOT            : The root directory of the LHAPDF installation (cache)
#  lhapdf_SEARCH_PATH     : Additional prefix paths to search (cache)
#  lhapdf_NO_DEFAULT_PATH : If set, skips searching the CMake default paths
#
# The following imported targets are defined:
#  lhapdf::lhapdf: The lhapdf library
#
# The following variables are set:
#  lhapdf_FOUND        : TRUE if library, headers, and all required components are found
#  lhapdf_VERSION      : The found version
#
#  lhapdf_LIBRARY      : Cache variable storing the found library
#  lhapdf_INCLUDE_DIR  : Cache variable storing the found include directory
#
# There are two search modes, depending on whether lhapdf_ROOT is set or not.
#
# If lhapdf_ROOT is specified, the library and headers are searched in
# lhapdf_ROOT/lib and lhapdf_ROOT/include. No other paths are searched.
# Whenever lhapdf_ROOT changes, previous search results are discarded
# (including when lhapdf_ROOT is unset). This allows the user to explicitly
# force using a specific installation.
#
# Otherwise, the search proceeds as follows:
# 1. No search is performed if lhapdf_LIBRARY or lhapdf_INCLUDE_DIR are
#    present from a previous search (or manually set by the user, but see below)
# 2. The default system paths according to CMake are searched (see the CMake
#    documentation) unless lhapdf_NO_DEFAULT_PATH is set.
# 3. The additional prefix paths in lhapdf_SEARCH_PATH are searched.
#
# Once both library and headers with an acceptable version are found, their
# location is stored and not searched again.
#
# To completely bypass any searching or override the found location, the user
# can directly set the lhapdf_LIBRARY and lhapdf_INCLUDE_DIR cache
# variables. This should only be used as a last resort in case of nonstandard
# or conflicting installations where specifying lhapdf_SEARCH_PATH or
# lhapdf_ROOT is insufficient. In this case the headers are still checked for
# a compatible version but otherwise it is up to the user to ensure that the
# library and headers exist and are consistent with each other.
#
# Versions 5 and 6 of lhapdf are considered different. By default version 6
# headers are searched for, ignoring version 5 headers. To search for version 5,
# a corresponding version must be specified via find_package, in which case
# version 6 headers are ignored and not found. At present, there is no way to
# identify the version of the library itself. This means that in principle it
# can happen that headers of the correct version are found while finding the
# wrong library version. This happens for example if the search path contains
# installations of both versions and the wrong version is seen first, which
# should thus be avoided. A specific installation can always be enforced using
# lhapdf_ROOT.
#
# If lhapdf versions 6.0 or 6.1 are found, the Boost headers (version 1.41 or
# higher) are searched for as well. If the Boost headers are not found,
# lhapdf_FOUND will be false, but the other variables will still be set. This is
# the minimal typical requirement for 6.1.x versions, older versions may have
# additional Boost requirements, which are not checked. For versions >= 6.2,
# Boost is not anymore needed and not searched for.
#
# Note that we're not searching for the data path since that is handled by the
# library itself.
#-------------------------------------------------------------------------------

include (FindPackageHelper)

if (lhapdf_FIND_VERSION AND lhapdf_FIND_VERSION VERSION_LESS 6)
   set (_lhapdf_header "LHAPDF/LHAPDFConfig.h")
else ()
   set (_lhapdf_header "LHAPDF/Version.h")
endif ()

find_package_helper (lhapdf
   LIBRARY LHAPDF
   HEADERS ${_lhapdf_header}
   VERSION_REGEX ".*#define LHAPDF_VERSION \"([0-9\\.]+)\"$"
)

unset (_lhapdf_header)

# find boost (lhapdf6 requires version >= 1.41)
if (TARGET lhapdf::lhapdf AND lhapdf_VERSION_MAJOR EQUAL 6 AND lhapdf_VERSION_MINOR LESS 2)
   find_package (Boost QUIET)
   # without asking for components Boost_FOUND is based on headers only
   if (Boost_FOUND)
      if (NOT lhapdf_FIND_QUIETLY)
         message (STATUS "Found Boost headers: ${Boost_INCLUDE_DIR}")
      endif ()
      set (Boost_VERSION_STRING ${Boost_VERSION})
      set (Boost_VERSION "${Boost_MAJOR_VERSION}.${Boost_MINOR_VERSION}.${Boost_SUBMINOR_VERSION}")
      set (Boost_FIND_VERSION "1.41")
      _find_package_check_version (Boost)
      unset (Boost_FIND_VERSION)
   endif ()
   if (Boost_FOUND)
      if (NOT TARGET Boost::boost)
         add_library (Boost::boost INTERFACE IMPORTED)
         set_target_properties (Boost::boost PROPERTIES INTERFACE_INCLUDE_DIRECTORIES "${Boost_INCLUDE_DIRS}")
      endif ()
      set_target_properties (lhapdf::lhapdf PROPERTIES INTERFACE_LINK_LIBRARIES Boost::boost)
      if (NOT lhapdf_FIND_QUIETLY)
         message (STATUS "Found suitable Boost version: ${Boost_VERSION} (lhapdf ${lhapdf_VERSION} requires minimum version 1.41)")
      endif ()
   else ()
      _find_package_error (lhapdf "Could NOT find Boost headers required by lhapdf.")
      if (NOT lhapdf_FIND_QUIETLY)
         message (">> Please specify the Boost installation path (e.g. by passing -DBOOST_ROOT=path/to/boost to cmake).")
      endif ()
      if (lhapdf_FIND_REQUIRED)
         message (FATAL_ERROR "Could NOT find required dependence for lhapdf ${lhapdf_VERSION}: Boost")
      endif ()
   endif ()
endif ()
