#-------------------------------------------------------------------------------
# CMake module to find the Rivet library and headers.
#
# Copyright (C) 2018  Frank Tackmann, Simone Alioli
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0.
#-------------------------------------------------------------------------------
#
# The following variables can be set by the user:
#  rivet_ROOT            : The root directory of the Rivet installation (cache)
#  rivet_SEARCH_PATH     : Additional prefix paths to search (cache)
#  rivet_NO_DEFAULT_PATH : If set, skips searching the CMake default paths
#
# The following imported targets are defined:
#  rivet::rivet: The rivet library
#
# The following variables are set:
#  rivet_FOUND        : TRUE if library and headers and all dependencies are found
#  rivet_VERSION      : The found version
#
#  rivet_LIBRARY      : Cache variable storing the found library
#  rivet_INCLUDE_DIR  : Cache variable storing the found include directory
#  rivet_DATA_DIR     : Cache variable storing the found data directory
#
# There are two search modes, depending on whether rivet_ROOT is set or not.
#
# If rivet_ROOT is specified, the library and headers are searched in
# rivet_ROOT/lib and rivet_ROOT/include. No other paths are searched.
# Whenever rivet_ROOT changes, previous search results are discarded
# (including when rivet_ROOT is unset). This allows the user to explicitly
# force using a specific installation.
#
# Otherwise, the search proceeds as follows:
# 1. No search is performed if rivet_LIBRARY or rivet_INCLUDE_DIR are
#    present from a previous search (or manually set by the user, but see below)
# 2. The default system paths according to CMake are searched (see the CMake
#    documentation) unless rivet_NO_DEFAULT_PATH is set.
# 3. The additional paths in rivet_SEARCH_PATH are searched with respective
#    suffixes 'lib' and 'include'
#
# Once both library and headers with an acceptable version are found, their
# location is stored and not searched again.
#
# In addition, the rivet data files are searched for in the location
# share/Rivet/ relative to either rivet_ROOT or the search paths.
# They are treated as required.
#
# To completely bypass any searching or override the found location, the user
# can directly set the rivet_LIBRARY and rivet_INCLUDE_DIR cache
# variables. This should only be used as a last resort in case of nonstandard
# or conflicting installations where specifying rivet_SEARCH_PATH or
# rivet_ROOT is insufficient. In this case the headers are still checked for
# a compatible version but otherwise it is up to the user to ensure that the
# library and headers exist and are consistent with each other.
#
# Rivet depends on the yoda, hepmc, and fastjet packages. These are searched for
# as well even if Rivet itself was not found in order to let the user know what
# else is missing. If the yoda_FOUND, hepmc_FOUND, or fastjet_FOUND variables
# are already defined, they are not searched for again. This allows callers to
# manually search for them in a more specific way and explicitly define these
# variables to be FALSE to avoid searching again. If the dependencies are found,
# they are added as public link dependencies to the rivet::rivet target. If some
# dependencies are not found, the rivet::rivet target is still defined, but
# rivet_FOUND is set to FALSE.
#-------------------------------------------------------------------------------

include (FindPackageHelper)

find_package_helper (rivet
   LIBRARY Rivet
   HEADERS "Rivet/Rivet.hh"
   DATA_FILES "MC_XS.info" SHARE_SUFFIXES "share/Rivet" "Rivet"
   VERSION_FILES "Rivet/Config/RivetConfig.hh"
   VERSION_REGEX ".*#define RIVET_VERSION \"([0-9\\.]+)\"$"
)

# forward QUIET and REQUIRED arguments
if (rivet_FIND_QUIETLY)
   set (find_arg QUIET)
endif ()
if (rivet_FIND_REQUIRED)
   set (find_arg ${find_arg} REQUIRED)
endif ()

if (NOT DEFINED yoda_FOUND)
   find_package (yoda ${find_arg})
endif ()
if (NOT TARGET yoda::yoda)
   set (msg "Could NOT find required dependence for rivet ${rivet_VERSION}: yoda")
   _find_package_error (rivet ${msg})
   if (rivet_FIND_REQUIRED)
      message (FATAL_ERROR ${msg})
   endif ()
endif ()

if (NOT DEFINED hepmc_FOUND)
   find_package (hepmc ${find_arg})
endif ()
if (NOT TARGET hepmc::hepmc)
   set (msg "Could NOT find required dependence for rivet ${rivet_VERSION}: hepmc")
   _find_package_error (rivet ${msg})
   if (rivet_FIND_REQUIRED)
      message (FATAL_ERROR ${msg})
   endif ()
endif ()

if (NOT DEFINED fastjet_FOUND)
   find_package (fastjet ${find_arg})
endif ()
if (NOT TARGET fastjet::fastjet)
   set (msg "Could NOT find required dependence for rivet ${rivet_VERSION}: fastjet")
   _find_package_error (rivet ${msg})
   if (rivet_FIND_REQUIRED)
      message (FATAL_ERROR ${msg})
   endif ()
endif ()

if (TARGET rivet::rivet)
   get_filename_component (rivet_LIB_DIR ${rivet_LIBRARY} PATH)
   set_property (TARGET rivet::rivet APPEND PROPERTY INTERFACE_LINK_LIBRARIES yoda::yoda)
   set_property (TARGET rivet::rivet APPEND PROPERTY INTERFACE_LINK_LIBRARIES hepmc::hepmc)
   set_property (TARGET rivet::rivet APPEND PROPERTY INTERFACE_LINK_LIBRARIES fastjet::fastjet)
endif ()
