#-------------------------------------------------------------------------------
# CMake module for checking C++ language and library features.
#
# Copyright (C) 2018  Frank Tackmann
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0.
#-------------------------------------------------------------------------------
#
# Provides functions to explicitly check for C++11/14/17 features using the
# officially recommended feature test macros from SD-6:
# https://isocpp.org/std/standing-documents/sd-6-sg10-feature-test-recommendations
#
# All functions behave similar in spirit to find_package() but for checking the
# availability of C++ features. They perform an active check (using try_compile)
# at configuration time and set a specific variable to TRUE or FALSE depending
# on whether a requested feature is available or not. As usual, the result is
# cached internally and only done once. To force redoing a check, the cache must
# be cleared (which is  anyway required to change the compiler).
#
# The functions explicitly honor the current settings for the variables
# CMAKE_CXX_STANDARD, CMAKE_CXX_STANDARD_REQUIRED, and CMAKE_CXX_EXTENSIONS.
# These are used by CMake to determine the appropriate compiler flags for the
# given standard and the current compiler and must be set to their desired
# values before calling the check functions provided here.
#
# Note that this module is orthogonal to CMake's built-in cmake-compile-features
# system, which uses target_compile_features to require certain features for a
# target, which CMake uses to compute the necessary C++ standard and flags. The
# problem with this approach is that it is not easy to require a specific
# C++ standard (or feature) to be available up front, since the diagnosis might
# only happen at compile time, the available features are not in one-to-one
# correspondence with SD-6, and it relies on CMake knowing a given compiler
# version and its supported features.
#
# The following functions are defined (with <std> = 11, 14, or 17):
#
# check_cxx<std>_support ([QUIET])
#
# Checks for C++11/14/17 support by checking the __cplusplus macro. The result
# is stored in the CXX<std>_lanuage_support variable. The result is also used
# to set the initial values for the CXX<std>_SUPPORT and
# CXX<std>_PARTIAL_SUPPORT variables, so usually this function is called before
# checking for individual features.
#
# check_cxx<std>_feature (<feature> <version> [QUIET])
#
# Checks if the core language <feature> is available at the given <version>.
# (i.e. if __cpp_<feature> >= <version>). The result is stored in the
# CXX<std>_<feature> variable. If the result is TRUE, also the
# CXX<std>_PARTIAL_SUPPORT variable is set to TRUE, while if FALSE, the
# CXX<std>_SUPPORT variable is set to FALSE.
#
# check_cxx<std>_lib_feature(<header> <feature> <version> [QUIET])
#
# Checks if the std library <feature> is available at the given <version>.
# (i.e. if __cpp_lib_<feature> >= <version>), where <header> is the name of the
# std header that must be included for <feature>. The result is stored in the
# CXX<std>_lib_<feature> variable.  If the result is TRUE, also the
# CXX<std>_PARTIAL_SUPPORT variable is set to TRUE, while if FALSE, the
# CXX<std>_SUPPORT variable is set to FALSE.
#
# Examples:
#  check_cxx11_feature (constexpr 200704) # checks for C++11 constexpr
#  check_cxx14_feature (constexpr 201304) # checks for C++14 constexpr
#
#  # checks for C++14 integer_sequence
#  check_cxx14_lib_feature (utility integer_sequence 201304)
#
# The relevant values for <header>, <feature>, <version> can be found on the
# above webpage. Note that the functions for different <std> all perform exactly
# the same check and are agnostic to whether a feature actually belongs to that
# version of the C++ standard. Their only difference is in the message they
# print and the names of the variables they set.
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
# Functions for C++11
#-------------------------------------------------------------------------------

# checks for C++11 support
function (check_cxx11_support)
   _check_cxx_feature_macro (
      "C++11 language support"
      ""
      "__cplusplus"
      201103L
      CXX11_language_support
      ${ARGV0}
   )
   set (CXX11_SUPPORT ${CXX11_language_support} PARENT_SCOPE)
   set (CXX11_PARTIAL_SUPPORT ${CXX11_language_support} PARENT_SCOPE)
endfunction ()

# checks for a specific C++11 language feature
function (check_cxx11_feature feature version)
   _check_cxx_feature_macro (
      "C++11 language feature ${feature}"
      ""
      "__cpp_${feature}"
      "${version}"
      CXX11_${feature}
      ${ARGV2}
   )
   if (CXX11_${feature})
      set (CXX11_PARTIAL_SUPPORT TRUE PARENT_SCOPE)
   else ()
      set (CXX11_SUPPORT FALSE PARENT_SCOPE)
   endif ()
endfunction ()

# checks for a C++11 library feature
function (check_cxx11_lib_feature header feature version)
   _check_cxx_feature_macro (
      "C++11 library feature ${feature}"
      "#include <${header}>"
      "__cpp_lib_${feature}"
      "${version}"
      CXX11_lib_${feature}
      ${ARGV3}
   )
   if (CXX11_lib_${feature})
      set (CXX11_PARTIAL_SUPPORT TRUE PARENT_SCOPE)
   else ()
      set (CXX11_SUPPORT FALSE PARENT_SCOPE)
   endif ()
endfunction ()

#-------------------------------------------------------------------------------
# Functions for C++14
#-------------------------------------------------------------------------------

# checks for C++14 support
function (check_cxx14_support)
   _check_cxx_feature_macro (
      "C++14 language support"
      ""
      "__cplusplus"
      201402L
      CXX14_language_support
      ${ARGV0}
   )
   set (CXX14_SUPPORT ${CXX14_language_support} PARENT_SCOPE)
   set (CXX14_PARTIAL_SUPPORT ${CXX14_language_support} PARENT_SCOPE)
endfunction ()

# checks for a specific C++14 language feature
function (check_cxx14_feature feature version)
   _check_cxx_feature_macro (
      "C++14 language feature ${feature}"
      ""
      "__cpp_${feature}"
      "${version}"
      CXX14_${feature}
      ${ARGV2}
   )
   if (CXX14_${feature})
      set (CXX14_PARTIAL_SUPPORT TRUE PARENT_SCOPE)
   else ()
      set (CXX14_SUPPORT FALSE PARENT_SCOPE)
   endif ()
endfunction ()

# checks for a C++14 library feature
function (check_cxx14_lib_feature header feature version)
   _check_cxx_feature_macro (
      "C++14 library feature ${feature}"
      "#include <${header}>"
      "__cpp_lib_${feature}"
      "${version}"
      CXX14_lib_${feature}
      ${ARGV3}
   )
   if (CXX14_lib_${feature})
      set (CXX14_PARTIAL_SUPPORT TRUE PARENT_SCOPE)
   else ()
      set (CXX14_SUPPORT FALSE PARENT_SCOPE)
   endif ()
endfunction ()

#-------------------------------------------------------------------------------
# Functions for C++17
#-------------------------------------------------------------------------------

# checks for C++17 support
function (check_cxx17_support)
   _check_cxx_feature_macro (
      "C++17 language support"
      ""
      "__cplusplus"
      201703L
      CXX17_language_support
      ${ARGV0}
   )
   set (CXX17_SUPPORT ${CXX17_language_support} PARENT_SCOPE)
   set (CXX17_PARTIAL_SUPPORT ${CXX17_language_support} PARENT_SCOPE)
endfunction ()

# checks for a specific C++17 language feature
function (check_cxx17_feature feature version)
   _check_cxx_feature_macro (
      "C++17 language feature ${feature}"
      ""
      "__cpp_${feature}"
      "${version}"
      CXX17_${feature}
      ${ARGV2}
   )
   if (CXX17_${feature})
      set (CXX17_PARTIAL_SUPPORT TRUE PARENT_SCOPE)
   else ()
      set (CXX17_SUPPORT FALSE PARENT_SCOPE)
   endif ()
endfunction ()

# checks for a C++17 library feature
function (check_cxx17_lib_feature header feature version)
   _check_cxx_feature_macro (
      "C++17 library feature ${feature}"
      "#include <${header}>"
      "__cpp_lib_${feature}"
      "${version}"
      CXX17_lib_${feature}
      ${ARGV3}
   )
   if (CXX17_lib_${feature})
      set (CXX17_PARTIAL_SUPPORT TRUE PARENT_SCOPE)
   else ()
      set (CXX17_SUPPORT FALSE PARENT_SCOPE)
   endif ()
endfunction ()

#-------------------------------------------------------------------------------
# Helper function
#-------------------------------------------------------------------------------

# This is the function that does the actual work. It checks for a generic C++
# feature macro.
function (_check_cxx_feature_macro feature header macro version variable)
   if (NOT DEFINED ${variable})
      set (testfile "${CMAKE_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp/test.cpp")
      set (source "${header}\n#if ${macro} >= ${version}\n int main(int, char**) { return 0; }\n #endif")
      file (WRITE ${testfile} "${source}\n")
      set (cmakeflags
         "-DCMAKE_CXX_STANDARD=${CMAKE_CXX_STANDARD}"
         "-DCMAKE_CXX_STANDARD_REQUIRED=${CMAKE_CXX_STANDARD_REQUIRED}"
         "-DCMAKE_CXX_EXTENSIONS=${CMAKE_CXX_EXTENSIONS}"
      )
      try_compile (${variable} ${CMAKE_BINARY_DIR} ${testfile} CMAKE_FLAGS ${cmakeflags})
      set (msg "Checking for ")
   endif ()
   if (NOT "${ARGV5}" STREQUAL "QUIET")
      if (${variable})
         message (STATUS "${msg}${feature}: available")
      else ()
         message (STATUS "${msg}${feature}: NOT available")
      endif ()
   endif ()
endfunction ()
