#-------------------------------------------------------------------------------
# CMake module for flexible handling of external data
#
# Author(s):
#    Frank Tackmann
#
# Copyright:
#    Copyright (C) 2016  Frank Tackmann
#
# Extends CMake's ExternalData module with the two functions described below,
# which provide a flexible and simple alternative interface to the ExternalData
# module. It is sufficient to include this module, which automatically includes
# the ExternalData module.
#
# ExternalData_add_files (
#    <target>    # Name of the data management target
#    <files...>  # Names of data files to add
# )
#
# Adds the give file(s) to be made available by ExternalData in the build tree
# (bypassing the whole DATA{...} mechanism) via <target>. The files are always
# made available in the build tree, even if a local file is present. Local files
# always take precedence over content links.
#
# ExternalData_add_store_target (
#    <target>    # Name of the data management target
# )
#
# If <target> doesn't exist yet it calls ExternalData_add_target (<target>). In
# addition, if a local store is set via ExternalData_LOCAL_STORE, it adds a
# custom target <target>-store, which replaces all present local files for
# <target> by content links and moves the content to the local store. Running
# the target automatically triggers a rerun of CMake to make the files available
# in the build tree.
#
# Warning: The <target>-store target will modify the source tree by moving the
# local files to the local store. Thus, be careful when using it and avoid
# running it automatically by having other targets depend on it.
#
# Note: Both functions currently only support .sha1 content links.
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
# interface for configure/generation mode
#-------------------------------------------------------------------------------

# setup
if (NOT ExternalData_ACTION OR CMAKE_GENERATOR)
   # Explicitly set the default used by ExternalData since we might need it earlier
   if (NOT ExternalData_BINARY_ROOT)
      set (ExternalData_BINARY_ROOT "${CMAKE_BINARY_DIR}")
   endif ()

   include (ExternalData)

   set(_external-data_MYSELF "${CMAKE_CURRENT_LIST_FILE}")

   if (NOT EXISTS "${ExternalData_BINARY_ROOT}/trigger-rerun")
      file (WRITE "${ExternalData_BINARY_ROOT}/trigger-rerun" "")
   endif ()
   include (${ExternalData_BINARY_ROOT}/trigger-rerun)
endif ()

# function to explicitly add external files to a data management target
function (ExternalData_add_files target)
   set (src ${ExternalData_SOURCE_ROOT})
   set (bin ${ExternalData_BINARY_ROOT})
   if ("${src}" STREQUAL "${bin}")
      set (in_source 1)
   endif ()
   foreach (relname IN LISTS ARGN)
      if (EXISTS "${src}/${relname}")
         if (NOT in_source)
            set (local "${bin}/${relname}|${src}/${relname}")
            set_property (GLOBAL APPEND PROPERTY _ExternalData_${target}_LOCAL "${local}")
         endif ()
      elseif (EXISTS "${src}/${relname}.sha1")
         set (fetch "${bin}/${relname}|${src}/${relname}|.sha1")
         set_property (GLOBAL APPEND PROPERTY _ExternalData_${target}_FETCH "${fetch}")
      else ()
         message ("WW Unable to provide content for external data: ${relname}")
      endif ()
   endforeach ()
endfunction ()

# function to add a custom target to move local files to a local store and
# replace them by content links
function (ExternalData_add_store_target target)
   if (NOT TARGET ${target})
      ExternalData_add_target (${target})
   endif ()
   if (NOT ExternalData_LOCAL_STORE)
      return ()
   endif ()
   get_property (local GLOBAL PROPERTY _ExternalData_${target}_LOCAL)
   if (local)
      if (NOT IS_DIRECTORY "${ExternalData_LOCAL_STORE}")
         message (FATAL_ERROR "Local store ${ExternalData_LOCAL_STORE} is not a directory.")
      endif ()

      add_custom_target (${target}-store)

      foreach (entry IN LISTS local)
         string (REPLACE "|" ";" entrydata "${entry}")
         list (GET entrydata 1 localfile)
         add_custom_command (
               TARGET ${target}-store PRE_BUILD
               COMMENT "Storing content of ${localfile}"
               COMMAND ${CMAKE_COMMAND} -Dfile=${localfile} -Dstore_dir=${ExternalData_LOCAL_STORE} -DExternalData_ACTION=store
                                        -P ${_external-data_MYSELF}
         )
      endforeach()

      add_custom_command (
         TARGET ${target}-store POST_BUILD
         COMMENT "Rerunning CMake and remaking ${target}"
         COMMAND ${CMAKE_COMMAND} -E touch "${ExternalData_BINARY_ROOT}/trigger-rerun"
         COMMAND ${CMAKE_COMMAND} --build ${CMAKE_BINARY_DIR} --target ${target}
      )
   endif ()
endfunction ()

#-------------------------------------------------------------------------------
# interface for script mode
#-------------------------------------------------------------------------------

# function to store content of local data into a local store
# file is the absolute path to the local file to be stored
# store_dir is the absolute path to the local storage
# This replaces file by file.sha1 containing the sha1 hash of file. The original
# file itself is moved to store_dir/SHA1/hash
function (_ExternalData_store_content file store_dir)
   set (algo SHA1)
   file ("${algo}" "${file}" hash)
   file (WRITE "${file}.sha1" "${hash}\n")
   file (MAKE_DIRECTORY "${store_dir}/${algo}")
   file (RENAME "${file}" "${store_dir}/${algo}/${hash}")
   message (STATUS "Stored object ${store_dir}/${algo}/${hash}")
endfunction()

if ("${ExternalData_ACTION}" STREQUAL "store")
   _ExternalData_store_content (${file} ${store_dir})
endif ()
