#-------------------------------------------------------------------------------
# CMake module for testing support
#
# Author(s):
#    Frank Tackmann
#
# Copyright:
#    Copyright (C) 2016  Frank Tackmann
#
# Provides support for a testing infrastructure including support for external
# test data (via external-data module).
#
# The module defines the following custom targets:
#  <testing>      : target to build all test binaries
#  <testing>-data : target that depends on all data management targets
#  <testing>-all  : depends on both testing and testing-data
#
# The actual name of the <testing> target is given by the testing_TARGET
# variable (see below) and defaults to <project>-testing, where <project> is
# the name of the current project (i.e. that of the most recent project() call).
#
# The user is free to add additional dependencies to these targets (e.g. custom
# tests).
#
# The module defines the following options that can be set by the user:
#  testing_LOCAL_STORE : Path to a local object store for external test data
#  testing_URL_STORE   : URL of a remote object store for external test data
#
#
# The following variables can be defined before including the module to adjust
# its behavior:
#  testing_TARGET          : The name to use for the global <testing> target
#
#  testing_DATA_SOURCE_DIR : The root directory under which all external test
#                            data (local files or content links) are located.
#                            Must be defined to turn on handling of external
#                            test data.
#  testing_DATA_BINARY_DIR : The build directory where external test data is
#                            provided (defaults to CMAKE_CURRENT_BINARY_DIR).
#
# These are also used to set ExternalData_SOURCE_ROOT and
# ExternalData_BINARY_ROOT.
#
# testing_LOCAL_STORE_DEFAULT : optional default to use for testing_LOCAL_STORE
# testing_URL_STORE_DEFAULT   : optional default to use for testing_URL_STORE
#
#
# The following functions are defined by the package
#
# add_testing_module (
#    <source_dir>         # the source directory to add
#    [MODULE <module>]    # optional name of the testing module (defaults to <source_dir>)
#    [DATADIR <data_dir>] # optional base directory holding external test data
# )
#
# Adds <source_dir> as a testing module with name <module>. It creates a
# <module>-testing target, which will depend on all tests added (see below).
#
# If <data_dir> is given, testing_DATA_SOURCE_DIR/<data_dir> is used as root
# directory for all data files. If <data_dir> is not specified and
# testing_DATA_SOURCE_DIR/<source_dir> exists then <source_dir> is used as
# default for <data_dir>. A <module>-testing-data is created which will depend
# on all data management targets created (see below).
#
# add_<project>_test (
#    <name>                  # name of test binary
#    [NAME <test_name>]      # optional name of test (defaults to <name>)
#    [SOURCES <sources...>]  # optional source files (defaults to "<name>.cpp")
#    [DEPENDS <libs...>]     # optional link dependencies
#    [FILES <datafiles...>]  # optional external test data files
# )
#
# Adds definitions to build the executable <name> from <sources...> with the
# optional link dependencies <libs...>. The corresponding <name> build target is
# added as dependence of the current <module>-testing target.
#
# If <datafiles...> are specified, a <name>-data target is created to handle
# them. The target is added as dependence of the current <module>-testing-data
# target. In addition, if testing_LOCAL_STORE is set, a <name>-data-store target
# is created, that allows to move local test data files to the local store and
# replace them by content links (see external-data documentation for details).
#
# The following variables are used by add_<project>_test:
#  testing_COMPILE_FLAGS  : Additional flags used to compile test binaries
#  testing_LINK_LIBRARIES : Additional libraries to link test binaries
#  testing_RUN_FLAGS      : Global command line arguments to run test binaries
#
#  testing_MODULE        : The <module> name of the last call to add_testing_module.
#  testing_MODULE_TARGET : The current module target, defaults to <module>-testing
#  testing_DATADIR       : The <datadir> of the last call to add_testing_module.
#
# If add_<project>_test is called outside a test module (i.e. if testing_MODULE
# is unset), the <name> and <name>-data targets are directly added as dependence
# of the global <testing> and <testing>-data targets.
#-------------------------------------------------------------------------------

if (NOT PROJECT_NAME)
   message (FATAL_ERROR "Not inside a project. The testing module can only be used after a call to project().")
endif ()

#-------------------------------------------------------------------------------
# global testing targets
#-------------------------------------------------------------------------------

# target to build all tests
if (NOT testing_TARGET)
   set (testing_TARGET ${PROJECT_NAME}-testing)
endif ()
add_custom_target (${testing_TARGET})

# testing-all target that depends on everything testing related
add_custom_target (${testing_TARGET}-all)
add_dependencies (${testing_TARGET}-all ${testing_TARGET})


#-------------------------------------------------------------------------------
# configure external data handling
#-------------------------------------------------------------------------------

if (testing_DATA_SOURCE_DIR)
   if (NOT IS_DIRECTORY "${testing_DATA_SOURCE_DIR}")
      message (FATAL_ERROR "Specified testing_DATA_SOURCE_DIR ${testing_DATA_SOURCE_DIR} is not a directory.")
   endif ()
   if (NOT testing_DATA_BINARY_DIR)
      set (testing_DATA_BINARY_DIR "${CMAKE_CURRENT_BINARY_DIR}")
   endif ()

   set (ExternalData_SOURCE_ROOT ${testing_DATA_SOURCE_DIR})
   set (ExternalData_BINARY_ROOT ${testing_DATA_BINARY_DIR})
   include (external-data)
   message (STATUS "Using external test data in ${testing_DATA_SOURCE_DIR}")

   set (ExternalData_OBJECT_STORES "${ExternalData_BINARY_ROOT}/.store")

   # define option for user to set local object store
   set (testing_LOCAL_STORE "${testing_LOCAL_STORE_DEFAULT}" CACHE PATH "Path to local object store for external test data (optional).")
   if (IS_DIRECTORY "${testing_LOCAL_STORE}")
      set (ExternalData_LOCAL_STORE "${testing_LOCAL_STORE}")
      set (ExternalData_OBJECT_STORES "${testing_LOCAL_STORE}" ${ExternalData_OBJECT_STORES})
      message (STATUS "Using local object store for external test data: ${testing_LOCAL_STORE}")
   else ()
      set (ExternalData_LOCAL_STORE "")
   endif ()

   # define option for user to set remote object store
   set (testing_URL_STORE "${testing_URL_STORE_DEFAULT}" CACHE STRING "URL of remote object store for external test data (optional).")
   if (testing_URL_STORE)
      message (STATUS "Using remote object store for external test data: ${testing_URL_STORE}")
      set (ExternalData_URL_TEMPLATES "${testing_URL_STORE}/%(algo)/%(hash)?raw" ${ExternalData_URL_TEMPLATES})
   endif ()

   # target to provide all test data
   add_custom_target (${testing_TARGET}-data)
   add_dependencies (${testing_TARGET}-all ${testing_TARGET}-data)
endif ()


#-------------------------------------------------------------------------------
# functions for adding tests
#-------------------------------------------------------------------------------

include (CMakeParseArguments)

# function to add a subdirectory as test module
function (add_testing_module source_dir)
   set (singleValueArgs MODULE DATADIR)
   cmake_parse_arguments (testing "" "${singleValueArgs}" "" ${ARGN})
   if (NOT testing_MODULE)
      set (testing_MODULE "${source_dir}")
   endif ()

   set (testing_MODULE_TARGET ${testing_MODULE}-testing)
   if (NOT TARGET ${testing_MODULE_TARGET})
      add_custom_target (${testing_MODULE_TARGET})
      add_dependencies (${testing_TARGET} ${testing_MODULE_TARGET})
   endif ()
   file (RELATIVE_PATH tmp "${CMAKE_SOURCE_DIR}" "${CMAKE_CURRENT_SOURCE_DIR}/${source_dir}")
   set (info "from ${tmp}")

   if (TARGET ${testing_TARGET}-data AND IS_DIRECTORY "${testing_DATA_SOURCE_DIR}")
      if (NOT testing_DATADIR AND IS_DIRECTORY "${testing_DATA_SOURCE_DIR}/${source_dir}")
         set (testing_DATADIR ${source_dir})
      endif ()
      if (testing_DATADIR)
         if (NOT TARGET ${testing_MODULE_TARGET}-data)
            add_custom_target (${testing_MODULE_TARGET}-data)
            add_dependencies (${testing_TARGET}-data ${testing_MODULE_TARGET}-data)
         endif ()
         file (RELATIVE_PATH tmp "${CMAKE_SOURCE_DIR}" "${testing_DATA_SOURCE_DIR}/${testing_DATADIR}")
         set (info "${info} using external data in ${tmp}")
      endif ()
   elseif (testing_DATADIR)
      message (FATAL_ERROR "Specified DATADIR but handling of external test data is not enabled (probably testing_DATA_SOURCE_DIR was not specified).")
   endif ()

   message (STATUS "Adding tests for module: ${testing_MODULE} (${info})")
   add_subdirectory (${source_dir})
endfunction ()

# function to add a test
function (add_${PROJECT_NAME}_test name)
   set (singleValueArgs NAME)
   set (multiValueArgs SOURCES DEPENDS FILES)
   cmake_parse_arguments (test "" "${singleValueArgs}" "${multiValueArgs}" ${ARGN})
   if (NOT test_SOURCES)
      set (test_SOURCES "${name}.cpp")
   endif ()
   if (NOT test_NAME)
      set (test_NAME ${name})
   endif ()

   add_executable (${name} ${test_SOURCES})
   target_include_directories (${name} PRIVATE ${testing_INCLUDE_DIRS})
   set_property (TARGET ${name} APPEND_STRING PROPERTY COMPILE_FLAGS " ${testing_COMPILE_FLAGS}")
   target_link_libraries (${name} ${testing_LINK_LIBRARIES} ${test_DEPENDS})
   add_test (NAME ${test_NAME} COMMAND ${name} ${testing_RUN_FLAGS})
   if (testing_MODULE)
      add_dependencies (${testing_MODULE_TARGET} ${name})
   else ()
      add_dependencies (${testing_TARGET} ${name})
   endif ()
   if (test_FILES)
      if (NOT COMMAND ExternalData_add_files)
         message (FATAL_ERROR "Specified data FILES but handling of external data is not enabled.")
      endif ()
      set (files "")
      if (testing_DATADIR)
         foreach (file ${test_FILES})
            set (files ${files} "${testing_DATADIR}/${file}")
         endforeach ()
      else ()
         set (files ${test_FILES})
      endif ()
      ExternalData_add_files (${name}-data ${files})
      ExternalData_add_store_target (${name}-data)
      if (testing_MODULE)
         add_dependencies (${testing_MODULE_TARGET}-data ${name}-data)
      else ()
         add_dependencies (${testing_TARGET}-data ${name}-data)
      endif ()
   endif ()
endfunction ()
