#-------------------------------------------------------------------------------
# Author(s):
#    Frank Tackmann
#
# Copyright:
#    Copyright (C) 2017, 2018  Geneva Collaboration
#
#    This file is part of the Geneva MC framework. Geneva is distributed under
#    the terms of the GNU General Public License version 3 (GPLv3), see the
#    COPYING file that comes with this distribution for details.
#    Please respect the academic usage guidelines in the GUIDELINES file.
#
# Description:
#    cmake script for global build configuration
#-------------------------------------------------------------------------------

# print cmake version we're running with
message (STATUS "Using CMake version: ${CMAKE_VERSION}")

# make sure we are not going to build in the main source directory
get_filename_component (src "${PROJECT_SOURCE_DIR}" REALPATH)
get_filename_component (bin "${PROJECT_BINARY_DIR}" REALPATH)
file (RELATIVE_PATH relbin ${src} ${bin})

if ("${relbin}/" MATCHES "^(cmake|doc|include|packages|python|share|src|testing|)/")
   message ("EE In-source builds are not allowed.")
   message ("EE CMake has already put some files (CMakeCache.txt, CMakeFiles/) into your source tree.")
   message (">> Delete these (or restart from a clean source), create a separate build directory (e.g. 'mkdir build'),")
   message (">> and call CMake from there (e.g. 'cd build && cmake ..')")
   message (FATAL_ERROR "In-source builds are not allowed.")
endif ()

unset (src)
unset (bin)
unset (relbin)

#-------------------------------------------------------------------------------
# Check compiler features and select compiler flags.
#-------------------------------------------------------------------------------

message (STATUS "Using C++ compiler: ${CMAKE_CXX_COMPILER_ID} ${CMAKE_CXX_COMPILER_VERSION} (${CMAKE_CXX_COMPILER})")

# require C++11
if (NOT CMAKE_CXX_STANDARD)
   set (CMAKE_CXX_STANDARD 11)
   set (CMAKE_CXX_STANDARD_REQUIRED ON)
   set (CMAKE_CXX_EXTENSIONS OFF)
endif ()

# check for C++11 availability
include (cxx-support)

check_cxx11_support ()

if (NOT CXX11_SUPPORT)
   message ("EE No C++11 support appears to be available.")
   set (fail TRUE)
endif ()

# select compiler flags.
if (CMAKE_CXX_COMPILER_ID MATCHES GNU OR CMAKE_CXX_COMPILER_ID MATCHES Clang)

   set (CMAKE_CXX_FLAGS_DEBUG "-O0 -g")
   set (CMAKE_CXX_FLAGS_RELWITHDEBINFO "-O2 -g")

   # turn on all warnings (currently not used: -Wsign-promo -Wold-style-cast)
   set (${PROJECT_NAME}_CXX_FLAGS "-pedantic -Wall -Wextra -Wconversion -Wsign-conversion -Wdeprecated")
   set (${PROJECT_NAME}_CXX_FLAGS "${${PROJECT_NAME}_CXX_FLAGS} -Wno-unused-parameter -Wno-implicit-fallthrough")

   if (CMAKE_CXX_COMPILER_ID MATCHES Clang)
      set (${PROJECT_NAME}_CXX_FLAGS "${${PROJECT_NAME}_CXX_FLAGS} -Wno-overloaded-virtual")
   endif ()

else ()
   message ("ww Unsupported C++ compiler: ${CMAKE_CXX_COMPILER_ID} (proceed at your own risk ...)")
   message ("ww Supported are GNU (g++) mand Clang/LLVM (clang++).")
   set (warn TRUE)

   # explicitly make sure that NDEBUG is properly (un)defined
   set (CMAKE_CXX_FLAGS_DEBUG "${CMAKE_CXX_FLAGS_DEBUG} -UNDEBUG")
   set (CMAKE_CXX_FLAGS_RELWITHDEBINFO "${CMAKE_CXX_FLAGS_RELWITHDEBINFO} -UNDEBUG")
   set (CMAKE_CXX_FLAGS_MINSIZEREL "${CMAKE_CXX_FLAGS_MINSIZEREL} -DNDEBUG")
   set (CMAKE_CXX_FLAGS_RELEASE "${CMAKE_CXX_FLAGS_RELEASE} -DNDEBUG")
endif ()

option (FORCE_CXX_COMPILER "Force using the configured C++ compiler even if it fails C++11 compatibility checks." OFF)
mark_as_advanced (FORCE_CXX_COMPILER)

if (warn OR fail)
   message (">> To use a different compiler, remove the current configuration ('rm -rf CMake*')")
   message (">> and specify the compiler by passing -DCMAKE_CXX_COMPILER=path/to/c++ to cmake.")
   if (fail)
      if (FORCE_CXX_COMPILER)
         message ("ww FORCE_CXX_COMPILER is turned on.")
         message ("ww Continuing with (likely) incompatible C++ compiler (you asked for it ...)")
      else ()
         message (">> (To force using the configured C++ compiler anyway pass -DFORCE_CXX_COMPILER=ON to cmake.)")
         message (FATAL_ERROR "Incompatible C++ compiler.")
      endif ()
   endif ()
endif ()

message (STATUS "Using C++ warning flags: ${${PROJECT_NAME}_CXX_FLAGS}")
separate_arguments (${PROJECT_NAME}_CXX_FLAGS)

#-------------------------------------------------------------------------------
# Build configuration
#-------------------------------------------------------------------------------

# set a default build type for single-configuration CMake generators if no build type is set.
if (NOT CMAKE_CONFIGURATION_TYPES)
   if (NOT CMAKE_BUILD_TYPE)
      set (CMAKE_BUILD_TYPE Release)
      message (STATUS "Using default build type: ${CMAKE_BUILD_TYPE}")
   else ()
      message (STATUS "Using configured build type: ${CMAKE_BUILD_TYPE}")
   endif ()
endif ()

# put all libraries and executables into common lib/ and bin/ directories in
# project build dir (unless output directories already have custom values)
if (NOT CMAKE_RUNTIME_OUTPUT_DIRECTORY)
   set (CMAKE_RUNTIME_OUTPUT_DIRECTORY ${PROJECT_BINARY_DIR}/bin)
endif ()
if (NOT CMAKE_LIBRARY_OUTPUT_DIRECTORY)
   set (CMAKE_LIBRARY_OUTPUT_DIRECTORY ${PROJECT_BINARY_DIR}/lib)
endif ()
if (NOT CMAKE_ARCHIVE_OUTPUT_DIRECTORY)
   set (CMAKE_ARCHIVE_OUTPUT_DIRECTORY ${PROJECT_BINARY_DIR}/lib)
endif ()

# option to build shared libs
option (BUILD_SHARED_LIBS "Whether to build shared libraries by default." ON)
if (BUILD_SHARED_LIBS)
   message (STATUS "Building libraries: shared")
else ()
   message (STATUS "Building libraries: static")
endif ()

#-------------------------------------------------------------------------------
# User options
#-------------------------------------------------------------------------------

# cache variable for the data directory
set (geneva_DATA_DIR ${PROJECT_SOURCE_DIR}/share/Geneva CACHE STRING "Geneva data directory.")
message (STATUS "Using data directory: ${geneva_DATA_DIR}")

# cache variables for default precision and accuracy (for comparing doubles)
set (geneva_DEFAULT_PRECISION "1e-10" CACHE STRING "Default relative precision for numerical checks (e.g. for comparing doubles).")
set (geneva_DEFAULT_ACCURACY "1e-10" CACHE STRING "Default absolute accuracy for numerical checks (e.g. for comparing doubles).")

# define option to set the target for refining momenta
option (REFINE_MOMENTA "Refine momenta" ON)
if (REFINE_MOMENTA)
   set (geneva_REFINE_TARGET_PRECISION "1e-12" CACHE STRING "Refine momenta precision")
   set (geneva_REFINE_TARGET_ACCURACY "1e-15" CACHE STRING "Refine momenta accuracy")
endif()

# option to turn on trapping floating point exceptions (turned off by default)
option (geneva_enable_fpe_trapping "Enable trapping of floating point exceptions" OFF)
set (geneva_USE_FPE_TRAPPING FALSE)
if (geneva_enable_fpe_trapping)
   # Only supported by gcc
   if (CMAKE_CXX_COMPILER_ID MATCHES GNU)
      message (STATUS "Enabling trapping of floating-point exceptions.")
      set (geneva_USE_FPE_TRAPPING TRUE)
   else ()
      message ("ww Trapping of floating point exceptions is only supported by GNU compiler.")
   endif ()
endif ()

# option to use OpenMP (turned off by default, currently unused)
option (geneva_enable_openmp "Enable use of OpenMP" OFF)
set (geneva_USE_OPENMP FALSE)
if (geneva_enable_openmp)
   find_package (OpenMP)
   if (OPENMP_FOUND)
      set (CMAKE_C_FLAGS "${CMAKE_C_FLAGS} ${OpenMP_C_FLAGS}")
      set (CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} ${OpenMP_CXX_FLAGS}")
      message (STATUS "Enabling OpenMP support.")
      set (geneva_USE_OPENMP TRUE)
   else ()
      message ("ww Could not find OpenMP support, continuing without.")
   endif()
endif()
