#-------------------------------------------------------------------------------
# Author(s):
#    Frank Tackmann
#
# Copyright:
#    Copyright (C) 2017 DESY
#
#    This file is part of the Geneva MC framework. Geneva is distributed under
#    the terms of the GNU General Public License version 3 (GPLv3), see the
#    COPYING file that comes with this distribution for details.
#    Please respect the academic usage guidelines in the GUIDELINES file.
#
# Description:
#    cmake script for finding external library dependencies
#-------------------------------------------------------------------------------

# user option for which beam function grids to use
# has to be defined already here as it is used by both lhapdf and beamfunc
set (beamfunc_PDF_SETS
   CT10nnlo
   MMHT2014nnlo68cl
   NNPDF31_nnlo_as_0118
   PDF4LHC15_nnlo_100
   CACHE STRING "beamfunc PDF sets to use (semicolon separated list)."
)

#-------------------------------------------------------------------------------
# external libraries
#-------------------------------------------------------------------------------

message ("== Configuring external libraries")

# user option to specify the prefix where automated local installations via
# 'make <package>' will be installed
set (geneva_external_prefix
   ${PROJECT_SOURCE_DIR}/external CACHE PATH
   "Install path prefix to use for automated local installations of external packages."
)

# user option whether to use geneva_external_prefix as the default library_ROOT for all libraries
option (geneva_external_only_default "Search for external packages by default only in geneva_external_prefix path." OFF)

# find the gsl library
find_package (gsl 1.13 REQUIRED OPTIONAL_COMPONENTS gslcblas)
# gsl is currently required and not an option, so this is always true
set (geneva_USE_GSL TRUE)

# define option to use the zlib library (turned on by default)
option (geneva_enable_zlib "Enable use of the zlib library" ON)
set (geneva_USE_ZLIB FALSE)
if (geneva_enable_zlib)
   find_package (zlib)
   if (TARGET zlib::zlib)
      set (geneva_USE_ZLIB TRUE)
   endif()
else ()
   message (STATUS "Skipping disabled: zlib")
endif ()

# define option whether to use FastJet (turned off by default)
option (geneva_enable_fastjet "Enable use of the FastJet library" OFF)
set (geneva_USE_FASTJET FALSE)
if (geneva_enable_fastjet)
   if (geneva_external_only_default)
      set (fastjet_ROOT ${geneva_external_prefix} CACHE PATH "Root path for FastJet installation.")
   endif ()
   find_package (fastjet 3.2)
   if (TARGET fastjet::fastjet)
      set (geneva_USE_FASTJET TRUE)
   else ()
      # Explicitly set fastjet_FOUND to false to avoid rivet searching for it again.
      set (fastjet_FOUND FALSE)
   endif ()
else ()
   message (STATUS "Skipping disabled: fastjet")
endif ()

# define option whether to use HepMC (turned on by default)
option (geneva_enable_hepmc "Enable use of the HepMC library" ON)
set (geneva_USE_HEPMC FALSE)
if (geneva_enable_hepmc)
   if (geneva_external_only_default)
      set (hepmc_ROOT ${geneva_external_prefix} CACHE PATH "Root path for HepMC installation.")
   endif ()
   find_package (hepmc 2.06.08)
   if (TARGET hepmc::hepmc)
      set (geneva_USE_HEPMC TRUE)
   else ()
      # Explicitly set hepmc_FOUND to false to avoid rivet searching for it again.
      set (hepmc_FOUND FALSE)
   endif ()
else ()
   message (STATUS "Skipping disabled: hepmc")
endif ()

# define option whether to use LHAPDF (turned on by default)
option (geneva_enable_lhapdf "Enable use of the LHAPDF library" ON)
set (geneva_USE_LHAPDF FALSE)
if (geneva_enable_lhapdf)
   if (geneva_external_only_default)
      set (lhapdf_ROOT ${geneva_external_prefix} CACHE PATH "Root path for LHAPDF installation.")
   endif ()
   find_package (lhapdf 6.2)
   if (TARGET lhapdf::lhapdf)
      set (geneva_USE_LHAPDF TRUE)
   endif ()
else ()
   message (STATUS "Skipping disabled: lhapdf")
endif ()

# define option whether to use OpenLoops (turned on by default)
option (geneva_enable_openloops "Enable use of the OpenLoops library" ON)
set (geneva_USE_OPENLOOPS FALSE)
if (geneva_enable_openloops)
   if (geneva_external_only_default)
      set (openloops_ROOT ${geneva_external_prefix} CACHE PATH "Root path for OpenLoops installation.")
   endif ()
   find_package (openloops OPTIONAL_COMPONENTS openloops_ppll_lt openloops_ppllj_lt openloops_pplljj_lt)
   if (TARGET openloops::openloops)
      set (geneva_USE_OPENLOOPS TRUE)
   endif ()
else ()
   message (STATUS "Skipping disabled: openloops")
endif ()

# define option whether to use Pythia8 (turned on by default)
option (geneva_enable_pythia8 "Enable use of the Pythia8 library" ON)
set (geneva_USE_PYTHIA8 FALSE)
if (geneva_enable_pythia8)
   if (geneva_external_only_default)
      set (pythia8_ROOT ${geneva_external_prefix} CACHE PATH "Root path for Pythia8 installation.")
   endif ()
   find_package (pythia8 8.230 OPTIONAL_COMPONENTS pythia8lhapdf6)
   if (TARGET pythia8::pythia8)
      set (geneva_USE_PYTHIA8 TRUE)
   endif ()
else ()
   message (STATUS "Skipping disabled: pythia8")
endif ()

# define option whether to use Rivet (turned off by default)
option (geneva_enable_rivet "Enable use of the Rivet library" OFF)
set (geneva_USE_RIVET FALSE)
if (geneva_enable_rivet)
   if (geneva_external_only_default)
      set (rivet_ROOT ${geneva_external_prefix} CACHE PATH "Root path for Rivet installation.")
      set (yoda_ROOT ${geneva_external_prefix} CACHE PATH "Root path for YODA installation.")
   endif ()
   find_package (yoda 1.6)
   if (NOT TARGET yoda::yoda)
      set (yoda_FOUND FALSE)
   endif ()
   find_package (rivet 2.5)
   if (rivet_FOUND)
      set (geneva_USE_RIVET TRUE)
   endif ()
else ()
   message (STATUS "Skipping disabled: rivet")
endif ()


# read available definitions for automated external installations
add_subdirectory (external)

#-------------------------------------------------------------------------------
# packages
#-------------------------------------------------------------------------------

message ("== Configuring packages")

# dive into packages/ subdirectory
add_subdirectory (packages)
