#-------------------------------------------------------------------------------
# CMake module to obtain version information from a git repository.
#
# Author(s):
#    Frank Tackmann
#
# Copyright:
#    Copyright (C) 2015  Frank Tackmann
#
# Defines the function
#
#  get_git_info(<workdir> [QUIET] [REQUIRED])
#
# It reads commit and version information from the git repository in <workdir>.
# The QUIET option suppresses all messages. The REQUIRED option causes a fatal
# error if Git is not found or <workdir> is not in a git repository.
#
# The function calls the git executable in the variable GIT_EXECUTABLE. If this
# is not set, it searches for Git with find_package(Git [options]), passing on
# all optional arguments past <workdir>.
#
# If successful, the following variables are defined in the parent's scope:
#
#  GIT_COMMIT  : The commit hash of HEAD
#  GIT_PREFIX  : The part of <workdir> relative to the repo's top-level
#  GIT_VERSION : A version string synthesized from the last tag and current commit
#  GIT_CHANGES : The summary of uncommitted changes (from 'git status --short')
#
# The generated GIT_VERSION has the form '<tag>[-<n>-g<commit>][-modified]',
# where <tag> is the most recent tag reachable from the commit, <n> is the
# number of commits since the tag, and <commit> is the same as GIT_COMMIT. This
# part is only added if n >= 1. If no tag is found, the form used is
# 'none-g<commit>[-modified]'.
# The '-modified' suffix is appended if there are uncomitted changes in tracked
# files in the working tree (untracked files are ignored), which is equivalent
# to GIT_CHANGES being nonempty.
#
# If the repository uses tags in the standard version format
# major[.minor[.patch[.tweak]]], then GIT_VERSION can be used as version string
# in CMake scripts and compared with VERSION_LESS, etc.
#-------------------------------------------------------------------------------

include (CMakeParseArguments)

function (get_git_info workdir)
   set (options QUIET REQUIRED)
   cmake_parse_arguments (git "${options}" "" "" ${ARGN})

   if (NOT GIT_EXECUTABLE)
      find_package (Git ${ARGN})
      if (NOT GIT_FOUND)
         return ()
      endif ()
   endif ()

   if (NOT git_QUIET)
      message (STATUS "Reading git working dir: ${workdir}")
   endif ()

   execute_process (
      COMMAND ${GIT_EXECUTABLE} rev-parse --short HEAD
      WORKING_DIRECTORY ${workdir}
      OUTPUT_VARIABLE commit
      OUTPUT_STRIP_TRAILING_WHITESPACE
      ERROR_QUIET
   )
   if (NOT commit)
      if (git_REQUIRED)
         message (FATAL_ERROR "${workdir} does not appear to be in a git repository.")
      endif ()
      if (NOT git_QUIET)
         message (">> Does not appear to be in a git repository.")
      endif ()
      return ()
   endif ()

   execute_process (
      COMMAND ${GIT_EXECUTABLE} describe
      WORKING_DIRECTORY ${workdir}
      OUTPUT_VARIABLE version
      OUTPUT_STRIP_TRAILING_WHITESPACE
      ERROR_QUIET
   )
   if (NOT version)
      set (version "g${commit}")
   endif ()

   execute_process (
      COMMAND ${GIT_EXECUTABLE} status -uno --porcelain
      WORKING_DIRECTORY ${workdir}
      OUTPUT_VARIABLE changes
      OUTPUT_STRIP_TRAILING_WHITESPACE
      ERROR_QUIET
   )
   if (changes)
      set (version "${version}-modified")
   endif()

   execute_process (
      COMMAND ${GIT_EXECUTABLE} rev-parse --show-prefix
      WORKING_DIRECTORY ${workdir}
      OUTPUT_VARIABLE prefix
      OUTPUT_STRIP_TRAILING_WHITESPACE
      ERROR_QUIET
   )

   if (NOT git_QUIET)
      if (prefix)
         message (STATUS "relative path to top-level: ${prefix}")
      endif ()
      message (STATUS "git commit:  ${commit}")
      message (STATUS "git version: ${version}")
      if (changes)
         message (STATUS "modifications:\n${changes}")
      endif ()
   endif ()

   # export results to scope of caller
   set (GIT_PREFIX ${prefix} PARENT_SCOPE)
   set (GIT_COMMIT ${commit} PARENT_SCOPE)
   set (GIT_VERSION ${version} PARENT_SCOPE)
   set (GIT_CHANGES ${changes} PARENT_SCOPE)
endfunction ()
