<!------------------------------------------------------------------------------
# Author(s):
#    Simone Alioli, Christian Bauer, Frank Tackmann
#
# Copyright:
#    Copyright (C) 2017 Geneva Collaboration
#
#    This file is part of the Geneva MC framework. Geneva is distributed under
#    the terms of the GNU General Public License version 3 (GPLv3), see the
#    COPYING file that comes with this distribution for details.
#    Please respect the academic usage guidelines in the GUIDELINES file.
#
# Description:
#    Documentation of user options and option file format
#------------------------------------------------------------------------------>
![GENEVA]
##### [Main] | [Installation] | [User Guide] | [Tutorial] | [Index] #####
--------------------------------------------------------------------------------

Option Files
=============

The option file is written in YAML and has 4 main sections:

 * **Run options:** Global run parameters

 * **Process options:** Define the details of the process

 * **Calculation options:** Define the details of the calculation

 * **Phase-space generator options:** Define generation cuts (and other options) for the phase-space generator

A full example is provided in the file [UserOptionsPPV.yml].

Run options
------------

This sets global parameters of the run

```yaml
global:
   process: pp_V            # process to generate
   run_name: "myRun"        # name of the run (automatically generated if absent)
   num_events: 1000         # number of events to generate
   max_time: 60             # maximum time to run in minutes

input_output:
   verbosity: info          # verbosity level
   setup_tag: ""            # tag added to setup files
   generate_tag: ""         # tag added to generation files
   file_path:
      setup: "setup"        # directory for setup files
      generate: "generate"  # directory for generation files

event_generation:
   unweighting:             # use partial unweighting, see below [optional but recommended]
      partial_relative: 30
   random:
      seed: 1               # random seed

event_analysis:             # options for event analysis, see below [optional]
   analyzer: Rates          # analyzer to use
```
If the `setup_tag` , `generate_tag`, or the `file_path` options are not
provided, the above default values are used.

The only process available at the moment is `pp_V`.

Only one of `num_events` or `max_time` needs to be specified. If both are
provided, the program stops after `max_time` minutes or when `num_events` have
been generated, whichever happens first.

If partial unweighting of the events is desired, the maximum weight to reweight 
against is needed. This can either be specified either as a multiple of the LO 
cross section via

```yaml
   unweighting:
      partial_relative: 30
```

or in terms of an absolute cross section weight (in pb) via

```yaml
   unweighting:
      partial_absolute: 60000
```

An Analyzer can be specified to perform an on-the-fly analysis of the generated
events. Currently, the only available choice is `Rates`, which calculates cross
sections for various parton and jet multiplicities and prints out the results.
The generated events can of course always be analyzed afterwards, obtaining any
desired observables.


Process options
----------------

This defines the details of the process selected in the global options. This
includes physical parameters, such as collider energy, PDF set, etc. In
addition, for each process different calculations, phase-space generators, and
matrix-element providers can be chosen.

```yaml
process:
   pp_V:
      initial_state:
         beams: pp                         # The initial states of the collider
         Ecm: 13000                        # collider energy
         pdf_provider:
            LHAPDF:
               set: "PDF4LHC15_nnlo_100"   # pdf set to use
      final_state:
         boson_type: Z                     # boson type
         boson_mass: 91.1876               # boson mass in GeV
         boson_width: 2.4952               # boson width in GeV
         decay: e+e-                       # decay channel of the boson
      calculation: SCETppV012              # calculation to perform
      phase_space: PP2BosonJets            # phase-space generator to use
      matrix_element_provider: OpenLoops   # matrix-element providers to use
```

Currently, the only available options for `calculation`, `phase_space`, and
`matrix_element_provider` are the ones given above.


Calculation options
--------------------

This specifies the options for the calculation selected in the process options.

```yaml
calculation:
   SCETppV012:
      precision: NNLO+NNLL0+NLL1           # order/precision of the calculation
      scale_settings:
         fixed_order:                      # type of fixed order scale, see below
            dynamic: TransverseMass
      couplings:
         alpha_s: fromPDF                  # alpha_s definition, see below
         alpha_em:
            scale: 91.1876                 # scale of alpha_em
            value: 7.556E-03               # value of alpha_em at above scale
         GF: 1.16639E-05                   # value of G_F
         sin2W: 0.2226459                  # value of sin^2 theta_W
```

The only setting for `precision` currently allowed is the one given above.

The `fixed_order` scale can be set either to a fixed value

```yaml
      scale_settings:
         fixed_order:
            fixed: 91.1876
```

or to a dynamic definition

```yaml
      scale_settings:
         fixed_order:
            dynamic: TransverseMass
```

The available dynamic choices are `InvariantMass` and `TransverseMass`.

The value of `alpha_s` can either be automatically taken from the chosen PDF set

```yaml
      couplings:
         alpha_s: fromPDF
```

or specified explicitly by providing an input value and scale

```yaml
      couplings:
         alpha_s:
            scale: 91.1876
            value: 0.118
```

Note that it is your responsibility to use values for the electroweak input
parameters that are consistent with each other.


Phase-space generator options
------------------------------

This specifies the options for the phase-space generator selected in the process
options. These include in particular cuts on the generated phase space.

```yaml
phase_space:
   PP2BosonJets:
      generation_cuts:
         boson_mass:
            min:  50          # minimum boson invariant mass
            max: 150          # maximum boson invariant mass
         boson_rapidity:      # boson rapidity cuts [optional]
            min: -2.5         # minimum boson rapidity
            max: +2.5         # maximum boson rapidity
```

The cuts on `boson_mass` (the dilepton invariant mass) are required. To avoid
the photon pole, a minimum boson mass of 1 GeV is always enforced. The cuts on
the `boson_rapidity` are optional.


[links]: #----------------------------------------------------------------------

[Main]:               ../README.md
[Installation]:       Installation.md
[User Guide]:         UserGuide.md
[Tutorial]:           tutorial/Tutorial.md
[Index]:              Index.md
[GENEVA]:             geneva.jpg

[UserOptionsPPV.yml]: ../share/Geneva/UserOptionsPPV.yml
