<!------------------------------------------------------------------------------
# Author(s):
#    Simone Alioli, Christian Bauer, Frank Tackmann
#
# Copyright:
#    Copyright (C) 2017 Geneva Collaboration
#
#    This file is part of the Geneva MC framework. Geneva is distributed under
#    the terms of the GNU General Public License version 3 (GPLv3), see the
#    COPYING file that comes with this distribution for details.
#    Please respect the academic usage guidelines in the GUIDELINES file.
#
# Description:
#    Documentation of Pythia interface
#------------------------------------------------------------------------------>
![GENEVA]
##### [Main] | [Installation] | [User Guide] | [Tutorial] | [Index] #####
--------------------------------------------------------------------------------

Pythia Interface
=================

The shower stage described in the [User Guide] and [Tutorial] uses a
`Pythia8::Pythia` instance created internally by Geneva. The LHEF files produced
by Geneva can also be showered using an externally managed instance, which is
described here.

The setup of Pythia has to be consistent with the calculation performed by
Geneva. This requires specific values for a few Pythia settings and to use a
Pythia8 UserHook that is specific to the Geneva calculation. Geneva provides a
PythiaInterface library for this purpose.


Pythia UserHook
----------------

The UserHook is obtained by calling one of the two static functions of the
`PythiaDriver` class

```cpp
UserHookRestriction* PythiaDriver::getUserHook(const std::string& lhefInputFile, const std::string& genevaOptionsFile);
UserHookRestriction* PythiaDriver::getUserHook(std::istream& lhefInputStream, const std::string& genevaOptionsFile);
```
The parameters of the functions are

 * `lhefInputFile`    : name of the LHEF file (only the header is needed)
 * `lhefInputStream`  : istream to read the LHEF header from
 * `genevaOptionsFile`: name of Geneva option file

Both functions return a pointer to a `Geneva::UserHookRestriction`, which
derives from `Pythia8::UserHooks`, and is configured to be consistent with
the calculation as defined by the `genevaOptionsFile`. The LHEF header is used
to check whether the provided option file is consistent with the options used
to generate the events.


Required Pythia settings
-------------------------

A list of all required Pythia8 settings is obtained by calling one of the two
static functions of the `PythiaDriver` class

```cpp
std::vector<std::string> PythiaDriver::getRequiredSettings(const std::string& lhefInputFile, const std::string& genevaOptionsFile)
std::vector<std::string> PythiaDriver::getRequiredSettings((std::istream& lhefInputStream, const std::string& genevaOptionsFile)
```

The function parameters are the same as for the `getUserHook` function above.
Each string in the returned vector is a Pythia8 option that must be set, e.g. by
passing it to Pythia8 via `pythia.readString()`. These options must not be
changed or overwritten. One way to ensure this is to set them right before
calling `pythia.init()`.


Consistency between LHEF file and Geneva options file
------------------------------------------------------

When showering events, it is recommended to always use the same Geneva version
and options used to generate them.

For convenience, the above functions check that the provided Geneva options are
consistent with the options used to generate the LHEF file (which are read from
the LHEF header). If inconsistencies are found, the program prints a warning
for each encountered option mismatch, and then throws a `Geneva::FatalError`
exception if there are any mismatches.

Note also that these checks are not necessarily exhaustive, and therefore extra
care should be taken when using different versions or option files for
generating and showering events.


Example code how to set up Pythia
----------------------------------

Here is a simple example main program that sets up Pythia8 to shower LHEF events
produced by Geneva.

```cpp
#include "Geneva/PythiaInterface/PythiaDriver.hpp"

#include "Pythia8/Pythia.h"

#include <iostream>

int main()
{
   // names of the event and pythia options file
   std::string lhefFile = "events.lhe";
   std::string genevaOptions = "input.yml";
   std::string pythiaOptions = "PYOptions.cmd";

   // our own Pythia instance
   Pythia8::Pythia pythia;

   // set the UserHook
   Pythia8::UserHooks* userHook = Geneva::PythiaDriver::getUserHook(lhefFile, genevaOptions);
   pythia.setUserHooksPtr(userHook);

   // set the Pythia options
   // first read in custom Pythia options
   pythia.readFile(pythiaOptions);

   // then set up Pythia options required by Geneva
   std::vector<std::string> requiredSettings = Geneva::PythiaDriver::getRequiredSettings(lhefFile, genevaOptions);
   for (std::vector<std::string>::iterator it = requiredSettings.begin(); it != requiredSettings.end(); ++it)
      pythia.readString((*it).c_str());

   // set up Pythia's LHEF reader
   Pythia8::LHAupLHEF* lhaupLhef = new Pythia8::LHAupLHEF(&pythia.info, lhefFile.c_str());
   pythia.setLHAupPtr(lhaupLhef);

   // initialize Pythia
   pythia.init();

   // call Pythia to shower the events and do something with them
   while (true) {
      if (pythia.next()) {
         // do something with showered event
         // ...
      } else {
         if (pythia.info.atEndOfFile())
            break;
         std::cout << "Could not shower event" << std::endl;
      }
   }
}
```

The easiest way to build this program is to use Geneva's CMake build system. To
do so, add the lines

```
add_executable (pythia-example path/to/pythia_example.cpp)
target_link_libraries (pythia-example PythiaInterface)
```

to the end of `src/CMakeLists.txt` and do

    build$ cmake .. && make pythia-example

This creates the binary `build/bin/pythia-example`.

To build the program from the outside, it must be compiled with the following
include paths

 * `<geneva_dir>/include`
 * `<build_dir>/include`
 * `<geneva_dir>/packages/gsl`
 * include path for GSL (if installed in a nonstandard location)
 * include path for Pythia8

where `<geneva_dir>` is the root of the Geneva source tree, and `<build_dir>` is
the build directory (typically `<geneva_dir>/build`). It needs to be linked
against

 * `<build_dir>/lib/libGenevaPythiaInterface.<ext>`
 * `<build_dir>/lib/libGenevaCore.<ext>`
 * GSL libraries (typically `-lgsl -lgslcblas`)
 * Pythia8 libraries

where `<ext>` is the appropriate library extension.

Note that the PythiaInterface library depends on most of the other Geneva
libraries, so using it requires a full build of Geneva. At present, it might
also require to include additional external libraries or to set the rpath to the
external library installation (e.g. for OpenLoops), depending on the specific
installation of the external libraries. In case of problems, use CMake as above
and do `make pythia-example VERBOSE=1` to see the full compile and link commands
generated by CMake.


[links]: #----------------------------------------------------------------------

[Main]:               ../README.md
[Installation]:       Installation.md
[User Guide]:         UserGuide.md
[Tutorial]:           tutorial/Tutorial.md
[Index]:              Index.md
[GENEVA]:             geneva.jpg
