<!------------------------------------------------------------------------------
# Author(s):
#    Christian Bauer, Simone Alioli, Frank Tackmann
#
# Copyright:
#    Copyright (C) 2017, 2018 Geneva Collaboration
#
#    This file is part of the Geneva MC framework. Geneva is distributed under
#    the terms of the GNU General Public License version 3 (GPLv3), see the
#    COPYING file that comes with this distribution for details.
#    Please respect the academic usage guidelines in the GUIDELINES file.
#
# Description:
#    User guide for how to run the code
#------------------------------------------------------------------------------>
![GENEVA]
##### [Main] | [Installation] | User Guide | [Tutorial] | [Index] #####
--------------------------------------------------------------------------------


User Guide
===========

Overview
---------

The running of the code proceeds through separate stages. A generic run requires
the following 4 main stages:

 1. **Setup stage:** Generates various grid files used in the generation of events

 2. **Generate stage:** Generates partonic LHEF events (which by default are partially unweighted)

 3. **Reweight stage:** Reweights the 0-jet bin to include power-suppressed NNLO terms

 4. **Shower stage:** Showers the LHEF events with Pythia8 to create HepMC events

The results from each of these stages are necessary for the next stage. While it
is in principle possible to do all these stages sequentially in a single run,
this is generally not advised. Instead, each stage is best run separately, such
that the results at each stage can be inspected and actions be taken in case of
problems. This also allows for an easier parallelization of the tasks that need
to be performed at each stage.

For production runs, we recommend the parallelization over O(100) cores via MPI,
where available. Running on a cluster might require to submit the commands to
the cluster's scheduler. An example script for the NERSC clusters, which use the
SLURM queuing system, is provided in `build/bin/geneva-NERSC`.

A step-by-step example run is provided in the [Tutorial].


Python interface
-----------------

### Overview ###

The recommended way of running the code is through the main `geneva` python
script, which after building the code will be in `build/bin/geneva`.

The Python interface requires the `genevapylib` Python library in
`build/lib/pyton2.7/site-packages` to be found via the system's `PYTHONPATH`.
When executing any of the Python scripts, they check if this is the case, and if
not will instruct you to run

    build$ source bin/geneva-env.sh

which sets the `PYTHONPATH` appropriately.

The general syntax for the Python interface is

    build$ bin/geneva <stage> <options>

The possible values for `<stage>` are given by

    $ geneva --help

The possible `<options>` for each stage are given by

    $ geneva <stage> --help

The following arguments are common to the four main stages:

 * `<option_file>`: Geneva option file

This is required for most stages. The layout of the `<option_file>` in YAML
format is explained in [Option Files].

 * `--start-seed <seed>`: First random number seed (seeds of subsequent runs are incremented by 1)
 * `--num-runs <runs>`: Number of parallel runs to perform
 * `--mpi`: Parallelize runs with MPI (see below for details)
 * `--dry-run`: Only print out commands without executing them
 * `--quiet`: Suppress all print output except errors

Both `--num-runs <runs>` and `--mpi` are options to parallelize the code.
Without `--mpi`, parallel subprocesses up to the number of `<runs>` are used,
so `<runs>` should typically be chosen to equal the number of available cores.

The output of individual commands is written to separate log files in the `log`
subdirectory of the output path of each stage. Only a minimal amount of status
information is written to the shell, which is suppressed with `--quiet` option.

In case of errors, a brief message is given including the name of the relevant
log file that should contain more detailed information. If one of several
parallel runs fails due to an error, the remaining runs belonging to the same
substage will continue running, but the program will stop after that substage.


#### Running with MPI ####

The Python interface can be used to run Geneva in parallel using MPI on any
MPI-aware machine or cluster. (This requires the `mpi4py` Python package.) To
execute a stage via MPI, use

    build$ mpirun <mpi_options> bin/geneva <stage> <options> --mpi

or an equivalent command to start MPI. In this case, the number of available MPI
processes must be at least as large as the number of `<runs>` (for the setup
stage it must match precisely). For the relevant `<mpi_options>` (e.g. how to
specify hosts, the number of MPI processes, etc.) please consult the
documentation of your MPI distribution.

Be aware that when running with MPI the `--mpi` option is required. It tells
Geneva that it is being run inside an MPI environment (i.e. that it was called
via `mpirun`), so it can distribute individual runs among the MPI processes.
When using OpenMPI, the MPI environment is detected automatically so the `--mpi`
may be omitted. (For other MPI distributions, omitting it would lead to
executing the identical stage duplicated on each MPI process, which is not what
you want.)

In case of runtime errors, a clean shutdown of the MPI environment is performed
after the substage in which the error occured. In particular, care is taken to
avoid MPI deadlocks, i.e. hanging MPI processes. (If you still encounter an MPI
deadlock, please send us a bug report.)


### 1. Setup stage ###

    build$ bin/geneva setup <option_file> --points <points> --iterations <iterations> [--start-seed <seed>] [--num-runs <runs>]

The setup stage produces several setup files, each starting with the `run_name`
specified in the `<option_file>`. These files are required for the subsequent
generate stage. Details about the various setup files produced can be found in
[Setup Files]. By default, they are located in the `setup` subdirectory. This
location can be changed with the `input_output::file_path::setup` option in the
`<option_file>`.

The integration grids are computed using the given number of `<iterations>`
and `<points>` per iteration (where the `<points>` are distributed over the
given number of `<runs>`). To obtain decent statistics, it is recommended to use
`<points>` >= 50M and `<iterations>` >= 3.


### 2. Generate stage ###

    build$ bin/geneva generate <option_file> [--start-seed <seed>] [--num-runs <runs>]

The generate stage produces LHEF files named
`<run_name>[_<generate_tag>]_<s>.lhe[.gz]`, where `<run_name>` and
`<generate_tag>` are taken from the `<option_file>`, and the seed `<s>` takes
the values in the range [`<seed>`, `<seed>+<runs>`). The LHEF files are
automaticaly gzipped if CMake's variable `geneva_enable_zlib` is ON and zlib
is found on the system.

All generated files are placed by default in the `generate` subdirectory. This
location can be changed with the `input_output::file_path::generate` option in
the `<option_file>`.

The generate stage requires the setup files produced in the setup stage. Their
location is given by the `input_output::file_path::setup` option in the
`<option_file>`, which defaults to the `setup` subdirectory.


### 3. Reweight stage ###

    build$ bin/geneva reweight <option_file> --sigma-below <central> <up> <down>  [--start-seed <seed>] [--num-runs <runs>]

Since Geneva's NNLO calculation is based on N-jettiness subtractions, the
0-jet cross section for Tau0 < Tau0cut at order alpha_s^2 is only correct up to
small power-suppressed terms. To achieve exact NNLO accuracy, these
contributions have to be included separately, which is achieved during this
stage by reweighting the 0-jet events in the LHEF files produced in the generate
stage. In addition, all events are normalized to the final total cross section
from combining the accumulated statistics across all runs.

The `--sigma-below` argument is required to provide the exact NNLO 0-jet
cross sections for Tau0 < Tau0cut. The `<central>`, `<up>`, and `<down>` values
are given in [NNLO Cross Section] or can be calculated as discussed there.

Here, the values of `<seed>` and `<runs>` are only used to determine which event
files are reweighted. So to reweight the files produced in the generate stage,
the same values must be used here.

The reweight stage expects the LHEF files in the `generate` subdirectory. This
location can be changed by the `input_output::file_path::generate` option in the
`<option_file>`. The LHEF files are reweighted in place, and no backup of the
original files are performed. The cross section before reweighting can be
obtained from the comments at the end of the file.


### 4. Shower stage ###

    build$ bin/geneva shower <option_file> <pythia_options> [--start-seed <seed>] [--num-runs <runs>] [...]

The shower stage showers the produced LHEF files with Pythia8. It takes the
following arguments:

 * `<pythia_options>`: Pythia8 option file

This is required and specifies the options for Pythia8 in its own format.

The values of `<seed>` and `<runs>` are used to determine which LHEF files are
showered. So to shower the files from previous stages, the same values must be
used here. Each event file is showered by a separate job, so the value of
`<runs>` also determines the number of parallel runs.

The shower stage also takes the following additional optional arguments:

 * `--outpath <path>`: Directory the output files are written to (defaults to `shower`)
 * `--shower-tag <tag>`: Extra tag added to the output files (defaults to empty)
 * `--max-time <minutes>`: Maximum runtime in minutes (unused by default)
 * `--max-events <num>`: Maximum number of events to shower (unused by default)
 * `--no-enforce-lhef-options`: Do not enforce consistency with the options in the LHEF files
 * `--no-hepmc`: Do not write out HepMC files

Unless `--no-hepmc` is specified, this produces HepMC files with names
`<run_name>[_<generate_tag>][_<shower_tag>]_<s>.hepmc[.gz]`, where
`<run_name>` and `<generate_tag>` are taken from the `<option_file>` and
`<shower_tag>` from the command line. If CMake's variable `geneva_enable_zlib`
is ON and zlib is found on the system the HepMC files are automatically gzipped.

The output of this stage is placed in the `<path>` directory, which defaults to
`shower`. By default all events in a LHEF file are showered. The `<minutes>` and
`<num>` arguments can be used to limit the runtime and/or number of events to
shower for each run.

The shower stage expects the LHEF files to be showered in the `generate`
subdirectory. This location can be changed with the
`input_output::file_path::generate` option in the `<option_file>`.


Pythia interface
-----------------

The shower stage described above utilizes a Pythia8 instance created internally
by Geneva. The LHEF files produced by Geneva can also be showered using an
external Pythia8 instance. This requires the external code that creates and
manages the Pythia8 instance to call and link to Geneva's Pythia8 interface
as explained in [Pythia Interface].


Rivet interface
----------------

The stages above are sufficient for a complete run producing physical results.
For convenience, an analyze-rivet stage to analyze the produced HepMC files with
Rivet is available when Rivet support is enabled (see [Installation]). It is
called as

    build$ bin/geneva analyze-rivet <filestem> -a <analysis> [-a <analysis> ...] [--start-seed <seed>] [--num-runs <runs>] [...]

The following arguments are required:

 * `<filestem>`: The full stem for the input files (may include a path)
 * `-a <analysis>`: Adds `<analysis>` to the list of Rivet analyses to perform

The following additional optional arguments are available:

 * `--last-seed <last>`: The last seed to consider (defaults to 0, which means automatic).
 * `--outpath <path>`: Directory the output files are written to (defaults to `rivet`)
 * `--num-weights <num>`: The number of weights to analyze (defaults to 1).

The files to operate on are constructed as `<filestem>_<s>.hepmc[.gz]`, where
`<s>` runs over the range of seeds [`<seed>`, `<last>`]. Each of the first
`<num>` weights in each file is analyzed by a separate job. If `<last>` is
specified, the number of parallel `<runs>` must be at least as large as `<num>`
times the number of files in the [`<seed>`, `<last>`] range. If `<last>` is not
given, it is automatically taken to be the largest possible value for the given
values of `<runs>` and `<num>`.

This stage produces `<filestem>_*.yoda` files, one for each weight and seed.
They are placed in the `<path>` directory, which defaults to `rivet`. These can
be merged and plotted as desired with the standard tools distributed with Rivet.


[links]: #----------------------------------------------------------------------

[Main]:               ../README.md
[Installation]:       Installation.md
[Tutorial]:           tutorial/Tutorial.md
[Option Files]:       OptionFiles.md
[Setup Files]:        SetupFiles.md
[Pythia Interface]:   PythiaInterface.md
[NNLO Cross Section]: NNLOCrossSection.md
[Index]:              Index.md
[GENEVA]:             geneva.jpg
