//------------------------------------------------------------------------------
/// \file AnalyzerBase.hpp
//
// Author(s):
//    Abel Ruiz, Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2013 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of AnalyzerBase
//------------------------------------------------------------------------------

#ifndef GENEVA_ANALYZER_BASE_HPP
#define GENEVA_ANALYZER_BASE_HPP

#include "Geneva/Analyzers/IAnalyzer.hpp"
#include "Geneva/Analyzers/Hist1D.hpp"
#include "Geneva/Analyzers/SingleBin.hpp"

#include <string>
#include <map>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class AnalyzerBase
 * \ingroup Analyzers
 *
 * \brief Base class implementing the IAnalyzer interface
 *
 * Provides a basic implementation of the IAnalyzer interface. It holds a map
 * of Hist1D and SingleBin and functions to create and access them by
 * name. Derived classes only need to create them and provide the actual
 * analysis routines via the analyzerBase hook functions.
 */
//------------------------------------------------------------------------------
class AnalyzerBase : public IAnalyzer
{
   private:
      std::map<std::string, SingleBin> _singleBins;
      std::map<std::string, Hist1D> _hists1D;

      /// whether the event has been properly initialized and can receive a new correlated weight
      bool _canAddCorrelatedWeight {false};

   public:
      virtual const SingleBin& singleBin(const std::string& name) const;
      virtual SingleBin& singleBin(const std::string& name);

      virtual const Hist1D& hist1D(const std::string& name) const;
      virtual Hist1D& hist1D(const std::string& name);

      virtual std::vector<SingleBin> getSingleBins() const;
      virtual std::vector<Hist1D> getHist1Ds() const;

      virtual void addSingleBin(const std::string& name);

      virtual void addHist1D(const std::string& name, size_t nbins, double binlow, double binhigh, bool isLinear);
      virtual void addHist1D(const std::string& name, const std::vector<double> binLimits, bool isLinear);

      virtual void initializeCorrelated();
      virtual void analyzeCorrelated(const Event& correlatedEvent);
      virtual void finalizeCorrelated();

      virtual void analyzePartonic(const Event& event);

      virtual void analyzeShowered(const Event& showeredEvent, const Configuration& hardProcess);
      virtual void analyzeShoweredNoMPI(const Event& showeredEvent, const Configuration& hardProcess);
      virtual void analyzeHadronized(const Event& showeredEvent, const Configuration& hardProcess);

      virtual void reset();

   protected:
      enum Level
      {
         kUnknown = 0,
         kFixedOrder = 1,
         kPartonic = 2,
         kShowered = 3,
         kShoweredNoMPI = 4,
         kHadronized = 5
      };

      /// Returns the default suffix for analyzer level.
      const char* levelSuffix(Level level) const;

   private:
      /// Analysis for a single \a event to be implemented by derived classes.
      virtual void analyzeBase(const Event& event, Level level) = 0;

      /// Analysis for a single (showered) \a event belonging to \a hardProcess.
      virtual void analyzeBase(const Event& event, const Configuration& hardProcess, Level level);

      virtual std::ostream& print(std::ostream& out) const;

      // helper function to initialize/finalize only Hists1Ds and SingleBins belonging to a given level
      void initializeLevel(Level level);
      void finalizeLevel(Level level);
};

////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Provides a default implementation that discards the provided hardProcess
// for analyzers that don't need to retain its information. Must be explicitly
// overwritten by derived classes that need hardProcess information.
inline void AnalyzerBase::analyzeBase(const Event& event, const Configuration&, Level level)
{
   analyzeBase(event, level);
}

} // namespace Geneva

#endif // GENEVA_ANALYZER_BASE_HPP
