//------------------------------------------------------------------------------
/// \file Hist1D.hpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 CERN, DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Hist1D
//------------------------------------------------------------------------------

#ifndef GENEVA_HIST_1D_HPP
#define GENEVA_HIST_1D_HPP

#include "Geneva/Core/Physics/Sigma.hpp"

#include <cstddef>
#include <string>
#include <vector>
#include <map>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class Hist1D
 * \ingroup Analyzers
 *
 * \brief Implements a one-dimensional histogram.
 *
 * The histogram is Sigma-valued and so automatically keeps track of the central
 * value and all variations. It supports linear, log binning, and variable user-
 * defined bin limits. Underflow and overflow bins are included.
 *
 * The basic operation is to fill in event weights (Sigma) according to a given
 * bin value and retrieve the accumulated binned cross sections at the end. It
 * supports filling with correlated weights.
 *
 * Basic algebraic operations are also supported.
 */
//------------------------------------------------------------------------------
class Hist1D
{
   private:
      std::string _name;                  ///< The name of the histogram
      size_t _nbins;                      ///< The number of bins
      std::vector<double> _binEdges;      ///< The bin edges
      bool _isLinear;                     ///< True if histogram is linearly binned

      std::vector<Sigma> _sumWeights;     ///< Sum of weights per bin
      std::vector<Sigma> _sumWeightsSq;   ///< Sum of squared weights per bin
      std::vector<size_t> _entries;       ///< Number of nonzero entries per bin
      size_t _totEntries;                 ///< Number of total entries in the histogram
      std::map<size_t, Sigma> _tmpwt;     ///< Sparse array-like temporary container for the correlated weights

      bool _canAddWeight;                 ///< Whether Hist1D can receive a new weight
      bool _variationsResized;            ///< Whether the vector of variations has been resized

   public:
      /// \name Constructors etc.
      ///@{
      Hist1D(const std::string& name, size_t nbins, double binlow, double binhigh, bool isLinear);
      Hist1D(const std::string& name, const std::vector<double> binlimits, bool isLinear);
      ///@}

      /// \name Queries and Accessors
      ///@{
      const std::string& name() const;                ///< Returns the name of the %Hist1D.
      bool canAddWeight() const;                      ///< Returns `true` if the %Hist1D can receive new (correlated) weights.

      const std::vector<Sigma>& sumWeights() const;   ///< Returns the sum of weights per bin.
      const std::vector<Sigma>& sumWeightsSq() const; ///< Returns the sum of squared weights per bin.

      const std::vector<size_t>& entries() const;     ///< Returns the nonzero entries per bin.
      size_t totEntries() const;                      ///< Returns the total number of entries.

      // Calculates the absolute bin cross sections.
      std::vector<Sigma> valuesAbs() const;

      // Calculates the statistical uncertainties on the absolute bin cross sections.
      std::vector<Sigma> errorsAbs() const;

      // Calculates the differential cross section.
      std::vector<Sigma> valuesDiff() const;

      // Calculates the statistical uncertainties on the differential cross section.
      std::vector<Sigma> errorsDiff() const;

      size_t nBins() const;                           ///< Returns the number of bins in the Hist1D (excluding underflow and overflow)
      const std::vector<double>& binEdges() const;    ///< Returns the bin edges.
      bool isLinear() const;                          ///< Returns `true` if the histogram is linearly binned and `false` otherwise.

      std::vector<double> binCenters() const;         ///< Returns the bin centers.
      ///@}

      /// \name Manipulation
      ///@{

      /// Start receiving new (correlated) weights.
      void initialize();

      /// Adds \a weight to the bin corresponding to \a value.
      void fill(double value, const Sigma& weight);

      /// Adds \a weight to the bin corresponding to \a value and all lower bins.
      void fillUpTo(double value, const Sigma& weight);

      /// Adds \a weight to the bin corresponding to \a value and all higher bins.
      void fillUpFrom(double value, const Sigma& weight);

      /// Finish receiving (correlated) weights.
      void finalize();

      // Assigns the given bin data to the Hist1D.
      void assign(const std::vector<Sigma>& sumWeights,
                  const std::vector<Sigma>& sumWeightsSq,
                  const std::vector<size_t>& entries, size_t totEntries);

      /// Resets the %Hist1D.
      void reset();
      ///@}

      /// \name Algebra
      ///@{
      // Hist1D addition assignment operator
      Hist1D& operator+=(const Hist1D& rhs);

      // Hist1D subtraction assignment operator
      Hist1D& operator-=(const Hist1D& rhs);

      // Hist1D multiplication with real number assignment operator
      Hist1D& operator*=(double rhs);

      // Hist1D division by real number assignment operator
      Hist1D& operator/=(double rhs);
      ///@}

      /// \name Output
      ///@{
      // Hist1D stream insertion operator
      friend std::ostream& operator<<(std::ostream& output, const Hist1D& hist);
      ///@}

   private:
      /// Streams a table with the central bin cross sections to \a out.
      std::ostream& print(std::ostream& out) const;

      void setBinEdges(double binlow, double binhigh);
      Sigma getBinError(size_t c) const;
};

// Hist1D addition operator
const Hist1D operator+(const Hist1D& lhs, const Hist1D& rhs);

// Hist1D subtraction operator
const Hist1D operator-(const Hist1D& lhs, const Hist1D& rhs);

// Hist1D multiplication operator with real number from the left
const Hist1D operator*(double lhs, const Hist1D& rhs);

// Hist1D multiplication operator with real number from the right
const Hist1D operator*(const Hist1D& lhs, double rhs);

// Hist1D division operator with real number
const Hist1D operator/(const Hist1D& lhs, double rhs);


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Queries and Accessors
//------------------------------------------------------------------------------

inline const std::string& Hist1D::name() const                { return _name; }

inline bool Hist1D::canAddWeight() const                      { return _canAddWeight; }

inline const std::vector<Sigma>& Hist1D::sumWeights() const   { return _sumWeights; }
inline const std::vector<Sigma>& Hist1D::sumWeightsSq() const { return _sumWeightsSq; }

inline const std::vector<size_t>& Hist1D::entries() const     { return _entries; }
inline size_t Hist1D::totEntries() const                      { return _totEntries; }

inline size_t Hist1D::nBins() const                           { return _nbins; }
inline const std::vector<double>& Hist1D::binEdges() const    { return _binEdges; }
inline bool Hist1D::isLinear() const                          { return _isLinear; }

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %Hist1D division by real number assignment operator
 * \param rhs The number to be divided by.
 * \return A reference to the %Hist1D.
 *
 * Each component of the %Hist1D is divided by \a rhs.
 * \note Division by zero is not checked.
 */
inline Hist1D& Hist1D::operator/=(double rhs)
{
   return operator*=(1. / rhs);
}

//------------------------------------------------------------------------------
/**
 * \brief %Hist1D addition operator
 * \param lhs The 1st %Hist1D to add.
 * \param rhs The 2nd %Hist1D to add.
 * \return A %Hist1D given by the sum of \a lhs and \a rhs.
 */
inline const Hist1D operator+(const Hist1D& lhs, const Hist1D& rhs)
{
   Hist1D tmp(lhs);
   tmp += rhs;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Hist1D subtraction operator
 * \param lhs The 1st %Hist1D
 * \param rhs The 2nd %Hist1D to be subtracted from \a lhs.
 * \return A %Hist1D given by the difference of \a lhs and \a rhs.
 */
inline const Hist1D operator-(const Hist1D& lhs, const Hist1D& rhs)
{
   Hist1D tmp(lhs);
   tmp -= rhs;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Hist1D multiplication operator with real number from the left
 * \param lhs The number to multiply with.
 * \param rhs The %Hist1D to be multiplied by \a lhs.
 * \return A %Hist1D given by the product of \a lhs times \a rhs.
 */
inline const Hist1D operator*(double lhs, const Hist1D& rhs)
{
   Hist1D tmp(rhs);
   tmp *= lhs;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Hist1D multiplication operator with real number from the right
 * \param lhs The %Hist1D to be multiplied by \a rhs.
 * \param rhs The number to multiply with.
 * \return A %Hist1D given by the product of \a lhs times \a rhs.
 */
inline const Hist1D operator*(const Hist1D& lhs, double rhs)
{
   Hist1D tmp(lhs);
   tmp *= rhs;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Hist1D division operator with real
 * \param lhs The %Hist1D to be divided by \a rhs.
 * \param rhs The number to divide by.
 * \return A %Hist1D given by \a lhs divided by \a rhs.
 * \note Division by zero is not checked.
 */
inline const Hist1D operator/(const Hist1D& lhs, double rhs)
{
   Hist1D tmp(lhs);
   tmp /= rhs;
   return tmp;
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %Hist1D stream insertion operator
 * \param out The output stream to be used.
 * \param sigma The %Hist1D to be streamed.
 * \return A reference to \a out.
 *
 * Streams a table with one row per bin (including underflow and overflow) in
 * the format "bin  entries  central value +/- error" to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const Hist1D& hist)
{
   return hist.print(out);
}

} // namespace Geneva

#endif // GENEVA_HIST_1D_HPP
