//------------------------------------------------------------------------------
/// \file IAnalyzer.hpp
//
// Author(s):
//    Erik Strand, Chris Vermilion, Frank Tackmann, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class IAnalyzer
//------------------------------------------------------------------------------

#ifndef GENEVA_IANALYZER_HPP
#define GENEVA_IANALYZER_HPP

#include <iosfwd>
#include <string>
#include <vector>

namespace Geneva
{

class SingleBin;
class Hist1D;
class Event;
class Sigma;
class Configuration;

//------------------------------------------------------------------------------
/**
 * \defgroup Analyzer Analyzers
 *
 * This group contains all internal analysis functionality.
 */
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
/**
 * \class IAnalyzer
 * \ingroup Analyzer
 *
 * \brief Abstract interface to analyzers.
 */
//------------------------------------------------------------------------------
class IAnalyzer
{
   public:
      virtual ~IAnalyzer() = default;

      /// Returns a reference to the rate with \a name.
      virtual const SingleBin& singleBin(const std::string& name) const = 0;
      virtual SingleBin& singleBin(const std::string& name) = 0;

      /// Returns a reference to the histogram with \a name.
      virtual const Hist1D& hist1D(const std::string& name) const = 0;
      virtual Hist1D& hist1D(const std::string& name) = 0;

      /// Returns a list of all rates and histograms
      virtual std::vector<SingleBin> getSingleBins() const = 0;
      virtual std::vector<Hist1D> getHist1Ds() const = 0;

      /// Adds a rate with name \a name.
      virtual void addSingleBin(const std::string& name) = 0;

      /// Adds a histogram with name \a name and fixed binning.
      virtual void addHist1D(const std::string& name, size_t nbins, double binlow, double binhigh, bool isLinear) = 0;

      /// Adds a histogram with name \a name using the variable \a binLimits.
      virtual void addHist1D(const std::string& name, const std::vector<double> binLimits, bool isLinear) = 0;

      /// Starts the analysis of correlated events.
      virtual void initializeCorrelated() = 0;

      /// Analyzes the \a correlatedEvent.
      virtual void analyzeCorrelated(const Event& correlatedEvent) = 0;

      /// Finishes the analysis of correlated events.
      virtual void finalizeCorrelated() = 0;

      /// Analyzes the partonic \a event.
      virtual void analyzePartonic(const Event& event) = 0;

      /// Analyzes the \a showeredEvent.
      virtual void analyzeShowered(const Event& showeredEvent, const Configuration& hardProcess) = 0;
      virtual void analyzeShoweredNoMPI(const Event& showeredEvent, const Configuration& hardProcess) = 0;
      virtual void analyzeHadronized(const Event& showeredEvent, const Configuration& hardProcess) = 0;

      /// Resets the analyzer.
      virtual void reset() = 0;

      /// Returns the name of the Analyzer.
      virtual std::string name() const = 0;

      /// Analyzer stream insertion operator.
      friend std::ostream& operator<<(std::ostream& out, const IAnalyzer& analyzer);

   protected:
      IAnalyzer() = default;
      IAnalyzer(const IAnalyzer&) = default;
      IAnalyzer(IAnalyzer&&) = default;
      IAnalyzer& operator=(const IAnalyzer&) = default;
      IAnalyzer& operator=(IAnalyzer&&) = default;

   private:
      /// Streams the Analyzer to \a out.
      virtual std::ostream& print(std::ostream& out) const = 0;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline std::ostream& operator<<(std::ostream& out, const IAnalyzer& analyzer)
{
   return analyzer.print(out);
}

} // namespace Geneva

#endif // GENEVA_IANALYZER_HPP
