//------------------------------------------------------------------------------
/// \file OutputStreamAnalyzer.hpp
//
// Author(s):
//    Jon Walsh
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class OutputStreamAnalyzer
//------------------------------------------------------------------------------

#ifndef GENEVA_OUTPUT_STREAM_ANALYZER_HPP
#define GENEVA_OUTPUT_STREAM_ANALYZER_HPP

#include "Geneva/Analyzers/Hist1D.hpp"
#include "Geneva/Analyzers/SingleBin.hpp"
#include "Geneva/Analyzers/IAnalyzer.hpp"

#include "Geneva/Core/Utils/MathUtils.hpp"

#include <iomanip>
#include <ostream>
#include <string>
#include <sstream>
#include <vector>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class OutputStreamAnalyzer
 * \ingroup Analyzer
 *
 * \brief An output stream that is used to output the Analyzer in Mathematica format.
 *
 * This class will output the Rate and Hist1D objects into a format that makes
 * them easy to manipulate in Mathematica, with the help of the Analyzer package.
 *
 * \note The actual Histograms and Rates are acquired when OutputStreamAnalyzer is contructed.
 * They can queried and streamed later, but their content is fixed at construction time.
 */
//------------------------------------------------------------------------------
class OutputStreamAnalyzer
{
   private:
      std::vector<SingleBin> _rates;  ///< SingleBins to be streamed
      std::vector<Hist1D> _hists;     ///< Histograms to be streamed
      std::string _stem;              ///< A stem identifying a particular run of Geneva

   public:
      OutputStreamAnalyzer(IAnalyzer* analyzer, std::string stem);
      OutputStreamAnalyzer(std::vector<SingleBin> rates, std::vector<Hist1D> hists, std::string stem);

      friend std::ostream& operator<<(std::ostream& out, const OutputStreamAnalyzer& osa);      ///< Specific print function accepting OutputStreamAnalyzer for input

   private:
      std::string rateString(const SingleBin& rate) const;              ///< Generates a string that is the Mathematica Analyzer output of the SingleObserv object
      std::string rateNames() const;                                    ///< Generates a Mathematica list of SingleObserv names
      std::string hist1DString(const Hist1D& hist) const;               ///< Generates a string that is the Mathematica Analyzer output of the Hist1D object
      std::string histNames() const;                                    ///< Generates a Mathematica list of Hist1D names

      std::string initialize() const;                                   ///< Readies the list to start accepting values
      std::string stringSciNotation(double value) const;                ///< Converts a double to a string in scientific notation in Mathematica format, with e -> *^
      std::string stringSciNotation(std::vector<double> value) const;   ///< Converts a vector of double to a list of strings in scientific notation in Mathematica format, with e -> *^

   protected:
      std::ostream& getOutput(std::ostream& stream) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Declarations
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
/**
 * Constructor
 * \param analyzer The analyzer whose output should be formatted
 * \param stem The stem which serves as an identifier of the particular run of Geneva
 */
inline OutputStreamAnalyzer::OutputStreamAnalyzer(IAnalyzer* analyzer, std::string stem)
   : _stem(stem)
{
   if (analyzer) {
      _rates = analyzer->getSingleBins();
      _hists = analyzer->getHist1Ds();
   }
}

//------------------------------------------------------------------------------
/**
 * Constructor
 * \param rates The list of rates whose output should be formatted
 * \param hists The list of histograms whose output should be formatted
 * \param stem The stem which serves as an identifier of the particular run of Geneva
 */
inline OutputStreamAnalyzer::OutputStreamAnalyzer(std::vector<SingleBin> rates, std::vector<Hist1D> hists, std::string stem)
: _rates(rates), _hists(hists), _stem(stem)
{}


//------------------------------------------------------------------------------
inline std::string OutputStreamAnalyzer::initialize() const
{
   return "(* This is the Analyzer output of Geneva *)\n\n";
}

//------------------------------------------------------------------------------
inline std::string OutputStreamAnalyzer::stringSciNotation(double value) const
{
   std::string expstr("e");
   std::string newexpstr("*^");

   std::stringstream valstream("");
   valstream << std::setprecision(15) << std::scientific << value;
   std::string valstr = valstream.str();

   size_t exppos = valstr.find(expstr);
   if (exppos != std::string::npos) {
      ptrdiff_t diff = ptrdiff_t_from_size_t(exppos);
      valstr.replace(valstr.begin() + diff, valstr.begin() + diff + 1, newexpstr);
   }

   return valstr;
}

//------------------------------------------------------------------------------
inline std::string OutputStreamAnalyzer::stringSciNotation(std::vector<double> value) const
{
   std::stringstream valstream("");

   valstream << "{";
   for (size_t c = 0; c < value.size(); c++) {
      valstream << OutputStreamAnalyzer::stringSciNotation(value[c]);
      if (c < value.size() - 1) { valstream << ", "; }
      else { valstream << "}"; }
   }

   return valstream.str();
}

} // namespace Geneva

#endif // GENEVA_OUTPUT_STREAM_ANALYZER_HPP
