//------------------------------------------------------------------------------
/// \file SingleBin.hpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 CERN, DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SingleBin
//------------------------------------------------------------------------------

#ifndef GENEVA_SINGLE_BIN_HPP
#define GENEVA_SINGLE_BIN_HPP

#include "Geneva/Core/Physics/Rate.hpp"

#include <cstddef>
#include <string>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class SingleBin
 * \ingroup Core
 *
 * \brief Implements a single cross section SingleBin.
 *
 * A SingleBin corresponds to a single histogram bin. Its value is a Rate, which
 * automatically keeps track of the central value and all variations.
 *
 * The basic operation is to fill in event weights (Sigma) and retrieve the
 * accumulated cross section at the end. It supports filling with correlated
 * weights.
 *
 * Basic algebraic operations are also supported.
 */
//------------------------------------------------------------------------------
class SingleBin
{
   private:
      std::string _name;          ///< name of the SingleBin
      Rate _rate;                 ///< rate of the SingleBin

      Sigma _tmpwt;               ///< temp container for the correlated weights
      bool _canAddWeight;         ///< whether the SingleBin can receive a new weight

   public:
      /// \name Constructors etc.
      ///@{
      /// Creates a SingleBin with \a name and zero Rate.
      SingleBin(const std::string& name);

      /// Creates a SingleBin with \a name and given data.
      SingleBin(const std::string& name, Sigma sumWeights, Sigma sumWeightsSq,
                size_t entries, size_t totEntries);

      ///@}

      /// \name Queries and Accessors
      ///@{
      const Rate& rate() const;             ///< Returns the rate of the %SingleBin.

      const std::string& name() const;      ///< Returns the name of the %SingleBin.

      bool canAddWeight() const;            ///< Returns `true` if the %SingleBin can receive new (correlated) weights.

      const Sigma& sumWeights() const;      ///< Returns the sum of weights.
      const Sigma& sumWeightsSq() const;    ///< Returns the sum of squared weights.

      size_t entries() const;               ///< Returns the number of nonzero entries.
      size_t totEntries() const;            ///< Returns the number of total entries.

      Sigma value() const;                  ///< Returns the value.
      Sigma error() const;                  ///< Returns the statistical uncertainty.
      ///@}

      /// \name Manipulation
      ///@{
      void initialize();                    ///< Start receiving new (correlated) weights.

      void fill(const Sigma& weight);       ///< Adds a (correlated) weight.

      void finalize();                      ///< Finish receiving (correlated) weights.

      void reset();                         ///< Resets the SingleBin.
      ///@}

      /// \name Algebra
      ///@{
      // SingleBin addition assignment operator
      SingleBin& operator+=(const SingleBin& rhs);

      // SingleBin subtraction assignment operator
      SingleBin& operator-=(const SingleBin& rhs);

      // SingleBin multiplication with real number assignment operator
      SingleBin& operator*=(double rhs);

      // SingleBin division by real number assignment operator
      SingleBin& operator/=(double rhs);
      ///@}

      /// \name Output
      ///@{
      // SingleBin stream insertion operator
      friend std::ostream& operator<<(std::ostream& out, const SingleBin& sb);
      ///@}

   private:
      /// Streams a string in the format "name  entries  central value +/- error" to \a out.
      std::ostream& print(std::ostream& out) const;
};

// SingleBin addition operator
const SingleBin operator+(const SingleBin& lhs, const SingleBin& rhs);

// SingleBin subtraction operator
const SingleBin operator-(const SingleBin& lhs, const SingleBin& rhs);

// SingleBin multiplication operator with real number from the left
const SingleBin operator*(double lhs, const SingleBin& rhs);

// SingleBin multiplication operator with real number from the right
const SingleBin operator*(const SingleBin& lhs, double rhs);

// SingleBin division operator with real number
const SingleBin operator/(const SingleBin& lhs, double rhs);


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Creates a SingleBin with \a name and zero Rate.
inline SingleBin::SingleBin(const std::string& name)
   : _name(name), _canAddWeight(false)
{}

//------------------------------------------------------------------------------
// Creates a SingleBin with name and given data.
inline SingleBin::SingleBin(const std::string& name, Sigma sumWeights, Sigma sumWeightsSq,
                            size_t entries, size_t totEntries)
   : _name(name), _rate(sumWeights, sumWeightsSq, entries, totEntries), _canAddWeight(false)
{}

//------------------------------------------------------------------------------
// Queries and Accessors
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
inline const Rate& SingleBin::rate() const
{
   assert(!_canAddWeight);
   return _rate;
}

//------------------------------------------------------------------------------
inline const std::string& SingleBin::name() const   { return _name; }

inline bool SingleBin::canAddWeight() const         { return _canAddWeight; }

inline const Sigma& SingleBin::sumWeights() const   { return rate().sumWeights(); }
inline const Sigma& SingleBin::sumWeightsSq() const { return rate().sumWeightsSq(); }

inline Sigma SingleBin::value() const               { return rate().value(); }
inline Sigma SingleBin::error() const               { return rate().error(); }

inline size_t SingleBin::entries() const            { return rate().entries(); }
inline size_t SingleBin::totEntries() const         { return rate().totEntries(); }

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %SingleBin addition assignment operator
 * \param rhs The %SingleBin to be added.
 * \return A reference to the %SingleBin.
 *
 * \a rhs is added to the %SingleBin.
 */
inline SingleBin& SingleBin::operator+=(const SingleBin& rhs)
{
   assert(!(_canAddWeight || rhs.canAddWeight()));
   _rate += rhs._rate;
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %SingleBin subtraction assignment operator
 * \param rhs The %SingleBin to be subtracted.
 * \return A reference to the %SingleBin.
 *
 * \a rhs is subtracted from the %SingleBin.
 */
inline SingleBin& SingleBin::operator-=(const SingleBin& rhs)
{
   assert(!(_canAddWeight || rhs.canAddWeight()));
   _rate -= rhs._rate;
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %SingleBin multiplication with real number assignment operator
 * \param rhs The number to be multiplied by.
 * \return A reference to the %SingleBin.
 *
 * Each component of the %SingleBin is multiplied by \a rhs.
 */
inline SingleBin& SingleBin::operator*=(double rhs)
{
   assert(!_canAddWeight);
   _rate *= rhs;
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %SingleBin division by real number assignment operator
 * \param rhs The number to be divided by.
 * \return A reference to the %SingleBin.
 *
 * Each component of the %SingleBin is divided by \a rhs.
 * \note Division by zero is not checked.
 */
inline SingleBin& SingleBin::operator/=(double rhs)
{
   return operator*=(1. / rhs);
}

//------------------------------------------------------------------------------
/**
 * \brief %SingleBin addition operator
 * \param lhs The 1st %SingleBin to add.
 * \param rhs The 2nd %SingleBin to add.
 * \return A %SingleBin given by the sum of \a lhs and \a rhs.
 */
inline const SingleBin operator+(const SingleBin& lhs, const SingleBin& rhs)
{
   SingleBin tmp(lhs);
   tmp += rhs;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %SingleBin subtraction operator
 * \param lhs The 1st %SingleBin
 * \param rhs The 2nd %SingleBin to be subtracted from \a lhs.
 * \return A %SingleBin given by the difference of \a lhs and \a rhs.
 */
inline const SingleBin operator-(const SingleBin& lhs, const SingleBin& rhs)
{
   SingleBin tmp(lhs);
   tmp -= rhs;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %SingleBin multiplication operator with real number from the left
 * \param lhs The number to multiply with.
 * \param rhs The %SingleBin to be multiplied by \a lhs.
 * \return A %SingleBin given by the product of \a lhs times \a rhs.
 */
inline const SingleBin operator*(double lhs, const SingleBin& rhs)
{
   SingleBin tmp(rhs);
   tmp *= lhs;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %SingleBin multiplication operator with real number from the right
 * \param lhs The %SingleBin to be multiplied by \a rhs.
 * \param rhs The number to multiply with.
 * \return A %SingleBin given by the product of \a lhs times \a rhs.
 */
inline const SingleBin operator*(const SingleBin& lhs, double rhs)
{
   SingleBin tmp(lhs);
   tmp *= rhs;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %SingleBin division operator with real
 * \param lhs The %SingleBin to be divided by \a rhs.
 * \param rhs The number to divide by.
 * \return A %SingleBin given by \a lhs divided by \a rhs.
 * \note Division by zero is not checked.
 */
inline const SingleBin operator/(const SingleBin& lhs, double rhs)
{
   SingleBin tmp(lhs);
   tmp /= rhs;
   return tmp;
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %SingleBin stream insertion operator
 * \param out The output stream to be used.
 * \param sigma The %SingleBin to be streamed.
 * \return A reference to \a out.
 *
 * Streams a string in the format "name  entries  central value +/- error" to
 * \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const SingleBin& bin)
{
   return bin.print(out);
}

} // namespace Geneva

#endif // GENEVA_SINGLE_BIN_HPP
