//------------------------------------------------------------------------------
/// \file CalcNLOppV.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class CalcNLOppV
//------------------------------------------------------------------------------

#ifndef GENEVA_CALC_NLO_PPV_HPP
#define GENEVA_CALC_NLO_PPV_HPP

#include "Geneva/Calculations/ICalculation.hpp"
#include "Geneva/Calculations/EngineNLOpp.hpp"
#include "Geneva/Calculations/PhiNLOppVFKS.hpp"

#include "Geneva/Core/Physics/SignatureSet.hpp"
#include "Geneva/Core/Physics/Configuration.hpp"

namespace Geneva
{

class OptionMap;

//------------------------------------------------------------------------------
/**
 * \class CalcNLOppV
 * \ingroup calc
 *
 * \brief NLO calculation for pp -> (V -> leptons)
 *
 * CalcNLOppV represents an NLO calculation for \f$\frac{\tex t{d}\sigma}{\text{d} \Phi}\f$
 * for the process \f$p p \to V \to leptons \f$.
 * For a given phase space point Phi2 it sums over all singular regions and for each region
 * performs the the integration over \f$\Phi_{\rm rad}\f$ using the FKS subtractions.
 */
//------------------------------------------------------------------------------
class CalcNLOppV : virtual public ICalculation
{
   private:
      /// helper class for NLO FKS calculations
      EngineNLOpp* _helper;
      /// The set of signatures this calculation will work with
      SignatureSet _sigSet;

      const PdgID  _idV;
      const PdgID  _idDecay;
      int _nLight;

      /// NLO definition of PhiN phase space
      PhiNLOppVFKS* _phiNLO;
      /// component that evaluates the electroweak coupling
      ICoupling* _alphaem;

      //------------------------------------------------------------------------------
      /**
       * \class CalcNLOppV::CommonCouplings
       * \ingroup calc
       *
       * \brief CommonCouplings functional definition for CalcNLOppV
       *
       * \todo: detailed description goes here
       */
      //------------------------------------------------------------------------------
      class CommonCouplings : public ICommonCouplings
      {
         private:
            ICoupling* _aem;

         public:
            CommonCouplings(ICoupling* aem)  : _aem(aem) {}

            virtual std::string name() const;
            virtual int numAlphaStrongPowers() const { return 0; };

         private:
            virtual double value(const Configuration& config, const double mu) const;
      };

      CommonCouplings* _commonCouplings;

   public:
      CalcNLOppV(const OptionMap& optMap,
                 ITreeME* treeME, ITreeMEComplete* treeMEComplete, IOneLoopME* oneLoopME,
                 UnitCubeRadGenerator* xRadGenerator, IPDF* pdf,
                 CouplingStrong* alphas, FixedOrder order, double Ecm, int nLight,
                 IScale* renScale, IScale* factScale, ICoupling* alphaem, double TauCut,
                 PdgID boson, PdgID lepton);

      virtual ~CalcNLOppV();

      virtual Sigma value(const Configuration& config);
      virtual Sigma fastEstimate(const Configuration& config);
      virtual const SignatureSet& neededSignatures() const;
      virtual IRestriction* getRestriction() const;
      virtual const std::string name() const;
      virtual const std::vector<NameValuePair> componentInfo() const;
      virtual void informGenerationWeight(const double w);
      virtual void initialize();
      virtual void finalize(const size_t numEvents);

      Sigma getLOCoeff(const Configuration& config, bool doVariations = true) const;
      NloCoefficients getFOCoeffs(const Configuration& config, bool doVariations = true) const;
      NloCoefficients getFOCoeffsBornKin(const Configuration& config, bool doVariations = true) const;

   private:
      friend class CalcNLOppVTest;

      SignatureSet getSignaturesN() const;
      SignatureSet getSignaturesNplusOne() const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Declarations
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline Sigma CalcNLOppV::value(const Configuration& config)
{
   return math_constants::gev2pb() * _helper->calculate(config, true);
}

//------------------------------------------------------------------------------
inline Sigma CalcNLOppV::fastEstimate(const Configuration& config)
{
   return math_constants::gev2pb() * _helper->getLOCoeff(config, false);
}

//------------------------------------------------------------------------------
inline const SignatureSet& CalcNLOppV::neededSignatures() const
{
   return _sigSet;
}

//------------------------------------------------------------------------------
inline IRestriction* CalcNLOppV::getRestriction() const
{
   return _helper->getRestriction();
}

//------------------------------------------------------------------------------
inline const std::string CalcNLOppV::name() const
{
   return "CalcNLOppV";
}

//------------------------------------------------------------------------------
inline void CalcNLOppV::informGenerationWeight(const double w)
{
   _helper->setGenerationWeight(w);
}

//------------------------------------------------------------------------------
inline void CalcNLOppV::initialize()
{
   _helper->resetAnalyzer();
}

//------------------------------------------------------------------------------
inline void CalcNLOppV::finalize(const size_t numEvents)
{
   _helper->finalize(numEvents);
}

//------------------------------------------------------------------------------
inline Sigma CalcNLOppV::getLOCoeff(const Configuration& config, bool doVariations) const
{
   return _helper->getLOCoeff(config, doVariations);
}

//------------------------------------------------------------------------------
inline NloCoefficients CalcNLOppV::getFOCoeffs(const Configuration& config, bool doVariations) const
{
   return _helper->getFOCoeffs(config, doVariations);
}

//------------------------------------------------------------------------------
inline NloCoefficients CalcNLOppV::getFOCoeffsBornKin(const Configuration& config, bool doVariations) const
{
   return _helper->getFOCoeffsBornKin(config, doVariations);
}

} // namespace Geneva

#endif // GENEVA_CALC_NLO_PPV_HPP
