//------------------------------------------------------------------------------
/// \file CalcSCETppV012.hpp
//
// Author(s):
//   Simone Alioli, Christian Bauer, Jon Walsh, Frank Tackmann
//
// Copyright:
//   Copyright (C) 2012 LBNL
//
//   This file is part of the Geneva MC framework. Geneva is distributed under
//   the terms of the GNU General Public License version 3 (GPLv3), see the
//   COPYING file that comes with this distribution for details.
//   Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//   Interface of class CalcSCETppV012
//------------------------------------------------------------------------------

#ifndef GENEVA_CALC_SCET_PPV012_HPP
#define GENEVA_CALC_SCET_PPV012_HPP

#include "Geneva/Calculations/ICalculation.hpp"
#include "Geneva/Calculations/CalcNLOppV.hpp"
#include "Geneva/Calculations/CalcNLOppVj.hpp"

#include "Geneva/Components/IPDF.hpp"
#include "Geneva/Components/IScale.hpp"

#include "Geneva/Core/Physics/Order.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/Utils/FileUtils.hpp"

#include <gsl/gsl_deriv.h>

namespace Geneva
{

class BeamFuncTauB;
class SplittingFunction;

//------------------------------------------------------------------------------
/**
 * \class CalcSCETppV012
 * \ingroup calc
 *
 * \brief Calculation to combine pp->V+0 and 1 jets at NLO
 *
 * \todo: detailed description goes here.
 */
//------------------------------------------------------------------------------
class CalcSCETppV012 : public ICalculation
{
   public:
      /// Named constants to specify scale choice
      enum Scale
      {
         kall,            ///< Run all scale choices
         kfastset,         ///< Run a subset of scale choices for quick error band estimation: central, up/down, Sup/Sdown
         khardset,         ///< Run a subset of scale choices for hard scale variations only: central, up/down
         kfixset,          ///< Run a subset of scale choices for fix scale choices only: central, kmuupfix, kmudownfix
         ksetupset,          ///< Run all calculations that are required for the setup of the calculation
         kmurun,          ///< default central running scales
         kmuup, kmudown,    ///< overall scale up and down (by factors of two)
         kmuBup, kmuBdown,   ///< Beam A,B scale up and down
         kmuSup, kmuSdown,   ///< soft scale up and down
         kmuxup, kmuxdown,   ///< transition points up and down
         kmuupfix, kmudownfix,   ///< up/down variation, with a XS fix to match the FO rate
         kmuhigherrun, kmuhigherup, kmuhigherdown,   ///< scales for an improved XS fix to match the FO rate
         kcanonical       ///< canonical scales
      };

      /// Named constants to specify rapidity dependence of profiles
      enum YProfiles
      {
         knone,             ///< Y-independent profiles
         knormal,           ///< normal moderate Y-dependence
         kstrong            ///< strong/aggressive Y-dependence
      };

      /// Named constants to specify the nonsingular profile
      enum NSProfile
      {
         khard,              ///< fixed at hard scale
         kbeam               ///< beam-like profile
      };

      /// Named constants to specify treatment of \f$ \Phi_1 \f$
      enum Phi1
      {
         kflat,           ///< Flat distribution in \f$ \Phi_1 \f$
         kmult,           ///< Multiplicative combination with \f$ d\sigma/d\Phi_1 \f$
         kadd             ///< Additive combination with \f$ d\sigma/d\Phi_1\f$
      };

   private:
      OptionMap _optMap;
      double _tau0cut;         ///< value of Tau separating 0 and 1 jets
      double _tau1cut;         ///< value of Tau separating 1 and 2 jets
      double _cmEnergy;         ///< Hadronic CM energy
      RunOrder _run0;           ///< order to which the running in Tau0 is performed
      RunOrder _run1;           ///< order to which the running in Tau1 is performed
      bool _split1Jet;           ///< whether to split the 1-jet
      bool _onlyExtraSplit1JetContribution;           ///< whether to split the 1-jet
      double _tau1IR;            ///< value to which the splitting of 1-jet is done
      FixedOrder _fixed0jet;     ///< fixed order to which 0 jet is included
      FixedOrder _fixed1jet;     ///< fixed order to which 1 jet is included
      Scale _inputScale;        ///< flag specifying the scale choices to use
      YProfiles _profilesY;      ///< flag specifying Y-dependence of profiles
      NSProfile _profileNS;      ///< flag specifying nonsingular profile
      Phi1 _phi1;             ///< flag specifying the \f$ \Phi_1 \f$ treatment to use

      bool _removeAlpha2;

      int _nLight;            ///< Number of light active flavors
      ICoupling* _alphaem;      ///< Ew coupling to use
      CouplingStrong* _alphas;   ///< running strong coupling

      static std::map<Scale, EngineNLOBase::ScaleVariation> fixedOrderScale; /// The static map connecting SCET scale variations with fixed-order ones
      static std::map<Scale, Scale> resummationScale; /// The static map connecting SCET scale variations with resummation ones (SCET variations include xs fix)

      IPDF* _pdf;             ///< pointer to a pdf
      IPDF* _positivePdf;

      ITreeME* _treeME;
      BeamFuncTauB* _beamA;         ///< beam function at fixed order for beam A
      BeamFuncTauB* _beamB;         ///< beam function at fixed order for beam B
      BeamFuncTauB* _beamAsing;      ///< beam function at fixed order for beam A, singular terms
      BeamFuncTauB* _beamBsing;      ///< beam function at fixed order for beam B, singular terms
      SignatureSet _sigs;       ///< Holds the signatures we work with

      CalcNLOppV* _calc0jet;       ///< V + 0-jet NLO calculation to use

      CalcNLOppVj* _calc1jet;      ///< V + 1-jet NLO calculation to use with _tau0cut

      UnitCubeRadGenerator* _xRadGen; ///< radiation variables generator

      std::map<Scale, double> _spectrumTau0Resummed;
      std::map<Scale, double> _spectrumTau0Expanded;
      std::map<Scale, double> _cumDerivTau0Resummed;
      std::map<Scale, double> _cumDerivTau0Expanded;

      bool _flavSummedBeams;  ///< whether or not the beam functions should be flavour summed when evaluating resummed/expanded in the spectrum
      bool _totalXSfix;      ///< whether or not we should use a resummation scheme that fixes the total XS
      bool _improvedXSfix;   ///< whether or not we should use an improved resummation scheme that fixes the total XS

      std::map<Scale, double> _profSmootherFac;   ///< the value of the constant rescaling in the profile smoother

      SplittingFunction* _splitting0To1;
      SplittingFunction* _splitting1To2;

      bool _includeResummation;
      bool _includeResumExpanded;
      bool _includeFixedOrder;

      bool _includeTau1Resummed;
      bool _includeTau1NonSingFixedOrder;
      bool _includeTau1NonSingFromTau1;
      bool _includeISRSplitting;
      bool _includeFSRSplitting;

      double _tau1Max;
      double _tau0OverTau1Max;
      bool _useU1ExpOndSigmaCdPhi1;       ///< flag specifying if the expansion of the Sudakov multiplies the whoel dSig
      bool _includeU1Kinematics;          ///< (temporary) flag to maintain backward compatibility in Tau1 resummation

      //Some debug level paramaters
      std::string _nonsingularPhi2Treatment;

      friend class CalcSCETppV012Test;

   public:
      CalcSCETppV012(const OptionMap& optMap, double Tau0cut, double Tau1cut, double Ecm,
                     RunOrder run0, RunOrder run1, FixedOrder fixed0jet, FixedOrder fixed1jet,
                     Scale scale, YProfiles profiles, Phi1 phi1, ICoupling* alphaem, CouplingStrong* alphas,
                     IPDF* pdf, ITreeME* tree, ITreeMEComplete* treeComplete, IOneLoopME* oneLoop,
                     UnitCubeRadGenerator* _xRadGen, int nLightFlavors, PdgID bosonId, PdgID decayChannel,
                     bool flavAverageBeam, bool totalXSfix, bool improvedXSfix, double mVLow, double mVHigh);

      virtual ~CalcSCETppV012();

      virtual Sigma fastEstimate(const Configuration& config);

      /// What signatures this calculation requests from a Generator
      virtual const SignatureSet& neededSignatures() const;

      /// The Restriction separating the various regions of phase space
      virtual IRestriction* getRestriction() const;

      virtual size_t getNumVariations() const;

      /// Returns the name of the Calculation.
      virtual const std::string name() const;

      /// Returns information about the components.
      virtual const std::vector<NameValuePair> componentInfo() const;

      virtual void informGenerationWeight(const double w);

      virtual void initialize();
      virtual void finalize(const size_t numEvents);

      /// Returns the value of the 0-jet MC cross section
      double sigma0(const Configuration& config, double Tau0cut, const NloCoefficients& sigma0Coeffs, Scale scale);
      /// Returns the value of the 1-jet MC cross section
      double sigma1(const Configuration& config, double Tau0, double Tau1cut, const NloCoefficients& sigma1Coeffs, Scale scale);
      /// Returns the value of the 2-jet MC cross section
      double sigma2(const Configuration& config, double Tau0, double Tau1, const NloCoefficients& sigma2Coeffs, Scale scale);

      /// Returns the value of the 2-jet MC cross section that comes from the splitting of the 1-jet cross-section
      double sigma2Split(double Tau1, Scale scale);

      double Tau0cut() {return _tau0cut;};

      // Provides the default filename for the improved xsec fix setup file.
      static std::string getSetupFilename(const OptionMap& options, bool addSeed = false);

   private:
      /// Stores the results of the 1->0 projection.
      struct
      {
         bool exists;
         Configuration config;
         double splittingProb;
      } _phi0Proj;

      /// Stores the results of the 1->2 projection.
      struct
      {
         bool exists;
         Configuration config;
         NloCoefficients coeffs;
         double splittingProb;
         double tau0;
      } _phi1Proj;

      //------------------------------------------------------------------------------
      // Helper class to set the nonsingular scale in the FO calculations
      class NonSingularScaleFixed : public IScale
      {
         private:
            double _value;

         public:
            NonSingularScaleFixed() : _value(0.) {}

            double value() const { return _value; }
            void setValue(double value) { _value = value; };

            virtual std::string name() const { return "CalcSCETppV012::NonSingularScaleFixed"; };

         private:
            virtual double value(const Configuration& config) const { return _value; };

      };

      NonSingularScaleFixed _scaleNS;
      IScale* _scaleFO;

      //------------------------------------------------------------------------------
      /**
      * \class CalcSCETppV012::Restriction
      * \ingroup res
      *
      * \brief Restriction defined by CalcSCETppV012
      *
      * \todo: detailed description goes here
      */
      //------------------------------------------------------------------------------
      class Restriction : public IRestriction
      {
      private:
         IRestriction* _res01jet;   ///< Restriction of the 0-jet NLO calculation used
         IRestriction* _res12jet;   ///< Restriction of the 1-jet NLO calculation used

      public:
         Restriction(IRestriction* res01jet, IRestriction* res12jet )
            : _res01jet(res01jet ), _res12jet(res12jet ) {}

         ~Restriction()
         {
            delete _res01jet;
            delete _res12jet;
         }

         bool contains(const Configuration& config) override;
         size_t countJets(const Configuration& config ) const override;

         IRestriction* clone() const override
         {
            return new Restriction(_res01jet->clone(), _res12jet->clone());
         }
      };

      ///Returns the result of the calculations for all scales
      virtual Sigma value(const Configuration& config)
      {
         return value(config, true);
      }

      virtual Sigma value(const Configuration& config, bool doVariations);

      ///Returns the result of the calculations for a given scale.
      //Note that Tau0 and Tau1 are only passed in for convenience
      double value(const Configuration& config, double Tau0, double Tau1,
                   const NloCoefficients& sigma0Coeffs, const NloCoefficients& sigma1Coeffs,
                   const NloCoefficients& sigma2Coeffs, Scale scale);

      //-----------------------------------------------------------------
      ///Fixed order calculations
      //-----------------------------------------------------------------
      ///Calculates and stores the fixed order coefficients for a 0-jet caluclation
      NloCoefficients calculateFO0Coeffs(const Configuration& config, bool doVariations);
      ///Calculates and stores the fixed order coefficients for a 1-jet caluclation
      NloCoefficients calculateFO1Coeffs(const Configuration& config, double Tau0, bool doVariations);
      ///Calculates and stores the fixed order coefficients for a 2-jet caluclation
      NloCoefficients calculateFO2Coeffs(const Configuration& config, double Tau0, bool doVariations);


      //-----------------------------------------------------------------
      ///N-Jettiness cumulants and spectra
      //-----------------------------------------------------------------

      ///Computes the resummed and expanded cumulant dsigma / dPhi0 (Tau0cut)
      std::pair<double, double> tau0Cumulant(const Configuration& config, double Tau0cut, Scale scale);
      ///Computes the resummed and expanded spectrum dsigma / dPhi0 dTau0
      std::pair<double, double> tau0Spectrum(const Configuration& config, double Tau0, Scale scale);

      //-----------------------------------------------------------------
      ///Pieces of the final formula
      //-----------------------------------------------------------------
      ///Computes the dSigma^C_{\ge q} / dPhi_1 from GIN 18
      double dSigmaCdPhi1(const Configuration& Phi1, double Tau0, const NloCoefficients& nloCoeffs, Scale scale);

      //-----------------------------------------------------------------
      ///Helper Functions
      //-----------------------------------------------------------------
      struct Phi0
      {
         double xA;
         PdgID idA;
         double xB;
         PdgID idB;
         double Q;
         double Y;

         Phi0(double xAin, PdgID idAin, double xBin, PdgID idBin, double Qin)
            : xA(xAin), idA(idAin), xB(xBin), idB(idBin), Q(Qin), Y(0.5 * log(xAin/xBin))
         {};
      };

      ///Calculates the resummed cumulant
      double sigma0Resummed(const Phi0& phi0, double Tau0cut, Scale scale, double muFO);
      ///Calculates the expansion of the resummed cumulant
      double sigma0ResumExpanded(const Phi0& phi0, double Tau0cut, Scale scale, double muFO);
      ///Calculates the resummed thrust spectrum
      double dSigma0Resummed(const Phi0& phi0, double TauB0, Scale scale, double muFO);
      ///Calculates the expansion of the resummed thrust spectrum
      double dSigma0ResumExpanded(const Phi0& phi0, double TauB0, Scale scale, double muFO);
      ///Calculates the difference between the resummed spectrum and the derivative of the cumulant
      double deltaSigmaR(const Phi0& phi0, double TauB0, Scale scale, double muFO);

      double getmuFO(const Configuration& config, double Tau0) const;
      bool getScales(Scale scale, double Tau, double Q, double Y, double muFO, double& muH, double& muB, double& muS);
      double murun(double x, double a, double y0, double x1, double x2, double x3 ) const;
      double getmuNS(Scale scale, double Tau0, double muFOCentral) const;

      double U0LL(double asH, double lnxH, double asBa, double lnxBa, double asBb,
               double lnxBb, double asS, double lnxS,  double as, int nf, double& eta);
      double U0NLL(double asH, double lnxH, double asBa, double lnxBa, double asBb,
                  double lnxBb, double asS, double lnxS,  double as, int nf, double& eta);
      double U0NNLL(double asH, double lnxH, double asBa, double lnxBa, double asBb,
                  double lnxBb, double asS, double lnxS,  double as, int nf, double& eta);

      //------------------------------------------------------------------------------
      // Helper class to hide the gsl machinery for the derivative of the resummed cumulant from the user
      friend class Sigma0ResummedDerivative;
      class Sigma0ResummedDerivative
      {
         private:
            typedef double(CalcSCETppV012::*F)(const Phi0&, double, Scale, double);

            const Phi0& _phi0;
            Scale _scale;
            double _muFO;
            CalcSCETppV012* const _obj;
            F const _func;
            gsl_function _gsl_func;

         public:
            Sigma0ResummedDerivative(CalcSCETppV012* obj, F func, const Phi0& phi0, Scale scale, double muFO)
               : _phi0(phi0), _scale(scale), _muFO(muFO), _obj(obj), _func(func)
            {
               _gsl_func.function = &static_call;
               _gsl_func.params = static_cast<void*>(this);
            }

            double operator()(double Tau0, double eps, double& abserr)
            {
               double result;
               if (Tau0 > eps)
                  gsl_deriv_central(&_gsl_func, Tau0, eps, &result, &abserr);
               else
                  gsl_deriv_forward(&_gsl_func, Tau0, eps, &result, &abserr);
               return result;
            }

         private:
            static double static_call(double x, void* p)
            {
               return static_cast<Sigma0ResummedDerivative*>(p)->call(x);
            }

            double call(double x)
            {
               return (_obj->*_func)(_phi0, x, _scale, _muFO);
            }
      };

      //------------------------------------------------------------------------------
      // Helper class to hide the gsl machinery for the derivative of the tau1 sudakov from the user
      friend class Tau1SudakovDerivative;
      class Tau1SudakovDerivative
      {
         private:
            const CalcSCETppV012* _obj;
            const Configuration& _phi1;
            double _tau1max;

         public:
            Tau1SudakovDerivative(const CalcSCETppV012* obj, const Configuration& phi1, double tau1Max)
               : _obj(obj), _phi1(phi1), _tau1max(tau1Max)
            {
               _gsl_func.function = &call;
               _gsl_func.params = this;
            }

            double operator()(double Tau1, double eps, double& abserr )
            {
               double result;
               if (Tau1 > eps) {
                  gsl_deriv_central(&_gsl_func, Tau1, eps, &result, &abserr);
               } else {
                  gsl_deriv_forward(&_gsl_func, Tau1, eps, &result, &abserr);
               }
               return result;
            }

         private:
            gsl_function _gsl_func;

            static double call(double x, void* p)
            {
               Tau1SudakovDerivative& me = *static_cast<Tau1SudakovDerivative*>(p);
               return me._obj->tau1Sudakov(me._phi1, x, me._tau1max);
            }
      };

      //------------------------------------------------------------------------------
      /**
      * \class CalcSCETppV012::ImprovedXSHandler
      * \ingroup res
      *  Helper class to manage the normalization and the
      *  reweighting to the total NNLO cross section
      * \todo: detailed description goes here
      */
      //------------------------------------------------------------------------------
      class ImprovedXSHandler
      {
         private:
            double _alphaCentral;
            double _alphaUp;
            double _alphaDown;
            SignatureSet _sigSet;

         public:
            ImprovedXSHandler(const std::string& filename);
            double alphaCentral() const {return _alphaCentral;}
            double alphaUp() const {return _alphaUp;}
            double alphaDown() const {return _alphaDown;}
      };

      ImprovedXSHandler* _improvedXSHandler;

      double tau1Sudakov(const Configuration& Phi1, double Tau1cut, double Tau1max) const;
      double tau1SudakovDerivative(const Configuration& Phi1, double Tau1, double Tau1max) const;

      double tau1SudakovExpanded(const Configuration& Phi1, double Tau1cut, double Tau1max, double muNS) const;
      double tau1SudakovExpandedDerivative(const Configuration& Phi1, double Tau1, double Tau1max, double muNS) const;

      double getTau1Measures(const Configuration& Phi1, double& sq, double &sg, double& wq2, double& wg2) const;

      double F0LO(BeamFuncTauB* beamfuncA, BeamFuncTauB* beamfuncB, double asH, double lnxH, double muBa, double muBb, double xA, double xB, PdgID idA, PdgID idB);
      double F0NLO(BeamFuncTauB* beamfuncA, BeamFuncTauB* beamfuncB, double asH, double lnxH, double asBa, double lnxBa, double muBa, double asBb, double lnxBb, double muBb, double asS, double lnxS, double eta, double xA, double xB, double Q, PdgID idA, PdgID idB);
      double F0NNLO(BeamFuncTauB* beamfuncA, BeamFuncTauB* beamfuncB, double asH, double lnxH, double asBa, double lnxBa, double muBa, double asBb, double lnxBb, double muBb, double asS, double lnxS, double eta, double xA, double xB, double Q, PdgID idA, PdgID idB);
      double dF0NLO(BeamFuncTauB* beamfuncA, BeamFuncTauB* beamfuncB, double asH, double lnxH, double asBa, double lnxBa, double muBa, double asBb, double lnxBb, double muBb, double asS, double lnxS, double eta, double xA, double xB, double Q, PdgID idA, PdgID idB);
      double dF0NNLO(BeamFuncTauB* beamfuncA, BeamFuncTauB* beamfuncB, double asH, double lnxH, double asBa, double lnxBa, double muBa, double asBb, double lnxBb, double muBb, double asS, double lnxS, double eta, double xA, double xB, double Q, PdgID idA, PdgID idB);

      std::pair<double, double> sigma0NLO(const Configuration& config, const NloCoefficients& sigma0Coeffs, Scale scale);

      double dsigma0NLO(const Configuration& config, double as, double x);

      double lepfac(PdgID id, double Q);
      double hard2(double asH, double lnXH, int order);
      double soft2(int dist, int order);

      void setBeamFuncXs(const Configuration& config, double& xBeamFuncA, double& xBeamFuncB);
      void setBeamFuncId(const Configuration& config, std::vector<PdgID>& idBeamFuncA, std::vector<PdgID>& idBeamFuncB);

      //The fully differential partonic 0-jet rate
      double fullyDifferential0JetNoPDF(const Configuration&  Phi0);

      //This determines the correct scale value given a Sigma with fixed-order scale variations and a SCET scale
      double extractFixedOrderScale(const Sigma& sigma, Scale scale);

      //The static map initializer
      static std::map<Scale, EngineNLOBase::ScaleVariation> initFixedOrderScaleMap();
      static std::map<Scale, Scale> initResummationScaleMap();
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline const SignatureSet& CalcSCETppV012::neededSignatures() const
{
   return _sigs;
}

//---------------------------------------------------------------------------
inline IRestriction* CalcSCETppV012::getRestriction() const
{
   return new Restriction(_calc0jet->getRestriction(), _calc1jet->getRestriction());
}

//------------------------------------------------------------------------------
/**
 * \brief Provides the default filename for the improved xsec fix setup file.
 * \param options The options to use.
 * \param addSeed Whether to append the seed as tag (defaults to `false`).
 * \return The full filename including the path and extension.
 *
 * If the CalcSCETppV012::ImprovedXSFix::fileStem option is nonempty it is used
 * for the main filestem. Otherwise it is constructed from the \a options as
 * 'fileStem_setupTag_CalcSCETppV012::ImprovedXSFix::fileTag_profile'.
 *
 * The full filename is then constructed as
 * 'filePath::setup/mainfilestem_seed.dat',
 * where the seed is only added if \a addSeed is `true`.
 */
inline std::string CalcSCETppV012::getSetupFilename(const OptionMap& options, bool addSeed)
{
   std::string stem = options.getOption("CalcSCETppV012::ImprovedXSFix::fileStem");
   if (stem.empty())
      stem = options.getOption("fileStem")
             + format_tag(options.getOption("setupTag"))
             + format_tag(options.getOption("CalcSCETppV012::ImprovedXSFix::fileTag"))
             + format_tag("profile");
   if (addSeed)
      stem += format_tag(options.getOption("seed"));
   return format_path(options.getOption("filePath::setup")) + stem + ".dat";
}

} // namespace Geneva

#endif // GENEVA_CALC_SCET_PPV012_HPP
