//------------------------------------------------------------------------------
/// \file EngineNLOBase.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of EngineNLOBase class
//------------------------------------------------------------------------------

#ifndef GENEVA_ENGINE_NLO_BASE_HPP
#define GENEVA_ENGINE_NLO_BASE_HPP

#include "Geneva/Calculations/IPhiNLO.hpp"
#include "Geneva/Calculations/ICalculation.hpp"

#include "Geneva/Components/ICommonCouplings.hpp"
#include "Geneva/Components/RegionInfoList.hpp"
#include "Geneva/Components/ITreeMEComplete.hpp"
#include "Geneva/Components/ITreeME.hpp"
#include "Geneva/Components/IOneLoopME.hpp"
#include "Geneva/Components/CouplingStrong.hpp"
#include "Geneva/Components/NloCoefficients.hpp"

#include "Geneva/Core/Physics/Configuration.hpp"
#include "Geneva/Core/Physics/SignatureSet.hpp"
#include "Geneva/Core/Physics/IRestriction.hpp"

#include <vector>
#include <string>

namespace Geneva
{

class UnitCubeRadGenerator;
class UnitCubeRad;
class PhiRadFKS;
class IAnalyzer;
class OptionMap;
class GenerationReport;

//------------------------------------------------------------------------------
/**
 * \class EngineNLOBase
 * \ingroup calc
 *
 * \brief   Helper class for a generic NLO calculation with FKS subtraction for generic collisions
 *
 * EngineNLOBase represents an base class for a NLO calculation for a generic process.
 * As such, it can return both the  LO contribution only, or the sum of LO+NLO. In this last case,
 * for a given phase space point PhiN, it sums over all emitting partons and for each parton
 * performs the the integration over \f$\Phi_{\rm rad}\f$ using the FKS subtractions.
 */
//------------------------------------------------------------------------------
class EngineNLOBase
{
   public:
      /// Named constants to specify scale choice
      enum ScaleVariation
      {
         kNone,
         kCorrelated,            ///< Run simultaneous up/down muR abd muF scale variations (3 combinations)
         kUncorrelated,          ///< Run uncorrelated up/down muR and muF scale variation  (9 combinations)
         kUncorrelatedNoExtremes ///< Run uncorrelated up/down muR and muF scale variation, excluding extremes (7 combinations)
      };

   protected:
      /// Holds the name of the NLO calculation
      std::string _calcname;

      /// Holds the number of points we will do the phiRad integration over
      size_t _numPointsReal;

      ///Holds the generation weight of the last Configuration
      double _lastGenWeight;

   private:
      /// Holds the signatures with N particles
      SignatureSet _signaturesBN;
      /// Holds the signatures with N+1 particles
      SignatureSet _signaturesBNp1;

   protected:
      /// order of the calculation (e.g. LO vs. NLO)
      FixedOrder _order;
      /// components that calculate tree-level
      ITreeME* _treeME;
      /// components that calculate tree-level with spin-correlated and color-correlated terms
      ITreeMEComplete* _treeMEComplete;
      /// components that calculate one-loop ME
      IOneLoopME* _oneLoopME;
      /// Generator for the phiRad splitting phase space for R-S integral
      UnitCubeRadGenerator* _xRadGenerator;
      /// component that evaluate the common couplings
      ICommonCouplings* _commonCouplings;
      /// strong coupling
      CouplingStrong* _alphas;

      /// Holds all the info concerning the singular regions
      RegionInfoList _regionsInfo;

      const std::vector<NameValuePair> componentInfo() const;

      //------------------------------------------------------------------------------
      /**
       * \class EngineNLOBase::Restriction
       * \ingroup res
       *
       * \brief Restriction
       *
       * \todo: detailed description goes here
       */
      //------------------------------------------------------------------------------
      class Restriction : public IRestriction
      {
         private:
            IPhiNLO* _phiNLO;

         public:
            Restriction(IPhiNLO* phiNLO) : _phiNLO(phiNLO) {}

            bool contains(const Configuration& config) override;
            size_t countJets(const Configuration& config) const override;

            IRestriction* clone() const override
            {
               return new Restriction(_phiNLO);
            }
      };

      Restriction* _restriction;

      /// Holds the fixed order Analyzer
      IAnalyzer* _analyzer;
      bool _doAnalyze;

  private:
      std::string _analyzerFile;
      GenerationReport* _report;

   public:
      EngineNLOBase(const OptionMap& optMap, const std::string& name, const SignatureSet& sigSet,
                    ITreeME* treeME, ITreeMEComplete* treeMEComplete, IOneLoopME* oneLoopME,
                    UnitCubeRadGenerator* xRadGenerator, ICommonCouplings* coupl,
                    IPhiNLO* phiNLO, CouplingStrong* alphas, FixedOrder order, int nLight);

      virtual ~EngineNLOBase();

      /// Calculate either LO or LO+NLO, depending on the request
      Sigma calculate(const Configuration& config, bool doVariations) const;

      /// Create a new copy of the Restriction. The caller owns it and is responsible for its deletion
      IRestriction* getRestriction() const;

      /// Get the LO coefficient.
      virtual Sigma getLOCoeff(const Configuration& config, bool doVariations) const = 0;

      /// Get the LO and NLO coefficients.
      virtual NloCoefficients getFOCoeffs(const Configuration& config, bool doVariations) const = 0;

      /// Get the LO and NLO coefficients with Born-like kinematics.
      virtual NloCoefficients getFOCoeffsBornKin(const Configuration& config, bool doVariations) const = 0;

      void setGenerationWeight(const double w) { _lastGenWeight = w; };

      void finalize(const size_t num);

      void resetAnalyzer() const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Declarations
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline IRestriction* EngineNLOBase::getRestriction() const
{
   return _restriction->clone();
}

} // namespace Geneva

#endif // GENEVA_ENGINE_NLO_BASE_HPP
