//------------------------------------------------------------------------------
/// \file EngineNLOpp.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class EngineNLOpp
//------------------------------------------------------------------------------

#ifndef GENEVA_ENGINE_NLO_PP_HPP
#define GENEVA_ENGINE_NLO_PP_HPP

#include "Geneva/Calculations//EngineNLOBase.hpp"
#include "Geneva/Components/IPDF.hpp"

#include <list>

namespace Geneva
{

class IScale;
class IPhiNLOwithPDF;
class SubtractionFKSpp;

//------------------------------------------------------------------------------
/**
 * \class EngineNLOpp
 * \ingroup calc
 *
 * \brief   Helper class for a generic NLO calculation with FKS subtraction for hadronic collisions.
 *
 * EngineNLOpp represents an NLO calculation for a generic process in hadronic collisions.
 * As such, it can return both the  LO contribution only, or the sum of LO+NLO. In this last case,
 * for a given phase space point PhiN, it sums over all emitting partons and for each parton
 * performs the the integration over \f$\Phi_{\rm rad}\f$ using the FKS subtractions.
 */
//------------------------------------------------------------------------------
class EngineNLOpp : public EngineNLOBase
{
   private:
      IScale* _renScale;                    ///< renormalization scale
      IScale* _factScale;                   ///< factorization scale

      ScaleVariation _scaleVariation;       ///< type of scale variation
      size_t _numScaleVariations;           ///< number of scale variations
      std::vector<double> _renScaleVaryCoeffs;
      std::vector<double> _factScaleVaryCoeffs;

      IPhiNLOwithPDF* _phiNLO;
      SubtractionFKSpp* _subtraction;

      IPDF* _pdf;
      double _cmEnergy;

      /// Number of points used in the collinear remnant integration
      size_t _numPointsCollRemn;
      bool _impSampleCollremn;

      /// Switches to turn off various contributions
      bool _includeBorn;
      bool _includeSoftVirt;
      bool _includeCollRemn;
      bool _includeReal;

   public:
      EngineNLOpp(const OptionMap& optMap, const std::string name, const SignatureSet& sigSet,
                  ITreeME* treeME, ITreeMEComplete* treeMEComplete, IOneLoopME* oneLoopME,
                  UnitCubeRadGenerator* phiRadFKSGenerator, IPDF* pdf, IScale* renScale, IScale* factScale,
                  ICommonCouplings* coupl, IPhiNLOwithPDF* phiNLO, CouplingStrong* alphas,
                  FixedOrder order, double Ecm, int nLight);

      virtual ~EngineNLOpp();

      const std::vector<NameValuePair> componentInfo() const;

      /// Calculate and returns the LO coefficient in GeV.
      virtual Sigma getLOCoeff(const Configuration& config, bool doVariations) const;

      /// Calculate and returns the LO and NLO coefficients in GeV.
      virtual NloCoefficients getFOCoeffs(const Configuration& config, bool doVariations) const;

      /// Calculate and returns the LO and the (subtracted) NLO coefficients with Born-like kinematics
      virtual NloCoefficients getFOCoeffsBornKin(const Configuration& config, bool doVariations) const;

   private:
      friend class CalcNLOppVTest;
      friend class CalcNLOppVjTest;
      friend class CalcNLOppVjjTest;

      std::vector<double> getScaleVariations(const Configuration& config, const PdgID idA, const double xA, const PdgID idB, double xB, const double couplStrongCentral, const int nPowers, const double luminosityCentral, const double mufc, const double murc, const double Q, const double w0, const double wf, const double wr) const;

      /// helper function that evaluates the tree-level contributions with scale variations
      Sigma calcTreeLevel(const Configuration& config, int asExp, bool doVariations) const;

      /// helper function that evaluates the NLO contributions with scale variations, allowing to switch off the single contributions
      NloCoefficients calcNLO(const Configuration& config, bool includeBorn, bool includeVirtSub, bool includeCollRemn, bool includeRealsSub, bool doVariations) const;

      /// Integrates the R-S contribution
      Sigma getIntegralRealSubtracted(const Configuration& config_CM, const double xBarA, const double xBarB, const double fluxBorn, const double mufBorn, const double murBorn, const double lumiBorn, const double alphasBorn, const int asExpBorn, const double couplStrongBorn, const double couplNonStrongBorn, bool doVariations) const;

      /// Generate the radiative phase space in the unitary cube
      void generateUnitCubeRad(const std::list<RegionInfo>::iterator map, UnitCubeRad& xRad) const;

         /// Integrates the CR contribution
      Sigma getIntegralCollinearRemnants(const Configuration& phiN, const double xBarA, const double xBarB, const double fluxBorn, const double mufBorn, const double murBorn, const double lumiBorn, const double alphasBorn, const int asExpBorn, const double couplStrongBorn, bool doVariations) const;

      ///The collinear remnants coefficients.The first elements in the vector is the coefficient of the non-logarithmic part, while the second element is the coefficient of the logarithm of the renormalization scale. The third element, if present, is the coeffcient of the logarithm of the factorization scale
      Sigma calculateCollinearRemnant(const Configuration& config, const Configuration::iterator initial, const double xBar, const double xBarOpposite, const double sBorn, const double mufc, const double murc, const double jac, const double remn, const double remnMuF, const double z, const double x, const double pdfcBorn, const double pdfcBornOpposite, const double alphasBorn, const int asExpBorn, const double couplStrongBorn, bool doVariations) const;
};

} // namespace Geneva

#endif // GENEVA_ENGINE_NLO_PP_HPP
