//------------------------------------------------------------------------------
/// \file ICalculation.hpp
//
// Author(s):
//    Chris Vermilion, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2013 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class ICalculation
//------------------------------------------------------------------------------

#ifndef GENEVA_I_CALCULATION_HPP
#define GENEVA_I_CALCULATION_HPP

#include "Geneva/Core/Physics/Sigma.hpp"

#include <cstdlib>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \defgroup calc Calculations
 *
 * Calculations are one of the basic building blocks of Geneva. A Calculation
 * computes the cross section for a given point in phase space. Each of the
 * classes above is a specific implementation. Click on a class name for the
 * documentation of that calculation.
 */
//------------------------------------------------------------------------------

typedef std::pair<std::string, std::string> NameValuePair;
typedef std::vector<NameValuePair> ComponentInfo;

class Configuration;
class SignatureSet;
class IRestriction;

//------------------------------------------------------------------------------
/**
 * \class ICalculation
 * \ingroup calc
 *
 * \brief Abstract interface to calculations.
 *
 * A calculation acts on a Configuration, which represents a point in phase
 * space, and returns a Sigma, which holds the cross section for that phase-
 * space point. Calculations are defined for a specific SignatureSet and a
 * specific region of phase space, where the latter can be obtained in form of
 * a IRestriction. Calculations are typically constructed from several more
 * basic components (matrix elements, couplings, etc.).
 */
//------------------------------------------------------------------------------
class ICalculation
{
   public:
      ICalculation() = default;
      ICalculation(const ICalculation&) = delete;
      ICalculation(ICalculation&&) = delete;
      ICalculation& operator=(const ICalculation&) = delete;
      ICalculation& operator=(ICalculation&&) = delete;
      virtual ~ICalculation() = default;

      /**
       * \brief Applies the calculation to a Configuration.
       * \param config The Configuration to act on.
       * \param doVariations Whether to perform variations.
       * \return The cross section in units of pb in form of a Sigma.
       *
       * The returned Sigma is set to be defined by the calculation if and only
       * if the computation was successful. In particular if \a config fails the
       * calculation's restriction or has a Signature not part of
       * neededSignatures() it should at the minimum return an undefined Sigma.
       * Calculations may also choose to report failures in other ways (e.g. by
       * printing an error message or throwing an exception for an unknown
       * Signature).
       *
       * If \a doVariations is set to false, the calculation may skip possible
       * variations and only perform the central value calculation. The result
       * for the central value itself must not be affected by this flag.
       */
      Sigma operator()(const Configuration& config, bool doVariations = true);

      /**
       * \brief Calculates a fast estimate of the cross section
       * \param config The Configuration to act on.
       * \return An estimate of the cross sectin in units of pb
       *
       * The returned cross section should provide a (typically lowest-order)
       * estimate of the cross section.
       */
      virtual Sigma fastEstimate(const Configuration& config);

      /**
       * \brief Returns the signatures the calculation needs.
       * \return A SignatureSet holding all needed signatures.
       *
       * The returned signatures should be precisely those which the calculation
       * needs in order to function properly and produce consistent results.
       */
      virtual const SignatureSet& neededSignatures() const = 0;

      /**
       * \brief Returns the calculation's phase-space restriction.
       * \return An IRestriction pointer.
       *
       * \note: The caller owns the returned pointer and is responsible for
       * deleting it.
       *
       * \todo: implement this in derived classes
       */
      virtual IRestriction* getRestriction() const { return 0; }

      /// Returns the number of variations that the calculation will return when called (in addition to the central value)
      virtual size_t getNumVariations() const { return 0; }

      /// Returns the name of the Calculation.
      virtual const std::string name() const = 0;

      /// Returns information about the components.
      virtual const ComponentInfo componentInfo() const = 0;

      /// Inform the calculation of the weight used to generate the last Configuration.
      virtual void informGenerationWeight(const double w) {};

      /// Initialize the calculation. Users can overload this function for specific purposes.
      virtual void initialize() {};

      /// Finalize the calculation. Users can overload this function for specific purposes.
      virtual void finalize(const size_t numEvents) {};

   private:
      /**
       * \brief Calculates the cross section for the given \p config.
       * \return The cross section in units of pb in form of a Sigma.
       *
       * This is used by operator() and must be implemented by specific
       * calculations to return the value of the cross section possibly along
       * with a set of variations.
       */
      virtual Sigma value(const Configuration& config) = 0;
      virtual Sigma value(const Configuration& config, bool doVariations);
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline Sigma ICalculation::operator()(const Configuration& config, bool doVariations)
{
   Sigma sigma = value(config, doVariations);
   return sigma;
}

//------------------------------------------------------------------------------
// default implementation in terms of value() for classes which don't explicitly
// implement this signature
inline Sigma ICalculation::value(const Configuration& config, bool doVariations)
{
   Sigma sigma = value(config);
   if (!doVariations)
      sigma.variations().clear();
   return sigma;
}

//------------------------------------------------------------------------------
// default implementation in terms of value() for classes which don't explicitly
// implement this
inline Sigma ICalculation::fastEstimate(const Configuration& config)
{
   return value(config, false);
}

} // namespace Geneva

#endif // GENEVA_I_CALCULATION_HPP
