//------------------------------------------------------------------------------
/// \file IPhiNLO.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface for PhiNLO classes
//------------------------------------------------------------------------------

#ifndef GENEVA_IPHI_NLO_HPP
#define GENEVA_IPHI_NLO_HPP

#include "Geneva/Core/Physics/Configuration.hpp"

#include <string>

namespace Geneva
{

class PhiRadFKS;

//------------------------------------------------------------------------------
/**
 * \class IPhiNLO
 * \ingroup calc
 *
 * \brief Interface to access the phase space definition PhiN at NLO for a given Calculation
 *
 * A PhiNLO is basically a set of restrictions to be applied on a given Calculation
 * to define its NLO phase space.
 */
//------------------------------------------------------------------------------
class IPhiNLO
{
   public:
      /// Checks if a Configuration is N-body and part of Phi_N.
      virtual bool isPhiN(const Configuration& phiN) const = 0;

      /// Checks if a Configuration is (N+1)-body and part of Phi_{N+1}.
      virtual bool isPhiNplus1(const Configuration& phiNp1) const = 0;

      virtual size_t countN(const Configuration& config) const = 0;

      virtual const std::string name() const = 0;

      // TODO: This should not be here but somehow be handled by constructSoftLimit directly
      virtual bool ThetaijRijSoft(const Configuration& configN, const Configuration::iterator emitter, const FourVector& pSoft) const = 0;

      // TODO: Remove these since IR safety requires them to always be true?
      bool ThetaijRijColl() const     { return true; }
      bool ThetaijRijSoftColl() const { return true; }

      virtual std::pair<bool, bool> cutoffFKS(const PhiRadFKS& phiRad) const = 0;

   protected:
      IPhiNLO() = default;
      IPhiNLO(const IPhiNLO&) = default;
      IPhiNLO(IPhiNLO&&) = default;
      IPhiNLO& operator=(const IPhiNLO&) = default;
      IPhiNLO& operator=(IPhiNLO&&) = default;
      ~IPhiNLO() = default;
};

} // namespace Geneva

#endif // GENEVA_IPHI_NLO_HPP
