//------------------------------------------------------------------------------
/// \file IPhiNLOwithPDF.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface for PhiNLO classes
//------------------------------------------------------------------------------

#ifndef GENEVA_IPHI_NLO_PDF_HPP
#define GENEVA_IPHI_NLO_PDF_HPP

#include "Geneva/Calculations/IPhiNLO.hpp"

namespace Geneva
{

class UnitCubeRad;
class FourVector;
class PhiRadFKS;

//------------------------------------------------------------------------------
/**
 * \class IPhiNLOwithPDF
 * \ingroup calc
 *
 * \brief Interface to access the phase space definition PhiN at NLO for a given Calculation that includes initial-state partons
 *
 * A PhiNLO is basically a set of restrictions to be applied on a given Calculation
 * to define its NLO phase space.
 */
//------------------------------------------------------------------------------
class IPhiNLOwithPDF : public virtual IPhiNLO
{
   public:
      /// Constructs the split Configuration. Return true if the Configuration is valid and passes all the restrictions. Several other objects are passed in by reference and need to be set inside:
      ///  - configNplus1, i.e. the split Configuration in partonic CM frame
      ///  - iterator to emitted parton belonging to configNplus1
      ///  - iterator to sister beloging to configNplus1
      ///  - xA, xB, i.e. Bjorken's x values needed to boost configNplus1 to the Lab frame.
      ///  - the jacobian of the splitting
      virtual bool constructPhiNplus1(const Configuration& configN, const Configuration::iterator emitter, double xBarA, double xBarB, const UnitCubeRad& xRad, Configuration& configNplus1, Configuration::iterator& emitted, Configuration::iterator& sister, double& xA, double& xB, PhiRadFKS& phiRad, double& jacobian) = 0;

      /// Constructs the split Configuration in the soft limit of the mapping. Sets the soft emitted momentum in the partonic CM frame by reference. Returns true when the construction works, false otherwise
      virtual bool constructSoftLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadSoft, double& jacobian) const = 0;

      virtual bool constructCollinearLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadColl, double& jacobian) const = 0;

      virtual bool constructSoftCollinearLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadSoftColl, double& jacobian) const = 0;

      virtual std::pair<double, double> xiYCollinearLimits(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const = 0;

      virtual std::pair<double, double> xiYSoftCollinearLimits(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const = 0;

      /// Returns the maximum energy that the emitted parton can achieve trough the soft limit of the
      /// mapping in the partonic CM frame
      /// NOTE: this is not necessarily identical to the physical maximum energy that a soft emission can reach, but it corresponds to the soft limit of the bound used in the mapping
      virtual double maxSoftEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const = 0;

      /// Returns the maximum energy that the emitted parton can achieve trough the collinear limit of the
      /// mapping in the partonic CM frame. Also return the maximum Mandelstam S for the maximum energy configuration.
      /// NOTE: this is not necessarily identical to the physical maximum energy that a collinear emission can reach, but it corresponds to the collinear limit of the bound used in the mapping
      virtual std::pair<double,double> maxCollEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const = 0;

      /// Returns the maximum energy that the emitted parton can achieve trough the soft-collinear limit of the
      /// mapping in the partonic CM frame
      /// NOTE: this is not necessarily identical to the physical maximum energy that a soft-collinear emission can reach, but it corresponds to the soft-collinear limit of the bound used in the mapping
      virtual double maxSoftCollEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const = 0;

   protected:
      IPhiNLOwithPDF() = default;
      IPhiNLOwithPDF(const IPhiNLOwithPDF&) = default;
      IPhiNLOwithPDF(IPhiNLOwithPDF&&) = default;
      IPhiNLOwithPDF& operator=(const IPhiNLOwithPDF&) = default;
      IPhiNLOwithPDF& operator=(IPhiNLOwithPDF&&) = default;
      ~IPhiNLOwithPDF() = default;
};

} // namespace Geneva

#endif // GENEVA_IPHI_NLO_PDF_HPP
