//------------------------------------------------------------------------------
/// \file PhiNLOppFKSBase.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interfaces of PhiNLOppFKSBase classes
//------------------------------------------------------------------------------

#ifndef GENEVA_PHI_NLO_PP_FKS_BASE_HPP
#define GENEVA_PHI_NLO_PP_FKS_BASE_HPP

#include "Geneva/Calculations/IPhiNLOwithPDF.hpp"

#include "Geneva/Components/FKSMapping.hpp"

namespace Geneva
{

class PhiRadFKS;

//------------------------------------------------------------------------------
/**
 * \class PhiNLOppFKSBase
 * \ingroup calc
 *
 * \brief Base class for NLO definitions of Phi in case FKS mapping is used for the splitting
 *
 * \todo document me!
 */
//------------------------------------------------------------------------------
class PhiNLOppFKSBase : public virtual IPhiNLOwithPDF
{
   protected:
      FKSMapping _mapFKS;
      double _cmEnergy;
      double _fksTinyCut;              /// The FKS cutoff to avoid getting to close to the divergence

   public:
      PhiNLOppFKSBase(double cmEnergy, double fksTinyCut = 0.)
         : _cmEnergy(cmEnergy), _fksTinyCut(fksTinyCut) {}

      virtual ~PhiNLOppFKSBase() = default;

      /// Constructs the split Configuration. Return true if Configuration is valid and passes all the restrictions.
      virtual bool constructPhiNplus1(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, Configuration& configNplus1CM, Configuration::iterator& emittedCM, Configuration::iterator& sisterCM, double& xA, double& xB, PhiRadFKS& phiRad, double& jacobian);

      /// Constructs the split Configuration in the soft limit of the mapping. Sets the soft emitted momentum in the partonic CM frame by reference. Returns true when the construction works, false otherwise
      virtual bool constructSoftLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadSoft, double& jacobian) const;

      /// Constructs the split Configuration in the collinear limit of the mapping. Sets the collinear emitted  and sister momenta in the partonic CM frame by reference. The Mandelstan S of the collinear split Configuration is passed in by reference and set internally. Returns true if the construction works, false otherwise
      virtual bool constructCollinearLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadColl, double& jacobian) const;

      /// Constructs the split Configuration in the soft-collinear limit of the mapping. Sets the soft-collinear emitted momentum in the partonic CM frame by reference. Returns true if the construction works, false otherwise
      virtual bool constructSoftCollinearLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadSoftColl, double& jacobian) const;

      /// Returns the maximum energy that the emitted parton can achieve trough the soft limit of the mapping in the partonic CM frame
      virtual double maxSoftEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const;

      /// Returns the maximum energy that the emitted parton can achieve trough the collinear limit of the mapping in the partonic CM frame. Also return the maximum Mandelstam S for the maximum energy configuration.
      virtual std::pair<double, double> maxCollEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const;

      /// Returns the maximum energy that the emitted parton can achieve trough the soft-collinear limit of the mapping in the partonic CM frame
      virtual double maxSoftCollEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const;

      virtual std::pair<double, double> xiYCollinearLimits(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const;

      virtual std::pair<double, double> xiYSoftCollinearLimits(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const;

      virtual std::pair<bool, bool> cutoffFKS(const PhiRadFKS& phiRad) const;

      virtual bool ThetaijRij(PhiNplus1 phiNp1CM) const = 0;

   protected:
      PhiNLOppFKSBase(const PhiNLOppFKSBase&) = default;
      PhiNLOppFKSBase(PhiNLOppFKSBase&&) = default;
      PhiNLOppFKSBase& operator=(const PhiNLOppFKSBase&) = default;
      PhiNLOppFKSBase& operator=(PhiNLOppFKSBase&&) = default;
};

} // namespace Geneva

#endif // GENEVA_PHI_NLO_PP_FKS_BASE_HPP
