//------------------------------------------------------------------------------
/// \file PhiNLOppVjBase.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interfaces of PhiNLOppVjBase classes
//------------------------------------------------------------------------------

#ifndef GENEVA_PHI_NLO_PPVJ_BASE_HPP
#define GENEVA_PHI_NLO_PPVJ_BASE_HPP

#include "Geneva/Calculations/IPhiNLOwithPDF.hpp"

#include "Geneva/Core/Physics/ConfigurationExtensions.hpp"
#include "Geneva/Core/Physics/SignatureSet.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class PhiNLOppVjBase
 * \ingroup calc
 *
 * \brief Base class for NLO definitions of Phi1 for p p -> V +1 jet at NLO
 *
 * \todo document me!
 */
//------------------------------------------------------------------------------
class PhiNLOppVjBase : public virtual IPhiNLOwithPDF
{
   protected:
      double _tau0Cut;
      double _tau1Cut;
      // list of needed LO Signatures, to be checked against when defining isPhiNPlus1
      SignatureSet _bornSignatures;
      double _cmEnergy;

   public:
      PhiNLOppVjBase(double Tau0cut, double Tau1cut, SignatureSet sigSet, double eCM);

      virtual ~PhiNLOppVjBase() = default;

      /// \name Accessors
      ///@{
      double Tau0cut() const;
      double Tau1cut() const;
      ///@}

      /// Checks if a Configuration is N-body and part of Phi_N.
      virtual bool isPhiN(const Configuration& configN) const;

      /// Checks if a Configuration is (N+1)-body and part of Phi_{N+1}.
      virtual bool isPhiNplus1(const Configuration& configNp1) const;

      virtual size_t countN(const Configuration& config) const;

      virtual bool ThetaijRijSoft(const Configuration& configNCM, const Configuration::iterator emitter, const FourVector& pSoft) const
      {
         return ThetaijSoft(configNCM, emitter, pSoft);
      }

   protected:
      PhiNLOppVjBase(const PhiNLOppVjBase&) = default;
      PhiNLOppVjBase(PhiNLOppVjBase&&) = default;
      PhiNLOppVjBase& operator=(const PhiNLOppVjBase&) = default;
      PhiNLOppVjBase& operator=(PhiNLOppVjBase&&) = default;

      bool Rij(const Configuration& configNp1CM) const;
      bool Thetaij(PhiNplus1 phiNp1CM) const;
      bool ThetaijSoft(const Configuration& configNCM, const Configuration::iterator emitter, const FourVector& pSoft) const;

      /// Checks whether a given Configuration maps back to a physical one, but the starting Configuration could have not possibly been generated trough a splitting, because the boundaries of the validity region for the map have been trespassed
      virtual bool projectionOutsideMapBoundaries(PhiNplus1 phiNp1CM, double xA, double xB) const = 0;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline PhiNLOppVjBase::PhiNLOppVjBase(double Tau0cut, double Tau1cut, SignatureSet sigSet, double cmEnergy)
   : _tau0Cut(Tau0cut), _tau1Cut(Tau1cut), _bornSignatures(sigSet), _cmEnergy(cmEnergy)
{}

//------------------------------------------------------------------------------
// Accessors
//------------------------------------------------------------------------------

inline double PhiNLOppVjBase::Tau0cut() const  { return _tau0Cut; }
inline double PhiNLOppVjBase::Tau1cut() const  { return _tau1Cut; }

} // namespace Geneva

#endif // GENEVA_PHI_NLO_PPVJ_HPP
