//------------------------------------------------------------------------------
/// \file PhiNLOppVjTau0FR.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interfaces of PhiNLOppVjTau0FR classes
//------------------------------------------------------------------------------

#ifndef GENEVA_PHI_NLO_PPVJ_TAU0FR_HPP
#define GENEVA_PHI_NLO_PPVJ_TAU0FR_HPP

#include "Geneva/Calculations/PhiNLOppVjBase.hpp"

#include "Geneva/Components/Tau0FRMapping.hpp"

#include "Geneva/Core/UI/Log.hpp"

namespace Geneva
{

class SignatureSet;
class PhiRadFKS;

//------------------------------------------------------------------------------
/**
 * \class PhiNLOppVjTau0FR
 * \ingroup calc
 *
 * \brief Defines Phi1 for p p -> V jet at NLO using recursive beam-thrust preserving map
 *
 * \todo document me!
 */
//------------------------------------------------------------------------------
class PhiNLOppVjTau0FR : public PhiNLOppVjBase
{
   private:
      Tau0FRMapping _mapTau0FR;

      double _fksTinyCut;              /// The FKS cutoff to avoid getting to close to the divergence

   public:
      PhiNLOppVjTau0FR(double Tau0cut, double Tau1cut, const SignatureSet& sigSet, double cmEnergy, bool fixqT, double tinyCut);

      virtual const std::string name() const;

      /// Constructs the split Configuration. Return true if Configuration is valid and passes all the restrictions.
      virtual bool constructPhiNplus1(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, Configuration& configNplus1CM, Configuration::iterator& emittedCM, Configuration::iterator& sisterCM, double& xA, double& xB, PhiRadFKS& phiRadFKS, double& jacobian);

      virtual bool constructSoftLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadSoftFks, double& jacobian) const;

      virtual bool constructCollinearLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadColl, double& jacobian) const;

      virtual bool constructSoftCollinearLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadSoftColl, double& jacobian) const;

      /// Returns the maximum energy that the emitted parton can achieve trough the soft limit of the mapping in the partonic CM frame
      /// NOTE: this is different from the physical maximum energy that a soft emission can reach, but it corresponds to the soft limit of the bound used in the mapping
      virtual double maxSoftEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const;

      /// Returns the maximum energy that the emitted parton can achieve trough the collinear limit of the mapping in the partonic CM frame. Also return the maximum Mandelstam S for the maximum energy configuration.
      /// NOTE: this is different from the physical maximum energy that a collinear emission can reach, but it corresponds to the collinear limit of the bound used in the mapping
      virtual std::pair<double, double> maxCollEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const;

      /// Returns the maximum energy that the emitted parton can achieve trough the soft-collinear limit of the mapping in the partonic CM frame
      /// NOTE: this is different from the physical maximum energy that a soft-collinear emission can reach, but it corresponds to the soft-collinear limit of the bound used in the mapping
      virtual double maxSoftCollEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const;

      virtual std::pair<double, double> xiYCollinearLimits(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const;

      virtual std::pair<double, double> xiYSoftCollinearLimits(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const;

      virtual std::pair<bool, bool> cutoffFKS(const PhiRadFKS& phiRad) const;

   private:
      virtual bool projectionOutsideMapBoundaries(PhiNplus1 phiNp1CM, double xA, double xB) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline PhiNLOppVjTau0FR::PhiNLOppVjTau0FR(double Tau0cut, double Tau1cut, const SignatureSet& sigSet, double cmEnergy, bool fixqT, double tinyCut)
   : PhiNLOppVjBase(Tau0cut, Tau1cut, sigSet, cmEnergy), _mapTau0FR(cmEnergy, fixqT), _fksTinyCut(tinyCut)
{
   if (_fksTinyCut < 0. or _fksTinyCut > 0.1)
      Log::error("PhiNLOppVjTau0FR") << "Tiny cut "<<_fksTinyCut<<" out of range [0, 0.1]";
}

} // namespace Geneva

#endif // GENEVA_PHI_NLO_PPVJ_TAU0FR_HPP
