//------------------------------------------------------------------------------
/// \file BeamFuncTauB.hpp
//
// Author(s):
//    Jon Walsh
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class BeamFuncTauB
//------------------------------------------------------------------------------

#ifndef GENEVA_BEAM_FUNC_TAU_B_HPP
#define GENEVA_BEAM_FUNC_TAU_B_HPP

#include "Geneva/Core/Physics/PdgID.hpp"

#include <string>

namespace beamfunc
{
class TauCoeffsQuark;
}

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class BeamFuncTauB
 * \ingroup Components
 *
 * \brief Beam thrust beam function in SCET.
 *
 * We write the beam function \f$ B_i (t,x,\mu) \f$ in terms of coefficients of
 * plus distributions in t, specifically \f$ \delta(t) \f$
 * and \f$ \frac{1}{\mu^2} \mathcal{L}_n \left(\frac{t}{\mu^2}\right) \f$.
 * These plus distributions are replaced with functions of \f$ \eta_{\rm tot} \f$
 * in resummed expresssions, and depend on components outside the beam function.
 *
 * The coefficients of these plus functions involve convolutions over
 * the perturbative \f$ I \f$ functions and the parton distribution functions.
 * An example can be found in the BeamFunctions.nb notebook in GenevaInternal
 * and in the literature in 1012.4480 in Eqs. 2.12 - 2.15 for the gluon beam function
 * and for the quark beam function in 1002.2213 in Eq. 3.48.
 *
 * We use the operator() to return the coefficients of these distributions,
 * times \f$ \alpha_s \f$ raised to the given order,
 * and the parameters that need to be supplied are
 *
 * PdgID pdgID    : Specifies the type of beam function
 * int dist       : Integer (-1, 0, ... ) specifying the distribution
 * int order      : The order in perturbation theory
 * double x       : The x value
 * double muB     : The scale
 *
 * for the dist argument, -1 gives the delta function coefficient,
 * 0 gives the coefficient of \f$ \frac{1}{\mu^2} \mathcal{L}_0 \left(\frac{t}{\mu^2}\right) \f$, etc.
 *
 * The way the component works is: the user asks for a given L_n coefficient
 * at a given order, for some flavor, x, and muB.
 * The component calls the beamfunc package for these coefficients
 */
//------------------------------------------------------------------------------
class BeamFuncTauB
{
   private:
      beamfunc::TauCoeffsQuark* _quarkTauBcoeffs;       ///< pointer to return the beam thrust coefficients

   public:
      BeamFuncTauB(const std::string& name, int member, int order, int nf);
      ~BeamFuncTauB();

      /// Returns the coefficient of the distributions in the beam function.
      double operator()(PdgID pdgID, int dist, int ord, double x, double muB)
      {
         return value(pdgID, dist, ord, x, muB);
      }

      std::string name() const;

   private:
      /// Returns the coefficient of the beam function according to the parameters given
      double value(PdgID pdgID, int dist, int ord, double x, double muB);
};

} // namespace Geneva

#endif // GENEVA_BEAM_FUNC_TAU_B_HPP
