//------------------------------------------------------------------------------
/// \file BoundingConstantsFSR.hpp
//
// Author(s):
//   Christian Bauer, Simone Alioli
//
// Copyright:
//   Copyright (C) 2014 LBNL, CERN
//
//   This file is part of the Geneva MC framework. Geneva is distributed under
//   the terms of the GNU General Public License version 3 (GPLv3), see the
//   COPYING file that comes with this distribution for details.
//   Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//   Interface of class BoundingConstantsFSR
//------------------------------------------------------------------------------

#ifndef GENEVA_BOUNDING_CONSTANTS_FSR_HPP
#define GENEVA_BOUNDING_CONSTANTS_FSR_HPP

#include "Geneva/Components/BoundingConstantsBase.hpp"

#include "Geneva/Core/Physics/PdgID.hpp"

#include <vector>
#include <map>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class BoundingConstantsFSR
 * \ingroup Components
 *
 * \brief This holds the information of the bounding constant for FSR splitting.
 *
 * The bounding constant ia a function of TauN and z, and it is
 * binned in these variables. We have a separate bounding function
 * for each flavor structure of the emission.
 *
 * We are creating a Grid of values in TauN and z, defined in the following way:
 *         t = 1. - Exp(-(Tau/10)^(1 / tauNSampling))
 * and 0 < z < 1.
 */
//------------------------------------------------------------------------------
class BoundingConstantsFSR : public BoundingConstantsBase
{
   private:
      size_t _numZBins;           ///< The number of bins for z
      double _zMin;               ///< The minimum value of z
      double _zMax;               ///< The maximum value of z
      double _binSizeZ;
      std::vector<double> _binZEdges;  ///< The values of the bin edges for z

      std::map<PdgID, std::vector<double> > _bound;   ///< The binned values of the bounding constant

   public:
      BoundingConstantsFSR(size_t numTauNBins, size_t numZBins, double tauNMin, double tauNMax, double zMin, double zMax);
      BoundingConstantsFSR(std::istream& initStream);

      /// Return the value of the bounding function for FSR, which depends on flavor structure of the emission
      double bound(Signature::iterator mother, size_t binTauN, size_t binZ);

      size_t binZ(double z);             ///< The bin in z given z

      size_t numZBins() { return _numZBins; };

      std::pair<double, double> binZEdges(size_t zBin);

      /// Access the bin edges for z
      std::vector<double>& binZEdges() { return _binZEdges; };

      /// Inform the bounding function of the ratio of function to bound given the flavor structure and z
      void informOfRatio(Signature::iterator mother, size_t tauNBin, size_t zBin, double ratio);

      virtual bool boundExists(Signature::iterator mother)const;
      virtual void write(std::ostream& outStream)const;
      virtual void printInfo() const;
};

} // namespace Geneva

#endif // GENEVA_BOUNDING_CONSTANTS_FSR_HPP
