//------------------------------------------------------------------------------
/// \file BoundingConstantsISR.hpp
//
// Author(s):
//   Christian Bauer, Simone Alioli
//
// Copyright:
//   Copyright (C) 2014 LBNL, CERN
//
//   This file is part of the Geneva MC framework. Geneva is distributed under
//   the terms of the GNU General Public License version 3 (GPLv3), see the
//   COPYING file that comes with this distribution for details.
//   Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//   Interface of class BoundingConstantsISR
//------------------------------------------------------------------------------

#ifndef GENEVA_BOUNDING_CONSTANTS_ISR_HPP
#define GENEVA_BOUNDING_CONSTANTS_ISR_HPP

#include "Geneva/Components/BoundingConstantsBase.hpp"

#include "Geneva/Core/Physics/PdgID.hpp"

#include <vector>
#include <map>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class BoundingConstantsISR
 * \ingroup Components
 *
 * \brief This holds the information of the bounding constant for ISR splitting.
 *
 * The bounding constant is a function of TauN, z and x, and it is
 * binned in these variables. We have a separate bounding function
 * for each flavor structure of the emission
 *
 * We are creating a Grid of values in TauN, z and x, defined in the following way:
 *         t = 1. - Exp(-(Tau/100)^(1 / tauNSampling))
 *         w = (Log(x / xMin) / Log(1 / xMin))^(1 / xSampling)
 *         v = (z - x) / (1 - x)
 *   This maps the volume 0 < TauN < inf, xMin < x < 1, x < z < 1 onto a unit cube
 *   Divide this cube into equally sized boxes with numTauNBins bins in t,
 *   numXbins in w and numZBins bins in v.
 *   Mapping these boxes back onto TauN, z and x defines the tiles in TauN, z and x
 */
//------------------------------------------------------------------------------
class BoundingConstantsISR : public BoundingConstantsBase
{
   private:
      size_t _numZBins;           ///< The number of bins for z
      size_t _numXBins;           ///< The number of bins for x
      double _zMin;               ///< The minimum value of z
      double _zMax;               ///< The maximum value of z
      double _xMin;               ///< The minimum value of x
      double _xMax;               ///< The maximum value of x
      int _xSampling;             ///< The power with which the x value is sampled

      std::map<PdgID, std::vector<double> > _boundA;   ///< The binned values of the bounding constant
      std::map<PdgID, std::vector<double> > _boundB;   ///< The binned values of the bounding constant

   public:
      BoundingConstantsISR(size_t numTauNBins, size_t numZBins, size_t numXBins, double tauNMin, double tauNMax, double zMin, double zMax, double xMin, double xMax);
      BoundingConstantsISR(std::istream& initStream);

      /// Return the value of the bounding function for ISR, which depends on flavor structure of the emission
      double bound(Signature::iterator mother, size_t binTauN, size_t binZ, size_t binX);

      size_t binX(double x);               ///< The bin in x given x
      size_t binZ(double z, double x);     ///< The bin in z given z and x

      size_t numXBins() { return _numXBins; };
      size_t numZBins() { return _numZBins; };

      std::pair<double, double> binXEdges(size_t xBin) const;
      std::pair<double, double> binZEdges(size_t zBin, double x) const;

      double xMin() { return _xMin; };
      double xMax() { return _xMax; };
      double zMin() { return _zMin; };
      double zMax() { return _zMax; };

      double xVal(double w) const;
      double zVal(double v, double w) const;
      double wVal(double x) const;
      double vVal(double z, double x) const;

      /// Inform the bounding function of the ratio of function to bound given the flavor structure, z and x
      void informOfRatio(Signature::iterator mother, size_t tauNBin, size_t zBin, size_t xBin, double ratio);

      virtual bool boundExists(Signature::iterator mother) const;
      virtual void write(std::ostream& outStream) const;
      virtual void printInfo() const;
};

} // namespace Geneva

#endif //  GENEVA_BOUNDING_CONSTANTS_ISR_HPP
