//------------------------------------------------------------------------------
/// \file Collider.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Collider
//------------------------------------------------------------------------------

#ifndef GENEVA_COLLIDER_HPP
#define GENEVA_COLLIDER_HPP

#include "Geneva/Core/Physics/PdgID.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include <string>

namespace Geneva
{

//------------------------------------------------------------------------------
class Collider
{
   public:
      /// Named constants to specify collider
      enum Type
      {
         kUnknown = 0,
         kEplusEminus = 1,
         kEminusEplus = 2,
         kProtonProton = 3,
         kProtonAntiproton = 4
      };

   private:
      Type _collider;
      PdgID _beamA;
      PdgID _beamB;

   public:
      Collider(const std::string& colliderstr);
      Collider(const OptionMap& optMapt);

      Type collider() const;
      PdgID beamA() const;
      PdgID beamB() const;
      bool isEE() const;
      bool isHadronic() const;

      friend std::ostream& operator<<(std::ostream& out, const Collider& c);

   private:
      void init(const std::string& colliderstr);

      std::ostream& print(std::ostream& out) const;
};

//------------------------------------------------------------------------------
inline Collider::Collider(const std::string& colliderstr)
{
   init(colliderstr);
}

//------------------------------------------------------------------------------
inline Collider::Collider(const OptionMap& optMap)
{
   init(optMap.getOption<std::string>("Collider"));
}

inline void Collider::init(const std::string& colliderstr)
{
   if (colliderstr == "pp") {
      _beamA = PdgID::kProton;
      _beamB = PdgID::kProton;
      _collider = kProtonProton;
   } else if (colliderstr == "ppbar") {
      _beamA = PdgID::kProton;
      _beamB = PdgID::kProtonBar;
      _collider = kProtonAntiproton;
   } else if (colliderstr == "e+e-") {
      _beamA = PdgID::kPositron;
      _beamB = PdgID::kElectron;
      _collider = kEplusEminus;
   } else if (colliderstr == "e-e+") {
      _beamA = PdgID::kElectron;
      _beamB = PdgID::kPositron;
      _collider = kEplusEminus;
   } else {
      _beamA = PdgID::kUnknown;
      _beamB = PdgID::kUnknown;
      _collider = kUnknown;
      Log::fatal("Collider::init") << "Unknown collider " << colliderstr << Log::end;
   }
}

//------------------------------------------------------------------------------
inline Collider::Type Collider::collider() const
{
   return _collider;
}

//------------------------------------------------------------------------------
inline PdgID Collider::beamA() const
{
   return _beamA;
}

//------------------------------------------------------------------------------
inline PdgID Collider::beamB() const
{
   return _beamB;
}

//------------------------------------------------------------------------------
inline bool Collider::isEE() const
{
   return _collider == kEminusEplus || _collider == kEplusEminus;
}

//------------------------------------------------------------------------------
inline bool Collider::isHadronic() const
{
   return _collider == kProtonProton || _collider == kProtonAntiproton;
}

//------------------------------------------------------------------------------
inline std::ostream& Collider::print(std::ostream& out) const
{
   std::string name;
   if (_collider == kProtonProton) {
      name = "pp";
   } else if (_collider == kProtonAntiproton) {
      name = "ppbar";
   } else if (_collider == kEplusEminus) {
      name = "e+e-";
   } else if (_collider == kEminusEplus) {
      name = "e-e+";
   } else
      name = "Unknown collider";
   return out << name;
}

//------------------------------------------------------------------------------
inline std::ostream& operator<<(std::ostream& out, const Collider& c)
{
   return c.print(out);
}

} // namespace Geneva

#endif // GENEVA_COLLIDER_HPP
