//------------------------------------------------------------------------------
/// \file CouplingEW.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class CouplingEW
//------------------------------------------------------------------------------

#ifndef GENEVA_COUPLING_EW_HPP
#define GENEVA_COUPLING_EW_HPP

#include "Geneva/Components/ICoupling.hpp"

#include <complex>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class CouplingEW
 * \ingroup Coupling
 *
 * \brief (Rudimentary) implementation of a running electromagnetic coupling constant.
 *
 * Uses alpha_Thomson below the electron mass, an interpolation between
 * the electron mass and the tau mass, and the evolution equation above the
 * tau mass, considering also the b quark threshold.
 *
 * \warning \todo Currently this is only crude and uses hard-coded values
 * for the coupling below m_b = 4.5 GeV, as well as for the running parameters
 * above that scale.
 */
//------------------------------------------------------------------------------
class CouplingEW : public ICoupling
{
   private:
      double _value0;        ///< The value of the coupling at the start scale mu0
      double _mu0;           ///< The start scale mu0

   public:
      /// \name Constructors etc.
      CouplingEW(double value0, double mu0);

      CouplingEW(const CouplingEW&) = delete;
      CouplingEW(CouplingEW&&) = delete;
      CouplingEW& operator=(const CouplingEW&) = delete;
      CouplingEW& operator=(CouplingEW&&) = delete;

      ~CouplingEW();
      ///@}

      /// \name ICoupling Interface
      ///@{
      virtual double beta(int n) const;

      virtual std::string name() const;
      ///@}

   private:
      // Computes the value of the running coupling.
      virtual double value(double mu) const;

      // Overloading value function to take a complex scale.
      virtual std::complex<double> value(std::complex<double> mu) const;

      struct _interp_type;
      struct _spline_type;
      _interp_type* _interp;
      _spline_type* _spline;

      static const double _oneOverAlphaEW[];
      static const double _logSqrtScale[];
      static const double _bMass;
};

} // namespace Geneva

#endif // GENEVA_COUPLING_EW_HPP
