//------------------------------------------------------------------------------
/// \file CouplingFixed.hpp
//
// Author(s):
//    Frank Tackmann, Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 MIT, LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class CouplingFixed
//------------------------------------------------------------------------------

#ifndef GENEVA_COUPLING_FIXED_HPP
#define GENEVA_COUPLING_FIXED_HPP

#include "Geneva/Components/ICoupling.hpp"

#include <complex>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class CouplingFixed
 * \ingroup Coupling
 *
 * \brief Implements a scale independent fixed coupling constant.
 *
 * The class simply returns a fixed value irrespective of the given scale.
 */
//------------------------------------------------------------------------------
class CouplingFixed : public ICoupling
{
   private:
      double _value;                   ///< value of the coupling

   public:
      CouplingFixed(double value);

      /// Since the coupling is fixed, it always returns 0.0 for the beta function for any n.
      virtual double beta(int n) const;

      virtual std::string name() const;

   private:
      // Returns the value of the coupling at the scale mu (in GeV)
      virtual double value(double mu) const;
      virtual std::complex<double> value(std::complex<double> mu) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors and Destructor
inline CouplingFixed::CouplingFixed(double value)
   : _value(value)
{}

//------------------------------------------------------------------------------
inline double CouplingFixed::beta(int n) const
{
   return 0.0;
}

//------------------------------------------------------------------------------
inline double CouplingFixed::value(double mu) const
{
   return _value;
}

//------------------------------------------------------------------------------
//Does not implement a complex fixed coupling as it is not currently needed.
inline std::complex<double> CouplingFixed::value(std::complex<double> mu) const
{
   return _value;
}

//------------------------------------------------------------------------------
inline std::string CouplingFixed::name() const
{
   std::stringstream s;
   s << "CouplingFixed at value "<< _value;
   return s.str();
}

} // namespace Geneva

#endif // GENEVA_COUPLING_FIXED_HPP
