//------------------------------------------------------------------------------
/// \file CouplingStrong.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2013 MIT, LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class CouplingStrong
//------------------------------------------------------------------------------

#ifndef GENEVA_COUPLING_STRONG_HPP
#define GENEVA_COUPLING_STRONG_HPP

#include "Geneva/Components/ICoupling.hpp"

#include "Geneva/Core/Physics/Order.hpp"

#include <complex>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class CouplingStrong
 * \ingroup Coupling
 *
 * \brief Implementation of the running strong coupling constant.
 */
//------------------------------------------------------------------------------
class CouplingStrong : public ICoupling
{
   private:
      double _value0;                  ///< The value of the coupling at the start scale mu0
      double _mu0;                     ///< The start scale mu0
      RunOrder _run;                   ///< The order used in the running
      int _nf;                         ///< The number of massless quark flavors used in the running
      double _lambdaQCD;               ///< The value of Lambda_QCD

   public:
      /// \name Constructors etc.
      ///@{
      CouplingStrong(double value0, double mu0, RunOrder run, int nf);
      ///@}

      /// \name Queries and Accessors
      ///@{
      double value0() const;           ///< Returns the value at the start scale mu0.
      double mu0() const;              ///< Returns the start scale mu0.
      RunOrder runOrder() const;       ///< Returns the running order.
      int nf() const;                  ///< Returns the number of light flavors.
      double lambdaQCD() const;        ///< Returns the value of Lambda_QCD
      ///@}

      /// \name ICoupling Interface
      ///@{
      // Returns the value of the QCD beta function at (n+1)-loop order.
      virtual double beta(int n) const;

      virtual std::string name() const;
      ///@}

   private:
      // Computes the value of the running coupling at the scale mu.
      virtual double value(double mu) const;
      virtual std::complex<double> value(std::complex<double> mu) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Queries
//------------------------------------------------------------------------------
inline double CouplingStrong::value0() const     { return _value0; }
inline double CouplingStrong::mu0() const        { return _mu0; }
inline RunOrder CouplingStrong::runOrder() const { return _run; }
inline int CouplingStrong::nf() const            { return _nf; }
inline double CouplingStrong::lambdaQCD() const  { return _lambdaQCD; }

} // namespace Geneva

#endif // GENEVA_COUPLING_STRONG_HPP
