//------------------------------------------------------------------------------
/// \file FKSKinematics.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2013 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class FKSKinematics
//------------------------------------------------------------------------------

#ifndef GENEVA_FKS_KINEMATICS_HPP
#define GENEVA_FKS_KINEMATICS_HPP

#include "Geneva/Core/Physics/Configuration.hpp"
#include "Geneva/Core/Physics/ConfigurationExtensions.hpp"

namespace Geneva
{

class PhiRadFKS;

//------------------------------------------------------------------------------
/**
 * \struct PhiNcmFKS
 * \ingroup comp
 *
 * \brief Container that serves as input to an FKS splitting.
 *
 * It contains a references to a Configuration in the partonic CM frame
 * together with an iterator pointing to the emitter.
 *
 * \note It is the user's responsiblity to ensure that the Configuration is the
 * partonic CM frame and that the emitter is a valid iterator.
 */
//------------------------------------------------------------------------------
struct PhiNcmFKS
{
   const Configuration& configCM;
   Configuration::iterator emitter;

   explicit PhiNcmFKS(const Configuration& configCM)
      : configCM(configCM), emitter(configCM.end()) {};

   PhiNcmFKS(const Configuration& configCM, Configuration::iterator emitter)
      : configCM(configCM), emitter(emitter) {};
};


//------------------------------------------------------------------------------
/**
 * \struct PhiNplus1FKS
 * \ingroup comp
 *
 * \brief The output of a FKSKinematics::split : an (N+1)-body phase space point in the CM and the particle produced by the split
 *
 * A %PhiNplus1FKS struct encapsulates the %Configuration in the CM frame which is the
 *  result of the splitting and two iterators to the FKS parton and its sister.
 */
//------------------------------------------------------------------------------
struct PhiNplus1FKS
{
   Configuration config;
   Configuration::iterator FKSparton;
   Configuration::iterator sister;

   PhiNplus1FKS()
      : config(Configuration()), FKSparton(config.end()), sister(config.end()) {};

   explicit PhiNplus1FKS(const Configuration& config)
      : config(config), FKSparton(config.end()), sister(config.end()) {};
};

//------------------------------------------------------------------------------
/**
 * \class FKSKinematics
 * \ingroup comp
 *
 * \brief Encapsulates the FKS kinematics between \f$\Phi_N$\f and \f$\Phi_{N+1}$\f.
 *
 * The primary use of FKSKinematics is to combine an N-parton  %Configuration
 * phiN with a set of splitting variables %PhiRadFKS and produce an N+1-parton
 * %Configuration.  The N-body %Configuration includes an emitting parton, and
 * the N+1-body %Configuration includes an "FKS parton" and a "sister parton":
 * the two new particles produced in the splitting.  To keep track of these
 * special particles, the input to a kinematics is a %PhiNcmFKS and the output is a
 * %PhiNplus1FKS. These two structs are simply %Configurations augmented with
 * iterators to the particles participating in the splitting. The initial
 * %Configuration is assumed to be provided in the partonic CM frame, and
 * checks are in place to assure that. The final %Configuration is also
 * returned in its own CM frame. In order for the user to boost the final
 * Configuration to the Lab frame when an ISR occurs,the new values of xA and xB
 * can be queried via the getXA() and getxB() methods. For FSR instead, the new values
 * of xA and xB corresponds to the old ones.
 */
//------------------------------------------------------------------------------
struct FKSKinematics
{
   /// \name Bjorken x's
   /// Get the momentum fractions for the N+1-body  and N-body configurations
   ///@{
   double getxA(const double xBarA, const double xi, const double y) const;
   double getxB(const double xBarB, const double xi, const double y) const;
   double getxBarA(const double xA, const double xi, const double y) const;
   double getxBarB(const double xB, const double xi, const double y) const;

   ///}@

   /// Get the maximum xi value allowed for the actual configuration
   double getXiMax(const PhiNcmFKS& phiN, double xBarA, double xBarB, double y) const;

   /// Get the maximum xi value allowed for the collinear limit of the actual configuration
   double getXiMaxColl(const PhiNcmFKS& phiN, double xBarA, double xBarB) const;

   /// generate a PhiNplus1FKS via ISR splitting. Assumes a valid PhiRadFKS is provided. The generated configuration and its iterators  are put in PhiNplus1FKS. The jacobian is returned.
   double splitISR(const PhiNcmFKS& phiN, const PhiRadFKS& phiRad, double xBarA, double xBarB, PhiNplus1FKS& phiNp1);

   /// generate a PhiNplus1FKS via FSR splitting. Assumes a valid PhiRadFKS is provided. The generated configuration and its iterators are put in PhiNplus1FKS .The jacobian is returned.
   double splitFSR(const PhiNcmFKS& phiN, const PhiRadFKS& validPhiRad, PhiNplus1FKS& phiNp1);

   /// Creates an N-body %Configuration from the N+1-body configuration by combining the the two FKS partons. It reverts an ISR splitting, assuming the PhiNp1 is given in the CM frame.
   Configuration::iterator projectConfigISR(PhiNplus1 phiNp1, Configuration& projConfig) const;
   const Configuration projectConfigISR(PhiNplus1 phiNp1) const;

   /// Creates an N-body %Configuration from the N+1-body configuration by combining the two FKS partons. It reverts a FSR splitting, assuming the PhiNp1 is given in the CM frame.
   Configuration::iterator projectConfigFSR(PhiNplus1 phiNp1, Configuration& projConfig) const;
   const Configuration projectConfigFSR(PhiNplus1 phiNp1) const;

   /// Creates the PhiRadFKS from the N+1-body configuration, assuming the PhiNp1 is given in the CM frame.
   const PhiRadFKS projectPhiRad(PhiNplus1 phiNp1) const;

   static double getPhiAngleFSR(const ThreeVector& fksDir, const ThreeVector& emitterDir);
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline const Configuration FKSKinematics::projectConfigISR(PhiNplus1 phiNp1) const
{
   Configuration tmp;
   projectConfigISR(phiNp1, tmp);
   return tmp;
}

//------------------------------------------------------------------------------
inline const Configuration FKSKinematics::projectConfigFSR(PhiNplus1 phiNp1) const
{
   Configuration tmp;
   projectConfigFSR(phiNp1, tmp);
   return tmp;
}

} // namespace Geneva

#endif // GENEVA_FKS_KINEMATICS_HPP
