//------------------------------------------------------------------------------
/// \file FKSMapping.hpp
//
// Author(s):
//    Jon Walsh, Saba Zuberi, Frank Tackmann, Christian Bauer, Chris Vermilion, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class FKSMapping
//------------------------------------------------------------------------------

#ifndef GENEVA_FKS_MAPPING_HPP
#define GENEVA_FKS_MAPPING_HPP

#include "Geneva/Components/FKSKinematics.hpp"

#include "Geneva/Core/Physics/ConfigurationExtensions.hpp"

namespace Geneva
{

class PhiRadFKS;

//------------------------------------------------------------------------------
/**
 * \class FKSMapping
 * \ingroup comp
 *
 * \brief Encapsulates the logic of the FKS mapping between \f$\Phi_N$\f and \f$\Phi_{N+1}$\f.
 *
 * The primary use of FKSMapping is to combine an N-parton %Configuration
 * phiN with a set of splitting variables %PhiRad and produce an N+1-parton
 * %Configuration.  The N-body %Configuration includes an emitting parton, and
 * the N+1-body %Configuration includes an "FKS parton" and a "sister parton":
 * the two new particles produced in the splitting.  To keep track of these
 * special particles, the input to a mapping is a %PhiNFKS and the output is a
 * %PhiNplus1FKS.  These two structs are simply %Configurations augmented with
 * iterators to the particles participating in the splitting.
 */
//------------------------------------------------------------------------------
class FKSMapping
{
   private:
      double _jacobian;             ///< jacobian from splitting

   public:
      /// Generates the splitting.
      bool generateSplit(const PhiNcmFKS& phiN, const PhiRadFKS& phiRadFKS, double x1, double x2, PhiNplus1FKS& phiNp1);

      /// Get the jacobian associated with the splitting
      double getJacobian() const;

      /// Get the momentum fractions for the N+1-body configuration
      double getxA(const double xBarA, const double xi, const double y) const;
      double getxB(const double xBarB, const double xi, const double y) const;

      /// Get the maximum xi value allowed for the actual configuration
      double getXiMaxReal(const PhiNcmFKS& phiN, const double xBarA, const double xBarB, const double y) const;

      /// Get the maximum xi value allowed for the collinear limit of the actual configuration
      double getXiMaxColl(const PhiNcmFKS& phiN, const double xBarA, const double xBarB) const;

      /// Creates an N-body configuration from the N+1-body configuration by combining the two FKS partons
      bool project(PhiNplus1 phiNp1, Configuration& phiN) const;

      std::string name() const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
/// Returns the Jacobian factor to go from dPhi_{n+1} to d\bar{Phi}_n dPhi_rad (x,cos theta,phi)
inline double FKSMapping::getJacobian() const
{
   return _jacobian;
}

//------------------------------------------------------------------------------
inline std::string FKSMapping::name() const
{
   return "FKSMapping";
}

} // namespace Geneva

#endif // GENEVA_FKS_MAPPING_HPP
