//------------------------------------------------------------------------------
/// \file ICommonCouplings.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class ICommonCouplings
//------------------------------------------------------------------------------

#ifndef GENEVA_I_COM_COUPL_HPP
#define GENEVA_I_COM_COUPL_HPP

#include <string>

namespace Geneva
{

/// \defgroup CommonCouplings
/// \ingroup comp

class Configuration;

//------------------------------------------------------------------------------
/// Defines the functional form of the couplings dependence.
/** \class ICommonCouplings
 *  \ingroup calc
 *
 * Defines the abstract interface to calculate the LO couplings for a given configuration.
 *
 */
//------------------------------------------------------------------------------
class ICommonCouplings
{
   public:
      virtual ~ICommonCouplings() = default;

      /// Returns the value of the couplings for the given %Configuration at the scale mu.
      double operator()(const Configuration& config, const double mu);

      /// Returns the name of the component.
      virtual std::string name() const = 0;

      /// Returns the number of powers of alpha strong present in common couplings
      virtual int numAlphaStrongPowers() const = 0;

   protected:
      ICommonCouplings() = default;
      ICommonCouplings(const ICommonCouplings&) = default;
      ICommonCouplings(ICommonCouplings&&) = default;
      ICommonCouplings& operator=(const ICommonCouplings&) = default;
      ICommonCouplings& operator=(ICommonCouplings&&) = default;

   private:
      /// Returns the value of the couplings for the given configuration.
      /**
      * This is a pure virtual hook function which is called by operator() to
      * do the actual computation and has to be implemented by derived classes.
      */
      virtual double value(const Configuration& config, const double mu) const = 0;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Declarations
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline double ICommonCouplings::operator()(const Configuration& config, const double mu)
{
   return value(config, mu);
}

} // namespace Geneva

#endif // GENEVA_I_COM_COUPL_HPP
