//------------------------------------------------------------------------------
/// \file ICoupling.hpp
//
// Author(s):
//    Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class ICoupling
//------------------------------------------------------------------------------

#ifndef GENEVA_I_COUPLING_HPP
#define GENEVA_I_COUPLING_HPP

#include <string>
#include <complex>

namespace Geneva
{

/// \defgroup Coupling
/// \ingroup comp

//------------------------------------------------------------------------------
/**
 * \class ICoupling
 * \ingroup Coupling
 *
 * \brief Abstract interface to coupling constant component
 *
 * Defines the abstract interface to any components that implement a scale-
 * dependent coupling constant. The renormalization scale mu is in GeV and is
 * provided as an argument to operator(mu) which returns the value of the
 * coupling at the scale mu.
 */
//------------------------------------------------------------------------------
class ICoupling
{
   public:
      virtual ~ICoupling() = default;

      /// Returns the value of the coupling at the scale mu (in GeV).
      double operator()(double mu);
      std::complex<double> operator()(std::complex<double> mu);

      /// Returns the beta function used to perform the calculation.
      virtual double beta(int n) const = 0;

      /// Returns the name of the component.
      virtual std::string name() const = 0;

   protected:
      ICoupling() = default;
      ICoupling(const ICoupling&) = default;
      ICoupling(ICoupling&&) = default;
      ICoupling& operator=(const ICoupling&) = default;
      ICoupling& operator=(ICoupling&&) = default;

   private:
      /**
       * \brief Returns the value of the coupling at the scale mu (in GeV).
       *
       * This is a pure virtual hook function which is called by operator() to
       * do the actual computation and has to be implemented by derived classes.
       */
      virtual double value(double mu) const = 0;
      virtual std::complex<double> value(std::complex<double> mu) const = 0;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Returns the value of the coupling at the scale mu (in GeV).
inline double ICoupling::operator()(double mu)
{
   return value(mu);
}

//------------------------------------------------------------------------------
// Returns the value of the complex coupling at the complex scale mu (in GeV).
inline std::complex<double> ICoupling::operator()(std::complex<double> mu)
{
   return value(mu);
}

} // namespace Geneva

#endif // GENEVA_I_COUPLING_HPP
