//------------------------------------------------------------------------------
/// \file IOneLoopME.hpp
//
// Author(s):
//    Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class IOneLoopME
//------------------------------------------------------------------------------

#ifndef GENEVA_I_ONE_LOOP_ME_HPP
#define GENEVA_I_ONE_LOOP_ME_HPP

#include <string>

namespace Geneva
{

/// \defgroup OneLoopME
/// \ingroup  Components

class Configuration;
class SignatureSet;

//------------------------------------------------------------------------------
/**
 * \ingroup OneLoopME
 * \brief   Calculates the interference term between a one-loop and tree-level ME.
 *
 * Defines the abstract interface to calculate the interference between an implentation of a
 * one-loop and tree-level matrix element.
 *
 * By convention this gives the standard squared invariant matrix-element but *without* including
 * the overall powers of coupling constants.
 *
 * All calculations have to be implemented in the MSbar scheme, and there is an overall prefactor
 * that has been ignored, which is given by
 *  1/Gamma(1-epsilon) * (4 \pi mu^2 / scale^2 )^ epsilon
 *
 * The IOneLoop matrix element is assumed to be evaluated at a scale set by the setScale() method,
 * which needs to be called before any calculation is possible.
 */
//------------------------------------------------------------------------------
class IOneLoopME
{
   public:
      virtual ~IOneLoopME() = default;

      /// Returns the value of the 1/eps^2 term.
      virtual double getDoublePole(const Configuration& config) const = 0;

      /// Returns the value of the 1/eps term.
      virtual double getSinglePole(const Configuration& config) const = 0;

      /// Returns the value of the finite term.
      virtual double getFiniteTerm(const Configuration& config) const = 0;

      /// Returns all terms in the provided arguments.
      virtual bool   getAllTerms(const Configuration& config,
                                 double& doublePole,
                                 double& singlePole,
                                 double& finiteTerm) const = 0;

      /// Checks compatibility with the given SignatureSet.
      virtual bool checkCompatibility(const SignatureSet& sigSet) const = 0;

      /// Returns the name of the one-loop matrix element.
      virtual std::string name() const = 0;

      /// Needed to define the common scale used in the evaluation
      void setScale(double scale);

   protected:
      IOneLoopME() = default;
      IOneLoopME(const IOneLoopME&) = default;
      IOneLoopME(IOneLoopME&&) = default;
      IOneLoopME& operator=(const IOneLoopME&) = default;
      IOneLoopME& operator=(IOneLoopME&&) = default;

      // Internal scale
      double _scale;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline void IOneLoopME::setScale(double scale)
{
   _scale = scale;
}

} // namespace Geneva

#endif // GENEVA_I_ONE_LOOP_ME_HPP
