//------------------------------------------------------------------------------
/// \file IPDF.hpp
//
// Author(s):
//    Chris Vermilion, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class IPDF
//------------------------------------------------------------------------------

#ifndef GENEVA_I_PDF_HPP
#define GENEVA_I_PDF_HPP

#include "Geneva/Core/Physics/PdgID.hpp"

#include <string>

namespace Geneva
{

/// \defgroup PDF
/// \ingroup comp

//------------------------------------------------------------------------------
/**
 * \class IPDF
 * \ingroup PDF
 *
 * \brief Defines the abstract interface for a PDF component.
 *
 * The user can query the particle density function \f$ f(x, Q) \f$ for
 * a given particle inside a beam particle, undergoing the
 * collision through  \c pdfA([idParticle,] x, Q)
 * and \c pdfB([idParticle,] x, Q).
 * A and B correspond to the \c initialA and \c initialB labeling in %Signatures
 * and %Configurations.
 */
//------------------------------------------------------------------------------
class IPDF
{
   public:
      virtual ~IPDF() = default;

       /**
        * \brief Returns the PDF for beam A/B for the given particle \a id, \a x, and \a muF.
        * \return the PDF value
        */
      virtual double pdfA(PdgID id, double x, double muF) const = 0;
      virtual double pdfB(PdgID id, double x, double muF) const = 0;

      /**
       * \brief Returns the PdgID for beam particle A/B.
       */
      virtual PdgID beamParticleA() const = 0;
      virtual PdgID beamParticleB() const = 0;

      /**
       * \brief Returns the name of the PDF for beam A/B.
       */
      virtual std::string nameA() const = 0;
      virtual std::string nameB() const = 0;

      /**
       * \brief Returns the member number for PDF set for A/B.
       *
       * Corresponds to the usual member number of the lhapdf set with 0 being
       * the central member.
       */
      virtual int memberA() const { return 0; }
      virtual int memberB() const { return 0; }

      /**
       * \brief Returns a unique identifier for the PDF for beam A/B.
       */
      virtual int pdfIDA() const { return -1; }
      virtual int pdfIDB() const { return -1; }

      /// Returns the number of flavours used.
      virtual int getNf() const = 0;

      /// Returns the name of the component.
      virtual std::string name() const = 0;

      /// Whether the PDF is enforced to be positive definite
      virtual bool isPositive() const { return false; }

      /// Returns the luminosity, i.e. the product of the two PDFs.
      double luminosity(PdgID idA, double xA, PdgID idB, double xB, double muF);

   protected:
      IPDF() = default;
      IPDF(const IPDF&) = default;
      IPDF(IPDF&&) = default;
      IPDF& operator=(const IPDF&) = default;
      IPDF& operator=(IPDF&&) = default;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline double IPDF::luminosity(PdgID idA, double xA, PdgID idB, double xB, double muF)
{
   return pdfA(idA, xA, muF) * pdfB(idB, xB, muF);
}

} // namespace Geneva

#endif // GENEVA_I_PDF_HPP
