//------------------------------------------------------------------------------
/// \file IScale.hpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2016 CERN, DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class IScale
//------------------------------------------------------------------------------

#ifndef GENEVA_I_SCALE_HPP
#define GENEVA_I_SCALE_HPP

#include <string>

namespace Geneva
{

class Configuration;

//------------------------------------------------------------------------------
/**
 * \class IScale
 * \ingroup comp
 *
 * \brief Abstract interface for renormalization/factorization scales.
 *
 * Defines the abstract interface to obtain a renormalization or factorization
 * scales for a given configuration.
 */
//------------------------------------------------------------------------------
class IScale
{
   public:
      virtual ~IScale() = default;

      /// Returns the value of the scale (in GeV) for the given %Configuration.
      double operator()(const Configuration& config);

      /// Returns the name of the component.
      virtual std::string name() const = 0;

   protected:
      IScale() = default;
      IScale(const IScale&) = default;
      IScale(IScale&&) = default;
      IScale& operator=(const IScale&) = default;
      IScale& operator=(IScale&&) = default;

   private:
      /**
       * \brief Returns the value of the scale (in GeV) for the given configuration.
       *
       * This is a pure virtual hook function which is called by operator() to
       * do the actual computation and has to be implemented by derived classes.
       */
      virtual double value(const Configuration& config) const = 0;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Returns the value of the scale (in GeV) for the given configuration.
inline double IScale::operator()(const Configuration& config)
{
   return value(config);
}

} // namespace Geneva

#endif // GENEVA_I_SCALE_HPP
