//------------------------------------------------------------------------------
/// \file ISplittingMapping.hpp
//
// Author(s):
//    Christian Bauer
//
// Copyright:
//    Copyright (C) 2014 <author's copyright holder>
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class ISplittingMapping
//------------------------------------------------------------------------------

#ifndef GENEVA_I_SPLITTING_MAPPING_HPP
#define GENEVA_I_SPLITTING_MAPPING_HPP

#include "Geneva/Components/RegionInfoList.hpp"

#include "Geneva/Core/Physics/Signature.hpp"
#include "Geneva/Core/Physics/SignatureSet.hpp"

#include <list>
#include <string>

namespace Geneva
{

class Configuration;
class RegionInfo;
class PhiN;
class PhiNplus1;
class PhiNplus1CMWritable;

//------------------------------------------------------------------------------
/**
 * \class ISPlittingFunctionMapping
 * \ingroup Components
 *
 * \brief Interface for the phase space mapping used in a splitting function
 */
//------------------------------------------------------------------------------
class ISplittingMapping
{
   public:
      virtual ~ISplittingMapping() = default;

      /**
       * \brief Returns the list of possible RegionInfos.
       */
      virtual RegionInfoList getRegionInfoList() = 0;

      /**
       * \brief Create the projected phase space
       *
       * Create the (N-1)-body config, the values of TauN, z and phi,
       * and the regionInfo from the N-body config
       * This functions returns true if the projection succeeds, and false otherwise
       * The function assumes that the emitted and sister in PhiNplus1 are the closest QCD pair, and that
       * this pair can come from a QCD splitting
       *
       * This function returns false if one of the following conditions are satisfied
       * 1) The Phi2 phase space can not be projected (it is not covered by the map)
       * 2) The (N+1)-body configuration does not allow for the required splitting
       */
      virtual bool calcPhiNValues(const PhiNplus1& config_Nplus1, Configuration& config_N, double& Tau_N, double& z, double& phi, RegionInfo& regionInfo) const = 0;

      /**
       * \brief Create the splitted phase space
       *
       * Create the (N+1)-body config, given the values of PhiN, TauN, z and phi,
       * This functions returns true if the splitting succeeds, and false otherwise
       */
      virtual bool generateSplit(const PhiN& phiN, const double Tau_N, const double z , const double phi, const RegionInfo& region, PhiNplus1CMWritable& phiNp1) const = 0;

      /**
       * \brief Check if PhiN, TauN, z, phi are a valid point for the splitting
       *
       * This checks the following conditions:
       * 1) The resulting Phi_N+1 phase space point is kinematically valid
       * 2) The emitted particle has the closest distance measure
       * 3) For ISR, the resulting z value is above the corresponding x value
       */
      virtual bool isValidPhiN(const Configuration& config_N, double Tau_N, double z, double phi, Signature::const_iterator mother) const = 0;

      /**
       * \brief Return the Jacobian dPhiN / dPhiNmin1 / dTauN / dz / dphi
       *
       * Assumes that the input is valid, i.e., that isValidPhiN is true for the input
       */
      virtual double jacobian(const Configuration& config_N, double Tau_N, double z, double phi, const RegionInfo& map) const = 0;

      /**
       * \brief Return the value of the transverse momentum of the emission
       */
      virtual double kTEmission(PdgID motherID, double TauN, double z) const = 0;

      /**
       * \brief Return a lower bound to the smallest value of z given PhiN, TauN, phi
       */
      virtual double zMin(const Configuration& config_N, const double Tau_N, const double phi, Signature::const_iterator mother, PdgID emittedID) = 0;

      /**
       * \brief Return an upper bound to the largest value of z given PhiN, TauN, phi
       */
      virtual double zMax(const Configuration& config_N, const double Tau_N, const double phi, Signature::const_iterator mother, PdgID emittedID) = 0;

      /**
       * \brief Return a lower bound to the smallest value of phi given PhiN, TauN
       */
      virtual double phiMin(const Configuration& config_N, const double Tau_N, Signature::const_iterator mother, PdgID emittedID) = 0;

      /**
       * \brief Return an upper bound to the largest value of phi given PhiN, TauN
       */
      virtual double phiMax(const Configuration& config_N, const double Tau_N, Signature::const_iterator mother, PdgID emittedID) = 0;

      /**
        * \brief Return the allowed splittings given a map
        *
        * Returns all splittings that are summed over when normalizing a splitting function.
        */
      virtual std::list<RegionInfo> getAllowedSplittings(const RegionInfo& map) = 0;

      /**
       * \brief Return the number of possible singular projections given a map
       */
      virtual std::list<RegionInfo> getAllowedProjections(const RegionInfo& map) = 0;

      /**
       * \brief Return the number of projections given a map
       *
       * This returns the number of possible singular projections
       */
      virtual size_t numQCDN() = 0;

      /**
       * \brief Return the N-1-body signature set
       */
      virtual SignatureSet sigsNmin1() = 0;

      /**
       * \brief Return the N-body signature set
       */
      virtual SignatureSet sigsN() = 0;

      /**
       * \brief Return the name of the splitting mapping
       */
      virtual std::string name() const = 0;

   protected:
      ISplittingMapping() = default;
      ISplittingMapping(const ISplittingMapping&) = default;
      ISplittingMapping(ISplittingMapping&&) = default;
      ISplittingMapping& operator=(const ISplittingMapping&) = default;
      ISplittingMapping& operator=(ISplittingMapping&&) = default;
};

} // namespace Geneva

#endif // GENEVA_I_SPLITTING_MAPPING_HPP
