//------------------------------------------------------------------------------
/// \file ITreeME.hpp
//
// Author(s):
//    Chris Vermilion, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2013 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class ITreeME
//------------------------------------------------------------------------------

#ifndef GENEVA_I_TREE_ME_HPP
#define GENEVA_I_TREE_ME_HPP

#include <string>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \defgroup TreeME Tree-level matrix elements
 * \ingroup comp
 */
//------------------------------------------------------------------------------

class SignatureSet;
class Configuration;

//------------------------------------------------------------------------------
/**
 * \class ITreeME
 * \ingroup TreeME
 *
 * \brief Abstract interface to tree-level matrix elements.
 *
 * Defines the abstract interface to any components that implement tree-level
 * matrix elements.
 *
 * By convention the value returned by operator() is the standard squared
 * invariant matrix-element but *without* including the overall powers of
 * coupling constants. That is, we write the tree-level matrix element squared
 * for N final state particles as
 *
 * \f$|M_N|^2 = (alpha_i)^N |M_N^0|^2\f$
 *
 * where \f$(\alpha_i)^N\f$ is the appropriate combination of \f$\alpha_\text{em}\f$ and \f$\alpha_\text{strong}\f$ and
 * the derived classes implement \f$|H_N^0|^2\f$.
 */
//------------------------------------------------------------------------------
class ITreeME
{
   public:
      virtual ~ITreeME() = default;

      /**
       * \brief Computes the squared tree-level matrix element for a Configuration.
       * \param config The Configuration to act on.
       * \return The result for the squared matrix element.
       *
       * The returned value excludes all overall powers of the coupling
       * constants.
       *
       * \note Evaluation is currently defined as a non-const operation.
       */
      double operator()(const Configuration& config) { return value(config); }

      /**
       * \brief Checks the compatibility with the given SignatureSet.
       * \param sigSet The SignatureSet to check
       * \return The level of compatibility (to be defined)
       *
       * This is a pure virtual hook function to be implemented by derived classes.
       *
       * The return value is 0 if the class is able to evaluate the tree-level
       * matrix elements for all signatures in the set. If it can't handle even
       * one of the signatures it returns and a nonzero error code (with some
       * yet-to-be defined meaning.
       *
       * \todo Define meaning of nonzero return values.
       * \todo We should probably also have a function which returns a set of
       * signatures it can do.
       */
      virtual int checkCompatibility(const SignatureSet& sigSet) const = 0;

      /// Returns the name of the tree-level matrix element.
      virtual const std::string name() const = 0;

   protected:
      ITreeME() = default;
      ITreeME(const ITreeME&) = default;
      ITreeME(ITreeME&&) = default;
      ITreeME& operator=(const ITreeME&) = default;
      ITreeME& operator=(ITreeME&&) = default;

   private:
      /**
       * \brief Returns the squared tree-level matrix element for a Configuration.
       * \param config The Configuration to act on.
       * \return The result for the squared matrix element.
       *
       * This is a pure virtual hook function which is called by operator() to
       * do the actual computation and has to be implemented by derived classes.
       * The returned value must exclude all overall powers of the coupling
       * constants.
       */
      virtual double value(const Configuration& config) = 0;
};

} // namespace Geneva

#endif // GENEVA_I_TREE_ME_HPP
