//------------------------------------------------------------------------------
/// \file ITreeMEColor.hpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2013 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class ITreeMEColor
//------------------------------------------------------------------------------

#ifndef GENEVA_I_TREE_ME_COLOR_HPP
#define GENEVA_I_TREE_ME_COLOR_HPP

#include "Geneva/Components/ITreeME.hpp"

#include "Geneva/Core/Physics/Configuration.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class ITreeMEColor
 * \ingroup TreeME
 *
 * \brief Abstract interface to tree-level matrix elements including color information.
 *
 * Given a Configuration config, one first calls the calculate(config) function,
 * which evaluates all required pieces. After that, one can extract the required
 * pieces through:
 *
 * value() : Returns the squared tree level matrix element
 * valueLargeNc(cf) : Returns the squared amplitude for color flow cf
 * colorIndixes(i,cf) : Returns the color-anticolor indices pair for
 *                      Particle i and color flow cf
 * numColorFlow() : Returns the number of color flows
 *
 * By convention the value returned is the standard squared
 * invariant matrix-element but *without* including the overall powers of
 * coupling constants. That is, we write the tree-level matrix element squared
 * for N final state particles as
 *
 * \f$|M_N|^2 = (alpha_i)^N |M_N^0|^2\f$
 *
 * where \f$(\alpha_i)^N\f$ is the appropriate combination of \f$\alpha_\text{em}\f$ and \f$\alpha_\text{strong}\f$ and
 * the derived classes implement \f$|H_N^0|^2\f$.
 */
//------------------------------------------------------------------------------
class ITreeMEColor : public ITreeME
{
   public:
      /**
       * \brief Performs all necessary calculations for a Configuration.
       * \param config The Configuration to use.
       *
       * Performs all necessary calculations to obtain the square of the tree-
       * level amplitude including the color subamplitudes for \p config.
       */
      virtual void calculateValue(const Configuration& config) = 0;

      /**
       * \brief Performs all necessary calculations for a Configuration.
       * \param config The Configuration to use.
       *
       * Performs all necessary calculations to obtain the square of the tree-
       * level amplitude including the color subamplitudes for \p config.
       */
      virtual void calculateColorInfo(const Configuration& config) = 0;

      /**
       * \brief Returns the result for the squared tree-level matrix element.
       *
       * Returns the result for the squared amplitude for the Configuration
       * given in the last call to \c calculate(). The returned value excludes
       * all overall powers of the coupling constants.
       */
      virtual double value() const = 0;

      /**
       * \brief Returns the contribution to the squared tree-level matrix element from a single color flow.
       * \param cf Index of the color flow to use.
       *
       * Returns the result for the squared amplitude for the Configuration
       * given in the last call to \c calculate() from the color flow \p cf.
       * The returned value excludes all overall powers of the coupling
       * constants.
       */
      typedef size_t colorFlowIndex;
      virtual double valueLargeNc(colorFlowIndex cf) const = 0;

      /**
       * \brief Returns the color indices for a particle and a color flow.
       * \param i Iterator to the particle in the Configuration.
       * \param cf Index of the color flow to use.
       * \return A pair of color-anticolor labels.
       *
       * The iterator \p i must be an iterator into the Configuration given in
       * the last call to \c calculate() and specifies the particle for which
       * to obtain the color-anticolor labels for the color flow \p cf.
       */
      virtual std::pair<int, int> colorIndices(Configuration::const_iterator i, colorFlowIndex cf) const = 0;

      /**
       * \brief Returns the number of possible color flows.
       *
       * Returns the number of color flows for the Configuration given in the
       * last call to \c calculate().
       */
      virtual size_t numColorFlows() const = 0;

   private:
      /**
       * \brief Returns the squared tree-level matrix element for a Configuration.
       * \param config The Configuration to act on.
       * \return The result for the squared matrix element.
       *
       * Default implementation to export the functionality to the ITreeME
       * interface.
       */
      virtual double value(const Configuration& config)
      {
         calculateValue(config);
         return value();
      };
};

} // namespace Geneva

#endif // GENEVA_I_TREE_ME_COLOR_HPP
