//------------------------------------------------------------------------------
/// \file ITreeMEComplete.hpp
//
// Author(s):
//    Christian Bauer, Chris Vermilion, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2013 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class ITreeMEComplete
//------------------------------------------------------------------------------

#ifndef GENEVA_I_TREE_ME_COMPLETE_HPP
#define GENEVA_I_TREE_ME_COMPLETE_HPP

#include "Geneva/Components/ITreeME.hpp"

#include "Geneva/Core/Physics/Configuration.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class ITreeMEComplete
 * \ingroup TreeME
 *
 * \brief Abstract interface to tree-level matrix elements including complete color and spin correlations.
 *
 * Given a Configuration config, one first calls the calculate(config) function,
 * which evaluates all required pieces. After that, one can extract the required
 * pieces through:
 *
 * value() : Returns the squared tree level matrix element
 * colorMatrix(i,j) : Returns the color correlated squared amplitude
 * polarizationMatrix(mu,nu) : Returns the polarization correlated squared amplitude
 *
 * By convention the value returned is the standard squared
 * invariant matrix-element but *without* including the overall powers of
 * coupling constants. That is, we write the tree-level matrix element squared
 * for N final state particles as
 *
 * \f$|M_N|^2 = (alpha_i)^N |M_N^0|^2\f$
 *
 * where \f$(\alpha_i)^N\f$ is the appropriate combination of \f$\alpha_\text{em}\f$ and \f$\alpha_\text{strong}\f$ and
 * the derived classes implement \f$|H_N^0|^2\f$.
 */
//------------------------------------------------------------------------------
class ITreeMEComplete : public ITreeME
{
   public:
      /**
       * \brief Performs all necessary calculations for a Configuration.
       * \param config The Configuration to use.
       *
       * Performs all necessary calculations to obtain the square of the tree-
       * level amplitude, the color-correlated amplitude, and the polarization-
       * correlated amplitude for \p config.
       */
      virtual void calculate (const Configuration& config) = 0;

      /**
       * \brief Returns the result for the squared tree-level matrix element.
       *
       * Returns the result for the squared amplitude for the Configuration
       * given in the last call to \c calculate(). The returned value excludes
       * all overall powers of the coupling constants.
       */
      virtual double value() const = 0;

      /**
       * \brief Returns the result for an entry in the color matrix.
       * \param i Iterator to the 1st particle in the Configuration.
       * \param j Iterator to the 2nd particle in the Configuration.
       *
       * Returns the result for color-correlated squared amplitude for the
       * Configuration given in the last call to \c calculate(). The iterators
       * \p i and \p j must be iterators into that Configuration and specify the
       * particles for which to evaluate the color correlation.
       *
       * The returned value excludes all overall powers of the coupling
       * constants.
       */
      virtual double colorMatrix(Configuration::const_iterator i, Configuration::const_iterator j) const = 0;


      /**
       * \brief Returns the result for an entry in the polarization matrix.
       * \param i Iterator to the particle in the Configuration.
       * \param mu The 1st polarization index.
       * \param nu The 2nd polarization index.
       *
       * Returns the result for the spin-correlated squared amplitude for the
       * Configuration given in the last call to \c calculate(). The iterator
       * \p i must be an iterator into that Configuration and specifies for
       * which particle to evaluate the polarization matrix.
       *
       * The returned value excludes all overall powers of the coupling
       * constants.
       *
       * \note If \p i is not a gluon, the matrix in mu nu is zero.
       */
      typedef size_t lorentzIndex;
      virtual double polarizationMatrix(Configuration::const_iterator i, lorentzIndex mu, lorentzIndex nu) const = 0;

   private:
      /**
       * \brief Returns the squared tree-level matrix element for a Configuration.
       * \param config The Configuration to act on.
       * \return The result for the squared matrix element.
       *
       * Default implementation to export the functionality to the ITreeME
       * interface.
       */
      virtual double value(const Configuration& config)
      {
         calculate(config);
         return value();
      };
};

} // namespace Geneva

#endif // GENEVA_I_TREE_ME_COMPLETE_HPP
