//------------------------------------------------------------------------------
/// \file LHAPDF.hpp
//
// Author(s):
//    Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class LHAPDF
//------------------------------------------------------------------------------

#ifndef GENEVA_LHAPDF_HPP
#define GENEVA_LHAPDF_HPP

#include "Geneva/Components/IPDF.hpp"

#include "Geneva/Core/Physics/PdgID.hpp"

#include <map>
#include <vector>

namespace LHAPDF
{
class PDF;
}

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class LHAPDF
 * \ingroup PDF
 *
 * \brief Implementation of IPDF using the LHAPDF library that automatically caches its recent results.
 */
//------------------------------------------------------------------------------
class LHAPDF : public IPDF
{
   private:
      PdgID _beamA;            ///< id of beam A
      int _memberA;            ///< member of PDG set for beam A (A, B can be same or different)
      ::LHAPDF::PDF* _pdfA;    ///< PDF handler for beam A
      std::string _nameA;      ///< name of PDF A

      PdgID _beamB;            ///< id of beam B
      int _memberB;            ///< member of PDG set for beam B (A, B can be same or different)
      ::LHAPDF::PDF* _pdfB;    ///< PDF handler for beam B (may point to same object as A!)
      std::string _nameB;      ///< name of PDF A

      const size_t _maxStackLength; ///< Maximum number of PDF values to cache (0 for no cache)
      bool _independentFlavourEvolution; ///< If true, singular flavours are evolved independently and no caching is used. Available from LHAPDF6.

   public:
      /// Arguments are passed to LHAPDF::initPDFSet()
      LHAPDF(PdgID beamA, PdgID beamB, const std::string& name, int member = 0, size_t maxStackLength = 0);
      LHAPDF(PdgID beamA, PdgID beamB, const std::string& nameA, const std::string& nameB, int memberA,  int memberB, size_t maxStackLength);

      LHAPDF(const LHAPDF&) = delete;
      LHAPDF(LHAPDF&&) = delete;
      LHAPDF& operator=(const LHAPDF&) = delete;
      LHAPDF& operator=(LHAPDF&&) = delete;
      ~LHAPDF();

      /**
       * \brief Returns the PDF for beam A/B for the given particle \a id, \a x, and \a muF.
       * \return the PDF value
       */
      virtual double pdfA(PdgID id, double x, double muF) const;
      virtual double pdfB(PdgID id, double x, double muF) const;

      /**
       * \brief Returns the PdgID for beam particle A/B.
       */
      virtual PdgID beamParticleA() const { return _beamA; };
      virtual PdgID beamParticleB() const { return _beamB; };

      /**
       * \brief Returns the name of the PDF for beam A/B.
       *
       * Corresponds to the usual name of the lhapdf set.
       */
      virtual std::string nameA() const { return _nameA; }
      virtual std::string nameB() const { return _nameB; }

      /**
       * \brief Returns the member number for PDF set for A/B.
       *
       * Corresponds to the usual member number of the lhapdf set with 0 being
       * the central member.
       */
      virtual int memberA() const { return _memberA; }
      virtual int memberB() const { return _memberB; }

      /**
       * \brief Returns a unique identifier for the PDF for beam A/B.
       *
       * The return value corresponds to the global lhapdf ID, which encodes
       * both PDF set and set member.
       */
      virtual int pdfIDA() const;
      virtual int pdfIDB() const;

      /// Returns the number of flavours used.
      virtual int getNf() const;

      /// Returns the name of the component.
      virtual std::string name() const;

      /// Returns \f$ \alpha_\mathrm{s} \f$ used by the central member
      double getAlphasQ(double Q) const;

      /// Returns the value of mZ used in the PDF
      double getmZ() const;

      /// Whether the PDF is enforced to be positive definite
      virtual bool isPositive() const;

   private:
      /**
       * \brief Handle internally the PDFs for initial beam A/B for given \a x and \a muF
       * \return map of PDF values for all parton flavors i = -6..6, with 0=gluon
       */
      std::map<int, double> pdfA(double x, double muF) const;
      std::map<int, double> pdfB(double x, double muF) const;

      /**
       * \brief Handle internally the transformations needed for non-proton beams
       * \param part the PdgID of the parton
       * \param beam the PdgID of the particle in the beam
       * \return the PdgID with which the proton PDF should be called to fake a non-proton beam
       */
      PdgID getFakePartonId(const PdgID part, const PdgID idBeam) const;
};

} // namespace Geneva

#endif // GENEVA_LHAPDF_HPP
