//------------------------------------------------------------------------------
/// \file NloCoefficients.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2014
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class NloCoefficients
//------------------------------------------------------------------------------

#ifndef GENEVA_NLO_COEFFS_HPP
#define GENEVA_NLO_COEFFS_HPP

#include "Geneva/Core/Physics/Sigma.hpp"
#include "Geneva/Core/UI/Log.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class NloCoefficients
 * \ingroup comp
 *
 * \brief Structure containing separate NLO coefficients.
 *
 * \todo document me!
 */
//------------------------------------------------------------------------------
class NloCoefficients
{
   private:
      Sigma _lo;
      Sigma _nloBornKin;
      Sigma _nloRealKin;

   public:
      /// \name Constructors etc.
      ///@{

      /// Creates a default empty %NloCoefficients.
      NloCoefficients() {};
      /// Creates a %NloCoefficients from given input.
      NloCoefficients(const Sigma& lo, const Sigma& nloB, const Sigma& nloR);

      ///@}

      /// \name Read and Write Access
      ///@{
      Sigma lo() const;                 ///< Returns leading order part.
      Sigma nloBornKin() const;         ///< Returns next-to-leading order part with Born-like kinematics.
      Sigma nloRealKin() const;         ///< Returns next-to-leading order part with real-like kinematics.
      Sigma nlo() const;                ///< Returns sum of next-to-leading parts.
      Sigma sum() const;                ///< Returns sum of leading and next-to-leading order parts.

      double loCentral() const;            ///< Returns central value of leading order part.
      double nloBornKinCentral() const;    ///< Returns central value of next-to-leading order part with Born-like kinematics.
      double nloRealKinCentral() const;    ///< Returns central value next-to-leading order part with real-like kinematics.
      double nloCentral() const;           ///< Returns central value of next-to-leading part.
      double sumCentral() const;           ///< Returns sum of central leading and next-to-leading order parts.

      double loVary(size_t n) const;         ///< Returns n-th variation for leading order part.
      double nloBornKinVary(size_t n) const; ///< Returns n-th variation for next-to-leading order part with Born-like kinematics.
      double nloRealKinVary(size_t n) const; ///< Returns n-th variation next-to-leading order part with real-like kinematics.
      double nloVary(size_t n) const;        ///< Returns n-th variation for next-to-leading part.
      double sumVary(size_t n) const;        ///< Returns sum of n-th leading and next-to-leading order variation.

      Sigma& lo();                      ///< Access leading order part.
      Sigma& nloBornKin();              ///< Access next-to-leading order part with Born-like kinematics.
      Sigma& nloRealKin();              ///< Access next-to-leading order part with real-like kinematics.
      ///@}

      /// \name Utility functions
      ///@{
      void clear();
      bool isDefined();
      ///@}
};

////////////////////////////////////////////////////////////////////////////////
// Inline Declarations
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------
// Creates a NloCoefficients from the given arguments.
inline NloCoefficients::NloCoefficients(const Sigma& lo, const Sigma& nloB, const Sigma& nloR)
   : _lo(lo), _nloBornKin(nloB), _nloRealKin(nloR)
{}

//------------------------------------------------------------------------------
// Read and Write Access
//------------------------------------------------------------------------------
inline Sigma NloCoefficients::lo() const
{
   return _lo;
}

//------------------------------------------------------------------------------
inline Sigma NloCoefficients::nloBornKin() const
{
   return _nloBornKin;
}

//------------------------------------------------------------------------------
inline Sigma NloCoefficients::nloRealKin() const
{
   return _nloRealKin;
}

//------------------------------------------------------------------------------
inline Sigma NloCoefficients::nlo() const
{
   if (_nloBornKin.isDefined() and _nloRealKin.isDefined())
      return _nloBornKin + _nloRealKin;
   else
      Log::error("NloCoefficients::nlo") << "Trying to combine one or more undefined Sigma" << Log::end;
   return Sigma();
}

//------------------------------------------------------------------------------
inline Sigma NloCoefficients::sum() const
{
   Sigma tmp = nlo();
   if (_lo.isDefined() and tmp.isDefined())
      return _lo + tmp;
   else
      Log::error("NloCoefficients::sum") << "Trying to combine one or more undefined Sigma" << Log::end;
   return Sigma();
}


//------------------------------------------------------------------------------
inline double NloCoefficients::loCentral() const
{
   return _lo.central();
}

//------------------------------------------------------------------------------
inline double NloCoefficients::loVary(size_t i) const
{
   return _lo.variations()[i];
}

//------------------------------------------------------------------------------
inline double NloCoefficients::nloCentral() const
{
   return _nloBornKin.central() + _nloRealKin.central();
}

//------------------------------------------------------------------------------
inline double NloCoefficients::nloVary(size_t i) const
{
   return _nloBornKin.variations()[i] + _nloRealKin.variations()[i];
}

//------------------------------------------------------------------------------
inline double NloCoefficients::sumCentral() const
{
   return _lo.central() + nloCentral();
}

//------------------------------------------------------------------------------
inline double NloCoefficients::sumVary(size_t i) const
{
   return _lo.variations()[i] + nloVary(i);
}

//------------------------------------------------------------------------------
inline Sigma& NloCoefficients::lo()
{
   return _lo;
}

//------------------------------------------------------------------------------
inline Sigma& NloCoefficients::nloBornKin()
{
   return _nloBornKin;
}

//------------------------------------------------------------------------------
inline Sigma& NloCoefficients::nloRealKin()
{
   return _nloRealKin;
}

//------------------------------------------------------------------------------
inline void NloCoefficients::clear()
{
   _lo.clear();
   _nloBornKin.clear();
   _nloRealKin.clear();
}

//------------------------------------------------------------------------------
inline bool NloCoefficients::isDefined()
{
   return _lo.isDefined() and _nloBornKin.isDefined() and _nloRealKin.isDefined();
}

} // namespace Geneva

#endif // GENEVA_NLO_COEFFS_HPP
