//------------------------------------------------------------------------------
/// \file RegionInfo.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definition of struct RegionInfo
//------------------------------------------------------------------------------

#ifndef GENEVA_REGION_INFO_HPP
#define GENEVA_REGION_INFO_HPP

#include "Geneva/Core/Physics/Signature.hpp"
#include "Geneva/Core/Physics/SignatureSet.hpp"
#include "Geneva/Core/Physics/Configuration.hpp"

#include <iosfwd>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief %RegionInfo contains information about the singular region mapping.
 *
 *  %RegionInfo contains informations about the singular region mapping. In case
 * there are no singularities, the underlying Born configuration is missing,
 * the iterator fN is automatically set to end()
 */
//------------------------------------------------------------------------------
class RegionInfo
{
   public:
      SignatureSet::iterator fN;    ///< The iterator to the n-body %Signature
      SignatureSet::iterator fNp1;  ///< The iterator n+1-body %Signature
      Signature::const_iterator i;        ///< The iterator to FKS parton in Phi_{N+1}
      Signature::const_iterator j;        ///< The iterator to the sister parton in Phi_{N+1}
      Signature::const_iterator k;        ///< The iterator to the emitter in Phi_N
      size_t multN;                ///< Multiplicity factor from identical emitters
      size_t multRad;              ///< MUltiplicity factor from identical maps
      int nLight;

      /// Check if it has a singularity associated to it
      bool hasInitialStateSingularity() const;
      bool hasFinalStateSingularity() const;

      bool operator== (const RegionInfo& rhs) const;

      Configuration::iterator getEmitter(const Configuration& config) const;

      /// RegionInfo stream insertion operator.
      friend std::ostream& operator<< (std::ostream& out, const RegionInfo& f);

   private:
      bool isLightQuark(PdgID id) const;
      /// Streams the %RegionInfo to \a output.
      std::ostream& print(std::ostream& output) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Equal operator
//------------------------------------------------------------------------------
/**
 * \brief %RegionInfo equlaity operator.
 * \param rhs The %RegionInfo that will be compared to.
 * \return A bool
 *
 * Compares two RegionInfo's to one another
 */
inline bool RegionInfo::operator== (const RegionInfo& rhs) const
{
   return (fN == rhs.fN)
      && (fNp1 == rhs.fNp1)
      && (i == rhs.i)
      && (j == rhs.j)
      && (k == rhs.k)
      && (multN == rhs.multN)
      && (multRad == rhs.multRad)
      && (nLight == rhs.nLight);
}

//------------------------------------------------------------------------------
inline bool RegionInfo::hasInitialStateSingularity() const
{
   if (k->isInitialA() || k->isInitialB()) {
      if (i->isGluon())
         return true;
      else if (isLightQuark(*i) && (i->isAntiPartner(*k) || *i == *j))
         return true;
      else
         return false;
   }
   return false;
}

//------------------------------------------------------------------------------
inline bool RegionInfo::hasFinalStateSingularity() const
{
   if (k->isFinal()) {
      if (i->isGluon())
         return true;
      else if (isLightQuark(*i) && (i->isAntiPartner(*j) || *i == *k))
         return true;
      else
         return false;
   }
   return false;
}

//------------------------------------------------------------------------------
inline std::ostream& operator<< (std::ostream& out, const RegionInfo& f)
{
   return f.print(out);
}


//------------------------------------------------------------------------------
inline bool RegionInfo::isLightQuark(PdgID id) const
{
   int tmp = abs(id.code());
   return tmp <= nLight && tmp != 0;
}

} // namespace Geneva

#endif // GENEVA_REGION_INFO_HPP
