//------------------------------------------------------------------------------
/// \file RegionInfoList.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definition of class RegionInfoList
//------------------------------------------------------------------------------

#ifndef GENEVA_REGION_INFO_LIST_HPP
#define GENEVA_REGION_INFO_LIST_HPP

#include "Geneva/Core/Physics/SignatureSet.hpp"

#include "Geneva/Components/RegionInfo.hpp"

#include <list>
#include <map>

/**
 * \brief %RegionsInfoList  contains information about the region mappings.
 *
 *  %RegionInfoList contains informations about the region mappings.
 */

namespace Geneva
{

class RegionInfoList
{
   public:
      typedef std::list<RegionInfo>::const_iterator  region_list_iterator;
      typedef std::multimap<Signature, RegionInfo>::const_iterator  singular_map_iterator;

      /// Sets RegionInfoList
      void setRegionInfoList(const SignatureSet& sigN, const SignatureSet& sigNplusOne, int nLight, bool verbose = false);

      /// Evaluate the multiplicity factors of the mapping from a N-body %Signature to a N+1-body %Signature
      size_t multiplicityFactor(const Signature& sigNp1, Signature::const_iterator sister, Signature::const_iterator fks) const;

      /// Get the list of singular regions that can be constructed from a given %Signature
      std::list<RegionInfo> getSingularRegionsFromSignature(const Signature& sigN) const;

      /// Get the list of singular regions that results in a  given %Signature
      std::list<RegionInfo> getSingularRegionsToSignature(const Signature& sigNplus1) const;

      /// Get the list of singular regions that results in a  given %Signature
      std::list<RegionInfo> getSingularRegionsBetweenSignatures(const Signature& sigN, const Signature& sigNplus1) const;

      /// Get the list of all regions
      std::list<RegionInfo> getRegionInfoList() const;

      /// Get the list of all ISR regions
      std::list<RegionInfo> getIsrRegionInfoList() const;

      /// Get the list of all FSR regions
      std::list<RegionInfo> getFsrRegionInfoList() const;

      bool isInitialized() const;
      void clear();
      region_list_iterator begin() const;
      region_list_iterator end() const;

      // RegionInfoList stream insertion operator.
      friend std::ostream& operator<<(std::ostream& out, const RegionInfoList& info);

   private:
      std::list<RegionInfo> _allRegions;
      SignatureSet _signaturesN;
      SignatureSet _signaturesNp1;
      SignatureSet _singularSignatures;
      SignatureSet _regularSignatures;
      std::multimap<Signature, RegionInfo> _singularRegionFromSigMap;
      std::multimap<Signature, RegionInfo> _singularRegionToSigMap;
      int _nLight = 0;
      bool _initialized = false;

      /// Streams the %RegionInfoList to \a output.
      std::ostream& print(std::ostream& output) const;
};

////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %RegionInfoList clear .
 */
inline void RegionInfoList::clear()
{
   _allRegions.clear();
   _signaturesN.clear();
   _signaturesNp1.clear();
   _singularSignatures.clear();
   _regularSignatures.clear();
   _singularRegionFromSigMap.clear();
   _singularRegionToSigMap.clear();
   _initialized = false;
   _nLight = 0;
}

//------------------------------------------------------------------------------
/**
 * \brief %RegionInfoList access operator.
 * \return A list of all %RegionInfo.
 */
inline std::list<RegionInfo> RegionInfoList::getRegionInfoList() const {
   return _allRegions;
}

//------------------------------------------------------------------------------
/**
 * \brief %RegionInfoList access operator.
 * \return A list of all %RegionInfo with a ISR splitting.
 */
inline std::list<RegionInfo> RegionInfoList::getIsrRegionInfoList() const
{
   std::list<RegionInfo> tmp;
   for (RegionInfoList::region_list_iterator iter = _allRegions.begin(); iter != _allRegions.end(); ++iter) {
     if (iter->hasInitialStateSingularity()) tmp.push_back(*iter);
   }
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %RegionInfoList access operator.
 * \return A list of all %RegionInfo with a FSR splitting.
 */
inline std::list<RegionInfo> RegionInfoList::getFsrRegionInfoList() const
{
   std::list<RegionInfo> tmp;
   for (RegionInfoList::region_list_iterator iter = _allRegions.begin(); iter != _allRegions.end(); ++iter) {
      if (iter->hasFinalStateSingularity()) tmp.push_back(*iter);
   }
   return tmp;
}


//------------------------------------------------------------------------------
/**
 * \brief Returns the list of valid %RegionInfo for the given %Signature.
 * \param sig The %Signature being inspected.
 * \return A list of all %RegionInfo which are valid for the given %Signature.
 *
 * This returns all the %RegionInfo that share the same N-Body %Signature
 */
inline std::list<RegionInfo> RegionInfoList::getSingularRegionsFromSignature(const Signature& sig)  const {

   std::list<RegionInfo> tmp;
   std::pair<RegionInfoList::singular_map_iterator, RegionInfoList::singular_map_iterator> it_pair = _singularRegionFromSigMap.equal_range(sig);
   RegionInfoList::singular_map_iterator iter;
   for (iter = it_pair.first; iter != it_pair.second; ++iter) {
      tmp.push_back(iter->second);
   }
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the list of valid %RegionInfo for the given %Signature.
 * \param sig The %Signature being inspected.
 * \return A list of all %RegionInfo which results in the the given %Signature.
 *
 * This returns all the %RegionInfo that give rise to the same the N+1-Body %Signature
 */
inline std::list<RegionInfo> RegionInfoList::getSingularRegionsToSignature(const Signature& sig)  const {

   std::list<RegionInfo> tmp;
   std::pair<RegionInfoList::singular_map_iterator, RegionInfoList::singular_map_iterator> it_pair = _singularRegionToSigMap.equal_range(sig);
   RegionInfoList::singular_map_iterator iter;
   for (iter = it_pair.first; iter != it_pair.second; ++iter) {
      tmp.push_back(iter->second);
   }
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the list of valid %RegionInfo for given %SignatureN and %SignatureNplus1.
 * \param sigN The n-body %Signature being inspected.
 * \param sig The (n+1)-body %Signature being inspected.
 * \return A list of all %RegionInfo which map between sigN and sigNplus1.
 *
 * This returns all the %RegionInfo that give rise to the same the N+1-Body %Signature given an N-Body %Signature
 */
inline std::list<RegionInfo> RegionInfoList::getSingularRegionsBetweenSignatures(const Signature& sigN, const Signature& sigNplus1)  const {

   std::list<RegionInfo> tmp;
   std::pair<RegionInfoList::singular_map_iterator, RegionInfoList::singular_map_iterator> it_pair_ToSig = _singularRegionToSigMap.equal_range(sigNplus1);
   std::pair<RegionInfoList::singular_map_iterator, RegionInfoList::singular_map_iterator> it_pair_FromSig = _singularRegionFromSigMap.equal_range(sigN);
   RegionInfoList::singular_map_iterator iterToSig, iterFromSig;
   for (iterToSig = it_pair_ToSig.first; iterToSig != it_pair_ToSig.second; ++iterToSig) {
      for (iterFromSig = it_pair_FromSig.first; iterFromSig != it_pair_FromSig.second; ++iterFromSig) {
         if (iterToSig->second == iterFromSig->second)
            tmp.push_back(iterFromSig->second);
      }
   }
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns whether RegionInfoList has been correctly initialized.
 * \return \a true if a setRegionInfoList has been called
 * and no errors occured, \a false otherwise
 */
inline bool RegionInfoList::isInitialized() const
{
   return _initialized;
}


//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator pointing to the first Signature in the %RegionInfoList.
 *
 * If the %RegionInfoList is empty the returned iterator will be equal to end(). It
 * is the user's responsibility to ensure that the iterator is not used for
 * reading in this case.
 */
inline RegionInfoList::region_list_iterator RegionInfoList::begin() const
{
   return _allRegions.begin();
}

//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator pointing beyond the last RegionInfo in the %RegionInfoList.
 *
 * The returned iterator points one step beyond the last RegionInfo and so should not be
 * used for reading. Instead, it is usually used to check that all RegionInfos have been
 * traversed.
 */
inline RegionInfoList::region_list_iterator RegionInfoList::end() const
{
   return _allRegions.end();
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------
/**
 * \brief %RegionInfoList stream insertion operator.
 * \param out The output stream to be used.
 * \param sig The %RegionInfoList to be streamed.
 * \return A reference to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const RegionInfoList& info)
{
   return info.print(out);
}

} // namespace Geneva

#endif // GENEVA_REGION_INFO_LIST_HPP
