//------------------------------------------------------------------------------
/// \file Scales.hpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2016 CERN, DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definitions of Scale classes
//------------------------------------------------------------------------------

#ifndef GENEVA_SCALES_HPP
#define GENEVA_SCALES_HPP

#include "Geneva/Components/IScale.hpp"

#include "Geneva/Core/Physics/Configuration.hpp"

#include <string>
#include <sstream>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class ScaleFixed
 * \ingroup comp
 *
 * \brief A fixed scale.
 *
 * Returns a fixed scale value times an optional prefactor.
 */
//------------------------------------------------------------------------------
class ScaleFixed : public IScale
{
   private:
      double _scale;
      double _prefactor;

   public:
      ScaleFixed(double scale, double prefactor = 1.0)
         : _scale(scale), _prefactor(prefactor)
      {}

      virtual std::string name() const;

   private:
      /// Returns prefactor * fixed scale independent of \a config.
      virtual double value(const Configuration& config) const;
};


//------------------------------------------------------------------------------
/**
 * \class ScaleHT
 * \ingroup comp
 *
 * \brief Scalar sum of transverse momenta of all particles satisfying a given condition.
 *
 * Returns the scalar sum of the transverse momenta of all final-state particles
 * satisfying a fixed UnaryCondition specified as template argument times an
 * optional prefactor.
 */
//------------------------------------------------------------------------------
template<typename UnaryCondition>
class ScaleHT : public IScale
{
   private:
      double _prefactor;

   public:
      ScaleHT(double prefactor = 1.0) : _prefactor(prefactor) {}

      virtual std::string name() const;

   private:
      /// Returns the scalar sum of |pT|.
      virtual double value(const Configuration& config) const;
};


//------------------------------------------------------------------------------
/**
 * \class ScaleInvariantMass
 * \ingroup comp
 *
 * \brief Total invariant mass of all particles satisfying a given condition.
 *
 * Returns the total invariant mass of all final-state particles satisfying a
 * fixed UnaryCondition specified as template argument times an optional
 * prefactor.
 */
//------------------------------------------------------------------------------
template<typename UnaryCondition>
class ScaleInvariantMass : public IScale
{
   private:
      double _prefactor;

   public:
      ScaleInvariantMass(double prefactor = 1.0) : _prefactor(prefactor) {}

      virtual std::string name() const;

   private:
      /// Returns the total invariant mass.
      virtual double value(const Configuration& config) const;
};


//------------------------------------------------------------------------------
/**
 * \class ScaleTransverseMass
 * \ingroup comp
 *
 * \brief Total transverse mass of all particles satisfying a given condition.
 *
 * Returns the total transverse mass of all final-state particles satisfying a
 * fixed UnaryCondition specified as template argument times an optional
 * prefactor.
 */
//------------------------------------------------------------------------------
template<typename UnaryCondition>
class ScaleTransverseMass : public IScale
{
   private:
      double _prefactor;

   public:
      ScaleTransverseMass(double prefactor = 1.0) : _prefactor(prefactor) {}

      virtual std::string name() const;

   private:
      /// Returns the total transverse mass.
      virtual double value(const Configuration& config) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// ScaleFixed
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Returns prefactor * fixed scale independent of \a config.
inline double ScaleFixed::value(const Configuration& config) const
{
   return _prefactor * _scale;
}

//------------------------------------------------------------------------------
inline std::string ScaleFixed::name() const
{
   std::stringstream s;
   s << "ScaleFixed (" << _prefactor*_scale << " GeV)";
   return s.str();
}

//------------------------------------------------------------------------------
// ScaleHT
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Returns the scalar sum of |pT|.
template<typename UnaryCondition>
inline double ScaleHT<UnaryCondition>::value(const Configuration& config) const
{
   double sum = 0.;
   for (Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it) {
      if (UnaryCondition()(*it))
         sum += it->pvec().perp();
   }
      return _prefactor * sum;
}

//------------------------------------------------------------------------------
template<typename UnaryCondition>
inline std::string ScaleHT<UnaryCondition>::name() const
{
   std::stringstream s;
   s << "ScaleHT";
   if (_prefactor != 1.0)
      s << "(* " << _prefactor << ")";
   return s.str();
}

//------------------------------------------------------------------------------
// ScaleInvariantMass
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Returns the total invariant mass.
template<typename UnaryCondition>
double ScaleInvariantMass<UnaryCondition>::value(const Configuration& config) const
{
   FourVector sum;
   for (Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it) {
      if (UnaryCondition()(*it))
         sum += it->p();
   }
      return _prefactor * sum.magnitude();
}

//------------------------------------------------------------------------------
template<typename UnaryCondition>
std::string ScaleInvariantMass<UnaryCondition>::name() const
{
   std::stringstream s;
   s << "ScaleInvariantMass";
   if (_prefactor != 1.0)
      s << "(* " << _prefactor << ")";
   return s.str();
}

//------------------------------------------------------------------------------
// ScaleTransverseMass
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Returns the total transverse mass.
template<typename UnaryCondition>
inline double ScaleTransverseMass<UnaryCondition>::value(const Configuration& config) const
{
   FourVector sum;
   for (Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it) {
      if (UnaryCondition()(*it))
         sum += it->p();
   }
   return _prefactor * sqrt(sum.plus() * sum.minus());
}

//------------------------------------------------------------------------------
template<typename UnaryCondition>
inline std::string ScaleTransverseMass<UnaryCondition>::name() const
{
   std::stringstream s;
   s << "ScaleTransverseMass";
   if (_prefactor != 1.0)
      s << "(* " << _prefactor << ")";
   return s.str();
}

} // namespace Geneva

#endif // GENEVA_SCALES_HPP
