//------------------------------------------------------------------------------
/// \file SignatureManager.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SignatureManager
//------------------------------------------------------------------------------

#ifndef GENEVA_SIGNATURE_MANAGER_HPP
#define GENEVA_SIGNATURE_MANAGER_HPP

#include <cstdlib>

namespace Geneva
{

class PdgID;
class Signature;
class SignatureSet;

//------------------------------------------------------------------------------
/**
 * \class SignatureManager
 * \ingroup comp
 *
 * \brief Provides SignatureSets for specific requirements.
 */
//------------------------------------------------------------------------------
class SignatureManager
{
   public:
      /// Generates SignatureSets for e+e- -> n partons.
      SignatureSet getSignatureseeJets(size_t numPartons, int numFlavors) const;

      /// Generates SignatureSets for pp -> W^(+/-) + n QCD partons
      /// If doDecay is false, W is not decayed
      SignatureSet getSignaturespp2WJets(PdgID lepton, size_t numPartons, int numFlavors, bool doDecay=true) const;

      /// Generates SignatureSets for pp -> W^+ + n QCD partons
      /// If doDecay is false, W is not decayed
      SignatureSet getSignaturespp2WPlusJets(PdgID lepton, size_t numPartons, int numFlavors, bool doDecay=true) const;

      /// Generates SignatureSets for pp -> W^- + n QCD partons
      /// If doDecay is false, W is not decayed
      SignatureSet getSignaturespp2WMinusJets(PdgID lepton, size_t numPartons, int numFlavors, bool doDecay=true) const;

      /// Generates SignatureSets for pp -> Z + n QCD partons
      /// If doDecay is false, Z is not decayed
      SignatureSet getSignaturespp2ZJets(PdgID lepton, size_t numPartons, int numFlavors, bool doDecay=true) const;

   private:
      /// Return the set of Signatures representing all possible single QCD emissions from \p sig
      SignatureSet QCDEmissions(const Signature& sig, int numFlavors) const;

      /// Return the set of Signatures representing all possible single QCD emissions from \p sigs
      SignatureSet QCDEmissions(const SignatureSet& sigs, int numFlavors) const;

      /// Return an "inverted" Signature -- with all IDs switched with their antiparticles
      Signature invertedSignature(const Signature& sig) const;

      /// Return an "inverted" SignatureSet -- same signatures but all IDs switched with their antiparticles
      SignatureSet invertedSignatures(const SignatureSet& sigs) const;

      /// Return a Signature with initialA <-> initialB
      Signature swapInitial(const Signature& sig) const;

      /// Return the set of Signatures for W+ + 2 partons
      SignatureSet getSignaturespp2WPlus2Jets(const PdgID& IDdecay, int numFlavors, bool doDecay=true) const;
};

} // namespace Geneva

#endif // GENEVA_SIGNATURE_MANAGER_HPP
