//------------------------------------------------------------------------------
/// \file SplittingFunction.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 <author's copyright holder>
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SplitingFunction
//------------------------------------------------------------------------------

#ifndef GENEVA_SPLITTING_FUNCTION_HPP
#define GENEVA_SPLITTING_FUNCTION_HPP

#include "Geneva/Components/ISplittingMapping.hpp"
#include "Geneva/Components/SplittingFunctionISR.hpp"
#include "Geneva/Components/SplittingFunctionFSR.hpp"
#include "Geneva/Components/RegionInfoList.hpp"

#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/Physics/ConfigurationExtensions.hpp"
#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"

namespace Geneva
{

class IPDF;

//------------------------------------------------------------------------------
/**
 * \class SplittingFunction
 * \ingroup Components
 *
 * \brief Implementation of a splitting function for phase space points with one QCD particle in the final state
 *
 * This implements a general function, together with information about bounding functions.
 */
//------------------------------------------------------------------------------
class SplittingFunction
{
   private:
      SplittingFunctionISR* _isrSplit;
      SplittingFunctionFSR* _fsrSplit;

   public:
      SplittingFunction(const OptionMap& optMap, ISplittingMapping* mapping, double Ecm, CouplingStrong* alphas, IPDF* pdf, double minX);

      SplittingFunction(const SplittingFunction&) = delete;
      SplittingFunction(SplittingFunction&&) = delete;
      SplittingFunction& operator=(const SplittingFunction&) = delete;
      SplittingFunction& operator=(SplittingFunction&&) = delete;

      ~SplittingFunction();

      /**
       * Implements the function p(PhiNplus1)
       * The z values of this function are distributed according to the splitting function f(z)
       * The phi values are distributed flat
       * It gives unity if integrated over all allowed values of z and phi, and summed over the allowed splittings
       *
       * \param PhiNplus1 The (N+1)-body phase space point
       *
       * This function does not require a (N+1)-body point that gives rise to a valid splitting.
       * If the (N+1)-body point is invalid, it returns zero
       */
      double operator()(const PhiNplus1& phiNplus1);

      //Project onto the Phi0 phase space
      bool projectToPhiN(const Configuration& configNPlus1, Configuration& configN);

      void printInfo() const;

      /**
       * Returns true if the phase space point satisfies all the theta
       * functions put on by the function normalized
       */
      bool isValid(const PhiNplus1& phiNplus1);

      //The Jacobian dPhi1 / (dPhi0 dTau0 dz dphi)
      double jacobian(const Configuration& configN, double TauN, double z, double phi, RegionInfo& map) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline SplittingFunction::SplittingFunction(const OptionMap& optMap, ISplittingMapping* mapping, double Ecm, CouplingStrong* alphas, IPDF* pdf,double minX) : _isrSplit(0), _fsrSplit(0)
{
   if (not mapping->getRegionInfoList().getIsrRegionInfoList().empty())
      _isrSplit = new SplittingFunctionISR(optMap, mapping, Ecm, alphas, pdf, minX);
   if (not mapping->getRegionInfoList().getFsrRegionInfoList().empty())
      _fsrSplit = new SplittingFunctionFSR(optMap, mapping, alphas, Ecm);
}

//------------------------------------------------------------------------------
inline SplittingFunction::~SplittingFunction()
{
   delete _isrSplit;
   delete _fsrSplit;
}

//------------------------------------------------------------------------------
inline double SplittingFunction::operator()(const PhiNplus1& phiNplus1)
{
   if (phiNplus1.sister()->isFinal())
      return _fsrSplit->normalized(phiNplus1);
   else
      return _isrSplit->normalized(phiNplus1);
}

//------------------------------------------------------------------------------
inline bool SplittingFunction::projectToPhiN(const Configuration& configNplus1, Configuration& configN)
{
   double TauN, z, phi;
   RegionInfo regionInfo;

   PhiNplus1 phiNplus1(configNplus1);

   //TODO: check whether we can use any other metric here
   NJettinessppRecursive(phiNplus1).cluster(phiNplus1.emitted(), phiNplus1.sister());

   if (phiNplus1.sister()->isFinal())
      return _fsrSplit->getMapping()->calcPhiNValues(phiNplus1, configN, TauN, z, phi, regionInfo);
   else
      return _isrSplit->getMapping()->calcPhiNValues(phiNplus1, configN, TauN, z, phi, regionInfo);
}

//------------------------------------------------------------------------------
inline bool SplittingFunction::isValid(const PhiNplus1& phiNplus1)
{
   //Calculate the ratio of the bounding integral ot the true integral statistically
   double TauN, z, phi;
   RegionInfo regionInfo;
   Configuration configN;
   ISplittingMapping* map = (phiNplus1.sister()->isFinal()) ? _fsrSplit->getMapping() :  _isrSplit->getMapping();

   // check that clustering is allowed by QCD flavor
   if (invalidFlavorProjection(map->sigsNmin1(), phiNplus1))
      return false;

   if (not map->calcPhiNValues(phiNplus1, configN, TauN, z, phi, regionInfo))
      return false;

   //If the resulting PhiN, TauN, z, phi values are not valid for the splitting, return 0
   if (!map->isValidPhiN(configN, TauN, z, phi, regionInfo.k))
      return false;

   return true;
}

//------------------------------------------------------------------------------
inline double SplittingFunction::jacobian(const Configuration& config, double TauN, double z, double phi, RegionInfo& map) const
{
   return map.hasFinalStateSingularity() ? _fsrSplit->getMapping()->jacobian(config, TauN, z, phi, map) : _isrSplit->getMapping()->jacobian(config, TauN, z, phi, map);
}

//------------------------------------------------------------------------------
inline void SplittingFunction::printInfo() const
{
   if (_isrSplit)
      _isrSplit->printInfo();
   if (_fsrSplit)
      _fsrSplit->printInfo();
}

} // namespace Geneva

#endif // GENEVA_SPLITTING_FUNCTION_HPP
