//------------------------------------------------------------------------------
/// \file SplittingFunctionBase.hpp
//
// Author(s):
//    Christian Bauer, Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 LBNL, CERN
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SplittingFunctionBase
//------------------------------------------------------------------------------

#ifndef GENEVA_SPLITTING_FUNCTION_BASE_HPP
#define GENEVA_SPLITTING_FUNCTION_BASE_HPP

#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/Utils/FileUtils.hpp"

#include <cstddef>
#include <vector>

namespace Geneva
{

class Configuration;
class RegionInfo;
class CouplingStrong;
class PhiNplus1;
class ISplittingMapping;

//------------------------------------------------------------------------------
/**
 * \class SplittingFunctionBase
 * \ingroup Components
 */
//------------------------------------------------------------------------------
class SplittingFunctionBase
{
   protected:
      const OptionMap& _optMap;
      double _cmEnergy;                ///< The value of the center of mass energy

      ISplittingMapping* _mapping;     ///< The phase space mapping for PhiN -> PhiNplus1

      CouplingStrong* _alphas;
      bool _useRunningAlphas;

      bool _isInitialized;             ///< A bool stating whether the initialization is complete

      std::vector<unsigned int> _passedZValues;
      double _zLower;
      double _zUpper;

      //Keep track of some statistics
      mutable size_t _numSplittingCalled;
      mutable size_t _numSplittingFailures;
      mutable size_t _numNormalizedCalled;
      mutable size_t _numRejectCalled;

   public:
      SplittingFunctionBase(const OptionMap& optMap, ISplittingMapping* mapping, CouplingStrong* alphas, double Ecm);

      virtual ~SplittingFunctionBase() = default;

      // Provides the default filename for the splitting function setup files.
      static std::string getSetupFilename(const OptionMap& options, const std::string& extraTag);

      /**
       * Implements the function p(PhiNplus1)
       * The z values of this function are distributed according to the splitting function f(z)
       * The phi values are distributed flat
       * It gives unity if integrated over all allowed values of z and phi, and summed over the allowed splittings
       *
       * \param PhiNplus1 The (N+1)-body phase space point
       *
       * This function does not require a (N+1)-body point that gives rise to a valid splitting.
       * If the (N+1)-body point is invalid, it returns zero
       */
      virtual double normalized(const PhiNplus1& phiNplus1) = 0;

      //Print information about the splitting function
      virtual void printInfo() const = 0;

      /**
       * Returns true if the phase space point satisfies all the theta
       * functions put on by the function normalized
       */
      bool isValid(const PhiNplus1& phiNplus1);

      ISplittingMapping* getMapping();

      double zMin(const Configuration& configN, double TauN, double phi, RegionInfo& regionInfo);
      double zMax(const Configuration& configN, double TauN, double phi, RegionInfo& regionInfo);

   private:
      virtual void initialize() = 0;

      /// Check if all necessary bounding functions exist
      virtual bool haveAllBoundingConstants() const = 0;

   protected:
      SplittingFunctionBase(const SplittingFunctionBase&) = default;
      SplittingFunctionBase(SplittingFunctionBase&&) = default;
      SplittingFunctionBase& operator=(const SplittingFunctionBase&) = default;
      SplittingFunctionBase& operator=(SplittingFunctionBase&&) = default;

      double jacobian(const Configuration& config, double TauN, double z, double phi, RegionInfo& map) const;

      ///< This initializes the failed z values vector
      void initializeFailedZValues(double zLower, double zUpper);

      ///< This adds a failed z value
      void addPassedZValue(double z);

      ///< This adjusts the zMax value based on the failed z values and the valid value of z
      double adjustedZMax(double z, double zMin, double zMax);
      ///< This adjusts the zMin value based on the failed z values and the valid value of z
      double adjustedZMin(double z, double zMin, double zMax);

      std::pair<double, double> adjustZMinZMax(double z, double zMin, double zMax, double phiMin, double phiMax, double TauN, Configuration& configN, RegionInfo& regionInfo, ISplittingMapping* mapping);
};

//------------------------------------------------------------------------------
/**
 * \brief Provides the default filename for the splitting function setup files.
 * \param options The options to use.
 * \param extraTag The additional (formatted) tag to append.
 * \return The full filename including the path and extension.
 *
 * If the SplittingFunction::fileStem option is nonempty it is used for the main
 * filestem. Otherwise it is constructed from the \a options as
 * 'fileStem_setupTag_SplittingFunction::fileTag_split'.
 *
 * The full filename is then constructed as
 * 'filePath::setup/mainfilestem_extraTag.dat".
 */
inline std::string SplittingFunctionBase::getSetupFilename(const OptionMap& options, const std::string& extraTag)
{
   std::string stem = options.getOption("SplittingFunction::fileStem");
   if (stem.empty())
      stem = options.getOption("fileStem")
             + format_tag(options.getOption("setupTag"))
             + format_tag(options.getOption("SplittingFunction::fileTag"))
             + format_tag("split");
   return format_path(options.getOption("filePath::setup")) + stem + extraTag + ".dat";
}

} // namespace Geneva

#endif // GENEVA_SPLITTING_FUNCTION_BASE_HPP
