//------------------------------------------------------------------------------
/// \file SplittingFunctionISR.hpp
//
// Author(s):
//    Christian Bauer, Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 LBNL, CERN
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SplittingFunctionISR
//------------------------------------------------------------------------------

#ifndef GENEVA_SPLITTING_FUNCTION_ISR_HPP
#define GENEVA_SPLITTING_FUNCTION_ISR_HPP

#include "Geneva/Components/SplittingFunctionBase.hpp"

#include "Geneva/Core/Physics/Signature.hpp"

#include <string>

namespace Geneva
{

class IPDF;
class BoundingConstantsISR;

//------------------------------------------------------------------------------
/**
 * \class SplittingFunctionISR
 * \ingroup Components
 */
//------------------------------------------------------------------------------
class SplittingFunctionISR : public SplittingFunctionBase
{
   private:
      IPDF* _pdf;
      BoundingConstantsISR* _boundingConstant;
      std::string _setupFile;  ///< The name of the grid file
      bool _useRunningPDFs;

   public:
      SplittingFunctionISR(const OptionMap& optMap, ISplittingMapping* mapping, double Ecm, CouplingStrong* alphas, IPDF* pdf, double minX);

      ~SplittingFunctionISR();

      /**
       * Implements the function p(PhiNplus1)
       * The z values of this function are distributed according to the splitting function f(z)
       * The phi values are distributed flat
       * It gives unity if integrated over all allowed values of z and phi, and summed over the allowed splittings
       *
       * \param PhiNplus1 The (N+1)-body phase space point
       *
       * This function does not require a (N+1)-body point that gives rise to a valid splitting.
       * If the (N+1)-body point is invalid, it returns zero
       */
      virtual double normalized(const PhiNplus1& phiNplus1);

      double ratioGSL(double s0, double s1, double s2, void* params) const;

      virtual void printInfo() const;

   private:
      virtual void initialize();

      /// Check if all necessary bounding functions exist
      virtual bool haveAllBoundingConstants() const;

      //-------------------------------------------------------------------------------------
      /**
       * Decides if the value of z should be rejected or not
       *
       * If the value of z is not valid, it should always reject the event
       * Otherwise, compare the sum over all possible splittings to the bounding function
       *
       */
      bool rejectEvent(const Configuration& config, const double TauN, const double z, const double phi, const RegionInfo& map);

      /// The splitting function written only as a function of z and x
      double splitting(Signature::const_iterator emitted, Signature::const_iterator sister, Signature::const_iterator mother, const double TauN, const double z, const double x) const;

      double sumAllowedSplittings(const double TauN, const double z, const double x, const RegionInfo& regionInfo) const;

      /// The functional form of the bounding function
      double bound(Signature::const_iterator mother, const double TauN, double z, double x) const;

      /// The functional form of the integral of the bounding function
      double intBound(Signature::const_iterator mother, const double TauN, double zMin, double z, double x) const;

      /// The functional form of the inverse of the integral of the bounding function
      double intInvBound(Signature::const_iterator mother, double intFraction, const double TauN, double x, double zMin, double zMax) const;

      double pickZ(Signature::const_iterator mother, const double TauN, double x, double zMin, double zMax) const;

      double ratio(const double TauN, const double z, const double x, RegionInfo* m) const;
};

} // namespace Geneva

#endif // GENEVA_SPLITTING_FUNCTION_ISR_HPP
