//------------------------------------------------------------------------------
/// \file SplittingMappingppV0To1.hpp
//
// Author(s):
//    Christian Bauer
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SplittingMappingppV0To1
//------------------------------------------------------------------------------

#ifndef GENEVA_SPLITTING_MAPPING_PPV_0_TO1_HPP
#define GENEVA_SPLITTING_MAPPING_PPV_0_TO1_HPP

#include "Geneva/Components/RegionInfoList.hpp"
#include "Geneva/Components/SplittingMappingBase.hpp"

#include "Geneva/Core/Physics/SignatureSet.hpp"

namespace Geneva
{

class OptionMap;
class PhiN;
class PhiNplus1;
class PhiNplus1CMWritable;

//------------------------------------------------------------------------------
/**
 * \class SplittingFunctionMapping
 * \ingroup Components
 *
 * \brief Implementation of the phase space map required for a SplittingFunction
 */
//------------------------------------------------------------------------------
class SplittingMappingppV0To1: public SplittingMappingBase
{
   public:
      SplittingMappingppV0To1(const OptionMap& optMap);

      /**
       * \brief Create the projected phase space
       *
       * This implements the purely virtual function of ISplittingMapping
       *
       * Create the (N-1)-body config, the values of TauN, z and phi,
       * and the regionInfo from the N-body config
       * This functions returns true if the projection succeeds, and false otherwise
       * The function assumes that the emitted and sister in PhiNplus1 are the closest QCD pair, and that
       * this pair can come from a QCD splitting
       *
       * This function returns false if one of the following conditions are satisfied
       * 1) The Phi2 phase space can not be projected (it is not covered by the map)
       * 2) The (N+1)-body configuration does not allow for the required splitting
       */
      virtual bool calcPhiNValues(const PhiNplus1& config_N, Configuration& config_Nmin1, double& TauN, double& z, double& phi, RegionInfo& regionInfo) const;

      /**
       * \brief Create the splitted phase space
       *
       * Create the (N+1)-body config, given the values of PhiN, TauN, z and phi,
       * This functions returns true if the splitting succeeds, and false otherwise
       */
      virtual bool generateSplit(const PhiN& phiN, const double Tau_N, const double z, const double phi, const RegionInfo& region, PhiNplus1CMWritable& phiNp1) const;

      /**
       * \brief Check if PhiN, TauN, z, phi are a valid point for the splitting
       *
       * This implements the purely virtual function from ISplittingMapping
       *
       * This checks the following conditions:
       * 1) The resulting Phi_N+1 phase space point is kinematically valid
       * 2) The emitted particle has the closest distance measure
       * 3) For ISR, the resulting z value is below the corresponding x value
       */
      virtual bool isValidPhiN(const Configuration& config_NMin1, double TauN, double z, double phi, Signature::const_iterator mother) const;

      /**
       * \brief Return the Jacobian dPhiN / dPhiNmin1 / dTauN / dz / dphi
       *
       * Assumes that the input is valid, i.e. that isValidPhiN is true for the input
       */
      virtual double jacobian(const Configuration& config, double Tau0, double z, double phi, const RegionInfo& map) const;

      /**
       * \brief Return the value of the transverse momentum of the emission
       */
      virtual double kTEmission(PdgID motherID, double Tau0, double z) const;

      /**
       * \brief Return a lower bound to the smallest value of z given PhiN, TauN, phi
       *
       * This implements the purely virtual function of ISplittingMapping
       */
      virtual double zMin(const Configuration& config_Nmin1, const double Tau_N, const double phi, Signature::const_iterator mother, PdgID emittedID);

      /**
       * \brief Return an upper bound to the largest value of z given PhiN, TauN, phi
       *
       * This implements the purely virtual function of ISplittingMapping
       */
      virtual double zMax(const Configuration& config_Nmin1, const double Tau_N, const double phi, Signature::const_iterator mother, PdgID emittedID);

      /**
       * \brief Return a lower bound to the smallest value of phi given PhiN, TauN
       *
       * This implements the purely virtual function of ISplittingMapping
       */
      virtual double phiMin(const Configuration& configNmin1, const double TauN, Signature::const_iterator mother, PdgID emittedID)
      {
         return 0.;
      };

      /**
       * \brief Return an upper bound to the largest value of phi given PhiN, TauN
       *
       * This implements the purely virtual function of ISplittingMapping
       */
      virtual double phiMax(const Configuration& configNmin1, const double TauN, Signature::const_iterator mother, PdgID emittedID)
      {
         return 2. * math_constants::pi();
      };

      /// Returns the number of QCD particles in the N-body configurations.
      virtual size_t numQCDN()
      {
         return 1;
      };

      virtual std::string name() const
      {
         return "ppV0To1";
      };

   private:
      /// Set the smallest allowed value of z
      double zMin(double mV, double expYV, double Tau0, Signature::const_iterator mother);

      /// Set the largest allowed value of z
      double zMax(double mV, double Tau0);
};

} // namespace Geneva

#endif // GENEVA_SPLITTING_MAPPING_PPV_0_TO1_HPP
