//------------------------------------------------------------------------------
/// \file SubtractionFKSBase.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2013 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SubtractionFKSBase
//------------------------------------------------------------------------------

#ifndef GENEVA_SUBTRACTION_FKS_BASE_HPP
#define GENEVA_SUBTRACTION_FKS_BASE_HPP

#include "Geneva/Core/Physics/PdgID.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <vector>

namespace Geneva
{

class ITreeMEComplete;
class FourVector;

//------------------------------------------------------------------------------
/**
 * \class SubtractionFKSBase
 * \ingroup subtraction
 *
 * \brief Base class for FKS subtraction. Contains common functionalities.

 * \todo document me!
 */
//------------------------------------------------------------------------------
class SubtractionFKSBase
{
   protected:
      /// component that calculates tree-level and one-loop ME's
      ITreeMEComplete* _treeMEComplete;
      /// number of light active flavours
      int _nLight;
      /// common scale used in subtraction
      double _commonScale;
      bool _scaleSet;

   public:
      ///Sets the common scale used in the subtraction
      void setScale(double scale);

      ///Gets the common scale used in the subtraction
      double getScale() const;

      ///Gets the number of light (active) flavors
      int nLight() const;

   protected:
      SubtractionFKSBase(ITreeMEComplete* treeMEComplete, int nLight);

      double softIntLightHeavyFin(const FourVector& k, const FourVector& p) const;
      double softIntLightHeavyEps(const FourVector& k, const FourVector& p) const;
      double softIntHeavyHeavyFin(const FourVector& p1, const FourVector& p2) const;
      double softIntHeavyHeavyEps(const FourVector& p1, const FourVector& p2) const;
      double softIntHeavyHeavyEps(const FourVector& p) const;

      /// Utility functions that tells which quark flavour must be considered light or heavy given the numebr of light active flavours,
      /// Set at construction time. Notice that the negation of isLightFlavour does not correspond to isHeavyFlavour, since non QCD particles may be present.
      bool isLightFlavour(PdgID id) const;
      bool isHeavyFlavour(PdgID id) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//---------------------------------------------------------------------------
inline void SubtractionFKSBase::setScale(double scale)
{
   _commonScale = scale;
   _scaleSet = true;
}

//---------------------------------------------------------------------------
inline int SubtractionFKSBase::nLight() const
{
   return _nLight;
}

//---------------------------------------------------------------------------
inline double SubtractionFKSBase::getScale() const
{
   if (_scaleSet) {
      return _commonScale;
   } else {
      Log::error("SubtractionFKSBase::getScale")<<"Trying to obtain the value of common scale before setting it."<<Log::end;
      return std::numeric_limits< double >::quiet_NaN();
   }
}

//---------------------------------------------------------------------------
inline bool SubtractionFKSBase::isLightFlavour(PdgID id) const
{
   return (id.isGluon() || (id.isQuark() && abs(id.code()) <= _nLight));
}

//---------------------------------------------------------------------------
inline bool SubtractionFKSBase::isHeavyFlavour(PdgID id) const
{
   return ((id.isQuark() && abs(id.code()) > _nLight));
}

} // namespace Geneva

#endif // GENEVA_SUBTRACTION_FKS_BASE_HPP
