//------------------------------------------------------------------------------
/// \file SubtractionFKSpp.hpp
//
// Author(s):
//    Christian Bauer, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SubtractionFKSpp
//------------------------------------------------------------------------------

#ifndef GENEVA_SUBTRACTION_FKS_PP_HPP
#define GENEVA_SUBTRACTION_FKS_PP_HPP

#include "Geneva/Components/SubtractionFKSBase.hpp"

#include "Geneva/Core/Physics/Configuration.hpp"

namespace Geneva
{

class ITreeMEComplete;
class PhiRadFKS;
struct PhiNcmFKS;

//------------------------------------------------------------------------------
/**
 * \class SubtractionFKSpp
 * \ingroup subtraction
 *
 * \brief The FKS subtraction for ISR and FSR, without S_ij functions
 *
 * \todo document me!
 *
 * \note This FKS subtraction class does not implement internally the S_ij (a.k.a. Theta_ij)
 * functions, as defined in the original FKS and FNO papers.
 * Henceforth, in order for it to work correctly, these - togheter with their corresponding
 * soft, collinear and soft-collinear limits - have to be provided elsewhere and
 * consistently applied to the various terms entering the subtracted result.
 */
//------------------------------------------------------------------------------
class SubtractionFKSpp : public SubtractionFKSBase
{
   private:
      /// FKS subtraction cutoff parameters (default values mean to always subtract)
      double _xiCut;
      double _yCutISR;
      double _yCutFSR;

   public:
      SubtractionFKSpp(ITreeMEComplete* treeMEComplete, int nlight, double xiCut, double yCutISR, double yCutFSR);

      double xiCut() const   { return _xiCut;   };
      double yCutISR() const { return _yCutISR; };
      double yCutFSR() const { return _yCutFSR; };

      ///The integrated subtraction coefficients (summed over soft and collinear subtractions). The first elements in the vector is the coefficient of the non-logarithmic part, while the second element is the coefficient of the logarithm of the renormalization scale. The third element, if present, is the coeffcient of the logarithm of the factorization scale
      std::vector<double> getIntegratedSubtractionCoefficients(const Configuration& phiN) const;

      std::pair<double, double> ggCollinearRemnant(const double sBorn, const double z) const;
      std::pair<double, double> ggSoftCollinearRemnant(const double sBorn, const double z) const;
      std::pair<double, double> qgCollinearRemnant(const double sBorn, const double z) const;
      std::pair<double, double> qqCollinearRemnant(const double sBorn, const double z) const;
      std::pair<double, double> qqSoftCollinearRemnant(const double sBorn, const double z) const;
      std::pair<double, double> gqCollinearRemnant(const double sBorn, const double z) const;

      ///The Soft subtraction
      double getSoft(const PhiNcmFKS& phiN, const PhiRadFKS& phiRad, double& jac) const;

      ///The Collinear subtraction for a given region
      double getColl(const PhiNcmFKS& phiN, const PhiRadFKS& phiRad, const double xBarA, const double xBarB, PdgID& idA, double& xA, PdgID& idB, double& xB, double& flux, double& jac) const;

      ///The Soft-Collinear subtraction for a given region
      double getSoftColl(const PhiNcmFKS& phiN, const PhiRadFKS& phiRad, double& jac) const;

      /// Returns the soft vector used in the soft subtraction for a given
      /// emitter momentum and phiRad
      FourVector getSoftVec(const FourVector& pk, const PhiRadFKS& phiRad, double Q) const;

      /// Returns the name of the component
      std::string name() const;

   private:
      /// Calculates the transverse vector used in the collinear subtraction for a given
      /// emitter and phiRad
      FourVector getKPerp(const Configuration::const_iterator& emitter,const PhiRadFKS& phiRad) const;

      /// Utility functions needed for Collinear Remnants
      double collRemnFunction(const double sBorn, const double z, const double scale) const;
      double softCollRemnFunction(const double sBorn, const double z, const double scale) const;
};

} // namespace Geneva

#endif // GENEVA_SUBTRACTION_FKS_PP_HPP
