//------------------------------------------------------------------------------
/// \file Tau0FRMapping.hpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann, Christan Bauer
//
// Copyright:
//    Copyright (C) 2013 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Tau0FR
//------------------------------------------------------------------------------

#ifndef GENEVA_REC_TAU0_FIXED_MAPPING_HPP
#define GENEVA_REC_TAU0_FIXED_MAPPING_HPP

#include "Geneva/Core/Physics/Configuration.hpp"
#include "Geneva/Core/Physics/ConfigurationExtensions.hpp"
#include "Geneva/Core/Physics/Signature.hpp"

namespace Geneva
{

class PhiRadFKS;
class FourVector;

typedef PhiRadFKS PhiRadTau0FR;

//------------------------------------------------------------------------------
/**
 * \class Tau0FR
 * \ingroup comp
 *
 * \brief Encapsulates the logic of a map which holds the value of Tau0 fixed
 *
 */
//------------------------------------------------------------------------------
class Tau0FRMapping
{
   private:
      double _cmEnergy;
      bool _qTfixed;

   public:
      Tau0FRMapping(double Ecm, bool fixqT) : _cmEnergy(Ecm), _qTfixed(fixqT) {}

      /// Get the max value for xi above which unphysical or nonsingular configurations would be produced
      double getXiMax(PhiN phiN, double y) const;

      /// Get the max value for xi above which unphysical or nonsingular configurations would be produced
      double getXiMax(PhiNCM phiNCM, double y) const;

      /// Get the max value for xi in the collinear limit
      double getXiMaxColl(PhiNCM phiNCM) const;

      // Performs a PhiN -> PhiN+1 splitting in the CM frame.
      bool generateSplit(PhiNCM phiNCM, const PhiRadTau0FR& phiRad, PhiNplus1CMWritable& phiNp1CM);

      // Performs a PhiN -> PhiN+1 splitting in the lab frame.
      bool generateSplit(PhiN phiN, const PhiRadTau0FR& phiRad, PhiNplus1CMWritable& phiNp1);

      // Performs a PhiN+1 -> PhiN projection in the lab frame.
      bool project(PhiNplus1 phiNp1CM, double xA, double xB, Configuration& configNCM, double& xBarA, double& xBarB) const;

      // Performs a PhiN+1 -> PhiN projection in the CM frame.
      bool project(PhiNplus1 phiNp1Lab, Configuration& configNLab) const;

      // Tests if a CM-frame projection PhiN+1 -> PhiN is possible.
      bool testProject(PhiNplus1 phiNp1Lab) const;

      // Tests if a lab-frame projection PhiN+1 -> PhiN is possible.
      bool testProject(PhiNplus1 phiNp1CM, double xA, double xB) const;

      std::string name() const;

      /// Obtain the phiRad for a given Phi1, Tau1, z, phi, and splitting
      double getPhiRad(const Configuration& Phi1, double Tau1, double z, double phi, Signature::const_iterator mother, PdgID emittedID, PdgID sisterID, PhiRadFKS& phiRad) const;

      /// Return a lower bound to the smallest value of z given PhiN, TauN, phi
      double zMin(const Configuration& Phi1, double Tau1, double phi, Signature::const_iterator mother, PdgID emittedID);

      /// Return an upper bound to the largest value of z given PhiN, TauN, phi
      double zMax(const Configuration& Phi1, double Tau1, double phi, Signature::const_iterator mother, PdgID emittedID);

      ///Return a lower bound to the smallest value of phi given PhiN, TauN
      double phiMin(const Configuration& config, double Tau0, Signature::const_iterator mother, PdgID emittedID);

      /// Return an upper bound to the largest value of phi given PhiN, TauN
      double phiMax(const Configuration& config, double Tau0, Signature::const_iterator mother, PdgID emittedID);

   private:
      // Generates the splitting in the CM frame for an ISR emission.
      bool splitISRCM(PhiNCM phiNCM, const PhiRadTau0FR& phiRad, PhiNplus1CMWritable& phiNp1CM);

      // Generates the splitting in the lab frame for an ISR emission.
      bool splitISRLab(PhiN phiN, const PhiRadTau0FR& phiRad, PhiNplus1CMWritable& phiNp1);

      // Implements the ISR splitting formulas in the lab frame.
      bool splitISR(FourVector qBar, double pBarPlus, double pBarMinus, double pBarx, double pBary, const PhiRadTau0FR& phiRad, PdgID spectatorID, PdgID otherInitial, PhiNplus1CMWritable& phiNp1);

      // Implements the ISR splitting formulas with fixed qT in the lab frame.
      bool splitISRfixqT(FourVector qBar, double pBarPlus, double pBarMinus, const PhiRadTau0FR& phiRad, PdgID spectatorID, PdgID otherInitial, PhiNplus1CMWritable& phiNp1);

      // Generates the splitting in the CM frame for an FSR emission.
      bool splitFSRCM(PhiNCM phiNCM, const PhiRadTau0FR& phiRad, PhiNplus1CMWritable& phiNp1CM);

      // Generates the splitting in the Lab frame for an FSR emission.
      bool splitFSRLab(PhiN phiN, const PhiRadTau0FR& phiRad, PhiNplus1CMWritable& phiNp1);

      // Implements the FSR splitting formulas in the lab frame.
      bool splitFSR(FourVector qBar, double pBarPlus, double pBarMinus, const PhiRadTau0FR& phiRad, PdgID idA, PdgID idB, PhiNplus1CMWritable& phiNp1);

      // Implements the FSR splitting formulas with fixed qT in the lab frame.
      bool splitFSRfixqT(FourVector qBar, double pBarPlus, double pBarMinus, const PhiRadTau0FR& phiRad, PdgID idA, PdgID idB, PhiNplus1CMWritable& phiNp1);

      // Performs the projection for ISR emissions in the CM frame.
      bool projectISRCM(const Configuration& configNp1CM, Configuration::iterator emitted, Configuration::iterator sister, double xA, double xB, Configuration& configNCM, double& xBarA, double& xBarB) const;

      // Performs the projection for ISR emissions in the Lab frame.
      bool projectISRLab(const Configuration& configNp1, Configuration::iterator emitted, Configuration::iterator sister, Configuration& configN) const;

      // Implements the ISR projection formulas for ISR emissions.
      bool projectISR(FourVector q, double p1Plus, double p1Minus, double p2Plus, double p2Minus, const double p1x, const double p1y, PdgID spectator, PdgID idA, PdgID idB, Configuration& configN) const;

      // Implements the ISR splitting formulas with fixed qT in the lab frame.
      bool projectISRfixqT(FourVector q, double p1Plus, double p1Minus, double p2Plus, double p2Minus, PdgID spectator, PdgID idA, PdgID idB, Configuration& configN) const;

      // Performs the projection for FSR emissions in the CM frame.
      bool projectFSRCM(const Configuration& configNp1CM, Configuration::iterator emitted, Configuration::iterator sister, double xA, double xB, Configuration& configNCM, double& xBarA, double& xBarB) const;

      // Performs the projection for FSR emissions in the lab frame.
      bool projectFSRLab(const Configuration& configNp1, Configuration::iterator emitted, Configuration::iterator sister, Configuration& configN) const;

      // Implements the FSR projection formulas in the lab frame.
      bool projectFSR(FourVector q, double p1Plus, double p1Minus, double p2Plus, double p2Minus, PdgID mother, PdgID idA, PdgID idB, Configuration& configN) const;

      // Implements the FSR projection formulas with fixed qT in the lab frame.
      bool projectFSRfixqT(FourVector q, double p1Plus, double p1Minus, double p2Plus, double p2Minus, PdgID mother, PdgID idA, PdgID idB, Configuration& configN) const;

      // Tests the projection for ISR emissions in the CM frame.
      bool testProjectISRCM(const Configuration& configNp1CM, Configuration::iterator emitted, double xA, double xB) const;

      // Tests the projection for ISR emissions in the lab frame.
      bool testProjectISRLab(const Configuration& configNp1, Configuration::iterator emitted) const;

      // Tests the projection for FSR emissions in the CM frame.
      bool testProjectFSRCM(const Configuration& configNp1CM, Configuration::iterator emitted, Configuration::iterator sister, double xA, double xB) const;

      // Tests the projection for FSR emissions in the lab frame.
      bool testProjectFSRLab(const Configuration& configNp1, Configuration::iterator emitted, Configuration::iterator sister) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline std::string Tau0FRMapping::name() const
{
   std::string str("Tau0FRMapping");
   if (_qTfixed)
      str += " fix qT";
   return str;
}

} // namespace Geneva

#endif // GENEVA_REC_TAU0_FIXED_MAPPING_HPP
