//------------------------------------------------------------------------------
/// \file TreeMEOpenLoops.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.

// Description:
//    Definition of class TreeMEOpenLoops
//------------------------------------------------------------------------------

#ifndef GENEVA_TREE_ME_OPENLOOPS_HPP
#define GENEVA_TREE_ME_OPENLOOPS_HPP

#include "Geneva/Components/ITreeME.hpp"
#include "Geneva/Components/ITreeMEColor.hpp"

#include "Geneva/Core/Physics/Signature.hpp"

#include <string>
#include <vector>
#include <map>

namespace Geneva
{

class Configuration;
class SignatureSet;
class OptionMap;

//------------------------------------------------------------------------------
/**
 * \class TreeMEOpenLoops
 * \ingroup gen
 *  Implementation of the TreeMEOpenLoops interface
 */
//------------------------------------------------------------------------------
class TreeMEOpenLoops : virtual public ITreeMEColor
{
   private:
      const Configuration* _config;       ///< pointer to the last evaluated Configuration
      double _value;                      ///< the last evaluated squared matrix element
      std::vector<double> _valueLargeNc;  ///< the array of squared amplitudes for each color flow (in the large-Nc limit).
      std::map<Signature, std::vector<std::vector<std::pair<int,int> > > > _colorFlowBasis;

   public:
      TreeMEOpenLoops(const OptionMap& optMap);

      virtual int checkCompatibility(const SignatureSet& sigSet) const;
      virtual const std::string name() const;

      /// Calculates all required value information for a given Configuration.
      virtual void calculateValue(const Configuration& config);

      /// Calculates all required color-related information for a given Configuration.
      virtual void calculateColorInfo(const Configuration& config);

       /// Returns the squared tree-level matrix element for the last calculated configuration.
      virtual double value() const;

      /// Returns the squared amplitude in the large-Nc limit for a given color flow cf. Color factor is not included.
      virtual double valueLargeNc(colorFlowIndex cf) const;

      /// Returns the color indexes for the given configuration, leg i and color flow cf
      virtual std::pair<int, int> colorIndices(Configuration::const_iterator i, colorFlowIndex cf) const;

      /// Returns the  number of color flows of the last calculated Configuration
      virtual size_t numColorFlows() const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline double TreeMEOpenLoops::value() const
{
   return _value;
}

//------------------------------------------------------------------------------
inline size_t TreeMEOpenLoops::numColorFlows() const
{
   return _colorFlowBasis.at(Signature(*_config)).size();
}

//------------------------------------------------------------------------------
inline double TreeMEOpenLoops::valueLargeNc(colorFlowIndex i) const
{
   return _valueLargeNc[i];
}

//------------------------------------------------------------------------------
inline int TreeMEOpenLoops::checkCompatibility(const SignatureSet& sigSet) const
{
   //TODO: remove this function ? never used and doing nothing
   return 0;
}

} // namespace Geneva

#endif // GENEVA_TREE_ME_OPENLOOPS_HPP
