//------------------------------------------------------------------------------
/// \file UnitCubeRadGenerator.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class UnitCubeRadGenerator
//------------------------------------------------------------------------------

#ifndef GENEVA_UNIT_CUBE_RAD_GENERATOR_HPP
#define GENEVA_UNIT_CUBE_RAD_GENERATOR_HPP

#include "Geneva/config.hpp"

#define DEFAULT_X1_CUTOFF 1e-10
#define DEFAULT_X2_CUTOFF 1e-10

namespace Geneva
{

class UnitCubeRad;

//------------------------------------------------------------------------------
/**
 * \class UnitCubeRadGenerator
 * \ingroup Generators
 *
 * \brief Randomly generates radiation variables in a unitary cube.
 *
 * The kinematic variables in UnitCubeRad are randomly generated within
 * the following limits:
 *
 * - \f$ _x1TinyCutoff \leq x1 \leq 1 - _x1TinyCutoff \f$
 * - \f$ _x2TinyCutoff \leq x2 \leq 1 - _x2TinyCutoff \f$
 * - \f$ 0 \leq x3 \leq 1 \f$
 *
 * The sampling is done considering x1 as an energy variable (singular as x1 -> 0), x2 as
 * a collinear angle (singular as x2 -> 0 and x2 -> 1) and x3 an azimuthal angle (not singular).
 * The jacobian from the importance sampling is included in the generated UnitCubeRad weight.
 *
 * In order to account for situations where the Jacobian can have singularities away from the
 * soft limit, if the singularJacobian variable is set to true, the sampling for x1 is done
 * considering also an integrable singularity in the limit x1 -> 1.
*/
//------------------------------------------------------------------------------
class UnitCubeRadGenerator
{
   private:
      ///< Cutoff values used for \f x1 \f and \f x2 \f to avoid numerical instabilities
      double _x1TinyCutoff;
      double _x2TinyCutoff;

      bool _independentTinyCuts;
      bool _singularJacobian;

   public:
      ///TODO: switch singularJacobian default to false when #ifdef NEW will be selected as the new default
      UnitCubeRadGenerator(double x1TinyCutoff = DEFAULT_X1_CUTOFF, double x2TinyCutoff = DEFAULT_X2_CUTOFF, bool independentTinyCuts = true, bool singularJacobian = true);

      /// Get the value of the tiny cutoff on x1 to avoid numerical instabilities.
      double x1TinyCutoff() const;

      /// Get the value of the tiny cutoff on x2 to avoid numerical instabilities.
      double x2TinyCutoff() const;

      /// Generates the kinematics of a UnitCubeRad for a generic emission
      void generate(double rand1, double rand2, double rand3, UnitCubeRad& xRad) const;

      /// Generates the kinematics of a UnitCubeRad for a generic emission. The sampling of the angle y depends on the sampling parameter ySampling
      void generate(double rand1, double rand2, double rand3, UnitCubeRad& xRad, double ySampling) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Declarations
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline UnitCubeRadGenerator::UnitCubeRadGenerator(double x1TinyCutoff, double x2TinyCutoff, bool independentTinyCuts, bool singularJacobian)
   : _x1TinyCutoff(x1TinyCutoff), _x2TinyCutoff(x2TinyCutoff), _independentTinyCuts(independentTinyCuts), _singularJacobian(singularJacobian)
{}

//------------------------------------------------------------------------------
inline double UnitCubeRadGenerator::x1TinyCutoff() const
{
   return _x1TinyCutoff;
}

//------------------------------------------------------------------------------
inline double UnitCubeRadGenerator::x2TinyCutoff() const
{
   return _x2TinyCutoff;
}

} // namespace Geneva

#endif // GENEVA_UNIT_CUBE_RAD_GENERATOR_HPP
