//------------------------------------------------------------------------------
/// \file Configuration.hpp
//
// Author(s):
//    Frank Tackmann, Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 Frank Tackmann, LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Configuration
//------------------------------------------------------------------------------

#ifndef GENEVA_CONFIGURATION_HPP
#define GENEVA_CONFIGURATION_HPP

#include "Geneva/Core/Physics/SignatureStatus.hpp"
#include "Geneva/Core/Physics/Particle.hpp"

#include <set>
#include <iosfwd>

namespace Geneva
{

class Signature;

//------------------------------------------------------------------------------
/**
 * \class Configuration
 * \ingroup Core
 *
 * \brief A set of Particles representing a phase-space configuration.
 *
 * The %Configuration distinguishes between initial-state and final-state
 * particles analogous to Signature IDs. There are at most two initial-state
 * particles that are distinguished as A and B, and are accessible via
 * initialA() and initialB() iterators. Identical particles in the final state
 * are allowed and are not distinguished (in particular their ordering inside
 * the %Configuration is meaningless and should not be relied upon). Internally
 * the %Configuration is implemented in terms of a STL multiset of Particles,
 * allowing for fast access and insertion/removal of Particles. The interface
 * closely resembles that of Signature
 *
 * A %Configuration can be traversed using iterators that have the type
 * Configuration::iterator and Configuration::reverse_iterator. Various member
 * functions use or return iterators. (Note that the iterators are always
 * const iterators, i.e. they are type-equivalent to Configuration::const_iterator)
 *
 * The Particle pointed to by an iterator 'i' is accessed using either *i or
 * i->. Iterators can be incremented, ++i, or decremented, --i, to point to the
 * next or previous particle in the %Configuration, and can be checked for
 * (in)equality (using == or !=). The relation operators (<, <=, >, >=) cannot
 * be used with %Configuration iterators. Iterators to Particles in a
 * %Configuration also support i->isInitialA(), i->isInitialB(), and
 * i->isFinal() queries, which allow to test as what kind of Particle the
 * pointed to Particle is stored.
 *
 * The members begin() and end() return iterators pointing to the first Particle
 * and one step beyond the last particle, respectively. Hence a %Configuration
 * can be traversed using
 * \code
 *    for (Configuration::iterator i = conf.begin(); i != conf.end(), ++i) {
 *       doSomethingwithParticle(*i);
 *       ...
 *    }
 * \endcode
 *
 * A reverse_iterator behaves exactly opposite which is useful if one wants to
 * traverse the %Configuration in descending order, i.e., rbegin() and rend()
 * point to the last Particle and one step before the first Particle,
 * respectively, and after incrementing a reverse_iterator it points to the
 * previous Particle.
 *
 * The above loop over both initial and final particles. To loop only over final
 * particles, use the otherwise similar beginFinal(), endFinal(), rbeginFinal(),
 * and rendFinal(). Iterators to the initial particles (A and B) can
 * be obtained from the initialA() and initialB() member functions.
 *
 * An existing iterator pointing to a Particle in a given %Configuration stays
 * valid if new Particles are inserted or other Particles are erased. Of course
 * it does become invalid if the Particle it points to is erased from the
 * Configuration. An iterator can only be used with the particular instance of a
 * Configuration it refers to, ie, it should be used like a pointer not like an
 * index.
 */
//------------------------------------------------------------------------------
class Configuration
{
   private:
      typedef SignatureStatus _status_type;

      /**
       * Internal type used by %Configuration which supplementes a Particle by a
       * Signature_Status indicating initial or final state.
       */
      class _status_particle : public Particle, public _status_type
      {
         friend class Configuration;
         friend class SignatureStatusID;

         private:
            _status_particle() = default;
            _status_particle(const Particle& particle, int status = kFinal)
               : Particle(particle), _status_type(status) {}
      };

      typedef SignatureStatusID::_compare_type _compare_type;
      typedef std::multiset<_status_particle, _compare_type> _impl_type;

      _impl_type _particles;                ///< the %Configuration's particles

      _impl_type::iterator _initialA;       ///< iterator to initial-state particle A
      _impl_type::iterator _initialB;       ///< iterator to initial-state particle B
      _impl_type::iterator _beginFinal;     ///< iterator to first final-state particle
      size_t _sizeFinal;                    ///< counter for the number of final states
      mutable _status_particle _lookupFinal;///< dummy entry for final-state searches

   public:
      /// \name Type Definitions
      ///@{
      typedef _impl_type::const_reference const_reference;

      /// const_iterator for \c Particle entries in the %Configuration.
      typedef _impl_type::const_iterator const_iterator;
      /// iterator is type equivalent to const_iterator
      typedef const_iterator iterator;
      ///@}

      /// \name Constructors etc.
      ///@{

      /// Creates an empty %Configuration.
      Configuration();

      /// Creates a %Configuration with \a particle as the only element in the final state.
      Configuration(const Particle& particle);

      // Creates a Configuration from a range of Particles.
      template <typename ConfigLikeIterator>
      Configuration(ConfigLikeIterator first, ConfigLikeIterator last);

      // Configuration copy constructor.
      Configuration(const Configuration& other);

      // Configuration assignment operator.
      Configuration& operator=(const Configuration& rhs);

      // Move assignment from another configuration.
      Configuration& moveFrom(Configuration& rhs);

      // Swaps the contents with another Configuration.
      void swap(Configuration& other);

      ~Configuration() = default;
      ///@}

      /// \name Accessing and Querying Particles
      ///@{
      // Returns an iterator to the first Particle in the Configuration.
      iterator begin() const;

      // Returns an iterator pointing beyond the last Particle in the Configuration.
      iterator end() const;

      // Returns an iterator to initial-state Particle A.
      iterator initialA() const;

      // Returns an iterator to initial-state Particle B.
      iterator initialB() const;

      // Returns an iterator to the first final-state Particle.
      iterator beginFinal() const;

      // Returns an iterator pointing beyond the last final-state Particle.
      iterator endFinal() const;

      /// Returns \c true if the %Configuration is empty, \c false otherwise.
      bool empty() const;

      /// Returns the total number of Particles in the %Configuration.
      size_t size() const;

      /// Returns the total number of Particles in the %Configuration's final state.
      size_t sizeFinal() const;

      /// Returns the maximum number of Particles the %Configuration could hold.
      size_t max_size() const;

      // Counts how often a Particle with some PdgID appears in the Configuration.
      size_t count(PdgID id) const;

      // Counts how often a Particle with some PdgID appears in the Configuration's final state.
      size_t countFinal(PdgID id) const;

      // Tries to find a Particle matching a Signature-like entry in the Configuration.
      template <typename SignatureLikeIterator>
      iterator find(SignatureLikeIterator it) const;

      // Tries to find a Particle with a given PdgID in the Configuration.
      iterator find(PdgID id) const;

      // Tries to find a Particle with a given PdgID in the Configuration's final state.
      iterator findFinal(PdgID id) const;

      // Finds the range of Particles with some PdgID in the Configuration's final state.
      std::pair<iterator, iterator> equalRangeFinal(PdgID id) const;

      // Tests if a Particle matching a Signature-like entry is contained in the Configuration.
      template <typename SignatureLikeIterator>
      bool includes(SignatureLikeIterator it) const;

      // Tests if a Particle with some PdgID is contained in the Configuration.
      bool includes(PdgID id) const;

      // Tests if a Particle with some PdgID is contained in the Configuration's final state.
      bool includesFinal(PdgID id) const;

      // Tests if a Signature is contained in the Configuration.
      bool includes(const Signature& other) const;

      // Tests if a Signature intersects with the Configuration.
      bool intersects(const Signature& other) const;
      ///@}

      /// \name Adding and Removing Particles
      ///@{
      // Inserts an initial- or final-state Particle into the Configuration.
      template <typename ConfigLikeIterator>
      iterator insert(ConfigLikeIterator it);

      // Inserts a range of Particles into the Configuration.
      template <typename ConfigLikeIterator>
      void insert(ConfigLikeIterator first, ConfigLikeIterator last);

      // Sets the initial-state Particle A of the Configuration.
      iterator setInitialA(const Particle& particle);

      // Sets the initial-state Particle B of the Configuration.
      iterator setInitialB(const Particle& particle);

      // Inserts a Particle into the Configuration's final state.
      iterator insertFinal(const Particle& particle);

      // Inserts a Particle into the Configuration's final state using a hint.
      iterator insertFinal(iterator hint, const Particle& particle);

      // Inserts a range of Particles into the Configuration's final state.
      template <typename InputIterator>
      void insertFinal(InputIterator first, InputIterator last);

      // Removes the Particle from the Configuration pointed to by it.
      void erase(iterator it);

      // Removes all Particles with some PdgID from the Configuration.
      size_t erase(PdgID id);

      // Removes a range of Particles from the Configuration.
      void erase(iterator first, iterator last);

      // Removes the Configuration's initial-state Particle A.
      bool eraseInitialA();

      // Removes the Configuration's initial-state Particle B.
      bool eraseInitialB();

      // Removes all Particles with some PdgID from the Configuration's final state.
      size_t eraseFinal(PdgID id);

      /// Removes all Particles from the %Configuration.
      void clear();
      ///@}

      /// \name Algebra
      ///@{
      // Checks if the Configuration is equal to a Signature.
      bool operator==(const Signature& rhs) const;

      // Checks if the Configuration is not equal to a Signature.
      bool operator!=(const Signature& rhs) const;

      // Applies a transformation to the momentum of each Particle in the Configuration.
      template <class Transformation>
      Configuration& apply(const Transformation& trans);

      // Returns a copy of %Configuration with a transformation applied to it.
      template <class Transformation>
      Configuration transformed(const Transformation& trans) const;
      ///@}

      /// \name Output
      ///@{
      // Configuration stream insertion operator.
      friend std::ostream& operator<<(std::ostream& out, const Configuration& config);
      ///@}

   private:
      /// Streams the %Configuration to \a output.
      std::ostream& print(std::ostream& output) const;

      /// Initializes the internal initial-state and final-state iterators.
      void init();

      /// Copies the content of another Configuration.
      void copy(const Configuration& other);
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Creates an empty Configuration.
inline Configuration::Configuration()
   : _particles(), _sizeFinal(0)
{
   _initialA   = end();
   _initialB   = end();
   _beginFinal = end();
}

//------------------------------------------------------------------------------
// Creates a Configuration with particle as the only element in the final state.
inline Configuration::Configuration(const Particle& particle)
   : _particles(), _sizeFinal(1)
{
   _initialA   = end();
   _initialB   = end();
   _beginFinal = _particles.insert(particle);
}

//------------------------------------------------------------------------------
/**
 * \brief Creates a %Configuration from a range of Particles.
 * \param first An iterator pointing to the first Particle in the range.
 * \param last  An iterator pointing beyond the last Particle in the range.
 *
 * Creates a new %Configuration by copying the particles from the range
 * specified by [\a first,\a last), keeping their initial-state and final-state
 * assignment.
 *
 * \note This is a template function that works for any %Configuration-like
 * iterator from which a particle can be read. To fill a %Configuration from a
 * generic range of particles (e.g. a std::vector) the insertFinal() member can
 * be used.
 */
template <typename ConfigLikeIterator>
inline Configuration::Configuration(ConfigLikeIterator first, ConfigLikeIterator last)
   : _particles()
{
   // here we rely on the Configuration ordering in [first, last)
   if (first != last && first->isInitialA()) {
      _initialA = _particles.insert(end(), _status_particle(*first, _status_particle::kInitialA));
      ++first;
   } else
      _initialA = end();
   if (first != last && first->isInitialB()) {
      _initialB = _particles.insert(end(), _status_particle(*first, _status_particle::kInitialB));
      ++first;
   } else
      _initialB = end();
   if (first != last) {
      const size_t oldsize = size();
      _beginFinal = _particles.insert(end(), *first);
      ++first;
      for (; first != last; ++first)
         _particles.insert(end(), *first);
      _sizeFinal = size() - oldsize;
   } else {
      _beginFinal = end();
      _sizeFinal = 0;
   }
}

//------------------------------------------------------------------------------
/**
 * \brief %Configuration copy constructor.
 * \param other The %Configuration that will be copied.
 *
 * Creates a new %Configuration as a copy of \a other.
 */
inline Configuration::Configuration(const Configuration& other)
   : _particles()
{
   copy(other);
}

//------------------------------------------------------------------------------
/**
 * \brief %Configuration assignment operator.
 * \param rhs The %Configuration that will be copied.
 * \return A reference to the %Configuration.
 *
 * Replaces the current content of the %Configuration with a copy of \a rhs.
 */
inline Configuration& Configuration::operator=(const Configuration& rhs)
{
   if (this != &rhs) {
      _particles.clear();
      copy(rhs);
   }
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief Move assignment from another configuration.
 * \param rhs The %Configuration that will be moved over.
 * \return A reference to the %Configuration.
 *
 * This mimics a move assignment in the absence of C++11 move semantics. It
 * replaces the current content of the %Configuration with that of \a rhs. (It
 * directly moves the content of \a rhs to this %Configuration without copying
 * of elements, which means it is fast and executes in constant time.)
 *
 * Iterators to \a rhs remain valid and now point to this %Configuration.
 * Previous iterators to this %Configuration become invalid.
 *
 * \warning After this operation \a rhs becomes invalid and must not be used
 * anymore.
 */
inline Configuration& Configuration::moveFrom(Configuration& rhs)
{
   _particles.swap(rhs._particles);
   _initialA = rhs._initialA;
   _initialB = rhs._initialB;
   _beginFinal = rhs._beginFinal;
   _sizeFinal = rhs._sizeFinal;
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief Swaps the contents with another %Configuration.
 * \param other The %Configuration being swapped with.
 *
 * This exchanges the contents of the %Configuration with that of \a other. The
 * swapping does not require the copying of elements, which means it is fast and
 * executes in constant time.
 */
inline void Configuration::swap(Configuration& other)
{
   _particles.swap(other._particles);
   init();
   other.init();
}

//------------------------------------------------------------------------------
// Accessing and Querying Particles
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator to the first Particle in the %Configuration.
 *
 * If the %Configuration is empty the returned iterator will be equal to end().
 * It is the user's responsibility to ensure that the iterator is not used for
 * reading in this case.
 */
inline Configuration::iterator Configuration::begin() const
{
   return _particles.begin();
}

//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator pointing beyond the last Particle in the %Configuration.
 *
 * The returned iterator points one step beyond the last Particle and must not
 * be used for reading. Instead, it is used to check the validity of other
 * iterators.
 */
inline Configuration::iterator Configuration::end() const
{
   return _particles.end();
}

//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator to initial Particle A.
 *
 * The returned iterator points to the Particle in the %Configuration that is
 * considered to be the initial Particle A. If no initial particle A is defined
 * end() is returned.
 */
inline Configuration::iterator Configuration::initialA() const
{
   return _initialA;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator to initial Particle B.
 *
 * The returned iterator points to the Particle in the %Configuration that is
 * considered to be the initial Particle B. If no initial particle B is defined
 * end() is returned.
 */
inline Configuration::iterator Configuration::initialB() const
{
   return _initialB;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator to the first final-state particle.
 *
 * Iterating from beginFinal() to endFinal() will iterator over all final-state
 * particles in the %Configuration and no others. If the %Configuration is empty
 * the returned iterator will be equal to endFinal(). It is the user's
 * responsibility to ensure that the iterator is not used for reading in this
 * case.
 */
inline Configuration::iterator Configuration::beginFinal() const
{
   return _beginFinal;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns an state iterator pointing beyond the last final-state particle.
 *
 * The returned iterator points one step beyond the last final-state Particle
 * and must not be used for reading. Instead, it is used to check the validity
 * of other iterators.
 */
inline Configuration::iterator Configuration::endFinal() const
{
   return end();
}

//------------------------------------------------------------------------------
// Returns true if the Configuration is empty, false otherwise.
inline bool Configuration::empty() const
{
   return _particles.empty();
}

//------------------------------------------------------------------------------
// Returns the total number of Particles in the Configuration.
inline size_t Configuration::size() const
{
   return _particles.size();
}

//------------------------------------------------------------------------------
// Returns the total number of final-state Particles in the Configuration.
inline size_t Configuration::sizeFinal() const
{
   return _sizeFinal;
}

//------------------------------------------------------------------------------
// Returns the maximum number of Particles the Configuration could hold.
inline size_t Configuration::max_size() const
{
   return _particles.max_size();
}

//------------------------------------------------------------------------------
/**
 * \brief Counts how often a Particle with some PdgID appears in the %Configuration.
 * \param id The PdgID to be looked for.
 * \return The number of times \a id appears in the %Configuration.
 *
 * Ocurrences of \a id in initial and final states are counted. If \a id is not
 * contained in the %Configuration at all then 0 is returned.
 */
inline size_t Configuration::count(PdgID id) const
{
   size_t temp = countFinal(id);
   if (_initialA != end() && _initialA->id() == id)
      ++temp;
   if (_initialB != end() && _initialB->id() == id)
      ++temp;
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief Counts how often a Particle with some PdgID appears in the %Configuration's final state.
 * \param id The PdgID to be looked for.
 * \return The number of final-state particles having \a id.
 *
 * If no Particle with \a id is contained in the %Configuration then 0 is
 * returned.
 */
inline size_t Configuration::countFinal(PdgID id) const
{
   _lookupFinal._id = id;
   return _particles.count(_lookupFinal);
}

//------------------------------------------------------------------------------
/**
 * \brief Tries to find a Particle matching a Signature-like entry in the %Configuration.
 * \param it An iterator pointing to the Signature-like entry to be searched for.
 * \return An iterator pointing to the first instance of a matching particle.
 *
 * If the PdgID \a it points to including its initial- or final-state
 * assignment is found in the %Configuration, the returned iterator points to
 * its first instance, otherwise end() is returned. Requires logarithmic time.
 *
 * \note This is a template function that works for any type of Signature-like
 * iterator (i.e. Signature::iterator or Configuration::iterator). For \a it
 * being a Configuration::iterator only the PdgID and initial/final assignment
 * are considered, but not the momentum of the Particle.
 */
template <typename SignatureLikeIterator>
inline Configuration::iterator Configuration::find(SignatureLikeIterator it) const
{
   if (it->isInitialA()) {
      if (_initialA != end() && _initialA->id() == it->id())
         return _initialA;
      else
         return end();
   }
   if (it->isInitialB()) {
      if (_initialB != end() && _initialB->id() == it->id())
         return _initialB;
      else
         return end();
   }
   return findFinal(it->id());
}

//------------------------------------------------------------------------------
/**
 * \brief Tries to find a Particle with a given PdgID in the %Configuration.
 * \param id The PdgID to be searched for.
 * \return An iterator pointing to the first instance of the sought particle.
 *
 * If a particle with \a id is found in the %Configuration the iterator points
 * to the first particle having \a id, otherwise end() is returned. Requires
 * logarithmic time. The search is performed over both initial and final states.
 */
inline Configuration::iterator Configuration::find(PdgID id) const
{
   if (_initialA != end() && _initialA->id() == id)
      return _initialA;
   if (_initialB != end() && _initialA->id() == id)
      return _initialA;
   return findFinal(id);
}

//------------------------------------------------------------------------------
/**
 * \brief Tries to find a Particle with a given PdgID in the %Configuration's final state.
 * \param id The PdgID to be searched for.
 * \return An iterator pointing to the first instance of a matching Particle.
 *
 * If \a id is found in the %Configuration the iterator points to the first
 * final-state Particle having \a id, otherwise endFinal() is returned. Requires
 * logarithmic time.
 */
inline Configuration::iterator Configuration::findFinal(PdgID id) const
{
   _lookupFinal._id = id;
   return _particles.find(_lookupFinal);
}

//------------------------------------------------------------------------------
/**
 * \brief Finds the range of Particles with some PdgID in the %Configuration's final state.
 * \param id The PdgID to be searched for.
 * \return A pair of iterators specifying the found range.
 *
 * If \a id is part of the %Configuration, then the [first, last) range
 * specified by the returned pair of iterators gives the complete range of
 * particles in the final state having \a id. If \a id is not found in
 * %Configuration then the two iterators will be equal, specifying an empty
 * range.
 */
inline std::pair<Configuration::iterator, Configuration::iterator>
Configuration::equalRangeFinal(PdgID id) const
{
   _lookupFinal._id = id;
   return _particles.equal_range(_lookupFinal);
}

//------------------------------------------------------------------------------
/**
 * \brief Tests if a Particle matching a Signature-like entry is contained in the %Configuration.
 * \param it An iterator pointing to the Signature-like entry to be searched for.
 * \return \c true if a particle matching \a it is contained in the
 *         %Configuration, \c false otherwise.
 *
 * This is identical to find(it) != end(), so if an iterator to the particle is
 * needed in case it is contained, it is faster to use find(it) directly. See
 * the documentation of find(it) for further details.
 */
template <typename SignatureLikeIterator>
inline bool Configuration::includes(SignatureLikeIterator it) const
{
   return find(it) != end();
}

//------------------------------------------------------------------------------
/**
 * \brief Tests if a Particle with some PdgID is contained in the %Configuration.
 * \param id The PdgID to be searched for.
 * \return \c true if a particle with \a id is contained in the %Configuration,
 *         \c false otherwise.
 *
 * This is identical to find(id) != end(), so if an iterator to the particle is
 * needed in case it is contained, it is faster to use find(id) directly. The
 * search is performed over both initial and final states.
 */
inline bool Configuration::includes(PdgID id) const
{
   return find(id) != end();
}

//------------------------------------------------------------------------------
/**
 * \brief Tests if a Particle with some PdgID is contained in the %Configuration's final state.
 * \param id The PdgID to be searched for.
 * \return \c true if a particle with \a id is contained in the %Configuration,
 *         \c false otherwise.
 *
 * This is identical to findFinal(id) != endFinal(), so if an iterator to the
 * particle is needed in case it is contained, it is faster to use findFinal(id)
 * directly.
 */
inline bool Configuration::includesFinal(PdgID id) const
{
   return findFinal(id) != endFinal();
}

//------------------------------------------------------------------------------
// Adding and Removing Particles
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Inserts an initial- or final-state Particle into the %Configuration.
 * \param it An iterator pointing to the Particle to be inserted.
 * \return An iterator pointing to the inserted Particle.
 *
 * If \a it points to an initial-state Particle the corresponding initial-state
 * (A or B) is set, otherwise \a it is inserted as a new final-state particle,
 * in which case the insertion requires logarithmic time.
 *
 * \note This is a template function that works for any %Configuration-like
 * iterator from which a particle can be read.
 */
template <typename ConfigLikeIterator>
inline Configuration::iterator Configuration::insert(ConfigLikeIterator it)
{
   if (it->isInitialA())
      return setInitialA(*it);
   if (it->isInitialB())
      return setInitialB(*it);
   return insertFinal(*it);
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts a range of Particles into the %Configuration.
 * \param first An iterator pointing to the first Particle in the range.
 * \param last  An iterator pointing beyond the last Particle in the range.
 *
 * Inserts the Particles from the range specified by [\a first,\a last). The
 * complexity is as for the analogous range constructor. The initial- and final-
 * state assignment of the Particles is kept. Possible initial states in the
 * range will overwrite the %Configuration's present initial states, while any
 * final states are added.
 *
 * \note This is a template function that works for any %Configuration-like
 * iterator from which a particle can be read.
 */
template <typename ConfigLikeIterator>
inline void Configuration::insert(ConfigLikeIterator first, ConfigLikeIterator last)
{
   // here we can keep initial-state information and rely on the ordering in [first, last)
   if (first != last && first->isInitialA()) {
      setInitialA(*first);
      ++first;
   }
   if (first != last && first->isInitialB()) {
      setInitialB(*first);
      ++first;
   }
   if (first != last) {
      const size_t oldsize = size();
      insertFinal(*first);
      ++first;
      for (; first != last; ++first)
         _particles.insert(end(), *first);
      _sizeFinal += size() - oldsize;
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Sets the initial-state Particle A of the %Configuration.
 * \param particle The particle to be set as initial-state A.
 * \return An iterator pointing to the inserted particle.
 *
 * If the %Configuration already has an initial-state A set, it is overwritten
 * with the given particle. If this is not desired, the user should check for
 * the existence of an initial state A first.
 */
inline Configuration::iterator Configuration::setInitialA(const Particle& particle)
{
   if (_initialA == end()) {
      _initialA = _particles.insert(_status_particle(particle, _status_particle::kInitialA));
   } else {
      // Forcing to overwrite the particle is faster than erasing and inserting. It is
      // legitimate here, since we enforce that there is only one initial-state
      // particle of a kind and whose ordering in the set is determined by its state only.
      const_cast<_status_particle&>(*_initialA).Particle::operator=(particle);
   }
   return _initialA;
}

//------------------------------------------------------------------------------
/**
 * \brief Sets the initial-state Particle B of the %Configuration.
 * \param particle The particle to be set as initial-state B.
 * \return An iterator pointing to the inserted particle.
 *
 * If the %Configuration already has an initial-state B set, it is overwritten
 * with the given particle. If this is not desired, the user should check for
 * the existence of an initial state B first.
 */
inline Configuration::iterator Configuration::setInitialB(const Particle& particle)
{
   if (_initialB == end()) {
      _initialB = _particles.insert(_status_particle(particle, _status_particle::kInitialB));
   } else {
      // Forcing to overwrite the particle is faster than erasing and inserting. It is
      // legitimate here, since we enforce that there is only one initial-state
      // particle of a kind and whose ordering in the set is determined by its state only.
      const_cast<_status_particle&>(*_initialB).Particle::operator=(particle);
   }
   return _initialB;
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts a Particle into the %Configuration's final state.
 * \param particle The Particle to be inserted.
 * \return An iterator pointing to the inserted Particle.
 *
 * Insertion requires logarithmic time.
 */
inline Configuration::iterator Configuration::insertFinal(const Particle& particle)
{
   iterator temp = _particles.insert(particle);
   ++_sizeFinal;
   // update _beginFinal if necessary, relies on the fact that for identical
   // elements new ones are inserted behind existing ones
   if (_beginFinal == end() || _compare_type()(*temp, *_beginFinal))
      _beginFinal = temp;
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts a Particle into the %Configuration's final state using a hint.
 * \param hint An iterator used as a hint where the Particle should be inserted.
 * \param particle The Particle to be inserted.
 * \return An iterator pointing to the inserted Particle.
 *
 * A good \a hint can speed up the insertion, while a bad hint will have no
 * speed gain, in which case insertion requires logarithmic time, just as
 * insert(particle) would.
 */
inline Configuration::iterator Configuration::insertFinal(Configuration::iterator hint, const Particle& particle)
{
   iterator temp = _particles.insert(hint, particle);
   ++_sizeFinal;
   // update _beginFinal if necessary, relies on the fact that for identical
   // elements new ones are inserted behind existing ones unless the hint
   // pointed to an identical element
   if (_beginFinal == end() || _compare_type()(*temp, *_beginFinal)
      || (hint == _beginFinal && hint->id() == particle.id()))
      _beginFinal = temp;
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts a range of Particles into the %Configuration's final state.
 * \param first An iterator pointing to the first Particle in the range.
 * \param last  An iterator pointing beyond the last Particle in the range.
 *
 * Inserts the Particles from the range specified by [\a first,\a last). The
 * complexity is as for the analogous range constructor.
 *
 * \note This is a template function that works for generic iterators for which
 * \c *it is a Particle. All Particles are put into the final state.
 */
template <typename InputIterator>
inline void Configuration::insertFinal(InputIterator first, InputIterator last)
{
   size_t oldsize = size();
   for (; first != last; ++first)
      _particles.insert(end(), _status_particle(*first, _status_particle::kFinal));
   _sizeFinal += size() - oldsize;
   if (_initialB != end()) {
      _beginFinal = _initialB;
      ++_beginFinal;
   } else if (_initialA != end()) {
      _beginFinal = _initialA;
      ++_beginFinal;
   } else
      _beginFinal = begin();
}

//------------------------------------------------------------------------------
/**
 * \brief Removes the Particle from the %Configuration pointed to by \a it.
 * \param it An iterator pointing to the Particle to be erased.
 *
 * The specific initial- or final-state Particle pointed to by \a it is removed
 * from the %Configuration.
 *
 * \note It is the user's responsibility to ensure that \a it is a valid
 * iterator into the %Configuration.
 */
inline void Configuration::erase(Configuration::iterator it)
{
   if (it == _initialA)
      _initialA = end();
   else if (it == _initialB)
      _initialB = end();
   else {
      --_sizeFinal;
      if (it == _beginFinal)
         ++_beginFinal;
   }
   _particles.erase(it);
}

//------------------------------------------------------------------------------
/**
 * \brief Removes all Particles with some PdgID from the %Configuration.
 * \param id The PdgID to be erased.
 * \return The total number of Particles erased.
 *
 * This removes all instances of the PdgID from the Configuration's initial and
 * final states. The return value is equal to what a prior call of count(id)
 * would have given.
 */
inline size_t Configuration::erase(PdgID id)
{
   size_t temp = 0;
   if (_initialA != end() && _initialA->id() == id) {
      _particles.erase(_initialA);
      _initialA = end();
      ++temp;
   }
   if (_initialB != end() && _initialB->id() == id) {
      _particles.erase(_initialB);
      _initialB = end();
      ++temp;
   }
   temp += eraseFinal(id);
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief Removes a range of Particles from the %Configuration.
 * \param first An iterator pointing to the first Particle in the %Configuration to be erased.
 * \param last  An iterator pointing beyond the last Particle in the %Configuration to be erased.
 *
 * The Particles in the %Configuration specified by [\a first,\a last) are
 * removed. The complexity is linear in the number of Particles to be erased.
 *
 * \note It is the user's responsibility to ensure that [\a first, \a last) is
 * a valid iterator range into the %Configuration.
 */
inline void Configuration::erase(Configuration::iterator first, Configuration::iterator last)
{
   if (_initialA == first && first != last) {
      _particles.erase(_initialA);
      _initialA = end();
      ++first;
   }
   if (_initialB == first && first != last) {
      _particles.erase(_initialB);
      _initialB = end();
      ++first;
   }
   if (_beginFinal == first) {
      _beginFinal = last;
   }
   size_t oldsize = size();
   _particles.erase(first, last);
   _sizeFinal -= oldsize - size();
}

//------------------------------------------------------------------------------
/**
 * \brief Removes the %Configuration's initial-state Particle A.
 * \return \c true if initial state A was present, \c false otherwise.
 */
inline bool Configuration::eraseInitialA()
{
   if (_initialA != end()) {
      _particles.erase(_initialA);
      _initialA = end();
      return true;
   }
   return false;
}

//------------------------------------------------------------------------------
/**
 * \brief Removes the %Configuration's initial-state Particle B.
 * \return \c true if initial state B was present, \c false otherwise.
 */
inline bool Configuration::eraseInitialB()
{
   if (_initialB != end()) {
      _particles.erase(_initialB);
      _initialB = end();
      return true;
   }
   return false;
}

//------------------------------------------------------------------------------
/**
 * \brief Removes all Particles with some PdgID from the %Configuration's final state.
 * \param id The PdgID to be erased.
 * \return The total number of Particles erased from the final state.
 *
 * Ocurrences of \a id in the Configuration's initial state are not affected.
 * The return value is equal to what a prior call of countFinal(id) would have
 * given.
 */
inline size_t Configuration::eraseFinal(PdgID id)
{
   _lookupFinal._id = id;
   if (_beginFinal != end() && _beginFinal->id() == id) {
      _beginFinal = _particles.upper_bound(_lookupFinal);
   }
   size_t temp;
   temp = _particles.erase(_lookupFinal);
   _sizeFinal -= temp;
   return temp;
}

//------------------------------------------------------------------------------
// Removes all Particles from the Configuration.
inline void Configuration::clear()
{
   _particles.clear();
   _initialA = end();
   _initialB = end();
   _beginFinal = end();
   _sizeFinal = 0;
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------
/**
 * \brief Checks if the %Configuration is not equal to a Signature.
 * \param rhs The Signature to be compared to
 * \return \c true if the %Configuration's Particles do not have exactly the
 *         same IDs as \a rhs, \c false otherwise.
 *
 * Identical to !operator==(rhs)
 */
inline bool Configuration::operator!=(const Signature& rhs) const
{
   return !(*this == rhs);
}

//------------------------------------------------------------------------------
/**
 * \brief Applies a transformation to the momentum of each Particle in the %Configuration.
 * \param trans The transformation to be applied.
 * \return A reference to the %Configuration.
 *
 * \a trans can be any function object which implements the function evaluation
 * operator with the signature
 * \code
 * const FourVector operator()(const FourVector& operand)
 * \endcode
 *
 * The transformation is applied to each Particle's momentum four-vector in the
 * %Configuration.
 */
template <class Transformation>
Configuration& Configuration::apply(const Transformation& trans)
{
   // changing the particle momentum will not invalidate Configuration's integrity.
   for (Configuration::iterator it = begin(); it != end(); ++it) {
      const_cast<_status_particle&>(*it).p().apply(trans);
   }
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns a copy of %Configuration with a transformation applied to it.
 * \param trans The transformation to be applied.
 * \return A transformed copy of the %Configuration.
 *
 * This applies \a trans to a new copy of the %Configuration and returns the
 * result. The %Configuration itself is not changed. See apply() for more
 * details.
 */
template <class Transformation>
Configuration Configuration::transformed(const Transformation& trans) const
{
   Configuration result(*this);
   result.apply(trans);
   return result;
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %Configuration stream insertion operator.
 * \param out The output stream to be used.
 * \param config The %Configuration to be streamed.
 * \return A reference to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const Configuration& config)
{
   return config.print(out);
}

} // namespace Geneva

#endif // GENEVA_CONFIGURATION_HPP
