//------------------------------------------------------------------------------
/// \file ConfigurationExtensions.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interfaces of Configuration Extensions
//------------------------------------------------------------------------------

#ifndef GENEVA_CONFIGURATION_EXTENSIONS_HPP
#define GENEVA_CONFIGURATION_EXTENSIONS_HPP

#include "Geneva/Core/Physics/ConfigurationWrappers.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class PhiN
 * \ingroup comp
 *
 * \brief Container for a PhiN point.
 *
 * A simple container for the input to a splitting. It contains a const
 * reference to the underlying %Configuration and an iterator to the emitter.
 *
 * \note It is the user's responsiblity to maintain the validity of the
 * iterator.
 */
//------------------------------------------------------------------------------
class PhiN : public const_config_ref
{
   private:
      Configuration::iterator _emitter;

   public:
      typedef Configuration::iterator iterator;

      explicit PhiN(const Configuration& config)
         : const_config_ref(config), _emitter(config.end())
      {};

      PhiN(const Configuration& configCM, Configuration::iterator emitter)
         : const_config_ref(configCM), _emitter(emitter)
      {};

      iterator emitter() const { return _emitter; }
      iterator& emitter()      { return _emitter; }
};

//------------------------------------------------------------------------------
/**
 * \class PhiNCM
 * \ingroup comp
 *
 * \brief Container for a PhiN point in the CM frame.
 *
 * A simple container for the input to a splitting. It contains a const
 * reference to the underlying %Configuration and an iterator to the emitter,
 * and a value of the boost required to go to the lab frame.
 *
 * \note It is the user's responsiblity to maintain the validity of the
 * iterator and to ensure that the Configuration is in the partonic CM frame.
 */
//------------------------------------------------------------------------------
class PhiNCM : public const_config_ref
{
   private:
      Configuration::iterator _emitter;

      double _lambda;

   public:
      typedef Configuration::iterator iterator;

      explicit PhiNCM(const Configuration& configCM, double xA, double xB)
         : const_config_ref(configCM), _emitter(configCM.end()), _lambda(sqrt(xA/xB))
      {};

      PhiNCM(const Configuration& configCM, Configuration::iterator emitter, double xA, double xB)
         : const_config_ref(configCM), _emitter(emitter), _lambda(sqrt(xA/xB))
      {};

      iterator emitter() const { return _emitter; }
      iterator& emitter()      { return _emitter; }

      double lambda() const    { return _lambda;  }
};


//------------------------------------------------------------------------------
/**
 * \class PhiNplus1
 * \ingroup comp
 *
 * \brief Container for a PhiN+1 point.
 *
 * A simple container for the input to a projection. It contains a const
 * reference to the underlying %Configuration and iterators to the emitted and
 * sister partons.
 *
 * \note It is the user's responsiblity to manage the iterators and ensure their
 * validity.
 */
//------------------------------------------------------------------------------
class PhiNplus1 : public const_config_ref
{
   protected:
      iterator _emitted;
      iterator _sister;

   public:
      explicit PhiNplus1(const Configuration& config)
         : const_config_ref(config), _emitted(config.end()), _sister(config.end())
      {}

      PhiNplus1(const Configuration& config, iterator emitted, iterator sister)
         : const_config_ref(config), _emitted(emitted), _sister(sister)
      {}

      iterator emitted() const { return _emitted; }
      iterator sister() const  { return _sister;  }

      iterator& emitted()      { return _emitted; }
      iterator& sister()       { return _sister;  }
};


//------------------------------------------------------------------------------
/**
 * \class PhiNplus1CMWritable
 * \ingroup comp
 *
 * \brief Container for a PhiN+1 point including a writable Configuration in the CM frame
 *
 * Equivalent to PhiNplus1 but containing an explicit Configuration object
 * that can be modified. It is meant to store the result of a splitting. As such
 * it contains an additional member to hold the jacobian as well as convenience
 * members to insert emitted and sister particles.
 *
 * When no write access to the contained Configuration object is needed, it can
 * be used in place of a normal PhiNplus1.
 *
 * \note It is the user's responsiblity to manage the contained Configuration
 * and iterators.
 */
//------------------------------------------------------------------------------
class PhiNplus1CMWritable : public config_writable, public PhiNplus1
{
   private:
      double _xA;
      double _xB;
      double _jacobian;

   public:
      typedef Configuration::iterator iterator;

      PhiNplus1CMWritable()
         : PhiNplus1(config_writable::_config), _xA(0.0), _xB(0.0), _jacobian(1.0) {}

      double xA() const       { return _xA;       }
      double xB() const       { return _xB;       }
      double jacobian() const { return _jacobian; }

      double& xA()            { return _xA;       }
      double& xB()            { return _xB;       }
      double& jacobian()      { return _jacobian; }

      iterator insertEmitted(const Particle& particle)
      {
         return _emitted = _config.insertFinal(particle);
      }

      iterator insertFinalSister(const Particle& particle)
      {
         return _sister = _config.insertFinal(particle);
      }

      iterator setInitialASister(const Particle& particle)
      {
         return _sister = _config.setInitialA(particle);
      }

      iterator setInitialBSister(const Particle& particle)
      {
         return _sister = _config.setInitialB(particle);
      }
};

//------------------------------------------------------------------------------
/**
 * \class PhiNplus1Writable
 * \ingroup comp
 *
 * \brief Container for a PhiN+1 point including a writable Configuration
 *
 * Equivalent to PhiNplus1 but containing an explicit Configuration object
 * that can be modified. It is meant to store the result of a splitting. As such
 * it contains an additional member to hold the jacobian as well as convenience
 * members to insert emitted and sister particles.
 *
 * When no write access to the contained Configuration object is needed, it can
 * be used in place of a normal PhiNplus1.
 *
 * \note It is the user's responsiblity to manage the contained Configuration
 * and iterators.
 */
//------------------------------------------------------------------------------
class PhiNplus1Writable : public config_writable, public PhiNplus1
{
   private:
      double _jacobian;

   public:
      typedef Configuration::iterator iterator;

      PhiNplus1Writable()
         : PhiNplus1(config_writable::_config), _jacobian(1.0) {}

      double jacobian() const { return _jacobian; }

      double& jacobian()      { return _jacobian; }

      iterator insertEmitted(const Particle& particle)
      {
         return _emitted = _config.insertFinal(particle);
      }

      iterator insertFinalSister(const Particle& particle)
      {
         return _sister = _config.insertFinal(particle);
      }

      iterator setInitialASister(const Particle& particle)
      {
         return _sister = _config.setInitialA(particle);
      }

      iterator setInitialBSister(const Particle& particle)
      {
         return _sister = _config.setInitialB(particle);
      }
};

} // namespace Geneva

#endif // GENEVA_CONFIGURATION_EXTENSIONS_HPP
