//------------------------------------------------------------------------------
/// \file ConfigurationWrappers.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definitions of internal Configuration wrappers
//------------------------------------------------------------------------------

#ifndef GENEVA_CONFIGURATION_WRAPPERS_HPP
#define GENEVA_CONFIGURATION_WRAPPERS_HPP

#include "Geneva/Core/Physics/Configuration.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class const_config_ref
 * \ingroup Core
 *
 * \brief A wrapper around a reference to a const Configuration.
 *
 * This class behaves essentially like a const Configuration. It wraps a
 * reference to a const Configuration and exports all Configuration members
 * allowed by const access. See the documentation of Configuration for more
 * details of the available member functions.
 *
 * It supports implicit type casting to a const Configuration&, and so can be
 * used wherever an argument to a const Configuration is needed.
 *
 * The default copy constructor assigns the reference. The assignment operator
 * is not available to avoid the semantic ambiguities.
 *
 * \note This is an internal class and should not be used directly.
 */
//------------------------------------------------------------------------------
class const_config_ref
{
   private:
      const Configuration& _config_ref;

   public:
      /// \name Type Definitions
      ///@{
      typedef Configuration::const_iterator const_iterator;
      typedef Configuration::const_iterator iterator;
      ///@}

      /// \name Constructors etc.
      ///@{
      explicit const_config_ref(const Configuration& config)
         : _config_ref(config) {};

      /// Type cast operator to const Configuration&.
      operator const Configuration&() const { return _config_ref; }
      ///@}

      /// \name Accessing and Querying Particles
      ///@{
      /// Returns an iterator to the first Particle in the Configuration.
      iterator begin() const      { return _config_ref.begin(); }

      /// Returns an iterator pointing beyond the last Particle in the Configuration.
      iterator end() const        { return _config_ref.end(); }

      /// Returns an iterator to initial-state Particle A.
      iterator initialA() const   { return _config_ref.initialA(); }

      /// Returns an iterator to initial-state Particle B.
      iterator initialB() const   { return _config_ref.initialB(); }

      /// Returns an iterator to the first final-state Particle.
      iterator beginFinal() const { return _config_ref.beginFinal(); }

      /// Returns an iterator pointing beyond the last final-state Particle.
      iterator endFinal() const   { return _config_ref.endFinal(); }

      /// Returns \c true if the %Configuration is empty, \c false otherwise.
      bool empty() const          { return _config_ref.empty(); }

      /// Returns the total number of Particles in the %Configuration.
      size_t size() const         { return _config_ref.size(); }

      /// Returns the total number of Particles in the %Configuration's final state.
      size_t sizeFinal() const    { return _config_ref.sizeFinal(); }

      /// Returns the maximum number of Particles the %Configuration could hold.
      size_t max_size() const     { return _config_ref.max_size(); }

      /// Counts how often a Particle with some PdgID appears in the Configuration.
      size_t count(PdgID id) const                  { return _config_ref.count(id); }

      /// Counts how often a Particle with some PdgID appears in the Configuration's final state.
      size_t countFinal(PdgID id) const             { return _config_ref.countFinal(id); }

      /// Tries to find a Particle matching a Signature-like entry in the Configuration.
      template <typename SignatureLikeIterator>
      iterator find(SignatureLikeIterator it) const { return _config_ref.find(it); }

      /// Tries to find a Particle with a given PdgID in the Configuration.
      iterator find(PdgID id) const                 { return _config_ref.find(id); }

      /// Tries to find a Particle with a given PdgID in the Configuration's final state.
      iterator findFinal(PdgID id) const            { return _config_ref.findFinal(id); }

      /// Finds the range of Particles with some PdgID in the Configuration's final state.
      std::pair<iterator, iterator> equalRangeFinal(PdgID id) const { return _config_ref.equalRangeFinal(id); }

      /// Tests if a Particle matching a Signature-like entry is contained in the Configuration.
      template <typename SignatureLikeIterator>
      bool includes(SignatureLikeIterator it) const { return _config_ref.includes(it); }

      /// Tests if a Particle with some PdgID is contained in the Configuration.
      bool includes(PdgID id) const                 { return _config_ref.includes(id); }

      /// Tests if a Particle with some PdgID is contained in the Configuration's final state.
      bool includesFinal(PdgID id) const            { return _config_ref.includesFinal(id); }

      /// Tests if a Signature is contained in the Configuration.
      bool includes(const Signature& other) const   { return _config_ref.includes(other); }

      /// Tests if a Signature intersects with the Configuration.
      bool intersects(const Signature& other) const { return _config_ref.intersects(other); }
      ///@}

      /// \name Algebra
      ///@{
      /// Checks if the Configuration is equal to a Signature.
      bool operator==(const Signature& rhs) const   { return _config_ref.operator==(rhs); }

      /// Checks if the Configuration is not equal to a Signature.
      bool operator!=(const Signature& rhs) const   { return _config_ref.operator!=(rhs); }

      /// Returns a copy of %Configuration with a transformation applied to it.
      template <class Transformation>
      Configuration transformed(const Transformation& trans) const { return _config_ref.transformed(trans); }
      ///@}
};


//------------------------------------------------------------------------------
/**
 * \class config_writable
 * \ingroup Core
 *
 * \brief A wrapper around an explicitly contained Configuration.
 *
 * This class wraps an explicitly contained Configuration and exports all
 * Configuration members requiring non const access. See the documentation of
 * Configuration for more details of its available member functions.
 *
 * It supports implicit type casting to a Configuration&, and so can be used
 * wherever an argument to a non-const Configuration is needed. The contained
 * Configuration can also be explicitly accessed using the config() member.
 *
 * \note This is an internal class and should not be used directly.
 */
//------------------------------------------------------------------------------
class config_writable
{
   protected:
      Configuration _config;

   public:
      /// \name Type Definitions
      ///@{
      typedef Configuration::const_iterator const_iterator;
      typedef Configuration::const_iterator iterator;
      ///@}

      /// Returns a non-const reference to the contained Configuration object.
      Configuration& config()   { return _config; }

      /// Type cast operator to Configuration&.
      operator Configuration&() { return _config; }

      /// \name Adding and Removing Particles
      ///@{
      /// Inserts an initial- or final-state Particle into the Configuration.
      template <typename ConfigLikeIterator>
      iterator insert(ConfigLikeIterator it)         { return _config.insert(it); }

      /// Inserts a range of Particles into the Configuration.
      template <typename ConfigLikeIterator>
      void insert(ConfigLikeIterator first, ConfigLikeIterator last) { return _config.insert(first, last); }

      /// Sets the initial-state Particle A of the Configuration.
      iterator setInitialA(const Particle& particle) { return _config.setInitialA(particle); }

      /// Sets the initial-state Particle B of the Configuration.
      iterator setInitialB(const Particle& particle) { return _config.setInitialB(particle); }

      /// Inserts a Particle into the Configuration's final state.
      iterator insertFinal(const Particle& particle) { return _config.insertFinal(particle); }

      /// Inserts a Particle into the Configuration's final state using a hint.
      iterator insertFinal(iterator hint, const Particle& particle) { return _config.insertFinal(hint, particle); }

      /// Inserts a range of Particles into the Configuration's final state.
      template <typename InputIterator>
      void insertFinal(InputIterator first, InputIterator last) { return _config.insertFinal(first, last); }

      /// Removes the Particle from the Configuration pointed to by it.
      void erase(iterator it)                       { return _config.erase(it); }

      /// Removes all Particles with some PdgID from the Configuration.
      size_t erase(PdgID id)                        { return _config.erase(id); }

      /// Removes a range of Particles from the Configuration.
      void erase(iterator first, iterator last)     { return _config.erase(first, last); }

      /// Removes the Configuration's initial-state Particle A.
      bool eraseInitialA()                          { return _config.eraseInitialA(); }

      /// Removes the Configuration's initial-state Particle B.
      bool eraseInitialB()                          { return _config.eraseInitialB(); }

      /// Removes all Particles with some PdgID from the Configuration's final state.
      size_t eraseFinal(PdgID id)                   { return _config.eraseFinal(id); };

      /// Removes all Particles from the %Configuration.
      void clear()                                  { return _config.clear(); }
      ///@}

      /// \name Algebra
      ///@{
      // Applies a transformation to the momentum of each Particle in the Configuration.
      template <class Transformation>
      Configuration& apply(const Transformation& trans) { return _config.apply(trans); }
      ///@}
};

} // namespace Geneva

#endif // GENEVA_CONFIGURATION_WRAPPERS_HPP
