//------------------------------------------------------------------------------
/// \file Emission.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Emission
//------------------------------------------------------------------------------

#ifndef GENEVA_EMISSION_HPP
#define GENEVA_EMISSION_HPP

#include "Geneva/Core/Physics/PdgID.hpp"

#include <iosfwd>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class Emission
 * \ingroup Core
 *
 * \brief Container for the flavor content of an emission.
 *
 * A container holding IDs for the mother (emitter), emitted, and sister
 * particles of an emission, with the convention:
 *
 * - ISR: sister -> (mother) + emitted
 * - FSR: (mother) -> sister + emitted
 *
 * The logical distinction between sister and emitted for FSR is up to the user.
 * (For example, the emitted sister should be treated as the "FKS" parton.)
 *
 * The emission type is specified by the #Type parameter with values \c kISRA,
 * \c kISRB, and \c kFSR.
 *
 * \note Currently the ID is just the PdgID, later it could contain the complete
 * set of quantum numbers (e.g. helicities) that are required to specify the
 * corresponding particle.
 */
//------------------------------------------------------------------------------
class Emission
{
   public:
      enum Type
      {
         kFSR = 0,
         kISRA = 1,
         kISRB = 2
      };

   private:
      Type _type;                      ///< Type of the emission

      PdgID _mother;                   ///< PdgID of the mother
      PdgID _emitted;                  ///< PdgID of the emitted particle
      PdgID _sister;                   ///< PdgID of the sister particle

   public:
      /// \name Constructors etc.
      ///@{

      /// Creates an %Emission of the given \a type with all IDs set to kUnknown.
      Emission(Type type = kFSR);

      /// Creates an %Emission of the given \a type with the given IDs.
      Emission(Type type, PdgID mother, PdgID emitted, PdgID sister);

      /// Creates a QCD emission from the given \a emitted and \a sister.
      template <typename SignatureEntry1, typename SignatureEntry2>
      Emission(const SignatureEntry1& emitted, const SignatureEntry2& sister);

      ///@}

      /// \name Queries and Accessors
      ///@{
      Type type() const;               ///< Returns the radiation type.

      PdgID mother() const;          ///< Returns the PdgID of the mother.
      PdgID emitted() const;         ///< Returns the PdgID of the emitted particle.
      PdgID sister() const;          ///< Returns the PdgID of the sister particle.

      Type& type();                    ///< Access the radiation type.

      PdgID& mother();               ///< Access the PdgID of the mother.
      PdgID& emitted();              ///< Access the PdgID of the emitted particle.
      PdgID& sister();               ///< Access the PdgID of the sister particle.
      ///@}

      /// \name Comparisons
      ///@{
      // Test if the Emission is FSR.
      bool isFSR() const;

      // Test if the Emission is ISR.
      bool isISR() const;

      // Test if the Emission is ISRA.
      bool isISRA() const;

      // Test if the Emission is ISRB.
      bool isISRB() const;

      // Emission equality comparison operator
      bool operator==(const Emission& rhs) const;

      // Emission inequality comparison operator
      bool operator!=(const Emission& rhs) const;
      ///@}

      /// \name Output
      ///@{
      // Emission stream insertion operator
      friend std::ostream& operator<<(std::ostream& out, const Emission& emission);
      ///@}

      /// Obtains the QCD mother id from \a emitted and \a sister.
      static PdgID getMotherQCD(PdgID emitted, PdgID sister, bool isFSR = true);

   protected:
      /// Streams a string in the format "sister -> (mother) emitted" (ISR) or "(mother) -> sister emitted" (FSR) to \a out.
      std::ostream& print(std::ostream& out) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------
// Creates a default Emission.
inline Emission::Emission(Type type)
   : _type(type), _mother(), _emitted(), _sister()
{}

//------------------------------------------------------------------------------
// Creates a Emission from the given arguments.
inline Emission::Emission(Type type, PdgID mother, PdgID emitted, PdgID sister)
   : _type(type), _mother(mother), _emitted(emitted), _sister(sister)
{}

//------------------------------------------------------------------------------
/**
 * \brief Creates an %Emission from \a emitted and \a sister.
 *
 * This uses \c isFinal(), \c isInitialA(), \c isInitialB() on the sister to
 * determine the emission type, and therefore can only be used with
 * Configuration or Signature entries as \c Emission(*it1, *it2), where
 * it1 and it2 are valid Configuration or Signature iterators.
 *
 * \note This asserts that emitted and sister correspond to a valid QCD
 * splitting.
 */
template <typename SignatureEntry1, typename SignatureEntry2>
Emission::Emission(const SignatureEntry1& emitted, const SignatureEntry2& sister)
   : _emitted(emitted.id()), _sister(sister.id())
{
   if (sister.isFinal())
      _type = kFSR;
   else if (sister.isInitialA())
      _type = kISRA;
   else
      _type = kISRB;

   _mother = getMotherQCD(_emitted, _sister, _type == kFSR);
}


//------------------------------------------------------------------------------
// Queries and Accessors
//------------------------------------------------------------------------------

inline Emission::Type Emission::type() const { return _type;    }
inline PdgID Emission::mother() const        { return _mother;  }
inline PdgID Emission::emitted() const       { return _emitted; }
inline PdgID Emission::sister() const        { return _sister;  }

inline Emission::Type& Emission::type()      { return _type;    }
inline PdgID& Emission::mother()             { return _mother;  }
inline PdgID& Emission::emitted()            { return _emitted; }
inline PdgID& Emission::sister()             { return _sister;  }

//------------------------------------------------------------------------------
// Comparisons
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Test if the %Emission is FSR.
 * \return \c true if %Emission is FSR, \c false otherwise.
 *
 * Equivalent to type() == \c kFSR.
 */
inline bool Emission::isFSR() const
{
   return _type == kFSR;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if the %Emission is ISR.
 * \return \c true if %Emission is ISR, \c false otherwise.
 *
 * Equivalent to type() != \c kFSR, so it returns \c true for both ISRA and ISRB
 * emissions, and \c false for FSR.
 */
inline bool Emission::isISR() const
{
   return _type != kFSR;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if the %Emission is ISRA.
 * \return \c true if %Emission is ISRA, \c false otherwise.
 *
 * Equivalent to type() == \c kISRA.
 */
inline bool Emission::isISRA() const
{
   return _type == kISRA;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if the %Emission is ISRB.
 * \return \c true if %Emission is ISRB, \c false otherwise.
 *
 * Equivalent to type() == \c kISRB.
 */
inline bool Emission::isISRB() const
{
   return _type == kISRB;
}

//------------------------------------------------------------------------------
/**
 * \brief Emission equality comparison operator
 * \param rhs The %Emission to compare to.
 * \return \c true if type and all IDs are equal, \c false otherwise.
 */
inline bool Emission::operator==(const Emission& rhs) const
{
   return _type == rhs._type && _mother == rhs._mother
          && _emitted == rhs._emitted && _sister == rhs._sister;
}

//------------------------------------------------------------------------------
/**
 * \brief Emission inequality comparison operator
 * \param rhs The %Emission to compare to.
 * \return \c true if either the type or IDs are not equal, \c false otherwise.
 *
 * Equivalent to \c !operator==(rhs).
 */
inline bool Emission::operator!=(const Emission& rhs) const
{
   return !operator==(rhs);
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------
/**
 * \brief %Emission stream insertion operator
 * \param out The output stream to be used.
 * \param Emission The %Emission to be streamed.
 * \return A reference to \a out.
 *
 * Streams the info about Emission to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const Emission& Emission)
{
   return Emission.print(out);
}

} // namespace Geneva

#endif // GENEVA_PHI_RAD_FKS_HPP
