//------------------------------------------------------------------------------
/// \file Event.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Event, which is a simple placeholder for a better Event class (eg, HepMC)
//------------------------------------------------------------------------------

#ifndef GENEVA_EVENT_HPP
#define GENEVA_EVENT_HPP

#include "Geneva/Core/Physics/Configuration.hpp"
#include "Geneva/Core/Physics/Sigma.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class Event
 * \ingroup Core
 *
 * \brief Holds the basic information of a Geneva event.
 *
 * An %Event consists of a phase-space point given by a Configuration, and a
 * weighted cross section (in units of pb) given by a Sigma, which holds the
 * cross section central value (and possibly variations) for that phase-space
 * weighted (i.e. divided by) the generation weight.
 */
//------------------------------------------------------------------------------
class Event
{
   protected:
      Configuration _config;                ///< the %Event's Configuration
      Sigma _weightedSigma;                 ///< the weighted cross section

   public:
      /// \name Constructors etc.
      ///@{
      Event() = default;
      Event(const Configuration& config, const Sigma& weightedSigma);
      ///@}

      /// \name Accessing and Querying
      ///@{
      const Configuration& config() const;  ///< Returns the %Event's Configuration.
      const Sigma& weightedSigma() const;   ///< Returns the %Event's weighted cross section
//      double genWeight() const;             ///< Returns the %Event's generation weight.

      // Gives access to the Event's Configuration.
      Configuration& config();

      // Sets the cross section and weight information.
      void setSigma(const Sigma& sigma);

      // Clears the cross section and weight information.
      void clearSigma();
      ///@}
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------
/**
 * \brief Constructs %event from a given Configuration and weight information
 * \param config The %Event's configuration
 * \param weightedSigma The final weighted cross section for the %Event
 *
 * The value for \a weightedSigma should already include the generation weight.
 */
inline Event::Event(const Configuration& config, const Sigma& weightedSigma)
   : _config(config), _weightedSigma(weightedSigma)
{}

//------------------------------------------------------------------------------
// Accessing and Querying
//------------------------------------------------------------------------------
inline const Configuration& Event::config() const { return _config;        }
inline const Sigma& Event::weightedSigma() const  { return _weightedSigma; }

//------------------------------------------------------------------------------
/**
 * \brief Gives access to the %Event's Configuration.
 * \return A reference to the %Event's Configuration.
 *
 * The event only acts as a container for its underlying Configuration. The user
 * is responsible for managing and filling the Configuration.
 */
inline Configuration& Event::config()
{
   return _config;
}

//------------------------------------------------------------------------------
/**
 * \brief Sets the cross section and weight information.
 * \param sigma The cross section in units of pb.
 *
 * Here, \a sigma is equal to the nominal cross section as returned by a calculation divided
 * by the weight with which the
 * %Event's Configuration was generated
 */
inline void Event::setSigma(const Sigma& sigma)
{
   _weightedSigma = sigma;
}

//------------------------------------------------------------------------------
/**
 * \brief Clears the cross section and weight information.
 *
 * This clears the %Event's cross section (which sets it to undefined) and sets
 * the generation weight to 1.0
 */
inline void Event::clearSigma()
{
   _weightedSigma.clear();
}

} // namespace Geneva

#endif // GENEVA_EVENT_HPP
