//------------------------------------------------------------------------------
/// \file FOConv.hpp
//
// Author(s):
//    Jon Walsh
//
// Copyright:
//    Copyright (C) 2013 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definitions of fixed order convolution functions for the RGE
//------------------------------------------------------------------------------

#ifndef GENEVA_FO_CONV_HPP
#define GENEVA_FO_CONV_HPP

#include "Geneva/Core/Utils/MathUtils.hpp"

#include <vector>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \struct FOConv
 * \ingroup Core
 *
 * \brief Definitions of functions needed for fixed order convolutions in cumulant, spectrum
 */
//------------------------------------------------------------------------------
class FOConv
{
   private:
      double _eta;            ///< eta value for convolutions
      double _psi0_1peta;     ///< psi0(1+eta)
      double _psi1_1peta;     ///< psi1(1+eta)
      double _psi2_1peta;     ///< psi2(1+eta)
      double _psi3_1peta;     ///< psi3(1+eta)

   public:
      /// constructor, destructor
      FOConv(double eta) : _eta(eta), _psi0_1peta(psi0(1. + eta)), _psi1_1peta(psi1(1. + eta)), _psi2_1peta(psin(2,1. + eta)), _psi3_1peta(psin(3,1. + eta)) {}

      // cumulant functions
      // every function is also implemented as static

      /// function coming from 1 derivative d_eta in cumulant
      static double d1Cumulant(double eta)
      {
         return psi0(1. + eta) + math_constants::gammaE();
      }
      double d1Cumulant()
      {
         return _psi0_1peta + math_constants::gammaE();
      }

      /// function coming from 2 derivatives d_eta^2 in cumulant
      static double d2Cumulant(double eta)
      {
         double d1 = d1Cumulant(eta);
         return d1*d1 - psi1(1. + eta);
      }
      double d2Cumulant()
      {
         double d1 = d1Cumulant();
         return d1*d1 - _psi1_1peta;
      }

      /// function coming from 3 derivatives d_eta^3 in cumulant
      static double d3Cumulant(double eta)
      {
         double d1 = d1Cumulant(eta);
         return d1*d1*d1 - 3.*d1*psi1(1. + eta) + psin(2,1. + eta);
      }
      double d3Cumulant()
      {
         double d1 = d1Cumulant();
         return d1*d1*d1 - 3.*d1*_psi1_1peta + _psi2_1peta;
      }

      /// function coming from 4 derivatives d_eta^4 in cumulant
      static double d4Cumulant(double eta)
      {
         double d1 = d1Cumulant(eta);
         return d1*d1*d1*d1 - 6.*d1*d1*psi1(1. + eta) + 3.*psi1(1. + eta)*psi1(1. + eta) + 4*d1*psin(2,1. + eta) - psin(3,1. + eta);
      }
      double d4Cumulant()
      {
         double d1 = d1Cumulant();
         return d1*d1*d1*d1 - 6.*d1*d1*_psi1_1peta + 3.*_psi1_1peta*_psi1_1peta + 4*d1*_psi2_1peta - _psi3_1peta;
      }

      /// function for the 1 derivative d_eta + L in cumulant, for a single FO function
      static double d1CumulantFO(double eta, double lnxF)
      {
         return -1.*(d1Cumulant(eta) + lnxF);
      }
      double d1CumulantFO(double lnxF)
      {
         return -1.*(d1Cumulant() + lnxF);
      }

      /// function for the 2 derivatives (d_eta + L)^2 in cumulant, for a single FO function
      static double d2CumulantFO(double eta, double lnxF)
      {
         double d2shift = d2Cumulant(eta) + 2.*lnxF*d1Cumulant(eta) + lnxF*lnxF;
         return (1./2.) * (d2shift + math_constants::zeta2());
      }
      double d2CumulantFO(double lnxF)
      {
         double d2shift = d2Cumulant() + 2.*lnxF*d1Cumulant() + lnxF*lnxF;
         return (1./2.) * (d2shift + math_constants::zeta2());
      }

      /// function for the 3 derivatives (d_eta + L)^3 in cumulant, for a single FO function
      static double d3CumulantFO(double eta, double lnxF)
      {
         double d3shift = d3Cumulant(eta) + 3.*lnxF*d2Cumulant(eta) + 3.*lnxF*lnxF*d1Cumulant(eta) + lnxF*lnxF*lnxF;
         double d1shift = d1Cumulant(eta) + lnxF;
         return (1./3.) * (-1.*d3shift - 3.*math_constants::zeta2()*d1shift - 2*math_constants::zeta3());
      }
      double d3CumulantFO(double lnxF)
      {
         double d3shift = d3Cumulant() + 3.*lnxF*d2Cumulant() + 3.*lnxF*lnxF*d1Cumulant() + lnxF*lnxF*lnxF;
         double d1shift = d1Cumulant() + lnxF;
         return (1./3.) * (-1.*d3shift - 3.*math_constants::zeta2()*d1shift - 2*math_constants::zeta3());
      }

      /// function for the 4 derivatives (d_eta + L)^4 in cumulant, for a single FO function
      static double d4CumulantFO(double eta, double lnxF)
      {
         double d4shift = d4Cumulant(eta) + 4.*lnxF*d3Cumulant(eta) + 6.*lnxF*lnxF*d2Cumulant(eta) + 4.*lnxF*lnxF*lnxF*d1Cumulant(eta) + lnxF*lnxF*lnxF*lnxF;
         double d2shift = d2Cumulant(eta) + 2.*lnxF*d1Cumulant(eta) + lnxF*lnxF;
         double d1shift = d1Cumulant(eta) + lnxF;
         return (1./4.) * (d4shift  + 6.*math_constants::zeta2()*d2shift + 8.*math_constants::zeta3()*d1shift + (3./20.)*math_constants::pi4());
      }
      double d4CumulantFO(double lnxF)
      {
         double d4shift = d4Cumulant() + 4.*lnxF*d3Cumulant() + 6.*lnxF*lnxF*d2Cumulant() + 4.*lnxF*lnxF*lnxF*d1Cumulant() + lnxF*lnxF*lnxF*lnxF;
         double d2shift = d2Cumulant() + 2.*lnxF*d1Cumulant() + lnxF*lnxF;
         double d1shift = d1Cumulant() + lnxF;
         return (1./4.) * (d4shift  + 6.*math_constants::zeta2()*d2shift + 8.*math_constants::zeta3()*d1shift + (3./20.)*math_constants::pi4());
      }

      /// function for the 1,1 derivative (d_eta + L_a)(d_eta + L_b) in cumulant, for mixing between two FO functions
      static double d11mixCumulantFO(double eta, double lnxA, double lnxB)
      {
         return d2Cumulant(eta) + d1Cumulant(eta)*(lnxA + lnxB) + lnxA*lnxB;
      }
      double d11mixCumulantFO(double lnxA, double lnxB)
      {
         return d2Cumulant() + d1Cumulant()*(lnxA + lnxB) + lnxA*lnxB;
      }

      /// function for the 2,1 derivative (d_eta + L_a)^2 (d_eta + L_b) in cumulant, for mixing between two FO functions
      static double d21mixCumulantFO(double eta, double lnxA, double lnxB)
      {
         double d21mixshift = d3Cumulant(eta) + d2Cumulant(eta)*(2.*lnxA + lnxB) + d1Cumulant(eta)*(lnxA*lnxA + 2.*lnxA*lnxB) + lnxA*lnxA*lnxB;
         double d1shiftB = d1Cumulant(eta) + lnxB;
         return (-1./2.) * (d21mixshift + math_constants::zeta2()*d1shiftB);
      }
      double d21mixCumulantFO(double lnxA, double lnxB)
      {
         double d21mixshift = d3Cumulant() + d2Cumulant()*(2.*lnxA + lnxB) + d1Cumulant()*(lnxA*lnxA + 2.*lnxA*lnxB) + lnxA*lnxA*lnxB;
         double d1shiftB = d1Cumulant() + lnxB;
         return (-1./2.) * (d21mixshift + math_constants::zeta2()*d1shiftB);
      }

      /// function for the 2,2 derivative (d_eta + L_a)^2 (d_eta + L_b)^2 in cumulant, for mixing between two FO functions
      static double d22mixCumulantFO(double eta, double lnxA, double lnxB)
      {
         double d22mixshift = d4Cumulant(eta) + d3Cumulant(eta)*2.*(lnxA + lnxB) + d2Cumulant(eta)*(lnxA*lnxA + lnxB*lnxB + 4.*lnxA*lnxB) + d1Cumulant(eta)*2.*lnxA*lnxB*(lnxA + lnxB) + lnxA*lnxA*lnxB*lnxB;
         double d2shiftA = d2Cumulant(eta) + 2.*lnxA*d1Cumulant(eta) + lnxA*lnxA;
         double d2shiftB = d2Cumulant(eta) + 2.*lnxB*d1Cumulant(eta) + lnxB*lnxB;
         return (1./4.) * (d22mixshift + math_constants::zeta2()*(d2shiftA + d2shiftB) + (math_constants::pi4())/36.);
      }
      double d22mixCumulantFO(double lnxA, double lnxB)
      {
         double d22mixshift = d4Cumulant() + d3Cumulant()*2.*(lnxA + lnxB) + d2Cumulant()*(lnxA*lnxA + lnxB*lnxB + 4.*lnxA*lnxB) + d1Cumulant()*2.*lnxA*lnxB*(lnxA + lnxB) + lnxA*lnxA*lnxB*lnxB;
         double d2shiftA = d2Cumulant() + 2.*lnxA*d1Cumulant() + lnxA*lnxA;
         double d2shiftB = d2Cumulant() + 2.*lnxB*d1Cumulant() + lnxB*lnxB;
         return (1./4.) * (d22mixshift + math_constants::zeta2()*(d2shiftA + d2shiftB) + (math_constants::pi4())/36.);
      }

      /// NLO cumulant FO function in terms of a vector of coefficients of L_n distributions.  Gives the coefficient of (alpha_s/4pi)
      static double funcCumulantNLO(std::vector<double> coeffs, double eta, double lnxF)
      {
         double Lm1coeff = coeffs[0];
         double L0coeff = coeffs[1];
         double L1coeff = coeffs[2];

         return L1coeff*d2CumulantFO(eta, lnxF) + L0coeff*d1CumulantFO(eta, lnxF) + Lm1coeff;
      }
      double funcCumulantNLO(std::vector<double> coeffs, double lnxF)
      {
         double Lm1coeff = coeffs[0];
         double L0coeff = coeffs[1];
         double L1coeff = coeffs[2];

         return L1coeff*d2CumulantFO(lnxF) + L0coeff*d1CumulantFO(lnxF) + Lm1coeff;
      }

      /// NNLO cumulant FO function in terms of a vector of coefficients of L_n distributions.  Gives the coefficient of (alpha_s/4pi)^2
      static double funcCumulantNNLO(std::vector<double> coeffs, double eta, double lnxF)
      {
         double Lm1coeff = coeffs[0];
         double L0coeff = coeffs[1];
         double L1coeff = coeffs[2];
         double L2coeff = coeffs[3];
         double L3coeff = coeffs[4];

         return L3coeff*d4CumulantFO(eta, lnxF) + L2coeff*d3CumulantFO(eta, lnxF) + L1coeff*d2CumulantFO(eta, lnxF) + L0coeff*d1CumulantFO(eta, lnxF) + Lm1coeff;
      }
      double funcCumulantNNLO(std::vector<double> coeffs, double lnxF)
      {
         double Lm1coeff = coeffs[0];
         double L0coeff = coeffs[1];
         double L1coeff = coeffs[2];
         double L2coeff = coeffs[3];
         double L3coeff = coeffs[4];

         return L3coeff*d4CumulantFO(lnxF) + L2coeff*d3CumulantFO(lnxF) + L1coeff*d2CumulantFO(lnxF) + L0coeff*d1CumulantFO(lnxF) + Lm1coeff;
      }

      /// mixing between NLO cumulant FO functions, in terms of vectors of coefficients of L_n distributions in each function.  Gives the coefficient of (alpha_s/4pi)^2
      static double funcCumulantmixNLO(std::vector<double> coeffsA, std::vector<double> coeffsB, double eta, double lnxA, double lnxB)
      {
         double Lm1coeffA = coeffsA[0];
         double L0coeffA = coeffsA[1];
         double L1coeffA = coeffsA[2];
         double Lm1coeffB = coeffsB[0];
         double L0coeffB = coeffsB[1];
         double L1coeffB = coeffsB[2];

         return L1coeffA*L1coeffB*d22mixCumulantFO(eta, lnxA, lnxB) + L1coeffA*L0coeffB*d21mixCumulantFO(eta, lnxA, lnxB) + L1coeffB*L0coeffA*d21mixCumulantFO(eta, lnxB, lnxA)
                  + L0coeffA*L0coeffB*d11mixCumulantFO(eta, lnxA, lnxB) + L1coeffA*Lm1coeffB*d2CumulantFO(eta, lnxA) + L1coeffB*Lm1coeffA*d2CumulantFO(eta, lnxB)
                  + L0coeffA*Lm1coeffB*d1CumulantFO(eta, lnxA) + L0coeffB*Lm1coeffA*d1CumulantFO(eta, lnxB) + Lm1coeffA*Lm1coeffB;
      }
      double funcCumulantmixNLO(std::vector<double> coeffsA, std::vector<double> coeffsB, double lnxA, double lnxB)
      {
         double Lm1coeffA = coeffsA[0];
         double L0coeffA = coeffsA[1];
         double L1coeffA = coeffsA[2];
         double Lm1coeffB = coeffsB[0];
         double L0coeffB = coeffsB[1];
         double L1coeffB = coeffsB[2];

         return L1coeffA*L1coeffB*d22mixCumulantFO(lnxA, lnxB) + L1coeffA*L0coeffB*d21mixCumulantFO(lnxA, lnxB) + L1coeffB*L0coeffA*d21mixCumulantFO(lnxB, lnxA)
                  + L0coeffA*L0coeffB*d11mixCumulantFO(lnxA, lnxB) + L1coeffA*Lm1coeffB*d2CumulantFO(lnxA) + L1coeffB*Lm1coeffA*d2CumulantFO(lnxB)
                  + L0coeffA*Lm1coeffB*d1CumulantFO(lnxA) + L0coeffB*Lm1coeffA*d1CumulantFO(lnxB) + Lm1coeffA*Lm1coeffB;
      }

      // spectrum functions

      /// function for the 1 derivative d_eta + L in distribution, for a single FO function
      static double d1DistFO(double eta, double lnxF)
      {
         return 1 + eta*d1CumulantFO(eta, lnxF);
      }
      double d1DistFO(double lnxF)
      {
         return 1 + _eta*d1CumulantFO(lnxF);
      }

      /// function for the 2 derivatives (d_eta + L)^2 in distribution, for a single FO function
      static double d2DistFO(double eta, double lnxF)
      {
         return d1CumulantFO(eta, lnxF) + eta*d2CumulantFO(eta, lnxF);
      }
      double d2DistFO(double lnxF)
      {
         return d1CumulantFO(lnxF) + _eta*d2CumulantFO(lnxF);
      }

      /// function for the 3 derivatives (d_eta + L)^3 in distribution, for a single FO function
      static double d3DistFO(double eta, double lnxF)
      {
         return 2.*d2CumulantFO(eta, lnxF) + eta*d3CumulantFO(eta, lnxF);
      }
      double d3DistFO(double lnxF)
      {
         return 2.*d2CumulantFO(lnxF) + _eta*d3CumulantFO(lnxF);
      }

      /// function for the 4 derivatives (d_eta + L)^4 in distribution, for a single FO function
      static double d4DistFO(double eta, double lnxF)
      {
         return 3.*d3CumulantFO(eta, lnxF) + eta*d4CumulantFO(eta, lnxF);
      }
      double d4DistFO(double lnxF)
      {
         return 3.*d3CumulantFO(lnxF) + _eta*d4CumulantFO(lnxF);
      }

      /// function for the 1,1 derivative (d_eta + L_a)(d_eta + L_b) in distribution, for mixing between two FO functions
      static double d11mixDistFO(double eta, double lnxA, double lnxB)
      {
         return eta * d11mixCumulantFO(eta, lnxA, lnxB) + d1CumulantFO(eta, lnxA) + d1CumulantFO(eta, lnxB);
      }
      double d11mixDistFO(double lnxA, double lnxB)
      {
         return _eta * d11mixCumulantFO(lnxA, lnxB) + d1CumulantFO(lnxA) + d1CumulantFO(lnxB);
      }

      /// function for the 2,1 derivative (d_eta + L_a)^2 (d_eta + L_b) in distribution, for mixing between two FO functions
      static double d21mixDistFO(double eta, double lnxA, double lnxB)
      {
         return eta * d21mixCumulantFO(eta, lnxA, lnxB) + 3.*d2CumulantFO(eta, lnxA) + (lnxA - lnxB)*d1CumulantFO(eta, lnxA) - math_constants::zeta2();
      }
      double d21mixDistFO(double lnxA, double lnxB)
      {
         return _eta * d21mixCumulantFO(lnxA, lnxB) + 3.*d2CumulantFO(lnxA) + (lnxA - lnxB)*d1CumulantFO(lnxA) - math_constants::zeta2();
      }

      /// function for the 2,2 derivative (d_eta + L_a)^2 (d_eta + L_b)^2 in distribution, for mixing between two FO functions
      static double d22mixDistFO(double eta, double lnxA, double lnxB)
      {
         return eta * d22mixCumulantFO(eta, lnxA, lnxB) + (3./2.)*(d3CumulantFO(eta, lnxA) + d3CumulantFO(eta, lnxB))
                  - ((1./2.)*(lnxA - lnxB)*(lnxA - lnxB) + math_constants::zeta2())*(d1CumulantFO(eta, lnxA) + d1CumulantFO(eta, lnxB)) + 2.*math_constants::zeta3();
      }
      double d22mixDistFO(double lnxA, double lnxB)
      {
         return _eta * d22mixCumulantFO(lnxA, lnxB) + (3./2.)*(d3CumulantFO(lnxA) + d3CumulantFO(lnxB))
                  - ((1./2.)*(lnxA - lnxB)*(lnxA - lnxB) + math_constants::zeta2())*(d1CumulantFO(lnxA) + d1CumulantFO(lnxB)) + 2.*math_constants::zeta3();
      }

      /// NLO distribution FO function in terms of a vector of coefficients of L_n distributions.  Gives the coefficient of (alpha_s/4pi)
      static double funcDistNLO(std::vector<double> coeffs, double eta, double lnxF)
      {
         double Lm1coeff = coeffs[0];
         double L0coeff = coeffs[1];
         double L1coeff = coeffs[2];

         return L1coeff*d2DistFO(eta, lnxF) + L0coeff*d1DistFO(eta, lnxF) + Lm1coeff * eta;
      }
      double funcDistNLO(std::vector<double> coeffs, double lnxF)
      {
         double Lm1coeff = coeffs[0];
         double L0coeff = coeffs[1];
         double L1coeff = coeffs[2];

         return L1coeff*d2DistFO(lnxF) + L0coeff*d1DistFO(lnxF) + Lm1coeff * _eta;
      }

      /// NNLO distribution FO function in terms of a vector of coefficients of L_n distributions.  Gives the coefficient of (alpha_s/4pi)^2
      static double funcDistNNLO(std::vector<double> coeffs, double eta, double lnxF)
      {
         double Lm1coeff = coeffs[0];
         double L0coeff = coeffs[1];
         double L1coeff = coeffs[2];
         double L2coeff = coeffs[3];
         double L3coeff = coeffs[4];

         return L3coeff*d4DistFO(eta, lnxF) + L2coeff*d3DistFO(eta, lnxF) + L1coeff*d2DistFO(eta, lnxF) + L0coeff*d1DistFO(eta, lnxF) + Lm1coeff*eta;
      }
      double funcDistNNLO(std::vector<double> coeffs, double lnxF)
      {
         double Lm1coeff = coeffs[0];
         double L0coeff = coeffs[1];
         double L1coeff = coeffs[2];
         double L2coeff = coeffs[3];
         double L3coeff = coeffs[4];

         return L3coeff*d4DistFO(lnxF) + L2coeff*d3DistFO(lnxF) + L1coeff*d2DistFO(lnxF) + L0coeff*d1DistFO(lnxF) + Lm1coeff*_eta;
      }

      /// mixing between NLO distribution FO functions, in terms of vectors of coefficients of L_n distributions in each function.  Gives the coefficient of (alpha_s/4pi)^2
      static double funcDistmixNLO(std::vector<double> coeffsA, std::vector<double> coeffsB, double eta, double lnxA, double lnxB)
      {
         double Lm1coeffA = coeffsA[0];
         double L0coeffA = coeffsA[1];
         double L1coeffA = coeffsA[2];
         double Lm1coeffB = coeffsB[0];
         double L0coeffB = coeffsB[1];
         double L1coeffB = coeffsB[2];

         return L1coeffA*L1coeffB*d22mixDistFO(eta, lnxA, lnxB) + L1coeffA*L0coeffB*d21mixDistFO(eta, lnxA, lnxB) + L1coeffB*L0coeffA*d21mixDistFO(eta, lnxB, lnxA)
                  + L0coeffA*L0coeffB*d11mixDistFO(eta, lnxA, lnxB) + L1coeffA*Lm1coeffB*d2DistFO(eta, lnxA) + L1coeffB*Lm1coeffA*d2DistFO(eta, lnxB)
                  + L0coeffA*Lm1coeffB*d1DistFO(eta, lnxA) + L0coeffB*Lm1coeffA*d1DistFO(eta, lnxB) + Lm1coeffA*Lm1coeffB*eta;
      }

      double funcDistmixNLO(std::vector<double> coeffsA, std::vector<double> coeffsB, double lnxA, double lnxB)
      {
         double Lm1coeffA = coeffsA[0];
         double L0coeffA = coeffsA[1];
         double L1coeffA = coeffsA[2];
         double Lm1coeffB = coeffsB[0];
         double L0coeffB = coeffsB[1];
         double L1coeffB = coeffsB[2];

         return L1coeffA*L1coeffB*d22mixDistFO(lnxA, lnxB) + L1coeffA*L0coeffB*d21mixDistFO(lnxA, lnxB) + L1coeffB*L0coeffA*d21mixDistFO(lnxB, lnxA)
                  + L0coeffA*L0coeffB*d11mixDistFO(lnxA, lnxB) + L1coeffA*Lm1coeffB*d2DistFO(lnxA) + L1coeffB*Lm1coeffA*d2DistFO(lnxB)
                  + L0coeffA*Lm1coeffB*d1DistFO(lnxA) + L0coeffB*Lm1coeffA*d1DistFO(lnxB) + Lm1coeffA*Lm1coeffB*_eta;
      }
};

} // namespace Geneva

#endif // GENEVA_FO_CONV_HPP
