//------------------------------------------------------------------------------
/// \file FourVector.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012  MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definition of class FourVector
//------------------------------------------------------------------------------

#ifndef GENEVA_FOUR_VECTOR_HPP
#define GENEVA_FOUR_VECTOR_HPP

#include "Geneva/Core/Physics/ThreeVector.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \defgroup Core Core Ingredients
 *
 * This group provides the core functionality used throughout the program.
 */
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
/**
 * \class FourVector
 * \ingroup Core
 *
 * \brief A fourvector with Minkowski metric (+,-,-,-) using cartesian coordinates
 *
 * The class implements a momentum fourvector \f$ p^mu \f$ or position
 * fourvector \f$ x^mu \f$ in terms of its cartesian coordinates
 * \f$ p^\mu = (p^0, p^1, p^2, p^3) \equiv (E, p_x, p_y, p_z) \f$ or
 * \f$ x^\mu = (x^0, x^1, x^2, x^3) \equiv (t, x, y, z)) \f$.
 * It uses the metric (+,-,-,-) so the  invariant square is
 * \f$ p^2 = E^2 - p_x^2 - p_y^2 - p_z^2 \f$ or
 * \f$ x^2 = t^2 - x^2 - y^2 - z^2 \f$.
 *
 * Most of the member functions and operators are implemented inline to make the
 * class lightweight. The members p0(), p1(), p2(), p3() give direct access to
 * the components for both reading and writing. (In the future we might want to
 * have separate FourMomentum and FourPosition classes that wrap FourVector and
 * use the correct semantics.)
 *
 * The following algebraic operations are defined with the obvious meanings
 * (where p, q, r are FourVectors, and c is a double):
 *
 * - p = q + r, p = q - r (component-wise addition/subtraction)
 * - p = c * q, p = q * c (multiplication by real scalar)
 * - c = q * p            (scalar product of p and q)
 * - p += q, p -= q       (computes p = p +/- q)
 * - p *= c               (computes p = c * p)
 * - p == q, p != q       (component-wise comparison)
 *
 * There are various physics related queries which return the invariant square
 * and the (signed) magnitude and checks if the fourvector is timelike,
 * lightlike, or spacelike. There are also members returning the square and
 * magnitude of the three-vector part only.
 *
 * Transformations (rotations, boosts) are implemented as separate function
 * objects (see Transformations.hpp) and can be applied to a %FourVector using
 * apply().
 */
//------------------------------------------------------------------------------
class FourVector
{
   protected:
      double _p0;                      ///< 0-component (E or t)
      ThreeVector _pvec;               ///< ThreeVector component (px, py, pz) or (x, y, z)

   public:
      /// Named constants to specify if a %FourVector is spacelike, lightlike, or timelike
      enum Likeness
      {
         kSpacelike = -1,
         kLightlike = 0,
         kTimelike = 1
      };

      /// \name Constructors etc.
      ///@{

      /// Creates a zero %FourVector.
      FourVector();

      /// Creates a %FourVector (p0, p1, p2, p3).
      FourVector(double p0, double p1, double p2, double p3);

      /// Creates a %FourVector (p0, pvec).
      FourVector(double p0, const ThreeVector& pvec);

      ///@}

      /// \name Accessing and Querying Components
      ///@{
      double p0() const;               ///< Returns 0-component (E or t).
      double p1() const;               ///< Returns 1-component (px or x).
      double p2() const;               ///< Returns 2-component (py or y).
      double p3() const;               ///< Returns 3-component (pz or z).
      const ThreeVector& pvec() const; ///< Returns ThreeVector component.

      double& p0();                    ///< Access 0-component (E or t)
      double& p1();                    ///< Access 1-component (px or x)
      double& p2();                    ///< Access 2-component (py or y)
      double& p3();                    ///< Access 3-component (pz or z)
      ThreeVector& pvec();             ///< Access ThreeVector component

      // Returns the component indexed by mu.
      double operator[](size_t mu) const;

      // Gives access to the component indexed by mu.
      double& operator[](size_t mu);

      // Returns all components.
      void getComponents(double& p0, double& p1, double& p2, double& p3) const;

      // Sets all components.
      void setComponents(double p0, double p1, double p2, double p3);

      /// Sets all components to zero.
      void setToZero();
      ///@}

      /// \name Algebra
      ///@{
      // FourVector equality comparison operator
      bool operator==(const FourVector& rhs) const;

      // FourVector inequality comparison operator
      bool operator!=(const FourVector& rhs) const;

      // Test if the FourVector is nonzero.
      bool isNonzero() const;

      // FourVector addition assignment operator
      FourVector& operator+=(const FourVector& rhs);

      // FourVector subtraction assignment operator
      FourVector& operator-=(const FourVector& rhs);

      // FourVector multiplication with real number assignment operator
      FourVector& operator*=(double rhs);

      // FourVector division by real number assignment operator
      FourVector& operator/=(double rhs);
      ///@}

      /// \name Physics
      /// Physics related queries and manipulations
      ///@{
      // Returns the Likeness of the FourVector.
      int likeness() const;

      // Test if the FourVector is timelike.
      bool isTimelike() const;

      // Test if the FourVector is lightlike.
      bool isLightlike() const;

      // Test if the FourVector is spacelike.
      bool isSpacelike() const;

      // Returns the invariant square of the FourVector.
      double square() const;

      // Returns the magnitude of the FourVector.
      double magnitude() const;

      //  Returns the signed magnitude of the FourVector.
      double signedMagnitude() const;

      // Returns the FourVector's rapidity relative to the z-direction.
      double rapidity() const;

      // Returns the FourVector's pseudo-rapidity relative to the z-direction.
      double pseudoRapidity() const;

      // Returns the FourVector's azimuthal angle around the z-direction relative to the x-direction in the range [-pi,pi].
      double phi() const;

      // Returns the FourVector's azimuthal angle around the z-direction relative to the x-direction in the range [0,2pi].
      double phi02pi() const;

      // Returns the FourVector's plus component relative to the z-direction as p0()-p3().
      double plus() const;

       // Returns the FourVector's minus component relative to the z-direction as p0()+p3().
      double minus() const;

      // Returns the magnitude of the FourVector's transverse component relative to the z-direction.
      double perp() const;

      // Returns the squared magnitude of the FourVector's transverse component relative to the z-direction.
      double perp2() const;

      // Applies a transformation to the FourVector.
      template <class Transformation>
      FourVector& apply(const Transformation& trans);

      // Returns a copy of FourVector with a transformation applied to it.
      template <class Transformation>
      FourVector transformed(const Transformation& trans) const;

      ///@}

      /// \name Output
      ///@{
      // FourVector stream insertion operator
      friend std::ostream& operator<<(std::ostream& out, const FourVector& p);
      ///@}

   protected:
      /// Streams a string in the format "(p0, p1, p2, p3)" to \a out.
      std::ostream& print(std::ostream& out) const;
};

// FourVector addition operator
const FourVector operator+(const FourVector& lhs, const FourVector& rhs);

// FourVector subtraction operator
const FourVector operator-(const FourVector& lhs, const FourVector& rhs);

// FourVector negation operator
const FourVector operator-(const FourVector& lhs);

// FourVector multiplication operator with real number from the left
const FourVector operator*(double lhs, const FourVector& rhs);

// FourVector multiplication operator with real number from the right
const FourVector operator*(const FourVector& lhs, double rhs);

// FourVector division operator with real number
const FourVector operator/(const FourVector& lhs, double rhs);

// FourVector multiplication operator corresponding to inner product
double operator*(const FourVector& lhs, const FourVector& rhs);


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Creates a zero FourVector.
inline FourVector::FourVector()
   : _p0(0.), _pvec()
{}

//------------------------------------------------------------------------------
// Creates a FourVector (p0, p1, p2, p3).
inline FourVector::FourVector(double p0, double p1, double p2, double p3)
   : _p0(p0), _pvec(p1, p2, p3)
{}

//------------------------------------------------------------------------------
// Creates a %FourVector (p0, pvec).
inline FourVector::FourVector(double p0, const ThreeVector& pvec)
   : _p0(p0), _pvec(pvec)
{}

//------------------------------------------------------------------------------
// Queries and Accessors
//------------------------------------------------------------------------------

inline double FourVector::p0() const               { return _p0;        }
inline double FourVector::p1() const               { return _pvec.p1(); }
inline double FourVector::p2() const               { return _pvec.p2(); }
inline double FourVector::p3() const               { return _pvec.p3(); }
inline const ThreeVector& FourVector::pvec() const { return _pvec;      }

inline double& FourVector::p0()                    { return _p0;        }
inline double& FourVector::p1()                    { return _pvec.p1(); }
inline double& FourVector::p2()                    { return _pvec.p2(); }
inline double& FourVector::p3()                    { return _pvec.p3(); }
inline ThreeVector& FourVector::pvec()             { return _pvec;      }

//------------------------------------------------------------------------------
/**
 * \brief Returns all components.
 * \param[out] p0 E or t component
 * \param[out] p1 px or x component
 * \param[out] p2 py or y component
 * \param[out] p3 pz or z component
 *
 * The components of the %FourVector are returned in the provided parameters.
 */
inline void FourVector::getComponents(double& p0, double& p1, double& p2, double& p3) const
{
   p0 = _p0;
   _pvec.getComponents(p1, p2, p3);
}

//------------------------------------------------------------------------------
/**
 * \brief Sets all components.
 * \param p0 E or t component
 * \param p1 px or x component
 * \param p2 py or y component
 * \param p3 pz or z component
 *
 * Sets all components of the %FourVector to the given values.
 */
inline void FourVector::setComponents(double p0, double p1, double p2, double p3)
{
   _p0 = p0;
   _pvec.setComponents(p1, p2, p3);
}

//------------------------------------------------------------------------------
// Sets all components to zero.
inline void FourVector::setToZero()
{
   setComponents(0., 0., 0., 0.);
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %FourVector equality comparison operator
 * \param rhs The %FourVector to compare to.
 * \return \c true if all components are equal, \c false otherwise.
 *
 * The comparison is done with the default accuracy and precision separately for
 * each of the components.
 */
inline bool FourVector::operator==(const FourVector& rhs) const
{
   return !operator!=(rhs);
}

//------------------------------------------------------------------------------
/**
 * \brief %FourVector addition assignment operator
 * \param rhs The %FourVector to be added.
 * \return A reference to the %FourVector.
 *
 * \a rhs is added to the %FourVector.
 */
inline FourVector& FourVector::operator+=(const FourVector& rhs)
{
   _p0 += rhs._p0;
   _pvec += rhs._pvec;
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %FourVector subtraction assignment operator
 * \param rhs The %FourVector to be subtracted.
 * \return A reference to the %FourVector.
 *
 * \a rhs is subtracted from the %FourVector.
 */
inline FourVector& FourVector::operator-=(const FourVector& rhs)
{
   _p0 -= rhs._p0;
   _pvec -= rhs._pvec;
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %FourVector multiplication with real number assignment operator
 * \param rhs The number to be multiplied by.
 * \return A reference to the %FourVector.
 *
 * Each component of the %FourVector is multiplied by \a rhs.
 */
inline FourVector& FourVector::operator*=(double rhs)
{
   _p0 *= rhs;
   _pvec *= rhs;
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %FourVector division by real number assignment operator
 * \param rhs The number to be divided by.
 * \return A reference to the %FourVector.
 *
 * Each component of the %FourVector is divided by \a rhs.
 * \note Division by zero is not checked.
 */
inline FourVector& FourVector::operator/=(double rhs)
{
   return operator*=(1. / rhs);
}

//------------------------------------------------------------------------------
/**
 * \brief %FourVector addition operator
 * \param lhs The 1st %FourVector to add.
 * \param rhs The 2nd %FourVector to add.
 * \return A %FourVector given by the sum of \a lhs and \a rhs.
 */
inline const FourVector operator+(const FourVector& lhs, const FourVector& rhs)
{
   return FourVector(lhs.p0() + rhs.p0(), lhs.pvec() + rhs.pvec());
}

//------------------------------------------------------------------------------
/**
 * \brief %FourVector subtraction operator
 * \param lhs The 1st %FourVector
 * \param rhs The 2nd %FourVector to be subtracted from \a lhs.
 * \return A %FourVector given by the difference of \a lhs and \a rhs.
 */
inline const FourVector operator-(const FourVector& lhs, const FourVector& rhs)
{
   return FourVector(lhs.p0() - rhs.p0(), lhs.pvec() - rhs.pvec());
}

//------------------------------------------------------------------------------
/**
 * \brief %FourVector negation operator
 * \param lhs The %FourVector to be negated.
 * \return A %FourVector given by the negative of \a lhs.
 */
inline const FourVector operator-(const FourVector& lhs)
{
   return FourVector(-lhs.p0(), -lhs.pvec());
}

//------------------------------------------------------------------------------
/**
 * \brief %FourVector multiplication operator with real number from the left
 * \param lhs The number to multiply with.
 * \param rhs The %FourVector to be multiplied by \a lhs.
 * \return A %FourVector given by the product of \a lhs times \a rhs.
 */
inline const FourVector operator*(double lhs, const FourVector& rhs)
{
   return FourVector(lhs * rhs.p0(), lhs * rhs.pvec());
}

//------------------------------------------------------------------------------
/**
 * \brief %FourVector multiplication operator with real number from the right
 * \param lhs The %FourVector to be multiplied by \a rhs.
 * \param rhs The number to multiply with.
 * \return A %FourVector given by the product of \a lhs times \a rhs.
 */
inline const FourVector operator*(const FourVector& lhs, double rhs)
{
   return rhs * lhs;
}

//------------------------------------------------------------------------------
/**
 * \brief %FourVector division operator with real
 * \param lhs The %FourVector to be divided by \a rhs.
 * \param rhs The number to divide by.
 * \return A %FourVector given by \a lhs divided by \a rhs.
 * \note Division by zero is not checked.
 */
inline const FourVector operator/(const FourVector& lhs, double rhs)
{
   return (1./rhs) * lhs;
}

//------------------------------------------------------------------------------
/**
 * \brief %FourVector multiplication operator corresponding to inner product
 * \param lhs The 1st %FourVector.
 * \param rhs The 2nd %FourVector.
 * \return The result from taking the inner product of \a lhs and \a rhs.
 *
 * The inner product is computed as
 * \f$ p \cdot q = p^0 q^0 - \vec{p}\cdot \vec{q} \f$.
 */
inline double operator*(const FourVector& lhs, const FourVector& rhs)
{
   return lhs.p0()*rhs.p0() - lhs.pvec()*rhs.pvec();
}

//------------------------------------------------------------------------------
// Physics
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Returns the #Likeness of the %FourVector.
 * \return An integer representing the #Likeness of the %FourVector.
 *
 * The returned integer can be compared with the named constants defined in
 * #Likeness to check if the %FourVector is spacelike, lightlike, or timelike.
 * It compares \f$ E^2 \f$ to \f$ \vec{p}^2 \f$ using the default precision and
 * arbitrary accuracy.
 */
inline int FourVector::likeness() const
{
   return compareRelative(_p0*_p0, _pvec*_pvec);
}

//------------------------------------------------------------------------------
/**
 * \brief Test if the %FourVector is timelike.
 * \return \c true if %FourVector is timelike, \c false otherwise.
 *
 * Equivalent to likeness() == \c kTimelike, which is \c true if
 * \f$ E^2 > \vec{p}^2 \f$. See likeness() for details.
 */
inline bool FourVector::isTimelike() const
{
   return likeness() == kTimelike;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if the %FourVector is lightlike.
 * \return \c true if %FourVector is lightlike, \c false otherwise.
 *
 * Equivalent to likeness() == \c kLightlike, which is \c true if
 * \f$ E^2 = \vec{p}^2 \f$. See likeness() for details.
 */
inline bool FourVector::isLightlike() const
{
   return likeness() == kLightlike;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if the %FourVector is spacelike.
 * \return \c true if %FourVector is spacelike, \c false otherwise.
 *
 * Equivalent to likeness() == \c kSpacelike, which is \c true if
 * \f$ E^2 < \vec{p}^2 \f$. See likeness() for details.
 */
inline bool FourVector::isSpacelike() const
{
   return likeness() == kSpacelike;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the invariant square of the %FourVector.
 * \return The value of the invariant square \f$ E^2 - \vec{p}^2 \f$.
 *
 * \note To check if the %FourVector is timelike, lightlike, or spacelike the
 * likeness(), isTimeLike(), isLightLike(), or isSpaceLike() members should be
 * used rather than directly comparing the square to zero by hand.
 */
inline double FourVector::square() const
{
   return _p0*_p0 - _pvec*_pvec;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the magnitude of the %FourVector.
 * \return The value of \f$ \sqrt{\lvert E^2 - \vec{p}^2 \rvert} \f$.
 */
inline double FourVector::magnitude() const
{
   return sqrt(abs(square()));
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the signed magnitude of the %FourVector.
 * \return The value of \f$ \mathrm{sgn}(E^2 - \vec{p}^2)\times
 *         \sqrt{\lvert E^2 - \vec{p}^2 \rvert} \f$.
 */
inline double FourVector::signedMagnitude() const
{
   const double s = square();
   return s >= 0.0 ? sqrt(s) : -sqrt(-s);
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the %FourVector's rapidity relative to the z-direction.
 * \return The value of \f$ 1/2 \ln((E+p_z)/(E-p_z)) \f$.
 *
 * The rapidity is not well-defined for \f$ |E| < |p_z| \f$. In this case, for
 * \f$ |E| \lesssim |p_z| (1 - \delta) \f$ the return value is nan (signaling)
 * and otherwise \f$ \mathrm{sgn}(E/p_z) \infty \f$, where \f$ \delta \f$  is the
 * default relative precision.
 */
inline double FourVector::rapidity() const
{
   const double arg = minus() / plus();
   if (arg >= 0.0)
      return 0.5 * log(arg);
   if (arg >= -0.5 * DEFAULT_PRECISION)
      return -std::numeric_limits<double>::infinity();
   if (1./arg >= -0.5 * DEFAULT_PRECISION)
      return std::numeric_limits<double>::infinity();
   return std::numeric_limits<double>::signaling_NaN();
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the %FourVector's pseudo-rapidity relative to the z-direction.
 * \return The value of \f$ 1/2 \ln((\lvert\vec{p}\rvert + p_z)/(\lvert\vec{p}\rvert - p_z)) \f$.
 *
 * The pseudo-rapidity is equivalent to the angle with respect to the
 * z-direction and is therefore always defined (as opposed to the rapidity()).
 */
inline double FourVector::pseudoRapidity() const
{
   const double tmp = pvec().magnitude();
   return 0.5 * log((tmp + p3()) / (tmp - p3()));
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the FourVector's azimuthal angle around the z-direction relative to the x-direction in the range [-pi,pi].
 * \return The value of the azimuthal angle in the range \f$ [-\pi,\pi] \f$ relative to the positive x-direction.
 *
 * Computes the inverse tangent of \f$ p_y/p_x \f$ using the signs of arguments
 * to correctly determine quadrant.
 */
inline double FourVector::phi() const
{
   return atan2(p2(), p1());
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the FourVector's azimuthal angle around the z-direction relative to the x-direction in the range [0,2pi].
 * \return The value of the azimuthal angle in the range \f$ [0,2pi] \f$ relative to the positive x-direction.
 *
 * This maps the [-pi, 0] range of phi to the [pi, 2pi] range.
 */
inline double FourVector::phi02pi() const
{
   double tmp = phi();
   if (tmp < 0.0)
      tmp += math_constants::_2pi();
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the FourVector's plus component relative to the z-direction as p0()-p3().
 * \return The value of \f$ p^+ = E - p_z \f$.
 */
inline double FourVector::plus() const
{
   return p0() - p3();
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the FourVector's minus component relative to the z-direction as p0()+p3().
 * \return The value of \f$ p^- = E + p_z \f$.
 */
inline double FourVector::minus() const
{
   return p0() + p3();
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the magnitude of the %FourVector's transverse component relative to the z-direction.
 * \return The value of \f$ \lvert \vec{p}_T \rvert = \sqrt{p_x^2 + p_y^2} \f$.
 */
inline double FourVector::perp() const
{
   return _pvec.perp();
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the squared magnitude of the %FourVector's transverse component relative to the z-direction.
 * \return The value of \f$ \lvert \vec{p}_T \rvert^2 = p_x^2 + p_y^2 \f$.
 */
inline double FourVector::perp2() const
{
   return _pvec.perp2();
}

//------------------------------------------------------------------------------
/**
 * \brief Applies a transformation to the %FourVector.
 * \param trans The transformation to be applied.
 * \return A reference to the %FourVector.
 *
 * \a trans can be any function object which implements the function evaluation
 * operator with the signature
 * \code
 * const FourVector operator()(const FourVector& operand)
 * \endcode
 *
 * The result of the transformation is assigned to the %FourVector.
 */
template <class Transformation>
inline FourVector& FourVector::apply(const Transformation& trans)
{
   return *this = trans(*this);
}

//------------------------------------------------------------------------------
/**
 * \brief Returns a copy of %FourVector with a transformation applied to it.
 * \param trans The transformation to be applied.
 * \return A transformed copy of the %FourVector.
 *
 * This applies \a trans to a new copy of the %FourVector and returns the
 * result. The %FourVector itself is not changed. See apply() for more details.
 */
template <class Transformation>
inline FourVector FourVector::transformed(const Transformation& trans) const
{
   return trans(*this);
}


//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %FourVector stream insertion operator
 * \param out The output stream to be used.
 * \param p The %FourVector to be streamed.
 * \return A reference to \a out.
 *
 * Streams a string in the format "(p0, p1, p2, p3)" to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const FourVector& p)
{
   return p.print(out);
}

} // namespace Geneva

#endif // GENEVA_FOUR_VECTOR_HPP
