//------------------------------------------------------------------------------
/// \file LHEvent.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class LHEvent
//------------------------------------------------------------------------------

#ifndef GENEVA_LH_EVENT_HPP
#define GENEVA_LH_EVENT_HPP

#include "Geneva/Core/Physics/Event.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class LHEvent
 * \ingroup Core
 *
 * \brief Extension of Event to supply the information required for LHEF format.
 *
 * Supplements an Event with the information required by the Les Houches event
 * file (LHEF) format, see Alwall et al. arXiv:hep-ph/0609017). Most information
 * is directly translated from the underlying Configuration.
 *
 * \note: aQCDup, aQEDup, vtimup, spinup are unused and set to default values.
 */
//------------------------------------------------------------------------------
class LHEvent : public Event
{
   private:
      double _scalup;              ///< the shower starting scale

   public:
      /// \name Constructors etc.
      ///@{
      LHEvent(const Event& event, double scalup)
         : Event(event), _scalup(scalup)
      {}

      ///@}

      /// \name LHEF fields
      ///@{
      size_t nup() const    { return config().size(); }
      double xwgtup() const { return weightedSigma().central(); }

      double scalup() const { return _scalup; }

      double aQEDup() const { return -1.; };
      double aQCDup() const { return -1.; };

      int istup(Configuration::iterator it) const
      {
         if (it->isFinal())
            return 1;
         // must be initialA or initialB, currently we do not support resonances
         return -1;
      }

      size_t mothup1(Configuration::iterator it) const
      {
         if (it->isFinal())
            return 1;
         // must be initialA or initialB, currently we do not support resonances
         return 0;
      }

      size_t mothup2(Configuration::iterator it) const
      {
         if (it->isFinal())
            return 2;
         // must be initialA or initialB, currently we do not support resonances
         return 0;
      }

      int icolup1(Configuration::iterator it) const { return it->color1(); }
      int icolup2(Configuration::iterator it) const { return it->color2(); }

      double vtimup() const { return 0.; }
      double spinup() const { return 9.; }
      ///@}
};

} // namespace Geneva

#endif // GENEVA_LH_EVENT_HPP
