//------------------------------------------------------------------------------
/// \file NJettiness.hpp
//
// Author(s):
//    Frank Tackmann, Chris Vermilion, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 Frank Tackmann, LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definitions for NJettiness metrics and interfaces of Njettiness user classes
//------------------------------------------------------------------------------

#ifndef GENEVA_N_JETTINESS_HPP
#define GENEVA_N_JETTINESS_HPP

#include "Geneva/Core/Physics/Configuration.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
// Definitions of metrics
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \struct NJettinessMetric
 * \ingroup Core
 *
 * \brief Functor implementing the NJettiness metrics.
 *
 * Contains both the clustering and measurement metrics for intial and final
 * states
 *
 * The FSR clustering metric is given by
 * \f$ d(p_i, p_j) = \lvert \vec p_i \rvert + \lvert \vec p_j \rvert - \lvert \vec p_i + \vec p_j \rvert \f$.
 *
 * The ISR (beam) clustering metric is given by
 * \f$ d(p) = \lvert \vec p \rvert - \lvert p^3 \rvert \f$.
 *
 */
//------------------------------------------------------------------------------
struct NJettinessMetric
{
   template <typename ConfigurationEntry1, typename ConfigurationEntry2>
   double operator()(const ConfigurationEntry1& p1, const ConfigurationEntry2& p2) const
   {
      if (p1.isFinal())
         if (p2.isFinal())
            return final(p1.p(), p2.p());
         else if (p2.isInitialA())
            return initialA(p1.p());
         else
            return initialB(p1.p());
      else if (p1.isInitialA())
         return initialA(p2.p());
      else
         return initialB(p2.p());
   }

   /// Clustering metric with beamA
   static double initialA(const FourVector& p)
   {
      return p.pvec().magnitude() - p.p3();
   }

   /// Clustering metric with beamB
   static double initialB(const FourVector& p)
   {
      return p.pvec().magnitude() + p.p3();
   }

   /// Combined clustering metric for both beams
   static double initial(const FourVector& p)
   {
      return p.pvec().magnitude() - abs(p.p3());
   }

   /// FSR clustering metric
   static double final(const FourVector& p1, const FourVector& p2)
   {
      return p1.pvec().magnitude() + p2.pvec().magnitude() - (p1.pvec() + p2.pvec()).magnitude();
   }

   /// Beam A measurement contribution
   static double initialMeasureA(const FourVector& p)
   {
      return p.plus();
   }

   /// Beam B measurement contribution
   static double initialMeasureB(const FourVector& p)
   {
      return p.minus();
   }

   /// Beam measurement contribution
   static double initialMeasure(const FourVector& p)
   {
      return p.p0() - abs(p.p3());
   }

   /// Final-state measurement contribution
   static double finalMeasure(const FourVector& p)
   {
      const double p0 = p.p0();
      const double pvec = p.pvec().magnitude();
      // only return something nonzero if E and |p| are nonequal within default relative precision.
      if (compareRelative(p0, pvec))
         return p0 - pvec;
      return 0.0;
   }
};

//------------------------------------------------------------------------------
/**
 * \struct NJettinessMetricFSR
 * \ingroup Core
 *
 * \brief Functor for the NJettiness FSR clustering metric.
 *
 * This exports the generic definitions of NJettinessMetric for the special case
 * of only final-state clusterings and momenta.
 */
//------------------------------------------------------------------------------
struct NJettinessMetricFSR
{
   /// Returns the clustering metric for generic particle types.
   template <typename ParticleType1, typename ParticleType2>
   double operator()(const ParticleType1& p1, const ParticleType2& p2) const
   {
      return operator()(p1.p(), p2.p());
   }

   /// Returns the clustering metric for \a p1 and \a p2.
   double operator()(const FourVector& p1, const FourVector& p2) const
   {
      return NJettinessMetric::final(p1, p2);
   }

   /// Returns the measurement contribution for \a p.
   static double measure(const FourVector& p)
   {
      return NJettinessMetric::finalMeasure(p);
   }
};

//------------------------------------------------------------------------------
/**
 * \struct NJettinessMetricFSRSoft
 * \ingroup Core
 *
 * \brief Functor for the NJettiness FSR clustering metric in the soft limit.
 *
 * \todo: document me!
 *
 * \note we are correctly assuming that the soft parton is massless,
 *       thus we can use pSoft.p0() instead of pSoft.pvec().magnitude()
 *       The order of the entries is crucial: There is indeed
 *        no requirement on pSoft.p0() < pHard.p0(). In fact, it may happen that
 *       in the subtraction we need to consider the combination of a gluon
 *       (which is the parton that can go soft) with whatever other parton, e.g.
 *       with quarks with happens to be softer (i.e. with less energy) than the
 *       gluon itself.
 */
//------------------------------------------------------------------------------
struct NJettinessMetricFSRSoft
{
   double operator()(const FourVector& pSoft, const FourVector& pHard) const
   {
      return pSoft.p0() * (1.0 - cosAngleBetween(pSoft.pvec(), pHard.pvec()));
   }

   double operator()(const Particle& pSoft, const Particle& pHard) const
   {
      return operator()(pSoft.p(), pHard.p());
   }
};


//------------------------------------------------------------------------------
/**
 * \struct NJettinessMetricISR
 * \ingroup Core
 *
 * \brief Functor for the NJettiness ISR clustering metric.
 *
 * This exports the generic definitions of NJettinessMetric for the special case
 * of only ISR emissions.
 */
//------------------------------------------------------------------------------
struct NJettinessMetricISR
{
   /// Returns the ISR (beam) clustering metric for generic particle types
   template <typename ParticleLike>
   double operator()(const ParticleLike& p) const
   {
      return operator()(p.p());
   }

   /// Returns the ISR (beam) clustering metric for \a p.
   double operator()(const FourVector& p) const
   {
      return NJettinessMetric::initial(p);
   }

   /// Returns the beam measurement contribution for \a p.
   static double measure(const FourVector& p)
   {
      return NJettinessMetric::initialMeasure(p);
   }
};


//------------------------------------------------------------------------------
// e+e- definitions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \struct TwoJettiness
 * \ingroup Core
 *
 * \brief A function object implementing exact 2-jettiness.
 *
 * Here 2-jettiness is defined by
 * \f[
 * \mathcal{T}_2 = \min_{\vec{n}} \sum_k E_k - \lvert\vec{n}\cdot \vec{p_k}\rvert
 * \f]
 *
 * where the sum runs over all particles in the final state and the resulting
 * \f$ \vec{n} \f$ from the minimization is equivalent to the thrust axis. For
 * massless particles (only) this definition is equivalent to
 * \f$ \mathcal{T}_2 = E_\mathrm{cm} (1 - T) \f$ where T is thrust.
 *
 * \note This implementation assumes that the given Configuration is in the
 * center-of-mass frame (i.e. that the total three-momentum in the final state
 * is zero.)
 *
 * \note Currently the axis determination is only implemented for 2 to 4
 * particles in the final state. Any other number will cause a runtime error.
 */
//------------------------------------------------------------------------------
struct TwoJettiness
{
   // Computes the exact 2-jettiness (Tau2).
   double operator()(const Configuration& config) const;

   // Computes the 2-jettiness (thrust) axis (for 2 to 4 particles).
   ThreeVector axis(const Configuration& config) const;
};


//------------------------------------------------------------------------------
/**
 * \struct NJettiness
 * \ingroup Core
 *
 * \brief A function object implementing exact N-jettiness.
 *
 * Here N-jettiness is defined by
 * \f[
 * \mathcal{T}_N = \min_{\vec{n}_i} \sum_k \min_i (E_k - \lvert\vec{n}_i \cdot \vec{p_k}\rvert
 * \f]
 *
 * where the sum runs over all particles in the final state and the minimization
 * runs over all axes i = 1 to N.
 *
 * For N = 2 this delegates all work to TwoJettiness.
 *
 * \note Currently, the axis determination is only implemented for N+1 particles
 * in the final state. Any other number will cause a runtime error.
 */
//------------------------------------------------------------------------------
struct NJettiness
{
   // Computes the exact N-jettiness (TauN).
   double operator()(const Configuration& config, size_t N) const;
};


//------------------------------------------------------------------------------
/**
 * \struct NJettinessRecursive
 * \ingroup Core
 *
 * \brief A function object implementing recursive N-jettiness.
 *
 * Recursive N-jettiness is defined by clustering down to N final-state momenta,
 * where in each clustering step (numFinal-1)-jettiness is minimized. In
 * particular, when starting with N+1 final states, the result is the same as
 * using NJettiness.
 */
//------------------------------------------------------------------------------
struct NJettinessRecursive
{
   // Computes the recursive N-jettiness (TauN).
   double operator()(const Configuration& config, size_t N) const;
};


//------------------------------------------------------------------------------
// pp definitions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \struct NJettinessppRecursive
 * \ingroup Core
 *
 * \brief A function object implementing recursive N-jettiness for pp collisions
 *
 * N-jettiness is defined as
 *
 * \f[
 * \mathcal{T}_N = \min_{\vec{n}_i} \sum_k \min_i \{E_k - \lvert\vec{n}_i \cdot \vec{p_k}\rvert \}
 * \f]
 *
 * where the sum runs over all particles in the final state and the minimization
 * runs over all axes i = a, b, 1 to N, where n_a and n_b are the positive and
 * negative beam directions.
 *
 * For N+1 final states the minimization over axes can be carried out.
 * Recursive N-jettiness is defined by clustering down from M > N to N final-
 * state momenta, where in each clustering step (M-1)-jettiness is minimized. In
 * particular, when starting with N+1 final states, the result is exact TauN.
 *
 * \note: By default this operates on all final-state particles for which
 * \c IsQCD() is true.
 */
//------------------------------------------------------------------------------
class NJettinessppRecursive
{
   protected:
      const Configuration& _config;
      double _y;
      size_t _numQCD;
      size_t _numNonQCD;

   public:
      NJettinessppRecursive(const Configuration& config);
      NJettinessppRecursive(const Configuration& config, double y);

      double y() const         { return _y; }
      size_t numQCD() const    { return _numQCD; }
      size_t numNonQCD() const { return _numNonQCD; }

      // Computes the recursive N-jettiness (TauN).
      double operator()(size_t N) const;

      // Performs a single N-jettiness clustering step.
      double cluster(Configuration::iterator& emitted, Configuration::iterator& sister) const;

      // Performs a single N-jettiness clustering step only taking into account QCD-allowed splittings.
      double clusterQCDAllowed(Configuration::iterator& emitted, Configuration::iterator& sister, bool allowqqbar) const;

      // Performs a single N-jettiness clustering step in the soft limit.
      double clusterSoft(Configuration::iterator& emitter, FourVector pSoft) const;
};

//------------------------------------------------------------------------------
/**
 * \struct ZeroJettinesspp
 * \ingroup Core
 *
 * \brief A function object implementing exact 0-jettiness (beam thrust) for pp collisions
 *
 * Here 0-jettiness is defined by
 * \f[
 * \mathcal{T}_0 = \sum_k E_k - \lvert\hat{z}\cdot \vec{p_k}\rvert
 * \f]
 *
 * where the sum runs over all particles in the final state and \f$ \hat{z} \f$
 * is the beam direction.
 *
 * \note: By default this operates on all final-state particles for which
 * \c IsQCD() is true.
 */
//------------------------------------------------------------------------------
class ZeroJettinesspp : private NJettinessppRecursive
{
   public:
      ZeroJettinesspp(const Configuration& config)
         : NJettinessppRecursive(config) {};

      ZeroJettinesspp(const Configuration& config, double y)
         : NJettinessppRecursive(config, y) {};

      double y() const         { return _y; }
      size_t numQCD() const    { return _numQCD; }
      size_t numNonQCD() const { return _numNonQCD; }

      // Computes the exact 0-jettiness (Tau0).
      double operator()() const;
};

} // namespace Geneva

#endif // GENEVA_N_JETTINESS_HPP
