//------------------------------------------------------------------------------
/// \file NJettiness_impl.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Template implementations for NJettiness algorithms
//------------------------------------------------------------------------------

#ifndef GENEVA_N_JETTINESS_IMPL_HPP
#define GENEVA_N_JETTINESS_IMPL_HPP

#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/Utils/Algorithms.hpp"

#include <list>
#include <vector>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief Template N-jettiness algorithms that do the actual work.
 *
 * \note These are for internal use. Do not use them directly unless you know
 * what you are doing.
 */
//------------------------------------------------------------------------------
struct NJettiness_impl
{
   template <typename ParticleLikeIterator, typename EndIterator, typename UnaryCondition>
   static double calculateTauNRecursive(size_t N, ParticleLikeIterator first, EndIterator last, double expy, UnaryCondition condFinal);

   template <typename ParticleLikeIterator, typename EndIterator, typename UnaryCondition, typename BinaryCondition>
   static double cluster(ParticleLikeIterator initialA, ParticleLikeIterator initialB,
                         ParticleLikeIterator firstFinal, EndIterator last,
                         ParticleLikeIterator& emitted, ParticleLikeIterator& sister,
                         double expy, size_t numFinal,
                         UnaryCondition condFinal, BinaryCondition cond2);

   template <typename ParticleLikeIterator, typename EndIterator, typename UnaryCondition>
   static double calculateTau0(ParticleLikeIterator first, EndIterator last, double expy, UnaryCondition cond);
};


//------------------------------------------------------------------------------
/**
 * \brief Template algorithm to calculate recursive TauN.
 */
//------------------------------------------------------------------------------
template <typename ParticleLikeIterator, typename EndIterator, typename UnaryCondition>
double NJettiness_impl::calculateTauNRecursive(size_t N, ParticleLikeIterator first, EndIterator last, double expy, UnaryCondition condFinal)
{
   std::list<FourVector> momenta;
   BoostZHyp boost(1./expy);
   for (; first != last; ++first) {
      if (condFinal(*first))
         momenta.push_back(boost(first->p()));
   }

   size_t nFinal = momenta.size();
   std::list<FourVector>::iterator it1, it2, it3;
   double TauN = 0.0;
   // cluster until nFinal == N
   while (nFinal > N) {
      const double dij = pairwise_minimum(momenta.begin(), momenta.end(), it1, it2, NJettinessMetricFSR());
      const double di = minimum(momenta.begin(), momenta.end(), it3, NJettinessMetricISR());
      if (di <= dij) {
         // clustering with beam: Add the beam contribution to TauN and remove the momentum
         TauN += NJettinessMetricISR::measure(*it3);
         momenta.erase(it3);
      } else {
         // final-state clustering: Combine the two momenta
         (*it1) += (*it2);
         momenta.erase(it2);
      }
      --nFinal;
   }

   // add contributions from the remaining N clustered final-state momenta
   for (it1 = momenta.begin(); it1 != momenta.end(); ++it1)
      TauN += NJettinessMetricFSR::measure(*it1);
   return TauN;
}

//------------------------------------------------------------------------------
/**
 * \brief Template algorithm to perform for a single TauN clustering step.
 */
//------------------------------------------------------------------------------
template <typename ParticleLikeIterator, typename EndIterator, typename UnaryCondition, typename BinaryCondition>
double NJettiness_impl::cluster(ParticleLikeIterator initialA, ParticleLikeIterator initialB,
                                ParticleLikeIterator firstFinal, EndIterator last,
                                ParticleLikeIterator& emitted, ParticleLikeIterator& sister,
                                double expy, size_t numFinal,
                                UnaryCondition condFinal, BinaryCondition cond2)
{
   typedef std::vector<std::pair<ParticleLikeIterator, FourVector> > momenta_type;
   momenta_type momenta;

   momenta.reserve(numFinal);
   BoostZHyp boost(1./expy);
   for (; firstFinal != last; ++firstFinal) {
      if (condFinal(*firstFinal))
         momenta.push_back(std::make_pair(firstFinal, boost(firstFinal->p())));
   }

   // effectively performs a conditional pairwise minimum for final-final pairs
   // and a conditional simple minimum for initial-final pairs
   double min = std::numeric_limits<double>::infinity();
   for (typename momenta_type::iterator i = momenta.begin(); i != momenta.end(); ++i) {
      if (initialA != last)
         if (cond2(*initialA, *(i->first))) {
            const double temp = NJettinessMetric::initialA(i->second);
            if (temp < min) {
               min = temp;
               emitted = i->first;
               sister = initialA;
            }
         }
      if (initialB != last)
         if (cond2(*initialB, *(i->first))) {
            const double temp = NJettinessMetric::initialB(i->second);
            if (temp < min) {
               min = temp;
               emitted = i->first;
               sister = initialB;
            }
         }
      typename momenta_type::iterator j = i;
      while (++j != momenta.end()) {
         if (cond2(*(i->first), *(j->first))) {
            const double temp = NJettinessMetric::final(i->second, j->second);
            if (temp < min) {
               emitted = i->first;
               sister = j->first;
               min = temp;
            }
         }
      }
   }
   return min;
}

//------------------------------------------------------------------------------
/**
 * \brief Template algorithm to calculate exact Tau0.
 */
//------------------------------------------------------------------------------
template <typename ParticleLikeIterator, typename EndIterator, typename UnaryCondition>
double NJettiness_impl::calculateTau0(ParticleLikeIterator first, EndIterator last, double expY, UnaryCondition cond)
{
   double Tau0 = 0.;
   const double etaplus = expY;
   const double etaminus = 1./etaplus;
   for (; first != last; ++first) {
      if (cond(*first)) {
         const double pplus = etaplus * first->p().plus();
         const double pminus = etaminus * first->p().minus();
         if (pplus < pminus)
            Tau0 += pplus;
         else
            Tau0 += pminus;
      }
   }
      return Tau0;
}

} // namespace Geneva

#endif // GENEVA_N_JETTINESS_IMPL_HPP
