//------------------------------------------------------------------------------
/// \file Particle.hpp
//
// Author(s):
//    Erik Strand, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 LBNL, MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Particle
//------------------------------------------------------------------------------

#ifndef GENEVA_PARTICLE_HPP
#define GENEVA_PARTICLE_HPP

#include "Geneva/Core/Physics/FourVector.hpp"
#include "Geneva/Core/Physics/PdgID.hpp"
#include "Geneva/Core/Physics/ParticleData.hpp"
#include "Geneva/Core/Utils/fixed_ptr.hpp"

#include <iosfwd>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class Particle
 * \ingroup Core
 *
 * \brief A particle with a PDG id and a four-momentum.
 *
 * The class implements a particle identified by its PDG id and four-momentum.
 * (and possibly other quantum numbers like helicity and color in the future).
 *
 * The particle's PDG id is represented as a PdgID and accessed by id(), and by
 * design cannot be changed. The particle's four- and three-momentum vectors are
 * accessed using p() and pvec(), respectively.
 *
 * Particles can be tested for equality (having the same id and four-momentum).
 * There are also a few physics-related queries which test the offshellness of
 * the particle by comparing its virtuality to its (pole)mass.
 *
 * \note There is no status information (created, intermediate, decayed, etc.)
 * implemented, which should be handled by higher-level classes.
 *
 * \todo Fixed physical data like mass and charge are currently hard coded.
 * Later they will be obtained via the particle's PDG data.
 */
//------------------------------------------------------------------------------
class Particle
{
   protected:
      // later _id will contain the complete set of quantum numbers necessary
      // to identify the particle including color, spin, etc.
      PdgID _id;                  ///< PDG id of the particle
      FourVector _p;              ///< momentum of the particle

      // later _color will become part of the complete set of quantum numbers
      int _color1;                ///< color label 1;
      int _color2;                ///< color label 2;

      fixed_ptr<const ParticleData> _data;  ///< pointer to PDG data of the particle

   public:
      /// Named constants to specify if a %Particle is on shell or spacelike or lightlike off shell.
      enum Offshellness
      {
         kSpacelike = -1,
         kOnShell = 0,
         kTimelike = 1
      };

      /// \name Constructors etc.
      ///@{

      /// Creates a %Particle with PdgID \a id and zero four-momentum.
      explicit Particle(PdgID id = PdgID::kUnknown);

      /// Creates a %Particle with PdgID \a id and four-momentum \a p.
      Particle(PdgID id, const FourVector& p);

      /// Creates a %Particle with PdgID \a id and the given momentum components.
      Particle(PdgID id, double energy, double px, double py, double pz);

      /// Creates a %Particle with PdgID \a id, and given four-momentum and color.
      Particle(PdgID id, const FourVector& p, int color1, int color2);

      /// Creates a %Particle with PdgID \a id, and given momentum components and color.
      Particle(PdgID id, double energy, double px, double py, double pz, int color1, int color2);

      ///@}

      /// \name Queries and Accessors
      ///@{
      PdgID id() const;                ///< Returns the particle's PdgID.

      const FourVector& p() const;     ///< Returns the particle's four-momentum vector.
      const ThreeVector& pvec() const; ///< Returns the particle's three-momentum vector.
      double energy() const;           ///< Returns the particle's energy.
      double px() const;               ///< Returns the particle's x momentum.
      double py() const;               ///< Returns the particle's y momentum.
      double pz() const;               ///< Returns the particle's z momentum.

      int color1() const;              ///< Returns the particle's first color label.
      int color2() const;              ///< Returns the particle's second color label.

      FourVector& p();                 ///< Access particle's four-momentum vector.
      ThreeVector& pvec();             ///< Access particle's three-momentum vector.
      double& energy();                ///< Access particle's energy.
      double& px();                    ///< Access particle's x momentum.
      double& py();                    ///< Access particle's y momentum.
      double& pz();                    ///< Access particle's z momentum.

      int& color1();                   ///< Access particle's first color label.
      int& color2();                   ///< Access particle's second color label.

      /// Returns a reference to the particle's PDG data.
      const ParticleData& data() const;

      double mass() const;             ///< Returns the particle's mass.
      double charge() const;           ///< Returns the particle's charge.
      ///@}

      /// \name Comparison Operators
      ///@{
      // Particle equality comparison operator
      bool operator==(const Particle& rhs) const;

      // Particle inequality comparison operator
      bool operator!=(const Particle& rhs) const;
      ///@}

      /// \name Physics Queries
      ///@{
      // Returns the offshellness of the Particle.
      int offshellness() const;

      /// Test if the Particle is on shell.
      bool isOnShell() const;

      /// Test if the Particle is timelike off shell.
      bool isTimelikeOffShell() const;

      /// Test if the Particle is spacelike off shell.
      bool isSpacelikeOffShell() const;

      /// Returns the virtuality of the Particle, i.e. the square of its four-momentum.
      double virtuality() const;

      /// Returns the squared mass of the %Particle. Equivalent to mass()*mass().
      double m2() const;
      ///@}

      /// \name Output
      ///@{
      // Particle stream insertion operator
      friend std::ostream& operator<<(std::ostream& out, const Particle& p);
      ///@}

   protected:
      /// Streams a string in the format "(id, p0, p1, p2, p3)" to \a out.
      std::ostream& print(std::ostream& out) const;

   private:
      bool lookupData();
};

} // namespace Geneva


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

#include "Geneva/Core/Utils/MathUtils.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------
// Creates a Particle with PdgID id and zero four-momentum.
inline Particle::Particle(PdgID id)
   : _id(id), _p(), _color1(0), _color2(0), _data()
{}

//------------------------------------------------------------------------------
// Creates a Particle with PdgID id and four-momentum p.
inline Particle::Particle(PdgID id, const FourVector& p)
   : _id(id), _p(p), _color1(0), _color2(0), _data()
{}

//------------------------------------------------------------------------------
// Creates a Particle with PdgID id and the given momentum components.
inline Particle::Particle(PdgID id, double energy, double px, double py, double pz)
   : _id(id), _p(energy, px, py, pz), _color1(0), _color2(0), _data()
{}

//------------------------------------------------------------------------------
// Creates a Particle with PdgID id, and given four-momentum and color.
inline Particle::Particle(PdgID id, const FourVector& p, int color1, int color2)
   : _id(id), _p(p), _color1(color1), _color2(color2), _data()
{}

//------------------------------------------------------------------------------
// Creates a Particle with PdgID id, and given momentum components and color.
inline Particle::Particle(PdgID id, double energy, double px, double py, double pz,
                          int color1, int color2)
   : _id(id), _p(energy, px, py, pz), _color1(color1), _color2(color2), _data()
{}

//------------------------------------------------------------------------------
// Queries and Accessors
//------------------------------------------------------------------------------
inline PdgID Particle::id() const                 { return _id;     }

inline const FourVector& Particle::p() const      { return _p;        }
inline const ThreeVector& Particle::pvec() const  { return _p.pvec(); }
inline double Particle::energy() const            { return _p.p0();   }
inline double Particle::px() const                { return _p.p1();   }
inline double Particle::py() const                { return _p.p2();   }
inline double Particle::pz() const                { return _p.p3();   }

inline int Particle::color1() const               { return _color1;   }
inline int Particle::color2() const               { return _color2;   }

inline FourVector& Particle::p()                  { return _p;        }
inline ThreeVector& Particle::pvec()              { return _p.pvec(); }
inline double& Particle::energy()                 { return _p.p0();   }
inline double& Particle::px()                     { return _p.p1();   }
inline double& Particle::py()                     { return _p.p2();   }
inline double& Particle::pz()                     { return _p.p3();   }

inline int& Particle::color1()                    { return _color1; }
inline int& Particle::color2()                    { return _color2; }

inline const ParticleData& Particle::data() const { return *_data; }

//------------------------------------------------------------------------------
// Comparison Operators
//------------------------------------------------------------------------------
/**
 * \brief %Particle equality comparison operator
 * \param rhs The %Particle to compare to.
 * \return \c true if both the PdgID and fourmomenta are equal, \c false otherwise.
 */
inline bool Particle::operator==(const Particle& rhs) const
{
   return (_id == rhs._id) && (_p == rhs._p);
}

//------------------------------------------------------------------------------
/**
 * \brief %Particle inequality comparison operator
 * \param rhs The %Particle to compare to.
 * \return \c true if the PdgID or fourmomenta are different, \c false otherwise.
 */
inline bool Particle::operator!=(const Particle& rhs) const
{
   return !(*this == rhs);
}

//------------------------------------------------------------------------------
// Physics Queries
//------------------------------------------------------------------------------
/**
 * \brief Returns the #Offshellness of the %Particle.
 * \return An integer representing the #Offshellness of the %Particle.
 *
 * The returned integer can be compared with the named constants defined in
 * #Offshellness to check if the %Particle is on shell, timelike off shell, or
 * spacelike off shell.
 *
 * It compares \f$ E^2 \f$ to \f$ \vec{p}^2 + m^2 \f$ using the default precision and
 * arbitrary accuracy.
 */
inline int Particle::offshellness() const
{
   return compareRelative(_p.p0()*_p.p0(), _p.p1()*_p.p1() + _p.p2()*_p.p2() + _p.p3()*_p.p3() + m2());
}

//------------------------------------------------------------------------------
/**
 * \brief Test if the %Particle is on shell.
 * \return \c true if %Particle is on shell, \c false otherwise.
 *
 * Equivalent to offshellness() == \c kOnShell, which is \c true if
 * \f$ E^2 = \vec{p}^2 + m^2 \f$. See offshellness() for details.
 */
inline bool Particle::isOnShell() const
{
   return offshellness() == kOnShell;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if the %Particle is timelike off shell.
 * \return \c true if %Particle is timelike off shell, \c false otherwise.
 *
 * Equivalent to offshellness() == \c kTimelike, which is \c true if
 * \f$ E^2 > \vec{p}^2 + m^2 \f$. See offshellness() for details.
 */
inline bool Particle::isTimelikeOffShell() const
{
   return offshellness() == kTimelike;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if the %Particle is spacelike off shell.
 * \return \c true if %Particle is spacelike off shell, \c false otherwise.
 *
 * Equivalent to offshellness() == \c kSpacelike, which is \c true if
 * \f$ E^2 < \vec{p}^2 + m^2 \f$. See offshellness() for details.
 */
inline bool Particle::isSpacelikeOffShell() const
{
   return offshellness() == kSpacelike;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the virtuality of the %Particle, i.e. the invariant square of its momentum.
 * \return The value of the invariant square \f$ E^2 - \vec{p}^2 \f$.
 *
 * \note To check if the %Particle is on or off shell, the offshellness(),
 * isOnShell(), isTimelikeOffShell(), or isSpacelikeOffShell() members should be
 * used rather than directly comparing the virtuality to the squared mass by
 * hand.
 */
inline double Particle::virtuality() const
{
   return _p.square();
}

//------------------------------------------------------------------------------
// Returns the squared mass of the Particle. Equivalent to mass()*mass().
inline double Particle::m2() const
{
   return mass() * mass();
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------
/**
 * \brief %Particle stream insertion operator
 * \param out The output stream to be used.
 * \param p The %Particle to be streamed.
 * \return A reference to \a out.
 *
 * Streams a string in the format "{[id], p0, p1, p2, p3, (color1, color2)}" to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const Particle& p)
{
   return p.print(out);
}

} // namespace Geneva

#endif // GENEVA_PARTICLE_HPP
