//------------------------------------------------------------------------------
/// \file ParticleData.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class ParticleData
//------------------------------------------------------------------------------

#ifndef GENEVA_PARTICLE_DATA_HPP
#define GENEVA_PARTICLE_DATA_HPP

#include "Geneva/Core/Physics/PdgID.hpp"

#include <string>

namespace Geneva
{

//------------------------------------------------------------------------------
/** \class ParticleData
 *  \ingroup Core
 *
 * \brief Data container that holds all fixed data about particles.
 *
 * The things stored are
 * - PDG ID
 * - PDG name
 * - mass (with error)
 * - width (with error)
 * - total spin J
 * - parity P
 * - charge conjugation C
 * - electric charge
 * - color representation
 *
 */
//------------------------------------------------------------------------------
class ParticleData
{
   public:
      typedef int Spin;      ///< Represents the spin s as 2*s+1
      typedef int Charge;    ///< Represents the electric Charge Q as 3*Q

      /// Represents parity quantum numbers
      enum Parity
      {
         kUndefined = 0,
         kPlus = +1,
         kMinus = -1
      };

      /// Color representation
      enum ColorRep
      {
         kSinglet = 1,
         kTriplet = 3,
         kAntiTriplet = -3,
         kOctet = 8
      };

   private:
      PdgID _id;             ///< PDG ID
      std::string _name;     ///< name
      double _mass;          ///< mass (including uncertainties)
      double _width;         ///< width (including uncertainties)
      Spin _spin;            ///< total spin
      Parity _pParity;       ///< space-time parity
      Parity _cParity;       ///< charge conjugation parity
      Charge _emCharge;      ///< electric charge
      ColorRep _colorRep;    ///< SU(3) color representation

   public:
      // Constructing and assigning
      // Creates a default ParticleData entry.
      ParticleData();

      /// Creates a %ParticleData object from the given data.
      ParticleData(PdgID id, const std::string& name, double mass, double width,
                   Spin spin, Parity pParity, Parity cParity, Charge emCharge,
                   ColorRep colorRep);

      // Read access
      PdgID id() const;
      const std::string& name() const;
      double mass() const;
      double width() const;
      Spin spin() const;
      Parity pParity() const;
      Parity cParity() const;
      Charge emCharge() const;
      ColorRep colorRep() const;

      // Write access
      PdgID& id();
      std::string& name();
      double& mass();
      double& width();
      Spin& spin();
      Parity& pParity();
      Parity& cParity();
      Charge& emCharge();
      ColorRep& colorRep();
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructing and assigning
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Creates a default %ParticleData entry.
 *
 * All data fields are set to their default values:
 * The PDG ID is set to PdgID::kUnknown, the name is set to an empty string,
 * mass and width to 0.0, spin to 0, parities to kUndefined, emCharge to 0 and
 * colorRep to kSinglet.
 */
inline ParticleData::ParticleData()
   : _id(PdgID::kUnknown), _name(), _mass(0.0), _width(0.0),
     _spin(1), _pParity(kUndefined), _cParity(kUndefined),
     _emCharge(0), _colorRep(kSinglet)
{}

//------------------------------------------------------------------------------
// Creates a ParticleData object from the given data.
inline ParticleData::ParticleData(PdgID id, const std::string& name,
                                  double mass, double width,
                                  Spin spin, Parity pParity, Parity cParity,
                                  Charge emCharge, ColorRep colorRep)
   : _id(id), _name(name), _mass(mass), _width(width),
     _spin(spin), _pParity(pParity), _cParity(cParity),
     _emCharge(emCharge), _colorRep(colorRep)
{}

//------------------------------------------------------------------------------
// Read access
//------------------------------------------------------------------------------

inline PdgID ParticleData::id() const                        { return _id;       }
inline const std::string& ParticleData::name() const         { return _name;     }
inline double ParticleData::mass() const                     { return _mass;     }
inline double ParticleData::width() const                    { return _width;    }
inline ParticleData::Spin ParticleData::spin() const         { return _spin;     }
inline ParticleData::Parity ParticleData::pParity() const    { return _pParity;  }
inline ParticleData::Parity ParticleData::cParity() const    { return _cParity;  }
inline ParticleData::Charge ParticleData::emCharge() const   { return _emCharge; }
inline ParticleData::ColorRep ParticleData::colorRep() const { return _colorRep; }

//------------------------------------------------------------------------------
// Write access
//------------------------------------------------------------------------------

inline PdgID& ParticleData::id()                        { return _id;       }
inline std::string& ParticleData::name()                { return _name;     }
inline double& ParticleData::mass()                     { return _mass;     }
inline double& ParticleData::width()                    { return _width;    }
inline ParticleData::Spin& ParticleData::spin()         { return _spin;     }
inline ParticleData::Parity& ParticleData::pParity()    { return _pParity;  }
inline ParticleData::Parity& ParticleData::cParity()    { return _cParity;  }
inline ParticleData::Charge& ParticleData::emCharge()   { return _emCharge; }
inline ParticleData::ColorRep& ParticleData::colorRep() { return _colorRep; }

} // namespace Geneva

#endif // GENEVA_PARTICLE_DATA_HPP
