//------------------------------------------------------------------------------
/// \file ParticleDataTable.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class ParticleDataTable
//------------------------------------------------------------------------------

#ifndef GENEVA_PARTICLE_DATA_TABLE_HPP
#define GENEVA_PARTICLE_DATA_TABLE_HPP

#include "Geneva/Core/Physics/PdgID.hpp"
#include "Geneva/Core/Physics/ParticleData.hpp"
#include "Geneva/Core/Utils/fixed_ptr.hpp"

#include <set>
#include <map>

namespace Geneva
{

//------------------------------------------------------------------------------
/** \class ParticleDataTable
 *  \ingroup Core
 *
 * \brief A table that holds the ParticleData of all particles.
 *
 * The data is indexed by the \c PdgID, and there can only be a single unique
 * entry for each \c PdgID. The data is accessed by providing either the
 * \c PdgID or the name of the particle as a \c std::string. The access time is
 * logarithmic in either case, but a bit faster when using the \c PdgID (since
 * comparing PdgIDs is faster than comparing strings.) In principle there can
 * be several entries with different PdgIDs that have the same particle name,
 * e.g. if there is user-defined special versions of a particle. Access by name
 * can then be used to find all entries with the same name.
 *
 * Direct write access to the entries in the table is not possible to ensure its
 * integrity. Entries can only be modified by overwriting existing entries.
 * (by clients that have write access to the table itself).
 */
//------------------------------------------------------------------------------
class ParticleDataTable
{
   private:
      /// Function object used to sort the entries in the table.
      struct _compare_type
      {
         bool operator()(const ParticleData& lhs, const ParticleData& rhs) const
         { return lhs.id().code() < rhs.id().code(); }
      };

      /// Set storing the ParticleData indexed by their \c PdgID.
      std::set<ParticleData, _compare_type> _table;

      /// Multimap to look up particles by their names.
      std::multimap<std::string, std::set<ParticleData, _compare_type>::const_iterator> _lookup;

   public:
      /// Constant pointer to a \c ParticleData entry in the table.
      typedef fixed_ptr<const ParticleData> const_ptr;

      /// Constant iterator for \c ParticleData entries in the table.
      typedef std::set<ParticleData, _compare_type>::const_iterator iterator;

      /// Constant reverse_iterator for \c ParticleData entries in the table.
      typedef std::set<ParticleData, _compare_type>::const_reverse_iterator reverse_iterator;

      /// Constant iterator for entries in the lookup table.
      typedef std::multimap<std::string, iterator>::const_iterator lookup_iterator;

      //Constructing and assigning
      /// Creates an empty table.
      ParticleDataTable();

      // Accessing and querying the data
      /// Returns an iterator pointing to the first entry in the table.
      iterator begin() const;

      /// Returns an iterator pointing beyond the last entry in the table.
      iterator end() const;

      /// Returns a reverse_iterator pointing to the last entry in the table.
      reverse_iterator rbegin() const;

      /// Returns a reverse_iterator pointing before the first entry in the table.
      reverse_iterator rend() const;

      /// Returns true if the table is empty.
      bool empty() const;

      /// Returns the total number of entries in the table.
      size_t size() const;

      // Checks if the table has an entry for a given PdgID.
      bool isAvailable(PdgID id) const;

      // Locates the entry for a given PdgID.
      const_ptr getData(PdgID id) const;

      // Counts how many entries of a given particle name exist.
      size_t numAvailable(const std::string& name) const;

      // Locates the nth entry with a given particle name.
      const_ptr getData(const std::string& name, size_t n) const;

      // Finds all entries with a given particle name.
      std::pair<lookup_iterator, lookup_iterator> find(const std::string& name) const;

      // Adding and deleting data
      /// Swaps contents with another %ParticleDataTable.
      void swap(ParticleDataTable& other);

      // Inserts a single ParticleData entry into the table.
      bool insert(const ParticleData& data, bool overwrite = false);

      // Inserts a range of ParticleData entries into the table.
      template <typename InputIterator>
      size_t insert(InputIterator first, InputIterator last, bool overwrite = false);

      // Removes an entry from the table
      void erase(iterator it);

      // Removes the entry for a PdgID from the table.
      bool erase(PdgID id);

      // Removes all entries with a given particle name from the table.
      size_t erase(const std::string& name);

      /// Removes all entries from the table.
      void clear();

   private:
      // Static dummy ParticleData object used during lookup.
      static ParticleData _entry;

      // Helper functions to manage the lookup table.
      void insertLookup(iterator it);
      void eraseLookup(iterator it);
};


////////////////////////////////////////////////////////////////////////////////
// Inline Declarations
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructing and assigning
//------------------------------------------------------------------------------
// Creates an empty table.
inline ParticleDataTable::ParticleDataTable()
   : _table(), _lookup()
{}

//------------------------------------------------------------------------------
// Accessing and querying the data
//------------------------------------------------------------------------------
// Returns an iterator pointing to the first entry in the table.
inline ParticleDataTable::iterator ParticleDataTable::begin() const
{
   return _table.begin();
}

//------------------------------------------------------------------------------
// Returns an iterator pointing beyond the last entry in the tablhiggssing_prod.m_e.
inline ParticleDataTable::iterator ParticleDataTable::end() const
{
   return _table.end();
}

//------------------------------------------------------------------------------
// Returns a reverse_iterator pointing to the last entry in the table.
inline ParticleDataTable::reverse_iterator ParticleDataTable::rbegin() const
{
   return _table.rbegin();
}

//------------------------------------------------------------------------------
// Returns a reverse_iterator pointing before the first entry in the table.
inline ParticleDataTable::reverse_iterator ParticleDataTable::rend() const
{
   return _table.rend();
}

//------------------------------------------------------------------------------
// Returns true if the table is empty.
inline bool ParticleDataTable::empty() const
{
   return _table.empty();
}

//------------------------------------------------------------------------------
// Returns the total number of entries in the table.
inline size_t ParticleDataTable::size() const
{
   return _table.size();
}

//------------------------------------------------------------------------------
/**
 * \brief Checks if the table has an entry for a given \c PdgID.
 * \param id The PdgID to be looked for.
 * \return True if the table has an entry for \p id, false otherwise.
 */
inline bool ParticleDataTable::isAvailable(PdgID id) const
{
   _entry.id() = id;
   return _table.find(_entry) != _table.end();
}

//------------------------------------------------------------------------------
/**
 * \brief Counts how many entries of a given particle name exist.
 * \param name The particle name to be looked for.
 * \return The total number of entries with \p name.
 *
 * If no entry with \p name exists, zero is returned.
 */
inline size_t ParticleDataTable::numAvailable(const std::string& name) const
{
   return _lookup.count(name);
}

//------------------------------------------------------------------------------
/**
 * \brief Finds all entries with a given particle name.
 * \param name The particle name to be looked for.
 * \return A [first, last) range specified by a std::pair of lookup_iterators.
 *
 * The lookup_iterators themselves point to std::pairs, whose first element is
 * a std:string corresponding to \p name and whose second element is an iterator
 * into the table. The first returned iterator specifies the first entry in the
 * table with \p name. It can be incremented to find all entries with \p name.
 * The second returned iterator specifies the end of range. If no entry with
 * \p name exists, the first and last iterators are equal.
 */
inline std::pair<ParticleDataTable::lookup_iterator, ParticleDataTable::lookup_iterator>
ParticleDataTable::find(const std::string& name) const
{
   return _lookup.equal_range(name);
}

//------------------------------------------------------------------------------
// Adding and deleting data
//------------------------------------------------------------------------------
// Swaps contents with another ParticleDataTable.
inline void ParticleDataTable::swap(ParticleDataTable& other)
{
   _table.swap(other._table);
   _lookup.swap(other._lookup);
}

//------------------------------------------------------------------------------
/**
 * \brief Removes an entry from the table.
 * \param it An iterator pointing to the entry to be erased.
 *
 * It is assumed that \p it is a valid iterator into the table; there are no
 * further checkes performed.
 */
inline void ParticleDataTable::erase(ParticleDataTable::iterator it)
{
   eraseLookup(it);
   _table.erase(it);
}


//------------------------------------------------------------------------------
// Removes all entries from the table.
inline void ParticleDataTable::clear()
{
   _table.clear();
   _lookup.clear();
}


} // namespace Geneva

#endif // GENEVA_PARTICLE_DATA_TABLE_HPP
