//------------------------------------------------------------------------------
/// \file PdgID.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012  MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class PdgID
//------------------------------------------------------------------------------

#ifndef GENEVA_PDG_ID_HPP
#define GENEVA_PDG_ID_HPP

#include "Geneva/Core/Utils/MathUtils.hpp"

#include <iosfwd>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class PdgID
 * \ingroup Core
 *
 * \brief The ID of a particle using the PDG numbering scheme.
 *
 * The primary purpose of the class is to represent the abstract type or ID of a
 * particle. The fact that the internal representation is based on the PDG
 * numbering scheme is secondary. (An electron will always be an electron,
 * irrespective of its internal integer code.)
 *
 * Various member functions are provided to test if the ID belongs to a certain
 * type of particles. Two IDs can be compared for (in)equality. Intentionally
 * no ordering operators are defined, since there is no sense in which an
 * electron is less than a gluon.
 *
 * Member functions are also provided to obtain the information that can be
 * deduced from the PDG code, such as the flavor, spin, antiparticle, #NPType.
 * The spin and antiparticles for elementary particles are assumed, since that
 * information is not actually part of their PDG code. Note that no information
 * about any charge quantum numbers is assumed and thus none is provided.
 *
 * The classification of particles follows the PDG classifications.
 * The PDG codes <= 102 are called fundamental. They are split into elementary
 * particles (<= 80) and special codes for interal use in MC (81-102, e.g. jets
 * are often 93). The particles in the fundamental range include all elementary
 * SM particles (<= 25) plus a variety of generic non-SM ones (e.g. 4th gen.
 * quarks/leptons, W', Z', two Higgs doublets, graviton). Quark and gluon bound
 * states (SM hadrons) and diquarks are encoded according to their flavor
 * content and spin. Particles are split up by convention into "particles"
 * and their "antiparticles" (which have negative code).
 *
 * Particles arising in specific new-physics scenarios (SUSY, Kaluza-Klein,
 * Technicolor) are associated with a specific NPType but otherwise mirror the
 * above organization. For this classification, all the above SM particles,
 * diquarks, and non-SM fundamental particles share a common #NPType \c kSMlike.
 * User-defined states have #NPType \c kUserDefined.
 *
 * For now, 10-digit nuclear codes are not supported and are treated as invalid.
 *
 * \note
 * The bare PDG code as an integer is obtained from code() and can be set
 * with setCode(), which checks the given code for validity. However, manual use
 * or manipulation of the PDG code by getting and setting it in this way is
 * \b strongly discouraged.
 */
//------------------------------------------------------------------------------
class PdgID
{
   public:
      /// Named constants for elementary particles and few other frequently used IDs.
      /// Note: (kMin < code) and (kMax > code) will be true for all valid PDG codes
      enum PdgCode
      {
         kUnknown      = 0,
         kDown         = 1,    kDownBar      = -1,
         kUp           = 2,    kUpBar        = -2,
         kStrange      = 3,    kStrangeBar   = -3,
         kCharm        = 4,    kCharmBar     = -4,
         kBottom       = 5,    kBottomBar    = -5,
         kTop          = 6,    kTopBar       = -6,
         kBottom4      = 7,    kBottom4Bar   = -7,
         kTop4         = 8,    kTop4Bar      = -8,
         kElectron     = 11,   kPositron     = -11,
         kENeutrino    = 12,   kENeutrinoBar = -12,
         kMuon         = 13,   kMuonBar      = -13,
         kMNeutrino    = 14,   kMNeutrinoBar = -14,
         kTau          = 15,   kTauBar       = -15,
         kTNeutrino    = 16,   kTNeutrinoBar = -16,
         kTau4         = 17,   kTau4Bar      = -17,
         kT4Neutrino   = 18,   kT4NeutrinoBar = -18,
         kGluon        = 21,
         kPhoton       = 22,
         kZ            = 23,
         kWPlus        = 24,   kWMinus       = -24,
         kHiggs        = 25,
         kZ2           = 32,
         kZ3           = 33,
         kW2Plus       = 34,   kW2Minus      = -34,
         kHiggs2       = 35,
         kHiggs3       = 36,
         kHiggsPlus    = 37,   kHiggsMinus   = -37,
         kGraviton     = 39,
         kJet          = 93,
         kProton       = 2212, kProtonBar    = -2212,
         kNeutron      = 2112, kNeutronBar   = -2112
      };

      /// Named constants identifying specific new-physics scenarios.
      enum NPType
      {
         kSMlike       = 0,  ///< SM and non-SM without their own NPType
         kSusy         = 1,  ///< Boson and left-handed fermion Susy partners, R-Hadrons
         kSusyR        = 2,  ///< Susy partners of right-handed fermions
         kTechniColor  = 3,  ///< Technicolor particles
         kExcited      = 40, ///< Excited (composite) quarks
         kDyon         = 41, ///< Magnetic monopoles and dyons
         kKaluzaKlein  = 5,  ///< Boson and left-handed fermion KK partners
         kKaluzaKleinR = 6,  ///< KK partners of right-handed fermions
         kUserDefined  = 9,  ///< User-defined states
         kQBall        = 10  ///< QBalls
      };

   protected:
      int _code;               ///< the PDG code

   public:
      /// \name Constructors etc.
      ///@{

      /// Constructs %PdgID using kUnknown.
      PdgID();

      // Constructs PdgID from a named constant.
      PdgID(PdgCode kname);

      // Constructs PdgID from an explicit integer code.
      explicit PdgID(int code);

      /// \name Simple Queries and Accessors
      ///@{

      /// Returns the integer code of the %PdgID.
      int code() const;

      // Tries to set PdgID from an explicit integer code.
      bool setCode(int code);
      ///@}

      /// \name Physics Tests
      /// Tests if %PdgID belongs to a specific physics type
      ///@{
      // Test if PdgID is a "particle".
      bool isParticle() const;

      // Test if PdgID is an "antiparticle".
      bool isAntiParticle() const;

      // Test if PdgID is a quark.
      bool isQuark() const;

      // Test if PdgID is a lepton.
      bool isLepton() const;

      // Test if PdgID is a gauge or Higgs boson.
      bool isGaugeBoson() const;

      // Test if PdgID is a gluon.
      bool isGluon() const;

      // Test if PdgID is an exotic particle.
      bool isExotic() const;

      // Test if PdgID is an internal MC particle.
      bool isMC() const;

      // Test if PdgID is an elementary particle.
      bool isElementary() const;

      // Test if PdgID is a meson.
      bool isMeson() const;

      // Test if PdgID is a baryon.
      bool isBaryon() const;

      // Test if PdgID is a diquark.
      bool isDiQuark() const;

      // Test if PdgID is a SM particle.
      bool isSM() const;

      // Test if PdgID is an elementary SM particle.
      bool isSMElementary() const;

      // Test if PdgID is a SM hadron.
      bool isSMHadron() const;

      // Test if PdgID is a particle that is its own antiparticle.
      bool isOwnAntiParticle() const;
      ///@}

      /// \name Additional Physics Queries
      ///@{
      // Checks if an integer represents a valid PDG code.
      bool isValid(int code) const;

      // Returns the AntiPartner of the PdgID.
      PdgID antiPartner() const;

      // Extracts the specific new-physics type of the PdgID.
      int getNPType() const;

      // Extracts the fundamental portion of the PdgID.
      PdgID getFundamental() const;

      // Extracts the total spin encoded by the PdgID.
      int getSpin() const;

      /// \name Comparison
      ///@{
      // PdgID equality comparison operator
      bool operator==(const PdgID& rhs) const;

      // PdgID inequality comparison operator
      bool operator!=(const PdgID& rhs) const;

      // PdgID ordering relation operator
      bool operator<(const PdgID& rhs) const;

      // Test if PdgID is the AntiPartner of other.
      bool isAntiPartner(PdgID other) const;
      ///@}

      /// \name Output
      ///@{
      // PdgID stream insertion operator.
      friend std::ostream& operator<<(std::ostream& out, PdgID id);
      ///@}

   private:
      // lookup table for codes <= 39
      static int _spins[40];

      /// Streams the integer code of the %PdgID to \a out.
      std::ostream& print(std::ostream& out) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Declarations
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------
// Constructs PdgID using kUnknown.
inline PdgID::PdgID()
   : _code(kUnknown)
{}

//------------------------------------------------------------------------------
/**
 * \brief Constructs %PdgID from the given named constant.
 *
 * This constructor allows implicit construction, such that any named constant
 * is automatically promoted to a %PdgID if needed. For example
 * \code
 *    PdgID id;
 *    id = kPositron;
 *    if (id == kElectron)
 * \endcode
 * are valid constructs.
 */
inline PdgID::PdgID(PdgCode kname)
   : _code(kname)
{}

//------------------------------------------------------------------------------
// Simple Queries and Accessors
//------------------------------------------------------------------------------
// Returns the integer code of the PdgID.
inline int PdgID::code() const
{
   return _code;
}

//------------------------------------------------------------------------------
/**
 * \brief Tries to set %PdgID from an explicit integer code.
 * \param code The integer code to be used.
 * \return \c true if successful and \a code is valid, \c false otherwise.
 *
 * If \a code is not valid, the %PdgID is not changed and \c false is returned.
 */
inline bool PdgID::setCode(int code)
{
   if (isValid(code)) {
      _code = code;
      return true;
   }
   return false;
}

//------------------------------------------------------------------------------
// Physics Tests
//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is a "particle".
 * \return \c true for a "particle" (i.e. positive PDG code), \c false otherwise.
 *
 * This refers to what are considered (by convention) "particles" as opposed
 * to "antiparticles". This includes particles that are their own antiparticles.
 */
inline bool PdgID::isParticle() const
{
   return _code > 0;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is an "antiparticle".
 * \return \c true for an "antiparticle" (i.e. negative PDG code), \c false otherwise.
 *
 * This refers to what are considered (by convention) "antiparticles" as opposed
 * to "particles".
 */
inline bool PdgID::isAntiParticle() const
{
   return _code < 0;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is a quark.
 * \return \c true for any quark or antiquark and \c false otherwise.
 *
 * This includes SM and 4th generation quarks and antiquarks (PDG codes 1-8).
 */
inline bool PdgID::isQuark() const
{
   int tmp = abs(_code);
   return tmp <= 8 && tmp != 0;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is a lepton.
 * \return \c true for any lepton or antilepton and \c false otherwise.
 *
 * This includes SM and 4th generation leptons and antileptons (PDG codes 11-18).
 */
inline bool PdgID::isLepton() const
{
   int tmp = abs(_code);
   return tmp >= 11 && tmp <= 18;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is a gauge or Higgs boson.
 * \return \c true for any gauge or Higgs boson and \c false otherwise.
 *
 * This includes SM gauge and Higgs bosons, gauge' bosons, 2-Higgs doublets, and
 * the graviton (PDG codes 21-39).
 */
inline bool PdgID::isGaugeBoson() const
{
   int tmp = abs(_code);
   return tmp >= 21 && tmp <= 39;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is a gluon.
 * \return \c true for a gluon and \c false otherwise.
 *
 * Since the gluon comes up frequently enough, it has its very own test function.
 */
inline bool PdgID::isGluon() const
{
   return _code == kGluon;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is an exotic particle.
 * \return \c true for PDG codes 40-80 and \c false otherwise.
 */
inline bool PdgID::isExotic() const
{
   int tmp = abs(_code);
   return tmp >= 40 && tmp <= 80;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is an internal MC particle.
 * \return \c true for PDG codes 81-102 and \c false otherwise.
 */
inline bool PdgID::isMC() const
{
   int tmp = abs(_code);
   return tmp >= 81 && tmp <= 102;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is an elementary particle.
 * \return \c true for any particle that is considered elementary, \c false otherwise.
 *
 * This includes all particles in the fundamental range (excluding special MC
 * particles), as well as their corresponding new-physics specific partners with
 * #NPType other than \c kSMlike.
 *
 * The return value is meaningless for dyons and QBalls, which fall outside this
 * classification.
 */
inline bool PdgID::isElementary() const
{
   int tmp = abs(_code) % 100000;
   return tmp <= 80 && tmp != 0;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is a meson.
 * \return \c true for any particle-antiparticle bound state, \c false otherwise.
 *
 * This includes the usual qqbar SM mesons (pi0, pi+, etc.), the reggeon and
 * pomeron (PDG codes 110 and 990), as well as new-physics specific
 * two-particle bound states, e.g. Technicolor mesons and SUSY R-mesons.
 *
 * The return value is meaningless for dyons and QBalls, which fall outside this
 * classification.
 */
inline bool PdgID::isMeson() const
{
   // test if nq1 is zero and nq2 is nonzero
   int tmp = abs(_code) % 10000;
   return tmp >= 110 && tmp <= 990;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is a baryon.
 * \return \c true for any bound state of three or more particles, \c false otherwise.
 *
 * This includes the usual SM baryons (proton, neutron, etc.), the odderon (PDG
 * code 9990), pentaquarks, and new-physics specific bound states, e.g.
 * Technicolor baryons and SUSY R-baryons.
 *
 * The return value is meaningless for dyons and QBalls, which fall outside this
 * classification.
 */
inline bool PdgID::isBaryon() const
{
   // test if nq1 is nonzero and nq3 is nonzero
   int tmp = abs(_code) % 10000;
   return tmp >= 1110 && ((tmp % 100) > 10);
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is a diquark.
 * \return \c true for any diquark, \c false otherwise.
 *
 * This includes diquarks, as well as their possible new-physics specific
 * versions.
 *
 * The return value is meaningless for dyons and QBalls, which fall outside this
 * classification.
 */
inline bool PdgID::isDiQuark() const
{
   // test if nq1 is nonzero and nq3 is zero
   int tmp = abs(_code) % 10000;
   return tmp >= 1100 && ((tmp % 100) < 10);
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is an elementary SM particle.
 * \return \c true for any SM quark, lepton, gauge, or Higgs boson, \c false otherwise.
 *
 * This only includes elementary SM particles and antiparticles, including the
 * SM Higgs boson (PDG codes 1-6, 11-16, 21-25). The graviton is not included.
 */
inline bool PdgID::isSMElementary() const
{
   int tmp = abs(_code);
   return tmp <= 25 && tmp != 0 && (tmp % 10 <= 6);
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is a SM particle.
 * \return \c true for any SM particle (elementary or hadron).
 *
 * This is identical to
 * \code isSMElementary() || isSMHadron() \endcode
 */
inline bool PdgID::isSM() const
{
   return isSMElementary() || isSMHadron();
}

//------------------------------------------------------------------------------
// Comparison
//------------------------------------------------------------------------------

/**
 * \brief %PdgID equality comparison operator
 * \param rhs The %PdgID to compare to
 * \return \c true if the PdgIDs are the same, \c false otherwise.
 */
inline bool PdgID::operator==(const PdgID& rhs) const
{
   return _code == rhs._code;
}

//------------------------------------------------------------------------------
/**
 * \brief %PdgID inequality comparison operator
 * \param rhs The %PdgID to compare to
 * \return \c true if the PdgIDs are not the same, \c false otherwise.
 */
inline bool PdgID::operator!=(const PdgID& rhs) const
{
   return _code != rhs._code;
}

//------------------------------------------------------------------------------
/**
 * \brief %PdgID ordering relation operator
 * \param rhs The %PdgID to be compare to
 * \return \c true if the %PdgID is considered "less than" \a rhs, \c false otherwise.
 */
inline bool PdgID::operator<(const PdgID& rhs) const
{
   return _code < rhs._code;
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is the AntiPartner of \a other
 * \param other The %PdgID to compare to.
 * \return \c true if \a other is the AntiPartner of the %PdgID, \c false otherwise.
 *
 * This is equivalent to
 * \code other == AntiPartner() \endcode
 * but faster in most cases.
 */
inline bool PdgID::isAntiPartner(PdgID other) const
{
   if (_code == -other._code)
      return true;
   if (_code == other._code)
      return isOwnAntiParticle();
   return false;
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------
/**
 * \brief %PdgID stream insertion operator.
 * \param out The output stream to be used.
 * \param id The %PdgID to be streamed.
 * \return A reference to \a out.
 *
 * Streams the integer code of the %PdgID to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, PdgID id)
{
   return id.print(out);
}

} // namespace Geneva

#endif // GENEVA_PDG_ID_HPP
