//------------------------------------------------------------------------------
/// \file PhiRadFKS.hpp
//
// Author(s):
//    Frank Tackmann, Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class PhiRadFKS
//------------------------------------------------------------------------------

#ifndef GENEVA_PHI_RAD_FKS_HPP
#define GENEVA_PHI_RAD_FKS_HPP

// #include "Geneva/Core/Physics/Emission.hpp"
#include "Geneva/Core/Physics/PdgID.hpp"

#include <iosfwd>

namespace Geneva
{

class FourVector;
class ThreeVector;
class UnitCubeRad;

//------------------------------------------------------------------------------
/**
 * \class PhiRadFKS
 * \ingroup Core
 *
 * \brief Container for the variables to fully describe the radiation phase space of an emission.
 *
 * The variables used to describe the emission phase space are
 *
 * - \f$ xi \f$: The rescaled energy of the emitted particle.
 * - \f$ y \f$: The cos of the emission angle of the emitted parton
 * - \f$ \phi \f$ The azimuthal angle of the emitted parton
 *
 * Since this is just a container, the precise interpretation of these variables
 * for a specific use case is up to the user. Usually the following phase space
 * limits for the angles should be adhered to (though their enforcement is the
 * responsibility of the user):
 *
 * - \f$ -1 \leq y \leq 1 \f$
 * - \f$ 0 \leq \phi \leq 2\pi \f$
 *
 * The RadiationType is defaulted to FSR in all constructors. This ensures backward
 * compatibility with old calculations that are calling PhiRadFKSGenerator
 * in a non-standard way.
 *
 * \note Currently the ID is just the PdgID, later it will contain the complete
 * set of quantum numbers (e.g. helicities) that are required to specify the
 * corresponding particle.
 */
//------------------------------------------------------------------------------
class PhiRadFKS
{
   public:
      enum RadiationType
      {
         kISRA, /// Initial state radiation from initial particle A
         kISRB, /// Initial state radiation from initial particle A
         kFSR  /// Final state radiation
      };

   private:
      double _xi;            ///< ratio of 2*Efks/Ecm
      double _y;             ///< Angle between daughters
      double _phi;           ///< Azimuthal angle around the boost axis

      PdgID _idMother;       ///< PdgID of the mother
      PdgID _idLeft;         ///< PdgID of the left (1st) daughter (sister)
      PdgID _idRight;        ///< PdgID of the right (2nd) daughter (FKS parton)

      RadiationType _radType;  ///< whether is a final or initial state splitting

   public:
      /// \name Constructors etc.
      ///@{

      /// Creates a default empty %PhiRadFKS.
      PhiRadFKS();

      /// Creates a %PhiRadFKS from the given arguments.
      PhiRadFKS(double xi, double y, double phi, PdgID idMother, PdgID idLeft, PdgID idRight, RadiationType radType);

      /// Creates a %PhiRadFKS from a UniteCubeRad, performing  the change of radiation variables between the unitary cube and the FKS one. Appropriately changing xiMax, it can be used also in the soft, collinear and soft-coll limits. Returns the jacobian of the transformation in the variable jac
      PhiRadFKS(const UnitCubeRad& xRad, double xiMax, double& jac);
      ///@}

      /// \name Read and Write Access
      ///@{
      double xi() const;           ///< Returns xi(y).
      double y() const;            ///< Returns y.
      double phi() const;          ///< Returns phi.

      PdgID idMother() const;      ///< Returns PdgID of the mother.
      PdgID idLeft() const;        ///< Returns PdgID of the left (1st) daughter.
      PdgID idRight() const;       ///< Returns PdgID of the right (2nd) daughter.

      RadiationType radType() const;      ///< Returns radiation type.

      double& xi();                ///< Access xi(y)
      double& y();                 ///< Access y
      double& phi();               ///< Access phi

      PdgID& idMother();          ///< Access PdgID of the mother
      PdgID& idLeft();            ///< Access PdgID of the left (1st) daughter
      PdgID& idRight();           ///< Access PdgID of the right (2nd) daughter

      RadiationType& radType();      ///< Access radiation type.
      ///@}

      /// \name Output
      ///@{
      // PhiRadFKS stream insertion operator
      friend std::ostream& operator<<(std::ostream& out, const PhiRadFKS& PhiRadFKS);
      ///@}

   protected:
      /// Streams info to \a out.
      std::ostream& print(std::ostream& out) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Declarations
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------
// Creates a default PhiRadFKS.
inline PhiRadFKS::PhiRadFKS()
   : _xi(0.), _y(0.), _phi(0.),_idMother(PdgID::kUnknown),
     _idLeft(PdgID::kUnknown), _idRight(PdgID::kUnknown), _radType(kFSR)
{}

//------------------------------------------------------------------------------
// Creates a PhiRadFKS from the given arguments.
inline PhiRadFKS::PhiRadFKS(double xi, double y, double phi, PdgID idMother, PdgID idLeft, PdgID idRight, RadiationType radType)
   : _xi(xi), _y(y), _phi(phi), _idMother(idMother), _idLeft(idLeft), _idRight(idRight), _radType(radType)
{}

//------------------------------------------------------------------------------
// Read and Write Access
//------------------------------------------------------------------------------
inline double PhiRadFKS::xi() const
{
   return _xi;
}

inline double PhiRadFKS::y() const
{
   return _y;
}
inline double PhiRadFKS::phi() const
{
   return _phi;
}
inline PdgID PhiRadFKS::idMother() const
{
   return _idMother;
}
inline PdgID PhiRadFKS::idLeft() const
{
   return _idLeft;
}
inline PdgID PhiRadFKS::idRight() const
{
   return _idRight;
}
inline PhiRadFKS::RadiationType PhiRadFKS::radType() const
{
   return _radType;
}
inline double& PhiRadFKS::xi()
{
   return _xi;
}

inline double& PhiRadFKS::y()
{
   return _y;
}
inline double& PhiRadFKS::phi()
{
   return _phi;
}

inline PdgID& PhiRadFKS::idMother()
{
   return _idMother;
}
inline PdgID& PhiRadFKS::idLeft()
{
   return _idLeft;
}
inline PdgID& PhiRadFKS::idRight()
{
   return _idRight;
}
inline PhiRadFKS::RadiationType& PhiRadFKS::radType()
{
   return _radType;
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------
/**
 * \brief %PhiRadFKS stream insertion operator
 * \param out The output stream to be used.
 * \param PhiRadFKS The %PhiRadFKS to be streamed.
 * \return A reference to \a out.
 *
 * Streams the info about PhiRadFKS to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const PhiRadFKS& PhiRadFKS)
{
   return PhiRadFKS.print(out);
}

} // namespace Geneva

#endif // GENEVA_PHI_RAD_FKS_HPP
