//------------------------------------------------------------------------------
/// \file QCD.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definitions of QCD specific constants and anomalous dimensions
//------------------------------------------------------------------------------

#ifndef GENEVA_QCD_HPP
#define GENEVA_QCD_HPP

#include "Geneva/Core/Physics/Particle.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \struct QCD
 * \ingroup Core
 *
 * \brief Definitions of QCD constants and anomalous dimensions
 *
 * Holds a collection of QCD specific constants and anomalous dimensions, where
 * the gauge group is SU(3).
 *
 * The group constants Nc, CA, CF, TF are accessed as \c QCD::constant.
 *
 * The beta function coefficients are accessed as \c QCD::beta0(nf) etc.
 * The coefficients of the cusp anomalous dimension are accessed as
 * \c QCD::Gamma0(nf) etc. Both are defined in MSbar and are functions of the
 * number of light flavors \a nf. The coefficients are defined with the
 * following conventions
 * \todo add formulas (B16) of physics.tex
 *
 */
//------------------------------------------------------------------------------
struct QCD : SU<3>
{
   /// \f$ \beta_0 \f$
   static double beta0(int nf)
   { return 11./3.*CA - 4./3.*TF*nf; }

   /// \f$ \beta_1 \f$
   static double beta1(int nf)
   { return 34./3.*CA*CA - 20./3.*CA*TF*nf - 4.*CF*TF*nf;}

   /// \f$ \beta_2 \f$
   static double beta2(int nf)
   {
      return 1./54.*(2857.*pow(CA, 3)
                     + (54.*CF*CF - 615.*CF*CA - 1415.*CA*CA) * 2.*TF*nf
                     + (66.*CF + 79.*CA) * pow(2.*TF*nf,2));
   }

   /// \f$ \Gamma_0 \f$
   static double Gamma0(int nf)
   { return 4.; }

   /// \f$ \Gamma_1 \f$
   static double Gamma1(int nf)
   { return (268./9. - 4./3.*math_constants::pi2()) * CA - 80./9. * TF*nf; }

   /// \f$ \Gamma_2 \f$
   static double Gamma2(int nf)
   {
      return (490./3. - 536./27.*math_constants::pi2() + 44./45.*math_constants::pi4() + 88./3.*math_constants::zeta3()) * CA*CA
             + ((-1672./27. + 160./27.*math_constants::pi2() - 224./3.*math_constants::zeta3()) * CA
                + (-220./3. + 64.*math_constants::zeta3()) * CF) * TF*nf
             - 64./27. * pow(TF*nf, 2);
   }

   //------------------------------------------------------------------------------
   // Casimirs, anomalous dimensions and AP splitting kernels with derivatives
   //------------------------------------------------------------------------------

   static double C(PdgID id)
   {
      if (id.isQuark())
         return CF;
      if (id.isGluon())
         return CA;
      return 0.;
   }

   static double gamma(PdgID id, int nf)
   {
      if (id.isQuark())
         return 3./2. * CF;
      if (id.isGluon())
         return 11./6. * CA - 2./3. * TF*nf;
      return 0.;
   }

   static double gammap(PdgID id, int nf)
   {
      if (id.isQuark())
         return (13./2. - 2./3.*math_constants::pi2()) * CF;
      if (id.isGluon())
         return (67./9. - 2./3.*math_constants::pi2()) * CA - 23./9. * TF*nf;
      else
         return 0.;
   }

   //------------------------------------------------------------------------------
   // AP splitting kernels with derivatives
   //------------------------------------------------------------------------------

   static double OneMinusZPgg(double z)
   {
      const double y2 = pow(1. - z, 2);
      return 2.*CA * (z + y2 / z + z*y2);
   }

   static double OneMinusZPqq(double z)
   {
      return CF * (1. + z*z);
   }

   static double OneMinusZPqg(double z)
   {
      const double y = 1. - z;
      return CF * (1. + y*y) / z * y;
   }

   static double OneMinusZPgq(double z)
   {
      const double y = 1. - z;
      return TF * (1. - 2.*z*y) * y;
   }

   static double Pprimeqq(double z)
   {
      return -CF * (1. - z);
   }

   static double Pprimeqg(double z)
   {
      return -CF *z;
   }

   static double Pprimegq(double z)
   {
      return -TF *2.*z*(1. - z);
   }
};

} // namespace Geneva

#endif // GENEVA_QCD_HPP
