//------------------------------------------------------------------------------
/// \file QCDUtils.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definitions of function objects to check for QCD particles and interactions.
//------------------------------------------------------------------------------

#ifndef GENEVA_QCD_UTILS_HPP
#define GENEVA_QCD_UTILS_HPP

#include "Geneva/Core/Physics/SignatureSet.hpp"
#include "Geneva/Core/Physics/ConfigurationExtensions.hpp"
#include "Geneva/Core/Physics/Particle.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \struct IsQCD
 * \ingroup Core
 *
 * \brief Check if a PdgID or Particle is strongly interacting.
 * \param id The ID (or particle) to check.
 * \return \c true if \a id is strongly interacting, \c false otherwise.
 *
 * Checks if the given \c PdgID or \c Particle is g, q or qbar.
 */
//------------------------------------------------------------------------------
struct IsQCD
{
   bool operator()(PdgID id) const
   {
      return id.isGluon() || id.isQuark();
   }

   bool operator()(const Particle& p) const
   {
      return operator()(p.id());
   }
};


//------------------------------------------------------------------------------
/**
 * \struct IsQCDPair
 * \ingroup Core
 *
 * \brief Checks if two PdgIDs or Particles are both strongly interacting.
 * \param id1 The 1st ID (or particle) to check.
 * \param id2 The 2nd ID (or particle) to check.
 * \return \c true if \a id1 and \a id2 are both strongly interacting, \c false
 *         otherwise.
 *
 * Checks if both given IDs satisfy \c IsQCD(), i.e., if both are g, q or qbar.
 */
//------------------------------------------------------------------------------
struct IsQCDPair
{
   bool operator()(PdgID id1, PdgID id2) const
   {
      return IsQCD()(id1) && IsQCD()(id2);
   }

   bool operator()(const Particle& p1, const Particle& p2) const
   {
      return IsQCD()(p1.id()) && IsQCD()(p2.id());
   }
};


//------------------------------------------------------------------------------
/**
 * \struct IsElectroWeak
 * \ingroup Core
 *
 * \brief Checks if a PdgID or Particle is an electroweak lepton or boson.
 * \param id The entry to check.
 * \return \c true if \a id is an electroweak lepton or boson, \c false otherwise.
 *
 * Checks if the ID is a lepton or a gauge boson other than a gluon or a Higgs
 * boson.
 *
 * \note This is not equivalent to \c !IsQCD(id).
 */
//------------------------------------------------------------------------------
struct IsElectroWeak
{
   bool operator()(PdgID id) const
   {
      return id.isLepton() || (id.isGaugeBoson() && !id.isGluon());
   }

   bool operator()(const Particle& p) const
   {
      return operator()(p.id());
   }
};


//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Checks if two particles can arise from a QCD splitting.
 * \param p1 The 1st particle to check.
 * \param p2 The 2nd particle to check.
 * \return \c true if \a p1 and \a p2 can come from QCD splitting, \c false
 *         otherwise.
 *
 * Checks if the two particles are either gg, gq, gqbar, qg, qbarg, or qqbar,
 * properly taking into account the crossing of flavour when one of the two
 * are in the initial state. If neither of them is in the final state it returns
 * \c false.
 *
 * Since this has to explicitly check for \c isFinal() it can only be used with
 * Configuration or Signature entries as \c IsQCDSplitting()(*it1, *it2), where
 * it1 and it2 are valid Configuration or Signature iterators.  If it is known
 * that \a p1 and \a p2 are both final states, IsQCDSplittingFSR can be used
 * instead, which does not have this restriction.
 *
 * \note This does not assume (and effectively includes checks) that \a p1 and
 * \a p2 satisfy \c IsQCD().
 */
//------------------------------------------------------------------------------
class IsQCDSplitting
{
   private:
      bool _allowqqbar;

   public:
      IsQCDSplitting() : _allowqqbar(true) {}
      IsQCDSplitting(bool allowqqbar) : _allowqqbar(allowqqbar) {}

      template <typename SignatureEntry1, typename SignatureEntry2>
      bool operator()(const SignatureEntry1& p1, const SignatureEntry2& p2) const
      {
         // discard the case in which both are initial states
         if (!(p1.isFinal() || p2.isFinal()))
            return false;

         // check combinations involving gluons (these are final vs. initial independent)
         if (p1.id().isGluon())
            return p2.id().isGluon() || p2.id().isQuark();
         if (p2.id().isGluon())
            return p1.id().isQuark();

         // check qqbar combinations
         if (_allowqqbar && p1.id().isQuark()) {
            if (p1.isFinal() && p2.isFinal())            // final-final
               return p1.id().isAntiPartner(p2.id());
            else                                         // initial-final
               return p1.id() == p2.id();
         }
         return false;
      }
};

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Checks if two Particles or IDs can arise from a FSR QCD splitting.
 * \param p1 The 1st Particle to check.
 * \param p2 The 2nd Particle to check.
 * \return \c true if \a p1 and \a p2 can come from a QCD splitting, \c false
 *         otherwise.
 *
 * Checks if the two particles are either gg, gq, gqbar, qg, qbarg, or qqbar.
 * qqbar splittings can be explicitly disallowed by passing an optional argument
 * \c IsQCDSplittingFSR(false).
 *
 * \note This assumes that p1 and p2 are final-state particles. It does not
 * assume (and effectively includes checks) that p1 and p2 satisfy \c IsQCD().
 */
//------------------------------------------------------------------------------
class IsQCDSplittingFSR
{
   private:
      bool _allowqqbar;

   public:
      IsQCDSplittingFSR() : _allowqqbar(true) {}
      IsQCDSplittingFSR(bool allowqqbar) : _allowqqbar(allowqqbar) {}

      template <typename SignatureEntry1, typename SignatureEntry2>
      bool operator()(const SignatureEntry1& p1, const SignatureEntry2& p2) const
      {
         return operator()(p1.id(), p2.id());
      }

      bool operator()(const PdgID& id1, const PdgID& id2) const
      {
         // check combinations involving gluons
         if (id1.isGluon())
            return id2.isGluon() || id2.isQuark();
         if (id2.isGluon())
            return id1.isQuark();

         // check if an allowed qqbar splitting
         return _allowqqbar && id1.isQuark() && id1.isAntiPartner(id2);
      }
};

//------------------------------------------------------------------------------
/**
 * \brief Checks if an emitted/sister pair can come from a valid QCD splitting.
 *
 * This is a more extensive check than \c IsQCDSplitting() taking into account
 * the allowed underlying PhiN signatures in case of an emitted quark.
 *
 * For an emitted gluon it effectively returns \c IsQCD()(sister).
 * For an emitted quark it explicitly checks whether the %Signature that results
 * from combining emitted and sister is included in the given %SignatureSet.
 *
 * \note: It is the user's responsibility to ensure that emitted and sister are
 * valid iterators.
 */
inline bool invalidFlavorProjection(const SignatureSet& sigSet, PhiNplus1 phiNp1)
{
   if (!IsQCDSplitting()(*phiNp1.emitted(), *phiNp1.sister()))
      return true;

   const PdgID emittedID = phiNp1.emitted()->id();
   const PdgID sisterID = phiNp1.sister()->id();

   // At this point we can assume a valid QCD splitting.
   // If a gluon was emitted the underlying Born will always be valid.
   if (emittedID.isGluon())
      return false;

   // emitted is a quark, so we need to construct the underlying Born signature.
   assert(emittedID.isQuark());
   Signature sigtmp(phiNp1);
   if (phiNp1.sister()->isInitialA()) {
      sigtmp.erase(sigtmp.findFinal(emittedID));
      if (sisterID.isQuark()) {
         sigtmp.setInitialA(PdgID::kGluon);
      } else {
         assert(sisterID.isGluon());
         sigtmp.setInitialA(emittedID.antiPartner());
      }
   } else if (phiNp1.sister()->isInitialB()) {
      sigtmp.erase(sigtmp.findFinal(emittedID));
      if (sisterID.isQuark()) {
         sigtmp.setInitialB(PdgID::kGluon);
      } else {
         assert(sisterID.isGluon());
         sigtmp.setInitialB(emittedID.antiPartner());
      }
   } else { // sister must be final
      if (sisterID.isGluon()) {
         sigtmp.erase(sigtmp.findFinal(sisterID));
      } else {
         assert(sisterID.isQuark());
         assert(emittedID.isAntiPartner(sisterID));
         sigtmp.erase(sigtmp.findFinal(emittedID));
         sigtmp.erase(sigtmp.findFinal(sisterID));
         sigtmp.insertFinal(PdgID::kGluon);
      }
   }

   assert(sigtmp.size() == phiNp1.size() - 1);
   return !sigSet.includes(sigtmp);
}

} // namespace Geneva

#endif // GENEVA_QCD_UTILS_HPP
