//------------------------------------------------------------------------------
/// \file RGE.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definitions of RGE evolution functions
//------------------------------------------------------------------------------

#ifndef GENEVA_RGE_HPP
#define GENEVA_RGE_HPP

#include "Geneva/Core/Physics/QCD.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <complex>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \struct RGE
 * \ingroup Core
 *
 * \brief Definitions of RGE evolution functions
 */
//------------------------------------------------------------------------------
struct RGE
{
   /// \f$ K_\Gamma \f$ to LL
   static double KGammaLL(double as0, double as, int nf)
   {
      const double r = as/as0;
      const double G0 = QCD::Gamma0(nf);
      const double b0 = QCD::beta0(nf);

      return -0.25*G0/(b0*b0) * math_constants::_4pi()/as0 * (1. - 1./r - log(r));
   }

   /// \f$ K_\Gamma \f$ to NLL
   static double KGammaNLL(double as0, double as, int nf) {
      const double r = as/as0;
      const double lnr = log(r);
      const double G0 = QCD::Gamma0(nf);
      const double G10 = QCD::Gamma1(nf) / G0;
      const double b0 = QCD::beta0(nf);
      const double b10 = QCD::beta1(nf) / b0;

      return - 0.25*G0/(b0*b0)
             * (math_constants::_4pi()/as0 * (1. - 1./r - lnr)
                + (G10 - b10) * (1. - r + lnr) + 0.5*b10 * lnr*lnr
               );
   }

   /// \f$ K_\Gamma \f$ to NNLL
   static double KGammaNNLL(double as0, double as, int nf)
   {
      const double r = as/as0;
      const double lnr = log(r);
      const double G0 = QCD::Gamma0(nf);
      const double G10 = QCD::Gamma1(nf) / G0;
      const double G20 = QCD::Gamma2(nf) / G0;
      const double b0 = QCD::beta0(nf);
      const double b10 = QCD::beta1(nf) / b0;
      const double b20 = QCD::beta2(nf) / b0;
      as0 *= math_constants::_1_4pi();

      return - 0.25*G0/(b0*b0)
             * (1./as0 * (1. - 1./r - lnr)
                + (G10 - b10) * (1. - r + lnr) + 0.5*b10 * lnr*lnr
                + as0 * ((b10*b10 - b20) * (0.5 - 0.5*r*r + lnr)
                         + (b10*G10 - b10*b10) * (1. - r + r*lnr)
                         - (G20 - b10*G10) * 0.5*pow(1.-r, 2))
               );
   }

   /// \f$ \eta_\Gamma \f$ to LL
   static double etaGammaLL(double as0, double as, int nf)
   {
      const double r = as/as0;
      const double G0 = QCD::Gamma0(nf);
      const double b0 = QCD::beta0(nf);

      return -0.5*G0/b0 * log(r);
   }

   /// \f$ \eta_\Gamma \f$ to NLL
   static double etaGammaNLL(double as0, double as, int nf)
   {
      const double r = as/as0;
      const double G0 = QCD::Gamma0(nf);
      const double G10 = QCD::Gamma1(nf) / G0;
      const double b0 = QCD::beta0(nf);
      const double b10 = QCD::beta1(nf) / b0;

      return -0.5*G0/b0 * (log(r) + as0*math_constants::_1_4pi() * (G10 - b10) * (r - 1.));
   }

   /// \f$ \eta_\Gamma \f$ to NNLL
   static double etaGammaNNLL(double as0, double as, int nf)
   {
      const double r = as/as0;
      const double G0 = QCD::Gamma0(nf);
      const double G10 = QCD::Gamma1(nf) / G0;
      const double G20 = QCD::Gamma2(nf) / G0;
      const double b0 = QCD::beta0(nf);
      const double b10 = QCD::beta1(nf) / b0;
      const double b20 = QCD::beta2(nf) / b0;
      as0 *= math_constants::_1_4pi();

      return -0.5*G0/b0
             * (log(r)
                + as0 * (G10 - b10) * (r - 1.)
                + as0*as0 * (G20 - b10*(G10 - b10) - b20) * 0.5*(r*r - 1.)
               );
   }

   /// \f$ K_\gamma \f$ to LL
   static double KgammaLL(double as0, double as, int nf)
   {
      return 0.0;
   }

   /// \f$ K_\gamma \f$ to NLL
   static double KgammaNLL(double as0, double as, int nf, double g0)
   {
      const double r = as/as0;
      const double b0 = QCD::beta0(nf);

      return -0.5*g0/b0 * log(r);
   }

   /// \f$ K_\gamma \f$ to NNLL
   static double KgammaNNLL(double as0, double as, int nf, double g0, double g1)
   {
      const double r = as/as0;
      const double b0 = QCD::beta0(nf);
      const double b10 = QCD::beta1(nf) / b0;

      return -0.5/b0 * (g0*log(r) + as0*math_constants::_1_4pi() * (g1 - g0*b10) * (r - 1.));
   }

   /// complex valued \f$ K_\Gamma \f$ to LL, defined to take complex as0
   static std::complex<double> KGammaLL(std::complex<double> as0, double as, int nf)
   {
      const std::complex<double> r = as/as0;
      const double G0 = QCD::Gamma0(nf);
      const double b0 = QCD::beta0(nf);

      return -0.25*G0/(b0*b0) * math_constants::_4pi()/as0 * (1. - 1./r - log(r));
   }

   /// complex valued \f$ K_\Gamma \f$ to NLL, defined to take complex as0
   static std::complex<double> KGammaNLL(std::complex<double> as0, double as, int nf) {
      const std::complex<double> r = as/as0;
      const std::complex<double> lnr = log(r);
      const double G0 = QCD::Gamma0(nf);
      const double G10 = QCD::Gamma1(nf) / G0;
      const double b0 = QCD::beta0(nf);
      const double b10 = QCD::beta1(nf) / b0;

      return - 0.25*G0/(b0*b0)
      * (math_constants::_4pi()/as0 * (1. - 1./r - lnr)
         + (G10 - b10) * (1. - r + lnr) + 0.5*b10 * lnr*lnr
         );
   }

   /// complex valued \f$ K_\Gamma \f$ to NNLL, defined to take complex as0
   static std::complex<double> KGammaNNLL(std::complex<double> as0, double as, int nf)
   {
      const std::complex<double> r = as/as0;
      const std::complex<double> lnr = log(r);
      const double G0 = QCD::Gamma0(nf);
      const double G10 = QCD::Gamma1(nf) / G0;
      const double G20 = QCD::Gamma2(nf) / G0;
      const double b0 = QCD::beta0(nf);
      const double b10 = QCD::beta1(nf) / b0;
      const double b20 = QCD::beta2(nf) / b0;
      as0 *= math_constants::_1_4pi();

      return - 0.25*G0/(b0*b0)
      * (1./as0 * (1. - 1./r - lnr)
         + (G10 - b10) * (1. - r + lnr) + 0.5*b10 * lnr*lnr
         + as0 * ((b10*b10 - b20) * (0.5 - 0.5*r*r + lnr)
                  + (b10*G10 - b10*b10) * (1. - r + r*lnr)
                  - (G20 - b10*G10) * 0.5*pow(1.-r, 2))
         );
   }

   /// complex valued \f$ \eta_\Gamma \f$ to LL, defined to take complex as0
   static std::complex<double> etaGammaLL(std::complex<double> as0, double as, int nf)
   {
      const std::complex<double> r = as/as0;
      const double G0 = QCD::Gamma0(nf);
      const double b0 = QCD::beta0(nf);

      return -0.5*G0/b0 * log(r);
   }

   /// complex valued \f$ \eta_\Gamma \f$ to NLL, defined to take complex as0
   static std::complex<double> etaGammaNLL(std::complex<double> as0, double as, int nf)
   {
      const std::complex<double> r = as/as0;
      const double G0 = QCD::Gamma0(nf);
      const double G10 = QCD::Gamma1(nf) / G0;
      const double b0 = QCD::beta0(nf);
      const double b10 = QCD::beta1(nf) / b0;

      return -0.5*G0/b0 * (log(r) + as0*math_constants::_1_4pi() * (G10 - b10) * (r - 1.));
   }

   ///complex valued \f$ \eta_\Gamma \f$ to NNLL, defined to take complex as0
   static std::complex<double> etaGammaNNLL(std::complex<double> as0, double as, int nf)
   {
      const std::complex<double> r = as/as0;
      const double G0 = QCD::Gamma0(nf);
      const double G10 = QCD::Gamma1(nf) / G0;
      const double G20 = QCD::Gamma2(nf) / G0;
      const double b0 = QCD::beta0(nf);
      const double b10 = QCD::beta1(nf) / b0;
      const double b20 = QCD::beta2(nf) / b0;
      as0 *= math_constants::_1_4pi();

      return -0.5*G0/b0
      * (log(r)
         + as0 * (G10 - b10) * (r - 1.)
         + as0*as0 * (G20 - b10*(G10 - b10) - b20) * 0.5*(r*r - 1.)
         );
   }

   /// complex valued \f$ K_\gamma \f$ to LL, defined to take complex as0
   static std::complex<double> KgammaLL(std::complex<double> as0, double as, int nf)
   {
      return 0.0;
   }

   /// complex valued \f$ K_\gamma \f$ to NLL, defined to take complex as0
   static std::complex<double> KgammaNLL(std::complex<double> as0, double as, int nf, double g0)
   {
      const std::complex<double> r = as/as0;
      const double b0 = QCD::beta0(nf);

      return -0.5*g0/b0 * log(r);
   }

   /// complex valued \f$ K_\gamma \f$ to NNLL, defined to take complex as0
   static std::complex<double> KgammaNNLL(std::complex<double> as0, double as, int nf, double g0, double g1)
   {
      const std::complex<double> r = as/as0;
      const double b0 = QCD::beta0(nf);
      const double b10 = QCD::beta1(nf) / b0;

      return -0.5/b0 * (g0*log(r) + as0*math_constants::_1_4pi() * (g1 - g0*b10) * (r - 1.));
   }
};

} // namespace Geneva

#endif // GENEVA_RGE_HPP
