//------------------------------------------------------------------------------
/// \file Rate.hpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 CERN, DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Rate
//------------------------------------------------------------------------------

#ifndef GENEVA_RATE_HPP
#define GENEVA_RATE_HPP

#include "Geneva/Core/Physics/Sigma.hpp"

#include <cstddef>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class Rate
 * \ingroup Core
 *
 * \brief An accumulated cross section including variations and uncertainties.
 *
 * A Rate corresponds to a single cross section. Its value is a Sigma, so it
 * automatically keeps track of the central value and all variations. It keeps
 * track of the accumulated weights, squared weights, and total number of
 * entries, which are used to obtain the value and statistical uncertainty.
 *
 * Basic algebraic operations are also supported.
 *
 * \note: A Rate does not support correlated weights, e.g. for a NLO calculation.
 */
//------------------------------------------------------------------------------
class Rate
{
   private:
      Sigma _sumWeights;          ///< Sigma for the sum of weights
      Sigma _sumWeightsSq;        ///< Sigma for the sum of squared weights
      size_t _entries;            ///< Number of nonzero entries
      size_t _totEntries;         ///< Total number of entries
      bool _variationsResized;    ///< Whether the number of variations has been initialized

   public:
      /// \name Constructors etc.
      ///@{

      /// Creates a zero Rate.
      Rate();

      /// Creates a Rate from the given data.
      Rate(Sigma sumWeights, Sigma sumWeightsSq, size_t entries, size_t totEntries);

      ///@}

      /// \name Queries and Accessors
      ///@{
      const Sigma& sumWeights() const;      ///< Returns the sum of weights.
      const Sigma& sumWeightsSq() const;    ///< Returns the sum of squared weights.

      size_t entries() const;               ///< Returns the number of nonzero entries.
      size_t totEntries() const;            ///< Returns the number of total entries.

      /// Calculates the value.
      Sigma value() const;

      // Calculates the statistical uncertainty.
      Sigma error() const;
      ///@}

      /// \name Manipulation
      ///@{

      // Adds a weight to the %Rate.
      void addWeight(const Sigma& weight);

      /// Adds a zero (empty) weight to the %Rate.
      void addWeight();

      /// Resets the %Rate to zero.
      void reset();
      ///@}

      /// \name Algebra
      ///@{
      // Rate addition assignment operator
      Rate& operator+=(const Rate& rhs);

      // Rate subtraction assignment operator
      Rate& operator-=(const Rate& rhs);

      // Rate multiplication with real number assignment operator
      Rate& operator*=(double rhs);

      // Rate division by real number assignment operator
      Rate& operator/=(double rhs);
      ///@}

      /// \name Output
      ///@{
      // Rate stream insertion operator
      friend std::ostream& operator<<(std::ostream& out, const Rate& rate);
      ///@}

   private:
      /// Streams the entries and central value +/- error to \a out.
      std::ostream& print(std::ostream& out) const;
};

// Rate addition operator
const Rate operator+(const Rate& lhs, const Rate& rhs);

// Rate subtraction operator
const Rate operator-(const Rate& lhs, const Rate& rhs);

// Rate multiplication operator with real number from the left
const Rate operator*(double lhs, const Rate& rhs);

// Rate multiplication operator with real number from the right
const Rate operator*(const Rate& lhs, double rhs);

// Rate division operator with real number
const Rate operator/(const Rate& lhs, double rhs);


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Creates a zero Rate.
inline Rate::Rate()
   : _sumWeights(0.0), _sumWeightsSq(0.0), _entries(0), _totEntries(0),
     _variationsResized(false)
{}

//------------------------------------------------------------------------------
// Creates a Rate from the given data.
inline Rate::Rate(Sigma sumWeights, Sigma sumWeightsSq, size_t entries, size_t totEntries)
   : _sumWeights(sumWeights), _sumWeightsSq(sumWeightsSq), _entries(entries), _totEntries(totEntries),
     _variationsResized(true)
{}

//------------------------------------------------------------------------------
// Queries and Accessors
//------------------------------------------------------------------------------

inline const Sigma& Rate::sumWeights() const   { return _sumWeights; }
inline const Sigma& Rate::sumWeightsSq() const { return _sumWeightsSq; }

inline size_t Rate::entries() const            { return _entries; }
inline size_t Rate::totEntries() const         { return _totEntries; }

//------------------------------------------------------------------------------
// Calculates the value.
inline Sigma Rate::value() const
{
   return _sumWeights/static_cast<double>(_totEntries);
}

//------------------------------------------------------------------------------
// Manipulation
//------------------------------------------------------------------------------

// Adds an empty weight to the %Rate.
inline void Rate::addWeight()
{
   _totEntries++;
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %Rate addition operator
 * \param lhs The 1st %Rate to add.
 * \param rhs The 2nd %Rate to add.
 * \return A %Rate given by the sum of \a lhs and \a rhs.
 */
inline const Rate operator+(const Rate& lhs, const Rate& rhs)
{
   Rate tmp(lhs);
   tmp += rhs;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Rate subtraction operator
 * \param lhs The 1st %Rate
 * \param rhs The 2nd %Rate to be subtracted from \a lhs.
 * \return A %Rate given by the difference of \a lhs and \a rhs.
 */
inline const Rate operator-(const Rate& lhs, const Rate& rhs)
{
   Rate tmp(lhs);
   tmp -= -1.*rhs;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Rate multiplication operator with real number from the left
 * \param lhs The number to multiply with.
 * \param rhs The %Rate to be multiplied by \a lhs.
 * \return A %Rate given by the product of \a lhs times \a rhs.
 */
inline const Rate operator*(double lhs, const Rate& rhs)
{
   Rate tmp(rhs);
   tmp *= lhs;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Rate multiplication operator with real number from the right
 * \param lhs The %Rate to be multiplied by \a rhs.
 * \param rhs The number to multiply with.
 * \return A %Rate given by the product of \a lhs times \a rhs.
 */
inline const Rate operator*(const Rate& lhs, double rhs)
{
   Rate tmp(lhs);
   tmp *= rhs;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Rate division operator with real
 * \param lhs The %Rate to be divided by \a rhs.
 * \param rhs The number to divide by.
 * \return A %Rate given by \a lhs divided by \a rhs.
 * \note Division by zero is not checked.
 */
inline const Rate operator/(const Rate& lhs, double rhs)
{
   Rate tmp(lhs);
   tmp /= rhs;
   return tmp;
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %Rate stream insertion operator
 * \param out The output stream to be used.
 * \param sigma The %Rate to be streamed.
 * \return A reference to \a out.
 *
 * Streams a string in the format "entries  value +/- error" to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const Rate& rate)
{
   return rate.print(out);
}

} // namespace Geneva

#endif // GENEVA_RATE_HPP
