//------------------------------------------------------------------------------
/// \file Sigma.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Sigma
//------------------------------------------------------------------------------

#ifndef GENEVA_SIGMA_HPP
#define GENEVA_SIGMA_HPP

#include <iostream>
#include <vector>
#include <cassert>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class Sigma
 * \ingroup Core
 *
 * \brief Represents a (cross section) central value with a number of variations.
 *
 * Behaves essentially as a container for a central value together with a
 * std::vector of variations, whose type is given by Sigma::Variations. The
 * management of the vector itself is the user's responsibility. The Sigma has a
 * flag isDefined() which indicates if the cross section is considered defined
 * or undefined.
 */
//------------------------------------------------------------------------------
class Sigma
{
   public:
      /// short-hand type definition for the variations
      typedef std::vector<double> Variations;
      /// const_iterator for the variations
      typedef Variations::const_iterator const_iterator;
      /// iterator for the variations
      typedef Variations::iterator iterator;

   private:
      double _central;                 ///< the %Sigma's central value
      Variations _variations;          ///< the %Sigma's variations
      bool _isDefined;                 ///< flag to indicate if the %Sigma is defined

   public:
      /// \name Constructors etc.
      ///@{

      /// Creates an empty and undefined Sigma.
      Sigma();

      /// Creates a defined %Sigma with the given central value and no variations.
      Sigma(double central, size_t numVariations = 0);

      /// Creates a defined %Sigma with the given central value and variations.
      Sigma(double central, Variations variations);

      ///@}

      /// \name Queries and Accessors
      ///@{
      double central() const;                    ///< Returns the central value.
      const Variations& variations() const;      ///< Returns the variations.

      /// Returns \c true if the %Sigma is defined, \c false otherwise.
      bool isDefined() const;

      double& central();                         ///< Access the central value.
      Variations& variations();                  ///< Access the std::vector of variations.

      /// Sets the %Sigma to be defined.
      void setDefined();

      /// Sets the %Sigma to be undefined.
      void setUndefined();

      // Clears the Sigma's central value and variations.
      void clear();
      ///@}

      /// \name Algebra
      ///@{
      // Sigma addition assignment operator
      Sigma& operator+=(const Sigma& rhs);

      // Sigma subtraction assignment operator
      Sigma& operator-=(const Sigma& rhs);

      // Sigma multiplication assignment operator
      Sigma& operator*=(const Sigma& rhs);

      // Sigma multiplication with real number assignment operator
      Sigma& operator*=(double rhs);

      // Sigma division by real number assignment operator
      Sigma& operator/=(double rhs);

      // Obtains the minimum and maximum of all variations.
      std::pair<double, double> getMinMaxVariations() const;
      ///@}

      /// \name Input
      ///@{
      // Sigma stream extraction operator
      friend std::istream& operator>>(std::istream& in, Sigma& sigma);
      ///@}

      /// \name Output
      ///@{
      // Sigma stream insertion operator
      friend std::ostream& operator<<(std::ostream& out, const Sigma& sigma);
      ///@}

   protected:
      /// Streams a string in the format "central (variations)" to \a out.
      std::ostream& print(std::ostream& out) const;
};

// Sigma addition operator
const Sigma operator+(const Sigma& lhs, const Sigma& rhs);

// Sigma subtraction operator
const Sigma operator-(const Sigma& lhs, const Sigma& rhs);

// Sigma multiplication operator
const Sigma operator*(const Sigma& lhs, const Sigma& rhs);

// Sigma negation operator
const Sigma operator-(const Sigma& lhs);

// Sigma multiplication operator with real number from the left
const Sigma operator*(double lhs, const Sigma& rhs);

// Sigma multiplication operator with real number from the right
const Sigma operator*(const Sigma& lhs, double rhs);

// Sigma division operator with real number
const Sigma operator/(const Sigma& lhs, double rhs);


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------
// Creates an empty and undefined Sigma.
inline Sigma::Sigma()
   : _central(0.0), _isDefined(false)
{}

//------------------------------------------------------------------------------
// Creates a defined Sigma with the given central value and no variations.
inline Sigma::Sigma(double central, size_t numVariations)
   : _central(central), _variations(numVariations), _isDefined(true)
{}

//------------------------------------------------------------------------------
// Creates a defined Sigma with the given central value and variations.
inline Sigma::Sigma(double central, Variations variations)
   : _central(central), _variations(variations), _isDefined(true)
{}

//------------------------------------------------------------------------------
// Queries and Accessors
//------------------------------------------------------------------------------
inline double Sigma::central() const                      { return _central;    }
inline const Sigma::Variations& Sigma::variations() const { return _variations; }
inline bool Sigma::isDefined() const                      { return _isDefined;  }

inline double& Sigma::central()               { return _central;    }
inline Sigma::Variations& Sigma::variations() { return _variations; }

//------------------------------------------------------------------------------
// Sets the Sigma to be defined.
inline void Sigma::setDefined()
{
   _isDefined = true;
}

//------------------------------------------------------------------------------
// Sets the Sigma to be undefined.
inline void Sigma::setUndefined()
{
   _isDefined = false;
}

//------------------------------------------------------------------------------
/**
 * \brief Clears the %Sigma's central value and variations.
 *
 * Sets the central value to 0, clears all variations, and sets the %Sigma to
 * be undefined.
 */
inline void Sigma::clear()
{
   _central = 0.;
   _variations.clear();
   setUndefined();
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------
/**
 * \brief %Sigma addition assignment operator
 * \param rhs The %Sigma to be added.
 * \return A reference to the %Sigma.
 *
 * The addition is performed for the central values and each variation.
 *
 * \note This is only well-defined if the %Sigma and \a rhs are both defined and
 * have the same number of variations.
 */
inline Sigma& Sigma::operator+=(const Sigma& rhs)
{
   assert(isDefined() && rhs.isDefined() && _variations.size() == rhs._variations.size());
   _central += rhs._central;
   const_iterator i2 = rhs._variations.begin();
   for (iterator i1 = _variations.begin(); i1 != _variations.end(); ++i1, ++i2) {
      *i1 += *i2;
   }
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %Sigma subtraction assignment operator
 * \param rhs The %Sigma to be subtracted.
 * \return A reference to the %Sigma.
 *
 * The subtraction is performed for the central values and each variation.
 *
 * \note This is only well-defined if the %Sigma and \a rhs are both defined and
 * have the same number of variations.
 */
inline Sigma& Sigma::operator-=(const Sigma& rhs)
{
   assert(isDefined() && rhs.isDefined() && _variations.size() == rhs._variations.size());
   _central -= rhs._central;
   const_iterator i2 = rhs._variations.begin();
   for (iterator i1 = _variations.begin(); i1 != _variations.end(); ++i1, ++i2) {
      *i1 -= *i2;
   }
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %Sigma multiplication assignment operator
 * \param rhs The %Sigma to be multiplied.
 * \return A reference to the %Sigma.
 *
 * The multiplication is performed for the central values and each variation.
 *
 * \note This is only well-defined if the %Sigma and \a rhs are both defined and
 * have the same number of variations.
 */
inline Sigma& Sigma::operator*=(const Sigma& rhs)
{
   assert(isDefined() && rhs.isDefined() && _variations.size() == rhs._variations.size());
   _central *= rhs._central;
   const_iterator i2 = rhs._variations.begin();
   for (iterator i1 = _variations.begin(); i1 != _variations.end(); ++i1, ++i2) {
      *i1 *= *i2;
   }
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %Sigma multiplication with real number assignment operator
 * \param rhs The number to be multiplied by.
 * \return A reference to the %Sigma.
 *
 * The %Sigma's central value and each variation are multiplied by \a rhs.
 */
inline Sigma& Sigma::operator*=(double rhs)
{
   _central *= rhs;
   for (iterator it = _variations.begin(); it != _variations.end(); ++it) {
      *it *= rhs;
   }
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %Sigma division by real number assignment operator
 * \param rhs The number to be divided by.
 * \return A reference to the %Sigma.
 *
 * The %Sigma's central value and each variation are divided by \a rhs.
 *
 * \note Division by zero is not checked.
 */
inline Sigma& Sigma::operator/=(double rhs)
{
   return operator*=(1. / rhs);
}

//------------------------------------------------------------------------------
/**
 * \brief %Sigma addition operator
 * \param lhs The 1st %Sigma to add.
 * \param rhs The 2nd %Sigma to add.
 * \return A %Sigma given by the sum of \a lhs and \a rhs.
 *
 * The addition is performed for the central values and each variation.
 *
 * \note This is only well-defined if \a lhs and \a rhs are both defined and
 * have the same number of variations.
 */
inline const Sigma operator+(const Sigma& lhs, const Sigma& rhs)
{
   Sigma temp(lhs);
   temp += rhs;
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Sigma subtraction operator
 * \param lhs The 1st %Sigma
 * \param rhs The 2nd %Sigma to be subtracted from \a lhs.
 * \return A %Sigma given by the difference of \a lhs and \a rhs.
 *
 * The subtraction is performed for the central values and each variation.
 *
 * \note This is only well-defined if \a lhs and \a rhs are both defined and
 * have the same number of variations.
 */
inline const Sigma operator-(const Sigma& lhs, const Sigma& rhs)
{
   Sigma temp(rhs);
   temp -= lhs;
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Sigma multiplication operator
 * \param lhs The 1st %Sigma to be multiplied
 * \param rhs The 2nd %Sigma to be multiplied
 * \return A %Sigma given by the product of \a lhs and \a rhs.
 *
 * The multiplication is performed for the central values and each variation.
 *
 * \note This is only well-defined if \a lhs and \a rhs are both defined and
 * have the same number of variations.
 */
inline const Sigma operator*(const Sigma& lhs, const Sigma& rhs)
{
   Sigma temp(rhs);
   temp *= lhs;
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Sigma negation operator
 * \param lhs The %Sigma to be negated.
 * \return A %Sigma given by the negative of \a lhs.
 *
 * The central value and each variation of \a lhs is negated.
 */
inline const Sigma operator-(const Sigma& lhs)
{
   Sigma temp(lhs);
   temp *= -1.;
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Sigma multiplication operator with real number from the left
 * \param lhs The number to multiply with.
 * \param rhs The %Sigma to be multiplied by \a lhs.
 * \return A %Sigma given by the product of \a lhs times \a rhs.
 */
inline const Sigma operator*(double lhs, const Sigma& rhs)
{
   Sigma temp(rhs);
   temp *= lhs;
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Sigma multiplication operator with real number from the right
 * \param lhs The %Sigma to be multiplied by \a rhs.
 * \param rhs The number to multiply with.
 * \return A %Sigma given by the product of \a lhs times \a rhs.
 */
inline const Sigma operator*(const Sigma& lhs, double rhs)
{
   return rhs * lhs;
}

//------------------------------------------------------------------------------
/**
 * \brief %Sigma division operator with real
 * \param lhs The %Sigma to be divided by \a rhs.
 * \param rhs The number to divide by.
 * \return A %Sigma given by \a lhs divided by \a rhs.
 * \note Division by zero is not checked.
 */
inline const Sigma operator/(const Sigma& lhs, double rhs)
{
   return (1./rhs) * lhs;
}

//------------------------------------------------------------------------------
// Input
//------------------------------------------------------------------------------
/**
 * \brief %Sigma stream extraction operator
 * \param in The input stream to read from.
 * \param sigma The %Sigma to be filled.
 * \return A reference to \a in.
 *
 * Read a string in the format "central (variations)" from \a in and fill \a sigma.
 * The input \a sigma is cleared before being filled.
 */
inline std::istream& operator>>(std::istream& in, Sigma& sigma)
{
   sigma.clear();
   in >> sigma.central() >> std::ws;
   if ((char) in.peek() == '(') {
      in.get();
      while (true) {
         double vary;
         in >> vary;
         sigma.variations().push_back(vary);
         // skip delimiter and exit when needed
         if ((char) in.get() == ')')
            break;
      }
   }
   sigma.setDefined();
   return in;
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------
/**
 * \brief %Sigma stream insertion operator
 * \param out The output stream to be used.
 * \param sigma The %Sigma to be streamed.
 * \return A reference to \a out.
 *
 * Streams a string in the format "central (variations)" to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const Sigma& sigma)
{
   return sigma.print(out);
}

} // namespace Geneva

#endif // GENEVA_SIGMA_HPP
