//------------------------------------------------------------------------------
/// \file Signature.hpp
//
// Author(s):
//    Frank Tackmann, Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 MIT, Frank Tackmann, LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Signature
//------------------------------------------------------------------------------

#ifndef GENEVA_SIGNATURE_HPP
#define GENEVA_SIGNATURE_HPP

#include "Geneva/Core/Physics/SignatureStatus.hpp"

#include <set>
#include <iosfwd>

namespace Geneva
{

class Configuration;

//------------------------------------------------------------------------------
/**
 * \class Signature
 * \ingroup Core
 *
 * \brief A set of PdgIDs representing the signature of a configuration.
 *
 * The %Signature distinguishes between initial-state and final-state IDs, with
 * at most two initial-state IDs that are distinguished as A and B, and are
 * accessible via initialA() and initialB() iterators. Identical IDs in the
 * final state are allowed and are not distinguished. Internally the %Signature
 * is implemented in terms of a STL multiset, allowing for fast access and
 * insertion/removal of IDs. It supports the standard interface for STL sets and
 * in addition a variety of specific set operations (union, etc.) that are
 * accessed by algebraic or logical operators with the obvious meanings:
 *
 * - s == t, s != t: Test for (in)equality
 * - s < t, s <= t, s > t, s >= t: Lexicographical comparison that can be used to
 * order signatures
 * - s -= t, s - t: The difference (complement) of s with respect to t
 * - s &= t, s & t: The intersection of s and t
 *
 * For example:
 *
 * - (ee~ -> qq~g) - (ee~ -> qq~)   = ( -> g)
 * - (ee~ -> qq~g) & (ee~ -> qq~)   = (ee~ -> qq~g)
 *
 * Since initial-state IDs are treated specially, a set union operation is not
 * supported, since its result is not well defined for the initial state IDs.
 *
 * Note that the %Signature only acts as a container and is agnostic to the IDs
 * it contains, so one can store unphysical signatures like (ee -> qg).
 *
 * A %Signature can be traversed using iterators that have the type
 * Signature::iterator and Signature::reverse_iterator. Various member functions
 * use or return iterators. All iterators are const iterators and do not allow
 * to change the ID they point to. To change an ID, it has to be erased and a
 * new one inserted. A special case are the initial-state IDs, which can be
 * set and overwritten using the setInitialA() and setInitialB() members.
 *
 * The ID pointed to by an iterator 'i' is accessed using either *i or i->.
 * Iterators can be incremented, ++i, or decremented, --i, to point to the
 * next or previous ID in the %Signature, and can be checked for (in)equality
 * (using == or !=). The relation operators (<, <=, >, >=) cannot be used with
 * %Signature iterators. Iterators support i->isInitialA(), i->isInitialB(), and
 * i->isFinal() queries, which allow to test as what kind of ID the pointed to
 * ID is stored.
 *
 * The members begin() and end() return iterators pointing to the first ID and
 * one step beyond the last ID, respectively. Hence a %Signature can be
 * traversed using
 * \code
 *    for (Signature::iterator i = sig.begin(); i != sig.end(), ++i) {
 *       doSomethingwithID(*i);
 *       ...
 *    }
 * \endcode
 *
 * A reverse_iterator behaves exactly opposite which is useful if one wants to
 * traverse the %Signature in descending order, i.e., rbegin() and rend() point
 * to the last ID and one step before the first ID, respectively, and after
 * incrementing a reverse_iterator it points to the previous ID. To just iterate
 * over final-state IDs, use beginFinal() and endFinal().
 *
 * An existing iterator pointing to an ID in a %Signature stays valid if new IDs
 * are inserted or other IDs are erased. Of course it does become invalid if the
 * ID it points to itself is erased.
 */
//------------------------------------------------------------------------------
class Signature
{
   private:
      typedef SignatureStatusID _status_id;
      typedef _status_id::_compare_type _compare_type;
      typedef std::multiset<_status_id, _compare_type> _impl_type;

      _impl_type _ids;                     ///< Set of IDs forming the %Signature

      _impl_type::iterator _initialA;      ///< Iterator to initial-state ID A
      _impl_type::iterator _initialB;      ///< Iterator to initial-state ID B
      _impl_type::iterator _beginFinal;    ///< Iterator to first final-state ID

   public:
      /// \name Type Definitions
      ///@{
      typedef _impl_type::const_reference const_reference;

      /// const_iterator for PdgID entries in the %Signature.
      typedef _impl_type::const_iterator const_iterator;
      /// iterator is type equivalent to const_iterator
      typedef const_iterator iterator;
      ///@}

      /// \name Constructors etc.
      ///@{

      /// Creates an empty %Signature.
      Signature();

      /// Creates a %Signature with \a id as the only element in the final state.
      Signature(PdgID id);

      /// Creates a %Signature with \a n copies of \a id in the final state.
      Signature(PdgID id, size_t n);

      // Creates a Signature from a range of IDs.
      template <typename SignatureLikeIterator>
      Signature(SignatureLikeIterator first, SignatureLikeIterator last);

      // Signature copy constructor.
      Signature(const Signature& other);

      // Creates a Signature from a Configuration.
      Signature(const Configuration& config);

      // Signature assignment operator.
      Signature& operator=(const Signature& rhs);

      // Signature assignment operator from a Configuration.
      Signature& operator=(const Configuration& rhs);

      ~Signature() = default;
      ///@}

      /// \name Accessing and Querying IDs
      ///@{
      // Returns an iterator to the first ID in the Signature.
      iterator begin() const;

      // Returns an iterator pointing beyond the last ID in the Signature.
      iterator end() const;

      // Returns an iterator to initial-state ID A.
      iterator initialA() const;

      // Returns an iterator to initial-state ID B.
      iterator initialB() const;

      // Returns an iterator to the first final-state ID.
      iterator beginFinal() const;

      // Returns an iterator pointing beyond the last final-state ID.
      iterator endFinal() const;

      /// Returns \c true if the %Signature is empty, \c false otherwise.
      bool empty() const;

      /// Returns the total number of IDs in the %Signature.
      size_t size() const;

      /// Returns the total number of IDs in the %Signature's final state.
      size_t sizeFinal() const;

      /// Returns the maximum number of IDs the %Signature could hold.
      size_t maxSize() const;

      // Counts how often an ID appears in the Signature.
      size_t count(PdgID id) const;

      // Counts how often an ID appears in the Signature's final state.
      size_t countFinal(PdgID id) const;

      // Returns the symmetry factor of the Signature.
      size_t symmetryFactor() const;

      // Tries to find a matching ID in the Signature.
      template <typename SignatureLikeIterator>
      iterator find(SignatureLikeIterator it) const;

      // Tries to find an ID in the Signature.
      iterator find(PdgID id) const;

      // Tries to find an ID in the Signature's final state.
      iterator findFinal(PdgID id) const;

      // Finds the range in the Signature's final state equal to an ID.
      std::pair<iterator, iterator> equalRangeFinal(PdgID id) const;

      // Tests if a matching ID is contained in the Signature.
      template <typename SignatureLikeIterator>
      bool includes(SignatureLikeIterator it) const;

      // Tests if an ID is contained in the Signature.
      bool includes(PdgID id) const;

      // Tests if an ID is contained in the Signature's final state.
      bool includesFinal(PdgID id) const;

      // Tests if a Signature is contained in the Signature.
      bool includes(const Signature& other) const;

      // Tests if a Signature intersects with the Signature.
      bool intersects(const Signature& other) const;
      ///@}

      /// \name Adding and Removing IDs
      ///@{
      // Swaps the contents with another Signature.
      void swap(Signature& other);

      // Inserts an initial- or final-state ID into the Signature.
      template <typename SignatureLikeIterator>
      iterator insert(SignatureLikeIterator it);

      // Inserts a range of IDs into the Signature.
      template <typename SignatureLikeIterator>
      void insert(SignatureLikeIterator first, SignatureLikeIterator last);

      // Sets the initial-state ID A of the Signature.
      iterator setInitialA(PdgID id);

      // Sets the initial-state ID B of the Signature.
      iterator setInitialB(PdgID id);

      // Inserts an ID into the Signature's final state.
      iterator insertFinal(PdgID id);

      // Inserts an ID into the Signature's final state using a hint.
      iterator insertFinal(iterator hint, PdgID id);

      // Inserts multiple copies of an ID into the Signature's final state.
      void insertFinal(PdgID id, size_t n);

      // Inserts a range of IDs into the Signature's final state.
      template <typename InputIterator>
      void insertFinal(InputIterator first, InputIterator last);

      // Removes the ID from the Signature pointed to by it.
      void erase(iterator it);

      // Removes all instances of an ID from the Signature.
      size_t erase(PdgID id);

      // Removes a range of IDs from the Signature.
      void erase(iterator first, iterator last);

      // Removes the Signature's initial-state ID A.
      bool eraseInitialA();

      // Removes the Signature's initial-state ID B.
      bool eraseInitialB();

      // Removes all instances of an ID from the Signature's final state.
      size_t eraseFinal(PdgID id);

      /// Removes all IDs from the %Signature.
      void clear();
      ///@}

      /// \name Algebra
      ///@{
      // Signature equality comparison operator
      bool operator==(const Signature& rhs) const;

      // Signature nonequality comparison operator
      bool operator!=(const Signature& rhs) const;

      // Signature ordering relation operator
      bool operator<(const Signature& rhs) const;

      /// Returns \c true if %Signature is less or equal to \a rhs.
      bool operator<=(const Signature& rhs) const;

      /// Returns \c true if %Signature is greater or equal to \a rhs.
      bool operator>=(const Signature& rhs) const;

      /// Returns \c true if %Signature is greater than \a rhs.
      bool operator>(const Signature& rhs) const;

      // Signature subtraction (complement) assignment operator
      Signature& operator-=(const Signature& rhs);

      // Signature AND (intersection) assignment operator
      Signature& operator&=(const Signature& rhs);

      // Signature subtraction (complement) operator
      friend const Signature operator-(const Signature& lhs, const Signature& rhs);

      // Signature AND (intersection) operator
      friend const Signature operator&(const Signature& lhs, const Signature& rhs);
      ///@}

      /// \name Output
      ///@{
      // Signature stream insertion operator.
      friend std::ostream& operator<<(std::ostream& out, const Signature& sig);
      ///@}

   private:
      /// Streams the %Signature to \a out.
      std::ostream& print(std::ostream& out) const;

      /// Initializes the internal initial-state and final-state iterators.
      void init();

      /// Copies the content of another %Signature-like class.
      template <class SignatureLike>
      void copy(const SignatureLike& other);
};

//------------------------------------------------------------------------------
/**
 * \struct CheckSignatureSize
 * \ingroup Core
 *
 * \brief An unary function object checking the size of a Signature.
 */
class CheckSignatureSize : public std::unary_function<Signature, bool>
{
   private:
      size_t _size;

   public:
      CheckSignatureSize(size_t size) : _size(size) {}

      bool operator()(const Signature& sig) const
      {
         return sig.size() == _size;
      }
};

} // namespace Geneva


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

#include "Geneva/Core/Physics/Configuration.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------
// Creates an empty Signature.
inline Signature::Signature()
   : _ids()
{
   _initialA = end();
   _initialB = end();
   _beginFinal = end();
}

//------------------------------------------------------------------------------
// Creates a Signature with id as the only element in the final state.
inline Signature::Signature(PdgID id)
   : _ids()
{
   _initialA = end();
   _initialB = end();
   _beginFinal = _ids.insert(id);
}

//------------------------------------------------------------------------------
// Creates a Signature with n copies of id in the final state.
inline Signature::Signature(PdgID id, size_t n)
   : _ids()
{
   _initialA = end();
   _initialB = end();
   while (n) {
      _ids.insert(end(), id);
      --n;
   }
   _beginFinal = begin();
}

//------------------------------------------------------------------------------
/**
 * \brief Creates a %Signature from a range of IDs.
 * \param first An iterator pointing to the first ID in the range.
 * \param last  An iterator pointing beyond the last ID in the range.
 *
 * Creates a new %Signature by copying the IDs from the Signature-like range
 * specified by [\a first,\a last), keeping their initial-state and final-state
 * assignment.
 *
 * \note This is a template function that works for any type of Signature-like
 * iterator (i.e. Signature::iterator or Configuration::iterator). To fill a
 * %Signature from a generic range of IDs (e.g. a std::vector) the insertFinal()
 * member can be used.
 */
template <typename SignatureLikeIterator>
inline Signature::Signature(SignatureLikeIterator first, SignatureLikeIterator last)
   : _ids()
{
   // here we rely on the Signature-like ordering in [first, last)
   if (first != last && first->isInitialA()) {
      _initialA = _ids.insert(_status_id(first->id(), _status_id::kInitialA));
      ++first;
   } else
      _initialA = end();
   if (first != last && first->isInitialB()) {
      _initialB = _ids.insert(_status_id(first->id(), _status_id::kInitialB));
      ++first;
   } else
      _initialB = end();
   if (first != last) {
      _beginFinal = _ids.insert(end(), first->id());
      ++first;
      for (; first != last; ++first)
         _ids.insert(end(), first->id());
   } else
      _beginFinal = end();
}

//------------------------------------------------------------------------------
/**
 * \brief %Signature copy constructor.
 * \param other The %Signature that will be copied.
 *
 * Creates a new %Signature as a copy of \a other.
 */
inline Signature::Signature(const Signature& other)
   : _ids()
{
   copy(other);
}

//------------------------------------------------------------------------------
/**
 * \brief Creates a Signature from a Configuration.
 * \param other The Configuration whose particle IDs will be copied.
 *
 * Creates a new %Signature by copying the particle IDs from \a other including
 * their initial-state and final-state assignments.
 */
inline Signature::Signature(const Configuration& other)
   : _ids()
{
   copy(other);
}

//------------------------------------------------------------------------------
/**
 * \brief %Signature assignment operator.
 * \param rhs The %Signature that will be copied.
 * \return A reference to the %Signature.
 *
 * Replaces the current content of the %Signature with a copy of \a rhs.
 */
inline Signature& Signature::operator=(const Signature& rhs)
{
   if (this != &rhs) {
      _ids.clear();
      copy(rhs);
   }
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief Signature assignment operator from a Configuration.
 * \param rhs The %Configuration whose particle IDs will be copied.
 * \return A reference to the %Signature.
 *
 * Replaces the current content of the %Signature with a copy of the particle
 * IDs of \a rhs including their initial-state and final-state assignments.
 */
inline Signature& Signature::operator=(const Configuration& rhs)
{
   _ids.clear();
   copy(rhs);
   return *this;
}

//------------------------------------------------------------------------------
// Accessing and Querying IDs
//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator to the first ID in the %Signature.
 *
 * If the %Signature is empty the returned iterator will be equal to end(). It
 * is the user's responsibility to ensure that the iterator is not used for
 * reading in this case.
 */
inline Signature::iterator Signature::begin() const
{
   return _ids.begin();
}

//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator pointing beyond the last ID in the %Signature.
 *
 * The returned iterator points one step beyond the last ID and must not be used
 * for reading. Instead, it is used to check the validity of other iterators.
 */
inline Signature::iterator Signature::end() const
{
   return _ids.end();
}

//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator to initial-state ID A.
 *
 * The returned iterator points to the ID in the %Signature that is
 * considered to be the ID of initial-state A. If no initial-state A is defined
 * end() is returned.
 */
inline Signature::iterator Signature::initialA() const
{
   return _initialA;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator to initial-state ID B.
 *
 * The returned iterator points to the ID in the %Signature that is
 * considered to be the ID of initial-state B. If no initial-state B is defined
 * end() is returned.
 */
inline Signature::iterator Signature::initialB() const
{
   return _initialB;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator to the first final-state ID.
 *
 * Iterating from beginFinal() to endFinal() will iterator over all final-state
 * IDs in the %Signature and no others. If the %Signature's final state is empty
 * the returned iterator will be equal to endFinal(). It is the user's
 * responsibility to ensure that the iterator is not used for reading in this
 * case.
 */
inline Signature::iterator Signature::beginFinal() const
{
   return _beginFinal;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns an state iterator pointing beyond the last final-state ID.
 *
 * The returned iterator points one step beyond the last final-state ID
 * and must not be used for reading. Instead, it is used to check the validity
 * of other iterators.
 */
inline Signature::iterator Signature::endFinal() const
{
   return end();
}

//------------------------------------------------------------------------------
// Returns true if the Signature is empty, false otherwise.
inline bool Signature::empty() const
{
   return _ids.empty();
}

//------------------------------------------------------------------------------
// Returns the total number of IDs in the Signature.
inline size_t Signature::size() const
{
   return _ids.size();
}

//------------------------------------------------------------------------------
// Returns the total number of final-state IDs in the Signature.
inline size_t Signature::sizeFinal() const
{
   size_t s = _ids.size();
   if (_initialA != end())
      --s;
   if (_initialB != end())
      --s;
   return s;
}

//------------------------------------------------------------------------------
// Returns the maximum number of IDs the Signature could hold.
inline size_t Signature::maxSize() const
{
   return _ids.max_size();
}

//------------------------------------------------------------------------------
/**
 * \brief Counts how often an ID appears in the %Signature.
 * \param id The ID to be looked for.
 * \return The number of times \a id appears in the %Signature.
 *
 * Ocurrences of \a id in initial and final states are counted. If \a id is not
 * contained in the %Signature at all then 0 is returned.
 */
inline size_t Signature::count(PdgID id) const
{
   size_t temp = countFinal(id);
   if (_initialA != end() && *_initialA == id)
      ++temp;
   if (_initialB != end() && *_initialB == id)
      ++temp;
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief Counts how often an ID appears as in the %Signature's final state.
 * \param id The ID to be looked for.
 * \return The number of times \a id appears in the %Signature's final state.
 *
 * Ocurrences of \a id in the %Signature's initial state are not counted. If
 * \a id is not contained in the %Signature's final state then 0 is returned.
 */
inline size_t Signature::countFinal(PdgID id) const
{
   return _ids.count(id);
}

//------------------------------------------------------------------------------
/**
 * \brief Tries to find a matching ID in the %Signature.
 * \param it An iterator pointing to the ID to be searched for.
 * \return An iterator pointing to the first instance of the sought ID.
 *
 * If the ID \a it points to is found in the %Signature including its initial-
 * or final-state assignment the returned iterator points to its first instance,
 * otherwise end() is returned. Requires logarithmic time.
 *
 * \note This is a template function that works for any type of Signature-like
 * iterator (i.e. Signature::iterator or Configuration::iterator).
 */
template <typename SignatureLikeIterator>
inline Signature::iterator Signature::find(SignatureLikeIterator it) const
{
   if (it->isInitialA()) {
      if (_initialA != end() && _initialA->id() == it->id())
         return _initialA;
      else
         return end();
   }
   if (it->isInitialB()) {
      if (_initialB != end() && _initialB->id() == it->id())
         return _initialB;
      else
         return end();
   }
   return findFinal(it->id());
}

//------------------------------------------------------------------------------
/**
 * \brief Tries to find an ID in the %Signature.
 * \param id The ID to be searched for.
 * \return An iterator pointing to the first instance of the sought ID.
 *
 * If \a id is found in the %Signature the iterator points to its first
 * instance, otherwise end() is returned. Requires logarithmic time.
 * The search is performed over both initial and final states.
 */
inline Signature::iterator Signature::find(PdgID id) const
{
   if (_initialA != end() && *_initialA == id)
      return _initialA;
   if (_initialB != end() && *_initialB == id)
      return _initialB;
   return findFinal(id);
}

//------------------------------------------------------------------------------
/**
 * \brief Tries to find an ID in the %Signature's final state.
 * \param id The ID to be searched for.
 * \return An iterator pointing to the first instance of the sought ID.
 *
 * If \a id is found in the %Signature's final state the returned iterator
 * points to the first instance, otherwise endFinal() is returned. Requires
 * logarithmic time.
 */
inline Signature::iterator Signature::findFinal(PdgID id) const
{
   return _ids.find(id);
}

//------------------------------------------------------------------------------
/**
 * \brief Finds the range in the %Signature's final state equal to an ID.
 * \param id The ID to be searched for.
 * \return A pair of iterators specifying the found range.
 *
 * If \a id is part of the %Signature, then the [first, last) range specified by
 * the returned pair of iterators gives the complete range of IDs in the final
 * state equal to \a id. If \a id is not found in %Signature then the two
 * iterators will be equal, specifying an empty range.
 */
inline std::pair<Signature::iterator, Signature::iterator>
Signature::equalRangeFinal(PdgID id) const
{
   return _ids.equal_range(id);
}

//------------------------------------------------------------------------------
/**
 * \brief Tests if a matching ID is contained in the %Signature.
 * \param it An iterator pointing to the ID to be searched for.
 * \return \c true if an ID as pointed to by \a it is contained in the %Signature,
 *         \c false otherwise.
 *
 * This is identical to find(it) != end(), so if an iterator to the ID is needed
 * in case it is contained, it is faster to use find(it) directly. See the
 * documentation of find(it) for further details.
 */
template <typename SignatureLikeIterator>
inline bool Signature::includes(SignatureLikeIterator it) const
{
   return find(it) != end();
}

//------------------------------------------------------------------------------
/**
 * \brief Tests if an ID is contained in the %Signature.
 * \param id The ID to be searched for.
 * \return \c true if \a id is contained in the %Signature, \c false otherwise.
 *
 * This is identical to find(id) != end(), so if an iterator to the ID is needed
 * in case it is contained, it is faster to use find(id) directly. The search is
 * performed over both initial and final states.
 */
inline bool Signature::includes(PdgID id) const
{
   return find(id) != end();
}

//------------------------------------------------------------------------------
/**
 * \brief Tests if an ID is contained in the %Signature's final state.
 * \param id The ID to be searched for.
 * \return \c true if \a id is contained in the %Signature, \c false otherwise.
 *
 * This is identical to findFinal(id) != endFinal(), so if an iterator to the ID
 * is needed in case it is contained, it is faster to use findFinal(id) directly.
 */
inline bool Signature::includesFinal(PdgID id) const
{
   return findFinal(id) != endFinal();
}

//------------------------------------------------------------------------------
// Adding and Removing IDs
//------------------------------------------------------------------------------
/**
 * \brief Swaps the contents with another %Signature.
 * \param other The %Signature being swapped with.
 *
 * This exchanges the contents of the %Signature with that of \a other. The
 * swapping does not require the copying of elements, which means it is fast and
 * executes in constant time.
 */
inline void Signature::swap(Signature& other)
{
   _ids.swap(other._ids);
   init();
   other.init();
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts an initial- or final-state ID into the %Signature.
 * \param it An iterator pointing to the ID to be inserted.
 * \return An iterator pointing to the inserted ID.
 *
 * If \a it points to an initial state the corresponding initial state (A or B)
 * is set, otherwise \a it is inserted as a new final-state ID, in which case
 * the insertion requires logarithmic time.
 *
 * \note This is a template function that works for any type of Signature-like
 * iterator (i.e. Signature::iterator or Configuration::iterator).
 */
template <typename SignatureLikeIterator>
inline Signature::iterator Signature::insert(SignatureLikeIterator it)
{
   if (it->isInitialA())
      return setInitialA(it->id());
   if (it->isInitialB())
      return setInitialB(it->id());
   return insertFinal(it->id());
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts a range of IDs into the %Signature.
 * \param first An iterator pointing to the first ID in the range.
 * \param last  An iterator pointing beyond the last ID in the range.
 *
 * Inserts the IDs from the range specified by [\a first,\a last). The
 * complexity is as for the analogous range constructor. The initial- and final-
 * state assignment of the IDs are kept. Possible initial states in the range
 * will overwrite the %Signature's present initial states, while any final
 * states are added.
 *
 * \note This is a template function that works for any type of Signature-like
 * iterator (i.e. Signature::iterator or Configuration::iterator).
 */
template <typename SignatureLikeIterator>
inline void Signature::insert(SignatureLikeIterator first, SignatureLikeIterator last)
{
   // here we can keep initial-state information and rely on the Signature-like
   // ordering in [first, last)
   if (first != last && first->isInitialA()) {
      setInitialA(first->id());
      ++first;
   }
   if (first != last && first->isInitialB()) {
      setInitialB(first->id());
      ++first;
   }
   if (first != last) {
      insertFinal(first->id());
      ++first;
      for (; first != last; ++first)
         _ids.insert(end(), first->id());
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Sets the initial-state ID A of the %Signature.
 * \param id The ID to be set as initial-state A.
 * \return An iterator pointing to the inserted ID.
 *
 * If the %Signature already has an initial-state A set, it is overwritten with
 * the given ID. If this is not desired, the user should check for the existence
 * of an initial state A first.
 */
inline Signature::iterator Signature::setInitialA(PdgID id)
{
   if (_initialA == end()) {
      _initialA = _ids.insert(_status_id(id, _status_id::kInitialA));
   } else {
      // Forcing to overwrite the id is faster than erasing and inserting. It is
      // legitimate here, since we enforce that there is only one initial-state
      // ID of a kind and whose ordering in the set is determined by its state only.
      const_cast<_status_id&>(*_initialA).PdgID::operator=(id);
   }
   return _initialA;
}

//------------------------------------------------------------------------------
/**
 * \brief Sets the initial-state ID B of the %Signature.
 * \param id The ID to be inserted as initial-satte B.
 * \return An iterator pointing to the inserted ID.
 *
 * If the %Signature already has an initial-state B set, it is overwritten with
 * the given ID. If this is not desired, the user should check for the existence
 * of an initial state B first.
 */
inline Signature::iterator Signature::setInitialB(PdgID id)
{
   if (_initialB == end()) {
      _initialB = _ids.insert(_status_id(id, _status_id::kInitialB));
   } else {
      // Forcing to overwrite the id is faster than erasing and inserting. It is
      // legitimate here, since we enforce that there is only one initial-state
      // ID of a kind and whose ordering in the set is determined by its state only.
      const_cast<_status_id&>(*_initialB).PdgID::operator=(id);
   }
   return _initialB;
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts an ID into the Signature's final state.
 * \param id The ID to be inserted.
 * \return An iterator pointing to the inserted ID.
 *
 * Insertion requires logarithmic time.
 */
inline Signature::iterator Signature::insertFinal(PdgID id)
{
   iterator temp = _ids.insert(id);
   // update _beginFinal if necessary, relies on the fact that for identical
   // elements new ones are inserted behind existing ones
   if (_beginFinal == end() || _compare_type()(*temp, *_beginFinal))
      _beginFinal = temp;
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts an ID into the %Signature's final state using a hint.
 * \param hint An iterator used as a hint where the ID should be inserted.
 * \param id The ID to be inserted.
 * \return An iterator pointing to the inserted ID.
 *
 * If the \a hint points to the correct location this will speed up the
 * insertion, otherwise there will be no speed gain over insert(id).
 */
inline Signature::iterator Signature::insertFinal(Signature::iterator hint, PdgID id)
{
   iterator temp = _ids.insert(hint, id);
   // update _beginFinal if necessary, relies on the fact that for identical
   // elements new ones are inserted behind existing ones unless the hint
   // pointed to an identical element
   if (_beginFinal == end() || _compare_type()(*temp, *_beginFinal)
      || (hint == _beginFinal && *hint == id))
      _beginFinal = temp;
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts a range of IDs into the %Signature's final state.
 * \param first An iterator pointing to the first ID in the range.
 * \param last  An iterator pointing beyond the last ID in the range.
 *
 * Inserts the IDs from the range specified by [\a first,\a last). The
 * complexity is as for multisets: Linear in N if the range is already sorted
 * (e.g. when obtained from another %Signature) and N log N otherwise (e.g. when
 * obtained from a std::vector of IDs), where N is the number of IDs in the
 * range.
 *
 * \note This is a template function that works for generic iterators for which
 * \c *it is a PdgID. All IDs are put into the final state.
 */
template <typename InputIterator>
inline void Signature::insertFinal(InputIterator first, InputIterator last)
{
   for (; first != last; ++first)
      _ids.insert(end(), _status_id(*first, _status_id::kFinal));
   if (_initialB != end()) {
      _beginFinal = _initialB;
      ++_beginFinal;
   } else if (_initialA != end()) {
      _beginFinal = _initialA;
      ++_beginFinal;
   } else
      _beginFinal = begin();
}

//------------------------------------------------------------------------------
/**
 * \brief Removes the ID from the %Signature pointed to by \a it.
 * \param it An iterator pointing to the ID to be erased.
 *
 * Only the specific initial- or final-state ID pointed to by \a it is removed
 * from the %Signature. Other identical IDs are kept.
 *
 * \note It is the user's responsibility to ensure that \a it is a valid
 * iterator into the %Signature.
 */
inline void Signature::erase(Signature::iterator it)
{
   if (it == _initialA)
      _initialA = end();
   else if (it == _initialB)
      _initialB = end();
   else if (it == _beginFinal)
      ++_beginFinal;
   _ids.erase(it);
}

//------------------------------------------------------------------------------
/**
 * \brief Removes all instances of an ID from the %Signature.
 * \param id The ID to be erased.
 * \return The total number of elements erased.
 *
 * This removes all instances of the ID from the Signature's initial and final
 * states. The return value is equal to what a prior call of count(id) would
 * have given.
 */
inline size_t Signature::erase(PdgID id)
{
   size_t temp = 0;
   if (_initialA != end() && *_initialA == id) {
      _ids.erase(_initialA);
      _initialA = end();
      ++temp;
   }
   if (_initialB != end() && *_initialB == id) {
      _ids.erase(_initialB);
      _initialB = end();
      ++temp;
   }
   temp += eraseFinal(id);
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief Removes a range of IDs from the %Signature.
 * \param first An iterator pointing to the first ID in the %Signature to be erased.
 * \param last  An iterator pointing beyond the last ID in the %Signature to be erased.
 *
 * The IDs in the %Signature specified by [\a first,\a last) are removed. The
 * complexity is linear in the number of IDs to be erased.
 *
 * \note It is the user's responsibility to ensure that [\a first, \a last) is
 * a valid iterator range into the %Signature.
 */
inline void Signature::erase(Signature::iterator first, Signature::iterator last)
{
   if (_initialA == first && first != last) {
      _ids.erase(_initialA);
      _initialA = end();
      ++first;
   }
   if (_initialB == first && first != last) {
      _ids.erase(_initialB);
      _initialB = end();
      ++first;
   }
   if (_beginFinal == first) {
      _beginFinal = last;
   }
   _ids.erase(first, last);
}

//------------------------------------------------------------------------------
/**
 * \brief Removes the %Signature's initial-state ID A.
 * \return \c true if initial state A was present, \c false otherwise.
 */
inline bool Signature::eraseInitialA()
{
   if (_initialA != end()) {
      _ids.erase(_initialA);
      _initialA = end();
      return true;
   }
   return false;
}

//------------------------------------------------------------------------------
/**
 * \brief Removes the %Signature's initial-state ID B.
 * \return \c true if initial state B was present, \c false otherwise.
 */
inline bool Signature::eraseInitialB()
{
   if (_initialB != end()) {
      _ids.erase(_initialB);
      _initialB = end();
      return true;
   }
   return false;
}

//------------------------------------------------------------------------------
/**
 * \brief Removes all instances of an ID from the %Signature's final state.
 * \param id The ID to be erased.
 * \return The total number of elements erased from the final state.
 *
 * Ocurrences of \a id in the Signature's initial state are not affected. The
 * return value is equal to what a prior call of countFinal(id) would have given.
 */
inline size_t Signature::eraseFinal(PdgID id)
{
   if (_beginFinal != end() && *_beginFinal == id) {
      _beginFinal = _ids.upper_bound(id);
   }
   return _ids.erase(id);
}

//------------------------------------------------------------------------------
// Removes all IDs from the Signature.
inline void Signature::clear()
{
   _ids.clear();
   _initialA = end();
   _initialB = end();
   _beginFinal = end();
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------
/**
 * \brief %Signature nonequality comparison operator
 * \param rhs The %Signature to be compared to
 * \return \c true if the %Signature does not contain exactly the same IDs as
 *         \a rhs, \c false otherwise.
 *
 * Identical to !operator==(rhs)
 */
inline bool Signature::operator!=(const Signature& rhs) const
{
   return !(*this == rhs);
}

//------------------------------------------------------------------------------
// Returns true if Signature is less or equal to rhs.
inline bool Signature::operator<=(const Signature& rhs) const
{
   return !(rhs < *this);
}

//------------------------------------------------------------------------------
// Returns true if Signature is greater or equal to rhs.
inline bool Signature::operator>=(const Signature& rhs) const
{
   return !(*this < rhs);
}

//------------------------------------------------------------------------------
// Returns true if Signature is greater than rhs.
inline bool Signature::operator>(const Signature& rhs) const
{
   return rhs < *this;
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------
/**
 * \brief %Signature stream insertion operator.
 * \param out The output stream to be used.
 * \param sig The %Signature to be streamed.
 * \return A reference to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const Signature& sig)
{
   return sig.print(out);
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Copies the content of another Signature-like class.
template <class SignatureLike>
void Signature::copy(const SignatureLike& other)
{
   if (other.initialA() != other.end())
      _initialA = _ids.insert(end(), _status_id(other.initialA()->id(), _status_id::kInitialA));
   else
      _initialA = end();
   if (other.initialB() != other.end())
      _initialB = _ids.insert(end(), _status_id(other.initialB()->id(), _status_id::kInitialB));
   else
      _initialB = end();
   typename SignatureLike::iterator it = other.beginFinal();
   if (it != other.endFinal()) {
      _beginFinal = _ids.insert(end(), it->id());
      ++it;
      for (; it != other.endFinal(); ++it)
         _ids.insert(end(), it->id());
   } else
      _beginFinal = end();
}

} // namespace Geneva

#endif // GENEVA_SIGNATURE_HPP
