//------------------------------------------------------------------------------
/// \file SignatureSet.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SignatureSet
//------------------------------------------------------------------------------

#ifndef GENEVA_SIGNATURE_SET_HPP
#define GENEVA_SIGNATURE_SET_HPP

#include "Geneva/Core/Physics/Signature.hpp"

#include <set>
#include <iosfwd>
#include <algorithm>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class SignatureSet
 * \ingroup Core
 *
 * \brief A sorted set of unique Signatures.
 *
 * The Signatures in the %SignatureSet are unique, i.e. the set cannot contain
 * duplicate entries. The signatures are automatically sorted by their size,
 * with equal-size Signatures sorted in ascending order using the < operator of
 * Signature. The %SignatureSet is implemented in terms of a STL set with its
 * standard interface, allowing for fast access, insertion/removal, and typical
 * set operations. It supports a variety of set operations (union, etc.) that
 * are accessed by algebraic or operators with the obvious meanings:
 *
 * - s += t, s + t: The union of the two SignatureSets s and t
 * - s -= t, s - t: The difference (complement) of s with respect to t
 * - s &= t, s & t: The intersection of s and t
 * - s ^= t, s ^ t: The symmetric difference of s and t: (s + t) - (s & t)
 * - s == t, s != t: Test for (in)equality
 * - s < t, s <= t, s > t, s >= t: Lexicographical comparison that can be used to order SignatureSets
 *
 * A %SignatureSet can be traversed using iterators that have the type
 * SignatureSet::iterator and SignatureSet::reverse_iterator. Various member
 * functions use or return iterators. %SignatureSet iterators behave essentially
 * like read-only pointers (they are const_iterators). The Signature pointed to
 * by an iterator 'i' is accessed (for reading only) using either *i or i->.
 * Iterators can incremented, i++, or decremented, i--, to point to the
 * next or previous Signature in the set, and can be checked for (in)equality
 * (using == or !=). The relation operators (<, <=, >, >=) cannot be used with
 * %SignatureSet iterators.
 *
 * The members begin() and end() return iterators pointing to the first and
 * one step beyond the last Signature, respectively. Hence a %SignatureSet can
 * be traversed using
 * \code
 *    for (SignatureSet::iterator i = set.begin(); i != set.end(), ++i) {
 *       doSomethingwithSignature(*i);
 *       ...
 *    }
 * \endcode
 *
 * A reverse_iterator behaves exactly opposite which is useful if one wants to
 * traverse the %SignatureSet in descending order, i.e., rbegin() and rend()
 * point to the last and one step before the first Signature, respectively, and
 * after incrementing a reverse_iterator it points to the previous Signature.
 *
 * An existing iterator pointing to a Signature in a %SignatureSet stays valid
 * if new Signatures are inserted or other Signatures are erased. Of course it
 * does become invalid if the Signature it points to is erased.
 */
//------------------------------------------------------------------------------
class SignatureSet
{
   private:
      std::set<Signature> _sigs;   ///< the set's signatures

   public:
      /// \name Type Definitions
      ///@{
      typedef std::set<Signature>::value_type       value_type;
      typedef std::set<Signature>::pointer          pointer;
      typedef std::set<Signature>::const_pointer    const_pointer;
      typedef std::set<Signature>::reference        reference;
      typedef std::set<Signature>::const_reference  const_reference;

      typedef std::set<Signature>::iterator         iterator;
      typedef std::set<Signature>::const_iterator   const_iterator;
      typedef std::set<Signature>::reverse_iterator reverse_iterator;
      typedef std::set<Signature>::const_reverse_iterator const_reverse_iterator;
      ///@}

      /// \name Constructors etc.
      ///@{

      /// Creates an empty %SignatureSet.
      SignatureSet() = default;

      /// Creates a %SignatureSet with \a sig as the only element.
      SignatureSet(const Signature& sig);

      // Creates a SignatureSet from a range of Signatures.
      template <typename InputIterator>
      SignatureSet(InputIterator first, InputIterator last);
      ///@}

      /// \name Accessing and Querying Signatures
      ///@{
      // Returns an iterator pointing to the first Signature in the SignatureSet.
      iterator begin() const;

      // Returns an iterator pointing beyond the last Signature in the SignatureSet.
      iterator end() const;

      // Returns a reverse_iterator pointing to the last Signature in the SignatureSet.
      reverse_iterator rbegin() const;

      // Returns a reverse_iterator pointing before the first Signature in the SignatureSet.
      reverse_iterator rend() const;

      /// Returns \c true if the %SignatureSet is empty, \c false otherwise.
      bool empty() const;

      /// Returns the total number of Signatures in the %SignatureSet.
      size_t size() const;

      /// Returns the maximum number of Signatures the %SignatureSet could hold.
      size_t max_size() const;

      // Tries to find a Signature in the SignatureSet.
      iterator find(const Signature& sig) const;

      // Finds the first Signature in the SignatureSet satisfying a lower bound.
      iterator lower_bound(const Signature& min) const;

      // Finds the first Signature in the SignatureSet that exceeds an upper bound.
      iterator upper_bound(const Signature& max) const;

      // Tests if a Signature is contained in the SignatureSet.
      bool includes(const Signature& sig) const;

      // Tests if a SignatureSet is contained in the SignatureSet.
      bool includes(const SignatureSet& other) const;

      // Tests if a SignatureSet intersects with the SignatureSet.
      bool intersects(const SignatureSet& other) const;
      ///@}

      /// \name Adding and Removing Signatures
      ///@{
      // Swaps the contents with another SignatureSet.
      void swap(SignatureSet& other);

      // Inserts a Signature into the SignatureSet.
      iterator insert(const Signature& sig);

      // Inserts a Signature into the SignatureSet using a hint.
      iterator insert(iterator hint, const Signature& sig);

      // Inserts a range of Signatures into the SignatureSet.
      template <typename InputIterator>
      void insert(InputIterator first, InputIterator last);

      // Removes the Signature from the SignatureSet pointed to by it.
      void erase(iterator it);

      // Removes a Signature from the SignatureSet.
      bool erase(const Signature& sig);

      // Removes a range of Signatures from the SignatureSet.
      void erase(iterator first, iterator last);

      /// Removes all Signatures from the %SignatureSet.
      void clear();

      // Counts the number of Signatures satisfying a condition.
      template <class Condition>
      size_t count_if(const Condition& cond) const;

      // Copies all Signatures satisfying a condition to a specified destination.
      template <class OutputIterator, class Condition>
      OutputIterator copy_if(OutputIterator result, const Condition& cond) const;

      // Copies all Signatures satisfying a condition,
      template <class Condition>
      const SignatureSet copy_if(const Condition& cond) const;

      // Removes all Signatures satisfying a condition.
      template <class Condition>
      size_t erase_if(const Condition& cond);

      // Moves all Signatures satisfying a condition to a specified destination.
      template <class OutputIterator, class Condition>
      OutputIterator move_if(OutputIterator it, const Condition& cond);

      // Moves all Signatures satisfying a condition.
      template <class Condition>
      const SignatureSet move_if(const Condition& cond);
      ///@}

      /// \name Algebra
      ///@{
      // SignatureSet equality comparison operator
      bool operator==(const SignatureSet& rhs) const;

      // SignatureSet nonequality comparison operator
      bool operator!=(const SignatureSet& rhs) const;

      // SignatureSet ordering relation operator
      bool operator<(const SignatureSet& rhs) const;

      /// Returns \c true if %SignatureSet is less or equal to \a rhs.
      bool operator<=(const SignatureSet& rhs) const;

      /// Returns \c true if %SignatureSet is greater or equal to \a rhs.
      bool operator>=(const SignatureSet& rhs) const;

      /// Returns \c true if %SignatureSet is greater than \a rhs.
      bool operator>(const SignatureSet& rhs) const;

      // SignatureSet addition (union) assignment operator
      SignatureSet& operator+=(const SignatureSet& rhs);

      // SignatureSet subtraction (complement) assignment operator
      SignatureSet& operator-=(const SignatureSet& rhs);

      // SignatureSet AND (intersection) assignment operator
      SignatureSet& operator&=(const SignatureSet& rhs);

      // SignatureSet XOR (symmetric difference) assignment operator
      SignatureSet& operator^=(const SignatureSet& rhs);
      ///@}

      /// \name Output
      ///@{
      // SignatureSet stream insertion operator.
      friend std::ostream& operator<<(std::ostream& out, const SignatureSet& sig);
      ///@}

   private:
      /// Streams the %SignatureSet to \a output.
      std::ostream& print(std::ostream& output) const;
};

// SignatureSet addition (union) operator
const SignatureSet operator+(const SignatureSet& lhs, const SignatureSet& rhs);

// SignatureSet subtraction (complement) operator
const SignatureSet operator-(const SignatureSet& lhs, const SignatureSet& rhs);

// SignatureSet AND (intersection) operator
const SignatureSet operator&(const SignatureSet& lhs, const SignatureSet& rhs);

// SignatureSet XOR (symmetric difference) operator
const SignatureSet operator^(const SignatureSet& lhs, const SignatureSet& rhs);


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Creates a %SignatureSet with sig as the only element.
inline SignatureSet::SignatureSet(const Signature& sig)
   : _sigs()
{
   insert(sig);
}

//------------------------------------------------------------------------------
/**
 * \brief Creates a %SignatureSet from a range of Signatures.
 * \param first An iterator pointing to the first Signature in the range.
 * \param last  An iterator pointing beyond the last Signature in the range.
 *
 * Create a new %SignatureSet by copying the Signatures from the range specified
 * by [\a first, \a last). The complexity is as for sets: Linear in N if the range is
 * already sorted (e.g. when obtained from another %SignatureSet) and N log N
 * otherwise (e.g. when obtained from a std::vector of Signaturess), where N is
 * the number of Signatures in the range.
 *
 * This is a template function that works for any type of iterator from which a
 * Signature can be read.
 */
template <typename InputIterator>
inline SignatureSet::SignatureSet(InputIterator first, InputIterator last)
   : _sigs(first, last)
{}

//------------------------------------------------------------------------------
// Accessing and Querying Signatures
//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator pointing to the first Signature in the %SignatureSet.
 *
 * If the %SignatureSet is empty the returned iterator will be equal to end(). It
 * is the user's responsibility to ensure that the iterator is not used for
 * reading in this case.
 */
inline SignatureSet::iterator SignatureSet::begin() const
{
   return _sigs.begin();
}

//------------------------------------------------------------------------------
/**
 * \brief Returns an iterator pointing beyond the last Signature in the %SignatureSet.
 *
 * The returned iterator points one step beyond the last Signature and so should not be
 * used for reading. Instead, it is usually used to check that all Signatures have been
 * traversed.
 */
inline SignatureSet::iterator SignatureSet::end() const
{
   return _sigs.end();
}

//------------------------------------------------------------------------------
/**
 * \brief Returns a reverse_iterator pointing to the last Signature in the %SignatureSet.
 *
 * If the %SignatureSet is empty the returned iterator will be equal to rend(). It
 * is the user's responsibility to ensure that the iterator is not used for
 * reading in this case.
 */
inline SignatureSet::reverse_iterator SignatureSet::rbegin() const
{
   return _sigs.rbegin();
}

//------------------------------------------------------------------------------
/**
 * \brief Returns a reverse_iterator pointing before the first Signature in the %SignatureSet.
 *
 * The returned iterator points one step before the last Signature and so should not be
 * used for reading. Instead, it is usually used to check that all Signatures have been
 * traversed.
 */
inline SignatureSet::reverse_iterator SignatureSet::rend() const
{
   return _sigs.rend();
}

//------------------------------------------------------------------------------
// Returns true if the SignatureSet is empty.
inline bool SignatureSet::empty() const
{
   return _sigs.empty();
}

//------------------------------------------------------------------------------
// Returns the total number of Signatures in the Signature.
inline size_t SignatureSet::size() const
{
   return _sigs.size();
}

//------------------------------------------------------------------------------
// Returns the maximum number of Signatures the SignatureSet could hold.
inline size_t SignatureSet::max_size() const
{
   return _sigs.max_size();
}

//------------------------------------------------------------------------------
/**
 * \brief Tries to find a Signature in the %SignatureSet.
 * \param sig The Signature to be searched for.
 * \return An iterator pointing to the first instance of the sought Signature.
 *
 * If \a sig is found in the %SignatureSet the returned iterator points to it,
 * otherwise end() is returned. Requires logarithmic time.
 */
inline SignatureSet::iterator SignatureSet::find(const Signature& sig) const
{
   return _sigs.find(sig);
}

//------------------------------------------------------------------------------
/**
 * \brief Finds the first Signature in the %SignatureSet satisfying a lower bound.
 * \param min The Signature to be used as lower bound.
 * \return An iterator pointing to the first Signature >= \a min
 *
 * If the %SignatureSet contains no Signatures >= \a min, then end() is returned.
 * Requires logarithmic time.
 *
 * Together with upper_bound(max) this function can be used to find a limited
 * range of Signatures in the %SignatureSet.
 * Setting first = lower_bound(min) and last = upper_bound(max) then the range
 * [first, last) specifies all Signatures that satisfy min <= Signature <= max.
 */
inline SignatureSet::iterator SignatureSet::lower_bound(const Signature& min) const
{
   return _sigs.lower_bound(min);
}

//------------------------------------------------------------------------------
/**
 * \brief Finds the first Signature in the %SignatureSet that exceeds an upper bound.
 * \param max The Signature to be used as upper bound.
 * \return An iterator pointing one step beyond the last Signature <= \a max
 *
 * If all Signatures in the %SignatureSet are <= \a max, then end() is returned.
 * See lower_bound(min) for a typical use case. Requires logarithmic time.
 */
inline SignatureSet::iterator SignatureSet::upper_bound(const Signature& max) const
{
   return _sigs.upper_bound(max);
}

//------------------------------------------------------------------------------
/**
 * \brief Tests if a Signature is contained in the %SignatureSet.
 * \param sig The %Signature to be searched for.
 * \return \c true if \a sig is contained in %SignatureSet, \c false otherwise.
 *
 * This is identical to find(id) != end(), so if an iterator to the Signature is
 * needed in case it is contained, it is faster to use find(id) directly.
 */
inline bool SignatureSet::includes(const Signature& sig) const
{
   return find(sig) != end();
}

//------------------------------------------------------------------------------
// Adding and Removing Signatures
//------------------------------------------------------------------------------
/**
 * \brief Swaps the contents with another %SignatureSet.
 * \param other The %SignatureSet being swapped with.
 *
 * This exchanges the entire contents of the %SignatureSet with that of \a other.
 * The swapping does not require the copying of elements, which means it is fast
 * and executes in constant time.
 */
inline void SignatureSet::swap(SignatureSet& other)
{
   _sigs.swap(other._sigs);
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts a Signature into the %SignatureSet.
 * \param sig The Signature to be inserted.
 * \return An iterator pointing to the inserted Signature.
 *
 * If \a sig is already part of %SignatureSet, it is not added again and the
 * returned iterator points to the existing Signature. Insertion requires
 * logarithmic time.
 */
inline SignatureSet::iterator SignatureSet::insert(const Signature& sig)
{
   return _sigs.insert(sig).first;
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts a Signature into the %SignatureSet using a hint.
 * \param hint An iterator used as a hint where the Signature should be inserted.
 * \param sig The Signature to be inserted.
 * \return An iterator pointing to the inserted Signature.
 *
 * If \a sig is already part of %SignatureSet, it is not added again and the
 * returned iterator points to the existing Signature. A good \a hint can speed
 * up the insertion, while a bad hint will have no speed gain, in which case
 * insertion requires logarithmic time, just as insert(sig) would.
 */
inline SignatureSet::iterator SignatureSet::insert(SignatureSet::iterator hint, const Signature& sig)
{
   return _sigs.insert(hint, sig);
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts a range of Signatures into the %SignatureSet.
 * \param first An iterator pointing to the first Signature in the range.
 * \param last  An iterator pointing beyond the last Signature in the range.
 *
 * Inserts the Signatures from the range specified by [\a first,\a last). Only
 * new Signatures that are not already part of the %SignatureSet are inserted.
 * The complexity is as for the analogous range constructor. This is a template
 * function that works for any type of iterator from which a Signature can be
 * read.
 */
template <typename InputIterator>
inline void SignatureSet::insert(InputIterator first, InputIterator last)
{
   _sigs.insert(first, last);
}

//------------------------------------------------------------------------------
/**
 * \brief Removes the Signature from the SignatureSet pointed to by \a it.
 * \param it An iterator pointing to the ID to be erased.
 *
 * \note It is the user's responsibility to ensure that \a it is a valid
 * iterator into the %SignatureSet.
 */
inline void SignatureSet::erase(SignatureSet::iterator it)
{
   _sigs.erase(it);
}

//------------------------------------------------------------------------------
/**
 * \brief Removes a Signature from the %SignatureSet.
 * \param sig The Signature to be erased.
 * \return \c true if \a sig was part of %SignatureSet and was erased, \c false
 *         otherwise.
 */
inline bool SignatureSet::erase(const Signature& sig)
{
   return _sigs.erase(sig);
}

//------------------------------------------------------------------------------
/**
 * \brief Removes a range of Signatures from the %SignatureSet.
 * \param first An iterator pointing to the first Signature in the %SignatureSet to be erased.
 * \param last  An iterator pointing beyond the last Signature in the %SignatureSet to be erased.
 *
 * The Signatures in the %SignatureSet specified by [\a first,\a last) are
 * removed. The complexity is linear in the number of Signatures to be erased.
 * \note It is the user's responsibility to ensure that [\a first, \a last) is
 * a valid iterator range into the %Signature.
 */
inline void SignatureSet::erase(SignatureSet::iterator first, SignatureSet::iterator last)
{
   _sigs.erase(first, last);
}

//------------------------------------------------------------------------------
// Removes all Signatures from the Signature.
inline void SignatureSet::clear()
{
   _sigs.clear();
}

//------------------------------------------------------------------------------
/**
 * \brief Counts the number of Signatures satisfying a condition.
 * \param cond The function object used as condition.
 * \return The number of found Signatures.
 *
 * Counts all Signatures for which the application of \a cond gives \c true. If
 * no such Signatures are found the return value is 0.
 */
template <class Condition>
size_t SignatureSet::count_if(const Condition& cond) const
{
   size_t count(0);
   for (iterator it = begin(); it != end(); ++it) {
      if (cond(*it)) {
         ++count;
      }
   }
   return count;
}

//------------------------------------------------------------------------------
/**
 * \brief Copies all Signatures satisfying a condition to a specified destination.
 * \param dest An output iterator where the Signatures will be copied to.
 * \param cond The function object used as condition.
 * \return An output iterator pointing one step beyond the last copied Signature.
 *
 * All Signatures for which the application of \a cond gives \c true are copied
 * to the iterator range starting at \a dest. The returned iterator marks the
 * end of the range where the Signatures were copied, i.e. the final range where
 * the copied Signatures are located is given by [\a dest, \a returnvalue).
 * Note that \a dest is an OutputIterator which must be writable. To use this
 * function to insert the selected Signatures into a container using it's
 * \c insert() function the \c std::inserter adapter can be used.
 */
template <class OutputIterator, class Condition>
OutputIterator SignatureSet::copy_if(OutputIterator dest, const Condition& cond) const
{
   for (iterator it = begin(); it != end(); ++it) {
      if (cond(*it)) {
         *dest = *it;
         ++dest;
      }
   }
   return dest;
}

//------------------------------------------------------------------------------
/**
 * \brief Copies all Signatures satisfying a condition.
 * \param cond The function object used as condition.
 * \return A %SignatureSet containing the copied Signatures.
 *
 * All Signatures for which the application of \a cond gives \c true are copied
 * into the returned %SignatureSet.
 */
template <class Condition>
const SignatureSet SignatureSet::copy_if(const Condition& cond) const
{
   SignatureSet temp;
   copy_if(std::inserter(temp, temp.end()), cond);
   return temp;
}

//------------------------------------------------------------------------------
/**
 * \brief Removes all Signatures satisfying a condition.
 * \param cond The function object used as condition.
 * \return The number of Signatures erased.
 *
 * All Signatures for which the application of \a cond gives \c true are erased
 * from the %SignatureSet.
 */
template <class Condition>
size_t SignatureSet::erase_if(const Condition& cond)
{
   iterator it = begin();
   size_t count = 0;
   while (it != end()) {
      if (cond(*it)) {
         erase(it++);
         ++count;
      } else
         ++it;
   }
   return count;
}

//------------------------------------------------------------------------------
/**
 * \brief Moves all Signatures satisfying a condition to a specified destination.
 * \param dest An output iterator where the Signatures will be moved to.
 * \param cond The function object used as condition.
 * \return An output iterator pointing one step beyond the last copied Signature.
 *
 * All Signatures for which the application of \a cond gives \c true are copied
 * to the iterator range starting at \a dest and are also erased from the
 * %SignatureSet. The returned iterator marks the end of the range where the
 * Signatures were moved, i.e. the final range where the moved Signatures are
 * located is given by [\a dest, \a returnvalue). Note that \a dest is an
 * OutputIterator which must be writable. To use this function to insert the
 * selected Signatures into a container using its \c insert() function the
 * \c std::inserter adapter can be used.
 */
template <class OutputIterator, class Condition>
OutputIterator SignatureSet::move_if(OutputIterator dest, const Condition& cond)
{
   iterator it = begin();
   while (it != end()) {
      if (cond(*it)) {
         *dest = *it;
         ++dest;
         erase(it++);
      } else
         ++it;
   }
   return dest;
}

//------------------------------------------------------------------------------
/**
 * \brief Moves all Signatures satisfying a condition.
 * \param cond The function object used as condition.
 * \return A %SignatureSet containing the moved Signatures.
 *
 * All Signatures for which the application of \a cond gives \c true are copied
 * into the returned SignatureSet and are also erased from the %SignatureSet.
 */
template <class Condition>
const SignatureSet SignatureSet::move_if(const Condition& cond)
{
   SignatureSet temp;
   move_if(std::inserter(temp, temp.end()), cond);
   return temp;
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------
/**
 * \brief %SignatureSet equality comparison operator
 * \param rhs The %SignatureSet to be compared to
 * \return \c true if the %SignatureSet contains exactly the same Signatures as
 *         \a rhs, \c false otherwise.
 *
 * Requires linear time.
 */
inline bool SignatureSet::operator==(const SignatureSet& rhs) const
{
   return _sigs == rhs._sigs;
}

//------------------------------------------------------------------------------
/**
 * \brief %SignatureSet nonequality comparison operator
 * \param rhs The %SignatureSet to be compared to
 * \return \c true if the %SignatureSet does not contain exactly the same
 *         Signatures as \a rhs, \c false otherwise.
 *
 * Identical to !operator==(rhs)
 */
inline bool SignatureSet::operator!=(const SignatureSet& rhs) const
{
   return !(_sigs == rhs._sigs);
}

//------------------------------------------------------------------------------
/**
 * \brief %SignatureSet ordering relation operator
 * \param rhs The %SignatureSet to be compared to
 * \return \c true if the %SignatureSet is considered "less than" \a rhs,
 *         \c false otherwise.
 *
 * Returns \c true if the %SignatureSet is "lexicographically" smaller than
 * \a rhs, and \c false otherwise, where the %SignatureSet is considered a "word"
 * and each of its Signatures a "letter".
 */
inline bool SignatureSet::operator<(const SignatureSet& rhs) const
{
   return _sigs < rhs._sigs;
}

//------------------------------------------------------------------------------
// Returns true if SignatureSet is less or equal to rhs.
inline bool SignatureSet::operator<=(const SignatureSet& rhs) const
{
   return !(rhs._sigs < _sigs);
}

//------------------------------------------------------------------------------
// Returns true if SignatureSet is greater or equal to rhs.
inline bool SignatureSet::operator>=(const SignatureSet& rhs) const
{
   return !(_sigs < rhs._sigs);
}

//------------------------------------------------------------------------------
// Returns true if SignatureSet is greater than rhs.
inline bool SignatureSet::operator>(const SignatureSet& rhs) const
{
   return rhs._sigs < _sigs;
}

//------------------------------------------------------------------------------
/**
 * \brief %SignatureSet addition (union) assignment operator
 * \param rhs The %SignatureSet to be added.
 *
 * Takes the union of the %SignatureSet with \a rhs. It inserts all Signatures
 * from \a rhs which are not already included in the %SignatureSet. Requires
 * linear time.
 */
inline SignatureSet& SignatureSet::operator+=(const SignatureSet& rhs)
{
   insert(rhs.begin(), rhs.end());
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %SignatureSet addition (union) operator
 * \param lhs The 1st %SignatureSet to be added.
 * \param rhs The 2nd %SignatureSet to be added.
 * \return The resulting %SignatureSet.
 *
 * Takes the union of \a lhs and \a rhs and returns the result in a new
 * %SignatureSet. Requires linear time.
 */
inline const SignatureSet operator+(const SignatureSet& lhs, const SignatureSet& rhs)
{
   SignatureSet temp(lhs);
   temp += rhs;
   return temp;
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------
/**
 * \brief %SignatureSet stream insertion operator.
 * \param out The output stream to be used.
 * \param sig The %SignatureSet to be streamed.
 * \return A reference to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const SignatureSet& sig)
{
   return sig.print(out);
}

} // namespace Geneva

#endif // GENEVA_SIGNATURE_SET_HPP
