//------------------------------------------------------------------------------
/// \file SignatureStatus.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SignatureStatus and SignatureStatusID
//------------------------------------------------------------------------------

#ifndef GENEVA_SIGNATURE_STATUS_HPP
#define GENEVA_SIGNATURE_STATUS_HPP

#include "Geneva/Core/Physics/PdgID.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class SignatureStatus
 * \ingroup Core
 *
 * \brief Indicates the status of an entry in a Signature-like class.
 *
 * \note This is an internal type used by Signature and should not be used
 * directly.
 */
//------------------------------------------------------------------------------
class SignatureStatus
{
   public:
      /// Returns \c true for initial-state A, \c false otherwise.
      bool isInitialA() const
      {
         return _status == kInitialA;
      }

      /// Returns \c true for initial-state B, \c false otherwise.
      bool isInitialB() const
      {
         return _status == kInitialB;
      }

      /// Returns \c true for a final state, \c false otherwise.
      bool isFinal() const
      {
         return _status == kFinal;
      }

   protected:
      enum
      {
         kInitialA = -2,
         kInitialB = -1,
         kFinal = 1
      };

      /// Default constructor using kFinal as the default status.
      SignatureStatus() : _status(kFinal) {}

      SignatureStatus(int status) : _status(status) {}

      int _status;
};


//------------------------------------------------------------------------------
/**
 * \class SignatureStatusID
 * \ingroup Core
 *
 * \brief Supplementes a PdgID by a SignatureStatus
 *
 * \note This is an internal type used by Signature and should not be used
 * directly.
 */
//------------------------------------------------------------------------------
class SignatureStatusID : public PdgID, public SignatureStatus
{
   friend class Signature;
   friend class Configuration;

   private:
      SignatureStatusID() = default;
      SignatureStatusID(PdgID id, int status = kFinal) : PdgID(id), SignatureStatus(status) {}

      PdgID id() const { return *this; }

      /// Binary function object defining an ordering relation.
      struct _compare_type
      {
         template <typename StatusType1, typename StatusType2>
         bool operator()(const StatusType1& lhs, const StatusType2& rhs) const
         {
            if (lhs._status == rhs._status)
               return lhs.id().code() < rhs.id().code();
            return lhs._status < rhs._status;
         }
      };
};

} // namespace Geneva

#endif // GENEVA_SIGNATURE_STATUS_HPP
