//------------------------------------------------------------------------------
/// \file Transformation.hpp
//
// Author(s):
//    Frank Tackmann, Chris Vermilion
//
// Copyright:
//    Copyright (C) 2014 MIT, LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface and definitions of basic vector transformations
//------------------------------------------------------------------------------

#ifndef GENEVA_TRANSFORMATION_HPP
#define GENEVA_TRANSFORMATION_HPP

#include "Geneva/Core/Physics/FourVector.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class Normalize
 * \ingroup Core
 *
 * \brief Normalization of a ThreeVector by its magnitude.
 *
 * Function object that normalizes a ThreeVector by its magnitude.
 *
 * When applied to a ThreeVector, it returns a normalized version of it, while
 * the ThreeVector itself is not changed.
 *
 * \note Application to a FourVector is currently not defined.
 */
//------------------------------------------------------------------------------
class Normalize
{
   public:
      Normalize();

      const ThreeVector operator()(const ThreeVector& operand) const;
};


//------------------------------------------------------------------------------
/**
 * \class RotationX
 * \ingroup Core
 *
 * \brief Rotation of a ThreeVector around the x-axis
 *
 * Function object that performs a rotation around the x-axis. The rotation is
 * performed in the mathematically positive direction around the x-axis
 * (right-hand rule).
 *
 * The %RotationX is constructed by providing either the rotation angle or the
 * cosine and sine of the rotation angle.
 *
 * When applied to a ThreeVector, it returns a rotated version of it, while the
 * ThreeVector itself is not changed. When applied to a FourVector, it returns
 * a version of it with its threevector part rotated, while the FourVector
 * itself is not changed.
 */
//------------------------------------------------------------------------------
class RotationX
{
   private:
      double _costheta;
      double _sintheta;

   public:
      RotationX(double theta);
      RotationX(double costheta, double sintheta);
      const RotationX inverse() const;

      const ThreeVector operator()(const ThreeVector& operand) const;
      const FourVector operator()(const FourVector& operand) const;
};

//------------------------------------------------------------------------------
/**
 * \class RotationY
 * \ingroup Core
 *
 * \brief Rotation of a ThreeVector around the y-axis
 *
 * Function object that performs a rotation around the y-axis. The rotation is
 * performed in the mathematically positive direction around the y-axis
 * (right-hand rule).
 *
 * The %RotationY is constructed by providing either the rotation angle or the
 * cosine and sine of the rotation angle.
 *
 * When applied to a ThreeVector, it returns a rotated version of it, while the
 * ThreeVector itself is not changed. When applied to a FourVector, it returns
 * a version of it with its threevector part rotated, while the FourVector
 * itself is not changed.
 */
//------------------------------------------------------------------------------
class RotationY
{
   private:
      double _costheta;
      double _sintheta;

   public:
      RotationY(double theta);
      RotationY(double costheta, double sintheta);
      const RotationY inverse() const;

      const ThreeVector operator()(const ThreeVector& operand) const;
      const FourVector operator()(const FourVector& operand) const;
};


//------------------------------------------------------------------------------
/**
 * \class RotationZ
 * \ingroup Core
 *
 * \brief Rotation of a ThreeVector around the z-axis
 *
 * Function object that performs a rotation around the z-axis. The rotation is
 * performed in the mathematically positive direction around the z-axis
 * (right-hand rule).
 *
 * The %RotationZ is constructed by providing either the rotation angle or the
 * cosine and sine of the rotation angle.
 *
 * When applied to a ThreeVector, it returns a rotated version of it, while the
 * ThreeVector itself is not changed. When applied to a FourVector, it returns
 * a version of it with its threevector part rotated, while the FourVector
 * itself is not changed.
 */
//------------------------------------------------------------------------------
class RotationZ
{
   private:
      double _costheta;
      double _sintheta;

   public:
      RotationZ(double theta);
      RotationZ(double costheta, double sintheta);
      const RotationZ inverse() const;

      const ThreeVector operator()(const ThreeVector& operand) const;
      const FourVector operator()(const FourVector& operand) const;
};


//------------------------------------------------------------------------------
/**
 * \class Rotation
 * \ingroup Core
 *
 * \brief Rotation of a ThreeVector around a general axis
 *
 * Function object that performs a rotation around a general axis. The rotation
 * is performed in the mathematically positive direction around the axis
 * (right-hand rule).
 *
 * When applied to a ThreeVector \f$ \vec{p} \f$, it returns a rotated version
 * of it, given by the Rodrigues' rotation formula
 * \f[
 * \vec{p}^{\,\prime} = \vec{p} \cos\theta + (\vec{n}\times\vec{p})\sin\theta + \vec{n} (\vec{n}\cdot\vec{p})(1-\cos\theta)
 * \f]
 * where \f$ \vec{n} \f$ is the rotation axis and \f$ \theta \f$ the rotation
 * angle. The ThreeVector itself is not changed. When applied to a FourVector,
 * it returns a version of it with its threevector part rotated, while the
 * FourVector itself is not changed.
 */
//------------------------------------------------------------------------------
class Rotation
{
   private:
      double _r11;
      double _r12;
      double _r13;
      double _r21;
      double _r22;
      double _r23;
      double _r31;
      double _r32;
      double _r33;

   public:
      Rotation(const ThreeVector& axis, double theta);
      Rotation(const ThreeVector& unitAxis, double costheta, double sintheta);
      Rotation(const ThreeVector& initial, const ThreeVector& final);
      const Rotation inverse() const;

      const ThreeVector operator()(const ThreeVector& operand) const;
      const FourVector operator()(const FourVector& operand) const;

   private:
      Rotation(double r11, double r12, double r13, double r21, double r22, double r23, double r31, double r32, double r33);
      void init(const ThreeVector& unitAxis, double costheta, double sintheta);
};


//------------------------------------------------------------------------------
/**
 * \class CrossProductWith
 * \ingroup Core
 *
 * \brief Cross product of a ThreeVector with another fixed ThreeVector.
 *
 * Function object that takes the cross product of the ThreeVector
 * \f$ \vec{p} \f$ it is applied to with a fixed ThreeVector \f$ \vec{q} \f$.
 * The %CrossProductWith object is constructed by providing the ThreeVector
 * \f$ \vec{q} \f$ with which to take the cross product.
 *
 * When applied to a ThreeVector \f$ \vec{p} \f$, it returns the result of
 * taking \f$ \vec{p} \times \vec{q} \f$, while the ThreeVector \f$ \vec{p} \f$
 * itself is not changed. When applied to a FourVector \f$ p^\mu \f$, it returns
 * the result of taking \f$ (p^0, \vec{p}\times \vec{q}) \f$, while the
 * FourVector \f$ p^\mu \f$ itself is not changed.
 */
//------------------------------------------------------------------------------
class CrossProductWith
{
   private:
      ThreeVector _rhs;

   public:
      CrossProductWith(const ThreeVector& rhs);

      const ThreeVector operator()(const ThreeVector& lhs) const;
      const FourVector operator()(const FourVector& lhs) const;
};


//------------------------------------------------------------------------------
/**
 * \class BoostX
 * \ingroup Core
 *
 * \brief Boost of a FourVector along the x-axis
 *
 * Function object that performs a boost along the x-axis. The %BoostX is
 * constructed by providing the boost factor \f$ \beta \f$.
 *
 * When applied to a FourVector, it returns a boosted version of it, while the
 * FourVector itself is not changed.
 */
//------------------------------------------------------------------------------
class BoostX
{
   private:
      double _beta;
      double _gamma;

   public:
      BoostX(double beta);
      const BoostX inverse() const;

      const FourVector operator()(const FourVector& operand) const;

   private:
      BoostX(double beta, double gamma);
};


//------------------------------------------------------------------------------
/**
 * \class BoostY
 * \ingroup Core
 *
 * \brief Boost of a FourVector along the y-axis
 *
 * Function object that performs a boost along the y-axis. The %BoostY is
 * constructed by providing the boost factor \f$ \beta \f$.
 *
 * When applied to a FourVector, it returns a boosted version of it, while the
 * FourVector itself is not changed.
 */
//------------------------------------------------------------------------------
class BoostY
{
   private:
      double _beta;
      double _gamma;

   public:
      BoostY(double beta);
      const BoostY inverse() const;

      const FourVector operator()(const FourVector& operand) const;

   private:
      BoostY(double beta, double gamma);
};


//------------------------------------------------------------------------------
/**
 * \class BoostZ
 * \ingroup Core
 *
 * \brief Boost of a FourVector along the z-axis
 *
 * Function object that performs a boost along the z-axis. The %BoostZ is
 * constructed by providing the boost factor \f$ \beta \f$.
 *
 * When applied to a FourVector, it returns a boosted version of it, while the
 * FourVector itself is not changed.
 *
 * \note: \c BoostZHyp should preferably be used, as it is numerically more
 * accurate and is expected to replace this version.
 */
//------------------------------------------------------------------------------
class BoostZ
{
   private:
      double _beta;
      double _gamma;

   public:
      BoostZ(double beta);
      const BoostZ inverse() const;

      const FourVector operator()(const FourVector& operand) const;

   private:
      BoostZ(double beta, double gamma);
};

//------------------------------------------------------------------------------
/**
 * \class BoostZHyp
 * \ingroup Core
 *
 * \brief Boost of a FourVector along the z-axis
 *
 * Function object that performs a boost along the z-axis. The %BoostZ is
 * constructed by providing the hyperpolic boost factor \f$ \lambda=\exp^{y} \f$
 * or a reference FourVector, which is used to determine the boost factor as
 * \f$ \lambda=\sqrt^{p^-/p^+} \f$.
 *
 * When applied to a FourVector, it returns a boosted version of it, while the
 * FourVector itself is not changed.
 */
//------------------------------------------------------------------------------
class BoostZHyp
{
   private:
      double _la;
      double _lainv;

   public:
      BoostZHyp(double la);
      BoostZHyp(const FourVector& p);

      const BoostZHyp inverse() const;

      const FourVector operator()(const FourVector& operand) const;

   private:
      BoostZHyp(double la, double lainv);
};


//------------------------------------------------------------------------------
/**
 * \class Boost
 * \ingroup Core
 *
 * \brief Boost of a FourVector along a general direction
 *
 * Function object that performs a boost along a general direction. The %Boost
 * is constructed by providing the boost vector \f$ \vec{\beta} \f$, which
 * describes both the direction and magnitude of the boost. For example, to
 * boost a FourVector \f$ p^\mu \f$ to its restframe, it must be boosted with
 * \f$ \vec{\beta} = -\vec{p}/p^0 \f$.
 *
 * When applied to a FourVector \f$ p^\mu \f$, it returns a boosted version of
 * it given by
 * \f[
 * \begin{aligned}
 *  (p^0)' &= \gamma(p^0 + \vec{\beta}\cdot\vec{p}) \\
 *  \vec{p}^{\,\prime} &= \gamma[ (\hat\beta\cdot \vec{p}) \hat\beta + \vec{\beta} p^0] + \vec{p} - (\hat\beta\cdot \vec{p}) \hat\beta
 * \quad\text{with}\quad \hat\beta = \frac{\vec{\beta}}{\lvert\vec{\beta}\rvert}
 * \end{aligned}
 * \f]
 * The FourVector itself is not changed.
 */
//------------------------------------------------------------------------------
class Boost
{
   private:
      ThreeVector _beta;
      double _gamma;
      double _gm1_b2;

   public:
      Boost(const ThreeVector& beta);
      Boost(const FourVector& p);
      const Boost inverse() const;

      const FourVector operator()(const FourVector& operand) const;

   private:
      Boost(const ThreeVector& beta, double gamma, double gm1_b2);
      void init();
};


////////////////////////////////////////////////////////////////////////////////
// Inline Declarations
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Normalize
//------------------------------------------------------------------------------
inline Normalize::Normalize()
{}

//------------------------------------------------------------------------------
inline const ThreeVector Normalize::operator()(const ThreeVector& operand) const
{
   return operand/operand.magnitude();
}


//------------------------------------------------------------------------------
// RotationX
//------------------------------------------------------------------------------
inline RotationX::RotationX(double theta)
   : _costheta(cos(theta)), _sintheta(sin(theta))
{}

//------------------------------------------------------------------------------
inline RotationX::RotationX(double costheta, double sintheta)
   : _costheta(costheta), _sintheta(sintheta)
{}

//------------------------------------------------------------------------------
inline const RotationX RotationX::inverse() const
{
   return RotationX(_costheta, -_sintheta);
}

//------------------------------------------------------------------------------
inline const ThreeVector RotationX::operator()(const ThreeVector& operand) const
{
   ThreeVector tmp(operand.p1(),
                   operand.p2() * _costheta - operand.p3() * _sintheta,
                   operand.p2() * _sintheta + operand.p3() * _costheta);
   return tmp;
}

//------------------------------------------------------------------------------
inline const FourVector RotationX::operator()(const FourVector& operand) const
{
   FourVector tmp(operand.p0(), operand.p1(),
                  operand.p2() * _costheta - operand.p3() * _sintheta,
                  operand.p2() * _sintheta + operand.p3() * _costheta);
   return tmp;
}


//------------------------------------------------------------------------------
// RotationY
//------------------------------------------------------------------------------
inline RotationY::RotationY(double theta)
   : _costheta(cos(theta)), _sintheta(sin(theta))
{}

//------------------------------------------------------------------------------
inline RotationY::RotationY(double costheta, double sintheta)
   : _costheta(costheta), _sintheta(sintheta)
{}

//------------------------------------------------------------------------------
inline const RotationY RotationY::inverse() const
{
   return RotationY(_costheta, -_sintheta);
}

//------------------------------------------------------------------------------
inline const ThreeVector RotationY::operator()(const ThreeVector& operand) const
{
   ThreeVector tmp(operand.p3() * _sintheta + operand.p1() * _costheta,
                   operand.p2(),
                   operand.p3() * _costheta - operand.p1() * _sintheta);
   return tmp;
}

//------------------------------------------------------------------------------
inline const FourVector RotationY::operator()(const FourVector& operand) const
{
   FourVector tmp(operand.p0(),
                  operand.p3() * _sintheta + operand.p1() * _costheta,
                  operand.p2(),
                  operand.p3() * _costheta - operand.p1() * _sintheta);
   return tmp;
}


//------------------------------------------------------------------------------
// RotationZ
//------------------------------------------------------------------------------
inline RotationZ::RotationZ(double theta)
   : _costheta(cos(theta)), _sintheta(sin(theta))
{}

//------------------------------------------------------------------------------
inline RotationZ::RotationZ(double costheta, double sintheta)
   : _costheta(costheta), _sintheta(sintheta)
{}

//------------------------------------------------------------------------------
inline const RotationZ RotationZ::inverse() const
{
   return RotationZ(_costheta, -_sintheta);
}

//------------------------------------------------------------------------------
inline const ThreeVector RotationZ::operator()(const ThreeVector& operand) const
{
   ThreeVector tmp(operand.p1() * _costheta - operand.p2() * _sintheta,
                   operand.p1() * _sintheta + operand.p2() * _costheta,
                   operand.p3());
   return tmp;
}

//------------------------------------------------------------------------------
inline const FourVector RotationZ::operator()(const FourVector& operand) const
{
   FourVector tmp(operand.p0(),
                  operand.p1() * _costheta - operand.p2() * _sintheta,
                  operand.p1() * _sintheta + operand.p2() * _costheta,
                  operand.p3());
   return tmp;
}

//------------------------------------------------------------------------------
// Rotation
//------------------------------------------------------------------------------
/**
 * \brief Construct a %Rotation from a rotation axis and angle
 * \param axis The ThreeVector to be used as the rotation axis.
 * \param theta The rotation angle.
 *
 * \a axis need not be normalized. The %Rotation is constructed using the usual
 * right-hand rule, i.e. it will be counter-clockwise by an angle \a theta in
 * the frame where \a axis is pointing towards the observer.
 */
inline Rotation::Rotation(const ThreeVector& axis, double theta)
{
   init(Normalize()(axis), cos(theta), sin(theta));
}

//------------------------------------------------------------------------------
/**
 * \brief Construct a %Rotation from a unit rotation axis and cos and sin of an angle
 * \param unitAxis The ThreeVector to be used as the rotation axis.
 * \param costheta The cosine of the rotation angle.
 * \param sintheta The sine of the rotation angle.
 *
 * \a unitAxis must be normalized. Also no checks are performed if \a costheta
 * and \a sintheta indeed correspond to the cosine and sine of the same angle.
 *
 * The %Rotation is constructed using the usual right-hand rule, i.e. it will be
 * counter-clockwise by an angle \a theta in the frame where \a axis is pointing
 * towards the observer.
 */
inline Rotation::Rotation(const ThreeVector& unitAxis, double costheta, double sintheta)
{
   init(unitAxis, costheta, sintheta);
}

//------------------------------------------------------------------------------
/**
 * \brief Construct a %Rotation that brings a %ThreeVector into another
 * \param initial The starting ThreeVector.
 * \param final The resulting ThreeVector.
 *
 * \a initial and  \a final need not be normalized. The %Rotation is constructed
 * using the usual right-hand rule, i.e. it will be counter-clockwise by an angle \a theta in
 * the frame where \a axis is pointing towards the observer. \a theta and \a axis are
 * determined via dot and cross products between \a initial and \a final.
 */
inline Rotation::Rotation(const ThreeVector& initial, const ThreeVector& final)
{
   const double theta = acos((initial*final)/(initial.magnitude()*final.magnitude()));
   const ThreeVector& axis = crossProduct(initial, final);
   init(Normalize()(axis), cos(theta), sin(theta));
}

//------------------------------------------------------------------------------
inline const Rotation Rotation::inverse() const
{
   return Rotation(_r11, _r21, _r31, _r12, _r22, _r32, _r13, _r23, _r33);
}

//------------------------------------------------------------------------------
inline const ThreeVector Rotation::operator()(const ThreeVector& operand) const
{
   ThreeVector tmp(_r11 * operand.p1() + _r12 * operand.p2() + _r13 * operand.p3(),
                   _r21 * operand.p1() + _r22 * operand.p2() + _r23 * operand.p3(),
                   _r31 * operand.p1() + _r32 * operand.p2() + _r33 * operand.p3());
   return tmp;
}

//------------------------------------------------------------------------------
inline const FourVector Rotation::operator()(const FourVector& operand) const
{
   FourVector tmp(operand.p0(),
                  _r11 * operand.p1() + _r12 * operand.p2() + _r13 * operand.p3(),
                  _r21 * operand.p1() + _r22 * operand.p2() + _r23 * operand.p3(),
                  _r31 * operand.p1() + _r32 * operand.p2() + _r33 * operand.p3());
   return tmp;
}

//------------------------------------------------------------------------------
inline Rotation::Rotation(double r11, double r12, double r13, double r21, double r22, double r23, double r31, double r32, double r33)
   : _r11(r11), _r12(r12), _r13(r13), _r21(r21), _r22(r22), _r23(r23), _r31(r31), _r32(r32), _r33(r33)
{}

//------------------------------------------------------------------------------
inline void Rotation::init(const ThreeVector& unitAxis, double costheta, double sintheta)
{
   ThreeVector nsin(sintheta * unitAxis);
   ThreeVector ncos((1. - costheta) * unitAxis);

   // diagonal elements
   _r11 = ncos.p1() * unitAxis.p1() + costheta;
   _r22 = ncos.p2() * unitAxis.p2() + costheta;
   _r33 = ncos.p3() * unitAxis.p3() + costheta;

   _r12 = _r21 = ncos.p1() * unitAxis.p2();
   _r13 = _r31 = ncos.p1() * unitAxis.p3();
   _r23 = _r32 = ncos.p2() * unitAxis.p3();

   _r12 -= nsin.p3();
   _r21 += nsin.p3();
   _r13 += nsin.p2();
   _r31 -= nsin.p2();
   _r23 -= nsin.p1();
   _r32 += nsin.p1();
}


//------------------------------------------------------------------------------
// CrossProductWith
//------------------------------------------------------------------------------
inline CrossProductWith::CrossProductWith(const ThreeVector& rhs)
   : _rhs(rhs)
{}

//------------------------------------------------------------------------------
inline const ThreeVector CrossProductWith::operator()(const ThreeVector& lhs) const
{
   return crossProduct(lhs, _rhs);
}

//------------------------------------------------------------------------------
inline const FourVector CrossProductWith::operator()(const FourVector& lhs) const
{
   FourVector tmp(lhs.p0(), crossProduct(lhs.pvec(), _rhs));
   return tmp;
}


//------------------------------------------------------------------------------
// BoostX
//------------------------------------------------------------------------------
inline BoostX::BoostX(double beta)
   : _beta(beta), _gamma(1./sqrt(1. - beta*beta))
{}

//------------------------------------------------------------------------------
inline const FourVector BoostX::operator()(const FourVector& operand) const
{
   FourVector tmp(_gamma*(operand.p0() + _beta*operand.p1()),
                  _gamma*(operand.p1() + _beta*operand.p0()),
                  operand.p2(), operand.p3());
   return tmp;
}

//------------------------------------------------------------------------------
inline const BoostX BoostX::inverse() const
{
   return BoostX(-_beta, _gamma);
}

//------------------------------------------------------------------------------
inline BoostX::BoostX(double beta, double gamma)
   : _beta(beta), _gamma(gamma)
{}


//------------------------------------------------------------------------------
// BoostY
//------------------------------------------------------------------------------
inline BoostY::BoostY(double beta)
   : _beta(beta), _gamma(1./sqrt(1. - beta*beta))
{}

//------------------------------------------------------------------------------
inline const FourVector BoostY::operator()(const FourVector& operand) const
{
   FourVector tmp(_gamma*(operand.p0() + _beta*operand.p2()),
                  operand.p1(),
                  _gamma*(operand.p2() + _beta*operand.p0()),
                  operand.p3());
   return tmp;
}

//------------------------------------------------------------------------------
inline const BoostY BoostY::inverse() const
{
   return BoostY(-_beta, _gamma);
}

//------------------------------------------------------------------------------
inline BoostY::BoostY(double beta, double gamma)
   : _beta(beta), _gamma(gamma)
{}


//------------------------------------------------------------------------------
// BoostZ
//------------------------------------------------------------------------------
inline BoostZ::BoostZ(double beta)
   : _beta(beta), _gamma(1./sqrt(1. - beta*beta))
{}

//------------------------------------------------------------------------------
inline const FourVector BoostZ::operator()(const FourVector& operand) const
{
   FourVector tmp(_gamma*(operand.p0() + _beta*operand.p3()),
                  operand.p1(), operand.p2(),
                  _gamma*(operand.p3() + _beta*operand.p0()));
   return tmp;
}

//------------------------------------------------------------------------------
inline const BoostZ BoostZ::inverse() const
{
   return BoostZ(-_beta, _gamma);
}

//------------------------------------------------------------------------------
inline BoostZ::BoostZ(double beta, double gamma)
   : _beta(beta), _gamma(gamma)
{}

//------------------------------------------------------------------------------
// BoostZ
//------------------------------------------------------------------------------
inline BoostZHyp::BoostZHyp(double la)
   : _la(la), _lainv(1./la)
{}

//------------------------------------------------------------------------------
inline BoostZHyp::BoostZHyp(const FourVector& p)
   : _la(sqrt(p.minus() / p.plus())), _lainv(1./_la)
{}

//------------------------------------------------------------------------------
inline const FourVector BoostZHyp::operator()(const FourVector& operand) const
{
   const double minus = operand.minus()*_la;
   const double plus = operand.plus()*_lainv;
   FourVector tmp(0.5*(minus + plus),
                  operand.p1(), operand.p2(),
                  0.5*(minus - plus));
   return tmp;
}

//------------------------------------------------------------------------------
inline const BoostZHyp BoostZHyp::inverse() const
{
   return BoostZHyp(_lainv, _la);
}

//------------------------------------------------------------------------------
inline BoostZHyp::BoostZHyp(double la, double lainv)
   : _la(la), _lainv(lainv)
{}

//------------------------------------------------------------------------------
// Boost
//------------------------------------------------------------------------------
/**
 * \brief Constructs a %Boost from a general boost vector.
 * \param beta The ThreeVector to be used as the boost vector.
 *
 * \note No checks are performed on the validity of \a beta. The user is
 * responsible to ensure that \a beta describes a valid boost, i.e. that the
 * square of \a beta is less than 1.
 */
inline Boost::Boost(const ThreeVector& beta)
   : _beta(beta)
{
   init();
}

//------------------------------------------------------------------------------
/**
 * \brief Constructs a %Boost from a timelike FourVector.
 * \param p The FourVector from which to obtain the boost vector.
 *
 * The boost vector is obtained from \a p as \f$ \vec\beta = \vec{p} / p^0 \f$.
 *
 * \note No checks are performed on the validity of \a p. The user is
 * responsible to ensure that \a p is timelike so it yields a valid boost vector.
 */
inline Boost::Boost(const FourVector& p)
   : _beta(p.pvec()/p.p0())
{
   init();
}

//------------------------------------------------------------------------------
inline const Boost Boost::inverse() const
{
   return Boost(-_beta, _gamma, _gm1_b2);
}

//------------------------------------------------------------------------------
inline const FourVector Boost::operator()(const FourVector& operand) const
{
   const double betadotp = _beta * operand.pvec();
   FourVector tmp(_gamma*(operand.p0() + betadotp),
                  operand.pvec() + _beta * (_gamma*operand.p0() + _gm1_b2*betadotp));
   return tmp;
}

//------------------------------------------------------------------------------
inline Boost::Boost(const ThreeVector& beta, double gamma, double gm1_b2)
   : _beta(beta), _gamma(gamma), _gm1_b2(gm1_b2)
{}

//------------------------------------------------------------------------------
inline void Boost::init()
{
   const double tmp = sqrt(1. - _beta.square());
   _gamma = 1./tmp;
   _gm1_b2 = _gamma / (1. + tmp);
}

} // namespace Geneva

#endif // GENEVA_TRANSFORMATION_HPP
