//------------------------------------------------------------------------------
/// \file UnitCubeRad.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class UnitCubeRad
//------------------------------------------------------------------------------

#ifndef GENEVA_UNIT_CUBE_RAD_HPP
#define GENEVA_UNIT_CUBE_RAD_HPP

#include "Geneva/Core/Physics/PdgID.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <cassert>
#include <iosfwd>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class UnitCubeRad
 * \ingroup Core
 *
 * \brief Data container that holds the 3 variables to fully describe a 1 -> 2 splitting.
 *
 * The variables used to describe the emission phase space are
 *
 * - \f$ x1 \f$ rescaled energy of the emitted parton
 * - \f$ x2 \f$ rescaled angle between emitted parton and sister
 * - \f$ x3 \f$ rescaled azimuthal angle of the daughters
 * - IDs of the mother particle, emitted parton (right daughter), and its sister (left daughter)
 * - RadiationType that distinguishes between ISR and FSR
 *
 *  The limits on x variables are
 * - \f$ 0 \leq x1 \leq 1 \f$
 * - \f$ 0 \leq x2 \leq 1 \f$
 * - \f$ 0 \leq x3 \leq 1 \f$
 *  These limits are enforced by explicit checks.
 *
 *  If an importance sampling weight has been used in the generation of a UnitCubeRad,
 *  e.g. from a UnitCubeGenerator, it must be stored internally and can be returned trough
 *  a call to weight(). A flat generation must return unit weight.
 *
 */
//------------------------------------------------------------------------------
class UnitCubeRad
{
   public:
      enum RadiationType
      {
         kISRA, /// Initial state radiation from initial particle A
         kISRB, /// Initial state radiation from initial particle A
         kFSR,  /// Final state radiation
         kUndefined
      };

   private:
      double _x1;
      double _x2;
      double _x3;

      PdgID _idMother;       ///< PdgID of the mother
      PdgID _idSister;       ///< PdgID of the left (1st) daughter (sister)
      PdgID _idEmitted;      ///< PdgID of the right (2nd) daughter (emitted parton)

      RadiationType _radType;  ///< whether is a final or initial state splitting

      double _weight;        ///<importance sample weight used in construction

   public:
      /// \name Constructors etc.
      ///@{

      /// Creates a default empty %UnitCubeRad.
      UnitCubeRad();

      /// Creates a %UnitCubeRad from the given arguments.
      UnitCubeRad(double x1, double x2, double  x3, PdgID idMother, PdgID idEmitted, PdgID idSister, RadiationType radType, double weight);

      ///Copy-constructor
      UnitCubeRad(const UnitCubeRad& other);
      ///@}

      /// \name Read and Write Access
      ///@{
      double x1() const;          ///< Returns x1.
      double x2() const;          ///< Returns x2.
      double x3() const;          ///< Returns x3.

      PdgID idMother() const;      ///< Returns PdgID of the mother.
      PdgID idEmitted() const;     ///< Returns PdgID of the emitted particle.
      PdgID idSister() const;       ///< Returns PdgID of the sister.

      RadiationType radType() const;      ///< Returns radiation type.

      double weight() const;      ///< Returns importance sample weight

      double& x1();               ///< Access x1
      double& x2();               ///< Access x2
      double& x3();               ///< Access x3

      PdgID& idMother();          ///< Access PdgID of the mother
      PdgID& idEmitted();         ///< Access PdgID of the emitted particle
      PdgID& idSister();           ///< Access PdgID of the sister

      RadiationType& radType();   ///< Access radiation type.

      double& weight();           ///< Access importance sample weight

      ///@}

      /// \name Output
      ///@{
      // UnitCubeRad stream insertion operator
      friend std::ostream& operator<<(std::ostream& out, const UnitCubeRad& UnitCubeRad);
      ///@}

   protected:
      /// Streams info to \a out.
      std::ostream& print(std::ostream& out) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Declarations
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------
// Creates a default UnitCubeRad.
inline UnitCubeRad::UnitCubeRad()
      : _x1(std::numeric_limits<double>::quiet_NaN()), _x2(std::numeric_limits<double>::quiet_NaN()), _x3(std::numeric_limits<double>::quiet_NaN()),_idMother(PdgID::kUnknown),
      _idSister(PdgID::kUnknown), _idEmitted(PdgID::kUnknown), _radType(kUndefined), _weight(1.0)
{}

//------------------------------------------------------------------------------
// Creates a UnitCubeRad from the given arguments.
inline UnitCubeRad::UnitCubeRad(double x1, double x2, double x3, PdgID idMother, PdgID idSister, PdgID idEmitted, RadiationType radType, double weight)
      : _x1(x1), _x2(x2), _x3(x3), _idMother(idMother), _idSister(idSister), _idEmitted(idEmitted), _radType(radType), _weight(weight)
{
   assert(0 <= x1 && x1<= 1);
   assert(0 <= x2 && x2<= 1);
   assert(0 <= x3 && x3<= 1);
   assert(radType != kUndefined);
   assert(std::isfinite(weight));
}

//------------------------------------------------------------------------------
// Creates a UnitCubeRad from another.
inline UnitCubeRad::UnitCubeRad(const UnitCubeRad& other)
: _x1(other.x1()), _x2(other.x2()), _x3(other. x3()), _idMother(other.idMother()), _idSister(other.idSister()), _idEmitted(other.idEmitted()), _radType(other.radType()),_weight(other.weight())
{
   assert(0 <= _x1 && _x1<= 1);
   assert(0 <= _x2 && _x2<= 1);
   assert(0 <= _x3 && _x3<= 1);
   assert(_radType != kUndefined);
   assert(std::isfinite(_weight));
}

//------------------------------------------------------------------------------
// Read and Write Access
//------------------------------------------------------------------------------
inline double UnitCubeRad::x1() const
{
   return _x1;
}

//------------------------------------------------------------------------------
inline double UnitCubeRad::x2() const
{
   return _x2;
}

//------------------------------------------------------------------------------
inline double UnitCubeRad:: x3() const
{
   return _x3;
}

//------------------------------------------------------------------------------
inline PdgID UnitCubeRad::idMother() const
{
   return _idMother;
}

//------------------------------------------------------------------------------
inline PdgID UnitCubeRad::idSister() const
{
   return _idSister;
}

//------------------------------------------------------------------------------
inline PdgID UnitCubeRad::idEmitted() const
{
   return _idEmitted;
}

//------------------------------------------------------------------------------
inline UnitCubeRad::RadiationType UnitCubeRad::radType() const
{
   return _radType;
}

//------------------------------------------------------------------------------
inline double UnitCubeRad::weight() const
{
   return _weight;
}

//------------------------------------------------------------------------------
inline double& UnitCubeRad::x1()
{
   return _x1;
}

//------------------------------------------------------------------------------
inline double& UnitCubeRad::x2()
{
   return _x2;
}

//------------------------------------------------------------------------------
inline double& UnitCubeRad:: x3()
{
   return _x3;
}

//------------------------------------------------------------------------------
inline PdgID& UnitCubeRad::idMother()
{
   return _idMother;
}

//------------------------------------------------------------------------------
inline PdgID& UnitCubeRad::idSister()
{
   return _idSister;
}

//------------------------------------------------------------------------------
inline PdgID& UnitCubeRad::idEmitted()
{
   return _idEmitted;
}

//------------------------------------------------------------------------------
inline UnitCubeRad::RadiationType& UnitCubeRad::radType()
{
   return _radType;
}

//------------------------------------------------------------------------------
inline double& UnitCubeRad::weight()
{
   return _weight;
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------
/**
 * \brief %UnitCubeRad stream insertion operator
 * \param out The output stream to be used.
 * \param UnitCubeRad The %UnitCubeRad to be streamed.
 * \return A reference to \a out.
 *
 * Streams the info about UnitCubeRad to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const UnitCubeRad& UnitCubeRad)
{
   return UnitCubeRad.print(out);
}

} // namespace Geneva

#endif // GENEVA_UNIT_CUBE_RAD_HPP
