//------------------------------------------------------------------------------
/// \file Exception.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    classes for error and exception handling
//------------------------------------------------------------------------------

#ifndef GENEVA_EXCEPTION_HPP
#define GENEVA_EXCEPTION_HPP

#include <exception>
#include <string>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief The main exception handler used at the outermost layer.
 *
 * By default this should be used in main programs as
 * \code
 * int main(int argc, const char* argv[])
 * {
 *    try {
 *
 *       <main code>
 *
 *    } catch (...) {
 *       return mainExceptionHandler();
 *    }
 *    return EXIT_SUCCESS;
 * }
 * \endcode
 */
int mainExceptionHandler();


//------------------------------------------------------------------------------
/**
 * \class NoError
 * \ingroup Core
 *
 * \brief Special exception type to signal immediate sucessfull exit.
 *
 * This is caught by the mainExceptionHandler to return from main with
 * EXIT_SUCESS.
 *
 * \note This should not be caught by intermediate layers.
 */
//------------------------------------------------------------------------------
struct NoError
{};


//------------------------------------------------------------------------------
/**
 * \class AnyError
 * \ingroup Core
 *
 * \brief Base exception class representing any type of error.
 *
 * This is the root of the exception class hierarchy.
 */
//------------------------------------------------------------------------------
class AnyError : public std::exception
{
   public:
      AnyError(const std::string& what, const std::string& where, bool logged)
         : _what(what), _where(where), _logged(logged) {}

      bool isLogged() const noexcept
      {
         return _logged;
      }

      const char* what() const noexcept override
      {
         return _what.c_str();
      }

      const char* where() const noexcept
      {
         return _where.c_str();
      }

      /// Returns the name of the Error.
      virtual const char* name() const noexcept = 0;

      /**
       * \brief Throws the Error as exception.
       *
       * To retain full type information this should be implemented by derived
       * classes as `{ throw *this; }`.
       */
      virtual void exception() const = 0;

   private:
      std::string _what;
      std::string _where;
      bool _logged;
};


//------------------------------------------------------------------------------
/**
 * \class NormalError
 * \ingroup Core
 *
 * \brief Represents any type of normal (recoverable) error.
 */
//------------------------------------------------------------------------------
class NormalError : public AnyError
{
   public:
      NormalError(const std::string& what, const std::string& where, bool logged)
         : AnyError(what, where, logged)
      {}

      const char* name() const noexcept override { return "Normal Error"; }
      [[noreturn]] void exception() const override { throw *this; }
};


//------------------------------------------------------------------------------
/**
 * \class NumericalError
 * \ingroup Core
 *
 * \brief Represents a numerical error.
 *
 * \note This is an example of a specialized normal (recoverable) error, it is
 * currently unused.
 */
class NumericalError : public NormalError
{
   public:
      NumericalError(const std::string& what, const std::string& where, bool logged)
         : NormalError(what, where, logged)
      {}

      const char* name() const noexcept override { return "Numerical Error"; }
      [[noreturn]] void exception() const override { throw *this; }
};


//------------------------------------------------------------------------------
/**
 * \class FatalError
 * \ingroup Core
 *
 * \brief Represents any type of fatal (unrecoverable) error.
 */
//------------------------------------------------------------------------------
class FatalError : public AnyError
{
   public:
      FatalError(const std::string& what, const std::string& where, bool logged)
         : AnyError(what, where, logged)
      {}

      const char* name() const noexcept override { return "Fatal Error"; }
      [[noreturn]] void exception() const override { throw *this; }
};

} // namespace Geneva

#endif // GENEVA_EXCEPTION_HPP
