//------------------------------------------------------------------------------
/// \file GenerationReport.hpp
//
// Author(s):
//    Calvin Berggren, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class GenerationReport
//------------------------------------------------------------------------------

#ifndef GENEVA_GENERATION_REPORT_HPP
#define GENEVA_GENERATION_REPORT_HPP

#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/Utils/FileUtils.hpp"

#include <sstream>
#include <string>

namespace Geneva
{

namespace tinyxml
{
class TiXmlElement;
}

//------------------------------------------------------------------------------
/**
 * \class GenerationReport
 * \ingroup Core
 *
 * \brief Creates a report of the current generation run, including the options and analyzer output.
 *
 * The produced XML file includes all information of the current generation run,
 * including all information needed to reproduce the same run conditions. The
 * information written includes
 * - version and git modification summary
 * - all options that were set
 * - information about run time, number of events, etc.
 * - any provided output from the run (normally the analyzer output).
 *
 * The format of the written XML file is such that it is a valid option file
 * that can be read back into Geneva to set the same options as used for this
 * run.
 *
 * The run output is provided by streaming it into the \c outputStream()
 * provided by the report, e.g,
 * \code
 *    report.outputStream() << analyzer;
 * \endcode
 * It is included in an XML element called <AnalyzerOutput>.
 */
//------------------------------------------------------------------------------
class GenerationReport
{
   private:
      const OptionMap& _options;       ///< Options to be included in the report
      std::string _pyoptions;          ///< Pythia options file whose content is added to the report
      std::stringstream _output;       ///< Stream which receives all run output

      std::string _title;              ///< Run title
      std::string _description;        ///< Run description
      std::string _executable;         ///< Name of the executable which performed the run

      double _runTime;                 ///< Run time in seconds
      size_t _numEvents;               ///< Number of events generated in the run

   public:
      /**
       * \brief Constructs the GenerationReport from the given arguments.
       *
       * \note This sets the title to getDefaultTitle(options). If a custom
       * title is required, e.g. to add additional tags, it can be set using
       * setTitle().
       */
      GenerationReport(const OptionMap& options, const std::string& executable,
                       const std::string& pythiaOptionsFile = "");

      /// \name Queries
      ///@{
      const std::string& title() const;
      const std::string& description() const;
      const std::string& executable() const;

      double runTimeInSec() const;
      size_t numEventsGenerated() const;

      const char* version() const;
      const char* modifiedFiles() const;
      ///@}

      /// \name Setting information
      ///@{
      // Returns the stream for receiving output.
      std::ostream& outputStream();

      /// Clears all previously received output.
      void clearOutput();

      void setTitle(const std::string& title);
      void setDescription(const std::string& description);
      void setRunTimeInSec(double runTime);
      void setNumEventsGenerated(size_t numGen);
      ///#}

      /// \name Output
      ///@{
      //
      friend std::ostream& operator<<(std::ostream& out, const GenerationReport& report);
      ///#}

      // Provides the default title for the report.
      static std::string getDefaultTitle(const OptionMap& options,
                                         const std::string& extraTag = "", bool addSeed = true);

      // Provides the default filename for the report file.
      static std::string getFilename(const OptionMap& options, const std::string& extraTag,
                                     const std::string& path, bool addSeed = true);
      static std::string getFilename(const OptionMap& options, const std::string& extraTag = "");

   private:
      // Prints the full report in XML format into the given output stream.
      std::ostream& print(std::ostream& out) const;

      void attachSummary(tinyxml::TiXmlElement* node) const;

      void addElement(const std::string& name, const std::string& qualifier,
                      tinyxml::TiXmlElement* top) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Queries
//------------------------------------------------------------------------------

inline const std::string& GenerationReport::title() const       { return _title;       }
inline const std::string& GenerationReport::description() const { return _description; }
inline const std::string& GenerationReport::executable() const  { return _executable;  }

inline double GenerationReport::runTimeInSec() const            { return _runTime;     }
inline size_t GenerationReport::numEventsGenerated() const      { return _numEvents;   }

//------------------------------------------------------------------------------
// Setting information
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Returns the stream for receiving output.
 * \return An \c ostream
 *
 * Any run output to be included in the %GenerationReport can be streamed into
 * the returned \c ostream. The received output will be included in the report
 * in XML element called <AnalyzerOutput>.
 */
inline std::ostream& GenerationReport::outputStream()
{
   return _output;
}

//------------------------------------------------------------------------------
inline void GenerationReport::setTitle(const std::string& title)
{
   _title = title;
}

//------------------------------------------------------------------------------
inline void GenerationReport::setDescription(const std::string& description)
{
   _description = description;
}

//------------------------------------------------------------------------------
inline void GenerationReport::setRunTimeInSec(double runTime)
{
   _runTime = runTime;
}

//------------------------------------------------------------------------------
inline void GenerationReport::setNumEventsGenerated(size_t numEvents)
{
   _numEvents = numEvents;
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------
/**
 * \brief Stream insertion operator
 * \param out The output stream to be used.
 * \param report The %GenerationReport to be streamed.
 * \return A reference to \a out.
 *
 * Prints the full report in human-readable XML format into the given output
 * stream. The output is primarily meant to be saved to a file.
 */
inline std::ostream& operator<<(std::ostream& out, const GenerationReport& report)
{
   return report.print(out);
}

//------------------------------------------------------------------------------
/**
 * \brief Provides the default title for the report.
 * \param options The options to use.
 * \param extraTag Additional tag to append (defaults to "").
 * \param addSeed Whether to append the seed as tag (defaults to `true`).
 * \return The full title.
 *
 * The title is constructed from the \a options as
 * 'fileStem_generateTag_Analyzer::fileTag_extraTag_seed',
 * where the seed is only added if \a addSeed is `true`.
 *
 * \note This is used with default arguments by the constructor to initialize
 * the title.
 */
inline std::string GenerationReport::getDefaultTitle(const OptionMap& options,
                                                     const std::string& extraTag, bool addSeed)
{
   std::string title = options.getOption("fileStem")
                       + format_tag(options.getOption("generateTag"))
                       + format_tag(options.getOption("Analyzer::fileTag"))
                       + format_tag(extraTag);
   if (addSeed)
      title += format_tag(options.getOption("seed"));
   return title;
}

//------------------------------------------------------------------------------
/**
 * \brief Provides the default filename for the report file.
 * \param options The options to use.
 * \param extraTag Additional tag to append (defaults to "").
 * \param path The filepath to use (defaults to filePath::generate).
 * \param addSeed Whether to append the seed as tag (defaults to `true`).
 * \return The full filename including the path and extension.
 *
 * The full filename is effectively constructed by prepending the path and
 * appending the '.xml' extensions to the default report title, resulting in
 * 'path/fileStem_generateTag_Analyzer::fileTag_extraTag_seed.xml',
 * where the seed is only added if \a addSeed is `true`.
 *
 * \note If only one string argument is provided it is used as \a extraTag with
 * the default \a path.
 */
inline std::string GenerationReport::getFilename(const OptionMap& options, const std::string& extraTag,
                                                 const std::string& path, bool addSeed)
{
   return format_path(path)
          + getDefaultTitle(options, extraTag, addSeed)
          + ".xml";
}

//------------------------------------------------------------------------------
inline std::string GenerationReport::getFilename(const OptionMap& options,
                                                 const std::string& extraTag)
{
   return getFilename(options, extraTag, options.getOption("filePath::generate"));
}

} // namespace Geneva

#endif // GENEVA_GENERATION_REPORT_HPP
