//------------------------------------------------------------------------------
/// \file Log.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Log
//------------------------------------------------------------------------------

#ifndef GENEVA_LOG_HPP
#define GENEVA_LOG_HPP

#include "Geneva/Core/UI/Exception.hpp"
#include "Geneva/Core/UI/MessageHandler.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class Log
 * \ingroup Core
 *
 * \brief Global Logger.
 *
 * This class acts as a container for several global static MessageHandlers
 * corresponding to different types of messages, in increasing verbosity:
 * - \c fatal() : Used for fatal errors that should immediately abort the program
 *                (e.g. "disk exploded")
 * - \c error() : Used for errors requiring aborting the current program flow,
 *                but not necessarily the program (generates exceptions that can
 *                be caught higher up)
 *                (e.g. generating an event failed, continue with next event)
 * - \c warning() : Used for warnings, namely abnormal conditions that do not
 *                  require  aborting the current program flow (no exceptions).
 * - \c info()    : Used for (minimal/important) information for the user.
 * - \c verboseInfo() : Used for more detailed information for the power user.
 * - \c debug() : Used only for debugging output.
 *
 * The handlers can be accessed by \c Log::handler(), e.g. to turn them on/off,
 * set individual output streams, colors etc. (though normally these should be
 * set by the main program or driver and not be changed by other classes.)
 *
 * Messages are generated by using the \c Log::handler("where") functions, e.g.
 * \code
 * Log::error("where") << "This is an error message" << Log::end;
 * Log::warning("class foo") << "Number of jets " << nJets << " is too large" << Log::end;
 *
 * Log::info("here") << "Some information prefaced by a header\n"
 * Log::info() << "Some more information without additional header\n"
 * \endcode
 *
 * Messages must be ended as shown by streaming \c Log::end.
 *
 * The \c ErrorHandler is used for \c fatal() and \c error(), which streams
 * the message and then throws a corresponding exception.
 * The \c WarningHandler is used for \c warning(), which streams the message and
 * also counts the number of occurrences.
 * The \c InfoHandler is used for \c info(), \c verboseInfo(), and \c debug(),
 * which just streams the message and can also directly accept streaming input.
 * See the documentation of the handlers for more detail of their use.
 *
 * Several static convenience functions can be used to set properties of all
 * handlers at once: \c setDefaultOutStream,\c setVerbosity, \c setColorsOff,
 * and \c setColorsDefault.
 */
//------------------------------------------------------------------------------
class Log
{
   public:
      /**
       * \brief Named constants to specify verbosity level.
       *
       * These constants are to be used with the \c setVerbosity() function.
       */
      enum Level
      {
         kOff = 0,
         kFatal = 1,
         kError = 2,
         kWarning = 3,
         kInfo = 4,
         kVerboseInfo = 5,
         kDebug = 6
      };

      static const Message::End end;

   private:
      static reference<std::ostream> _default;

      static ErrorHandler _fatalHandler;
      static ErrorHandler _errorHandler;
      static WarningHandler _warningHandler;
      static InfoHandler _infoHandler;
      static InfoHandler _verboseInfoHandler;
      static InfoHandler _debugHandler;

   public:
      static ErrorHandler& fatal();
      static ErrorHandler& error();
      static WarningHandler& warning();
      static InfoHandler& info();
      static InfoHandler& verboseInfo();
      static InfoHandler& debug();

      static Message fatal(const std::string& where);
      static Message error(const std::string& where);
      static Message warning(const std::string& where);
      static Message info(const std::string& where);
      static Message verboseInfo(const std::string& where);
      static Message debug(const std::string& where);

      // Sets the output stream of all handlers.
      static std::ostream& setDefaultOutStream(std::ostream& defaultOut);

      // Turns handlers on/off according to given verbosity level.
      static void setVerbosity(Level level);

      // Turns off all coloring.
      static void setColorsOff();

      // Sets the colors of all of handlers to their default values.
      static void setColorsDefault();

   private:
      Log() = delete;
      Log(const Log&) = delete;
      Log(Log&&) = delete;
      Log& operator=(const Log&) = delete;
      Log& operator=(Log&&) = delete;
      ~Log() = delete;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline ErrorHandler& Log::fatal()      { return _fatalHandler;       }
inline ErrorHandler& Log::error()      { return _errorHandler;       }
inline WarningHandler& Log::warning()  { return _warningHandler;     }
inline InfoHandler& Log::info()        { return _infoHandler;        }
inline InfoHandler& Log::verboseInfo() { return _verboseInfoHandler; }
inline InfoHandler& Log::debug()       { return _debugHandler;       }

//------------------------------------------------------------------------------
/**
 * \brief Generates a fatal error message.
 * \param where A string specifying the origin of the message.
 * \return A Message object.
 *
 * The error message is to be streamed into the returned Message object and must
 * be concluded by \c Log::end.
 */
inline Message Log::fatal(const std::string& where)
{
   return _fatalHandler.startError<FatalError>(where);
}

//------------------------------------------------------------------------------
/**
 * \brief Generates an error message.
 * \param where A string specifying the origin of the message.
 * \return A Message object.
 *
 * The error message is to be streamed into the returned Message object and must
 * be terminated by \c Log::end.
 */
inline Message Log::error(const std::string& where)
{
   return _errorHandler.startError<NormalError>(where);
}

//------------------------------------------------------------------------------
/**
 * \brief Generates a warning message.
 * \param where A string specifying the origin of the message.
 * \return A Message object.
 *
 * The warning message is to be streamed into the returned Message object and
 * must be terminated by \c Log::end.
 */
inline Message Log::warning(const std::string& where)
{
   return _warningHandler.startMessage(where);
}

//------------------------------------------------------------------------------
/**
 * \brief Generates an info message.
 * \param where A string specifying the origin of the message.
 * \return A Message object.
 *
 * The info message is to be streamed into the returned Message object. This is
 * to be used for minimal and important information for the user. For more
 * detailed verbose information use \c verboseInfo().
 *
 * \note Normally info messages do not require an explicit header and can be
 * directly streamed into the \c info() handler.
 */
inline Message Log::info(const std::string& where)
{
   return _infoHandler.startMessage(where);
}

//------------------------------------------------------------------------------
/**
 * \brief Generates a detailed info message.
 * \param where A string specifying the origin of the message.
 * \return A Message object.
 *
 * The info message is to be streamed into the returned Message object. This is
 * to be used for detailed and verbose information for the user. For minimal and
 * important information use \c info().
 *
 * \note Normally info messages do not require an explicit header and can be
 * directly streamed into the \c verboseInfo() handler.
 */
inline Message Log::verboseInfo(const std::string& where)
{
   return _verboseInfoHandler.startMessage(where);
}

//------------------------------------------------------------------------------
/**
 * \brief Generates a debug message.
 * \param where A string specifying the origin of the message.
 * \return A Message object.
 *
 * The debug message is to be streamed into the returned Message object. This is
 * to be used for debugging information only and not for normal production use.
 * For user information relevant during normal production use \c info() or
 * \c verboseInfo(), for warnings/errors use one of \c warning(), \c error(), or
 * \c fatal().
 *
 * \note Debug messages not requiring an explicit header and origin information
 * can also be directly streamed into the \c debug() handler.
 */
inline Message Log::debug(const std::string& where)
{
   return _debugHandler.startMessage(where);
}

} // namespace Geneva

#endif // GENEVA_LOG_HPP
