//------------------------------------------------------------------------------
/// \file MessageHandler.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interfaces of MessageHandler classes
//------------------------------------------------------------------------------

#ifndef GENEVA_MESSAGE_HANDLER_HPP
#define GENEVA_MESSAGE_HANDLER_HPP

#include "Geneva/Core/UI/Exception.hpp"
#include "Geneva/Core/UI/TextColor.hpp"
#include "Geneva/Core/Utils/fixed_ptr.hpp"

#include <stdexcept>
#include <sstream>
#include <map>

namespace Geneva
{

class MessageHandlerBase;
class InfoHandler;
class ErrorHandler;
class WarningHandler;

//------------------------------------------------------------------------------
/**
 * \class Message
 * \ingroup Core
 *
 * \brief Represents a streamed error, warning, or info message.
 *
 * This basically acts as a wrapper for an output stream. Objects of type
 * %Message are provided by the \c MessageHandler classes into which the user
 * then streams error, warning, info, etc. messages.
 *
 * \c std::endl and \c std::flush cannot be streamed into a %Message. Instead
 * the end of a message is signaled by streaming the \c Message::End object,
 * resulting in a call to the MessageHandler's \c endMessage() function, which
 * in turn takes the appropriate actions like flushing etc.
 *
 * \note Only the \c MessageHandler classes can construct a new %Message, and
 * users should not have to do so themselves. Normal operation requires users
 * to be able to copy construct Messages. However, users should not store
 * Message objects for later use. This is not supported and will interfere with
 * the message handling.
 */
//------------------------------------------------------------------------------
class Message
{
   public:
      /**
       * \struct Message::End
       * \brief Empty helper object signaling the end of a message.
       */
      struct End {};

      Message(const Message&) = default;
      Message(Message&&) = default;
      ~Message() = default;

      /// Stream insertion operator.
      template <class T>
      Message& operator<<(const T& message);

      /// Stream insertion for end of message object.
      void operator<<(Message::End);

      /**
       * \brief Stream insertion operator to catch \c std::endl etc.
       *
       * This is intentionally only declared but not defined, which produces a
       * linker error when trying to stream \c std::endl into a %Message (which
       * tends to give much less cryptic error messages than with the compiler
       * due to the heavy template use in the stdlib).
       */
      void operator<<(std::ostream& undefined(std::ostream&));

   protected:
      friend class MessageHandlerBase;
      friend class InfoHandler;
      friend class WarningHandler;
      friend class ErrorHandler;

      Message();
      Message(std::ostream& out, MessageHandlerBase* handler);

      // disable assignment
      Message& operator=(const Message&) = delete;

      std::ostream& _out;
      MessageHandlerBase* _handler;

      std::ostream& null();
};


//------------------------------------------------------------------------------
/**
 * \class MessageHandlerBase
 * \ingroup Core
 *
 * \brief Base class with basic functionality for MessageHandlers
 *
 * A message is started by calling \c startMessage supplying the origin of the
 * message as an argument. If the handler is turn on, this prints a header with
 * the handler's tag and the message's origin. It returns a \c Message object
 * into which additional information can be streamed. The end of the message
 * is signaled by streaming the \c Message::End object into the \c Message,
 * which results in a call to the handler's \c endMessage() function.
 *
 * If the handler is turned off, then no header is printed and all input
 * streamed into the \c Message object returned by \c startMessage is discarded.
 *
 * Derived classes can redefine the \c startMessage() and \c endMessage()
 * functions to implement special behaviour.
 */
//------------------------------------------------------------------------------
class MessageHandlerBase
{
   protected:
      reference<std::ostream> _out;
      bool _isOn;

      std::string _tag;
      TextColor _color;

   public:
      MessageHandlerBase(std::ostream& out, bool isOn, const std::string& tag,
                         TextColor color);

      /// \name Queries and Accessors
      ///@{

      /// Returns a reference to the output stream messages are streamed to.
      std::ostream& outStream() const;

      /// Returns \c true if the handler is turned on, \c false otherwise.
      bool isOn() const;

      /// Returns the message tag used in message headers.
      const std::string& tag() const;

      /// Returns the TextColor used for message headers.
      TextColor color() const;

      // Sets the output stream messages are streamed to.
      std::ostream& setOutStream(std::ostream& out);

      /// Turns the handler on.
      void turnOn();

      /// Turns the handler off.
      void turnOff();

      /// Sets the message tag used in message headers.
      void setTag(const std::string& tag);

      /// Sets the TextColor used for message headers.
      void setColor(TextColor color);
      ///@}

      /// \name Messaging Interface
      ///@{
      // Starts a message.
      virtual Message startMessage(const std::string& where);

      // Concludes a message.
      virtual void endMessage();
      ///@}
};


//------------------------------------------------------------------------------
/**
 * \class InfoHandler
 * \ingroup Core
 *
 * \brief Handler for information messages.
 *
 * The handler uses the \c MessageHandlerBase for all message handling. When
 * turned on it directly forwards any streamed messages to the output stream.
 * When turned off all output is discarded.
 *
 * In addition, the InfoHandler also allows to directly stream into it, without
 * going through \c startMessage().
 */
//------------------------------------------------------------------------------
class InfoHandler : public MessageHandlerBase
{
   public:
      InfoHandler(std::ostream& out, bool isOn, const std::string& tag,
                  TextColor color = TextColor::kOff);

      // Stream insertion operator.
      template <class T>
      Message operator<<(const T& message);
};


//------------------------------------------------------------------------------
/**
 * \class WarningHandler
 * \ingroup Core
 *
 * \brief Handler for warning messages.
 *
 * The basic behavior is similar to \c MessageHandlerBase.
 *
 * Warning messages are generated by an explicit call to \c startMessage(). They
 * must be concluded by streaming the \c Message::End object (or by an explicit
 * call to \c endMessage()). Once an error message is concluded, the handler
 * increases the counter for the particular warning. If the counter passes a
 * threshold a fatal error is thrown with its \c what member initialized with
 * the warning message.
 *
 * \note The fatal error is always thrown after too many warnings,
 * the on/off state of the handler only affects whether error messages
 * are also sent to the output stream before throwing.
 * See the documentation of \c startMessage() and
 * \c endMessage() for more details.
 */
//------------------------------------------------------------------------------
class WarningHandler : public MessageHandlerBase
{
   public:
      WarningHandler(std::ostream& out, bool isOn, size_t maxCounts,
                     const std::string& tag, TextColor color = TextColor::kOff);

      ~WarningHandler();

      // Starts a warning message.
      virtual Message startMessage(const std::string& where);

      // Concludes a warning message.
      virtual void endMessage();

      /// Returns the current maximum allowed number of warnings.
      size_t maxCounts() const;

      /// Sets the maximum allowed number of warnings.
      void setMaxCounts(size_t maxCounts);

      /// Resets all counters.
      void resetCounts();

      /// Writes a summary of the total number of warnings received so far.
      void writeCounts();

   private:
      std::stringstream _what;

      typedef std::map<const std::string, size_t> _impl_type;
      _impl_type _counts;
      _impl_type::iterator _current;

      size_t _maxCounts;
};

//------------------------------------------------------------------------------
/**
 * \class ErrorHandler
 * \ingroup Core
 *
 * \brief Handler for error messages.
 *
 * The basic behavior is similar to \c MessageHandlerBase.
 *
 * Error messages are generated by an explicit call to \c startError(). They
 * must be concluded by streaming the \c Message::End object (or by an explicit
 * call to \c endMessage()). Once an error message is concluded, the handler
 * throws an exception containing the received message.
 *
 * \note The exception is always thrown, the on/off state of the handler only
 * affects whether error messages are also sent to the output stream before
 * throwing. See the documentation of \c startError() and \c endMessage() for
 * more details.
 */
//------------------------------------------------------------------------------
class ErrorHandler : public MessageHandlerBase
{
   protected:
      std::stringstream _what;
      std::string _where;

      void (*_thrower)(const std::string&, const std::string&, bool);

   public:
      ErrorHandler(std::ostream& out, bool isOn, const std::string& tag,
                   TextColor color = TextColor::kOff);

      // Starts an error message of type ErrorType
      template <class ErrorType>
      Message startError(const std::string& where)
      {
         _thrower = &_throwError<ErrorType>;
         return startMessage(where);
      }

      // Concludes an error message and throws an exception.
      virtual void endMessage();

   private:
      // Starts an error message.
      virtual Message startMessage(const std::string& where);

      template <class ErrorType>
      static void _throwError(const std::string& what, const std::string& where, bool logged)
      {
         throw ErrorType(what, where, logged);
      }
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Message
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
template <class T>
Message& Message::operator<<(const T& message)
{
   _out << message;
   return *this;
}

//------------------------------------------------------------------------------
inline void Message::operator<<(Message::End)
{
   if (_handler)
      _handler->endMessage();
}

//------------------------------------------------------------------------------
inline Message::Message()
   : _out(null()), _handler(0)
{}

//------------------------------------------------------------------------------
inline Message::Message(std::ostream& out, MessageHandlerBase* handler)
   : _out(out), _handler(handler)
{}

//------------------------------------------------------------------------------
// MessageHandlerBase
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
inline MessageHandlerBase::MessageHandlerBase(std::ostream& out, bool isOn,
                                              const std::string& tag, TextColor color)
   : _out(out), _isOn(isOn), _tag(tag), _color(color)
{}

//------------------------------------------------------------------------------
inline std::ostream& MessageHandlerBase::outStream() const { return _out;   }
inline bool MessageHandlerBase::isOn() const               { return _isOn;  }
inline const std::string& MessageHandlerBase::tag() const  { return _tag;   }
inline TextColor MessageHandlerBase::color() const         { return _color; }

//------------------------------------------------------------------------------
/**
 * \brief Sets the output stream messages are streamed to.
 * \param out The new output stream to use.
 * \return A reference to the currently used output stream.
 */
inline std::ostream& MessageHandlerBase::setOutStream(std::ostream& out)
{
   std::ostream& tmp(_out);
   _out.set(out);
   return tmp;
}

//------------------------------------------------------------------------------
inline void MessageHandlerBase::turnOn()
{
   _isOn = true;
}

//------------------------------------------------------------------------------
inline void MessageHandlerBase::turnOff()
{
   _isOn = false;
}

//------------------------------------------------------------------------------
inline void MessageHandlerBase::setTag(const std::string& tag)
{
   _tag = tag;
}

//------------------------------------------------------------------------------
inline void MessageHandlerBase::setColor(TextColor color)
{
   _color = color;
}

//------------------------------------------------------------------------------
// InfoHandler
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
inline InfoHandler::InfoHandler(std::ostream& out, bool isOn,
                                const std::string& tag, TextColor color)
   : MessageHandlerBase(out, isOn, tag, color)
{}

//------------------------------------------------------------------------------
/**
 * \brief Stream insertion operator.
 * \param message The information to be streamed.
 * \return A \c Message object for further streaming.
 *
 * In addition to using \c startMessage(), %InfoHandler also allows directly
 * streaming into it, which generates a message without header.
 */
template <class T>
Message InfoHandler::operator<<(const T& message)
{
   if (isOn()) {
      _out.get() << message;
      return Message(_out, this);
   }
   return Message();
}

//------------------------------------------------------------------------------
// WarningHandler
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
inline WarningHandler::WarningHandler(std::ostream& out, bool isOn, size_t maxCounts,
                                      const std::string& tag, TextColor color)
   : MessageHandlerBase(out, isOn, tag, color),
     _counts(), _current(_counts.end()), _maxCounts(maxCounts)
{}

//------------------------------------------------------------------------------
inline WarningHandler::~WarningHandler()
{
   writeCounts();
}

//------------------------------------------------------------------------------
inline size_t WarningHandler::maxCounts() const
{
   return _maxCounts;
}

//------------------------------------------------------------------------------
inline void WarningHandler::setMaxCounts(size_t maxCounts)
{
   _maxCounts = maxCounts;
}

//------------------------------------------------------------------------------
inline void WarningHandler::resetCounts()
{
   _counts.clear();
}

//------------------------------------------------------------------------------
// ErrorHandler
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
inline ErrorHandler::ErrorHandler(std::ostream& out, bool isOn,
                                  const std::string& tag, TextColor color)
   : MessageHandlerBase(out, isOn, tag, color), _thrower(0)
{}

} // namespace Geneva

#endif // GENEVA_MESSAGE_HANDLER_HPP
