//------------------------------------------------------------------------------
/// \file Option.hpp
//
// Author(s):
//    Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Option
//------------------------------------------------------------------------------

#ifndef GENEVA_OPTION_HPP
#define GENEVA_OPTION_HPP

#include <vector>
#include <string>
#include <iosfwd>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class OptionBase
 * \ingroup Core
 *
 * \brief Option class, essential a <name, value> pair, where value can be any type
 */
//------------------------------------------------------------------------------
class OptionBase
{
   protected:
      std::string _name;        ///< short name for this option, eg, "CenterOfMassEnergy"
      std::string _description; ///< longer description, eg, "Center of mass energy for the collision"
      bool _changed;            ///< whether this option has been changed from its default value

      OptionBase(std::string n, std::string d);

   public:
      virtual ~OptionBase() = default;

      virtual void setValue(std::string) = 0; ///< parse value from a string

      std::string         name()               const {return _name;}
      std::string         description()        const {return _description;}
      virtual std::string getValueStr()        const = 0;
      virtual std::string getDefaultValueStr() const = 0;
      bool                changed()            const {return _changed;}

      virtual OptionBase* clone() const = 0;

      /// Print option to a stream (polymorphically via print()).  Just prints value, not name.
      friend std::ostream& operator<<(std::ostream& os, const OptionBase& opt);
      virtual std::ostream& print(std::ostream& os) const = 0;

   protected:
      OptionBase(const OptionBase&) = default;
      OptionBase(OptionBase&&) = default;
      OptionBase& operator=(const OptionBase&) = default;
      OptionBase& operator=(OptionBase&&) = default;

      /// Read a T from val -- specialized for many types below
      /*
       * Used by Option<T>; put here so I don't have to specialize all of
       * Option<T> for every T, just this function.  (E.g., to specialize for bool
       * would require specializing the class Option<bool>.)
       */
      template <class T>
      T _parseString(std::string val) const;
};

//------------------------------------------------------------------------------
/**
 * \class Option
 * \ingroup Core
 *
 * \brief Template class holding a value of type T, as well as a default value
 */
template <class T>
class Option : public OptionBase
{
   private:
      T _value;        ///< the current value of this option
      T _defaultValue; ///< the default value of this option

   public:
      Option(std::string name, std::string val, std::string desc="");

      virtual std::string getValueStr()        const;
      virtual std::string getDefaultValueStr() const;

      void setValue(std::string val) { _value = _parseString<T>(val); _changed = (val!=getDefaultValueStr()); }
      T value()        const         {return _value;}
      T defaultValue() const         {return _defaultValue;}

      virtual OptionBase* clone() const { return new Option<T>(*this); }
      virtual std::ostream& print(std::ostream& os) const;
};

// partial specialization for vector<T>
template <class T>
class Option<std::vector<T> > : public OptionBase
{
   private:
      std::vector<T> _value;        ///< the current value of this option
      std::vector<T> _defaultValue; ///< the default value of this option

   public:
      Option(std::string name, std::string val, std::string desc="");

      virtual std::string getValueStr()        const;
      virtual std::string getDefaultValueStr() const;

      void setValue(std::string val) { _value = _parseStringVector(val); _changed = (val!=getDefaultValueStr()); }
      std::vector<T> value()        const         {return _value;}
      std::vector<T> defaultValue() const         {return _defaultValue;}

      virtual OptionBase* clone() const { return new Option<std::vector<T> >(*this); }
      virtual std::ostream& print(std::ostream& os) const;

   protected:
      /// Read a vector of T's from a string by using sstream >> t repeatedly
      std::vector<T> _parseStringVector(std::string val) const;
};

} // namespace Geneva


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

#include "Geneva/Core/UI/Log.hpp"

#include <sstream>
#include <ostream>
#include <iomanip>
#include <cstddef>

namespace Geneva
{

//------------------------------------------------------------------------------
inline OptionBase::OptionBase(std::string n, std::string d)
   : _name(n), _description(d), _changed(false)
{}


//------------------------------------------------------------------------------
inline std::ostream& operator<<(std::ostream& os, const OptionBase& opt)
{
   return opt.print(os);
}

//------------------------------------------------------------------------------
/**
 * \brief Generic implementation of _parseString
 *
 * Will work for string, char, int, double
 * Specializations for bool below
 */
template <class T>
inline T OptionBase::_parseString(std::string val) const
{
   if (val.empty())
      Log::fatal("OptionBase::_parseString") << "Trying to parse an empty string in " << _name << Log::end;

   std::stringstream ss(val);
   T t;
   ss >> t;
   if (ss.fail())
      Log::fatal("OptionBase::_parseString") << "Could not convert \'" << val << "\' to the appropriate type in " << _name << Log::end;
   return t;
}

//------------------------------------------------------------------------------
/// Specialization of _parseString to string which allows to parse empty strings avoiding the checks
/// NOTE: this is needed because some of the string options would allow for a default empty argument, e.g. optionsFile
template <>
inline std::string OptionBase::_parseString<std::string>(std::string val) const
{
  return val;
}

//------------------------------------------------------------------------------
/// Specialization of _parseString to bool which you can't just sstream >> t
template <>
inline bool OptionBase::_parseString<bool>(std::string val) const
{
   if (!val.empty()) {
      if (val == "T" || val == "t" || val == "true" || val == "True" || val == "on" || val == "ON")
         return true;
      else if (val == "F" || val == "f" || val == "False" || val == "false" || val == "off" || val == "OFF")
         return false;
      Log::fatal("OptionBase::_parseString<bool>") << "Bool option in " << _name << " is not amongst {T,t,true,True,on,ON,F,f,False,false,off,OFF}" << Log::end;
      return false;
   } else {
      Log::fatal("OptionBase::_parseString<bool>") << "Trying to parse an empty string in " << _name << Log::end;
      return false;
   }
}

//------------------------------------------------------------------------------
// Read a vector of T's from a string by using sstream >> t repeatedly
/// This version is used by Option<std::vector<T> > instead of _parseString(val)
template<class T>
inline std::vector<T> Option<std::vector<T> >::_parseStringVector(std::string val) const
{
   std::stringstream ss(val);
   std::vector<T> vec;
   std::string t;
   while (ss.good()) {
      ss >> t;
      vec.push_back(_parseString<T>(t));
   }
   return vec;
}

//------------------------------------------------------------------------------
template <class T>
inline Option<T>::Option(std::string name, std::string val, std::string desc)
   : OptionBase(name, desc), _value(_parseString<T>(val)), _defaultValue(_parseString<T>(val))
{}

//------------------------------------------------------------------------------
template <class T>
inline Option<std::vector<T> >::Option(std::string name, std::string val, std::string desc)
   : OptionBase(name, desc), _value(_parseStringVector(val)), _defaultValue(_parseStringVector(val))
{}

//------------------------------------------------------------------------------
template <class T>
inline std::string Option<T>::getValueStr() const
{
   std::stringstream ss;
   ss << std::boolalpha << std::setprecision(20) << _value;
   return ss.str();
}

//------------------------------------------------------------------------------
template <class T>
inline std::string Option<T>::getDefaultValueStr() const
{
   std::stringstream ss;
   ss << std::boolalpha << std::setprecision(20) << _defaultValue;
   return ss.str();
}

//------------------------------------------------------------------------------
template <class T>
inline std::ostream& Option<T>::print(std::ostream& os) const
{
   os << std::boolalpha << std::setprecision(20) << _value << " (" << _defaultValue << ")" << std::noboolalpha;
   return os;
}

//------------------------------------------------------------------------------
template <class T>
inline std::string Option<std::vector<T> >::getValueStr() const
{
   std::stringstream ss;
   ss << std::boolalpha << std::setprecision(20);
   for (size_t i = 0; i < _value.size(); ++i) {
      if (i > 0) ss << " ";
      ss << _value[i];
   }
   return ss.str();
}

//------------------------------------------------------------------------------
template <class T>
inline std::string Option<std::vector<T> >::getDefaultValueStr() const
{
   std::stringstream ss;
   ss << std::boolalpha << std::setprecision(20);
   for (size_t i = 0; i < _defaultValue.size(); ++i) {
      if (i > 0) ss << " ";
      ss << _defaultValue[i];
   }
   return ss.str();
}

//------------------------------------------------------------------------------
template <class T>
inline std::ostream& Option<std::vector<T> >::print(std::ostream& os) const
{
   for (size_t i = 0; i < _value.size(); ++i) {
      if (i > 0) os << " ";
      os << std::boolalpha << _value[i];
   }
   os << " (";
   for (size_t i = 0; i < _defaultValue.size(); ++i) {
      if (i > 0) os << " ";
      os << _defaultValue[i];
   }
   os << ")" << std::noboolalpha;
   return os;
}

} // namespace Geneva

#endif // GENEVA_OPTION_HPP
