//------------------------------------------------------------------------------
/// \file OptionMap.hpp
//
// Author(s):
//    Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definition of class OptionMap, as well as OptionBase and Option<T>
//------------------------------------------------------------------------------

#ifndef GENEVA_OPTION_MAP_HPP
#define GENEVA_OPTION_MAP_HPP

#include "Geneva/Core/UI/Option.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include <map>
#include <iosfwd>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class OptionMap
 * \ingroup Core
 *
 * \brief Class to hold options of any type for use by the rest of the code
 */
//------------------------------------------------------------------------------
class OptionMap
{
   private:
      typedef std::map<std::string, OptionBase*> _impl_type;

      _impl_type _optionMap; ///< The container for options

   public:
      typedef _impl_type::iterator iterator;
      typedef _impl_type::const_iterator const_iterator;

      OptionMap() {};

      ~OptionMap();

      // OptionMap copy constructor
      OptionMap(const OptionMap& other);

      // OptionMap assignment operator
      OptionMap& operator=(const OptionMap& rhs);

      // Iterator for options
      /// Returns an iterator pointing to the first Option in the map.
      const_iterator begin() const;

      /// Returns an iterator pointing beyond the last Option in the map.
      const_iterator end() const;

      /// Get the value of an option in string form. The type does not need to be known. Throws an exception if not set
      std::string getOption(const std::string& name) const;

      /// Get the value of an option.  Throws an exception if not set
      template <class T>
      T getOption(const std::string& name) const;

      /// Get the default value of an option.
      template <class T>
      T getOptionDefault(const std::string& name) const;

      /// Whether this option has been changed from default
      bool isDefault(const std::string& name) const;

      /// Whether this option exists in the current list
      bool optionExists(const std::string& name) const;

      /// Creates an option with a default value.
      template <class T>
      void createOption(const std::string& name, const std::string& val, const std::string& description = "");

      /// Creates options with default values by reading from a parser.
      template <class T>
      void createDefaultOptions(T& parser);

      /// Set an option that has already been created. Throws an exception if option does not exist.
      void setOption(const std::string& name, const std::string& val);

      /// Sets options by reading from a parser.
      template <class T>
      void setUserOptions(T& parser, bool enforceExists);

      /// Sets options from a given OptionMap
      void setUserOptions(const OptionMap& options, bool enforceExists);

      /// User-friendly printout of available options with default values
      void printOptions() const;

      /// Print all options to a stream
      friend std::ostream& operator<<(std::ostream& output, const OptionMap& o);
};

} // namespace Geneva


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

#include <utility>

namespace Geneva
{

//------------------------------------------------------------------------------
inline OptionMap::OptionMap(const OptionMap& other)
{
   for (const_iterator it = other.begin(); it != other.end(); ++it)
      _optionMap.insert(_optionMap.end(), std::make_pair(it->first, it->second->clone()));
}

//------------------------------------------------------------------------------
inline OptionMap::~OptionMap()
{
   for (const_iterator it = begin(); it != end(); ++it)
      delete it->second;
}

//------------------------------------------------------------------------------
inline OptionMap& OptionMap::operator=(const OptionMap& rhs)
{
   OptionMap tmp(rhs);
   _optionMap.swap(tmp._optionMap);
   return *this;
}

//------------------------------------------------------------------------------
/**
 * If the %OptionMap is empty the returned iterator will be equal to end(). It
 * is the user's responsibility to ensure that the iterator is not used for
 * reading in this case.
 */
inline OptionMap::const_iterator OptionMap::begin() const
{
   return _optionMap.begin();
}

//------------------------------------------------------------------------------
/**
 * The returned iterator points one step beyond the last Option and so should not be
 * used for reading. Instead, it is usually used to check that all Options have been
 * traversed.
 */
inline OptionMap::const_iterator OptionMap::end() const
{
   return _optionMap.end();
}

//------------------------------------------------------------------------------
template <class T>
void OptionMap::createOption(const std::string& name, const std::string& value, const std::string& description)
{
   const_iterator it = _optionMap.find(name);
   if (it != _optionMap.end())
      Log::error("OptionMap::createOption") << "Could not create already existing option \'" << name << "\'." << Log::end;

   OptionBase* opt = new Option<T>(name, value, description);
   _optionMap.insert(std::make_pair(name, opt));
}

//------------------------------------------------------------------------------
template <class T>
void OptionMap::createDefaultOptions(T& parser)
{
   while (parser.hasNext()) {
      OptionBase* opt = parser.getNext();
      // check that it doesn't exist yet
      iterator it = _optionMap.find(opt->name());
      if (it != _optionMap.end()) {
         delete it->second;
         _optionMap.erase(it);
      }
      _optionMap.insert(std::make_pair(opt->name(), opt));
   }
}

//------------------------------------------------------------------------------
template <class T>
void OptionMap::setUserOptions(T& parser, bool enforceExists)
{
   while (parser.hasNext()) {
      std::string name, val;
      parser.getNext(name, val);
      if (optionExists(name))
         setOption(name, val);
      else {
         if (enforceExists)
            Log::error("OptionMap::setUserOptions") << "Could not set unknown option \'"<< name << "\'." << Log::end;
         else
            Log::warning("OptionMap::setUserOptions") << "Ignoring unknown option \'"<< name << "\'." << Log::end;
      }
   }
}

//------------------------------------------------------------------------------
/// Get the value of an option.  Throws an exception if not set
template <class T>
T OptionMap::getOption(const std::string& name) const
{
   const_iterator it = _optionMap.find(name);
   if (it == _optionMap.end())
      Log::error("OptionMap::getOption") << "Could not retrieve value of unknown option \'" << name << "\'." << Log::end;

   const Option<T>* opt = dynamic_cast<const Option<T>*>(it->second);
   if (opt == 0)
      Log::error("OptionMap::getOption") << "Existing option \'" << name << "\' does not have the required type." << Log::end;
   return opt->value();
}

//------------------------------------------------------------------------------
/// Get the default value of an option.
template <class T>
T OptionMap::getOptionDefault(const std::string& name) const
{
   const_iterator it = _optionMap.find(name);
   if (it == _optionMap.end())
      Log::error("OptionMap::getOptionDefault") << "Could not retrieve default value of unknown option \'" << name << "\'." << Log::end;

   const Option<T>* opt = dynamic_cast<const Option<T>*>(it->second);
   if (opt == 0)
      Log::error("OptionMap::getOptionDefault") << "Existing option \'" << name << "\' does not have the required type." << Log::end;
   return opt->defaultValue();
}

} //namespace Geneva

#endif // GENEVA_OPTION_MAP_HPP
