//------------------------------------------------------------------------------
/// \file OptionParserXML.hpp
//
// Author(s):
//    Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class OptionParserXML
//------------------------------------------------------------------------------

#ifndef GENEVA_OPTION_PARSER_XML_HPP
#define GENEVA_OPTION_PARSER_XML_HPP

#include <string>
#include <vector>

namespace Geneva
{

namespace tinyxml
{
class TiXmlDocument;
class TiXmlElement;
}

class OptionBase;

//------------------------------------------------------------------------------
/**
 * \class OptionParserXML
 * \ingroup Core
 *
 * \brief Parser for reading options from an XML file
 */
//------------------------------------------------------------------------------
class OptionParserXML
{
   private:
      std::string _fileName;           ///< name of the xml file

   public:
      /// Loads the default xml option file
      OptionParserXML();

      /// Loads the xml option file \a fileName
      OptionParserXML(const std::string& fileName);

      ~OptionParserXML();

      /// Return \c true if there are more options to retrieve.
      bool hasNext() const { return !_elementStack.empty(); }

      /**
       * \brief Gets the next option.
       * \return A new OptionBase object.
       *
       * Parses the next option including type and default values and creates
       * a complete OptionBase object from it. The caller owns the object.
       *
       * This is used for inserting new options into an OptionMap.
       */
      OptionBase* getNext();

      /**
       * \brief Gets the next option.
       * \param[out] name name of the option
       * \param[out] value value of the option
       *
       * Parses the next option and returns its name and value as strings in
       * \a name and \a value.
       *
       * This is used for setting the value of existing options in an OptionMap.
       */
      void getNext(std::string& name, std::string& value);

      /// Returns the name of the xml file being read.
      const std::string& fileName() { return _fileName; }

   private:
      tinyxml::TiXmlDocument* _document;
      std::vector<tinyxml::TiXmlElement*> _elementStack;

      void _loadFile();
      void _advanceToNextElement();
      std::string _getScopedElementName();
};

} // namespace Geneva

#endif // GENEVA_OPTION_PARSER_XML_HPP
