//------------------------------------------------------------------------------
/// \file TextColor.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class TextColor
//------------------------------------------------------------------------------

#ifndef GENEVA_TEXT_COLOR_HPP
#define GENEVA_TEXT_COLOR_HPP

#include <iosfwd>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class TextColor
 * \ingroup Core
 *
 * \brief Lightweight class representing the color of text.
 *
 * The TextColor can only be initialized from the named color codes defined in
 * \c TextColor::ColorCode.
 *
 * The color of an output stream can be set by using the \c set() member
 * function or by streaming a TextColor object or a raw \c TextColor::ColorCode.
 * The color attributes can be reset using \c TextColor::kReset color code or
 * by calling the \c reset() member function.
 *
 * The class uses ANSI escape sequences to set the color. This only makes sense
 * for streams which will eventually go to output devices that understand these
 * (most screen terminals do). The special code \c TextColor::kOff can be used
 * to suppress all color handling and to avoid cluttering the output with color
 * codes for output devices that don't understand colors.
 *
 * \note The class has the size of an int and so should be passed by value.
 */
//------------------------------------------------------------------------------
class TextColor
{
   public:
      /**
       * \brief Named constants for specific text colors.
       */
      enum ColorCode
      {
         /// Resets all previously set text attributes
         kReset = -1,

         /// Does nothing
         kOff = 0,

         /// Specify bold or bright text
         kBold = 100,

         // Explicit colors
         kBlack = '0',       kDarkGray = kBlack + kBold,
         kRed = '1',         kBoldRed = kRed + kBold,
         kGreen = '2',       kBoldGreen = kGreen + kBold,
         kBrown = '3',       kYellow = kBrown + kBold,
         kBlue = '4',        kBoldBlue = kBlue + kBold,
         kMagenta = '5',     kBoldMagenta = kMagenta + kBold,
         kCyan = '6',        kBoldCyan = kCyan + kBold,
         kLightGray = '7',   kWhite = kLightGray + kBold
      };

   private:
      int _code;

   public:
      // Default constructor.
      TextColor();

      // Creates a TextColor from a given ColorCode.
      TextColor(ColorCode color);

      // Sets the color of the given output stream.
      std::ostream& set(std::ostream& out);

      // Resets the color of the given output stream.
      std::ostream& reset(std::ostream& out);
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
/**
 * \brief Default constructor.
 *
 * Initializes the TextColor to \c TextColor::kOff.
 */
inline TextColor::TextColor()
   : _code(kOff)
{}

//------------------------------------------------------------------------------
/**
 * \brief Creates a TextColor from a given ColorCode.
 * \param color The TextColor::ColorCode to use.
 */
inline TextColor::TextColor(ColorCode color)
   : _code(color)
{}

//------------------------------------------------------------------------------
/**
 * \name Stream insertion operator for TextColor.
 * \param out The output stream.
 * \param color The TextColor to set.
 * \return A reference to \a out.
 *
 * Set the color of \a out by streaming a TextColor object to it:
 * \code out << color; \endcode
 */
inline std::ostream& operator<<(std::ostream& out, TextColor color)
{
   return color.set(out);
}

//------------------------------------------------------------------------------
/**
 * \name Stream insertion operator for TextColor::ColorCode.
 * \param out The output stream.
 * \param colorCode The color code to set.
 * \return A reference to \a out.
 *
 * Set the color of \a out by streaming a TextColor::ColorCode to it:
 * \code out << TextColor::kRed; \endcode
 * It sets the color of \a out to the TextColor constructed from \a colorCode.
 */
inline std::ostream& operator<<(std::ostream& out, TextColor::ColorCode colorCode)
{
   return TextColor(colorCode).set(out);
}

} // namespace Geneva

#endif // GENEVA_TEXT_COLOR_HPP
