//------------------------------------------------------------------------------
/// \file Timing.hpp
//
// Author(s):
//    Christian Bauer, Frank Tackmann, Jesse Thaler, Simone Alioli
//
// Copyright:
//    Copyright (C) 2017, 2018 Geneva Collaboration
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Timing
//------------------------------------------------------------------------------

#ifndef GENEVA_TIMING_HPP
#define GENEVA_TIMING_HPP

#include <string>
#include <chrono>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 * \brief A timing class that keeps track of the walltime spent on generating events
 *
 * A class that keeps track of the timing information of the event generation
 * (or any loop for that matter)
 *
 * It is used in the following way:
 * 1) Before using it set the targetValue to the total number of events being generated
 *    and the currentValue to zero.
 * 2) At the start of the event loop, start the timer
 * 3) For each new event, update the currentValue
 * 4) At certain times (say for each 1000 events) print the progressString to the screen
 * 5) After the event loop, stop the timer
 * 6) One can then print out the timeElapsedString
 */
//------------------------------------------------------------------------------
class Timing
{
   private:
      std::chrono::time_point<std::chrono::system_clock> _start; ///< Time when event loop started
      std::chrono::time_point<std::chrono::system_clock> _stop;  ///< Time when event loop finished
      bool _isRunning {false};         ///< True if the timer is running

      size_t _targetValue {1};         ///< The value the timer will take when it will be turned off
      size_t _currentValue {0};        ///< The current value

   public:
      void start();                    ///< Starts the timer
      void stop();                     ///< Stops the timer

      void setTargetValue(size_t);     ///< Sets the total number of events to be generated
      void setCurrentValue(size_t);    ///< Updates the current number of events

      size_t targetValue() const;      ///< Returns the total number of events
      size_t currentValue() const;     ///< Returns the current number of events

      double timeElapsedInSec() const; ///< Returns the elapsed time in seconds
      double timeElapsedInMin() const; ///< Returns the elapsed time in minutes

      /// Return a string with the elapsed time in seconds.
      std::string timeElapsedString(std::string prefix=std::string("")) const;

      /// Return a formatted progress string for output to the screen
      std::string progressString();

   private:
      /// Convert number of events and seconds into formated string
      std::string eventtime2string(size_t events, size_t secs) const;
      std::string int2string(size_t n) const;
};

} //namespace Geneva

#endif // GENEVA_TIMING_HPP
