//------------------------------------------------------------------------------
/// \file Algorithms.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2013 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Collection of basic algorithms.
//------------------------------------------------------------------------------

#ifndef GENEVA_ALGORITHMS_HPP
#define GENEVA_ALGORITHMS_HPP

#include <limits>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \struct TrueCondition
 * \ingroup Core
 *
 * \brief Helper functor that always returns \c true.
 *
 * Can be used in template algorithms as unary or binary condition that always
 * yields \c true.
 */
//------------------------------------------------------------------------------
struct TrueCondition
{
   template <typename T1>
   bool operator()(T1) const { return true; }

   template <typename T1, typename T2>
   bool operator()(T1, T2) const { return true; }
};


//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Checks if an iterator is contained in an iterator range.
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param it    The iterator to check.
 * \return \c true if \a it is in the range \c [first, last), \c false otherwise.
 *
 * It is the user's responsibility to ensure that all iterators actually belong
 * to the same container.
 *
 * \note Requires linear time. For random access iterators a faster method is to
 * simply check that \c first <= it < last.
 */
template <class Iterator>
inline bool range_contains(Iterator first, Iterator last, Iterator it)
{
   while (first != last) {
      if (it == first)
         return true;
      ++first;
   }
   return false;
}


//------------------------------------------------------------------------------
// minimum
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Finds the minimum value of all elements in a given range.
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param func The value function to be used.
 * \return The minimum value that was found.
 *
 * Finds the minimum function value of all elements in the range specified by
 * \c [first, last). The function to evaluate for each element is specified by
 * the \a func function object, which should take an element as argument and
 * return a double. To also obtain the element with the minimum value
 * \c minimum(first, last, it, func) can be used instead.
 *
 * \note Requires linear time, where \a func is evaluated precisely once for
 * each element in the range. If the range is empty then +inf is returned.
 */
template <typename ForwardIterator, typename UnaryFunction>
inline double minimum(ForwardIterator first, ForwardIterator last, UnaryFunction func)
{
   if (first == last) {
      return std::numeric_limits<double>::infinity();
   }
   double min = func(*first);
   while (++first != last) {
      const double temp = func(*first);
      if (temp < min) {
         min = temp;
      }
   }
   return min;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Finds the element with the minimum value of all elements in a given range.
 * \param[in] first An iterator pointing to the first element in the range.
 * \param[in] last  An iterator pointing beyond the last element in the range.
 * \param[out] it Iterator pointing to the first element with the minimum value.
 * \param[in] func The value function to be used.
 * \return The minimum value that was found.
 *
 * Finds the element with the minimum function value of all elements in the
 * range specified by \c [first, last). The function to evaluate for each
 * element is specified by the \a func function object, which should take an
 * element as argument and return a double. If several elements have the same
 * minimum value then \a it will point to the first one found.
 *
 * \note Requires linear time, where \a func is evaluated precisely once for
 * each element in the range. If the range is empy then \a it will point to
 * \a last and +inf is returned.
 */
template <typename ForwardIterator, typename UnaryFunction>
inline double minimum(ForwardIterator first, ForwardIterator last,
                      ForwardIterator& it, UnaryFunction func)
{
   it = first;
   if (first == last) {
      return std::numeric_limits<double>::infinity();
   }
   double min = func(*first);
   while (++first != last) {
      const double temp = func(*first);
      if (temp < min) {
         it = first;
         min = temp;
      }
   }
   return min;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Finds the minimum value of all elements in a given range satisfying a condition..
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param func The value function to be used.
 * \param cond The condition each considered element has to satisfy.
 * \return The minimum value that was found.
 *
 * Finds the minimum function value of all elements in the range specified by
 * \c [first, last). The function to evaluate for each element is specified by
 * the \a func function object, which should take an element as argument and
 * return a double. Only elements are considered for which the \a cond function
 * object returns \c true. To also obtain the element with the minimum value
 * \c minimum_if(first, last, it, func) can be used instead.
 *
 * \note Requires linear time, where \a cond is evaluated precisely once for
 * each element in the range and \a func only for those satisfying \a cond. If
 * the range is empty or no element satisfies \a cond then +inf is returned.
 */
template <typename ForwardIterator, typename UnaryFunction, typename UnaryCondition>
inline double minimum_if(ForwardIterator first, ForwardIterator last,
                         UnaryFunction func, UnaryCondition cond)
{
   double min = std::numeric_limits<double>::infinity();
   for (; first != last; ++first) {
      if (cond(*first)) {
         const double temp = func(*first);
         if (temp < min) {
            min = temp;
         }
      }
   }
   return min;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Finds the element with the minimum value of all elements in a given range satisfying a condition.
 * \param[in] first An iterator pointing to the first element in the range.
 * \param[in] last  An iterator pointing beyond the last element in the range.
 * \param[out] it Iterator pointing to the first element with the minimum value.
 * \param[in] func The value function to be used.
 * \param[in] cond The condition each considered element has to satisfy.
 * \return The minimum value that was found.
 *
 * Finds the element with the minimum function value of all elements in the
 * range specified by \c [first, last). The function to evaluate for each
 * element is specified by the \a func function object, which should take an
 * element as argument and return a double. Only elements are considered for
 * which the \a cond function object returns \c true. If several elements have
 * the same minimum value then \a it will point to the first one found.
 *
 * \note Requires linear time, where \a cond is evaluated precisely once for
 * each element in the range and \a func only for those satisfying \a cond.
 * If the range is empy then \a it will point to \a last and +inf is returned.
 */
template <typename ForwardIterator, typename UnaryFunction, typename UnaryCondition>
inline double minimum_if(ForwardIterator first, ForwardIterator last,
                         ForwardIterator& it, UnaryFunction func, UnaryCondition cond)
{
   double min = std::numeric_limits<double>::infinity();
   it = last;
   for (; first != last; ++first) {
      if (cond(*first)) {
         const double temp = func(*first);
         if (temp < min) {
            min = temp;
            it = first;
         }
      }
   }
   return min;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Checks if a given element has the minimum value of all elements in a given range.
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param it Iterator pointing to the element.
 * \param func The value function to be used.
 * \return \c true if \a it corresponds to the element with the minimum value,
 *         \c false otherwise.
 *
 * The algorithm checks if any other element in the range specified by
 * \c [first, last) has a smaller value than \a it. The algorithm stops as soon
 * as one element with a smaller value is found, so in this case it is faster
 * than finding the overall minimum. The function to evaluate for each element
 * is specified by the \a func function object, which should take an element as
 * argument and return a double.
 *
 * \note Requires linear time. The user is responsible to make sure that \a it
 * is a valid iterator into the range.
 */
template <typename ForwardIterator, typename UnaryFunction>
inline bool is_minimum(ForwardIterator first, ForwardIterator last,
                       ForwardIterator it,  UnaryFunction func)
{
   const double min = func(*it);
   for (; first != last; ++first) {
      if (first == it)
         continue;
      if (func(*first) < min)
         return false;
   }
   return true;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Checks if a given element has the minimum value of all elements in a given range satisfying a condition.
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param it Iterator pointing to the element.
 * \param func The value function to be used.
 * \param cond The unary condition each considered element has to satisfy.
 * \return \c true if \a it corresponds to the element with the minimum value,
 *         \c false otherwise.
 *
 * The algorithm checks if any other element in the range specified by
 * \c [first, last) has a smaller value than \a it. The algorithm stops as soon
 * as one element with a smaller value is found, so in this case it is faster
 * than finding the overall minimum. The function to evaluate for each element
 * is specified by the \a func function object, which should take an element as
 * argument and return a double. Only elements are considered for which the
 * \c cond function object returns \c true. If \a it itself does not satisfy
 * \a cond then \c false is returned.
 *
 * \note Requires linear time. The user is responsible to make sure that \a it
 * is a valid iterator into the range.
 */
template <typename ForwardIterator, typename UnaryFunction, typename UnaryCondition>
inline bool is_minimum_if(ForwardIterator first, ForwardIterator last,
                          ForwardIterator it, UnaryFunction func, UnaryCondition cond)
{
   if (!cond(*it))
      return false;
   const double min = func(*it);
   for (; first != last; ++first) {
      if (first == it)
         continue;
      if (cond(*first) && func(*first) < min)
         return false;
   }
   return true;
}


//------------------------------------------------------------------------------
// maximum
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Finds the maximum value of all elements in a given range.
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param func The value function to be used.
 * \return The maximum value that was found.
 *
 * Finds the maximum function value of all elements in the range specified by
 * \c [first, last). The function to evaluate for each element is specified by
 * the \a func function object, which should take an element as argument and
 * return a double. To also obtain the element with the maximum value
 * \c maximum(first, last, it, func) can be used instead.
 *
 * \note Requires linear time, where \a func is evaluated precisely once for
 * each element in the range. If the range is empty then -inf is returned.
 */
template <typename ForwardIterator, typename UnaryFunction>
inline double maximum(ForwardIterator first, ForwardIterator last, UnaryFunction func)
{
   if (first == last) {
      return -std::numeric_limits<double>::infinity();
   }
   double max = func(*first);
   while (++first != last) {
      const double temp = func(*first);
      if (temp > max) {
         max = temp;
      }
   }
   return max;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Finds the element with the maximum value of all elements in a given range.
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param[out] it Iterator pointing to the first element with the maximum value.
 * \param func The value function to be used.
 * \return The maximum value that was found.
 *
 * Finds the element with the maximum function value of all elements in the
 * range specified by \c [first, last). The function to evaluate for each
 * element is specified by the \a func function object, which should take an
 * element as argument and return a double. If several elements have the same
 * maximum value then \a it will point to the first one found.
 *
 * \note Requires linear time, where \a func is evaluated precisely once for
 * each element in the range. If the range is empy then \a it will point to
 * \a first and -inf is returned.
 */
template <typename ForwardIterator, typename UnaryFunction>
inline double maximum(ForwardIterator first, ForwardIterator last,
                      ForwardIterator& it, UnaryFunction func)
{
   it = first;
   if (first == last) {
      return -std::numeric_limits<double>::infinity();
   }
   double max = func(*first);
   while (++first != last) {
      const double temp = func(*first);
      if (temp > max) {
         it = first;
         max = temp;
      }
   }
   return max;
}


//------------------------------------------------------------------------------
// pairwise_minimum
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Finds the pairwise minimum distance between elements in a given range.
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param metric The pairwise distance measure to be used.
 * \return The minimum distance that was found.
 *
 * Finds the pairwise minimum distance between any two elements in the range
 * specified by \c [first, last). The pairwise distance measure is specified by
 * the \a metric function object, which should take two elements as arguments
 * and return a double. To also obtain the elements with the minimum distance
 * \c pairwise_minimum(first, last, it1, it2, metric) can be used instead.
 *
 * \note It is assumed that \a metric is commutative. Requires quadratic time,
 * where \a metric is evaluated precisely once for each pair of elements in the
 * range. If the range has less than two elements then +inf is returned.
 */
template <typename ForwardIterator, typename BinaryMetric>
inline double pairwise_minimum(ForwardIterator first, ForwardIterator last, BinaryMetric metric)
{
   double min = std::numeric_limits<double>::infinity();
   for (; first != last; ++first) {
      ForwardIterator i = first;
      while (++i != last) {
         const double temp = metric(*first, *i);
         if (temp < min) {
            min = temp;
         }
      }
   }
   return min;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Finds the elements with the pairwise minimum distance of all elements in a given range.
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param[out] it1 Iterator pointing to the 1st found element.
 * \param[out] it2 Iterator pointing to the 2nd found element.
 * \param metric The pairwise distance measure to be used.
 * \return The minimum distance that was found.
 *
 * Finds the two elements in the range specified by \c [first, last) that have
 * the smallest pairwise distance. The pairwise distance measure is specified by
 * the \a metric function object, which should take two elements as arguments
 * and return a double.
 *
 * \note It is assumed that \a metric is commutative. Requires quadratic time,
 * where \a metric is evaluated precisely once for each pair of elements in the
 * range. If the range has less than two elements then \a it1 and \a it2 will
 * both point to \a last and +inf is returned.
 */
template <typename ForwardIterator, typename BinaryMetric>
inline double pairwise_minimum(ForwardIterator first, ForwardIterator last,
                               ForwardIterator& it1, ForwardIterator& it2, BinaryMetric metric)
{
   double min = std::numeric_limits<double>::infinity();
   it1 = last;
   it2 = last;
   for (; first != last; ++first) {
      ForwardIterator i = first;
      while (++i != last) {
         const double temp = metric(*first, *i);
         if (temp < min) {
            it1 = first;
            it2 = i;
            min = temp;
         }
      }
   }
   return min;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Finds the pairwise minimum distance between elements in a given range satisfying a condition.
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param metric The pairwise distance measure to be used.
 * \param cond The condition each considered pair has to satisfy.
 * \return The minimum distance that was found.
 *
 * Finds the pairwise minimum distance between any two elements in the range
 * specified by \c [first, last). The pairwise distance measure is specified by
 * the \a metric function object, which should take two elements as arguments
 * and return a double. Only pairs of elements are considered for which the
 * \a cond function object, which should take two elements as arguments, returns
 * \c true. To also obtain the elements with the minimum distance
 * \c pairwise_minimum_if(first, last, it1, it2, metric, cond) can be used
 * instead.
 *
 * \note It is assumed that both \a metric and \a cond are commutative. Requires
 * quadratic time, where \a cond is evaluated precisely once for each pair of
 * elements in the range and \a metric only for those satisifying \a cond. If
 * the range has less than two elements then +inf is returned.
 */
template <typename ForwardIterator, typename BinaryMetric, typename BinaryCondition>
inline double pairwise_minimum_if(ForwardIterator first, ForwardIterator last,
                                  BinaryMetric metric, BinaryCondition cond)
{
   double min = std::numeric_limits<double>::infinity();
   for (; first != last; ++first) {
      ForwardIterator i = first;
      while (++i != last) {
         if (cond(*first, *i)) {
            const double temp = metric(*first, *i);
            if (temp < min) {
               min = temp;
            }
         }
      }
   }
   return min;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Finds the elements with the pairwise minimum distance of all elements in a given range satisfying a condition.
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param[out] it1 Iterator pointing to the 1st found element.
 * \param[out] it2 Iterator pointing to the 2nd found element.
 * \param metric The pairwise distance measure to be used.
 * \param cond The condition each considered pair has to satisfy.
 * \return The minimum distance that was found.
 *
 * Finds the two elements in the range specified by \c [first, last) that have
 * the smallest pairwise distance. The pairwise distance measure is specified
 * by the \a metric function object, which should take two elements as arguments
 * and return a double. Only pairs of elements are considered for which the
 * \a cond function object, which should take two elements as arguments, returns
 * \c true.
 *
 * \note It is assumed that both \a metric and \a cond are commutative. Requires
 * quadratic time, where \a cond is evaluated precisely once for each pair of
 * elements in the range and \a metric only for those satisifying \a cond. If
 * the range has less than two elements or no pair of elements satisfies \a cond
 * then \a it1 and \a it2 will both point to \a last +inf is returned.
 */
template <typename ForwardIterator, typename BinaryMetric, typename BinaryCondition>
inline double pairwise_minimum_if(ForwardIterator first, ForwardIterator last,
                                  ForwardIterator& it1, ForwardIterator& it2,
                                  BinaryMetric metric, BinaryCondition cond)
{
   double min = std::numeric_limits<double>::infinity();
   it1 = last;
   it2 = last;
   for (; first != last; ++first) {
      ForwardIterator i = first;
      while (++i != last) {
         if (cond(*first, *i)) {
            const double temp = metric(*first, *i);
            if (temp < min) {
               it1 = first;
               it2 = i;
               min = temp;
            }
         }
      }
   }
   return min;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Checks if a given pair has the pairwise minimum distance of all elements in a given range.
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param it1 Iterator pointing to the 1st element of the pair.
 * \param it2 Iterator pointing to the 2nd element of the pair.
 * \param metric The pairwise distance measure to be used.
 * \return \c true if \a it1 and \a it2 correspond to the pair with the minimum
 *         distance, \c false otherwise.
 *
 * The algorithm checks if any other pair of elements in the range specified by
 * \c [first, last) has a smaller pairwise distance than the \a it1 \a it2 pair.
 * The algorithm stops as soon as one pair with a smaller distance is found, so
 * in this case it is faster than finding the overall minimum.
 * The pairwise distance measure is specified by the \a metric function object,
 * which should take two elements as arguments and return a double.
 *
 * \note It is assumed that \a metric is commutative. Requires quadratic time.
 * The user is responsible to make sure that \a it1 and \a it2 are valid
 * iterators into the range.
 */
template <typename ForwardIterator, typename BinaryMetric>
inline bool is_pairwise_minimum(ForwardIterator first, ForwardIterator last,
                                ForwardIterator it1, ForwardIterator it2, BinaryMetric metric)
{
   const double min = metric(*it1, *it2);
   for (; first != last; ++first) {
      ForwardIterator i = first;
      if (first == it1) {
         while (++i != last) {
            if (i != it2 && metric(*first, *i) < min)
               return false;
         }
      } else if (first == it2) {
         while (++i != last) {
            if (i != it1 && metric(*first, *i) < min)
               return false;
         }
      } else {
         while (++i != last) {
            if (metric(*first, *i) < min)
               return false;
         }
      }
   }
   return true;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Checks if a given pair has the pairwise minimum distance of all elements in a given range satisfying a condition.
 * \param first An iterator pointing to the first element in the range.
 * \param last  An iterator pointing beyond the last element in the range.
 * \param it1 Iterator pointing to the 1st element of the pair.
 * \param it2 Iterator pointing to the 2nd element of the pair.
 * \param metric The pairwise distance measure to be used.
 * \param cond The condition each considered pair has to satisfy.
 * \return \c true if \a it1 and \a it2 correspond to the pair with the minimum
 *         distance, \c false otherwise.
 *
 * The algorithm checks if any other pair of elements in the range specified by
 * \c [first, last) has a smaller pairwise distance than the \a it1 \a it2 pair.
 * The algorithm stops as soon as one pair with a smaller distance is found, so
 * in this case it is faster than finding the overall minimum.
 * The pairwise distance measure is specified by the \a metric function object,
 * which should take two elements as arguments and return a double. Only pairs
 * of elements are considered for which the \c cond function object, which
 * should take two elements as arguments, returns \c true. If \a it1 and \a it2
 * themselves do not satisfy \a cond then \c false is returned.
 *
 * \note It is assumed that \a metric is commutative. Requires quadratic time.
 * The user is responsible to make sure that \a it1 and \a it2 are valid
 * iterators into the range.
 */
template <typename ForwardIterator, typename BinaryMetric, typename BinaryCondition>
inline bool is_pairwise_minimum_if(ForwardIterator first, ForwardIterator last,
                                   ForwardIterator it1, ForwardIterator it2,
                                   BinaryMetric metric, BinaryCondition cond)
{
   if (!cond(*it1, *it2))
      return false;
   const double min = metric(*it1, *it2);
   for (; first != last; ++first) {
      ForwardIterator i = first;;
      if (first == it1) {
         while (++i != last) {
            if (i != it2 && cond(*first, *i) && metric(*first, *i) < min)
               return false;
         }
      } else if (first == it2) {
         while (++i != last) {
            if (i != it1 && cond(*first, *i) && metric(*first, *i) < min)
               return false;
         }
      } else {
         while (++i != last) {
            if (cond(*first, *i) && metric(*first, *i) < min)
               return false;
         }
      }
   }
   return true;
}

} // namespace Geneva

#endif // GENEVA_ALGORITHMS_HPP
