//------------------------------------------------------------------------------
/// \file CaptureOutput.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2017 UNIMIB
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class CaptureOutput
//------------------------------------------------------------------------------

#ifndef GENEVA_CAPTURE_OUTPUT_HPP
#define GENEVA_CAPTURE_OUTPUT_HPP

#include <streambuf>
#include <sstream>
#include <fstream>

namespace Geneva
{
//------------------------------------------------------------------------------
/**
 * \class CaptureOutput
 * \ingroup Core
 *
 * \brief Captures the screen output and allows directing it to a log file.
 *
 * The capture is activated by the constructor and deactivated by the destructor.
 * All output on Log and all output on std::cout (e.g. from external libraries)
 * is captured and directed for the lifetime of an instance of this class.
 *
 * The captured and future output can be directed to a log file at any given
 * time using the \direct_to method. An empty file name is interpreted as
 * directing to std::cout. The output is by default appended to the given log
 * file, unless explicitly requested otherwise.
 *
 * Upon destruction of the class, the default output to std::cout is restored
 * and all output that has been captured and not redirected yet is sent there.
 */
//------------------------------------------------------------------------------
class CaptureOutput
{
   private:
      std::stringstream _captured; ///< Temporary stream to store the screen output
      std::ofstream _logFile;      ///< Log file where the output is redirected to (when requested)
      std::streambuf* _coutBuf;    ///< Store original std::cout buffer

   public:
      CaptureOutput();
      ~CaptureOutput();

      CaptureOutput(const CaptureOutput&) = delete;
      CaptureOutput(CaptureOutput&&) = delete;
      CaptureOutput& operator=(const CaptureOutput&) = delete;
      CaptureOutput& operator=(CaptureOutput&&) = delete;

      /// Directs the output to \a logFileName.
      void direct_to(const std::string& logFileName, bool append = true);
};

} // namespace Geneva

#endif // GENEVA_CAPTURE_OUTPUT_HPP
