//------------------------------------------------------------------------------
/// \file FPEUtils.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Utilities to handle floating point exceptions
//------------------------------------------------------------------------------

#ifndef GENEVA_FPE_UTILS_HPP
#define GENEVA_FPE_UTILS_HPP

#include "Geneva/config.hpp"

#include <fenv.h>
#include <stdexcept>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class TrapFPE
 * \ingroup Core
 *
 * \brief Trap to generate SIGFPE signal from floating point exceptions.
 *
 * Turns on trapping of floating point exceptions (FPEs). GCC has a feature
 * where specific FPEs can be set to automatically cause a SIGFPE generated at
 * OS level, which immediately aborts the program.
 *
 * The Trap is activated by the constructor and deactivated by the destructor.
 * This means SIGFPEs for a given set of FPEs will be generated for the lifetime
 * an instance of this class.
 *
 * If several traps are nested, then always the innermost is active and shadows
 * the outer ones. This also means an inner trap with fewer FPEs set to signal
 * can be used to temporarily turn off signals from FPEs that were turned on by
 * an outer trap. This only refers to the signalling of FPEs, they are always
 * raised by the CPU.
 *
 * \note This class should only be used for debugging to know exactly where an
 * FPE occurred. To check for raised FPEs under normal circumstances use the
 * \c CatchFPE class.
 */
//------------------------------------------------------------------------------
class TrapFPE
{
#ifdef geneva_USE_FPE_TRAPPING
   private:
      int _oldFlags;

   public:
      // by default enable invalid, division by zero, and overflow FPEs
      TrapFPE(int flags = FE_INVALID | FE_DIVBYZERO | FE_OVERFLOW);

      ~TrapFPE();
#else
   public:
      TrapFPE() = default;
      TrapFPE(int) {};
      ~TrapFPE() = default;
#endif

      TrapFPE(const TrapFPE&) = delete;
      TrapFPE(TrapFPE&&) = delete;
      TrapFPE& operator=(const TrapFPE&) = delete;
      TrapFPE& operator=(TrapFPE&&) = delete;
};


//------------------------------------------------------------------------------
/**
 * \class floating_point_error
 * \ingroup Core
 *
 * \brief Exception class to report floating point errors.
 *
 * This is derived from \c std::runtime_error and thrown by \c CatchFPE when an
 * FPE was detected.
 */
//------------------------------------------------------------------------------
class floating_point_error : public std::runtime_error
{
   public:
      explicit floating_point_error(const std::string& what)
         : std::runtime_error(what)
      {};
};


//------------------------------------------------------------------------------
/**
 * \class CatchFPE
 * \ingroup Core
 *
 * \brief Detects and reports floating point exceptions (FPEs).
 *
 * This class "catches" a given set of FPEs and "releases" them as ordinary C++
 * exceptions. FPEs are raised by the CPU and stay so until explicitly cleared.
 * Here, "catching" means that the specified FPEs are prevented from being
 * trapped and can instead be tested for, cleared, or released.
 *
 * The CatchFPE is activated by the constructor and starts from a clean state.
 * Subsequently occurring FPEs can be explicitly tested for (in which case the
 * user can decide what to do). They will stay raised until they are either
 * cleared or released. When FPEs are released, they are cleared and an ordinary
 * ordinary C++ exception of type \c floating_point_error is thrown.
 *
 * The Catch is deactivated by its destructor (i.e. when it ceases to exist). At
 * that point, all caught FPEs are cleared and the previous FPE state is
 * restored.
 *
 * If several CatchFPEs are nested they are effectively combined, where each can
 * test, clear, and release its set of FPEs. Clearing a type of FPE clears it
 * for all active catches. (except when a CatchFPE ceases to exist, it does
 * restore any FPEs that were raised before it started its life).
 *
 * If a TrapFPE is active, this class prevents all FPEs from reaching
 * an active outer TrapFPE. Any FPEs that occur while the CatchFPE is active and
 * are not to be caught by it, will reach the TrapFPE only after the CatchFPE
 * ceases to exist. On the other hand, an inner TrapFPE takes precedence over an
 * outer CatchFPE and will immediately trap and signal its FPEs, while all other
 * FPEs are seen by outer CatchFPEs.
 */
//------------------------------------------------------------------------------
class CatchFPE
{
   private:
      fenv_t _env;
      int _flags;

   public:
      /**
       * \brief Activates the catching of FPEs.
       * \param flags Bitmask of which FPEs should be caught.
       *
       * Valid FPE flags defined by the C library are
       * - FE_INVALID
       * - FE_DIVBYZERO
       * - FE_INEXACT
       * - FE_OVERFLOW
       * - FE_UNDERFLOW
       *
       * By default invalid, division by zero, and overflow are caught.
       */
      CatchFPE(int flags = FE_INVALID | FE_DIVBYZERO | FE_OVERFLOW);

      CatchFPE(const CatchFPE&) = delete;
      CatchFPE(CatchFPE&&) = delete;
      CatchFPE& operator=(const CatchFPE&) = delete;
      CatchFPE& operator=(CatchFPE&&) = delete;

      /**
       * \brief Deactivates the catching of FPEs.
       *
       * The previous floating point environment is restored. Currently caught
       * FPEs are cleared, other raised FPEs are reraised after restoring the
       * environment.
       */
      ~CatchFPE();

      /**
       * \brief Tests if there are caught FPEs.
       * \return Bitmask of currently caught FPEs.
       *
       * The returned bitmask only includes raised FPEs that are "caught" by the
       * CatchFPE.
       */
      int test();

      /**
       * \brief Clears caught FPEs.
       *
       * Only raised FPEs that are "caught" by the CatchFPE are cleared.
       */
      void clear();

      /**
       * \brief Releases caught FPEs.
       * \throw floating_point_error in case there are caught FPEs.
       *
       * Tests if there are caught FPEs (see \c test()). If so, they are cleared
       * and a \c floating_point_error is thrown. The \c what() message of the
       * exception includes information about which type of FPEs were caught.
       */
      void release();
};

} // namespace Geneva

#endif // GENEVA_FPE_UTILS_HPP
