//------------------------------------------------------------------------------
/// \file MathUtils.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Math definitions and support functions
//------------------------------------------------------------------------------

#ifndef GENEVA_MATH_UTILS_HPP
#define GENEVA_MATH_UTILS_HPP

#include "Geneva/config.hpp"

#ifdef geneva_USE_GSL
#include <gsl/gsl_sys.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_sf_dilog.h>
#include <gsl/gsl_sf_psi.h>
#include <gsl/gsl_sf_erf.h>
#endif

#include <cstddef>
#include <cstdlib>
#include <cassert>
#include <cmath>
#include <limits>
#include <utility>

namespace Geneva
{

// pull C++ standard types into our namespace
using std::size_t;
using std::ptrdiff_t;

//------------------------------------------------------------------------------
// Define Gamma function gamma(x).
// C99 has two functions lgamma(x) giving ln(Gamma(x)) and tgamma(x) giving the
// true Gamma(x). These definitions pull them into our namespace. Some compilers
// (including gcc and clang) still have a deprecated built-in gamma(x) returning
// lgamma(x), which is very misleading and should not be used.
//------------------------------------------------------------------------------

inline double gamma(double x)
{
   return ::tgamma(x);
}

inline double lngamma(double x)
{
   return ::lgamma(x);
}

//------------------------------------------------------------------------------
// Math functions where we want to enforce to use the versions in namespace std
//------------------------------------------------------------------------------

// std::pow defines a pow(double, int) signature
using std::pow;
using std::exp;
using std::log;
using std::sqrt;
using std::sin;
using std::cos;
using std::tan;
using std::asin;
using std::acos;
using std::atan;
using std::isnan;

// std::abs defines double abs(double) signature in addition to the int abs(int)
// signature in cstdlib
using std::abs;

//------------------------------------------------------------------------------
// Include special function definitions from gsl and put them into our
// namespace so the user code is independent of where they are coming from.
//------------------------------------------------------------------------------

#ifdef geneva_USE_GSL

namespace gsl
{

//------------------------------------------------------------------------------
/// Returns n!
inline double factorial(const unsigned int n)
{
   return gsl_sf_fact(n);
}

//------------------------------------------------------------------------------
/// Returns Li_2(x).
inline double dilog(const double x)
{
   return gsl_sf_dilog(x);
}

//------------------------------------------------------------------------------
/// Returns Di-Gamma psi^(0)(x).
inline double psi0(const double x)
{
   return gsl_sf_psi(x);
}

//------------------------------------------------------------------------------
/// Returns Tri-Gamma psi^(1)(x).
inline double psi1(const double x)
{
   return gsl_sf_psi_1(x);
}

//------------------------------------------------------------------------------
/// Returns Poly-Gamma psi^(n)(x).
inline double psin(const int n, const double x)
{
   return gsl_sf_psi_n(n, x);
}

//-----------------------------------------------------------------------------
/// Returns the error function Erf(x).
inline double erf(const double x)
{
   return gsl_sf_erf(x);
}

} // namespace gsl

using gsl::factorial;
using gsl::dilog;
using gsl::psi0;
using gsl::psi1;
using gsl::psin;
using gsl::erf;

#endif // geneva_USE_GSL


//------------------------------------------------------------------------------
/**
 * \struct math_constants
 * \ingroup Core
 *
 * \brief Definitions of mathematical constants
 *
 * Holds a collection of mathematical constants and in case of \f$\pi\f$ several
 * commonly needed combinations.
 *
 * It is modeled after the std::numerical_limits, except that the only type
 * supported right now is double. (It is easy to make it into a template later.)
 *
 * Note that names starting with numbers start with an additional underscore.
 */
//------------------------------------------------------------------------------
struct math_constants
{
   // factors of pi
   /// \f$\pi\f$
   static double pi()
   { return static_cast<double>(3.1415926535897932384626433832795029L); }
   /// \f$2 \pi\f$
   static double _2pi()
   { return static_cast<double>(6.2831853071795864769252867665590058L); }
   /// \f$3 \pi\f$
   static double _3pi()
   { return static_cast<double>(9.4247779607693797153879301498385087L); }
   /// \f$4 \pi\f$
   static double _4pi()
   { return static_cast<double>(12.566370614359172953850573533118012L); }
   /// \f$8 \pi\f$
   static double _8pi()
   { return static_cast<double>(25.132741228718345907701147066236023L); }
   /// \f$\pi/2\f$
   static double pi_2()
   { return static_cast<double>(1.5707963267948966192313216916397514L); }
   /// \f$\pi/3\f$
   static double pi_3()
   { return static_cast<double>(1.0471975511965977461542144610931676L); }
   /// \f$\pi/4\f$
   static double pi_4()
   { return static_cast<double>(0.7853981633974483096156608458198757L); }
   /// \f$\pi/8\f$
   static double pi_8()
   { return static_cast<double>(0.3926990816987241548078304229099379L); }

   // powers of pi
   /// \f$\pi^2\f$
   static double pi2()
   { return static_cast<double>(9.8696044010893586188344909998761511L); }
   ///  \f$\pi^2/6\f$
   static double pi2_6()
   { return static_cast<double>(1.6449340668482264364724151666460252L); }
   /// \f$\pi^3\f$
   static double pi3()
   { return static_cast<double>(31.006276680299820175476315067101395L); }
   /// \f$\pi^4\f$
   static double pi4()
   { return static_cast<double>(97.409091034002437236440332688705111L); }

   // inverse powers of pi
   /// \f$1 / \pi\f$
   static double _1_pi()
   { return static_cast<double>(0.31830988618379067153776752674502872L); }
   /// \f$1 / (2\pi)\f$
   static double _1_2pi()
   { return static_cast<double>(0.15915494309189533576888376337251436L); }
   /// \f$1 / (4\pi)\f$
   static double _1_4pi()
   { return static_cast<double>(7.95774715459476678844418816862571810e-2L); }
   /// \f$1 / \pi^2\f$
   static double _1_pi2()
   { return static_cast<double>(0.10132118364233777144387946320972764L); }
   /// \f$1 / (16\pi^2)\f$
   static double _1_16pi2()
   { return static_cast<double>(6.33257397764611071524246645060797743e-3L); }
   /// \f$1 / (64\pi^3)\f$
   static double _1_64pi3()
   { return static_cast<double>(5.03930225518742018506594573258807634e-4L); }

   // square roots
   ///  \f$\sqrt{2}\f$
   static double sqrt2()
   { return static_cast<double>(1.4142135623730950488016887242096981L); }
   ///  \f$\sqrt{3}\f$
   static double sqrt3()
   { return static_cast<double>(1.7320508075688772935274463415058723L); }
   ///  \f$1 / \sqrt{2}\f$
   static double sqrt1_2()
   { return static_cast<double>(0.7071067811865475244008443621048490L); }
   ///  \f$1 / \sqrt{3}\f$
   static double sqrt1_3()
   { return static_cast<double>(0.5773502691896257645091487805019575L); }
   ///  \f$\sqrt{\pi}\f$
   static double sqrtpi()
   { return static_cast<double>(1.7724538509055160272981674833411452L); }

   // other constants
   ///  Euler-Mascheroni constant \f$e\f$
   static double e()
   { return static_cast<double>(2.7182818284590452353602874713526625L); }
   ///  Euler's constant \f$\gamma_E\f$
   static double gammaE()
   { return static_cast<double>(0.5772156649015328606065120900824024L); }
   ///  \f$\ln 2\f$
   static double ln2()
   { return static_cast<double>(0.6931471805599453094172321214581766L); }
   ///  \f$\ln 3\f$
   static double ln3()
   { return static_cast<double>(1.0986122886681096913952452369225257L); }
   ///  \f$\ln 10\f$
   static double ln10()
   { return static_cast<double>(2.3025850929940456840179914546843642L); }
   ///  GeV to pb conversion factor, value from PDG
   static double gev2pb()
   { return static_cast<double>(0.389379338e9L); }

   // zeta function
   ///  \f$\zeta(2) = \pi^2/6\f$
   static double zeta2()
   { return static_cast<double>(1.6449340668482264364724151666460252L); }
   ///  \f$\zeta(3)\f$
   static double zeta3()
   { return static_cast<double>(1.2020569031595942853997381615114500L); }
   ///  \f$\zeta(4) = \pi^4/90\f$
   static double zeta4()
   { return static_cast<double>(1.0823232337111381915160036965411679L); }
   ///  \f$\zeta(5)\f$
   static double zeta5()
   { return static_cast<double>(1.0369277551433699263313654864570342L); }
};


//------------------------------------------------------------------------------
/**
 * \struct SU
 * \ingroup Core
 *
 * \brief Definition of group constants for SU(N)
 *
 * The template parameter specifies the number of generators. The constants
 * are accessed as \c SU<N>::constant, where \c constant can be \c Nc, \c CA,
 * \c CF, \c TF corresponding to are \f$ N_c = N \f$, \f$ C_A = N \f$,
 * \f$ C_F = (N^2 - 1)/(2N) \f$, and \f$ T_F = 1/2 \f$, respectively.
 */
//------------------------------------------------------------------------------
template <int N>
struct SU
{
   // definitions of various constants
   static const int Nc = N;         ///< number of colors \f$ N_c \f$
   static const double CA;          ///< adjoint Casimir \f$ C_A \f$
   static const double CF;          ///< fundamental Casimir \f$ C_F \f$
   static const double TF;          ///< generator normalization \f$ T_F \f$
};

template <int N>
const double SU<N>::CA = N;

template <int N>
const double SU<N>::CF = (N*N - 1.)/(2.*N);

template <int N>
const double SU<N>::TF = 0.5;


//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Approximate comparison of two doubles to given precision and accuracy
 * \param x1 The 1st double to compare.
 * \param x2 The 2nd double to compare.
 * \param delta The relative precision \f$ \delta \f$ to be used. The internal
 *        default precision is used as default value.
 * \param epsilon The absolute accuracy \f$ \epsilon \f$ to be used. The
 *        internal default accuracy is used as default value.
 * \return An integer 1, 0, or -1 corresponding to the sign of \f$ x_1 - x_2 \f$
 *
 * Checks if \a x1 and \a x2 are approximately equal in terms of both absolute
 * accuracy and relative precision (for the latter using the algorithm by
 * D.E. Knuth in Seminumerical and used in the floating-point comparison from
 * the GSL). That is, \a x1 and \a x2 are considered equal (return value 0) if
 * their absolute difference is less or equal either than \a epsilon or roughly
 * \a delta times the larger of them.
 *
 * Specifically, the return value is computed as follows:
 * \f[
 * \begin{aligned}
 * -1 &\qquad\text{for}\quad& x_1 - x_2 &< -\varepsilon \\
 * 0  &\qquad\text{for}\quad& -\varepsilon \leq x_1 - x_2 &\leq \varepsilon \\
 * 1  &\qquad\text{for}\quad& \varepsilon < x_1 - x_2 &
 * \end{aligned}
 * \f]
 *
 * where \f$ \varepsilon \f$ is given as
 * \f$ \varepsilon = \mathrm{max} \{ \epsilon, \delta \times 2^n \}\f$
 * where \f$ n \f$ is the maximum base-2 exponent of \a x1 and \a x2.
 *
 * \note Since \a x1 and \a x2 are also compared to absolute accuracy, the value
 * of \a epsilon should be roughly chosen as \a delta times the typical size of
 * \a x1 and \a x2. To compare using relative precision only (i.e. to arbitrary
 * accuracy) use compareRelative().
 */
inline int compare(const double x1, const double x2,
                   const double delta = DEFAULT_PRECISION,
                   const double epsilon = DEFAULT_ACCURACY)
{
   // compute sign and absolute value of x1 - x2
   int sign;
   double difference = x1 - x2;
   if (difference >= 0.)
      sign = 1;
   else {
      sign = -1;
      difference = -difference;
   }
   // if |x1 - x2| <= epsilon (absolute precision) we are done
   if (difference <= epsilon)
      return 0;

   // check the relative precision
   int exponent;
   frexp(abs(x1) > abs(x2) ? x1 : x2, &exponent);
   const double eps = ldexp(delta, exponent);
   if (difference <= eps)
      return 0;
   return sign;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Approximate comparison of two doubles to given precision and accuracy
 *
 * Identical to \c compare() (see the documentation there). This version returns
 * a std::pair. The first entry is the integer with the result of the compare.
 * The second entry is a double giving the value of \f$ \varepsilon \f$ to which
 * \f$ \lvert x_1 - x_2 \rvert \f$ was compared, and which is the
 * larger of \a epsilon and the value computed from \a delta.
 */
inline std::pair<int, double> compareEps(const double x1, const double x2,
                                         const double delta = DEFAULT_PRECISION,
                                         const double epsilon = DEFAULT_ACCURACY)
{
   // compute sign and absolute value of x1 - x2
   int sign;
   double difference = x1 - x2;
   if (difference >= 0.)
      sign = 1;
   else {
      sign = -1;
      difference = -difference;
   }

   // compute eps from relative precision
   int exponent;
   frexp(abs(x1) > abs(x2) ? x1 : x2, &exponent);
   double eps = ldexp(delta, exponent);
   if (epsilon > eps)
      eps = epsilon;

   // check |x1 - x2| <= eps
   if (difference <= eps)
      return std::pair<int, double>(0, eps);
   else
      return std::pair<int, double>(sign, eps);
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Approximate comparison of a double with zero to a given accuracy
 * \param x The double to compare with zero.
 * \param epsilon The absolute accuracy \f$ \epsilon \f$ to be used. If
 *        \a epsilon is omitted, the internal default accuracy is used.
 * \return An integer 1, 0, or -1 corresponding to the sign of \a x
 *
 * Checks if \a x is approximately equal to zero.
 *
 * The return value is given by
 * \f[
 * \begin{aligned}
 * -1 &\qquad\text{for}\quad& x &< -\epsilon \\
 * 0  &\qquad\text{for}\quad& -\epsilon \leq x &\leq \epsilon \\
 * 1  &\qquad\text{for}\quad& \epsilon < x &
 * \end{aligned}
 * \f]
 *
 * \warning Comparing to an absolute accuracy is often not very meaningful
 * without further information where that zero came from. Hence, direct use of
 * this function should be avoided in favor of compare(). For example, instead of
 * \code compareToZero(x1 - x2 - a*x3) \endcode
 * one should rather use one of
 * \code
 * compare(x1 - x2, a*x3)
 * compare(x1, x2 + a*x3)
 * \endcode
 */
inline int compareToZero(const double x, const double epsilon = DEFAULT_ACCURACY)
{
   if (x > epsilon)
      return 1;
   if (x < -epsilon)
      return -1;
   return 0;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Approximate comparison of two doubles to given relative precision only
 * \param x1 The 1st double to compare.
 * \param x2 The 2nd double to compare.
 * \param delta The relative precision \f$ \delta \f$ to be used. If \a delta
 *        is omitted, the internal default precision is used.
 * \return An integer 1, 0, or -1 corresponding to the sign of \f$ x_1 - x_2 \f$
 *
 * Checks if \a x1 and \a x2 are approximately equal using relative precision
 * only as given by \a delta. This is equivalent to using compare() with
 * \a epsilon set to 0.0 (i.e. using arbitrary accuracy).
 *
 * \note Since \a x1 and \a x2 are only compared to relative precision, this
 * function should not be used to compare a double to 0.0.
 */
inline int compareRelative(const double x1, const double x2,
                           const double delta = DEFAULT_PRECISION)
{
   int exponent;
   frexp(abs(x1) > abs(x2) ? x1 : x2, &exponent);
   const double eps = ldexp(delta, exponent);
   return compareToZero(x1 - x2, eps);
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Approximate comparison of two doubles to given absolute accuracy only
 * \param x1 The 1st double to compare.
 * \param x2 The 2nd double to compare.
 * \param epsilon The absolute accuracy \f$ \epsilon \f$ to be used. If
 *        \a epsilon is omitted, the internal default accuracy is used.
 * \return An integer 1, 0, or -1 corresponding to the sign of \f$ x_1 - x_2 \f$
 *
 * Checks if \a x1 and \a x2 are approximately equal using absolute accuracy
 * only as given by \a epsilon. This is equivalent to using compare() with
 * \a delta set to 0.0 (i.e. using arbitrary precision)
 */
inline int compareAbsolute(const double x1, const double x2,
                           const double epsilon = DEFAULT_ACCURACY)
{
   return compareToZero(x1 - x2, epsilon);
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Approximate comparison whether a double is within a given [min, max] range
 * \param x The double to check
 * \param min The lower limit of the range.
 * \param max The upper limit of the range.
 * \param delta The relative precision \f$ \delta \f$ to be used. The internal
 *        default precision is used as default value.
 * \param epsilon The absolute accuracy \f$ \epsilon \f$ to be used. If
 *        \a epsilon is omitted, the internal default accuracy is used.
 * \return An integer 2, 1, 0, -1 or -2.
 *
 * Checks if \a x is approximately within the [min, max] interval, using both
 * the relative accuracy \a delta and the absolute accuracy \a epsilon.
 *
 * The return value is given by
 * \f[
 * \begin{aligned}
 * -1 &\qquad\text{for}\quad& x &< min -\varepsilon \\
 * 0  &\qquad\text{for}\quad& min - \varepsilon \leq x &\leq max + \varepsilon \\
 * 1  &\qquad\text{for}\quad& x > max + \varepsilon &
 * \end{aligned}
 * \f]
 * where \f$ \varepsilon \f$ is given as
 * \f$ \varepsilon = \mathrm{max} \{ \epsilon, \delta \times 2^n \}\f$
 * where \f$ n \f$ is the maximum base-2 exponent of \a x and \a min or \a max.
 */
inline int compareRange(const double x, const double min, const double max,
                        const double delta = DEFAULT_PRECISION,
                        const double epsilon = DEFAULT_ACCURACY)
{
   if (compare(x, max, delta, epsilon) == 1)
      return 1;
   if (compare(x, min, delta, epsilon) == -1)
      return -1;
   return 0;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Approximate comparison whether a double is in the unit range [0.0,1.0]
 * \param x The double to check.
 * \param epsilon The absolute accuracy \f$ \epsilon \f$ to be used. If
 *        \a epsilon is omitted, the internal default accuracy is used.
 * \return An integer 1, 0, or -1
 *
 * Checks if \a x is approximately within the [0.0,1.0] interval, using absolute
 * accuracy only as given by \a epsilon.
 *
 * The return value is given by
 * \f[
 * \begin{aligned}
 * -1 &\qquad\text{for}\quad& x &< -\epsilon \\
 * 0  &\qquad\text{for}\quad& -\epsilon \leq x &\leq 1 + \epsilon \\
 * 1  &\qquad\text{for}\quad& x > 1 + \epsilon &
 * \end{aligned}
 * \f]
 */
inline int compareUnitRange(const double x, const double epsilon = DEFAULT_ACCURACY)
{
   if (x > 1.0 + epsilon)
      return 1;
   if (x < -epsilon)
      return -1;
   return 0;
}


//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Safe, portable conversion from \c size_t to \c int
 */
inline int int_from_size_t(size_t i)
{
   assert(sizeof(size_t) < sizeof(int) || i < static_cast<size_t>(std::numeric_limits<int>::max()));
   return static_cast<int>(i);
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Safe, portable conversion from \c int to \c size_t
 */
inline size_t size_t_from_int(int i)
{
   assert(i >= 0 && (sizeof(int) <= sizeof(size_t) || i < static_cast<int>(std::numeric_limits<size_t>::max())));
   return static_cast<size_t>(i);
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Safe, portable conversion from \c ptrdiff_t to \c size_t
 */
inline size_t size_t_from_ptrdiff_t(const ptrdiff_t i)
{
   assert(i >= 0 && (sizeof(ptrdiff_t) <= sizeof(size_t) || i < static_cast<ptrdiff_t>(std::numeric_limits<size_t>::max())));
   return static_cast<size_t>(i);
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Safe, portable conversion from \c size_t to \c ptrdiff_t
 */
inline ptrdiff_t ptrdiff_t_from_size_t(const size_t i)
{
   assert(sizeof(size_t) < sizeof(ptrdiff_t) || i < static_cast<size_t>(std::numeric_limits<ptrdiff_t>::max()));
   return static_cast<ptrdiff_t>(i);
}

} // namespace Geneva

#endif // GENEVA_MATH_UTILS_HPP
