//------------------------------------------------------------------------------
/// \file NumericsUtils.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Utility functions to enforce numerical limits and ranges
//------------------------------------------------------------------------------

#ifndef GENEVA_NUMERICS_UTILS_HPP
#define GENEVA_NUMERICS_UTILS_HPP

#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <iomanip>

namespace Geneva
{

//------------------------------------------------------------------------------
inline bool forceGreaterEqualZero(double& x, std::string where)
{
   if (compareToZero(x) == -1)
      Log::error(where) << std::setprecision(16) << "Unrecoverable numerical imprecision: x = " << x
            << " is below allowed min = 0" << Log::end;

   if (x < 0.0) {
      Log::warning(where) << std::setprecision(16) << "Numerical imprecision: Setting x = " << x << " to allowed min = 0" << Log::end;
      x = 0.0;
      return false;
   }
   return true;
}

//------------------------------------------------------------------------------
inline bool forceGreaterEqual(double& x, double min, std::string where)
{
   if (compare(x, min) == -1)
      Log::error(where) << std::setprecision(16) << "Unrecoverable numerical imprecision: x = "
         << x << " is below allowed min = " << min << Log::end;

   if (x < min) {
      Log::warning(where) << std::setprecision(16) << "Numerical imprecision: Setting x = " << x << " to allowed min = " << min << Log::end;
      x = min;
      return false;
   }
   return true;
}

//------------------------------------------------------------------------------
inline bool forceLessEqual(double& x, double max, std::string where)
{
   if (compare(x, max) == 1)
      Log::error(where) << std::setprecision(16) << "Unrecoverable numerical imprecision: x = "
         << x << " is above allowed max = " << max << Log::end;

   if (x > max) {
      Log::warning(where) << std::setprecision(16) << "Numerical imprecision: Setting x = " << x << " to allowed max = " << max << Log::end;
      x = max;
      return false;
   }
   return true;
}

//------------------------------------------------------------------------------
inline bool forceInRange(double& x, double min, double max, std::string where)
{
   if (min > max)
      Log::error(where) << std::setprecision(16) << "Allowed min = " << min << " is larger than allowed max = " << max << Log::end;

   if (compareRange(x, min, max) != 0)
      Log::error(where) << std::setprecision(16) << "Unrecoverable numerical imprecision: x = " << x
            << " is outside allowed [" << min << ", " << max << "] interval." << Log::end;

   if (x > max) {
      Log::warning(where) << std::setprecision(16) << "Numerical imprecision: Setting x = " << x << " to allowed max = " << max << Log::end;
      x = max;
      return false;
   }
   if (x < min) {
      Log::warning(where) << std::setprecision(16) << "Numerical imprecision: Setting x = " << x << " to allowed min = " << min << Log::end;
      x = min;
      return false;
   }
   return true;
}

//------------------------------------------------------------------------------
inline bool forceInUnitRange(double& x, std::string where)
{
   if (compareUnitRange(x) != 0)
      Log::error(where) << std::setprecision(16) << "Unrecoverable numerical imprecision: x = " << x
            << " is outside allowed [0, 1] interval." << Log::end;

   if (x > 1.0) {
      Log::warning(where) << std::setprecision(16) << "Numerical imprecision: Setting x = " << x << " to allowed max = 1.0" << Log::end;
      x = 1.0;
      return false;
   }
   if (x < 0.0) {
      Log::warning(where) << std::setprecision(16) << "Numerical imprecision: Setting x = " << x << " to allowed min = 0.0" << Log::end;
      x = 0.0;
      return false;
   }
   return true;
}

} // namespace Geneva

#endif // GENEVA_NUMERICS_UTILS_HPP
