//------------------------------------------------------------------------------
/// \file Random.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 MIT
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Random number generator definitions
//------------------------------------------------------------------------------

#ifndef GENEVA_RANDOM_HPP
#define GENEVA_RANDOM_HPP

#include <string>
#include <cstddef>

namespace Geneva
{

#define NTAB 32

/// Holds the state of the generator
struct RandomState
{
   long kseed;
   long kidum2;
   long kiy;
   long kiv[NTAB];
};

#undef NTAB

// TODO: At the moment these are just a collection of free functions using a global
// random number generator and global seed. These should really become a Random class

/// Random angle between 0 and 2pi
double randomAngle();

/// Random number between 0 and 1
double randomInterval();

/// Random integer in [0, N)
int randomInteger(int N);

/// Random integer in [0, N); specialization for size_t
size_t randomInteger(size_t N);

/// Functor for random integer in [0, N)
struct RandomInteger
{
   int operator()(int N)
   {
      return randomInteger(N);
   }
};

/// Default random number generator; called by randomAngle() and randomInterval()
double Ran2(long*);

/// Set global random number seed
void setRandomSeed(long newSeed);

/// Stores state of generator to internal savedState object
void storeRandomState();
/// Returns the state of generator
RandomState getRandomState();

/// Initialize generator from a given RandomState object
void initializeRandomState(RandomState state);

/// Saves state of generator to the file in trunk/data/ (defaults to RandomState.dat)
void saveRandomState(std::string filename="RandomState.dat");

/// Reads state of generator from the file trunk/data/ (defaults to RandomState.dat), initializes it
void initializeFromStatefile(std::string filename="RandomState.dat");

/// Check if the generator is in the same state of the RandomState object passed, i.e. no call to random has been performed
bool equalToStoredState(RandomState state);

} // namespace Geneva

#endif // GENEVA_RANDOM_HPP

