//------------------------------------------------------------------------------
/// \file Singleton.hpp
//
// Author(s):
//    Chris Vermilion, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definition of class Singleton
//------------------------------------------------------------------------------

#ifndef GENEVA_SINGLETON_HPP
#define GENEVA_SINGLETON_HPP

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class Singleton
 * \ingroup Core
 *
 * \brief A template for Singleton classes.
 *
 * This is a template implementation of a Singleton. An example how to use the
 * Singleton is as follows:
 *
 * \code
 * class Foo : public Singleton<Foo>
 * {
 *     protected:
 *        friend class Singleton<Foo>; // only Singleton can call the constructor
 *        Foo() {...}
 *     public:
 *     ...
 * };
 * \endcode
 */
//------------------------------------------------------------------------------
template<typename Type>
class Singleton
{
   public:
      /// holds a static Type (the singleton), and returns a reference on request
      static Type& instance()
      {
         static Type _theInstance;
         return _theInstance;
      }

      template <typename ArgType>
      static Type& instance(ArgType arg)
      {
         static Type _theInstance(arg);
         return _theInstance;
      }

      template <typename ArgType1, typename ArgType2>
      static Type& instance(ArgType1 arg1, ArgType2 arg2)
      {
         static Type _theInstance(arg1, arg2);
         return _theInstance;
      }

      template <typename ArgType1, typename ArgType2, typename ArgType3>
      static Type& instance(ArgType1 arg1, ArgType2 arg2, ArgType3 arg3)
      {
         static Type _theInstance(arg1, arg2, arg3);
         return _theInstance;
      }

   protected:
      /// only derived classes can create or destroy
      Singleton() = default;
      ~Singleton() = default;

   private:
      /// no copying
      Singleton(const Singleton&) = delete;
      Singleton(Singleton&&) = delete;
      Singleton& operator=(const Singleton&) = delete;
      Singleton& operator=(Singleton&&) = delete;
};

} // namespace Geneva

#endif // GENEVA_SINGLETON_HPP
