//------------------------------------------------------------------------------
/// \file fixed_ptr.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Definitions of some very basic support classes.
//------------------------------------------------------------------------------

#ifndef GENEVA_FIXED_PTR_HPP
#define GENEVA_FIXED_PTR_HPP

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class fixed_ptr
 * \ingroup Core
 *
 * \brief A fixed pointer that cannot be modified or deleted by users.
 *
 * The pointer can be dereferenced using \c *ptr and \c ptr-> as usual, as well
 * as tested for equality. The only way for the user to modify it is via a copy
 * or assignment operator. In particular, the pointed to object cannot be
 * deleted by the user.
 *
 * Classes can use this to give clients access to their private data (usually
 * read-only by using
 *
 * \code fixed_ptr<const T> \endcode.
 *
 * This has a similar effect as providing a const reference to private data,
 * with the advantage that clients can keep and pass around the access pointer.
 * To do so, one simply uses an ordinary pointer to initialize a \c fixed_ptr.
 * The integrity of a \c fixed_ptr can be guaranteed to clients since they have
 * no direct access to the underlying ordinary pointer. (There is of course no
 * guard against explicit abuses like casting by the user ...)
 */
//------------------------------------------------------------------------------
template<typename T>
class fixed_ptr
{
   private:
      T* _ptr;

   public:
      typedef T element_type;

      fixed_ptr() : _ptr(0) {};

      fixed_ptr(T* ptr) : _ptr(ptr) {};

      T& operator*() const  { return *_ptr;     }
      T* operator->() const { return _ptr;      }
      operator bool() const { return _ptr != 0; }

      bool operator==(const fixed_ptr<T>& rhs) const
      {
         return _ptr == rhs._ptr;
      }

      bool operator!=(const fixed_ptr<T>& rhs) const
      {
         return _ptr != rhs._ptr;
      }
};


//------------------------------------------------------------------------------
/**
 * \class reference
 * \ingroup Core
 *
 * \brief A reference that can be copied and reassigned.
 *
 * A wrapper which behaves like a reference to an object of type \c T, which is
 * CopyConstructible and CopyAssignable.
 *
 * In function calls the \c reference<T> is implicitly convertible to \c T&, and
 * so can be used like an ordinary reference to an object of type \c T. In cases
 * where the compiler cannot use implicit conversion, \c get() can be used to
 * explicitly obtain a reference to the underlying object. The reference can be
 * assigned a new object by using \c set() or by explicit assignment from
 * another \c reference<T>. For example,
 *
 * \code
 *    int i, j, k;
 *    reference<int> ref(i);      // ref points to i
 *    ref++;                      // i is incremented
 *    ref.set(j);                 // ref points to j
 *    ref++;                      // j is incremented
 *
 *    // ref = k;                 // not possible to avoid ambiguity whether to
 *                                // assign k to ref or to j
 *    ref.get() = k;              // assigns j = k
 *    ref = reference<int>(k)     // assigns ref to point to k
 *
 * \endcode
 *
 * \note The implicit conversion from \c T to \c reference<T> is not allowed to
 * avoid any ambiguity whether in an assignment the reference or underlying
 * object should is meant to be assigned as in \c ref = k above.
 */
//------------------------------------------------------------------------------
template <typename T>
class reference
{
   private:
      T* _object;

   public:
      /**
       * \brief Creates a reference to a given object.
       * \param object The object to refer to.
       */
      explicit reference(T& object)
         : _object(&object) {}

      /**
       * \brief Returns a reference to the underlying referred to object.
       */
      T& get() const { return *_object; }

      /**
       * \brief Assigns a new object to the reference.
       * \param object The new object to refer to.
       */
      void set(T& object) { _object = &object; }

      /**
       * \brief Conversion operator to \c T&.
       * \return A reference to the underlying referred to object.
       */
      operator T&() const { return get(); }

      /**
       * \brief Equality comparison operator.
       * \param rhs The reference to compare to.
       * \return \c true if \a rhs refers to the same object, \c false otherwise.
       */
      bool operator==(const reference<T>& rhs) const
      {
         return _object == rhs._object;
      }

      /**
       * \brief Nonequality comparison operator.
       * \param rhs The reference to compare to.
       * \return \c true if \a rhs refers to a different object, \c false otherwise.
       */
      bool operator!=(const reference<T>& rhs) const
      {
         return _object != rhs._object;
      }
};

} // namespace Geneva

#endif // GENEVA_FIXED_PTR_HPP
