//------------------------------------------------------------------------------
/// \file EventGeneratorAdaptive.hpp
//
// Author(s):
//    Christian Bauer, Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2016 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class EventGeneratorAdaptive
//------------------------------------------------------------------------------

#ifndef GENEVA_EVENT_GENERATOR_ADAPTIVE_HPP
#define GENEVA_EVENT_GENERATOR_ADAPTIVE_HPP

#include "Geneva/Driver/IEventGenerator.hpp"

#include "Geneva/Calculations/ICalculation.hpp"

#include "Geneva/Generators/IGenerator.hpp"
#include "Geneva/Generators/IntegrationChannel.hpp"
#include "Geneva/Generators/SmartRandomHandler.hpp"

#include "Geneva/Core/Physics/Event.hpp"
#include "Geneva/Core/Physics/SignatureSet.hpp"
#include "Geneva/Core/Physics/Rate.hpp"

#include <vector>
#include <utility>

namespace Geneva
{

class OptionMap;

//------------------------------------------------------------------------------
/**
 * \class EventGeneratorAdaptive
 * \ingroup EventGenerator
 *
 * \brief Generates both weighted and unweighted events.
 *
 * Generates unweighted events, one at a time. The generation of unweighted
 * events is performed by MINT, which stores the phase space mapping from
 * random numbers to a point in phase space, and the calculation.
 */
//------------------------------------------------------------------------------
class EventGeneratorAdaptive: public IEventGenerator
{
   protected:
      typedef std::vector<IntegrationChannel>::iterator ChannelIterator;

      IGenerator* _gen;             ///< The phase space generator used
      ICalculation* _calc;          ///< The calculation used
      const OptionMap& _optMap;

      size_t _calcFailures;         ///< number of failed calculations
      size_t _maxCalcFailures;      ///< number of allowed failed calculations before throwing a fatal error
      size_t _maxGenFailures;       ///< number of allowed failed generation attempts before throwing a fatal error

      size_t _numBins;              ///< number of bins per integration channel
      size_t _numIterations;        ///< number of iterations for adapting the integration grid

      SignatureSet _sigSet;         ///< The main Signature set that the EventGenerator handles

      std::vector<IntegrationChannel> _channels;    ///< The vector of IntegrationChannels that each SmartRandom is responsible for
      IntegrationChannel* _lastChannel;

      std::vector<double> _cumulativeWeights; /// The cumulative weights for the channels (normalized to the total weighy);

      Event _event;                    ///< the most recent event generated

      Rate _totalRate;                 ///< Total rate of all events requested

      std::string _stem;               ///< The stem used for the grid files
      std::string _seedTag;            ///< The seed tag appended to grid file names

   private:
      SmartRandomHandler _smartRandomHandler;
      Signature _signature;

   public:
      EventGeneratorAdaptive(IGenerator* gen, ICalculation* calc, const OptionMap& optMap);

      virtual const Event& next();                ///< Create one event

      virtual void finalize(const size_t numEvents);

      /// Returns the total rate generated
      virtual const Rate& totalRate() const;

   protected:
      /// Generates a Configuration
      double generateConfig(ChannelIterator channel, double channelWeight);

      /// Picks a channel based on the relative probabilities
      std::pair<ChannelIterator, double> pickChannel();

   private:
      virtual void next(ChannelIterator it, double channelWeight);

      bool loadSignatureProbabilities();
      void setupOptimization(); ///< Set up the grid
      void createChannels(const SignatureSet& sigSet);
      void writeCrossSection(ChannelIterator it);
      void initializeChannelWeights();
      void updateChannelWeights();
};

////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline const Rate& EventGeneratorAdaptive::totalRate() const
{
   return _totalRate;
}

} // namespace Geneva

#endif // GENEVA_EVENT_GENERATOR_ADAPTIVE_HPP
