//------------------------------------------------------------------------------
/// \file EventGeneratorWeighted.hpp
//
// Author(s):
//    Chris Vermilon, Christian Bauer
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class EventGenerator
//------------------------------------------------------------------------------

#ifndef GENEVA_EVENT_GENERATOR_WEIGHTED_HPP
#define GENEVA_EVENT_GENERATOR_WEIGHTED_HPP

#include "Geneva/Driver/IEventGenerator.hpp"

#include "Geneva/Calculations/ICalculation.hpp"

#include "Geneva/Generators/IGenerator.hpp"

#include "Geneva/Core/Physics/Event.hpp"
#include "Geneva/Core/Physics/Rate.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class EventGeneratorWeighted
 * \ingroup EventGenerator
 *
 * \brief Generates weighted events.
 *
 * Generates weighted events, one at a time.  Stores a Generator and a
 * Calculation, and for every event requested, uses the Generator to generate
 * a point in phase space, and the Calculation to determine the appropriate
 * cross section.
 */
//------------------------------------------------------------------------------
class EventGeneratorWeighted : public IEventGenerator
{
   private:
      IGenerator* _generator;
      ICalculation* _calculation;
      Event _event;                    ///< the most recent event generated
      size_t _calcFailures;            ///< number of failed calculations
      size_t _maxCalcFailures;         ///< number of allowed failed calculations before throwing a fatal error
      size_t _maxGenFailures;          ///< number of allowed failed generation attempts before throwing a fatal error
      Rate _totalRate;                 ///< Total rate of all events requested

   public:
      EventGeneratorWeighted(IGenerator* gen, ICalculation* calc, const OptionMap& optMap);

      virtual void finalize(const size_t numEvents);

      /// Returns the total rate generated
      virtual const Rate& totalRate() const;

      virtual const Event& next();
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline EventGeneratorWeighted::EventGeneratorWeighted(IGenerator* gen, ICalculation* calc, const OptionMap& optMap)
   : _generator(gen), _calculation(calc), _calcFailures(0)
{
   _maxCalcFailures = size_t_from_int(optMap.getOption<int>("EventGenerator::maxCalcFailures"));
   _maxGenFailures = size_t_from_int(optMap.getOption<int>("EventGenerator::maxGenFailures"));
}

//------------------------------------------------------------------------------
inline void EventGeneratorWeighted::finalize(const size_t numEvents)
{
   _generator->finalize();
   _calculation->finalize(numEvents);
}

//------------------------------------------------------------------------------
inline const Rate& EventGeneratorWeighted::totalRate() const
{
   return _totalRate;
}

} // namespace Geneva

#endif // GENEVA_EVENT_GENERATOR_WEIGHTED_HPP
