//------------------------------------------------------------------------------
/// \file Geneva.hpp
//
// Author(s):
//    Chris Vermilion, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class Geneva
//------------------------------------------------------------------------------

#ifndef GENEVA_GENEVA_HPP
#define GENEVA_GENEVA_HPP

#include "Geneva/Driver/IEventGenerator.hpp"

#include "Geneva/Core/Physics/Rate.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"

#include <iosfwd>

namespace Geneva
{

/**
 * \mainpage Geneva documentation
 *
 * GenEvA is an event generator that can combine fixed-order and resummed
 * perturbative calculations with fully exclusive parton shower Monte Carlos.
 * It is based on the methods described in
 * <a href="http://arxiv.org/abs/0801.4026">arXiv:0801.4026</a> and
 * <a href="http://arxiv.org/abs/0801.4028">arXiv:0801.4028</a>.
 * The basic code flow is:
 *
 * - A \ref gen "Generator" generates a point in phase space, basically a set of
 *   particle types and a momentum for each particle.  This corresponds to an
 *   "event" at a particle collider like the LHC.
 * - A \ref calc "Calculation" calculates a weight for this event.  The
 *   calculation typically includes a fixed-order matrix element, possibly
 *   supplemented with factors for logarithmic resummation.
 * - The resulting weighted events are passed to an Analyzer, which calculates
 *   observables (e.g., jet \f$p_T\f$ distributions, total cross sections, etc.)
 *   and stores the results.
 * - The weighted events might also be passed to an external showering program
 *   such as Pythia.  Currently \ref pythia "Pythia8" is available.
 *
 * Internally, the first two steps are grouped in the EventGenerator class,
 * which generates weighted events.  This in turn is wrapped by the
 * Geneva::Geneva driver class, which provides the user with a simple event
 * generation interface while handling option reading and other global
 * facilities.
 *
 * Currently it is the user's responsibility to set up and use an Analyzer, as
 * well as to pass events to Pythia or another parton shower.  These features
 * may eventually move inside the Geneva::Geneva interface.  To see how it
 * all works right now, see the main_pythia.cpp example.
 *
 * More information on specific classes can be found by clicking on class names
 * throughout these pages.  For a hierchical listing of major classes, check out
 * the Modules tab.  A complete (flat) list can be found at the Classes tab.
 */
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \defgroup Driver Main Driver
 *
 * This group contains the main classes to drive the event generation.
 */
//------------------------------------------------------------------------------

class Event;

//------------------------------------------------------------------------------
/**
 * \class Geneva::Geneva
 * \ingroup Driver
 *
 * \brief Main ("driver") interface for the Geneva library
 *
 * The main interface class for the Geneva library. Loads default and user
 * options, takes care of initialization, and wraps EventGenerator.
 */
//------------------------------------------------------------------------------
class Geneva
{
   private:
      IEventGenerator* _evtGen; ///< This guy does all the work
      OptionMap _options;       ///< Hold global option set
      int _eventNum;          ///< Holds the number of the current event. Can be used for debugging a particular event.
      bool _isInitialized;

   public:
      // Constructor using only default options.
      Geneva();

      // Constructor using command line options.
      Geneva(int argc, const char* argv[]);

      // Constructor using a string of options.
      Geneva(const std::string& options);

      Geneva(const Geneva&) = delete;
      Geneva(Geneva&&) = delete;
      Geneva& operator=(const Geneva&) = delete;
      Geneva& operator=(Geneva&&) = delete;

      ~Geneva();

      /// Wrapper for optionMap.getOption()
      template <class T>
      T getOption(const std::string& name) const;

      /// Returns a reference to the global option map.
      const OptionMap& optionMap() const;

      /// Generate the next event.
      const Event& next();

      /// Initializes the run after options have been set.
      void initialize();

      /// Finalize the run.
      void finalize(const size_t numEvents);

      /// Return the total rate generated
      virtual const Rate& totalRate() const;

      void setEventNum(int eventNum) {_eventNum = eventNum;};

      /// Returns a header string including banner and current start time.
      static std::string header();

      /// Returns a footer string including current end time.
      static std::string footer();

   private:
      /// Helper function to construct from the command line
      void construct(const std::string& cmdline = "");

      /// Initialize the random sequence
      void setRandom();

      /// Helper function to parse the options from the command line
      void parse(const std::string& cmdline);
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
template <class T>
inline T Geneva::getOption(const std::string& name) const
{
   return _options.getOption<T>(name);
}

//------------------------------------------------------------------------------
inline const OptionMap& Geneva::optionMap() const
{
   return _options;
}

//------------------------------------------------------------------------------
inline const Rate& Geneva::totalRate() const
{
   return _evtGen->totalRate();
}

} // namespace Geneva

#endif // GENEVA_GENEVA_HPP
