//------------------------------------------------------------------------------
/// \file LHEFInterface.hpp
//
// Author(s):
//     Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2013 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class LHEFInterface
//------------------------------------------------------------------------------

#ifndef GENEVA_LHEF_INTERFACE_HPP
#define GENEVA_LHEF_INTERFACE_HPP

#include "Geneva/config.hpp"

#include "Geneva/Core/Physics/PdgID.hpp"
#include "Geneva/Core/Physics/Rate.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/Utils/Singleton.hpp"
#include "Geneva/Core/Utils/FileUtils.hpp"

#ifdef geneva_USE_ZLIB
#include "Geneva/gzstream.hpp"
#endif

#include <vector>
#include <fstream>

namespace Geneva
{

class ITreeMEColor;
class Event;
class LHEvent;
class Configuration;

//------------------------------------------------------------------------------
/**
 * \class LHEFInterface
 *
 * \brief Interface for writing out Geneva events in the LHEF format
 *
 * Handles both the initialization of LHEvents and formatted writeout in the
 * LHEF format. The LHEF header and footer are supplemented by additional run
 * information.
 */
//------------------------------------------------------------------------------
class LHEFInterface : public Singleton<LHEFInterface>
{
   private:
#ifdef geneva_USE_ZLIB
      typedef gzstream::igzstream _istream_type; ///< zipped input stream
      typedef gzstream::ogzstream _ostream_type; ///< zipped output stream
#else
      typedef std::ifstream _istream_type;       ///< plain input stream
      typedef std::ofstream _ostream_type;       ///< plain output stream
#endif

      std::string _fileName;                ///< name of the output file
      _ostream_type _out;                   ///< output stream

      ITreeMEColor* _colorCalculator;       ///< color calculator
      double _tauNcut;                      ///< N-jettiness cuts
      double _tauNp1cut;
      std::string _calc;                    ///< Calculation name

      std::pair<PdgID, PdgID> _beam;        ///< PdgID of the beams (IDBMUP)
      std::pair<double, double> _beamEnergy;///< beam energies (EBMUP)
      std::pair<int, int> _pdfID;           ///< lhapdf ids for the beams (PDFSUP)

      std::vector<double> _xsecup;          ///< cross section for each subprocess in pb (unused)
      std::vector<double> _xerrup;          ///< xsec error for each subprocess in pb (unused)
      std::vector<double> _xmaxup;          ///< maximum weight for each subprocess (unused)
      std::vector<size_t> _lprup;           ///< identifiers for each subprocess (unused)

      Rate _eventRate;                      ///< total cross section of all received events
      Rate _totalRate;                      ///< total cross section as calculated by the EventGenerator

      bool _normalizeWeights;               ///< flag whether to normalize the event weights to the total xsec
      int _runSeed;                         ///< seed used for this run (included in the footer info)

   public:
      /// Writes out the \a event in LHEF format.
      void writeEvent(const Event& evt);

      /// Sets the total cross section.
      void setTotalRate(const Rate& totalRate) { _totalRate = totalRate; }

      PdgID beamA() const { return _beam.first; }
      PdgID beamB() const { return _beam.second; }

      double beamAEnergy() const { return _beamEnergy.first; }
      double beamBEnergy() const { return _beamEnergy.second; }

      int beamApdfID() const { return _pdfID.first; }
      int beamBpdfID() const { return _pdfID.second; }

      int idwtup() const { return -4; }

      size_t nprup() const { return 1; }

      double xsecup(size_t i = 0) const { return _xsecup[i]; }
      double xerrup(size_t i = 0) const { return _xerrup[i]; }
      double xmaxup(size_t i = 0) const { return _xmaxup[i]; }

      size_t lprup(size_t i = 0) const { return _lprup[i]; }

      // Provides the default filename for the LHEF file.
      static std::string getFilename(const OptionMap& options, bool addSeed = true);

   protected:
      friend class Singleton<LHEFInterface>;
      LHEFInterface(const OptionMap& optionMap, std::string fileName);
      ~LHEFInterface() { finalize(); }

   private:
      void finalize();

      /// Writes the header info to \a out.
      std::ostream& printHeader(std::ostream& out, const OptionMap& options) const;

      /// Writes the weight info for the header to \a out.
      std::ostream& printWeightInfo(std::ostream& out, const OptionMap& options) const;

      /// Writes the event info to \a out.
      std::ostream& printEvent(std::ostream& out, const LHEvent& event) const;

      /// Writes the footer info to \a out.
      std::ostream& printFooter(std::ostream& out) const;

      /// Normalizes the event weights in the file to the total cross section.
      void normalizeWeights() const;

      /// Determines the shower starting scale (scalup) for \a config.
      double getShowerScale(const Configuration& config) const;

      /// Sets the color flow for \a config.
      void setColorFlow(Configuration& config) const;
};

//------------------------------------------------------------------------------
/**
 * \brief Provides the default filename for the LHEF file.
 * \param options The options to use.
 * \param addSeed Whether to append the seed as tag (defaults to `true`).
 * \return The full filename including the path and extension.
 *
 * The full filename is constructed as
 * 'filePath::generate/fileStem_generateTag_seed.lhe[.gz]',
 * where the seed is only added if \a addSeed is `true`, and the .gz extension
 * applies when ZLIB is enabled.
 */
inline std::string LHEFInterface::getFilename(const OptionMap& options, bool addSeed)
{
   std::string name = format_path(options.getOption("filePath::generate"))
                      + options.getOption("fileStem")
                      + format_tag(options.getOption("generateTag"));
   if (addSeed)
      name += format_tag(options.getOption("seed"));
   name += ".lhe";
#ifdef geneva_USE_ZLIB
   name += ".gz";
#endif
   return name;
}

} // namespace Geneva

#endif // GENEVA_LHEF_INTERFACE_HPP
