//------------------------------------------------------------------------------
/// \file YamlToXmlTranslation.hpp
//
// Author(s):
//    Christian Bauer, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 LBNL, CERN, DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Translation of Yaml options in new format to Xml options in old format
//------------------------------------------------------------------------------

#ifndef GENEVA_YAML_TO_XML_TRANSLATION_HPP
#define GENEVA_YAML_TO_XML_TRANSLATION_HPP

#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/Utils/FileUtils.hpp"

namespace Geneva
{


//------------------------------------------------------------------------------
/**
 * \brief Retrieve an option for the yamlOptions
 *
 * Performs check that the option exists and throws error otherwise
 */
std::string getOption(const OptionMap& yamlOptions, std::string name)
{
   if (!yamlOptions.optionExists(name))
      Log::fatal("yaml options") << "Need to provide \"" << name << "\""<< Log::end;
   return yamlOptions.getOption(name);
}
//------------------------------------------------------------------------------
/**
 * \brief Converts the options from new (YAML) layout to old (XML) layout.
 *
 * This requires all options to be present. If an option is not present,
 * an error is thrown.
 */
OptionMap convertYAMLToXML(const OptionMap& yamlOptions)
{
   OptionMap xmlOptions;

   // global options
   // First obtain calculation, generator, and ME provider
   // TODO: this assumes each process has an entry calculation, phase_space, matrix_element_provider
   std::string processName = "process::" + getOption(yamlOptions, "global::process");
   if (processName != "process::pp_V")
      Log::fatal("yaml options") << "Currently only allowed value for process is [pp_V]. File contains " << processName << Log::end;

   std::string calcName = "calculation::" + getOption(yamlOptions, processName + "::calculation");
   std::string genName = "phase_space::" + getOption(yamlOptions, processName + "::phase_space");
   std::string providerName = getOption(yamlOptions, processName + "::matrix_element_provider");

   if (yamlOptions.optionExists("input_output::generate_tag"))
      xmlOptions.createOption<std::string>("generateTag", getOption(yamlOptions, "input_output::generate_tag"));
   if (yamlOptions.optionExists("input_output::setup_tag"))
      xmlOptions.createOption<std::string>("setupTag", getOption(yamlOptions, "input_output::setup_tag"));
   bool hasNumberOfEvents = yamlOptions.optionExists("global::num_events");
   bool hasMaxTime = yamlOptions.optionExists("global::max_time");
   if (!hasNumberOfEvents && !hasMaxTime)
      Log::fatal("yaml options") << "Need to specify either global::num_events or global::max_time" << Log::end;
   if (hasNumberOfEvents)
      xmlOptions.createOption<std::string>("num", getOption(yamlOptions, "global::num_events"));
   if (hasMaxTime)
      xmlOptions.createOption<std::string>("maxTime", getOption(yamlOptions, "global::max_time"));

   // input_output options
   if (yamlOptions.optionExists("input_output::file_path::setup"))
      xmlOptions.createOption<std::string>("filePath::setup", getOption(yamlOptions, "input_output::file_path::setup"));
   if (yamlOptions.optionExists("input_output::file_path::generate"))
      xmlOptions.createOption<std::string>("filePath::generate", getOption(yamlOptions, "input_output::file_path::generate"));
   if (yamlOptions.optionExists("input_output::overwrite_output"))
      xmlOptions.createOption<std::string>("overwriteOutput", getOption(yamlOptions, "input_output::overwrite_output"));

   std::string verbosity = getOption(yamlOptions, "input_output::verbosity");
   if (verbosity == "verbose" || verbosity == "debug")
      xmlOptions.createOption<std::string>("verbose", "true");
   else
      xmlOptions.createOption<std::string>("verbose", "false");

   if (yamlOptions.optionExists("input_output::redirect_logs::file")) {
      xmlOptions.createOption<std::string>("logFile", getOption(yamlOptions, "input_output::redirect_logs::file"));
      if (yamlOptions.optionExists("input_output::redirect_logs::append"))
         xmlOptions.createOption<std::string>("logFile::append", getOption(yamlOptions, "input_output::redirect_logs::append"));
   } else {
      xmlOptions.createOption<std::string>("logFile", "");
   }

   // event_generation options
   bool useUnweighting = yamlOptions.optionExists("event_generation::unweighting::partial_relative") || yamlOptions.optionExists("event_generation::unweighting::partial_absolute");
   if (useUnweighting)
      xmlOptions.createOption<std::string>("EventGenerator", "AdaptivePartialUnweighted");
   else
      xmlOptions.createOption<std::string>("EventGenerator", "Adaptive");

   if (useUnweighting) {
      // absolute or relative unweighting
      bool useAbsolute = yamlOptions.optionExists("event_generation::unweighting::partial_absolute");
      bool useRelative = yamlOptions.optionExists("event_generation::unweighting::partial_relative");
      if (useAbsolute && useRelative)
         Log::fatal("yaml options") << "Can only define partial_absolute or partial_relative for event_generation::unweighting, not both." << Log::end;
      if (useRelative) {
         xmlOptions.createOption<std::string>("EventGenerator::AdaptivePartialUnweighted::targetWeight", "0.");
         xmlOptions.createOption<std::string>("EventGenerator::AdaptivePartialUnweighted::targetMultiplier", getOption(yamlOptions, "event_generation::unweighting::partial_relative"));
         if (yamlOptions.optionExists("event_generation::unweighting::estimate_precision"))
            xmlOptions.createOption<std::string>("EventGenerator::AdaptivePartialUnweighted::estimatePrecision", getOption(yamlOptions, "event_generation::unweighting::estimate_precision"));
      } else if (useAbsolute) {
         xmlOptions.createOption<std::string>("EventGenerator::AdaptivePartialUnweighted::targetWeight", getOption(yamlOptions, "event_generation::unweighting::partial_absolute"));
         xmlOptions.createOption<std::string>("EventGenerator::AdaptivePartialUnweighted::targetMultiplier", "0.");
      } else
         Log::fatal("yaml options") << "Need to provide event_generation::unweighting::partial_absolute or event_generation::unweighting::partial_relative" << Log::end;
   }

   xmlOptions.createOption<std::string>("seed", getOption(yamlOptions, "event_generation::random::seed"));

   // event_analysis options
   bool useAnalyzer =  yamlOptions.optionExists("event_analysis::analyzer");
   if (useAnalyzer) {
      std::string analyzerName = getOption(yamlOptions, "event_analysis::analyzer");
      if (analyzerName == "Rates")
         xmlOptions.createOption<std::string>("Analyzer", "Rates");
      else
         Log::fatal("yaml options") << "Allowed value for event_analysis::analyzer is [Rates]. File contains " << analyzerName << Log::end;
   } else {
      if (yamlOptions.optionExists("event_analysis") && getOption(yamlOptions, "event_analysis") != "off")
         Log::fatal("yaml options") << "Allowed value for event_analysis is [off]. File contains " << getOption(yamlOptions, "event_analysis") << Log::end;
      xmlOptions.createOption<std::string>("Analyzer", "None");
   }

   // process options
   if (processName == "process::pp_V") {
      // intitial_state
      std::string initialStateName = processName + "::initial_state";
      xmlOptions.createOption<std::string>("Collider", getOption(yamlOptions, initialStateName + "::beams"));
      xmlOptions.createOption<std::string>("Ecm", getOption(yamlOptions, initialStateName + "::Ecm"));
      // PDF provider options
      xmlOptions.createOption<std::string>("PDFSet", getOption(yamlOptions, initialStateName + "::pdf_provider::LHAPDF::set"));
      xmlOptions.createOption<std::string>("PDF", "LHAPDF");

      // final_state
      std::string finalStateName = processName + "::final_state";
      std::string bosonType = getOption(yamlOptions, finalStateName + "::boson_type");
      xmlOptions.createOption<std::string>("Process::DrellYan::BosonType", bosonType);
      if (bosonType == "Z") {
         xmlOptions.createOption<std::string>("ZMass", getOption(yamlOptions, finalStateName + "::boson_mass"));
         xmlOptions.createOption<std::string>("ZWidth", getOption(yamlOptions, finalStateName + "::boson_width"));
         std::string decayType = getOption(yamlOptions, finalStateName + "::decay");
         if (decayType == "e+e-")
            xmlOptions.createOption<std::string>("Process::DrellYan::DecayChannel", "electrons");
         else if (decayType == "mu+mu-")
            xmlOptions.createOption<std::string>("Process::DrellYan::DecayChannel", "muons");
         else
            Log::fatal("yaml options") << "Only allowed values for final_state::decay are [e+e-, mu+mu-]. File contains " << decayType << Log::end;
      } else
         Log::fatal("yaml options") << "Only allowed value for final_state::boson_type is [Z]. File contains " << bosonType << Log::end;
   }
   else
      Log::fatal("yaml options") << "Currently only allowed value for process is [pp_V]. File contains " << processName << Log::end;

   // calculation options
   if (calcName == "calculation::SCETppV012") {
      // set the calculation
      xmlOptions.createOption<std::string>("calc", "CalcSCETppV012");

      // calculation::SCETppV012 options
      // precision
      std::string precision = getOption(yamlOptions, calcName + "::precision");
      if (precision == "NNLO+NNLL0+NLL1") {
         xmlOptions.createOption<std::string>("CalcSCETppV012::fixed0jet", "NLO");
         xmlOptions.createOption<std::string>("CalcSCETppV012::fixed1jet", "NLO");
         xmlOptions.createOption<std::string>("CalcSCETppV012::run", "NNLL");
         xmlOptions.createOption<std::string>("CalcSCETppV012::run1", "NLL");
         xmlOptions.createOption<std::string>("CalcSCETppV012::run1::split1Jet", "true");
         xmlOptions.createOption<std::string>("CalcSCETppV012::ImprovedXSFix", "true");
      } else
         Log::fatal("yaml options") << "Allowed value for calculation::SCETppV012::precision is [NNLO+NNLL0+NLL1]. File contains " << precision << Log::end;

      // scale_settings
      if (yamlOptions.optionExists(calcName + "::scale_settings::fixed_order::fixed")) {
         xmlOptions.createOption<std::string>("RenormalizationScale::Type", "Fixed");
         xmlOptions.createOption<std::string>("FactorizationScale::Type", "Fixed");
         xmlOptions.createOption<std::string>("RenormalizationScale::Fixed::Value", getOption(yamlOptions, calcName + "::scale_settings::fixed_order::fixed"));
         xmlOptions.createOption<std::string>("FactorizationScale::Fixed::Value", getOption(yamlOptions, calcName + "::scale_settings::fixed_order::fixed"));
      } else if (yamlOptions.optionExists(calcName + "::scale_settings::fixed_order::dynamic")) {
         xmlOptions.createOption<std::string>("RenormalizationScale::Type", getOption(yamlOptions, calcName + "::scale_settings::fixed_order::dynamic"));
         xmlOptions.createOption<std::string>("FactorizationScale::Type", getOption(yamlOptions, calcName + "::scale_settings::fixed_order::dynamic"));
      }

      // couplings
      if (yamlOptions.optionExists(calcName + "::couplings::alpha_s")) {
         // only allowed scalar is fromPDF
         if (getOption(yamlOptions, calcName + "::couplings::alpha_s") != "fromPDF")
            Log::fatal("yaml options") << "Unknown option for couplings::alpha_s" << Log::end;
         xmlOptions.createOption<std::string>("alphas::valueAtmZFromPDF", "true");
      } else {
         xmlOptions.createOption<std::string>("alphas::valueAtmZFromPDF", "false");
         xmlOptions.createOption<std::string>("alphas::value0", getOption(yamlOptions, calcName + "::couplings::alpha_s::value"));
         xmlOptions.createOption<std::string>("alphas::mu0", getOption(yamlOptions, calcName + "::couplings::alpha_s::scale"));
      }
      xmlOptions.createOption<std::string>("alphaem::value0", getOption(yamlOptions, calcName + "::couplings::alpha_em::value"));
      xmlOptions.createOption<std::string>("alphaem::mu0", getOption(yamlOptions, calcName + "::couplings::alpha_em::scale"));
      xmlOptions.createOption<std::string>("GFermi", getOption(yamlOptions, calcName + "::couplings::GF"));
      xmlOptions.createOption<std::string>("SinSquaredThetaW", getOption(yamlOptions, calcName + "::couplings::sin2W"));
   } else
      Log::fatal("yaml options") << "Currently only allowed value for calculation is [SCETppV012]. File contains " << calcName << Log::end;


   // phase-space generator options
   if (genName == "phase_space::PP2BosonJets") {
      // set the Generator
      xmlOptions.createOption<std::string>("Generator", "PP2BosonJets");

      // generation_cuts
      std::string cutsName = genName + "::generation_cuts";
      xmlOptions.createOption<std::string>("PP2BosonJets::BInvMassCutLow", getOption(yamlOptions, cutsName + "::boson_mass::min"));
      xmlOptions.createOption<std::string>("PP2BosonJets::BInvMassCutHigh", getOption(yamlOptions, cutsName + "::boson_mass::max"));
      bool useBosonRapidity = yamlOptions.optionExists(cutsName + "::boson_rapidity::min") || yamlOptions.optionExists(cutsName + "::boson_rapidity::max");
      if (useBosonRapidity) {
         xmlOptions.createOption<std::string>("PP2BosonJets::BRapidityCut", "true");
         xmlOptions.createOption<std::string>("PP2BosonJets::BRapidityCut::Low", getOption(yamlOptions, cutsName + "::boson_rapidity::min"));
         xmlOptions.createOption<std::string>("PP2BosonJets::BRapidityCut::High", getOption(yamlOptions, cutsName + "::boson_rapidity::max"));
      } else
            xmlOptions.createOption<std::string>("PP2BosonJets::BRapidityCut", "false");
   } else
      Log::fatal("yaml options") << "Currently only allowed phase_space is PP2BosonJets. File contains " << genName << Log::end;

   // matrix-element provider options
   if (providerName == "OpenLoops") {
      xmlOptions.createOption<std::string>("TreeME", "OpenLoops");
      xmlOptions.createOption<std::string>("TreeMEColor", "OpenLoops");
      xmlOptions.createOption<std::string>("TreeMEComplete", "OpenLoops");
      xmlOptions.createOption<std::string>("OneLoopME", "OpenLoops");
   } else if (providerName == "Custom") {
      std::string MEName = "matrix_element_provider";
      xmlOptions.createOption<std::string>("TreeME", getOption(yamlOptions, MEName + "::tree"));
      xmlOptions.createOption<std::string>("TreeMEColor", getOption(yamlOptions, MEName + "::tree_with_correlations"));
      xmlOptions.createOption<std::string>("TreeMEComplete", getOption(yamlOptions, MEName + "::tree_with_largeNc"));
      xmlOptions.createOption<std::string>("OneLoopME", getOption(yamlOptions, MEName + "::one_loop"));
   } else
      Log::fatal("yaml options") << "Allowed values for matrix_element_provider are [OpenLoops, Custom]. File contains " << providerName << Log::end;

   // any extra options provided in the yaml file
   int i = 0;
   while (true) {
      std::stringstream ss_extraOptionName;
      ss_extraOptionName << "extraOption_" << i++;
      if (yamlOptions.optionExists(ss_extraOptionName.str())) {
         std::string extraOption = yamlOptions.getOption<std::string>(ss_extraOptionName.str());
         std::stringstream ss;
         ss << extraOption;
         std::string name, value;
         ss >> name;
         std::string tmpVal;
         ss >> value;
         while (ss >> tmpVal)
            value += " " + tmpVal;
         if (xmlOptions.optionExists(name))
            xmlOptions.setOption(name, value);
         else
            xmlOptions.createOption<std::string>(name, value);
      }
      else
         break;
   }

   // Create a default run_name if left empty.
   std::string runname;
   if (yamlOptions.optionExists("global::run_name"))
      runname = yamlOptions.getOption("global::run_name");
   if (runname.empty()) {
      const std::string process = getOption(yamlOptions, "global::process");
      int Ecm;
      std::stringstream(yamlOptions.getOption("process::" + process + "::initial_state::Ecm")) >> Ecm;
      std::stringstream ecmtag;
      if (Ecm % 1000 == 0)
         ecmtag << Ecm / 1000 << "TeV";
      else
         ecmtag << Ecm << "GeV";

      runname = "gva" + format_tag(process) + format_tag(ecmtag.str());
      Log::info("yaml options") << "Using autogenerated run_name: " << runname << Log::end;
   }
   xmlOptions.createOption<std::string>("fileStem", runname);

   return xmlOptions;
}

} // namespace Geneva

#endif // GENEVA_YAML_TO_XML_TRANSLATION_HPP
