//------------------------------------------------------------------------------
/// \file GeneratorPP2BosonJets.hpp
//
// Author(s):
//    Simone Alioli, Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class GeneratorPP2BosonJets
//------------------------------------------------------------------------------

#ifndef GENEVA_GENERATOR_PP2BOSON_JETS_HPP
#define GENEVA_GENERATOR_PP2BOSON_JETS_HPP

#include "Geneva/Generators/IGenerator.hpp"

#include "Geneva/Core/Physics/SignatureSet.hpp"
#include "Geneva/Core/Physics/Configuration.hpp"

namespace Geneva
{

class UnitCubeRad;

//------------------------------------------------------------------------------
/**
 * \class GeneratorppBosonJets
 * \ingroup gen
 *
 * \brief Generator for pp -> B + Jets Configurations
 *
 * B denotes a color single boson or its decay products.
 */
//------------------------------------------------------------------------------
class GeneratorPP2BosonJets : virtual public IGenerator
{
   public:
      /// Sampling mode for choice of Signature.  By default it uses \c kRandom
      /// to access them randomly. Use \c kSequential
      /// to access them sequentially.
      enum SignatureAccessMode
      {
         kRandom,      /// Access Signatures randomly.
         kSequential /// Access Signaures sequentially.
      };

      /// Strategy to generate Configurations with many particles in the final state. By default it uses \c kRambo
      enum LargeMultiplicitiesGenerationMode
      {
         kRambo,
         kRecursiveFKS
      };

      /// Strategy to sample the invariant mass
      enum SamplingMass
      {
         kFlatMass,
         kBreitWignerMass
      };

      /// Strategy to sample the total cm energy tau = x1 x2
      enum SamplingTau
      {
         kFlatTau,
         kExponentialTau,
         kPowerTau,
         kSqrtTau
      };

      /// Strategy to sample the total rapidity y = log(x1/x2)
      enum SamplingY
      {
         kFlatY,
         kPowerY
      };

      /// Strategy to sample the remaining variables
      enum SamplingThetaJet
      {
         kFlatThetaJet,
         kImportanceThetaJet
      };

   private:
      double _lastWeight;              ///< the weight for the last generated Configuration
      SignatureSet _sigSet;            ///< the signatures we will generate
      Configuration _lastConfig;       ///< the last configuration generated
      double _cmEnergy, _massB, _widthB;
      PdgID _idB;
      int _nLight;
      double _lowBosonMassSquared, _highBosonMassSquared;
      SignatureAccessMode _sigAccess;
      LargeMultiplicitiesGenerationMode _genModeLargeN;
      SamplingMass _importanceSampleMass;
      SamplingTau _importanceSampleTau;
      SamplingY _importanceSampleY;
      SamplingThetaJet _importanceSampleThetaJet;
      double _ktJetCut;
      double _mJetJetCut;
      double _yBcutLow;
      double _yBcutHigh;
      size_t _signatureNumber; ///< keep track of the current  %Signature for sequential generation
      unsigned int _numGenerated, _numFail;
      bool _useBjjFKS;
      std::string _bjjTau1cut;
      double _tau1cut;

      bool _useTau0Generator;

   public:
      GeneratorPP2BosonJets(double Ecm, int nLight, double massB, double widthB, PdgID idB , double lowBMass, double HighBmass, double ktJetcut, double mJetJetCut, double yVcutLow, double yVcutHigh, SignatureAccessMode sigAccess, LargeMultiplicitiesGenerationMode genMode, SamplingMass importSampleMass, SamplingTau importSampleTau, SamplingY importSampleY, SamplingThetaJet importSampleThetaJet, bool useBjjFKS, bool useTau0Generator, std::string _bjjTau1cut = "none", double _tau1cut = 0.);

      GeneratorPP2BosonJets(const GeneratorPP2BosonJets&) = delete;
      GeneratorPP2BosonJets(GeneratorPP2BosonJets&&) = delete;
      GeneratorPP2BosonJets& operator=(const GeneratorPP2BosonJets&) = delete;
      GeneratorPP2BosonJets& operator=(GeneratorPP2BosonJets&&) = delete;

      ~GeneratorPP2BosonJets();

      /// \name IGenerator Interface
      ///@{
      /// See IGenerator for detailed documentation

      /// Returns the current set of Signatures used by the generator.
      virtual const SignatureSet& getSignatures() const;

      /// Sets the Signatures for which to generate Configurations.
      virtual bool setSignatures(const SignatureSet& sigSet);

      /// Adds a Signature to generate Configurations for.
      virtual bool addSignature(const Signature& sig);

      /// Checks if the generator can handle a Signature.
      virtual bool checkSignature(const Signature& sig) const;

      /// Generates a new Configuration.
      virtual bool generate(Configuration& config);

      /// Generates a new Configuration.
      virtual bool generate(const Signature& sig, std::vector<double> rand, Configuration& config);

      virtual size_t numberOfRandomsNeeded(const Signature& sig);

      /// Returns the weight of the last generated configuration.
      virtual double getLastWeight();
      ///@}

      /// Returns the name "GeneratorPP2BosonJets" of the Generator.
      virtual const std::string name() const;

      virtual void finalize();

   private:
      /// Generates a new Configuration with only B  in the final state, given external randoms. Sets the jacobian.
      bool generateB(const Signature& signature, Configuration& configuration, std::vector<double> rand, double& jac);

      /// Generates a new Configuration with B and 1 parton in the final state, given external randoms. Sets the jacobian.
      bool generateBj(const Signature& signature, Configuration& configuration, std::vector<double> rand, double& jac);

      /// Generates a new Configuration with B and 2 partons in the final state, given external randoms.Sets the jacobian.
      bool generateBjj(const Signature& signature, Configuration& config, std::vector<double> rand, double& jac);

      struct _solver_type;
      _solver_type* _solver;

      /// Generates a new Configuration with B and 1 parton in the final state, given external randoms. It is using Tau0 as one of the random variables.
      bool generateBjTau0(const Signature& signature, Configuration& configuration, std::vector<double> rand, double& jac);

      /// Generates a new Configuration  with B and more than 2 extra parton in the final state using recursive FKS splittings.
      bool generateBjjRecursiveFKS(const Signature& signature, Configuration& config, std::vector<double> rand, double& jac);

      /// Generates a new Configuration  with B and more than 2 extra parton in the final state using RAMBO (CURRENTLY UNTESTED) .
      bool generateBjetsRAMBO(const Signature& signature, Configuration& config, std::vector<double> rand, double& jac);

      // helper functions
      double generateInvMassSquared(double ran, bool doDecay, double& massSquared);

      double generateTau(double ranTau, double InvMass1Squared,  double InvMass2Squared, double transverseMomentum, double& tau);

      double generateTau0(double ranTau0, double mV,  double expAbsY, double& tau0);

      double generateZ(double ranZ, double mV,  double expY, double tau0, std::string rad, double& z);
      double getZMin(double mV,  double expY, double tau0, std::string rad);

      double generateY(double ranY, double tau, double yMin, double yMax, double& y);

      void setxBarAxBarB(double tau, double y, double& xBarA, double& xBarB);

      bool includesDecay(const Signature& sig) const;

      double sampleCosTheta(double ran, double sampling, double& cosTheta);

      double bosonRapidity(const Configuration& config);

      bool split(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, Configuration& configNplus1CM, Configuration::iterator& emittedCM, Configuration::iterator& sisterCM, double& xA, double& xB, double& jacobian);
};

} // namespace Geneva

#endif // GENEVA_GENERATOR_PP2BOSON_JETS_HPP
