//------------------------------------------------------------------------------
/// \file GeneratorUtils.hpp
//
// Author(s):
//    Simon Berman, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2016 LBNL, DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class GeneratorUtils
//------------------------------------------------------------------------------

#ifndef GENEVA_GENERATOR_UTILS_HPP
#define GENEVA_GENERATOR_UTILS_HPP

#include "Geneva/Core/Physics/Signature.hpp"
#include "Geneva/Core/Physics/FourVector.hpp"
#include "Geneva/Core/Physics/Particle.hpp"
#include "Geneva/Core/Physics/PdgID.hpp"

#include <gsl/gsl_roots.h>

#include <vector>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief Utitlity functions needed by several generators.
 *
 * Includes functions for converting massless particle four momenta to massive
 * particle four momenta, boosting massless partcile four momenta into a frame in
 * which energy conservation holds, and functions for generating color information.
 */
//------------------------------------------------------------------------------
class GeneratorUtils
{
   public:
      /// Tests if the given Signature is of the form e+e-.
      static bool isEe(const Signature& sig);

      /**
       * \brief Calculates the value xi required to included masses
       *
       * Follows implementation by Ellis, Kleiss, and Stirling.
       */
      static double findxi(double Ecm, std::vector<double> E, std::vector<double> m, gsl_root_fdfsolver* s);

      /// Uses xi to appropriately adjust FVs to include mass
      static std::vector<FourVector> massMomentaFactor(double xi, std::vector<double> m, std::vector<FourVector>& zeroMass);

      /// Calculates factor for mass correction in differential weight
      static double massWeightFactor(std::vector<FourVector> k, double Ecm, size_t numGenerated);

      /// Boosts massless four momenta into a frame in which they satisfy energy conservation
      static std::vector<FourVector> physicalFV(std::vector<FourVector> q, double Ecm);

      static int getNewColorLabel(int& numColorLabel);

      static bool isValidSplit(PdgID idi, PdgID idj, int nTriplet);

      // Generate the color information of a color history.
      static void generateColorFlowHistory(const Signature& sig, std::vector<int>& color, std::vector<int>& antiColor, int& numColorLabel);

   private:
      struct _splitting_type;

      static void generateColorFlow(_splitting_type* mother, std::vector<int>& color, std::vector<int>& antiColor, int& numColorLabel);

      // Create the color history and return the start of the linked list.
      static _splitting_type* createColorHistory(const Signature& sig, std::vector<int>& color, std::vector<int>& antiColor);

      // Create the color history and return the start of the linked list directly using a Configuration
      static _splitting_type* createColorHistory(Configuration config);

      // Generate the color information of an element of a colorHist.
      static void generateColor(_splitting_type* color, int& numColorLabel);
};

} // namespace Geneva

#endif // GENEVA_GENERATOR_UTILS_HPP
