//------------------------------------------------------------------------------
/// \file IGenerator.hpp
//
// Author(s):
//    Erik Strand, Frank Tackmann, Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class IGenerator
//------------------------------------------------------------------------------

#ifndef GENEVA_I_GENERATOR_HPP
#define GENEVA_I_GENERATOR_HPP

#include <string>
#include <vector>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \defgroup gen Generators
 *
 * Collection of classes dealing with the generation of phase space.
 */
//------------------------------------------------------------------------------

class Configuration;
class SignatureSet;
class Signature;

//------------------------------------------------------------------------------
/**
 * \class IGenerator
 * \ingroup gen
 *
 * \brief Abstract interface to all phase-space generators
 *
 * Defines the abstract interface to all phase-space generators. A generator
 * generates a phase-space configuration with an associated generation weight.
 * The weight represents the probability with which the generator generated the
 * configuration.
 *
 */
//------------------------------------------------------------------------------
class IGenerator
{
   public:
      virtual ~IGenerator() = default;

      /// \name Handling of Signatures
      ///@{

      /// Returns the current set of Signatures used by the generator.
      virtual const SignatureSet& getSignatures() const = 0;

      /**
       * \brief Sets all Signatures for which to generate Configurations.
       * \param sigSet The SignatureSet to be used.
       * \return \c true if the generator can handle all Signatures, \c false
       *         otherwise.
       *
       * Replaces the current Signatures of the generator by those in \a sigSet.
       * Returns \c true if the generator is able to generate Configurations
       * from all signatures in \a sigSet. If \a sigSet contains at least one
       * Signature the generatore is not able to handle, \c false is returned
       * and none of the Signatures is set.
       */
      virtual bool setSignatures(const SignatureSet& sigSet) = 0;

      /**
       * \brief Adds a Signature to generate Configurations for.
       * \param sig The Signature to be added.
       * \return \c true if the generator can handle the Signature, \c false
       *         otherwise.
       *
       * Adds \a sig to the list of Signatures used by the generator. If the
       * generator is not able to handle \a sig, it is not added and \c false
       * is returned.
       *
       * \note: If \a sig is already part of the generators Signatures, it is up
       * to each generator whether to add \a sig again or not. The return value
       * should be \c true regardless, so it is always equivalent to that of
       * checkSignature().
       */
      virtual bool addSignature(const Signature& sig) = 0;

      /**
       * \brief Checks if the generator can handle a Signature.
       * \param sig The Signature to be checked.
       * \return \c true if the generator can handle the Signature, \c false
       *         otherwise.
       *
       * The return value is guaranteed to be the same as that of addSignature().
       */
      virtual bool checkSignature(const Signature& sig) const = 0;
      ///@}

      /// \name Generation
      ///@{

      /**
       * \brief Generates a new Configuration.
       * \param[out] config Place to store the generated Configuration.
       * \return \c true if a valid Configuration has been generated, \c false
       *         otherwise.
       *
       * The newly generated Configuration is stored in \a config, overwriting
       * any possible previous content.
       *
       * If the return value is \c false, the generated Configuration was
       * invalid (e.g. did not pass generation cuts). In this case, the content
       * of \a config is undefined (\todo make it well-defined).
       *
       * An error is thrown in case of errors during the generation.
       */
      virtual bool generate(Configuration& config) = 0;

      /**
       * \brief Generates a new Configuration.
       * \param[in] sig The signature to be generated.
       * \param[in] rand The random numbers to be used in the generation.
       * \param[out] config Place to store the generated Configuration.
       *
       * \return \c true if a valid Configuration has been generated, \c false
       *         otherwise.
       *
       * The newly generated Configuration is stored in \a config, overwriting
       * any possible previous content.
       *
       * The generation uses the random numbers passed in via \a rand.
       *
       * If the return value is \c false, the generated Configuration was
       * invalid (e.g. did not pass generation cuts). In this case, the content
       * of \a config is undefined (\todo make it well-defined).
       *
       * An error is thrown in case of errors during the generation.
       */
      virtual bool generate(const Signature& sig, std::vector<double> rand, Configuration& config) = 0;

      /**
       * \brief Determines the number of random numbers needed to generate the Configuration
       * \param sigSize The size of the signature to be generated.
       * \return The number of random numbers needed.
       */
      virtual size_t numberOfRandomsNeeded(const Signature& sig) = 0;

      /**
       * \brief Returns the weight of the last generated configuration.
       * \return The generation weight.
       *
       * The returned weight corresponds to the last successful call to generate().
       */
      virtual double getLastWeight() = 0;
      ///@}

      /// Returns the name of the generator.
      virtual const std::string name() const = 0;

      /// Finalizes the generator, prints the statistics etc.
      virtual void finalize() {};

   protected:
      IGenerator() = default;
      IGenerator(const IGenerator&) = default;
      IGenerator(IGenerator&&) = default;
      IGenerator& operator=(const IGenerator&) = default;
      IGenerator& operator=(IGenerator&&) = default;
};

} // namespace Geneva

#endif // GENEVA_I_GENERATOR_HPP
