//------------------------------------------------------------------------------
/// \file ISmartRandom.hpp
//
// Author(s):
//    Christian Bauer
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SmartRandomMint.hpp
//------------------------------------------------------------------------------

#ifndef GENEVA_I_SMART_RANDOM_HPP
#define GENEVA_I_SMART_RANDOM_HPP

#include <vector>
#include <string>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class ISmartRandom
 * \ingroup Generators
 *
 * \brief Interface for a Smart Random Generator
 *
 * Generates Random numbers, while being able to optimize based on a given function.
 * Keeps track of the integral of the function.
 *
 * Once the importance sampling has been set up
 * \code
 *     bool nextRandom(std::vector<double>& rand, double& vol)
 *     result = f(rand)
 *     informOfCalculation(result * vol)
 * \endcode
 * This will fill the vector rand with random numbers, and hands back the
 * volume of the cell from which the random number is drawn from.
 *
 * To set up a grid using importance sampling, one does a sequence of calls
 * from somewhere that has a pointer to mint
 * \code
 *     stream = mint->initialize(fileStem);
 *     for (int it = 0; it < numIterations; it++) {
 *        mint->resetOptimizeParameters();
 *        for (int call = 0; i < numCalls; i++) {
 *           mint->nextRandom(rand, vol);
 *           funct = function(rand);
 *           mint->informOfCalculation(funct / vol);
 *        }
 *        mint->optimize(it, numCalls, stream);
 *     }
 *     mint->finalize(stream, numIterations);
 * \endcode
 * It is important to reset the optimization parameters after each iteration.
 *
 * The obtained grids can be saved to a file. These files can be read
 * in later and will set up the grid exactly how it was when written out.
 *
 * To use SmartRandomMint with a weighted event generator, you just have to obtain random
 * numbers using the getNextRandom() function.
 */
//------------------------------------------------------------------------------
class ISmartRandom
{
   public:
      virtual ~ISmartRandom() = default;

      /// Obtain the next set of random numbers. prob is the probability with which the random number was generated
      virtual bool nextRandom(std::vector<double>& rand, double& prob) = 0;
      /// Get the overall weight of the SmartRandom
      virtual double overallWeight() = 0;

      /// \name Functions to set up the optimization
      ///@{
      /// Initializes the optimization
      virtual void initializeOptimization() = 0;
      /// Inform the SmartRandom of the result for the function to be integrated
      virtual void informOfCalculation(double) = 0;
      /// Inform the SmartRandom that an error has occurred
      virtual void informOfFailure() = 0;
      /// Optimizes SmartRandom using the information tracked internally
      virtual void optimize() = 0;
      /// Finalizes the optimization
      virtual void finalizeOptimization() = 0;
      ///@}

      /// \name Functions that store the resulting grids and bounds
      ///@{
      /// Writes the optimization information to \a out.
      virtual bool storeOptimization(std::ostream& out) = 0;
      /// Reads the optimization information from \a in.
      virtual bool loadOptimization(std::istream& in) = 0;
      ///@}

      /// \name Functions needed for the folding of random numbers
      ///@{
      /// Turns on folding in a given dimension with number of foldings = numFolding
      virtual void setFolding(const size_t dim, const size_t numFolding) = 0; ///< Set the folding in a given dimension to a given number. This has to happen right after setting up SmartRandomMint
      /// Obtain the next set of random numbers for folded variables
      virtual bool nextRandomFolded(std::vector<double>& rand, double& vol) = 0;
      ///@}

      /// \name Functions that return stored information
      ///@{
      /// Returns the dimension
      virtual size_t dimension() = 0;
      /// Returns the name
      virtual std::string name() = 0;
      /// Returns whether the function is optimized or not
      virtual bool isOptimized() = 0;

   protected:
      ISmartRandom() = default;
      ISmartRandom(const ISmartRandom&) = default;
      ISmartRandom(ISmartRandom&&) = default;
      ISmartRandom& operator=(const ISmartRandom&) = default;
      ISmartRandom& operator=(ISmartRandom&&) = default;
};

} // namespace Geneva

#endif // defined GENEVA_I_SMART_RANDOM_HPP
