//------------------------------------------------------------------------------
/// \file IntegrationChannel.hpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class IntegrationChannel
//------------------------------------------------------------------------------

#ifndef GENEVA_INTEGRATION_CHANNEL_HPP
#define GENEVA_INTEGRATION_CHANNEL_HPP

#include "Geneva/Core/Physics/SignatureSet.hpp"

#include <map>
#include <string>
#include <vector>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class IntegrationChannel
 *
 * \brief A class that implements an integration channel
 */
//------------------------------------------------------------------------------
class IntegrationChannel
{

   private:
      SignatureSet _signatures;                   ///< The set of %Signatures in the channel
      std::map<Signature, double> _weight;        ///< The total weight associated to each %Signature
      std::map<Signature, double> _weightSq;      ///< The total weightSq associated to each %Signature
      size_t _numEvents;                          ///< The total number of events
      std::string _name;                          ///< The name of the channel
      size_t _signatureSize;                      ///< The size of the %Signatures in the channel NOTE: all %Signatures must have the same size
      std::vector<double> _cumulativeWeights;     ///< The cumulative weights for the signatures (normalized to total weight)

   public:
      IntegrationChannel(const SignatureSet& sigSet, const std::string& name);

      SignatureSet signatures();                 ///< Returns a copy of the %SignatureSet
      size_t signatureSize();

      double weight(const Signature& sig) const;               ///< Returns the weight of a given Signature
      double weightSq(const Signature& sig) const;             ///< Returns the weightSq of a given Signature
      size_t numEvents() const;                                ///< Returns the total number of events

      void setWeight(const Signature& sig, double weight);     ///< Set weight of a given Signature
      void setWeightSq(const Signature& sig, double weightSq); ///< Set weightSq of weight to a given Signature
      void setNumEvents(size_t numEvents);
      void addWeight(const Signature& sig, double weight);     ///< Add a weight to a given Signature

      double crossSection(const Signature& sig) const;         ///< Returns the cross-section for a given %Signature
      double crossSectionError(const Signature& sig) const;    ///< Returns the error of the cross-section for a given %Signature

      double crossSection() const;                             ///< Returns the total cross-section in the channel (sum of %Signature's cross sections)
      double crossSectionError() const;                        ///< Returns the total cross-section error in the channel (sum of %Signature's cross sections)

      double totalWeight() const;                              ///< Returns the total weight of the channel (sum of absolute values of %Signature's cross sections)

      double probability(const Signature& sig) const;          ///< Returns the probability for a given %Signature

      bool operator<(const IntegrationChannel& other) const;

      const std::string name() const;

      /**
       * \brief Picks a signature based on the relative probabilities
       * \return A pair with a %Signature and the probability with which it is picked.
       */
      std::pair<Signature, double> pickSignature() const;

      void initialize();
      void reset();

      /// \name Output
      ///@{
      // IntegrationChannel stream insertion operator.
      friend std::ostream& operator<<(std::ostream& out, const IntegrationChannel& ch);
      ///@}

   private:
      /// Streams the %IntegrationChannel to \a output.
      std::ostream& print(std::ostream& output) const;
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline IntegrationChannel::IntegrationChannel(const SignatureSet& sigSet, const std::string& name)
   : _signatures(sigSet), _name(name)
{
   _signatureSize = sigSet.begin()->size();
}

//------------------------------------------------------------------------------
inline SignatureSet IntegrationChannel::signatures()
{
   return _signatures;
}

//------------------------------------------------------------------------------
inline size_t IntegrationChannel::signatureSize()
{
   return _signatureSize;
}

//------------------------------------------------------------------------------
inline double IntegrationChannel::weight(const Signature& sig) const
{
   return _weight.find(sig)->second;
}

//------------------------------------------------------------------------------
inline double IntegrationChannel::weightSq(const Signature& sig) const
{
   return _weightSq.find(sig)->second;
}

//------------------------------------------------------------------------------
inline size_t IntegrationChannel::numEvents() const
{
   return _numEvents;
}

//------------------------------------------------------------------------------
inline void IntegrationChannel::setWeight(const Signature& sig, double weight)
{
   _weight[sig] = weight;
}

//------------------------------------------------------------------------------
inline void IntegrationChannel::setWeightSq(const Signature& sig, double weightSq)
{
   _weightSq[sig] = weightSq;
}

//------------------------------------------------------------------------------
inline void IntegrationChannel::setNumEvents(size_t numEvents)
{
   _numEvents = numEvents;
}

//------------------------------------------------------------------------------
inline void IntegrationChannel::addWeight(const Signature& sig, double weight)
{
   _weight[sig] += weight;
   _weightSq[sig] += weight * weight;
   _numEvents++;
}

//------------------------------------------------------------------------------
inline double IntegrationChannel::crossSection(const Signature& sig) const
{
   if (_numEvents == 0)
      return 0.;
   else
      return _weight.find(sig)->second / static_cast<double>(_numEvents);
}

//------------------------------------------------------------------------------
inline double IntegrationChannel::crossSectionError(const Signature& sig) const
{
   double avWeight = _weight.find(sig)->second / static_cast<double>(_numEvents);
   double avWeightSq = _weightSq.find(sig)->second / static_cast<double>(_numEvents);
   return sqrt(abs(avWeight * avWeight - avWeightSq) / static_cast<double>(_numEvents));
}

//------------------------------------------------------------------------------
inline double IntegrationChannel::crossSection() const
{
   double result = 0.;
   for (std::map<Signature, double>::const_iterator it = _weight.begin(); it != _weight.end(); ++it) {
      result += it->second;
   }
   return result / static_cast<double>(_numEvents);
}

//------------------------------------------------------------------------------
inline double IntegrationChannel::crossSectionError() const
{
   double result = 0.;
   for (std::map<Signature, double>::const_iterator it = _weight.begin(); it != _weight.end(); ++it) {
      result += it->second;
   }
   double resultSq = 0.;
   for (std::map<Signature, double>::const_iterator it = _weightSq.begin(); it != _weightSq.end(); ++it) {
      resultSq += it->second;
   }
   double avResult = result / static_cast<double>(_numEvents);
   double avResultSq = resultSq / static_cast<double>(_numEvents);
   return sqrt(abs(avResult * avResult - avResultSq) / static_cast<double>(_numEvents));
}

//------------------------------------------------------------------------------
inline double IntegrationChannel::totalWeight() const
{
   return _cumulativeWeights.back();
}

//------------------------------------------------------------------------------
inline double IntegrationChannel::probability(const Signature& sig) const
{
   size_t pos = size_t_from_ptrdiff_t(std::distance(_signatures.begin(), _signatures.find(sig)));
   return (pos == 0) ? _cumulativeWeights[0] / _cumulativeWeights.back() : (_cumulativeWeights[pos] - _cumulativeWeights[pos - 1]) / _cumulativeWeights.back();
}

//------------------------------------------------------------------------------
inline bool IntegrationChannel::operator<(const IntegrationChannel& rhs) const
{
   if (_signatures == rhs._signatures)
      return name() < rhs.name();
   else
      return _signatures < rhs._signatures;
}

//------------------------------------------------------------------------------
inline const std::string IntegrationChannel::name() const
{
   return _name;
}

//------------------------------------------------------------------------------
// Output
//------------------------------------------------------------------------------
/**
 * \brief %IntegrationChannel stream insertion operator.
 * \param out The output stream to be used.
 * \param ch The %IntegrationChannel to be streamed.
 * \return A reference to \a out.
 */
inline std::ostream& operator<<(std::ostream& out, const IntegrationChannel& ch)
{
   return ch.print(out);
}

}

#endif // GENEVA_INTEGRATION_CHANNEL_HPP
