//------------------------------------------------------------------------------
/// \file SmartRandomHandler.hpp
//
// Author(s):
//    Christian Bauer, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017, 2018 Geneva Collaboration
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SmartRandomHandler
//------------------------------------------------------------------------------

#ifndef GENEVA_SMART_RANDOM_HANDLER_HPP
#define GENEVA_SMART_RANDOM_HANDLER_HPP

#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/Utils/FileUtils.hpp"

#include <vector>
#include <map>
#include <string>

namespace Geneva
{

class ISmartRandom;
class IntegrationChannel;

//------------------------------------------------------------------------------
/**
 * \class SmartRandomHandler
 * \ingroup gen
 *
 * \brief Handler for SmartRandoms
 *
 * This class manages a number of SmartRandoms. It owns the SmartRandoms and is
 * reponsible for their construction and destruction. It also keeps track of
 * whether the SmartRandoms have been setup or not.
 */
//------------------------------------------------------------------------------
class SmartRandomHandler
{
   public:
      enum SmartRandomType
      {
         kMint
      };

      SmartRandomHandler() = default;
      SmartRandomHandler(const SmartRandomHandler&) = delete;
      SmartRandomHandler(SmartRandomHandler&&) = delete;
      SmartRandomHandler& operator=(const SmartRandomHandler&) = delete;
      SmartRandomHandler& operator=(SmartRandomHandler&&) = delete;
      ~SmartRandomHandler();

      // Creates a new SmartRandom for channel.
      bool makeSmartRandom(SmartRandomType type, const IntegrationChannel& channel,
                           size_t dimension, size_t numBins);

      // Obtains random numbers from the SmartRandom for channel.
      void getRandom(const IntegrationChannel& channel, std::vector<double>& rand, double& prob);

      // Informs the SmartRandom for channel of the result of the calculation.
      void informOfCalculation(const IntegrationChannel& channel, double calcResult);

      // Informs the SmartRandom for channel that the calculation failed.
      void informOfFailure(const IntegrationChannel& channel);

      /// Returns \c true if the SmartRandom for \a channel is set up.
      bool isSetup(const IntegrationChannel& channel);

      /// Returns \c true if all SmartRandoms are set up.
      bool isSetup();

      // Initializes the optimization for all SmartRandoms.
      void initializeOptimization();

      // Optimizes all SmartRandoms.
      void optimize();

      // Finalizes the optimization for all SmartRandoms.
      void finalizeOptimization();

      // Reads the optimization for all SmartRandoms from files.
      bool loadOptimization(const std::string& fileStem);

      // Writes the optimization for all SmartRandoms to files.
      void storeOptimization(const std::string& fileStem, const std::string& seedTag);

      // Provides the default filestem to be used for the grid files.
      static std::string getGridStem(const OptionMap& options);

   private:
      typedef std::map<std::string, ISmartRandom*> _impl_type;
      typedef _impl_type::iterator _iterator;

      typedef std::map<ISmartRandom*, bool> _impl_setup_type;
      typedef _impl_setup_type::iterator _setup_iterator;

      _impl_type _smart_randoms;       ///< Stores all the SmartRandoms.
      _impl_setup_type _is_setup;      ///< Stores whether a SmartRandom is set up or not.

      /// Gets a pointer to the ISmartRandom for \a channel.
      ISmartRandom* _get_smart_random(const IntegrationChannel& channel);
};

//------------------------------------------------------------------------------
/**
 * \brief Provides the default filestem to be used for the grid files.
 * \param options The options to use.
 * \return The full filestem including the path.
 *
 * If the Mint::fileStem option is nonempty it is used for the file stem.
 * Otherwise the returned file stem is constructed from the \a options as
 * 'fileStem_setupTag_Mint::fileTag_grid'.
 *
 * In each case the filePath::setup path is prepended.
 */
inline std::string SmartRandomHandler::getGridStem(const OptionMap& options)
{
   std::string gridStem = options.getOption("Mint::fileStem");
   if (gridStem.empty())
      gridStem = options.getOption("fileStem")
                 + format_tag(options.getOption("setupTag"))
                 + format_tag(options.getOption("Mint::fileTag"))
                 + format_tag("grid");
   return format_path(options.getOption("filePath::setup")) + gridStem;
}

} // namespace Geneva

#endif // GENEVA_SMART_RANDOM_HANDLER_HPP
