//------------------------------------------------------------------------------
/// \file SmartRandomMint.hpp
//
// Author(s):
//    Christian Bauer
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class SmartRandomMint.hpp
//------------------------------------------------------------------------------

#ifndef GENEVA_SMART_RANDOM_MINT_HPP
#define GENEVA_SMART_RANDOM_MINT_HPP

#include "Geneva/Generators/ISmartRandom.hpp"

#include <vector>
#include <string>

namespace Geneva
{

class Event;
//------------------------------------------------------------------------------
/**
 * \class SmartRandomMint
 * \ingroup Generators
 *
 * \brief Implementation of a SmartRandom based on the Vegas or MINT algorithm
 *
 * This implements the algorithm of Nason, described in hep-ph/0709.2085
 *
 * The SmartRandomMint class is responsible for creating a set of random numbers
 * that are then later used to calculate a function with these numbers as
 * input. MINT is obtaining these numbers using a binned grid for each
 * dimenion. The result from this function can be passed back into MINT, which
 * can then adjust its grid according to that information.
 *
 * The algorithm is based on the MINT algorithm by Nason in the above
 * paper, but the implementation into Geneva is quite different. In particular,
 * Mint as derived in this class does not provide the complete implementation
 * of the event generator, but it has to be used in conjunction with the
 * event generator EventGeneratorMint
 *
 * SmartRandomMint needs to be initialized with the number of dimensions that are integrated
 * over and the number of bins used to sample each dimension. It does not need
 * to know about the function it is integrating.
 */
//------------------------------------------------------------------------------
class SmartRandomMint: public ISmartRandom
{
   protected:
      //Global information about SmartRandomMint
      const std::string _name;  ///< The name of SmartRandomMint
      size_t _dimension;     ///< The dimension of the function to be integrated
      size_t _numbins;       ///< The number of bins for each dimension
      bool _hasChanged;    ///< Whether SmartRandomMint has received any new optimization information since the last call to optimize (or the beginning)

      //Information that is needed to perform importance sampling
      std::vector< std::vector <double> > _xGrid;  ///< The grid according to which I sample the random points. It has _numBins+1 entries for the grid edges
      std::vector< std::vector <double> > _cumIntegral;  ///< The fraction of the total integral at a given xrid location. It has _numBins+1 entries for the grid edge
      std::vector< std::vector <double> > _cellIntegral;  ///< The total integral in each cell
      std::vector< std::vector <size_t> > _cellCount;  ///< The total number of points thrown in each cell
      std::vector< std::vector <double> > _cumBinProb;  ///< The cumulative probabilty with which each cell is sampled.

      size_t _iteration;   ///< The current iteration

      //Information about dealing with folded variables
      std::vector<size_t> _foldingInfo;  ///< The information on which dimensions are folded
      std::vector<size_t> _currentFolding;  ///< The vector telling me the current folding being used
      bool _firstFolding;           ///< True if we need a new set of folding

      //Information to hold the set of random numbers used
      std::vector<size_t> _currentBin;   ///< The current bin in a given dimension
      std::vector<size_t> _currentFoldedBin;   ///< The current bin in a given folded dimension
      std::vector<double> _posInBin;  ///<The position within the bin in a given dimension
      std::vector<double> _randomPoint;   ///<The value of the random number in a given dimension

      //Information about the result of SmartRandomMint
      double _weightPrevious;     ///< The average weight of the function in the previous iteration to setup the grids
      double _weightSqPrevious;   ///< The average weight^2 of the function in the previous iteration to setup the grids
      double _weightNow;           ///< The average weight of the function in this iteration to setup the grids
      double _weightSqNow;         ///< The average weight^2 of the function n this iteration to setup the grids
      size_t _numEventsPrevious;///< The number of events used in the previous iteration to setup the grids
      size_t _numEventsNow;     ///< The number of events used in this iteration to setup the grids

      bool _gridExists;         ///< A flag saying if the grid exists
      bool _wasInformed;        ///< Whether SmartRandomMint received information back about the last point generated
      bool _inGridSetup;        ///< Is true if we are currently setting up the grid

   public:
      /// Creates a SmartRandomMint of the given dimensions and number of bins.
      SmartRandomMint(size_t dimension, size_t numBins);

      /// Obtain the next set of random numbers for unfolded variables.
      virtual bool nextRandom(std::vector<double>& rand, double& prob);
      /// Get the overall weight of the SmartRandom
      virtual double overallWeight();

      /// \name Functions needed to set up the grid
      ///@{
      /// Initializes the Grid before importance sampling can begin
      virtual void initializeOptimization();
      /// Inform the grid of the result for the function to be integrated
      virtual void informOfCalculation(double);
      /// Inform the grid an error has occurred
      virtual void informOfFailure();
      /// Adjusts the grid using the information tracked internally
      virtual void optimize();
      /// Finalizes the Grid after importance sampling is finished
      virtual void finalizeOptimization();

      /// \name Functions that store the resulting grids and bounds
      ///@{
      // Writes the grid information to out.
      virtual bool storeOptimization(std::ostream& out);
      // Reads the grid information from in.
      virtual bool loadOptimization(std::istream& in);
     ///@}

      ///@}
      /// \name Functions needed for the folding of random numbers
      ///@{
      /// Turns on folding in a given dimension with number of foldings = numFolding
      virtual void setFolding(const size_t dim, const size_t numFolding); ///< Set the folding in a given dimension to a given number. This has to happen right after setting up SmartRandomMint
      /// Obtain the next set of random numbers for folded variables
      virtual bool nextRandomFolded(std::vector<double>& rand, double& jac);
      ///@}

      /// \name Functions that return stored information
      ///@{
      /// Returns the dimension
      virtual size_t dimension() {return _dimension;};
      /// Returns the name
      virtual std::string name() {return _name;};
      /// Returns whether the function is optimized or not
      virtual bool isOptimized() {return (not _inGridSetup);};

   private:
      /// Resets all grid parameters after each iteration
      void resetGridParameters();
      double getBinSize(const size_t dim, const size_t bin); ///< Calculate the bin size in a given dimension
      bool findNextFolding(); //Get the next folding, returs false if next folding is not available any more
      /// Access whether SmartRandomMint has been updated
      bool hasChanged() {return _hasChanged;};

      void setCumIntegral(std::vector< std::vector <double> >cellIntegral, std::vector< std::vector <size_t> > cellCount);
      void setNewxGrid(std::vector< std::vector <double> >xGridOld, std::vector< std::vector <double> > cumIntegral);
};

} // namespace Geneva

#endif // GENEVA_SMART_RANDOM_MINT_HPP
