//------------------------------------------------------------------------------
/// \file PythiaDriver.hpp
//
// Author(s):
//    Christian Bauer
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class PythiaDriver
//------------------------------------------------------------------------------

#ifndef GENEVA_PYTHIA_DRIVER_HPP
#define GENEVA_PYTHIA_DRIVER_HPP

#include "Geneva/Core/Physics/Event.hpp"
#include "Geneva/Core/Physics/Sigma.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"

#include "Geneva/PythiaInterface/UserHookRestriction.hpp"

#include "Pythia8/Pythia.h"

#include <sstream>
#include <string>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \class PythiaDriver
 * \ingroup PythiaInterface
 *
 * \brief Handles the showering of events in Geneva
 *
 * The class is in charge of setting up the Pythia class in a way that is
 * consistent with the Geneva events being showered. It requires as an input
 * a LHEF file generated by Geneva, and optionally a Pythia option file. All
 * Pythia settings are contained in the second file.
 *
 * It ensures that none of the Pythia options chosen interfere with showering
 * an event file produced by Geneva.
 *
 * After the class is set up, one can retrieve the Pythia object, either as
 * const or non-const. There are two ways of generating the events contained
 * in the LHEF file
 *
 * 1) Call the next() routine of the Pythia driver, and read out the results
 *    through a const Pythia object which can be obtained from the class
 * 2) Obtain a non-const version of the Pythia object, and then call next()
 *    directly on that.
 *
 * The advantage of the first is that one can be sure that no accidental changes
 * can be made to the Pythia object, since it is const. The disadvantage is that
 * Pythia is written in a way that many public functions do not work on a const
 * object due to poor coding.
 *
 * For this reason we are providign the second option. Once the Pythia object
 * has been retrieved, one can use it like in any other Pythia program, but
 * the Pythia driver can no longer guarantee that it remains correctly set up.
 */
//------------------------------------------------------------------------------
class PythiaDriver
{
   private:
      Pythia8::Pythia _pythia;

      OptionMap _optMap;

      size_t _numEvents;                   ///< The number of events in the LHE file

      size_t _nPythiaEvents;               ///< The total number of events seen so far
      size_t _nPythiaFailAllowed;          ///< The number of times the shower is allowed to fail before terminating
      size_t _nPythiaFail;                 ///< The number of times the shower has failed so far
      size_t _nOverVeto;                   ///< The number of times that an event has over-vetoed from the UHR
      std::map<size_t, size_t> _nOverVetoFromMultiplicity; ///< The breakdown of overvetoes according to partonic multiplicity

      UserHookRestriction* _userhook;      ///< The userhook to veto pythia at the parton level

      Sigma _lastWeight;                   ///< Stores the Sigma of the last showered event

   public:
      PythiaDriver(const std::string& lhefInputFile, const std::string& genevaOptionsFile, bool enforceLhefOptions, const std::string& pyOptionsFile,
                   size_t maxEvents, double fractionPythiaFailAllowed);

      PythiaDriver(const PythiaDriver&) = delete;
      PythiaDriver(PythiaDriver&&) = delete;
      PythiaDriver& operator=(const PythiaDriver&) = delete;
      PythiaDriver& operator=(PythiaDriver&&) = delete;

      ~PythiaDriver();

      /// Return a const reference to Pythia8
      const Pythia8::Pythia& getPythia() const;
      /// Return a reference to Pythia8
      Pythia8::Pythia& getPythia();

      // The two static functions that return information required for Pythia8 object consistent with Calculation
      static UserHookRestriction* getUserHook(const std::string& lfefInputFile, const std::string& genevaOptionsFile, bool enforceLhefOptions = true);
      static UserHookRestriction* getUserHook(std::istream& lhefInputStream, const std::string& genevaOptionsFile, bool enforceLhefOptions = true);

      static std::vector<std::string> getRequiredSettings(const std::string& lhefInputFile, const std::string& genevaOptionsFile, bool enforceLhefOptions = true);
      static std::vector<std::string> getRequiredSettings(std::istream& lhefInputStream, const std::string& genevaOptionsFile, bool enforceLhefOptions = true);

      const OptionMap& getOptions() const;  ///< Returns the internally stored option map

      size_t numEvents() const;             ///< Returns the total number of events to be read from the LHE file

      /// Returns \c true as long as the maximum number of events or the end of the file is not reached.
      bool hasNext() const;

      bool next();                          ///< Showers the next event

      /// Returns the weight of the last event.
      const Sigma& lastWeight() const;

      /// Returns the last event before showering.
      Event getPartonicEvent() const;

      /// Returns the last event after showering but before hadronization.
      Event getShoweredEvent() const;

      /// Returns the last event after showering with possible MPI contributions removed.
      Event getShoweredNoMPIEvent() const;

      /// Returns the last event after showering and hadronization.
      Event getHadronizedEvent() const;

      /// Returns the final event in Pythia8 format.
      const Pythia8::Event& pythiaEvent() const;

      void stat(); ///< Print the statistics info

   private:
      void init(const std::string& lhefInputFile, const std::string& genevaOptionsFile, bool enforceLhefOptions, const std::string& pyOptionsFile, bool verbose);
      void setRequiredOptions(const std::vector<std::string>& requiredOptions);
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline const Pythia8::Pythia& PythiaDriver::getPythia() const
{
   return _pythia;
}

//------------------------------------------------------------------------------
inline Pythia8::Pythia& PythiaDriver::getPythia()
{
   return _pythia;
}

//------------------------------------------------------------------------------
inline const OptionMap& PythiaDriver::getOptions() const
{
   return _optMap;
}

//------------------------------------------------------------------------------
inline size_t PythiaDriver::numEvents() const
{
   return _numEvents;
}

//------------------------------------------------------------------------------
inline bool PythiaDriver::hasNext() const
{
   return (_nPythiaEvents < _numEvents) && !_pythia.info.atEndOfFile();
}

//------------------------------------------------------------------------------
inline const Sigma& PythiaDriver::lastWeight() const
{
   return _lastWeight;
}

//------------------------------------------------------------------------------
inline const Pythia8::Event& PythiaDriver::pythiaEvent() const
{
   return _pythia.event;
}

} //namespace Geneva

#endif // GENEVA_PYTHIA_DRIVER_HPP
