//------------------------------------------------------------------------------
/// \file ShowerScaleSCETppV012.hpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class ShowerScaleSCETppV012
//------------------------------------------------------------------------------

#ifndef GENEVA_SHOWER_SCALE_SCET_PPV012_HPP
#define GENEVA_SHOWER_SCALE_SCET_PPV012_HPP

#include "Geneva/Core/Physics/NJettiness.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \struct ShowerScaleSCETppV012
 *
 * \brief Determines the effective NJettiness showering scale as relevant for
 * the SCETppV012 calculation.
 *
 * \note This is used by both generation (LHEF interface) and showering (UHR).
 * The implementation is inlined to avoid link dependence between both parts.
 */
//------------------------------------------------------------------------------
class ShowerScaleSCETppV012
{
   private:
      size_t _maxFinal;
      double _tau0cut;
      double _tau1cut;
      double _inclFactor;
      double _inclMin;

   public:
      ShowerScaleSCETppV012(size_t maxFinal, double Tau0cut, double Tau1cut,
                            double inclFactor = 1.0, bool inclMinCut = true)
         : _maxFinal(maxFinal), _tau0cut(Tau0cut), _tau1cut(Tau1cut),
           _inclFactor(inclFactor), _inclMin(0.)
      {
         assert(maxFinal <= 2);
         if (inclMinCut) {
            if (maxFinal == 1)
               _inclMin = _tau0cut;
            else if (maxFinal == 2)
               _inclMin = _tau1cut;
         }
      }

      double getTauScale(const Configuration& config) const
      {
         NJettinessppRecursive TauFR(config);
         const size_t numFinalQCD = TauFR.numQCD();

         // exclusive bins
         if (numFinalQCD < _maxFinal) {
            if (numFinalQCD == 0)
               return _tau0cut;
            if (numFinalQCD == 1) {
               const double tau0 = ZeroJettinesspp(config)();
               if (tau0 > _tau0cut)
                  return _tau1cut;
               else
                  return _tau0cut;
            }
         }

         // inclusive bin
         assert(numFinalQCD == _maxFinal);

         const bool allowqqbar = (numFinalQCD == 2) && !config.includes(PdgID(PdgID::kGluon));
         Configuration::iterator emitted, sister;
         // Note: This relies on the fact that the minimum clustering metric for
         // N+1 massless particles directly gives TauN
         const double tauScale = _inclFactor * TauFR.clusterQCDAllowed(emitted, sister, allowqqbar);
         if (tauScale > _inclMin)
            return tauScale;
         return _inclMin;
      }

      double getpTScale(const Configuration& config) const
      {
         const double Q = 2. * sqrt(config.initialA()->energy() * config.initialB()->energy());
         const double pTScale = sqrt(getTauScale(config) * Q);
         return pTScale;
      }
};

} // namespace Geneva

#endif // GENEVA_SHOWER_SCALE_SCET_PPV012_HPP
