//------------------------------------------------------------------------------
/// \file UHRppB012.hpp
//
// Author(s):
//    Christian Bauer, Chris Vermilion, Calvin Berggren
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class UHRppB012
//------------------------------------------------------------------------------

#ifndef GENEVA_UHR_PP_B012_HPP
#define GENEVA_UHR_PP_B012_HPP

#include "Geneva/PythiaInterface/UserHookRestriction.hpp"

namespace Geneva
{

class IRestriction;
class Configuration;

//------------------------------------------------------------------------------
/**
 * \class UHRppB012
 * \ingroup pythia
 *
 * \brief UserHookRestriction for pp -> Boson + 0/1/2 jets
 *
 * The UHR vetoes ensures that Pythia does not shower out of the jet
 * multiplicities, essentially mimicking a Tau-ordered shower.
 *
 * We distinguish between the full veto and the step veto. The full veto is
 * applied after the whole event has been showered, while the step veto is
 * applied after each emission.
 *
 * The showering attempt is vetoed if Tau_PY > Tau_Hard, where Tau_PY and
 * Tau_Hard are determined based on the parton multiplicity of the hard event.
 *
 * To avoid a hard cutoff, the constraint is smoothed by a factor lambda as
 * Tau_PY > (1 + randomInterval() * lambda) * Tau_Hard
 *
 * We use several counters to keep track of how often each veto is applied:
 * Veto if number of particles after showering is less than maxFinal
 * NLO veto (numPY <= maxFinal)
 *    nveto[0]:  numHard = 0
 *    nveto[1]:  numHard = 1
 *    nveto[2]:  numHard = 2
 * Full Veto
 *    nveto[3]:  numHard = 0
 *    nveto[4]:  numHard = 1
 *    nveto[5]:  numHard = 2
 * Step Veto
 *    nveto[6]:  numHard = 0
 *    nveto[7]:  numHard = 1
 *    nveto[8]:  numHard = 2
 * where numHard is the number of partons in the hard process.
 */
//------------------------------------------------------------------------------
class UHRppB012 : public UserHookRestriction
{
   private:
      IRestriction* _restriction;
      size_t _maxFinal;
      double _tau0cut;
      double _tau1cut;
      double _lambda;
      bool _removeMPI;
      double _pTmaxMPI;
      size_t _maxISRstep;
      size_t _maxFSRstep;
      size_t _numHardFinal;
      double _yBoost;
      double _tau0;
      double _tauShowerScale;
      bool _useTauClust;
      bool _sumTauClust;
      bool _useHardDirections;
      double _tau1MaxRatio;

      std::vector<size_t> _nveto;

   public:
      UHRppB012(IRestriction* restriction, size_t maxAttempts, size_t maxFinal,
                double tau0cut, double tau1cut, double lambda, bool removeMPI, double pTmaxMPI,
                size_t maxISRstep, size_t maxFSRstep, bool useTauClust, bool sumTauClust,
                bool useHardDirections, double tau1MaxRatio);

      virtual std::ostream& printStatistics(std::ostream& out);

   private:
      virtual void setHardObservables();
      virtual bool vetoPartonLevelEarly(const Pythia8::Event& event);
      virtual bool vetoStep(int iPos, int nISR, int nFSR, const Pythia8::Event& event);

      bool fractionalRejection(double TauPY, double TauHard);
};

} // namespace Geneva

#endif // GENEVA_UHR_PP_B012_HPP
