//------------------------------------------------------------------------------
/// \file UserHookRestriction.hpp
//
// Author(s):
//    Christian Bauer, Chris Vermilion, Calvin Berggren
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Interface of class UserHookRestriction
//------------------------------------------------------------------------------

#ifndef GENEVA_USER_HOOK_RESTRICTION_HPP
#define GENEVA_USER_HOOK_RESTRICTION_HPP

#include "Geneva/Core/Physics/Configuration.hpp"

#include "Pythia8/UserHooks.h"
#include "Pythia8/Event.h"

#include <iosfwd>

namespace Geneva
{

//------------------------------------------------------------------------------
/** \class UserHookRestriction
 *  \ingroup pythia
 *
 * A user hook restriction provides an implementation for a veto on the Pythia
 * shower which is based on the calculation used in Geneva. This is a base class
 * from which other UHR's can be derived.
 */
//------------------------------------------------------------------------------
class UserHookRestriction : public Pythia8::UserHooks
{
   private:
      size_t _numVetoes;               ///< Total number of vetoes
      size_t _numVetoesEvent;          ///< Number of vetoes for current event
      size_t _maxAttempts;             ///< Maximum number of showering attempts per event

      Configuration _hardProcess;      ///< Hard process before any showering
      Pythia8::Event _partonLevel;     ///< Event after showering but before hadronization
      bool _init;                      ///< Whether _partonLevel has been initialized yet

   public:
      UserHookRestriction(size_t maxAttempts = 0);

      /// \name Pythia8::UserHooks interface
      /// @{
      virtual bool canVetoProcessLevel() { return true; };
      virtual bool doVetoProcessLevel(Pythia8::Event& event);

      virtual bool canVetoPartonLevelEarly() { return true; };
      virtual bool doVetoPartonLevelEarly(const Pythia8::Event& event);
      virtual bool retryPartonLevel() { return true; }

      virtual bool canVetoStep() { return true; }
      virtual int numberVetoStep() { return 1000000; }
      virtual bool doVetoStep(int iPos, int nISR, int nFSR, const Pythia8::Event& event);
      /// @}

      size_t numVetoes() const;
      size_t numVetoesEvent() const;

      const Pythia8::Event& partonLevel() const;
      const Configuration& hardProcess() const;

      bool isOverVetoed() const;

      virtual std::ostream& printStatistics(std::ostream& out) { return out; };

   private:
      virtual void setHardObservables() {};
      virtual bool vetoPartonLevelEarly(const Pythia8::Event& event) = 0;
      virtual bool vetoStep(int iPos, int nISR, int nFSR, const Pythia8::Event& event) = 0;
};


//------------------------------------------------------------------------------
/** \class UserHookRestrictionNone
 *  \ingroup pythia
 *
 *  Implementation of a UserHookRestriction that does nothing
 */
class UserHookRestrictionNone : public UserHookRestriction
{
   public:
      UserHookRestrictionNone() {}
      virtual void printStatistics() {}

   protected:
      virtual bool vetoPartonLevelEarly(const Pythia8::Event& event) { return false; }
      virtual bool vetoStep(int iPos, int nISR, int nFSR, const Pythia8::Event& event) { return false; }
};


////////////////////////////////////////////////////////////////////////////////
// Inline Definitions
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
inline UserHookRestriction::UserHookRestriction(size_t maxAttempts)
   : _numVetoes(0), _numVetoesEvent(0), _maxAttempts(maxAttempts), _init(false)
{}

//------------------------------------------------------------------------------
inline size_t UserHookRestriction::numVetoes() const                  { return _numVetoes; }
inline size_t UserHookRestriction::numVetoesEvent() const             { return _numVetoesEvent; }
inline const Pythia8::Event& UserHookRestriction::partonLevel() const { return _partonLevel; }
inline const Configuration& UserHookRestriction::hardProcess()  const { return _hardProcess; }

//------------------------------------------------------------------------------
inline bool UserHookRestriction::isOverVetoed() const
{
   return _numVetoesEvent >= _maxAttempts;
}

} // namespace Geneva

#endif // GENEVA_USER_HOOK_RESTRICTION_HPP
