//------------------------------------------------------------------------------
/// \file InterpolateConv.hpp
//
// Author(s):
//    Jon Walsh, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017  DESY
//
//    This file is part of the SCETlib library. SCETlib is distributed under the
//    terms of the GNU General Public License version 3 (GPLv3), see the COPYING
//    file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class InterpolateConv
//------------------------------------------------------------------------------

#include "beamfunc/InterpolateConv.hpp"
#include "beamfunc/ConvOptions.hpp"

#include "LHAPDF/PDF.h"

#include <iostream>

namespace beamfunc
{

namespace
{

// ----------------------------------------------------------------------
// Function to convert enum Convname to string
// ----------------------------------------------------------------------
std::string ToString(Convname convname)
{
   switch (convname) {
      case k_fx: return "fx";
      // quark QCD convolutions
      // LO
      case kPqx_fx: return "Pqx_fx";
      // NLO
      case kP1qqV_fq: return "P1qqV_fq";
      case kP1qqbarV_fq: return "P1qqbarV_fq";
      case kP1qqS_fq: return "P1qqS_fq";
      case kP1qg_fg: return "P1qg_fg";
      // LO x LO
      case kPqqPqx_fx: return "PqqPqx_fx";
      case kPqgPgx_fx: return "PqgPgx_fx";

      // quark Tau convolutions
      // NLO
      case kIqx_fx: return "Iqx_fx";
      // NLO x LO
      case kIqqPqx_fx: return "IqqPqx_fx";
      case kIqgPgx_fx: return "IqgPgx_fx";
      // NNLO
      case kI2qqV_fq: return "I2qqV_fq";
      case kI2qqbarV_fq: return "I2qqbarV_fq";
      case kI2qqS_fq: return "I2qqS_fq";
      case kI2qg_fg: return "I2qg_fg";

      default: return "[Unknown ConvName]";
   }
}

// ----------------------------------------------------------------------
// Function by which convolution is rescaled
// ----------------------------------------------------------------------
double RescaleFactor(double x)
{
   return pow(x,1.5)*log(x);
   //return 1.0;
}

// ----------------------------------------------------------------------
// Function by which fq and fg are rescaled
// ----------------------------------------------------------------------
double RescaleFactorPDF(double x)
{
   return x;
}

} // anonymous namespace

//------------------------------------------------------------------------------
InterpolateConv::InterpolateConv(Convname convname, const std::string pdfset, int imem, const std::string method)
{
   int v = ::LHAPDF::verbosity();
   ::LHAPDF::setVerbosity(0);

   const std::string tmp = ::LHAPDF::getConfig().get_entry("Interpolator");
   ::LHAPDF::getConfig().set_entry("Interpolator", method);
   _conv = ::LHAPDF::mkPDF(pdfset + '_' + ToString(convname), imem);
   ::LHAPDF::getConfig().set_entry("Interpolator", tmp);

   ::LHAPDF::setVerbosity(v);

   _ispdf = (convname == k_fx);
}

//------------------------------------------------------------------------------
InterpolateConv::~InterpolateConv()
{
   delete _conv;
}

//------------------------------------------------------------------------------
// Function to perform interpolation and return value at given x and muB
double InterpolateConv::interpolate(int pid, double x, double muB)
{
   double prefactor;
   if (_ispdf)
      prefactor = RescaleFactorPDF(x);
   else
      prefactor = RescaleFactor(x);

   if ((pid != 0 && std::abs(pid) <= 5) || pid == kGluon || pid == kSinglet)
      return (1./prefactor) * _conv->xfxQ2(pid, x, muB*muB);

   std::cerr << "beamfunc: Not a valid particle ID. Returning 0.0" << std::endl;
   return 0.0;
}

} // namespace beamfunc
