//------------------------------------------------------------------------------
/// \file TauCoeffsQuark.cpp
//
// Author(s):
//    Jon Walsh, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017  DESY
//
//    This file is part of the SCETlib library. SCETlib is distributed under the
//    terms of the GNU General Public License version 3 (GPLv3), see the COPYING
//    file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of classes TauCoeffsQuark
//------------------------------------------------------------------------------

#include "beamfunc/beamfunc.hpp"

#include "beamfunc/TauCoeffsQuark.hpp"

#include "LHAPDF/LHAPDF.h"

#include <iostream>
#include <cassert>

namespace beamfunc
{

namespace
{

//------------------------------------------------------------------------------
// constants
//------------------------------------------------------------------------------
struct math_constants
{
   // pi
   static double pi() { return 3.14159265358979; }
   // zeta3
   static double zeta3() { return 1.20205690315959; }
   // ln2
   static double ln2() { return 0.693147180559945; }
};

//------------------------------------------------------------------------------
struct QCD
{
   // color factors
   static double CA() { return 3.; }
   static double CF() { return 4./3.; }
   static double TF() { return 0.5; }

   // beta0, beta1
   static double beta0(int nf) { return 11./3.*CA() - 4./3.*TF()*nf; }
   static double beta1(int nf) { return 34./3.*CA()*CA() - 20./3.*CA()*TF()*nf - 4.*CF()*TF()*nf; }

   // Gamma0, Gamma1
   static double Gamma0(int nf) { return 4.; }
   static double Gamma1(int nf) { return 4. * ((67./9. - math_constants::pi() * math_constants::pi() / 3.) * CA() - 20./9. * TF()*nf); }

   // quark anomalous dimensions: gammaBq0, gammaBq1
   static double gammaBq0(int nf) { return 6. * CF(); }
   static double gammaBq1(int nf) { return CF() * ((146./9. - 80. * math_constants::zeta3()) * CA() + (3. - 4. * math_constants::pi() * math_constants::pi() + 48. * math_constants::zeta3()) * CF() + (121./9. + math_constants::pi() * math_constants::pi() * 2./3.) * beta0(nf)); }
};

} // anonymous namespace

//------------------------------------------------------------------------------
TauCoeffsQuark::TauCoeffsQuark(int order, const std::string pdfset, int imem, int nf)
   : _nf(nf)
{
   // Make sure there is a valid dummy alphas in the global config, because
   // lhapdf >= 6.2 creates the PDF's alphas eagerly and fails otherwise.
   if (!::LHAPDF::getConfig().has_key("AlphaS_Type")) {
      ::LHAPDF::getConfig().set_entry("AlphaS_Type", "analytic");
      ::LHAPDF::getConfig().set_entry("AlphaS_Lambda5", -1.0);
   } else if (::LHAPDF::getConfig().get_entry("AlphaS_Type") == "analytic" &&
              !(::LHAPDF::getConfig().has_key("AlphaS_Lambda5")
                || ::LHAPDF::getConfig().has_key("AlphaS_Lambda4")
                || ::LHAPDF::getConfig().has_key("AlphaS_Lambda3"))) {
      ::LHAPDF::getConfig().set_entry("AlphaS_Lambda5", -1.0);
   }

   std::string lhapdf_path;
   const char* env_cstr = std::getenv("GENEVA_DATA_DIR");
   if (env_cstr != 0 && *env_cstr != '\0')
      lhapdf_path = std::string(env_cstr) + "/beamfunc";
   else
      lhapdf_path = std::string(BEAMFUNC_DATA_DIR);
   lhapdf_path += "/" + pdfset + "_beamfunc/";

   // read LHAPDF_DATA_PATH and if not defined older LHAPATH
   const char* lhapdf_path_cstr = std::getenv("LHAPDF_DATA_PATH");
   if (lhapdf_path_cstr == 0)
      lhapdf_path_cstr = getenv("LHAPATH");
   if (lhapdf_path_cstr != 0) {
      lhapdf_path += ':';
      lhapdf_path += std::string(lhapdf_path_cstr);
   }

   ::LHAPDF::setPaths(lhapdf_path);

   _convsQCD = new QCDQuarkConvs(order, pdfset, imem, "logcubic");
   _convsTau = new TauQuarkConvs(order, pdfset, imem, "logcubic");
}

//------------------------------------------------------------------------------
TauCoeffsQuark::~TauCoeffsQuark()
{
   delete _convsTau;
   delete _convsQCD;
}

//------------------------------------------------------------------------------
// calculate a coefficient for a given point
double TauCoeffsQuark::Biqj(int order, int logorder, int pid, double x, double muB)
{
   double result = 0;
   double prefactor = 1. / (4. * math_constants::pi()); // prefactor conventions, include (1/4pi)^n

   if (order == 0) {
      if (logorder == -1) { result = B0qm1(pid, x, muB); }
      else {
         std::cout << "at LO, only log order = -1 defined!" << std::endl;
         assert(false);
      }

   } else if (order == 1) {
      switch (logorder) {
         case - 1:
            result = prefactor * B1qm1(pid, x, muB);
            break;
         case 0:
            result = prefactor * B1q0(pid, x, muB);
            break;
         case 1:
            result = prefactor * B1q1(pid, x, muB);
            break;
         default:
            std::cout << "at NLO, only log order = -1, 0, 1 defined!" << std::endl;
            break;
      }
   } else if (order == 2) {
      switch (logorder) {
         case - 1:
            result = prefactor * prefactor * B2qm1(pid, x, muB);
            break;
         case 0:
            result = prefactor * prefactor * B2q0(pid, x, muB);
            break;
         case 1:
            result = prefactor * prefactor * B2q1(pid, x, muB);
            break;
         case 2:
            result = prefactor * prefactor * B2q2(pid, x, muB);
            break;
         case 3:
            result = prefactor * prefactor * B2q3(pid, x, muB);
            break;
         default:
            std::cout << "at NNLO, only log order = -1, 0, 1, 2, 3 defined!" << std::endl;
            break;
      }
   } else {
      std::cout << "only order = 0, 1, 2 supported!" << std::endl;
      assert(false);
   }
   return result;
}

// LO, delta function
double TauCoeffsQuark::B0qm1(int pid, double x, double muB)
{
   double fq = _convsQCD->convValue(k_fx, pid, x, muB);
   return fq;
}

// NLO, delta function
double TauCoeffsQuark::B1qm1(int pid, double x, double muB)
{
   double Iqq1delXfq = _convsTau->convValue(kIqx_fx, pid, x, muB);
   double Iqg1delXfg = _convsTau->convValue(kIqx_fx, kGluon, x, muB);
   return 2. * QCD::CF() * Iqq1delXfq + 2. * QCD::TF() * Iqg1delXfg;
}

// NLO, L0
double TauCoeffsQuark::B1q0(int pid, double x, double muB)
{
   double PqqXfq = _convsQCD->convValue(kPqx_fx, pid, x, muB);
   double PqgXfg = _convsQCD->convValue(kPqx_fx, kGluon, x, muB);
   double fq = _convsQCD->convValue(k_fx, pid, x, muB);
   double gamma0term = -0.5 * QCD::gammaBq0(_nf) * fq;
   return 2. * QCD::CF() * PqqXfq + 2. * QCD::TF() * PqgXfg + gamma0term;
}

// NLO, L1
double TauCoeffsQuark::B1q1(int pid, double x, double muB)
{
   double fq = _convsQCD->convValue(k_fx, pid, x, muB);
   return QCD::CF() * QCD::Gamma0(_nf) * fq;
}

// NNLO, delta function
double TauCoeffsQuark::B2qm1(int pid, double x, double muB)
{
   double I2qqVXfq = _convsTau->convValue(kI2qqV_fq, pid, x, muB);
   double I2qqbarVXfq = _convsTau->convValue(kI2qqbarV_fq, -pid, x, muB);
   double I2qqSXfq = _convsTau->convValue(kI2qqS_fq, kSinglet, x, muB);
   double I2qgXfg = _convsTau->convValue(kI2qg_fg, kGluon, x, muB);

   return 4. * QCD::CF() * (I2qqVXfq + I2qqbarVXfq + I2qqSXfq) + 4 * QCD::TF() * I2qgXfg;
}

// NNLO, L0
double TauCoeffsQuark::B2q0(int pid, double x, double muB)
{
   double fq = _convsQCD->convValue(k_fx, pid, x, muB);
   double PqqXfq = _convsQCD->convValue(kPqx_fx, pid, x, muB);
   double PqgXfg = _convsQCD->convValue(kPqx_fx, kGluon, x, muB);
   double Iqq1delXfq = _convsTau->convValue(kIqx_fx, pid, x, muB);
   double Iqg1delXfg = _convsTau->convValue(kIqx_fx, kGluon, x, muB);
   // combined NNLO Quark Convolutions
   double P1qqVXfq = _convsQCD->convValue(kP1qqV_fq, pid, x, muB);
   double P1qqbarVXfq = _convsQCD->convValue(kP1qqbarV_fq, -pid, x, muB);
   double P1qqSXfq = _convsQCD->convValue(kP1qqS_fq, kSinglet, x, muB);
   double P1qgXfg = _convsQCD->convValue(kP1qg_fg, kGluon, x, muB);
   double Pqi1Xfi = QCD::CF() * (P1qqVXfq + P1qqbarVXfq + P1qqSXfq) + QCD::TF() * P1qgXfg;

   double Iqq1delXPqqXfq = _convsTau->convValue(kIqqPqx_fx, pid, x, muB);
   double Iqq1delXPqgXfg = _convsTau->convValue(kIqqPqx_fx, kGluon, x, muB);
   double Iqg1delXPgqXfq = _convsTau->convValue(kIqgPgx_fx, kSinglet, x, muB);
   double Iqg1delXPggXfg = QCD::CA() * (_convsTau->convValue(kIqgPgx_fx, kGluon, x, muB)) + 0.5*QCD::beta0(_nf) * Iqg1delXfg;
   double Iqi1delXPijXfj = QCD::CF() * QCD::CF() * Iqq1delXPqqXfq + QCD::CF() * QCD::TF() * Iqg1delXPgqXfq + QCD::TF() * QCD::CF() * Iqq1delXPqgXfg + QCD::TF() * Iqg1delXPggXfg;
   double pisq = math_constants::pi() * math_constants::pi();

   return (math_constants::zeta3() * QCD::CF() * QCD::CF() * QCD::Gamma0(_nf) * QCD::Gamma0(_nf) - (1. / 2.) * QCD::gammaBq1(_nf)
            + (1. / 12.) * pisq * QCD::gammaBq0(_nf) * QCD::CF() * QCD::Gamma0(_nf)) * fq
          - (1. / 3.) * pisq * QCD::CF() * QCD::Gamma0(_nf) * QCD::CF() * PqqXfq
          - (1. / 3.) * pisq * QCD::CF() * QCD::Gamma0(_nf) * QCD::TF() * PqgXfg
          - (2 * QCD::beta0(_nf) + QCD::gammaBq0(_nf)) * QCD::CF() * Iqq1delXfq
          - (2 * QCD::beta0(_nf) + QCD::gammaBq0(_nf)) * QCD::TF() * Iqg1delXfg
          + 4. * Pqi1Xfi
          + 4. * Iqi1delXPijXfj;
}

// NNLO, L1
double TauCoeffsQuark::B2q1(int pid, double x, double muB)
{
   double fq = _convsQCD->convValue(k_fx, pid, x, muB);
   double PqqXfq = _convsQCD->convValue(kPqx_fx, pid, x, muB);
   double PqgXfg = _convsQCD->convValue(kPqx_fx, kGluon, x, muB);
   double Iqq1delXfq = _convsTau->convValue(kIqx_fx, pid, x, muB);
   double Iqg1delXfg = _convsTau->convValue(kIqx_fx, kGluon, x, muB);
   // combined NNLO Quark Convolutions
   double PqqXPqqXfq = _convsQCD->convValue(kPqqPqx_fx, pid, x, muB);
   double PqqXPqgXfg = _convsQCD->convValue(kPqqPqx_fx, kGluon, x, muB);
   double PqgXPgqXfq = _convsQCD->convValue(kPqgPgx_fx, kSinglet, x, muB);
   double PqgXPggXfg = QCD::CA() * (_convsQCD->convValue(kPqgPgx_fx, kGluon, x, muB)) + 0.5*QCD::beta0(_nf) * PqgXfg;
   double PqiXPijXfj = QCD::CF() * QCD::CF() * PqqXPqqXfq + QCD::CF() * QCD::TF() * PqgXPgqXfq + QCD::CF() * QCD::TF() * PqqXPqgXfg + QCD::TF() * PqgXPggXfg;
   double pisq = math_constants::pi() * math_constants::pi();

   return (((-1. / 6.) * pisq * QCD::CF() * QCD::CF() * QCD::Gamma0(_nf) * QCD::Gamma0(_nf) + QCD::CF() * QCD::Gamma1(_nf)
               + (1./4.) * QCD::gammaBq0(_nf) * QCD::gammaBq0(_nf) + (1./2.) * QCD::beta0(_nf) * QCD::gammaBq0(_nf)) * fq
           - 2. * (QCD::beta0(_nf) + QCD::gammaBq0(_nf)) * QCD::CF() * PqqXfq
           - 2. * (QCD::beta0(_nf) + QCD::gammaBq0(_nf)) * QCD::TF() * PqgXfg
           + 4. * PqiXPijXfj
           + 2. * QCD::CF() * QCD::Gamma0(_nf) * QCD::CF() * Iqq1delXfq
           + 2. * QCD::CF() * QCD::Gamma0(_nf) * QCD::TF() * Iqg1delXfg);
}

// NNLO, L2
double TauCoeffsQuark::B2q2(int pid, double x, double muB)
{
   double fq = _convsQCD->convValue(k_fx, pid, x, muB);
   double PqqXfq = _convsQCD->convValue(kPqx_fx, pid, x, muB);
   double PqgXfg = _convsQCD->convValue(kPqx_fx, kGluon, x, muB);

   return (((-1. / 2.) * QCD::beta0(_nf) + (-3. / 4.) * QCD::gammaBq0(_nf)) * QCD::CF() * QCD::Gamma0(_nf) * fq
           + (3. * QCD::CF() * QCD::CF() * QCD::Gamma0(_nf)) * PqqXfq
           + (3. * QCD::CF() * QCD::TF() * QCD::Gamma0(_nf)) * PqgXfg);
}

// NNLO, L3
double TauCoeffsQuark::B2q3(int pid, double x, double muB)
{
   double fq = _convsQCD->convValue(k_fx, pid, x, muB);
   return ((1. / 2.) * QCD::CF() * QCD::CF() * QCD::Gamma0(_nf) * QCD::Gamma0(_nf) * fq);
}

} // namespace beamfunc
