// ============================================================================
// gzstream, C++ iostream classes wrapping the zlib compression library.
// Copyright (C) 2001  Deepak Bandyopadhyay, Lutz Kettner
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// ============================================================================
//
// File          : gzstream.h
// Revision      : $Revision: 1.5 $
// Revision_date : $Date: 2002/04/26 23:30:15 $
// Author(s)     : Deepak Bandyopadhyay, Lutz Kettner
//
// Standard streambuf implementation following Nicolai Josuttis, "The
// Standard C++ Library".
// ============================================================================

// Modified by Simone Alioli to add functionality and to fix a bug in igzstream and ogzstream
// constructors that resulted in the rdstate of the gzstreambuf being reset after istream
// and ostream ctor calls and therefore not reflected by the igzstream and ogzstream objects.
//
// Frank Tackmann: Fix integer conversion warnings.

#ifndef GENEVA_GZSTREAM_HPP
#define GENEVA_GZSTREAM_HPP

// standard C++ with new header file names and std:: namespace
#include <iostream>
#include <fstream>
#include <zlib.h>
#include <cstring>
#include <algorithm>

namespace Geneva
{

namespace gzstream {

// ----------------------------------------------------------------------------
// Internal classes to implement gzstream. See below for user classes.
// ----------------------------------------------------------------------------

class gzstreambuf : public std::streambuf {
  private:
    static const int _bufferSize = 47+256;    // size of data buff
    // totals 512 bytes under g++ for igzstream at the end.

    gzFile           _file;                // file handle for compressed file
    char             _buffer[_bufferSize]; // data buffer
    bool             _opened;              // open/close state of stream
    std::ios_base::openmode _mode;         // I/O mode


    int flush_buffer() {
        // Separate the writing of the buffer from overflow() and
        // sync() operation.
        int w = static_cast<int>(pptr() - pbase());
        if (gzwrite(_file, pbase(), static_cast<unsigned>(w)) != w)
            return EOF;
        pbump(-w);
        return w;
    }

  public:
    gzstreambuf() : _opened(false) {
        setp(_buffer, _buffer + (_bufferSize-1));
        setg(_buffer + 4,     // beginning of putback area
              _buffer + 4,     // read position
              _buffer + 4);    // end position
        // ASSERT: both input & output capabilities will not be used together
    }

    bool is_open() { return _opened; }

    int rewind() { return gzrewind(_file); }

    ~gzstreambuf() { close(); }

    gzstreambuf* open(const char* name, std::ios_base::openmode open_mode) {
        if (is_open())
            return (gzstreambuf*)0;
        _mode = open_mode;
        // no append nor read/write mode
        if ((_mode & std::ios::ate) || (_mode & std::ios::app)
            || ((_mode & std::ios::in) && (_mode & std::ios::out)))
            return (gzstreambuf*)0;
        char  fmode[10];
        char* fmodeptr = fmode;
        if (_mode & std::ios::in)
            *fmodeptr++ = 'r';
        else if (_mode & std::ios::out)
            *fmodeptr++ = 'w';
        *fmodeptr++ = 'b';
        *fmodeptr = '\0';
        _file = gzopen(name, fmode);
        if (_file == 0)
            return (gzstreambuf*)0;
        _opened = true;
        return this;
    }

    gzstreambuf* close() {
        if (is_open()) {
            sync();
            _opened = false;
            if (gzclose(_file) == Z_OK)
                return this;
        }
        return (gzstreambuf*)0;
    }

    virtual int underflow() { // used for input buffer only
        if (gptr() && (gptr() < egptr()))
            return * reinterpret_cast<unsigned char *>(gptr());

        if (!(_mode & std::ios::in) || ! _opened)
            return EOF;
        // Josuttis' implementation of inbuf
        size_t n_putback = static_cast<size_t>(gptr() - eback());
        if (n_putback > 4)
            n_putback = 4;
        memcpy(_buffer + (4 - n_putback), gptr() - n_putback, n_putback);

        int num = gzread(_file, _buffer+4, _bufferSize-4);
        if (num <= 0) // ERROR or EOF
            return EOF;

        // reset buffer pointers
        setg(_buffer + (4 - n_putback),   // beginning of putback area
             _buffer + 4,                 // read position
             _buffer + 4 + num);          // end of buffer

        // return next character
        return * reinterpret_cast<unsigned char *>(gptr());
    }

    virtual int overflow(int c=EOF) { // used for output buffer only
        if (! (_mode & std::ios::out) || ! _opened)
            return EOF;
        if (c != EOF) {
            *pptr() = static_cast<char_type>(c);
            pbump(1);
        }
        if (flush_buffer() == EOF)
            return EOF;
        return c;
    }

    virtual int sync() {
        // Changed to use flush_buffer() instead of overflow(EOF)
        // which caused improper behavior with std::endl and flush(),
        // bug reported by Vincent Ricard.
        if (pptr() && pptr() > pbase()) {
            if (flush_buffer() == EOF)
                return -1;
        }
        return 0;
    }
};

class gzstreambase : virtual public std::ios {

  protected:
    gzstreambuf buf;

  public:
    gzstreambase() { init(&buf); }

    gzstreambase(const char* name, std::ios_base::openmode mode) {
        init(&buf);
        open(name, mode);
    }

    ~gzstreambase() {
        buf.close();
    }

    void open(const char* name, std::ios_base::openmode open_mode) {
        if (! buf.open(name, open_mode)) {
            clear(rdstate() | std::ios::badbit);
        }
    }

    void close() {
        if (buf.is_open()) {
            if (! buf.close()) {
                clear(rdstate() | std::ios::badbit);
            }
        }
    }

    gzstreambuf* rdbuf() { return &buf; }

    bool is_open() { return buf.is_open(); }

    int rewind() { return buf.rewind(); }
};

// ----------------------------------------------------------------------------
// User classes. Use igzstream and ogzstream analogously to ifstream and
// ofstream respectively. They read and write files based on the gz*
// function interface of the zlib. Files are compatible with gzip compression.
// ----------------------------------------------------------------------------

class igzstream : public gzstreambase, public std::istream {
public:
    igzstream() : std::istream(&buf) {}
    igzstream(const char* name, std::ios_base::openmode open_mode = std::ios::in) : gzstreambase(name, open_mode), std::istream(good() ? &buf : 0) { }
    gzstreambuf* rdbuf() { return gzstreambase::rdbuf(); }
    void open(const char* name, std::ios_base::openmode open_mode = std::ios::in) { gzstreambase::open(name, open_mode); }
};

class ogzstream : public gzstreambase, public std::ostream {
public:
    ogzstream() : std::ostream(&buf) {}
    ogzstream(const char* name, std::ios_base::openmode mode = std::ios::out) : gzstreambase(name, mode), std::ostream(good() ? &buf : 0) {}
    gzstreambuf* rdbuf() { return gzstreambase::rdbuf(); }
    void open(const char* name, std::ios_base::openmode open_mode = std::ios::out) { gzstreambase::open(name, open_mode); }
};

} // namespace zgstream

} // namespace Geneva

#endif // GENEVA_GZSTREAM_HPP
