//------------------------------------------------------------------------------
/// \file AnalyzerBase.cpp
//
// Author(s):
//    Jon Walsh, Frank Tackmann, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class AnalyzerBase
//------------------------------------------------------------------------------

#include "Geneva/Analyzers/AnalyzerBase.hpp"

#include "Geneva/Core/Utils/MathUtils.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include <string>
#include <iterator>
#include <algorithm>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief Returns a reference to the SingleBin \a name.
 *
 * If no SingleBin with \a name is found, a fatal error is thrown.
 */
SingleBin& AnalyzerBase::singleBin(const std::string& name)
{
   std::map<std::string, SingleBin>::iterator it = _singleBins.find(name);
   if (it == _singleBins.end())
      Log::fatal("AnalyzerBase::singleBin") << "SingleBin " << name << " not found." << Log::end;
   return it->second;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns a const reference to the SingleBin \a name.
 *
 * If no SingleBin with \a name is found, a fatal error is thrown.
 */
const SingleBin& AnalyzerBase::singleBin(const std::string& name) const
{
   std::map<std::string, SingleBin>::const_iterator it = _singleBins.find(name);
   if (it == _singleBins.end())
      Log::fatal("AnalyzerBase::singleBin") << "SingleBin " << name << " not found." << Log::end;
   return it->second;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns a reference to the Hist1D \a name.
 *
 * If no Hist1D with \a name is found, a fatal error is thrown.
 */
Hist1D& AnalyzerBase::hist1D(const std::string& name)
{
   std::map<std::string, Hist1D>::iterator it = _hists1D.find(name);
   if (it == _hists1D.end())
      Log::fatal("AnalyzerBase::hist1D") << "Histogram " << name << " not found." << Log::end;
   return it->second;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns a const reference to the Hist1D \a name.
 *
 * If no Hist1D with \a name is found, a fatal error is thrown.
 */
const Hist1D& AnalyzerBase::hist1D(const std::string& name) const
{
   std::map<std::string, Hist1D>::const_iterator it = _hists1D.find(name);
   if (it == _hists1D.end())
      Log::fatal("AnalyzerBase::hist1D") << "Histogram " << name << " not found." << Log::end;
   return it->second;
}

//------------------------------------------------------------------------------
std::vector<SingleBin> AnalyzerBase::getSingleBins() const
{
   std::vector<SingleBin> bins;
   bins.reserve(_singleBins.size());
   for (std::map<std::string, SingleBin>::const_iterator it = _singleBins.begin(); it != _singleBins.end(); ++it)
      bins.push_back(it->second);
   return bins;
}

//------------------------------------------------------------------------------
std::vector<Hist1D> AnalyzerBase::getHist1Ds() const
{
   std::vector<Hist1D> hists;
   hists.reserve(_hists1D.size());
   for (std::map<std::string, Hist1D>::const_iterator it = _hists1D.begin(); it != _hists1D.end(); ++it)
      hists.push_back(it->second);
   return hists;
}

//------------------------------------------------------------------------------
/// Adds a singleBin to the analyzer
void AnalyzerBase::addSingleBin(const std::string& name)
{
   std::pair<std::map<std::string, SingleBin>::iterator, bool> it = _singleBins.insert(std::make_pair(name, SingleBin(name)));
   if (!it.second)
      Log::fatal("AnalyzerBase::addSingleBin") << "SingleBin " << name << " already exists." << Log::end;
}

//------------------------------------------------------------------------------
/// Adds a Hist1D to the analyzer
void AnalyzerBase::addHist1D(const std::string& name, size_t nbins, double binlow, double binhigh, bool isLinear)
{
   std::pair<std::map<std::string, Hist1D>::iterator, bool> it = _hists1D.insert(std::make_pair(name, Hist1D(name, nbins, binlow, binhigh, isLinear)));
   if (!it.second)
      Log::fatal("AnalyzerBase::addHist1D") << "Hist1D " << name << " already exists." << Log::end;
}

//------------------------------------------------------------------------------
/// Adds a Hist1D to the analyzer
void AnalyzerBase::addHist1D(const std::string& name, const std::vector<double> binLimits, bool isLinear)
{
   std::pair<std::map<std::string, Hist1D>::iterator, bool> it = _hists1D.insert(std::make_pair(name, Hist1D(name, binLimits, isLinear)));
   if (!it.second)
      Log::fatal("AnalyzerBase::addHist1D") << "Hist1D " << name << " already exists." << Log::end;
}

//------------------------------------------------------------------------------
void AnalyzerBase::initializeCorrelated()
{
   assert(!_canAddCorrelatedWeight);
   initializeLevel(kFixedOrder);
   _canAddCorrelatedWeight = true;
}

//------------------------------------------------------------------------------
void AnalyzerBase::analyzeCorrelated(const Event& correlatedEvent)
{
   assert(_canAddCorrelatedWeight);
   analyzeBase(correlatedEvent, kFixedOrder);
}

//------------------------------------------------------------------------------
void AnalyzerBase::finalizeCorrelated()
{
   assert(_canAddCorrelatedWeight);
   finalizeLevel(kFixedOrder);
   _canAddCorrelatedWeight = false;
}

//------------------------------------------------------------------------------
void AnalyzerBase::analyzePartonic(const Event& event)
{
   initializeLevel(kPartonic);
   analyzeBase(event, kPartonic);
   finalizeLevel(kPartonic);
}

//------------------------------------------------------------------------------
void AnalyzerBase::analyzeShowered(const Event& showeredEvent, const Configuration& hardProcess)
{
   initializeLevel(kShowered);
   analyzeBase(showeredEvent, hardProcess, kShowered);
   finalizeLevel(kShowered);
}

//------------------------------------------------------------------------------
void AnalyzerBase::analyzeShoweredNoMPI(const Event& showeredEvent, const Configuration& hardProcess)
{
   initializeLevel(kShoweredNoMPI);
   analyzeBase(showeredEvent, hardProcess, kShoweredNoMPI);
   finalizeLevel(kShoweredNoMPI);
}

//------------------------------------------------------------------------------
void AnalyzerBase::analyzeHadronized(const Event& showeredEvent, const Configuration& hardProcess)
{
   initializeLevel(kHadronized);
   analyzeBase(showeredEvent, hardProcess, kHadronized);
   finalizeLevel(kHadronized);
}

//------------------------------------------------------------------------------
void AnalyzerBase::reset()
{
   for (std::map<std::string, SingleBin>::iterator it = _singleBins.begin(); it != _singleBins.end(); ++it)
      it->second.reset();
   for (std::map<std::string, Hist1D>::iterator it = _hists1D.begin(); it != _hists1D.end(); ++it)
      it->second.reset();
}

//------------------------------------------------------------------------------
// Protected and private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Returns the default suffix for analyzer level.
const char* AnalyzerBase::levelSuffix(Level level) const
{
   if (level == kFixedOrder)
      return " FO";
   if (level == kPartonic)
      return " noPY";
   if (level == kShowered)
      return " PY";
   if (level == kShoweredNoMPI)
      return " PYnoMPI";
   if (level == kHadronized)
      return " PYh";
   Log::fatal("AnalyzerBase") << "Unknown analysis level " << level << Log::end;
   return "";
}

//------------------------------------------------------------------------------
std::ostream& AnalyzerBase::print(std::ostream& out) const
{
   out << "Rate                     Entries     Sigma [pb]       +/- Error\n";
   out << "---------------------------------------------------------------------------\n";

   for (std::map<std::string, SingleBin>::const_iterator it = _singleBins.begin(); it != _singleBins.end(); ++it)
      out << it->second;

   return out;
}

//------------------------------------------------------------------------------
namespace
{

// helper function to check if string ends with suffix
bool ends_with(const std::string& str, const std::string& suffix)
{
   return str.size() >= suffix.size() && !str.compare(str.size()-suffix.size(), suffix.size(), suffix);
}

}

//------------------------------------------------------------------------------
void AnalyzerBase::initializeLevel(Level level)
{
   std::string suffix = levelSuffix(level);
   for (std::map<std::string, SingleBin>::iterator it = _singleBins.begin(); it != _singleBins.end(); ++it) {
      if (ends_with(it->first, suffix))
         it->second.initialize();
   }
      for (std::map<std::string, Hist1D>::iterator it = _hists1D.begin(); it != _hists1D.end(); ++it) {
         if (ends_with(it->first, suffix))
            it->second.initialize();
      }
}

//------------------------------------------------------------------------------
void AnalyzerBase::finalizeLevel(Level level)
{
   std::string suffix = levelSuffix(level);
   for (std::map<std::string, SingleBin>::iterator it = _singleBins.begin(); it != _singleBins.end(); ++it) {
      if (ends_with(it->first, suffix))
         it->second.finalize();
   }
      for (std::map<std::string, Hist1D>::iterator it = _hists1D.begin(); it != _hists1D.end(); ++it) {
         if (ends_with(it->first, suffix))
            it->second.finalize();
      }
}

} // namespace Geneva
