//------------------------------------------------------------------------------
/// \file AnalyzerFactory.cpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class AnalyzerFactory
//------------------------------------------------------------------------------

#include "Geneva/Analyzers/AnalyzerFactory.hpp"
#include "Geneva/Analyzers/AnalyzerRates.hpp"

#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"

using std::string;

namespace Geneva
{

//------------------------------------------------------------------------------
IAnalyzer* AnalyzerFactory::getAnalyzer(const OptionMap& optMap)
{
   OptionMap tmpMap(optMap);
   tmpMap.setOption("Analyzer::FixedOrder", "false");
   string name = tmpMap.getOption<string>("Analyzer");
   if (name == "") {
      Log::error("AnalyzerFactory::getAnalyzer") << "No Analyzer specified." << Log::end;
   }
   if (name == "Rates") {
      _toDelete.push_back(new AnalyzerRates(optMap));
   } else if (name == "None") {
      Log::warning("AnalyzerFactory::getAnalyzer") << "Explicitly asking for no analyzer" << Log::end;
      return 0;
   } else {
      Log::error("AnalyzerFactory::getAnalyzer") << "Unknown Analyzer: " << name << Log::end;
   }

   IAnalyzer* tmp = _toDelete.back();
   return tmp;
}

//------------------------------------------------------------------------------
IAnalyzer* AnalyzerFactory::getInternalAnalyzer(const OptionMap& optMap)
{
   OptionMap tmpMap(optMap);
   tmpMap.setOption("Analyzer::Unshowered", "false");
   tmpMap.setOption("Analyzer::Showered", "false");
   tmpMap.setOption("Analyzer::ShoweredNoMPI", "false");
   tmpMap.setOption("Analyzer::Hadronized", "false");
   string name = tmpMap.getOption<string>("Analyzer");
   if (name == "") {
      Log::error("AnalyzerFactory::getInternalAnalyzer") << "No Analyzer specified." << Log::end;
   }
   if (name == "None") {
      Log::warning("AnalyzerFactory::getInternalAnalyzer") << "Explicitly asking for no analyzer" << Log::end;
      return 0;
   } else if (name == "Rates") {
      _toDelete.push_back(new AnalyzerRates(tmpMap));
   } else {
      Log::error("AnalyzerFactory::getInternalAnalyzer") << "Analyzer " << name << " is either unknown or not suited to be used internally. "<< Log::end;
   }

   IAnalyzer* tmp = _toDelete.back();
   return tmp;
}

//------------------------------------------------------------------------------
AnalyzerFactory::~AnalyzerFactory()
{
   for (size_t i = 0; i < _toDelete.size(); ++i) {
      delete _toDelete[i];
   }
}

} // namespace Geneva
