//------------------------------------------------------------------------------
/// \file AnalyzerRates.cpp
//
// Author(s):
//    Jon Walsh, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class AnalyzerRates
//------------------------------------------------------------------------------

#include "Geneva/Analyzers/AnalyzerRates.hpp"
#include "Geneva/Analyzers/SingleBin.hpp"

#include "Geneva/Components/Collider.hpp"

#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/Physics/Event.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <string>
#include <algorithm>

namespace Geneva
{

//------------------------------------------------------------------------------
AnalyzerRates::AnalyzerRates(const OptionMap& optmap)
{
   // get relevant Tau cuts
   _eeCollider = Collider(optmap).isEE();
   if (_eeCollider) {
      _tauNcut = optmap.getOption<double>("NJettiness::Tau2cut");
      _tauNplusOnecut = optmap.getOption<double>("NJettiness::Tau3cut");
   } else {
      _tauNcut = optmap.getOption<double>("NJettiness::Tau0cut");
      _tauNplusOnecut = optmap.getOption<double>("NJettiness::Tau1cut");
   }

   // define rates
   if (optmap.getOption<bool>("Analyzer::FixedOrder"))
      addSingleBins(kFixedOrder);
   if (optmap.getOption<bool>("Analyzer::Unshowered"))
      addSingleBins(kPartonic);
   if (optmap.getOption<bool>("Analyzer::Showered"))
      addSingleBins(kShowered);
   if (optmap.getOption<bool>("Analyzer::ShoweredNoMPI"))
      addSingleBins(kShoweredNoMPI);
   if (optmap.getOption<bool>("Analyzer::Hadronized"))
      addSingleBins(kHadronized);
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
void AnalyzerRates::addSingleBins(Level level)
{
   const std::string app = levelSuffix(level);

   addSingleBin("sigma" + app);
   if (_eeCollider) {
      addSingleBin("sigma2" + app);
      addSingleBin("sigma3" + app);
      addSingleBin("sigma4" + app);
      addSingleBin("sigma5 incl" + app);

      addSingleBin("sigmaJets2" + app);
      addSingleBin("sigmaJets3" + app);
      addSingleBin("sigmaJets>=4" + app);
   } else {
      addSingleBin("sigma0" + app);
      addSingleBin("sigma1" + app);
      addSingleBin("sigma2" + app);
      addSingleBin("sigma3 incl" + app);

      addSingleBin("sigmaJets0" + app);
      addSingleBin("sigmaJets1" + app);
      addSingleBin("sigmaJets>=2" + app);
   }
}

//------------------------------------------------------------------------------
void AnalyzerRates::analyzeBase(const Event& event, Level level)
{
   const std::string app = levelSuffix(level);
   // get the config, sigma from the event
   const Sigma& sigma = event.weightedSigma();
   const Configuration& config = event.config();

   if (!sigma.isDefined() || config.empty())
      return;

   // get number of final state QCD particles
   size_t nfinal = size_t_from_ptrdiff_t(std::count_if(config.beginFinal(), config.endFinal(), IsQCD()));

   // total xsec
   singleBin("sigma" + app).fill(sigma);

   if (_eeCollider) {
      // partonic rates
      if (nfinal == 2)
         singleBin("sigma2" + app).fill(sigma);
      if (nfinal == 3)
         singleBin("sigma3" + app).fill(sigma);
      if (nfinal == 4)
         singleBin("sigma4" + app).fill(sigma);
      if (nfinal >= 5)
         singleBin("sigma5 incl" + app).fill(sigma);

      // jet rates
      if (nfinal == 2 || NJettinessRecursive()(config, 2) < _tauNcut) {
         singleBin("sigmaJets2" + app).fill(sigma);
      } else if (nfinal == 3 || NJettinessRecursive()(config, 3) < _tauNplusOnecut) {
         singleBin("sigmaJets3" + app).fill(sigma);
      } else {
         singleBin("sigmaJets>=4" + app).fill(sigma);
      }
   } else {
      // partonic rates
      if (nfinal == 0)
         singleBin("sigma0" + app).fill(sigma);
      if (nfinal == 1)
         singleBin("sigma1" + app).fill(sigma);
      if (nfinal == 2)
         singleBin("sigma2" + app).fill(sigma);
      if (nfinal >= 3)
         singleBin("sigma3 incl" + app).fill(sigma);

      // numbers of jets classified by TauN
      NJettinessppRecursive TauN(config);
      const double Tau0 = ZeroJettinesspp(config, TauN.y())();
      const double Tau1 = TauN(1);

      // jet rates
      if (nfinal == 0 ||  Tau0 <= _tauNcut) {
         singleBin("sigmaJets0" + app).fill(sigma);
      } else if (nfinal == 1 || Tau1 <= _tauNplusOnecut) {
         singleBin("sigmaJets1" + app).fill(sigma);
      } else {
         singleBin("sigmaJets>=2" + app).fill(sigma);
      }
   }
}

} // namespace Geneva
