//------------------------------------------------------------------------------
/// \file Hist1D.cpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 CERN, DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class Hist1D
//------------------------------------------------------------------------------

#include "Geneva/Analyzers/Hist1D.hpp"

#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <functional>
#include <algorithm>
#include <numeric>
#include <iterator>
#include <ostream>
#include <iomanip>

namespace Geneva
{

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * The constructor initializes all bins with value zero an no variations. It
 * also checks that for a log histogram the lowest bin is greater than zero.
 */
Hist1D::Hist1D(const std::string& name, size_t nbins, double binlow, double binhigh, bool isLinear)
   : _name(name), _nbins(nbins), _isLinear(isLinear), _totEntries(0),
     _canAddWeight(false), _variationsResized(false)
{
   if (binlow <= 0. && !isLinear)
      Log::fatal("Hist1D") << "In definition of " << name << ": Log binning requires positive lowest bin boundary." << Log::end;
   setBinEdges(binlow, binhigh);
   _entries.assign(_nbins+2, 0);
   _sumWeights.assign(_nbins+2, Sigma(0.0));
   _sumWeightsSq.assign(_nbins+2, Sigma(0.0));
}

//------------------------------------------------------------------------------
/**
 * The constructor initializes all bins with value zero an no variations. It
 * also checks that the provided binEdges are sorted and that for a log
 * histogram the lowest bin is greater than zero.
 */
Hist1D::Hist1D(const std::string& name, const std::vector<double> binEdges, bool isLinear)
   : _name(name), _binEdges(binEdges), _isLinear(isLinear), _totEntries(0),
     _canAddWeight(false), _variationsResized(false)
{
   //NOTE: replace with std::is_sorted after switch to C++11
   if (std::adjacent_find(_binEdges.begin(),_binEdges.end(), std::greater_equal<double>()) != _binEdges.end())
      Log::fatal("Hist1D") << "Provided bins are not sorted in strictly increasing order." << Log::end;
   if (_binEdges.front() <= 0. && !isLinear)
      Log::fatal("Hist1D") << "In definition of " << name << ": Log binning requires positive lowest bin boundary." << Log::end;
   _nbins = binEdges.size() - 1;
   _entries.assign(_nbins + 2, 0);
   _sumWeights.assign(_nbins + 2, Sigma(0.0));
   _sumWeightsSq.assign(_nbins + 2, Sigma(0.0));
}

//------------------------------------------------------------------------------
// Queries and Accessors
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Calculates the absolute bin cross sections.
 * \return A `std::vector` with a Sigma for each bin.
 *
 * The Sigma contains central value and all variations (if applicable) per bin.
 * The overflow and underflow bins are given by the first and last entry in the
 * returned vector.
 */
std::vector<Sigma> Hist1D::valuesAbs() const
{
   assert(!_canAddWeight);
   std::vector<Sigma> values;
   values.reserve(_nbins + 2);
   const double norm = 1./static_cast<double>(_totEntries);
   for (size_t i = 0; i < _nbins + 2; i++)
      values.push_back(_sumWeights[i] * norm);
   return values;
}

//------------------------------------------------------------------------------
/**
 * \brief Calculates the statistical uncertainties on the absolute bin cross sections.
 * \return A `std::vector` with a Sigma for each bin.
 *
 * The Sigma contains central value and all variations (if applicable) per bin.
 * The overflow and underflow bins are given by the first and last entry in the
 * returned vector.
 */
std::vector<Sigma> Hist1D::errorsAbs() const
{
   assert(!_canAddWeight);
   std::vector<Sigma> errors;
   errors.reserve(_nbins + 2);
   for (size_t i = 0; i < _nbins + 2; i++)
      errors.push_back(getBinError(i));
   return errors;
}

//------------------------------------------------------------------------------
/**
 * \brief Calculates the differential cross section.
 * \return A `std::vector` with a Sigma for each bin.
 *
 * The cross section for the main bins is divided by the respective bin width.
 * The Sigma contains central value and all variations (if applicable) per bin.
 * For the overflow and underflow bins, given by the first and last entry in the
 * returned vector, the absolute cross section is used.
 */
std::vector<Sigma> Hist1D::valuesDiff() const
{
   assert(!_canAddWeight);
   std::vector<Sigma> values;
   values.reserve(_nbins + 2);
   const double norm = 1./static_cast<double>(_totEntries);
   values.push_back(_sumWeights.front() * norm);
   for (size_t c = 1; c < _nbins + 1; c++)
      values.push_back(_sumWeights[c] * norm / (_binEdges[c] - _binEdges[c-1]));
   values.push_back(_sumWeights.back() * norm);
   return values;
}

//------------------------------------------------------------------------------
/**
 * \brief Calculates the statistical uncertainties on the differential cross section.
 * \return A vector with a Sigma for each bin.
 *
 * The uncertainties for the main bins are divided by the respective bin width.
 * The Sigma contains central value and all variations (if applicable) per bin.
 * For the overflow and underflow bins, given by the first and last entry in the
 * returned vector, the absolute cross section uncertainties are used.
 */
std::vector<Sigma> Hist1D::errorsDiff() const
{
   assert(!_canAddWeight);
   std::vector<Sigma> errors;
   errors.reserve(_nbins + 2);
   errors.push_back(getBinError(0));
   for (size_t c = 1; c < _nbins + 1; c++)
      errors.push_back(getBinError(c) / (_binEdges[c] - _binEdges[c-1]));
   errors.push_back(getBinError(_nbins + 1));
   return errors;
}

//------------------------------------------------------------------------------
std::vector<double> Hist1D::binCenters() const
{
   std::vector<double> centers(_nbins, 0.0);
   if (_isLinear) {
      for (size_t i = 0; i < _nbins; i++) {
         centers[i] = _binEdges[i] + 0.5 * (_binEdges[i + 1] - _binEdges[i]);
      }
   } else {
      for (size_t i = 0; i < _nbins; i++) {
         centers[i] = _binEdges[i] * sqrt(_binEdges[i + 1] / _binEdges[i]);
      }
   }
   return centers;
}

//------------------------------------------------------------------------------
// Manipulation
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
void Hist1D::initialize()
{
   assert(!_canAddWeight);
   _tmpwt.clear();
   _canAddWeight = true;
}

//------------------------------------------------------------------------------
void Hist1D::fill(double value, const Sigma& weight)
{
   assert(_canAddWeight);
   if (weight.isDefined()) {
      size_t binIndex = size_t_from_ptrdiff_t(std::distance(_binEdges.begin(), std::upper_bound(_binEdges.begin(), _binEdges.end(), value)));
      assert(binIndex < _nbins+2);
#ifndef NDEBUG
      if(binIndex == 0)
         assert(value < _binEdges[0]);
      else
         assert(value >= _binEdges[binIndex-1]);
#endif
      Sigma& tmpWt = _tmpwt[binIndex];
      if (!tmpWt.isDefined()) {
         tmpWt = weight;
      } else {
         tmpWt += weight;
      }
   }
}

//------------------------------------------------------------------------------
void Hist1D::fillUpTo(double value, const Sigma& weight)
{
   assert(_canAddWeight);
   if (weight.isDefined()) {
      size_t binIndex = size_t_from_ptrdiff_t(std::distance(_binEdges.begin(), std::upper_bound(_binEdges.begin(), _binEdges.end(), value)));
      assert(binIndex < _nbins+2);
      #ifndef NDEBUG
      if(binIndex == 0)
         assert(value < _binEdges[0]);
      else
         assert(value >= _binEdges[binIndex-1]);
      #endif
      for (size_t c = 0; c <= binIndex; c++){
         Sigma& tmpWt = _tmpwt[c];
         if (!tmpWt.isDefined()) {
            tmpWt = weight;
         } else {
            tmpWt += weight;
         }
      }
   }
}

//------------------------------------------------------------------------------
void Hist1D::fillUpFrom(double value, const Sigma& weight)
{
   assert(_canAddWeight);
   if (weight.isDefined()) {
      size_t binIndex = size_t_from_ptrdiff_t(std::distance(_binEdges.begin(), std::upper_bound(_binEdges.begin(), _binEdges.end(), value)));
      assert(binIndex < _nbins+2);
      #ifndef NDEBUG
      if(binIndex == 0)
         assert(value < _binEdges[0]);
      else
         assert(value >= _binEdges[binIndex-1]);
      #endif
      for (size_t c = binIndex; c < _nbins+2; c++){
         Sigma& tmpWt = _tmpwt[c];
         if (!tmpWt.isDefined()) {
            tmpWt = weight;
         } else {
            tmpWt += weight;
         }
      }
   }
}

//------------------------------------------------------------------------------
void Hist1D::finalize()
{
   assert(_canAddWeight);
   for (std::map<size_t,Sigma>::const_iterator it = _tmpwt.begin(); it != _tmpwt.end(); ++it) {
      const Sigma& weight = it->second;
      if (weight.isDefined()) {
         // Only after the first defined Sigma is found the vector of variations can be resized
         if (!_variationsResized) {
            size_t n = weight.variations().size();
            for (std::vector<Sigma>::iterator i = _sumWeights.begin(); i != _sumWeights.end(); ++i) {
               assert(i->variations().empty());
               i->variations().resize(n);
            }
            for (std::vector<Sigma>::iterator i = _sumWeightsSq.begin(); i != _sumWeightsSq.end(); ++i) {
               assert(i->variations().empty());
               i->variations().resize(n);
            }
            _variationsResized = true;
         }

         const size_t bin = it->first;
         _sumWeights[bin] += weight;
         _sumWeightsSq[bin] += weight * weight;
         _entries[bin]++;
      }
   }
   // always increase the total counter
   _totEntries++;
   _canAddWeight = false;
}

//------------------------------------------------------------------------------
/**
 * \brief Assigns the given bin data to the %Hist1D.
 * \param weights The sum of weights for all bins
 * \param weightsSq The sum of the squared weights for all bins
 * \param entries The nonzero entries for all bins
 * \param totEntries The total number entries
 *
 * The vectors must have nBins() + 2 entries, where the first and last entries
 * are interpreted as underflow and overflow bins.
 */
void Hist1D::assign(const std::vector<Sigma>& sumWeights, const std::vector<Sigma>& sumWeightsSq,
                    const std::vector<size_t>& entries, size_t totEntries)
{
   assert(sumWeights.size() == _nbins + 2);
   assert(sumWeightsSq.size() == _nbins + 2);
   assert(entries.size() == _nbins + 2);
   _sumWeights = sumWeights;
   _sumWeightsSq = sumWeightsSq;
   _entries = entries;
   _totEntries = totEntries;
}

//------------------------------------------------------------------------------
void Hist1D::reset()
{
   _totEntries = 0;
   _tmpwt.clear();
   _entries.assign(_nbins + 2, 0);
   _sumWeights.assign(_nbins + 2, Sigma(0.0));
   _sumWeightsSq.assign(_nbins + 2, Sigma(0.0));
   _canAddWeight = false;
   _variationsResized = false;
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %Hist1D addition assignment operator
 * \param rhs The %Hist1D to be added.
 * \return A reference to the %Hist1D.
 *
 * \a rhs is added to the %Hist1D.
 */
Hist1D& Hist1D::operator+=(const Hist1D& rhs)
{
   assert(!(_canAddWeight || rhs.canAddWeight()));
   assert(_nbins == rhs._nbins);
   assert(_binEdges.front() == rhs._binEdges.front());
   assert(_binEdges.back() == rhs._binEdges.back());
   assert(_isLinear == rhs._isLinear);
   _totEntries += rhs._totEntries;
   std::transform(_sumWeights.begin(), _sumWeights.end(), rhs._sumWeights.begin(),
                  _sumWeights.begin(), std::plus<Sigma>());
   std::transform(_sumWeightsSq.begin(), _sumWeightsSq.end(), rhs._sumWeightsSq.begin(),
                  _sumWeightsSq.begin(), std::plus<Sigma>());
   std::transform(_entries.begin(), _entries.end(), rhs._entries.begin(),
                  _entries.begin(), std::plus<size_t>());
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %Hist1D subtraction assignment operator
 * \param rhs The %Hist1D to be subtracted.
 * \return A reference to the %Hist1D.
 *
 * \a rhs is subtracted from the %Hist1D.
 */
Hist1D& Hist1D::operator-=(const Hist1D& rhs)
{
   *this += -1.*rhs;
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %Hist1D multiplication with real number assignment operator
 * \param rhs The number to be multiplied by.
 * \return A reference to the %Hist1D.
 *
 * Each component of the %Hist1D is multiplied by \a rhs.
 */
Hist1D& Hist1D::operator*=(double rhs)
{
   assert(!_canAddWeight);
   for (std::vector<Sigma>::iterator it = _sumWeights.begin(); it != _sumWeights.end(); ++it)
      *it *= rhs;
   for (std::vector<Sigma>::iterator it = _sumWeightsSq.begin(); it != _sumWeightsSq.end(); ++it)
      *it *= rhs * rhs;
   return *this;
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Streams a table with the central bin cross sections to out.
std::ostream& Hist1D::print(std::ostream& out) const
{
   assert(!_canAddWeight);
   out << "Hist1D: " << _name << " (total entries: " << _totEntries << ")\n";
   out << "Bin  Entries     Sigma [pb]       +/- Error [pb]\n";
   out << "-------------------------------------------------------\n";
   out.precision(10);
   std::vector<Sigma> val = valuesAbs();
   std::vector<Sigma> err = errorsAbs();
   for (size_t i = 0; i < _nbins + 2; ++i) {
      out << std::setw(5) << i;
      out << std::setw(10) << entries()[i] << "  ";
      out << std::scientific << val[i].central() << " +/- " << err[i].central() << '\n';
   }
   return out;
}

//------------------------------------------------------------------------------
void Hist1D::setBinEdges(double binlow, double binhigh)
{
   if (binlow >= binhigh)
      Log::fatal("Hist1D::setBinLimits") << "Provided lowest bin "<<binlow<<" not strictly less than largest one "<<binhigh<<Log::end;
   _binEdges.reserve(_nbins + 1);
   double bin = binlow;
   if (_isLinear) {
      const double step = (binhigh - binlow) / static_cast<double>(_nbins);
      for (size_t c = 0; c < _nbins; c++, bin += step)
         _binEdges.push_back(bin);
   } else {
      const double step = pow(binhigh / binlow, 1. / static_cast<double>(_nbins));
      for (size_t c = 0; c < _nbins; c++, bin *= step) {
         _binEdges.push_back(bin);
      }
   }
   _binEdges.push_back(binhigh);
}

//------------------------------------------------------------------------------
Sigma Hist1D::getBinError(size_t i) const
{
   assert(!_canAddWeight);
   const double n = static_cast<double>(_totEntries);
   Sigma error(sqrt(abs(_sumWeightsSq[i].central()/n - pow(_sumWeights[i].central()/n, 2)) / (n - 1.0)));
   error.variations().reserve(_sumWeights[i].variations().size());
   Sigma::const_iterator it2 = _sumWeightsSq[i].variations().begin();
   for (Sigma::const_iterator it = _sumWeights[i].variations().begin(); it != _sumWeights[i].variations().end(); ++it, ++it2) {
      error.variations().push_back(sqrt(abs((*it2)/n - pow(*it/n, 2)) / (n - 1.0)));
   }
   return error;
}

} // namespace Geneva
