//------------------------------------------------------------------------------
/// \file OutputStreamAnalyzer.cpp
//
// Author(s):
//    Jon Walsh, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class OutputStreamAnalyzer
//------------------------------------------------------------------------------

#include "Geneva/Analyzers/OutputStreamAnalyzer.hpp"
#include "Geneva/Analyzers/Hist1D.hpp"
#include "Geneva/Analyzers/SingleBin.hpp"

#include "Geneva/Core/UI/Option.hpp"

#include <iostream>
#include <algorithm>

namespace Geneva
{

//------------------------------------------------------------------------------
std::ostream& operator<<(std::ostream& out, const OutputStreamAnalyzer& osa)
{
   return osa.getOutput(out);
}

//------------------------------------------------------------------------------
std::string OutputStreamAnalyzer::rateString(const SingleBin& rate) const
{
   std::stringstream sostream("");

   // get the name
   std::string name = rate.name();

   // get the value, error, weight, weightSq, and entries
   // for central and scale variation
   // convert the doubles to scientific notation strings
   const size_t nentries = rate.totEntries();
   const size_t binentries = rate.entries();
   std::string centralvalueStr = stringSciNotation(rate.value().central());
   std::string centralerrorStr = stringSciNotation(rate.error().central());
   std::string centralweightStr = stringSciNotation(rate.sumWeights().central());
   std::string centralweightSqStr = stringSciNotation(rate.sumWeightsSq().central());

   size_t nvariations = rate.value().variations().size();
   std::vector<double> varyvalue = rate.value().variations();
   std::vector<double> varyerror = rate.error().variations();
   std::vector<double> varyweight = rate.sumWeights().variations();
   std::vector<double> varyweightSq = rate.sumWeightsSq().variations();

   std::vector<std::string> varyvalueStr, varyerrorStr, varyweightStr, varyweightSqStr;
   for (size_t i = 0; i < nvariations; i++) {
      varyvalueStr.push_back(stringSciNotation(varyvalue[i]));
      varyerrorStr.push_back(stringSciNotation(varyerror[i]));
      varyweightStr.push_back(stringSciNotation(varyweight[i]));
      varyweightSqStr.push_back(stringSciNotation(varyweightSq[i]));
   }

   // write the output string for the central scale
   sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"central\", name] = \"" << name << "\";" << std::endl;
   sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"central\", value] = " << centralvalueStr << ";" << std::endl;
   sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"central\", error] = " << centralerrorStr << ";" << std::endl;
   sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"central\", weight] = " << centralweightStr << ";" << std::endl;
   sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"central\", weightSq] = " << centralweightSqStr << ";" << std::endl;
   sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"central\", binEntries] = " << binentries << ";" << std::endl;
   sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"central\", entries] = " << nentries << ";" << std::endl;

   // write the number of variations
   sostream << "rate[\"" << _stem << "\", \"" << name << "\", variations] = " << nvariations << ";" << std::endl;
   // now do the variations
   for (size_t i = 0; i < nvariations; i++) {
      sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", name] = \"" << name << "\";" << std::endl;
      sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", value] = " << varyvalueStr[i] << ";" << std::endl;
      sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", error] = " << varyerrorStr[i] << ";" << std::endl;
      sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", weight] = " << varyweightStr[i] << ";" << std::endl;
      sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", weightSq] = " << varyweightSqStr[i] << ";" << std::endl;
      sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", binEntries] = " << binentries << ";" << std::endl;
      sostream << "rate[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", entries] = " << nentries << ";" << std::endl;
   }

   return sostream.str();
}

namespace
{

//------------------------------------------------------------------------------
// helper functions used by hist1DString
// they currently drop the underflow and overflow bins
std::vector<double> central(const std::vector<Sigma>& sigma)
{
   std::vector<double> tmp;
   tmp.reserve(sigma.size() - 2);
   for (std::vector<Sigma>::const_iterator it = ++sigma.begin(); it != --sigma.end(); ++it)
      tmp.push_back(it->central());
   return tmp;
}

//------------------------------------------------------------------------------
std::vector<std::vector<double> > variations(const std::vector<Sigma>& sigma)
{
   std::vector<std::vector<double> > tmp;
   size_t nvariations  = sigma.begin()->variations().size();
   tmp.reserve(nvariations);
   for (size_t i = 0; i < nvariations; i++) {
      std::vector<double> varyi;
      varyi.reserve(sigma.size() - 2);
      for (std::vector<Sigma>::const_iterator it = ++sigma.begin(); it != --sigma.end(); ++it) {
         assert(it->variations().size() == nvariations);
         varyi.push_back(it->variations()[i]);
      }
      tmp.push_back(varyi);
   }
   return tmp;
}

} // anonymous namespace

//------------------------------------------------------------------------------
std::string OutputStreamAnalyzer::hist1DString(const Hist1D& hist) const
{
   std::stringstream histstream("");

   // get the name, binning
   std::string name = hist.name();
   std::string binning = "linear";
   if (!hist.isLinear()) { binning = "log"; }

   std::string binLimitsStr = stringSciNotation(hist.binEdges());

   size_t nentries = hist.totEntries();
   std::vector<double> binEdges = hist.binEdges();

   // get the value, error, weight, weightSq, bin centers, bin entries, # entries, bin width
   // convert the doubles to strings of formatted lists

   std::string centralvalueStr = stringSciNotation(central(hist.valuesDiff()));
   std::string centralerrorStr = stringSciNotation(central(hist.errorsDiff()));
   std::vector<double> tmpc = central(hist.sumWeights());
   std::vector<double> tmpcSq = central(hist.sumWeightsSq());
   for(size_t c=0; c<binEdges.size()-1; c++) {
      double w=binEdges[c+1]-binEdges[c];
      tmpc[c]/=w;
      tmpcSq[c]/=w*w;
   }
   std::string centralweightStr = stringSciNotation(tmpc);
   std::string centralweightSqStr = stringSciNotation(tmpcSq);
   std::string centersStr = stringSciNotation(hist.binCenters());

   size_t nvariations = hist.sumWeights().begin()->variations().size();
   std::vector< std::vector<double> > varyvalue = variations(hist.valuesDiff());
   std::vector< std::vector<double> > varyerror = variations(hist.errorsDiff());
   std::vector< std::vector<double> > varyweight = variations(hist.sumWeights());
   std::vector< std::vector<double> > varyweightSq = variations(hist.sumWeightsSq());

   std::vector<std::string> varyvalueStr, varyerrorStr, varyweightStr, varyweightSqStr;
   for (size_t i = 0; i < nvariations; i++) {
      varyvalueStr.push_back(stringSciNotation(varyvalue[i]));
      varyerrorStr.push_back(stringSciNotation(varyerror[i]));
      for(size_t c=0; c<binEdges.size()-1; c++) {
         double w=binEdges[c+1]-binEdges[c];
         varyweight[i][c]/=w;
         varyweightSq[i][c]/=w*w;
      }
      varyweightStr.push_back(stringSciNotation(varyweight[i]));
      varyweightSqStr.push_back(stringSciNotation(varyweightSq[i]));
   }

   const std::vector<size_t>& entries = hist.entries();
   std::stringstream entrystream("");
   entrystream << "{";
   for (size_t c = 1; c < entries.size() - 1; c++) {
      entrystream << std::setprecision(15) << entries[c];
      if (c < entries.size() - 2) { entrystream << ", "; }
      else { entrystream << "}"; }
   }
   std::string entriesStr = entrystream.str();

   //TODO: this binwidth is useless and wrong when bins are not equally spaced
   double binwidth = binEdges[1] - binEdges[0];

   //TODO: decide what we really need

   // write the output string for the central scale
   histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"central\", name] = \"" << name << "\";" << std::endl;
   histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"central\", value] = " << centralvalueStr << ";" << std::endl;
   histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"central\", error] = " << centralerrorStr << ";" << std::endl;
   histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"central\", weight] = " << centralweightStr << ";" << std::endl;
   histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"central\", weightSq] = " << centralweightSqStr << ";" << std::endl;
   histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"central\", bins] = " << centersStr << ";" << std::endl;
   histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"central\", binEntries] = " << entriesStr << ";" << std::endl;
   histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"central\", binWidth] = " << binwidth << ";" << std::endl;
   histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"central\", entries] = " << nentries << ";" << std::endl;
   histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"central\", binning] = " << binning << ";" << std::endl;

   // write the number of variations
   histstream << "hist[\"" << _stem << "\", \"" << name << "\", variations] = " << nvariations << ";" << std::endl;
   // now do the variations
   for (size_t i = 0; i < nvariations; i++) {
      histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", name] = \"" << name << "\";" << std::endl;
      histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", value] = " << varyvalueStr[i] << ";" << std::endl;
      histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", error] = " << varyerrorStr[i] << ";" << std::endl;
      histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", weight] = " << varyweightStr[i] << ";" << std::endl;
      histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", weightSq] = " << varyweightSqStr[i] << ";" << std::endl;
      histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", bins] = " << centersStr << ";" << std::endl;
      histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", binEntries] = " << entriesStr << ";" << std::endl;
      histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", binWidth] = " << binwidth << ";" << std::endl;
      histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", entries] = " << nentries << ";" << std::endl;
      histstream << "hist[\"" << _stem << "\", \"" << name << "\", \"vary" << i + 1 << "\", binning] = " << binning << ";" << std::endl;
   }

   return histstream.str();
}

//------------------------------------------------------------------------------
std::string OutputStreamAnalyzer::rateNames() const
{
   std::stringstream namestream("");

   if (not _rates.empty()) {
      namestream << "rate[\"" << _stem << "\", \"names\"] = ";
      namestream << "{";
      for (std::vector<SingleBin>::const_iterator it = _rates.begin(); it != --_rates.end(); ++it) {
         namestream << "\"" << it->name() << "\", ";
      }
      namestream << "\"" << _rates.rbegin()->name() << "\"";
      namestream << "};" << std::endl;
   }
   return namestream.str();
}

//------------------------------------------------------------------------------
std::string OutputStreamAnalyzer::histNames() const
{
   std::stringstream namestream("");

   if (not _hists.empty()) {
      namestream << "hist[\"" << _stem << "\", \"names\"] = ";
      namestream << "{";
      for (std::vector<Hist1D>::const_iterator it = _hists.begin(); it != --_hists.end(); ++it) {
         namestream << "\"" << it->name() << "\", ";
      }
      namestream << "\"" << _hists.rbegin()->name() << "\"";
      namestream << "};" << std::endl;
   }
   return namestream.str();
}

//------------------------------------------------------------------------------
std::ostream& OutputStreamAnalyzer::getOutput(std::ostream& stream) const
{
   stream << initialize();

   stream << rateNames() << std::endl;

   for (std::vector<SingleBin>::const_iterator it = _rates.begin(); it != _rates.end(); ++it) {
      stream << rateString(*it);
   }
   stream << std::endl;

   stream << histNames() << std::endl;

   for (std::vector<Hist1D>::const_iterator it = _hists.begin(); it != _hists.end(); ++it) {
      stream << hist1DString(*it);
   }

   return stream;
}

} // namespace Geneva
