//------------------------------------------------------------------------------
/// \file CalcNLOppVj.cpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class CalcNLOppVj
//------------------------------------------------------------------------------

#include "Geneva/Calculations/CalcNLOppVj.hpp"
#include "Geneva/Calculations/PhiNLOppVjFKS.hpp"
#include "Geneva/Calculations/PhiNLOppVjTau0FR.hpp"

#include "Geneva/Components/SignatureManager.hpp"

#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"

#include <cassert>

namespace Geneva
{

//------------------------------------------------------------------------------
CalcNLOppVj::CalcNLOppVj(const OptionMap& optMap, Tau0NLO tau0NLO,
                         ITreeME* treeME, ITreeMEComplete* treeMEComplete, IOneLoopME* oneLoopME,
                         UnitCubeRadGenerator* xRadGenerator, IPDF* pdf,
                         CouplingStrong* alphas, FixedOrder order, double Ecm, int nLight,
                         IScale* renScale, IScale* factScale,
                         ICoupling* alphaem, double Tau0Cut, double Tau1Cut,
                         PdgID boson, PdgID lepton)
   : _idV(boson), _idDecay(lepton), _nLight(nLight), _alphaem(alphaem), _alphas(alphas)
{
   // check that at least one process is selected
   assert(_idV == PdgID::kZ || _idV == PdgID::kWPlus || _idV == PdgID::kWMinus);
   // check that at least one leptonic decay is chosen
   assert(_idDecay == PdgID::kElectron || _idDecay == PdgID::kMuon || _idDecay == PdgID::kTau);
   // check that the calculation is performed for 5 light flavours
   assert(_nLight == 5);

   // Sets the Signatures
   _sigSet.clear();
   _sigSet = getSignaturesN();

   if (order == FixedOrder::kNLO)
      _sigSet += getSignaturesNplusOne();

   double fksTinyCut = optMap.getOption<double>("CalcNLOppVj::FKSTinyCut");
   if (tau0NLO == kTau0FR) {
      _phiNLO = new PhiNLOppVjTau0FR(Tau0Cut, Tau1Cut, getSignaturesN(), Ecm, optMap.getOption<bool>("NJettiness::Tau0NLO::Tau0FR::fixqT"), fksTinyCut);
   } else {
      _phiNLO = new PhiNLOppVjFKS(Tau0Cut, Tau1Cut, getSignaturesN(), Ecm, fksTinyCut);
   }

   //Sets the leading order couplings
   _commonCouplings = new CommonCouplings(_alphas, _alphaem);

   // Construct helper
   _helper = new EngineNLOpp(optMap, name(), _sigSet, treeME, treeMEComplete, oneLoopME, xRadGenerator, pdf, renScale, factScale, _commonCouplings, _phiNLO, alphas, order, Ecm, nLight);
}

//------------------------------------------------------------------------------
/** Explicitly destroys objects created at construction, in the opposite order of creation,
 * to avoid memory leaks
*/
CalcNLOppVj::~CalcNLOppVj()
{
   delete _helper;
   delete _commonCouplings;
   delete _phiNLO;
}

//------------------------------------------------------------------------------
const std::vector<NameValuePair> CalcNLOppVj::componentInfo() const
{
   std::vector<NameValuePair> pairs, common;
   // Add common NLO components
   common = _helper->componentInfo();
   std::copy(common.begin(), common.end(), std::back_inserter(pairs));
   //add process specific ones
   pairs.push_back(NameValuePair("alphaem", _alphaem->name()));
   pairs.push_back(NameValuePair("alphas", _alphas->name()));
   pairs.push_back(NameValuePair("phiNLO", _phiNLO->name()));
   return pairs;
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
/**
 * \brief Calculate the couplings.
 * \param config The configuration to act on.
 * \return The product of the couplings for the given process.
 *
 * Calculate the common couplings between the LO and NLO coefficient, for the
 * given \a config .
 */
double CalcNLOppVj::CommonCouplings::value(const Configuration& config, const double muR) const
{
   const double alphaem =  (*_aem)(muR);
   const double alphas = (*_as)(muR);
   double coupl = alphaem * alphaem * alphas;
   return coupl;
}

//------------------------------------------------------------------------------
std::string CalcNLOppVj::CommonCouplings::name() const
{
   return "alphaem^2 * alphas";
}

//------------------------------------------------------------------------------
/**
 * This returns the subprocesses with n final state particles required for this calculation.
 */
SignatureSet CalcNLOppVj::getSignaturesN() const
{
   bool doDecay = true;
   if (_idV == PdgID::kZ)
      return SignatureManager().getSignaturespp2ZJets(_idDecay, 1, _nLight, doDecay);
   else if (_idV == PdgID::kWPlus)
      return SignatureManager().getSignaturespp2WPlusJets(_idDecay, 1, _nLight, doDecay);
   else if (_idV == PdgID::kWMinus)
      return SignatureManager().getSignaturespp2WMinusJets(_idDecay, 1, _nLight, doDecay);
   else {
      Log::error("CalcNLOppVj::getSignaturesN") << "Unknown boson " << _idV << Log::end;
      return SignatureSet();
   }
}

//------------------------------------------------------------------------------
/**
 * This returns the subprocesses with n+1 final state particles required for this calculation.
 */
SignatureSet CalcNLOppVj::getSignaturesNplusOne() const
{
   bool doDecay = true;
   if (_idV == PdgID::kZ)
      return SignatureManager().getSignaturespp2ZJets(_idDecay, 2, _nLight, doDecay);
   else if (_idV == PdgID::kWPlus)
      return SignatureManager().getSignaturespp2WPlusJets(_idDecay, 2, _nLight, doDecay);
   else if (_idV == PdgID::kWMinus)
      return SignatureManager().getSignaturespp2WMinusJets(_idDecay, 2, _nLight, doDecay);
   else {
      Log::error("CalcNLOppVj::getSignaturesNplusOne") << "Unknown boson " << _idV << Log::end;
      return SignatureSet();
   }
}

} // namespace Geneva
