//------------------------------------------------------------------------------
/// \file CalcSCETppV012.cpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class CalcSCETppV012
//------------------------------------------------------------------------------

#include "Geneva/Calculations/CalcSCETppV012.hpp"

#include "Geneva/Components/ComponentFactory.hpp"
#include "Geneva/Components/ICoupling.hpp"
#include "Geneva/Components/CouplingStrong.hpp"
#include "Geneva/Components/FluxFactor.hpp"
#include "Geneva/Components/BeamFuncTauB.hpp"
#include "Geneva/Components/SplittingFunction.hpp"
#include "Geneva/Components/SplittingMappingppV0To1.hpp"
#include "Geneva/Components/SplittingMappingppV1To2.hpp"
#include "Geneva/Components/Collider.hpp"
#include "Geneva/Components/SignatureManager.hpp"
#include "Geneva/Components/UnitCubeRadGenerator.hpp"
#include "Geneva/Components/Scales.hpp"

#include "Geneva/Core/Physics/RGE.hpp"
#include "Geneva/Core/Physics/FOConv.hpp"
#include "Geneva/Core/Physics/PdgID.hpp"
#include "Geneva/Core/Physics/ConfigurationUtils.hpp"
#include "Geneva/Core/Physics/QCD.hpp"
#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/Random.hpp"

#include <cassert>
#include <iomanip>
#include <map>
#include <fstream>
#include <string>


#define REFINE_PHIPROJ_PRECISION 1e-10
#define REFINE_PHIPROJ_ACCURACY 1e-13

namespace Geneva
{

//------------------------------------------------------------------------------
std::map<CalcSCETppV012::Scale, EngineNLOBase::ScaleVariation> CalcSCETppV012::fixedOrderScale =  initFixedOrderScaleMap();

//------------------------------------------------------------------------------
std::map<CalcSCETppV012::Scale, EngineNLOBase::ScaleVariation> CalcSCETppV012::initFixedOrderScaleMap()
{
   std::map<CalcSCETppV012::Scale, EngineNLOBase::ScaleVariation> map;
   map[kall] = EngineNLOBase::kCorrelated;
   map[kfastset] = EngineNLOBase::kCorrelated;
   map[khardset] = EngineNLOBase::kCorrelated;
   map[kfixset] = EngineNLOBase::kCorrelated;
   map[ksetupset] = EngineNLOBase::kCorrelated;
   map[kmurun] = EngineNLOBase::kNone;
   map[kmuup] = EngineNLOBase::kNone;
   map[kmuupfix] = EngineNLOBase::kNone;
   map[kmudown] = EngineNLOBase::kNone;
   map[kmudownfix] = EngineNLOBase::kNone;
   map[kmuBup] = EngineNLOBase::kNone;
   map[kmuBdown] = EngineNLOBase::kNone;
   map[kmuSup] = EngineNLOBase::kNone;
   map[kmuSdown] = EngineNLOBase::kNone;
   map[kmuxup] = EngineNLOBase::kNone;
   map[kmuxdown] = EngineNLOBase::kNone;
   map[kcanonical] = EngineNLOBase::kNone;
   return map;
}

//------------------------------------------------------------------------------
std::map<CalcSCETppV012::Scale, CalcSCETppV012::Scale> CalcSCETppV012::resummationScale =  initResummationScaleMap();

//------------------------------------------------------------------------------
std::map<CalcSCETppV012::Scale, CalcSCETppV012::Scale> CalcSCETppV012::initResummationScaleMap()
{
   std::map<CalcSCETppV012::Scale, CalcSCETppV012::Scale> map;
   //NOTE: only meaningful entries are mapped. If one tries to retrieve the value
   // of keys whish would correspond to multiple entries, e.g. kall it would get back an out-of-map error
   map[kmurun] = kmurun;
   map[kmuup] = kmuup;
   map[kmuupfix] = kmuup;
   map[kmudown] = kmudown;
   map[kmudownfix] = kmudown;
   map[kmuBup] = kmuBup;
   map[kmuBdown] = kmuBdown;
   map[kmuSup] = kmuSup;
   map[kmuSdown] = kmuSdown;
   map[kmuxup] = kmuxup;
   map[kmuxdown] = kmuxdown;
   map[kcanonical] = kcanonical;
   map[kmuhigherrun] = kmuhigherrun;
   map[kmuhigherup] = kmuhigherup;
   map[kmuhigherdown] = kmuhigherdown;
   return map;
}

//------------------------------------------------------------------------------
CalcSCETppV012::CalcSCETppV012(const OptionMap& optMap, double Tau0cut, double Tau1cut, double Ecm,
                               RunOrder run0, RunOrder run1, FixedOrder fixed0jet, FixedOrder fixed1jet,
                               Scale scale, YProfiles profilesY, Phi1 phi1, ICoupling* alphaem, CouplingStrong* alphas,
                               IPDF* pdf, ITreeME* tree, ITreeMEComplete* treeComplete, IOneLoopME* oneLoop,
                               UnitCubeRadGenerator* xRadGen, int nLightFlavors, PdgID bosonId, PdgID decayChannel,
                               bool flavSumBeams, bool totalXSfix, bool improvedXSfix, double mVLow, double mVHigh)
   : _optMap(optMap), _tau0cut(Tau0cut), _tau1cut(Tau1cut), _cmEnergy(Ecm),
     _run0(run0), _run1(run1), _fixed0jet(fixed0jet), _fixed1jet(fixed1jet),
     _inputScale(scale), _profilesY(profilesY), _phi1(phi1), _nLight(nLightFlavors), _alphaem(alphaem), _alphas(alphas),
     _pdf(pdf), _positivePdf(0), _treeME(tree), _xRadGen(xRadGen), _flavSummedBeams(flavSumBeams), _totalXSfix(totalXSfix),
     _improvedXSfix(improvedXSfix), _splitting0To1(0), _splitting1To2(0), _improvedXSHandler(0)
{
   // what we can and  can't (won't) do
   // 0-jet LO, 1-jet LO
   // only NLL running
   if (_fixed0jet == FixedOrder::kLO && _fixed1jet == FixedOrder::kLO) {
      if (_run0 != RunOrder::kNLL) {
         Log::error("CalcSCETppV012") << "0-jet at LO, 1-jet at LO, running not at NLL is not supported" << Log::end;
      }
   }
   // 0-jet NLO, 1-jet LO
   // only NLL or NNLL running
   else if (_fixed0jet == FixedOrder::kNLO && _fixed1jet == FixedOrder::kLO) {
      if (_run0 != RunOrder::kNLL && _run0 != RunOrder::kNNLL) {
         Log::error("CalcSCETppV012") << "0-jet at NLO, 1-jet at LO, running not at NLL or NNLL is not supported" << Log::end;
      }
   }
   // 0-jet LO, 1-jet NLO
   // not supported
   else if (_fixed0jet == FixedOrder::kLO && _fixed1jet == FixedOrder::kNLO) {
      Log::error("CalcSCETppV012") << "0-jet at LO, 1-jet at NLO is not supported" << Log::end;
   }
   // 0-jet NLO, 1-jet NLO
   // only NNLL which is treated as NNLL'
   else if (_fixed0jet == FixedOrder::kNLO && _fixed1jet == FixedOrder::kNLO) {
      if (_run0 != RunOrder::kNNLL) {
         Log::error("CalcSCETppV012") << "0-jet at NLO, 1-jet at NLO, running not at NNLL is not supported" << Log::end;
      }
   }

   else if (_fixed0jet == FixedOrder::kNNLO) {
      Log::error("CalcSCETppV012") << "0-jet at NNLO is not supported" << Log::end;
   }

   assert(bosonId == PdgID::kZ); // TODO: resummation is only valid for Z + jets production, not W + jets

   _sigs.clear();

   // We use the renormalization scale as the nonsingular scale.
   // TODO: Warn the user that factScale != renScale is being ignored.
   _scaleFO = ComponentFactory::instance().getRenormalizationScale(optMap);

   //Check if the correct scale variation set is selected in the fixed-order part
   std::string vary = optMap.getOption("FixedOrder::ScaleVariations");
   EngineNLOBase::ScaleVariation scaleVar = EngineNLOBase::kNone;
   if (vary == "none") scaleVar = EngineNLOBase::kNone;
   else if (vary == "corr") scaleVar = EngineNLOBase::kCorrelated;
   else if (vary == "uncorr") scaleVar = EngineNLOBase::kUncorrelated;
   else if (vary == "uncorrNoExtremes") scaleVar = EngineNLOBase::kUncorrelatedNoExtremes;
   else {
      Log::fatal("CalcSCETppV012") << "Unknown choice of scale variations " << vary << Log::end;
   }
   if (scaleVar != fixedOrderScale[scale])
      Log::error("CalcSCETppV012") << "Mismatch between options for SCET scale "<<optMap.getOption("CalcSCETppV012::scale")<<" and FO scale variations " << vary << Log::end;

   _calc0jet = new CalcNLOppV(optMap, tree, treeComplete, oneLoop, _xRadGen, pdf, alphas, fixed0jet, Ecm, nLightFlavors, &_scaleNS, &_scaleNS, alphaem, _tau0cut, bosonId, decayChannel);

   _sigs += SignatureManager().getSignaturespp2ZJets(decayChannel, 0, _nLight);
   if(fixed0jet == FixedOrder::kNLO)
      _sigs += SignatureManager().getSignaturespp2ZJets(decayChannel, 1, _nLight);

   CalcNLOppVj::Tau0NLO tau0NLO;
   std::string opt = optMap.getOption<std::string>("NJettiness::Tau0NLO");
   if (opt == "Tau0FR")       tau0NLO = CalcNLOppVj::kTau0FR;
   else if (opt == "Tau0FKS") tau0NLO = CalcNLOppVj::kTau0FKS;
   else {
      tau0NLO = CalcNLOppVj::kTau0FKS;
      Log::error("CalcSCETppV012") << "Unknown NJettiness::Tau0NLO " << opt << Log::end;
   }

   opt = optMap.getOption("CalcSCETppV012::nonSingularScale");
   if (opt == "hard")
      _profileNS = khard;
   else if (opt == "beam")
      _profileNS = kbeam;
   else
      Log::error("CalcSCETppV012") << "Unknown CalcSCETppV012::nonSingularScale " << opt << Log::end;

   _calc1jet = new CalcNLOppVj(optMap, tau0NLO, tree, treeComplete, oneLoop, _xRadGen, pdf, alphas, fixed1jet, Ecm, nLightFlavors, &_scaleNS, &_scaleNS, alphaem, _tau0cut, _tau1cut, bosonId, decayChannel);

   _sigs += SignatureManager().getSignaturespp2ZJets(decayChannel, 1, _nLight);
   if ((fixed1jet == FixedOrder::kNLO) or (run1 != RunOrder::kNone)) {
      //If we are running with Tau1 resummation turned on, we need to generate 2-parton signatures even if doing a LO1 calculation.
      _sigs += SignatureManager().getSignaturespp2ZJets(decayChannel, 2, _nLight);
   }

   int order = 0;
   if (_fixed0jet == FixedOrder::kNLO && _fixed1jet == FixedOrder::kLO) { order = 1; }
   else if (_fixed0jet == FixedOrder::kNLO && _fixed1jet == FixedOrder::kNLO) { order = 2; }
   _beamA = new BeamFuncTauB(_pdf->nameA(), _pdf->memberA(), order, _nLight);
   _beamB = new BeamFuncTauB(_pdf->nameB(), _pdf->memberB(), order, _nLight);
   _beamAsing = new BeamFuncTauB(_pdf->nameA(), _pdf->memberA(), order, _nLight);
   _beamBsing = new BeamFuncTauB(_pdf->nameB(), _pdf->memberB(), order, _nLight);

   if (_phi1 == kadd) {
      _includeResummation = optMap.getOption<bool>("CalcSCETppV012::DebugAdditive::IncludeResum");
      _includeResumExpanded = optMap.getOption<bool>("CalcSCETppV012::DebugAdditive::IncludeResummExpanded");
      _includeFixedOrder = optMap.getOption<bool>("CalcSCETppV012::DebugAdditive::IncludeFixedOrder");

      static SplittingMappingppV0To1 mapping0To1(optMap);
      if (pdf->isPositive())
         _splitting0To1 = new SplittingFunction(optMap, &mapping0To1, Ecm, _alphas, _pdf, pow(mVLow/Ecm, 2));
      else {
         Log::warning("CalcSCETppV012") << _pdf->name() << " is not positive-definite and cannot be used in 0 -> 1 SplittingFunction. Using CT10nnlo instead!" << Log::end;
         Collider collider(optMap);
         _positivePdf = ComponentFactory::makePDF(collider.beamA(), collider.beamB(), "CT10nnlo");
         _splitting0To1 = new SplittingFunction(optMap, &mapping0To1, Ecm, _alphas, _positivePdf, pow(mVLow/Ecm, 2));
      }
      assert(_splitting0To1);

      if (_improvedXSfix) {
         double profsmootherfacCentral, profsmootherfacUp, profsmootherfacDown;
         if (!_totalXSfix)
            Log::error("CalcSCETppV012::CalcSCETppV012") << "CalcSCETppV012::ImprovedXSFix requires CalcSCETppV012::TotalXSFix." << Log::end;
         if (optMap.getOption<bool>("CalcSCETppV012::ImprovedXSFix::needsSetup")) {
            if (_includeResumExpanded || _includeFixedOrder) {
               Log::error("CalcSCETppV012::CalcSCETppV012") << "The setupset should only be used with includeResummedExpanded and includeFixedOrder set to false" << Log::end;
            }
            profsmootherfacCentral = 1.;
            profsmootherfacUp = 1.;
            profsmootherfacDown = 1.;
         } else {
            _improvedXSHandler = new ImprovedXSHandler(getSetupFilename(optMap));
            profsmootherfacCentral = _improvedXSHandler->alphaCentral();
            profsmootherfacUp = _improvedXSHandler->alphaUp();
            profsmootherfacDown = _improvedXSHandler->alphaDown();
         }

         _profSmootherFac[kmurun] = profsmootherfacCentral;
         _profSmootherFac[kmuup] = profsmootherfacUp;
         _profSmootherFac[kmuupfix] = profsmootherfacUp;
         _profSmootherFac[kmudown] = profsmootherfacDown;
         _profSmootherFac[kmudownfix] = profsmootherfacDown;
         _profSmootherFac[kmuBup] = profsmootherfacCentral;
         _profSmootherFac[kmuBdown] = profsmootherfacCentral;
         _profSmootherFac[kmuSup] = profsmootherfacCentral;
         _profSmootherFac[kmuSdown] = profsmootherfacCentral;
         _profSmootherFac[kmuxup] = profsmootherfacCentral;
         _profSmootherFac[kmuxdown] = profsmootherfacCentral;
         _profSmootherFac[kcanonical] = profsmootherfacCentral;
         _profSmootherFac[kmuhigherrun] = profsmootherfacCentral;
         _profSmootherFac[kmuhigherup] = profsmootherfacCentral;
         _profSmootherFac[kmuhigherdown] = profsmootherfacCentral;
      }
   } else if (_improvedXSfix)
      Log::error("CalcSCETppV012") << "The improved cross section fix only works with additive matching." << Log::end;

   _split1Jet = false;
   if (_run1 != RunOrder::kNone) {
      if (_run1 != RunOrder::kLL && _run1 != RunOrder::kNLL)
         Log::error("CalcSCETppV012") << "Tau1 resummation is only known up to NLL." << Log::end;

      if (_phi1 != kadd)
         Log::error("CalcSCETppV012") << "Tau1 resummation only works for additive matching." << Log::end;

      _split1Jet = optMap.getOption<bool>("CalcSCETppV012::run1::split1Jet");
      _tau1IR = std::numeric_limits<double>::quiet_NaN();
      if (_split1Jet) {
         _tau1IR = std::max(optMap.getOption<double>("CalcSCETppV012::run1::split1Jet::Tau1IR"), 1.1*_alphas->lambdaQCD());
         if (_tau1IR > _tau1cut)
            Log::error("CalcSCETppV012") << "Tau1IR = " << _tau1IR << " is larger than Tau1cut = " << _tau1cut << Log::end;
         _onlyExtraSplit1JetContribution = optMap.getOption<bool>("CalcSCETppV012::run1::split1Jet::OnlyExtraContribution");
      }
      _useU1ExpOndSigmaCdPhi1 = optMap.getOption<bool>("CalcSCETppV012::run1::useU1ExpOndSigmaCdPhi1");
      _includeU1Kinematics = optMap.getOption<bool>("CalcSCETppV012::run1::includeKinematics");

      static SplittingMappingppV1To2 mapping1To2(optMap);
      if (pdf->isPositive())
         _splitting1To2 = new SplittingFunction(optMap, &mapping1To2, Ecm, _alphas, _pdf, pow(mVLow/Ecm, 2));
      else {
         Log::warning("CalcSCETppV012") << _pdf->name() << " is not positive-definite and cannot be used in 1 -> 2 SplittingFunction. Using CT10nnlo instead!" << Log::end;
         if (_positivePdf == 0) {
            Collider collider(optMap);
            _positivePdf = ComponentFactory::makePDF(collider.beamA(), collider.beamB(), "CT10nnlo");
         }
         _splitting1To2 = new SplittingFunction(optMap, &mapping1To2, Ecm, _alphas, _positivePdf, pow(mVLow/Ecm, 2));
      }
      assert(_splitting1To2);

      _tau1Max = optMap.getOption<double>("CalcSCETppV012::run1::tau1Max");
      _tau0OverTau1Max = optMap.getOption<double>("CalcSCETppV012::DebugTau1Res::Tau0OverTau1Max");
   }

   // TODO: Get rid of these
   _includeTau1Resummed = optMap.getOption<bool>("CalcSCETppV012::DebugTau1Res::IncludeResummed");
   _includeTau1NonSingFixedOrder = optMap.getOption<bool>("CalcSCETppV012::DebugTau1Res::IncludeNonSingFO");
   _includeTau1NonSingFromTau1 = optMap.getOption<bool>("CalcSCETppV012::DebugTau1Res::IncludeNonSingFromTau1");
   _includeISRSplitting = optMap.getOption<bool>("CalcSCETppV012::DebugTau1Res::IncludeISRSplitting");
   _includeFSRSplitting = optMap.getOption<bool>("CalcSCETppV012::DebugTau1Res::IncludeFSRSplitting");

   _nonsingularPhi2Treatment = "fixed";

   _removeAlpha2 = optMap.getOption<bool>("CalcSCETppV012::DebugTau1Res::RemoveAlpha2");
}

//------------------------------------------------------------------------------
CalcSCETppV012::~CalcSCETppV012()
{
   if (_improvedXSHandler)
      delete _improvedXSHandler;
   delete _splitting0To1;
   delete _splitting1To2;
   delete _positivePdf;
   delete _beamA;
   delete _beamB;
   delete _beamAsing;
   delete _beamBsing;
   delete _calc0jet;
   delete _calc1jet;
   delete _xRadGen;
}

//------------------------------------------------------------------------------
Sigma CalcSCETppV012::fastEstimate(const Configuration& config)
{
   _scaleNS.setValue(getmuFO(config, 0.));
   return math_constants::gev2pb() * _calc0jet->getLOCoeff(config, false);
}

//------------------------------------------------------------------------------
size_t CalcSCETppV012::getNumVariations() const
{
   switch (_inputScale) {
      case kall:     return 10;
      case kfastset: return 4;
      case khardset: return 2;
      case kfixset:  return 2;
      default: return 0;
   }
}

//------------------------------------------------------------------------------
const std::string CalcSCETppV012::name() const
{
   return "CalcSCETppV012";
}

//------------------------------------------------------------------------------
const std::vector<NameValuePair> CalcSCETppV012::componentInfo() const
{
   std::vector<NameValuePair> components;
   components.push_back(NameValuePair("alphas", _alphas->name()));
   components.push_back(NameValuePair("alphaem", _alphaem->name()));
   components.push_back(NameValuePair("treeME", _treeME->name()));
   components.push_back(NameValuePair("hard function", "my own"));
   components.push_back(NameValuePair("beam function A", _beamA->name()));
   components.push_back(NameValuePair("beam function B", _beamB->name()));
   components.push_back(NameValuePair("soft function", "my own"));
   components.push_back(NameValuePair("########## 0-jet components #########", ""));
   components.push_back(NameValuePair("0-jet NLO calculation", _calc0jet->name()));
   const std::vector<NameValuePair>& ZeroJetComponents = _calc0jet->componentInfo() ;
   components.insert(components.end(), ZeroJetComponents.begin(), ZeroJetComponents.end());
   components.push_back(NameValuePair("########## 1-jet components #########", ""));
   components.push_back(NameValuePair("1-jet NLO calculation", _calc1jet->name()));
   const std::vector<NameValuePair>& OneJetComponents1 = _calc1jet->componentInfo() ;
   components.insert(components.end(), OneJetComponents1.begin(), OneJetComponents1.end());
   return components;
}

//---------------------------------------------------------------------------
void CalcSCETppV012::informGenerationWeight(const double w)
{
   _calc0jet->informGenerationWeight(w);
   _calc1jet->informGenerationWeight(w);
}

//------------------------------------------------------------------------------
void CalcSCETppV012::initialize()
{
   _calc0jet->initialize();
   _calc1jet->initialize();
}

//------------------------------------------------------------------------------
void CalcSCETppV012::finalize(size_t numEvents)
{
   _calc0jet->finalize(numEvents);
   _calc1jet->finalize(numEvents);
   if (_splitting0To1)
      _splitting0To1->printInfo();
   if (_splitting1To2)
      _splitting1To2->printInfo();
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/// \todo: this should really check if config has the correct signature as well
bool CalcSCETppV012::Restriction::contains(const Configuration& config)
{
   return _res01jet->contains(config) || _res12jet->contains(config);
}

//------------------------------------------------------------------------------
/// \todo: this is undefined if config does not have a correct signature
size_t CalcSCETppV012::Restriction::countJets(const Configuration& config) const
{
   const ptrdiff_t nFinalQCD = std::count_if(config.beginFinal(), config.endFinal(), IsQCD());
   if (nFinalQCD == 0)
      return 0;
   if (nFinalQCD == 1)
      return _res01jet->countJets(config);
   return _res12jet->countJets(config);
}

//------------------------------------------------------------------------------
CalcSCETppV012::ImprovedXSHandler::ImprovedXSHandler(const std::string& fileName)
{
   size_t numEvents;
   double weightGE1[6] = {0., 0., 0., 0., 0., 0.};
   double weightGE1Sq[6] = {0., 0., 0., 0., 0., 0.};

   std::ifstream XSFile(fileName.c_str());
   if (!XSFile.good())
      Log::fatal("CalcSCETppV012::ImprovedXSHandler") << "Could not open the cross-section info file " << fileName << Log::end;

   std::stringstream ss;
   std::string line;

   XSFile >> numEvents;
   for (size_t i = 0; i < 6; ++i)
      XSFile >> weightGE1[i];

   for (size_t i = 0; i < 6; ++i)
      XSFile >> weightGE1Sq[i];

   double average[6];
   double error[6];
   const double norm = static_cast<double>(numEvents);
   for (int i = 0; i < 6; ++i) {
      average[i] = weightGE1[i] / norm;
      error[i] = sqrt((weightGE1Sq[i] / norm - average[i]*average[i]) / (norm - 1.));
   }

   _alphaCentral = average[3] / average[0];
   _alphaUp = average[4] / average[1];
   _alphaDown = average[5] / average[2];

   Log::verboseInfo("CalcSCETppV012::ImprovedXSHandler") << "The alpha factors are:"
      << "\ncentral: " << _alphaCentral << " +- "
      << _alphaCentral * sqrt(pow(error[3]/average[3],2) + pow(error[0]/average[0],2))
      << "\nup:      " << _alphaUp << " +- "
      << _alphaUp * sqrt(pow(error[4]/average[4],2) + pow(error[1]/average[1],2))
      << "\ndown:    " << _alphaDown << " +- "
      << _alphaDown * sqrt(pow(error[5]/average[5],2) + pow(error[2]/average[2],2))
      << Log::end;
}

//------------------------------------------------------------------------------
Sigma CalcSCETppV012::value(const Configuration& config, bool doVariations)
{
   _spectrumTau0Resummed.clear();
   _spectrumTau0Expanded.clear();
   _cumDerivTau0Resummed.clear();
   _cumDerivTau0Expanded.clear();

   //Let's evaluate the fixed order coefficients since they will be needed multiple times
   NJettinessppRecursive TauN(config);
   const double Tau0 = ZeroJettinesspp(config, TauN.y())();
   const double Tau1 = TauN(1);

   //Discard 2-parton events with Tau0 < Tau0cut
   ptrdiff_t numFinalQCD = std::count_if(config.beginFinal(), config.endFinal(), IsQCD());
   if (numFinalQCD == 2 and Tau0 < _tau0cut)
      return Sigma();

   //First, calculate the various fixed order coefficients that are needed in various pieces of the calculation
   NloCoefficients sigma0Coeffs, sigma1Coeffs, sigma2Coeffs;

   //First, consider multiplicative matching
   if (_phi1 == kmult) {
      if (Tau0 < _tau0cut) {
         sigma0Coeffs = calculateFO0Coeffs(config, doVariations);
         if (not sigma0Coeffs.isDefined()) {
            return Sigma();
         }
      } else if (Tau1 < _tau1cut) {
         sigma1Coeffs = calculateFO1Coeffs(config, Tau0, doVariations);
         if (not sigma1Coeffs.isDefined()) {
            return Sigma();
         }
      } else {
         sigma2Coeffs = calculateFO2Coeffs(config, Tau0, doVariations);
         if (not sigma2Coeffs.isDefined()) {
            return Sigma();
         }
      }
   }

   //Second, consider additive matching
   else if (_phi1 == kadd) {
      if (_includeFixedOrder) {
         if (Tau0 < _tau0cut) {
            sigma0Coeffs = calculateFO0Coeffs(config, doVariations);
            if (not sigma0Coeffs.isDefined()) {
               return Sigma();
            }
         } else if (Tau1 < _tau1cut) {
            sigma1Coeffs = calculateFO1Coeffs(config, Tau0, doVariations);
            if (!sigma1Coeffs.isDefined()) {
               // these checks assume that this can only happen for numFinalQCD == 2
               assert(numFinalQCD == 2);
               if (!_split1Jet || Tau1 < _tau1IR)
                  return Sigma();
            }
         } else {
            sigma2Coeffs = calculateFO2Coeffs(config, Tau0, doVariations);
            if (not sigma2Coeffs.isDefined()) {
               return Sigma();
            }
         }
      }

      // Do the projections and required NLO calculations.
      if (numFinalQCD == 1) {
         // Do the 1->0 projection and corresponding 0->1 splitting probability.
         _phi0Proj.config.clear();
         _phi0Proj.exists = _splitting0To1->projectToPhiN(config, _phi0Proj.config);
         if (_phi0Proj.exists) {
            // Get the 0->1 splitting probability
            if (_includeResummation || _includeResumExpanded) {
               PhiNplus1 phi(config);
               TauN.cluster(phi.emitted(), phi.sister());
               _phi0Proj.splittingProb = (*_splitting0To1)(phi);
            }
         }
      } else if (numFinalQCD == 2) {
         // Do the 2->1 projection, but only if it might be needed.
         _phi1Proj.exists = false;
         const bool needSigma2 = (Tau1 >= _tau1cut) && sigma2Coeffs.lo().isDefined() && (_run1 != RunOrder::kNone);
         const bool needSigma2Split = _split1Jet && (Tau1 >= _tau1IR) && (Tau1 < _tau1cut);
         if (needSigma2 || needSigma2Split) {
            _phi1Proj.config.clear();
            _phi1Proj.exists = _splitting1To2->projectToPhiN(config, _phi1Proj.config);
         }

         if (_phi1Proj.exists) {
#ifdef REFINE_MOMENTA
            enforceMomentumConservation(_phi1Proj.config, "CalcSCETppV012::value", REFINE_PHIPROJ_PRECISION, REFINE_PHIPROJ_ACCURACY);
#endif
            if (needSigma2Split) {
               _phi1Proj.tau0 = ZeroJettinesspp(_phi1Proj.config)();
               // Check if we actually need it.
               if (_phi1Proj.tau0 >= _tau0cut && Tau1 * _tau0OverTau1Max < _phi1Proj.tau0) {
                  _phi1Proj.coeffs = calculateFO1Coeffs(_phi1Proj.config, _phi1Proj.tau0, doVariations);
                  assert(_phi1Proj.coeffs.isDefined());
               } else
                  _phi1Proj.exists = false;
            } else {
               // If it is only needed by sigma2 avoid calculating R-S.
               assert (needSigma2);
               _scaleNS.setValue(getmuNS(_inputScale, Tau0, getmuFO(_phi1Proj.config, Tau0)));
               _phi1Proj.coeffs = _calc1jet->getFOCoeffsBornKin(_phi1Proj.config, doVariations);
            }
         }
         if (_phi1Proj.exists) {
            // Get the corresponding 1->2 splitting probability.
            PhiNplus1 phi2(config);
            TauN.cluster(phi2.emitted(), phi2.sister());
            _phi1Proj.splittingProb = (*_splitting1To2)(phi2);

            // Do the 1->0 projection and corresponding 0->1 splitting probability
            // on the projected Phi1 point
            _phi0Proj.config.clear();
            _phi0Proj.exists = _splitting0To1->projectToPhiN(_phi1Proj.config, _phi0Proj.config);
            if (_phi0Proj.exists) {
               if (_includeResummation || _includeResumExpanded) {
                  PhiNplus1 phi(_phi1Proj.config);
                  NJettinessppRecursive(phi).cluster(phi.emitted(), phi.sister());
                  _phi0Proj.splittingProb = (*_splitting0To1)(phi);
               }
            }
         }
      }
   } else {
      Log::error("CalcSCETppV012::valueset") << "Only multiplicative and additive matching is implemented" << Log::end;
      return 0.;
   }

   // TODO: is there a better way to achieve the independence from the random below?
   RandomState ranstate = getRandomState();
   Sigma result;
   if (_optMap.getOption<bool>("CalcSCETppV012::ImprovedXSFix::needsSetup")) {

      //First get the result for the improved cross-section fix with alpha = 1
      _improvedXSfix = true;
      _totalXSfix = true;
      double centralAlpha1 = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmurun);
      initializeRandomState(ranstate);
      double upAlpha1 = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuupfix);
      initializeRandomState(ranstate);
      double downAlpha1 = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmudownfix);

      //Now get the result with total cross-section fix but without the improved cross-section fix
      _improvedXSfix = false;
      _totalXSfix = true;
      initializeRandomState(ranstate);
      double centralNoImprovedFix = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmurun);
      initializeRandomState(ranstate);
      double upNoImprovedFix = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuupfix);
      initializeRandomState(ranstate);
      double downNoImprovedFix = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmudownfix);

      //Subtract from the above the result without any cross-section fix
      _improvedXSfix = false;
      _totalXSfix = false;
      initializeRandomState(ranstate);
      double centralNoFix = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmurun);
      initializeRandomState(ranstate);
      double upNoFix = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuupfix);
      initializeRandomState(ranstate);
      double downNoFix = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmudownfix);

      // reset the flags
      _totalXSfix = true;
      _improvedXSfix = true;

      //fill the results
      result.central() = centralAlpha1 - centralNoFix;
      result.variations().resize(5);
      result.variations()[0] = upAlpha1 - upNoFix;
      result.variations()[1] = downAlpha1 - downNoFix;
      result.variations()[2] = centralNoImprovedFix - centralNoFix;
      result.variations()[3] = upNoImprovedFix - upNoFix;
      result.variations()[4] = downNoImprovedFix  - downNoFix;

   } else if (_inputScale == kall) {
      result.central() = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmurun);
      if (doVariations) {
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuup));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmudown));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuBup));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuBdown));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuSup));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuSdown));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuxup));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuxdown));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuupfix));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmudownfix));
      }
   } else if (_inputScale == kfastset) {
      result.central() = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmurun);
      if (doVariations) {
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuup));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmudown));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuSup));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuSdown));
      }
   } else if (_inputScale == khardset) {
      result.central() = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmurun);
      if (doVariations) {
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuup));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmudown));
      }
   } else if (_inputScale == kfixset) {
      result.central() = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmurun);
      if (doVariations) {
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmuupfix));
         initializeRandomState(ranstate);
         result.variations().push_back(value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, kmudownfix));
      }
   } else {
      result.central() = value(config, Tau0, Tau1, sigma0Coeffs, sigma1Coeffs, sigma2Coeffs, _inputScale);
   }

   // ensure that we don't get nan's
   if (!std::isfinite(result.central())) {
      Log::error("CalcSCETppV012::valueset") << "result= " << result << " in " << name() <<  Log::end;
   }

   // If the central and all variations are exactly zero, return an undefined Sigma.
   // In principle, this can still happen in some cases with split1Jet.
   if (result.central() == 0.0) {
      bool isZero = true;
      for (Sigma::const_iterator it = result.variations().begin(); it != result.variations().end(); ++it) {
         if (*it != 0.0) {
            isZero = false;
            break;
         }
      }
      if (isZero)
         return Sigma();
   }
   result.setDefined();
   result *= math_constants::gev2pb();
   return result;
}

//------------------------------------------------------------------------------
double CalcSCETppV012::value(const Configuration& config, double Tau0, double Tau1,
                             const NloCoefficients& sigma0Coeffs, const NloCoefficients& sigma1Coeffs,
                             const NloCoefficients& sigma2Coeffs, Scale scale)
{
   ptrdiff_t numFinalQCD = std::count_if(config.beginFinal(), config.endFinal(), IsQCD());
   //First, consider multiplicative matching
   if (_phi1 == kmult) {
      if ((Tau0 < _tau0cut) and (numFinalQCD == 0 or numFinalQCD == 1))
         return sigma0(config, _tau0cut, sigma0Coeffs, scale);
      else if ((Tau0 >= _tau0cut) and (Tau1 < _tau1cut) and (numFinalQCD == 1 or numFinalQCD == 2)) {
         return sigma1(config, Tau0, _tau1cut, sigma1Coeffs, scale);
      } else if ((Tau0 >= _tau0cut) and (Tau1 >= _tau1cut) and (numFinalQCD == 2)) {
         return sigma1(config, Tau0, _tau1cut, sigma2Coeffs, scale);
      } else {
         ///NOTE: 2-parton events with Tau0 < Tau0cut are discarded because they only contribute at NNLO
         assert(numFinalQCD == 2);
         return 0.0;
      }
   }
   //Second, consider additive matching
   else if (_phi1 == kadd) {
      if ((Tau0 < _tau0cut) and (numFinalQCD == 0 or numFinalQCD == 1)) {
         return sigma0(config, _tau0cut, sigma0Coeffs, scale);
      } else if ((Tau0 >= _tau0cut) and (Tau1 < _tau1cut)) {
         assert(numFinalQCD == 1 || numFinalQCD == 2);
         if (_split1Jet) {
            //If we are running with _split1jet, we split the 1-jet into an exclusive 1-jet contribution and inclusive 2-jet constribution
            //The phi1 contribution is the usual one multiplied by a Sudakov factor
            if (numFinalQCD == 1) {
               double sig1 = sigma1(config, Tau0, _tau1cut, sigma1Coeffs, scale);

               const double Tau1max = std::min(_tau1cut, Tau0 / _tau0OverTau1Max);
               const double U1 = tau1Sudakov(config, _tau1IR, Tau1max);
               if (_onlyExtraSplit1JetContribution)
                  return sig1 * (U1 - 1.);
               else
                  return sig1 * U1;
            }
            assert(numFinalQCD == 2);
            //The phi2 contribution is the usual one, plus a contribution for Tau1 > Tau1IR
            double result = sigma2Split(Tau1, scale);
            if (sigma1Coeffs.lo().isDefined() && (!_onlyExtraSplit1JetContribution))
               result += sigma1(config, Tau0, _tau1cut, sigma1Coeffs, scale);
            return result;
         } else {
            return sigma1(config, Tau0, _tau1cut, sigma1Coeffs, scale);
         }
      } else if ((Tau0 >= _tau0cut) and (Tau1 >= _tau1cut)) {
         assert(numFinalQCD == 2);
         if (sigma2Coeffs.lo().isDefined())
            return sigma2(config, Tau0, Tau1, sigma2Coeffs, scale);
         else
            return 0.;
      } else {
         ///NOTE: 2-parton events with Tau0 < Tau0cut are discarded because they only contribute at NNLO
         assert(numFinalQCD == 2);
         return 0.0;
      }
   } else {
      Log::error("CalcSCETppV012::value") << "Only multiplicative and additive matching is implemented" << Log::end;
      return 0.;
   }
}

// ~~~ checked against Mathematica ~~~ //
//------------------------------------------------------------------------------
/// Computes the resummed integrated 0-jet rate as function of TauB0cut.
double CalcSCETppV012::sigma0(const Configuration& config, double Tau0cut, const NloCoefficients& sigma0Coeffs, Scale scale)
{
   //First, consider multiplicative matching
   if (_phi1 == kmult) {
      // Get the fixed order terms
      std::pair<double, double> sig0FixedOrder(0., 0.);
      sig0FixedOrder = sigma0NLO(config, sigma0Coeffs, scale);
      assert(std::isfinite(sig0FixedOrder.first));
      assert(std::isfinite(sig0FixedOrder.second));

      //Get the resummed pieces
      std::pair<double, double> tau0Cum = tau0Cumulant(config, _tau0cut, scale);

      const ptrdiff_t nQCDFinal = std::count_if(config.beginFinal(), config.endFinal(), IsQCD());

      // final combination: the QCD non singular terms (nQCDFinal=1) does not receive any resummation constribution
      if (nQCDFinal  == 0) {
         double fullyDifferential = fullyDifferential0JetNoPDF(config);
         return sig0FixedOrder.first + sig0FixedOrder.second + (tau0Cum.first - tau0Cum.second) * fullyDifferential;
      } else {
         return sig0FixedOrder.second ;
      }
   }

   //Second, consider additive matching
   else if (_phi1 == kadd) {
      //The entire sigma0 contribution goes into the term included with _includeTau1Resummed
      if (!_includeTau1Resummed)
         return 0.;

      // Get the fixed order terms
      std::pair<double, double> sig0FixedOrder(0., 0.);
      if (_includeFixedOrder)
         sig0FixedOrder = sigma0NLO(config, sigma0Coeffs, scale);
      assert(std::isfinite(sig0FixedOrder.first));
      assert(std::isfinite(sig0FixedOrder.second));
      //Get the resummed pieces
      std::pair<double, double> tau0Cum(0., 0.);
      if (_includeResummation || _includeResumExpanded) {
         tau0Cum = tau0Cumulant(config, _tau0cut, scale);
      }

      double tau0CumulantResummed = 0.;
      double tau0CumulantResummedExpanded = 0.;
      if (_includeResummation)
         tau0CumulantResummed = tau0Cum.first;
      if (_includeResumExpanded)
         tau0CumulantResummedExpanded = tau0Cum.second;

      const ptrdiff_t nQCDFinal = std::count_if(config.beginFinal(), config.endFinal(), IsQCD());

      // final combination: the QCD non singular terms (nQCDFinal=1) does not receive any resummation constribution
      if (nQCDFinal  == 0) {
         double fullyDifferential = fullyDifferential0JetNoPDF(config);
         return sig0FixedOrder.first + sig0FixedOrder.second + (tau0CumulantResummed - tau0CumulantResummedExpanded) * fullyDifferential;
      } else {
         return sig0FixedOrder.second ;
      }
   } else {
      Log::error("CalcSCETppV012::sigma0") << "Only multiplicative and additive matching is implemented" << Log::end;
      return 0.;
   }

}

//------------------------------------------------------------------------------
/// Computes the final expression for the exclusive 1-jet cross section
double CalcSCETppV012::sigma1(const Configuration& config, double Tau0, double Tau1cut, const NloCoefficients& sigma1Coeffs, Scale scale)
{
   //First, consider multiplicative matching
   if (_phi1 == kmult) {
      double result = 0.;

      double xA, xB;
      setBeamFuncXs(config, xA, xB);

      double fixedOrder = 0.;

      fixedOrder = extractFixedOrderScale(sigma1Coeffs.sum(), scale);
      if (fixedOrder == 0.0)
         return 0.0;
      std::pair<double, double> tau0Spec(0., 0.);
      tau0Spec = tau0Spectrum(config, Tau0, scale);
      if (tau0Spec.second == 0.0) {
         assert(tau0Spec.first == 0.0); // should never have expanded 0.0 and not also resummed (except for dsig1 = 0.0)
         // In a region with exactly no resummation both resummed and expanded can be exactly equal to 0.0,
         // in which case their ratio is meant to be unity, so we just don't multiply by it.
         result = fixedOrder;
      } else {
         result = fixedOrder * tau0Spec.first / tau0Spec.second;
      }
      return result;
   }

   //Second, consider additive matching
   else if (_phi1 == kadd) {

      double fixedOrderLO = 0.;
      double fixedOrderBornKin = 0.;
      double fixedOrderRealKin = 0.;
      // Get the fixed order terms
      if (_includeFixedOrder) {
         fixedOrderLO = extractFixedOrderScale(sigma1Coeffs.lo(), scale);
         fixedOrderBornKin = extractFixedOrderScale(sigma1Coeffs.lo(), scale) + extractFixedOrderScale(sigma1Coeffs.nloBornKin(), scale);
         fixedOrderRealKin = extractFixedOrderScale(sigma1Coeffs.nloRealKin(), scale);
      }

      ptrdiff_t nQCDFinal = std::count_if(config.beginFinal(), config.endFinal(), IsQCD());
      if (nQCDFinal == 2) {
         //These events are non-singular in Tau1, so I can turn off Tau1 resummation.
         //The simplest is to give these events a fixed order weight. If I want to keep the Tau0 resummation, I could use multiplicative matching for this.
         //There is a debug parameter we can set to decide which option to use here.
         if (_nonsingularPhi2Treatment == "fixed") {
            if (_includeFixedOrder and _includeTau1NonSingFixedOrder)
               return fixedOrderBornKin + fixedOrderRealKin;
            else
               return 0.;
         } else if (_nonsingularPhi2Treatment == "zero") {
            return 0.;
         }
      }

      double dSig1InclusiveSingular = dSigmaCdPhi1(config, Tau0, sigma1Coeffs, scale);

      //Now compute the U1 Sudakov and its expansion
      double tau1Sud = tau1Sudakov(config, Tau1cut, Tau0 / _tau0OverTau1Max);
      double tau1SudExp = 0.;
      if (_includeFixedOrder and !_removeAlpha2 and (_fixed1jet == FixedOrder::kNLO))
         tau1SudExp = tau1SudakovExpanded(config, Tau1cut, Tau0 / _tau0OverTau1Max, getmuNS(scale, Tau0, getmuFO(config, Tau0)));

      double resummedCompleteTau1 = dSig1InclusiveSingular * tau1Sud;

      double result = 0.;
      if (_includeTau1Resummed)
         result += resummedCompleteTau1;
      if (_includeTau1NonSingFixedOrder)
         result += fixedOrderRealKin;
      if (_includeTau1NonSingFromTau1) {
         if (_useU1ExpOndSigmaCdPhi1)
            result -= dSig1InclusiveSingular * tau1SudExp;
         else
            result -= fixedOrderLO * tau1SudExp;
      }

      return  result;

   } else {
      Log::error("CalcSCETppV012::value") << "Only multiplicative and additive matching is implemented" << Log::end;
      return 0.;
   }

}

//------------------------------------------------------------------------------
/// Computes the final expression for the inclusive 2-jet cross section
double CalcSCETppV012::sigma2(const Configuration& config, double Tau0, double Tau1, const NloCoefficients& sigma2Coeffs, Scale scale)
{
   //Make sure we have only 2 parton events
   assert(std::count_if(config.beginFinal(), config.endFinal(), IsQCD()) == 2);

   //We should only get here for additive matching
   assert(_phi1 == kadd);

   //Get the fixed order term
   double B2 = 0.;
   //since at the moment we are only including the 2 jet at FO, if not requested it returns 0
   if (_includeFixedOrder and _includeTau1NonSingFixedOrder) {
      switch(_fixed1jet) {
         case  FixedOrder::kLO :
            B2 = 0.;
            break;
         case FixedOrder::kNLO:
            switch(resummationScale[scale]) {
               case kmuup: {
                  B2 = (_inputScale == kmuup || _inputScale == kmuupfix) ? sigma2Coeffs.nlo().central() : sigma2Coeffs.nlo().variations()[0];
                  break;
               }
               case kmudown: {
                  B2 = (_inputScale == kmudown || _inputScale == kmudownfix) ?  sigma2Coeffs.nlo().central() : sigma2Coeffs.nlo().variations()[1];
                  break;
               }
               default:
                  B2 = sigma2Coeffs.nlo().central();
                  break;
            }
            break;
         default: {
            Log::error("CalcSCETppV012::sigma2") << "2-jet rate only available at LO!" << Log::end;
            return std::numeric_limits<double>::quiet_NaN();
         }
      }
   }

   // Return the fixed-order result without Tau1 resummation.
   if (_run1 == RunOrder::kNone)
      return B2;

   // Return the fixed-order result if there is no valid 2->1 projection.
   if (!_phi1Proj.exists)
      return B2;

   double dSig1InclusiveSingular = dSigmaCdPhi1(_phi1Proj.config, Tau0, _phi1Proj.coeffs, scale);

   //Now calculate the Sudakov factor
   //Get the derivative of the Tau1 Sudakov and its expansion
   double tau1SudakovDeriv = tau1SudakovDerivative(_phi1Proj.config, Tau1, Tau0 / _tau0OverTau1Max);
   double tau1SudakovExpDeriv = 0.;
   if (!_removeAlpha2 and (_fixed1jet == FixedOrder::kNLO))
      tau1SudakovExpDeriv = tau1SudakovExpandedDerivative(_phi1Proj.config, Tau1, Tau0 / _tau0OverTau1Max, getmuNS(scale, Tau0, getmuFO(_phi1Proj.config, Tau0)));

   double B1;
   if (not _includeFixedOrder)
      B1 = 0.;
   else {
      switch(resummationScale[scale]) {
         case kmuup: {
            B1 = (_inputScale == kmuup || _inputScale == kmuupfix) ? _phi1Proj.coeffs.lo().central() : _phi1Proj.coeffs.lo().variations()[0];
            break;
         }
         case kmudown: {
            B1 = (_inputScale == kmudown || _inputScale == kmudownfix) ?  _phi1Proj.coeffs.lo().central() : _phi1Proj.coeffs.lo().variations()[1];
            break;
         }
         default:
            B1 = _phi1Proj.coeffs.lo().central();
            break;
      }
   }

   double result = 0.;
   if (_includeTau1Resummed)
      result += dSig1InclusiveSingular * _phi1Proj.splittingProb * tau1SudakovDeriv;
   if (_includeTau1NonSingFixedOrder)
      result += B2;
   if (_includeTau1NonSingFromTau1) {
      if (_useU1ExpOndSigmaCdPhi1)
         result -= dSig1InclusiveSingular * tau1SudakovExpDeriv;
      else
         result -= B1 * _phi1Proj.splittingProb * tau1SudakovExpDeriv;
   }

   return result;
}

//------------------------------------------------------------------------------
/// Computes the final expression for the exclusive 1-jet cross section
double CalcSCETppV012::sigma2Split(double Tau1, Scale scale)
{
   // We should only get here for additive matching and split1Jet
   assert(_phi1 == kadd && _split1Jet);

   // Return if there is no valid 2->1 projection.
   if (!_phi1Proj.exists)
      return 0.;

   // These must hold if _phi1Proj exists.
   assert(Tau1 >= _tau1IR && Tau1 < _tau1cut);
   assert(_phi1Proj.tau0 >= _tau0cut);

   const double sig1 = sigma1(_phi1Proj.config, _phi1Proj.tau0, _tau1cut, _phi1Proj.coeffs, scale);

   const double Tau1max = std::min(_tau1cut, _phi1Proj.tau0 / _tau0OverTau1Max);
   const double U1deriv = tau1SudakovDerivative(_phi1Proj.config, Tau1, Tau1max);
   return sig1 * U1deriv * _phi1Proj.splittingProb;
}

//------------------------------------------------------------------------------
double CalcSCETppV012::dSigmaCdPhi1(const Configuration& Phi1, double Tau0, const NloCoefficients& nloCoeffs, Scale scale)
{
   //Make sure that we really have a Phi1 event with nQCDFinal == 1
   assert(std::count_if(Phi1.beginFinal(), Phi1.endFinal(), IsQCD()) == 1);

   //First, get the fixed order pieces with Born kinematics
   double fixedOrderBornKin = 0.;
   // Get the fixed order terms
   if (_includeFixedOrder) {
      if (_removeAlpha2)
         fixedOrderBornKin = extractFixedOrderScale(nloCoeffs.lo(), scale);
      else
         fixedOrderBornKin = extractFixedOrderScale(nloCoeffs.lo(), scale) + extractFixedOrderScale(nloCoeffs.nloBornKin(), scale);
   }

   // Return the fixed-order result if there is no valid 1->0 projection.
   if (!_phi0Proj.exists)
      return fixedOrderBornKin;

   // Get the resummed pieces.
   double resummedTau0 = 0.;
   double resummedExpandedTau0 = 0.;

   if (_includeResummation || _includeResumExpanded) {
      const std::pair<double, double> tau0Spec = tau0Spectrum(Phi1, Tau0, scale);
      double sigma0NoPDF = fullyDifferential0JetNoPDF(_phi0Proj.config);
      if (_includeResummation) {
         resummedTau0 = sigma0NoPDF * tau0Spec.first * _phi0Proj.splittingProb;
      }
      if (_includeResumExpanded) {
         resummedExpandedTau0 = sigma0NoPDF * tau0Spec.second * _phi0Proj.splittingProb;
      }
   }

   return resummedTau0 + fixedOrderBornKin - resummedExpandedTau0;
}

// ~~~ checked against POWHEG Z ~~~ //
//------------------------------------------------------------------------------
/// Total contributions to the cumulant at fixed NLO
std::pair<double, double> CalcSCETppV012::sigma0NLO(const Configuration& config, const NloCoefficients& sigma0Coeffs, Scale scale)
{
   switch(resummationScale[scale]) {
      case kmuup: {
         IRestriction* restrUp = _calc0jet->getRestriction();
         const size_t nJetsFound = restrUp->countJets(config);
         delete restrUp;
         return (nJetsFound != std::numeric_limits<size_t>::max()) ?
                ((_inputScale == kmuup || _inputScale == kmuupfix) ?
                 std::make_pair(sigma0Coeffs.loCentral(), sigma0Coeffs.nloCentral()) : std::make_pair(sigma0Coeffs.loVary(0), sigma0Coeffs.nloVary(0)))
                   : std::make_pair(0.0, 0.0);
      }
      case kmudown: {
         IRestriction* restrDown = _calc0jet->getRestriction();
         const size_t nJetsFound = restrDown->countJets(config);
         delete restrDown;
         return (nJetsFound != std::numeric_limits<size_t>::max()) ?
                ((_inputScale == kmudown || _inputScale == kmudownfix) ?
                 std::make_pair(sigma0Coeffs.loCentral(), sigma0Coeffs.nloCentral()) : std::make_pair(sigma0Coeffs.loVary(1), sigma0Coeffs.nloVary(1)))
                   :  std::make_pair(0.0, 0.0);
      }
      default: {
         IRestriction* restrCentral =  _calc0jet->getRestriction();
         const size_t nJetsFound = restrCentral->countJets(config);
         delete restrCentral;
         return (nJetsFound != std::numeric_limits<size_t>::max()) ?  std::make_pair(sigma0Coeffs.loCentral(), sigma0Coeffs.nloCentral()) :  std::make_pair(0.0, 0.0);
      }
   }
}

// ~~~ checked against POWHEG Z ~~~ //
//------------------------------------------------------------------------------
/// compute the sigma0 NLO coefficient and store it to save time.
/// \return True if \a config passes the restriction according to the NLO calculation
NloCoefficients CalcSCETppV012::calculateFO0Coeffs(const Configuration& config, bool doVariations)
{
   _scaleNS.setValue(getmuNS(_inputScale, _tau0cut, getmuFO(config, _tau0cut)));
   return _calc0jet->getFOCoeffs(config, doVariations);
}

// ~~~ checked against POWHEG Z ~~~ //
//------------------------------------------------------------------------------
/// compute the dsigma1 fixed-order coefficients and store them to save time.
/// \return True if \a config passes the restriction according to the NLO calculation
NloCoefficients CalcSCETppV012::calculateFO1Coeffs(const Configuration& config, double Tau0, bool doVariations)
{
   _scaleNS.setValue(getmuNS(_inputScale, Tau0, getmuFO(config, Tau0)));
   switch(_fixed1jet) {
      case FixedOrder::kLO: {
         //If we do Tau1 resummation, we need to set the LO coefficient for 2-jet events to zero by hand. If we were to call the calculation, we would get an undefined coefficient, and the remaining calculation would never perform the Tau1 resummation
         ptrdiff_t numFinalQCD = std::count_if(config.beginFinal(), config.endFinal(), IsQCD());
         if (numFinalQCD == 2 && _run1 != RunOrder::kNone) {
            Sigma zero;
            if (_inputScale == kall) {
               zero = Sigma(0.0, std::vector<double>(10));
            } else if (_inputScale == kfastset) {
               zero = Sigma(0.0, std::vector<double>(4));
            } else if (_inputScale == khardset) {
               zero = Sigma(0.0, std::vector<double>(2));
            } else {
               zero = Sigma(0.0);
            }
            return NloCoefficients(zero, zero, zero);
         }
         assert(std::count_if(config.beginFinal(), config.endFinal(), IsQCD()) == 1);
         return _calc1jet->getFOCoeffs(config, doVariations);
      }
      break;
      case FixedOrder::kNLO: {
         // set the NLO coefficients to zero if we are below Tau0cut or removeAlpha2 is set to true
         NloCoefficients nloCoeffs;
         if (_removeAlpha2) {
            Sigma LOCoeff = _calc1jet->getLOCoeff(config, doVariations);
            Sigma zero = Sigma(0.0, std::vector<double>(LOCoeff.variations().size()));
            nloCoeffs = NloCoefficients(LOCoeff, zero, zero);
         } else {
            nloCoeffs = _calc1jet->getFOCoeffs(config, doVariations);
         }
         return nloCoeffs;
      }
      break;
      default: {
         Log::error("CalcSCETppV012::calculateFO1Coeffs") << "1-jet rate only available at LO or NLO!" << Log::end;
         return NloCoefficients();
      }
   }
}

// ~~~ checked against POWHEG Z ~~~ //
//------------------------------------------------------------------------------
/// compute the dsigma1 fixed-order coefficients and store them to save time.
NloCoefficients CalcSCETppV012::calculateFO2Coeffs(const Configuration& config, double Tau0, bool doVariations)
{
   return calculateFO1Coeffs(config, Tau0, doVariations);
}

//------------------------------------------------------------------------------
/**
 * \brief Computes the resummed and expanded differential spectrum as function of TauB0.
 * \return A pair of doubles, the first contains the result for the resummed
 *         spectrum and the second the result for the expanded spectrum.
 *
 * This takes the simple cumulant defined below, and sums over the flavors of the beams
 * if _flavSummedBeams = true
 */
std::pair<double, double> CalcSCETppV012::tau0Cumulant(const Configuration& config, double Tau0cut, Scale scale)
{
   double sig0Resummed = 0;
   double sig0ResumExpanded = 0;

   // set the flavors
   std::vector<PdgID> idBeamFuncA, idBeamFuncB;
   setBeamFuncId(config, idBeamFuncA, idBeamFuncB);

   // set the x values
   double xA, xB;
   setBeamFuncXs(config, xA, xB);

   // get the q value
   FourVector leptons;
   for(Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it)
      if (it->id().isLepton()) leptons += it->p();
   const double Q = leptons.magnitude();
   assert(Q > 0.0);

   double muFO = getmuFO(config, Tau0cut);

   // loop over all possible flavor combinations given the configuration
   for(size_t idindex = 0; idindex < idBeamFuncA.size(); idindex++) {
      // get the id for each iteration
      PdgID idA = idBeamFuncA[idindex];
      PdgID idB = idBeamFuncB[idindex];

      // set the Phi0 phase space
      //TODO: This keeps recalculating the same Y for every flavor combination
      Phi0 phi0(xA, idA, xB, idB, Q);

      //Now, get the resummed contributions
      Scale projscale = resummationScale[scale];
      sig0Resummed += sigma0Resummed(phi0, Tau0cut, projscale, muFO);
      sig0ResumExpanded += sigma0ResumExpanded(phi0, Tau0cut, projscale, muFO);
   }
   // checks for Nan's
   assert(std::isfinite(sig0Resummed));
   assert(std::isfinite(sig0ResumExpanded));
   return std::pair<double, double>(sig0Resummed, sig0ResumExpanded);
}

// ~~~ checked against Mathematica ~~~ //
//------------------------------------------------------------------------------
/**
 * \brief Computes the resummed and expanded differential spectrum as function of TauB0.
 * \return A pair of doubles, the first contains the result for the resummed
 *         spectrum and the second the result for the expanded spectrum.
 *
 * This takes the simple thrust spectrum defined below, and sums over the flavors of the beams
 * if _flavSummedBeams == true, and performs the cross section fix if _totalXSfix == true
 */
std::pair<double, double> CalcSCETppV012::tau0Spectrum(const Configuration& config, double TauB0, Scale scale)
{
   // set the x values
   double xA, xB;
   setBeamFuncXs(config, xA, xB);

   // set the flavors
   std::vector<PdgID> idBeamFuncA, idBeamFuncB;
   if (_flavSummedBeams) {
      // flavors for beam A
      idBeamFuncA.push_back(PdgID::kDown); idBeamFuncA.push_back(PdgID::kUp); idBeamFuncA.push_back(PdgID::kStrange); idBeamFuncA.push_back(PdgID::kCharm); idBeamFuncA.push_back(PdgID::kBottom);
      idBeamFuncA.push_back(PdgID::kDownBar); idBeamFuncA.push_back(PdgID::kUpBar); idBeamFuncA.push_back(PdgID::kStrangeBar); idBeamFuncA.push_back(PdgID::kCharmBar); idBeamFuncA.push_back(PdgID::kBottomBar);
      // flavors for beam B (complement of A)
      idBeamFuncB.push_back(PdgID::kDownBar); idBeamFuncB.push_back(PdgID::kUpBar); idBeamFuncB.push_back(PdgID::kStrangeBar); idBeamFuncB.push_back(PdgID::kCharmBar); idBeamFuncB.push_back(PdgID::kBottomBar);
      idBeamFuncB.push_back(PdgID::kDown); idBeamFuncB.push_back(PdgID::kUp); idBeamFuncB.push_back(PdgID::kStrange); idBeamFuncB.push_back(PdgID::kCharm); idBeamFuncB.push_back(PdgID::kBottom);
   } else {
      setBeamFuncId(config, idBeamFuncA, idBeamFuncB);
   }

   // get the q value
   FourVector leptons;
   for(Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it)
      if (it->id().isLepton()) leptons += it->p();

   const double Q = leptons.magnitude();
   assert(Q > 0.0);

   double muFO = getmuFO(config, TauB0);

   double dsigma0resum = 0.0;
   double dsigma0exp = 0.0;

   // loop over all possible flavor combinations given the configuration
   for(size_t idindex = 0; idindex < idBeamFuncA.size(); idindex++) {
      // get the id for each iteration
      PdgID idA = idBeamFuncA[idindex];
      PdgID idB = idBeamFuncB[idindex];

      // set the Phi0 phase space
      Phi0 phi0(xA, idA, xB, idB, Q);

      double fullyDifferential;

      if (_phi1 == kmult)
         fullyDifferential = lepfac(idA, Q);
      else if (_phi1 == kadd) {
         fullyDifferential = 1.;
      } else {
         fullyDifferential = 0.;
         Log::error("CalcSCETppV012::dsigma0") << "Only multiplicative or additive matching is currently implemented" << Log::end;
      }

      // for most parts of the calculation, we need kmuup/down instead of the XS fix scales
      Scale projscale = resummationScale[scale];
      // now get the resummed, resummed expanded spectrum
      if (_spectrumTau0Resummed.find(projscale) == _spectrumTau0Resummed.end())
         _spectrumTau0Resummed[projscale] = dSigma0Resummed(phi0, TauB0, projscale, muFO);
      if (_spectrumTau0Expanded.find(projscale) == _spectrumTau0Expanded.end())
         _spectrumTau0Expanded[projscale] = dSigma0ResumExpanded(phi0, TauB0, projscale, muFO);

      dsigma0resum += fullyDifferential * _spectrumTau0Resummed[projscale];
      dsigma0exp += fullyDifferential * _spectrumTau0Expanded[projscale];

      // if we are fixing the total XS, do it here
      // for the kmuupfix and kmudownfix scales, we correct for the muup/mudown variations
      // otherwise we correct for the central scale
      if (_totalXSfix) {
         if (scale != kmuupfix && scale != kmudownfix) {
            dsigma0resum += fullyDifferential * deltaSigmaR(phi0, TauB0, kmurun, muFO);
         } else if (scale == kmuupfix) {
            dsigma0resum += fullyDifferential * deltaSigmaR(phi0, TauB0, kmuup, muFO);
         } else if (scale == kmudownfix) {
            dsigma0resum += fullyDifferential * deltaSigmaR(phi0, TauB0, kmudown, muFO);
         }
      }
   }

   return std::pair<double, double>(dsigma0resum, dsigma0exp);
}

//------------------------------------------------------------------------------
///Calculates the resummed cumulant
double CalcSCETppV012::sigma0Resummed(const Phi0& phi0, double Tau0cut, Scale scale, double muFO)
{
   double xA = phi0.xA;
   PdgID idA = phi0.idA;
   double xB = phi0.xB;
   PdgID idB = phi0.idB;
   double Q = phi0.Q;
   double Y = phi0.Y;

   double muH, muB, muS;
   getScales(scale, Tau0cut, Q, Y, muFO, muH, muB, muS);

   // alphas at the hard scale we're running to
   const double as = (*_alphas)(muH);
   const int nf = _alphas->nf();

   // NOTE: Arbitrary scale in resummation xi set to xi = Tau0cut everywhere
   const double asH  = (*_alphas)(muH);
   const double lnxH = log(Q * Q / (muH * muH));

   const double asS  = (*_alphas)(muS);
   const double lnxS = log(Tau0cut / muS);

   const double asBa  = (*_alphas)(muB);
   const double lnxBa = log(Tau0cut * Q  / (muB * muB));

   const double asBb  = (*_alphas)(muB);
   const double lnxBb = log(Tau0cut * Q / (muB * muB));

   double eta = std::numeric_limits<double>::quiet_NaN();

   // set the evolution factor, which is independent of the beam info (x, flavor)
   double U0;
   switch(_run0) {
      case RunOrder::kLL: U0 = U0LL(asH, lnxH, asBa, lnxBa, asBb, lnxBb, asS, lnxS, as, nf, eta); break;
      case RunOrder::kNLL: U0 = U0NLL(asH, lnxH, asBa, lnxBa, asBb, lnxBb, asS, lnxS, as, nf, eta); break;
      case RunOrder::kNNLL: U0 = U0NNLL(asH, lnxH, asBa, lnxBa, asBb, lnxBb, asS, lnxS, as, nf, eta); break;
      default: {
         Log::error("CalcSCETppV012::sigma0resum") << "Unknown run order" << _run0 << Log::end;
         U0 = std::numeric_limits<double>::quiet_NaN();
      }
   }

   // now set the fixed order factors for resummation
   double F0 = 0;

   // if 0-jet is NLO, 1-jet is LO, we need to add the NLO fixed order coefficients
   if (_fixed0jet == FixedOrder::kNLO && _fixed1jet == FixedOrder::kLO) {
      F0 = F0NLO(_beamA, _beamB, asH, lnxH, asBa, lnxBa, muB, asBb, lnxBb, muB, asS, lnxS, eta, xA, xB, Q, idA, idB)
           + (*_beamA)(idA, -1, 0, xA, muB) * (*_beamB)(idB, -1, 0, xB, muB);
   }
   // if 0-jet is NLO, 1-jet is NLO, we need to add the NNLO fixed order coefficients
   if (_fixed0jet == FixedOrder::kNLO && _fixed1jet == FixedOrder::kNLO) {
      F0 = F0NLO(_beamA, _beamB, asH, lnxH, asBa, lnxBa, muB, asBb, lnxBb, muB, asS, lnxS, eta, xA, xB, Q, idA, idB)
          + (*_beamA)(idA, -1, 0, xA, muB) * (*_beamB)(idB, -1, 0, xB, muB);
      if (!_removeAlpha2)
         F0 += F0NNLO(_beamA, _beamB, asH, lnxH, asBa, lnxBa, muB, asBb, lnxBb, muB, asS, lnxS, eta, xA, xB, Q, idA, idB);
   }

   return F0 * U0;
}

//------------------------------------------------------------------------------
// resummed expanded cumulant
double CalcSCETppV012::sigma0ResumExpanded(const Phi0& phi0, double Tau0cut, Scale scale, double muFO)
{
   double xA = phi0.xA;
   PdgID idA = phi0.idA;
   double xB = phi0.xB;
   PdgID idB = phi0.idB;
   double Q = phi0.Q;

   // alphas at the fixed-order nonsingular scale
   const double muNS = getmuNS(scale, Tau0cut, muFO);
   const double as = (*_alphas)(muNS);

   // compute the FO singular at the same time
   // the FO singular is given by the FO factors in the resummation with all scales set to muNS (U0 = 1)
   // this is only true if we are working to a prime order counting
   // if we are working to, e.g., NNLL, we should only have the singular to NLO
   const double lnxFOH = log(Q * Q / (muNS * muNS));
   const double lnxFOB = log(Tau0cut * Q / (muNS * muNS));
   const double lnxFOS = log(Tau0cut / muNS);
   double etaFO = 0;

   // now set the fixed order factors for resummed expanded
   double F0sing = std::numeric_limits<double>::quiet_NaN();
   // We compute the singular terms to the same order as the FO
   if (_fixed0jet == FixedOrder::kNNLO) {
      F0sing = F0NLO(_beamAsing, _beamBsing, as, lnxFOH, as, lnxFOB, muNS, as, lnxFOB, muNS, as, lnxFOS, etaFO, xA, xB, Q, idA, idB)
              + (*_beamAsing)(idA, -1, 0, xA, muNS) * (*_beamBsing)(idB, -1, 0, xB, muNS);
      if (!_removeAlpha2)
         F0sing += F0NNLO(_beamAsing, _beamBsing, as, lnxFOH, as, lnxFOB, muNS, as, lnxFOB, muNS, as, lnxFOS, etaFO, xA, xB, Q, idA, idB);
   } else if (_fixed0jet == FixedOrder::kNLO) {
      F0sing = F0NLO(_beamAsing, _beamBsing, as, lnxFOH, as, lnxFOB, muNS, as, lnxFOB, muNS, as, lnxFOS, etaFO, xA, xB, Q, idA, idB)
               + (*_beamAsing)(idA, -1, 0, xA, muNS) * (*_beamBsing)(idB, -1, 0, xB, muNS);
   } else {
      F0sing = (*_beamAsing)(idA, -1, 0, xA, muNS) * (*_beamBsing)(idB, -1, 0, xB, muNS);
   }
   return F0sing;
}

//------------------------------------------------------------------------------
// resummed spectrum
double CalcSCETppV012::dSigma0Resummed(const Phi0& phi0, double TauB0, Scale scale, double muFO)
{
   assert(TauB0 > _tau0cut);

   double xA = phi0.xA;
   PdgID idA = phi0.idA;
   double xB = phi0.xB;
   PdgID idB = phi0.idB;
   double Q = phi0.Q;
   double Y = phi0.Y;

   double dsigma0resum = 0.0;

   double muH, muB, muS;
   bool resumregion = getScales(scale, TauB0, Q, Y, muFO, muH, muB, muS);

   if (resumregion) {
      // nontrivial scale-separation and so nontrivial resummation
      // NOTE: Arbitrary scale in resummation xi = TauB0 everywhere
      // We use the fixed-order scale as the scale we're running to.
      const double as = (*_alphas)(muH);
      const int nf = _alphas->nf();

      const double asH  = (*_alphas)(muH);
      const double lnxH = log((Q * Q) / (muH * muH));

      const double asS  = (*_alphas)(muS);
      const double lnxS = log(TauB0 / (muS));

      const double asBa  = (*_alphas)(muB);
      const double lnxBa = log(TauB0 * Q / (muB * muB));

      const double asBb  = (*_alphas)(muB);
      const double lnxBb = log(TauB0 * Q / (muB * muB));

      double eta;
      double U0;
      switch(_run0) {
         case RunOrder::kLL: U0 = U0LL(asH, lnxH, asBa, lnxBa, asBb, lnxBb, asS, lnxS, as, nf, eta); break;
         case RunOrder::kNLL: U0 = U0NLL(asH, lnxH, asBa, lnxBa, asBb, lnxBb, asS, lnxS, as, nf, eta); break;
         case RunOrder::kNNLL: U0 = U0NNLL(asH, lnxH, asBa, lnxBa, asBb, lnxBb, asS, lnxS, as, nf, eta); break;
         default: {
            Log::error("CalcSCETppV012::dsigma0Resummed") << "Unknown run order" << _run0 << Log::end;
            U0 = std::numeric_limits<double>::quiet_NaN();
         }
      }

      // now set the fixed order factors for resummation
      double dF0 = 0;

      if (_fixed0jet == FixedOrder::kNLO) {
         if (_fixed1jet == FixedOrder::kLO) {
            // 0-jet at NLO, 1-jet at LO: NLL' or NNLL resummation with 1-loop FO terms
            if (_run0 == RunOrder::kNLL || _run0 == RunOrder::kNNLL) {
               dF0 += dF0NLO(_beamA, _beamB, asH, lnxH, asBa, lnxBa, muB, asBb, lnxBb, muB, asS, lnxS, eta, xA, xB, Q, idA, idB)
                      + eta * (*_beamA)(idA, -1, 0, xA, muB) * (*_beamB)(idB, -1, 0, xB, muB);
            } else {
               Log::error("CalcSCETppV012::dsigma0Resummed") << "inconsistent running/matching: 0-jet NLO, 1-jet LO needs NLL or NNLL!" << Log::end;
            }
         } else { // _fixed1jet == FixedOrder::kNLO
            // 0-jet at NLO, 1-jet at NLO: NNLL' resummation with 2-loop FO terms
            if (_run0 == RunOrder::kNNLL) {
               dF0 += dF0NLO(_beamA, _beamB, asH, lnxH, asBa, lnxBa, muB, asBb, lnxBb, muB, asS, lnxS, eta, xA, xB, Q, idA, idB)
                      + eta * (*_beamA)(idA, -1, 0, xA, muB) * (*_beamB)(idB, -1, 0, xB, muB);
               if (TauB0 > _tau0cut)
                  //Only include the NNLO matching contribution if TauB0 > _tau0cut
                  if (!_removeAlpha2)
                     dF0 += dF0NNLO(_beamA, _beamB, asH, lnxH, asBa, lnxBa, muB, asBb, lnxBb, muB, asS, lnxS, eta, xA, xB, Q, idA, idB);
            } else {
               Log::error("CalcSCETppV012::dsigma0Resummed") << "inconsistent running/matching: 0-jet NLO, 1-jet NLO needs NNLL!" << Log::end;
            }
         }
      } else { // _fixed0jet == FixedOrder::kLO ( => _fixed1jet == FixedOrder::kLO)
         // 0-jet at LO, 1-jet at LO: NLL' resummation with 1-loop FO terms
         // note that with additive matching we would actually do tree level FO terms
         // but with multiplicative matching this is not an option since the tree level resummed expanded is 0
         if (_fixed1jet == FixedOrder::kLO && _run0 == RunOrder::kNLL) {
            dF0 += dF0NLO(_beamA, _beamB, asH, lnxH, asBa, lnxBa, muH, asBb, lnxBb, muH, asS, lnxS, eta, xA, xB, Q, idA, idB)
                   + eta * (*_beamA)(idA, -1, 0, xA, muH) * (*_beamB)(idB, -1, 0, xB, muH);
         } else {
            Log::error("CalcSCETppV012::dsigma0Resummed") << "inconsistent running/matching: 0-jet LO, 1-jet LO needs NLL!" << Log::end;
         }
      }

      dsigma0resum = dF0 * U0 / TauB0;
   } else {
      // resummed = resummed expanded, so call that function
      dsigma0resum = dSigma0ResumExpanded(phi0, TauB0, scale, muFO);
   }

   return dsigma0resum;
}

//------------------------------------------------------------------------------
// resummed expanded spectrum
double CalcSCETppV012::dSigma0ResumExpanded(const Phi0& phi0, double TauB0, Scale scale, double muFO)
{
   double xA = phi0.xA;
   PdgID idA = phi0.idA;
   double xB = phi0.xB;
   PdgID idB = phi0.idB;
   double Q = phi0.Q;

   // set FO scales, alphas
   double muNS = getmuNS(scale, TauB0, muFO);
   const double as = (*_alphas)(muNS);
   const double lnxFOH = log(Q * Q / (muNS * muNS));
   const double lnxFOB = log(TauB0 * Q / (muNS * muNS));
   const double lnxFOS = log(TauB0 / muNS);
   double etaFO = 0;

   // we only need the FO singular terms
   double dF0sing = 0;

   if (_fixed0jet == FixedOrder::kNLO) {
      if (_fixed1jet == FixedOrder::kLO) {
         // 0-jet at NLO, 1-jet at LO: NLL' or NNLL resummation with 1-loop FO terms
         if (_run0 == RunOrder::kNLL || _run0 == RunOrder::kNNLL) {
            dF0sing += dF0NLO(_beamAsing, _beamBsing, as, lnxFOH, as, lnxFOB, muNS, as, lnxFOB, muNS, as, lnxFOS, etaFO, xA, xB, Q, idA, idB);
         } else {
            Log::error("CalcSCETppV012::dsigma0ResumExpanded") << "inconsistent running/matching: 0-jet NLO, 1-jet LO needs NLL or NNLL!" << Log::end;
         }
      } else { // _fixed1jet == FixedOrder::kNLO
         // 0-jet at NLO, 1-jet at NLO: NNLL' resummation with 2-loop FO terms
         if (_run0 == RunOrder::kNNLL) {
            dF0sing += dF0NLO(_beamAsing, _beamBsing, as, lnxFOH, as, lnxFOB, muNS, as, lnxFOB, muNS, as, lnxFOS, etaFO, xA, xB, Q, idA, idB);
            if (!_removeAlpha2)
               dF0sing += dF0NNLO(_beamAsing, _beamBsing, as, lnxFOH, as, lnxFOB, muNS, as, lnxFOB, muNS, as, lnxFOS, etaFO, xA, xB, Q, idA, idB);
         } else {
            Log::error("CalcSCETppV012::dsigma0ResumExpanded") << "inconsistent running/matching: 0-jet NLO, 1-jet NLO needs NNLL!" << Log::end;
         }
      }
   } else { // _fixed0jet == FixedOrder::kLO ( => _fixed1jet == FixedOrder::kLO)
      // 0-jet at LO, 1-jet at LO: NLL' resummation with 1-loop FO terms
      // note that with additive matching we would actually do tree level FO terms
      // but with multiplicative matching this is not an option since the tree level resummed expanded is 0
      if (_fixed1jet == FixedOrder::kLO && _run0 == RunOrder::kNLL) {
         dF0sing += dF0NLO(_beamAsing, _beamBsing, as, lnxFOH, as, lnxFOB, muNS, as, lnxFOB, muNS, as, lnxFOS, etaFO, xA, xB, Q, idA, idB);
      } else {
         Log::error("CalcSCETppV012::dsigma0ResumExpanded") << "inconsistent running/matching: 0-jet LO, 1-jet LO needs NLL!" << Log::end;
      }
   }

   return dF0sing / TauB0;
}

//------------------------------------------------------------------------------
// spectrum: resummed and resummed expanded
double CalcSCETppV012::deltaSigmaR(const Phi0& phi0, double TauB0, Scale scale, double muFO)
{
   assert(scale == kmurun || scale == kmuup || scale == kmudown);

   // force to return 0. outside resummation region
   double muH, muB, muS;
   if (!getScales(scale, TauB0, phi0.Q, phi0.Y, muFO, muH, muB, muS))
      return 0.;

   // function to smooth the profiles as they turn off to the FO
   double profsmoother = 1;

   Scale dsigmaRscale = scale;
   if (_improvedXSfix) {
      // base function for the profile smoother
      double profsmootherbase = (0.5 - 0.5 * tanh((8. / phi0.Q) * (TauB0 - 10.)));

      if (scale == kmurun)
         dsigmaRscale = kmuhigherrun;
      else if (scale == kmuup)
         dsigmaRscale = kmuhigherup;
      else
         dsigmaRscale = kmuhigherdown;

      profsmoother = profsmootherbase * _profSmootherFac[scale];
   }

   // take the derivative of the cumulant for the given scale.
   // Save the information so it can be used later
   if (_cumDerivTau0Resummed.find(dsigmaRscale) == _cumDerivTau0Resummed.end()) {
      Sigma0ResummedDerivative deriveCumulant(this, &CalcSCETppV012::sigma0Resummed, phi0, dsigmaRscale, muFO);
      double abserr;
      const double result = deriveCumulant(TauB0, 1e-6, abserr);
      if ((abs(result) > 1. && abserr > 1e-3 * abs(result)) || (abs(result) <= 1. && abserr > 1e-3)) {
         Log::warning("CalcSCETppV012::deltasigmaR") << "Large error " << abserr << " in taking the numerical derivative of the cumulant at Tau0 = " << TauB0 << ", result = " << result << Log::end;
      }
      _cumDerivTau0Resummed[dsigmaRscale] = result;
   }

   // calculate the spectrum at the given scale
   // Save the information so it can be used later
   if (_spectrumTau0Resummed.find(dsigmaRscale) == _spectrumTau0Resummed.end())
      _spectrumTau0Resummed[dsigmaRscale] = dSigma0Resummed(phi0, TauB0, dsigmaRscale, muFO);

   // delta sigma_R is the difference
   return profsmoother * (_cumDerivTau0Resummed[dsigmaRscale] - _spectrumTau0Resummed[dsigmaRscale]);
}

//------------------------------------------------------------------------------
double CalcSCETppV012::getmuFO(const Configuration& config, double Tau0) const
{
   const double muBelow = ScaleInvariantMass<IsElectroWeak>()(config);
   const double x = Tau0/muBelow;
   const double x3 = 0.85;
   const double x4 = 1.15;
   if (x < x3)
      return muBelow;

   double muAbove = (*_scaleFO)(config);
   if (muAbove == muBelow || x >= x4)
      return muAbove;

   // at this point we have x3 <= x < x4
   // smoothly interpolate with quadratic equations
   double smooth = 0.0;
   if (x < 1.0)
      smooth = 0.5*pow((x - x3)/(1.0 - x3), 2);
   else
      smooth = 1.0 - 0.5*pow((x - x4)/(1.0 - x4), 2);
   return muBelow * (1.0-smooth) + muAbove * smooth;
}

//------------------------------------------------------------------------------
/// Computes the values for the running scales.
/// Returns \c true if the scales are different (so resummation is nontrivial)
/// and \c false if all the scales are exactly equal (in which case resummation
/// is trivial and should be treated special).
bool CalcSCETppV012::getScales(Scale scale, double TauB0, double Q, double Y, double muFO,
                               double& muH, double& muB, double& muS)
{
   // canonical scales
   if (scale == kcanonical) {
      muH = Q;
      muS = TauB0;
      muB = sqrt(muH * muS);
      return (TauB0 < Q);
   }

   // profile scales
   // rapidity-independent base parameters
   double a = 1;
   double y0 = 2.5 / Q;
   double x1 = 0.2;
   double x2 = 0.5;
   double x3 = 0.8;
   double dx = 0.05;

   if (_profilesY == knormal) {
      // normal moderate rapidity dependence
      x1 = 0.2;
      x3 = 0.8;
      dx = 0.05;
      if (abs(Y) > 2.) {
         const double dY = std::min(abs(Y) - 2., 2.);
         x1 -= 0.05 * dY;
         x3 -= 0.25 * dY;
      }
      x2 = 0.5 * (x1 + x3);
   } else if (_profilesY == kstrong) {
      // strong/aggressive rapidity dependence
      x1 = 0.15;
      x3 = 0.65;
      dx = 0.05;
      if (abs(Y) > 2.) {
         const double dY = std::min(abs(Y) - 2., 2.5);
//       y0 *= 1. - 0.08 * dY;
         x1 -= 0.03 * dY;
         x3 -= 0.22 * dY;
         dx -= 0.01 * dY;
      }
      x2 = 0.5 * (x1 + x3);
   }

   // main parameter
   double x = TauB0 / Q;

   // central scale
   muH = muFO;
   muS = muH * murun(x, a, y0, x1, x2, x3);
   muB = sqrt(muH * muS);

   // various profiles
   // mu up
   if (scale == kmuup) {
      muH *= 2.;
      muS *= 2.;
      muB *= 2.;
   }
   // mu down
   else if (scale == kmudown) {
      muH *= 0.5;
      muS *= 0.5;
      muB *= 0.5;
   }
   // muB up, down
   else if (scale == kmuBup || scale == kmuBdown) {
      double beta = 1./6.;
      // muB up
      if (scale == kmuBup) {
         muB = pow(muH, 0.5 + beta) * pow(muS, 0.5 - beta);
      }
      // muB down
      else if (scale == kmuBdown) {
         muB = pow(muH, 0.5 - beta) * pow(muS, 0.5 + beta);
      }
   }
   // muS up, down
   else if (scale == kmuSup || scale == kmuSdown) {
      double varyfactor = 1.;
      if (x < 0.)
         varyfactor = 2.;
      else if (x < 0.5*x3)
         varyfactor = 2.*(1. - pow(x/x3, 2));
      else if (x < x3)
         varyfactor = 1. + 2.*pow(1. - x/x3, 2);

      if (scale == kmuSup)       // muS up
         muS *= varyfactor;
      else                       // muS down
         muS /= varyfactor;
      muB = sqrt(muH * muS);
   }
   // mux up
   else if (scale == kmuxup) {
      muS = muH * murun(x, a, y0, x1 + dx, x2 + dx, x3 + dx);
      muB = sqrt(muH * muS);
      return (x < x3 + dx);
   }
   // mux down
   else if (scale == kmuxdown) {
      muS = muH * murun(x, a, y0, x1 - dx, x2 - dx, x3 - dx);
      muB = sqrt(muH * muS);
      return (x < x3 - dx);
   }
   // higher order profiles
   else if (scale == kmuhigherrun) {
      double hmurun = -0.004 + (0.5 + 0.002) * (1. + tanh(7. * (x - 0.25)));
      muS = muH * hmurun;
      muB = sqrt(muH * muS);
   } else if (scale == kmuhigherup) {
      muH = 2. * muFO;
      double hmurun = -0.004 + (0.5 + 0.002) * (1. + tanh(7. * (x - 0.25)));
      muS = muH * hmurun;
      muB = sqrt(muH * muS);
   } else if (scale == kmuhigherdown) {
      muH = 0.5 * muFO;
      double hmurun = -0.004 + (0.5 + 0.002) * (1. + tanh(7. * (x - 0.25)));
      muS = muH * hmurun;
      muB = sqrt(muH * muS);
   }
   return (x < x3);
}

// ~~~ checked against Mathematica ~~~ //
//------------------------------------------------------------------------------
/**
 * \brief Profile function
 * \param x The rescaled observable (e.g., Tau), rescaled by the canonical hard scale
 * \param a Speficies the scaling with the observable in the canonical regime for the soft scale
 * \param y0 The cutoff value for the profile scales as x->0
 * \param x1 The cutoff for the canonical scaling region (in x)
 * \param x2 The midpoint of the transition region (in x)
 * \param x3 The cutoff for the resummation (in x)
 * \return The value of murun(x)
 *
 * Defines the murun function, which determines the central soft scale via
 * muS = Q * murun.
 *
 * The nonperturbative region is determined by x0 = 2y0/a. Normally, this should
 * yield x0 < x1. In rare cases when x0 > x1, we switch from the nonperturbative
 * to the transition profile whereever they intersect, which leads to a kink at
 * that point.
 *
 * \note: It is the user's responsibility to ensure that 0 < x1 < x2 < x3 and
 * that a > 0 and y0 >= 0!
 */
double CalcSCETppV012::murun(double x, double a, double y0, double x1, double x2, double x3) const
{
   assert(0. <= x1 && x1 <= x2 && x2 <= x3 && a > 0. && y0 >= 0.);

   if (x >= x3)                        // nothing to do for fixed-order region
      return 1.0;
   if (x <= 0.)
      return y0;

   // nonperturbative cutoff and canonical region for x <= x1
   const double y1 = a*x1;
   if (2.*y0 <= y1) {                  // x0 < x1
      const double y = a*x;
      if (y < 2.*y0)                   // quadratic cutoff for x < x0 = 2y0/a < x1
         return y0 * (1. + pow(y/(2.*y0), 2));
      if (y <= y1)                     // linear canonical region for x0 <= x <= x1
         return y;
   } else if (y0 < 0.99 * y1) {
      if (x <= x1)                     // x^n cutoff for x <= x1
         return (y1 - y0) * pow(x/x1, y1/(y1 - y0)) + y0;
   } else {
      Log::warning("CalcSCETppV012::murun") << "Nonperturbative cutoff y0 = " << y0
         << " close or above transition region y1 = " << y1
         << ", using flat cutoff with kinked profile" << Log::end;
      if (x <= x1)                     // flat cutoff for x <= x1
         return y0;
   }

   // transition region x1 < x <= x3
   const double b = (2. - a * (x1 + x2)) / (x3 - x1);
   double y;
   if (x <= x2)                        // lower part of transition x1 < x <= x2
      y = a * x + (b - a) * pow(x - x1, 2) / (2. * (x2 - x1));
   else                                // upper part of transition x2 < x <= x3
      y = b * (x - pow(x - x2, 2) / (2 * (x3 - x2))) + 0.5 * (a - b) * (x1 + x2);

   if (y > y0)
      return y;
   else
      return y0;
}

//------------------------------------------------------------------------------
/// Computes the nonsingular scale according to \p scale.
double CalcSCETppV012::getmuNS(Scale scale, double Tau0, double muFO) const
{
   double muNS;
   if (_profileNS == khard)
      muNS = muFO;
   else { // _profilesNS == kbeam
      Log::fatal("CalcSCETppV012::getmuNS") << "Nonsingular beam profile is not yet fully supported." << Log::end;
      // TODO: need to check _profilesY and x should be Tau0/Q not Tau0/muFO
      double x = Tau0 / muFO;
      double a = 1;
      double y0 = 2.5 / muFO;
      double x1 = 0.2;
      double x2 = 0.5;
      double x3 = 0.75;
      muNS = muFO * sqrt(murun(x, a, y0, x1, x2, x3));
   }

   switch (scale) {
      case kmuupfix:
      case kmuhigherup:
      case kmuup:
         return 2. * muNS;
      case kmudownfix:
      case kmuhigherdown:
      case kmudown:
         return 0.5 * muNS;
      default:
         return muNS;
   }
}

//------------------------------------------------------------------------------
/// Evolution factor for 0-jet resummation to LL
double CalcSCETppV012::U0LL(double asH, double lnxH, double asBa, double lnxBa, double asBb,
                            double lnxBb, double asS, double lnxS, double as, int nf, double& eta)
{

   const double KH   = -2.*QCD::CF * RGE::KGammaLL(asH, as, nf);
   const double etaH = QCD::CF * RGE::etaGammaLL(asH, as, nf);

   const double KBa   =  4.*QCD::CF * RGE::KGammaLL(asBa, as, nf);
   const double etaBa = -2.*QCD::CF * RGE::etaGammaLL(asBa, as, nf);

   const double KBb   =  4.*QCD::CF * RGE::KGammaLL(asBb, as, nf);
   const double etaBb = -2.*QCD::CF * RGE::etaGammaLL(asBb, as, nf);

   const double KS   = -4.*QCD::CF * RGE::KGammaLL(asS, as, nf);
   const double etaS =  4.*QCD::CF * RGE::etaGammaLL(asS, as, nf);

   // return total eta
   eta = etaBa + etaBb + etaS;

   //lnxH = ln (- mH^2 / muH^2), which is real
   const double U0H = abs(exp(2.*KH + 2.*etaH * lnxH));

   return U0H * exp(KBa + KBb + KS - math_constants::gammaE() * eta
                    + etaBa * lnxBa + etaBb * lnxBb + etaS * lnxS) / gamma(1. + eta);
}

//------------------------------------------------------------------------------
/// Evolution factor for 0-jet resummation to NLL
double CalcSCETppV012::U0NLL(double asH, double lnxH, double asBa, double lnxBa, double asBb,
                             double lnxBb, double asS, double lnxS, double as, int nf, double& eta)
{
   const double gH0 = -6.*QCD::CF;
   const double KH   = -2.*QCD::CF * RGE::KGammaNLL(asH, as, nf)
                       + RGE::KgammaNLL(asH, as, nf, gH0);
   const double etaH = QCD::CF * RGE::etaGammaNLL(asH, as, nf);

   const double gB0 = 6.*QCD::CF;
   const double KBa   =  4.*QCD::CF * RGE::KGammaNLL(asBa, as, nf)
                         + RGE::KgammaNLL(asBa, as, nf, gB0);
   const double etaBa = -2.*QCD::CF * RGE::etaGammaNLL(asBa, as, nf);

   const double KBb   =  4.*QCD::CF * RGE::KGammaNLL(asBb, as, nf)
                         + RGE::KgammaNLL(asBb, as, nf, gB0);
   const double etaBb = -2.*QCD::CF * RGE::etaGammaNLL(asBb, as, nf);

   // gammaS0 = 0
   const double KS   = -4.*QCD::CF * RGE::KGammaNLL(asS, as, nf);
   const double etaS =  4.*QCD::CF * RGE::etaGammaNLL(asS, as, nf);

   // return total eta
   eta = etaBa + etaBb + etaS;

   //lnxH = ln (- mH^2 / muH^2), which is real
   const double U0H = abs(exp(2.*KH + 2.*etaH * lnxH));

   return U0H * exp(KBa + KBb + KS - math_constants::gammaE() * eta
                    + etaBa * lnxBa + etaBb * lnxBb + etaS * lnxS) / gamma(1. + eta);
}

// ~~~ checked against Mathematica ~~~ //
//------------------------------------------------------------------------------
/// Evolution factor for 2-jet resummation to NNLL
double CalcSCETppV012::U0NNLL(double asH, double lnxH, double asBa, double lnxBa, double asBb,
                              double lnxBb, double asS, double lnxS, double as, int nf, double& eta)
{

   const double gH0 = -6 * QCD::CF;
   const double gH1 = QCD::CF * QCD::CF * (-3. + 4 * math_constants::pi2() - 48.*math_constants::zeta3())
                      + QCD::CF * QCD::CA * (-961. / 27. - 11 * math_constants::pi2() / 3. + 52.*math_constants::zeta3())
                      + QCD::CF * QCD::TF * nf * (260. / 27. + 4.*math_constants::pi2() / 3.);

   const double KH = -2.*QCD::CF * RGE::KGammaNNLL(asH, as, nf)
                     + RGE::KgammaNNLL(asH, as, nf, gH0, gH1);
   const double etaH = QCD::CF * RGE::etaGammaNNLL(asH, as, nf);

   const double gB0 = 6 * QCD::CF;
   const double gB1 = QCD::CF * QCD::CF * (3. - 4 * math_constants::pi2() + 48.*math_constants::zeta3())
                      + QCD::CF * QCD::CA * (1769. / 27. + 22.*math_constants::pi2() / 9. - 80.*math_constants::zeta3())
                      + QCD::CF * QCD::TF * nf * (-484. / 27. - 8.*math_constants::pi2() / 9.);

   const double KBa   =  4.*QCD::CF * RGE::KGammaNNLL(asBa, as, nf)
                         + RGE::KgammaNNLL(asBa, as, nf, gB0, gB1);
   const double etaBa = -2.*QCD::CF * RGE::etaGammaNNLL(asBa, as, nf);

   const double KBb   =  4.*QCD::CF * RGE::KGammaNNLL(asBb, as, nf)
                         + RGE::KgammaNNLL(asBb, as, nf, gB0, gB1);
   const double etaBb = -2.*QCD::CF * RGE::etaGammaNNLL(asBb, as, nf);

   const double gS0 = 0.0;
   const double gS1 = - 2.*gH1 - 2.*gB1;

   const double KS   = -4.*QCD::CF * RGE::KGammaNNLL(asS, as, nf)
                       + RGE::KgammaNNLL(asS, as, nf, gS0, gS1);
   const double etaS =  4.*QCD::CF * RGE::etaGammaNNLL(asS, as, nf);

   // return total eta
   eta = etaBa + etaBb + etaS;

   //lnxH = ln (- mH^2 / muH^2), which is real
   const double U0H = abs(exp(2.*KH + 2.*etaH * lnxH));


   return U0H * exp(KBa + KBb + KS - math_constants::gammaE() * eta
                    + etaBa * lnxBa + etaBb * lnxBb + etaS * lnxS) / gamma(1. + eta);
}

////------------------------------------------------------------------------------
///// Evolution factor for 0-jet resummation to LL
//double CalcSCETppV012::U0LLExpand(double lnxH, double lnxBa,
//                                  double lnxBb, double lnxS, double muH, double muBa, double muBb, double muS, double as, int nf)
//{
//   //Hard function
//   const double KHExp   = - 2. * QCD::CF * as * math_constants::_1_4pi() * 0.5 * QCD::Gamma0(nf) * log(muH / muH) * log(muH / muH);
//   const double etaHExp = QCD::CF * as * math_constants::_1_4pi() * QCD::Gamma0(nf) * log(muH / muH);
//
//   //Beam function A
//   const double KBaExp   =  4. * QCD::CF * as * math_constants::_1_4pi() * 0.5 * QCD::Gamma0(nf) * log(muBa / muH) * log(muBa / muH);
//   const double etaBaExp = - 2. * QCD::CF * as * math_constants::_1_4pi() * QCD::Gamma0(nf) * log(muBa / muH);
//
//   const double KBbExp   =  4. * QCD::CF * as * math_constants::_1_4pi() * 0.5 * QCD::Gamma0(nf) * log(muBb / muH) * log(muBb / muH);
//   const double etaBbExp = - 2. * QCD::CF  * as * math_constants::_1_4pi() * QCD::Gamma0(nf) * log(muBb / muH);
//
//   //Soft function
//   const double KSExp   = - (4. * QCD::CF + 2. * QCD::CA) * as * math_constants::_1_4pi() * 0.5 * QCD::Gamma0(nf) * log(muS / muH) * log(muS / muH);
//   const double etaSExp = (4. * QCD::CF + 2. * QCD::CA) * as * math_constants::_1_4pi() * QCD::Gamma0(nf) * log(muS / muH);
//
//   return 2. * KHExp + KBaExp + KBbExp + KSExp + 2. * etaHExp * lnxH + etaBaExp * lnxBa + etaBbExp * lnxBb + etaSExp * lnxS;
//
//
//}// ~~~ checked against Mathematica ~~~ //
////------------------------------------------------------------------------------
///// Evolution factor for 0-jet resummation to LL
//double CalcSCETppV012::U0NLLExpand(double lnxH, double lnxBa,
//                                   double lnxBb, double lnxS, double muH, double muBa, double muBb, double muS, double as, int nf)
//{
//   //Hard function
//   const double gH0 = - 6. * QCD::CF;
//   const double KHExp   = - 2. * QCD::CF * as * math_constants::_1_4pi() * 0.5 * QCD::Gamma0(nf) * log(muH / muH) * log(muH / muH) - as * math_constants::_1_4pi() * gH0 * log(muH / muH);
//   const double etaHExp = QCD::CF * as * math_constants::_1_4pi() * QCD::Gamma0(nf) * log(muH / muH);
//
//   const double gB0A = 6. * QCD::CF;
//   const double KBaExp   =  4. * QCD::CF * as * math_constants::_1_4pi() * 0.5 * QCD::Gamma0(nf) * log(muBa / muH) * log(muBa / muH) - as * math_constants::_1_4pi() * gB0A * log(muBa / muH);
//   const double etaBaExp = - 2. * QCD::CF * as * math_constants::_1_4pi() * QCD::Gamma0(nf) * log(muBa / muH);
//
//   const double gB0B = 6. * QCD::CF;
//   const double KBbExp   =  4. * QCD::CF * as * math_constants::_1_4pi() * 0.5 * QCD::Gamma0(nf) * log(muBb / muH) * log(muBb / muH) - as * math_constants::_1_4pi() * gB0B * log(muBa / muH);
//   const double etaBbExp = - 2. * QCD::CF  * as * math_constants::_1_4pi() * QCD::Gamma0(nf) * log(muBb / muH);
//
//   //Soft function
//   const double KSExp   = - 4. * QCD::CF * as * math_constants::_1_4pi() * 0.5 * QCD::Gamma0(nf) * log(muS / muH) * log(muS / muH);
//   const double etaSExp =  4. * QCD::CF * as * math_constants::_1_4pi() * QCD::Gamma0(nf) * log(muS / muH);
//
//   return 2. * KHExp + KBaExp + KBbExp + KSExp + 2. * etaHExp * lnxH + etaBaExp * lnxBa + etaBbExp * lnxBb + etaSExp * lnxS;
//
//
//}// ~~~ checked against Mathematica ~~~ //
////------------------------------------------------------------------------------
///// Evolution factor for 0-jet resummation to LL
//double CalcSCETppV012::U0NNLLExpand(double lnxH, double lnxBa,
//                                    double lnxBb, double lnxS, double muH, double muBa, double muBb, double muS, double as, int nf)
//{
//   Log::error("CalcSCETppV012::U0NNLLExpand") << "Expanded Sudakov factor at NNLL not available yet" << Log::end;
//   return -1.;
//}

// ~~~ checked against Mathematica ~~~ //
//------------------------------------------------------------------------------
/// NLO matching contributions for cumulant for 0-jet resummation
double CalcSCETppV012::F0LO(BeamFuncTauB* beamfuncA, BeamFuncTauB* beamfuncB, double asH, double lnxH, double muBa, double muBb, double xA, double xB, PdgID idA, PdgID idB)
{
   // hard function at different orders
   double hLO = hard2(asH, lnxH, 0);

   // beam functions at different orders
   double Ba0qm1 = (*beamfuncA)(idA, -1, 0, xA, muBa);
   double baLO = Ba0qm1;

   double Bb0qm1 = (*beamfuncB)(idB, -1, 0, xB, muBb);
   double bbLO = Bb0qm1;

   // soft function at different orders
   std::vector<double> SoftNLOcoeffs, SoftNNLOcoeffs;
   // LO
   double sLO = soft2(-1, 0);
   double F0NLO = hLO * baLO * bbLO * sLO ;

   return F0NLO;
}

// ~~~ checked against Mathematica ~~~ //
//------------------------------------------------------------------------------
/// NLO matching contributions for cumulant for 0-jet resummation
double CalcSCETppV012::F0NLO(BeamFuncTauB* beamfuncA, BeamFuncTauB* beamfuncB, double asH, double lnxH, double asBa, double lnxBa, double muBa, double asBb, double lnxBb, double muBb,
                             double asS, double lnxS, double eta, double xA, double xB, double Q, PdgID idA, PdgID idB)
{
   // hard function at different orders
   double hLO = hard2(asH, lnxH, 0);
   double hNLO = hard2(asH, lnxH, 1);

   // beam functions at different orders
   std::vector<double> BaNLOcoeffs, BbNLOcoeffs;
   // now calculate and add the coefficients
   // NLO, delta function
   double Ba1qm1 = (*beamfuncA)(idA, -1, 1, xA, muBa);
   double Bb1qm1 = (*beamfuncB)(idB, -1, 1, xB, muBb);
   BaNLOcoeffs.push_back(Ba1qm1);
   BbNLOcoeffs.push_back(Bb1qm1);
   // NLO, L0
   double Ba1q0 = (*beamfuncA)(idA, 0, 1, xA, muBa);
   double Bb1q0 = (*beamfuncB)(idB, 0, 1, xB, muBb);
   BaNLOcoeffs.push_back(Ba1q0);
   BbNLOcoeffs.push_back(Bb1q0);
   // NLO, L1
   double Ba1q1 = (*beamfuncA)(idA, 1, 1, xA, muBa);
   double Bb1q1 = (*beamfuncB)(idB, 1, 1, xB, muBb);
   BaNLOcoeffs.push_back(Ba1q1);
   BbNLOcoeffs.push_back(Bb1q1);
   // LO
   double Ba0qm1 = (*beamfuncA)(idA, -1, 0, xA, muBa);
   double Bb0qm1 = (*beamfuncB)(idB, -1, 0, xB, muBb);

   // set up the FO convolutions
   FOConv convFO(eta);

   // now get the beam functions
   double baLO = Ba0qm1;
   double bbLO = Bb0qm1;
   double baNLO = asBa * convFO.funcCumulantNLO(BaNLOcoeffs, -lnxBa);
   double bbNLO = asBb * convFO.funcCumulantNLO(BbNLOcoeffs, -lnxBb);

   // soft function at different orders
   std::vector<double> SoftNLOcoeffs, SoftNNLOcoeffs;
   // LO
   double sLO = soft2(-1, 0);
   // NLO
   SoftNLOcoeffs.push_back(soft2(-1, 1));
   SoftNLOcoeffs.push_back(soft2(0, 1));
   SoftNLOcoeffs.push_back(soft2(1, 1));
   double sNLO = asS * convFO.funcCumulantNLO(SoftNLOcoeffs, -lnxS);

   double F0NLO = hNLO * baLO * bbLO * sLO + hLO * baNLO * bbLO * sLO + hLO * baLO * bbNLO * sLO + hLO * baLO * bbLO * sNLO;

   return F0NLO;
}

// ~~~ checked against Mathematica ~~~ //
//------------------------------------------------------------------------------
/// NNLO matching contributions for cumulant for 0-jet resummation
double CalcSCETppV012::F0NNLO(BeamFuncTauB* beamfuncA, BeamFuncTauB* beamfuncB, double asH, double lnxH, double asBa, double lnxBa, double muBa, double asBb, double lnxBb, double muBb,
                              double asS, double lnxS, double eta, double xA, double xB, double Q, PdgID idA, PdgID idB)
{
   // hard function at different orders
   double hLO = hard2(asH, lnxH, 0);
   double hNLO = hard2(asH, lnxH, 1);
   double hNNLO = hard2(asH, lnxH, 2);

   // beam functions at different orders
   std::vector<double> BaNLOcoeffs, BbNLOcoeffs;
   std::vector<double> BaNNLOcoeffs, BbNNLOcoeffs;
   // now calculate and add the coefficients
   // NNLO, delta function
   double Ba2qm1 = (*beamfuncA)(idA, -1, 2, xA, muBa);
   double Bb2qm1 = (*beamfuncB)(idB, -1, 2, xB, muBb);
   BaNNLOcoeffs.push_back(Ba2qm1);
   BbNNLOcoeffs.push_back(Bb2qm1);
   // NNLO, L0
   double Ba2q0 = (*beamfuncA)(idA, 0, 2, xA, muBa);
   double Bb2q0 = (*beamfuncB)(idB, 0, 2, xB, muBb);
   BaNNLOcoeffs.push_back(Ba2q0);
   BbNNLOcoeffs.push_back(Bb2q0);
   // NNLO, L1
   double Ba2q1 = (*beamfuncA)(idA, 1, 2, xA, muBa);
   double Bb2q1 = (*beamfuncB)(idB, 1, 2, xB, muBb);
   BaNNLOcoeffs.push_back(Ba2q1);
   BbNNLOcoeffs.push_back(Bb2q1);
   // NNLO, L2
   double Ba2q2 = (*beamfuncA)(idA, 2, 2, xA, muBa);
   double Bb2q2 = (*beamfuncB)(idB, 2, 2, xB, muBb);
   BaNNLOcoeffs.push_back(Ba2q2);
   BbNNLOcoeffs.push_back(Bb2q2);
   // NNLO, L3
   double Ba2q3 = (*beamfuncA)(idA, 3, 2, xA, muBa);
   double Bb2q3 = (*beamfuncB)(idB, 3, 2, xB, muBb);
   BaNNLOcoeffs.push_back(Ba2q3);
   BbNNLOcoeffs.push_back(Bb2q3);
   // NLO, delta function
   double Ba1qm1 = (*beamfuncA)(idA, -1, 1, xA, muBa);
   double Bb1qm1 = (*beamfuncB)(idB, -1, 1, xB, muBb);
   BaNLOcoeffs.push_back(Ba1qm1);
   BbNLOcoeffs.push_back(Bb1qm1);
   // NLO, L0
   double Ba1q0 = (*beamfuncA)(idA, 0, 1, xA, muBa);
   double Bb1q0 = (*beamfuncB)(idB, 0, 1, xB, muBb);
   BaNLOcoeffs.push_back(Ba1q0);
   BbNLOcoeffs.push_back(Bb1q0);
   // NLO, L1
   double Ba1q1 = (*beamfuncA)(idA, 1, 1, xA, muBa);
   double Bb1q1 = (*beamfuncB)(idB, 1, 1, xB, muBb);
   BaNLOcoeffs.push_back(Ba1q1);
   BbNLOcoeffs.push_back(Bb1q1);
   // LO
   double Ba0qm1 = (*beamfuncA)(idA, -1, 0, xA, muBa);
   double Bb0qm1 = (*beamfuncB)(idB, -1, 0, xB, muBb);

   // set up the FO convolutions
   FOConv convFO(eta);

   // now get the beam functions
   double baLO = Ba0qm1;
   double bbLO = Bb0qm1;
   double baNLO = asBa * convFO.funcCumulantNLO(BaNLOcoeffs, -lnxBa);
   double bbNLO = asBb * convFO.funcCumulantNLO(BbNLOcoeffs, -lnxBb);
   double baNNLO = asBa * asBa * convFO.funcCumulantNNLO(BaNNLOcoeffs, -lnxBa);
   double bbNNLO = asBb * asBb * convFO.funcCumulantNNLO(BbNNLOcoeffs, -lnxBb);

   // soft function at different orders
   std::vector<double> SoftNLOcoeffs, SoftNNLOcoeffs;
   // LO
   double sLO = soft2(-1, 0);
   // NLO
   SoftNLOcoeffs.push_back(soft2(-1, 1));
   SoftNLOcoeffs.push_back(soft2(0, 1));
   SoftNLOcoeffs.push_back(soft2(1, 1));
   double sNLO = asS * convFO.funcCumulantNLO(SoftNLOcoeffs, -lnxS);
   // NNLO
   SoftNNLOcoeffs.push_back(soft2(-1, 2));
   SoftNNLOcoeffs.push_back(soft2(0, 2));
   SoftNNLOcoeffs.push_back(soft2(1, 2));
   SoftNNLOcoeffs.push_back(soft2(2, 2));
   SoftNNLOcoeffs.push_back(soft2(3, 2));
   double sNNLO = asS * asS * convFO.funcCumulantNNLO(SoftNNLOcoeffs, -lnxS);

   // beam-beam mixing and soft-beam mixing
   double baNLObbNLO = asBa * asBb * convFO.funcCumulantmixNLO(BaNLOcoeffs, BbNLOcoeffs, -lnxBa, -lnxBb);
   double baNLOsNLO = asBa * asS * convFO.funcCumulantmixNLO(BaNLOcoeffs, SoftNLOcoeffs, -lnxBa, -lnxS);
   double bbNLOsNLO = asBb * asS * convFO.funcCumulantmixNLO(BbNLOcoeffs, SoftNLOcoeffs, -lnxBb, -lnxS);

   double F0NNLO = hNNLO * baLO * bbLO * sLO + hLO * baNNLO * bbLO * sLO + hLO * baLO * bbNNLO * sLO + hLO * baLO * bbLO * sNNLO
                   + hNLO * (baNLO * bbLO * sLO + baLO * bbNLO * sLO + baLO * bbLO * sNLO)
                   + hLO * (baNLObbNLO * sLO + baNLOsNLO * bbLO + bbNLOsNLO * baLO);

   return F0NNLO;
}

// ~~~ checked against Mathematica ~~~ //
//------------------------------------------------------------------------------
/// NLO matching contributions for spectrum for 0-jet resummation
double CalcSCETppV012::dF0NLO(BeamFuncTauB* beamfuncA, BeamFuncTauB* beamfuncB, double asH, double lnxH, double asBa, double lnxBa, double muBa, double asBb, double lnxBb, double muBb,
                              double asS, double lnxS, double eta, double xA, double xB, double Q, PdgID idA, PdgID idB)
{
   // hard function at different orders
   double hLO = hard2(asH, lnxH, 0);
   double hNLO = hard2(asH, lnxH, 1);

   // beam functions at different orders
   std::vector<double> BaNLOcoeffs, BbNLOcoeffs;
   std::vector<double> BaNNLOcoeffs, BbNNLOcoeffs;
   // now calculate and add the coefficients
   // NLO, delta function
   double Ba1qm1 = (*beamfuncA)(idA, -1, 1, xA, muBa);
   double Bb1qm1 = (*beamfuncB)(idB, -1, 1, xB, muBb);
   BaNLOcoeffs.push_back(Ba1qm1);
   BbNLOcoeffs.push_back(Bb1qm1);
   // NLO, L0
   double Ba1q0 = (*beamfuncA)(idA, 0, 1, xA, muBa);
   double Bb1q0 = (*beamfuncB)(idB, 0, 1, xB, muBb);
   BaNLOcoeffs.push_back(Ba1q0);
   BbNLOcoeffs.push_back(Bb1q0);
   // NLO, L1
   double Ba1q1 = (*beamfuncA)(idA, 1, 1, xA, muBa);
   double Bb1q1 = (*beamfuncB)(idB, 1, 1, xB, muBb);
   BaNLOcoeffs.push_back(Ba1q1);
   BbNLOcoeffs.push_back(Bb1q1);
   // LO
   double Ba0qm1 = (*beamfuncA)(idA, -1, 0, xA, muBa);
   double Bb0qm1 = (*beamfuncB)(idB, -1, 0, xB, muBb);

   // set up the FO convolutions
   FOConv convFO(eta);

   // now get the beam functions
   double baLO = Ba0qm1;
   double bbLO = Bb0qm1;
   double baNLO = asBa * convFO.funcDistNLO(BaNLOcoeffs, -lnxBa);
   double bbNLO = asBb * convFO.funcDistNLO(BbNLOcoeffs, -lnxBb);

   // soft function at different orders
   std::vector<double> SoftNLOcoeffs, SoftNNLOcoeffs;
   // LO
   double sLO = soft2(-1, 0);
   // NLO
   SoftNLOcoeffs.push_back(soft2(-1, 1));
   SoftNLOcoeffs.push_back(soft2(0, 1));
   SoftNLOcoeffs.push_back(soft2(1, 1));
   double sNLO = asS * convFO.funcDistNLO(SoftNLOcoeffs, -lnxS);

   double dF0NLO = hNLO * baLO * bbLO * sLO * eta + hLO * baNLO * bbLO * sLO + hLO * baLO * bbNLO * sLO + hLO * baLO * bbLO * sNLO;

   return dF0NLO;
}

// ~~~ checked against Mathematica ~~~ //
//------------------------------------------------------------------------------
/// NNLO matching contributions for spectrum for 0-jet resummation
double CalcSCETppV012::dF0NNLO(BeamFuncTauB* beamfuncA, BeamFuncTauB* beamfuncB, double asH, double lnxH, double asBa, double lnxBa, double muBa, double asBb, double lnxBb, double muBb,
                               double asS, double lnxS, double eta, double xA, double xB, double Q, PdgID idA, PdgID idB)
{
   // hard function at different orders
   double hLO = hard2(asH, lnxH, 0);
   double hNLO = hard2(asH, lnxH, 1);
   double hNNLO = hard2(asH, lnxH, 2);

   // beam functions at different orders
   std::vector<double> BaNLOcoeffs, BbNLOcoeffs;
   std::vector<double> BaNNLOcoeffs, BbNNLOcoeffs;
   // now calculate and add the coefficients
   // NNLO, delta function
   double Ba2qm1 = (*beamfuncA)(idA, -1, 2, xA, muBa);
   double Bb2qm1 = (*beamfuncB)(idB, -1, 2, xB, muBb);
   BaNNLOcoeffs.push_back(Ba2qm1);
   BbNNLOcoeffs.push_back(Bb2qm1);
   // NNLO, L0
   double Ba2q0 = (*beamfuncA)(idA, 0, 2, xA, muBa);
   double Bb2q0 = (*beamfuncB)(idB, 0, 2, xB, muBb);
   BaNNLOcoeffs.push_back(Ba2q0);
   BbNNLOcoeffs.push_back(Bb2q0);
   // NNLO, L1
   double Ba2q1 = (*beamfuncA)(idA, 1, 2, xA, muBa);
   double Bb2q1 = (*beamfuncB)(idB, 1, 2, xB, muBb);
   BaNNLOcoeffs.push_back(Ba2q1);
   BbNNLOcoeffs.push_back(Bb2q1);
   // NNLO, L2
   double Ba2q2 = (*beamfuncA)(idA, 2, 2, xA, muBa);
   double Bb2q2 = (*beamfuncB)(idB, 2, 2, xB, muBb);
   BaNNLOcoeffs.push_back(Ba2q2);
   BbNNLOcoeffs.push_back(Bb2q2);
   // NNLO, L3
   double Ba2q3 = (*beamfuncA)(idA, 3, 2, xA, muBa);
   double Bb2q3 = (*beamfuncB)(idB, 3, 2, xB, muBb);
   BaNNLOcoeffs.push_back(Ba2q3);
   BbNNLOcoeffs.push_back(Bb2q3);
   // NLO, delta function
   double Ba1qm1 = (*beamfuncA)(idA, -1, 1, xA, muBa);
   double Bb1qm1 = (*beamfuncB)(idB, -1, 1, xB, muBb);
   BaNLOcoeffs.push_back(Ba1qm1);
   BbNLOcoeffs.push_back(Bb1qm1);
   // NLO, L0
   double Ba1q0 = (*beamfuncA)(idA, 0, 1, xA, muBa);
   double Bb1q0 = (*beamfuncB)(idB, 0, 1, xB, muBb);
   BaNLOcoeffs.push_back(Ba1q0);
   BbNLOcoeffs.push_back(Bb1q0);
   // NLO, L1
   double Ba1q1 = (*beamfuncA)(idA, 1, 1, xA, muBa);
   double Bb1q1 = (*beamfuncB)(idB, 1, 1, xB, muBb);
   BaNLOcoeffs.push_back(Ba1q1);
   BbNLOcoeffs.push_back(Bb1q1);
   // LO
   double Ba0qm1 = (*beamfuncA)(idA, -1, 0, xA, muBa);
   double Bb0qm1 = (*beamfuncB)(idB, -1, 0, xB, muBb);

   // set up the FO convolutions
   FOConv convFO(eta);

   // now get the beam functions
   double baLO = Ba0qm1;
   double bbLO = Bb0qm1;
   double baNLO = asBa * convFO.funcDistNLO(BaNLOcoeffs, -lnxBa);
   double bbNLO = asBb * convFO.funcDistNLO(BbNLOcoeffs, -lnxBb);
   double baNNLO = asBa * asBa * convFO.funcDistNNLO(BaNNLOcoeffs, -lnxBa);
   double bbNNLO = asBb * asBb * convFO.funcDistNNLO(BbNNLOcoeffs, -lnxBb);

   // soft function at different orders
   std::vector<double> SoftNLOcoeffs, SoftNNLOcoeffs;
   // LO
   double sLO = soft2(-1, 0);
   // NLO
   SoftNLOcoeffs.push_back(soft2(-1, 1));
   SoftNLOcoeffs.push_back(soft2(0, 1));
   SoftNLOcoeffs.push_back(soft2(1, 1));
   double sNLO = asS * convFO.funcDistNLO(SoftNLOcoeffs, -lnxS);
   // NNLO
   SoftNNLOcoeffs.push_back(soft2(-1, 2));
   SoftNNLOcoeffs.push_back(soft2(0, 2));
   SoftNNLOcoeffs.push_back(soft2(1, 2));
   SoftNNLOcoeffs.push_back(soft2(2, 2));
   SoftNNLOcoeffs.push_back(soft2(3, 2));
   double sNNLO = asS * asS * convFO.funcDistNNLO(SoftNNLOcoeffs, -lnxS);

   // beam-beam mixing and soft-beam mixing
   double baNLObbNLO = asBa * asBb * convFO.funcDistmixNLO(BaNLOcoeffs, BbNLOcoeffs, -lnxBa, -lnxBb);
   double baNLOsNLO = asBa * asS * convFO.funcDistmixNLO(BaNLOcoeffs, SoftNLOcoeffs, -lnxBa, -lnxS);
   double bbNLOsNLO = asBb * asS * convFO.funcDistmixNLO(BbNLOcoeffs, SoftNLOcoeffs, -lnxBb, -lnxS);

   double dF0NNLO = hNNLO * baLO * bbLO * sLO * eta + hLO * baNNLO * bbLO * sLO + hLO * baLO * bbNNLO * sLO + hLO * baLO * bbLO * sNNLO
                    + hNLO * (baNLO * bbLO * sLO + baLO * bbNLO * sLO + baLO * bbLO * sNLO)
                    + hLO * (baNLObbNLO * sLO + baNLOsNLO * bbLO + bbNLOsNLO * baLO);

   return dF0NNLO;
}

// ~~~ checked against Mathematica ~~~ //
//------------------------------------------------------------------------------
/** Leptonic factor in Born cross section for qqbar > Z
 *  A prefactor to the hard function for a given flavor, Q value
 */
double CalcSCETppV012::lepfac(PdgID id, double Q)
{
   // only defined for quarks
   if (!id.isQuark()) { return 0.0; }

   // charges, couplings, and constants
   const double MZ = _optMap.getOption<double>("ZMass");
   const double GammaZ = _optMap.getOption<double>("ZWidth");
   const double sinSqw = _optMap.getOption<double>("SinSquaredThetaW");
   const double cw = sqrt(1.0 - sinSqw);
   const double sw = sqrt(sinSqw);
   const double sin2w = 2. * sw * cw;

   const double Qlep = -1.;
   const double T3lep = -1. / 2.;
   const double vlep = (T3lep - 2.*Qlep * sw * sw) / sin2w;
   const double alep = (T3lep) / sin2w;

   double Qq = -1. / 3.;
   double T3q = -1. / 2.;
   if (abs(id.code()) == 2 || abs(id.code()) == 4) {
      Qq = 2. / 3.;
      T3q = 1. / 2.;
   }

   double vq = (T3q - 2.*Qq * sw * sw) / sin2w;
   double aq = (T3q) / sin2w;

   double leptonicfactor = Qq * Qq + ((vq * vq + aq * aq) * (vlep * vlep + alep * alep) - 2 * Qq * vq * vlep * (1 - MZ * MZ / (Q * Q))) / ((1 - MZ * MZ / (Q * Q)) * (1 - MZ * MZ / (Q * Q)) + GammaZ * GammaZ * MZ * MZ / (Q * Q * Q * Q));

   return leptonicfactor;
}


// ~~~ checked against Mathematica ~~~ //
//------------------------------------------------------------------------------
/** Hard function at 0 and 1 loop, standard O2 hard function without pi^2 resummation.
 * \param ord The order in perturbation theory, i.e. the term with \f$\alpha^{\text{ord}+1}\f$
 * \return The real valued \f$\alpha^{\text{ord}}\f$ correction.
*/
double CalcSCETppV012::hard2(double asH, double lnXH, int order)
{
   // LO
   if (order == 0) {
      return 1.0;
   }

   // NLO
   if (order == 1) {
      return 2. * (asH / math_constants::_4pi()) * QCD::CF * (-lnXH * lnXH + 3.*lnXH - 8. + (7. / 6.) * math_constants::pi2());
   }

   // NNLO
   if (order == 2) {
      double CH1sq = (64. - (29. / 3.) * math_constants::pi2() + (49. / 36.) * math_constants::pi4())
                     + (-48. - 5.*math_constants::pi2()) * lnXH + (25. + (5. / 3.) * math_constants::pi2()) * lnXH * lnXH
                     - 6.*lnXH * lnXH * lnXH + lnXH * lnXH * lnXH * lnXH;
      double CH2CFCF = ((255. / 4.) - 18.*math_constants::pi2() + (157. / 180.) * math_constants::pi4() - 60.*math_constants::zeta3())
                       + (-45. + 15.*math_constants::pi2() + 48.*math_constants::zeta3()) * lnXH + (25. - (19. / 3.) * math_constants::pi2()) * lnXH * lnXH
                       - 6.*lnXH * lnXH * lnXH + lnXH * lnXH * lnXH * lnXH;
      double CH2CFCA = ((-51157. / 324.) + (1061. / 54.) * math_constants::pi2() - (8. / 45.) * math_constants::pi4() + (626. / 9.) * math_constants::zeta3())
                       + ((2545. / 27.) - (44. / 9.) * math_constants::pi2() - 52.*math_constants::zeta3()) * lnXH + ((-233. / 9.) + (2. / 3.) * math_constants::pi2()) * lnXH * lnXH
                       + (22. / 9.) * lnXH * lnXH * lnXH;
      double CH2CFNF = ((4085. / 81.) - (182. / 27.) * math_constants::pi2() + (8. / 9.) * math_constants::zeta3()) + ((-836. / 27.) + (16. / 9.) * math_constants::pi2()) * lnXH
                       + (76. / 9.) * lnXH * lnXH - (8. / 9.) * lnXH * lnXH * lnXH;
      return (asH / math_constants::_4pi()) * (asH / math_constants::_4pi()) * (QCD::CF * QCD::CF * (CH1sq + CH2CFCF) + QCD::CF * QCD::CA * CH2CFCA + QCD::CF * QCD::TF * _alphas->nf() * CH2CFNF);
   }

   // all other cases
   return 0.0;
}

// ~~~ checked against Mathematica ~~~ //
//------------------------------------------------------------------------------
/** Beam Thrust soft function, given in 1012.4480.
 * It is given as coefficients S(-1,0) delta(Tau) + aS * ( S(-1,1) delta(Tau) + S(0,1) 1/muS L0(Tau/muS) + S(1,1) 1/muS L1(Tau/muS))
 * \param dist Specifies the order of the distribution that the coefficient is to be returned for.
 * \param ord The order in perturbation theory, i.e. the term with \f$\alpha^{\text{ord}}\f$
 * \return The coefficient of \f$\alpha^{\text{ord}}\f$ multiplied by the distribution of order dist.
 */
double CalcSCETppV012::soft2(int dist, int order)
{
   // LO
   if (order == 0) {
      if (dist == -1)
         return 1.0;
      else
         return 0.0;
   }
   // NLO
   if (order == 1) {
      if (dist == -1)
         return math_constants::_1_4pi() * QCD::CF * math_constants::pi2() / 3.;
      else if (dist == 1)
         return math_constants::_1_4pi() * QCD::CF * (-16.);
      else
         return 0.0;
   }
   // NNLO
   if (order == 2) {
      if (dist == -1)
         return math_constants::_1_4pi() * math_constants::_1_4pi() *
                (QCD::CF * QCD::CF * (-3. / 10.) * math_constants::pi4()
                 + QCD::CF * QCD::CA * ((-2140. / 81.) + (-335. / 54.) * math_constants::pi2() + (638. / 9.) * math_constants::zeta3() + (22. / 45.) * math_constants::pi4())
                 + QCD::CF * QCD::TF * _alphas->nf() * ((80. / 81.) + (74. / 27.) * math_constants::pi2() + (-232. / 9.) * math_constants::zeta3()));
      else if (dist == 0)
         return math_constants::_1_4pi() * math_constants::_1_4pi() *
                (QCD::CF * QCD::CF * (256. * math_constants::zeta3())
                 + QCD::CF * QCD::CA * ((1616. / 27.) + (-44. / 9.) * math_constants::pi2() + (-56.) * math_constants::zeta3())
                 + QCD::CF * QCD::TF * _alphas->nf() * ((-448. / 27.) + (16. / 9.) * math_constants::pi2()));
      else if (dist == 1)
         return math_constants::_1_4pi() * math_constants::_1_4pi() *
                (QCD::CF * QCD::CF * (-48. * math_constants::pi2())
                 + QCD::CF * QCD::CA * ((-1072. / 9.) + (16. / 3.) * math_constants::pi2())
                 + QCD::CF * QCD::TF * _alphas->nf() * (320. / 9.));
      else if (dist == 2)
         return math_constants::_1_4pi() * math_constants::_1_4pi() *
                (QCD::CF * QCD::CA * (176. / 3.) + QCD::CF * QCD::TF * _alphas->nf() * (-64. / 3.));
      else if (dist == 3)
         return math_constants::_1_4pi() * math_constants::_1_4pi() * 128. * QCD::CF * QCD::CF;
      else
         return 0.0;
   }

   // all other cases
   return 0.0;
}

//------------------------------------------------------------------------------
/** Function to calculate the X's of the two beam functions (for resummation in SCET)
 * \param config the Configuration for the event
 * \param xbeamA the value of x for beam A
 * \param xbeamB the value of x for beam B
 * \return nothing, it saves the x's into the variables given by the argument
 */
void CalcSCETppV012::setBeamFuncXs(const Configuration& config, double& xbeamA, double& xbeamB)
{
   FourVector leptons;
   for(Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it)
      if (it->id().isLepton()) leptons += it->p();

   double Qlep = leptons.magnitude();
   double Ylep = leptons.rapidity();

   xbeamA = (Qlep / _cmEnergy) * exp(Ylep);
   xbeamB = (Qlep / _cmEnergy) * exp(-Ylep);
   //xbeamA = sqrt(leptons.square() - leptons.perp2())/_cmEnergy * exp(Ylep);
   //xbeamB = sqrt(leptons.square() - leptons.perp2()/_cmEnergy  * exp(-Ylep);
}

//------------------------------------------------------------------------------
/** Function to calculate the PdgID's of the partons inputting into the beam functions
 * \param config the Configuration for the event
 * \return nothing, it saves the beam function coefficients
 */
void CalcSCETppV012::setBeamFuncId(const Configuration& config, std::vector<PdgID>& idBeamFuncA, std::vector<PdgID>& idBeamFuncB)
{
   PdgID idpdfA = config.initialA()->id();
   PdgID idpdfB = config.initialB()->id();

   // get a list of the final QCD parton ids
   std::vector<PdgID> finalQCD;
   for(Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it) {
      if (IsQCD()(*it)) { finalQCD.push_back(it->id()); }
   }

   // now set the beam ids by looking at the intial/final state
   // for V+0, pdfs give the ids
   if (finalQCD.size() == 0) {
      idBeamFuncA.push_back(idpdfA);
      idBeamFuncB.push_back(idpdfB);
   }
   // for V+1, there are 2 cases:
   // 1) both pdfs are quarks, which set the ids
   // 2) one pdf is a quark which sets one beam id,
   //     and the other outgoing particle is the anti-particle of the other beam
   // note: here we use V=Z for the qg channels
   else if (finalQCD.size() == 1) {
      if (idpdfA.isQuark() && idpdfB.isQuark()) {
         idBeamFuncA.push_back(idpdfA);
         idBeamFuncB.push_back(idpdfB);
      } else if (idpdfA.isQuark() && !idpdfB.isQuark()) {
         idBeamFuncA.push_back(idpdfA);
         idBeamFuncB.push_back(idpdfA.antiPartner());
      } else if (!idpdfA.isQuark() && idpdfB.isQuark()) {
         idBeamFuncA.push_back(idpdfB.antiPartner());
         idBeamFuncB.push_back(idpdfB);
      } else {
         Log::error("CalcSCETppV012::setBeamFuncId") << "bad Z+1-jet flavor combination \n" << config << Log::end;
      }
   }
   // for V+2, there are 4 cases, and we only deal with V=Z here:
   // 1) both pdfs are quarks of the same flavor (and the two final state partons are gluons), which set the ids
   // 2) both pdfs are quarks of different flavors.  In this case one quark has undergone the ISR chain
   //        q -> q g -> q q' qbar', where the Z is formed by q' qbar' (with the q' or qbar' coming from a pdf)
   //      since we can't tell which ISR chain this is (pdfA or B), or which beam is associated with what parton,
   //      there are 4 possibilities:
   //        a) A = q, B = qbar
   //        b) A = qbar, A = q
   //        c) A = q', B = qbar'
   //        d) A = qbar', B = q'
   //      note that if the flavors of q, q' are the same, then there are only two choices (c&d are the same as a&b)
   // 3) one pdf is a quark which sets one beam id and the other is a gluon,
   //     and the other outgoing quark is the anti-particle of the other beam
   // 4) neither pdf is a quark, in which case the final state partons are a q-qbar pair, which give two possiblities:
   //        a) A = q, B = qbar
   //        b) A = qbar, B = q
   //
   // note for V=W, we have to deal with all possible flavor combinations on internal quark lines
   else if (finalQCD.size() == 2) {
      // case 1
      if (idpdfA.isQuark() && idpdfB.isQuark() && finalQCD[0].isGluon() && finalQCD[1].isGluon()) {
         idBeamFuncA.push_back(idpdfA);
         idBeamFuncB.push_back(idpdfB);
      }
      // case 2
      else if (idpdfA.isQuark() && idpdfB.isQuark() && finalQCD[0].isQuark() && finalQCD[1].isQuark()) {
         if (abs(idpdfA.code()) == abs(idpdfB.code())) {
            idBeamFuncA.push_back(idpdfA);
            idBeamFuncB.push_back(idpdfA.antiPartner());
            idBeamFuncA.push_back(idpdfA.antiPartner());
            idBeamFuncB.push_back(idpdfA);
         } else {
            idBeamFuncA.push_back(idpdfA);
            idBeamFuncB.push_back(idpdfA.antiPartner());
            idBeamFuncA.push_back(idpdfA.antiPartner());
            idBeamFuncB.push_back(idpdfA);
            idBeamFuncA.push_back(idpdfB);
            idBeamFuncB.push_back(idpdfB.antiPartner());
            idBeamFuncA.push_back(idpdfB.antiPartner());
            idBeamFuncB.push_back(idpdfB);
         }
      }
      // case 3
      else if (idpdfA.isQuark() && idpdfB.isGluon()) {
         idBeamFuncA.push_back(idpdfA);
         idBeamFuncB.push_back(idpdfA.antiPartner());
      } else if (idpdfB.isQuark() && idpdfA.isGluon()) {
         idBeamFuncA.push_back(idpdfB.antiPartner());
         idBeamFuncB.push_back(idpdfB);
      }
      // case 4
      else if (idpdfA.isGluon() && idpdfB.isGluon()) {
         if (finalQCD[0].isQuark() && finalQCD[1].isQuark()) {
            idBeamFuncA.push_back(finalQCD[0]);
            idBeamFuncB.push_back(finalQCD[0].antiPartner());
            idBeamFuncA.push_back(finalQCD[0].antiPartner());
            idBeamFuncB.push_back(finalQCD[0]);
         }
      }
      // cross check - we should never get here
      else {
         Log::error("CalcSCETppV012::setBeamFuncId") << "bad Z+2-jet flavor combination\n" << config << Log::end;
      }
   }
   return;
}

//------------------------------------------------------------------------------
//Return the fully differential partonic cross section
double CalcSCETppV012::fullyDifferential0JetNoPDF(const Configuration& Phi0)
{
   FourVector negLepton, posLepton;
   for(Configuration::iterator it = Phi0.beginFinal(); it != Phi0.endFinal(); ++it) {
      if (it->id().isParticle())
         negLepton = it->p();
      else
         posLepton = it->p();
   }
   double Q = (negLepton + posLepton).magnitude();

   return (*_alphaem)(Q) * (*_alphaem)(Q) * FluxFactor()(Phi0) * (*_treeME)(Phi0);
}

//------------------------------------------------------------------------------
double CalcSCETppV012::extractFixedOrderScale(const Sigma& sigma, Scale scale)
{
   switch(resummationScale[scale]) {
      case kmuup:
         return (fixedOrderScale[_inputScale] == EngineNLOBase::kNone) ? sigma.central() : sigma.variations()[0];
      case kmudown:
         return (fixedOrderScale[_inputScale] == EngineNLOBase::kNone) ? sigma.central() : sigma.variations()[1];
      default:
         return sigma.central();
   }
}

////------------------------------------------------------------------------------
//double CalcSCETppV012::Tau0Sudakov(const PdgID idA, const PdgID idB, const double Q, const double Tau0, Scale scale)
//{
//
//   if (_run0 == RunOrder::kNone)
//      return 1.;
//
//   double muH;
//   double muB, muS, muFO;
//
//   bool resumregion = true;
//   resumregion = getScales(scale, Tau0, Q, muH, muB, muS, muFO);
//
//   double U0 = 0.;
//   if (resumregion) {
//      // nontrivial scale-separation and so nontrivial resummation
//      // NOTE: Arbitrary scale in resummation xi = TauB0 everywhere
//      // We use the fixed-order scale as the scale we're running to.
//      const double as = (*_alphas)(muFO);
//      const int nf = _alphas->nf();
//
//      const double asH  = (*_alphas)(muH);
//      const double lnxH = log((Q * Q) / (muH * muH));
//
//      const double asS  = (*_alphas)(muS);
//      const double lnxS = log(Tau0 / (muS));
//
//      const double asBa  = (*_alphas)(muB);
//      const double lnxBa = log(Tau0 * Q / (muB * muB));
//
//      const double asBb  = (*_alphas)(muB);
//      const double lnxBb = log(Tau0 * Q / (muB * muB));
//
//      double eta;
//
//      switch (_run0) {
//         case RunOrder::kLL: U0 = U0LL(asH, lnxH, asBa, lnxBa, asBb, lnxBb, asS, lnxS, as, nf, eta); break;
//         case RunOrder::kNLL: U0 = U0NLL(asH, lnxH, asBa, lnxBa, asBb, lnxBb, asS, lnxS, as, nf, eta); break;
//         case RunOrder::kNNLL: U0 = U0NNLL(asH, lnxH, asBa, lnxBa, asBb, lnxBb, asS, lnxS, as, nf, eta); break;
//         default: {
//            Log::error("CalcSCETppV012::dsigma0Resummed") << "Unknown run order" << _run0 << Log::end;
//            U0 = std::numeric_limits<double>::quiet_NaN();
//         }
//      }
//   }
//   return U0;
//}
//
////------------------------------------------------------------------------------
//double CalcSCETppV012::Tau0SudakovExpanded(const PdgID idA, const PdgID idB, const double Q, const double Tau0, Scale scale)
//{
//
//   if (_run0 == RunOrder::kNone)
//      return 0.;
//
//   double muH;
//   double muB, muS, muFO;
//
//   bool resumregion = true;
//   resumregion = getScales(scale, Tau0, Q, muH, muB, muS, muFO);
//
//   double U0Exp = 0.;
//   if (resumregion) {
//      // nontrivial scale-separation and so nontrivial resummation
//      // NOTE: Arbitrary scale in resummation xi = TauB0 everywhere
//      // We use the fixed-order scale as the scale we're running to.
//      const double as = (*_alphas)(muFO);
//      const int nf = _alphas->nf();
//
//      const double lnxH = log((Q * Q) / (muH * muH));
//
//      const double lnxS = log(Tau0 / (muS));
//
//      const double lnxBa = log(Tau0 * Q / (muB * muB));
//
//      const double lnxBb = log(Tau0 * Q / (muB * muB));
//
//      switch (_run0) {
//         case RunOrder::kLL: U0Exp = U0LLExpand(lnxH, lnxBa, lnxBb, lnxS, muH, muB, muB, muS, as, nf); break;
//         case RunOrder::kNLL: U0Exp = U0NLLExpand(lnxH, lnxBa, lnxBb, lnxS, muH, muB, muB, muS, as, nf); break;
//         case RunOrder::kNNLL: U0Exp = U0NNLLExpand(lnxH, lnxBa, lnxBb, lnxS, muH, muB, muB, muS, as, nf); break;
//         default: {
//            Log::error("CalcSCETppV012::dsigma0Resummed") << "Unknown run order" << _run0 << Log::end;
//            U0Exp = std::numeric_limits<double>::quiet_NaN();
//         }
//      }
//   }
//   return U0Exp;
//}
//
////------------------------------------------------------------------------------
//double CalcSCETppV012::Tau0SudakovDerivative(const PdgID idA, const PdgID idB, const double Q, const double Tau0, Scale scale)
//{
//
//   if (_run0 == RunOrder::kNone)
//      return 0.;
//
//   double eps = 1e-6;
//   return (Tau0Sudakov(idA, idB, Q, Tau0, scale) - Tau0Sudakov(idA, idB, Q, Tau0 * (1. - eps), scale)) / Tau0 / eps;
//}
//
////------------------------------------------------------------------------------
//double CalcSCETppV012::Tau0SudakovExpandedDerivative(const PdgID idA, const PdgID idB, const double Q, const double Tau0, Scale scale)
//{
//
//   if (_run0 == RunOrder::kNone)
//      return 0.;
//
//   double eps = 1e-6;
//   return (Tau0SudakovExpanded(idA, idB, Q, Tau0, scale) - Tau0SudakovExpanded(idA, idB, Q, Tau0 * (1. - eps), scale)) / Tau0 / eps;
//}

//------------------------------------------------------------------------------
double CalcSCETppV012::tau1Sudakov(const Configuration& Phi1, double Tau1cut, double Tau1max) const
{
   if (_run1 == RunOrder::kNone)
      return 1.;

   Tau1max = std::min(Tau1max, _tau1Max);
   if (Tau1cut >= Tau1max)
      return 1.;

   // Get the scales. We use mu = Tau1Max as the arbitrary scale we're running to.
   const double mu = Tau1max;
   const double muH = Tau1max;
   const double muJ2 = Tau1cut * Tau1max;
   const double muS = Tau1cut;

   const double as = (*_alphas)(mu);
   const int nf = _alphas->nf();

   const double asH = (*_alphas)(muH);
   const double asJ = (*_alphas)(sqrt(muJ2));
   const double asS = (*_alphas)(muS);

   double KH, etaH, KJ, etaJ, KS, etaS;
   double lnU = 0.;
   if (_run1 == RunOrder::kLL) {
      KH = RGE::KGammaLL(asH, as, nf);
      etaH = RGE::etaGammaLL(asH, as, nf);

      KJ = RGE::KGammaLL(asJ, as, nf);
      etaJ = RGE::etaGammaLL(asJ, as, nf);

      KS = RGE::KGammaLL(asS, as, nf);
      etaS = RGE::etaGammaLL(asS, as, nf);
   } else {
      assert(_run1 == RunOrder::kNLL);
      KH = RGE::KGammaNLL(asH, as, nf);
      etaH = RGE::etaGammaNLL(asH, as, nf);

      KJ = RGE::KGammaNLL(asJ, as, nf);
      etaJ = RGE::etaGammaNLL(asJ, as, nf);

      KS = RGE::KGammaNLL(asS, as, nf);
      etaS = RGE::etaGammaNLL(asS, as, nf);

      const double jetNonCusp = 12.*QCD::CF + 2.*QCD::beta0(nf);
      lnU += RGE::KgammaNLL(asH, as, nf, -jetNonCusp);
      lnU += RGE::KgammaNLL(asJ, as, nf, jetNonCusp);
   }

   const double casimir = 2.*QCD::CF + QCD::CA;

   double sq, sg, wq2, wg2;
   getTau1Measures(Phi1, sq, sg, wq2, wg2);

   const double xH = 1./(muH*muH);
   lnU += -2.*casimir * KH;
   if (_includeU1Kinematics)
      lnU += etaH * (2.*QCD::CF * log(xH*sq) + QCD::CA * log(xH*sg));

   const double xJ = pow(Tau1cut/muJ2, 2);
   lnU += +4.*casimir * KJ;
   if (_includeU1Kinematics)
      lnU += etaJ * (-2.*QCD::CF * log(xJ*wq2) + -QCD::CA * log(xJ*wg2));

   const double xS = pow(Tau1cut/muS, 2);
   lnU += -2.*casimir * KS;
   if (_includeU1Kinematics)
      lnU += etaS * (2.*QCD::CF * log(xS*wq2/sq) + QCD::CA * log(xS*wg2/sg));

   const double eta = 2.*casimir * (etaS - etaJ);
   lnU += -math_constants::gammaE() * eta;
   return exp(lnU) / gamma(1. + eta);
}

//------------------------------------------------------------------------------
double CalcSCETppV012::tau1SudakovDerivative(const Configuration& Phi1, double Tau1, double Tau1max) const
{
   if (_run1 == RunOrder::kNone)
      return 0.;

   Tau1max = std::min(Tau1max, _tau1Max);
   if (Tau1 >= Tau1max)
      return 0.;

   double abserr;
   Tau1SudakovDerivative derivative(this, Phi1, Tau1max);
   const double result = derivative(Tau1, 1e-6, abserr);
   if ((abs(result) > 1. && abserr > 1e-3 * abs(result)) || (abs(result) <= 1. && abserr > 1e-3)) {
      Log::warning("CalcSCETppV012::tau1SudakovDerivative") << "Large error " << abserr << " in taking the numerical derivative of Tau1 Sudakov at Tau1 = " << Tau1 << ", result = " << result << Log::end;
   }
   return result;
}

//------------------------------------------------------------------------------
double CalcSCETppV012::tau1SudakovExpanded(const Configuration& Phi1, double Tau1cut, double Tau1max, double muNS) const
{
   if (_run1 == RunOrder::kNone)
      return 0.;

   Tau1max = std::min(Tau1max, _tau1Max);
   if (Tau1cut >= Tau1max)
      return 0.;

   double sq, sg, wq2, wg2;
   getTau1Measures(Phi1, sq, sg, wq2, wg2);

   // the alphas at muNS in terms of which we expand
   const double asFO = (*_alphas)(muNS);
   const int nf = _alphas->nf();

   // the resummation scales and FO logarithms
   const double muH = Tau1max;
   const double muJ2 = Tau1cut * Tau1max;
   const double muS = Tau1cut;

   const double lnH = log(muNS / muH);
   const double lnJ = 0.5*log(muNS*muNS / muJ2);
   const double lnS = log(muNS / muS);

   const double casimir = 2.*QCD::CF + QCD::CA;

   double result = 0.;

   const double xH = 1./(muH*muH);
   result += -2.*casimir * 2. * lnH * lnH;
   if (_includeU1Kinematics)
      result += 4.*lnH * (2.*QCD::CF * log(xH*sq) + QCD::CA * log(xH*sg));

   const double xJ = pow(Tau1cut/muJ2, 2);
   result += +4.*casimir * 2. * lnJ * lnJ;
   if (_includeU1Kinematics)
      result += 4.*lnJ * (-2.*QCD::CF * log(xJ*wq2) + -QCD::CA * log(xJ*wg2));

   const double xS = pow(Tau1cut/muS, 2);
   result += -2.*casimir * 2. * lnS * lnS;
   if (_includeU1Kinematics)
      result += 4.*lnS * (2.*QCD::CF * log(xS*wq2/sq) + QCD::CA * log(xS*wg2/sg));

   if (_run1 >= RunOrder::kNLL) {
      // At 1 loop, softNonCusp = 0 and hardNonCusp = - jetNonCusp
      const double jetNonCusp = 12.*QCD::CF + 2.*QCD::beta0(nf);
      result += jetNonCusp * (lnJ - lnH);
   }

   return asFO * math_constants::_1_4pi() * result;
}

//------------------------------------------------------------------------------
double CalcSCETppV012::tau1SudakovExpandedDerivative(const Configuration& Phi1, double Tau1, double Tau1max, double muNS) const
{
   if (_run1 == RunOrder::kNone)
      return 0.;

   Tau1max = std::min(Tau1max, _tau1Max);
   if (Tau1 >= Tau1max)
      return 0.;

   double sq, sg, wq2, wg2;
   getTau1Measures(Phi1, sq, sg, wq2, wg2);

   // The alphas at muNS in terms of which we expand
   const double asFO = (*_alphas)(muNS);
   const int nf = _alphas->nf();

   // the resummation scales and FO logarithms
   const double muJ2 = Tau1 * Tau1max;
   const double muS = Tau1;

   const double lnJ = 0.5*log(muNS*muNS / muJ2);
   const double lnS = log(muNS / muS);

   const double casimir = 2.*QCD::CF + QCD::CA;

   double result = 0.;

   // here we rely on canonical scaling such that xJ = 1/Tau1max^2 is Tau1 independent
   const double xJ = 1./(Tau1max*Tau1max);
   result += -8.*casimir * lnJ;
   if (_includeU1Kinematics)
      result += -2. * (-2.*QCD::CF * log(xJ*wq2) + -QCD::CA * log(xJ*wg2));

   // here we rely on canonical scaling such that xS = 1 is Tau1 independent
   const double xS = 1.;
   result += 8.*casimir * lnS;
   if (_includeU1Kinematics)
      result += -4. * (2.*QCD::CF * log(xS*wq2/sq) + QCD::CA * log(xS*wg2/sg));

   if (_run1 >= RunOrder::kNLL) {
      // At 1 loop, softNonCusp = 0 and hardNonCusp = - jetNonCusp
      const double jetNonCusp = 12.*QCD::CF + 2.*QCD::beta0(nf);
      result += -0.5 * jetNonCusp;
   }

   return asFO * math_constants::_1_4pi() * result / Tau1;
}

//------------------------------------------------------------------------------
/**
 * \return The magnitude Q of the total leptonic momentum
 */
double CalcSCETppV012::getTau1Measures(const Configuration& Phi1, double& sq, double &sg, double& wq2, double& wg2) const
{
   FourVector leptons;
   Configuration::iterator final1;
   for (Configuration::iterator it = Phi1.beginFinal(); it != Phi1.endFinal(); ++it) {
      if (!IsQCD()(*it))
         leptons += it->p();
      else
         final1 = it;
   }
   double wa = Phi1.initialA()->p().minus();
   double wb = Phi1.initialB()->p().plus();

   // the sij
   const double sab = wa * wb;
   const double sa1 = wa * final1->p().plus();
   const double sb1 = wb * final1->p().minus();

   // the Tau1 measures in the leptonic cms frame
   const double lambda = sqrt(leptons.minus()/leptons.plus());
   wa /= lambda;
   wb *= lambda;
   const double w1 = final1->p().plus() * lambda + final1->p().minus() / lambda;

   // pick the quark and gluon measures according to partonic channel
   if (final1->id().isGluon()) {
      assert(Phi1.initialA()->id().isQuark());
      assert(Phi1.initialB()->id().isQuark());
      sq = sab;
      sg = sa1 * sb1 / sab;
      wq2 = wa * wb;
      wg2 = w1 * w1;
   } else if (Phi1.initialA()->id().isGluon()) {
      assert(final1->id().isQuark());
      assert(Phi1.initialB()->id().isQuark());
      sq = sb1;
      sg = sab * sa1 / sb1;
      wq2 = wb * w1;
      wg2 = wa * wa;
   } else {
      assert(final1->id().isQuark());
      assert(Phi1.initialA()->id().isQuark());
      assert(Phi1.initialB()->id().isGluon());
      sq = sa1;
      sg = sab * sb1 / sa1;
      wq2 = wa * w1;
      wg2 = wb * wb;
   }
   return leptons.magnitude();
}

} // namespace Geneva
