//------------------------------------------------------------------------------
/// \file CalculationFactory.cpp
//
// Author(s):
//    Chris Vermilion, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class CalculationFactory
//------------------------------------------------------------------------------

#include "Geneva/Calculations/CalculationFactory.hpp"
#include "Geneva/Calculations/CalcNLOppV.hpp"
#include "Geneva/Calculations/CalcNLOppVj.hpp"
#include "Geneva/Calculations/CalcSCETppV012.hpp"

#include "Geneva/Components/ComponentFactory.hpp"
#include "Geneva/Components/FKSMapping.hpp"
#include "Geneva/Components/UnitCubeRadGenerator.hpp"
#include "Geneva/Components/SignatureManager.hpp"
#include "Geneva/Components/IScale.hpp"

#include "Geneva/Core/Physics/Order.hpp"
#include "Geneva/Core/Physics/SignatureSet.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"

using std::string;

namespace Geneva
{
//------------------------------------------------------------------------------
ICalculation* CalculationFactory::getCalculation(const OptionMap& optionMap)
{
   if (optionMap.getOption<bool>("verbose"))
      Log::setVerbosity(Log::kVerboseInfo);

   ComponentFactory& compFactory = ComponentFactory::instance();

   string calcName = optionMap.getOption<string>("calc");
   if (calcName == "") {
      Log::error("CalculationFactory::getCalculation") << "No Calculation specified." << Log::end;
   }

   if (calcName == "CalcSCETppV012") {
      double Ecm = optionMap.getOption<double>("Ecm");
      double Tau0cut = optionMap.getOption<double>("NJettiness::Tau0cut");
      double Tau1cut = optionMap.getOption<double>("NJettiness::Tau1cut");
      int    nLight = optionMap.getOption<int>("NLightFlavors");

      RunOrder run0;
      string opt = optionMap.getOption<string>("CalcSCETppV012::run");
      if (opt == "LL")         run0 = RunOrder::kLL;
      else if (opt == "NLL")   run0 = RunOrder::kNLL;
      else if (opt == "NNLL")  run0 = RunOrder::kNNLL;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unknown CalcSCETppV012::run0 running order " << opt << Log::end;
      }
      RunOrder run1;
      opt = optionMap.getOption<string>("CalcSCETppV012::run1");
      if (opt == "none")       run1 = RunOrder::kNone;
      else if (opt == "LL")    run1 = RunOrder::kLL;
      else if (opt == "NLL")   run1 = RunOrder::kNLL;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unknown CalcSCETppV012::run1 running order " << opt << Log::end;
      }
      FixedOrder fixed0jet;
      opt = optionMap.getOption<string>("CalcSCETppV012::fixed0jet");
      if (opt == "LO")       fixed0jet = FixedOrder::kLO;
      else if (opt == "NLO") fixed0jet = FixedOrder::kNLO;
      else if (opt == "NNLO") fixed0jet = FixedOrder::kNNLO;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unknown CalcSCETppV012::fixed0jet fixed order " << opt << Log::end;
      }
      FixedOrder fixed1jet;
      opt = optionMap.getOption<string>("CalcSCETppV012::fixed1jet");
      if (opt == "LO")       fixed1jet = FixedOrder::kLO;
      else if (opt == "NLO") fixed1jet = FixedOrder::kNLO;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unknown CalcSCETppV012::fixed1jet fixed order " << opt << Log::end;
      }
      CalcSCETppV012::Scale scale = CalcSCETppV012::kall;
      opt = optionMap.getOption<string>("CalcSCETppV012::scale");
      if (opt == "all")             scale = CalcSCETppV012::kall;
      else if (opt == "fastset")    scale = CalcSCETppV012::kfastset;
      else if (opt == "hardset")    scale = CalcSCETppV012::khardset;
      else if (opt == "fixset")     scale = CalcSCETppV012::kfixset;
      else if (opt == "setupset")   scale = CalcSCETppV012::ksetupset;
      else if (opt == "murun")      scale = CalcSCETppV012::kmurun;
      else if (opt == "muup")       scale = CalcSCETppV012::kmuup;
      else if (opt == "mudown")     scale = CalcSCETppV012::kmudown;
      else if (opt == "muBup")      scale = CalcSCETppV012::kmuBup;
      else if (opt == "muBdown")    scale = CalcSCETppV012::kmuBdown;
      else if (opt == "muSup")      scale = CalcSCETppV012::kmuSup;
      else if (opt == "muSdown")    scale = CalcSCETppV012::kmuSdown;
      else if (opt == "muxup")      scale = CalcSCETppV012::kmuxup;
      else if (opt == "muxdown")    scale = CalcSCETppV012::kmuxdown;
      else if (opt == "muupfix")      scale = CalcSCETppV012::kmuupfix;
      else if (opt == "mudownfix")    scale = CalcSCETppV012::kmudownfix;
      else if (opt == "canonical")  scale = CalcSCETppV012::kcanonical;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unknown CalcSCETppV012::scale choice " << opt << Log::end;
      }

      opt = optionMap.getOption<string>("CalcSCETppV012::Yprofiles");
      CalcSCETppV012::YProfiles profiles = CalcSCETppV012::knone;
      if (opt == "none")        profiles = CalcSCETppV012::knone;
      else if (opt == "normal") profiles = CalcSCETppV012::knormal;
      else if (opt == "strong") profiles = CalcSCETppV012::kstrong;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unsupported CalcSCETppV012::Yprofiles " << opt << Log::end;
      }

      opt = optionMap.getOption<string>("CalcSCETppV012::phi1");
      CalcSCETppV012::Phi1 phi1 = CalcSCETppV012::kmult;
      if (opt == "mult")      phi1 = CalcSCETppV012::kmult;
      else if (opt == "add")      phi1 = CalcSCETppV012::kadd;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unsupported CalcSCETppV012::phi1 " << opt << Log::end;
      }
      ITreeME*         tree       = compFactory.getTreeME(optionMap);
      ITreeMEComplete* treeMEComp = compFactory.getTreeMEComplete(optionMap);
      IOneLoopME*      oneLoopME  = compFactory.getOneLoopME(optionMap);
      IPDF*            pdf        = compFactory.getPDF(optionMap);

      PdgID boson;
      opt = optionMap.getOption<string>("Process::DrellYan::BosonType");
      if (opt == "W+") boson = PdgID::kWPlus;
      else if (opt == "W-") boson = PdgID::kWMinus;
      else if (opt == "Z")  boson = PdgID::kZ;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unknown BosonType " << opt << Log::end;
      }
      if (boson != PdgID::kZ)
         Log::fatal("CalculationFactory::getCalculation") << "BosonType " << opt << " not yet implemented" << Log::end;
      PdgID lepton;
      opt = optionMap.getOption<string>("Process::DrellYan::DecayChannel");
      if (opt == "electrons") lepton = PdgID::kElectron;
      else if (opt == "muons")     lepton = PdgID::kMuon;
      else if (opt == "taus")      lepton = PdgID::kTau;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unknown DecayChannel " << opt << Log::end;
      }

      ICoupling* alphaem = compFactory.getCouplingEM(optionMap);
      // reads the value of alpha_s at MZ from PDF set
      CouplingStrong* alphas = compFactory.getCouplingStrong(optionMap);

      bool flavSum;
      if (phi1 == CalcSCETppV012::kmult)
         flavSum = true;
      else
         flavSum = false;

      const bool totalXSfix = optionMap.getOption<bool> ("CalcSCETppV012::TotalXSFix");
      const bool improvedXSfix = optionMap.getOption<bool> ("CalcSCETppV012::ImprovedXSFix");

      // The cuts to avoid numerical instabilities must be set at construction of the FKS generator.
      // For the current Madgraph package, values smaller than 1e-6 results in FPE in the HELAS routines
      double x1TinyCut = optionMap.getOption<double>("UnitCubeRadGenerator::x1TinyCut");
      double x2TinyCut = optionMap.getOption<double>("UnitCubeRadGenerator::x2TinyCut");
      bool independentCuts = optionMap.getOption<bool>("UnitCubeRadGenerator::independentCuts");

      UnitCubeRadGenerator* xRadGen  = new UnitCubeRadGenerator(x1TinyCut, x2TinyCut, independentCuts, false);

      double mVLow = optionMap.getOption<double>("PP2BosonJets::BInvMassCutLow");
      double mVHigh = optionMap.getOption<double>("PP2BosonJets::BInvMassCutHigh");

      _toDelete.push_back(new CalcSCETppV012(optionMap, Tau0cut, Tau1cut, Ecm, run0, run1, fixed0jet, fixed1jet, scale, profiles, phi1, alphaem, alphas,  pdf, tree, treeMEComp, oneLoopME, xRadGen, nLight,  boson, lepton, flavSum, totalXSfix, improvedXSfix, mVLow, mVHigh));

   } else if (calcName == "CalcNLOppV") {
      int                 nlight     = optionMap.getOption<int>("NLightFlavors");
      ITreeME*            treeME     = compFactory.getTreeME(optionMap);
      ITreeMEComplete*    treeMEComp = compFactory.getTreeMEComplete(optionMap);
      IOneLoopME*         oneLoopME  = compFactory.getOneLoopME(optionMap);

      double x1TinyCut = optionMap.getOption<double>("UnitCubeRadGenerator::x1TinyCut");
      double x2TinyCut = optionMap.getOption<double>("UnitCubeRadGenerator::x2TinyCut");
      bool independentCuts = optionMap.getOption<bool>("UnitCubeRadGenerator::independentCuts");

      UnitCubeRadGenerator* xRadGen  = new UnitCubeRadGenerator(x1TinyCut, x2TinyCut, independentCuts, false);

      IPDF*               pdf        = compFactory.getPDF(optionMap);
      double              Ecm        = optionMap.getOption<double>("Ecm");
      double              TauB0cut   = optionMap.getOption<double>("NJettiness::Tau0cut");

      PdgID boson;
      string opt = optionMap.getOption<string>("Process::DrellYan::BosonType");
      if (opt == "W+") boson = PdgID::kWPlus;
      else if (opt == "W-") boson = PdgID::kWMinus;
      else if (opt == "Z")  boson = PdgID::kZ;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unknown BosonType " << opt << Log::end;
      }
      assert(boson == PdgID::kZ); //untested otherwise
      PdgID lepton;
      opt = optionMap.getOption<string>("Process::DrellYan::DecayChannel");
      if (opt == "electrons") lepton = PdgID::kElectron;
      else if (opt == "muons")     lepton = PdgID::kMuon;
      else if (opt == "taus")      lepton = PdgID::kTau;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unknown DecayChannel " << opt << Log::end;
      }

      FixedOrder order;
      opt = optionMap.getOption<string>("FixedOrder");
      if (opt == "LO")  order = FixedOrder::kLO;
      else if (opt == "NLO") order = FixedOrder::kNLO;
      else {
         Log::error("CalculationFactory::getCalculation") << "Order not allowed " << opt << Log::end;
      }

      IScale* renScale = compFactory.getRenormalizationScale(optionMap);
      IScale* factScale = compFactory.getFactorizationScale(optionMap);

      ICoupling* alphaem = compFactory.getCouplingEM(optionMap);
      // reads the value of alpha_s at MZ from PDF set
      CouplingStrong* alphas = compFactory.getCouplingStrong(optionMap);

      _toDelete.push_back(new CalcNLOppV(optionMap, treeME, treeMEComp, oneLoopME, xRadGen, pdf, alphas, order, Ecm, nlight, renScale, factScale, alphaem, TauB0cut, boson, lepton));

   } else if (calcName == "CalcNLOppVj") {
      int                 nlight     = optionMap.getOption<int>("NLightFlavors");
      ITreeME*            treeME     = compFactory.getTreeME(optionMap);
      ITreeMEComplete*    treeMEComp = compFactory.getTreeMEComplete(optionMap);
      IOneLoopME*         oneLoopME  = compFactory.getOneLoopME(optionMap);

      double x1TinyCut = optionMap.getOption<double>("UnitCubeRadGenerator::x1TinyCut");
      double x2TinyCut = optionMap.getOption<double>("UnitCubeRadGenerator::x2TinyCut");
      bool independentCuts = optionMap.getOption<bool>("UnitCubeRadGenerator::independentCuts");

      UnitCubeRadGenerator* xRadGen  = new UnitCubeRadGenerator(x1TinyCut, x2TinyCut, independentCuts, false);

      IPDF*               pdf        = compFactory.getPDF(optionMap);
      double              Ecm        = optionMap.getOption<double>("Ecm");
      double              TauB0cut   = optionMap.getOption<double>("NJettiness::Tau0cut");
      double              TauB1cut   = optionMap.getOption<double>("NJettiness::Tau1cut");

      PdgID boson;
      string opt = optionMap.getOption<string>("Process::DrellYan::BosonType");
      if (opt == "W+") boson = PdgID::kWPlus;
      else if (opt == "W-") boson = PdgID::kWMinus;
      else if (opt == "Z")  boson = PdgID::kZ;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unknown BosonType " << opt << Log::end;
      }
      assert(boson == PdgID::kZ); //untested otherwise
      PdgID lepton;
      opt = optionMap.getOption<string>("Process::DrellYan::DecayChannel");
      if (opt == "electrons") lepton = PdgID::kElectron;
      else if (opt == "muons")     lepton = PdgID::kMuon;
      else if (opt == "taus")      lepton = PdgID::kTau;
      else {
         Log::error("CalculationFactory::getCalculation") << "Unknown DecayChannel " << opt << Log::end;
      }

      FixedOrder order;
      opt = optionMap.getOption<string>("FixedOrder");
      if (opt == "LO")  order = FixedOrder::kLO;
      else if (opt == "NLO") order = FixedOrder::kNLO;
      else {
         Log::error("CalculationFactory::getCalculation") << "Order not allowed " << opt << Log::end;
      }

      IScale* renScale = compFactory.getRenormalizationScale(optionMap);
      IScale* factScale = compFactory.getFactorizationScale(optionMap);

      ICoupling* alphaem = compFactory.getCouplingEM(optionMap);
      // reads the value of alpha_s at MZ from PDF set
      CouplingStrong* alphas = compFactory.getCouplingStrong(optionMap);

      CalcNLOppVj::Tau0NLO tau0NLO;
      opt = optionMap.getOption<string>("NJettiness::Tau0NLO");
      if (opt == "Tau0FR")       tau0NLO = CalcNLOppVj::kTau0FR;
      else if (opt == "Tau0FKS") tau0NLO = CalcNLOppVj::kTau0FKS;
      else {
         tau0NLO = CalcNLOppVj::kTau0FKS;
         Log::error("CalculationFactory::getCalculation") << "Unknown NJettiness::Tau0NLO " << opt << Log::end;
      }

      _toDelete.push_back(new CalcNLOppVj(optionMap, tau0NLO, treeME, treeMEComp, oneLoopME, xRadGen, pdf, alphas, order, Ecm, nlight, renScale, factScale, alphaem, TauB0cut, TauB1cut, boson, lepton));

   } else {
      Log::error("CalculationFactory::getCalculation") << "Unknown Calculation " << calcName << Log::end;
      return 0;
   }
   return _toDelete.back();
}

//------------------------------------------------------------------------------
CalculationFactory::~CalculationFactory()
{
   for (size_t i = 0; i < _toDelete.size(); ++i) {
      delete _toDelete[i];
   }
}

} // namespace Geneva
