//------------------------------------------------------------------------------
/// \file EngineNLOBase.cpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of EngineNLOBase class
//------------------------------------------------------------------------------

#include "Geneva/Calculations/EngineNLOBase.hpp"

#include "Geneva/Analyzers/AnalyzerFactory.hpp"
#include "Geneva/Analyzers/OutputStreamAnalyzer.hpp"

#include "Geneva/Components/FluxFactor.hpp"
#include "Geneva/Components/UnitCubeRadGenerator.hpp"

#include "Geneva/Core/Physics/UnitCubeRad.hpp"
#include "Geneva/Core/Physics/Signature.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/Physics/ConfigurationUtils.hpp"
#include "Geneva/Core/Physics/PhiRadFKS.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/UI/GenerationReport.hpp"
#include "Geneva/Core/Utils/Random.hpp"
#include "Geneva/Core/Utils/FileUtils.hpp"

#include <stdexcept>
#include <cassert>
#include <fstream>
#include <numeric>

namespace Geneva
{

//------------------------------------------------------------------------------
EngineNLOBase::EngineNLOBase(const OptionMap& optMap, const std::string& name, const SignatureSet& sigSet,
                             ITreeME* treeME, ITreeMEComplete* treeMEComplete, IOneLoopME* oneLoopME,
                             UnitCubeRadGenerator* xRadGenerator, ICommonCouplings* coupl,
                             IPhiNLO* phiNLO, CouplingStrong* alphas, FixedOrder order, int nLight)
   : _calcname(name), _lastGenWeight(std::numeric_limits<double>::quiet_NaN()), _order(order),
     _treeME(treeME), _treeMEComplete(treeMEComplete), _oneLoopME(oneLoopME),
     _xRadGenerator(xRadGenerator), _commonCouplings(coupl), _alphas(alphas),
     _analyzer(0), _report(0)
{
   if (!(_order == FixedOrder::kLO || _order == FixedOrder::kNLO))
      Log::fatal("EngineNLOBase") << "Order " << _order << " is not available."<< Log::end;

   _restriction = new Restriction(phiNLO);
   _numPointsReal = size_t_from_int(optMap.getOption<int>("EngineNloFks::NumPointsReal"));
   if (!_regionsInfo.isInitialized()) {
      Log::debug("EngineNLOBase") << "Initializing Signatures and Mappings for "<< _calcname << Log::end;
      // signatures with fewer particles will be always ordered first
      size_t nParticlesAtLO = sigSet.begin()->size();
      // NOTE: _signaturesBN and  _signaturesBNp1 must live as long as EngineNLOBase, and definitively more than derived classes, since RegionsInfo methods in derived classes use pointers to them
      _signaturesBN = sigSet.copy_if(CheckSignatureSize(nParticlesAtLO));

      if (_order == FixedOrder::kNLO) {
         _signaturesBNp1 = sigSet.copy_if(CheckSignatureSize(nParticlesAtLO + 1));
         //sanity check
         assert(_signaturesBN.size() + _signaturesBNp1.size() == sigSet.size());
         _regionsInfo.setRegionInfoList(_signaturesBN, _signaturesBNp1, nLight);
      }
   }

   if (optMap.getOption<bool>("Analyzer::FixedOrder"))
      _analyzer = AnalyzerFactory::instance().getInternalAnalyzer(optMap);
   _doAnalyze = (_analyzer != 0);
   if (_doAnalyze) {
      Log::verboseInfo("EngineNLOBase") << "Using " << _analyzer->name() << " for " << _calcname << Log::end;
      std::string extraTag;
      if (_order == FixedOrder::kLO)
         extraTag = "fixedLO";
      else
         extraTag = "fixedNLO";
      _analyzerFile = GenerationReport::getFilename(optMap, extraTag);
      _report = new GenerationReport(optMap, "FixedOrder_" +  _calcname);
      _report->setTitle(GenerationReport::getDefaultTitle(optMap, extraTag));
      _report->setRunTimeInSec(optMap.getOption<double>("maxTime") * 60.);
   }
}

//------------------------------------------------------------------------------
EngineNLOBase::~EngineNLOBase()
{
   delete _report;
}

//------------------------------------------------------------------------------
/**
 * Basing on the calculation order requested, this calculates the inclusive
 * result for X + N-jet @LO  or the results for the exclusive X + N jet production rate @ NLO
 * combined with the X + (N+1)-jet rate @ LO. In the latter case it uses the FKS subtractions
 * discussed in detail in the accompanying physics file.
 */
Sigma EngineNLOBase::calculate(const Configuration& config, bool doVariations) const
{
   if (_order == FixedOrder::kNLO) {
      NloCoefficients coeff = getFOCoeffs(config, doVariations);
      if (not coeff.lo().isDefined()) {
         assert(not coeff.nloBornKin().isDefined() and not coeff.nloRealKin().isDefined());
         return Sigma();
      }
      if (not coeff.nloBornKin().isDefined() or not coeff.nloRealKin().isDefined()) {
         assert(not coeff.lo().isDefined());
         return Sigma();
      }
      return coeff.sum();
   } else if (_order == FixedOrder::kLO) {
      return getLOCoeff(config, doVariations);
   } else {
      Log::fatal("EngineNLOBase::calculate")<< "The requested order: " << _order << " is not available for "<< _calcname << Log::end;
      return Sigma();
   }
}

//------------------------------------------------------------------------------
/**
 * Components used in the engine base class, in common between all NLO calculations with FKS subtraction
 */
const std::vector<NameValuePair> EngineNLOBase::componentInfo() const
{
   std::vector<NameValuePair> components;
   components.push_back(NameValuePair("treeME", _treeME->name()));
   components.push_back(NameValuePair("commonCouplings", _commonCouplings->name()));
   if (_order == FixedOrder::kNLO) {
      components.push_back(NameValuePair("treeMEComplete", _treeMEComplete->name()));
      components.push_back(NameValuePair("oneLoopME", _oneLoopME->name()));
      components.push_back(NameValuePair("alphas", _alphas->name()));
   }
   return components;
}

//-----------------------------------------------------------------------------
void EngineNLOBase::resetAnalyzer() const
{
   if (_doAnalyze)
      _analyzer->reset();
}

//------------------------------------------------------------------------------
/// Performs the final operations, like saving the fixed-order analyzer to a file.
void EngineNLOBase::finalize(const size_t numEvents)
{
   if (_doAnalyze) {
      _report->setNumEventsGenerated(numEvents);
      _report->outputStream() << OutputStreamAnalyzer(_analyzer, _report->title());
      std::ofstream file(_analyzerFile.c_str());
      file << *_report;
      file.close();
      if (!file)
         Log::warning("EngineNLOBase::finalize") << "Could not write the fixed-order analyzer report to file: "
                                                 << _analyzerFile << Log::end;
      else
         Log::info("EngineNLOBase::finalize") << "Wrote fixed-order analyzer report to file: "
                                              << _analyzerFile << Log::end;
      Log::verboseInfo("EngineNLOBase::finalize") << "Results from " << _report->title() << Log::end;
      Log::verboseInfo() << *_analyzer << Log::end;
   }
   //After dumping the Analyzer make sure that no further weight is fed into it
   _doAnalyze = false;
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/// \todo: this should really check if config has the correct signature as well
bool EngineNLOBase::Restriction::contains(const Configuration& config)
{
   const Configuration config_CM = config.transformed(BoostZHyp(sqrt(config.initialB()->energy()/config.initialA()->energy())));
   return _phiNLO->isPhiN(config_CM) || _phiNLO->isPhiNplus1(config_CM);
}

//------------------------------------------------------------------------------
/// \todo: this is undefined if config does not have a correct signature
size_t EngineNLOBase::Restriction::countJets(const Configuration& config) const
{
   const Configuration config_CM = config.transformed(BoostZHyp(sqrt(config.initialB()->energy()/config.initialA()->energy())));
   return _phiNLO->countN(config_CM);
}

} // namespace Geneva
