//------------------------------------------------------------------------------
/// \file  EngineNLOpp.cpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class EngineNLOpp
//------------------------------------------------------------------------------

#include "Geneva/Calculations/EngineNLOpp.hpp"
#include "Geneva/Calculations/IPhiNLOwithPDF.hpp"

#include "Geneva/Components/IScale.hpp"
#include "Geneva/Components/SubtractionFKSpp.hpp"
#include "Geneva/Components/FluxFactor.hpp"
#include "Geneva/Components/UnitCubeRadGenerator.hpp"
#include "Geneva/Components/FKSKinematics.hpp"

#include "Geneva/Analyzers/IAnalyzer.hpp"

#include "Geneva/Core/Physics/QCD.hpp"
#include "Geneva/Core/Physics/UnitCubeRad.hpp"
#include "Geneva/Core/Physics/PhiRadFKS.hpp"
#include "Geneva/Core/Physics/Signature.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/Physics/ConfigurationUtils.hpp"
#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/Event.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/Utils/Random.hpp"

#include <cassert>
#include <numeric>

//#define CHECK_SOFT
//#define CHECK_COLL
//#define CHECK_SOFTCOLL

#define INCLUDE_SOFT
#define INCLUDE_COLL
#define INCLUDE_SOFTCOLL

//#define ONLY_ISRA
//#define ONLY_ISRB
//#define ONLY_FSR


#define XICUTDIV 0.
#define YCUTDIV 1.

#define REFINE_ENGINENLO_PRECISION 1e-10
#define REFINE_ENGINENLO_ACCURACY 1e-13

namespace Geneva
{

//------------------------------------------------------------------------------
EngineNLOpp::EngineNLOpp(const OptionMap& optMap,  std::string name, const SignatureSet& sigSet,
                         ITreeME* treeME, ITreeMEComplete* treeMEComplete, IOneLoopME* oneLoopME,
                         UnitCubeRadGenerator* radGenerator, IPDF* pdf, IScale* renScale, IScale* factScale,
                         ICommonCouplings* LoCouplings, IPhiNLOwithPDF* phiNLO, CouplingStrong* alphas,
                         FixedOrder order, double Ecm, int nLight)
   :  EngineNLOBase(optMap, name, sigSet, treeME, treeMEComplete, oneLoopME, radGenerator, LoCouplings, phiNLO, alphas, order, nLight),
      _renScale(renScale), _factScale(factScale), _phiNLO(phiNLO), _subtraction(0),
      _pdf(pdf), _cmEnergy(Ecm), _numPointsCollRemn(0), _impSampleCollremn(false)
{
   double renfac = optMap.getOption<double>("RenormalizationScale::VaryFactor");
   double factfac = optMap.getOption<double>("FactorizationScale::VaryFactor");
   std::string vary = optMap.getOption("FixedOrder::ScaleVariations");
   if (vary == "none") {
      _scaleVariation = kNone;
      _numScaleVariations = 0;
   } else if (vary == "corr") {
      _scaleVariation = kCorrelated;
      _numScaleVariations = 2;
      _renScaleVaryCoeffs.push_back(renfac);
      _renScaleVaryCoeffs.push_back(1./renfac);
      _factScaleVaryCoeffs.push_back(factfac);
      _factScaleVaryCoeffs.push_back(1./factfac);
   } else if (vary == "uncorr") {
      _scaleVariation = kUncorrelated;
      _numScaleVariations = 8;
      _renScaleVaryCoeffs.push_back(renfac);
      _renScaleVaryCoeffs.push_back(1./renfac);
      _factScaleVaryCoeffs.push_back(factfac);
      _factScaleVaryCoeffs.push_back(1./factfac);
   } else if (vary == "uncorrNoExtremes") {
      _scaleVariation = kUncorrelatedNoExtremes;
      _numScaleVariations = 6;
      _renScaleVaryCoeffs.push_back(renfac);
      _renScaleVaryCoeffs.push_back(1./renfac);
      _factScaleVaryCoeffs.push_back(factfac);
      _factScaleVaryCoeffs.push_back(1./factfac);
   } else {
      Log::fatal("EngineNLOpp") << "Unknown choice of scale variations" << vary << Log::end;
   }

   if (order == FixedOrder::kNLO) {
      _numPointsCollRemn = size_t_from_int(optMap.getOption<int>("EngineNloFks::NumPointsCollRemn"));
      _impSampleCollremn = optMap.getOption<bool>("EngineNloFks::ImportanceSampleCollRemn");

      const double xiCut = optMap.optionExists("SubtractionFKS::xiCut") ? optMap.getOption<double>("SubtractionFKS::xiCut") : 1.0;
      const double yCutISR = optMap.optionExists("SubtractionFKS::yCutISR") ? optMap.getOption<double>("SubtractionFKS::yCutISR") : 2.0;
      const double yCutFSR = optMap.optionExists("SubtractionFKS::yCutFSR") ? optMap.getOption<double>("SubtractionFKS::yCutFSR") : 2.0;
      _subtraction = new SubtractionFKSpp(treeMEComplete, nLight, xiCut, yCutISR, yCutFSR);
   }

   _includeBorn = optMap.getOption<bool>("EngineNloFks::IncludeBorn");
   _includeSoftVirt = optMap.getOption<bool>("EngineNloFks::IncludeSoftVirt") && order == FixedOrder::kNLO;
   _includeCollRemn = optMap.getOption<bool>("EngineNloFks::IncludeCollRemn") && order == FixedOrder::kNLO;
   _includeReal = optMap.getOption<bool>("EngineNloFks::IncludeReal") && order == FixedOrder::kNLO;
}

//------------------------------------------------------------------------------
/**
 * Explicitly destroys objects created at initialization, in the opposite order of creation,
 * to avoid memory leaks.
 */
EngineNLOpp::~EngineNLOpp()
{
   delete _subtraction;
}

//------------------------------------------------------------------------------
/**
 * Components used in the engine class, in common between all NLO calculations with FKS subtraction
 */
const std::vector<NameValuePair> EngineNLOpp::componentInfo() const
{
   std::vector<NameValuePair> components, base;
   base = EngineNLOBase::componentInfo();
   std::copy(base.begin(), base.end(), std::back_inserter(components));
   components.push_back(NameValuePair("PDF", _pdf->name()));
   components.push_back(NameValuePair("muR", _renScale->name()));
   components.push_back(NameValuePair("muF", _factScale->name()));
   return components;
}

//------------------------------------------------------------------------------
/**
* \brief Calculate the LO coefficient.
* \param config The configuration to act on.
* \return The LO coefficient in GeV.
*
* Calculate the LO coefficient, cLO, including the overall flux-factor and
* coupling prefactors.
* If \a config does not satisfy the LO  phase space cuts then 0.0 is
* returned for cLO. This also includes the configurations  which are
* relative NLO.
* To ensure consistency with the NLO part, when a non boost-invariant metric is used
* it is crucial that isPhiN acts on a configuration in the partonic CM frame.
* However, the information in a Configuration in the partonic CM frame might
* not be enough, so we always pass a Configuration in the Lab frame and perform
* the boost to the CM frame internally.
*
*/
Sigma EngineNLOpp::getLOCoeff(const Configuration& config, bool doVariations) const
{
   if (_doAnalyze)
      _analyzer->initializeCorrelated();
   Sigma res;
   if (_phiNLO->isPhiN(config)) {
#ifndef NDEBUG
      if (!checkMomentumConservation(config, 1e-6, 1e-8))
         Log::warning("EngineNLOpp::getLOCoeff") << "PhiN point violates momentum conservation." << Log::end;
#endif
      res = calcTreeLevel(config, _commonCouplings->numAlphaStrongPowers(), doVariations);
      if (_doAnalyze && std::isnormal(_lastGenWeight)) {
         _analyzer->analyzeCorrelated(Event(config, res * math_constants::gev2pb()/ _lastGenWeight));
      }
   }
   if (_doAnalyze)
      _analyzer->finalizeCorrelated();
   return res;
}


//------------------------------------------------------------------------------
/**
 * \brief Calculate the NLO coefficients.
 * \param config The configuration to act on.
 * \return A pair containing the LO and NLO coefficients in GeV.
 *
 * Calculate the  NLO coefficients, including the overall
 * flux-factor and coupling prefactors.
 * If \a config does not satisfy the N-body or (N+1)-body phase space cuts then
 * 0.0 is returned for both cLO and cNLO coefficients. Note that if \a config
 * is purely NLO (e.g. for a (N+1)-body config) then the born  will be 0.0 while
 * other purely NLO coefficients can be nonzero.
 *
 * To perform the NLO subtraction and to evaluate the phase space
 * restrictions when a non boost-invariant metric is used (in particular
 * the theta functions that performs the separation of the singular regions)
 * it is crucial to act on a configuration in the partonic CM frame, where the
 * FKS variables are defined.
 *
 * Since the information in a Configuration in the partonic CM frame might
 * not be enough to evaluate isPhin or isPhinplus1, we always pass a Configuration
 * in the Lab frame and perform the boost to the CM frame internally.
 */
NloCoefficients EngineNLOpp::getFOCoeffs(const Configuration& config, bool doVariations) const
{
   return calcNLO(config, _includeBorn, _includeSoftVirt, _includeCollRemn, _includeReal, doVariations);
}


//------------------------------------------------------------------------------
/**
 * \brief Calculate the NLO coefficients with Born-like kinematics.
 * \param config The configuration to act on.
 * \return A pair containing the LO and NLO coefficients in GeV.
 *
 * Calculate the  NLO coefficients with Born-like kinematics, including the overall
 * flux-factor and coupling prefactors.
 * If \a config does not satisfy the N-body  phase space cuts then
 * 0.0 is returned for both cLO and cNLO coefficients.
 *
 * To perform the NLO subtraction and to evaluate the phase space
 * restrictions when a non boost-invariant metric is used (in particular
 * the theta functions that performs the separation of the singular regions)
 * it is crucial to act on a configuration in the partonic CM frame, where the
 * FKS variables are defined.
 *
 * Since the information in a Configuration in the partonic CM frame might
 * not be enough to evaluate isPhin, we always pass a Configuration
 * in the Lab frame and perform the boost to the CM frame internally.
 */
NloCoefficients EngineNLOpp::getFOCoeffsBornKin(const Configuration& config, bool doVariations) const
{
   return calcNLO(config, _includeBorn, _includeSoftVirt, _includeCollRemn, false, doVariations);
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
std::vector<double> EngineNLOpp::getScaleVariations(const Configuration& config, const PdgID idA, const double xA, const PdgID idB, double xB, const double couplStrongCentral, const int nPowers, const double luminosityCentral, const double mufc, const double murc, const double Q, const double w0, const double wf, const double wr) const
{
   size_t length = _renScaleVaryCoeffs.size();
   if (length != _factScaleVaryCoeffs.size())
      Log::error("EngineNLOpp::getScaleVariations") << "Different sizes of scale variations vectors" << length << " " << _factScaleVaryCoeffs.size() << Log::end;
   std::vector<double> variations;
   if (_scaleVariation == kCorrelated) {
      for (size_t i = 0; i < length; ++i) {
         const double muf = _factScaleVaryCoeffs[i] * mufc;
         const double luminosityVary = _pdf->luminosity(idA, xA, idB, xB, muf);
         const double mur = _renScaleVaryCoeffs [i] * murc;
         const double couplVary = pow((*_alphas)(mur), nPowers);
         variations.push_back(couplVary * luminosityVary * (w0 + 2.* wf * log(muf / Q) + 2.* wr * log(mur / Q)));
      }
   } else if (_scaleVariation == kUncorrelatedNoExtremes) {
      for (size_t i = 0; i < length; ++i) {
         const double muf = _factScaleVaryCoeffs[i] * mufc;
         const double luminosityVary = _pdf->luminosity(idA, xA, idB, xB, muf);
         const double mur = _renScaleVaryCoeffs[i] * murc;
         const double couplVary = pow((*_alphas)(mur), nPowers);
         variations.push_back(couplVary * luminosityVary * (w0 + 2.* wf * log(muf / Q) + 2.* wr * log(mur / Q)));
         variations.push_back(couplStrongCentral * luminosityVary * (w0 + 2.* wf * log(muf / Q) + 2.* wr * log(murc / Q)));
         variations.push_back(couplVary * luminosityCentral * (w0 + 2.* wf * log(mufc / Q) + 2.* wr * log(mur / Q)));
      }
   } else if (_scaleVariation == kUncorrelated) {
      for (size_t i = 0; i < length; ++i) {
         const double muf = _factScaleVaryCoeffs[i] * mufc;
         const double luminosityVary = _pdf->luminosity(idA, xA, idB, xB, muf);
         const double mur = _renScaleVaryCoeffs[i] * murc;
         const double couplVary = pow((*_alphas)(mur), nPowers);
         variations.push_back(couplVary * luminosityVary * (w0 + 2.* wf * log(muf / Q) + 2.* wr * log(mur / Q)));
         variations.push_back(couplStrongCentral * luminosityVary * (w0 + 2.* wf * log(muf / Q) + 2.* wr * log(murc / Q)));
         variations.push_back(couplVary * luminosityCentral * (w0 + 2.* wf * log(mufc / Q) + 2.* wr * log(mur / Q)));
         for (size_t j = 0; j < length; ++j) {
            if (i != j) {
               const double mur = _renScaleVaryCoeffs[j] * murc;
               const double couplVary = pow((*_alphas)(mur), nPowers);
               variations.push_back(couplVary * luminosityVary * (w0 + 2.* wf * log(muf / Q) + 2.* wr * log(mur / Q)));
            }
         }
      }
   }
   assert(variations.size() == _numScaleVariations);
   return variations;
}

//------------------------------------------------------------------------------
/**
 * \brief helper function to calculate the tree-level contribution
 * \param config The configuration to act on.
 * \return The tree-level result, including scale variations..
 *
 * Calculate the tree level contribution, including the overall flux-factor, luminosity  and
 * coupling prefactors, without the overall GeV to pb conversion factor
 */
Sigma EngineNLOpp::calcTreeLevel(const Configuration& config, int asExp, bool doVariations) const
{
   const double xA = 2.*config.initialA()->energy() / _cmEnergy;
   const double xB = 2.*config.initialB()->energy() / _cmEnergy;
   assert(xA >= 0. && xA <= 1.0);
   assert(xB >= 0. && xB <= 1.0);
   const double muRc = (*_renScale)(config);
   const double couplStrong = pow((*_alphas)(muRc), asExp);
   const double couplStrongBorn = pow((*_alphas)(muRc), _commonCouplings->numAlphaStrongPowers());
   const double couplNonStrongBorn = (*_commonCouplings)(config, muRc) / couplStrongBorn;
   const double pre = FluxFactor()(config) * couplNonStrongBorn;
   const double muFc = (*_factScale)(config);
   const double luminosity = _pdf->luminosity(config.initialA()->id(), xA, config.initialB()->id(), xB, muFc);
   const double tree = (*_treeME)(config);
   const double central = couplStrong * luminosity * tree;
   std::vector<double> variations;
   if (doVariations) {
      variations = getScaleVariations(config, config.initialA()->id(), xA, config.initialB()->id(), xB, couplStrong, asExp, luminosity, muFc, muRc, muRc, tree, 0.0, 0.0);
   }
   return pre * Sigma(central, variations);
}

//------------------------------------------------------------------------------
NloCoefficients EngineNLOpp::calcNLO(const Configuration& config, bool includeBorn, bool includeVirtSub, bool includeCollRemn, bool includeRealsSub, bool doVariations) const
{
   size_t numScaleVariations = doVariations ? _numScaleVariations : 0;
   if (_doAnalyze)
      _analyzer->initializeCorrelated();
   Sigma loCoeff;
   Sigma nloCoeffBornKin;
   Sigma nloCoeffRealKin;
   if (_phiNLO->isPhiN(config)) {
      const double xBarA = 2.*config.initialA()->energy() / _cmEnergy;
      const double xBarB = 2.*config.initialB()->energy() / _cmEnergy;
      assert(xBarA >= 0. && xBarA <= 1.0);
      assert(xBarB >= 0. && xBarB <= 1.0);
      Configuration config_CM = config.transformed(BoostZHyp(sqrt(xBarB/xBarA)));
#ifdef REFINE_MOMENTA
      enforceMomentumConservation(config_CM, "EngineNLOpp::calcNLO", REFINE_ENGINENLO_PRECISION, REFINE_ENGINENLO_ACCURACY);
#endif
      const double muRc = (*_renScale)(config);
      const double alphas = (*_alphas)(muRc);
      const int asExp = _commonCouplings->numAlphaStrongPowers();
      const double couplStrong = pow(alphas, asExp);
      const double couplNonStrong = (*_commonCouplings)(config, muRc) / couplStrong;
      const double muFc = (*_factScale)(config);
      const double luminosity = _pdf->luminosity(config.initialA()->id(), xBarA, config.initialB()->id(), xBarB, muFc);
      const double flux = FluxFactor()(config);
      const double pre = flux * couplNonStrong;
      const double tree = (*_treeME)(config);
      const double muES = muRc;
      if (includeBorn) {
         const double resultB = tree * couplStrong * luminosity;
         if (not std::isfinite(resultB)) {  //Make sure once again that we don't get nan's
            Log::error("EngineNLOpp::calcNLO") << "B = " << resultB << " in " << _calcname << "\n"<<std::setprecision(16) << config << Log::end;
         }
         std::vector<double> bornVariations;
         if (doVariations) {
            bornVariations = getScaleVariations(config, config.initialA()->id(), xBarA, config.initialB()->id(), xBarB, couplStrong, asExp, luminosity, muFc, muRc, muES, tree, 0.0, 0.0);
         }
         loCoeff = pre * Sigma(resultB, bornVariations);
      } else {
         loCoeff = Sigma(0., numScaleVariations);
      }
      if (includeVirtSub || includeRealsSub || includeCollRemn) {
         _subtraction->setScale(muES);
         // Performs the required tree level calculations (pieces of this are used below)
         _treeMEComplete->calculate(config_CM);
      }
      if (includeVirtSub) {
         // Evaluates the Virtual term and Integrated Subtraction
         _oneLoopME->setScale(_subtraction->getScale());
         const double wV0 = _oneLoopME->getFiniteTerm(config);
         const double wVR = math_constants::_1_4pi() * QCD::beta0(_subtraction->nLight()) * _commonCouplings->numAlphaStrongPowers() * tree;
         std::vector<double> intSub = _subtraction->getIntegratedSubtractionCoefficients(config_CM);
         const double centralVI = couplStrong * luminosity * alphas * (wV0 + intSub[0] + 2. * intSub[2] * log(muFc / _subtraction->getScale()) + 2. * wVR * log(muRc / _subtraction->getScale()));
         if (not std::isfinite(centralVI)) {  //Make sure once again that we don't get nan's
            Log::error("EngineNLOpp::calcNLO") << "V+I = " << centralVI << " V= "<< wV0 <<" in " << _calcname << "\n"<<std::setprecision(16) << config <<  Log::end;
         }
         std::vector<double> varyVI;
         if (doVariations) {
            varyVI = getScaleVariations(config, config.initialA()->id(), xBarA, config.initialB()->id(), xBarB, couplStrong * alphas, asExp + 1, luminosity, muFc, muRc, _subtraction->getScale(), wV0 + intSub[0], intSub[2], wVR);
         }
         if (not nloCoeffBornKin.isDefined())
            nloCoeffBornKin = pre * Sigma(centralVI, varyVI);
         else
            nloCoeffBornKin += pre * Sigma(centralVI, varyVI);
      }
      if (includeRealsSub) {
         // Evaluates the R-S integral including pdf, coupling constants and prefactors
         nloCoeffRealKin = getIntegralRealSubtracted(config_CM, xBarA, xBarB, flux, muFc, muRc, luminosity, alphas, asExp, couplStrong, couplNonStrong, doVariations);
      } else {
         nloCoeffRealKin = Sigma(0., numScaleVariations);
      }
      if (includeCollRemn) {
         // Evaluates the Collinear Remnants integral including the pdfs
         const Sigma resultCR = getIntegralCollinearRemnants(config, xBarA, xBarB, flux, muFc, muRc, luminosity, alphas, asExp, couplStrong, doVariations);
         if (not std::isfinite(resultCR.central())) {  //Make sure once again that we don't get nan's
            Log::error("EngineNLOpp::calcNLO") << "CR = "<< resultCR.central() << " in " << _calcname << "\n"<<std::setprecision(16) << config <<  Log::end;
         }
         if (not nloCoeffBornKin.isDefined())
            nloCoeffBornKin = pre * resultCR;
         else
            nloCoeffBornKin += pre * resultCR;
      }
      if (not includeVirtSub && not includeCollRemn)
         nloCoeffBornKin = Sigma(0., numScaleVariations);

      if (_doAnalyze && std::isnormal(_lastGenWeight)) {
         Event event(config, (loCoeff + nloCoeffBornKin) * math_constants::gev2pb() / _lastGenWeight);
         _analyzer->analyzeCorrelated(event);
      }
   }
   if (_phiNLO->isPhiNplus1(config) && includeRealsSub) {
      loCoeff = Sigma(0., numScaleVariations);
      nloCoeffBornKin = Sigma(0., numScaleVariations);
      nloCoeffRealKin = calcTreeLevel(config, _commonCouplings->numAlphaStrongPowers() + 1, doVariations);
      if (not std::isfinite(nloCoeffRealKin.central())) {  //Make sure once again that we don't get nan's
         Log::error("EngineNLOpp::calcNLO") << "R = " << nloCoeffRealKin.central() << " in " << _calcname << "\n"<<std::setprecision(16) << config <<  Log::end;
      }
      if (_doAnalyze && std::isnormal(_lastGenWeight)) {
         Event event(config, nloCoeffRealKin  * math_constants::gev2pb() / _lastGenWeight);
         _analyzer->analyzeCorrelated(event);
      }
   }
   if (_doAnalyze)
      _analyzer->finalizeCorrelated();

   //Make sure if one coeff is defined the other ones are too
   if (loCoeff.isDefined() or nloCoeffBornKin.isDefined() or nloCoeffRealKin.isDefined()) {
      if (not loCoeff.isDefined())
         loCoeff = Sigma(0., numScaleVariations);
      if (not nloCoeffBornKin.isDefined())
         nloCoeffBornKin = Sigma(0., numScaleVariations);
      if (not nloCoeffRealKin.isDefined())
         nloCoeffRealKin = Sigma(0., numScaleVariations);
   }
   return NloCoefficients(loCoeff, nloCoeffBornKin, nloCoeffRealKin);
}

//------------------------------------------------------------------------------
Sigma EngineNLOpp::getIntegralRealSubtracted(const Configuration& config_CM, const double xBarA, const double xBarB, const double fluxBorn, const double mufBorn, const double murBorn, const double lumiBorn, const double alphasBorn, const int asExpBorn, const double couplStrongBorn, const double couplNonStrongBorn, bool doVariations) const
{
//    Log::debug().turnOn();
   Log::debug() << "====================================" << Log::end;
   Log::debug() << "======      NEW EVENT       ========" << Log::end;
   Log::debug() << "====================================" << Log::end;
   Sigma result(0., doVariations ? _numScaleVariations : 0);
   PhiNcmFKS phiN(config_CM);
   double Q = (config_CM.initialA()->p() + config_CM.initialB()->p()).magnitude();
   std::list<RegionInfo> Mappings = _regionsInfo.getSingularRegionsFromSignature(Signature(config_CM));
   for (std::list<RegionInfo>::iterator m = Mappings.begin(); m != Mappings.end(); ++m) {
      //Select the emitter and set in PhiN
      Log::debug() << "====================================" << Log::end;
      phiN.emitter = m->getEmitter(phiN.configCM);
#ifdef ONLY_ISRA
      if (!phiN.emitter->isInitialA()) continue;
#endif
#ifdef ONLY_ISRB
      if (!phiN.emitter->isInitialB()) continue;
#endif
#ifdef ONLY_FSR
      if (!phiN.emitter->isFinal()) continue;
#endif
      for (unsigned n = 0; n < _numPointsReal; n++) {
         Sigma event;
         Sigma counterEvent;
         //Generate importance-sampled variables in the unit cube
         UnitCubeRad xRad;
         generateUnitCubeRad(m, xRad);
         // Get the real contribution
         Configuration configNp1_CM;
         Configuration::iterator emitted, sister;
         double xAreal = 0.;
         double xBreal = 0.;
         double jacReal = 0.;
         PhiRadFKS phiRadReal;
         bool didSplit = _phiNLO->constructPhiNplus1(phiN.configCM, phiN.emitter, xBarA, xBarB, xRad, configNp1_CM, emitted, sister, xAreal, xBreal, phiRadReal, jacReal);
#ifdef REFINE_MOMENTA
         enforceMomentumConservation(configNp1_CM, "EngineNLOpp::getIntegralRealSubtracted", REFINE_ENGINENLO_PRECISION, REFINE_ENGINENLO_ACCURACY);
#endif

#ifndef ONLYSUBT
         // Get the real contributions to the event weight
         std::pair<bool, bool> removeReal = _phiNLO->cutoffFKS(phiRadReal);
         Log::debug("Engine real") << Log::end;
         if (didSplit and !removeReal.first and !removeReal.second) {
            assert(jacReal > 0.0);
            double mufReal = (*_factScale)(configNp1_CM);
            double fluxReal = FluxFactor()(configNp1_CM);
            double lumiReal = _pdf->luminosity(configNp1_CM.initialA()->id(), xAreal, configNp1_CM.initialB()->id(), xBreal, mufReal);
            double wR0 = (*_treeME)(configNp1_CM);
            double murReal = (*_renScale)(configNp1_CM);
            double alphasReal = (*_alphas)(murReal);
            double couplStrongReal = pow(alphasReal, asExpBorn + 1);
            double couplNonStrongReal = (*_commonCouplings)(configNp1_CM, murReal) / couplStrongReal * alphasReal;
            double realc = wR0 * lumiReal * couplStrongReal;
            if (not std::isfinite(realc)) {
               Log::error("EngineNLOpp::getIntegralRealSubtracted") << "getReal " << realc << " on \n" << configNp1_CM << Log::end;
            }
            std::vector<double> variations;
            if (doVariations) {
               variations = getScaleVariations(configNp1_CM, configNp1_CM.initialA()->id(), xAreal, configNp1_CM.initialB()->id(), xBreal, couplStrongReal, asExpBorn + 1, lumiReal, mufReal, murReal, murReal, wR0, 0.0, 0.0);
            }
            event = Sigma(realc, variations) * couplNonStrongReal * jacReal * fluxReal * static_cast<double>(m->multRad);
         }
#endif

#ifdef INCLUDE_SOFT
         // Get the soft contribution
         Log::debug("Engine soft") << Log::end;
         Sigma soft;
         if (m->i->isGluon()) {
            Log::debug("EngineNLOpp::getIntegralRealSubtracted") << "Qbar " << Q << Log::end;
            double jacSoft = 0.;
            PhiRadFKS phiRadSoft;
            _phiNLO->constructSoftLimit(phiN.configCM, phiN.emitter, xBarA, xBarB, xRad, phiRadSoft, jacSoft);
            std::pair<bool, bool> removeSoft = _phiNLO->cutoffFKS(phiRadSoft);

            double xiMaxS = 2.*_phiNLO->maxSoftEmittedEnergyCm(phiN.configCM, phiN.emitter, xBarA, xBarB, xRad) / Q;
            Log::debug("phiRadSoft NEW xiMaxS=") << xiMaxS << Log::end;
            Log::debug("jacsoft NEW") << jacSoft << Log::end;

            //Evaluate the soft if ThetaijRijSoft passes and the the soft is not removed due to the y being too small.
            //We need to calculate if the soft is removed due to xi being too small, since it is needed for the analytical integration piece later
            if (!removeSoft.second and _phiNLO->ThetaijRijSoft(phiN.configCM, phiN.emitter, _subtraction->getSoftVec(phiN.emitter->p(), phiRadSoft, Q))) {
               double jacTmp = 0.;
               double wS0 = _subtraction->getSoft(phiN, phiRadSoft, jacTmp);
               jacSoft *= jacTmp;
               double softc = wS0 * lumiBorn * couplStrongBorn * alphasBorn;
               if (not std::isfinite(softc)) {
                  Log::error("EngineNLOpp::getIntegralRealSubtracted") << "getSoft " << soft << " on \n" << phiN.configCM << Log::end;
               }
               std::vector<double> variations;
               if (doVariations) {
                  variations = getScaleVariations(phiN.configCM, phiN.configCM.initialA()->id(), xBarA, phiN.configCM.initialB()->id(), xBarB, couplStrongBorn * alphasBorn, asExpBorn + 1, lumiBorn, mufBorn, murBorn, _subtraction->getScale(), wS0, 0.0, 0.0);
               }
               soft = Sigma(softc, variations) * couplNonStrongBorn * jacSoft * fluxBorn  * static_cast<double>(m->multN);
# if defined CHECK_SOFT
               if (didSplit) {
                  FourVector pSoft = _subtraction->getSoftVec(phiN.emitter->p(), phiRadSoft, Q);
                  FourVector diff = emitted->p() - pSoft;
                  Log::debug() << *m << Log::end;
                  Log::debug() << phiRadSoft.xi() << Log::end;
                  Log::debug() << "Tiny Cutoff: " << _xRadGenerator->x1TinyCutoff() << Log::end;
                  Log::debug() << "p Emitted: " << emitted->p() << Log::end;
                  Log::debug() << "p Soft   : " << pSoft << Log::end;
                  Log::debug() << "Diff   : " << diff << Log::end;
                  Log::debug() << "PhiRadFks real : " << FKSKinematics().projectPhiRad(PhiNplus1(configNp1_CM, emitted, sister)) << Log::end;
                  Log::debug() << "PhiRadFks soft : " << phiRadSoft << Log::end;
                  Log::debug() << "jacReal : " << jacReal << Log::end;
                  Log::debug() << "jacSoft : " << jacSoft* Q* Q* phiRadSoft.xi() * math_constants::_1_64pi3() << Log::end;
                  Log::debug() << "Real : " << real << Log::end;
                  Log::debug() << "Soft : " << soft << Log::end;
                  if (compareToZero(real / soft - 1.0, 0.1) != 0) {
                     Log::debug() << "Press enter to continue" << Log::end;
                     std::cin.get();
                  }
               }
# endif
            }
            // Add  the soft contributions to the counter-event weight. The xiAnalyticFactorSoft takes into account the potential different limits on xi for the real and the soft.
            //This also takes care of replacing the real with the subtraction in some singular regions
            bool thetaYSoft;
            if (xRad.radType() == UnitCubeRad::kISRB)
               thetaYSoft = (phiRadSoft.y() > -YCUTDIV);
            else
               thetaYSoft = (phiRadSoft.y() < YCUTDIV);

            //This removes the differential soft
            double thetaXiSoft = (phiRadSoft.xi() > XICUTDIV and !removeSoft.first) ? 1. : 0.;
            double xiAnalyticFactorSoft = (thetaXiSoft + log(_subtraction->xiCut() / std::max(xiMaxS, XICUTDIV)) * xRad.x1());
            if (soft.isDefined() && thetaYSoft)
               counterEvent = -xiAnalyticFactorSoft * soft;
         }
#endif

#ifdef INCLUDE_COLL
         // Get the collinear contribution
         Log::debug("Engine coll") << Log::end;
         double jacColl = 0.;
         PhiRadFKS phiRadColl;
         bool didSplitColl = _phiNLO->constructCollinearLimit(phiN.configCM, phiN.emitter, xBarA, xBarB, xRad, phiRadColl, jacColl);
         std::pair<bool, bool> removeColl = _phiNLO->cutoffFKS(phiRadColl);
         std::pair<double, double> xiYboundsColl = _phiNLO->xiYCollinearLimits(phiN.configCM, phiN.emitter, xBarA, xBarB, xRad);
         double yBoundColl = xiYboundsColl.second;
         Sigma coll;
         if (!removeColl.first and didSplitColl && _phiNLO->ThetaijRijColl()) {
            if (phiRadColl.xi() > FKSKinematics().getXiMaxColl(phiN, xBarA, xBarB)) {
               Log::error("EngineNLOpp::getIntegralRealSubtracted") << "xi " << phiRadColl.xi() << " xiMaxC_fks " << FKSKinematics().getXiMaxColl(phiN, xBarA, xBarB) << Log::end;
            }
            const double z = 1. - phiRadColl.xi() * Q / 2. / phiN.emitter->energy();
            if (!(phiN.emitter->isFinal() &&  phiRadColl.idLeft().isGluon() && phiRadColl.idRight().isGluon() && z < 0.5)) {
               PdgID idA, idB;
               double xAcoll, xBcoll, fluxColl, jacTmp;
               double wC0 = _subtraction->getColl(phiN, phiRadColl, xBarA, xBarB, idA, xAcoll, idB, xBcoll, fluxColl, jacTmp);
               jacColl *= jacTmp;
               double lumiColl = _pdf->luminosity(idA, xAcoll, idB, xBcoll, mufBorn);
               double collc =  wC0 * lumiColl * couplStrongBorn * alphasBorn;
               if (not std::isfinite(collc)) {
                  Log::error("EngineNLOpp::getIntegralRealSubtracted") << "getColl " << coll << " on \n" <<  phiN.configCM <<  Log::end;
               }
               std::vector<double> variations;
               if (doVariations) {
                  variations = getScaleVariations(phiN.configCM, idA, xAcoll, idB, xBcoll, couplStrongBorn * alphasBorn, asExpBorn + 1, lumiColl, mufBorn, murBorn, _subtraction->getScale(), wC0, 0.0, 0.0);
               }
               coll = Sigma(collc, variations) * couplNonStrongBorn * jacColl * fluxColl * static_cast<double>(m->multN);
# if defined CHECK_COLL
               if (didSplit && didSplitColl) {
                  Log::debug() << *m << Log::end;
                  Log::debug() << "Tiny Cutoff: " << _xRadGenerator->x2TinyCutoff() << Log::end;
                  Log::debug() << "Emitted CM: " << emitted->p() << Log::end;
                  Log::debug() << "PhiRadFks real : " << FKSKinematics().projectPhiRad(PhiNplus1(configNp1_CM, emitted, sister)) << Log::end;
                  Log::debug() << "PhiRadFks coll : " << phiRadColl << Log::end;
                  Log::debug() << "Jac Real : " << jacReal << Log::end;
                  double tmp;
                  if (phiN.emitter->isFinal()) {
                     double q0 = config_CM.initialA()->energy() + config_CM.initialB()->energy();
                     double kNBar = phiN.emitter->p().pvec().magnitude();
                     tmp = q0 * q0 * phiRadColl.xi() * (kNBar - phiRadColl.xi() * q0 / 2.) / kNBar * math_constants::_1_64pi3();
                  } else {
                     tmp = Q * Q * phiRadColl.xi() / (1. - phiRadColl.xi()) / (1. - phiRadColl.xi()) * math_constants::_1_64pi3();
                  }
                  Log::debug() << "Jac Coll : " << jacColl* tmp << Log::end;
                  Log::debug() << "Real : " << real << Log::end;
                  Log::debug() << "Coll : " << coll << Log::end;
                  if (compareToZero(real / coll - 1.0, 0.1) != 0) {
                     Log::debug() << "Press enter to continue" << Log::end;
                     std::cin.get();
                  }
               }
# endif
            }
         }

         // Add the collinear contributions to the counter-event weight. The yAnalyticFactorColl takes into account the potential different limits on y for the collinear
         //This also takes care of replacing the real with the subtraction in some singular regions
         double yAnalyticFactorColl = std::numeric_limits< double >::quiet_NaN();
         if (phiN.emitter->isInitialA()) {
            double ycut = 1. - _subtraction->yCutISR();
            double thetaY = ((!removeColl.second) and phiRadColl.y() < YCUTDIV) ? 1. : 0.;
            yAnalyticFactorColl = (thetaY + (1. - xRad.x2()) * log((1. - ycut) / (1. - std::min(yBoundColl, YCUTDIV))));
         } else if (phiN.emitter->isInitialB()) {
            double ycut = -1. + _subtraction->yCutISR();
            double thetaY = ((!removeColl.second) and phiRadColl.y() > -YCUTDIV) ? 1. : 0.;
            yAnalyticFactorColl = (thetaY + (1. - xRad.x2()) * log((1. + ycut) / (1. + std::max(yBoundColl, -YCUTDIV))));
         } else if (phiN.emitter->isFinal()) {
            double ycut = 1. - _subtraction->yCutFSR();
            double thetaY = ((!removeColl.second) and phiRadColl.y() < YCUTDIV) ? 1. : 0.;
            yAnalyticFactorColl = (thetaY + (1. - xRad.x2()) * log((1. - ycut) / (1. - std::min(yBoundColl, YCUTDIV))));
         } else {
            Log::fatal("EngineNLOpp::getIntegralRealSubtracted") << " wrong emitter!" << Log::end;
         }
         bool thetaXiColl = (phiRadColl.xi() > XICUTDIV);
         if (coll.isDefined() && thetaXiColl) {
            if (counterEvent.isDefined())
               counterEvent += -yAnalyticFactorColl * coll;
            else
               counterEvent = -yAnalyticFactorColl * coll;
         }
#endif

#ifdef INCLUDE_SOFTCOLL
         Sigma softColl;
         if (m->i->isGluon()) {
            // Get the soft-collinear contribution
            Log::debug("Engine softcoll") << Log::end;
            double jacSoftColl = 0.;
            PhiRadFKS phiRadSoftColl;
            _phiNLO->constructSoftCollinearLimit(phiN.configCM, phiN.emitter, xBarA, xBarB, xRad, phiRadSoftColl, jacSoftColl);
            std::pair<bool, bool> removeSoftColl = _phiNLO->cutoffFKS(phiRadSoftColl);
            std::pair<double, double> xiYboundsSoftColl = _phiNLO->xiYSoftCollinearLimits(phiN.configCM, phiN.emitter, xBarA, xBarB, xRad);
            const double xiMaxSoftColl = xiYboundsSoftColl.first;
            double yBoundSoftColl = xiYboundsSoftColl.second;
            if (_phiNLO->ThetaijRijSoftColl()) {
               double jacTmp = 0.;
               double wSC0 = _subtraction->getSoftColl(phiN, phiRadSoftColl, jacTmp);
               jacSoftColl *= jacTmp;
               double softcollc = wSC0 * lumiBorn * couplStrongBorn * alphasBorn;
               if (not std::isfinite(softcollc)) {
                  Log::error("EngineNLOpp::getIntegralRealSubtracted") << "getSoftColl " << softcollc << " on \n" <<  phiN.configCM << Log::end;
               }
               std::vector<double> variations;
               if (doVariations) {
                  variations = getScaleVariations(phiN.configCM, phiN.configCM.initialA()->id(), xBarA, phiN.configCM.initialB()->id(), xBarB, couplStrongBorn * alphasBorn, asExpBorn + 1, lumiBorn, mufBorn, murBorn, _subtraction->getScale(), wSC0, 0.0, 0.0);
               }
               softColl = Sigma(softcollc, variations) * couplNonStrongBorn * jacSoftColl * fluxBorn * static_cast<double>(m->multN);
# if defined CHECK_SOFTCOLL
               if (didSplit) {
                  Log::debug() << *m << Log::end;
                  Log::debug() << "Tiny Cutoff: " << _xRadGenerator->x1TinyCutoff() << Log::end;
                  Log::debug() << "Emitted CM: " << emitted->p() << Log::end;
                  Log::debug() << "PhiRadFks real : " << FKSKinematics().projectPhiRad(PhiNplus1(configNp1_CM, emitted, sister)) << Log::end;
                  Log::debug() << "PhiRadFks softcoll : " << phiRadSoftColl << Log::end;
                  Log::debug() << "jacReal : " << jacReal << Log::end;
                  Log::debug() << "jacCollSoft : " << jacSoftColl* Q* Q* phiRadSoftColl.xi() * math_constants::_1_64pi3() << Log::end;
                  Log::debug() << "Real : " << real << Log::end;
                  Log::debug() << "CollSoft : " << softcoll << Log::end;
                  if (compareToZero(real / softcoll - 1.0, 0.1) != 0) {
                     Log::debug() << "Press enter to continue" << Log::end;
                     std::cin.get();
                  }
               };
# endif
            }
            // Add the soft-collinear contributions to the counter-event weight. The yAnalyticFactorSoftColl and xiAnalyticFactorSoftColl take into account the potential different limits on xi and y for the soft-collinear
            //This also takes care of replacing the real with the subtraction in some singular regions
            double thetaXiSoftColl = ((!removeSoftColl.first) and phiRadSoftColl.xi() > XICUTDIV) ? 1. : 0.;
            double xiAnalyticSoftColl = (thetaXiSoftColl + log(_subtraction->xiCut() / std::max(xiMaxSoftColl, XICUTDIV)) * xRad.x1());
            double yAnalyticSoftColl = std::numeric_limits< double >::quiet_NaN();
            if (phiN.emitter->isInitialA()) {
               double ycut = 1. - _subtraction->yCutISR();
               double thetaY = ((!removeSoftColl.second) and phiRadSoftColl.y() < YCUTDIV) ? 1. : 0.;
               yAnalyticSoftColl = (thetaY + (1. - xRad.x2()) * log((1. - ycut) / (1. - std::min(yBoundSoftColl, YCUTDIV))));
            } else if (phiN.emitter->isInitialB()) {
               double ycut = -1. + _subtraction->yCutISR();
               double thetaY = ((!removeSoftColl.second) and phiRadSoftColl.y() > -YCUTDIV) ? 1. : 0.;
               yAnalyticSoftColl = (thetaY + (1. - xRad.x2()) * log((1. + ycut) / (1. + std::max(yBoundSoftColl, -YCUTDIV))));
            } else if (phiN.emitter->isFinal()) {
               double ycut = 1. - _subtraction->yCutFSR();
               double thetaY = ((!removeSoftColl.second) and phiRadSoftColl.y() < YCUTDIV) ? 1. : 0.;
               yAnalyticSoftColl = (thetaY + (1. - xRad.x2()) * log((1. - ycut) / (1. - std::min(yBoundSoftColl, YCUTDIV))));
            } else {
               Log::fatal("EngineNLOpp::getIntegralRealSubtracted") << " wrong emitter!" << Log::end;
            }
            if (softColl.isDefined()) {
               if (counterEvent.isDefined())
                  counterEvent += softColl * xiAnalyticSoftColl * yAnalyticSoftColl;
               else
                  counterEvent = softColl * xiAnalyticSoftColl * yAnalyticSoftColl;
            }
         }
#endif
         event *= xRad.weight();
         counterEvent *= xRad.weight();
         if (_doAnalyze && std::isnormal(_lastGenWeight)) {
            _analyzer->analyzeCorrelated(Event(configNp1_CM.transformed(BoostZHyp(sqrt(xAreal/xBreal))), event  * math_constants::gev2pb() / (_lastGenWeight * static_cast<double>(_numPointsReal))));
            _analyzer->analyzeCorrelated(Event(phiN.configCM.transformed(BoostZHyp(sqrt(xBarA / xBarB))), counterEvent  * math_constants::gev2pb() / (_lastGenWeight * static_cast<double>(_numPointsReal))));
         }
         if (event.isDefined()) {
            if (counterEvent.isDefined())
               result += event + counterEvent;
            else
               result += event;
         } else if (counterEvent.isDefined())
            result += counterEvent;
      }
   }
   result /= static_cast<double>(_numPointsReal);
   if (not std::isfinite(result.central())) {  //Make sure once again that we don't get nan's
      Log::error("EngineNLOpp::getIntegralRealSubtracted") << "result=" << result << " in " << _calcname <<  Log::end;
   }
   return result;
}

//------------------------------------------------------------------------------
void EngineNLOpp::generateUnitCubeRad(const std::list<RegionInfo>::iterator map, UnitCubeRad& xRad) const
{
   const double rand1 = randomInterval();
   const double rand2 = randomInterval();
   const double rand3 = randomInterval();

#if defined CHECK_SOFT or defined CHECK_COLL or defined CHECK_SOFTCOLL
   double smallRandom = 1e-6;
#endif

#ifdef CHECK_SOFT
   Log::debug().turnOn();
   Log::debug() << "\n------------------------\n" << Log::end;
   rand1 = (map->hasInitialStateSingularity()) ? smallRandom :  1. - smallRandom;
#endif

#ifdef CHECK_COLL
   Log::debug().turnOn();
   Log::debug() << "\n------------------------\n" << Log::end;
   rand2 = smallRandom;
#endif

#ifdef CHECK_SOFTCOLL
   Log::debug().turnOn();
   Log::debug() << "\n------------------------\n" << Log::end;
   rand1 = (map->hasInitialStateSingularity()) ? smallRandom : 1. - smallRandom;
   rand2 = smallRandom;
#endif

   xRad.idMother() = *map->k;
   xRad.idEmitted() = *map->i;
   xRad.idSister() = *map->j;
   if (map->hasFinalStateSingularity()) {
      xRad.radType() = UnitCubeRad::kFSR;
      assert(xRad.idEmitted().isParticle());
      assert(xRad.idEmitted().isGluon() || xRad.idSister().isAntiPartner(xRad.idEmitted()));
   } else if (map->hasInitialStateSingularity()) {
      xRad.radType() = (map->k->isInitialA()) ? UnitCubeRad::kISRA : UnitCubeRad::kISRB;
   }
   _xRadGenerator->generate(rand1, rand2, rand3, xRad);
   return;
}

#define COLLREMN_TINY_CUTOFF 1e-6
//------------------------------------------------------------------------------
/**
 * Calculates the integral over the collinear remnants.
 *
 * TODO: this should be separated into two different components A and B,
 *       in order to be able to deal with the e p collision scenario
 *
 * NOTE: COLLREMN_TINY_CUTOFF is a small cutoff parameter to avoid integrating
 *       down to the singular limits
 *
 */
//------------------------------------------------------------------------------
Sigma EngineNLOpp::getIntegralCollinearRemnants(const Configuration& config_CM, const double xBarA, const double xBarB, const double fluxBorn, const double mufBorn, const double murBorn, const double lumiBorn, const double alphasBorn, const int asExpBorn, const double couplStrongBorn, bool doVariations) const
{
   const double sBorn = (config_CM.initialA()->p() + config_CM.initialB()->p()).square();
   const double mufc = (*_factScale)(config_CM);
   const double omxPlus = (1. - xBarA);
   const double omxMinus = (1. - xBarB);
   const double RemnantPlus = log(omxPlus / _subtraction->xiCut()) * log((sBorn * _subtraction->yCutISR()) / (2.0 * pow(_subtraction->getScale(), 2))) + (pow(log(omxPlus), 2) - pow(log(_subtraction->xiCut()), 2));
   const double RemnantMinus = log(omxMinus / _subtraction->xiCut()) * log((sBorn * _subtraction->yCutISR()) / (2.0 * pow(_subtraction->getScale(), 2))) + (pow(log(omxMinus), 2) - pow(log(_subtraction->xiCut()), 2));
   const double RemnantPlusMuF = -log(omxPlus / _subtraction->xiCut());
   const double RemnantMinusMuF = -log(omxMinus / _subtraction->xiCut());
   const double fABornPlus = _pdf->pdfA(config_CM.initialA()->id(), xBarA, mufc);
   const double fBBornMinus = _pdf->pdfB(config_CM.initialB()->id(), xBarB, mufc);
   Sigma result(0., doVariations ? _numScaleVariations : 0);
   double murc = (*_renScale)(config_CM);
   for (size_t n = 0; n < _numPointsCollRemn; n++) {
      double ran = randomInterval();
      double x, jac;
      if (_impSampleCollremn) {
         /// Importance sampling at 0 and 1
         x = pow(ran, 5) * (70.*pow(ran, 4) - 315.*pow(ran, 3) + 540. * pow(ran, 2) - 420.*ran + 126.);
         jac = 630.* pow((1. - ran) * ran, 4);
      } else {
         /// No importance sampling
         x = ran;
         jac = 1.;
      }
      assert(ran >= 0 && ran <= 1.0);
      assert(x >= 0.0 && x <= 1.0);
      assert(omxPlus >= 0.0 && omxPlus <= 1.0);
      assert(omxMinus >= 0.0 && omxMinus <= 1.0);

      double zPlus = 1. - omxPlus * COLLREMN_TINY_CUTOFF - omxPlus * (1. - COLLREMN_TINY_CUTOFF) * x;
      double jacPlus = jac *  omxPlus;
      double zMinus = 1. - omxMinus * COLLREMN_TINY_CUTOFF - omxMinus * (1. - COLLREMN_TINY_CUTOFF) * x;
      double jacMinus = jac * omxMinus;

      double xRealPlus = xBarA / zPlus;
      assert(xRealPlus >= 0.0 && xRealPlus <= 1.0);
      double xRealMinus = xBarB / zMinus;
      assert(xRealMinus >= 0.0 && xRealMinus <= 1.0);
      Sigma CRPlus = calculateCollinearRemnant(config_CM, config_CM.initialA(), xBarA, xBarB, sBorn, mufc, murc, jacPlus, RemnantPlus, RemnantPlusMuF, zPlus, xRealPlus, fABornPlus, fBBornMinus, alphasBorn, asExpBorn, couplStrongBorn, doVariations);
      Sigma CRMinus = calculateCollinearRemnant(config_CM, config_CM.initialB(), xBarB, xBarA, sBorn, mufc, murc, jacMinus, RemnantMinus, RemnantMinusMuF, zMinus, xRealMinus, fBBornMinus, fABornPlus, alphasBorn, asExpBorn, couplStrongBorn, doVariations);
      result += (CRPlus + CRMinus) * _treeMEComplete->value() * math_constants::_1_2pi();
   }
   result /= static_cast<double>(_numPointsCollRemn);
   return result;
}

//------------------------------------------------------------------------------
Sigma EngineNLOpp::calculateCollinearRemnant(const Configuration& config, const Configuration::iterator initial, const double xBar, const double xBarOpposite, const double sBorn, const double mufc, const double murc, const double jac, const double remn, const double remnMuF, const double z, const double x, const double pdfcBorn, const double pdfcBornOpposite, const double alphasBorn, const int asExpBorn, const double couplStrongBorn, bool doVariations) const
{
   Sigma result(0., doVariations ? _numScaleVariations : 0);
   assert(initial->isInitialA() or initial->isInitialB());
   PdgID id = initial->id();
   PdgID idOpposite = initial->isInitialA() ? config.initialB()->id() : config.initialA()->id();
   double pdfc = initial->isInitialA() ? _pdf->pdfA(id, x, mufc) : _pdf->pdfB(id, x, mufc);
   if (id.isGluon()) {
      // g -> (g) g
      std::pair<double, double> coeffs1 = _subtraction->ggCollinearRemnant(sBorn, z);
      double wCR0gg = coeffs1.first * jac;
      double wCRFgg = coeffs1.second * jac;
      std::pair<double, double> coeffs2 = _subtraction->ggSoftCollinearRemnant(sBorn, z);
      double wSCR0gg = coeffs2.first * jac + QCD::OneMinusZPgg(1.0) * remn;
      double wSCRFgg = coeffs2.second * jac + QCD::OneMinusZPgg(1.0) * remnMuF;
      double centralCRgg = couplStrongBorn * alphasBorn * pdfc * pdfcBornOpposite * (wCR0gg + 2. * wCRFgg * log(mufc / _subtraction->getScale()));
      std::vector<double> varyCRgg;
      if (doVariations) {
         varyCRgg = getScaleVariations(config, PdgID::kGluon, x, idOpposite, xBarOpposite, couplStrongBorn * alphasBorn, asExpBorn + 1, pdfc * pdfcBornOpposite, mufc, murc, _subtraction->getScale(), wCR0gg, wCRFgg, 0.0);
      }
      result += Sigma(centralCRgg, varyCRgg);

      double centralSCRgg = couplStrongBorn * alphasBorn * pdfcBorn * pdfcBornOpposite * (wSCR0gg + 2. * wSCRFgg * log(mufc / _subtraction->getScale()));
      std::vector<double> varySCRgg;
      if (doVariations) {
         varySCRgg = getScaleVariations(config, PdgID::kGluon, xBar, idOpposite, xBarOpposite, couplStrongBorn * alphasBorn, asExpBorn + 1, pdfcBorn * pdfcBornOpposite, mufc, murc, _subtraction->getScale(), wSCR0gg, wSCRFgg, 0.0);
      }
      result += Sigma(centralSCRgg, varySCRgg);

      // q -> (g) q
      std::pair<double, double> coeffs3 = _subtraction->qgCollinearRemnant(sBorn, z);
      double wCR0qg = coeffs3.first *  jac;
      double wCRFqg = coeffs3.second * jac;
      for (int iid = -6; iid <= 6; iid++) {
         if (abs(iid) <= _subtraction->nLight() && iid != 0) {
            PdgID idTmp(iid);
            double tmpPdf = 0.0;
            if (initial->isInitialA()) {
               tmpPdf = _pdf->pdfA(idTmp, x, mufc);
            } else {
               tmpPdf = _pdf->pdfB(idTmp, x, mufc);
            }
            double centralCRqg = couplStrongBorn * alphasBorn * tmpPdf * pdfcBornOpposite * (wCR0qg + 2. * wCRFqg * log(mufc / _subtraction->getScale()));
            std::vector<double> varyCRqg;
            if (doVariations) {
               varyCRqg = getScaleVariations(config, idTmp, x, idOpposite, xBarOpposite, couplStrongBorn * alphasBorn, asExpBorn + 1, tmpPdf * pdfcBornOpposite, mufc, murc, _subtraction->getScale(), wCR0qg, wCRFqg, 0.0);
            }
            result += Sigma(centralCRqg, varyCRqg);
         }
      }
   } else {
      // q -> (q) g
      std::pair<double, double> coeffs1 = _subtraction->qqCollinearRemnant(sBorn, z);
      double wCR0qq = coeffs1.first * jac;
      double wCRFqq = coeffs1.second * jac;
      std::pair<double, double> coeffs2 = _subtraction->qqSoftCollinearRemnant(sBorn, z);
      double wSCR0qq = coeffs2.first * jac + QCD::OneMinusZPqq(1.0) * remn;
      double wSCRFqq = coeffs2.second * jac + QCD::OneMinusZPqq(1.0) * remnMuF;
      double centralCRqq = couplStrongBorn * alphasBorn * pdfc * pdfcBornOpposite * (wCR0qq + 2. * wCRFqq * log(mufc / _subtraction->getScale()));
      std::vector<double> varyCRqq;
      if (doVariations) {
         varyCRqq = getScaleVariations(config, id, x, idOpposite, xBarOpposite, couplStrongBorn * alphasBorn, asExpBorn + 1, pdfc * pdfcBornOpposite, mufc, murc, _subtraction->getScale(), wCR0qq, wCRFqq, 0.0);
      }
      result += Sigma(centralCRqq, varyCRqq);

      double centralSCRqq = couplStrongBorn * alphasBorn * pdfcBorn * pdfcBornOpposite * (wSCR0qq + 2. * wSCRFqq * log(mufc / _subtraction->getScale()));
      std::vector<double> varySCRqq;
      if (doVariations) {
         varySCRqq = getScaleVariations(config, id, xBar, idOpposite, xBarOpposite, couplStrongBorn * alphasBorn, asExpBorn + 1, pdfcBorn * pdfcBornOpposite, mufc, murc, _subtraction->getScale(), wSCR0qq, wSCRFqq, 0.0);
      }
      result += Sigma(centralSCRqq, varySCRqq);

      // g -> (q) q~
      double pdfGluon = initial->isInitialA() ? _pdf->pdfA(PdgID::kGluon, x, mufc) : _pdf->pdfB(PdgID::kGluon, x, mufc);
      std::pair<double, double> coeffs3 = _subtraction->gqCollinearRemnant(sBorn, z);
      double wCR0gq = coeffs3.first *  jac;
      double wCRFgq = coeffs3.second * jac;

      double centralCRgq = couplStrongBorn * alphasBorn * pdfGluon * pdfcBornOpposite * (wCR0gq + 2. * wCRFgq * log(mufc / _subtraction->getScale()));
      std::vector<double> varyCRgq;
      if (doVariations) {
         varyCRgq = getScaleVariations(config, PdgID::kGluon, x, idOpposite, xBarOpposite, couplStrongBorn * alphasBorn, asExpBorn + 1, pdfGluon * pdfcBornOpposite, mufc, murc, _subtraction->getScale(), wCR0gq, wCRFgq, 0.0);
      }
      result += Sigma(centralCRgq, varyCRgq);
   }
   return result;
}

} // namespace Geneva
