//------------------------------------------------------------------------------
/// \file PhiNLOppFKSBase.cpp
//
// Author(s):
//   Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementations of PhiNLOppFKSBase classes
//------------------------------------------------------------------------------

#include "Geneva/Calculations/PhiNLOppFKSBase.hpp"

#include "Geneva/Core/Physics/PhiRadFKS.hpp"
#include "Geneva/Core/Physics/UnitCubeRad.hpp"
#include "Geneva/Core/Physics/ConfigurationUtils.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include <cassert>

namespace Geneva
{

//------------------------------------------------------------------------------
bool PhiNLOppFKSBase::constructPhiNplus1(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, Configuration& configNplus1CM, Configuration::iterator& emittedCM, Configuration::iterator& sisterCM, double& xA, double& xB, PhiRadFKS& phiRad, double& jacobian)
{
   double y = -1. + 2.*xRad.x2();
#ifndef NEW
   if (xRad.radType() == UnitCubeRad::kISRB) y *= -1;
#endif
   double ximax = _mapFKS.getXiMaxReal(PhiNcmFKS(configNCM, emitterCM), xBarA, xBarB, y);
   phiRad = PhiRadFKS(xRad, ximax, jacobian);
   PhiNplus1FKS phiNplus1FKS;
   // test if the splitting is valid and get the split config
   if (_mapFKS.generateSplit(PhiNcmFKS(configNCM, emitterCM), phiRad, xBarA, xBarB, phiNplus1FKS)) {
      // the FKS map returns a N+1 Configuration in its partonic CM frame
      //NOTE: in order to be consistent with ThetaijSoft, ThetaijRij  must be evaluated in CM frame
      if (ThetaijRij(PhiNplus1(phiNplus1FKS.config, phiNplus1FKS.FKSparton, phiNplus1FKS.sister))) {
         jacobian *= _mapFKS.getJacobian();
         if (emitterCM->isFinal()) {
            xA = xBarA;
            xB = xBarB;
         } else {
            xA = _mapFKS.getxA(xBarA, phiRad.xi(), phiRad.y());
            xB = _mapFKS.getxB(xBarB, phiRad.xi(), phiRad.y());
         }
         configNplus1CM.moveFrom(phiNplus1FKS.config);
         emittedCM = phiNplus1FKS.FKSparton;
         sisterCM = phiNplus1FKS.sister;
         return true;
      }
      xA = std::numeric_limits<double>::quiet_NaN();
      xB = std::numeric_limits<double>::quiet_NaN();
      configNplus1CM.clear();
      emittedCM = configNplus1CM.end();
      sisterCM = configNplus1CM.end();
      jacobian = std::numeric_limits<double>::quiet_NaN();
      return false;
   }
   // generateSplit should never fail for the FKS map
   assert(false);
   return false;
}

//------------------------------------------------------------------------------
double PhiNLOppFKSBase::maxSoftEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad)  const
{
   double y = -1. + 2.*xRad.x2();
#ifndef NEW
   if (xRad.radType() == UnitCubeRad::kISRB) y *= -1;
#endif
   ///NOTE: since we have use the physical xiMax for the real, the soft limit of ximax corresponds with ximax itself
   double ximax = _mapFKS.getXiMaxReal(PhiNcmFKS(configNCM, emitterCM), xBarA, xBarB, y);
   return 0.5*ximax*sqrt(xBarA*xBarB)*_cmEnergy;
}

//------------------------------------------------------------------------------
std::pair<double, double> PhiNLOppFKSBase::maxCollEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad)  const
{
   assert(false);
   double ximaxColl = _mapFKS.getXiMaxColl(PhiNcmFKS(configNCM, emitterCM), xBarA, xBarB);
   double partonicS=xBarA*xBarB/(1.-ximaxColl)*_cmEnergy*_cmEnergy;
   return std::make_pair(0.5*ximaxColl*sqrt(partonicS), partonicS);
}

//------------------------------------------------------------------------------
double PhiNLOppFKSBase::maxSoftCollEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad)  const
{
   double ximaxSoftColl = _mapFKS.getXiMaxColl(PhiNcmFKS(configNCM, emitterCM), xBarA, xBarB);
   return 0.5*ximaxSoftColl*sqrt(xBarA*xBarB)*_cmEnergy;
}

//------------------------------------------------------------------------------
bool PhiNLOppFKSBase::constructSoftLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadSoft, double& jacobian) const
{
   double y = -1. + 2.*xRad.x2();
#ifndef NEW
   if (xRad.radType() == UnitCubeRad::kISRB) y *= -1;
#endif
   double xiMax = _mapFKS.getXiMaxReal(PhiNcmFKS(configNCM, emitterCM), xBarA, xBarB, y);
   phiRadSoft = PhiRadFKS(xRad, xiMax, jacobian);
   return true;
}

//------------------------------------------------------------------------------
bool PhiNLOppFKSBase::constructCollinearLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadColl, double& jacobian) const
{
   double xiMax = _mapFKS.getXiMaxColl(PhiNcmFKS(configNCM, emitterCM), xBarA, xBarB);
   phiRadColl = PhiRadFKS(xRad, xiMax, jacobian);
   return true;
}

//------------------------------------------------------------------------------
bool PhiNLOppFKSBase::constructSoftCollinearLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadSoftColl, double& jacobian) const
{
   double xiMax = _mapFKS.getXiMaxColl(PhiNcmFKS(configNCM, emitterCM), xBarA, xBarB);
   phiRadSoftColl = PhiRadFKS(xRad, xiMax, jacobian);
   return true;
}

//------------------------------------------------------------------------------
std::pair<double, double> PhiNLOppFKSBase::xiYCollinearLimits(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const
{
   double xiMax = _mapFKS.getXiMaxColl(PhiNcmFKS(configNCM, emitterCM), xBarA, xBarB);
   double yLimit = emitterCM->isInitialB() ? 1.0 :  -1.0;
   return std::pair<double, double>(xiMax, yLimit);
}

//------------------------------------------------------------------------------
std::pair<double, double> PhiNLOppFKSBase::xiYSoftCollinearLimits(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const
{
   double xiMax = _mapFKS.getXiMaxColl(PhiNcmFKS(configNCM, emitterCM), xBarA, xBarB);
   double yLimit = emitterCM->isInitialB() ? 1.0 :  -1.0;
   return std::pair<double, double>(xiMax, yLimit);
}

//------------------------------------------------------------------------------
std::pair<bool, bool> PhiNLOppFKSBase::cutoffFKS(const PhiRadFKS& phiRad) const
{
   bool cutXi = phiRad.xi() < _fksTinyCut;
   bool cutY;
   if (phiRad.radType() == PhiRadFKS::kISRB)
      cutY = (1. + phiRad.y()) < _fksTinyCut;
   else
      cutY = (1. - phiRad.y()) < _fksTinyCut;
   return std::make_pair(cutXi, cutY);
}

} // namespace Geneva
