//------------------------------------------------------------------------------
/// \file PhiNLOppVFKS.cpp
//
// Author(s):
//   Simone Alioli, Chris Vermilion, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of PhiNLOppVFKS class
//------------------------------------------------------------------------------

#include "Geneva/Calculations/PhiNLOppVFKS.hpp"

#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include <cassert>
#include <algorithm>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief Checks if a Configuration is N-body and part of \f$ \Phi_N \f$.
 * \param phiN The Configuration to check.
 * \return \c true if \a phiN is N-body and part of $\f$ \Phi_N \f$, \c false
 *         otherwise.
 *
 * Returns \c true if and only if \c phiN is an N-body configuration and is
 * considered to be part of \f$ \Phi_N \f$.  For V+0/1 jet, this simply tests
 * that there are no QCD final-state particles.
 */
bool PhiNLOppVFKS::isPhiN(const Configuration& configN) const
{
   return std::count_if(configN.beginFinal(), configN.endFinal(), IsQCD()) == 0;
}

//------------------------------------------------------------------------------
/**
 * \brief Checks if a Configuration is (N+1)-body and part of \f$ \Phi_{N+1} \f$.
 * \param phiNp1 The Configuration to check.
 * \return \c true if \a phiNp1 is (N+1)-body and part of $\f$ \Phi_{N+1} \f$,
 *         \c false otherwise.
 *
 * Returns \c true if and only if \c phiNp1 is an (N+1)-body configuration and
 * is considered to be part of \f$ \Phi_{N+1} \f$, i.e. it contributes to the
 * (N+1)-jet cross section and is outside the (IR-singular) region of
 * \f$ \Phi_{N+1} \f$ that is included in either the N-jet or (N-1)-jet cross
 * sections.
 */
bool PhiNLOppVFKS::isPhiNplus1(const Configuration& configNp1) const
{
   NJettinessppRecursive TauN(configNp1);

   // not an (N+1)-body state
   if (TauN.numQCD() != 1)
      return false;

   Configuration::iterator emitted;
   Configuration::iterator sister;
   const double min = TauN.cluster(emitted, sister);

   // final state quarks collinear to initial state quarks of the same flavour are not singular
   // since the boson must be emitted from the quark line
   if (sister->id() == emitted->id())
      return true;

   // sanity check
   const double Tau0 = TauN(0);
   if (compare(min, Tau0) != 0) {
      Log::debug("PhiNLOppVFKS::isPhiNplus1") << "min metric/Tau0 - 1 " << min/Tau0 - 1 << Log::end;
   };
   return Tau0 > _tau0Cut;
}

//------------------------------------------------------------------------------
/**
 * Returns the number of jets the given config is considered to have, which can
 * be either 0 or 1 for configurations with 0 or 1 QCD final-state parton. For
 * anything else an exception is thrown.
 */
size_t PhiNLOppVFKS::countN(const Configuration& config) const
{
   ptrdiff_t numQCDFinal = std::count_if(config.beginFinal(), config.endFinal(), IsQCD());
   switch (numQCDFinal) {
      case 0:
         return 0;
      case 1:
         if (isPhiNplus1(config))
            return 1;
         else
            return 0;
      default:
         Log::error("PhiNLOppVFKS::countN") << "Invalid or unknown number of jets" << Log::end;
         return size_t_from_int(-1);
   }
}

//------------------------------------------------------------------------------
const std::string PhiNLOppVFKS::name() const
{
   return "PhiNLOppVFKS";
}

//------------------------------------------------------------------------------
bool PhiNLOppVFKS::ThetaijRij(PhiNplus1 phiNp1CM) const
{
   assert(IsQCDSplitting()(*phiNp1CM.emitted(), *phiNp1CM.sister()));
   NJettinessppRecursive TauN(phiNp1CM);
   assert(TauN.numQCD() == 1);

   // Rij
   if (ZeroJettinesspp(phiNp1CM, TauN.y())() >= _tau0Cut)
      return false;

   // Thetaij
   Configuration::iterator emitted, sister;
   TauN.cluster(emitted, sister);
   return phiNp1CM.emitted() == emitted && phiNp1CM.sister() == sister;
}

//------------------------------------------------------------------------------
bool PhiNLOppVFKS::ThetaijRijSoft(const Configuration& configNCM, const Configuration::iterator emitter, const FourVector& pSoft) const
{
   NJettinessppRecursive TauN(configNCM);
   assert(TauN.numQCD() == 0);
   Configuration::iterator it;
   TauN.clusterSoft(it, pSoft);
   return emitter == it;
}

} // namespace Geneva
