//------------------------------------------------------------------------------
/// \file PhiNLOppVjBase.cpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementations of PhiNLOppVjBase class
//------------------------------------------------------------------------------

#include "Geneva/Calculations/PhiNLOppVjBase.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"

#include "Geneva/Components/FKSKinematics.hpp"

#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/PhiRadFKS.hpp"
#include "Geneva/Core/UI/Log.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief Checks if a Configuration is N-body and part of \f$ \Phi_N \f$.
 * \param phiN The Configuration to check.
 * \return \c true if \a phiN is N-body and part of \f$ \Phi_N \f$, \c false
 *         otherwise.
 *
 * Returns \c true if and only if \c phiN is an N-body configuration and is
 * considered to be part of \f$ \Phi_N \f$, i.e. it contributes to the N-jet
 * cross section and is outside the (IR-singular) region of \f$ \Phi_N \f$
 * that is included in the (N-1)-jet cross section.
 */
bool PhiNLOppVjBase::isPhiN(const Configuration& configN) const
{
   // not an N-body state
   if (std::count_if(configN.beginFinal(), configN.endFinal(), IsQCD()) != 1)
      return false;

   // Tau0 is longitudinal-boost-invariant, so we can safely calculate it on the configuration in the Lab Frame.
   const double Tau0 = ZeroJettinesspp(configN)();

   return Tau0 > _tau0Cut;
}

//------------------------------------------------------------------------------
/**
 * \brief Checks if a Configuration is (N+1)-body and part of \f$ \Phi_{N+1} \f$.
 * \param phiNp1 The Configuration to check.
 * \return \c true if \a phiNp1 is (N+1)-body and part of \f$ \Phi_{N+1} \f$,
 *         \c false otherwise.
 *
 * Returns \c true if and only if \c phiNp1 is an (N+1)-body configuration and
 * is considered to be part of \f$ \Phi_{N+1} \f$, i.e. it contributes to the
 * (N+1)-jet cross section and is outside the (IR-singular) region of
 * \f$ \Phi_{N+1} \f$ that is included in either the N-jet or (N-1)-jet cross
 * sections.
 */
bool PhiNLOppVjBase::isPhiNplus1(const Configuration& configNp1) const
{
   ZeroJettinesspp Tau0(configNp1);

   // not an (N+1)-body state
   if (Tau0.numQCD() != 2)
      return false;

   // Configurations that fail the Tau0 cut are not considered here, since
   // they would contribute to NNLO
   if (Tau0() < _tau0Cut)
      return false;

   // Once we get rid of the additional boost to CM we can avoid this copy
   const double xA = 2. * (configNp1.initialA()->energy() / _cmEnergy);
   const double xB = 2. * (configNp1.initialB()->energy() / _cmEnergy);
   const Configuration configNp1CM = configNp1.transformed(BoostZHyp(sqrt(xB/xA)));

   PhiNplus1 phiNp1CM(configNp1CM);
   NJettinessppRecursive TauN(phiNp1CM);
   TauN.cluster(phiNp1CM.emitted(), phiNp1CM.sister());

   // Configurations that can't be mapped back to an underlying Born cannot be generated through the splitting
   if (invalidFlavorProjection(_bornSignatures, phiNp1CM))
      return true;

   bool thetaTau1AboveCut = TauN(1) > _tau1Cut;

   bool thetaBarMap = projectionOutsideMapBoundaries(phiNp1CM, xA, xB);

   PhiRadFKS phiRadFKS = FKSKinematics().projectPhiRad(phiNp1CM);
   std::pair<bool, bool> cutReal = cutoffFKS(phiRadFKS);

   if ((cutReal.first || cutReal.second) && (thetaBarMap || thetaTau1AboveCut)) {
      Log::warning(name()+"::isPhiNplus1") << "NLO1/LO2 boundary is determined by FKS cutoff and not by Tau1 value." << Log::end;
      return false;
   }

   return thetaBarMap || thetaTau1AboveCut;
}

//------------------------------------------------------------------------------
/**
 * Returns the number of jets the given config is considered to have, which can
 * be either 0, 1, or 2 for configurations with 1 or 2 QCD final-state partons.
 * For other configurations an exception is thrown.
 */
size_t PhiNLOppVjBase::countN(const Configuration& config) const
{
   ptrdiff_t numQCDFinal = std::count_if(config.beginFinal(), config.endFinal(), IsQCD());
   if (!(numQCDFinal == 1 || numQCDFinal == 2))
      Log::error("PhiNLOppVjBase::countN") << "Invalid or unknown number of jets." << Log::end;

   // if it is part of PhiN+1 we are done
   if (numQCDFinal == 2 && isPhiNplus1(config))
      return 2;

   const double Tau0 = ZeroJettinesspp(config)();
   if (Tau0 > _tau0Cut)
      return 1;
   return 0;
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
bool PhiNLOppVjBase::Rij(const Configuration& configNp1CM) const
{
   NJettinessppRecursive TauN(configNp1CM);
   return TauN(1) < _tau1Cut && ZeroJettinesspp(configNp1CM, TauN.y())() >= _tau0Cut;
}

//------------------------------------------------------------------------------
bool PhiNLOppVjBase::Thetaij(PhiNplus1 phiNp1CM) const
{
   assert(IsQCDSplitting()(*phiNp1CM.emitted(), *phiNp1CM.sister()));
   NJettinessppRecursive TauN(phiNp1CM);
   Configuration::iterator emitted, sister;
   TauN.cluster(emitted, sister);
   return phiNp1CM.emitted() == emitted && phiNp1CM.sister() == sister;
}

//------------------------------------------------------------------------------
bool PhiNLOppVjBase::ThetaijSoft(const Configuration& configNCM, const Configuration::iterator emitter, const FourVector& pSoft) const
{
   NJettinessppRecursive TauN(configNCM);
   assert(TauN.numQCD() == 1);
   Configuration::iterator it;
   TauN.clusterSoft(it, pSoft);
   return emitter == it;
}

} // namespace Geneva
