//------------------------------------------------------------------------------
/// \file PhiNLOppVjTau0FR.cpp
//
// Author(s):
//   Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementations of PhiNLOppVjTau0FR classes
//------------------------------------------------------------------------------

#include "Geneva/Calculations/PhiNLOppVjTau0FR.hpp"

#include "Geneva/Components/FKSKinematics.hpp"

#include "Geneva/Core/Physics/PhiRadFKS.hpp"
#include "Geneva/Core/Physics/UnitCubeRad.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/Physics/ConfigurationUtils.hpp"
#include "Geneva/Core/UI/Log.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
bool PhiNLOppVjTau0FR::constructPhiNplus1(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad,
                                          Configuration& configNplus1CM, Configuration::iterator& emittedCM, Configuration::iterator& sisterCM,
                                          double& xA, double& xB, PhiRadFKS& phiRadFks, double& jacobian)
{
   double y = -1. + 2.*xRad.x2();
#ifndef NEW
   if (xRad.radType() == UnitCubeRad::kISRB)
      y = -y;
#endif
   PhiNCM phiNCM(configNCM, emitterCM, xBarA, xBarB);
   const double xiUbound = _mapTau0FR.getXiMax(phiNCM, y);
   const PhiRadTau0FR phiRad(xRad, xiUbound, jacobian);
   PhiNplus1CMWritable phiNp1CM;
   if (_mapTau0FR.generateSplit(phiNCM, phiRad, phiNp1CM)) {
#ifndef NDEBUG
      if (not _mapTau0FR.testProject(phiNp1CM, phiNp1CM.xA(), phiNp1CM.xB())) {
         Log::fatal("PhiNLOppVjTau0FR::constructPhiNplus1") << "The projection after the splitting fails" << Log::end;
      }
#endif
      // the Tau0FR map returns a N+1 Configuration in its partonic CM frame
      //NOTE: in order to be consistent with ThetaijSoft, ThetaijRij  must be evaluated in CM frame
      if (Thetaij(phiNp1CM) && Rij(phiNp1CM.config())) {
         phiRadFks = FKSKinematics().projectPhiRad(phiNp1CM);
         // NOTE: since we do not know the maximum xiMaxFks value that would correspond to xiMaxFR
         // (we only have an upper limit) we need to make sure we are never getting too close to it to avoid spikes.
         Configuration configCMfks;
         PhiNcmFKS phiNFKS(configCMfks);
         if (phiNp1CM.sister()->isFinal())
            phiNFKS.emitter = FKSKinematics().projectConfigFSR(phiNp1CM, configCMfks);
         else {
            phiNFKS.emitter = FKSKinematics().projectConfigISR(phiNp1CM, configCMfks);
         }
         double xBarAfks = FKSKinematics().getxBarA(phiNp1CM.xA(), phiRadFks.xi(), phiRadFks.y());
         double xBarBfks = FKSKinematics().getxBarB(phiNp1CM.xB(), phiRadFks.xi(), phiRadFks.y());
         double xiMaxFks = FKSKinematics().getXiMax(phiNFKS, xBarAfks, xBarBfks, phiRadFks.y());
         if (compare(phiRadFks.xi(), xiMaxFks) == -1) {
            jacobian *= phiNp1CM.jacobian();
            emittedCM = phiNp1CM.emitted();
            sisterCM = phiNp1CM.sister();
            xA = phiNp1CM.xA();
            xB = phiNp1CM.xB();
            configNplus1CM.moveFrom(phiNp1CM);
            return true;
         }
      }
   }
   xA = std::numeric_limits<double>::quiet_NaN();
   xB = std::numeric_limits<double>::quiet_NaN();
   configNplus1CM.clear();
   emittedCM = configNplus1CM.end();
   sisterCM = configNplus1CM.end();
   jacobian = std::numeric_limits<double>::quiet_NaN();
   return false;
}

//------------------------------------------------------------------------------
const std::string PhiNLOppVjTau0FR::name() const
{
   return "PhiNLOppVjTau0FR";
}

//------------------------------------------------------------------------------
bool PhiNLOppVjTau0FR::constructSoftLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadSoftFks, double& jacobian) const
{
   double y = -1 + 2.*xRad.x2();
#ifndef NEW
   if (xRad.radType() == UnitCubeRad::kISRB)
      y = -y;
#endif
   const double xiUbound = _mapTau0FR.getXiMax(PhiNCM(configNCM, emitterCM, xBarA, xBarB), y);
   const PhiRadTau0FR phiRadSoft(xRad, xiUbound, jacobian);
//    Log::debug("soft limit ") << "PhiRadTau0FR " << phiRadSoft << Log::end;
   double Q = _cmEnergy * sqrt(xBarA * xBarB);
   BoostZHyp fromSoftCmToLab(sqrt(xBarA/xBarB));
   FourVector emitterLab = fromSoftCmToLab(emitterCM->p());
   FourVector softCm;
   if (xRad.radType() == UnitCubeRad::kFSR) {
      const ThreeVector emitterDir = Normalize()(emitterLab.pvec());
      //Now determine the three vector direction
      const double y =  phiRadSoft.y();
      ///NOTE: here (and in Tau0FRmapping) we have selected the value of phi respect to the x-axis, while the FKS map assumes phi=0 to be the y-axis
      ThreeVector softDir(sqrt(1. - y * y)*cos(phiRadSoft.phi()), sqrt(1. - y * y)*sin(phiRadSoft.phi()), y);
      const ThreeVector zDir(0, 0, 1);
      const Rotation alignZToEmitter(zDir, emitterDir);
      double emittedEnergyLab = phiRadSoft.xi() * _cmEnergy / 2.;
      FourVector softLab(emittedEnergyLab, emittedEnergyLab * emitterLab.pvec().magnitude() / emitterLab.p0() * alignZToEmitter(softDir));
      Log::debug("soft limit FSR") << "soft Lab Frame " << softLab << Log::end;
      Log::debug("soft limit FSR") << "xBarA Lab " << xBarA << Log::end;
      Log::debug("soft limit FSR") << "xBarB Lab " << xBarB << Log::end;
      Log::debug("soft limit FSR") << "beta CmToLab " << (xBarA - xBarB) / (xBarA + xBarB) << Log::end;
      Log::debug("soft limit FSR") << "gamma CmToLab " << 1. / sqrt(1. - pow((xBarA - xBarB) / (xBarA + xBarB), 2)) << Log::end;
      Log::debug("soft limit FSR") << "soft CM Frame " << softLab.transformed(fromSoftCmToLab.inverse()) << Log::end;

      softCm = softLab.transformed(fromSoftCmToLab.inverse());
      const double xiSoftFks = 2.*softCm.p0() / Q;
      const double ySoftFks = (softCm.pvec() * emitterCM->pvec()) / (softCm.pvec().magnitude() * emitterCM->pvec().magnitude());
      ThreeVector k = Normalize()(emitterCM->p().pvec());
      ThreeVector fksDir = Normalize() (softCm.pvec());
      const double phiSoftFks = FKSKinematics::getPhiAngleFSR(fksDir, k);
      phiRadSoftFks = PhiRadFKS(xiSoftFks, ySoftFks, phiSoftFks, xRad.idMother(), xRad.idSister(), xRad.idEmitted(), PhiRadFKS::kFSR);
   } else {
      double Eg = phiRadSoft.xi() * _cmEnergy / 2.;
      double y = phiRadSoft.y();
      double phi = phiRadSoft.phi();
      ///NOTE: here (and in Tau0FRmapping) we have selected the value of phi respect to the x-axis, while the FKS map assumes phi=0 to be the y-axis
      ThreeVector ps(sqrt(1. - y * y)*cos(phi), sqrt(1. - y * y)*sin(phi), y);
      const FourVector softLab(Eg, Eg * ps);
      Log::debug("soft limit ISR") << "Q " << Q << Log::end;
      Log::debug("soft limit ISR") << "xi " << phiRadSoft.xi() << Log::end;
      Log::debug("soft limit ISR") << "xBarA Lab " << xBarA << Log::end;
      Log::debug("soft limit ISR") << "xBarB Lab " << xBarB << Log::end;
      Log::debug("soft limit ISR") << "soft Lab Frame " << softLab << Log::end;
      Log::debug("soft limit ISR") << "soft CM Frame " << softLab.transformed(fromSoftCmToLab.inverse()) << Log::end;
      softCm = softLab.transformed(fromSoftCmToLab.inverse());
      const double xiSoftFks = 2.*softCm.p0() / Q;
      const double ySoftFks = softCm.p3() / softCm.p0();
      // NOTE: phi02pi returns the value of phi respect to the x-axis, while the FKS map  assumes phi=0 to be the y-axis
      double phiYaxis = math_constants::pi_2() - softCm.phi02pi();
      if (phiYaxis < 0) phiYaxis = math_constants::_2pi() + phiYaxis;
      const double phiSoftFks = phiYaxis;
      phiRadSoftFks = PhiRadFKS(xiSoftFks, ySoftFks, phiSoftFks, xRad.idMother(), xRad.idSister(), xRad.idEmitted(),
                                emitterCM->isInitialA() ? PhiRadFKS::kISRA : PhiRadFKS::kISRB);
   }
   jacobian *=  phiRadSoft.xi() * _cmEnergy * _cmEnergy / (2. * Q * softCm.p0());
   return true;
}

//------------------------------------------------------------------------------
bool PhiNLOppVjTau0FR::constructCollinearLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadCollFKS, double& jacobian) const
{
   double xiBound =  _mapTau0FR.getXiMaxColl(PhiNCM(configNCM, emitterCM, xBarA, xBarB));

   double xiFKS, yFKS, phiFKS;
   PhiRadFKS::RadiationType radType;

   if (xRad.radType() == UnitCubeRad::kFSR) {
      double beta = -(xBarA - xBarB) / (xBarA + xBarB);
      double gamma = 1. / sqrt(1. - beta * beta);
      BoostZHyp fromCmToLab(sqrt(xBarA/xBarB));
      FourVector emitterLab = fromCmToLab(emitterCM->p());
      double cosThetaLab = emitterLab.p3() / emitterLab.p0();
      double xiFR = xiBound * xRad.x1();
      double yFR = -1. + 2. * xRad.x2();
      double phiFR = math_constants::_2pi() * xRad.x3();

      xiFKS = xiFR / sqrt(xBarA * xBarB) * gamma * (1. + beta * cosThetaLab);
      yFKS = 1. - (1. - yFR) * pow(gamma * (1. + beta * cosThetaLab), -2);

      //Finally, determine the phi value. Again, all formulas are expanded around the limit x2->1
      double E2lab = _cmEnergy / 2. * xiFR;
      double E1lab = emitterLab.p0() - E2lab;
      double y2Lab = 1. - 2. * (1. - xRad.x2()) * E1lab / emitterLab.p0();

      ThreeVector softDir(sqrt(1. - y2Lab * y2Lab)*cos(phiFR), sqrt(1. - y2Lab * y2Lab)*sin(phiFR), y2Lab);
      const ThreeVector emitterDir = Normalize()(emitterLab.pvec());
      const ThreeVector zDir(0, 0, 1);
      const Rotation alignZToEmitter(zDir, emitterDir);
      FourVector collLab(E2lab, E2lab * emitterLab.pvec().magnitude() / emitterLab.p0() * alignZToEmitter(softDir));
      FourVector collCm = collLab.transformed(fromCmToLab.inverse());
      Log::debug("constructCollinearLimit") << "pColl CM " << collCm << Log::end;
      ThreeVector k = Normalize()(emitterCM->p().pvec());
      ThreeVector fksDir = Normalize()(collCm.pvec());
      phiFKS = FKSKinematics::getPhiAngleFSR(fksDir, k);
      radType = PhiRadFKS::kFSR;
      jacobian = 4. * math_constants::pi() * xiBound / (sqrt(xBarA * xBarB) * gamma * (1. + beta * cosThetaLab));
   } else {
      double xA = xBarA;
      double xB = xBarB;
      if (xRad.radType() == UnitCubeRad::kISRA) {
         xA += xiBound * xRad.x1();
      } else if (xRad.radType() == UnitCubeRad::kISRB) {
         xB += xiBound * xRad.x1();
      } else {
         Log::fatal("PhiNLOppVjTau0FR::constructSoftAndCollinearFV") << "Radiation type can be either ISRA or ISRB at this point." << Log::end;
      }
      xiFKS = xiBound * xRad.x1();
      xiFKS /= (xRad.radType() == UnitCubeRad::kISRA) ? xA : xB;

#ifndef NEW
      yFKS = (xRad.radType() == UnitCubeRad::kISRA) ? 1. - 2.*(xA / xB) * (1. - xRad.x2()) : -1. +  2.*(xB / xA) * (1. - xRad.x2());
      ///NOTE: the FR map returns the value of phi respect to the x-axis, while the FKS map assumes phi=0 to be the y-axis
      double phiFR = math_constants::_2pi() * xRad.x3();
      double phiYaxis = math_constants::pi_2() - phiFR;
      phiFKS =  (phiYaxis < 0.) ? math_constants::_2pi() + phiYaxis : phiYaxis;

      if (xRad.radType() == UnitCubeRad::kISRA)
         jacobian = 4. * math_constants::pi() * xiBound * xBarA / xBarB / xA;
      else
         jacobian = 4. * math_constants::pi() * xiBound * xBarB / xBarA / xB;

      radType = (xRad.radType() == UnitCubeRad::kISRA) ? PhiRadFKS::kISRA : PhiRadFKS::kISRB;
#else
      double xA = xBarA;
      double xB = xBarB;
      if (xRad.radType() == UnitCubeRad::kISRA) {
         xA += xiBound * xRad.x1();
      } else if (xRad.radType() == UnitCubeRad::kISRB) {
         xB += xiBound * xRad.x1();
      } else {
         Log::fatal("PhiNLOppVjTau0FR::constructSoftAndCollinearFV") << "Radiation type can be either ISRA or ISRB at this point." << Log::end;
      }
      if (xRad.radType() == UnitCubeRad::kISRA) {
         xiFKS = xiBound * xRad.x1();
         xiFKS /= xA;
         jacobian = xiBound * xBarA;
         jacobian /= xA * xA;
      } else {
         xiFKS = xiBound * xRad.x1();
         xiFKS /= xB;

         jacobian = xiBound * xBarB;
         jacobian /= xB * xB;
      }

#ifndef NEW
      if (xRad.radType() == UnitCubeRad::kISRA) {
         yFKS = xRad.x2() - xA / xB * (1. - xRad.x2());
         yFKS /= xRad.x2() + xA / xB * (1. - xRad.x2());
         jacobian *= 2. * xA / xB;
         jacobian /= pow(xRad.x2() + xA / xB * (1. - xRad.x2()), 2);
      } else {
         yFKS = -1.;
         yFKS *= xRad.x2() - xB / xA * (1. - xRad.x2());
         yFKS /= xRad.x2() + xB / xA * (1. - xRad.x2());
         jacobian *= 2. * xB / xA;
         jacobian /= pow(xRad.x2() + xB / xA * (1. - xRad.x2()), 2);
      }
#else
      if (xRad.radType() == UnitCubeRad::kISRA) {
         yFKS = xRad.x2() - xA / xB * (1. - xRad.x2());
         yFKS /= xRad.x2() + xA / xB * (1. - xRad.x2());
         jacobian *= 2. * xA / xB;
         jacobian /= pow(xRad.x2() + xA / xB * (1. - xRad.x2()), 2);
      } else {
         yFKS = -1.;
         yFKS = (1. - xRad.x2()) - xB / xA * xRad.x2();
         yFKS /= (1. - xRad.x2()) + xB / xA * xRad.x2();
         jacobian *= 2. * xB / xA;
         jacobian /= pow((1. - xRad.x2()) + xB / xA * xRad.x2(), 2);

      }
#endif

      ///NOTE: the FR map returns the value of phi respect to the x-axis, while the FKS map assumes phi=0 to be the y-axis
      double phiYaxis = math_constants::pi_2() - phiFR;
      phiFKS =  (phiYaxis < 0.) ? math_constants::_2pi() + phiYaxis : phiYaxis;

      jacobian *= 2. * math_constants::pi();
      radType = (xRad.radType() == UnitCubeRad::kISRA) ? PhiRadFKS::kISRA : PhiRadFKS::kISRB;
#endif
   }

   phiRadCollFKS = PhiRadFKS(xiFKS, yFKS, phiFKS, xRad.idMother(), xRad.idSister(), xRad.idEmitted(), radType);
   ///NOTE: make sure we are never getting too close to the physical xiMaxColl limit to avoid spikes.
   double xiMaxColl = FKSKinematics().getXiMaxColl(PhiNcmFKS(configNCM, emitterCM), xBarA, xBarB);
   return (compare(xiFKS, xiMaxColl) == -1);
}

//------------------------------------------------------------------------------
bool PhiNLOppVjTau0FR::constructSoftCollinearLimit(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, PhiRadFKS& phiRadSoftCollFKS, double& jacobian) const
{
   double xiBound =  _mapTau0FR.getXiMaxColl(PhiNCM(configNCM, emitterCM, xBarA, xBarB));

   double xiFKS, yFKS, phiFKS;
   PhiRadFKS::RadiationType radType;

   if (xRad.radType() == UnitCubeRad::kFSR) {
      double beta = -(xBarA - xBarB) / (xBarA + xBarB);
      double gamma = 1. / sqrt(1. - beta * beta);
      BoostZHyp fromCmToLab(sqrt(xBarA/xBarB));
      FourVector emitterLab = fromCmToLab(emitterCM->p());
      double cosThetaLab = emitterLab.p3() / emitterLab.p0();
      double xiFR = xiBound * xRad.x1();
      double yFR = -1. + 2. * xRad.x2();
      double phiFR = math_constants::_2pi() * xRad.x3();

      xiFKS = xiFR / sqrt(xBarA * xBarB) * gamma * (1. + beta * cosThetaLab);
      yFKS = 1. - (1. - yFR) * pow(gamma * (1. + beta * cosThetaLab), -2);

      //Now construct the phiFKS angle
      const ThreeVector emitterDir = Normalize()(emitterLab.pvec());
      ///NOTE: here (and in Tau0FRmapping) we have selected the value of phi respect to the x-axis, while the FKS map assumes phi=0 to be the y-axis
      ThreeVector softDir(sqrt(1. - yFR * yFR)*cos(phiFR), sqrt(1. - yFR * yFR)*sin(phiFR), yFR);
      const ThreeVector zDir(0, 0, 1);
      const Rotation alignZToEmitter(zDir, emitterDir);
      double emittedEnergyLab = xiFR * _cmEnergy / 2.;
      FourVector softCollLab(emittedEnergyLab, emittedEnergyLab * emitterLab.pvec().magnitude() / emitterLab.p0() * alignZToEmitter(softDir));
      FourVector softCollCm = softCollLab.transformed(fromCmToLab.inverse());
      Log::debug("constructSoftCollinearLimit") << "pSoftColl CM " << softCollCm << Log::end;
      ThreeVector k = Normalize()(emitterCM->p().pvec());
      ThreeVector fksDir = Normalize()(softCollCm.pvec());
      phiFKS = FKSKinematics::getPhiAngleFSR(fksDir, k);
      radType = PhiRadFKS::kFSR;
      jacobian = 4. * math_constants::pi() * xiBound / (sqrt(xBarA * xBarB) * gamma * (1. + beta * cosThetaLab));
   } else {
      xiFKS = xiBound * xRad.x1();
      xiFKS /= (xRad.radType() == UnitCubeRad::kISRA) ? xBarA : xBarB;

#ifndef NEW
      yFKS = (xRad.radType() == UnitCubeRad::kISRA) ? 1. - 2.*(xBarA / xBarB) * (1. - xRad.x2()) : -1. +  2.*(xBarB / xBarA) * (1. - xRad.x2());
#else
      assert(false):
         yFKS = (xRad.radType() == UnitCubeRad::kISRA) ? 1. - 2.*(xBarA / xBarB) * (1. - xRad.x2()) : -1. +  2.*(xBarB / xBarA) * xRad.x2();
#endif
      ///NOTE: the FR map returns the value of phi respect to the x-axis, while the FKS map assumes phi=0 to be the y-axis
      double phiFR = math_constants::_2pi() * xRad.x3();
      double phiYaxis = math_constants::pi_2() - phiFR;
      phiFKS =  (phiYaxis < 0.) ? math_constants::_2pi() + phiYaxis : phiYaxis;

      if (xRad.radType() == UnitCubeRad::kISRA)
         jacobian = 4. * math_constants::pi() * xiBound / xBarB;
      else
         jacobian = 4. * math_constants::pi() * xiBound / xBarA;

      radType = (xRad.radType() == UnitCubeRad::kISRA) ? PhiRadFKS::kISRA : PhiRadFKS::kISRB;
   }

   phiRadSoftCollFKS = PhiRadFKS(xiFKS, yFKS, phiFKS, xRad.idMother(), xRad.idSister(), xRad.idEmitted(), radType);
   return true;
}

//------------------------------------------------------------------------------
std::pair<double, double> PhiNLOppVjTau0FR::xiYCollinearLimits(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const
{
   double xiBound =  _mapTau0FR.getXiMaxColl(PhiNCM(configNCM, emitterCM, xBarA, xBarB));

   double xiMax, yMin;
   if (xRad.radType() == UnitCubeRad::kFSR) {
      double beta = -(xBarA - xBarB) / (xBarA + xBarB);
      double gamma = 1. / sqrt(1. - beta * beta);
      BoostZHyp fromCmToLab(sqrt(xBarA/xBarB));
      FourVector emitterLab = fromCmToLab(emitterCM->p());
      double cosThetaLab = emitterLab.p3() / emitterLab.p0();

      xiMax = xiBound / sqrt(xBarA * xBarB) * gamma * (1. + beta * cosThetaLab);
      yMin = 1. - 2. * pow(gamma * (1. + beta * cosThetaLab), -2);

   } else {
      if (xRad.radType() == UnitCubeRad::kISRA) {
         xiMax = xiBound / (xBarA + xiBound);
         yMin = 1. - 2. * (xBarA + xiBound * xRad.x1()) / xBarB;
      } else {
         xiMax = xiBound / (xBarB + xiBound);
         yMin = -1. + 2. * (xBarB + xiBound * xRad.x1()) / xBarA;
      }
   }
   return std::pair<double, double>(xiMax, yMin);
}

//------------------------------------------------------------------------------
std::pair<double, double> PhiNLOppVjTau0FR::xiYSoftCollinearLimits(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad) const
{
   double xiBound = _mapTau0FR.getXiMaxColl(PhiNCM(configNCM, emitterCM, xBarA, xBarB));

   double xiMax, yMin;
   if (xRad.radType() == UnitCubeRad::kFSR) {
      double beta = -(xBarA - xBarB) / (xBarA + xBarB);
      double gamma = 1. / sqrt(1. - beta * beta);
      BoostZHyp fromCmToLab(sqrt(xBarA/xBarB));
      FourVector emitterLab = fromCmToLab(emitterCM->p());
      double cosThetaLab = emitterLab.p3() / emitterLab.p0();

      xiMax = xiBound / sqrt(xBarA * xBarB) * gamma * (1. + beta * cosThetaLab);
      yMin = 1. - 2. * pow(gamma * (1. + beta * cosThetaLab), -2);
   } else {
      if (xRad.radType() == UnitCubeRad::kISRA) {
         xiMax = xiBound / xBarA;
         yMin = 1. - 2. * xBarA / xBarB;
      } else {
         xiMax = xiBound / xBarB;
         yMin = -1. + 2. * xBarB / xBarA;
      }
   }
   return std::pair<double, double>(xiMax, yMin);
}

//------------------------------------------------------------------------------
double PhiNLOppVjTau0FR::maxSoftEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad)  const
{
   UnitCubeRad xRadMaxSoft(xRad);
   xRadMaxSoft.x1() = 1.;
   PhiRadFKS phiRad;
   double jac;
   constructSoftLimit(configNCM, emitterCM, xBarA, xBarB, xRadMaxSoft, phiRad, jac);
   return 0.5 * phiRad.xi() * _cmEnergy * sqrt(xBarA * xBarB);
}

//------------------------------------------------------------------------------
std::pair<double, double> PhiNLOppVjTau0FR::maxCollEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad)  const
{
   double xiUbound =  _mapTau0FR.getXiMaxColl(PhiNCM(configNCM, emitterCM, xBarA, xBarB));

   double xA = xBarA;
   double xB = xBarB;
   if (xRad.radType() == UnitCubeRad::kISRA) {
      xA += xiUbound;
   } else if (xRad.radType() == UnitCubeRad::kISRB) {
      xB += xiUbound;
   }
   BoostZHyp fromCmToLab(sqrt(xA/xB));
   FourVector pEmitter = emitterCM->p().transformed(fromCmToLab);
   FourVector pCollLab = xiUbound * _cmEnergy * 0.5 / pEmitter.p0() * pEmitter;
   double maxCollEmittedEnergy = pCollLab.transformed(fromCmToLab.inverse()).p0();
   if (xRad.radType() == UnitCubeRad::kISRA)
      assert(compare(maxCollEmittedEnergy, sqrt((xBarB) / (xBarA + xiUbound))*xiUbound * _cmEnergy * 0.5) == 0);
   if (xRad.radType() == UnitCubeRad::kISRB)
      assert(compare(maxCollEmittedEnergy, sqrt((xBarA) / (xBarB + xiUbound))*xiUbound * _cmEnergy * 0.5) == 0);
   double partonicScoll = xA * xB * _cmEnergy * _cmEnergy;
   return std::make_pair(maxCollEmittedEnergy, partonicScoll);
}

//------------------------------------------------------------------------------
double PhiNLOppVjTau0FR::maxSoftCollEmittedEnergyCm(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad)  const
{
   UnitCubeRad xRadMaxSoftColl(xRad);
   xRadMaxSoftColl.x1() = 1.;
   FourVector pMaxSoftCollCm;
   double jac;
   PhiRadFKS phiRad;
   bool split = constructSoftCollinearLimit(configNCM, emitterCM, xBarA, xBarB, xRadMaxSoftColl,  phiRad, jac);
   if (!split)
      Log::error(name() + "::maxSoftCollEmittedEnergyCm") << "constructSoftLimit false" << Log::end;
   return 0.5 * phiRad.xi() * _cmEnergy * sqrt(xBarA * xBarB);
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
bool PhiNLOppVjTau0FR::projectionOutsideMapBoundaries(PhiNplus1 phiNp1CM, double xA, double xB) const
{
   return !_mapTau0FR.testProject(phiNp1CM, xA, xB);
}

//------------------------------------------------------------------------------
std::pair<bool, bool> PhiNLOppVjTau0FR::cutoffFKS(const PhiRadFKS& phiRad) const
{
   bool cutXi = phiRad.xi() < _fksTinyCut;
   bool cutY;
   if (phiRad.radType() == PhiRadFKS::kISRB)
      cutY = (1. + phiRad.y()) < _fksTinyCut;
   else
      cutY = (1. - phiRad.y()) < _fksTinyCut;
   return std::make_pair(cutXi, cutY);
}

} // namespace Geneva
