//------------------------------------------------------------------------------
/// \file BeamFuncTauB.cpp
//
// Author(s):
//    Jon Walsh
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class BeamFuncTauB
//------------------------------------------------------------------------------

#include "Geneva/Components/BeamFuncTauB.hpp"

#include "Geneva/Core/UI/Log.hpp"

#include "beamfunc/TauCoeffsQuark.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
BeamFuncTauB::BeamFuncTauB(const std::string& pdfName, int pdfMember, int order, int nf)
{
   try {
      _quarkTauBcoeffs = new beamfunc::TauCoeffsQuark(order, pdfName, pdfMember, nf);
   } catch (std::exception& ex) {
      Log::error("BeamFuncTauB") << ex.what() << Log::end;
   }
   Log::verboseInfo("BeamFuncTauB") << "Initialized TauB quark beam function for " << pdfName << " member " << pdfMember << Log::end;

   if (nf != 5) {
      Log::warning("BeamFuncTauB") << "You have set nf != 5, but the LHAPDF grids for the convolutions currently only exist for nf = 5! You will get (hopefully mildly) wrong answers." << Log::end;
   }
}

//------------------------------------------------------------------------------
BeamFuncTauB::~BeamFuncTauB()
{
   delete _quarkTauBcoeffs;
}

//------------------------------------------------------------------------------
std::string BeamFuncTauB::name() const
{
   return "BeamFuncTauB";
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \param pdgID The PDG id for the flavor of the beam function
 * \param dist A code specifying the order of the distribution that the coefficient is to be returned for.
 * \param ord The order in perturbation theory, i.e. the term with \f$\alpha^{\text{ord}}\f$
 * \return The coefficient of \f$\alpha^{\text{ord}}\f$ multiplied by the distribution of order dist.
 * \todo Need to implement two loop corrections.
 */
double BeamFuncTauB::value(PdgID pdgID, int dist, int ord, double x, double muB)
{
   return _quarkTauBcoeffs->Biqj(ord, dist, pdgID.code(), x, muB);
}

} // namespace Geneva
