//------------------------------------------------------------------------------
/// \file ComponentFactory.cpp
//
// Author(s):
//    Chris Vermilion, Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2013 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class ComponentFactory
//------------------------------------------------------------------------------

#include "Geneva/config.hpp"

#include "Geneva/Components/ComponentFactory.hpp"
#include "Geneva/Components/FluxFactor.hpp"
#include "Geneva/Components/CouplingFixed.hpp"
#include "Geneva/Components/CouplingStrong.hpp"
#include "Geneva/Components/CouplingEW.hpp"
#include "Geneva/Components/Collider.hpp"
#include "Geneva/Components/Scales.hpp"
#include "Geneva/Components/IPDF.hpp"

#ifdef geneva_USE_LHAPDF
#include "Geneva/Components/LHAPDF.hpp"
#endif

#ifdef geneva_USE_OPENLOOPS
#include "Geneva/Components/TreeMEOpenLoops.hpp"
#include "Geneva/Components/TreeMECompleteOpenLoops.hpp"
#include "Geneva/Components/OneLoopMEOpenLoops.hpp"
#endif

#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
// private helper functions
// use anomymous namespace instead of private member functions to completely
// contain external dependence here in the implementation
namespace
{

#ifdef geneva_USE_LHAPDF
LHAPDF* getLHAPDF(const OptionMap& optMap);
#endif

} // anonymous namespace

//------------------------------------------------------------------------------
ITreeME* ComponentFactory::getTreeME(const OptionMap& optMap)
{
   const std::string name = optMap.getOption("TreeME");

   if (name == "OpenLoops")
#ifdef geneva_USE_OPENLOOPS
      return new TreeMEOpenLoops(optMap);
#else
   Log::fatal("ComponentFactory")
      << "OpenLoops support is disabled. Please compile with OpenLoops enabled to use TreeMEOpenLoops." << Log::end;
#endif

   Log::error("ComponentFactory::getTreeME") << "Unknown TreeME " << name << Log::end;
   return 0;
}

//------------------------------------------------------------------------------
ITreeMEComplete* ComponentFactory::getTreeMEComplete(const OptionMap& optMap)
{
   const std::string name = optMap.getOption("TreeMEComplete");

   if (name == "OpenLoops")
#ifdef geneva_USE_OPENLOOPS
      return new TreeMECompleteOpenLoops(optMap);
#else
      Log::fatal("ComponentFactory")
         << "OpenLoops support is disabled. Please compile with OpenLoops enabled to use TreeMECompleteOpenLoops." << Log::end;
#endif

   Log::error("ComponentFactory::getTreeMEComplete") << "Unknown TreeMEComplete " << name << Log::end;
   return 0;
}

//------------------------------------------------------------------------------
ITreeMEColor* ComponentFactory::getTreeMEColor(const OptionMap& optMap)
{
   const std::string name = optMap.getOption("TreeMEColor");

   if (name == "OpenLoops")
#ifdef geneva_USE_OPENLOOPS
      return new TreeMEOpenLoops(optMap);
#else
   Log::fatal("ComponentFactory")
      << "OpenLoops support is disabled. Please compile with OpenLoops enabled to use TreeMEOpenLoops." << Log::end;
#endif

   Log::error("ComponentFactory::getTreeMEComplete") << "Unknown TreeMEColor " << name << Log::end;
   return 0;
}

//------------------------------------------------------------------------------
IOneLoopME* ComponentFactory::getOneLoopME(const OptionMap& optMap)
{
   const std::string name = optMap.getOption("OneLoopME");

   if (name == "OpenLoops")
#ifdef geneva_USE_OPENLOOPS
      return new OneLoopMEOpenLoops(optMap);
#else
      Log::fatal("ComponentFactory")
         << "OpenLoops support is disabled. Please compile with OpenLoops enabled to use OneLoopMEOpenLoops." << Log::end;
#endif

   Log::error("ComponentFactory::getOneLoopME") << "Unknown OneLoopME " << name << Log::end;
   return 0;
}

//------------------------------------------------------------------------------
CouplingStrong* ComponentFactory::getCouplingStrong(const OptionMap& optMap)
{
   //Get the input values
   double value0;
   double mu0;

   if (optMap.getOption<bool>("alphas::valueAtmZFromPDF")) {
#ifdef geneva_USE_LHAPDF
      // this option assumes LHAPDF as the PDF provider
      LHAPDF* lhapdf = getLHAPDF(optMap);
      try {
         mu0 = lhapdf->getmZ();
      } catch (...)  {
         mu0 = optMap.getOption<double>("ZMass");
         Log::warning("ComponentFactory::getCouplingStrong") << "Cannot get mZ value from LHAPDF. Falling back to mu0 = " << mu0 << Log::end;
      }
      value0 = lhapdf->getAlphasQ(mu0);
      delete lhapdf;
#else
      Log::fatal("ComponentFactory::getCouplingStrong") << "No LHAPDF support. Please compile with LHAPDF enabled to obtain alphas(mZ) from PDF." << Log::end;
      return 0;
#endif
   } else {
      value0 = optMap.getOption<double>("alphas::value0");
      mu0 = optMap.getOption<double>("alphas::mu0");
   }

   //Construct a CouplingStrong from the input values
   const std::string opt = optMap.getOption("alphas::run");
   RunOrder run;
   if (opt == "none")      run = RunOrder::kNone;
   else if (opt == "LL")   run = RunOrder::kLL;
   else if (opt == "NLL")  run = RunOrder::kNLL;
   else if (opt == "NNLL") run = RunOrder::kNNLL;
   else
      Log::error("ComponentFactory::getCouplingStrong") << "Unknown alphas running order " << opt << Log::end;

   int nf = optMap.getOption<int>("NLightFlavors");
   return new CouplingStrong(value0, mu0, run, nf);
}

//------------------------------------------------------------------------------
ICoupling* ComponentFactory::getCouplingEM(const OptionMap& optMap)
{
   const double value0 = optMap.getOption<double>("alphaem::value0");
   const bool run = optMap.getOption<bool>("alphaem::run");
   if (run) {
      const double mu0 = optMap.getOption<double>("alphaem::mu0");
      return new CouplingEW(value0, mu0);
   } else {
      return new CouplingFixed(value0);
   }
}

//------------------------------------------------------------------------------
IPDF* ComponentFactory::getPDF(const OptionMap& optMap)
{
   const std::string name = optMap.getOption("PDF");
#ifdef geneva_USE_LHAPDF
   if (name == "LHAPDF")
      return getLHAPDF(optMap);
#else
   Log::fatal("ComponentFactory::getPDF") << "No LHAPDF support. Please compile with LHAPDF enabled to use it as PDF provider." << Log::end;
   return 0;
#endif

   Log::error("ComponentFactory::getPDF") << "Unknown PDF " << name << Log::end;
   return 0;
}

//------------------------------------------------------------------------------
IScale* ComponentFactory::getRenormalizationScale(const OptionMap& optMap)
{
   std::string name = optMap.getOption("RenormalizationScale::Type");
   double preFactor = optMap.getOption<double>("RenormalizationScale::PreFactor");

   if (name == "Fixed")
      return new ScaleFixed(optMap.getOption<double>("RenormalizationScale::Fixed::Value"), preFactor);
   if (name == "InvariantMass")
      return new ScaleInvariantMass<IsElectroWeak>(preFactor);
   if (name == "TransverseMass")
      return new ScaleTransverseMass<IsElectroWeak>(preFactor);
   if (name == "HT")
      return new ScaleHT<IsQCD>(preFactor);

   Log::error("ComponentFactory::getRenormalizationScale") << "Unknown Scale " << name << Log::end;
   return 0;
}

//------------------------------------------------------------------------------
IScale* ComponentFactory::getFactorizationScale(const OptionMap& optMap)
{
   std::string name = optMap.getOption("FactorizationScale::Type");
   double preFactor = optMap.getOption<double>("FactorizationScale::PreFactor");
   if (name == "Fixed")
      return new ScaleFixed(optMap.getOption<double>("FactorizationScale::Fixed::Value"), preFactor);
   if (name == "InvariantMass")
      return new ScaleInvariantMass<IsElectroWeak>(preFactor);
   if (name == "TransverseMass")
      return new ScaleTransverseMass<IsElectroWeak>(preFactor);
   if (name == "HT")
      return new ScaleHT<IsQCD>(preFactor);

   Log::error("ComponentFactory::getFactorizationScale") << "Unknown Scale " << name << Log::end;
   return 0;
}

//------------------------------------------------------------------------------
IPDF* ComponentFactory::makePDF(PdgID beamA, PdgID beamB, const std::string& name, int member, size_t maxStackLength)
{
#ifdef geneva_USE_LHAPDF
   return new LHAPDF(beamA, beamB, name, member, maxStackLength);
#else
   Log::fatal("ComponentFactory::makeLHAPDF") << "No LHAPDF support. Please compile with LHAPDF enabled to use it as PDF provider." << Log::end;
   return 0;
#endif
}

//------------------------------------------------------------------------------
// private helper functions
//------------------------------------------------------------------------------

namespace
{

#ifdef geneva_USE_LHAPDF
//------------------------------------------------------------------------------
LHAPDF* getLHAPDF(const OptionMap& optMap)
{
   const std::string pdfsetA = optMap.getOption("PDFSetA");
   const std::string pdfsetB = optMap.getOption("PDFSetB");
   const size_t cacheSize = size_t_from_int(optMap.getOption<int>("PDFCacheSize"));

   Collider collider(optMap);
   if (!collider.isHadronic())
      Log::error("ComponentFactory::getLHAPDF") << "LHAPDFs not supported for this collider." << Log::end;

   if (pdfsetA != pdfsetB) {
      // different sets for beamA and beamB
      const int memberA = optMap.getOption<int>("PDFMemberA");
      const int memberB = optMap.getOption<int>("PDFMemberB");
      return new LHAPDF(collider.beamA(), collider.beamB(), pdfsetA, pdfsetB, memberA, memberB, cacheSize);
   }

   // same set for both beams
   std::string pdfset;
   int member;
   if (optMap.isDefault("PDFSetA")) {
      pdfset = optMap.getOption("PDFSet");
      member = optMap.getOption<int>("PDFMember");
   } else {
      pdfset = pdfsetA;
      member = optMap.getOption<int>("PDFMemberA");
   }
   return new LHAPDF(collider.beamA(), collider.beamB(), pdfset, member, cacheSize);
}
#endif // geneva_USE_LHAPDF

} // anonymous namespace

} // namespace Geneva
