//------------------------------------------------------------------------------
/// \file CouplingEW.cpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class CouplingEW
//------------------------------------------------------------------------------

#include "Geneva/Components/CouplingEW.hpp"

#include "Geneva/Core/Utils/MathUtils.hpp"

#include <gsl/gsl_spline.h>

namespace Geneva
{

//------------------------------------------------------------------------------
struct CouplingEW::_spline_type : gsl_spline {};
struct CouplingEW::_interp_type : gsl_interp_accel {};

//------------------------------------------------------------------------------
CouplingEW::CouplingEW(double value0, double mu0)
   : _value0(value0), _mu0(mu0)
{
   const int nPoints = 3;
   _interp = static_cast<_interp_type*>(gsl_interp_accel_alloc());
   _spline = static_cast<_spline_type*>(gsl_spline_alloc(gsl_interp_cspline, nPoints));
   gsl_spline_init(_spline, _logSqrtScale, _oneOverAlphaEW, nPoints);
}

//------------------------------------------------------------------------------
CouplingEW::~CouplingEW()
{
   gsl_spline_free(_spline);
   gsl_interp_accel_free(_interp);
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the value of the beta function.
 *
 * \warning Currently this is not supported and simply returns 0.0;
 */
double CouplingEW::beta(int n) const
{
   return 0;
}

//------------------------------------------------------------------------------
std::string CouplingEW::name() const
{
   std::stringstream s;
   s << "CouplingEW. Using alpha_em(" << _mu0 << ")^-1 = " << 1/_value0;
   return s.str();
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Computes the value of the running EW coupling \f$ \alpha_em \f$.
 * \param mu Scale in GeV at which to evaluate the coupling
 * \return \f$ \alpha_em \f$ at the scale \a mu
 */
double CouplingEW::value(double mu) const
{
   if (mu < exp(_logSqrtScale[2]))
      return 1./_oneOverAlphaEW[2];
   if (mu < exp(_logSqrtScale[0])) {
      double logScale = log(mu);
      double result = 1./gsl_spline_eval(_spline, logScale, _interp);
      return result;
   }
   if (mu < _bMass) {
      double b = 3. + 2.*SU<3>::Nc*(1./9.) + 2.*SU<3>::Nc*(4./9.);
      double logX = 2.*(log(mu) - _logSqrtScale[0]);
      double result = (1./_oneOverAlphaEW[0])/(1. - (b*logX)/(3.*math_constants::pi()*_oneOverAlphaEW[0]));
      return result;
   } else {
      double b = 3. + 3.*SU<3>::Nc*(1./9.) + 2.*SU<3>::Nc*(4./9.);
      double logX = 2.*log(mu/_mu0);
      double result = _value0/(1. - (b*logX*_value0)/(3.*math_constants::pi()));
      return result;
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Computes the value of the running EW coupling \f$ \alpha_em \f$, taking
 * the real part of a complex scale. Does not currently implement complex EW coupling
 * as this is not needed.
 * \param Allows for complex mu Scale in GeV at which to evaluate the coupling
 * \return Returns \f$ \alpha_em \f$ at the scale \f$ Re(mu) \f$.
 *
 */
std::complex<double> CouplingEW::value(std::complex<double> mu) const
{
   return CouplingEW::value(real(mu));
}

//------------------------------------------------------------------------------
// 1/alpha_em at  m_e=0.000511, m_mu=0.1056,  m_tau=1.777
const double CouplingEW::_oneOverAlphaEW[] = {137.036, 135.95, 133.513};

// logs of sqrt(m^2) at m_e=0.000511, m_mu=0.1056,  m_tau=1.777
const double CouplingEW::_logSqrtScale[] = {-7.57914, -2.2481, 0.574927};

// hardcoded bottom quark mass threshold in GeV
const double CouplingEW::_bMass = 4.5;

} // namespace Geneva
