//------------------------------------------------------------------------------
/// \file CouplingStrong.cpp
//
// Author(s):
//    Christian Bauer, Chris Vermilion, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2013 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class CouplingStrong
//------------------------------------------------------------------------------

#include "Geneva/Components/CouplingStrong.hpp"

#include "Geneva/Core/Physics/QCD.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <limits>

namespace Geneva
{

//------------------------------------------------------------------------------
// Constructors etc.
CouplingStrong::CouplingStrong(double value0, double mu0, RunOrder run, int nf)
   : _value0(value0), _mu0(mu0), _run(run), _nf(nf)
{
   switch (_run) {
      case RunOrder::kNone:
         _lambdaQCD = std::numeric_limits< double >::quiet_NaN();
         break;
      case RunOrder::kLL: {
          double b = beta(0)*math_constants::_1_4pi();
          double t = 1.0/b/_value0;
          double tOld;
          size_t count = 0;
          do {
            count++;
            if (count > 10000)
               Log::error("CouplingStrong") << "Lambda QCD calculation doesn't converge at LL"<<Log::end;
            tOld = t;
            double as0 = 1.0/b/t;
            double as1 = -1.0/b/t/t;
            t += (_value0-as0)/as1;
          } while (abs(tOld-t)/tOld > 1e-8);
          _lambdaQCD = _mu0/exp(t/2.);
         }
          break;
      case RunOrder::kNLL: {
         double b = beta(0)*math_constants::_1_4pi();
         double bp = (153. - 19. * _nf) * math_constants::_1_2pi() / (3.*beta(0));
         double t = 1.0/b/_value0;
         double tOld;
         size_t count = 0;
         do {
            count++;
            if (count > 10000)
               Log::error("CouplingStrong") << "Lambda QCD calculation doesn't converge at NLL"<<Log::end;
            tOld = t;
            double xlt = log(t);
            double as0 = 1.0/b/t - bp * xlt* pow(b*t, -2);
            double as1 = -1.0/b/t/t - bp/b/b * (1.-2.*xlt)*pow(t,-3);
            t += (_value0-as0)/as1;
         } while (abs(tOld-t)/tOld > 1e-8);
         _lambdaQCD = _mu0/exp(t/2.);
         }
         break;
      case RunOrder::kNNLL: {
         double b0 = beta(0)*math_constants::_1_4pi();
         double b1 = (153. - 19. * _nf) / (24.*math_constants::pi2());
         double b2 = (2857./2 - 5033./18 * _nf + 325./54 *_nf*_nf)*math_constants::_1_64pi3();
         double t = 1.0/b0/_value0;
         double tOld;
         size_t count = 0;
         do {
            count++;
            if (count > 10000)
               Log::error("CouplingStrong") << "Lambda QCD calculation doesn't converge at NNLL"<<Log::end;
            tOld = t;
            double xlt = log(t);
            double as0 = 1.0/b0/t * (1. - b1/b0/b0 * xlt/t + pow(b1/b0/b0 * xlt/t, 2) - (b1*b1 * (1.+xlt) - b0*b2)*pow(b0,-4)/t/t);
            double as1 = (-2.*b1*b1*xlt*xlt*pow(b0,-4)*pow(t,-3) + 2.*(b1*b1*(1.+xlt)- b0*b2)*pow(b0,-4)*pow(t,-3) + b1*xlt/b0/b0/t/t + 2.*b1*b1*xlt*pow(b0,-4)*pow(t,-3)-b1/b0/b0/t/t - b1*b1*pow(b0,-4)*pow(t,-3))/b0/t - (b1*b1*xlt*xlt*pow(b0,-4)*pow(t,-2) - (b1*b1*(1.+xlt) - b0*b2)*pow(b0,-4)*pow(t,-2) -b1*xlt/b0/b0/t + 1)/b0/t/t;
            t += (_value0-as0)/as1;
         } while (abs(tOld-t)/tOld > 1e-8);
         _lambdaQCD = _mu0/exp(t/2.);
         }
         break;
      default:
         Log::error("CouplingStrong") << "Running coupling only implemented up to NNLL (3 loops running)" << Log::end;
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the value of the QCD beta function at (n+1)-loop order.
 *
 * The returned coefficients \f$\beta_n\f$ are defined according to the convention
 * \f$-\alpha_s \sum _{n=0}^{\infty} \beta _n (\frac{\alpha _s}{4\pi})^{n+1}\f$
 */
double CouplingStrong::beta(int n) const
{
   switch (n) {
      case 0:
         return QCD::beta0(_nf);
      case 1:
         return QCD::beta1(_nf);
      case 2:
         return QCD::beta2(_nf);
      default:
         Log::error("CouplingStrong::beta") << "Beta function only implemented up to 3 loops (beta_2)" << Log::end;
   }
   return 0.;
}

//------------------------------------------------------------------------------
std::string CouplingStrong::name() const
{
   std::stringstream s;
   s << "CouplingStrong. Using alphas(" << _mu0 << ") = " << _value0;
   s << " with ";
   switch (_run) {
      case RunOrder::kNone: s << "no running."; break;
      case RunOrder::kNNLL: s << "N";
      case RunOrder::kNLL:  s << "N";
      case RunOrder::kLL:   s << "LL running and nf = " << _nf;
   }
   return s.str();
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Computes the value of the running strong coupling \f$ \alpha_s \f$.
 * \param mu Scale in GeV at which to evaluate the coupling
 * \return \f$ \alpha_s \f$ at the scale \a mu
 *
 * The RGE is performed to the running order specified at construction time.
 * The RGE is not solved exactly, but is given expanded up to order
 * \f$ \alpha_s^{loop\ order+1} \f$.
 */
double CouplingStrong::value(double mu) const
{
   if (mu < _lambdaQCD)
      Log::error("CouplingStrong::value") << "mu below LQCD" << Log::end;

   const double b0 = QCD::beta0(_nf);
   const double b10 = QCD::beta1(_nf) / b0;
   const double b20 = QCD::beta2(_nf) / b0;
   const double r = 1. + _value0*math_constants::_1_2pi()*b0*log(mu/_mu0);
   switch (_run) {
      case RunOrder::kNone:
         return _value0;
      case RunOrder::kLL:
         return _value0 / r;
      case RunOrder::kNLL:
         return 1. / (r/_value0 + b10*math_constants::_1_4pi()*log(r));
      case RunOrder::kNNLL:
         return 1. / (r/_value0 + b10*math_constants::_1_4pi()*log(r)
                      + _value0*math_constants::_1_16pi2()*(b20*(1. - 1./r) + b10*b10*(log(r)/r + 1./r - 1.))
                     );
      default:
         Log::error("CouplingStrong::value") << "Running coupling only implemented up to NNLL (3 loops running)" << Log::end;
   }
   return 0.;
}

//------------------------------------------------------------------------------
/**
 * \brief Computes the value of the running strong coupling \f$ \alpha_s \f$, with a complex scale mu.
 * \param mu complex scale in GeV at which to evaluate the coupling
 * \return complex valued \f$ \alpha_s \f$ at the scale \a mu
 *
 * The RGE is performed to the running order specified at construction time.
 * The RGE is not solved exactly, but is given expanded up to order
 * \f$ \alpha_s^{loop\ order+1} \f$.
 */
std::complex<double> CouplingStrong::value(std::complex<double> mu) const
{
   const double b0 = QCD::beta0(_nf);
   const double b10 = QCD::beta1(_nf) / b0;
   const double b20 = QCD::beta2(_nf) / b0;
   const std::complex<double> r = 1. + _value0*math_constants::_1_2pi()*b0*log(mu/_mu0);
   switch (_run) {
      case RunOrder::kNone:
         return _value0;
      case RunOrder::kLL:
         return _value0 / r;
      case RunOrder::kNLL:
         return 1. / (r/_value0 + b10*math_constants::_1_4pi()*log(r));
      case RunOrder::kNNLL:
         return 1. / (r/_value0 + b10*math_constants::_1_4pi()*log(r)
                        + _value0*math_constants::_1_16pi2()*(b20*(1. - 1./r)
                        + b10*b10*(log(r)/r + 1./r - 1.))
                        );
      default:
         Log::error("CouplingStrong::value") << "Running coupling only implemented up to NNLL (3 loops running)" << Log::end;
   }
   return 0.;
}

} // namespace Geneva
