//------------------------------------------------------------------------------
/// \file FKSKinematics.cpp
//
// Author(s):
//     Simone Alioli
//
// Copyright:
//    Copyright (C) 2013 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class FKSKinematics
//------------------------------------------------------------------------------

#include "Geneva/Components/FKSKinematics.hpp"

#include "Geneva/Core/Physics/PhiRadFKS.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/Physics/ConfigurationUtils.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"
#include "Geneva/Core/Utils/NumericsUtils.hpp"

#include <cassert>
#include <iomanip>

#ifndef NDEBUG
#define FKSKINEMATICS_PRECISION 1e-8
#define FKSKINEMATICS_ACCURACY 1e-11
#endif

namespace Geneva
{

//------------------------------------------------------------------------------
double FKSKinematics::getxA(const double xBarA, const double xi, const double y) const
{
   double tmp = xBarA / sqrt(1. - xi) * sqrt(2. - xi * (1. - y)) / sqrt(2. - xi * (1. + y));
   forceInUnitRange(tmp, "FKSKinematics::getxA");
   return tmp;
}

//------------------------------------------------------------------------------
double FKSKinematics::getxB(const double xBarB, const double xi, const double y) const
{
   double tmp = xBarB / sqrt(1. - xi) * sqrt(2. - xi * (1. + y)) / sqrt(2. - xi * (1. - y));
   forceInUnitRange(tmp, "FKSKinematics::getxB");
   return tmp;
}


//------------------------------------------------------------------------------
double FKSKinematics::getxBarA(const double xA, const double xi, const double y) const
{
   double tmp = xA * sqrt(1. - xi) * sqrt(2. - xi * (1. + y)) / sqrt(2. - xi * (1. - y));
   forceInUnitRange(tmp, "FKSKinematics::getxBarA");
   return tmp;
}

//------------------------------------------------------------------------------
double FKSKinematics::getxBarB(const double xB, const double xi, const double y) const
{
   double tmp = xB * sqrt(1. - xi) * sqrt(2. - xi * (1. - y)) / sqrt(2. - xi * (1. + y));
   forceInUnitRange(tmp, "FKSKinematics::getxBarB");
   return tmp;
}

//------------------------------------------------------------------------------
double FKSKinematics::getXiMax(const PhiNcmFKS& phiN, double xBarA, double xBarB, double y) const
{
   if (phiN.emitter->isFinal()) {
      const double Q = phiN.configCM.initialA()->energy() + phiN.configCM.initialB()->energy();
      // take into account potential offshellness of phiN.emitter due to limited numerical accuracy
      return (2.*phiN.emitter->energy() - phiN.emitter->p().square()/Q)/Q;
//       return 2.*phiN.emitter->energy()/Q;
   }
   if (phiN.emitter->isInitialA() || phiN.emitter->isInitialB())
      return 1. - std::max((2.*(1. + y) * xBarA * xBarA) / (sqrt(pow((1. + xBarA * xBarA) * (1. - y), 2) + 16.*y * xBarA * xBarA) + (1. - y) * (1. - xBarA * xBarA)), (2.*(1. - y) * xBarB * xBarB) / (sqrt(pow((1. + xBarB * xBarB) * (1. + y), 2) - 16.*y * xBarB * xBarB) + (1. + y) * (1. - xBarB * xBarB)));

   Log::fatal("FKSKinematics::getXiMaxReal") << "Emitter must be either initial or final." << Log::end;
   return 0.;
}

//------------------------------------------------------------------------------
/**
 * Calculates the maximum xi value  for a collinear configuration.
 * This quantity does not depend on the Mapping
 */
double FKSKinematics::getXiMaxColl(const PhiNcmFKS& phiN, double xBarA, double xBarB) const
{
   if (phiN.emitter->isFinal()) {
      const double Q = phiN.configCM.initialA()->energy() + phiN.configCM.initialB()->energy();
      // take into account potential offshellness of phiN.emitter due to limited numerical accuracy
      return (2.*phiN.emitter->energy() - phiN.emitter->p().square()/Q)/Q;
//       return 2.*phiN.emitter->energy()/Q;
   }
   if (phiN.emitter->isInitialA())
      return 1. - xBarA; // collinear limit of _xiMax for y = 1
   if (phiN.emitter->isInitialB())
      return 1. - xBarB; // collinear limit of _xiMax for y = -1
   Log::fatal("FKSKinematics::getXiMaxColl") << "Emitter must be initial or final." << Log::end;
   return 0.;
}

//------------------------------------------------------------------------------
/**
 * Generates the split config for Initial State Radiation in the CM frame
 * The PhiNplus1 structure containing the splitted configuration,
 * the iterators to the FKS partons and its sister is taken as input and
 * filled internally.
 * \param phiN the N-body %Configuration and the %Particle to split
 * \param phiRad the kinematic and flavor variables in the splitting
 * \param hadS the hadronic center-of-mass energy squared
 * \param x1, \param x2 the partonic energy fraction wrt the beams
 * \param phiRadNplus1FKS the (N+1)-body splitted %Configuration, the iterators to the splitted particles
 * \returns  the jacobian of the splitting.
 */
double FKSKinematics::splitISR(const PhiNcmFKS& phiN, const PhiRadFKS& phiRad, double xBarA, double xBarB, PhiNplus1FKS& phiNp1)
{
   // sanity check this is only ever used for ISR
   assert((phiN.emitter->isInitialA() && phiRad.radType() == PhiRadFKS::kISRA)
          || (phiN.emitter->isInitialB() && phiRad.radType() == PhiRadFKS::kISRB));

   // sanity check that phiN is in the CM frame
   assert(isInCoMFrame(phiN.configCM, 1e-3, 1e-5));

   // here lambda = sqrt(xB/xA * xBarA/xBarB)
   const double lambda = sqrt((2. - phiRad.xi()*(1. + phiRad.y())) / (2. - phiRad.xi()*(1. - phiRad.y())));
   double jacobian = 1. / (1. - phiRad.xi());
   const double tmp = sqrt(jacobian);

   const double QBar = phiN.configCM.initialA()->energy() + phiN.configCM.initialB()->energy();
   const double Q = tmp * QBar;

   // construct emitted p2 in the partonic CM frame of N+1 body
   const double p0 = (0.5*Q) * phiRad.xi();
   const double pT = p0 * sqrt(1. - phiRad.y() * phiRad.y());
   const FourVector p(p0, pT * sin(phiRad.phi()), pT * cos(phiRad.phi()), p0 * phiRad.y());

   // Longitudinal boost from old PhiN CM to new PhiN=1 CM
   const BoostZHyp boostToNewCM(lambda);

   // Transverse boost to compensate the pT of the FKS parton
   ThreeVector beta(-p.p1(), -p.p2(), 0.);
   beta /= sqrt(p.perp2() + QBar*QBar);
   const Boost transverseBoost(beta);

   // add the recoiling final state particles
   phiNp1.config.clear();
   for (Configuration::iterator it = phiN.configCM.beginFinal(); it != phiN.configCM.endFinal(); ++it) {
      Particle finalParticle(*it);
      finalParticle.p().apply(transverseBoost).apply(boostToNewCM);
      phiNp1.config.insertFinal(phiNp1.config.end(), finalParticle);
   }

   const FourVector kA(0.5*Q, 0., 0., 0.5*Q);
   const FourVector kB(0.5*Q, 0., 0., -0.5*Q);

   // add initial particles
   if (phiRad.radType() == PhiRadFKS::kISRA) {
      phiNp1.sister = phiNp1.config.setInitialA(Particle(phiRad.idLeft(), kA, 0, 0));
      phiNp1.config.setInitialB(Particle(phiN.configCM.initialB()->id(), kB, 0, 0));
   } else {
      phiNp1.config.setInitialA(Particle(phiN.configCM.initialA()->id(), kA, 0, 0));
      phiNp1.sister = phiNp1.config.setInitialB(Particle(phiRad.idLeft(), kB, 0, 0));
   }

   // add the FKS parton
   phiNp1.FKSparton = phiNp1.config.insertFinal(Particle(phiRad.idRight(), p, 0, 0));

   // final accuracy checks
#ifndef NDEBUG
   if (!checkMomentumConservation(phiNp1.config, FKSKINEMATICS_PRECISION, FKSKINEMATICS_ACCURACY))
      Log::warning("FKSKinematics::splitISR") << "Phi_N+1 point from splitting violates momentum conservation." << Log::end;
   if (!checkOnShellConditions(phiNp1.config, FKSKINEMATICS_PRECISION, FKSKINEMATICS_ACCURACY))
      Log::warning("FKSKinematics::splitISR") << "Phi_N+1 point from splitting violates on-shell conditions." << Log::end;
#endif

   jacobian *= Q*Q * phiRad.xi() * math_constants::_1_64pi3();
   return jacobian;
}

//------------------------------------------------------------------------------
/**
 * Generates the split config for Final State Radiation in the CM frame.
 * The PhiNplus1 structure containing the splitted configuration,
 * the iterators to the FKS partons and its sister is taken as input
 * and filled internally.
 * \param phiN the N-body %Configuration and the %Particle to split
 * \param phiRadFKS the kinematic and flavor variables in the splitting
 * \param phiRadNplus1FKS the (N+1)-body splitted %Configuration, the iterators to the splitted particles
 * \returns the jacobian of the splitting.
 */
double FKSKinematics::splitFSR(const PhiNcmFKS& phiN, const PhiRadFKS& phiRad, PhiNplus1FKS& phiNp1)
{
   // sanity check this is only ever used for FSR
   assert(phiN.emitter->isFinal() && phiRad.radType() == PhiRadFKS::kFSR);

   // sanity check that phiN is in the CM frame
   assert(isInCoMFrame(phiN.configCM, 1e-3, 1e-5));

   // Q is the partonic Ecm
   const double Q = phiN.configCM.initialA()->energy() + phiN.configCM.initialB()->energy();
   const double kBar = phiN.emitter->pvec().magnitude();
   const ThreeVector n12 = phiN.emitter->pvec() * (1./kBar);

   // energies of emitted and sisters
   const double tmp12eps = 1. / (1. - 0.5*phiRad.xi()*(1. - phiRad.y()));
   const double k20 = 0.5 * phiRad.xi() * Q;
   // take into account potential offshellness of phiN.emitter due to limited numerical accuracy
//    const double k10 = (phiN.emitter->energy() - k20) * tmp12eps;
   const double k10 = (phiN.emitter->energy() - 0.5*phiN.emitter->p().square()/Q - k20) * tmp12eps;
   double jacobian = k10 * (1./kBar) * tmp12eps;

   // this is equivalent to xi > xiMax
   if (k10 < 0.0)
      Log::error("FKSKinematics::splitFSR") << "Too large xi = " << phiRad.xi() << " > xiMax gives k1^0 = " << k10 << " < 0" << Log::end;

   // construct k12 = k1 + k2, where \vec{k12} = |k12|/|kbar| \vec{kbar}
   const double k12sqExact = 2.*k10*k20*(1. - phiRad.y());
   const double k120 = k10 + k20;
   const FourVector k12(k120, sqrt(k120*k120 - k12sqExact) * n12);
   // recompute k12^2 and |vec{k12}| from k12
   const double k12sq = k12.square();
   const double k12vecmag = k12.pvec().magnitude();
   assert(k12sq >= 0.0);
   assert(k12vecmag > 0.0);
   if (k12sq < 0.0) {
      // warn if we get spacelike p12 (can only be due to numerical inaccuracy)
      Log::warning("FKSKinematics::splitFSR") << "p12^2 = " << k12sq << "is spacelike for xi = " << phiRad.xi() << ", 1-y = " << 1. - phiRad.y() << Log::end;
      return std::numeric_limits<double>::quiet_NaN();
   }
   if (compareRelative(k12sq, k12sqExact, 1e-4)) {
      Log::warning("FKSKinematics::splitFSR") << "Loss of numerical precision for xi = " << phiRad.xi() << ", 1-y = " << 1.-phiRad.y()
         << ": Constructed (k1+k2)^2 = " << k12sq << " differs from exact value " << k12sqExact << " by relative " << k12sq/k12sqExact - 1.
         << Log::end;
   }

   // construct k2
   const double tmp = k20 / k12vecmag;
   const double k2T = tmp * k10 * sqrt(1. - phiRad.y()*phiRad.y());
   const double k2L = tmp * (k10 * phiRad.y() + k20);

   // construct unit vector perpendicular to n12 with reference direction z
   // this is \vec{n12} x \hat{z} normalized
   double norm = n12.perp2();
   norm = (norm != 0.0) ? 1./sqrt(norm) : 1.0;
   const ThreeVector perpsin(norm * n12.p2(), - norm * n12.p1(), 0.);
   const ThreeVector perpcos = crossProduct(perpsin, n12);
   const ThreeVector perp = (perpsin * sin(phiRad.phi()) + perpcos * cos(phiRad.phi()));

   // construct full k1 and k2
   const FourVector k2(k20, k2T * perp + k2L * n12);
   const FourVector k1(k10, k12.pvec() - k2.pvec());

   // determine the recoil boost (have to be careful with numerical accuracy here)
   const double k12plus = k120 - k12vecmag;
   const double tmp2 = k12plus * (2.*Q - k12plus);
   const double beta = tmp2 / (2.*Q*Q - tmp2);
   const Boost boostrec(beta * n12);

   phiNp1.config.clear();

   // add initial particles - copy them in, they don't change
   phiNp1.config.setInitialA(*phiN.configCM.initialA());
   phiNp1.config.setInitialB(*phiN.configCM.initialB());

   // add final particles, first do the recoil ones
   // copy them in and then change their momenta
   for (Configuration::iterator it = phiN.configCM.beginFinal(); it != phiN.configCM.endFinal(); ++it) {
      if (it != phiN.emitter) {
         Particle finalParticle(*it);
         finalParticle.p().apply(boostrec);
         phiNp1.config.insertFinal(phiNp1.config.end(), finalParticle);
      }
   }

   phiNp1.sister    = phiNp1.config.insertFinal(Particle(phiRad.idLeft(), k1, 0, 0));
   phiNp1.FKSparton = phiNp1.config.insertFinal(Particle(phiRad.idRight(), k2, 0, 0));

   // final accuracy checks
#ifndef NDEBUG
   if (!checkMomentumConservation(phiNp1.config, FKSKINEMATICS_PRECISION, FKSKINEMATICS_ACCURACY))
      Log::warning("FKSKinematics::splitFSR") << "Phi_N+1 point from splitting violates momentum conservation." << Log::end;
   if (!checkOnShellConditions(phiNp1.config, FKSKINEMATICS_PRECISION, FKSKINEMATICS_ACCURACY))
      Log::warning("FKSKinematics::splitFSR") << "Phi_N+1 point from splitting violates on-shell conditions." << Log::end;
#endif

   jacobian *= Q*Q * phiRad.xi() * math_constants::_1_64pi3();
   return jacobian;
}

//------------------------------------------------------------------------------
/// Creates an N-body %Configuration from the N+1-body configuration by combining the two FKS partons. It assumes an ISR splitting has taken place.
Configuration::iterator FKSKinematics::projectConfigISR(PhiNplus1 phiNp1, Configuration& projectedconfig) const
{
   // sanity check that phiNp1 is in the CM
   assert(isInCoMFrame(phiNp1, 1e-3, 1e-5));

   // sanity check that this is only ever used for ISR
   assert(phiNp1.sister()->isInitialA() || phiNp1.sister()->isInitialB());

   FourVector ktot;
   for (Configuration::iterator it = phiNp1.beginFinal(); it != phiNp1.endFinal(); ++it) {
      if (it != phiNp1.emitted()) {
         ktot += it->p();
      }
   }

   // calulate xi and y
   const double s = (phiNp1.initialA()->p() + phiNp1.initialB()->p()).square();
   const double xi = 2.*(phiNp1.emitted()->energy()) / sqrt(s);
   const double y = phiNp1.emitted()->pz() / phiNp1.emitted()->energy();

   // longitudinal boost such that BL kTot has zero rapidty
   BoostZHyp longB(exp(-ktot.rapidity()));
   // transverse boost such that BT BL kTot has zero transverse momentum
   ThreeVector beta(ktot.p1(), ktot.p2(), 0.0);
   beta /= sqrt(ktot.perp2() + s * (1. - xi));
   Boost boostTransverse(Boost(-beta));

   projectedconfig.clear();
   for (Configuration::iterator it = phiNp1.beginFinal(); it != phiNp1.endFinal(); ++it) {
      if (it != phiNp1.emitted()) {
         Particle particle(*it);
         particle.p().apply(longB);
         particle.p().apply(boostTransverse);
         particle.p().apply(longB.inverse());
         projectedconfig.insertFinal(particle);
      }
   }

   // now do the projected pair flavor
   PdgID projectedID;
   if (phiNp1.emitted()->id().isGluon())
      projectedID = phiNp1.sister()->id();
   else if (phiNp1.sister()->id().isGluon())
      projectedID = (phiNp1.emitted()->id()).antiPartner();
   else
      projectedID = PdgID::kGluon;

   const double tmp = sqrt((2. - xi * (1. + y)) / (2. - xi * (1. - y)));
   const double EA = phiNp1.initialA()->energy() * sqrt(1. - xi) * tmp;
   const double EB = phiNp1.initialB()->energy() * sqrt(1. - xi) / tmp;

   Configuration::iterator emitter;
   if (phiNp1.sister()->isInitialA()) {
      emitter = projectedconfig.setInitialA(Particle(projectedID, EA, 0., 0., EA));
      projectedconfig.setInitialB(Particle(phiNp1.initialB()->id(), EB, 0., 0., -EB));
   } else {
      projectedconfig.setInitialA(Particle(phiNp1.initialA()->id(), EA, 0., 0., EA));
      emitter = projectedconfig.setInitialB(Particle(projectedID, EB, 0., 0., -EB));
   }

   const BoostZHyp toCMframe(sqrt(EB/EA));
   projectedconfig.apply(toCMframe);

#ifndef NDEBUG
   if (!checkMomentumConservation(projectedconfig, FKSKINEMATICS_PRECISION, FKSKINEMATICS_ACCURACY))
      Log::warning("FKSKinematics::projectConfigISR") << "PhiN point from projection violates momentum conservation." << Log::end;
   if (!checkOnShellConditions(projectedconfig, FKSKINEMATICS_PRECISION, FKSKINEMATICS_ACCURACY))
      Log::warning("FKSKinematics::projectConfigISR") << "PhiN point from projection violates on-shell conditions." << Log::end;
#endif

   return emitter;
}

//------------------------------------------------------------------------------
/// Creates an N-body %Configuration from the N+1-body configuration by combining the two FKS partons. It assumes a FSR splitting has taken place.
Configuration::iterator FKSKinematics::projectConfigFSR(PhiNplus1 phiNp1, Configuration& projectedconfig) const
{
   //sanity check that is in the CM
   assert(isInCoMFrame(phiNp1, 1e-3, 1e-5));

   // sanity check this is only ever used for FSR
   assert(phiNp1.sister()->isFinal());

   // now we define k = k_i + k_j
   // and q =  kA + kB
   // and krec = sum_{l != i,j} k_l = q - k
   FourVector k = phiNp1.emitted()->p() + phiNp1.sister()->p();
   FourVector q = phiNp1.initialA()->p() + phiNp1.initialB()->p();
   FourVector krec = q - k;

   // now find the beta needed for the boost
   // the boost B makes q - B.krec massless
   // beta is parallel to krec
   // mangitude of beta = {alpha * sqrt[alpha^2 + x^2 - 1] - x}/{alpha^2 + x^2}
   // where alpha = {q^2 + krec^2}/{2q.krec}
   // and x = {q0*krec0 * (krecmom/krec0) + qvec.krecvec * (krec0/krecmom)}/{q.krec}
   double alpha = (q.square() + krec.square()) / (2 * (q * krec));
   double betarec = krec.pvec().magnitude() / krec.p0();
   double x = (q.p0() * krec.p0() * betarec + q.pvec() * krec.pvec() * (1 / betarec)) / (q * krec);
   double betamag = (alpha * sqrt(alpha * alpha + x * x - 1) - x) / (alpha * alpha + x * x);

   ThreeVector beta(krec.p1() * betamag / krec.pvec().magnitude(),
                    krec.p2() * betamag / krec.pvec().magnitude(),
                    krec.p3() * betamag / krec.pvec().magnitude());

   // define the boost by beta
   Boost boostrec(beta);

   // now make the new config
   // all of the final state particles \bar{Phi}_n are defined as follows:
   // l != i,j: \bar{k}_l = B.k_l
   // \bar{k}_{i+j} = q - B.krec
   // these new momenta should still sum to q, each be massless
   projectedconfig.clear();

   // add initial particles - copy them in, they don't change
   projectedconfig.setInitialA(*phiNp1.initialA());
   projectedconfig.setInitialB(*phiNp1.initialB());

   // add final particles, first do the recoil ones
   // copy them in and then change their momenta
   for (Configuration::iterator it = phiNp1.beginFinal(); it != phiNp1.endFinal(); ++it) {
      if (it != phiNp1.emitted() && it != phiNp1.sister()) {
         PdgID particleid = it->id();
         FourVector newmom = boostrec(it->p());
         projectedconfig.insertFinal(Particle(particleid, newmom.p0(), newmom.p1(), newmom.p2(), newmom.p3()));
      }
   }

   // now do the projected pair
   // flavor
   PdgID projectedID;
   if (phiNp1.emitted()->id().isGluon())
      projectedID = phiNp1.sister()->id();
   else if (phiNp1.sister()->id().isGluon())
      projectedID = phiNp1.emitted()->id();
   else
      projectedID = PdgID::kGluon;

   FourVector projectedmom = q - boostrec(krec);
   Configuration::iterator emitter = projectedconfig.insertFinal(Particle(projectedID, projectedmom));

#ifndef NDEBUG
   if (!checkMomentumConservation(projectedconfig, FKSKINEMATICS_PRECISION, FKSKINEMATICS_ACCURACY))
      Log::warning("FKSKinematics::projectConfigFSR") << "PhiN point from projection violates momentum conservation." << Log::end;
   if (!checkOnShellConditions(projectedconfig, FKSKINEMATICS_PRECISION, FKSKINEMATICS_ACCURACY))
      Log::warning("FKSKinematics::projectConfigFSR") << "PhiN point from projection violates on-shell conditions." << Log::end;
#endif

   return emitter;
}

//------------------------------------------------------------------------------
/// Creates the PhiRadFKS from the N+1-body configuration, assuming the PhiNp1 is given in the CM frame.
const PhiRadFKS FKSKinematics::projectPhiRad(PhiNplus1 phiNp1) const
{
   //sanity check that is in the CM
   assert(isInCoMFrame(phiNp1, 1e-3, 1e-5));

   //Construct the radiation type
   PhiRadFKS::RadiationType radType;
   if (phiNp1.sister()->isInitialA())
      radType = PhiRadFKS::kISRA;
   else if (phiNp1.sister()->isInitialB())
      radType = PhiRadFKS::kISRB;
   else
      radType = PhiRadFKS::kFSR;

   const double Q = phiNp1.initialA()->energy() + phiNp1.initialB()->energy();
   const double xi = 2. * phiNp1.emitted()->energy() / Q;
   double y;
   if (radType == PhiRadFKS::kFSR)
      y = cosAngleBetween(phiNp1.emitted()->pvec(), phiNp1.sister()->pvec());
   else
      // ISR
      y = phiNp1.emitted()->p().p3() / phiNp1.emitted()->p().p0();

   // determine the phi angle
   double phi;
   if (radType == PhiRadFKS::kFSR) {
      ThreeVector k = Normalize()(phiNp1.emitted()->pvec() + phiNp1.sister()->pvec());
      ThreeVector fksDir = Normalize()(phiNp1.emitted()->pvec());
      phi = getPhiAngleFSR(fksDir, k);
   } else {
      // For ISR, phi is determined with respect to the y-axis.
      phi = math_constants::pi_2() - phiNp1.emitted()->p().phi02pi();
      if (phi < 0.)
         phi += math_constants::_2pi();
   }

   // now do the projected pair flavor
   PdgID idEmitter;
   if ((radType == PhiRadFKS::kISRA) || (radType == PhiRadFKS::kISRB)) {
      if (phiNp1.emitted()->id().isGluon())
         idEmitter = phiNp1.sister()->id();
      else if (phiNp1.sister()->id().isGluon())
         idEmitter = (phiNp1.emitted()->id()).antiPartner();
      else
         idEmitter = PdgID::kGluon;
   } else if (radType == PhiRadFKS::kFSR) {
      if (phiNp1.emitted()->id().isGluon())
         idEmitter = phiNp1.sister()->id();
      else if (phiNp1.sister()->id().isGluon())
         idEmitter = phiNp1.emitted()->id();
      else
         idEmitter = PdgID::kGluon;
   }

   return PhiRadFKS(xi, y, phi, idEmitter, phiNp1.sister()->id(), phiNp1.emitted()->id(), radType);
}

//------------------------------------------------------------------------------
/**
 * Calculates the phi angle for a FSR emission in the CM frame.
 * It requires the emitter and fks parton %ThreeVectors to be defined in the CM frame
 * and to have unit lenght
 */
double FKSKinematics::getPhiAngleFSR(const ThreeVector& fksDir, const ThreeVector& emitterDir)
{
   // make sure we are working with unit vectors
   assert(compare(emitterDir.magnitude(), 1.) == 0);
   assert(compare(fksDir.magnitude(), 1.) == 0);
   const ThreeVector eta(0., 0., 1.);
   // calculate the component of fks transverse to eimmeter by subtracting the parallel component
   const ThreeVector fksPerp = fksDir - (fksDir * emitterDir) * emitterDir;
   const ThreeVector etaPerp = crossProduct(emitterDir, eta);
   double phi = atan2(fksPerp*etaPerp, fksPerp*eta);
   // make sure we return a phi in [0,2pi]
   if (phi < 0.) phi += math_constants::_2pi();
   return phi;
}

} // namespace Geneva
