//------------------------------------------------------------------------------
/// \file FKSMapping.cpp
//
// Author(s):
//    Jon Walsh, Saba Zuberi, Frank Tackmann, Christian Bauer, Chris Vermilion, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class FKSMapping
//------------------------------------------------------------------------------

#include "Geneva/Components/FKSMapping.hpp"

#include "Geneva/Core/Physics/PhiRadFKS.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
double FKSMapping::getxA(const double xBarA, const double xi, const double y) const
{
   return FKSKinematics().getxA(xBarA, xi, y);
}

//------------------------------------------------------------------------------
double FKSMapping::getxB(const double xBarB, const double xi, const double y) const
{
   return FKSKinematics().getxB(xBarB, xi, y);
}

//------------------------------------------------------------------------------
double FKSMapping::getXiMaxReal(const PhiNcmFKS& phiN, const double xBarA, const double xBarB, const double y) const
{
   return FKSKinematics().getXiMax(phiN, xBarA, xBarB, y);
}

//------------------------------------------------------------------------------
double FKSMapping::getXiMaxColl(const PhiNcmFKS& phiN, const double xBarA, const double xBarB) const
{
   return FKSKinematics().getXiMaxColl(phiN, xBarA, xBarB);
}

//-------------------------------------------------------------------------------------
/**
 * \brief Generates the split config
 * Returns false if splitting could not be performed because the emitter is ambigous.
 * \param phiN the N-body %Configuration and the %Particle to split
 * \param phiRad the kinematic and flavor variables in the splitting
 * \param x1, \param x2 the momentum fractions of incoming particles (if any)
 * \param[out] phiNp1 the generated phase-space point
 */
bool FKSMapping::generateSplit(const PhiNcmFKS& phiN, const PhiRadFKS& phiRadFKS, double x1, double x2, PhiNplus1FKS& phiNp1)
{
   if (phiN.emitter->isFinal()) {
      _jacobian = FKSKinematics().splitFSR(phiN, phiRadFKS, phiNp1);
   } else if (phiN.emitter->isInitialA() || phiN.emitter->isInitialB()) {
      _jacobian = FKSKinematics().splitISR(phiN, phiRadFKS, x1, x2, phiNp1);
   } else {
      return false;
   }
   return true;
}

//------------------------------------------------------------------------------
bool FKSMapping::project(PhiNplus1 phiNp1, Configuration& phiN) const
{
   if (phiNp1.sister()->isFinal())
      FKSKinematics().projectConfigFSR(phiNp1, phiN);
   else
      FKSKinematics().projectConfigISR(phiNp1, phiN);
   return true;
}

} // namespace Geneva
