//------------------------------------------------------------------------------
/// \file FluxFactor.cpp
//
// Author(s):
//    Christian Bauer, Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class FluxFactor
//------------------------------------------------------------------------------

#include "Geneva/Components/FluxFactor.hpp"

#include "Geneva/Core/Physics/Configuration.hpp"
#include "Geneva/Core/Physics/SignatureSet.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
// Checks compatibility with the given SignatureSet.
/**
 * Checks to see if the given signatures can be calculated with this component.
 * If so, then 0 is returned
 * If not, then 1 is returned
 * Possibly could have other return codes if a subset is doable, for example
 */
int FluxFactor::checkCompatibility(const SignatureSet& sigSet) const
{
   return 0;
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------
/**
 * \brief Computes the flux factor part of the cross section
 * Symbolically, gives the part of the cross section represented by \f$\frac{1}{2 E_A 2 E_B |v_A-v_B|}\f$
 */
double FluxFactor::operator()(const Configuration& config) const
{
   const FourVector fvA = config.initialA()->p();
   const FourVector fvB = config.initialB()->p();
   const double mAsquared = (fvA).square();
   const double mBsquared = (fvB).square();

   return 1./ (4.* sqrt(pow(fvA*fvB,2) - mAsquared*mBsquared));
}

} // namespace Geneva
