//------------------------------------------------------------------------------
/// \file LHAPDF.cpp
//
// Author(s):
//    Chris Vermilion, Simone Alioli
//
// Copyright:
//    Copyright (C) 2013 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class LHAPDF
//------------------------------------------------------------------------------

//#define USE_ANALYTIC_PDF

#define __STDC_CONSTANT_MACROS // needed for C99 INT64_C macros
#include <stdint.h>
#include <cstdlib>

namespace Geneva
{

//------------------------------------------------------------------------------
class HashKey
{
   public:
      bool operator==(const HashKey& key) const {
         return eq(x, key.x) && eq(Q, key.Q);
      };

      HashKey(double x_ = 0, double Q_ = 0)
         : x(x_), Q(Q_) {}

   private:
      double x, Q;
      typedef union {double d64; int64_t i64;} DI64;
      // static const double mdelta=1e-14;
      // static const int64_t idelta=1+mdelta/DBL_EPSILON; // needs cfloat
      // Comparison tolerance for double
      // idelta fdelta(%)
      // 5 - 1.11e-15
      // 46 - 1.02e-14
      // 451 - 1.00e-13
      // 4504 - 1.00e-12
      // 45036 - 1.00e-11
      static const uint64_t idelta = UINT64_C(5); // unsigned !!!

      bool eq(double a, double b) const {
         const DI64 ia = {a};
         const DI64 ib = {b};
         const int64_t diff = ia.i64 - ib.i64;
         return diff == INT64_C(0) || static_cast<uint64_t>(llabs(diff)) <= idelta;
      }

      bool neq(double a, double b) const {
         const DI64 ia = {a};
         const DI64 ib = {b};
         const int64_t diff = ia.i64 - ib.i64;
         return diff != INT64_C(0) && static_cast<uint64_t>(llabs(diff)) > idelta;
      }
};

} // namespace Geneva

#undef __STDC_CONSTANT_MACROS

#include "Geneva/Components/LHAPDF.hpp"

#include "Geneva/Core/UI/Log.hpp"

#include "LHAPDF/PDF.h"

#include <sstream>
#include <deque>
#include <cassert>

namespace Geneva
{

typedef std::pair<HashKey, std::map<int, double> > HashPair;

//------------------------------------------------------------------------------
std::map<int, double> LHAPDF::pdfA(double x, double muF)  const
{
#ifdef USE_ANALYTIC_PDF
   Log::fatal("must not be called") << Log::end;
#endif
   HashKey hash(x, muF);
   static std::deque<HashPair> PDFStack;
   std::deque<HashPair>::iterator it;
   for (it = PDFStack.begin(); it != PDFStack.end(); ++it) {
      if (hash == it->first)
         return it->second;
   }
   if (PDFStack.size() > _maxStackLength) PDFStack.pop_back();

   // not found in stack - calculate it
   std::map<int, double> r = _pdfA->xfxQ(x, muF);
   std::map<int, double>::iterator mit, mitE = r.end();
   // maps are ordered by key, so this should result in same vector as before
   for (mit = r.begin(); mit != mitE; ++mit) {
      mit->second /= x;
   }

   // add to stack
   PDFStack.push_front(std::make_pair(hash, r));
   return PDFStack.begin()->second;
}

//------------------------------------------------------------------------------
std::map<int, double> LHAPDF::pdfB(double x, double muF) const
{
#ifdef USE_ANALYTIC_PDF
   Log::fatal("must not be called") << Log::end;
#endif
   HashKey hash(x, muF);
   static std::deque<HashPair> PDFStack;
   std::deque<HashPair>::iterator it;
   for (it = PDFStack.begin(); it != PDFStack.end(); ++it) {
      if (hash == it->first)
         return it->second;
   }
   if (PDFStack.size() > _maxStackLength) PDFStack.pop_back();

   // not found in stack - calculate it
   std::map<int, double> r = _pdfB->xfxQ(x, muF);
   std::map<int, double>::iterator mit, mitE = r.end();
   // maps are ordered by key, so this should result in same vector as before
   for (mit = r.begin(); mit != mitE; ++mit) {
      mit->second /= x;
   }

   // add to stack
   PDFStack.push_front(std::make_pair(hash, r));
   return PDFStack.begin()->second;
}

//------------------------------------------------------------------------------
/// Return the PdgId that would be required to fake the parton content of a proton
///NOTE: it assumes symmetry of sea-quarks content
PdgID LHAPDF::getFakePartonId(const PdgID part, const PdgID beam) const
{
  if (beam == PdgID::kProtonBar) {
      return part.antiPartner();
  } else if (beam == PdgID::kPhoton) {
      //TODO: in case the of photon beams no trasformation is required, but only few LHAPDf sets support photons. Implement check on that.
      return part;
  } else if (beam == PdgID::kNeutron) {
      if (part == PdgID::kUp) {
          return PdgID::kDown;
      } else if (part == PdgID::kDown) {
          return PdgID::kUp;
      } else if (part ==PdgID::kUpBar) {
          return PdgID::kDownBar;
      } else if (part == PdgID::kDownBar) {
          return PdgID::kUpBar;
      } else {
          return part;
      }
  } else if (beam == PdgID::kNeutronBar) {
      if (part == PdgID::kUp) {
          return PdgID::kDownBar;
      } else if (part == PdgID::kDown) {
          return PdgID::kUpBar;
      } else if (part ==PdgID::kUpBar) {
          return PdgID::kDown;
      } else if (part == PdgID::kDownBar) {
          return PdgID::kUp;
      } else {
          return part.antiPartner();
      }
  } else {
    Log::error("LHAPDF::getFakePartonId") << "Beam "<<beam<<" not yet implemented."<<Log::end;
  }
  return PdgID::kUnknown;
}

//------------------------------------------------------------------------------
double LHAPDF::pdfA(PdgID id, double x, double muF) const
{
#ifdef USE_ANALYTIC_PDF
   return 1./x;
#else
   assert(id.isGluon() || id.isQuark());
   double pdf = 0.;
   try {
      PdgID part = (_beamA == PdgID::kProton) ? id : getFakePartonId(id, _beamA);
      pdf = (_independentFlavourEvolution) ? (_pdfA->xfxQ(part.code(), x, muF)) / x : pdfA(x, muF)[part.code()];
   } catch  (std::exception& lhapdfEx) {
      Log::error("LHAPDF::pdfA") << lhapdfEx.what() << Log::end;
   }
   return pdf;
#endif
}

//------------------------------------------------------------------------------
double LHAPDF::pdfB(PdgID id, double x, double muF) const
{
#ifdef USE_ANALYTIC_PDF
   return 1./x;
#else
   assert(id.isGluon() || id.isQuark());
   double pdf = 0;
   try {
      PdgID part = (_beamB == PdgID::kProton) ? id : getFakePartonId(id, _beamB);
      pdf = (_independentFlavourEvolution) ? (_pdfB->xfxQ(part.code(), x, muF)) / x : pdfB(x, muF)[part.code()];
   } catch  (std::exception& lhapdfEx) {
      Log::error("LHAPDF::pdfB") << lhapdfEx.what() << Log::end;
   }
   return pdf;
#endif
}

//------------------------------------------------------------------------------
LHAPDF::LHAPDF(PdgID beamA, PdgID beamB, const std::string& name, int member,  size_t maxStackLength)
   : _beamA(beamA), _memberA(member), _nameA(name), _beamB(beamB), _memberB(member), _nameB(name), _maxStackLength(maxStackLength)
{
   if (_maxStackLength == 0) {
      _independentFlavourEvolution = true;
      Log::verboseInfo("LHAPDF") << "Using independent flavour evolution for " <<name << Log::end;
   } else {
      _independentFlavourEvolution = false;
      Log::verboseInfo("LHAPDF") << "LHAPDF: Using combined flavour evolution and caching of results for "<< name  << Log::end;
   }
   ::LHAPDF::setVerbosity(1);
   try {
      _pdfA = ::LHAPDF::mkPDF(name, member);
   } catch (std::exception& lhapdfEx) {
      Log::error("LHAPDF") << lhapdfEx.what() << Log::end;
   }
   ::LHAPDF::setVerbosity(0);
   _pdfB = _pdfA;
}

//------------------------------------------------------------------------------
LHAPDF::LHAPDF(PdgID beamA, PdgID beamB, const std::string& nameA, const std::string& nameB,
               int memberA, int memberB, size_t maxStackLength)
   : _beamA(beamA), _memberA(memberA), _nameA(nameA), _beamB(beamB), _memberB(memberB), _nameB(nameB), _maxStackLength(maxStackLength)
{
   if (_maxStackLength == 0) {
      _independentFlavourEvolution = true;
      Log::verboseInfo("LHAPDF") << "Using independent flavour evolution for "<< nameA <<" and "<< nameB << Log::end;
   } else {
      _independentFlavourEvolution = false;
      Log::verboseInfo("LHAPDF") << "LHAPDF: Using combined flavour evolution and caching of results for " << nameA <<" and "<< nameB << Log::end;
   }
   ::LHAPDF::setVerbosity(1);
   try {
      _pdfA = ::LHAPDF::mkPDF(nameA, memberA);
      _pdfB = ::LHAPDF::mkPDF(nameB, memberB);
   } catch  (std::exception& lhapdfEx) {
      Log::error("LHAPDF") << lhapdfEx.what() << Log::end;
   }
   ::LHAPDF::setVerbosity(0);
}

//------------------------------------------------------------------------------
LHAPDF::~LHAPDF()
{
   if (_nameA != _nameB) delete _pdfB;
   delete _pdfA;
}

//------------------------------------------------------------------------------
int LHAPDF::pdfIDA() const
{
   return _pdfA->lhapdfID();
}

//------------------------------------------------------------------------------
int LHAPDF::pdfIDB() const
{
   return _pdfB->lhapdfID();
}

//------------------------------------------------------------------------------
int LHAPDF::getNf() const
{
   // New LHAPDF allows non-contiguous flavors? Take highest quark ID as "Nf"
   std::vector<int> flavorsA = _pdfA->flavors();
   int nfA = -1;
   for (size_t i = 0; i < flavorsA.size(); ++i) {
      if (flavorsA[i] > nfA && flavorsA[i] <= 8) {
         nfA = flavorsA[i];
      }
   }
   std::vector<int> flavorsB = _pdfB->flavors();
   int nfB = -1;
   for (size_t i = 0; i < flavorsB.size(); ++i) {
      if (flavorsB[i] > nfB && flavorsB[i] <= 8) {
         nfB = flavorsB[i];
      }
   }
   if (nfA != nfB)
      Log::error("LHAPDF::getNf") << "The two pdf's have different values of nf" << Log::end;
   if (nfA == -1) {
      Log::warning("LHAPDF::getNf") << "Returned nf = -1, using nf = 5 instead." << Log::end;
      nfA = 5;
   }
   return nfA;
}

//------------------------------------------------------------------------------
std::string LHAPDF::name() const
{
   std::string result;
   if (_maxStackLength > 1) {
      // TODO: Can't cache independent flavours for now
      assert(!_independentFlavourEvolution);
      result += "Cached version of ";
   }
      result += "LHAPDF " + ::LHAPDF::version() + " with ";
      result += (_independentFlavourEvolution) ? "independent " : "combined ";
      result += "flavour evolution. ";
      std::ostringstream ssA;
      _pdfA->print(ssA,1);
      std::string infoA = ssA.str();
      result += infoA.substr(0, infoA.length()-1);
      if (nameA() != nameB()) {
         std::ostringstream ssB;
         _pdfA->print(ssB,1);
         std::string infoB = ssB.str();
         result += " ";
         result += infoB.substr(0, infoB.length()-1);
      }
      return result;
}

//------------------------------------------------------------------------------
double LHAPDF::getAlphasQ(double Q) const
{
   if (nameA() != nameB()) {
      Log::error("LHAPDF::getAlphasQ") << "Cannot get alphas from PDF when using two different PDFs." << Log::end;
   }
   return _pdfA->alphasQ(Q);
}

//------------------------------------------------------------------------------
double LHAPDF::getmZ() const
{
   if (nameA() != nameB())
      Log::error("LHAPDF::getmZ") << "Cannot get mZ from PDF when using two different PDFs." << Log::end;
   double mZ = std::numeric_limits<double>::quiet_NaN();
   if (_pdfA->info().has_key("MZ")) {
      mZ = _pdfA->info().get_entry_as<double>("MZ");
   }
   if (!std::isfinite(mZ)) {
      Log::error("LHAPDF::getmZ") << "Failed to get sensible mZ from LHAPDF." << Log::end;
   }
   return mZ;
}

//------------------------------------------------------------------------------
bool LHAPDF::isPositive() const
{
   return _pdfA->forcePositive() > 0 && _pdfB->forcePositive() > 0;
}

} // namespace Geneva
