//------------------------------------------------------------------------------
/// \file OpenLoopsInterface.cpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class OpenLoopsInterface
//------------------------------------------------------------------------------

#include "Geneva/config.hpp"

#include "Geneva/Components/OpenLoopsInterface.hpp"
#include "Geneva/Components/SignatureManager.hpp"

#include "Geneva/Core/Physics/SignatureSet.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include <fstream>
#include <sstream>
#include <iterator>

//------------------------------------------------------------------------------
/**
 *  Functions defined in the openLoops interface
 */
//------------------------------------------------------------------------------

extern "C" {
   void ol_setparameter_int(const char* param, int val);
   void ol_setparameter_double(const char* param, double val);
   void ol_setparameter_string(const char* param, const char* val);
   void ol_getparameter_int(const char* param, int* val);
   void ol_getparameter_double(const char* param, double* val);
   int ol_register_process(const char* process, int amptype);
   int ol_n_external(int id);
   void ol_phase_space_point(int id, double sqrt_s, double* pp);
   void ol_start();
   void ol_finish();
   void ol_tree_colbasis_dim(int id, int* ncolb, int* colelemsz, int* nheltot);
   void ol_tree_colourflow(int id, int* flowbasis);
   void ol_evaluate_tree_colvect2(int id, double* pp, double* m2arr);
   void ol_evaluate_tree(int id, double* pp, double* m2tree);
   void ol_evaluate_cc(int id, double* pp, double* m2tree, double* m2cc, double* m2ew);
   void ol_evaluate_sc(int id, double* pp, int emitter, double* polvect, double* m2sc);
   void ol_evaluate_scpowheg(int id, double* pp, int emitter, double* m2tree, double* m2munu);
   void ol_evaluate_loop(int id, double* pp, double* m2tree, double* m2loop, double* acc);
   void ol_evaluate_loop2(int id, double* pp, double* m2loop2, double* acc);
   void ol_evaluate_ct(int id, double* pp, double* m2_tree, double* m2_ct);
}

namespace Geneva
{

//------------------------------------------------------------------------------
OpenLoopsInterface::OpenLoopsInterface()
   : _codesMaps(kAmplitudeTypeLEN)
{
   std::string install_path;
   const char* env_cstr = std::getenv("OpenLoopsPath");
   if (env_cstr != 0 && *env_cstr != '\0')
      install_path = std::string(env_cstr);
   else
      install_path = std::string(config::openloops_install_path);

   Log::verboseInfo("OpenLoopsInterface") << "Setting openloops install_path to: "
      << install_path << Log::end;
   ol_setparameter_string("install_path", install_path.c_str());
}

//------------------------------------------------------------------------------
OpenLoopsInterface::~OpenLoopsInterface()
{
   ol_finish();
}

//------------------------------------------------------------------------------
void OpenLoopsInterface::addLoopProcesses(const OptionMap& optMap)
{
   // Increase verbosity level to list loaded libraries
   if (optMap.getOption<bool>("verbose")) {
      ol_setparameter_int("verbose", 2);
   } else {
      ol_setparameter_int("verbose", 0);
      ol_setparameter_double("psp_tolerance", 1e-5);
   }
   const std::string calcName = optMap.getOption("calc");
   int nf = optMap.getOption<int>("NLightFlavors");
   std::vector<SignatureSet> sigSets;
   std::vector<int> alphasPowers;
   std::vector<int> alphaEmPowers;
   PdgID lepton;
   std::string decay = optMap.getOption<std::string>("Process::DrellYan::DecayChannel");
   if (decay == "electrons")  lepton = PdgID::kElectron;
   else if (decay == "muons") lepton = PdgID::kMuon;
   else if (decay == "taus")  lepton = PdgID::kTau;
   else {
      Log::error("OpenLoopsInterface::addLoopProcesses") << "Unknown DecayChannel " << decay << Log::end;
   }
   if (calcName == "CalcNLOppV") {
      sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 0, nf));
      alphasPowers.push_back(1);
      alphaEmPowers.push_back(2);
   } else if (calcName == "CalcNLOppVj") {
      sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 1, nf));
      alphasPowers.push_back(2);
      alphaEmPowers.push_back(2);
   } else if (calcName == "CalcNLOppVjj") {
      sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 2, nf));
      alphasPowers.push_back(3);
      alphaEmPowers.push_back(2);
   } else if (calcName == "CalcSCETppV012") {
      sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 0, nf));
      alphasPowers.push_back(1);
      alphaEmPowers.push_back(2);
      sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 1, nf));
      alphasPowers.push_back(2);
      alphaEmPowers.push_back(2);
   } else {
      Log::error("OpenLoopsInterface:addLoopProcesses") << "OpenLoops not available for this calculation " << calcName << Log::end;
   }

   for (size_t i = 0; i < sigSets.size(); i++) {
      ol_setparameter_int("order_ew", alphaEmPowers[i]);
      //NOTE:  order_qcd must only be specified when doing NLO EW corrections
      //ol_setparameter_int("order_qcd", alphasPowers[i]);
      for (SignatureSet::iterator it = sigSets[i].begin(); it != sigSets[i].end(); ++it) {
         std::stringstream ss;
         ss << *(it->initialA()) <<" "<< *(it->initialB()) << " ->";
         for (Signature::iterator ii = it->beginFinal(); ii != it->endFinal(); ++ii) {
            ss << " " << *ii;
         }
         std::string name = ss.str();
         int code = ol_register_process(name.c_str(), 11);
         if (code <= 0) {
            Log::error("OpenLoopsInterface::addLoopProcesses") << "could not register "<<name<<". Openloops returns code "<<code<<Log::end;
         } else {
            if (_codesMaps[kLoop].find(*it) == _codesMaps[kLoop].end()) {
               _codesMaps[kLoop][*it] = code;
               Log::debug("OpenLoopsInterface::addLoopProcesses") << "Registered loop process "<< *it << "with code "<< code<<Log::end;
            }
         }
      }
   }

   _optMap = optMap;
}

//------------------------------------------------------------------------------
void OpenLoopsInterface::addTreeProcesses(const OptionMap& optMap)
{
   // Increase verbosity level to list loaded libraries
   if (optMap.getOption<bool>("verbose")) {
      ol_setparameter_int("verbose", 2);
   } else {
      ol_setparameter_int("verbose", 0);
      ol_setparameter_double("psp_tolerance", 1e-5);
   }
   const std::string calcName = optMap.getOption("calc");
   int nf = optMap.getOption<int>("NLightFlavors");
   std::vector<SignatureSet> sigSets;
   std::vector<int> alphasPowers;
   std::vector<int> alphaEmPowers;
   PdgID lepton;
   std::string decay = optMap.getOption<std::string>("Process::DrellYan::DecayChannel");
   if (decay == "electrons")  lepton = PdgID::kElectron;
   else if (decay == "muons") lepton = PdgID::kMuon;
   else if (decay == "taus")  lepton = PdgID::kTau;
   else {
      Log::error("OpenLoopsInterface::addTreeProcesses") << "Unknown DecayChannel " << decay << Log::end;
   }
   if (calcName == "CalcNLOppV") {
      sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 0, nf));
      alphasPowers.push_back(0);
      alphaEmPowers.push_back(2);
      if (optMap.getOption("FixedOrder") == "NLO") {
         sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 1, nf));
         alphasPowers.push_back(1);
         alphaEmPowers.push_back(2);
      }
   } else if (calcName == "CalcNLOppVj") {
      sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 1, nf));
      alphasPowers.push_back(1);
      alphaEmPowers.push_back(2);
      if (optMap.getOption("FixedOrder") == "NLO") {
         sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 2, nf));
         alphasPowers.push_back(2);
         alphaEmPowers.push_back(2);
      }
   } else if (calcName == "CalcNLOppVjj") {
      sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 2, nf));
      alphasPowers.push_back(2);
      alphaEmPowers.push_back(2);
      if (optMap.getOption("FixedOrder") == "NLO") {
         sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 3, nf));
         alphasPowers.push_back(3);
         alphaEmPowers.push_back(2);
      }
   } else if (calcName == "CalcSCETppV012") {
      sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 0, nf));
      alphasPowers.push_back(0);
      alphaEmPowers.push_back(2);
      sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 1, nf));
      alphasPowers.push_back(1);
      alphaEmPowers.push_back(2);
      if ((optMap.getOption("CalcSCETppV012::fixed1jet") == "NLO") or (optMap.getOption("CalcSCETppV012::run1") != "none")) {
         sigSets.push_back(SignatureManager().getSignaturespp2ZJets(lepton, 2, nf));
         alphasPowers.push_back(2);
         alphaEmPowers.push_back(2);
      }
   } else {
      Log::error("OpenLoopsInterface::addTreeProcesses") << "OpenLoops not available for this calculation " << calcName << Log::end;
   }

   for (size_t i = 0; i < sigSets.size(); i++) {
      ol_setparameter_int("order_ew", alphaEmPowers[i]);
      //NOTE:  order_qcd must only be specified when doing NLO EW corrections
      //ol_setparameter_int("order_qcd", alphasPowers[i]);
      for (SignatureSet::iterator it = sigSets[i].begin(); it != sigSets[i].end(); ++it) {
         std::stringstream ss;
         ss << *(it->initialA()) <<" "<< *(it->initialB()) << " ->";
         for (Signature::iterator ii = it->beginFinal(); ii != it->endFinal(); ++ii) {
            ss << " " << *ii;
         }
         std::string name = ss.str();
         int code = ol_register_process(name.c_str(), 1);
         if (code <= 0) {
            Log::error("OpenLoopsInterface::addTreeProcesses") << "could not register "<<name<<". Openloops returns code "<<code<<Log::end;
         } else {
            if (_codesMaps[kTree].find(*it) == _codesMaps[kTree].end()) {
               _codesMaps[kTree][*it] = code;
               Log::debug("OpenLoopsInterface::addLoopProcesses") << "Registered tree process "<< *it << "with code "<< code<<Log::end;
            }
         }
      }
   }

   _optMap = optMap;
}

//------------------------------------------------------------------------------
void OpenLoopsInterface::addLoopInducedProcesses(const OptionMap& optMap)
{
   Log::fatal("OpenLoopsInterface::addLoopInducedProcesses") << "OpenLoops interface not available for this yet" << Log::end;
   std::string name = "none";
   // Increase verbosity level to list loaded libraries
   if (optMap.getOption<bool>("verbose")) {
      ol_setparameter_int("verbose", 2);
   } else {
      ol_setparameter_int("verbose", 0);
      ol_setparameter_double("psp_tolerance", 1e-5);
   }
   ol_register_process(name.c_str(), 12);
}

//------------------------------------------------------------------------------
void OpenLoopsInterface::initialize()
{
   if (_isInitialized) {
      return;
   }

   //means stability_mode=23, redlib1=1, redlib2=7, redlib_qp=5
   ol_setparameter_int("preset", 2);

   double mz = _optMap.getOption<double>("ZMass");
   ol_setparameter_double("mass(23)", mz);
   double gz = _optMap.getOption<double>("ZWidth");
   ol_setparameter_double("width(23)", gz);
   double mw = _optMap.getOption<double>("WMass");
   ol_setparameter_double("mass(24)", mw);
   double gw = _optMap.getOption<double>("WWidth");
   ol_setparameter_double("width(24)", gw);
   int nf = _optMap.getOption<int>("NLightFlavors");
   ol_setparameter_int("minnf_alphasrun", nf);
   // NOTE: The following is creating problems with the fact that nf=6 was used for the generation
//    ol_setparameter_int("n_quarks", nf);

   // Initialize OpenLoops
   ol_start();

   _isInitialized = true;
}

//------------------------------------------------------------------------------
/**
 * Converts a configuration of Geneva into a pointer to momenta
 */
void OpenLoopsInterface::configToOpenLoops(const Configuration& config, double* momenta) const
{
   // assign momenta
   momenta[0] = config.initialA()->energy();
   momenta[1] = config.initialA()->px();
   momenta[2] = config.initialA()->py();
   momenta[3] = config.initialA()->pz();
   momenta[4] = config.initialA()->mass();

   momenta[5] = config.initialB()->energy();
   momenta[6] = config.initialB()->px();
   momenta[7] = config.initialB()->py();
   momenta[8] = config.initialB()->pz();
   momenta[9] = config.initialB()->mass();

   size_t i = 10;
   for (Configuration::iterator it = config.beginFinal(); it != config.end(); ++it) {
      momenta[i++] = it->energy();
      momenta[i++] = it->px();
      momenta[i++] = it->py();
      momenta[i++] = it->pz();
      momenta[i++] = it->mass();
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Evaluate the value of the squared tree-level matrix element for the given configuration.
 * \param config The %Configuration to use
 * \param renormScale The value of the renormalization scale
 * \param alphaS The value of the strong coupling constant
 * \param alphaEM The value of the electromagnetic coupling constant
 * \param m2 The double to store the result in
 */
void OpenLoopsInterface::evaluateTree(const Configuration& config, double renormScale, double alphaS, double alphaEM, double* m2)
{
   initialize();
   size_t nParticles = config.size();
   double* p = new double [nParticles * 5];
   configToOpenLoops(config, p);
   Signature sig(config);
   _codes_map_type::iterator found = _codesMaps[kTree].find(sig);
   if (found == _codesMaps[kTree].end())
      Log::error("OpenLoopsInterface::evaluateTree") << "Could not find Signature " << sig << "in amplitude list" << Log::end;
   const int code = found->second;
   assert(ol_n_external(code) == int_from_size_t(nParticles));
   // Set parameter: strong coupling
   ol_setparameter_double("alpha_s", alphaS);
   // Set parameter: EM coupling
   ol_setparameter_double("alpha", alphaEM);
   // Set parameter: renormalization scale
   ol_setparameter_double("mu", renormScale);

   ol_evaluate_tree(code, p, m2);

   delete[] p;
}

//------------------------------------------------------------------------------
/**
 * \brief Evaluate the value of the squared tree-level matrix element  and the color info for the given configuration.
 * \param config The %Configuration to use
 * \param renormScale The value of the renormalization scale
 * \param alphaS The value of the strong coupling constant
 * \param alphaEM The value of the electromagnetic coupling constant
 * \param m2 The double to store the result in
 */
void OpenLoopsInterface::evaluateTreeLargeNc(const Configuration& config, double renormScale, double alphaS, double alphaEM, double* m2)
{
   initialize();
   int nParticles = int_from_size_t(config.size());
   double* p = new double [nParticles * 5];
   configToOpenLoops(config, p);
   Signature sig(config);
   std::map<Signature,int>::iterator found = _codesMaps[kTree].find(sig);
   if (found == _codesMaps[kTree].end())
      Log::error("OpenLoopsInterface::evaluateTree") << "Could not find Signature " << sig << "in amplitude list" << Log::end;
   const int code = found->second;
   assert(ol_n_external(code) == nParticles);
   // Set parameter: strong coupling
   ol_setparameter_double("alpha_s", alphaS);
   // Set parameter: EM coupling
   ol_setparameter_double("alpha", alphaEM);
   // Set parameter: renormalization scale
   ol_setparameter_double("mu", renormScale);

   ol_evaluate_tree_colvect2(code, p, m2);

   delete[] p;
}

//------------------------------------------------------------------------------
/**
 * \brief Retrieve the color indices associated with all possible color flows.
 * \return The color flow basis.
 */
std::map<Signature, std::vector<std::vector<std::pair<int, int> > > > OpenLoopsInterface::getColorFlowBasis()
{
   std::map<Signature, std::vector<std::vector<std::pair<int, int> > > > flowbasis;
   for (std::map<Signature,int>::iterator it = _codesMaps[kTree].begin(); it != _codesMaps[kTree].end(); ++it) {
      int id, ncolb, colelemsz, nheltot, nex;
      id = it->second;
      // number of external particles in the process
      nex = ol_n_external(id);
      // dimensions of the colour basis (colelemsz and nheltot can be ignored here)
      ol_tree_colbasis_dim(id, &ncolb, &colelemsz, &nheltot);
      std::vector<int> cfb(size_t_from_int(ncolb * nex * 2));
      // get the colour flow basis
      ol_tree_colourflow(id, cfb.data());
      std::vector<std::vector<std::pair<int, int> > > values;
      for (size_t i = 0; i < size_t_from_int(ncolb); ++i) {
         std::vector<std::pair<int, int> > valuecol;
         for (size_t j = 0; j < size_t_from_int(nex); ++j) {
            // NOTE: this gymnastic is needed because OL does not follow the LHEF
            // convention and returns inverted indexes for outgoing partons.
            // We assume all indexes to be inverted below and only
            // invert the incoming (minimizing the number of inversions).
            std::pair<int, int> colIndexes(cfb[i*size_t_from_int(2*nex) + 2*j + 1], cfb[i*size_t_from_int(2*nex) + 2*j]);
            if (colIndexes.first != 0)
               colIndexes.first += 500;
            if (colIndexes.second != 0)
               colIndexes.second += 500;
            if (j < 2)
               std::swap(colIndexes.first, colIndexes.second);
            valuecol.push_back(colIndexes);
         }
         values.push_back(valuecol);
      }
      flowbasis[it->first] = values;
   }
   return flowbasis;
}

//------------------------------------------------------------------------------
/**
 * \brief Evaluate the value of the squared tree-level matrix element for the given configuration, including color and spin correlations.
 * \param config The %Configuration to use
 * \param renormScale The value of the renormalization scale
 * \param alphaS The value of the strong coupling constant
 * \param alphaEM The value of the electromagnetic coupling constant
 * \param m2 The double to store the result in
 * \param m2cc The array to store the color-correlated results in
 * \param m2munu The array to store the spin-correlated results in
 * \param m2ccew The array to store the charge-conjugated results in
 */
void OpenLoopsInterface::evaluateTreeComplete(const Configuration& config, double renormScale, double alphaS, double alphaEM, double* m2, double* m2cc, double* m2munu, double* m2ccew)
{
   initialize();
   size_t nParticles = config.size();
   double* p = new double [nParticles * 5];
   configToOpenLoops(config, p);
   Signature sig(config);
   _codes_map_type::iterator found = _codesMaps[kTree].find(sig);
   if (found == _codesMaps[kTree].end())
      Log::error("OpenLoopsInterface::evaluateTreeComplete") << "Could not find Signature " << sig << "in amplitude list" << Log::end;
   const int code = found->second;
   assert(ol_n_external(code) == int_from_size_t(nParticles));
   // Set parameter: strong coupling
   ol_setparameter_double("alpha_s", alphaS);
   // Set parameter: EM coupling
   ol_setparameter_double("alpha", alphaEM);
   // Set parameter: renormalization scale
   ol_setparameter_double("mu", renormScale);

   double m2tmp1 = std::numeric_limits<double>::quiet_NaN();
   ol_evaluate_cc(code, p, &m2tmp1, m2cc, m2ccew);

   int position = 1;
   for (Configuration::iterator it = config.begin(); it != config.end(); ++it, position++) {
      double m2tmp2 = 0.;
      if (it->id().isGluon()) {
         ol_evaluate_scpowheg(code, p, position, &m2tmp2, &m2munu[(position-1)*16]);
#ifndef NDEBUG
         if (compare(m2tmp1, m2tmp2, 1e-6) != 0)
            Log::warning("OpenLoopsInterface::evaluateTreeComplete") << " LO results from ol_evaluate_cc " << m2tmp1 << " and ol_evaluate_scpowheg " << m2tmp2 << " are not numerically compatible." << Log::end;
#endif
      }
   }
   *m2 = m2tmp1;
   delete[] p;
}

//------------------------------------------------------------------------------
/**
 * \brief Evaluate the values of the one-loop and the squared tree-level matrix element for the given configuration.
 * \param config The %Configuration to use
 * \param renormScale The value of the renormalization scale
 * \param alphaS The value of the strong coupling constant
 * \param alphaEM The value of the electromagnetic coupling constant
 * \param m2tree The double to store the tree-level result in
 * \param m2loop The double to store the one-loop result in
 * \param accuracy The doubel to store the accuracy of the result
 */
void OpenLoopsInterface::evaluateLoop(const Configuration& config, double renormScale, double alphaS, double alphaEM, double* m2tree, double* m2loop, double* accuracy)
{
   initialize();
   size_t nParticles = config.size();
   double* p = new double [nParticles * 5];
   configToOpenLoops(config, p);
   Signature sig(config);
   _codes_map_type::iterator found = _codesMaps[kLoop].find(sig);
   if (found == _codesMaps[kLoop].end())
      Log::error("OpenLoopsInterface::evaluateLoop") << "Could not find Signature " << sig << "in amplitude list" << Log::end;
   const int code = found->second;
   assert(ol_n_external(code) == int_from_size_t(nParticles));
   // Set parameter: strong coupling
   ol_setparameter_double("alpha_s", alphaS);
   // Set parameter: EM coupling
   ol_setparameter_double("alpha", alphaEM);
   // Set parameter: renormalization scale
   ol_setparameter_double("mu", renormScale);

   ol_evaluate_loop(code, p, m2tree, m2loop, accuracy);

   delete[] p;
}

} // end namespace Geneva
