//------------------------------------------------------------------------------
/// \file RegionInfo.cpp
//
// Author(s):
//   Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of RegionInfo class
//------------------------------------------------------------------------------

#include "Geneva/Components/RegionInfo.hpp"

#include "Geneva/Core/Physics/Signature.hpp"
#include "Geneva/Core/Physics/Configuration.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/Random.hpp"

#include <ostream>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * Picks the correct emitter from the Configuration or selects it randomly when
 * multiple particles of the same kind are present in the final state.
 */
Configuration::iterator RegionInfo::getEmitter(const Configuration& config) const
{
   const PdgID idEmitter = *k;
   if (hasFinalStateSingularity()) {
      // pick a random emitter between the equivalent ones
      std::pair<Configuration::iterator, Configuration::iterator> emitters = config.equalRangeFinal(idEmitter);
      size_t numEmitters = size_t_from_ptrdiff_t(std::distance(emitters.first, emitters.second));
      if (numEmitters >= 2)
         std::advance(emitters.first, ptrdiff_t_from_size_t(randomInteger(numEmitters)));
      return emitters.first;
   } else if (hasInitialStateSingularity()) {
      if (k->isInitialA())
         return config.initialA();
      else
         return config.initialB();
   } else {
      Log::error("RegionInfo::getEmitter") << "Neither final nor initial state singularity." << Log::end;
   }
   return config.end();
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Streams the RegionInfo to output.
std::ostream& RegionInfo::print(std::ostream& out) const
{
   out << "Signature N : " << *fN << '\n'
       << "Mult N: " << multN << '\n'
       << "Signature N+1 : " << *fNp1 << '\n';
   if (hasFinalStateSingularity()) {
      out << "Splitting FSR: (" << *k << ") -> " << *j << " " << *i << "_FKS\n";
   } else if (hasInitialStateSingularity()) {
      out << "Splitting ISR: " << *j << " -> (" << *k << ") " << *i << "_FKS\n";
   }
   out << "Mult Splitting :" << multRad << '\n';
   return out;
}

} // namespace Geneva
