//------------------------------------------------------------------------------
/// \file RegionInfoList.cpp
//
// Author(s):
//   Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementations of RegionInfoList class
//------------------------------------------------------------------------------

#include "Geneva/Components/RegionInfoList.hpp"

#include "Geneva/Core/Physics/Signature.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include <cassert>
#include <iterator>
#include <algorithm>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * This function identifies all the singular regions, given the n-particles and
 * n+1-particles signatures and fill the _allRegions internal list with
 * all the info for each region.
 */
void RegionInfoList::setRegionInfoList(const SignatureSet& sigN, const SignatureSet& sigNplusOne, int nLight, bool verbose)
{
   // clear previous inizalizations
   if (_initialized) this->clear();
   _nLight = nLight;
   _signaturesN = sigN;
   _signaturesNp1 = sigNplusOne;
   _allRegions.clear();
   _singularRegionFromSigMap.clear();

   for (SignatureSet::iterator sigIter = sigN.begin(); sigIter != sigN.end(); ++sigIter) {
      std::set<int> finalEmitters;
      Signature tempSig; // new signature to be constructed
      SignatureSet::iterator sigBNp1; // iterator to Set of N+1 Signatures
      RegionInfo info;
      info.nLight = nLight;
      info.fN = sigIter;
      info.multN = sigIter->symmetryFactor();

      // initial state emitter A
      info.k = sigIter->initialA();
      if (info.k->isGluon()) {
         // g -> (g) g initial state splitting
         tempSig.clear();
         tempSig = *sigIter;
         tempSig.insertFinal(PdgID::kGluon);
         sigBNp1 = sigNplusOne.find(tempSig);
         if (sigBNp1 != sigNplusOne.end()) {
            info.fNp1 = sigBNp1;
            info.i = sigBNp1->findFinal(PdgID::kGluon);
            info.j = sigBNp1->initialA();
            info.multRad = multiplicityFactor(tempSig, info.j, info.i);
            _allRegions.push_back(info);
            _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
            _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
         } else {
            Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
         }

         for (int fl = 1; fl <= nLight; fl++) {
            // q -> (g) q initial state splitting
            PdgID quarkID(fl);
            tempSig.clear();
            tempSig = *sigIter;
            tempSig.insertFinal(quarkID);
            tempSig.setInitialA(quarkID);
            sigBNp1 = sigNplusOne.find(tempSig);
            if (sigBNp1 != sigNplusOne.end()) {
               info.fNp1 = sigBNp1;
               info.i = sigBNp1->findFinal(quarkID);
               info.j = sigBNp1->initialA();
               info.multRad = multiplicityFactor(tempSig, info.j, info.i);
               _allRegions.push_back(info);
               _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
               _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
            } else {
               Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
            }

            // qbar -> (g) qbar initial state splitting
            tempSig.clear();
            tempSig = *sigIter;
            tempSig.insertFinal(quarkID.antiPartner());
            tempSig.setInitialA(quarkID.antiPartner());
            sigBNp1 = sigNplusOne.find(tempSig);
            if (sigBNp1 != sigNplusOne.end()) {
               info.fNp1 = sigBNp1;
               info.i = sigBNp1->findFinal(quarkID.antiPartner());
               info.j = sigBNp1->initialA();
               info.multRad = multiplicityFactor(tempSig, info.j, info.i);
               _allRegions.push_back(info);
               _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
               _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
            } else {
               Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
            }
         }
      } else if (abs(info.k->code()) <= nLight && info.k->code() != 0) {

         // g -> (q) qbar initial state splitting
         tempSig.clear();
         tempSig = *sigIter;
         tempSig.insertFinal(info.k->antiPartner());
         tempSig.setInitialA(PdgID::kGluon);
         sigBNp1 = sigNplusOne.find(tempSig);
         if (sigBNp1 != sigNplusOne.end()) {
            info.fNp1 = sigBNp1;
            info.i = sigBNp1->findFinal(info.k->antiPartner());
            info.j = sigBNp1->initialA();
            info.multRad = multiplicityFactor(tempSig, info.j, info.i);
            _allRegions.push_back(info);
            _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
            _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
         } else {
            Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
         }

         // q -> (q) g initial state splitting
         tempSig.clear();
         tempSig = *sigIter;
         tempSig.insertFinal(PdgID::kGluon);
         sigBNp1 = sigNplusOne.find(tempSig);
         if (sigBNp1 != sigNplusOne.end()) {
            info.fNp1 = sigBNp1;
            info.i = sigBNp1->findFinal(PdgID::kGluon);
            info.j = sigBNp1->initialA();
            info.multRad = multiplicityFactor(tempSig, info.j, info.i);
            _allRegions.push_back(info);
            _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
            _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
         } else {
            Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
         }
      }

      // initial state emitter B
      info.k = sigIter->initialB();
      if (info.k->isGluon()) {
         // g -> (g) g initial state splitting
         tempSig.clear();
         tempSig = *sigIter;
         tempSig.insertFinal(PdgID::kGluon);
         sigBNp1 = sigNplusOne.find(tempSig);
         if (sigBNp1 != sigNplusOne.end()) {
            info.fNp1 = sigBNp1;
            info.i = sigBNp1->findFinal(PdgID::kGluon);
            info.j = sigBNp1->initialB();
            info.multRad = multiplicityFactor(tempSig, info.j, info.i);
            _allRegions.push_back(info);
            _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
            _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
         } else {
            Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
         }

         for (int fl = 1; fl <= nLight; fl++) {
            // q -> (g) q initial state splitting
            PdgID quarkID(static_cast<int>(fl));
            tempSig.clear();
            tempSig = *sigIter;
            tempSig.insertFinal(quarkID);
            tempSig.setInitialB(quarkID);
            sigBNp1 = sigNplusOne.find(tempSig);
            if (sigBNp1 != sigNplusOne.end()) {
               info.fNp1 = sigBNp1;
               info.i = sigBNp1->findFinal(quarkID);
               info.j = sigBNp1->initialB();
               info.multRad = multiplicityFactor(tempSig, info.j, info.i);
               _allRegions.push_back(info);
               _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
               _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
            } else {
               Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
            }

            // qbar -> (g) qbar initial state splitting
            tempSig.clear();
            tempSig = *sigIter;
            tempSig.insertFinal(quarkID.antiPartner());
            tempSig.setInitialB(quarkID.antiPartner());
            sigBNp1 = sigNplusOne.find(tempSig);
            if (sigBNp1 != sigNplusOne.end()) {
               info.fNp1 = sigBNp1;
               info.i = sigBNp1->findFinal(quarkID.antiPartner());
               info.j = sigBNp1->initialB();
               info.multRad = multiplicityFactor(tempSig, info.j, info.i);
               _allRegions.push_back(info);
               _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
               _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
            } else {
               Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
            }
         }
      } else if (abs(info.k->code()) <= static_cast<int>(nLight) && info.k->code() != 0) {
         // g -> (q) qbar initial state splitting
         tempSig.clear();
         tempSig = *sigIter;
         tempSig.insertFinal(info.k->antiPartner());
         tempSig.setInitialB(PdgID::kGluon);
         sigBNp1 = sigNplusOne.find(tempSig);
         if (sigBNp1 != sigNplusOne.end()) {
            info.fNp1 = sigBNp1;
            info.i = sigBNp1->findFinal(info.k->antiPartner());
            info.j = sigBNp1->initialB();
            info.multRad = multiplicityFactor(tempSig, info.j, info.i);
            _allRegions.push_back(info);
            _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
            _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
         } else {
            Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
         }

         // q -> (q) g initial state splitting
         tempSig.clear();
         tempSig = *sigIter;
         tempSig.insertFinal(PdgID::kGluon);
         sigBNp1 = sigNplusOne.find(tempSig);
         if (sigBNp1 != sigNplusOne.end()) {
            info.fNp1 = sigBNp1;
            info.i = sigBNp1->findFinal(PdgID::kGluon);
            info.j = sigBNp1->initialB();
            info.multRad = multiplicityFactor(tempSig, info.j, info.i);
            _allRegions.push_back(info);
            _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
            _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
         } else {
            Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
         }
      }

      // final state emitters
      for (Signature::iterator k = sigIter->beginFinal(); k != sigIter->endFinal(); ++k) {
         if (finalEmitters.find(k->code()) == finalEmitters.end()) {
            finalEmitters.insert(k->code());
            info.k = k;
            if (k->isGluon()) {
               info.k = k;
               // (g) -> g g final state splitting
               tempSig.clear();
               tempSig = *sigIter;
               tempSig.insertFinal(PdgID::kGluon);
               sigBNp1 = sigNplusOne.find(tempSig);
               if (sigBNp1 != sigNplusOne.end()) {
                  info.fNp1 = sigBNp1;
                  std::pair<Signature::iterator, Signature::iterator> range = sigBNp1->equalRangeFinal(PdgID::kGluon);
                  //NOTE: we make the specific choice that the emitter iterator corresponds to the first occurrence of PdgID::kGluon
                  // and that the emitted iterator corresponds to the last. Any other pairing would be equally allowed.
                  info.i = range.first;
                  info.j = --range.second;
                  info.multRad = multiplicityFactor(tempSig, info.j, info.i);
                  _allRegions.push_back(info);
                  _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
                  _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
               } else {
                  Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
               }

               for (int fl = 1; fl <= nLight; fl++) {
                  // (g) -> q qbar final state splitting
                  PdgID quarkID(static_cast<int>(fl));
                  tempSig.clear();
                  tempSig = *sigIter;
                  tempSig.erase(tempSig.findFinal(PdgID::kGluon));
                  tempSig.insertFinal(quarkID);
                  tempSig.insertFinal(quarkID.antiPartner());
                  sigBNp1 = sigNplusOne.find(tempSig);
                  if (sigBNp1 != sigNplusOne.end()) {
                     info.fNp1 = sigBNp1;
                     info.i = sigBNp1->findFinal(quarkID);
                     info.j = sigBNp1->findFinal(quarkID.antiPartner());
                     info.multRad = multiplicityFactor(tempSig, info.j, info.i);
                     _allRegions.push_back(info);
                     _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
                     _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
                  } else {
                     Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
                  }
               }
            } else if (abs(info.k->code()) <= static_cast<int>(nLight) && info.k->code() != 0) {
               // (q) -> g q final state splitting
               tempSig.clear();
               tempSig = *sigIter;
               tempSig.insertFinal(PdgID::kGluon);
               sigBNp1 = sigNplusOne.find(tempSig);
               if (sigBNp1 != sigNplusOne.end()) {
                  info.fNp1 = sigBNp1;
                  info.i = sigBNp1->findFinal(PdgID::kGluon);
                  info.j = sigBNp1->findFinal(*info.k);
                  info.multRad = multiplicityFactor(tempSig, info.j, info.i);
                  _allRegions.push_back(info);
                  _singularRegionFromSigMap.insert(std::make_pair(*sigIter, info));
                  _singularRegionToSigMap.insert(std::make_pair(*sigBNp1, info));
               } else {
                  Log::warning("RegionInfoList::setRegionInfoList") << "SignatureSet N+1 does not contain " << tempSig << Log::end;
               }
            }
         }
      }
   }
   Log::debug("RegionInfoList::setRegionInfoList") << _allRegions.size() << " singular regions found " << Log::end;
   for (RegionInfoList::region_list_iterator infoIter = _allRegions.begin(); infoIter != _allRegions.end(); ++infoIter) {
      _singularSignatures.insert(*(infoIter->fNp1));
      if (verbose)
         Log::debug() << *infoIter;
   }
   Log::debug("RegionInfoList::setRegionInfoList") << _singularSignatures.size() << " singular signatures found" << Log::end;
   if (verbose)
      Log::debug() << _singularSignatures << Log::end;

   _regularSignatures = sigNplusOne - _singularSignatures;
   if (!_regularSignatures.empty()) {
      Log::warning("RegionInfoList::setRegionInfoList") << _regularSignatures.size() << " regular signatures found" << Log::end;
      if (verbose)
         Log::debug() << _regularSignatures << Log::end;
   }
   //sanity check
   assert((_signaturesNp1 - _singularSignatures - _regularSignatures).empty());

   //if everything ok, set the status
   _initialized = true;
}

/**
 * This function returns the multiplicity factor for a splitting from a N-body
 * %Signature towards a given N+1-body %Signature. Once the emitter has been fixed,
 * the multiplicity depends only upon the possible assignments of the FKS parton
 * and its sister amongst the available PdgID's in the %Signature. The initial and
 * final-state emitter case are handled differently.
 */
//------------------------------------------------------------------------------
size_t RegionInfoList::multiplicityFactor(const Signature& sigNp1, Signature::const_iterator sister, Signature::const_iterator fks) const
{
   size_t tmp = 1;
   if (sister->isFinal()) {
      if (*sister != *fks) {
         tmp = sigNp1.countFinal(*sister);
         tmp *= sigNp1.countFinal(*fks);
      } else {
         size_t N = sigNp1.countFinal(*fks);
         tmp = N * (N - 1);
      }
   } else {
      tmp = sigNp1.countFinal(*fks);
   }
   return tmp;
}

//------------------------------------------------------------------------------
// Streams the RegionInfoList to output.
std::ostream& RegionInfoList::print(std::ostream& out) const
{
   std::copy(begin(), end(), std::ostream_iterator<RegionInfo>(out));
   return out;
}

} // namespace Geneva
