//------------------------------------------------------------------------------
/// \file SplittingFunctionBase
//
// Author(s):
//    Christian Bauer, Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 LBNL, CERN
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class SplittingFunctionBase
//------------------------------------------------------------------------------

#include "Geneva/Components/SplittingFunctionBase.hpp"
#include "Geneva/Components/ISplittingMapping.hpp"
#include "Geneva/Components/CouplingStrong.hpp"

#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/Physics/ConfigurationExtensions.hpp"
#include "Geneva/Core/Utils/NumericsUtils.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
SplittingFunctionBase::SplittingFunctionBase(const OptionMap& optMap, ISplittingMapping* mapping, CouplingStrong* alphas, double Ecm)
   : _optMap(optMap), _cmEnergy(Ecm), _mapping(mapping), _alphas(alphas)
{
   _useRunningAlphas = optMap.getOption<bool>("SplittingFunction::runningAlpha");
   _numSplittingCalled = 0;
   _numNormalizedCalled = 0;
   _numSplittingFailures = 0;
   _numRejectCalled = 0;
}

//-------------------------------------------------------------------------------------
/**
 * Returns true if the phase space point satisfies all the theta
 * functions put on by the function normalized
 */
bool SplittingFunctionBase::isValid(const PhiNplus1& phiNplus1)
{
   //Calculate the ratio of the bounding integral ot the true integral statistically
   double TauN, z, phi;
   RegionInfo regionInfo;
   Configuration configN;

   // check that clustering is allowed by QCD flavor
   if (invalidFlavorProjection(_mapping->sigsNmin1(), phiNplus1))
      return false;

   if (not _mapping->calcPhiNValues(phiNplus1, configN, TauN, z, phi, regionInfo))
      return false;

   //If the resulting PhiN, TauN, z, phi values are not valid for the splitting, return 0
   if (!_mapping->isValidPhiN(configN, TauN, z, phi, regionInfo.k))
      return false;

   return true;
}

//------------------------------------------------------------------------------
ISplittingMapping* SplittingFunctionBase::getMapping()
{
   return _mapping;
}

//-------------------------------------------------------------------------------------
double SplittingFunctionBase::zMin(const Configuration& configN, double TauN, double phi, RegionInfo& regionInfo)
{
   return _mapping->zMin(configN, TauN, phi, regionInfo.k, *regionInfo.i);
}

//-------------------------------------------------------------------------------------
double SplittingFunctionBase::zMax(const Configuration& configN, double TauN, double phi, RegionInfo& regionInfo)
{
   return _mapping->zMax(configN, TauN, phi, regionInfo.k, *regionInfo.i);
}

//------------------------------------------------------------------------------
// Protected member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
double SplittingFunctionBase::jacobian(const Configuration& config, double TauN, double z, double phi, RegionInfo& map) const
{
   return _mapping->jacobian(config, TauN, z, phi, map);
}

//------------------------------------------------------------------------------
void SplittingFunctionBase::initializeFailedZValues(double zLower, double zUpper)
{
   _zLower = zLower;
   _zUpper = zUpper;
   _passedZValues.clear();
   for (size_t i = 0; i < 50; ++i) {
      _passedZValues.push_back(0);
   }
}

//------------------------------------------------------------------------------
void SplittingFunctionBase::addPassedZValue(double z)
{
   int zBin = static_cast<int>(50. * log((1. - z) / (1. - _zLower)) / log((1. - _zUpper) / (1. - _zLower)));
   //Deal with under and overflow values
   if (zBin > 49)
      zBin = 49;
   else if (zBin < 0)
      zBin = 0;
   _passedZValues[static_cast<size_t>(zBin)] += 1;
}

//------------------------------------------------------------------------------
double SplittingFunctionBase::adjustedZMax(double z, double zMin, double zMax)
{
   int bin = 49;
   for (; bin >= 0; --bin) {
      if (_passedZValues[static_cast<size_t>(bin)] > 0)
         break;
   }
   double zHit = zMax;
   if (bin < 49)
      zHit = 1. - pow((1. - zMax) / (1. - zMin), static_cast<double>(bin + 1) / 50.) * (1. - zMin);
   forceInRange(zHit,zMin,zMax,"SplittingFunctionBase::adjustedZMax");
   if (zHit < z)
      zHit = z;
   return zHit + 0.5 * (zMax - zHit);
}

//------------------------------------------------------------------------------
double SplittingFunctionBase::adjustedZMin(double z, double zMin, double zMax)
{
   size_t bin = 0;
   for (; bin <= 49; ++bin) {
      if (_passedZValues[bin] > 0)
         break;
   }
   double zHit = zMax;
   if (bin < 49)
      zHit = 1. - pow((1. - zMax) / (1. - zMin), static_cast<double>(bin + 1) / 50.) * (1. - zMin);
   forceInRange(zHit,zMin,zMax,"SplittingFunctionBase::adjustedZMin");
   if (zHit > z)
      zHit = z;
   return zMin + 0.5 * (zHit - zMin);
}

//------------------------------------------------------------------------------
std::pair<double, double> SplittingFunctionBase::adjustZMinZMax(double z, double zMin, double zMax, double phiMin, double phiMax, double TauN, Configuration& configN, RegionInfo& regionInfo, ISplittingMapping* mapping) {
   //Try to reduce the size of the z Range
   double zMaxNew = adjustedZMax(z, zMin, zMax);
   //Now check that this new zMaxvalue is really above the allowed range
   bool isAbove = false;
   size_t numCheckIsGood = 0;
   while (not isAbove) {
      numCheckIsGood++;
      if (numCheckIsGood>10)
         break;
      for (int i = 0; i < 1000; ++i) {
         isAbove = true;
         double phiTest = phiMin + 1. * i / 1000 * (phiMax - phiMin);
         if (mapping->isValidPhiN(configN, TauN, zMax, phiTest, regionInfo.k)) {
            isAbove = false;
            zMaxNew = zMaxNew + 0.1 * (zMax - zMaxNew);
            break;
         }
      }
   }
   double zMinNew = adjustedZMin(z, zMin, zMax);
   //Now check that this new zMaxvalue is really above the allowed range
   bool isBelow = false;
   numCheckIsGood = 0;
   while (not isBelow) {
      numCheckIsGood++;
      if (numCheckIsGood>10)
         break;
      for (int i = 0; i < 1000; ++i) {
         isBelow = true;
         double phiTest = phiMin + 1. * i / 1000 * (phiMax - phiMin);
         if (mapping->isValidPhiN(configN, TauN, zMax, phiTest, regionInfo.k)) {
            isBelow = false;
            zMinNew = zMin + 0.1 * (zMinNew - zMin);
            break;
         }
      }
   }
   return std::pair<double, double>(zMinNew, zMaxNew);
}

} // namespace Geneva
