//------------------------------------------------------------------------------
/// \file SplittingFunctionFSR.cpp
//
// Author(s):
//    Christian Bauer, Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 LBNL, CERN
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class SplittingFunctionFSR
//------------------------------------------------------------------------------

#include "Geneva/Components/SplittingFunctionFSR.hpp"
#include "Geneva/Components/ISplittingMapping.hpp"
#include "Geneva/Components/RegionInfoList.hpp"
#include "Geneva/Components/BoundingConstantsFSR.hpp"
#include "Geneva/Components/CouplingStrong.hpp"

#include "Geneva/Core/Physics/QCD.hpp"
#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/Random.hpp"
#include "Geneva/Core/Utils/NumericsUtils.hpp"
#include "Geneva/Core/Utils/FileUtils.hpp"

#include <fstream>
#include <stdexcept>

#define CUTOFF_FSR 1.0e-10

namespace Geneva
{

//------------------------------------------------------------------------------
SplittingFunctionFSR::SplittingFunctionFSR(const OptionMap& optMap, ISplittingMapping* mapping, CouplingStrong* alphas, double Ecm)
   : SplittingFunctionBase(optMap, mapping, alphas, Ecm)
{
   //Store the random state, so that we get the same sequence whether we create the bounding grid or not
   RandomState ranstate = getRandomState();
   _setupFile = getSetupFilename(optMap, format_tag(_mapping->name()) + format_tag("FSR"));
   std::ifstream setupFile(_setupFile.c_str());
   if (setupFile.good()) {  //Read from the setup File
      _boundingConstant = new BoundingConstantsFSR(setupFile);
      //Check that we have all necessary bounding constants
      if (haveAllBoundingConstants()) {
         _isInitialized = true;
         Log::verboseInfo() << "Loaded FSR bounding function for "<< _mapping->name() <<" from file " << _setupFile << Log::end;
      } else {
         Log::warning("SplittingFunctionFSR") << "Not all bounding constants for "<< _mapping->name() <<" found in file " << _setupFile << Log::end;
         _isInitialized = false;
      }
   } else {
      _boundingConstant = new BoundingConstantsFSR((_useRunningAlphas) ? 20 : 1, 50, 0.1, Ecm, 0.0, 1.0 - CUTOFF_FSR);
      _isInitialized = false;
   }
   if (not _isInitialized) {
      initialize();
      _isInitialized = true;
   }
   double boundFactor = optMap.getOption<double>("SplittingFunction::boundFactor");
   if (boundFactor < 1.0) {
      Log::fatal("SplittingFunction") << "In " << _mapping->name() << " SplittingFunctionFSR::boundFactor must be larger or equal to 1" << Log::end;
   }
   _boundingConstant->setBoundFactor(boundFactor);
   //Restore the random state, so that we get the same sequence whether we create the bounding grid or not
   initializeRandomState(ranstate);
}

//-------------------------------------------------------------------------------------
SplittingFunctionFSR::~SplittingFunctionFSR()
{
   delete _boundingConstant;
}

//-------------------------------------------------------------------------------------
/**
 * Implements the function p(PhiNplus1)
 * The z values of this function are distributed according to the splitting function f(z)
 * The phi values are distributed flat
 * It gives unity if integrated over all allowed values of z and phi, and summed over the allowed splittings
 *
 * \param PhiNplus1 The (N+1)-body phase space point
 *
 * This function does not require a (N+1)-body point that gives rise to a valid splitting.
 * If the (N+1)-body point is invalid, it returns zero
 */
double SplittingFunctionFSR::normalized(const PhiNplus1& phiNplus1)
{
   _numNormalizedCalled++;
   //Calculate the ratio of the bounding integral ot the true integral statistically
   double TauN, z, phi;
   RegionInfo regionInfo;
   Configuration configN;

   // check that clustering is allowed by QCD flavor
   if (invalidFlavorProjection(_mapping->sigsNmin1(), phiNplus1)) {
      return 0.;
   }

   if (not _mapping->calcPhiNValues(phiNplus1, configN, TauN, z, phi, regionInfo)) {
      return 0.;
   }

   //If the resulting PhiN, TauN, z, phi values are not valid for the splitting, return 0
   if (not _mapping->isValidPhiN(configN, TauN, z, phi, regionInfo.k)) {
      return 0.;
   }

   // Now that I have valid values, calculate the splitting function
   try {

      // Determine the value of phiMin and phiMax
      double phiMin = _mapping->phiMin(configN, TauN, regionInfo.k, *regionInfo.i);
      double phiMax = _mapping->phiMax(configN, TauN, regionInfo.k, *regionInfo.i);

      // Determine the value of zMin and zMax
      double zMin = _mapping->zMin(configN, TauN, phi, regionInfo.k, *regionInfo.i);
      double zMax = _mapping->zMax(configN, TauN, phi, regionInfo.k, *regionInfo.i);
      forceInRange(z, zMin, zMax, "SplittingFunctionFSR::normalized");

      //Call the splitting function
      _numSplittingCalled++;

      double result = 0.0;
      //Determine the value of the un-normalized splitting function
      result = splitting(regionInfo.i, regionInfo.j, TauN, z);

      //If that is zero, no need to determine the normalization factor
      if (result == 0) {
         return 0.;
      }

      // Normalize to the total integral of the bounding function
      result /= intBound(regionInfo.k, TauN, zMin, zMax);
      // Count the number of times I am creating an invalid configuration by randomly choosing z and phi
      size_t intRatio = 1;
      double randomZ = pickZ(regionInfo.k, TauN, zMin, zMax);
      double randomPhi = phiMin + randomInterval() * (phiMax - phiMin);
      result /= (phiMax - phiMin);

      //Keep track which z values I have tried
      initializeFailedZValues(zMin, zMax);
      unsigned int adjustZMinMax = 0;

      //Check if the resulting z, phi values are rejected or not
      while (rejectEvent(configN, TauN, randomZ, randomPhi, regionInfo)) {
         _numRejectCalled++;
         //Increase the counter
         intRatio++;
         //Create new z and phi values
         randomZ = pickZ(regionInfo.k, TauN, zMin, zMax);
         randomPhi = phiMin + randomInterval() * (phiMax - phiMin);

         if (intRatio % 5000 == 0) {
            adjustZMinMax++;
            std::pair<double, double> zMinMax = adjustZMinZMax(z, zMin, zMax, phiMin, phiMax, TauN, configN, regionInfo, _mapping);
            zMin = zMinMax.first;
            zMax = zMinMax.second;
            intRatio = 1;
            initializeFailedZValues(zMin, zMax);
         }
         //If we have adjusted the zMax value more than 100 times, throw an error
         if (adjustZMinMax > 100) {
            Log::error("SplittingFunctionFSR::normalized") << "In " << _mapping->name() << " Got stuck calculating the ratio of the integrals.\n The limits are {zMin, zMax} = {" << zMin << ", " << zMax << "}" << "{phiMin, phiMax} = {" << phiMin << ", " << phiMax << "}, and the event is\n" << configN << Log::end;
         }
      }

      //Multiply the result by the ratio of the true integral over the integral of the bounding function
      result *= static_cast<double>(intRatio);
      if (result == 0.) {
         return 0.;
      }
      result /= jacobian(configN, TauN, z, phi, regionInfo);
      result /= static_cast<double>(_mapping->numQCDN() + 1);
      return result;

   } catch (const NormalError&) {
      _numSplittingFailures++;
      if (10000*_numSplittingFailures > _numSplittingCalled && (_numSplittingCalled > 500000))
         Log::fatal("SplittingFunctionFSR::normalized") << "In " << _mapping->name() << " " << _numSplittingFailures << " failures in " << _numSplittingCalled << " tries." << Log::end;
      else
         Log::warning("SplittingFunctionFSR::normalized") << "In " << _mapping->name() << " Error caught, returning 0 and continuing." << Log::end;
      return 0.;
   }
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
void SplittingFunctionFSR::initialize()
{
   Log::info("SplittingFunctionFSR::initialize") << "Determining upper-bounding grids for splitting functions for "<< _mapping->name() <<" ..." << Log::end;

   //Loop over all possible mappings
   std::list<RegionInfo> Mappings = _mapping->getRegionInfoList().getFsrRegionInfoList();
   for (std::list<RegionInfo>::iterator m = Mappings.begin(); m != Mappings.end(); ++m) {
      if (not m->k->isFinal()) {
         Log::fatal("SplittingFunctionFSR::init") << "In " << _mapping->name() << " Only works for FSR" << Log::end;
      }
      std::string rad;
      rad = "FSR";
      //Skip it if we already have a bounding constant for it
      if (_boundingConstant->boundExists(m->k)) {
         continue;
      }
      //FSR emissions
      //For this case I know the analytical form, so I can easily obtain an upper bounding function
      for (size_t tauNBin = 0; tauNBin < _boundingConstant->numTauNBins(); tauNBin++) {
         //The maximum in TauN is at the lowest edge, since it only enter in alphas
         double lowTauNBinEdge = _boundingConstant->binTauNEdges(tauNBin).first;
         //For all splitting functions, (1-z) * P(z) is concave, such that the maximum is at the edges
         for (size_t zBin = 0; zBin < _boundingConstant->numZBins(); zBin++) {
            std::pair<double, double> zBinLimits = _boundingConstant->binZEdges(zBin);
            double maxRatio = std::max(sumAllowedSplittings(lowTauNBinEdge, zBinLimits.first, *m) / bound(m->k, lowTauNBinEdge, zBinLimits.first), sumAllowedSplittings(lowTauNBinEdge, zBinLimits.second, *m) / bound(m->k, lowTauNBinEdge, zBinLimits.second));
            _boundingConstant->informOfRatio(m->k, tauNBin, zBin, maxRatio);
         }
      }
      Log::info("SplittingFunctionFSR::init") << "Finished FSR bounding constant for id " << m->k->code() << " " << rad << Log::end;
   }

   std::ofstream setupFile(_setupFile.c_str());
   if (setupFile.good()) {
      Log::verboseInfo() << "Writing bounding constant for FSR to file "<< _setupFile << Log::end;
      _boundingConstant->write(setupFile);
   } else {
      Log::warning("SplittingFunctionFSR::init") << "In " << _mapping->name() << " Problem in writing bounding constant" << Log::end;
   }
   setupFile.close();
}

//------------------------------------------------------------------------------
bool SplittingFunctionFSR::haveAllBoundingConstants() const
{
   //Loop over all possible FSR mappings
   std::list<RegionInfo> Mappings = _mapping->getRegionInfoList().getFsrRegionInfoList();
   for (std::list<RegionInfo>::iterator m = Mappings.begin(); m != Mappings.end(); ++m) {
      if (not _boundingConstant->boundExists(m->k)) {
         return false;
      }
   }
   return true;
}

//------------------------------------------------------------------------------
bool SplittingFunctionFSR::rejectEvent(const Configuration& configN, const double TauN, const double z, const double phi, const RegionInfo& map)
{
   assert(map.hasFinalStateSingularity());
   if (not _mapping->isValidPhiN(configN, TauN, z, phi, map.k)) {
      return true;
   }
   addPassedZValue(z);

   double ratio;
   double numerator = sumAllowedSplittings(TauN, z, map);
   ratio = numerator / bound(map.k, TauN, z);
   if (ratio > 1.) {
      Log::warning("SplittingFunctionFSR::rejectEvent") << "In " << _mapping->name() << " Got a z = " << z << " for tauN = " << TauN << " that violates bound with ratio = " << ratio << "." << Log::end;
   }
   return randomInterval() > ratio;
}

//------------------------------------------------------------------------------
void SplittingFunctionFSR::printInfo() const
{
   Log::info() << "SplittingFunctionFSR "<< _mapping->name()<< Log::end;
   Log::info() << "Number of calls to normalized(): " << _numNormalizedCalled << Log::end;
   Log::info() << "Number of calls to splitting() : " << _numSplittingCalled << Log::end;
   Log::info() << "Number of calls to reject() : " << _numRejectCalled << Log::end;
   _boundingConstant->printInfo();
}

//------------------------------------------------------------------------------
//This returns the splitting function for FSR emissions
double SplittingFunctionFSR::splitting(Signature::const_iterator emitted, Signature::const_iterator sister, const double TauN, const double z) const
{
   double alphas = 1.0;
   if (_useRunningAlphas) {
      double kT = _mapping->kTEmission(*sister, TauN, z);
      // cutoff alphas running at 500 Mev
      alphas = (kT > 0.5) ? (*_alphas) (kT) : (*_alphas)(0.5);
   }
   if (emitted->isGluon() && sister->isGluon()) {
      //g->gg splitting
      if (z < 0.5)
         return 0.;
      return QCD::OneMinusZPgg(z) * alphas / (1. - z);
   } else if (emitted->isGluon()) {
      //q->qg splitting
//      if (z > 0.5)
//         return 0.;
      return QCD::OneMinusZPqq(z) * alphas / (1. - z);
   } else if (sister->isGluon()) {
      //q->gq splitting
//      if (z < 0.5)
//         return 0.;
      return QCD::OneMinusZPqg(z) * alphas / (1. - z);
   } else {
      //g->qqbar splitting
      return QCD::OneMinusZPgq(z) * alphas / (1. - z);
   }
}


//------------------------------------------------------------------------------
// The sum over the various FSR splittings
double SplittingFunctionFSR::sumAllowedSplittings(const double tauN, const double z, const RegionInfo& regionInfo) const
{
   //Find all RegionInfo's that correspond to a given mother and Phi0 Signature
   //First, find all splittings given Phi0 (note that they will all have the same x
   std::list<RegionInfo>  possibleSplittings = _mapping->getAllowedSplittings(regionInfo);
   assert(possibleSplittings.begin() != possibleSplittings.end());

   double result = 0.;
   for (std::list<RegionInfo>::iterator it = possibleSplittings.begin(); it != possibleSplittings.end(); ++it) {
      double tmp = splitting(it->i, it->j,tauN, z);
      if (not std::isfinite(tmp)) {
         Log::error("SplittingFunctionFSR::sumAllowedSplittings") << "In " << _mapping->name() << " splitting "<<*(it->k)<<" -> "<<*(it->j)<<" "<<*(it->i)<<" is not finite for tauN = "<<tauN<<" z= "<<z<< Log::end;
      }
      result += tmp;
   }
   return result;
}

//------------------------------------------------------------------------------
double SplittingFunctionFSR::bound(Signature::const_iterator mother, const double TauN, double z) const
{
   // Only include the bounding constant after initialization
   if (_isInitialized) {
      size_t zBin = _boundingConstant->binZ(z);
      size_t tauNBin = _boundingConstant->binTauN(TauN);
      return _boundingConstant->bound(mother, tauNBin, zBin) / (1. - z);
   }
   return 1. / (1. - z);
}

//------------------------------------------------------------------------------
double SplittingFunctionFSR::intBound(Signature::const_iterator mother, const double TauN, double zMin, double z) const
{
   size_t binTauN = _boundingConstant->binTauN(TauN);

   // Find the zbin containing zlow
   size_t bin = _boundingConstant->binZ(zMin);

   // Find the zbin containing z
   size_t binZ = _boundingConstant->binZ(z);

   double result = 0;

   // Sum up the contributions of all bins below the bin that z is in
   while (bin < binZ) {
      double zup = _boundingConstant->binZEdges()[bin + 1];
      const double bound = _boundingConstant->bound(mother, binTauN, bin);
      result += bound * log((1. - zMin) / (1. - zup));
      zMin = zup;
      ++bin;
   };

   // Add the contribution to the integral from the binEdge below z to z
   const double bound = _boundingConstant->bound(mother, binTauN, binZ);
   result += bound * log((1. - zMin) / (1. - z));
   return result;
}

//------------------------------------------------------------------------------
double SplittingFunctionFSR::intInvBound(Signature::const_iterator mother, double intFraction, const double TauN, double zMin, double zMax) const
{
   size_t binTauN = _boundingConstant->binTauN(TauN);

   if (zMin > zMax) {
      Log::error("SplittingFunctionFSR::intInvBound") << "In " << _mapping->name() << " Got zmin = " << zMin << " > zmax = " << zMax << Log::end;
   }
   // Get the target for the integral value
   double integral = intFraction * intBound(mother, TauN, zMin, zMax);

   // Find the bin inside of which the integral exceeds the target integral value
   size_t bin = _boundingConstant->binZ(zMin);

   double zup = _boundingConstant->binZEdges()[bin + 1];
   double bound = _boundingConstant->bound(mother, binTauN,  bin);
   double intBelow = bound * log((1. - zMin) / (1. - zup));
   while (intBelow < integral) {
      integral -= intBelow;
      zMin = zup;
      ++bin;
      zup = _boundingConstant->binZEdges()[bin + 1];
      bound = _boundingConstant->bound(mother, binTauN, bin);
      intBelow = bound * log((1. - zMin) / (1. - zup));
   }

   // Return the final z value by inverting the remaining integral inside the current bin
   return 1. - exp(-integral / bound) * (1. - zMin);
}

//------------------------------------------------------------------------------
///This returns a z value distributed according to bounding function
double SplittingFunctionFSR::pickZ(Signature::const_iterator mother, const double TauN, double zMin, double zMax) const
{
   return intInvBound(mother, randomInterval(), TauN, zMin, zMax);
}

} // namespace Geneva
