//------------------------------------------------------------------------------
/// \file SplittingFunctionISR.cpp
//
// Author(s):
//    Christian Bauer, Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 LBNL, CERN
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class SplittingFunctionISR
//------------------------------------------------------------------------------

#include "Geneva/Components/SplittingFunctionISR.hpp"
#include "Geneva/Components/ISplittingMapping.hpp"
#include "Geneva/Components/RegionInfoList.hpp"
#include "Geneva/Components/BoundingConstantsISR.hpp"
#include "Geneva/Components/IPDF.hpp"
#include "Geneva/Components/CouplingStrong.hpp"

#include "Geneva/Core/Physics/QCD.hpp"
#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/Random.hpp"
#include "Geneva/Core/Utils/NumericsUtils.hpp"
#include "Geneva/Core/Utils/FileUtils.hpp"

#include <fstream>
#include <stdexcept>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_blas.h>

namespace Geneva
{

namespace GoldLinkerWorkaround
{
// somewhat ugly workaround to explicitly require a cblas symbol to force some
// linkers (like ld.gold linker) to link against libgslcblas
// this is really a bug in how some distributions build the gsl
double(*cblas_dnrm2_ptr)(const int, const double*, const int) = &cblas_dnrm2;
}

namespace
{

//------------------------------------------------------------------------------
// param structure for the gsl functions
struct SplittingFunctionISRParameters
{
   RegionInfo* regionInfo;
   size_t binTauN;
   size_t binX;
   size_t binZ;
   SplittingFunctionISR* splittingFunction;
};

//------------------------------------------------------------------------------
// Wrapper that points to member function
// Trick: The SplittingFunction is an element of the gsl_f_pars struct
// so we can tease the value of the function out of there.
double SplittingFunctionGSLWrapperISR(const gsl_vector* var, void* pp)
{
   SplittingFunctionISRParameters* p = (SplittingFunctionISRParameters*)pp;
   return p->splittingFunction->ratioGSL(gsl_vector_get(var, 0), gsl_vector_get(var, 1), gsl_vector_get(var, 2), p);
}

} // anonymous namespace

//------------------------------------------------------------------------------
SplittingFunctionISR::SplittingFunctionISR(const OptionMap& optMap, ISplittingMapping* mapping, double Ecm, CouplingStrong* alphas, IPDF* pdf, double minX)
   : SplittingFunctionBase(optMap, mapping, alphas, Ecm), _pdf(pdf)
{
   //Store the random state, so that we get the same sequence whether we create the bounding grid or not
   RandomState ranstate = getRandomState();
   _useRunningPDFs = optMap.getOption<bool>("SplittingFunction::runningPDFs");

   _setupFile = getSetupFilename(optMap, format_tag(_mapping->name()) + format_tag("ISR"));
   std::ifstream setupFile(_setupFile.c_str());
   if (setupFile.good()) {  //Read from the setup File
      _boundingConstant = new BoundingConstantsISR(setupFile);
      //Check that we have all necessary bounding constants
      if (haveAllBoundingConstants()) {
         _isInitialized = true;
         Log::verboseInfo() << "Loaded ISR bounding function for "<< _mapping->name() <<" from file " << _setupFile << Log::end;
      } else {
         Log::warning("SplittingFunctionISR") << "Not all bounding constants for "<< _mapping->name() <<" found in file " << _setupFile << Log::end;
         _isInitialized = false;
      }
   } else {
      size_t numTauNBins = 1;
      if (_useRunningPDFs or _useRunningAlphas) numTauNBins = 20;
      _boundingConstant = new BoundingConstantsISR(numTauNBins, 20, 20, 0.1, Ecm, 0., 1., minX, 1.);
      _isInitialized = false;
   }
   if (not _isInitialized) {
      initialize();
      _isInitialized = true;
   }
   double boundFactor = optMap.getOption<double>("SplittingFunction::boundFactor");
   if (boundFactor < 1.0) {
      Log::fatal("SplittingFunction") << "SplittingFunctionISR::boundFactor must be larger or equal to 1" << Log::end;
   }
   _boundingConstant->setBoundFactor(boundFactor);
   //Restore the random state, so that we get the same sequence whether we create the bounding grid or not
   initializeRandomState(ranstate);
}

//------------------------------------------------------------------------------
SplittingFunctionISR::~SplittingFunctionISR()
{
   delete _boundingConstant;
}

//-------------------------------------------------------------------------------------
/**
 * Implements the function p(PhiNplus1)
 * The z values of this function are distributed according to the splitting function f(z)
 * The phi values are distributed flat
 * It gives unity if integrated over all allowed values of z and phi, and summed over the allowed splittings
 *
 * \param PhiNplus1 The (N+1)-body phase space point
 *
 * This function does not require a (N+1)-body point that gives rise to a valid splitting.
 * If the (N+1)-body point is invalid, it returns zero
 */
double SplittingFunctionISR::normalized(const PhiNplus1& phiNplus1)
{
   _numNormalizedCalled++;
   //Calculate the ratio of the bounding integral ot the true integral statistically
   double TauN, z, phi;
   RegionInfo regionInfo;
   Configuration configN;

   // check that clustering is allowed by QCD flavor
   if (invalidFlavorProjection(_mapping->sigsNmin1(), phiNplus1)) {
      return 0.;
   }

   if (not _mapping->calcPhiNValues(phiNplus1, configN, TauN, z, phi, regionInfo)) {
      return 0.;
   }

   //If the resulting PhiN, TauN, z, phi values are not valid for the splitting, return 0
   if (not _mapping->isValidPhiN(configN, TauN, z, phi, regionInfo.k)) {
      return 0.;
   }

   // Now that I have valid values, calculate the splitting function
   try {

      // Determine the value of phiMin and phiMax
      double phiMin = _mapping->phiMin(configN, TauN, regionInfo.k, *regionInfo.i);
      double phiMax = _mapping->phiMax(configN, TauN, regionInfo.k, *regionInfo.i);

      // Determine the value of zMin and zMax
      double zMin = _mapping->zMin(configN, TauN, phi, regionInfo.k, *regionInfo.i);
      double zMax = _mapping->zMax(configN, TauN, phi, regionInfo.k, *regionInfo.i);
      forceInRange(z, zMin, zMax, "SplittingFunctionISR::normalized");

      //Call the splitting function
      _numSplittingCalled++;

      double result = 0.0;

      //Determine the value of x
      double x;
      if (regionInfo.k->isInitialA()) {
         x = 2. * configN.initialA()->energy() / _cmEnergy;
      } else {
         x = 2. * configN.initialB()->energy() / _cmEnergy;
      }
      forceGreaterEqual(z, x, "SplittingFunctionISR::normalized");

      result = splitting(regionInfo.i, regionInfo.j, regionInfo.k, TauN, z, x);

      //If that is zero, no need to determine the normalization factor
      if (result == 0) {
         return 0.;
      }

      // Normalize to the total integral of the bounding function
      result /= intBound(regionInfo.k, TauN, zMin, zMax, x);

      // Count the number of times I am creating an invalid configuration by randomly choosing z
      size_t intRatio = 1;
      double randomZ = pickZ(regionInfo.k, TauN, x, zMin, zMax);
      double randomPhi = phiMin + randomInterval() * (phiMax - phiMin);
      result /= (phiMax - phiMin);

      //Keep track which z values I have tried
      initializeFailedZValues(zMin, zMax);
      unsigned int adjustZMinMax = 0;

      while (rejectEvent(configN, TauN, randomZ, randomPhi, regionInfo)) {
         _numRejectCalled++;
         //Increase the counter
         intRatio++;
         //Create new z and phi values
         randomZ = pickZ(regionInfo.k, TauN, x, zMin, zMax);
         randomPhi = phiMin + randomInterval() * (phiMax - phiMin);

         if (intRatio % 5000 == 0) {
            adjustZMinMax++;
            std::pair<double, double> zMinMax = adjustZMinZMax(z, zMin, zMax, phiMin, phiMax, TauN, configN, regionInfo, _mapping);
            zMin = zMinMax.first;
            zMax = zMinMax.second;
            intRatio = 1;
            initializeFailedZValues(zMin, zMax);
         }
         //If we have adjusted the zMax value more than 100 times, throw an error
         if (adjustZMinMax > 100) {
            Log::error("SplittingFunctionISR::normalized") << "In " << _mapping->name() << " Adjusted the zMin too many times.\n The limits are {zMin, zMax} = {" << zMin << ", " << zMax << "}" << "{phiMin, phiMax} = {" << phiMin << ", " << phiMax << "}, and the event is\n" << configN << Log::end;
         }
      }
      //Multiply the result by the ratio of the true integral over the integral of the bounding function
      result *= static_cast<double>(intRatio);
      if (result == 0.) {
         return 0.;
      }

      result /= jacobian(configN, TauN, z, phi, regionInfo);
      result /= static_cast<double>(_mapping->numQCDN() + 1);
      return result;

   } catch (const NormalError&) {
      _numSplittingFailures++;
      if (10000 * _numSplittingFailures > _numSplittingCalled && (_numSplittingCalled > 500000))
         Log::fatal("SplittingFunctionISR::normalized") << "In " << _mapping->name() << " " << _numSplittingFailures << " failures in " << _numSplittingCalled << " tries. Halting execution!" << Log::end;
      else
         Log::warning("SplittingFunctionISR::normalized") << "In " << _mapping->name() << " Error caught, returning 0 and continuing." << Log::end;
      return 0.;
   }
}

//------------------------------------------------------------------------------
// The sum over the various ISR splittings in a form that can be used in gsl. Since gsl minimizes a function, return the negative of the splitting
double SplittingFunctionISR::ratioGSL(double s0, double s1, double s2, void* params) const
{
   struct SplittingFunctionISRParameters* parameters = (struct SplittingFunctionISRParameters*) params;
   RegionInfo regionInfo = *parameters->regionInfo;

   double tMin = static_cast<double>(parameters->binTauN) / static_cast<double>(_boundingConstant->numTauNBins());
   double tMax = static_cast<double>(1 + parameters->binTauN) / static_cast<double>(_boundingConstant->numTauNBins());
   double vMin = static_cast<double>(parameters->binZ) / static_cast<double>(_boundingConstant->numZBins());
   double vMax = static_cast<double>(1 + parameters->binZ) / static_cast<double>(_boundingConstant->numZBins());
   double wMin = static_cast<double>(parameters->binX) / static_cast<double>(_boundingConstant->numXBins());
   double wMax = static_cast<double>(1 + parameters->binX) / static_cast<double>(_boundingConstant->numXBins());
   double t = (tMin + tMax) / 2. - (tMax - tMin) * atan(s0) / math_constants::pi();
   double v = (vMin + vMax) / 2. - (vMax - vMin) * atan(s1) / math_constants::pi();
   double w = (wMin + wMax) / 2. - (wMax - wMin) * atan(s2) / math_constants::pi();

   return -ratio(_boundingConstant->tauNVal(t), _boundingConstant->zVal(v, w), _boundingConstant->xVal(w), &regionInfo);
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

double SplittingFunctionISR::ratio(const double tauN, const double z, const double x, RegionInfo* m) const
{
   return sumAllowedSplittings(tauN, z, x, *m) / bound(m->k, tauN, z, x);
}

//------------------------------------------------------------------------------
void SplittingFunctionISR::initialize()
{
   Log::info("SplittingFunctionISR::initialize") << "Determining upper-bounding grids for splitting functions for "<< _mapping->name() <<" ..." << Log::end;

   //Loop over all possible mappings
   std::list<RegionInfo> Mappings = _mapping->getRegionInfoList(). getIsrRegionInfoList();
   for (std::list<RegionInfo>::iterator m = Mappings.begin(); m != Mappings.end(); ++m) {
      if (not (m->k->isInitialA() or m->k->isInitialB())) {
         Log::fatal("SplittingFunctionISR::init") << "In " << _mapping->name() << " Only works for ISRA or ISRB" << Log::end;
      }
      std::string rad = "";
      if (m->k->isInitialA()) {
         rad = "ISRA";
      } else {
         rad = "ISRB";
      }
      //Skip it if we already have a bounding constant for it
      if (_boundingConstant->boundExists(m->k)) {
         continue;
      }

      //ISR emissions
      for (size_t tauNBin = 0; tauNBin < _boundingConstant->numTauNBins(); tauNBin++) {
         double tMin = static_cast<double>(tauNBin) / static_cast<double>(_boundingConstant->numTauNBins());
         double tMax = static_cast<double>(tauNBin + 1) / static_cast<double>(_boundingConstant->numTauNBins());
         for (size_t xBin = 0; xBin < _boundingConstant->numXBins(); xBin++) {
            double wMin = static_cast<double>(xBin) / static_cast<double>(_boundingConstant->numXBins());
            double wMax = static_cast<double>(xBin + 1) / static_cast<double>(_boundingConstant->numXBins());
            for (size_t zBin = 0; zBin < _boundingConstant->numZBins(); zBin++) {
               double vMin = static_cast<double>(zBin) / static_cast<double>(_boundingConstant->numZBins());
               double vMax = static_cast<double>(zBin + 1) / static_cast<double>(_boundingConstant->numZBins());
               double tStart = tMin + 0.5 * (tMax -tMin);
               double wStart = wMin + 0.5 * (wMax - wMin);
               double vStart = vMin + 0.5 * (vMax - vMin);
               double num = 0.;
               try {
                  num = sumAllowedSplittings(_boundingConstant->tauNVal(tStart), _boundingConstant->zVal(vStart, wStart), _boundingConstant->xVal(wStart), *m);
               } catch (const NormalError&) {
                  Log::warning("SplittingFunctionISR::initialize")
                     << "In " << _mapping->name()
                     << " Error caught. Continuing with result 0 for TauN = " << _boundingConstant->tauNVal(tStart)
                     << " z = " << _boundingConstant->zVal(vStart, wStart)
                     << " x = " << _boundingConstant->xVal(wStart) << Log::end;
               }
               double ratioStart =  num / bound(m->k,_boundingConstant->tauNVal(tStart), _boundingConstant->zVal(vStart, wStart), _boundingConstant->xVal(wStart));

               //Now try positions inside the box to find a good starting point
               for (double tPos = 0.0001; tPos <= 10; tPos+=9.9998/10) {
                  double t = tMin + tPos / 10 * (tMax - tMin);
                  for (double wPos = 0.0001; wPos <= 10; wPos+=9.9998/10) {
                     double w = wMin + wPos / 10 * (wMax - wMin);
                     for (double vPos = 0.0001; vPos <= 10; vPos+=9.9998/10) {
                        double v = vMin + vPos /10 * (vMax - vMin);
                        double tmp = 0.;
                        try {
                           tmp = sumAllowedSplittings(_boundingConstant->tauNVal(t), _boundingConstant->zVal(v, w), _boundingConstant->xVal(w), *m);
                        } catch (const NormalError&) {
                           Log::warning("SplittingFunctionISR::initialize")
                              << "In " << _mapping->name()
                              << " Error caught. Continuing with result 0 for TauN = " << _boundingConstant->tauNVal(t)
                              << " z = " << _boundingConstant->zVal(v, w)
                              << " x = " << _boundingConstant->xVal(w) << Log::end;
                        }
                        double ratio = tmp / bound(m->k,_boundingConstant->tauNVal(t), _boundingConstant->zVal(v, w), _boundingConstant->xVal(w));
                        if (ratio > ratioStart) {
                           ratioStart = ratio;
                           tStart = t;
                           wStart = w;
                           vStart = v;
                        }
                     }
                  }
               }
               if (ratioStart == 0.) {
                  _boundingConstant->informOfRatio(m->k, tauNBin, zBin, xBin, 0.);
                  continue;
               }
               //Now run the gsl routing to find the maximum
               //Note that for the gsl routine, we make a variable transformation
               //to the variables
               //   s0 = Tan[Pi(0.5 - (t-tMin)/(tMax-tMin))]
               //   s1 = Tan[Pi(0.5 - (v-vMin)/(vMax-vMin))]
               //   s2 = Tan[Pi(0.5 - (w-wMin)/(wMax-wMin))]
               //since these variables are unbounded
               double s0 = tan(math_constants::pi() * (0.5 - (tStart - tMin) / (tMax - tMin)));
               double s1 = tan(math_constants::pi() * (0.5 - (vStart - vMin) / (vMax - vMin)));
               double s2 = tan(math_constants::pi() * (0.5 - (wStart - wMin) / (wMax - wMin)));

               gsl_vector* variables;
               variables = gsl_vector_alloc(3);
               gsl_vector_set(variables, 0, s0);
               gsl_vector_set(variables, 1, s1);
               gsl_vector_set(variables, 2, s2);

               gsl_vector* stepSizes;
               stepSizes = gsl_vector_alloc(3);
               gsl_vector_set(stepSizes, 0, .001);
               gsl_vector_set(stepSizes, 1, .001);
               gsl_vector_set(stepSizes, 2, .001);

               const gsl_multimin_fminimizer_type* T =
               gsl_multimin_fminimizer_nmsimplex2rand;
               gsl_multimin_fminimizer* s = NULL;

               SplittingFunctionISRParameters parameters = {&(*m), tauNBin,  xBin, zBin, this};

               gsl_multimin_function F = {&SplittingFunctionGSLWrapperISR, 3, &parameters};

               s = gsl_multimin_fminimizer_alloc(T, 3);
               gsl_multimin_fminimizer_set(s, &F, variables, stepSizes);

               int iter = 0;
               bool success;

               do {
                  iter++;
                  double sInit = s->fval;

                  gsl_multimin_fminimizer_iterate(s);

                  double sFinal = s->fval;

                  success = abs((sInit - sFinal) / sInit) < 0.000001;
               } while (!success && iter < 1000);

               if (!success) {
                  Log::warning("SplittingFunctionISR::init") << "In " << _mapping->name() << " Did not find minimum in 1000 iterations for TauN= "<< _boundingConstant->tauNVal(tStart)<<" z = "<< _boundingConstant->zVal(vStart, wStart)<<" x = "<< _boundingConstant->xVal(wStart)<<". Using last value = "<<  -s->fval <<" instead" << Log::end;
               }

               _boundingConstant->informOfRatio(m->k, tauNBin, zBin, xBin, -s->fval);

               gsl_vector_free(variables);
               gsl_vector_free(stepSizes);
               gsl_multimin_fminimizer_free(s);
            }
         }
      }
      Log::info("SplittingFunctionISR::init") << "Finished ISR bounding constant for "<< _mapping->name() <<" for id " << m->k->code() << " " << rad << Log::end;
   }

   std::ofstream setupFile(_setupFile.c_str());
   if (setupFile.good()) {
      Log::verboseInfo() << "Writing bounding constant for ISR to file " << _setupFile << Log::end;
      _boundingConstant->write(setupFile);
   } else {
      Log::warning("SplittingFunctionISR::init") << "In " << _mapping->name() << " Problem in writing bounding constant" << Log::end;
   }
   setupFile.close();
}

//------------------------------------------------------------------------------
bool SplittingFunctionISR::haveAllBoundingConstants() const
{
   //Loop over all possible ISR mappings
   std::list<RegionInfo> Mappings = _mapping->getRegionInfoList().getIsrRegionInfoList();
   for (std::list<RegionInfo>::iterator m = Mappings.begin(); m != Mappings.end(); ++m) {
      if (not _boundingConstant->boundExists(m->k)) {
         return false;
      }
   }
   return true;
}

//------------------------------------------------------------------------------
bool SplittingFunctionISR::rejectEvent(const Configuration& configN, const double TauN, const double z, const double phi, const RegionInfo& map)
{
   assert(map.hasInitialStateSingularity());
   if (not _mapping->isValidPhiN(configN, TauN, z, phi, map.k)) {
      return true;
   }
   addPassedZValue(z);

   double ratio;
   double numerator = 0.;
   double x = 0.;
   if (map.k->isInitialA()) {
      x = 2. * configN.initialA()->energy() / _cmEnergy;
      numerator = sumAllowedSplittings(TauN, z, x, map);
   } else if (map.k->isInitialB()) {
      x = 2. * configN.initialB()->energy() / _cmEnergy;
      numerator = sumAllowedSplittings(TauN, z, x, map);
   } else {
      Log::fatal("SplittingFunctionISR::rejectEvent") << "In " << _mapping->name() << " Emitter for ISR needs to be either initialA or initialB" << Log::end;
   }
   ratio = numerator / bound(map.k, TauN, z, x);

   if (ratio > 1.) {
      Log::warning("SplittingFunctionISR::rejectEvent") << "In " << _mapping->name() << " Got a z = " << z << " for tauN = " << TauN << ", x = " << x << " that violates bound with ratio = " << ratio << "." << Log::end;
   }
   return randomInterval() > ratio;
}

//------------------------------------------------------------------------------
void SplittingFunctionISR::printInfo() const
{
   Log::info() << "SplittingFunctionISR "<< _mapping->name()<< Log::end;
   Log::info() << "Number of calls to normalized(): " << _numNormalizedCalled << Log::end;
   Log::info() << "Number of calls to splitting() : " << _numSplittingCalled << Log::end;
   Log::info() << "Number of calls to reject() : " << _numRejectCalled << Log::end;
   _boundingConstant->printInfo();
}


//------------------------------------------------------------------------------
//This returns the splitting function for ISR emissions
double SplittingFunctionISR::splitting(Signature::const_iterator emitted, Signature::const_iterator sister, Signature::const_iterator mother, const double TauN,  const double z, const double x) const
{
   double kT = _mapping->kTEmission(*mother, TauN, z);
   double alphas = 1.0;
   if (_useRunningAlphas) {
      // cutoff alphas running at 500 Mev
      alphas = (kT > 0.5) ? (*_alphas) (kT) : (*_alphas)(0.5);
   }
   //TODO: Magic Number
   double muF = 90.;
   if (_useRunningPDFs) {
      muF = kT;
   }
   double pdfNum = 0.;
   double pdfDen = 0.;
   if (mother->isInitialA()) {
      if (z < x) {
         Log::error("SplittingFunctionISR::splittingISR") << "In " << _mapping->name() << " Invalid A z = " << z << " < x = " << x << Log::end;
      }
      pdfNum = _pdf->pdfA(*sister, x / z, muF);
      if (compareToZero(pdfNum) == 0) {
         return 0.0;
      }
      pdfDen = _pdf->pdfA(*mother, x, muF);
      if (compareToZero(pdfDen) == 0) {
         return 0.0;
      }
   } else {
      if (z < x) {
         Log::error("SplittingFunctionISR::splittingISR") << "In " << _mapping->name() << " Invalid B z = " << z << " < x = " << x << Log::end;
      }
      pdfNum = _pdf->pdfB(*sister, x / z, muF);
      if (compareToZero(pdfNum) == 0) {
         return 0.0;
      }
      pdfDen = _pdf->pdfB(*mother, x, muF);
      if (compareToZero(pdfDen) == 0) {
         return 0.0;
      }
   }
   double pdfRatio = pdfNum / pdfDen;
   assert(pdfRatio != 0.0);
   if (pdfRatio < 0.0) {
      Log::warning("SplittingFunctionISR::splitting") << "In " << _mapping->name() << " Negative pdf value!. Setting it to zero. f(" << *sister << "," << x / z << "," << muF << ") = " << pdfNum << " /  f(" << *mother << "," << x << "," << muF << ") = " << pdfDen <<Log::end;
      return 0.0;
   }
   if (emitted->isGluon() && sister->isGluon()) {
      //g->(g)g splitting
      return QCD::OneMinusZPgg(z) / (1. - z) * alphas * pdfRatio;
   } else if (emitted->isGluon()) {
      //q->(q)g splitting
      return QCD::OneMinusZPqq(z) / (1. - z) * alphas * pdfRatio;
   } else if (sister->isGluon()) {
      //g->(qbar)q splitting
      return QCD::OneMinusZPgq(z) / (1. - z) * alphas * pdfRatio;
   } else {
      //q->(g)q splitting
      return QCD::OneMinusZPqg(z) / (1. - z) * alphas * pdfRatio;
   }
}

//------------------------------------------------------------------------------
// The sum over the various ISR splittings
double SplittingFunctionISR::sumAllowedSplittings(const double tauN, const double z, const double x, const RegionInfo& regionInfo) const
{
   //Find all RegionInfo's that correspond to a given mother and Phi0 Signature
   //First, find all splittings given Phi0 (note that they will all have the same x
   std::list<RegionInfo>  possibleSplittings = _mapping->getAllowedSplittings(regionInfo);
   assert(possibleSplittings.begin() != possibleSplittings.end());

   double result = 0.;
   for (std::list<RegionInfo>::iterator it = possibleSplittings.begin(); it != possibleSplittings.end(); ++it) {
      double tmp = splitting(it->i, it->j, it->k, tauN, z, x);
      if (not std::isfinite(tmp)) {
         splitting(it->i, it->j, it->k, tauN, z, x);
         Log::error("SplittingFunctionISR::sumAllowedSplittings") << "In " << _mapping->name() << " splitting "<<*(it->k)<<" -> "<<*(it->j)<<" "<<*(it->i)<<" is not finite for tauN = "<<tauN<<" z= "<<z<<" x= "<<x<<Log::end;
      }
      result += tmp;
   }
   return result;
}

//------------------------------------------------------------------------------
double SplittingFunctionISR::bound(Signature::const_iterator mother, const double TauN, double z, double x) const
{
   // Only include the bounding constant after initialization
   if (_isInitialized) {
      size_t zBin = _boundingConstant->binZ(z, x);
      size_t xBin = _boundingConstant->binX(x);
      size_t tauNBin = _boundingConstant->binTauN(TauN);
      return _boundingConstant->bound(mother, tauNBin, zBin, xBin) / (1. - z);
   }
   return 1. / (1. - z);
}

//------------------------------------------------------------------------------
double SplittingFunctionISR::intBound(Signature::const_iterator mother, const double TauN, double zMin, double z, double x) const
{
   size_t binTauN = _boundingConstant->binTauN(TauN);

   size_t binX = _boundingConstant->binX(x);

   // Find the zbin containing zlow
   size_t bin = _boundingConstant->binZ(zMin, x);

   // Find the zbin containing z
   size_t binZ = _boundingConstant->binZ(z, x);

   double result = 0;

   // Sum up the contributions of all bins below the bin that z is in
   while (bin < binZ) {
      double zup = _boundingConstant->binZEdges(bin, x).second;
      const double bound = _boundingConstant->bound(mother, binTauN, bin, binX);
      result += bound * log((1. - zMin) / (1. - zup));
      zMin = zup;
      ++bin;
   };

   // Add the contribution to the integral from the binEdge below z to z
   const double bound = _boundingConstant->bound(mother, binTauN, binZ, binX);
   result += bound * log((1. - zMin) / (1. - z));
   return result;
}

//------------------------------------------------------------------------------
double SplittingFunctionISR::intInvBound(Signature::const_iterator mother, double intFraction, const double TauN, double x, double zMin, double zMax) const
{
   size_t binTauN = _boundingConstant->binTauN(TauN);
   size_t binX = _boundingConstant->binX(x);

   if (zMin > zMax) {
      Log::error("SplittingFunctionISR::intInvBound") << "In " << _mapping->name() << " Got zmin = " << zMin << " > zmax = " << zMax << Log::end;
   }

   // Get the target for the integral value
   double integral = intFraction * intBound(mother, TauN, zMin, zMax, x);

   // Find the bin inside of which the integral exceeds the target integral value
   size_t bin = _boundingConstant->binZ(zMin, x);

   double zup = _boundingConstant->binZEdges(bin, x).second;
   double bound = _boundingConstant->bound(mother, binTauN, bin, binX);
   double intBelow = bound * log((1. - zMin) / (1. - zup));
   while (intBelow < integral) {
      integral -= intBelow;
      zMin = zup;
      ++bin;
      zup = _boundingConstant->binZEdges(bin, x).second;
      bound = _boundingConstant->bound(mother, binTauN, bin, binX);
      intBelow = bound * log((1. - zMin) / (1. - zup));
   }

   // Return the final z value by inverting the remaining integral inside the current bin
   return 1. - exp(-integral / bound) * (1. - zMin);
}

//------------------------------------------------------------------------------
///This returns a z value distributed according to bounding function
double SplittingFunctionISR::pickZ(Signature::const_iterator mother, const double TauN, double x, double zMin, double zMax) const
{
   return intInvBound(mother, randomInterval(), TauN, x, zMin, zMax);
}

} // namespace Geneva
