//------------------------------------------------------------------------------
/// \file SplittingMappingppV0To1.cpp
//
// Author(s):
//    Christian Bauer
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class SplittingMappingppV0To1
//------------------------------------------------------------------------------

#include "Geneva/Components/SplittingMappingppV0To1.hpp"
#include "Geneva/Components/SignatureManager.hpp"

#include "Geneva/Core/Physics/Emission.hpp"
#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/Physics/Particle.hpp"
#include "Geneva/Core/Physics/ConfigurationUtils.hpp"
#include "Geneva/Core/Physics/ConfigurationExtensions.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"

namespace Geneva
{

//-----------------------------------------------------------------------------
SplittingMappingppV0To1::SplittingMappingppV0To1(const OptionMap& optMap)
{
   PdgID lepton;
   std::string opt = optMap.getOption("Process::DrellYan::DecayChannel");
   if (opt == "electrons")  lepton = PdgID::kElectron;
   else if (opt == "muons") lepton = PdgID::kMuon;
   else if (opt == "taus")  lepton = PdgID::kTau;
   else {
      Log::error("SplittingMappingppV1To2") << "Unknown DecayChannel " << opt << Log::end;
   }

   //Set up the RegionInfoList
   const int nlight = 5;
   _cmEnergy = optMap.getOption<double>("Ecm");
   _signaturesNmin1 = SignatureManager().getSignaturespp2ZJets(lepton, 0, nlight, true);
   _signaturesN = SignatureManager().getSignaturespp2ZJets(lepton, 1, nlight, true);
   _regionsInfo.setRegionInfoList(_signaturesNmin1, _signaturesN, nlight, false);
}

//------------------------------------------------------------------------------
bool SplittingMappingppV0To1::calcPhiNValues(const PhiNplus1& config, Configuration& configNmin1, double& Tau0, double& z, double& phi, RegionInfo& regionInfo)const
{
   assert(config.sizeFinal() == 3);
   configNmin1.clear();

   //Get the QCD particle and the momenta of the leptons
   Configuration::iterator emitted;
   FourVector leptons(0., 0., 0., 0.);
   FourVector negLeptonFV, posLeptonFV;
   PdgID negLeptonID, posLeptonID;
   for (Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it) {
      if (it->id().isLepton()) {
         leptons += it->p();
         if (it->id().isParticle()) {
            negLeptonFV = it->p();
            negLeptonID = it->id();
         } else {
            posLeptonFV = it->p();
            posLeptonID = it->id();
         }
      } else
         emitted = it;
   }

   Configuration::iterator sister;
   Emission::Type radType;
   double mV = leptons.magnitude();
   double YV = leptons.rapidity();
   double expY = exp(YV);
   double qPlus = mV / expY;
   double qMinus = mV * expY;
   double pPlus = emitted->p().plus();
   double pMinus = emitted->p().minus();
   if (pMinus / pPlus > expY * expY) {
      // p.rapidity() > YV
      z = qMinus / (qMinus + pMinus);
      sister = config.initialA();
      radType = Emission::kISRA;
   } else {
      z = qPlus / (qPlus + pPlus);
      sister = config.initialB();
      radType = Emission::kISRB;
   }
   //The phi angle is defined to be zero for px=0, py>0
   phi = atan2(emitted->px(), emitted->py());
   if (phi < 0)
      phi += 2. * math_constants::pi();

   //Finally, set the regionInfo
   std::list<RegionInfo> singularRegions = _regionsInfo.getSingularRegionsToSignature(Signature(config));
   std::list<RegionInfo>::iterator it;
   PdgID idA, idB;
   for (it = singularRegions.begin(); it != singularRegions.end(); ++it) {
      Emission::Type currentRadiation;
      if (it->k->isInitialA()) {
         currentRadiation = Emission::kISRA;
         idA = *(it->k);
         idB = it->k->antiPartner();
      } else if (it->k->isInitialB()) {
         idB = *(it->k);
         idA = it->k->antiPartner();
         currentRadiation = Emission::kISRB;
      } else {
         idA = config.initialA()->id();
         idB = config.initialB()->id();
         currentRadiation = Emission::kFSR;
      }
      if ((emitted->id() == *(it->i)) && (sister->id() == *(it->j)) && currentRadiation == radType) {
         break;
      }
   }
   if (it == singularRegions.end()) {
      return false;
   }

   regionInfo = *it;

   Tau0 = ZeroJettinesspp(config, YV)();

   //Calculate the luminosity of the underlying 0-jet emission
   double xA = mV / _cmEnergy * exp(YV);
   double xB = mV / _cmEnergy * exp(-YV);

   FourVector pABar(xA, 0., 0., xA);
   pABar *= _cmEnergy;
   Boost toZRestFrame = Boost(leptons).inverse();
   posLeptonFV.apply(toZRestFrame);
   negLeptonFV.apply(toZRestFrame);

   //Rescale the leptons to the correct center of mass energy
   negLeptonFV *= mV / 2. / negLeptonFV.p0();
   posLeptonFV *= mV / 2. / posLeptonFV.p0();

   configNmin1.clear();
   configNmin1.setInitialA(Particle(idA, mV / 2., 0., 0., mV / 2.));
   configNmin1.setInitialB(Particle(idB, mV / 2., 0., 0., - mV / 2.));
   configNmin1.insertFinal(Particle(negLeptonID, negLeptonFV));
   configNmin1.insertFinal(Particle(posLeptonID, posLeptonFV));

   BoostZHyp boostZ(sqrt(xA / xB));
   configNmin1.apply(boostZ);

   return true;
}

//-----------------------------------------------------------------------------
bool SplittingMappingppV0To1::generateSplit(const PhiN& phi0, const double Tau0, const double z, const double phi, const RegionInfo& map, PhiNplus1CMWritable& phiNp1) const
{
   double mV = (phi0.initialA()->p() + phi0.initialB()->p()).magnitude();
   double expY = sqrt(phi0.initialA()->energy() / phi0.initialB()->energy());

   PdgID idA, idB;
   double pPlus = 0.;
   double pMinus = 0.;
   double qPlus = 0.;
   double qMinus = 0.;
   double pPerpSq = 0.;

   phiNp1.clear();

   if (phi0.emitter()->isInitialA()) {
//      if (z < phi0.initialA()->p().minus() / _cmEnergy)
//         return false;
      pPlus = Tau0 / expY;
      pMinus = mV * expY * (1. / z - 1);
      pPerpSq = pPlus * pMinus;
      qPlus = sqrt(pPerpSq + mV * mV) / expY;
      qMinus = sqrt(pPerpSq + mV * mV) * expY;
      if (not (pPlus + qPlus <= _cmEnergy) || not (pMinus + qMinus <= _cmEnergy) || not (pMinus / pPlus >= expY * expY))
         return false;
      phiNp1.setInitialA(Particle(*map.j, (pMinus + qMinus) / 2., 0., 0., (pMinus + qMinus) / 2.));
      phiNp1.setInitialB(Particle(phi0.initialB()->id(), (pPlus + qPlus) / 2., 0., 0., -(pPlus + qPlus) / 2.));

   }
   if (phi0.emitter()->isInitialB()) {
//      if (z < phi0.initialB()->p().plus() / _cmEnergy)
//         return false;
      pMinus = Tau0 * expY;
      pPlus = mV / expY * (1. / z - 1);
      pPerpSq = pPlus * pMinus;
      qPlus = sqrt(pPerpSq + mV * mV) / expY;
      qMinus = sqrt(pPerpSq + mV * mV) * expY;
      if (not (pPlus + qPlus <= _cmEnergy) || not (pMinus + qMinus <= _cmEnergy) || not (pMinus / pPlus <= expY * expY))
         return false;
      phiNp1.setInitialA(Particle(phi0.initialA()->id(), (pMinus + qMinus) / 2., 0., 0., (pMinus + qMinus) / 2.));
      phiNp1.setInitialB(Particle(*map.j, (pPlus + qPlus) / 2., 0., 0., -(pPlus + qPlus) / 2.));
   }

   //Get the transverse components
   double px = sqrt(pPerpSq) * sin(phi);
   double py = sqrt(pPerpSq) * cos(phi);

   //Now add the new parton to the final state
   phiNp1.insertFinal(Particle(*map.i, (pPlus + pMinus) / 2., px, py, (pMinus - pPlus) / 2.));

   //Now add the leptons to the final state
   //First, get the total q momentum
   FourVector q((qPlus + qMinus) / 2., -px, -py, (qMinus - qPlus) / 2.);
   //I need to figure out the boost from the Phi0 q momentum to the Phi1 q momentum
   BoostZHyp toCM(sqrt(phi0.initialB()->energy()/phi0.initialA()->energy()));
   Boost toNewLab = Boost(q);
   FourVector qNew(0., 0., 0., 0.);
   for (Configuration::iterator it = phi0.beginFinal(); it != phi0.endFinal(); ++it) {
      if (it->id().isLepton()) {
         FourVector lep = it->p().transformed(toCM);
         lep.apply(toNewLab);
         phiNp1.insertFinal(Particle(it->id(), lep));
         qNew += lep;
      }
   }
#ifndef NDEBUG
   if (!checkMomentumConservation(phiNp1.config(), 1e-8, 1e-10))
      Log::warning("SplittingMappingppV0To1::generateSplit") << "Phi_N+1 point from splitting violates momentum conservation." << Log::end;
#endif
   phiNp1.jacobian() = jacobian(phi0, Tau0, z, phi, map);
   return true;
}

//------------------------------------------------------------------------------
bool SplittingMappingppV0To1::isValidPhiN(const Configuration& config, double Tau0, double z, double phi, Signature::const_iterator mother) const
{
   double mV = (config.initialA()->p() + config.initialB()->p()).magnitude();
   double expY = sqrt(config.initialA()->energy() / config.initialB()->energy());

   if (mother->isInitialA()) {
      if (z < config.initialA()->p().minus() / _cmEnergy)
         return false;
      double pPlus = Tau0 / expY;
      double pMinus = mV * expY * (1. / z - 1);
      double pPerpSq = pPlus * pMinus;
      double qPlus = sqrt(pPerpSq + mV * mV) / expY;
      double qMinus = sqrt(pPerpSq + mV * mV) * expY;
      return (pPlus + qPlus <= _cmEnergy) && (pMinus + qMinus <= _cmEnergy) && (pMinus / pPlus >= expY * expY);
   }
   if (mother->isInitialB()) {
      if (z < config.initialB()->p().plus() / _cmEnergy)
         return false;
      double pMinus = Tau0 * expY;
      double pPlus = mV / expY * (1. / z - 1);
      double pPerpSq = pPlus * pMinus;
      double qPlus = sqrt(pPerpSq + mV * mV) / expY;
      double qMinus = sqrt(pPerpSq + mV * mV) * expY;
      return (pPlus + qPlus <= _cmEnergy) && (pMinus + qMinus <= _cmEnergy) && (pMinus / pPlus <= expY * expY);
   }
   Log::error("SplittingFunctionZSplitSum::isValidPhi1") << "Got an invalid z: " << z << "Tau0 = " << Tau0 << " mV = " << mV << " expY = " << expY << Log::end;
   return false;
}

//------------------------------------------------------------------------------
double SplittingMappingppV0To1::kTEmission(PdgID motherID, double Tau0, double z) const
{
   double kT = sqrt(90.* Tau0 * (1.-z)/z);
   if ((motherID == PdgID::kBottom or motherID == PdgID::kBottomBar) and kT < 10.)
      kT = 10.;
   if ((motherID == PdgID::kCharm or motherID == PdgID::kCharmBar) and kT < 3.)
      kT = 3.;
   return kT;
}

//------------------------------------------------------------------------------
double SplittingMappingppV0To1::jacobian(const Configuration& Phi0, double Tau0, double z, double phi, const RegionInfo& map) const
{

   double qMinusBar = 2. * Phi0.initialA()->energy();
   double qPlusBar = 2. * Phi0.initialB()->energy();
   double mV2 = qPlusBar * qMinusBar;

   double jac1 = (2. * mV2) / (z * z * _cmEnergy * _cmEnergy) / (256. * math_constants::pi4()) / 4. / math_constants::pi();
   double jac2 = 1. / 16. / math_constants::pi2() * sqrt(mV2) / _cmEnergy / _cmEnergy;
   return jac1 / jac2;
}

//------------------------------------------------------------------------------
double SplittingMappingppV0To1::zMin(const Configuration& config, const double Tau0, const double phi, Signature::const_iterator mother, PdgID emittedID)
{
   double qMinusBar = 2. * config.initialA()->energy();
   double qPlusBar = 2. * config.initialB()->energy();
   double mV = sqrt(qPlusBar * qMinusBar);
   double expY = sqrt(qMinusBar / qPlusBar);
   return zMin(mV, expY, Tau0, mother);
}

//------------------------------------------------------------------------------
double SplittingMappingppV0To1::zMin(double mV, double expY, double Tau0, Signature::const_iterator mother)
{
   //Determine the lower limit of the z integration
   double zMin = 0.;
   //The formulas for ISRA and ISRB are the same, after the replacement YV->-YV
   if (mother->isInitialB()) {
      expY = 1. / expY;
   }
   double zlim = ((mV * Tau0) / (_cmEnergy * _cmEnergy * expY * expY - mV * mV - 2.*_cmEnergy * expY * Tau0 + mV * Tau0 + Tau0 * Tau0));
   if (zlim < 0) {
      zMin = 0.;
      return zMin;
   }
   if (zlim >= zMin)
      zMin = zlim;

   zlim = (2. * _cmEnergy * expY * mV + 2. * pow(expY, 2) * pow(mV, 2) + pow(expY, 2) * mV * Tau0);
   zlim /= (2. * (pow(_cmEnergy, 2) + 2. * _cmEnergy * expY * mV + pow(expY, 2) * mV * Tau0));
   zlim += sqrt((4. * pow(expY, 4) * pow(mV, 4) + 4. * _cmEnergy * pow(expY, 3) * pow(mV, 2) * Tau0 + pow(expY, 4) * pow(mV, 2) * pow(Tau0, 2)) /
                pow(pow(_cmEnergy, 2) + 2 * _cmEnergy * expY * mV + pow(expY, 2) * mV * Tau0, 2)) / 2.;
   // NOTE: negative zlim means there is no solution in [0,1], so return a lower limit > 1. to indicate invalid phase space
   if (zlim < 0) {
      zMin = 0.;
      return zMin;
   }
   if (zlim >= zMin)
      zMin = zlim;
   return zMin;
}

//------------------------------------------------------------------------------
double SplittingMappingppV0To1::zMax(const Configuration& config, const double Tau0, const double phi, Signature::const_iterator mother, PdgID emittedID)
{
   double qMinusBar = 2. * config.initialA()->energy();
   double qPlusBar = 2. * config.initialB()->energy();
   double mV = sqrt(qPlusBar * qMinusBar);
   return zMax(mV, Tau0);
}

//------------------------------------------------------------------------------
double SplittingMappingppV0To1::zMax(double mV, double Tau0)
{
   return mV / (mV + Tau0);
}

} // namespace Geneva
