//------------------------------------------------------------------------------
/// \file SplittingMappingppV1To2.cpp
//
// Author(s):
//    Christian Bauer
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class SplittingMappingppV1To2
//------------------------------------------------------------------------------

#include "Geneva/Components/SplittingMappingppV1To2.hpp"
#include "Geneva/Components/SignatureManager.hpp"
#include "Geneva/Components/Tau0FRMapping.hpp"
#include "Geneva/Components/FKSKinematics.hpp"

#include "Geneva/Core/Physics/Emission.hpp"
#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/Physics/PhiRadFKS.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
SplittingMappingppV1To2::SplittingMappingppV1To2(const OptionMap& optMap)
{
   PdgID lepton;
   std::string opt = optMap.getOption("Process::DrellYan::DecayChannel");
   if (opt == "electrons")  lepton = PdgID::kElectron;
   else if (opt == "muons") lepton = PdgID::kMuon;
   else if (opt == "taus")  lepton = PdgID::kTau;
   else {
      Log::error("SplittingMappingppV1To2") << "Unknown DecayChannel " << opt << Log::end;
   }

   //Set up the RegionInfoList
   int nlight = 5;
   _cmEnergy = optMap.getOption<double>("Ecm");
   _signaturesNmin1 = SignatureManager().getSignaturespp2ZJets(lepton, 1, nlight, true);
   _signaturesN = SignatureManager().getSignaturespp2ZJets(lepton, 2, nlight, true);
   _regionsInfo.setRegionInfoList(_signaturesNmin1, _signaturesN, nlight, false);

   _mapping = new Tau0FRMapping(_cmEnergy, optMap.getOption<bool>("NJettiness::Tau0NLO::Tau0FR::fixqT"));
}

//------------------------------------------------------------------------------
SplittingMappingppV1To2::~SplittingMappingppV1To2()
{
   delete _mapping;
}

//------------------------------------------------------------------------------
bool SplittingMappingppV1To2::calcPhiNValues(const PhiNplus1& Phi2, Configuration& Phi1, double& Tau1, double& z, double& phi, RegionInfo& regionInfo)const
{
   // return false if there is no singular region corresponding to Phi2
   std::list<RegionInfo> singularRegions = _regionsInfo.getSingularRegionsToSignature(Signature(Phi2));
   std::list<RegionInfo>::iterator it;
   for (it = singularRegions.begin(); it != singularRegions.end(); ++it) {
      bool radTypeMatches;
      if (it->k->isInitialA())
         radTypeMatches = Phi2.sister()->isInitialA();
      else if (it->k->isInitialB())
         radTypeMatches = Phi2.sister()->isInitialB();
      else
         radTypeMatches = Phi2.sister()->isFinal();
      if ( (Phi2.emitted()->id() == *(it->i)) && (Phi2.sister()->id() == *(it->j)) && radTypeMatches) {
         break;
      }
   }
   if (it == singularRegions.end())
      return false;
   regionInfo = *it;

   //Determine the projected Configuration. Return false if the projection fails
   if (!_mapping->project(Phi2, Phi1))
      return false;

   // obtain values for Tau1, z, phi
   Tau1 = NJettinessppRecursive(Phi2)(1);
   if (Phi2.sister()->isInitialA()) {
      z = Phi1.initialA()->p().minus() / (Phi1.initialA()->p().minus() + Phi2.emitted()->p().minus());
      phi = Phi2.emitted()->p().phi02pi();
   } else if (Phi2.sister()->isInitialB()) {
      z = Phi1.initialB()->p().plus() / (Phi1.initialB()->p().plus() + Phi2.emitted()->p().plus());
      phi = Phi2.emitted()->p().phi02pi();
   } else {
      z = Phi2.sister()->energy() / (Phi2.emitted()->energy() + Phi2.sister()->energy());
      FourVector p12Lab = Phi2.emitted()->p() + Phi2.sister()->p();
      const Boost fromLabtoP12RestFrame = Boost(p12Lab).inverse();
      const FourVector p2Inp12RestFrame = fromLabtoP12RestFrame(Phi2.emitted()->p());
      const ThreeVector zDir(0, 0, 1);
      const Rotation alignP12ToZ(p12Lab.pvec(), zDir);
      const FourVector p2Rotated = alignP12ToZ(p2Inp12RestFrame);
      phi = p2Rotated.phi02pi();
   }

   return true;
}

//------------------------------------------------------------------------------
bool SplittingMappingppV1To2::isValidPhiN(const Configuration& Phi1, double Tau1, double z, double phi, Signature::const_iterator mother) const
{
   //To check if the resulting PhiNPlus1 phase space point is kinematically valid,
   //I will use the generateSplit function of the Tau0FR mapping, which is written
   //in terms of phiRad

   //Convert the Tau1, z, phi values into phiRad
   //Note that I am using kGluon as the PdgID's of the emitted and sister particle, since later I will need to find the closest QCD particle. The actual flavor is no important
   PhiRadFKS phiRad;
   _mapping->getPhiRad(Phi1, Tau1, z, phi, mother, PdgID::kGluon, PdgID::kGluon, phiRad);

   //Return false if we get unphysical values of xi or y
   if (phiRad.xi() != phiRad.xi() || phiRad.y() != phiRad.y() || phiRad.xi() < 0.  || phiRad.xi() > 1. || phiRad.y() < -1. || phiRad.y() > 1.)
      return false;

   //To test if the Phi_N+1 point is valid, perform the splitting.
   //Before splitting, check if the value of z is below x
   PhiNplus1CMWritable phi2;
   if (mother->isInitialA()) {
      double x = Phi1.initialA()->p().minus() / _cmEnergy;
      if (z < x)
         return false;
      if (phiRad.xi() > _mapping->getXiMax(PhiN(Phi1, Phi1.initialA()), phiRad.y()))
          return false;
      if (!_mapping->generateSplit(PhiN(Phi1, Phi1.initialA()), phiRad, phi2))
         return false;
   } else if (mother->isInitialB()) {
      double x = Phi1.initialB()->p().plus() / _cmEnergy;
      if (z < x)
         return false;
      if (phiRad.xi() > _mapping->getXiMax(PhiN(Phi1, Phi1.initialB()), phiRad.y()))
         return false;
      if (!_mapping->generateSplit(PhiN(Phi1, Phi1.initialB()), phiRad, phi2))
          return false;
   } else {
      Configuration::iterator emitter = Phi1.find(mother);
      if (phiRad.xi() > _mapping->getXiMax(PhiN(Phi1, emitter), phiRad.y()))
         return false;
      if (!_mapping->generateSplit(PhiN(Phi1, emitter), phiRad, phi2))
         return false;
   }

   //There are some xi values for which the splitting succeeds, but the projection fails (these are points where xi > xiMax, but xi is still valid. This can happen since we put some arbitrary cut on xi and move everything above there into the non-singular)
   if (not _mapping->testProject(phi2))
      return false;

   // check that emitted and sister are the closest according to our clustering
   Configuration::iterator emitted, sister;
   NJettinessppRecursive(phi2.config()).cluster(emitted, sister);


   return ((emitted == phi2.emitted() && sister == phi2.sister()) or (sister == phi2.emitted() && emitted == phi2.sister()));
}

//------------------------------------------------------------------------------
bool SplittingMappingppV1To2::generateSplit(const PhiN& phi1, const double Tau1, const double z, const double phi,  const RegionInfo& map, PhiNplus1CMWritable& phi2) const
{
   PhiRadFKS phiRad;
   double jacobian = _mapping->getPhiRad(phi1, Tau1, z, phi, map.k, *map.i, *map.j, phiRad);

   if (not _mapping->generateSplit(phi1, phiRad, phi2))
      return false;

   phi2.jacobian() *= jacobian;
   return true;
}

//------------------------------------------------------------------------------
double SplittingMappingppV1To2::kTEmission(PdgID motherID, double Tau1, double z) const
{
   //TODO: fix this with the correct formula
   static bool ini=true;
   if (ini) {
      Log::verboseInfo("SplittingMappingppV1To2::kTEmission") <<" Using a fixed value of 90 for scales in 1->2 splitting "<<Log::end;
      ini=false;
   }
   return 90.;
}

//TODO: Try to avoid doing the whole splitting just to get the Jacobian
//------------------------------------------------------------------------------
double SplittingMappingppV1To2::jacobian(const Configuration& Phi1, double Tau1, double z, double phi, const RegionInfo& map) const
{
   double jacobian;
   //First, get the jacobian dPhi2 / dPhi1 / dPhiRad from th TauFR mapping
   //Then add the Jacobian dPhiRad / dTau1 / dz / dphi
   PhiRadFKS phiRad;
   jacobian = _mapping->getPhiRad(Phi1, Tau1, z, phi, map.k, PdgID::kUnknown, PdgID::kUnknown, phiRad);
   PhiNplus1CMWritable phi2;
   if (map.k->isInitialA()) {
      if (phiRad.radType() != PhiRadFKS::kISRA)
         Log::error("SplittingMappingppV1To2::jacobian") << "Wrong radiation type" << Log::end;
      if (!_mapping->generateSplit(PhiN(Phi1, Phi1.initialA()), phiRad, phi2))
         Log::error("SplittingMappingppV1To2::jacobian") << "No Jacobian exists" << Log::end;

   } else if (map.k->isInitialB()) {
      if (phiRad.radType() != PhiRadFKS::kISRB)
         Log::error("SplittingMappingppV1To2::jacobian") << "Wrong radiation type" << Log::end;
      if (!_mapping->generateSplit(PhiN(Phi1, Phi1.initialB()), phiRad, phi2))
         Log::error("SplittingMappingppV1To2::jacobian") << "No Jacobian exists" << Log::end;
   } else {
      Configuration::iterator emitter = Phi1.find(map.k);
      if (phiRad.radType() != PhiRadFKS::kFSR)
         Log::error("SplittingMappingppV1To2::jacobian") << "Wrong radiation type" << Log::end;
      if (!_mapping->generateSplit(PhiN(Phi1, emitter), phiRad, phi2))
         Log::error("SplittingMappingppV1To2::jacobian") << "No Jacobian exists" << Log::end;
   }
   jacobian *= phi2.jacobian();
   jacobian *= static_cast<double>(map.multRad);

   return jacobian;
}

//------------------------------------------------------------------------------
double SplittingMappingppV1To2::zMin(const Configuration& Phi1, const double Tau1, const double phi, Signature::const_iterator mother, PdgID emittedID)
{
   return _mapping->zMin(Phi1, Tau1, phi, mother, emittedID);
}

//------------------------------------------------------------------------------
double SplittingMappingppV1To2::zMax(const Configuration& Phi1, const double Tau1, const double phi, Signature::const_iterator mother, PdgID emittedID)
{
   return _mapping->zMax(Phi1, Tau1, phi, mother, emittedID);
}

double SplittingMappingppV1To2::phiMin(const Configuration& configNmin1, const double TauN, Signature::const_iterator mother, PdgID emittedID)
{
   return _mapping->phiMin(configNmin1, TauN, mother, emittedID);
}

double SplittingMappingppV1To2::phiMax(const Configuration& configNmin1, const double TauN, Signature::const_iterator mother, PdgID emittedID)
{
   return _mapping->phiMax(configNmin1, TauN, mother, emittedID);
}

//------------------------------------------------------------------------------
/**
 * Calculates the phi angle for a FSR emission in the CM frame.
 * It requires the emitter and fks parton %ThreeVectors to be defined in the CM frame
 * and to have unit lenght
 */
double SplittingMappingppV1To2::getPhiAngleFSR(const ThreeVector& fksDir, const ThreeVector& emitterDir) const
{
   // make sure we are working with unit vectors
   assert(compare(emitterDir.magnitude(), 1.) == 0);
   assert(compare(fksDir.magnitude(), 1.) == 0);
   const ThreeVector eta(0., 0., 1.);
   // calculate the component of fks transverse to eimmeter by subtracting the parallel component
   const ThreeVector fksPerp = fksDir - (fksDir * emitterDir) * emitterDir;
   const ThreeVector etaPerp = crossProduct(emitterDir, eta);
   double phi = atan2(fksPerp*etaPerp, fksPerp*eta);
   // make sure we return a phi in [0,2pi]
   if (phi < 0.)
      phi += math_constants::_2pi();
   return phi;
}

} // namespace Geneva
