//------------------------------------------------------------------------------
/// \file SubtractionFKSBase.cpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of base class SubtractionFKSBase
//------------------------------------------------------------------------------

#include "Geneva/Components/SubtractionFKSBase.hpp"
#include "Geneva/Components/ITreeMEComplete.hpp"

#include "Geneva/Core/Physics/FourVector.hpp"

#include <cassert>

namespace Geneva
{
//------------------------------------------------------------------------------
SubtractionFKSBase::SubtractionFKSBase(ITreeMEComplete* treeMEcomp, int nLight)
   : _treeMEComplete(treeMEcomp), _nLight(nLight), _commonScale(0.), _scaleSet(false)
{
   assert(nLight >= 3 && nLight <= 6);
}

//------------------------------------------------------------------------------
double SubtractionFKSBase::softIntLightHeavyEps(const FourVector& k, const FourVector& p) const
{
   double kHatdotpHat = (k / k.p0()) * (p / p.p0());
   double beta = sqrt(1. - kHatdotpHat);
   return -2.*(0.25 * pow(log((1. - beta) / (1. + beta)), 2) + log(kHatdotpHat / (1. + beta)) * log(kHatdotpHat / (1. - beta)) + dilog(1. - kHatdotpHat / (1. + beta)) + dilog(1. - kHatdotpHat / (1. - beta)));
}

//------------------------------------------------------------------------------
double SubtractionFKSBase::softIntLightHeavyFin(const FourVector& k, const FourVector& p) const
{
   double kHatdotpHat = (k / k.p0()) * (p / p.p0());
   double pHatdotpHat = (p / p.p0()) * (p / p.p0());
   return log(kHatdotpHat * kHatdotpHat / pHatdotpHat);
}

//------------------------------------------------------------------------------
double SubtractionFKSBase::softIntHeavyHeavyEps(const FourVector& p) const
{
   double beta = sqrt(1. - (p.pvec() / p.p0()).square());
   return 2.*log((1. + beta) / (1. - beta)) / beta;
}

//------------------------------------------------------------------------------
double SubtractionFKSBase::softIntHeavyHeavyEps(const FourVector& p1, const FourVector& p2) const
{
   ThreeVector b1Hat = p1.pvec() / p1.p0();
   ThreeVector b2Hat = p2.pvec() / p2.p0();
   double b1HatSquared =  b1Hat.square();
   double b2HatSquared =  b2Hat.square();
   double b1Hatdotb2Hat = b1Hat * b2Hat;
   double a = b1HatSquared + b2HatSquared + 2.*b1Hatdotb2Hat;
   double x1 = (b1HatSquared - b1Hatdotb2Hat);
   double x2 = (b2HatSquared - b1Hatdotb2Hat);
   double b = b1HatSquared * b2HatSquared - b1Hatdotb2Hat * b1Hatdotb2Hat;
   double zPlus = sqrt(a) + sqrt(a - b);
   double zMinus = sqrt(a) - sqrt(a - b);
   double zMinusOvb = 1. / zPlus;
   double z1 = sqrt(x1 * x1 + b) - x1;
   double z1Ovb = 1. / (sqrt(x1 * x1 + b) + x1);
   double z2 = sqrt(x2 * x2 + b) + x2;
   double Kappaz2 = -0.5 * pow(log((z2 - zMinus) * (zPlus - z2) / ((z2 + zPlus) * (z2 + zMinus))), 2) - 2.*dilog(2.*zMinus * (zPlus - z2) / ((zPlus - zMinus) * (z2 + zMinus))) - 2.*dilog(-2.*zPlus * (zMinus + z2) / ((zPlus - zMinus) * (zPlus - z2)));
   double result = Kappaz2 + 0.5 * pow(log((z1Ovb - zMinusOvb) * (zPlus - z1) / ((zPlus + z1) * (z1Ovb + zMinusOvb))), 2) + 2.*dilog(2.*zMinusOvb / (zPlus - zMinus) * (zPlus - z1) / (zMinusOvb + z1Ovb)) + 2.*dilog(-2.*zPlus / (zPlus - zMinus) * (zMinus + z1) / (zPlus - z1));
   result *= (1. - b1Hatdotb2Hat) / sqrt(a - b);
   return result;
}

//------------------------------------------------------------------------------
double SubtractionFKSBase::softIntHeavyHeavyFin(const FourVector& p1, const FourVector& p2) const
{
   ThreeVector b1Hat = p1.pvec() / p1.p0();
   ThreeVector b2Hat = p2.pvec() / p2.p0();
   double b1HatSquared =  b1Hat.square();
   double b2HatSquared =  b2Hat.square();
   double b1Hatdotb2Hat = b1Hat * b2Hat;
   double beta = sqrt(1. - (1. - b1HatSquared) * (1. - b2HatSquared) / b1Hatdotb2Hat / b1Hatdotb2Hat);
   return log((1. + beta) / (1. - beta)) / beta;
}

} // namespace Geneva
