//------------------------------------------------------------------------------
/// \file SubtractionFKSpp.cpp
//
// Author(s):
//    Christian Bauer, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class SubtractionFKSpp
//------------------------------------------------------------------------------

#include "Geneva/Components/SubtractionFKSpp.hpp"
#include "Geneva/Components/ITreeMEComplete.hpp"
#include "Geneva/Components/FluxFactor.hpp"
#include "Geneva/Components/FKSKinematics.hpp"

#include "Geneva/Core/Physics/PhiRadFKS.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/Physics/ConfigurationUtils.hpp"
#include "Geneva/Core/Physics/QCD.hpp"
#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/Random.hpp"

#include <cassert>
#include <algorithm>
#include <functional>

namespace Geneva
{

//------------------------------------------------------------------------------
SubtractionFKSpp::SubtractionFKSpp(ITreeMEComplete* treeMEComplete, int nlight, double xiCut, double yCutISR, double yCutFSR)
   :  SubtractionFKSBase(treeMEComplete, nlight), _xiCut(xiCut), _yCutISR(yCutISR), _yCutFSR(yCutFSR)
{
   // sanity checks
   assert((xiCut >= 0.0) && (xiCut <= 1.0));
   assert((yCutISR >= 0.0) && (yCutISR <= 2.0));
   assert((yCutFSR >= 0.0) && (yCutFSR <= 2.0));
}

//------------------------------------------------------------------------------
/**
 * Calculates the coeffcients of the integral of the subtraction counterterms.
 * It assumes that a factor 1/Gamma(1-epsilon) * (4 \pi)^epsilon * alphas
 * has been factored out and ignored.
 */
std::vector<double> SubtractionFKSpp::getIntegratedSubtractionCoefficients(const Configuration& phiN_CM) const
{
   assert(isInCoMFrame(phiN_CM, 1e-3, 1e-5)); //sanity check

   double Q = 0;  // collinear
   double I = 0;  // soft
   const double Etot = 2.0 * phiN_CM.initialA()->energy();
   const double sBorn = Etot * Etot;
   double logXiCut = log(_xiCut);
   const double MuSquared = pow(getScale(), 2);
   double logXiCut2SovMu2 = log(pow(_xiCut, 2) * sBorn / MuSquared);
   PdgID idA = phiN_CM.initialA()->id();
   PdgID idB = phiN_CM.initialB()->id();
   const double tree = _treeMEComplete->value();
   // Initial state (massless!) parton contribution
   double QmuF = -tree*( QCD::gamma(idA, _nLight) + 2.*QCD::C(idA) * logXiCut + QCD::gamma(idB, _nLight) + 2.*QCD::C(idB) * logXiCut );
   // Final state massless partons
   for (Configuration::const_iterator it = phiN_CM.beginFinal(); it != phiN_CM.endFinal(); ++it) {
      if (isLightFlavour(it->id())) {
         double Ek = it->energy();
         double log2EkovEtot = log(2.*Ek / Etot);
         PdgID idk = it->id();
         Q += QCD::gammap(idk, _nLight);
         Q -= log((sBorn * _yCutFSR) / (2.0 * MuSquared)) * (QCD::gamma(idk, _nLight) - 2.*QCD::C(idk) * (log2EkovEtot - logXiCut));
         Q += 2.*QCD::C(idk) * (log2EkovEtot + logXiCut) * (log2EkovEtot - logXiCut);
         Q -= 2.*QCD::gamma(idk, _nLight) * log2EkovEtot;
      }
   }
   Q *= tree;
   Configuration::const_iterator it_i, it_j;
   for (it_i = phiN_CM.begin(); it_i != phiN_CM.end(); ++it_i) {
      for (it_j = it_i, ++it_j; it_j != phiN_CM.end(); ++it_j) {
         // massless-massless
         if (isLightFlavour(it_i->id()) && isLightFlavour(it_j->id())) {
            double Kijov2Eij = 0.5 * (it_i->p() / it_i->energy() * it_j->p() / it_j->energy());
            double tmp = abs(1.0 - Kijov2Eij);
            // careful in evaluating the logarithm for too tiny values to avoid numerical instabilities
            double LogLogTerm = (tmp > 1e-14) ? log(tmp) * log(Kijov2Eij) : 0.0;
            I += (0.5 * logXiCut2SovMu2 * (logXiCut2SovMu2 + 2.*log(Kijov2Eij))
                  - dilog(Kijov2Eij)
                  + 0.5 * pow(log(Kijov2Eij), 2)
                  - LogLogTerm) * _treeMEComplete->colorMatrix(it_i, it_j);
         }
         //massless-massive
         if (isLightFlavour(it_i->id()) && isHeavyFlavour(it_j->id())) {
            assert(false); //TODO: not yet tested
            I += 0.5 * ((0.5 * logXiCut2SovMu2 * logXiCut2SovMu2 - math_constants::pi2_6())
                        + softIntLightHeavyFin(it_j->p(), it_i->p()) * logXiCut2SovMu2 - softIntLightHeavyEps(it_j->p(), it_i->p())) * _treeMEComplete->colorMatrix(it_i, it_j);
         }
         //massive-massless
         if (isLightFlavour(it_j->id()) && isHeavyFlavour(it_i->id())) {
            assert(false); //TODO: not yet tested
            I += 0.5 * ((0.5 * logXiCut2SovMu2 * logXiCut2SovMu2 - math_constants::pi2_6())
                        + softIntLightHeavyFin(it_i->p(), it_j->p()) * logXiCut2SovMu2 - softIntLightHeavyEps(it_i->p(), it_j->p())) * _treeMEComplete->colorMatrix(it_i, it_j);
         }
         //massive-massive
         if (isHeavyFlavour(it_i->id()) && isHeavyFlavour(it_j->id())) {
            assert(false); //TODO: not yet tested
            I += 0.5 * (softIntHeavyHeavyFin(it_i->p(), it_j->p()) * logXiCut2SovMu2 - softIntHeavyHeavyEps(it_i->p(), it_j->p())) * _treeMEComplete->colorMatrix(it_i, it_j);
         }
      }
   }
   I *= 2.;
   // Final state massive partons
   for (Configuration::const_iterator it = phiN_CM.beginFinal(); it != phiN_CM.endFinal(); ++it) {
      if (isHeavyFlavour(it->id())) {
         assert(false); //TODO: not yet tested
         I -= QCD::C(it->id()) * (logXiCut2SovMu2 - 0.5 * softIntHeavyHeavyEps(it->p()));
      }
   }
   std::vector<double> coeffs;
   coeffs.push_back((Q + I) * math_constants::_1_2pi());
   coeffs.push_back(0.);
   coeffs.push_back(QmuF * math_constants::_1_2pi());
   return coeffs;
}

//------------------------------------------------------------------------------
std::pair<double, double> SubtractionFKSpp::ggCollinearRemnant(const double sBorn, const double z) const
{
   // g -> (g) g
   double ggCR = QCD::OneMinusZPgg(z) * collRemnFunction(sBorn, z, getScale()) / z;
   double ggCRmuF = -QCD::OneMinusZPgg(z) / (1.-z) / z;
   return std::make_pair(ggCR, ggCRmuF);
}

//------------------------------------------------------------------------------
std::pair<double, double> SubtractionFKSpp::ggSoftCollinearRemnant(const double sBorn, const double z) const
{
   // g -> (g) g
   double ggCR = -QCD::OneMinusZPgg(1.0) * softCollRemnFunction(sBorn, z, getScale());
   double ggCRmuF = +QCD::OneMinusZPgg(1.0) / (1.-z);
   return std::make_pair(ggCR, ggCRmuF);
}

//------------------------------------------------------------------------------
std::pair<double, double> SubtractionFKSpp::qgCollinearRemnant(const double sBorn, const double z) const
{
   // q -> (g) q
   double qgCR = (QCD::OneMinusZPqg(z) * collRemnFunction(sBorn, z, getScale()) - QCD::Pprimeqg(z)) / z;
   double qgCRmuF = -QCD::OneMinusZPqg(z) / (1.-z) / z;
   return std::make_pair(qgCR, qgCRmuF);
}

//------------------------------------------------------------------------------
std::pair<double, double> SubtractionFKSpp::qqCollinearRemnant(const double sBorn, const double z) const
{
   // q -> (q) g
   double qqCR = (QCD::OneMinusZPqq(z) * collRemnFunction(sBorn, z, getScale()) - QCD::Pprimeqq(z)) / z;
   double qqCRmuF = -QCD::OneMinusZPqq(z) / (1.-z) / z;
   return std::make_pair(qqCR, qqCRmuF);
}

//------------------------------------------------------------------------------
std::pair<double, double> SubtractionFKSpp::qqSoftCollinearRemnant(const double sBorn, const double z) const
{
   // q -> (q) g
   double qqCR = -QCD::OneMinusZPqq(1.0) * softCollRemnFunction(sBorn, z, getScale());
   double qqCRmuF = +QCD::OneMinusZPqq(1.0) / (1.-z);
   return std::make_pair(qqCR, qqCRmuF);
}

//------------------------------------------------------------------------------
std::pair<double, double> SubtractionFKSpp::gqCollinearRemnant(const double sBorn, const double z) const
{
   // g -> (q) q~
   double gqCR = (QCD::OneMinusZPgq(z) * collRemnFunction(sBorn, z, getScale()) - QCD::Pprimegq(z)) / z;
   double gqCRmuF = -QCD::OneMinusZPgq(z) / (1.-z) / z;
   return std::make_pair(gqCR, gqCRmuF);
}

//------------------------------------------------------------------------------
/**
 * Calculates the soft subtraction emission for a given emitter, N-body configurations
 * PhiN in the CM frame and radiation phase space PhiRad. Returns zero if the soft
 * particle is not a gluon. The soft limits of the restriction are implemented elsewhere
 * and are not checked here.
 */
double SubtractionFKSpp::getSoft(const PhiNcmFKS& phiN, const PhiRadFKS& phiRad, double& jac) const
{
   if (phiRad.idRight().isGluon()) {
      assert(isInCoMFrame(phiN.configCM, 1e-3, 1e-5));  // sanity check

      double soft = 0.;
      double sConfig = 4.*phiN.configCM.initialA()->energy() * phiN.configCM.initialA()->energy();
      jac = sConfig * phiRad.xi() * math_constants::_1_64pi3();
      const double Ecm = 2.0 * phiN.configCM.initialA()->energy();

      /// construct SoftVec in CM frame
      FourVector pSoft = getSoftVec(phiN.emitter->p(), phiRad, Ecm);

      /// evaluates the soft limits of the restrictions on configurations in the  CM frame
      /// Massless and massive emitters
      for (Configuration::const_iterator i = phiN.configCM.begin(); i != phiN.configCM.end(); ++i) {
         if (IsQCD()(*i)) {
            Configuration::const_iterator j;
            for (j = i, j++; j != phiN.configCM.end(); ++j) {
               if (IsQCD()(*j)) {
                  double temp = i->p() * j->p();
                  temp /= i->p() * pSoft;
                  temp /= j->p() * pSoft;
                  temp *= _treeMEComplete->colorMatrix(i, j);
                  soft += temp;
               }
            }
         }
      }
      ///Bornjk is symmetric, the sum above runs on j>i and we multiply by 2 here
      soft *= 2.;
      ///Massive emitters only
      for (Configuration::const_iterator i = phiN.configCM.begin(); i != phiN.configCM.end(); ++i) {
         if (i->id().isQuark() && !isLightFlavour(i->id())) {
            double temp = i->p().square();
            temp /= i->p() * pSoft;
            temp /= i->p() * pSoft;
            temp *= 2.*QCD::CF * _treeMEComplete->value();
            soft -= temp;
         }
      }
      soft *= math_constants::_4pi();
      return soft;
   }
   return 0.0;
}


//------------------------------------------------------------------------------
/**
 * Calculates the collinear subtraction for a given emitter, N-body configurations
 * PhiN in the CM frame and radiation phase space. Sets the values of the PdgId's
 * and Bjorken x's after the splitting by reference. The collinear limits of the
 * restrictions are implemented elsewhere and are not checked here.
 */
double SubtractionFKSpp::getColl(const PhiNcmFKS& phiN, const PhiRadFKS& phiRad, const double xBarA, const double xBarB, PdgID& idA, double& xA, PdgID& idB, double& xB, double& flux, double& jac) const
{
   assert(isInCoMFrame(phiN.configCM, 1e-3, 1e-5)); // sanity check

   if (phiN.emitter->isFinal()) {
      double q0 = phiN.configCM.initialA()->energy() + phiN.configCM.initialB()->energy();
      double kNBar = phiN.emitter->p().pvec().magnitude();
      jac = q0 * q0 * phiRad.xi() * (kNBar - phiRad.xi() * q0 / 2.) / kNBar * math_constants::_1_64pi3();
      idA =  phiN.configCM.initialA()->id();
      idB =  phiN.configCM.initialB()->id();
      xA = xBarA;
      xB = xBarB;
      flux = FluxFactor()(phiN.configCM);
   } else {
      //Construct the luminosity
      double sConfig = 4.*phiN.configCM.initialA()->energy() * phiN.configCM.initialA()->energy();
      jac = sConfig * phiRad.xi() / (1. - phiRad.xi()) / (1. - phiRad.xi()) * math_constants::_1_64pi3();
      if (phiN.emitter->isInitialA()) {
         idA = phiRad.idLeft();
         xA = xBarA / (1. - phiRad.xi());
         idB = phiN.configCM.initialB()->id();
         xB = xBarB;
      } else if (phiN.emitter->isInitialB()) {
         idB = phiRad.idLeft();
         xB = xBarB / (1. - phiRad.xi());
         idA = phiN.configCM.initialA()->id();
         xA = xBarA;
      } else {
         //sanity check
         Log::error("SubtractionFKSpp::getColl") << "Emitter must be initialA or intialB." << Log::end;
      }
      // now the couplings and prefactors
      flux = FluxFactor()(phiN.configCM) * (1. - phiRad.xi());
   }
   double coll = 0.;
   double z;
   assert(phiN.emitter->id() == phiRad.idMother());
   if (phiN.emitter->isInitialA() || phiN.emitter->isInitialB()) {
      //Calculate the AP functions (expressions in (B.3), (B.4), (B.5), (B.6) of 1002.2581, without factor of 8 pi alphas / k^2 )
      z = 1. - phiRad.xi();
      double split = 0.;
      if (phiRad.idMother().isGluon()) {
         const double g[4][4] = {{1, 0, 0, 0}, {0, -1, 0, 0}, {0, 0, -1, 0}, {0, 0, 0, -1}};
         FourVector kperp = getKPerp(phiN.emitter, phiRad);
         if (phiRad.idLeft().isGluon()) {
            // g -> (g) g splitting (Formula (B.3))
            for (size_t mu = 0; mu < 4; mu++) {
               for (size_t nu = 0; nu < 4; nu++) {
                  double temp = -2. * g[mu][nu] * (z / (1. - z) + z * (1. - z));
                  temp += 4. * (1. - z) / z * kperp[mu] * kperp[nu];
                  temp *= _treeMEComplete->polarizationMatrix(phiN.emitter, mu, nu);
                  split += temp;
               }
            }
            split *= QCD::CA;
         } else if (phiRad.idLeft().isQuark()) {
            // q -> (g) q splitting (Formula (B.5))
            for (size_t mu = 0; mu < 4; mu++) {
               for (size_t nu = 0; nu < 4; nu++) {
                  double temp = -g[mu][nu] * z;
                  temp += 4.*(1. - z) / z * kperp[mu] * kperp[nu];
                  temp *= _treeMEComplete->polarizationMatrix(phiN.emitter, mu, nu);
                  split += temp;
               }
            }
            split *= QCD::CF;
         } else {
            assert(false);
         }
      } else if (phiRad.idMother().isQuark()) {
         if (phiRad.idLeft().isQuark()) {
            // q -> (q) g splitting (Formula (B.4))
            split = QCD::CF * (1. + z * z) / (1. - z) * _treeMEComplete->value();

         } else if (phiRad.idLeft().isGluon()) {
            // g - > (q) q~ splitting (Formula (B.6))
            split = QCD::TF * (z * z + (1. - z) * (1. - z)) * _treeMEComplete->value();
         }
      }
      //Now multiply by the 1/k^2
      coll = split / (2. * phiN.configCM.initialA()->energy() * phiN.configCM.initialB()->energy() * (1. - z));
      if (phiN.emitter->isInitialA())  coll /= 1. - phiRad.y();
      if (phiN.emitter->isInitialB())  coll /= 1. + phiRad.y();
      //Now multiply by the 8 pi
      coll *=  8.* math_constants::pi();
      return coll;

   } else if (phiN.emitter->isFinal()) {
      const double Ecm = 2.0 * phiN.configCM.initialA()->energy();
      z = 1. - phiRad.xi() * Ecm / 2. / phiN.emitter->energy();
      double split = 0.;
      if (phiRad.idMother().isGluon()) {
         const double g[4][4] = {{1, 0, 0, 0}, {0, -1, 0, 0}, {0, 0, -1, 0}, {0, 0, 0, -1}};
         FourVector kperp = getKPerp(phiN.emitter, phiRad);
         if (phiRad.idLeft().isQuark()) {
            // (g) -> q q~ splitting
            for (size_t mu = 0; mu < 4; mu++) {
               for (size_t nu = 0; nu < 4; nu++) {
                  double temp = -g[mu][nu];
                  temp -= 4.*z * (1. - z) * kperp[mu] * kperp[nu];
                  temp *= _treeMEComplete->polarizationMatrix(phiN.emitter, mu, nu);
                  split += temp;
               }
            }
            split *= QCD::TF;
         } else {
            // (g) -> g g splitting
            for (size_t mu = 0; mu < 4; mu++) {
               for (size_t nu = 0; nu < 4; nu++) {
                  double temp = -g[mu][nu] * (z / (1. - z) + (1. - z) / z);
                  temp += 2.*z * (1. - z) * kperp[mu] * kperp[nu];
                  temp *= _treeMEComplete->polarizationMatrix(phiN.emitter, mu, nu);
                  split += temp;
               }
            }
            split *= 2.*QCD::CA;
         }
      } else {
         // (q) -> q g splitting
         split = QCD::CF * (1. + z * z) / (1. - z);
         split *= _treeMEComplete->value();
      }
      //multiply by the 1/k^2
      coll = split / (2. * phiN.emitter->energy() * phiN.emitter->energy() * z * (1 - z) * (1 - phiRad.y()));
      //multiply by the 8 pi
      coll *= 8. * math_constants::pi();
      return coll;
   } else {
      Log::error("SubtractionFKSpp::getColl") << "Emitter must be initial or final." << Log::end;
   }
   return 0.0;
}

//------------------------------------------------------------------------------
/**
 * Calculates the soft-collinear subtraction for a given emitter, N-body configurations
 * PhiN in the CM frame and radiation phase space PhiRad. The soft-collinear limits of the
 * restrictions are implemented elsewhere and are not checked here.
 */
double SubtractionFKSpp::getSoftColl(const PhiNcmFKS& phiN, const PhiRadFKS& phiRad, double& jac) const
{
   assert(isInCoMFrame(phiN.configCM, 1e-3, 1e-5));

   double softcoll = 0;
   if (phiRad.idRight().isGluon()) { // only evaluate soft limit if the FKS parton is a gluon

      const FourVector& pk = phiN.emitter->p();
      double z;
      if (phiN.emitter->isInitialA() || phiN.emitter->isInitialB()) {
         double sConfig = 4.*phiN.configCM.initialA()->energy() * phiN.configCM.initialA()->energy();
         jac = sConfig * phiRad.xi() * math_constants::_1_64pi3();
      } else if (phiN.emitter->isFinal()) {
         double q0 = phiN.configCM.initialA()->energy() + phiN.configCM.initialB()->energy();
         jac = q0 * q0 * phiRad.xi() * math_constants::_1_64pi3();
      }
      assert(phiN.emitter->id() == phiRad.idMother()); // sanity check
//Calculate the the soft limit of the AP functions ( cfr. expressions in (B.3), (B.4), (B.5), (B.6) of 1002.2581, without factor of 8 pi alphas / k^2 )
      if (phiN.emitter->isInitialA() || phiN.emitter->isInitialB()) {
         z = 1. - phiRad.xi();
         double split = 0.;
         if (phiRad.idMother().isGluon()) {
            const double g[4][4] = {{1, 0, 0, 0}, {0, -1, 0, 0}, {0, 0, -1, 0}, {0, 0, 0, -1}};
            if (phiRad.idLeft().isGluon()) {
               // g -> (g) g splitting (Formula (B.3))
               for (size_t mu = 0; mu < 4; mu++) {
                  for (size_t nu = 0; nu < 4; nu++) {
                     double temp = -2. * g[mu][nu] / (1. - z);
                     temp *= _treeMEComplete->polarizationMatrix(phiN.emitter, mu, nu);
                     split += temp;
                  }
               }
               split *= QCD::CA;
            } else if (phiRad.idLeft().isQuark()) {
               // q -> (g) q splitting (Formula (B.5))
               for (size_t mu = 0; mu < 4; mu++) {
                  for (size_t nu = 0; nu < 4; nu++) {
                     double temp = -g[mu][nu];
                     temp *= _treeMEComplete->polarizationMatrix(phiN.emitter, mu, nu);
                     split += temp;
                  }
               }
               split *= QCD::CF;
            }
         } else if (phiRad.idMother().isQuark()) {
            if (phiRad.idLeft().isQuark()) {
               // q -> (q) g splitting (Formula (B.4))
               split = QCD::CF * 2. / (1. - z) * _treeMEComplete->value();

            } else if (phiRad.idLeft().isGluon()) {
               // g - > (q) q~ splitting (Formula (B.6))
               split = QCD::TF * 2. * _treeMEComplete->value();
            }
         } else {
            Log::error("SubtractionFKSpp::getSoftColl") << "Unammissible splitting in ISR." << Log::end;
         }
         //Now multiply by the 1/k^2
         softcoll = split / (2. * phiN.configCM.initialA()->energy() * phiN.configCM.initialA()->energy() * (1. - z));
         if (phiN.emitter->isInitialA()) softcoll /= 1. - phiRad.y();
         if (phiN.emitter->isInitialB()) softcoll /= 1. + phiRad.y();
         //Now multiply by the 8 pi
         softcoll *= 8.* math_constants::pi();
         return softcoll;
      } else if (phiN.emitter->isFinal()) {
         const double Ecm = 2.0 * phiN.configCM.initialA()->energy();
         z = 1. -  phiRad.xi() * Ecm / 2. / pk.p0();
         double split = 0.;
         if (phiRad.idMother().isGluon()) {
            const double g[4][4] = {{1, 0, 0, 0}, {0, -1, 0, 0}, {0, 0, -1, 0}, {0, 0, 0, -1}};
            if (phiRad.idLeft().isQuark()) {
               // (g) -> q q~ splitting
               for (size_t mu = 0; mu < 4; mu++) {
                  for (size_t nu = 0; nu < 4; nu++) {
                     double temp = -g[mu][nu];
                     temp *= _treeMEComplete->polarizationMatrix(phiN.emitter, mu, nu);
                     split += temp;
                  }
               }
               split *= QCD::TF;
            } else {
               // (g) -> g g splitting
               for (size_t mu = 0; mu < 4; mu++) {
                  for (size_t nu = 0; nu < 4; nu++) {
                     double temp = -g[mu][nu] / (1. - z);
                     temp *= _treeMEComplete->polarizationMatrix(phiN.emitter, mu, nu);
                     split += temp;
                  }
               }
               split *= 2.*QCD::CA;
            }
         } else {
            // (q) -> q g splitting
            split = QCD::CF * 2. / (1. - z);
            split *= _treeMEComplete->value();
         }
         //multiply by the 1/k^2
         softcoll = split / (2. * pk.p0() * pk.p0() * (1. - z) * (1. - phiRad.y()));
         //multiply by the 8 pi
         softcoll *= 8. * math_constants::pi();
         return softcoll;
      }
   }
   return 0.0;
}

//------------------------------------------------------------------------------
/**
 * Calculates the soft vector needed in the soft subtraction in the CM frame.
 * It requires the emitter momentum pk to be defined in the CM frame.
 * This results in the FourVector of the gluon that would have been obtained
 * by the FKSMapping in the limit \f$ \xi \to 0\f$.
 */
FourVector SubtractionFKSpp::getSoftVec(const FourVector& pk, const PhiRadFKS& phiRad, double  Q) const
{
   if (phiRad.radType() == PhiRadFKS::kFSR) {
      // define vector etaperp which is perpendicular to pk and the z-direction
      ThreeVector etaperp = crossProduct(ThreeVector(0., 0., 1.), pk.pvec());
      //Rotate etaperp by angle phi around pk direction
      etaperp.apply(Rotation(pk.pvec(), phiRad.phi()));
      etaperp /= etaperp.magnitude();
      double Eg = phiRad.xi() * Q / 2.;

      //Now determine the three vector ps
      ThreeVector ps = Eg / pk.pvec().magnitude() * (sqrt(1 - phiRad.y() * phiRad.y()) * crossProduct(pk.pvec(), etaperp)
                                       + pk.pvec() * phiRad.y());
      FourVector softVec(Eg, ps);
      return softVec;
   } else if ((phiRad.radType() == PhiRadFKS::kISRA) || (phiRad.radType() == PhiRadFKS::kISRB)) {
      double Eg = phiRad.xi() * Q / 2.;
      double y = phiRad.y();
      double phi = phiRad.phi();
      ThreeVector ps(sqrt(1. - y * y)*sin(phi), sqrt(1. - y * y)*cos(phi), y);
      ps *= Eg;
      FourVector softVec(Eg, ps);
      return softVec;
   }
   //sanity check
   assert(false);
   return FourVector();
}

//---------------------------------------------------------------------------
std::string SubtractionFKSpp::name() const
{
   std::stringstream s;
   s << "FKS ISR and FSR subtraction: xiCut = " << _xiCut << " ycutISR = " << _yCutISR << " yCutFSR = " << _yCutFSR;
   return s.str();
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * Calculates the transverse vector needed in the collinear subtraction in the CM frame.
 * It requires the emitter momentum to be defined in the CM frame.
 */
FourVector SubtractionFKSpp::getKPerp(const Configuration::const_iterator& emitter, const PhiRadFKS& phiRad) const
{
   if (emitter->isInitialA() || emitter->isInitialB()) {
      return FourVector(0., sin(phiRad.phi()), cos(phiRad.phi()), 0.);
   } else if (emitter->isFinal()) {
      ThreeVector kPerp(emitter->px(), emitter->py(), -emitter->p().perp2() / emitter->pz());
      kPerp /= kPerp.magnitude();
      ThreeVector dir = emitter->pvec();
      kPerp.apply(Rotation(dir, phiRad.phi()));
      return FourVector(0, kPerp);
   }
   Log::error("SubtractionFKSpp::getKPerp") << "Emitter must be initial or final." << Log::end;
   return FourVector();
}

//-----------------------------------------------------------------------------
double SubtractionFKSpp::collRemnFunction(const double sBorn, const double z, const double scale) const
{
   return 1. / (1. - z) * log(sBorn * _yCutISR / (2. * z * pow(scale, 2))) + 2.*log(1. - z) / (1. - z);
}

//------------------------------------------------------------------------------
double SubtractionFKSpp::softCollRemnFunction(const double sBorn, const double z, const  double scale) const
{
   return 1. / (1. - z) * log(sBorn * _yCutISR / (2. * pow(scale, 2))) + 2.*log(1. - z) / (1. - z);
}

} // namespace Geneva
