//------------------------------------------------------------------------------
/// \file Tau0FRMapping.cpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann, Christan Bauer
//
// Copyright:
//    Copyright (C) 2014 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class Tau0FRMapping
//------------------------------------------------------------------------------

#include "Geneva/Components/Tau0FRMapping.hpp"

#include "Geneva/Core/Physics/PhiRadFKS.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/Physics/ConfigurationUtils.hpp"
#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/Emission.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/NumericsUtils.hpp"

#include <cassert>

#ifndef NDEBUG
#define TAU0FR_MAPPING_PRECISION 1e-7
#define TAU0FR_MAPPING_ACCURACY 1e-10
#endif

namespace Geneva
{

//------------------------------------------------------------------------------
// Public member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
double Tau0FRMapping::getXiMax(PhiN phiN, double y) const
{
   FourVector qBar;
   Configuration::iterator finalQCD;
   for (Configuration::iterator it = phiN.beginFinal(); it != phiN.endFinal(); it++) {
      if (!IsQCD()(*it))
         qBar += it->p();
      else
         finalQCD = it;
   }

   const double pPlus = finalQCD->p().plus();
   const double pMinus = finalQCD->p().minus();
   assert(pPlus >= 0.0);
   assert(pMinus >= 0.0);

   double xiMax = 1.0;
   if (phiN.emitter()->isFinal()) {
      if (_qTfixed) {
         const double tmp = (pPlus * qBar.minus() < pMinus * qBar.plus()) ? pPlus : pMinus;
         xiMax = 2. * std::min(0.5*(pMinus+pPlus)/_cmEnergy, tmp / (_cmEnergy * (1. - y)));
      } else {
         xiMax = (pMinus + pPlus) / _cmEnergy;
      }
   } else {
      const double exp2Y = qBar.minus() / qBar.plus();
      const double tmp = (pPlus * exp2Y < pMinus) ? pPlus * exp2Y : pMinus;
      if ((1. - y)*qBar.minus() < (1. + y)*qBar.plus())
         // hp2+ < hp2-
         xiMax = std::min(xiMax, 2.*tmp / (_cmEnergy * (1. - y) * exp2Y));
      else
         // hp2- <= hp2+
         xiMax = std::min(xiMax, 2.*tmp / (_cmEnergy * (1. + y)));
   }

   assert(xiMax >= 0.0 && xiMax <= 1.0);
   return xiMax;
}

//------------------------------------------------------------------------------
double Tau0FRMapping::getXiMax(PhiNCM phiNCM, double y) const
{
   BoostZHyp toLabFrame(phiNCM.lambda());
   FourVector qBar;
   Configuration::iterator finalQCD;
   for (Configuration::iterator it = phiNCM.beginFinal(); it != phiNCM.endFinal(); it++) {
      if (!IsQCD()(*it))
         qBar += toLabFrame(it->p());
      else
         finalQCD = it;
   }

   const double pPlus = finalQCD->p().plus() / phiNCM.lambda();
   const double pMinus = finalQCD->p().minus() * phiNCM.lambda();
   assert(pPlus >= 0.0);
   assert(pMinus >= 0.0);

   double xiMax = 1.0;
   if (phiNCM.emitter()->isFinal()) {
      if (_qTfixed) {
         const double tmp = (pPlus * qBar.minus() < pMinus * qBar.plus()) ? pPlus : pMinus;
         xiMax = 2. * std::min(0.5*(pMinus+pPlus)/_cmEnergy, tmp / (_cmEnergy * (1. - y)));
      } else {
         xiMax = (pMinus + pPlus) / _cmEnergy;
      }
   } else {
      const double exp2Y = qBar.minus() / qBar.plus();
      const double tmp = (pPlus * exp2Y < pMinus) ? pPlus * exp2Y : pMinus;
      if ((1. - y)*qBar.minus() < (1. + y)*qBar.plus())
         // hp2+ < hp2-
         xiMax = std::min(xiMax, 2.*tmp / (_cmEnergy * (1. - y) * exp2Y));
      else
         // hp2- <= hp2+
         xiMax = std::min(xiMax, 2.*tmp / (_cmEnergy * (1. + y)));
   }

   assert(xiMax >= 0.0 && xiMax <= 1.0);
   return xiMax;
}

//------------------------------------------------------------------------------
double Tau0FRMapping::getXiMaxColl(PhiNCM phiNCM) const
{
   if (phiNCM.emitter()->isFinal()) {
      const double pPlus = phiNCM.emitter()->p().plus() / phiNCM.lambda();
      const double pMinus = phiNCM.emitter()->p().minus() * phiNCM.lambda();
      assert(pPlus >= 0.0);
      assert(pMinus >= 0.0);
      return (pMinus + pPlus) / _cmEnergy;
   } else
      return 1.0;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Performs a PhiN -> PhiN+1 splitting in the CM frame.
 * \param phiNCM The N-body %Configuration in the CM frame and the %Particle to split
 * \param phiRad The kinematic and flavor variables in the splitting
 * \param[out] phiNp1CM The split N+1-body phase-space point in the CM frame
 * \return \c true if the splitting could be performed, \c false otherwise
 *
 */
bool Tau0FRMapping::generateSplit(PhiNCM phiNCM, const PhiRadTau0FR& phiRad, PhiNplus1CMWritable& phiNp1CM)
{
   // sanity checks
   assert(phiRad.radType() == PhiRadTau0FR::kFSR || phiRad.radType() == PhiRadTau0FR::kISRA || phiRad.radType() == PhiRadTau0FR::kISRB);
   assert(std::count_if(phiNCM.beginFinal(), phiNCM.endFinal(), IsQCD()) == 1);

   if (phiRad.radType() == PhiRadTau0FR::kFSR)
      return splitFSRCM(phiNCM, phiRad, phiNp1CM);
   else
      return splitISRCM(phiNCM, phiRad, phiNp1CM);
}

//-------------------------------------------------------------------------------------
/**
 * \brief Performs a PhiN -> PhiN+1 splitting in the lab frame.
 * \param phiN The N-body %Configuration in the lab frame and the %Particle to split
 * \param phiRad The kinematic and flavor variables in the splitting
 * \param[out] phiNp1 The split N+1-body phase-space point in the lab frame
 * \return \c true if the splitting could be performed, \c false otherwise
 */
bool Tau0FRMapping::generateSplit(PhiN phiN, const PhiRadTau0FR& phiRad, PhiNplus1CMWritable& phiNp1)
{
   // sanity checks
   assert(phiRad.radType() == PhiRadTau0FR::kFSR || phiRad.radType() == PhiRadTau0FR::kISRA || phiRad.radType() == PhiRadTau0FR::kISRB);
   assert(std::count_if(phiN.beginFinal(), phiN.endFinal(), IsQCD()) == 1);

   if (phiRad.radType() == PhiRadTau0FR::kFSR)
      return splitFSRLab(phiN, phiRad, phiNp1);
   else
      return splitISRLab(phiN, phiRad, phiNp1);
}

//-------------------------------------------------------------------------------------
/**
 * \brief Performs a PhiN+1 -> PhiN projection in the CM frame.
 * \param phiNp1CM The (N+1)-body %Configuration in the CM frame and the two particles to combine
 * \param xA, \param xB The values of xA and xB of the PhiN+1 point
 * \param[out] configNCM The projected N-body %Configuration in the CM frame
 * \param[out] xBarA, \param[out] xBarB The values of xA and xB of the projected PhiN point
 * \return \c true if the projection could be performed, \c false otherwise
 */
bool Tau0FRMapping::project(PhiNplus1 phiNp1CM, double xA, double xB,
                            Configuration& configNCM, double& xBarA, double& xBarB) const
{
   // sanity checks
   assert(phiNp1CM.sister()->isFinal() || phiNp1CM.sister()->isInitialA() || phiNp1CM.sister()->isInitialB());
   assert(phiNp1CM.emitted()->isFinal());
   assert(std::count_if(phiNp1CM.beginFinal(), phiNp1CM.endFinal(), IsQCD()) == 2);

   if (phiNp1CM.sister()->isFinal())
      return projectFSRCM(phiNp1CM, phiNp1CM.emitted(), phiNp1CM.sister(), xA, xB, configNCM, xBarA, xBarB);
   else
      return projectISRCM(phiNp1CM, phiNp1CM.emitted(), phiNp1CM.sister(), xA, xB, configNCM, xBarA, xBarB);
}

//-------------------------------------------------------------------------------------
/**
 * \brief Performs a PhiN+1 -> PhiN projection in the lab frame.
 * \param phiNp1Lab The (N+1)-body %Configuration in the lab frame and the two particles to combine
 * \param[out] configNLab The projected N-body %Configuration in the lab frame
 * \return \c true if the projection is possible, \c false otherwise
 */
bool Tau0FRMapping::project(PhiNplus1 phiNp1Lab, Configuration& configNLab) const
{
   // sanity checks
   assert(phiNp1Lab.sister()->isFinal() || phiNp1Lab.sister()->isInitialA() || phiNp1Lab.sister()->isInitialB());
   assert(phiNp1Lab.emitted()->isFinal());
   assert(std::count_if(phiNp1Lab.beginFinal(), phiNp1Lab.endFinal(), IsQCD()) == 2);

   if (phiNp1Lab.sister()->isFinal())
      return projectFSRLab(phiNp1Lab, phiNp1Lab.emitted(), phiNp1Lab.sister(), configNLab);
   else
      return projectISRLab(phiNp1Lab, phiNp1Lab.emitted(), phiNp1Lab.sister(), configNLab);
}

//-------------------------------------------------------------------------------------
/**
 * \brief Tests if a CM-frame projection PhiN+1 -> PhiN is possible.
 * \param phiNp1CM The (N+1)-body %Configuration in the CM frame and the two particles to combine
 * \param xA, \param xB The values of xA and xB of the PhiN+1 point
 * \return \c true if the projection is possible, \c false otherwise
 */
bool Tau0FRMapping::testProject(PhiNplus1 phiNp1CM, double xA, double xB) const
{
   // sanity checks
   assert(phiNp1CM.sister()->isFinal() || phiNp1CM.sister()->isInitialA() || phiNp1CM.sister()->isInitialB());
   assert(phiNp1CM.emitted()->isFinal());
   assert(std::count_if(phiNp1CM.beginFinal(), phiNp1CM.endFinal(), IsQCD()) == 2);
   if (phiNp1CM.sister()->isFinal())
      return testProjectFSRCM(phiNp1CM, phiNp1CM.emitted(), phiNp1CM.sister(), xA, xB);
   else
      return testProjectISRCM(phiNp1CM, phiNp1CM.emitted(), xA, xB);
}

//-------------------------------------------------------------------------------------
/**
 * \brief Tests if a lab-frame projection PhiN+1 -> PhiN is possible.
 * \param phiNp1Lab The (N+1)-body %Configuration in the lab frame and the two particles to combine
 * \return \c true if the projection is possible, \c false otherwise
 */
bool Tau0FRMapping::testProject(PhiNplus1 phiNp1Lab) const
{
   // sanity checks
   assert(phiNp1Lab.sister()->isFinal() || phiNp1Lab.sister()->isInitialA() || phiNp1Lab.sister()->isInitialB());
   assert(phiNp1Lab.emitted()->isFinal());
   assert(std::count_if(phiNp1Lab.beginFinal(), phiNp1Lab.endFinal(), IsQCD()) == 2);
   if (phiNp1Lab.sister()->isFinal())
      return testProjectFSRLab(phiNp1Lab, phiNp1Lab.emitted(), phiNp1Lab.sister());
   else
      return testProjectISRLab(phiNp1Lab, phiNp1Lab.emitted());
}

//------------------------------------------------------------------------------
/**
 * \brief Obtain the phiRad for a given Phi1, Tau1, z, phi, and splitting
 *
 * \param[in] mother iterator to the mother particle in Phi11
 * \param[out] phiRad The resulting phiRad
 * \return The Jacobian dPhiRad / (dTau1 dz dphi)
 *
 * \todo: There should be documentation here whether the function assumes
 * that mother is a valid iterator and whether emittedID and sisterID
 * should be valid or not. I.e., whether the caller or the function is
 * responsible for this.
 */
double Tau0FRMapping::getPhiRad(const Configuration& Phi1, double Tau1, double z, double phi, Signature::const_iterator mother, PdgID emittedID, PdgID sisterID, PhiRadFKS& phiRad) const
{
   const double yBoost = NJettinessppRecursive(Phi1).y();
   const double expY = exp(yBoost);
   const double expMinY = 1. / expY;
   //Determine the values of xi and y and the RadiationType
   double xi, y, jacobian;
   //Set the Jacobian to dxi dy dphi / dTau1 dz dphi
   jacobian = 0.;
   if (mother->isInitialA()) {
      phiRad.radType() = PhiRadFKS::kISRA;
      const double kBarA = Phi1.initialA()->p().minus();
      xi = (expY * kBarA * (1. - z) + z * Tau1) / _cmEnergy / z / expY;
      y = (expY * kBarA * (1. - z) - z * Tau1) / (expY * kBarA * (1. - z) + z * Tau1);
      jacobian = 2. * kBarA / _cmEnergy / z / (expY * kBarA * (1. - z) + Tau1 * z);
   } else if (mother->isInitialB()) {
      phiRad.radType() = PhiRadFKS::kISRB;
      const double kBarB = Phi1.initialB()->p().plus();
      xi = (kBarB * (1. - z) + z * Tau1 * expY) / _cmEnergy / z;
      y = (z * Tau1 * expY - kBarB * (1. - z)) / (z * Tau1 * expY + kBarB * (1. - z));
      jacobian = 2. * expY * kBarB / _cmEnergy / z / (kBarB * (1. - z) + expY * Tau1 * z);
   } else {
      phiRad.radType() = PhiRadFKS::kFSR;
      FourVector q, pBar;
      for (Configuration::iterator it = Phi1.beginFinal(); it != Phi1.endFinal(); ++it) {
         if (it->id().isLepton()) {
            q += it->p();
         } else
            pBar = it->p();
      }
      if (_qTfixed) {
         double p12Plus, p12Minus, E12, p12Sq;
         double PartialE12PartialTau1, Partialp12SqPartialTau1;
         if (sqrt(pBar.minus() / pBar.plus()) > expY) {
            double qPerpSq = q.p1() * q.p1() + q.p2() * q.p2();
            double pBarPlus = pBar.plus();
            p12Minus = expY * (Tau1 + qPerpSq / (expY * pBarPlus - Tau1));
            p12Plus = pBarPlus;
            p12Sq = p12Plus * p12Minus - qPerpSq;
            E12 = (p12Plus + p12Minus) / 2.;
            double Partialp12MinusPartialTau1 = expY * (1. + qPerpSq / (expY * pBarPlus - Tau1) / (expY * pBarPlus - Tau1));
            PartialE12PartialTau1 = 0.5 * Partialp12MinusPartialTau1;
            Partialp12SqPartialTau1 = p12Plus * Partialp12MinusPartialTau1;
         } else {
            double qPerpSq = q.p1() * q.p1() + q.p2() * q.p2();
            double pBarMinus = pBar.minus();
            p12Plus = expMinY * (Tau1 + qPerpSq / (expMinY * pBarMinus - Tau1));
            p12Minus = pBarMinus;
            p12Sq = p12Plus * p12Minus - qPerpSq;
            E12 = (p12Plus + p12Minus) / 2.;
            double Partialp12PlusPartialTau1 = expMinY * (1. + qPerpSq / (expMinY * pBarMinus - Tau1) / (expMinY * pBarMinus - Tau1));
            PartialE12PartialTau1 = 0.5 * Partialp12PlusPartialTau1;
            Partialp12SqPartialTau1 = p12Minus * Partialp12PlusPartialTau1;
         }
         //There are two solutions to the equation I am solving, and only if tau1 = hatE12 - hatp12 did I find the correct solution
         if (expY * p12Plus + expMinY * p12Minus < 2. * Tau1) {
            xi = -1.;
            y = -2.;
         } else {
            xi = 2. * E12 * (1. - z) / _cmEnergy;
            y = 1. - p12Sq / (xi * _cmEnergy * (E12 - xi * _cmEnergy / 2.));
            //Some partial derivatives
            double PartialxiPartialE12 = 2. / _cmEnergy * (1. - z);
            double PartialyPartialxi = 4. * p12Sq * (E12 - _cmEnergy * xi) / (_cmEnergy * xi * xi * pow(_cmEnergy * xi - 2. * E12, 2));
            double PartialyPartialp12Sq = - 1. / (xi * _cmEnergy * (E12 - xi * _cmEnergy / 2.));
            double PartialyPartialE12 = p12Sq / (xi * _cmEnergy * pow(E12 - 0.5 * xi * _cmEnergy, 2));

            //The individual derivates of the Jacobian matrix
            double DxiDz = - 2. * E12 / _cmEnergy;
            double DxiDTau1 = PartialxiPartialE12 * PartialE12PartialTau1;

            double DyDz = PartialyPartialxi * DxiDz;
            double DyDTau1 = PartialyPartialp12Sq * Partialp12SqPartialTau1 + (PartialyPartialE12 + PartialyPartialxi*PartialxiPartialE12) * PartialE12PartialTau1;

            jacobian = abs(DyDz * DxiDTau1 - DyDTau1 * DxiDz);
         }
      } else {
         const double E12 = 0.5 * (pBar.plus() + pBar.minus());
         xi = 2. * E12 * (1. - z) / _cmEnergy;
         const double E2 = xi * _cmEnergy / 2.;
         const double E1 = E12 - E2;
         y = 1. - 0.5 * Tau1 * (expY * pBar.plus() + expMinY * pBar.minus() - Tau1) / E1 / E2;
         jacobian = E12 * (2. * E1 * E2 * (1. - y) - Tau1 * Tau1) / (_cmEnergy * E1 * E2 * Tau1);
      }
   }

   phiRad.xi() = xi;
   phiRad.y() = y;
   phiRad.phi() = phi;
   phiRad.idMother() = *mother;
   phiRad.idLeft() = sisterID;
   phiRad.idRight() = emittedID;

   return abs(jacobian);
}

//------------------------------------------------------------------------------
double Tau0FRMapping::zMin(const Configuration& Phi1, double Tau1, double phi, Signature::const_iterator mother, PdgID emittedID)
{
   if (mother->isInitialA())
      return Phi1.initialA()->p().minus() / _cmEnergy;

   if (mother->isInitialB())
      return Phi1.initialB()->p().plus() / _cmEnergy;

   // final-state mother
   FourVector q;
   FourVector pBar;
   for (Configuration::iterator it = Phi1.beginFinal(); it != Phi1.endFinal(); ++it) {
      if (!IsQCD()(it->id()))
         q += it->p();
      else
         pBar = it->p();
   }
   double expY = sqrt(q.minus() / q.plus());
   double expMinY = 1. / expY;
   double beta0;
   if (_qTfixed) {
      if (pBar.minus() > pBar.plus() * expY * expY) {
         double qPerpSq = q.p1() * q.p1() + q.p2() * q.p2();
         double pBarPlus = pBar.plus();
         double p12Minus = expY * (Tau1 + qPerpSq / (expY * pBarPlus - Tau1));
         double p12Plus = pBarPlus;
         double p12Sq = p12Plus * p12Minus - qPerpSq;
         double E12 = (p12Plus + p12Minus) / 2.;
         double p12 = sqrt(E12 * E12 - p12Sq);
         beta0 = p12 / E12;
      } else {
         double qPerpSq = q.p1() * q.p1() + q.p2() * q.p2();
         double pBarMinus = pBar.minus();
         double p12Plus = expMinY * (Tau1 + qPerpSq / (expMinY * pBarMinus - Tau1));
         double p12Minus = pBarMinus;
         double p12Sq = p12Plus * p12Minus - qPerpSq;
         double E12 = (p12Plus + p12Minus) / 2.;
         double p12 = sqrt(E12 * E12 - p12Sq);
         beta0 = p12 / E12;
      }
   } else {
      const double E12 = 0.5 * (pBar.plus() + pBar.minus());
      const double p12Sq = Tau1 * (expY * pBar.plus() + expMinY * pBar.minus() - Tau1);
      beta0 = sqrt(E12*E12 - p12Sq) / E12;
   }
   double tmp = (1. - beta0) / 2.;
   forceGreaterEqualZero(tmp, "Tau0FRMapping::zMin");
   return tmp;
}

//------------------------------------------------------------------------------
double Tau0FRMapping::zMax(const Configuration& Phi1, const double Tau1, const double phi, Signature::const_iterator mother, PdgID emittedID)
{
   FourVector q;
   FourVector pBar;
   for (Configuration::iterator it = Phi1.beginFinal(); it != Phi1.endFinal(); ++it) {
      if (!IsQCD()(it->id()))
         q += it->p();
      else
         pBar = it->p();
   }
   double expY = sqrt(q.minus() / q.plus());
   double expMinY = 1. / expY;
   if (mother->isInitialA()) {
      double xA = Phi1.initialA()->p().minus() / _cmEnergy;
      return 1. / (1. + Tau1 / _cmEnergy / xA * expY);
   }

   if (mother->isInitialB()) {
      double xB = Phi1.initialB()->p().plus() / _cmEnergy;
      return 1. / (1. + Tau1 / _cmEnergy / xB / expY);
   }

   // final-state mother
   double beta0;
   if (_qTfixed) {
      if (pBar.minus() > pBar.plus() * expY * expY) {
         double qPerpSq = q.p1() * q.p1() + q.p2() * q.p2();
         double pBarPlus = pBar.plus();
         double p12Minus = expY * (Tau1 + qPerpSq / (expY * pBarPlus - Tau1));
         double p12Plus = pBarPlus;
         double p12Sq = p12Plus * p12Minus - qPerpSq;
         double E12 = (p12Plus + p12Minus) / 2.;
         double p12 = sqrt(E12 * E12 - p12Sq);
         beta0 = p12 / E12;
      } else {
         double qPerpSq = q.p1() * q.p1() + q.p2() * q.p2();
         double pBarMinus = pBar.minus();
         double p12Plus = expMinY * (Tau1 + qPerpSq / (expMinY * pBarMinus - Tau1));
         double p12Minus = pBarMinus;
         double p12Sq = p12Plus * p12Minus - qPerpSq;
         double E12 = (p12Plus + p12Minus) / 2.;
         double p12 = sqrt(E12 * E12 - p12Sq);
         beta0 = p12 / E12;
      }
   } else {
      const double E12 = 0.5 * (pBar.plus() + pBar.minus());
      const double p12Sq = Tau1 * (expY * pBar.plus() + expMinY * pBar.minus() - Tau1);
      beta0 = sqrt(E12*E12 - p12Sq) / E12;
   }
   double tmp = (1. + beta0) / 2.;
   forceLessEqual(tmp, 1.0, "Tau0FRMapping::zMax");
   return tmp;
}

//------------------------------------------------------------------------------
double Tau0FRMapping::phiMin(const Configuration& config, const double Tau0, Signature::const_iterator mother, PdgID emittedID)
{
   return 0.;
}

//------------------------------------------------------------------------------
double Tau0FRMapping::phiMax(const Configuration& config, const double Tau0, Signature::const_iterator mother, PdgID emittedID)
{
   return math_constants::_2pi();
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------
/**
 * \brief Generates the splitting in the CM frame for an ISR emission.
 *
 * \param phiNCM The N-body %Configuration in the CM frame and the %Particle to split
 * \param phiRad The kinematic and flavor variables in the splitting
 * \param[out] phiNp1CM The resulting (N+1)-body Configuration
 * \return \c true if the splitting could be performed, \c false otherwise
 *
 * Most of the actual formulas for the splitting are relegated to another function
 */
bool Tau0FRMapping::splitISRCM(PhiNCM phiNCM, const PhiRadTau0FR& phiRad, PhiNplus1CMWritable& phiNp1CM)
{
   BoostZHyp toLabFrame(phiNCM.lambda());
   phiNp1CM.clear();
   FourVector qBar;
   Configuration::iterator spectator;
   for (Configuration::iterator it = phiNCM.beginFinal(); it != phiNCM.endFinal(); it++) {
      if (!IsQCD()(*it))
         // move over the leptons and boost them to lab frame
         qBar += phiNp1CM.insertFinal(Particle(it->id(), toLabFrame(it->p())))->p();
      else
         spectator = it;
   }

   double pBarPlus = spectator->p().plus() / phiNCM.lambda();
   double pBarMinus = spectator->p().minus() * phiNCM.lambda();

   PdgID otherInitial;
   if (phiRad.radType() == PhiRadFKS::kISRA) {
      otherInitial = phiNCM.initialB()->id();
   } else {
      otherInitial = phiNCM.initialA()->id();
   }

   if (_qTfixed) {
      if (!splitISRfixqT(qBar, pBarPlus, pBarMinus, phiRad, spectator->id(), otherInitial, phiNp1CM))
         return false;
   } else {
      if (!splitISR(qBar, pBarPlus, pBarMinus, spectator->px(), spectator->py(), phiRad, spectator->id(), otherInitial, phiNp1CM))
         return false;
   }
   const BoostZHyp toCMframe(sqrt(phiNp1CM.initialB()->p().plus()/phiNp1CM.initialA()->p().minus()));
   phiNp1CM.apply(toCMframe);

   // final accuracy checks
//    assert(isInCoMFrame(phiNp1CM, "Tau0FRMapping::splitISRCM", 1e-12));
   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Generates the splitting in the lab frame for an ISR emission.
 *
 * \param phiNCM The N-body %Configuration in the CM frame and the %Particle to split
 * \param phiRad The kinematic and flavor variables in the splitting
 * \param[out] phiNp1CM The resulting (N+1)-body Configuration
 * \return \c true if the splitting could be performed, \c false otherwise
 *
 * Most of the actual formulas for the splitting are relegated to another function
 */
bool Tau0FRMapping::splitISRLab(PhiN phiN, const PhiRadTau0FR& phiRad, PhiNplus1CMWritable& phiNp1)
{
   phiNp1.clear();
   FourVector qBar;
   Configuration::iterator spectator;
   for (Configuration::iterator it = phiN.beginFinal(); it != phiN.endFinal(); it++) {
      if (!IsQCD()(*it))
         // move over the leptons
         qBar += phiNp1.insertFinal(*it)->p();
      else
         spectator = it;
   }

   double pBarPlus = spectator->p().plus();
   double pBarMinus = spectator->p().minus();

   PdgID otherInitial;
   if (phiRad.radType() == PhiRadFKS::kISRA) {
      otherInitial = phiN.initialB()->id();
   } else {
      otherInitial = phiN.initialA()->id();
   }
   if (_qTfixed)
      return splitISRfixqT(qBar, pBarPlus, pBarMinus, phiRad, spectator->id(), otherInitial, phiNp1);
   else
      return splitISR(qBar, pBarPlus, pBarMinus, spectator->px(), spectator->py(), phiRad, spectator->id(), otherInitial, phiNp1);
}

//-------------------------------------------------------------------------------------
/**
 * \brief Implements the ISR splitting formulas in the Lab frame.
 *
 * \param qBar The momentum of the vector boson
 * \param pBarPlus, \param pBarMinus The components of the emitter momentum
 * \param phiRad The kinematic and flavor variables of the splitting
 * \param spectatorID The PdgID of the final state particle not partipating in the splitting
 * \param otherInitial The flavor of the beam not participating in the splitting
 * \param[out][ phiNp1 The final (N+1) body configuration in the lab frame
 * \return \c true if the splitting could be performed, \c false otherwise
 */
bool Tau0FRMapping::splitISR(FourVector qBar, double pBarPlus, double pBarMinus, double pBarx, double pBary, const PhiRadTau0FR& phiRad, PdgID spectatorID, PdgID otherInitial, PhiNplus1CMWritable& phiNp1)
{
   assert(pBarPlus >= 0.0);
   assert(pBarMinus >= 0.0);

   const double qBarPlus = qBar.plus();
   const double qBarMinus = qBar.minus();
   const double exp2Y = qBarMinus / qBarPlus;
   assert(qBarPlus >= 0.0);
   assert(qBarMinus >= 0.0);

   // construct emitted p2 from phiRad
   const double p20 = 0.5 * _cmEnergy * phiRad.xi();
   const double p2T = p20 * sqrt(1. - phiRad.y() * phiRad.y());
   const double p2Plus = p20 * (1. - phiRad.y());
   const double p2Minus = p20 * (1. + phiRad.y());
   const FourVector p2(p20, p2T * cos(phiRad.phi()), p2T * sin(phiRad.phi()), p20 * phiRad.y());

   double p1Minus;
   double p1Plus;
   if (pBarPlus * exp2Y < pBarMinus) {
      if (p2Plus < p2Minus / exp2Y) {
         p1Plus = pBarPlus - p2Plus;
      } else {
         p1Plus = pBarPlus - p2Minus / exp2Y;
      }
      assert(p1Plus >= 0.0);
      if (p1Plus < 0.0) {
         return false;
      }
      p1Minus = p1Plus * pBarMinus / pBarPlus;
   } else {
      if (p2Plus * exp2Y < p2Minus) {
         p1Minus = pBarMinus - p2Plus * exp2Y;
      } else {
         p1Minus = pBarMinus - p2Minus;
      }
      assert(p1Minus >= 0.0);
      if (p1Minus < 0.0) {
         return false;
      }
      p1Plus = p1Minus * pBarPlus / pBarMinus;
   }

   const double p1PerpSq = p1Plus * p1Minus;
   const double rescale = sqrt(p1PerpSq / qBar.perp2());
   const FourVector p1(0.5 * (p1Minus + p1Plus), rescale * pBarx, rescale * pBary, 0.5 * (p1Minus - p1Plus));

   const double qPerpSq = (p1Plus + p2Plus) * (p1Minus + p2Minus) - (p1 + p2).square();
   if (qPerpSq < 0.0)
      return false;
   const double mLLSq = qBar.square();
   const double rescaleqPlusMinus = sqrt((mLLSq + qPerpSq) / (mLLSq + qBar.perp2()));
   const double qPlus = rescaleqPlusMinus * qBar.plus();
   const double qMinus = rescaleqPlusMinus * qBar.minus();
   const FourVector q(0.5 * (qMinus + qPlus), - p1.p1() - p2.p1(), - p1.p2() - p2.p2(), 0.5 * (qMinus - qPlus));

   const double minus = p1Minus + p2Minus + qMinus;
   const double plus = p1Plus + p2Plus + qPlus;
   // check that we do not exceed available Ecm (too large xi)
   if (minus > _cmEnergy || plus > _cmEnergy)
      return false;

   // after all checks, construct the N+1 phase space point
   // first modify the momenta of the already-inserted non-strongly interacting particles
   assert(std::count_if(phiNp1.beginFinal(), phiNp1.endFinal(), IsQCD()) == 0);
   Boost toCMqBar = Boost(qBar).inverse();
   Boost toLabq(q);
   phiNp1.apply(toCMqBar);
   phiNp1.apply(toLabq);

   // add final state partons
   phiNp1.insertEmitted(Particle(phiRad.idRight(), p2, 0, 0));
   phiNp1.insertFinal(Particle(spectatorID, p1, 0, 0));

   // add initial particles
   const FourVector pA(0.5*minus, 0.0, 0.0, 0.5*minus);
   const FourVector pB(0.5*plus, 0.0, 0.0, -0.5*plus);
   phiNp1.xA() = minus / _cmEnergy;
   phiNp1.xB() = plus / _cmEnergy;

   if (phiRad.radType() == PhiRadTau0FR::kISRA) {
      phiNp1.setInitialASister(Particle(phiRad.idLeft(), pA, 0, 0));
      phiNp1.setInitialB(Particle(otherInitial, pB, 0, 0));
   } else { // ISRB
      phiNp1.setInitialA(Particle(otherInitial, pA, 0, 0));
      phiNp1.setInitialBSister(Particle(phiRad.idLeft(), pB, 0, 0));
   }

   // final accuracy checks
#ifndef NDEBUG
   if (!checkMomentumConservation(phiNp1.config(), TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::splitISR") << "Phi_N+1 point from splitting violates momentum conservation." << Log::end;
   if (!checkOnShellConditions(phiNp1.config(), TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::splitISR") << "Phi_N+1 point from splitting violates on-shell conditions." << Log::end;
#endif

   phiNp1.jacobian() = p1.perp() / sqrt(pBarPlus * pBarMinus);
   phiNp1.jacobian() *= _cmEnergy * _cmEnergy * phiRad.xi() * math_constants::_1_64pi3();
   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Implements the ISR splitting formulas with fixed qT in the lab frame.
 *
 * \param q The momentum of the vector boson
 * \param pBarPlus, \param pBarMinus The components of the emitter momentum
 * \param phiRad The kinematic and flavor variables of the splitting
 * \param spectatorID The PdgID of the final state particle not partipating in the splitting
 * \param otherInitial The flavor of the beam not participating in the splitting
 * \param[out] phiNp1 The final (N+1) body configuration in the lab frame
 * \return \c true if the splitting could be performed, \c false otherwise
 */
bool Tau0FRMapping::splitISRfixqT(FourVector q, double pBarPlus, double pBarMinus, const PhiRadTau0FR& phiRad, PdgID spectatorID, PdgID otherInitial, PhiNplus1CMWritable& phiNp1)
{
   assert(pBarPlus >= 0.0);
   assert(pBarMinus >= 0.0);

   const double qPlus = q.plus();
   const double qMinus = q.minus();
   const double exp2Y = qMinus / qPlus;
   assert(qPlus >= 0.0);
   assert(qMinus >= 0.0);

   // construct emitted p2 from phiRad
   const double p20 = 0.5 * _cmEnergy * phiRad.xi();
   const double p2T = p20 * sqrt(1. - phiRad.y() * phiRad.y());
   const double p2Plus = p20 * (1. - phiRad.y());
   const double p2Minus = p20 * (1. + phiRad.y());
   const FourVector p2(p20, p2T * cos(phiRad.phi()), p2T * sin(phiRad.phi()), p20 * phiRad.y());

   double p1Minus;
   double p1Plus;
   const double p1perp2 = pow(q.p1() + p2.p1(), 2) + pow(q.p2() + p2.p2(), 2);
   if (pBarPlus * exp2Y < pBarMinus) {
      // hatp+ < hatp-
      if (p2Plus < p2Minus / exp2Y)
         // hatp2+ < hatp2- (same hemisphere case 1)
         p1Plus = pBarPlus - p2Plus;
      else
         // hatp2- <= hatp2+ (opposite hemisphere case 2)
         p1Plus = pBarPlus - p2Minus / exp2Y;
      assert(p1Plus >= 0.0);
      if (p1Plus < 0.0) {
//         Log::debug().turnOn();
//         Log::debug() << "q+, q-: " << qPlus << ", " << qMinus << ", exp2Y = " << exp2Y << '\n';
//         Log::debug() << "p+, p-: " << pPlus << ", " << pMinus << ", p+ e^2Y = " << pPlus*exp2Y << '\n';
//         Log::debug() << "p2+, p2-: " << p2Plus << ", " << p2Minus << ", p2- e^-2Y = " << p2Minus / exp2Y << '\n';
//         Log::debug() << "p1+     : " << p1Plus << '\n';
//         Log::debug() << "xi = " << phiRad.xi() << ", 1. - y = " << 1. - phiRad.y() << '\n';
//         Log::debug().turnOff();
         return false;
      }
      p1Minus = p1perp2 / p1Plus;
      // p1 must be in the same hemisphere as pbar by definition of the ISR map
      if (p1Minus < p1Plus * exp2Y)
         return false;
      phiNp1.jacobian() = pBarPlus / p1Plus;
   } else {
      // hatp- <= hatp+
      if (p2Plus * exp2Y < p2Minus)
         // hatp2+ < hatp2- (opposite hemisphere case 3)
         p1Minus = pBarMinus - p2Plus * exp2Y;
      else
         // hatp2- <= hatp2+  (same hemisphere case 4)
         p1Minus = pBarMinus - p2Minus;
      assert(p1Minus >= 0.0);
      if (p1Minus < 0.0)
         return false;
      p1Plus = p1perp2 / p1Minus;
      // p1 must be in the same hemisphere as pbar by definition of the ISR map
      if (p1Minus > p1Plus * exp2Y)
         return false;
      phiNp1.jacobian() = pBarMinus / p1Minus;
   }

   const double minus = p1Minus + p2Minus + qMinus;
   const double plus = p1Plus + p2Plus + qPlus;
   // check that we do not exceed available Ecm (too large xi)
   if (minus > _cmEnergy || plus > _cmEnergy)
      return false;

   // after all checks, construct the N+1 phase space point
   // add final state partons
   phiNp1.insertEmitted(Particle(phiRad.idRight(), p2, 0, 0));

   const FourVector p1(0.5 * (p1Minus + p1Plus), -p2.p1() - q.p1(), -p2.p2() - q.p2(), 0.5 * (p1Minus - p1Plus));
   phiNp1.insertFinal(Particle(spectatorID, p1, 0, 0));

   // add initial particles
   const FourVector pA(0.5*minus, 0.0, 0.0, 0.5*minus);
   const FourVector pB(0.5*plus, 0.0, 0.0, -0.5*plus);
   phiNp1.xA() = minus / _cmEnergy;
   phiNp1.xB() = plus / _cmEnergy;

   if (phiRad.radType() == PhiRadTau0FR::kISRA) {
      phiNp1.setInitialASister(Particle(phiRad.idLeft(), pA, 0, 0));
      phiNp1.setInitialB(Particle(otherInitial, pB, 0, 0));
   } else { // ISRB
      phiNp1.setInitialA(Particle(otherInitial, pA, 0, 0));
      phiNp1.setInitialBSister(Particle(phiRad.idLeft(), pB, 0, 0));
   }

   // final accuracy checks
#ifndef NDEBUG
   if (!checkMomentumConservation(phiNp1.config(), TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::splitISRfixqT") << "Phi_N+1 point from splitting violates momentum conservation." << Log::end;
   if (!checkOnShellConditions(phiNp1.config(), TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::splitISRfixqT") << "Phi_N+1 point from splitting violates on-shell conditions." << Log::end;
#endif

   // TODO: discuss whether to include also the Jacobian to rescale the PhiRad variables to match the definition in the Subtractions
   phiNp1.jacobian() *= _cmEnergy * _cmEnergy * phiRad.xi() * math_constants::_1_64pi3();
   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Generates the splitting in the CM frame for an FSR emission.
 *
 * \param phiNCM The N-body %Configuration in the CM frame and the %Particle to split
 * \param phiRad The kinematic and flavor variables in the splitting
 * \param[out] phiNp1CM The resulting (N+1)-body Configuration
 * \return \c true if the splitting could be performed, \c false otherwise
 *
 * Most of the actual formulas for the splitting are relegated to another function
 */
//------------------------------------------------------------------------------
bool Tau0FRMapping::splitFSRCM(PhiNCM phiNCM, const PhiRadTau0FR& phiRad, PhiNplus1CMWritable& phiNp1CM)
{
   BoostZHyp toLabFrame(phiNCM.lambda());
   phiNp1CM.clear();
   FourVector qBar;
   for (Configuration::iterator it = phiNCM.beginFinal(); it != phiNCM.endFinal(); it++) {
      if (!IsQCD()(*it))
         // move over the leptons and boost them to lab frame
         qBar += phiNp1CM.insertFinal(Particle(it->id(), toLabFrame(it->p())))->p();
   }
   const double pBarPlus = phiNCM.emitter()->p().plus() / phiNCM.lambda();
   const double pBarMinus = phiNCM.emitter()->p().minus() * phiNCM.lambda();

   PdgID initialAID = phiNCM.initialA()->id();
   PdgID initialBID = phiNCM.initialB()->id();

   if (_qTfixed) {
      if (!splitFSRfixqT(qBar, pBarPlus, pBarMinus, phiRad, initialAID, initialBID, phiNp1CM))
         return false;
   } else {
      if (!splitFSR(qBar, pBarPlus, pBarMinus, phiRad, initialAID, initialBID, phiNp1CM))
         return false;
   }
   // boost into the new CM frame
   const BoostZHyp toCMframe(sqrt(phiNp1CM.initialB()->p().plus()/phiNp1CM.initialA()->p().minus()));
   phiNp1CM.apply(toCMframe);

   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Generates the splitting in the Lab frame for an FSR emission.
 *
 * \param phiNCM The N-body %Configuration in the CM frame and the %Particle to split
 * \param phiRad The kinematic and flavor variables in the splitting
 * \param[out] phiNp1CM The resulting (N+1)-body Configuration
 * \return \c true if the splitting could be performed, \c false otherwise
 *
 * Most of the actual formulas for the splitting are relegated to another function
 */
bool Tau0FRMapping::splitFSRLab(PhiN phiN, const PhiRadTau0FR& phiRad, PhiNplus1CMWritable& phiNp1)
{
   phiNp1.clear();
   FourVector qBar;
   for (Configuration::iterator it = phiN.beginFinal(); it != phiN.endFinal(); it++) {
      if (!IsQCD()(*it))
         // move over the leptons
         qBar += phiNp1.insertFinal(*it)->p();
   }

   const double pBarPlus = phiN.emitter()->p().plus();
   const double pBarMinus = phiN.emitter()->p().minus();

   PdgID initialAID = phiN.initialA()->id();
   PdgID initialBID = phiN.initialB()->id();

   if (_qTfixed)
      return splitFSRfixqT(qBar, pBarPlus, pBarMinus, phiRad, initialAID, initialBID, phiNp1);
   else
      return splitFSR(qBar, pBarPlus, pBarMinus, phiRad, initialAID, initialBID, phiNp1);
}

//-------------------------------------------------------------------------------------
/**
 * \brief Implements the FSR splitting formulas in the lab frame.
 *
 * \param qBar The momentum of the vector boson
 * \param pBarPlus, \param pBarMinus The components of the emitter momentum
 * \param phiRad The kinematic and flavor variables of the splitting
 * \param idA The PDG id of initial parton A
 * \param idB The PDG id of initial parton B
 * \param[out] phiNp1 The final (N+1) body configuration in the lab frame
 * \return \c true if the splitting could be performed, \c false otherwise
 */
bool Tau0FRMapping::splitFSR(FourVector qBar, double pBarPlus, double pBarMinus, const PhiRadTau0FR& phiRad, PdgID idA, PdgID idB, PhiNplus1CMWritable& phiNp1)
{
   assert(pBarPlus >= 0.0);
   assert(pBarMinus >= 0.0);

   const double p12Plus = pBarPlus;
   const double p12Minus = pBarMinus;

   //Get the energies of p1 and p2, as well as p12Sq
   const double p20 = 0.5 * phiRad.xi() *_cmEnergy;
   const double p10 = 0.5 * (p12Plus + p12Minus) - p20;
   const double p12Sq = 2. * p10 * p20 * (1. - phiRad.y());
   if (p12Sq < 0.0 || p12Sq > p12Plus * p12Minus)
      return false;

   const double qPerpSq = p12Plus * p12Minus - p12Sq;
   const double mLLSq = qBar.square();
   const double qPlusMinusRescale = sqrt((mLLSq + qPerpSq) / (mLLSq + qBar.perp2()));
   const double qPlus = qBar.plus() * qPlusMinusRescale;
   const double qMinus = qBar.minus() * qPlusMinusRescale;
   assert(qPlus >= 0.0);
   assert(qMinus >= 0.0);

   const double qPerpRescale = sqrt(qPerpSq / qBar.perp2());
   const FourVector q(0.5 * (qMinus + qPlus), qPerpRescale * qBar.p1(), qPerpRescale * qBar.p2(), 0.5 * (qMinus - qPlus));
   assert(compareRelative(q.square(), mLLSq) == 0);

   // first modify the momenta of the already-inserted non-strongly interacting particles
   assert(std::count_if(phiNp1.beginFinal(), phiNp1.endFinal(), IsQCD()) == 0);
   Boost toCMqBar = Boost(qBar).inverse();
   Boost toLabq(q);
   phiNp1.apply(toCMqBar);
   phiNp1.apply(toLabq);

   //Construct p12Mu
   const FourVector p12(0.5 * (p12Minus + p12Plus), -q.p1(), -q.p2(), 0.5 * (p12Minus - p12Plus));

   //Construct p2Mu
   double cosTheta = (2. * p20 - p12.p0()) / p12.pvec().magnitude();
   double p2z = p20 - 0.5 * (p12.p0() - p12.pvec().magnitude()) * (1. - cosTheta);
   const double p2T2 = p20*p20 - p2z*p2z;
   if (p2T2 < 0.0)
      return false;
   const double p2T = sqrt(p2T2);
   FourVector p2(p20, p2T * cos(phiRad.phi()), p2T * sin(phiRad.phi()), p2z);

   //Construct p1Mu
   const Rotation alignZToP12(ThreeVector(0., 0., 1.), p12.pvec());
   p2.apply(alignZToP12);
   FourVector p1(p12 - p2);
   //In case p1Plus or p1Minus is negative, adjust the p10 component such that p1 is on-shell
   // NOTE: This assumes massless particles
   if ((p1.plus() < 0) || (p1.minus() < 0))
      p1.p0() = p1.pvec().magnitude();

   // add initial particles
   // recompute total plus and minus to improve total momentum conservation in the config
   const double minus = p1.minus() + p2.minus() + qMinus;
   const double plus = p1.plus() + p2.plus() + qPlus;
   if (minus > _cmEnergy || plus > _cmEnergy)
      return false;

   // after all checks, construct the N+1 phase space point
   // add the emitted and sister partons
   phiNp1.insertEmitted(Particle(phiRad.idRight(), p2, 0, 0));
   phiNp1.insertFinalSister(Particle(phiRad.idLeft(), p1, 0, 0));

   phiNp1.setInitialA(Particle(idA, 0.5*minus, 0.0, 0.0, 0.5*minus));
   phiNp1.setInitialB(Particle(idB, 0.5*plus, 0.0, 0.0, -0.5*plus));
   phiNp1.xA() = minus / _cmEnergy;
   phiNp1.xB() = plus / _cmEnergy;

   // final accuracy checks
#ifndef NDEBUG
   if (!checkMomentumConservation(phiNp1.config(), TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::splitFSR") << "Phi_N+1 point from splitting violates momentum conservation." << Log::end;
   if (!checkOnShellConditions(phiNp1.config(), TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::splitFSR") << "Phi_N+1 point from splitting violates on-shell conditions." << Log::end;
#endif

   // TODO: discuss whether to include also the Jacobian to rescale the PhiRad variables to match the definition in the Subtractions
   phiNp1.jacobian() = p10 / p12.pvec().magnitude();
   phiNp1.jacobian() *= _cmEnergy * _cmEnergy * phiRad.xi() * math_constants::_1_64pi3();

   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Implements the FSR splitting formulas with fixed qT in the lab frame.
 *
 * \param q The momentum of the vector boson
 * \param pBarPlus, \param pBarMinus The components of the emitter momentum
 * \param phiRad The kinematic and flavor variables of the splitting
 * \param idA The PDG id of initial parton A
 * \param idB The PDG id of initial parton B
 * \param[out] phiNp1 The final (N+1) body configuration in the lab frame
 * \return \c true if the splitting could be performed, \c false otherwise
 */
bool Tau0FRMapping::splitFSRfixqT(const FourVector q, double pBarPlus, double pBarMinus, const PhiRadTau0FR& phiRad, PdgID idA, PdgID idB, PhiNplus1CMWritable& phiNp1)
{
   assert(pBarPlus >= 0.0);
   assert(pBarMinus >= 0.0);

   const double qPlus = q.plus();
   const double qMinus = q.minus();
   const double exp2Y = qMinus / qPlus;
   assert(qPlus >= 0.0);
   assert(qMinus >= 0.0);

   // this is the small parameter that determines the scaling of p12^2
   const double p20 = 0.5 * phiRad.xi() *_cmEnergy;
   const double p12eps = p20 * (1. - phiRad.y());
   double p12Plus;
   double p12Minus;
   if (pBarMinus > pBarPlus * exp2Y) {
      p12Plus = pBarPlus;
      p12Minus = (q.perp2() + (pBarPlus - 2.0*p20) * p12eps) / (pBarPlus - p12eps);
      assert(p12Minus >= 0.0);
      if (p12Minus < 0.0) {
         return false;
      }
      // check that we do not exceed available Ecm (too large xi)
      if (p12Minus + qMinus > _cmEnergy)
         return false;
      phiNp1.jacobian() = pBarPlus / (pBarPlus - p12eps);
   } else {
      p12Minus = pBarMinus;
      p12Plus = (q.perp2() + (pBarMinus - 2.0*p20) * p12eps) / (pBarMinus - p12eps);
      assert(p12Plus >= 0.0);
      if (p12Plus < 0.0) {
         return false;
      }
      // check that we do not exceed available Ecm (too large xi)
      if (p12Plus + qPlus > _cmEnergy)
         return false;
      phiNp1.jacobian() = pBarMinus / (pBarMinus - p12eps);
   }

   const FourVector p12(0.5 * (p12Minus + p12Plus), -q.p1(), -q.p2(), 0.5 * (p12Minus - p12Plus));
//    assert(p12.square() >= 0.0);
   if (p12.square() < 0.0) {
      // warn if we get spacelike p12 (can only be due to numerical inaccuracy)
      Log::warning("Tau0FRMapping::splitFSRfixqT") << "p12^2 = " << p12.square() << "is spacelike for xi = " << phiRad.xi() << ", 1-y = " << 1. -  phiRad.y() << Log::end;
      return false;
   }

   const double p10 = p12.p0() - p20;
   //  p2T = sqrt(p20*p20 - p2z*p2z);
   const double tmp = p20 / p12.pvec().magnitude();
   const double p2T = tmp * p10 * sqrt((1. - phiRad.y()*phiRad.y()));
   const double p2z = tmp * (p10 * phiRad.y() + p20);
   FourVector p2(p20, p2T * cos(phiRad.phi()), p2T * sin(phiRad.phi()), p2z);

   const Rotation alignZToP12(ThreeVector(0., 0., 1.), p12.pvec());
   p2.apply(alignZToP12);
   const FourVector p1(p12 - p2);
   phiNp1.jacobian() *= p1.pvec().magnitude() / p12.pvec().magnitude();

   // after all checks, construct the N+1 phase space point
   // add the emitted and sister partons
   phiNp1.insertEmitted(Particle(phiRad.idRight(), p2, 0, 0));
   phiNp1.insertFinalSister(Particle(phiRad.idLeft(), p1, 0, 0));

   // add initial particles
   // recompute total plus and minus to improve total momentum conservation in the config
   const double minus = p1.minus() + p2.minus() + qMinus;
   const double plus = p1.plus() + p2.plus() + qPlus;
   phiNp1.setInitialA(Particle(idA, 0.5*minus, 0.0, 0.0, 0.5*minus));
   phiNp1.setInitialB(Particle(idB, 0.5*plus, 0.0, 0.0, -0.5*plus));
   phiNp1.xA() = minus / _cmEnergy;
   phiNp1.xB() = plus / _cmEnergy;

   // final accuracy checks
#ifndef NDEBUG
   if (!checkMomentumConservation(phiNp1.config(), TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::splitFSRfixqT") << "Phi_N+1 point from splitting violates momentum conservation." << Log::end;
   if (!checkOnShellConditions(phiNp1.config(), TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::splitFSRfixqT") << "Phi_N+1 point from splitting violates on-shell conditions." << Log::end;
#endif

   // TODO: discuss whether to include also the Jacobian to rescale the PhiRad variables to match the definition in the Subtractions
   phiNp1.jacobian() *= _cmEnergy * _cmEnergy * phiRad.xi() * math_constants::_1_64pi3();
   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Performs the projection for ISR emissions in the CM frame.
 *
 * \param configNp1CM The (N+1)-body Configuration in the CM frame
 * \param emitted, \param sister Iterators to the emitted and sister particle
 * \param xA, \param xB The values of xA and xB
 * \param[out] configNCM The projected config in the CM frame
 * \param[out] xBarA, \param[out] xBarB The values of xBarA and xBarB to be returned
 * \return \c true if the projection could be performed, \c false otherwise
 *
 * Most of the actual formulas for the splitting are relegated to another function
 */
bool Tau0FRMapping::projectISRCM(const Configuration& configNp1CM, Configuration::iterator emitted, Configuration::iterator sister, double xA, double xB, Configuration& configNCM, double& xBarA, double& xBarB) const
{
   const double lambdaCM = sqrt(xA / xB);
   const BoostZHyp toLabFrame(lambdaCM);
   configNCM.clear();
   xBarA = std::numeric_limits<double>::quiet_NaN();
   xBarB = std::numeric_limits<double>::quiet_NaN();

   FourVector q;
   Configuration::iterator spectator;
   for (Configuration::iterator it = configNp1CM.beginFinal(); it != configNp1CM.endFinal(); ++it) {
      if (!IsQCD()(*it))
         // move over the leptons and boost them to lab frame
         q += configNCM.insertFinal(Particle(it->id(), toLabFrame(it->p())))->p();
      else if (it != emitted)
         spectator = it;
   }

   const double p1Plus = spectator->p().plus() / lambdaCM;
   const double p1Minus = spectator->p().minus() * lambdaCM;
   assert(p1Plus >= 0.0);
   assert(p1Minus >= 0.0);

   double p2Plus = emitted->p().plus() / lambdaCM;
   double p2Minus = emitted->p().minus() * lambdaCM;
   assert(p2Plus >= 0.0);
   assert(p2Minus >= 0.0);

   // get ids of initial particles
   PdgID idA, idB;
   Emission emission(*emitted, *sister);
   if (emission.isISRA()) {
      idA = emission.mother();
      idB = configNp1CM.initialB()->id();
   } else {
      idA = configNp1CM.initialA()->id();
      idB = emission.mother();
   }

   if (_qTfixed) {
      if (!projectISRfixqT(q, p1Plus, p1Minus, p2Plus, p2Minus, spectator->id(), idA, idB, configNCM))
         return false;
   } else {
      if (!projectISR(q, p1Plus, p1Minus, p2Plus, p2Minus, spectator->p().p1(), spectator->p().p2(), spectator->id(), idA, idB, configNCM))
         return false;
   }
   // boost into the new CM frame
   xBarA = configNCM.initialA()->p().minus() / _cmEnergy;
   xBarB = configNCM.initialB()->p().plus() / _cmEnergy;
   const BoostZHyp toCMframe(sqrt(xBarB/xBarA));
   configNCM.apply(toCMframe);

   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Performs the projection for ISR emissions in the Lab frame.
 *
 * \param configNp1 The (N+1)-body Configuration in the lab frame
 * \param emitted, \param sister Iterators to the emitted and sister particle
 * \param[out] configN The projected config in the lab frame
 * \return \c true if the projection could be performed, \c false otherwise
 *
 * Most of the actual formulas for the splitting are relegated to another function
 */
bool Tau0FRMapping::projectISRLab(const Configuration& configNp1, Configuration::iterator emitted, Configuration::iterator sister, Configuration& configN) const
{
   configN.clear();
   FourVector q;
   Configuration::iterator spectator;
   for (Configuration::iterator it = configNp1.beginFinal(); it != configNp1.endFinal(); ++it) {
      if (!IsQCD()(*it))
         // move over the leptons and boost them to lab frame
         q += configN.insertFinal(*it)->p();
      else if (it != emitted)
         spectator = it;
   }

   const double p1Plus = spectator->p().plus();
   const double p1Minus = spectator->p().minus();
   assert(p1Plus >= 0.0);
   assert(p1Minus >= 0.0);

   double p2Plus = emitted->p().plus();
   double p2Minus = emitted->p().minus();
   assert(p2Plus >= 0.0);
   assert(p2Minus >= 0.0);

   // get ids of initial particles
   PdgID idA, idB;
   Emission emission(*emitted, *sister);
   if (emission.isISRA()) {
      idA = emission.mother();
      idB = configNp1.initialB()->id();
   } else {
      idA = configNp1.initialA()->id();
      idB = emission.mother();
   }

   if (_qTfixed)
      return projectISRfixqT(q, p1Plus, p1Minus, p2Plus, p2Minus, spectator->id(), idA, idB, configN);
   else
      return projectISR(q, p1Plus, p1Minus, p2Plus, p2Minus, spectator->p().p1(), spectator->p().p2(), spectator->id(), idA, idB, configN);
}

//-------------------------------------------------------------------------------------
/**
 * \brief Implements the ISR projection formulas for ISR emissions.
 *
 * \param q The momentum of the vector boson
 * \param p1Plus, \param p1Minus The components of the emitted momentum
 * \param p2Plus, \param p2Minus The components of the sister momentum
 * \param spectator The PDG id of the final state particle not taking part in the splitting
 * \param idA The PDG id of initial parton A
 * \param idB The PDG id of initial parton B
 * \param[out] configN The final (N+1) body configuration in the lab frame
 * \return \c true if the projection could be performed, \c false otherwise
 */
bool Tau0FRMapping::projectISR(FourVector q, double p1Plus, double p1Minus, double p2Plus, double p2Minus, const double p1x, const double p1y, PdgID spectator, PdgID idA, PdgID idB, Configuration& configN) const
{
   const double qPlus = q.plus();
   const double qMinus = q.minus();
   assert(qPlus >= 0.0);
   assert(qMinus >= 0.0);
   const double exp2Y = qMinus / qPlus;

   double pPlus, pMinus;

   if (p1Plus * exp2Y < p1Minus) {
      // hatp1+ < hatp1-
      if (p2Plus < p2Minus / exp2Y) {
         // hatp2^+ < hatp2^- (same hemisphere)
         pPlus = p1Plus + p2Plus;
      } else {
         // hatp2- <= hatp2+ (opposite hemisphere)
         pPlus = p1Plus + p2Minus / exp2Y;
      }
      pMinus = pPlus * p1Minus / p1Plus;
   } else {
      // hatp1- <= hatp1+
      if (p2Plus * exp2Y < p2Minus) {
         // hatp2+ < hatp2- (opposite hemisphere)
         pMinus = p1Minus + p2Plus * exp2Y;
      } else {
         // hatp2- <= hatp2+ (same hemisphere)
         pMinus = p1Minus + p2Minus;
      }
      pPlus = pMinus * p1Plus / p1Minus;
   }

   const double qBarPerpSq = pPlus * pMinus;
   const double mLLSq = q.square();
   const double qRescale = sqrt((mLLSq + qBarPerpSq) / (mLLSq + q.perp2()));
   const double qBarPlus = qRescale * qPlus;
   const double qBarMinus = qRescale * qMinus;

   const double minus = pMinus + qBarMinus;
   const double plus = pPlus + qBarPlus;
   // if we exceed Ecm (possible for large exp2Y) there is no Tau0-preserving projection
   if (minus > _cmEnergy || plus > _cmEnergy)
      return false;

   const double pPerpRescale = sqrt(qBarPerpSq / (p1Plus * p1Minus));
   const FourVector p(0.5 * (pMinus + pPlus), pPerpRescale * p1x, pPerpRescale * p1y, 0.5 * (pMinus - pPlus));

   // modify the momenta of the already-inserted non-strongly-interacting particles
   const FourVector qBar(0.5 * (qBarMinus + qBarPlus), -p.p1(), -p.p2(), 0.5 * (qBarMinus - qBarPlus));
   Boost toLabqBar(qBar);
   Boost toLabq(q);
   configN.apply(toLabq.inverse());
   configN.apply(toLabqBar);

   // set initial state particles
   configN.setInitialA(Particle(idA, 0.5 * minus, 0., 0., 0.5 * minus));
   configN.setInitialB(Particle(idB, 0.5 * plus, 0., 0., -0.5 * plus));

   // insert finale state parton
   configN.insertFinal(Particle(spectator, p));

  // final accuracy checks
#ifndef NDEBUG
   if (!checkMomentumConservation(configN, TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::projectISR") << "Phi_N point from projection violates momentum conservation." << Log::end;
   if (!checkOnShellConditions(configN, TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::projectISR") << "Phi_N point from projection violates on-shell conditions." << Log::end;
#endif

   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Implements the ISR splitting formulas with fixed qT in the lab frame.
 *
 * \param q The momentum of the vector boson
 * \param p1Plus, \param p1Minus The components of the emitted momentum
 * \param p2Plus, \param p2Minus The components of the sister momentum
 * \param spectator The PDG id of the final state particle not taking part in the splitting
 * \param idA The PDG id of initial parton A
 * \param idB The PDG id of initial parton B
 * \param[out] configN The final (N+1) body configuration in the lab frame
 * \return \c true if the projection could be performed, \c false otherwise
 */
bool Tau0FRMapping::projectISRfixqT(FourVector q, double p1Plus, double p1Minus, double p2Plus, double p2Minus, PdgID spectator, PdgID idA, PdgID idB, Configuration& configN) const
{
   const double qPlus = q.plus();
   const double qMinus = q.minus();
   assert(qPlus >= 0.0);
   assert(qMinus >= 0.0);
   const double exp2Y = qMinus / qPlus;
   const double qT2 = q.perp2();

   double pPlus, pMinus;

   if (p1Plus * exp2Y < p1Minus) {
      // hatp1+ < hatp1-
      if (p2Plus < p2Minus / exp2Y) {
         // hatp2^+ < hatp2^- (same hemisphere)
         pPlus = p1Plus + p2Plus;
      } else {
         // hatp2- <= hatp2+ (opposite hemisphere)
         pPlus = p1Plus + p2Minus / exp2Y;
      }
      pMinus = qT2 / pPlus;
      // we must have hatp+ < hatp-, otherwise there is no Tau0-preserving projection
      if (pMinus < pPlus * exp2Y)
         return false;
   } else {
      // hatp1- <= hatp1+
      if (p2Plus * exp2Y < p2Minus) {
         // hatp2+ < hatp2- (opposite hemisphere)
         pMinus = p1Minus + p2Plus * exp2Y;
      } else {
         // hatp2- <= hatp2+ (same hemisphere)
         pMinus = p1Minus + p2Minus;
      }
      pPlus = qT2 / pMinus;
      // we must have hatp- < hatp+, otherwise there is no Tau0-preserving projection
      if (pMinus > pPlus * exp2Y)
         return false;
   }

   const double minus = pMinus + qMinus;
   const double plus = pPlus + qPlus;
   // if we exceed Ecm (possible for large exp2Y) there is no Tau0-preserving projection
   if (minus > _cmEnergy || plus > _cmEnergy)
      return false;

   configN.setInitialA(Particle(idA, 0.5 * minus, 0., 0., 0.5 * minus));
   configN.setInitialB(Particle(idB, 0.5 * plus, 0., 0., -0.5 * plus));

   // insert finale state parton
   const FourVector p(0.5 * (pMinus + pPlus), -q.p1(), -q.p2(), 0.5 * (pMinus - pPlus));
   configN.insertFinal(Particle(spectator, p));

   // final accuracy checks
#ifndef NDEBUG
   if (!checkMomentumConservation(configN, TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::projectISRfixqT") << "Phi_N point from projection violates momentum conservation." << Log::end;
   if (!checkOnShellConditions(configN, TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::projectISRfixqT") << "Phi_N point from projection violates on-shell conditions." << Log::end;
#endif

   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Performs the projection for FSR emissions in the CM frame.
 *
 * \param configNp1CM The (N+1)-body Configuration in the CM frame
 * \param emitted, \param sister Iterators to the emitted and sister particle
 * \param xA, \param xB The values of xA and xB
 * \param[out] configNCM The projected config in the CM frame
 * \param[out] xBarA, \param[out] xBarB The values of xBarA and xBarB to be returned
 * \return \c true if the projection could be performed, \c false otherwise
 *
 * Most of the actual formulas for the splitting are relegated to another function
 */
bool Tau0FRMapping::projectFSRCM(const Configuration& configNp1CM, Configuration::iterator emitted, Configuration::iterator sister, double xA, double xB, Configuration& configNCM, double& xBarA, double& xBarB) const
{
   const double lambdaCM = sqrt(xA / xB);
   const BoostZHyp toLabFrame(lambdaCM);
   configNCM.clear();
   xBarA = std::numeric_limits<double>::quiet_NaN();
   xBarB = std::numeric_limits<double>::quiet_NaN();

   FourVector q;
   for (Configuration::iterator it = configNp1CM.beginFinal(); it != configNp1CM.endFinal(); ++it) {
      if (!IsQCD()(*it))
         // move over the leptons and boost them to lab frame
         q += configNCM.insertFinal(Particle(it->id(), toLabFrame(it->p())))->p();
   }

   const double p1Plus = sister->p().plus() / lambdaCM;
   const double p1Minus = sister->p().minus() * lambdaCM;
   assert(p1Plus >= 0.0);
   assert(p1Minus >= 0.0);

   double p2Plus = emitted->p().plus() / lambdaCM;
   double p2Minus = emitted->p().minus() * lambdaCM;
   assert(p2Plus >= 0.0);
   assert(p2Minus >= 0.0);

   const PdgID mother = Emission(*emitted, *sister).mother();

   if (_qTfixed) {
      if (!projectFSRfixqT(q, p1Plus, p1Minus, p2Plus, p2Minus, mother, configNp1CM.initialA()->id(), configNp1CM.initialB()->id(), configNCM))
         return false;
   } else {
      if (!projectFSR(q, p1Plus, p1Minus, p2Plus, p2Minus, mother, configNp1CM.initialA()->id(), configNp1CM.initialB()->id(), configNCM))
         return false;
   }

   // boost into the new CM frame
   xBarA = configNCM.initialA()->p().minus() / _cmEnergy;
   xBarB = configNCM.initialB()->p().plus() / _cmEnergy;
   const BoostZHyp toCMframe(sqrt(xBarB/xBarA));
   configNCM.apply(toCMframe);

   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Performs the projection for FSR emissions in the Lab frame.
 *
 * \param configNp1 The (N+1)-body Configuration in the CM frame
 * \param emitted, \param sister Iterators to the emitted and sister particle
 * \param[out] configN The projected config in the lab frame
 * \return \c true if the projection could be performed, \c false otherwise
 *
 * Most of the actual formulas for the splitting are relegated to another function
 */
bool Tau0FRMapping::projectFSRLab(const Configuration& configNp1, Configuration::iterator emitted, Configuration::iterator sister, Configuration& configN) const
{
   configN.clear();
   FourVector q;
   for (Configuration::iterator it = configNp1.beginFinal(); it != configNp1.endFinal(); ++it) {
      if (!IsQCD()(*it))
         // move over the leptons
         q += configN.insertFinal(Particle(*it))->p();
   }

   const double p1Plus = sister->p().plus();
   const double p1Minus = sister->p().minus();
   assert(p1Plus >= 0.0);
   assert(p1Minus >= 0.0);

   double p2Plus = emitted->p().plus();
   double p2Minus = emitted->p().minus();
   assert(p2Plus >= 0.0);
   assert(p2Minus >= 0.0);

   const PdgID mother = Emission(*emitted, *sister).mother();

   if (_qTfixed)
      return projectFSRfixqT(q, p1Plus, p1Minus, p2Plus, p2Minus, mother, configNp1.initialA()->id(), configNp1.initialB()->id(), configN);
   else
      return projectFSR(q, p1Plus, p1Minus, p2Plus, p2Minus, mother, configNp1.initialA()->id(), configNp1.initialB()->id(), configN);
}

//-------------------------------------------------------------------------------------
/**
 * \brief Implements the FSR projection formulas in the lab frame.
 *
 * \param q The momentum of the vector boson
 * \param p1Plus, \param p1Minus The components of the emitted momentum
 * \param p2Plus, \param p2Minus The components of the sister momentum
 * \param motherID The PDG id of the mother
 * \param idA The PDG id of initial parton A
 * \param idB The PDG id of initial parton B
 * \param[out] configN The final (N+1) body configuration in the lab frame
 * \return \c true if the projection could be performed, \c false otherwise
 */
bool Tau0FRMapping::projectFSR(FourVector q, double p1Plus, double p1Minus, double p2Plus, double p2Minus, PdgID motherID, PdgID idA, PdgID idB, Configuration& configN) const
{
   //Obtain the information for p12
   const double p12Plus = p1Plus + p2Plus;
   const double p12Minus = p1Minus + p2Minus;
   const double p12PerpSq = q.perp2();

   //Calculate the information for qBar
   double qBarPerpSq = q.perp2() + p12Plus * p12Minus - p12PerpSq;
   const double mllSq = q.square();
   const double qPlusMinusRescale = sqrt((mllSq + qBarPerpSq) / (mllSq + q.perp2()));
   const double qBarPlus = q.plus() * qPlusMinusRescale;
   const double qBarMinus = q.minus() * qPlusMinusRescale;

   //Calculate the information for pBar
   const double pBarPlus = p12Plus;
   const double pBarMinus = p12Minus;

   const double minus = pBarMinus + qBarMinus;
   const double plus = pBarPlus + qBarPlus;
   if (minus > _cmEnergy || plus > _cmEnergy)
      return false;

   const double qPerpRescale = sqrt(qBarPerpSq / q.perp2());
   const FourVector pBar(0.5 * (pBarMinus + pBarPlus), - qPerpRescale * q.p1(), - qPerpRescale * q.p2(), 0.5 * (pBarMinus - pBarPlus));

   // modify the momenta of the already-inserted non-strongly-interacting particles
   const FourVector qBar(0.5 * (qBarMinus + qBarPlus), qPerpRescale * q.p1(), qPerpRescale * q.p2(), 0.5 * (qBarMinus - qBarPlus));
   Boost toLabqBar(qBar);
   Boost toLabq(q);
   configN.apply(toLabq.inverse());
   configN.apply(toLabqBar);

   // set initial state partons
   configN.setInitialA(Particle(idA, 0.5 * minus, 0., 0., 0.5 * minus));
   configN.setInitialB(Particle(idB, 0.5 * plus, 0., 0., -0.5 * plus));

   // insert final state parton
   configN.insertFinal(Particle(motherID, pBar));

   // final accuracy checks
#ifndef NDEBUG
   if (!checkMomentumConservation(configN, TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::projectFSR") << "Phi_N point from projection violates momentum conservation." << Log::end;
   if (!checkOnShellConditions(configN, TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::projectFSR") << "Phi_N point from projection violates on-shell conditions." << Log::end;
#endif

   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Implements the FSR projection formulas with fixed qT in the lab frame.
 *
 * \param q The momentum of the vector boson
 * \param p1Plus, \param p1Minus The components of the emitted momentum
 * \param p2Plus, \param p2Minus The components of the sister momentum
 * \param motherID The PDG id of the mother
 * \param idA The PDG id of initial parton A
 * \param idB The PDG id of initial parton B
 * \param[out] configN The final (N+1) body configuration in the lab frame
 * \return \c true if the projection could be performed, \c false otherwise
 */
bool Tau0FRMapping::projectFSRfixqT(FourVector q, double p1Plus, double p1Minus, double p2Plus, double p2Minus, PdgID motherID, PdgID idA, PdgID idB, Configuration& configN) const
{
   const double qPlus = q.plus();
   const double qMinus = q.minus();
   assert(qPlus >= 0.0);
   assert(qMinus >= 0.0);
   const double exp2Y = qMinus / qPlus;
   const double qT2 = q.perp2();

   double pPlus;
   double pMinus;
   if ((p1Plus + p2Plus) * exp2Y < p1Minus + p2Minus) {
      // hatp+ < hatp-
      // same hemisphere
      pPlus = p1Plus + p2Plus;
      pMinus = qT2 / pPlus;
      // we must have hatp+ < hatp-, otherwise there is no Tau0-preserving projection
      if (pMinus < pPlus * exp2Y)
         return false;
   } else {
      // same hemisphere
      pMinus = p1Minus + p2Minus;
      pPlus = qT2 / pMinus;
      // we must have hatp- < hatp+, otherwise there is no Tau0-preserving projection
      if (pMinus > pPlus * exp2Y)
         return false;
   }

   // Fail the projection for p20 > p0, which means xi > xiMax. This is purely
   // nonsingular and can only happen for p20 > p10
   if (pPlus + pMinus < p2Plus + p2Minus)
      return false;

   const double minus = pMinus + qMinus;
   const double plus = pPlus + qPlus;
   // for FSR we cannot exceed Ecm, since we also use same-hemisphere kinematics
   assert(minus <= _cmEnergy && plus <= _cmEnergy);

   // insert initial state partons
   configN.setInitialA(Particle(idA, 0.5 * minus, 0., 0., 0.5 * minus));
   configN.setInitialB(Particle(idB, 0.5 * plus, 0., 0., -0.5 * plus));

   const FourVector p(0.5 * (pMinus + pPlus), -q.p1(), -q.p2(), 0.5 * (pMinus - pPlus));
   configN.insertFinal(Particle(motherID, p));

   // final accuracy checks
#ifndef NDEBUG
   if (!checkMomentumConservation(configN, TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::projectFSRfixqT") << "Phi_N point from projection violates momentum conservation." << Log::end;
   if (!checkOnShellConditions(configN, TAU0FR_MAPPING_PRECISION, TAU0FR_MAPPING_ACCURACY))
      Log::warning("Tau0FRMapping::projectFSRfixqT") << "Phi_N point from projection violates on-shell conditions." << Log::end;
#endif
   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Tests the projection for ISR emissions in the CM frame.
 *
 * \param configNp1CM The (N+1)-body Configuration in the CM frame
 * \param emitted Iterator to the emitted particle
 * \param xA, \param xB The values of xA and xB
 * \return \c true if the projection is possible, \c false otherwise
 */
bool Tau0FRMapping::testProjectISRCM(const Configuration& configNp1CM, Configuration::iterator emitted, double xA, double xB) const
{
   const double lambdaCM = sqrt(xA / xB);
   const BoostZHyp toLabFrame(lambdaCM);

   FourVector q;
   Configuration::iterator spectator;
   for (Configuration::iterator it = configNp1CM.beginFinal(); it != configNp1CM.endFinal(); ++it) {
      if (!IsQCD()(*it))
         q += toLabFrame(it->p());
      else if (it != emitted)
         spectator = it;
   }

   double qPlus = q.plus();
   double qMinus = q.minus();
   const double exp2Y = qMinus / qPlus;
   const double qT2 = q.perp2();
   assert(qPlus >= 0.0);
   assert(qMinus >= 0.0);

   const double p1Plus = spectator->p().plus() / lambdaCM;
   const double p1Minus = spectator->p().minus() * lambdaCM;
   assert(p1Plus >= 0.0);
   assert(p1Minus >= 0.0);

   double p2Plus = emitted->p().plus() / lambdaCM;
   double p2Minus = emitted->p().minus() * lambdaCM;
   assert(p2Plus >= 0.0);
   assert(p2Minus >= 0.0);

   double pPlus;
   double pMinus;
   if (_qTfixed) {
      if (p1Plus * exp2Y < p1Minus) {
         // hatp1+ < hatp1-
         if (p2Plus < p2Minus / exp2Y) {
            // hatp2^+ < hatp2^- (same hemisphere)
            pPlus = p1Plus + p2Plus;
            assert(pPlus <= spectator->p().perp() / sqrt(exp2Y) + p2Minus / exp2Y);
         } else {
            // hatp2- <= hatp2+ (opposite hemisphere)
            pPlus = p1Plus + p2Minus / exp2Y;
            assert(pPlus <= spectator->p().perp() / sqrt(exp2Y) + p2Plus);
         }
         pMinus = qT2 / pPlus;
         // we must have hatp+ < hatp-, otherwise there is no Tau0-preserving projection
         if (pMinus < pPlus * exp2Y)
            return false;
      } else {
         // hatp1- <= hatp1+
         if (p2Plus * exp2Y < p2Minus) {
            // hatp2+ < hatp2- (opposite hemisphere)
            pMinus = p1Minus + p2Plus * exp2Y;
            assert(pMinus <= spectator->p().perp() * sqrt(exp2Y) + p2Minus);
         } else {
            // hatp2- <= hatp2+ (same hemisphere)
            pMinus = p1Minus + p2Minus;
            assert(pMinus <= spectator->p().perp() * sqrt(exp2Y) + p2Plus * exp2Y);
         }
         pPlus = qT2 / pMinus;
         // we must have hatp- < hatp+, otherwise there is no Tau0-preserving projection
         if (pMinus > pPlus * exp2Y)
            return false;
      }
   } else {
      if (p1Plus * exp2Y < p1Minus) {
         if (p2Plus < p2Minus / exp2Y) {
            pPlus = p1Plus + p2Plus;
         } else {
            pPlus = p1Plus + p2Minus / exp2Y;
         }
         pMinus = pPlus * p1Minus / p1Plus;
      } else {
         if (p2Plus * exp2Y < p2Minus) {
            pMinus = p1Minus + p2Plus * exp2Y;
         } else {
            pMinus = p1Minus + p2Minus;
         }
         pPlus = pMinus * p1Plus / p1Minus;
      }
      const double qBarPerpSq = pPlus * pMinus;
      const double mLLSq = q.square();
      const double qRescale = sqrt((mLLSq + qBarPerpSq) / (mLLSq + qT2));
      qPlus *= qRescale;
      qMinus *= qRescale;
   }

   const double minus = pMinus + qMinus;
   const double plus = pPlus + qPlus;

   // if we exceed Ecm (possible for large exp2Y) there is no Tau0-preserving projection
   if (minus > _cmEnergy || plus > _cmEnergy)
      return false;

   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Tests the projection for ISR emissions in the lab frame.
 *
 * \param configNp1 The (N+1)-body Configuration in the lab frame
 * \param emitted Iterator to the emitted particle
 * \return \c true if the projection is possible, \c false otherwise
 */
//------------------------------------------------------------------------------
bool Tau0FRMapping::testProjectISRLab(const Configuration& configNp1, Configuration::iterator emitted) const
{
   FourVector q;
   Configuration::iterator spectator;
   for (Configuration::iterator it = configNp1.beginFinal(); it != configNp1.endFinal(); ++it) {
      if (!IsQCD()(*it))
         q += it->p();
      else if (it != emitted)
         spectator = it;
   }

   double qPlus = q.plus();
   double qMinus = q.minus();
   const double exp2Y = qMinus / qPlus;
   const double qT2 = q.perp2();
   assert(qPlus >= 0.0);
   assert(qMinus >= 0.0);

   const double p1Plus = spectator->p().plus();
   const double p1Minus = spectator->p().minus();
   assert(p1Plus >= 0.0);
   assert(p1Minus >= 0.0);

   double p2Plus = emitted->p().plus();
   double p2Minus = emitted->p().minus();
   assert(p2Plus >= 0.0);
   assert(p2Minus >= 0.0);

   double pPlus;
   double pMinus;
   if (_qTfixed) {
      if (p1Plus * exp2Y < p1Minus) {
         // hatp1+ < hatp1-
         if (p2Plus < p2Minus / exp2Y) {
            // hatp2^+ < hatp2^- (same hemisphere)
            pPlus = p1Plus + p2Plus;
            assert(pPlus <= spectator->p().perp() / sqrt(exp2Y) + p2Minus / exp2Y);
         } else {
            // hatp2- <= hatp2+ (opposite hemisphere)
            pPlus = p1Plus + p2Minus / exp2Y;
            assert(pPlus <= spectator->p().perp() / sqrt(exp2Y) + p2Plus);
         }
         pMinus = qT2 / pPlus;
         // we must have hatp+ < hatp-, otherwise there is no Tau0-preserving projection
         if (pMinus < pPlus * exp2Y)
            return false;
      } else {
         // hatp1- <= hatp1+
         if (p2Plus * exp2Y < p2Minus) {
            // hatp2+ < hatp2- (opposite hemisphere)
            pMinus = p1Minus + p2Plus * exp2Y;
            assert(pMinus <= spectator->p().perp() * sqrt(exp2Y) + p2Minus);
         } else {
            // hatp2- <= hatp2+ (same hemisphere)
            pMinus = p1Minus + p2Minus;
            assert(pMinus <= spectator->p().perp() * sqrt(exp2Y) + p2Plus * exp2Y);
         }
         pPlus = qT2 / pMinus;
         // we must have hatp- < hatp+, otherwise there is no Tau0-preserving projection
         if (pMinus > pPlus * exp2Y)
            return false;
      }
   } else {
      if (p1Plus * exp2Y < p1Minus) {
         if (p2Plus < p2Minus / exp2Y) {
            pPlus = p1Plus + p2Plus;
         } else {
            pPlus = p1Plus + p2Minus / exp2Y;
         }
         pMinus = pPlus * p1Minus / p1Plus;
      } else {
         if (p2Plus * exp2Y < p2Minus) {
            pMinus = p1Minus + p2Plus * exp2Y;
         } else {
            pMinus = p1Minus + p2Minus;
         }
         pPlus = pMinus * p1Plus / p1Minus;
      }

      const double qBarPerpSq = pPlus * pMinus;
      const double mLLSq = q.square();
      const double qRescale = sqrt((mLLSq + qBarPerpSq) / (mLLSq + qT2));
      qPlus *= qRescale;
      qMinus *= qRescale;
   }

   const double minus = pMinus + qMinus;
   const double plus = pPlus + qPlus;

   // if we exceed Ecm (possible for large exp2Y) there is no Tau0-preserving projection
   if (minus > _cmEnergy || plus > _cmEnergy)
      return false;

   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Tests the projection for FSR emissions in the CM frame.
 *
 * \param configNp1CM The (N+1)-body Configuration in the CM frame
 * \param emitted, \param sister Iterators to the emitted and sister particle
 * \param xA, \param xB The values of xA and xB
 * \return \c true if the projection is possible, \c false otherwise
 */
bool Tau0FRMapping::testProjectFSRCM(const Configuration& configNp1CM, Configuration::iterator emitted, Configuration::iterator sister, double xA, double xB) const
{
   const double lambdaCM = sqrt(xA / xB);
   const BoostZHyp toLabFrame(lambdaCM);

   FourVector q;
   for (Configuration::iterator it = configNp1CM.beginFinal(); it != configNp1CM.endFinal(); ++it) {
      if (!IsQCD()(*it))
         q += toLabFrame(it->p());
   }

   const double p1Plus = sister->p().plus() / lambdaCM;
   const double p1Minus = sister->p().minus() * lambdaCM;
   assert(p1Plus >= 0.0);
   assert(p1Minus >= 0.0);

   const double p2Plus = emitted->p().plus() / lambdaCM;
   const double p2Minus = emitted->p().minus() * lambdaCM;
   assert(p2Plus >= 0.0);
   assert(p2Minus >= 0.0);

   if (_qTfixed) {
      const double qPlus = q.plus();
      const double qMinus = q.minus();
      const double exp2Y = qMinus / qPlus;
      assert(qPlus >= 0.0);
      assert(qMinus >= 0.0);
      const double qT2 = q.perp2();

      double pPlus;
      double pMinus;
      if ((p1Plus + p2Plus) * exp2Y < p1Minus + p2Minus) {
         // hatp+ < hatp-
         // same hemisphere
         pPlus = p1Plus + p2Plus;
         pMinus = qT2 / pPlus;
         // we must have hatp+ < hatp-, otherwise there is no Tau0-preserving projection
         if (pMinus < pPlus * exp2Y)
            return false;
      } else {
         // same hemisphere
         pMinus = p1Minus + p2Minus;
         pPlus = qT2 / pMinus;
         // we must have hatp- < hatp+, otherwise there is no Tau0-preserving projection
         if (pMinus > pPlus * exp2Y)
            return false;
      }

      // Fail the projection for p20 > p0, which means xi > xiMax. This is purely
      // nonsingular and can only happen for p20 > p10
      if (pPlus + pMinus < p2Plus + p2Minus)
         return false;

      // for FSR we cannot exceed Ecm, since we also use same-hemisphere kinematics
      assert(pMinus + qMinus <= _cmEnergy && pPlus + qPlus <= _cmEnergy);
   } else {

      //Obtain the information for p12
      const double p12Plus = p1Plus + p2Plus;
      const double p12Minus = p1Minus + p2Minus;
      const double p12PerpSq = q.perp2();

      //Calculate the information for qBar
      double qBarPerpSq = q.perp2() + p12Plus * p12Minus - p12PerpSq;
      const double mllSq = q.square();
      const double qPlusMinusRescale = sqrt((mllSq + qBarPerpSq) / (mllSq + q.perp2()));
      const double qBarPlus = q.plus() * qPlusMinusRescale;
      const double qBarMinus = q.minus() * qPlusMinusRescale;

      //Calculate the information for pBar
      const double pBarPlus = p12Plus;
      const double pBarMinus = p12Minus;

      const double minus = pBarMinus + qBarMinus;
      const double plus = pBarPlus + qBarPlus;
      if (minus > _cmEnergy || plus > _cmEnergy)
         return false;
   }
   return true;
}

//-------------------------------------------------------------------------------------
/**
 * \brief Tests the projection for FSR emissions in the lab frame.
 *
 * \param configNp1 The (N+1)-body Configuration in the CM frame
 * \param emitted, \param sister Iterators to the emitted and sister particle
 * \return \c true if the projection is possible, \c false otherwise
 */
bool Tau0FRMapping::testProjectFSRLab(const Configuration& configNp1, Configuration::iterator emitted, Configuration::iterator sister) const
{
   FourVector q;
   Configuration::iterator spectator;
   for (Configuration::iterator it = configNp1.beginFinal(); it != configNp1.endFinal(); ++it) {
      if (!IsQCD()(*it))
         q += it->p();
      else if (it != emitted)
         spectator = it;
   }

   const double p1Plus = sister->p().plus();
   const double p1Minus = sister->p().minus();
   assert(p1Plus >= 0.0);
   assert(p1Minus >= 0.0);

   const double p2Plus = emitted->p().plus();
   const double p2Minus = emitted->p().minus();
   assert(p2Plus >= 0.0);
   assert(p2Minus >= 0.0);

   if (_qTfixed) {
      const double qPlus = q.plus();
      const double qMinus = q.minus();
      const double exp2Y = qMinus / qPlus;
      assert(qPlus >= 0.0);
      assert(qMinus >= 0.0);
      const double qT2 = q.perp2();

      double pPlus;
      double pMinus;
      if ((p1Plus + p2Plus) * exp2Y < p1Minus + p2Minus) {
         // hatp+ < hatp-
         // same hemisphere
         pPlus = p1Plus + p2Plus;
         pMinus = qT2 / pPlus;
         // we must have hatp+ < hatp-, otherwise there is no Tau0-preserving projection
         if (pMinus < pPlus * exp2Y)
            return false;
      } else {
         // same hemisphere
         pMinus = p1Minus + p2Minus;
         pPlus = qT2 / pMinus;
         // we must have hatp- < hatp+, otherwise there is no Tau0-preserving projection
         if (pMinus > pPlus * exp2Y)
            return false;
      }

      // Fail the projection for p20 > p0, which means xi > xiMax. This is purely
      // nonsingular and can only happen for p20 > p10
      if (pPlus + pMinus < p2Plus + p2Minus)
         return false;

      // for FSR we cannot exceed Ecm, since we also use same-hemisphere kinematics
      assert(pMinus + qMinus <= _cmEnergy && pPlus + qPlus <= _cmEnergy);
   } else {

      //Obtain the information for p12
      const double p12Plus = p1Plus + p2Plus;
      const double p12Minus = p1Minus + p2Minus;
      const double p12PerpSq = q.perp2();

      //Calculate the information for qBar
      double qBarPerpSq = q.perp2() + p12Plus * p12Minus - p12PerpSq;
      const double mllSq = q.square();
      const double qPlusMinusRescale = sqrt((mllSq + qBarPerpSq) / (mllSq + q.perp2()));
      const double qBarPlus = q.plus() * qPlusMinusRescale;
      const double qBarMinus = q.minus() * qPlusMinusRescale;

      //Calculate the information for pBar
      const double pBarPlus = p12Plus;
      const double pBarMinus = p12Minus;

      const double minus = pBarMinus + qBarMinus;
      const double plus = pBarPlus + qBarPlus;
      if ((minus > _cmEnergy) || (plus > _cmEnergy))
         return false;
   }
   return true;
}

} // namespace Geneva
