//------------------------------------------------------------------------------
/// \file UnitCubeRadGenerator.cpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class UnitCubeRadGenerator
//------------------------------------------------------------------------------

#include "Geneva/Components/UnitCubeRadGenerator.hpp"

#include "Geneva/Core/Physics/UnitCubeRad.hpp"
#include "Geneva/Core/Physics/PdgID.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <cassert>
#include <iostream>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief Generates %UnitCubeRad kinematics
 * \param rand1,\param rand2, \param rand3 flat random numbers
 * \param xRad The %UnitCubeRad to store the result
 *  and the generation weight in
 *
 */
void UnitCubeRadGenerator::generate(double rand1, double rand2, double rand3, UnitCubeRad& xRad) const
{
   double jac = 1.;
   if (xRad.radType() == UnitCubeRad::kFSR) {
      if (_singularJacobian) {
#ifdef NEW
         xRad.x1() = (1. - _x1TinyCutoff) - (1. - 2.*_x1TinyCutoff) * pow(1.0 - rand1, 2);
         jac *= (1. - 2.*_x1TinyCutoff) * 2.*(1.0 - rand1);
#else
         xRad.x1() = (1. - 2.*_x1TinyCutoff) * (1. - rand1 * rand1) + _x1TinyCutoff;
         jac *= 2.0 * rand1 * (1. - 2.*_x1TinyCutoff);
#endif
      } else {
#ifdef NEW
         xRad.x1() = _x1TinyCutoff + (1. - 2.*_x1TinyCutoff) * rand1;
         jac *= (1. - 2.*_x1TinyCutoff);
#else
         xRad.x1() = (1. - 2.*_x1TinyCutoff) * rand1 * rand1 + _x1TinyCutoff;
         jac *= (1. - 2.*_x1TinyCutoff) * 2.*rand1;
#endif
      }
   } else {
      xRad.x1() = (3. - 2.*rand1) * pow(rand1, 2) * (1. - 2.* _x1TinyCutoff) + _x1TinyCutoff;
      jac *= 6.0 * (1.0 - rand1) * rand1 * (1.0 - 2.0 * _x1TinyCutoff);
   }

   double newX2tinyCutoff =  _independentTinyCuts ? _x2TinyCutoff :  std::max(_x2TinyCutoff, _x1TinyCutoff / xRad.x1());

#ifdef NEW
   double tmp = 1. - 2.*rand2;
   xRad.x2() = 0.5 + 0.75 * (tmp - pow(tmp, 3) / 3.) * (1. - newX2tinyCutoff);
   jac *= (1.0 - newX2tinyCutoff) * 1.5 * (tmp * tmp - 1.);
#else
   double tmp = rand2 * rand2 * rand2;
   xRad.x2() = 0.5 * (1. + (1.0 - newX2tinyCutoff) * (1.0 - 2.*tmp));
   jac *= 3 * rand2 * rand2 * (1.0 - newX2tinyCutoff);
#endif

   xRad.x3() = rand3;
   assert(xRad.x1() > 0.0 && (xRad.x1() < 1.0));
   assert((xRad.x2() > 0.0) && (xRad.x2() < 1.0));
   assert((xRad.x3() > 0.0) && (xRad.x3() < 1.0));
   xRad.weight() = jac;
   return;
}

//------------------------------------------------------------------------------
/**
 * \brief Generates %UnitCubeRad kinematics
 * \param rand1,\param rand2, \param rand3 flat random numbers
 * \param xRad The %UnitCubeRad to store the result
 *  and the generation weight in
 *
 * The sampling of the y depends on the sampling parameter ySampling
 *
 */
void UnitCubeRadGenerator::generate(double rand1, double rand2, double rand3, UnitCubeRad& xRad, double ySampling) const
{
   double jac = 1.;
   //Let's first transform variables from randi to zi. This is where we do the importance sampling towards the singular values

   double z1;
   if (_singularJacobian) {
      z1 = (1. - 2.*_x1TinyCutoff) * (1. - rand1 * rand1) + _x1TinyCutoff;
      jac *= 2.0 * rand1 * (1. - 2.*_x1TinyCutoff);

   } else {
      z1 = pow(rand1, 2) * (1. - 2.* _x1TinyCutoff) + _x1TinyCutoff;
      jac *= 2. * rand1 * (1.0 - 2.0 * _x1TinyCutoff);

   }
   double newX2tinyCutoff =  _independentTinyCuts ? _x2TinyCutoff :  std::max(_x2TinyCutoff, _x1TinyCutoff / xRad.x1());
   double tmp = pow(rand2, ySampling);
   jac *= ySampling * pow(rand2, ySampling - 1.);
//   double tmp = rand2 * rand2 * rand2;
//   jac *= 3. * rand2 * rand2;
   double z2 = 0.5 * (1. + (1.0 - newX2tinyCutoff) * (1.0 - 2.*tmp));
   jac *= (1.0 - newX2tinyCutoff);

   double z3 = rand3;

   //Now assign the values of xi from the zi. This is where we do the importance sampling that depnds on the radiation.
   xRad.x1() = z1;
   xRad.x2() = z2 * ySampling / (1. + z2 * (ySampling - 1.));
   xRad.x3() = z3;
   jac *= ySampling / pow((1. + z2 * (ySampling - 1.)), 2.);
   if (xRad.x1() < 0.)
      xRad.x1() = 0.;
   if (xRad.x1() > 1.)
      xRad.x1() = 1.;
   if (xRad.x2() < 0.)
      xRad.x2() = 0.;
   if (xRad.x2() > 1.)
      xRad.x2() = 1.;
   if (xRad.x3() < 0.)
      xRad.x3() = 0.;
   if (xRad.x3() > 1.)
      xRad.x3() = 1.;

   assert(xRad.x1() >= 0.0 && (xRad.x1() <= 1.0));
   assert((xRad.x2() >= 0.0) && (xRad.x2() <= 1.0));
   assert((xRad.x3() >= 0.0) && (xRad.x3() <= 1.0));
   xRad.weight() = jac;
   return;
}

} // namespace Geneva
