//------------------------------------------------------------------------------
/// \file Configuration.cpp
//
// Author(s):
//    Frank Tackmann, Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class Configuration
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/Configuration.hpp"
#include "Geneva/Core/Physics/Signature.hpp"

#include <iostream>
#include <iterator>
#include <algorithm>

namespace Geneva
{

//------------------------------------------------------------------------------
// Accessing and Querying Particles
//------------------------------------------------------------------------------
/**
 * \brief Tests if a Signature is contained in the %Configuration.
 * \param other The Signature to be searched for.
 * \return \c true if all IDs in \a other are contained in the %Configuration,
 *         \c false otherwise.
 *
 * If \c true is returned, \a other is a subsignature of the %Configuration. The
 * number of duplicate IDs is taken into account, that is, the %Configuration
 * must have at least as many Particles of the same ID as are in \a other for
 * \c true to be returned. Requires linear time.
 */
bool Configuration::includes(const Signature& other) const
{
   return std::includes(begin(), end(), other.begin(), other.end(), _compare_type());
}

//------------------------------------------------------------------------------
/**
 * \brief Tests if a Signature intersects with the %Configuration.
 * \param other The Signature to be tested.
 * \return \c true if at least one ID in \a other is also contained in the
 *         %Configuration, \c false otherwise.
 *
 * If \c true is returned, the %Configuration has a non-empty intersection with
 * \a other, i.e. at least one particle in the %Configuration corresponds to a
 * Signature entry in \a other, including initial- and final-state assignments.
 */
bool Configuration::intersects(const Signature& other) const
{
   iterator i1 = begin();
   const iterator end1 = end();
   Signature::iterator i2 = other.begin();
   const Signature::iterator end2 = other.end();

   // Since we only test for intersection we can stop as soon as we find one
   // common element.
   while (i1 != end1 && i2 != end2) {
      if (_compare_type()(*i1, *i2))
         ++i1;
      else if (_compare_type()(*i2, *i1))
         ++i2;
      else
         return true;
   }
   return false;
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------
/**
 * \brief Checks if the %Configuration is equal to a Signature.
 * \param rhs The Signature to be compared to
 * \return \c true if the %Configuration's Particles have exactly the same IDs
 *         as \a rhs, \c false otherwise.
 *
 * Initial- and final-state assignment as well as duplicate IDs are taken into
 * account. Requires linear time.
 */
bool Configuration::operator==(const Signature& rhs) const
{
   if (size() != rhs.size())
      return false;

   iterator i1 = begin();
   if (i1 == end())
      return true;
   Signature::iterator i2 = rhs.begin();

   if (i1->isInitialA()) {
      if ((!i2->isInitialA()) || i1->id() != *i2)
         return false;
      ++i1;
      if (i1 == end())
         return true;
      ++i2;
   }
   if (i1->isInitialB()) {
      if ((!i2->isInitialB()) || i1->id() != *i2)
         return false;
      ++i1;
      if (i1 == end())
         return true;
      ++i2;
   }
   if (!i2->isFinal())
      return false;
   while (true) {
      if (i1->id() != *i2)
         return false;
      ++i1;
      if (i1 == end())
         return true;
      ++i2;
   };
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Streams the Configuration to output.
std::ostream& Configuration::print(std::ostream& out = std::cout) const
{
   if (_initialA != end())
      out << "A: " << *_initialA << '\n';
   if (_initialB != end())
      out << "B: " << *_initialB << '\n';
   std::copy(beginFinal(), endFinal(), std::ostream_iterator<Particle>(out, "\n"));
   return out;
}

//------------------------------------------------------------------------------
// Initializes the internal initial-state and final-state iterators.
void Configuration::init()
{
   _beginFinal = begin();
   _sizeFinal = _particles.size();
   if (_sizeFinal && _beginFinal->isInitialA()) {
      _initialA = _beginFinal;
      ++_beginFinal;
      --_sizeFinal;
   } else
      _initialA = end();
   if (_sizeFinal && _beginFinal->isInitialB()) {
      _initialB = _beginFinal;
      ++_beginFinal;
      --_sizeFinal;
   } else
      _initialB = end();
}

//------------------------------------------------------------------------------
// Copies the content of another Configuration.
void Configuration::copy(const Configuration& other)
{
   if (other.initialA() != other.end())
      _initialA = _particles.insert(end(), *other.initialA());
   else
      _initialA = end();
   if (other.initialB() != other.end())
      _initialB = _particles.insert(end(), *other.initialB());
   else
      _initialB = end();
   if (other._sizeFinal != 0) {
      Configuration::iterator it = other.beginFinal();
      _beginFinal = _particles.insert(end(), *it);
      ++it;
      for (; it != other.endFinal(); ++it)
         _particles.insert(end(), *it);
   } else
      _beginFinal = end();
   _sizeFinal = other._sizeFinal;
}

} // namespace Geneva
