//------------------------------------------------------------------------------
/// \file ConfigurationUtils.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Utility functions for checking numerical accuracy of Configurations
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/ConfigurationUtils.hpp"
#include "Geneva/Core/Physics/RefineConfiguration.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <iomanip>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Checks four-momentum conservation for a Configuration.
 * \param config The Configuration to check.
 * \param where The place from where this was called.
 * \param precision Relative precision to use
 * \param accuracy Absolute accuracy to use
 * \return \c true if \a config satisfies energy-momentum conservation to the
 *         required precision and accuracy, \c false otherwise.
 *
 * Checks if the total four-momentum in the initial and final state sums to zero
 * to the required \a relative precision (with respect to the largest single
 * component in either initial or final state) and absolute \a accuracy. The
 * check is performed separately for each momentum component.
 *
 * \note If the configuration is empty, \c true is returned.
 */
bool checkMomentumConservation(const Configuration& config, double precision, double accuracy)
{
   if (config.size() == 0)
      return true;

   FourVector sum;
   FourVector max;
   // sum up the final-state momentum components excluding the largest ones, which are kept separate.
   for (Configuration::iterator it = config.begin(); it != config.end(); ++it) {
      for (size_t i = 0; i <= 3; ++i) {
         if (abs(it->p()[i]) > abs(max[i])) {
            sum[i] += max[i];
            if (it->isFinal())
               max[i] = it->p()[i];
            else
               max[i] = -it->p()[i];
         } else {
            if (it->isFinal())
               sum[i] += it->p()[i];
            else
               sum[i] -= it->p()[i];
         }
      }
   }

   bool success = true;
   double allowed = 0.0;
   double achieved = 0.0;
   for (size_t i = 0; i <= 3; ++i) {
      // compare each max component with the sum of the others
      std::pair<int, double> tmp = compareEps(sum[i], -max[i], precision, accuracy);
      if (tmp.first && tmp.second > allowed) {
         success = false;
         allowed = tmp.second;
         achieved = sum[i] + max[i];
      }
   }

   if (success) {
      // give more details to Log::debug
//       Log::debug("checkMomentumConservation") << "passed at relative precision " << precision << " and accuracy " <<  accuracy <<  Log::end;
      return true;
   }
   if (allowed > accuracy)
      Log::debug("checkMomentumConservation")
         << "failed at relative precision " << achieved/allowed*precision << " (allowed: " << precision << ')' << Log::end;
   else
      Log::debug("checkMomentumConservation")
         << "failed at absolute accuracy " << achieved << " (allowed: " << accuracy << ')' << Log::end;

   // give more details to Log::debug
   Log::debug()
      << "4-momentum sum = " << sum + max << " (allowed: " << allowed
      << ")\nFull configuration:\n"  << std::setprecision(16) << config << std::setprecision(6) << Log::end;
   return false;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Checks onshellness for a Configuration.
 * \param config The Configuration to check.
 * \param where The place from where this was called.
 * \param precision Relative precision to use
 * \param accuracy Absolute accuracy to use
 * \return \c true if each %Particle in \a config satisfies on-shell conditions to
 * the required precision and accuracy, \c false otherwise.
 *
 * \note If the configuration is empty, \c true is returned.
 */
bool checkOnShellConditions(const Configuration& config, double precision, double accuracy)
{
   if (config.size() == 0)
      return true;
   bool success = true;
   double allowed = 0.0;
   double achieved = 0.0;
   Configuration::iterator it;
   for (it = config.begin(); it != config.end(); ++it) {
      const double energy2 = it->energy()*it->energy();
      const double pvec2 = it->pvec().square();
      std::pair<int, double> tmp = compareEps(energy2, pvec2 + it->m2(), precision, accuracy);
      if (tmp.first) {
         success = false;
         allowed = tmp.second;
         achieved = energy2 - pvec2 - it->m2();
         break;
      }
   }
   if (success) {
      // give more details to Log::debug
//       Log::debug("checkOnShellConditions") << "passed at relative precision " << precision << " and accuracy " <<  accuracy <<  Log::end;
      return true;
   }
   if (allowed > accuracy)
      Log::debug("checkOnShellConditions") << "failed at relative precision " << achieved/allowed*precision << " (allowed: " << precision << ')' << Log::end;
   else
      Log::debug("checkOnShellConditions") << "failed at absolute accuracy " << achieved << " (allowed: " << accuracy << ')' << Log::end;

   Log::debug()
      << "p^2 - m^2 = " << std::setprecision(16) << achieved  << " (allowed: " << allowed
      << ")\nFor particle:\n" << *it << "\nFull configuration:\n" << config << std::setprecision(6) << Log::end;
   return false;
}

//------------------------------------------------------------------------------
/**
 * \ingroup Core
 *
 * \brief Checks if a Configuration is in its CM frame.
 * \param config The Configuration to check.
 * \param where The place from where this was called.
 * \param precision Relative precision to use
 * \param accuracy Absolute accuracy to use
 * \return \c true if \a config is in its center-of-momentum frame to the
 *         required precision and accuracy, \c false otherwise.
 *
 * Checks if the total three-momenta in the initial and final states sum to zero
 * to the required \a relative precision (with respect to the largest single
 * component) and absolute \a accuracy. The check is performed separately for
 * for the px, py, pz components.
 *
 * \note Assumes that the configuration has initialA, initialB, and at least
 * 1 final state.
 */
bool isInCoMFrame(const Configuration& config, double precision, double accuracy)
{
   Configuration::iterator it = config.beginFinal();
   ThreeVector finalSum;
   ThreeVector finalMax = it->pvec();
   ++it;
   // sum up the final-state momentum component while keeping the largest one separate.
   for (; it != config.endFinal(); ++it) {
      for (size_t i = 1; i <= 3; ++i)
         if (abs(it->pvec()[i]) > abs(finalMax[i])) {
            finalSum[i] += finalMax[i];
            finalMax[i] = it->pvec()[i];
         } else
            finalSum[i] += it->pvec()[i];
   }

   bool initialFailed = false;
   bool finalFailed = false;
   double initialAllowed = 0.0;
   double finalAllowed = 0.0;
   double initial = 0.0;
   double final = 0.0;
   for (size_t i = 1; i <= 3; ++i) {
      std::pair<int, double> tmp = compareEps(config.initialA()->pvec()[i], -config.initialB()->pvec()[i], precision, accuracy);
      if (tmp.first && tmp.second > initialAllowed) {
         initialFailed = true;
         initialAllowed = tmp.second;
         initial = config.initialA()->pvec()[i] + config.initialB()->pvec()[i];
      }
      tmp = compareEps(finalSum[i], -finalMax[i], precision, accuracy);
      if (tmp.first && tmp.second > finalAllowed) {
         finalFailed = true;
         finalAllowed = tmp.second;
         final = finalSum[i] + finalMax[i];
      }
   }

   if (!(initialFailed || finalFailed))
      return true;

   // figure out if initial or final was worse
   double allowed;
   double achieved;
   if (initialAllowed > finalAllowed) {
      allowed = initialAllowed;
      achieved = initial;
   } else {
      allowed = finalAllowed;
      achieved = final;
   }
   if (allowed > accuracy)
      Log::debug("isInCoMFrame")
         << "fails at relative precision " << achieved/allowed*precision << " (allowed: " << precision << ')' << Log::end;
   else
      Log::debug("isInCoMFrame")
         << "fails at absolute accuracy " << achieved << " (allowed: " << accuracy << ')' << Log::end;

   if (initialFailed)
      Log::debug() << "Initial 3-momentum sum = " << config.initialA()->pvec() + config.initialB()->pvec() << " (allowed: " << initialAllowed << ')' << Log::end;
   if (finalFailed)
      Log::debug() << "Final 3-momentum sum = " << finalSum + finalMax << " (allowed: " << finalAllowed << ')' << Log::end;
   Log::debug() << std::setprecision(16) << "Full configuration: sqrt(s) = " << (config.initialA()->p() + config.initialB()->p()).magnitude() << '\n'
      << config << std::setprecision(6) << Log::end;
   return false;
}

//------------------------------------------------------------------------------
/**
 * \brief Enforces a physical configuration.
 *
 * Returns \c true if both momentum conservation and on-shell conditions are
 * satisfied, \c false if they were violated and had to be refined.
 *
 * Throws an error if the refinement itself failed or didn't improve (the latter
 * only on DEBUG).
 */
bool enforceMomentumConservation(Configuration& config, std::string where,
                                 double precision, double accuracy, double buffer)
{
   if (checkMomentumConservation(config, precision, accuracy) && checkOnShellConditions(config, precision, accuracy))
      return true;

   if (!refineConfiguration(config))
      Log::error(where) << "Failed to refine Configuration." << Log::end;
#ifndef NDEBUG
   // throw an error if still not good after refining
   if (!checkMomentumConservation(config, buffer*precision, buffer*accuracy)
       || !checkOnShellConditions(config, buffer*precision, buffer*accuracy))
      Log::error(where) << "Configuration does not satisfy momentum after refinement" << Log::end;
#endif
   return false;
}

} // namespace Geneva
