//------------------------------------------------------------------------------
/// \file Emission.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class Emission
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/Emission.hpp"

#include <ostream>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief Determines the QCD mother id from \a emitted and \a sister.
 *
 * \param emitted PdgID of the emitted parton
 * \param sister PdgID of the sister parton
 * \param isFSR whether the splitting is FSR (\c true) or ISR (\c false)
 * \return The mother id
 *
 * This determines the mother id from \a emitted and \a sister based on QCD
 * splitting rules for FSR or ISR emissions (depending whether \a isFSR is
 * \c true or \c false). It is symmetric in emitted and sister (i.e. there is no
 * logical distinction between them).
 *
 * \note This asserts that emitted and sister correspond to a valid QCD
 * splitting, i.e. that \c ISQCDSplitting(emitted, sister) is \c true.
 */
PdgID Emission::getMotherQCD(PdgID emitted, PdgID sister, bool isFSR)
{
   // FSR: (q) -> qg, (g) -> gg
   // ISR: q -> (q)g, g -> (g)g
   if (emitted.isGluon()) {
      assert(sister.isGluon() || sister.isQuark());
      return sister;
   }

   if (sister.isGluon()) {
      assert(emitted.isQuark());
      if (isFSR)
         // FSR: (q) -> gq
         return emitted;
      else
         // ISR: g -> (qbar)q or g -> (q)qbar
         return emitted.antiPartner();
   }

   // (g) -> qqbar or q -> (g) q
   assert(emitted.isQuark());
   assert((isFSR && sister.isAntiPartner(emitted)) || (!isFSR && sister == emitted));
   return PdgID::kGluon;
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Streams a string in the format "ISR[A|B]: sister -> (mother) emitted" or "FSR: (mother) -> sister emitted" to \a out.
std::ostream& Emission::print(std::ostream& out) const
{
   if (_type == kFSR)
      out << "FSR: (" << _mother << ") -> " << _sister << " " << _emitted;
   else {
      if (_type == kISRA)
         out << "ISRA: ";
      else
         out << "ISRB: ";
      out << _sister << " -> (" << _mother <<") " << _emitted;
   }
   return out;
}

} // namespace Geneva
