//------------------------------------------------------------------------------
/// \file FourVector.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class FourVector
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/FourVector.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include <ostream>

namespace Geneva
{

//------------------------------------------------------------------------------
// Accessing and Querying Components
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Returns the component indexed by mu.
 * \param mu The index of the component 0 <= mu <= 3.
 * \return The value of the component.
 *
 * If mu is out of range a \c std::out_of_range exception is thrown.
 */
double FourVector::operator[](size_t mu) const
{
   if (mu == 0)
      return _p0;
   if (mu == 1)
      return _pvec.p1();
   if (mu == 2)
      return _pvec.p2();
   if (mu != 3)
      Log::error("FourVector::operator[]") << "mu index " << mu << " is out of range" << Log::end;
   return _pvec.p3();
}

//------------------------------------------------------------------------------
/**
 * \brief Gives access to the component indexed by mu.
 * \param mu The index of the component 0 <= mu <= 3.
 * \return A reference to the component.
 *
 * If mu is out of range an out_of_range error is thrown.
 */
double& FourVector::operator[](size_t mu)
{
   if (mu == 0)
      return _p0;
   if (mu == 1)
      return _pvec.p1();
   if (mu == 2)
      return _pvec.p2();
   if (mu != 3)
      Log::error("FourVector::operator[]") << "mu index " << mu << " is out of range" << Log::end;
   return _pvec.p3();
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %FourVector inequality comparison operator
 * \param rhs The %FourVector to compare to.
 * \return \c true if at least one of the components are different, \c false otherwise.
 *
 * The comparison is done with the default accuracy and precision separately for
 * each of the components.
 */
bool FourVector::operator!=(const FourVector& rhs) const
{
   return compare(_p0, rhs._p0) || (_pvec != rhs._pvec);
}

//------------------------------------------------------------------------------
/**
 * \brief Test if the %FourVector is nonzero.
 * \return \c true if at least one component is nonzero, \c false otherwise.
 *
 * Compares each components to 0.0 using the default absolute accuracy.
 *
 * \warning This function is here for completeness. Directly comparing to zero
 * is usually not very meaningfull for doubles since the overall scale is
 * unknown, so use of this function should be avoided. For example, one should
 * compare two %FourVectors with each other rather than their difference to zero.
 */
bool FourVector::isNonzero() const
{
   return compareToZero(_p0) || _pvec.isNonzero();
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Streams a string in the format "(p0, p1, p2, p3)" to out.
std::ostream& FourVector::print(std::ostream& out) const
{
   return out << "(" << _p0 << ", " << _pvec.p1() << ", " << _pvec.p2() << ", " << _pvec.p3() << ")";
}

} // namespace Geneva
