//------------------------------------------------------------------------------
/// \file NJettiness.cpp
//
// Author(s):
//    Frank Tackmann, Chris Vermilion, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 Frank Tackmann, LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of NJettiness classes
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/NJettiness_impl.hpp"
#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/Algorithms.hpp"

#include <list>
#include <vector>
#include <algorithm>

namespace Geneva
{

//------------------------------------------------------------------------------
// e+e- definitions
//------------------------------------------------------------------------------


namespace
{

// Unary function object returning the square of a Particle's Threevector.
struct ThreeVectorSquare
{
   double operator()(const Particle& p) const
   {
      return p.pvec().square();
   }
};

} // anonymous namespace

//------------------------------------------------------------------------------
/**
 * \brief Computes the 2-jettiness (thrust) axis (for 2 to 4 particles).
 * \param config The Configuration to use.
 * \return A ThreeVector with the resulting axis.
 *
 * Returns a normalized ThreeVector pointing along the 2-jettiness axis, which
 * is equivalent to the thrust axis. The orientation of the axis is arbitrary
 * and will be given by the first particle or pair of particles by which the
 * direction is determined.
 *
 * \note Currently this is only implemented for 2 to 4 particles in the final
 * state. Any other number will cause a runtime error.
 * \note This currently assumes that all the final state particles contribute
 *  to the definition of axis
 */
ThreeVector TwoJettiness::axis(const Configuration& config) const
{
   const size_t nFinal = config.sizeFinal();
   if (nFinal == 2) {
      // thrust axis is parallel to the final particles, choose first one
      return Normalize()(config.beginFinal()->pvec());
   }
   if (nFinal == 3) {
      // thrust axis is parallel to largest momentum particle
      Configuration::iterator it;
      double maxp2 = maximum(config.beginFinal(), config.endFinal(), it, ThreeVectorSquare());
      return it->pvec() / sqrt(maxp2);
   }
   if (nFinal == 4) {
      // thrust axis is parallel to largest momentum particle or pair of particles
      const Configuration::iterator first = config.beginFinal();
      Configuration::iterator it;
      double maxp2 = maximum(first, config.endFinal(), it, ThreeVectorSquare());
      ThreeVector result = it->pvec();
      it = first;
      while (++it != config.endFinal()) {
         ThreeVector tempAxis = first->pvec() + it->pvec();
         const double temp = tempAxis.square();
         if (temp > maxp2) {
            maxp2 = temp;
            result = tempAxis;
         }
      }
      return result / sqrt(maxp2);
   }
   Log::error("TwoJettiness::axis") << "Only implemented for 2 to 4 final particles, called for " << nFinal << " final particles." << Log::end;
   return ThreeVector();
}

//------------------------------------------------------------------------------
/**
 * \brief Computes the exact 2-jettiness (Tau2).
 * \param config The Configuration to use.
 * \return The result for Tau2.
 * \note This currently assumes that all the final state particles contribute
 *  to the definition of Tau2
 */
double TwoJettiness::operator()(const Configuration& config) const
{
   const size_t nFinal = config.sizeFinal();
   if (nFinal == 2) {
      // Tau2 = E1 - |pvec1| + E2 - |pvec2| = Etotal - 2|pvec|
      Configuration::iterator it = config.beginFinal();
      double Etotal = it->energy();
      ++it;
      Etotal += it->energy();
      double p = 2.*it->pvec().magnitude();
      // return exact 0.0 if E and p are equal within our default relative precision.
      if (compareRelative(Etotal, p))
         return Etotal - 2.*it->pvec().magnitude();
      else
         return 0.0;
   }
   if (nFinal == 3) {
      // thrust axis is parallel to largest momentum particle so
      // Tau2 = Etotal - 2|pvecmax|
      Configuration::iterator it = config.beginFinal();
      double maxp2 = it->pvec().square();
      double Etotal = it->energy();
      while (++it != config.endFinal()) {
         const double temp = it->pvec().square();
         if (temp > maxp2) {
            maxp2 = temp;
         }
         Etotal += it->energy();
      }
      return Etotal - 2.*sqrt(maxp2);
   }
   if (nFinal == 4) {
      // thrust axis is parallel to largest momentum particle or pair of particles so
      // Tau2 = Etotal - 2|pvecmax|
      const Configuration::iterator first = config.beginFinal();
      Configuration::iterator it = first;
      double maxp2 = it->pvec().square();
      double Etotal = it->energy();
      while (++it != config.endFinal()) {
         double temp = it->pvec().square();
         const double temp2 = (first->pvec() + it->pvec()).square();
         if (temp2 > temp) {
            temp = temp2;
         }
         if (temp > maxp2) {
            maxp2 = temp;
         }
         Etotal += it->energy();
      }
      return Etotal - 2.*sqrt(maxp2);
   }

   // right now getting here will fail since axis only works for 2 to 4 particles
   ThreeVector thrustAxis = axis(config);
   double Tau2 = 0.0;
   for (Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it) {
      Tau2 += it->energy() - abs(thrustAxis*it->pvec());
   }
   return Tau2;
}

//------------------------------------------------------------------------------
/**
 * \brief Computes the exact N-jettiness (TauN).
 * \param config The Configuration to use.
 * \return The result for TauN.
 *
 * For N = 2 the result is computed by calling \c TwoJettiness()(config).
 *
 * For <= N final state particles the result is the same and given by the sum of
 * \f$ E - |\vec{p}| \f$ of all particles (which will be zero for massless final
 * states).
 * \note Currently this is only implemented for up to N+1 final state particles.
 * Any other number will cause a runtime error.
 * \note This currently assumes that all the final state particles contribute
 *  to the definition of NJettiness
 */
double NJettiness::operator()(const Configuration& config, size_t N) const
{
   if (N == 2) {
      return TwoJettiness()(config);
   }

   const size_t nFinal = config.sizeFinal();
   if (nFinal <= N) {
      // TauN = sum_k E_k - |pvec_k|
      double TauN = 0.0;
      for (Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it) {
         const double E = it->energy();
         const double p = it->pvec().magnitude();
         // only add something nonzero to TauN if E and p are equal within our default relative precision.
         if (compareRelative(E, p)) {
            TauN += E - p;
         }
      }
      return TauN;
   }
   if (nFinal == N + 1) {
      double TauN = NJettiness()(config, N+1);
      TauN += pairwise_minimum(config.beginFinal(), config.endFinal(), NJettinessMetricFSR());
      return TauN;
   }
   Log::error("NJettiness") << "Only implemented for up to N + 1 final particles, called for N = " << N << " and N + " << nFinal - N << " final particles." << Log::end;
   return std::numeric_limits<double>::quiet_NaN();
}

//------------------------------------------------------------------------------
/**
 * \brief Computes the recursive N-jettiness (TauN).
 * \param config The Configuration to use.
 * \return The result for TauN.
 *
 * If the number of final states is <= N+1 the result is computed by calling
 * \c NJettiness()(config). For more final states the momenta are clustered
 * down by minimizing (numFinal-1)-jettiness in each clustering step.
 */
double NJettinessRecursive::operator()(const Configuration& config, size_t N) const
{
   size_t nFinal = config.sizeFinal();
   if (nFinal <= N + 1) {
      return NJettiness()(config, N);
   }

   // Cluster to N
   std::list<FourVector> momenta;
   for (Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it) {
         momenta.push_back(it->p());
   }
   std::list<FourVector>::iterator it1, it2;
   // cluster until nFinal == N
   while (nFinal > N) {
      pairwise_minimum(momenta.begin(), momenta.end(), it1, it2, NJettinessMetricFSR());
      (*it1) += (*it2);
      momenta.erase(it2);
      --nFinal;
   }

   double TauN = 0.0;
   for (it1 = momenta.begin(); it1 != momenta.end(); ++it1) {
      const double E = it1->p0();
      const double p = it1->pvec().magnitude();
      // only add something nonzero to TauN if E and p are equal within our default relative precision.
      if (compareRelative(E, p)) {
         TauN += E - p;
      }
   }
   return TauN;
}


//------------------------------------------------------------------------------
// pp definitions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
NJettinessppRecursive::NJettinessppRecursive(const Configuration& config)
   : _config(config), _y(0.), _numQCD(_config.sizeFinal()), _numNonQCD(0)
{
   FourVector nonQCD;
   for (Configuration::iterator it = _config.beginFinal(); it != _config.endFinal(); ++it) {
      if (!IsQCD()(*it)) {
         nonQCD += it->p();
         _numNonQCD++;
      }
   }
   if (_numNonQCD) {
      _y = nonQCD.rapidity();
      _numQCD -= _numNonQCD;
   }
}

//------------------------------------------------------------------------------
NJettinessppRecursive::NJettinessppRecursive(const Configuration& config, double y)
   : _config(config), _y(y), _numQCD(_config.sizeFinal()), _numNonQCD(0)
{
   for (Configuration::iterator it = _config.beginFinal(); it != _config.endFinal(); ++it) {
      if (!IsQCD()(*it))
         ++_numNonQCD;
   }
   _numQCD -= _numNonQCD;
}

//------------------------------------------------------------------------------
/**
 * \brief Computes recursive N-jettiness (TauN) for pp collisions.
 * \param config The Configuration to use.
 * \return The result for TauN.
 *
 * If the number of final states is <= N+1 the result is equivalent to the exact
 * TauN (i.e. with fully minimized axes). For more final states the momenta are
 * clustered down by minimizing (numFinal-1)-jettiness in each clustering step.
 */
//------------------------------------------------------------------------------
double NJettinessppRecursive::operator()(size_t N) const
{
   return NJettiness_impl::calculateTauNRecursive(N, _config.beginFinal(), _config.endFinal(), exp(_y), IsQCD());
}

//------------------------------------------------------------------------------
/**
 * \brief Performs a single N-jettiness clustering step.
 * \param[out] emitted Iterator pointing to the emitted particle.
 * \param[out] sister Iterator pointing to the sister particle.
 * \return The minimal value of the clustering metric found.
 *
 * This minimizes the NJettiness clustering metric over all pairs of particles
 * (final-state clustering) and single particles (initial-state clustering). For
 * final-state clustering the assignment of which particle is \a emitted and
 * which is \a sister is in principle arbitrary. (At the moment it uses our
 * default conventions.)
 * For initial-state clustering the sister will point to \c initialA() or
 * \c initialB() depending on which beam the clustering is with.
 *
 * \warning: The returned minimum metric is equal to (N-1)-jettiness if and only
 * if all particles are strictly massless.
 */
//------------------------------------------------------------------------------
double NJettinessppRecursive::cluster(Configuration::iterator& emitted, Configuration::iterator& sister) const
{
   emitted = _config.end();
   sister = _config.end();
   const double min = NJettiness_impl::cluster(_config.initialA(), _config.initialB(),
                                               _config.beginFinal(), _config.endFinal(),
                                               emitted, sister, exp(_y), _numQCD,
                                               IsQCD(), TrueCondition());

   // TODO: This logic should go somewhere else, since the energy condition is frame-dependent.
   // Implements the current default logic for emitted/sister assignment for FSR clusterings.
   // (which corresponds to the way the Thetaij and the RegionInfo mapping have been set up):
   // For qg, the gluon is taken as the emitted.
   // For qqbar, the quark is taken as the emitted.
   // For gg, the one with the lower energy is taken as the emitted.
   if (sister != _config.end() && sister->isFinal()) {
      if ( (sister->id().isGluon() && (!(emitted->id().isGluon()) || sister->energy() < emitted->energy()))
           || emitted->id().isAntiParticle() )
         std::swap(emitted, sister);
   }
   return min;
}

//------------------------------------------------------------------------------
/**
 * \brief Performs a single N-jettiness clustering step only taking into account QCD-allowed splittings.
 * \param[out] emitted Iterator pointing to the emitted particle.
 * \param[out] sister Iterator pointing to the sister particle.
 * \return The minimal value of the clustering metric found.
 *
 * Analogous to \c cluster(), but only considers emitted/sisters pairs that
 * correspond to valid QCD splittings, i.e., for which
 * \c IsQCDSplitting(allowqqbar)(*emitted, *sister) is \c true.
 *
 * \warning: The returned minimum metric is equal to (N-1)-jettiness if and only
 * if all particles are strictly massless.
 */
//------------------------------------------------------------------------------
double NJettinessppRecursive::clusterQCDAllowed(Configuration::iterator& emitted, Configuration::iterator& sister, bool allowqqbar) const
{
   emitted = _config.end();
   sister = _config.end();
   const double min = NJettiness_impl::cluster(_config.initialA(), _config.initialB(),
                                               _config.beginFinal(), _config.endFinal(),
                                               emitted, sister, exp(_y), _numQCD,
                                               IsQCD(), IsQCDSplitting(allowqqbar));

   // same logic as for cluster
   if (sister != _config.end() && sister->isFinal()) {
      if ( (sister->id().isGluon() && (!(emitted->id().isGluon()) || sister->energy() < emitted->energy()))
         || emitted->id().isAntiParticle() )
         std::swap(emitted, sister);
   }
   return min;
}

//------------------------------------------------------------------------------
/**
 * \brief Performs a single N-jettiness clustering step in the soft limit.
 * \param[out] sister Iterator pointing to the sister particle.
 * \return The minimal value of the clustering metric found.
 */
//------------------------------------------------------------------------------
double NJettinessppRecursive::clusterSoft(Configuration::iterator& emitter, FourVector pSoft) const
{
   BoostZHyp boost(exp(-_y));
   pSoft.apply(boost);
   double min = NJettinessMetricISR()(pSoft);
   if (pSoft.p3() > 0.0)
      emitter = _config.initialA();
   else
      emitter = _config.initialB();

   for (Configuration::iterator it = _config.beginFinal(); it != _config.endFinal(); ++it) {
      if (IsQCD()(*it)) {
         const double temp = NJettinessMetricFSRSoft()(pSoft, boost(it->p()));
         if (temp < min) {
            emitter = it;
            min = temp;
         }
      }
   }
   return min;
}

//------------------------------------------------------------------------------
/**
 * \brief Computes the exact 0-jettiness (TauN) for pp collisions.
 * \param config The Configuration to use.
 * \return The result for TauN.
 *
 * If the number of final states is <= N+1 the result is equivalent to the exact
 * TauN (i.e. with fully minimized axes). For more final states the momenta are
 * clustered down by minimizing (numFinal-1)-jettiness in each clustering step.
 */
//------------------------------------------------------------------------------
double ZeroJettinesspp::operator()() const
{
   return NJettiness_impl::calculateTau0(_config.beginFinal(), _config.endFinal(), exp(_y), IsQCD());
}

} // namespace Geneva
