//------------------------------------------------------------------------------
/// \file ParticleDataTable.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 MIT
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class ParticleDataTable
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/ParticleDataTable.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
// Accessing and querying the data
//------------------------------------------------------------------------------
/**
 * \brief Locates the entry for a given \c PdgID.
 * \param id The \c PdgID to be looked for.
 * \return A const_ptr to the date entry corresponding to \p id.
 *
 * If no entry for \p id is found, a zero pointer is returned.
 */
ParticleDataTable::const_ptr ParticleDataTable::getData(PdgID id) const
{
   _entry.id() = id;
   iterator it = _table.find(_entry);
   if (it != _table.end())
      return const_ptr(&(*it));
   else
      return const_ptr();
}

//------------------------------------------------------------------------------
/**
 * \brief Locates the nth entry with a given particle name.
 * \param name The particle name to be looked for.
 * \return A \c const_ptr to the nth entry with \p name.
 *
 * If \p n is not given, by default the first entry (n = 1) is returned. If less
 * than \p n entries with \p name exist, a zero pointer is returned. To find all
 * entries with the same name use \code find(name) \endcode.
 */
ParticleDataTable::const_ptr ParticleDataTable::getData(const std::string& name, size_t n) const
{
   std::pair<lookup_iterator, lookup_iterator> range = _lookup.equal_range(name);
   // if not found or n doesn't make sense return zero pointer.
   if (n < 1)
      return const_ptr();
   --n;
   while (range.first != range.second && n != 0) {
      --n;
      ++range.first;
   }
   if (range.first != range.second)
      return const_ptr(&(*(range.first->second)));
   else
      return const_ptr();
}

//------------------------------------------------------------------------------
// Adding and deleting data
//------------------------------------------------------------------------------
/**
 * \brief Inserts a single \c ParticleData entry into the table.
 * \param data The entry to be inserted.
 * \param overwrite Flag to specify if existing entries should be overwritten.
 * \return True if the entry was inserted, false otherwise.
 *
 * If no entry with the same PdgID as in \p data exists, the \p data is inserted
 * as a new entry and true is returned. Otherwise, the behavior depends on the
 * value of \p overwrite. If \p overwrite is true the data of the existing entry
 * is overwritten with the new values in \p data and true is returned. Existing
 * iterators/pointers to this entry remain valid. If \p overwrite is false, the
 * existing entry is kept unchanged and false is returned.
 */
bool ParticleDataTable::insert(const ParticleData& data, bool overwrite)
{
   iterator it = _table.lower_bound(data);
   // not found, so just insert
   if (it == _table.end() || it->id() != data.id()) {
      it = _table.insert(it, data);
      insertLookup(it);
      return true;
   }
   // found, overwrite existing entry if overwrite is true
   if (overwrite) {
      eraseLookup(it);
      // By default sets don't allow modifying keys. In our case it's okay since
      // we know that the PdgIDs are the same, so we have to do some
      // const casting here.
      *const_cast<ParticleData*>(&(*it)) = data;
      insertLookup(it);
      return true;
   }
   return false;
}

//------------------------------------------------------------------------------
/**
 * \brief Inserts a range of \c ParticleData entries into the table.
 * \param first Iterator pointing to the first entry to be inserted.
 * \param last, Iterator pointing beyond the last entry to be inserted.
 * \param overwrite Flag to specify if existing entries should be overwritten.
 * \return The number of entries that were inserted.
 *
 * The function reads \c ParticleData objects from the range [\p first, \p last)
 * and inserts them into the table using the same logic as
 * \code insert(data, overwrite) \endcode. The return value specifies how often
 * an entry was actually inserted/overwritten.
 *
 * This is a template function that works for any type of iterator from which a
 * ParticleData object can be read.
 */
template <typename InputIterator>
size_t ParticleDataTable::insert(InputIterator first, InputIterator last, bool overwrite)
{
   iterator it = _table.lower_bound(*first);
   // not found, so just insert
   if (it == _table.end() || it->id() != first->id()) {
      it = _table.insert(it, *first);
      insertLookup(it);
      return true;
   }
   // found, overwrite existing entry if overwrite is true
   if (overwrite) {
      eraseLookup(it);
      // By default sets don't allow modifying keys. In our case it's okay since
      // we know that the PdgIDs are the same, so we have to do some
      // const casting here.
      *const_cast<ParticleData*>(&(*it)) = *first;
      insertLookup(it);
      return true;
   }
   return false;
}

//------------------------------------------------------------------------------
/**
 * \brief Removes the entry for a \c PdgID from the table.
 * \param id The \c PdgID of the entry to be erased.
 * \return True if an entry was erased, false otherwise.
 */
bool ParticleDataTable::erase(PdgID id)
{
   _entry.id() = id;
   iterator it = _table.find(_entry);
   if (it != _table.end()) {
      erase(it);
      return true;
   }
   return false;
}

//------------------------------------------------------------------------------
/**
 * \brief Removes all entries with a given particle name from the table.
 * \param name The particle name of the entries to be erased.
 * \return The number of entries that were erased.
 */
size_t ParticleDataTable::erase(const std::string& name)
{
   std::pair< std::multimap<std::string, iterator>::iterator,
       std::multimap<std::string, iterator>::iterator > range;
   range = _lookup.equal_range(name);
   size_t num = 0;
   for (; range.first != range.second; ++range.first) {
      _table.erase(range.first->second);
      _lookup.erase(range.first);
      ++num;
   }
   return num;
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
ParticleData ParticleDataTable::_entry;

//------------------------------------------------------------------------------
// Inserts an entry into the lookup table for the table entry pointed to by it.
void ParticleDataTable::insertLookup(iterator it)
{
   std::multimap<std::string, iterator>::value_type value(it->name(), it);
   _lookup.insert(value);
}

//------------------------------------------------------------------------------
// Delets the entry in the lookup table for the table entry pointed to by it.
// Assumes that *it has an entry in the lookup table.
void ParticleDataTable::eraseLookup(iterator it)
{
   std::multimap<std::string, iterator>::iterator li;
   li = _lookup.lower_bound(it->name());
   while (li->second->id() != it->id())
      ++li;
   _lookup.erase(li);
}

} // namespace Geneva
