//------------------------------------------------------------------------------
/// \file PdgID.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012  MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class PdgID
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/PdgID.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include <ostream>

namespace Geneva
{

//------------------------------------------------------------------------------
// Constructors etc.
//------------------------------------------------------------------------------
/**
 * \brief Constructs %PdgID from an explicit integer code.
 * \param code The integer code to be used as PDG ID.
 *
 * This constructor can only be used explicitly, so integers are not
 * automatically converted to a %PdgID to enforce type safety. For example, in
 * \code
 *    PdgID id;
 *    id = kElectron;   // good
 *    id = PdgID(11);   // bad but compiles
 *    id = 11;          // compile error
 * \endcode
 * the last line produces a compile error.
 *
 * If \a code is not valid a warning is issued and the %PdgID initialized to
 * \c PdgID::kUnknown.
 */
PdgID::PdgID(int code)
   : _code(kUnknown)
{
   if (!setCode(code))
      Log::warning("PdgID") << "Unrecognized PDG code " << code << ", using PdgID::kUnknown instead." << Log::end;
}

//------------------------------------------------------------------------------
// Physics Tests
//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is a SM hadron.
 * \return \c true for any SM meson or baryon, \c false otherwise.
 *
 * This only includes SM hadrons (including pomeron and odderon) but no
 * diquarks. The meson and baryon quarks are checked to be SM quarks (up...top).
 * (Baryons can include pentaquarks, but the quark content of the 4th and 5th
 * quark is not checked as their codes could in principle overlap with excited
 * baryons.)
 */
bool PdgID::isSMHadron() const
{
   int tmp = abs(_code);
   if ((tmp > 1000000 && tmp < 9000000) || tmp > 9900000)
      return false;
   // veto fundamental and diquark
   tmp %= 10000;
   if (tmp == 990 || tmp == 9990) // check glue balls
      return true;
   if (tmp >= 7000)               // q1 = 0-6
      return false;
   tmp %= 1000;
   if (tmp < 100 || tmp >= 700)   // q2 = 1-6
      return false;
   tmp %= 100;
   return tmp >= 10 && tmp < 70;  // q3 = 1-6
}

//------------------------------------------------------------------------------
/**
 * \brief Test if %PdgID is a particle that is its own antiparticle.
 * \return \c true for any particle that is known to be its own antiparticle,
 *         \c false otherwise.
 *
 * A particle is its own antiparticle if it is described by a real field (i.e.
 * neutral bosons and majorana fermions). This is only tested for quarks,
 * leptons, gauge and Higgs bosons and their new-physics specific partners,
 * as well as SM hadrons. All others, including MC internal ones, are assumed to
 * have a valid distinct antiparticle.
 */
bool PdgID::isOwnAntiParticle() const
{
   int tmp2 = abs(_code);
   // first check fundamental-like
   int tmp = tmp2 % 100000;
   if (tmp <= 39)
      return _spins[tmp] > 0;
   // next check for SM
   tmp2 /= 100000;
   if (tmp2 >= 10 && tmp2 != 90)
      return false;
   tmp %= 10000;
   if (tmp == 9990)  // odderon
      return true;
   if (tmp > 990)    // not a meson
      return false;
   return (tmp / 10) % 11 == 0;  // check for nq2 == nq3
}

//------------------------------------------------------------------------------
// Additional Physics Queries
//------------------------------------------------------------------------------
/**
 * \brief Checks if an integer represents a valid PDG code.
 * \param code The integer code to be checked.
 * \return \c true if \a code is valid, \c false otherwise.
 *
 * All codes are checked for a valid #NPType.
 * The validity of codes in the fundamental range is checked explicitly,
 * including new-physics specific partners. The validity of new-physics specific
 * codes is not further checked (since they could in principle take any form.)
 * For SM hadrons a minimal check is performed for mesons and baryons to have
 * two and three entries in the right place and no invalid antiparticles. (The
 * ordering of the quarks is not checked since there are exceptions and it is
 * not critical.
 */
bool PdgID::isValid(int code) const
{
   if (!code)                // 0 is valid
      return true;
   int tmp = abs(code);
   int tmp2 = tmp / 100000;
   // check and separate NPtype. Here we have to be more careful and filter
   // codes that we can otherwise gloss over and assume to not exist
   if (tmp2 <= 9)       // nr <= 9 is SM
      tmp2 = 0;
   else if (tmp2 <= 40)      // nnr 10-40 is SUSY, Technicolor, excited
      tmp %= 100000;
   else if (tmp2 == 41)      // dyons are weird ...
      return (tmp > 4110000);
   else if (tmp2 <= 49)      // 42-49 are not allowed
      return false;
   else if (tmp2 <= 69)      // 50-69 are KK
      tmp %= 100000;
   else if (tmp2 <= 89)      // 70-89 are not allowed
      return false;
   else if (tmp2 <= 98) {
      tmp2 = 0;              // n = 9 are unassigned mesons or pentaquarks
   } else
      return (tmp2 <= 100);  // user-defined and QBalls are not further checked
   // next check fundamental
   if (tmp <= 39)
      return _spins[tmp] != 0 && (code >= 0 || _spins[tmp] < 0);
   if (tmp <= 102)           // valid exotic or MC
      return true;
   if (tmp2)                 // don't check NP any further
      return true;
   tmp %= 10000;
   int q2 = (tmp % 1000) / 100;
   if (tmp >= 1000)
      return q2;             // nonzero q1 requires nonzero q2
   // zero 1 defines a meson for which we need
   // q2 and q3 nonzero (and unequal for code < 0)
   int q3 = (tmp % 100) / 10;
   return (q2 && q3) && (code > 0 || q2 != q3);
}

//------------------------------------------------------------------------------
/**
 * \brief Returns the AntiPartner of the %PdgID.
 * \return The %PdgID of the particle described by the conjugated field.
 *
 * Here, the AntiPartner is defined by field conjugation. That is, for charged
 * particles (e+, W+, pi+) the %PdgID for the particle with the opposite charge
 * (e-, W-, pi-) is returned (i.e. negating the PDG code). For the left-handed
 * neutrinos the corresponding right-handed antineutrinos are returned. For
 * neutral particles with real fields (gamma, pi0, etc.), their own %PdgID is
 * returned.
 */
PdgID PdgID::antiPartner() const
{
   PdgID result(kUnknown);
   if (isOwnAntiParticle())
      result._code = _code;
   else
      result._code = -_code;
   return result;
}

//------------------------------------------------------------------------------
/**
 * \brief Extracts the specific new-physics type of the %PdgID.
 * \return An integer representing the #NPType
 *
 * The returned integer can be compared with the named constants defined in
 * #NPType.
 */
int PdgID::getNPType() const
{
   int tmp = abs(_code) / 100000;
   if (tmp >= 90 && tmp < 99)
      return 0;
   if (tmp <= 39 || tmp >= 50)
      return tmp / 10;
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief Extracts the fundamental portion of the %PdgID.
 * \return If successful, the resulting fundamental %PdgID, otherwise
 *         \c PdgID::kUnknown.
 *
 * If the %PdgID corresponds to a new-physics specific partner of a fundamental
 * particle, the corresponding fundamental particle (with #NPType \c kSMlike) is
 * returned. If the %PdgID already belongs to the fundamental range it is simply
 * returned. In all other cases \c PdgID::kUnknown is returned.
 */
PdgID PdgID::getFundamental() const
{
   PdgID result(kUnknown);
   int tmp = _code % 100000;
   if (abs(tmp) <= 102)
      result._code = tmp;
   return result;
}

//------------------------------------------------------------------------------
/**
 * \brief Extracts the total spin encoded by the %PdgID.
 * \return 2s+1 where s is the total spin, or 0 if the spin is not known.
 *
 * Elementary new-physics specific particles are assumed to have the same spin
 * as their SM-like partners, except for SUSY, where the spin of elementary
 * particles is adjusted appropriately. If the spin is not known 0 is returned.
 */
int PdgID::getSpin() const
{
   int tmp2 = abs(_code);
   int tmp = tmp2 % 100000;
   if (tmp <= 39) {
      int s = abs(_spins[tmp]);
      if (tmp2 < 1000000 || tmp2 >= 3000000 || s == 0)  // no SUSY or unknown spin
         return s;
      else {
         if (s == 1)    // Higgs get spin + 1/2
            return 2;
         return s - 1;  // everybody else gets spin -1/2
      };
   };
   if (tmp <= 102)      // exotics and special MC have unknown spin
      return 0;
   return tmp % 10;
}


//------------------------------------------------------------------------------
// Private member functions and static data
//------------------------------------------------------------------------------

// lookup table for codes <=39
// magnitude is the spin 2s+1, 0 denotes an invalid code, and negative means
// has an antiparticle
int PdgID::_spins[40] = {0,
                         -2, -2, -2, -2, -2, -2, -2, -2, 0, 0,
                         -2, -2, -2, -2, -2, -2, -2, -2, 0, 0,
                         3,  3,  3, -3,  1,  0,  0,  0, 0, 0,
                         0,  3,  3, -3,  1,  1, -1,  0, 5
                        };

//------------------------------------------------------------------------------
// Streams the integer code of the PdgID to out.
std::ostream& PdgID::print(std::ostream& out) const
{
   return out << _code;
}

} // namespace Geneva
