//------------------------------------------------------------------------------
/// \file PhiRadFKS.cpp
//
// Author(s):
//    Frank Tackmann, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class PhiRadFKS
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/PhiRadFKS.hpp"
#include "Geneva/Core/Physics/UnitCubeRad.hpp"
#include "Geneva/Core/Physics/FourVector.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <cassert>
#include <ostream>

namespace Geneva
{

//------------------------------------------------------------------------------
PhiRadFKS::PhiRadFKS(const UnitCubeRad& xRad, double xiMax, double& jac)
{
   _idMother = xRad.idMother();
   _idRight = xRad.idEmitted();
   _idLeft = xRad.idSister();
   if (xRad.radType() == UnitCubeRad::kFSR)
      _radType = PhiRadFKS::kFSR;
   else if (xRad.radType() == UnitCubeRad::kISRA)
      _radType = PhiRadFKS::kISRA;
   else if (xRad.radType() == UnitCubeRad::kISRB)
      _radType = PhiRadFKS::kISRB;
   else
      Log::fatal("PhiRadFKS")<<"Radiation region unknown."<<Log::end;
   _xi = xRad.x1() * xiMax;
   assert(_xi <= xiMax);
   _y = -1. + 2.*xRad.x2();
   if (xRad.radType() == UnitCubeRad::kISRB)
      _y = -_y;
   _phi = math_constants::_2pi() * xRad.x3();
   jac = xiMax * math_constants::_4pi();
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Streams info about PhiRadFKS to out.
std::ostream& PhiRadFKS::print(std::ostream& out) const
{
   if (_radType == kFSR) {
      out << "PhiRadFKS FSR: (" << _idMother << ") -> " << _idLeft << " " << _idRight << "_FKS\n";
   } else if (_radType == kISRA) {
      out << "PhiRadFKS ISRA: " << _idLeft << " -> (" << _idMother << ") " << _idRight << "_FKS\n";
   } else if (_radType == kISRB) {
      out << "PhiRadFKS ISRB: " << _idLeft << " -> (" << _idMother << ") " << _idRight << "_FKS\n";
   } else {
      Log::fatal("PhiRadFKS::print") << "Radiation region unknown." << Log::end;
   }
   out << "xi: " << _xi << " y: " << _y << " phi: " << _phi << '\n';
   return out;
}

} // namespace Geneva
