//------------------------------------------------------------------------------
/// \file Rate.cpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 CERN, DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class Rate
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/Rate.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <iomanip>
#include <algorithm>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief Calculates the statistical uncertainty.
 * \return A Sigma representing the uncertainty of central value and all variations
 *
 * Calculates the statistical uncertainty from the sum of squared weights and
 * total entries for the central value and all variations.
 */
Sigma Rate::error() const
{
   const double n = static_cast<double>(_totEntries);
   Sigma error(sqrt(abs(_sumWeightsSq.central()/n - pow(_sumWeights.central()/n, 2)) / (n - 1.0)));
   error.variations().reserve(_sumWeights.variations().size());
   Sigma::const_iterator it2 = _sumWeightsSq.variations().begin();
   for (Sigma::const_iterator it = _sumWeights.variations().begin(); it != _sumWeights.variations().end(); ++it, ++it2) {
      error.variations().push_back(sqrt(abs((*it2)/n - pow(*it/n, 2)) / (n - 1.0)));
   }

   return error;
}

//------------------------------------------------------------------------------
/**
 * \brief Adds a weight to the %Rate.
 * \param weight The Sigma to be added.
 *
 * If \a weight is a defined Sigma it is added to the %Rate and the counters for
 * entries and total entries are increased.
 *
 * If \a weight is undefined, only the total entries are increased, equivalent
 * to adding a zero weight.
 */
void Rate::addWeight(const Sigma& weight)
{
   // only fill if the weight is defined
   if (weight.isDefined()) {
      // initialize variations from the first defined weight we encounter
      if (!_variationsResized) {
         assert(_sumWeights.variations().empty());
         assert(_sumWeightsSq.variations().empty());
         size_t n = weight.variations().size();
         _sumWeights.variations().resize(n);
         _sumWeightsSq.variations().resize(n);
         _variationsResized = true;
      }
      _sumWeights += weight;
      _sumWeightsSq += weight*weight;
      _entries++;
   }
   // always increase the total counter
   _totEntries++;
}

//------------------------------------------------------------------------------
// Resets the Rate to zero.
void Rate::reset()
{
   _sumWeights = Sigma(0.0);
   _sumWeightsSq = Sigma(0.0);
   _entries = 0;
   _totEntries = 0;
   _variationsResized = false;
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %Rate addition assignment operator
 * \param rhs The %Rate to be added.
 * \return A reference to the %Rate.
 *
 * \a rhs is added to the %Rate.
 */
Rate& Rate::operator+=(const Rate& rhs)
{
   _sumWeights += rhs._sumWeights;
   _sumWeightsSq += rhs._sumWeightsSq;
   _entries += rhs._entries;
   _totEntries += rhs._totEntries;
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %Rate subtraction assignment operator
 * \param rhs The %Rate to be subtracted.
 * \return A reference to the %Rate.
 *
 * \a rhs is subtracted from the %Rate.
 */
Rate& Rate::operator-=(const Rate& rhs)
{
   _sumWeights -= rhs._sumWeights;
   _sumWeightsSq += rhs._sumWeightsSq;
   _entries += rhs._entries;
   _totEntries += rhs._totEntries;
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %Rate multiplication with real number assignment operator
 * \param rhs The number to be multiplied by.
 * \return A reference to the %Rate.
 *
 * Each component of the %Rate is multiplied by \a rhs.
 */
Rate& Rate::operator*=(double rhs)
{
   _sumWeights *= rhs;
   _sumWeightsSq *= rhs * rhs;
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %Rate division by real number assignment operator
 * \param rhs The number to be divided by.
 * \return A reference to the %Rate.
 *
 * Each component of the %Rate is divided by \a rhs.
 * \note Division by zero is not checked.
 */
Rate& Rate::operator/=(double rhs)
{
   return operator*=(1. / rhs);
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Streams the entries and central value +/- error to out.
std::ostream& Rate::print(std::ostream& out) const
{
   out.precision(10);
   out << std::setw(10) << entries() << "  ";
   out << std::scientific << value().central() << " +/- " << error().central() << '\n';
   return out;
}

} // namespace Geneva
