//------------------------------------------------------------------------------
/// \file RefineConfiguration.cpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2017 CERN
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implements a utility function to refine a Configuration
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/RefineConfiguration.hpp"
#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include <vector>
#include <cassert>

namespace Geneva
{

namespace
{

#ifdef geneva_USE_QUADMATH
#include <quadmath.h>
//------------------------------------------------------------------------------
// Overload stream operator for __float128 quad type
// NOTE: only possible if linking against libquadmath
std::ostream& operator<<(std::ostream& out, const __float128& value)
{
   char* str = new char[50];
   quadmath_snprintf(str, 50, "%Qf", value);
   out<<str;
   return out;
}
#define FABS fabsq
#define SQRT sqrtq
#else
#define FABS fabs
#define SQRT sqrt
#endif // geneva_USE_QUADMATH

//------------------------------------------------------------------------------
// Ancillary function
template <typename T, typename U>
T truncate(const T val, const T acc, const std::vector<U>& scales)
{
   T ans = val;
   bool known = false;
   for (size_t i=0; i < scales.size(); i++) {
      if (abs(val - T(scales[i])) < acc) {
         ans = T(scales[i]);
         known = true;
         break;
      }
   }
   if (!known) {
      Log::warning("truncate") << "Unknown scale " << val << Log::end;
   }
   return ans;
}

//------------------------------------------------------------------------------
// NOTE: Adaptated from original version in NJet
// This is doing the actual work.
// Refine momenta by forcing on-shell conditions that change the energy of the
// first N-2 momenta and by modifying the last two momenta to fulfill momentum
// conservation. In particular, all the components of the last momemtum and
// only the 0 and 3 components of the second-but-last are modified for
// enforcing momentun conservation.
template <typename SRC, typename DST, typename SCT>
bool refineMomenta(const std::vector<SRC>& pin, std::vector<DST>& pout, const std::vector<SCT>& scales2, SCT scales2_thr, size_t max_iter)
{
   /*
    * F = {
    *  s0 + q0 + p0 == 0,
    *  s3 + q3 + p3 == 0,
    *  q0^2 - q12sq - q3^2 - qM^2,
    *  p0^2 - p12sq - p3^2 - pM^2}
    *  X = {q0, q3, p0, p3}
    * J = {{1, 0, 1, 0},
    *     {0, 1, 0, 1},
    *     {2 q0, -2 q3, 0, 0},
    *     {0, 0, 2 p0, -2 p3}}
    * J(X_n) (X_n+1 - X_n) = -F(X_n)
    */

   // Find the two particles to be modified last
   size_t NN = pin.size() / 4;
   size_t np = NN-2;
   size_t nq = NN-1;

   SRC x1x2maxp = 0;
   SRC x1x2maxq = 0;

   DST scale = 0.;
   for (size_t n=0; n<NN; n++) {
      const DST x0abs = FABS(pin[0+4*n]);
      if (x0abs > scale) {
         scale = x0abs;
      }
      const SRC x1x2abs = FABS(pin[1+4*n]*pin[2+4*n]);
      if (x1x2abs > x1x2maxq) {
         if (nq != n) {
            x1x2maxp = x1x2maxq;
            np = nq;
         }
         if (np != n) {
            x1x2maxq = x1x2abs;
            nq = n;
         }
      } else if (x1x2abs > x1x2maxp) {
         if (nq != n) {
            x1x2maxp = x1x2abs;
            np = n;
         }
      }
   }
   assert(nq != np);

   // Fix all the other particles
   const DST M2acc = (scale*scale)*scales2_thr;

   DST s0 = DST();
   DST s1 = DST();
   DST s2 = DST();
   DST s3 = DST();

   for (size_t n=0; n<NN; n++) {
      if ((n == np) or (n == nq)) continue;
      const DST nP2 = DST(pin[1+4*n])*pin[1+4*n]
      + DST(pin[2+4*n])*pin[2+4*n]
      + DST(pin[3+4*n])*pin[3+4*n];
      DST nM2 = DST(pin[0+4*n])*pin[0+4*n] - nP2;

      // if we are close enough to a known scale, assume it is equal to it
      nM2 = truncate(nM2, M2acc, scales2);

      // fix on-shellness of particles other than np or nq
      DST energy = SQRT(nP2 + nM2);
      if (pin[0+4*n] < 0.) {
         energy = -energy;
      }
      pout[0+4*n] = energy;
      pout[1+4*n] = pin[1+4*n];
      pout[2+4*n] = pin[2+4*n];
      pout[3+4*n] = pin[3+4*n];

      DST sign = 1.;
      if (n<2)
         sign=-1.;

      s0 += sign*pout[0+4*n];
      s1 += sign*pout[1+4*n];
      s2 += sign*pout[2+4*n];
      s3 += sign*pout[3+4*n];
   }

   //Now do the last 2 particles
   const DST p1 = pout[1+4*np] = pin[1+4*np];
   const DST p2 = pout[2+4*np] = pin[2+4*np];
   const DST q1 = pout[1+4*nq] = -(pin[1+4*np] + s1);
   const DST q2 = pout[2+4*nq] = -(pin[2+4*np] + s2);

   DST p0 = pin[0+4*np];
   DST p3 = pin[3+4*np];
   DST q0 = pin[0+4*nq];
   DST q3 = pin[3+4*nq];

   const DST p12sq = p1*p1 + p2*p2;
   const DST q12sq = q1*q1 + q2*q2;

   DST pM2 = p0*p0 - (p12sq + p3*p3);
   DST qM2 = q0*q0 - (q12sq + q3*q3);

   pM2 = truncate(pM2, M2acc, scales2);
   qM2 = truncate(qM2, M2acc, scales2);

   DST dist = 1e30;
   DST prevdist = 1e31;
   size_t count = 0;

   DST newq0 = q0;
   DST newq3 = q3;
   DST newp0 = p0;
   DST newp3 = p3;

   do {
      q0 = newq0;
      q3 = newq3;
      p0 = newp0;
      p3 = newp3;

      const DST detJ = 4.*(q0*p3 - q3*p0);
      if (detJ == 0.) {
         Log::warning("refineMomenta") << "det(J) == 0., expect to get nan ..." << Log::end;
      }

      const DST ex1 = 2.*p12sq - 2.*p3*p3 + 2.*pM2 + 4.*p0*s0 - 4.*p3*s3;
      const DST exq12 = 2.*q12sq + 2.*qM2;
      const DST exr12 = 2.*p12sq + 2.*pM2;
      const DST dq0 = (p3*(-2.*q0*q0 + exq12) + q3*(2.*p0*p0 - 2.*q3*p3 + 4.*q0*p0 + ex1))/detJ;
      const DST dq3 = (p0*( 2.*q3*q3 + exq12) + q0*(2.*p0*p0 + 2.*p0*q0 - 4.*q3*p3 + ex1))/detJ;
      const DST dp0 = (p3*(-2.*q0*q0 - exq12 - 4.*q0*p0 + 2.*q3*q3 - 4.*q0*s0)
      + q3*(-exr12 + 2.*p0*p0 + 2.*p3*p3 + 4.*p3*s3))/detJ;
      const DST dp3 = (q0*(-2.*p0*p0 - exr12 - 2.*q0*p0 - 2.*p3*p3 - 4.*p0*s0)
      + p0*(-exq12 + 2.*q3*q3 + 4.*q3*p3 + 4.*q3*s3))/detJ;
      newq0 = q0 + dq0;
      newq3 = q3 + dq3;
      newp0 = p0 + dp0;
      newp3 = p3 + dp3;
      prevdist = dist;
      dist = FABS(s0 + newp0 + newq0) + FABS(newp0*newp0 - (p12sq + newp3*newp3 + pM2))
      + FABS(s3 + newp3 + newq3) + FABS(newq0*newq0 - (q12sq + newq3*newq3 + qM2));
   } while (prevdist > dist && ++count < max_iter);

   pout[0+4*np] = p0;
   pout[3+4*np] = p3;
   pout[0+4*nq] = q0;
   pout[3+4*nq] = q3;

   return !(prevdist > dist && count == max_iter);
}

} // anonymous namespace

//------------------------------------------------------------------------------
bool refineConfiguration(Configuration& config, size_t max_iter)
{
   size_t n = config.size();

   // if config has not enough particles we cannot do anything
   if (n <= 3)
      return false;

//    Log::debug().turnOn();
//       Log::debug("Configuration::refine")
//       << "Original configuration:\n"  << std::setprecision(16) << *this << std::setprecision(6) << Log::end;
//    Log::debug().turnOff();

   //Order the final state particles such that the last two correspond to the particle being refined
   //Choose the two leptons to be the particles that get refined
   std::vector<double> pin(4 * n), pout(4 * n);

   for (size_t mu = 0; mu <= 3; ++mu) {
      pin[mu] = config.initialA()->p()[mu];
      pin[4 + mu] = config.initialB()->p()[mu];
   }
   size_t nonRefinePos = 2;
   size_t refinePos = n - 2;
   for (Configuration::iterator it = config.beginFinal(); it != config.endFinal(); it++) {
      size_t index;
      if (IsQCD()(it->id()))
         index = nonRefinePos++;
      else
         index = refinePos++;
      assert(index < n);
      for (size_t mu = 0; mu <= 3; ++mu) {
         pin[4 * index + mu] = it->p()[mu];
      }
   }
   //Sanity checks that we had the right number of particles
   assert(nonRefinePos == n - 2);

   //TODO: this assumes that there are only massless particles
   std::vector<double> scales(1, 0.0);

   const double scales_thr = 1e-5;

   // refine the momenta
   if (!refineMomenta(pin, pout, scales, scales_thr, max_iter)) {
      Log::warning("refineConfiguration") << "Failed to converge in " << max_iter << " iterations." << Log::end;
   }

   // Overwite original momenta with refined ones
   // Changing the particle momenta will not invalidate Configuration's integrity.
   for (size_t mu = 0; mu <= 3; mu++) {
      const_cast<Configuration::iterator::value_type&>(*config.initialA()).p()[mu] = pout[mu];
      const_cast<Configuration::iterator::value_type&>(*config.initialB()).p()[mu] = pout[4 + mu];
   }

   nonRefinePos = 2;
   refinePos = n - 2;
   for (Configuration::iterator it = config.beginFinal(); it != config.endFinal(); it++) {
      size_t index;
      if (IsQCD()(it->id()))
         index = nonRefinePos++;
      else
         index = refinePos++;
      assert(index < n);
      for (size_t mu = 0; mu <= 3; ++mu) {
         const_cast<Configuration::iterator::value_type&>(*it).p()[mu] = pout[4 * index + mu];
      }
   }

//       Log::debug().turnOn();
//       Log::debug("refineConfiguration")
//             << "Refined configuration:\n"  << std::setprecision(16) << config << std::setprecision(6) << Log::end;
//       Log::debug().turnOff();

   //Check that we didn't change the Configuration by too much
   //Check can be loose, since particles involved in smallest measure not changed
   const double Q = config.initialA()->energy() + config.initialB()->energy();
   const double epsilon = 1e-4 * Q;
   for (size_t i = 0; i < 4 * n; i++) {
      if (compareAbsolute(pin[i], pout[i], epsilon)) {
         Log::warning("Configuration::refine") << "Momentum component " << pin[i] << " changed by more than " << epsilon << Log::end;
         return false;
      }
   }
   return true;
}

} // namespace Geneva
