//------------------------------------------------------------------------------
/// \file Sigma.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class Sigma
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/Sigma.hpp"

#include <ostream>
#include <iterator>

namespace Geneva
{

//------------------------------------------------------------------------------
// Queries and Accessors
//------------------------------------------------------------------------------
/**
 * \brief Computes the minimum and maximum of the %Sigma's variations.
 * \return A pair of doubles containing the found {min, max}.
 *
 * This includes the central value in finding the minimum and maximum variation,
 * such that always min <= central <= max (including the cases when the %Sigma
 * has zero or one variations). The complexity is linear in the number of
 * variations. The result is recomputed on each call, so it is the user's
 * responsibility to store it in case it is needed frequently.
 */
std::pair<double, double> Sigma::getMinMaxVariations() const
{
   double min = _central;
   double max = _central;
   for (const_iterator it = _variations.begin(); it != _variations.end(); ++it) {
      if (*it < min)
         min = *it;
      else if (*it > max)
         max = *it;
   }
   return std::make_pair(min, max);
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Streams a string in the format "central (variations)" to out.
std::ostream& Sigma::print(std::ostream& out) const
{
   out << _central;
   if (_variations.size() > 0) {
      out << " (";
      std::copy(_variations.begin(), --_variations.end(), std::ostream_iterator<double>(out, ", "));
      out << *_variations.rbegin() << ")";
   }
   out << std::endl;
   return out;
}

} // namespace Geneva
