//------------------------------------------------------------------------------
/// \file Signature.cpp
//
// Author(s):
//    Frank Tackmann, Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 MIT, Frank Tackmann, LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class Signature
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/Signature.hpp"
#include "Geneva/Core/Physics/Configuration.hpp"

#include <algorithm>
#include <ostream>
#include <iterator>
#include <cassert>

namespace Geneva
{

//------------------------------------------------------------------------------
// Accessing and Querying IDs
//------------------------------------------------------------------------------
/**
 * \brief Calculates the symmetry factor of the %Signature.
 * \return The symmetry factor.
 *
 * The symmetry factor is obtained by the product of the factorials of the
 * multiplicities of each final-state ID in %Signature.
 *
 * \note This uses \c size_t and therefore only works for small multiplicities
 * (usually up to 12).
 */
size_t Signature::symmetryFactor() const
{
   if (_beginFinal == end())
      return 1;

   iterator it = _beginFinal;
   _status_id id = *it;
   ++it;
   size_t result = 1;
   unsigned count = 1;

   while (it != end()) {
      if (_compare_type()(id, *it)) {
         id = *it;
         if (count > 1) {
            double fac = factorial(count);
            assert(fac * static_cast<double>(result) < static_cast<double>(std::numeric_limits<size_t>::max()));
            result *= static_cast<size_t>(fac);
            count = 1;
         }
      } else
         ++count;
      ++it;
   }
   if (count > 1) {
      double fac = factorial(count);
      assert(fac * static_cast<double>(result) < static_cast<double>(std::numeric_limits<size_t>::max()));
      result *= static_cast<size_t>(fac);
   }
   return result;
}

//------------------------------------------------------------------------------
/**
 * \brief Tests if a %Signature is contained in the %Signature.
 * \param other The %Signature to be searched for.
 * \return \c true if all IDs in \a other are contained in the %Signature,
 *         \c false otherwise.
 *
 * If true is returned, \a other is a subsignature of the %Signature, i.e. their
 * intersection will be equal to \a other. The number of duplicate IDs is taken
 * into account, that is, the %Signature must have at least as many of the same
 * ID as are in \a other for \c true to be returned. Requires linear time.
 */
bool Signature::includes(const Signature& other) const
{
   return std::includes(begin(), end(), other.begin(), other.end(), _compare_type());
}

//------------------------------------------------------------------------------
/**
 * \brief Tests if a %Signature intersects with the %Signature.
 * \param other The %Signature to be tested.
 * \return \c true if at least one ID in \a other is also contained in the the
 *         %Signature, \c false otherwise.
 *
 * If \c true is returned, the %Signature has a non-empty intersection with
 * \a other. In this case this is faster than computing the intersection (since
 * the algorithm can stop at the first common ID). For \c false, it takes
 * exactly as long as computing the (empty) intersection (because all IDs have
 * to be checked in this case). Hence, if the intersection is later needed, it
 * is overall faster to directly compute it and check for its emptiness.
 */
bool Signature::intersects(const Signature& other) const
{
   iterator i1         = begin();
   iterator i2         = other.begin();
   const iterator end1 = end();
   const iterator end2 = other.end();

   // Since we only test for intersection we can stop as soon as we find one
   // common element.
   while (i1 != end1 && i2 != end2) {
      if (_compare_type()(*i1, *i2))
         ++i1;
      else if (_compare_type()(*i2, *i1))
         ++i2;
      else
         return true;
   }
   return false;
}

//------------------------------------------------------------------------------
// Adding and Removing IDs
//------------------------------------------------------------------------------
/**
 * \brief Inserts multiple copies of an ID into the %Signature's final state.
 * \param id The ID to be inserted.
 * \param n The number of copies to be inserted.
 *
 * This is somewhat faster than calling insert(id) \a n times. If n is zero,
 * nothing happens.
 */
void Signature::insertFinal(PdgID id, size_t n)
{
   if (n == 0)
      return;
   iterator it = insertFinal(id);
   // ensures that additional elements are placed behind the first inserted one
   // so _beginFinal does not need to be updated again.
   ++it;
   --n;
   while (n) {
      _ids.insert(it, id);
      --n;
   }
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------
/**
 * \brief %Signature equality comparison operator
 * \param rhs The %Signature to be compared to
 * \return \c true if the %Signature contains exactly the same IDs as \a rhs,
 *         \c false otherwise.
 *
 * Initial-state and final-state information as well as duplicate IDs are taken
 * into account. Requires linear time.
 */
bool Signature::operator==(const Signature& rhs) const
{
   if (size() != rhs.size())
      return false;

   iterator i1 = begin();
   if (i1 == end())
      return true;
   iterator i2 = rhs.begin();

   if (i1->isInitialA()) {
      if ((!i2->isInitialA()) || *i1 != *i2)
         return false;
      ++i1;
      if (i1 == end())
         return true;
      ++i2;
   }
   if (i1->isInitialB()) {
      if ((!i2->isInitialB()) || *i1 != *i2)
         return false;
      ++i1;
      if (i1 == end())
         return true;
      ++i2;
   }
   if (!i2->isFinal())
      return false;
   while (true) {
      if (*i1 != *i2)
         return false;
      ++i1;
      if (i1 == end())
         return true;
      ++i2;
   };
}

//------------------------------------------------------------------------------
/**
 * \brief %Signature ordering relation operator
 * \param rhs The %Signature to be compared to
 * \return \c true if the %Signature is considered "less than" \a rhs, \c false
 *         otherwise.
 *
 * For Signatures of different size the smaller is considered "less". For equal
 * size Signatures, the comparison is done lexicographically, where each
 * %Signature is considered a "word" and each of its IDs a "letter".
 */
bool Signature::operator<(const Signature& rhs) const
{
   if (size() < rhs.size())
      return true;
   if (size() > rhs.size())
      return false;
   return std::lexicographical_compare(begin(), end(), rhs.begin(), rhs.end(), _compare_type());
}

//------------------------------------------------------------------------------
/**
 * \brief %Signature subtraction (complement) assignment operator
 * \param rhs The %Signature to take the complement with.
 * \return A reference to the %Signature
 *
 * Takes the complement of the %Signature with \a rhs. As a result the
 * %Signature will only contain IDs that are not contained in \a rhs. IDs in
 * \a rhs that are not part of the %Signature have no effect (apart from
 * maybe slowing down the operation). The number of duplicate IDs is taken into
 * account. That is, if \a rhs has N instances of the same ID, then N will be
 * removed from the %Signature. (If it has less than N, then all will be
 * removed.) Requires linear time.
 *
 */
Signature& Signature::operator-=(const Signature& rhs)
{
   iterator         i1 = begin();
   iterator         i2 = rhs.begin();
   const iterator end1 = end();
   const iterator end2 = rhs.end();

   // We exploit here that set iterators stay valid if elements are erased,
   // but it's important to use erase(i1++), which advances i1 and returns its
   // old value, which is then erased.
   while (i1 != end1 && i2 != end2) {
      if (_compare_type()(*i1, *i2))
         ++i1;
      else if (_compare_type()(*i2, *i1))
         ++i2;
      else {
         erase(i1++);
         ++i2;
      }
   }
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %Signature AND (intersection) assignment operator
 * \param rhs The %Signature to take the intersection with
 * \return A reference to the %Signature
 *
 * Takes the intersection of the %Signature with \a rhs. It only keeps IDs in
 * the %Signature that are also part of \a rhs, removing all others. IDs in
 * \a rhs that are not part of the %Signature have no effect (apart from
 * maybe slowing down the operation). The number of duplicate IDs is taken into
 * account. That is, only as many instances of the same ID in the %Signature are
 * kept as there are in \a rhs. Requires linear time.
 *
 */
Signature& Signature::operator&=(const Signature& rhs)
{
   iterator         i1 = begin();
   iterator         i2 = rhs.begin();
   const iterator end1 = end();
   const iterator end2 = rhs.end();

   // We exploit here that set iterators stay valid if elements are erased,
   // but it's important to use erase(i1++), which advances i1 and returns its
   // old value, which is then erased.
   while (i1 != end1 && i2 != end2) {
      if (_compare_type()(*i1, *i2))
         erase(i1++);
      else if (_compare_type()(*i2, *i1))
         ++i2;
      else {
         ++i1;
         ++i2;
      }
   }
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %Signature subtraction (complement) operator
 * \param lhs The %Signature to take the complement of.
 * \param rhs The %Signature to take the complement with.
 * \return The resulting %Signature.
 *
 * Takes the complement of \a lhs with \a rhs and returns the result in a new
 * %Signature. See operator-= for how the complement is taken. Requires linear
 * time.
 */
const Signature operator-(const Signature& lhs, const Signature& rhs)
{
   Signature result;
   std::set_difference(lhs.begin(), lhs.end(), rhs.begin(), rhs.end(),
                       std::inserter(result._ids, result.end()), Signature::_compare_type());
   result.init();
   return result;
}

//------------------------------------------------------------------------------
/**
 * \brief %Signature AND (intersection) operator
 * \param lhs The 1st %Signature to take the intersection with.
 * \param rhs The 2nd %Signature to take the intersection with.
 * \return The resulting %Signature.
 *
 * Takes the intersection of \a lhs and \a rhs and returns the result in a new
 * %Signature. See operator&= for how the intersection is taken. Requires linear
 * time.
 */
const Signature operator&(const Signature& lhs, const Signature& rhs)
{
   Signature result;
   std::set_intersection(lhs.begin(), lhs.end(), rhs.begin(), rhs.end(),
                         std::inserter(result._ids, result.end()), Signature::_compare_type());
   result.init();
   return result;
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Streams the Signature to out.
std::ostream& Signature::print(std::ostream& out) const
{
   std::copy(begin(), end(), std::ostream_iterator<PdgID>(out, " "));
   return out;
}

//------------------------------------------------------------------------------
// Initializes the internal initial-state and final-state iterators.
void Signature::init()
{
   _beginFinal = begin();
   if (_beginFinal != end() && _beginFinal->isInitialA()) {
      _initialA = _beginFinal;
      ++_beginFinal;
   } else
      _initialA = end();
   if (_beginFinal != end() && _beginFinal->isInitialB()) {
      _initialB = _beginFinal;
      ++_beginFinal;
   } else
      _initialB = end();
}

} // namespace Geneva
