//------------------------------------------------------------------------------
/// \file ThreeVector.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class ThreeVector
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/ThreeVector.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include <ostream>

namespace Geneva
{

//------------------------------------------------------------------------------
// Accessing and Querying Components
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Returns the component indexed by mu.
 * \param mu The index of the component 1 <= mu <= 3.
 * \return The value of the component.
 *
 * If mu is out of range an out_of_range error is thrown.
 */
double ThreeVector::operator[](size_t mu) const
{
   if (mu == 1)
      return _p1;
   if (mu == 2)
      return _p2;
   if (mu != 3)
      Log::error("ThreeVector::operator[]") << "mu index " << mu << " is out of range" << Log::end;
   return _p3;
}

//------------------------------------------------------------------------------
/**
 * \brief Gives access to the component indexed by mu.
 * \param mu The index of the component 1 <= mu <= 3.
 * \return A reference to the component.
 *
 * If mu is out of range an out_of_range error is thrown.
 */
double& ThreeVector::operator[](size_t mu)
{
   if (mu == 1)
      return _p1;
   if (mu == 2)
      return _p2;
   if (mu != 3)
      Log::error("ThreeVector::operator[]") << "mu index " << mu << " is out of range" << Log::end;
   return _p3;
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief %ThreeVector inequality comparison operator
 * \param rhs The %ThreeVector to compare to.
 * \return \c true if at least one of the components are different,
 *         \c false otherwise.
 *
 * The comparison is done with the default accuracy and precision separately for
 * each of the components.
 */
bool ThreeVector::operator!=(const ThreeVector& rhs) const
{
   return compare(_p1, rhs._p1) || compare(_p2, rhs._p2) || compare(_p3, rhs._p3);
}

//------------------------------------------------------------------------------
/**
 * \brief Test if the %ThreeVector is nonzero.
 * \return \c true if at least one component is nonzero, \c false otherwise.
 *
 * Compares each components to 0.0 using the default absolute accuracy.
 *
 * \warning This function is here for completeness. Directly comparing to zero
 * is usually not very meaningfull for doubles since the overall scale is
 * unknown, so use of this function should be avoided. For example, one should
 * compare two %ThreeVectors with each other rather than their difference to
 * zero.
 */
bool ThreeVector::isNonzero() const
{
   return compareToZero(_p1) || compareToZero(_p2) || compareToZero(_p3);
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Streams a string in the format (p1, p2, p3) to out.
std::ostream& ThreeVector::print(std::ostream& out) const
{
   return out << "(" << _p1 << ", " << _p2 << ", " << _p3 << ")";
}

} // namespace Geneva
