//------------------------------------------------------------------------------
/// \file GenerationReport.cpp
//
// Author(s):
//    Calvin Berggren, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class GenerationReport
//------------------------------------------------------------------------------

#include "Geneva/version.hpp"

#include "Geneva/Core/UI/GenerationReport.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include "tinyxml/tinyxml.hpp"

#include <ctime>
#include <fstream>

namespace Geneva
{

using std::string;

//------------------------------------------------------------------------------
GenerationReport::GenerationReport(const Geneva::OptionMap& options, const std::string& executable,
                                   const std::string& pythiaOptionsFile)
   : _options(options), _pyoptions(pythiaOptionsFile), _output("(* no output *)"),
     _title(getDefaultTitle(options)),
     _executable(executable), _runTime(0.), _numEvents(0)
{
}

//------------------------------------------------------------------------------
void GenerationReport::clearOutput()
{
   _output.str("(* no output *)");
}

//------------------------------------------------------------------------------
const char* GenerationReport::version() const
{
   return config::version;
}

//------------------------------------------------------------------------------
const char* GenerationReport::modifiedFiles() const
{
   return config::gitChanges;
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Prints the full report in XML format into the given output stream.
 * \param out The output stream to be used.
 * \return A reference to \a out.
 */
std::ostream& GenerationReport::print(std::ostream& out) const
{
   tinyxml::TiXmlDocument report;

   tinyxml::TiXmlDeclaration* decl = new tinyxml::TiXmlDeclaration("1.0", "", "");
   report.LinkEndChild(decl);

   tinyxml::TiXmlElement* root = new tinyxml::TiXmlElement("GenevaOptionSet");
   report.LinkEndChild(root);

   attachSummary(root);

   for (OptionMap::const_iterator it = _options.begin(); it != _options.end(); ++it) {
      addElement(it->second->name(), "", root);
   }

   if(!_pyoptions.empty()) {
      std::ifstream t(_pyoptions.c_str());
      if (t.good()) {
         tinyxml::TiXmlElement* pyop = new tinyxml::TiXmlElement("PythiaOptions");
         std::stringstream buffer;
         buffer << t.rdbuf();
         tinyxml::TiXmlText* pyoutput = new tinyxml::TiXmlText("\n"+buffer.str());
         pyoutput->SetCDATA(true);
         pyop->LinkEndChild(pyoutput);
         root->LinkEndChild(pyop);
      }
   }

   // add the output from the main run
   tinyxml::TiXmlElement* anal = new tinyxml::TiXmlElement("AnalyzerOutput");
   tinyxml::TiXmlText* output = new tinyxml::TiXmlText(_output.str());
   output->SetCDATA(true);
   anal->LinkEndChild(output);
   root->LinkEndChild(anal);

   tinyxml::TiXmlPrinter printer;
   printer.SetIndent("   ");
   report.Accept(&printer);

   return out << printer.Str();
}

//------------------------------------------------------------------------------
void GenerationReport::attachSummary(tinyxml::TiXmlElement* node) const
{
   tinyxml::TiXmlElement* summary = new tinyxml::TiXmlElement("Summary");
   node->LinkEndChild(summary);

   // run title
   tinyxml::TiXmlElement* titleEl = new tinyxml::TiXmlElement("Title");
   titleEl->LinkEndChild(new tinyxml::TiXmlText(_title));
   summary->LinkEndChild(titleEl);

   // run description
   tinyxml::TiXmlElement* description = new tinyxml::TiXmlElement("Description");
   description->LinkEndChild(new tinyxml::TiXmlText(_description));
   summary->LinkEndChild(description);

   // date and time
   {
      time_t rawtime = std::time(0);
      char formattedDate[80];
      char formattedTime[80];

      std::strftime(formattedDate, 80, "%Y-%m-%d", std::localtime(&rawtime));
      // TODO: XML Time data type prefers time zone as an offset from UTC
      std::strftime(formattedTime, 80, "%H:%M:%S %Z", std::localtime(&rawtime));

      tinyxml::TiXmlElement* date = new tinyxml::TiXmlElement("Date");
      date->LinkEndChild(new tinyxml::TiXmlText(formattedDate));
      summary->LinkEndChild(date);

      tinyxml::TiXmlElement* time = new tinyxml::TiXmlElement("Time");
      time->LinkEndChild(new tinyxml::TiXmlText(formattedTime));
      summary->LinkEndChild(time);
   }

   // run time
   std::stringstream ss;
   ss << static_cast<size_t>(_runTime);
   tinyxml::TiXmlElement* runTime = new tinyxml::TiXmlElement("RuntimeInSec");
   runTime->LinkEndChild(new tinyxml::TiXmlText(ss.str()));
   summary->LinkEndChild(runTime);

   // number of generated events
   ss.str("");
   ss << _numEvents;
   tinyxml::TiXmlElement* numEvents = new tinyxml::TiXmlElement("NumEvents");
   numEvents->LinkEndChild(new tinyxml::TiXmlText(ss.str()));
   summary->LinkEndChild(numEvents);

   // name of executable
   tinyxml::TiXmlElement* executable = new tinyxml::TiXmlElement("Executable");
   executable->LinkEndChild(new tinyxml::TiXmlText(_executable));
   summary->LinkEndChild(executable);

   // version
   tinyxml::TiXmlElement* genevaVersion = new tinyxml::TiXmlElement("Version");
   genevaVersion->LinkEndChild(new tinyxml::TiXmlText(version()));
   summary->LinkEndChild(genevaVersion);

   // modified files (only included if not empty)
   if (*modifiedFiles()) {
      tinyxml::TiXmlElement* modFiles= new tinyxml::TiXmlElement("ModifiedFiles");
      tinyxml::TiXmlText* data = new tinyxml::TiXmlText("\n"+std::string(modifiedFiles()));
      data->SetCDATA(true);
      modFiles->LinkEndChild(data);
      summary->LinkEndChild(modFiles);
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Adds an element according to the name given and connects it to the node given.
 * \param name The name of the option to be created.
 * \param qualifier The prefix needed by the OptionMap to fully specify the option.
 * \param top The parent node to link to after creation
 *
 * The function is used recursively to handle nested options of the form
 * "parent::child". The qualifier is the full prefix "grandparent::" (including
 * the trailing "::") needed by the OptionMap to find the full option
 * "grandparent::parent::child".
 *
 * The function extracts the leading parent name at the first occurring "::",
 * create a node for it if one doesn't already exist, and then recurse to the
 * child level.
 *
 * \todo Factorize this functionality into a separate OptionWriterXML class.
 */
void GenerationReport::addElement(const std::string& name, const std::string& qualifier,
                                  tinyxml::TiXmlElement* top) const
{
   // split nested option into parent and child at first occurence of "::"
   string parentName;
   string childName;
   size_t pos = name.find("::");
   if (pos != string::npos) {
      parentName = name.substr(0, pos);
      childName = name.substr(pos + 2);
   } else {
      parentName = name;
   }

   // try to find the parent in case it has already been created
   tinyxml::TiXmlElement* parent = top->FirstChildElement(parentName);
   if (!parent) {
      // create the node for the option and link it
      tinyxml::TiXmlElement* optionElement = new tinyxml::TiXmlElement(parentName);
      const string val = _options.getOption(qualifier + parentName);
      const bool changed = !_options.isDefault(qualifier + parentName);
      if (changed)
         optionElement->SetAttribute("changed", "true");
      if (!val.empty())
         optionElement->LinkEndChild(new tinyxml::TiXmlText(val));
      top->LinkEndChild(optionElement);
      parent = optionElement;
   }

   // for nested option, recurse to the next level
   if (!childName.empty())
      addElement(childName, qualifier + parentName + "::", parent);
}

} // namespace Geneva
