//------------------------------------------------------------------------------
/// \file Log.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class Log
//------------------------------------------------------------------------------

#include "Geneva/Core/UI/Log.hpp"

#include <iostream>

namespace Geneva
{

//------------------------------------------------------------------------------
reference<std::ostream> Log::_default(std::cout);
const Message::End Log::end = Message::End();
ErrorHandler Log::_fatalHandler(_default, true, "FATAL ERROR in ", TextColor::kBoldMagenta);
ErrorHandler Log::_errorHandler(_default, true, "ERROR in ", TextColor::kBoldRed);
WarningHandler Log::_warningHandler(_default, true, 1000, "WARNING in ", TextColor::kBrown);
InfoHandler Log::_infoHandler(_default, true, "INFO from ", TextColor::kBold);
InfoHandler Log::_verboseInfoHandler(_default, true, "INFO from ", TextColor::kBold);
InfoHandler Log::_debugHandler(_default, false, "DEBUG from ", TextColor::kBoldBlue);

//------------------------------------------------------------------------------
/**
 * \brief Sets the output stream of all handlers.
 * \param defaultOut The new default output stream to use.
 * \return A reference to the current default output stream.
 *
 * This sets the output stream of all handlers to \a defaultOut. To set the
 * output stream of individual handlers use \c handler().setOutputStream where
 * 'handler' is one of the handlers.
 *
 * \note At program start the default output stream is to \c std::cout.
 */
std::ostream& Log::setDefaultOutStream(std::ostream& defaultOut)
{
   std::ostream& tmp(_default);
   _default.set(defaultOut);
   fatal().setOutStream(_default);
   error().setOutStream(_default);
   warning().setOutStream(_default);
   info().setOutStream(_default);
   verboseInfo().setOutStream(_default);
   debug().setOutStream(_default);
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief Turns handlers on/off according to given verbosity level.
 * \param level The verbosity level to set.
 *
 * All handlers at and below the given verbosity level are turn on, while all
 * handlers of higher verbosity are turned off. The used verbosity ordering from
 * lowest to highest is fatal < error < warning < info < verboseInfo < debug.
 *
 * \note At program start all handlers except debug are turned on (corresponding
 * to verbosity level \c kVerboseInfo).
 */
void Log::setVerbosity(Level level)
{
   fatal().turnOff();
   error().turnOff();
   warning().turnOff();
   info().turnOff();
   verboseInfo().turnOff();
   debug().turnOff();
   switch (level)
   {
      case kDebug:
         debug().turnOn();
      case kVerboseInfo:
         verboseInfo().turnOn();
      case kInfo:
         info().turnOn();
      case kWarning:
         warning().turnOn();
      case kError:
         error().turnOn();
      case kFatal:
         fatal().turnOn();
      case kOff: ;
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Turns off all coloring.
 *
 * This sets the color of all handlers to \c TextColor::kOff. Use this to avoid
 * cluttering output with color codes when the default output stream goes to a
 * file rather than screen.
 */
void Log::setColorsOff()
{
   fatal().setColor(TextColor::kOff);
   error().setColor(TextColor::kOff);
   warning().setColor(TextColor::kOff);
   info().setColor(TextColor::kOff);
   verboseInfo().setColor(TextColor::kOff);
   debug().setColor(TextColor::kOff);
}

//------------------------------------------------------------------------------
/**
 * \brief Sets the colors of all of handlers to their default values.
 *
 * This sets the color of all handlers to their default values:
 * - fatal: \c kBoldMagenta
 * - error: \c kRed
 * - warning: \c kBrown
 * - info: \c kOff
 * - verboseInfo: \c kOff
 * - debug: \c kBlue
 *
 * By default info and verboseInfo are uncolored.
 */
void Log::setColorsDefault()
{
   fatal().setColor(TextColor::kBoldMagenta);
   error().setColor(TextColor::kBoldRed);
   warning().setColor(TextColor::kBrown);
   info().setColor(TextColor::kBold);
   verboseInfo().setColor(TextColor::kBold);
   debug().setColor(TextColor::kBoldBlue);
}

} // namespace Geneva
