//------------------------------------------------------------------------------
/// \file MessageHandler.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of MessageHandler classes
//------------------------------------------------------------------------------

#include "Geneva/Core/UI/MessageHandler.hpp"

#include <utility>

namespace Geneva
{

//------------------------------------------------------------------------------
// Message
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * Internal member returning a reference to a null ostream that discards
 * anything it receives.
 */
std::ostream& Message::null()
{
   static std::ostream _null(0);
   return _null;
}

//------------------------------------------------------------------------------
// MessageHandlerBase
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Starts a message.
 * \param where A string specifying the origin of the message.
 * \return A Message object.
 *
 * This outputs a header of the form "tag where: " in the handlers color,
 * followed by any output streamed to the returned Message object. The output is
 * directly forwarded to the output stream.
 *
 * If the handler is turned off, no header is output and any output streamed to
 * returned Message is discarded.
 *
 * \note This is the default implementation which can be modified by derived
 * handler classes.
 */
Message MessageHandlerBase::startMessage(const std::string& where)
{
   if (isOn()) {
      _out << _color << _tag << where << ": ";
      _color.reset(_out);
      return Message(_out, this);
   }
   return Message();
}

//------------------------------------------------------------------------------
/**
 * \brief Concludes a message.
 *
 * This is called upon streaming the \c Message::End object to a previously
 * obtained Message object.
 *
 * This simply inserts a new line '\n' into the output stream. Normally this
 * only causes a flush for output streams going to the screen but not file or
 * memory.
 *
 * \note This is the default implementation which can be modified by derived
 * handler classes.
 */
void MessageHandlerBase::endMessage()
{
   if (isOn())
      _out.get() << '\n';
}

//------------------------------------------------------------------------------
// WarningHandler
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Starts a warning message.
 * \param where A string specifying the origin of the message.
 * \return A Message object.
 *
 * This outputs a header of the form "tag where: " in the handlers color,
 * followed by any output streamed to the returned Message object. The message
 * is kept by the handler and only sent to the output stream by the concluding
 * \c endMessage() call.
 *
 * If the handler is turned off, no header is output. The output streamed to the
 * returned Message is still used to construct the warning message.
 *
 * \note If \c startMessage() is called again before concluding a warning message
 * the text of the previous message is discarded.
 */
Message WarningHandler::startMessage(const std::string& where)
{
   if (isOn()) {
      _out << _color << _tag << where << ": ";
      _color.reset(_out);
   }
   _current = _counts.insert(std::make_pair(where, 0)).first;
   _what.str("");
   return Message(_what, this);
}

//------------------------------------------------------------------------------
/**
 * \brief Concludes a warning message.
 *
 * This is called upon streaming the \c Message::End object to a previously
 * obtained Message object.
 *
 * If the handler is turned on, the received warning message is sent to the output
 * stream, which is then flushed. If the handler is turned off, the message is not sent
 * to the output stream.
 *
 * If the number of allowed warnings is exceeded, a FatalError is raised.
 */
void WarningHandler::endMessage()
{
   if (isOn()) {
      _out.get() << _what.str() << std::endl;
   }
   if (_current != _counts.end() && ++_current->second > _maxCounts) {
      ErrorHandler fatal(_out, isOn(), "FATAL ERROR in ", TextColor::kBoldMagenta);
      fatal.startError<FatalError>("WarningHandler")
         << "Exceeded " << _maxCounts << " allowed warnings from: " << _current->first;
      _current = _counts.end();
      fatal.endMessage();
   }
}

//------------------------------------------------------------------------------
void WarningHandler::writeCounts()
{
   if (!_counts.empty() && isOn()) {
      _out.get() << "Total number of warnings:\n";
      for (std::map<const std::string, size_t>::iterator it = _counts.begin(); it != _counts.end(); ++it)
         _out.get() << it->second << "\tfrom " << it->first << "\n";
   }
}

//------------------------------------------------------------------------------
// ErrorHandler
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Starts an error message.
 * \param where A string specifying the origin of the message.
 * \return A Message object.
 *
 * This outputs a header of the form "tag where: " in the handlers color,
 * followed by any output streamed to the returned Message object. The message
 * is kept by the handler and only sent to the output stream by the concluding
 * \c endMessage() call.
 *
 * If the handler is turned off, no header is output. The output streamed to the
 * returned Message is still used to construct the error message.
 *
 * \note If \c startMessage() is called again before concluding an error message
 * the text of the previous message is discarded.
 */
Message ErrorHandler::startMessage(const std::string& where)
{
   if (isOn()) {
      _out << _color << _tag << where << ": ";
      _color.reset(_out);
   }
   _what.str("");
   _where = where;
   return Message(_what, this);
}

//------------------------------------------------------------------------------
/**
 * \brief Concludes an error message and throws an exception.
 *
 * This is called upon streaming the \c Message::End object to a previously
 * obtained Message object.
 *
 * If the handler is turned on, the received error message is sent to the output
 * stream, which is then flushed. Then an \c ErrorType exception is thrown,
 * whose \c what() member is initialized with the text of the received
 * error message. If the handler is turned off, the message is not sent to the
 * output stream, but the exception is still thrown. The exception also contains
 * the information where the error came from and if it was output or not.
 */
void ErrorHandler::endMessage()
{
   bool logged = false;
   if (isOn()) {
      if (_out.get() << _what.str() << std::endl)
         logged = true;
   }
   if (_thrower) {
      void (*tmp)(const std::string&, const std::string&, bool) = _thrower;
      _thrower = 0;
      (*tmp)(_what.str(), _where, logged);
   }
}

} // namespace Geneva
