//------------------------------------------------------------------------------
/// \file OptionMap.cpp
//
// Author(s):
//    Chris Vermilion, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class OptionMap
//------------------------------------------------------------------------------

#include "Geneva/Core/UI/OptionMap.hpp"

#include <ostream>
#include <iomanip>

namespace Geneva
{

//------------------------------------------------------------------------------
std::string OptionMap::getOption(const std::string& name) const
{
   const_iterator it = _optionMap.find(name);
   if (it == _optionMap.end())
      Log::error("OptionMap::getOption") << "Could not retrieve value of unknown option \'" << name << "\'." << Log::end;
   return it->second->getValueStr();
}

//------------------------------------------------------------------------------
/// Whether this option has been changed from its default.
bool OptionMap::isDefault(const std::string& name) const
{
   const_iterator it = _optionMap.find(name);
   if (it == _optionMap.end())
      Log::error("OptionMap::isDefault") << "Looking for unknown option \'" << name << "\'." << Log::end;

   return !(((*it).second)->changed());
}

//------------------------------------------------------------------------------
bool OptionMap::optionExists(const std::string& name) const
{
   return _optionMap.find(name) != _optionMap.end();
}

//------------------------------------------------------------------------------
/**
 * \brief Sets an existing option to a given value.
 *
 * Throws an exception if the option does not exist.
 */
void OptionMap::setOption(const std::string& name, const std::string& val)
{
   const_iterator it = _optionMap.find(name);
   if (it == _optionMap.end())
      Log::error("OptionMap::setOption") << "Could not set unknown option \'"<< name << "\'." << Log::end;
   it->second->setValue(val);
}

//------------------------------------------------------------------------------
void OptionMap::setUserOptions(const OptionMap& options, bool enforceExists)
{
   for (const_iterator it = options.begin(); it != options.end(); ++it) {
      if (optionExists(it->first))
         setOption(it->first, it->second->getValueStr());
      else  {
         if (enforceExists)
            Log::error("OptionMap::setUserOptions") << "Could not set unknown option \'"<< it->first << "\'." << Log::end;
         else
            Log::warning("OptionMap::setUserOptions") << "Ignoring unknown option \'"<< it->first << "\'." << Log::end;
      }
   }
}

//------------------------------------------------------------------------------
void OptionMap::printOptions() const
{
   Log::info() << "Available options and default values:\n";
   for (const_iterator it = begin(); it != end(); ++it)
      Log::info() << "  --" << std::setw(50) << std::left << it->first << "   ("
      << it->second->getDefaultValueStr() << ")\n";
}

//------------------------------------------------------------------------------
/// Loop over all options and print
std::ostream& operator<<(std::ostream& os, const OptionMap& optMap)
{
   for (OptionMap::const_iterator it = optMap.begin(); it != optMap.end(); ++it) {
      std::stringstream value, defaultvalue;
      value << it->second->getValueStr();
      defaultvalue << it->second->getDefaultValueStr();
      if (it->second->changed())
         os << "\033[1;36m"; // display cyan if changed

      os << " " << std::setw(50) << std::left << it->first
      << ": "  << value.str()
      << " (" << defaultvalue.str() << ")"
      << "\033[0m" << std::endl;
   }
   return os;
}

} //namespace Geneva
