//------------------------------------------------------------------------------
/// \file OptionParserCL.cpp
//
// Author(s):
//    Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class OptionParserCL
//------------------------------------------------------------------------------

#include "Geneva/Core/UI/OptionParserCL.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <algorithm>

namespace Geneva
{

//------------------------------------------------------------------------------
OptionParserCL::OptionParserCL(int argc, const char* argv[])
{
   // put all command line arguments in one stream, separated by spaces
   std::stringstream stream;
   for (int i = 1; i < argc; i++) {
      stream << argv[i] << ' ';
   }
   parse(stream);
}

//------------------------------------------------------------------------------
OptionParserCL::OptionParserCL(const std::string& optionString)
{
   std::stringstream stream(optionString);
   parse(stream);
}

//------------------------------------------------------------------------------
// Checks if a specific option is present.
bool OptionParserCL::hasOption(const std::string& name) const
{
   const_iterator it;
   return find(name, it);
}

//------------------------------------------------------------------------------
void OptionParserCL::getNext(std::string& name, std::string& value)
{
   name = _options.front().first;
   value = _options.front().second;
   _options.pop_front();
}

//------------------------------------------------------------------------------
namespace
{

struct NameEqual
{
   const std::string& _name;
   NameEqual(const std::string& tmp) : _name(tmp) {};
   bool operator()(const std::pair<std::string, std::string>& lhs)
   {
      return lhs.first == _name;
   };
};

}

//------------------------------------------------------------------------------
// Finds a specific option.
bool OptionParserCL::find(const std::string& name, const_iterator& it) const
{
   it = std::find_if(_options.begin(), _options.end(), NameEqual(name));
   return it != _options.end();
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
void OptionParserCL::parse(std::stringstream& in)
{
   std::string tempString = "";
   std::string name = "";
   std::string value = "";

   // treat first option specially
   if (in.peek() != '-') {
      Log::fatal("OptionParserCL::parseCommandLineOptions") << "First argument is not a valid command line option. Type --help for a list of valid options." << Log::end;
   }
   in >> tempString;
   // trim '-' or '--' from option name
   if (tempString.substr(1, 1) != "-") {
      Log::fatal("OptionParserCL::parseCommandLineOptions") << "Short options (like -h) are not supported." << Log::end;
      name = tempString.substr(1, tempString.size() - 1);
   } else {
      name = tempString.substr(2, tempString.size() - 2);
   }

   while (in >> tempString) {
      if ((tempString.substr(0, 1) == "-")  && (!std::isdigit(tempString.substr(1, 1)[0]))){
         // start a new option
         _options.push_back(std::make_pair(name, value));
         value = "";

         // trim '-' or '--' from option name
         if (tempString.substr(1, 1) != "-") {
            Log::fatal("OptionParserCL::parseCommandLineOptions") <<  "Short options (like -h) are not supported." << Log::end;
            name = tempString.substr(1, tempString.size() - 1);
         } else {
            name = tempString.substr(2, tempString.size() - 2);
         }

      } else {
         if (value != "")
            value += ' ';
         value += tempString;
      }
   }

   _options.push_back(std::make_pair(name, value));
}

} //namespace Geneva
