//------------------------------------------------------------------------------
/// \file OptionParserLHEFHeader.cpp
//
// Author(s):
//    Christian Bauer
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class OptionParserLHEFHeader
//------------------------------------------------------------------------------

#include "Geneva/Core/UI/OptionParserLHEFHeader.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <algorithm>
#include <istream>

namespace Geneva
{

//------------------------------------------------------------------------------
OptionParserLHEFHeader::OptionParserLHEFHeader(std::istream& optionStream)
{
   parse(optionStream);
}

//------------------------------------------------------------------------------
OptionParserLHEFHeader::OptionParserLHEFHeader(const std::string& optionString)
{
   std::stringstream stream(optionString);
   parse(stream);
}

//------------------------------------------------------------------------------
// Checks if a specific option is present.
bool OptionParserLHEFHeader::hasOption(const std::string& name) const
{
   const_iterator it;
   return find(name, it);
}

//------------------------------------------------------------------------------
void OptionParserLHEFHeader::getNext(std::string& name, std::string& value)
{
   name = _options.front().first;
   value = _options.front().second;
   _options.pop_front();
}

//------------------------------------------------------------------------------
namespace
{

struct NameEqual
{
   const std::string& _name;
   NameEqual(const std::string& tmp) : _name(tmp) {};
   bool operator()(const std::pair<std::string, std::string>& lhs)
   {
      return lhs.first == _name;
   };
};

}

//------------------------------------------------------------------------------
// Finds a specific option.
bool OptionParserLHEFHeader::find(const std::string& name, const_iterator& it) const
{
   it = std::find_if(_options.begin(), _options.end(), NameEqual(name));
   return it != _options.end();
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
void OptionParserLHEFHeader::parse(std::istream& in)
{
   std::string tempString = "";
   std::string name = "";
   std::string value = "";
   std::string delimiter = " : ";

   //Discard all lines until we get to "## Options"
   while (tempString != "  ## Options:") {
      std::getline(in, tempString);
      if (!(in)) {
         if (in.eof())
            Log::error("OptionParserLHEFHeader::parse") << "The LHEF file does not have proper header information" << Log::end;
         else
            Log::error("OptionParserLHEFHeader::parse") << "Error reading the LHEF file" << Log::end;
      }
   }

   std::getline(in, tempString);
   if (!(in)) {
      if (in.eof())
         Log::error("OptionParserLHEFHeader::parse") << "The LHEF file does not have proper header information" << Log::end;
      else
         Log::error("OptionParserLHEFHeader::parse") << "Error reading the LHEF file" << Log::end;
   }
   //Read until we get to "-->"
   while (tempString != "-->") {
      if (!(tempString.at(2) == '#'))
         Log::error("OptionParserLHEFHeader::parse") << "The LHEF file does not have proper header information" << Log::end;
      if (tempString == "<header>")
         Log::error("OptionParserLHEFHeader::parse") << "The LHEF file does not have proper header information" << Log::end;
      size_t pos = tempString.find(delimiter);
      name = tempString.substr(4, pos - 4);
      tempString.erase(0, pos + delimiter.length());
      value = tempString.substr(0, tempString.length());

      _options.push_back(std::make_pair(name, value));
      std::getline(in, tempString);
      if (!(in)) {
         if (in.eof())
            Log::error("OptionParserLHEFHeader::parse") << "The LHEF file does not have proper header information" << Log::end;
         else
            Log::error("OptionParserLHEFHeader::parse") << "Error reading the LHEF file" << Log::end;
      }
   }
}

} //namespace Geneva
