//------------------------------------------------------------------------------
/// \file OptionParserXML.cpp
//
// Author(s):
//    Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class OptionParserXML
//------------------------------------------------------------------------------

#include "Geneva/config.hpp"

#include "Geneva/Core/UI/OptionParserXML.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/Option.hpp"

#include "tinyxml/tinyxml.hpp"

#include <cstdlib>

namespace Geneva
{

//------------------------------------------------------------------------------
OptionParserXML::OptionParserXML()
{
   const char* env_cstr = std::getenv("GENEVA_DATA_DIR");
   if (env_cstr != 0 && *env_cstr != '\0')
      _fileName = std::string(env_cstr);
   else
      _fileName = std::string(config::data_dir);
   _fileName += "/DefaultOptions.xml";

   _loadFile();
}

//------------------------------------------------------------------------------
OptionParserXML::OptionParserXML(const std::string& fileName)
   : _fileName(fileName)
{
   _loadFile();
}

//------------------------------------------------------------------------------
OptionParserXML::~OptionParserXML()
{
   delete _document;
}

//------------------------------------------------------------------------------
OptionBase* OptionParserXML::getNext()
{
   const tinyxml::TiXmlElement* el = _elementStack.back();
   std::string name = _getScopedElementName();
   const char* v = el->GetText();
   const char* d = el->Attribute("desc");
   const char* t = el->Attribute("type");
   std::string val = v ? v : "";
   std::string type = t ? t : "";
   std::string desc = d ? d : "";

   _advanceToNextElement(); // do this before returning...

   if (type == "int")
      return new Option<int>(name, val, desc);
   else if (type == "double")
      return new Option<double>(name, val, desc);
   else if (type == "bool")
      return new Option<bool>(name, val, desc);
   else if (type == "string" || type == "")  // string is default
      return new Option<std::string>(name, val, desc);
   else if (type == "vector<int>")
      return new Option<std::vector<int> >(name, val, desc);
   else if (type == "vector<double>")
      return new Option<std::vector<double> >(name, val, desc);
   else if (type == "vector<bool>")
      return new Option<std::vector<bool> >(name, val, desc);
   else if (type == "vector<string>")
      return new Option<std::vector<std::string> >(name, val, desc);
   else {
      Log::error("OptionParserXML::getNext")<<"Unrecognized type " << type <<Log::end;
      return 0;
   }
}

//------------------------------------------------------------------------------
void OptionParserXML::getNext(std::string& name, std::string& value)
{
   const tinyxml::TiXmlElement* el = _elementStack.back();

   name = _getScopedElementName();
   const char* text = el->GetText(); // text will be NULL if another element immediately follows
   value = text ? std::string(text) : "";

   _advanceToNextElement();
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/// Loads and parses the xml file into an element tree.
void OptionParserXML::_loadFile()
{
   _document = new tinyxml::TiXmlDocument(_fileName);
   if (!_document->LoadFile())
      Log::fatal("OptionParserXML") << "Error parsing file "<< _fileName << '\n'
         << _document->ErrorDesc() <<" at line " << _document->ErrorRow() << " column " <<_document->ErrorCol() << Log::end;

   tinyxml::TiXmlElement* root = _document->RootElement();
   if (root == 0 || root->ValueStr() != "GenevaOptionSet")
      Log::fatal("OptionParserXML") << _fileName << " does not have a valid root element." << Log::end;

   if (root->FirstChildElement())
      _elementStack.push_back(root->FirstChildElement());
}

//------------------------------------------------------------------------------
/**
 * \brief Advance _currentElement to the next element in the tree
 *
 * The tree is traversed in the following order:
 *
 * 1) 1st child, if available
 * 2) Next sibling, if available
 * 3) "Next uncle", if available.  That is, track back through parents looking
 *    for siblings.
 */
void OptionParserXML::_advanceToNextElement()
{
   if (_elementStack.empty())
      return;

   tinyxml::TiXmlElement* el = _elementStack.back();
   tinyxml::TiXmlElement* next = el->FirstChildElement();
   if (next)
      _elementStack.push_back(next);
   else {
      _elementStack.pop_back(); // done with this element and its children
      next = el->NextSiblingElement();
      while (!next && !_elementStack.empty()) {
         tinyxml::TiXmlElement* parent = _elementStack.back();
         _elementStack.pop_back();
         next = parent->NextSiblingElement();
      }
      if (next)
         _elementStack.push_back(next);
   }
}

//------------------------------------------------------------------------------
/// Returns a "scoped" name for the current element prepended to option names
std::string OptionParserXML::_getScopedElementName()
{
   if (!hasNext())
      return "";
   else {
      std::string name;
      for (size_t i = 0; i < _elementStack.size() - 1; ++i) {
         name += _elementStack[i]->Value();
         name += "::";
      }
      name += _elementStack.back()->Value();
      return name;
   }
}

} // namespace Geneva
