//------------------------------------------------------------------------------
/// \file OptionParserYAML.cpp
//
// Author(s):
//    Christian Bauer, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class OptionParserYAML
//------------------------------------------------------------------------------

#include "Geneva/Core/UI/OptionParserYAML.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/Option.hpp"

#include "yaml-cpp/yaml.h"

#include <stdexcept>

namespace Geneva
{

//-----------------------------------------------------------------------
/**
 * \brief Construct the OptionParserYAML
 *
 * This loads and parses the given \a file. After construction, the options can
 * be retrieved via hasNext() and getNext()
 */
OptionParserYAML::OptionParserYAML(const std::string& file)
{
   // load the file
   Geneva_YAML::Node head;
   try {
      head = Geneva_YAML::LoadFile(file);
   } catch(const std::exception& ex) {
      Log::fatal("OptionParserYAML::OptionParserYAML") << "Could not parse the file " << file << "\nPlease check carefully its content (paying particular attention to TABS which are not valid YAML syntax) or validate it with a YAML validator, e.g. http://www.yamllint.com/" << Log::end;
   }

   // make sure that the head option makes sense
   if (head.IsNull())
      Log::error("OptionParserYAML::OptionParserYAML") << "The head node of the file " << file << " is empty." << Log::end;

   // parse the file
   parseOption("", head);
   _currentOption = _optionList.begin();
}

//-----------------------------------------------------------------------
/**
 * \brief Checks if there are more options to retrieve.
 *
 * Uses from the internal list _optionList and iterator _currentOption
 */
bool OptionParserYAML::hasNext() const
{
   return _currentOption != _optionList.end();
}

//-----------------------------------------------------------------------
/**
 * \brief Gets the next option
 *
 * Reads from the internal list _optionList and iterator _currentOption.
 */
void OptionParserYAML::getNext(std::string& name, std::string& val)
{
   if (_currentOption == _optionList.end())
      Log::error("OptionParserYAML::getNext") << "No further option available" << Log::end;
   name = _currentOption->first;
   val = _currentOption->second;
   _currentOption++;
}

//-----------------------------------------------------------------------
/**
 * \brief Gets the next option
 *
 * Reads from the internal list _optionList and iterator _currentOption.
 */
OptionBase* OptionParserYAML::getNext()
{
   std::string name;
   std::string val;
   getNext(name, val);
   return new Option<std::string>(name, val, "");
}

//-----------------------------------------------------------------------
// Private member functions
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
/**
 * \brief Parses one option at a time.
 *
 * This does all the work. If the value is either a scalar or a sequence, the
 * information is stored, otherwise uses recursion to parse the suboptions.
 */
void OptionParserYAML::parseOption(const std::string& name, const Geneva_YAML::Node& node)
{
   if ((name == "extraOptionsOldFormat") && !node.IsSequence())
      Log::fatal("OptionParserYAML::parseOption") << "extraOptionsOldFormat can only be a list. Check your input card!" << Log::end;
   switch (node.Type()) {
      case Geneva_YAML::NodeType::Sequence: {
         if (name == "extraOptionsOldFormat") {
            int i = 0;
            for (Geneva_YAML::const_iterator it = node.begin(); it != node.end(); ++it) {
               std::string val = it->as<std::string>();
               std::stringstream ss;
               ss << "extraOption_" << i++;
               addOption(ss.str(), val);
            }
         } else {
            std::stringstream ss;
            Geneva_YAML::const_iterator it = node.begin();
            ss << it->as<std::string>();
            ++it;
            for (; it != node.end(); ++it) {
               ss << " " << it->as<std::string>();
            }
            addOption(name, ss.str());
         }
         return;
      }
      case Geneva_YAML::NodeType::Scalar:
         // add the option
         addOption(name, node.as<std::string>());
         return;
      case Geneva_YAML::NodeType::Map:
         // recursively process the subobtions
         for (Geneva_YAML::const_iterator it = node.begin(); it != node.end(); ++it) {
            std::stringstream ss;
            ss << name;
            if (name != "")
               ss << "::";
            ss << it->first.as<std::string>();
            parseOption(ss.str(), it->second);
         }
         return;
      case Geneva_YAML::NodeType::Null:
         // end of the option file
         return;
      case Geneva_YAML::NodeType::Undefined:
         Log::fatal("OptionParserYAML::parseOption") << "Found node with undefined type." << Log::end;
   }
}

//-----------------------------------------------------------------------
/**
 * \brief Adds the option with key \a name and value \a value.
 *
 * Appends the option to the list _optionList
 */
void OptionParserYAML::addOption(std::string name, std::string value)
{
   _optionList.push_back(std::pair<std::string, std::string>(name, value));
}

} // namespace Geneva
