//------------------------------------------------------------------------------
/// \file TextColor.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class TextColor
//------------------------------------------------------------------------------

#include "Geneva/Core/UI/TextColor.hpp"

#include <ostream>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief Sets the color of the given output stream.
 * \param out The output stream.
 * \return A reference to \a out.
 *
 * Sets the color of \a out by streaming an appropriate ANSI escape sequence
 * to it. If the TextColor is set to \c TextColor::kOff then \a out is returned
 * untouched.
 */
std::ostream& TextColor::set(std::ostream& out)
{
   if (_code == kOff)
      return out;
   out << "\033[";
   if (_code < kBold)
      out << '0';
   else
      out << '1';
   const int color = _code % kBold;
   if (color > 0)
      out << ";3" << static_cast<char>(color);
   out << 'm';
   return out;
}

//------------------------------------------------------------------------------
/**
 * \brief Resets the color of the given output stream.
 * \param out The output stream.
 * \return A reference to \a out.
 *
 * This resets all previously set color attributes of \a out. If the TextColor
 * is set to either \c TextColor::kReset or \c TextColor::kOff then \a out is
 * returned untouched.
 */
std::ostream& TextColor::reset(std::ostream& out)
{
   if (_code > 0)
      out << "\033[0m";
   return out;
}

} // namespace Geneva
