//------------------------------------------------------------------------------
/// \file Timing.cpp
//
// Author(s):
//    Christian Bauer, Frank Tackmann, Jesse Thaler, Simone Alioli
//
// Copyright:
//    Copyright (C) 2017, 2018 Geneva Collaboration
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class Timing
//------------------------------------------------------------------------------

#include "Geneva/Core/UI/Timing.hpp"

#include <sstream>

namespace Geneva
{

//------------------------------------------------------------------------------
// Starts the timer
/**
 * This should be executed at the beginning of the event loop
 */
void Timing::start()
{
   _isRunning = true;
   _start = std::chrono::system_clock::now();
}

//------------------------------------------------------------------------------
// Stops the timer
/**
 * This should be executed at the end of the event loop
 */
void Timing::stop()
{
   _isRunning = false;
   _stop = std::chrono::system_clock::now();
}

//------------------------------------------------------------------------------
// Returns the elapsed time in seconds
double Timing::timeElapsedInSec() const
{
   if (_isRunning) {
      auto now = std::chrono::system_clock::now();
      return std::chrono::duration<double>(now - _start).count();
   }

   return std::chrono::duration<double>(_stop - _start).count();
}

//------------------------------------------------------------------------------
// Returns the elapsed time in minutes
double Timing::timeElapsedInMin() const
{
   return timeElapsedInSec() / 60.0;
}

//------------------------------------------------------------------------------
// Returns a string containing the elapsed time in seconds
/**
 * This will be formatted nicely and present the time in seconds, minutes, hours or day,
 * whatever is most appropriate
 */
std::string Timing::timeElapsedString(std::string prefix) const
{
   std::stringstream timingMessage("");
   timingMessage << "Total time elapsed: " << timeElapsedInSec() << " sec";
   return timingMessage.str();
}

//------------------------------------------------------------------------------
// Returns the total number of events
size_t Timing::targetValue() const
{
   return _targetValue;
}

//------------------------------------------------------------------------------
// Returns the current number of events
size_t Timing::currentValue() const
{
   return _currentValue;
}

//------------------------------------------------------------------------------
// Sets the total number of events to be generated
/**
 * This should be done before the start of the event loop
 */
void Timing::setTargetValue(size_t newTarget)
{
   _targetValue = newTarget;
}

//------------------------------------------------------------------------------
// Sets the current number of events
/**
 * This should be executed every time a new event is generated
 */
void Timing::setCurrentValue(size_t newCurrent)
{
   _currentValue = newCurrent;
}

//------------------------------------------------------------------------------
// Returns a progress string that can be output to the screen
/**
 * Gives a nicely formated string that contains the current number of events generated,
 * the current time it has taken, the total number of events, with the expected total time
 * and a progress bar that moves from left to right.
 * Note that this progress string should always overwright itself, such that one should not
 * issue an endl after outputting this progress string.
 */
std::string Timing::progressString()
{
   auto now = std::chrono::system_clock::now();
   size_t timeSoFar = static_cast<size_t>(std::chrono::duration<double>(now - _start).count());

   size_t estimatedTotalTime = timeSoFar * targetValue();
   if (currentValue() > 0)
      estimatedTotalTime /= currentValue();
   size_t currentPosition = 63;
   if (currentValue() < targetValue()) {
      currentPosition *= currentValue();
      currentPosition /= targetValue();
   }
   std::string eventString = eventtime2string(currentValue(), timeSoFar);
   std::string timeString = eventtime2string(targetValue() - currentValue(), estimatedTotalTime - timeSoFar);
   std::string summaryString = eventtime2string(targetValue(), estimatedTotalTime);

   std::string displayString = "|"; // 1 char
   for (size_t i = 0; i < 63; i++) { // + 63 chars
      if (i <  currentPosition)
         displayString += "-";
      if (i == currentPosition)
         displayString += ">";
      if (i >  currentPosition)
         displayString += " ";
   }
   displayString += "|" ; // + 1 char = 65 chars
   displayString.replace(7, eventString.size(), eventString);
   displayString.replace(59 - timeString.size(), timeString.size(), timeString);

   //            65 chars         + 1 char + (<=13 chars) <= 79
   return displayString + ' ' + summaryString + '\r';
}

//------------------------------------------------------------------------------
// Helper function that converts number of events and seconds into a nicely formated string
/**
 * Takes the number of events and seconds and presents them in a nicely formated string.
 * It figures out if time should be presented in seconds, minutes, hours, etc and if the
 * number of events should be presented in thousands, millions etc.
 *
 * For fewer than a trillion events and a time less than 1000 weeks, returns a
 * string between 4 and 13 characters long.
 */
std::string Timing::eventtime2string(size_t events, size_t secs) const
{
   std::string eventString = "";
   std::string timeString = "";

   // min characters: 1; max characters: 6
   if (events >= 1 * 1000000000) {
      eventString = int2string(events/1000000000)  + "." + int2string((events % 1000000000) / (1000000000/10)) + "B";
   } else if (events >= 1 * 1000000) {
      eventString = int2string(events/1000000) + "." + int2string((events % 1000000)/(1000000/10)) + "M";
   } else if (events >= 1 * 1000) {
      eventString = int2string(events/1000) + "K";
   } else {
      eventString = int2string(events);
   }

   // min characters: 2; max characters: 6
   if (secs >= 1 * 604800) { // 1 week
      timeString = int2string(secs/604800) + "w" + int2string((secs % 604800)/86400) + "d";
   } else if (secs >= 1 * 86400) { // 1 day
      timeString = int2string(secs/86400) + "d" + int2string((secs % 86400)/3600) + "h";
   } else if (secs >= 1 * 3600) { // 1 hour
      timeString = int2string(secs/3600) + "h" + int2string((secs % 3600)/60) + "m";
   } else if (secs >= 1 * 60) { // 1 minute
      timeString = int2string(secs/60) + "m" + int2string(secs % 60) + "s";
   } else {
      timeString = int2string(secs) + "s";
   }

   // min characters: 4; max characters: 13
   return (eventString + "/" + timeString);
}

//------------------------------------------------------------------------------
// Helper function that converts an integer to a string
std::string Timing::int2string(size_t n) const
{
   std::ostringstream stream("");
   stream << n;
   return stream.str();
}

} //namespace Geneva
