//------------------------------------------------------------------------------
/// \file FPEUtils.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Utilities to handle floating point exceptions
//------------------------------------------------------------------------------

#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/FPEUtils.hpp"

#ifdef geneva_USE_FPE_TRAPPING
#define _GNU_SOURCE 1
#include <fenv.h>
#endif

namespace Geneva
{

//------------------------------------------------------------------------------
// TrapFPE
//------------------------------------------------------------------------------

#ifdef geneva_USE_FPE_TRAPPING

//------------------------------------------------------------------------------
TrapFPE::TrapFPE(int flags)
{
   _oldFlags = feenableexcept(flags);
   if (_oldFlags == -1)
      Log::warning("TrapFPE") << "Failed to enable floating point exceptions with bitmask " << flags << Log::end;
   else
      Log::debug("TrapFPE") << "Enabled floating point exceptions with bitmask " << flags << Log::end;
}

//------------------------------------------------------------------------------
TrapFPE::~TrapFPE()
{
   if (_oldFlags != -1) {
      if (feenableexcept(_oldFlags) == -1)
         Log::warning("~TrapFPE") << "Failed to restore floating point exceptions with previous bitmask " << _oldFlags << Log::end;
      else
         Log::debug("~TrapFPE") << "Restored previous floating point exceptions with bitmask " << _oldFlags << Log::end;
   }
}

#endif // geneva_USE_FPE_TRAPPING

//------------------------------------------------------------------------------
// CatchFPE
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
CatchFPE::CatchFPE(int flags)
   : _flags(flags)
{
   feholdexcept(&_env);
}

//------------------------------------------------------------------------------
CatchFPE::~CatchFPE()
{
   clear();
   feupdateenv(&_env);
}

//------------------------------------------------------------------------------
int CatchFPE::test()
{
   return fetestexcept(_flags);
}

//------------------------------------------------------------------------------
void CatchFPE::clear()
{
   feclearexcept(_flags);
}

//------------------------------------------------------------------------------
void CatchFPE::release()
{
   const int raised = test();
   if (raised) {
      clear();
      std::string what = "Floating point exception";
      if (raised & FE_INVALID)
         what += " (invalid operation)";
      if (raised & FE_DIVBYZERO)
         what += " (division by zero)";
      if (raised & FE_OVERFLOW)
         what += " (overflow)";
      if (raised & FE_UNDERFLOW)
         what += " (underflow)";
      if (raised & FE_INEXACT)
         what += " (inexact result)";
      throw floating_point_error(what);
   }
}

} // namespace Geneva
