//------------------------------------------------------------------------------
/// \file Random.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 MIT
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Random number generator implementation
//    This is just the implementation of rand2 from numerical recipes
//------------------------------------------------------------------------------

#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/Random.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <sstream>
#include <fstream>

namespace Geneva
{

#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

static long idum2 = 123456789;
static long iy = 0;
static long iv[NTAB];

long testSeed = -1;

static RandomState storedState; // stored state

//------------------------------------------------------------------------------
void setRandomSeed(long newSeed)
{
   if (newSeed > 0) testSeed = -newSeed;
   else if (newSeed == 0) testSeed = -1;
   else testSeed = newSeed;
}

//------------------------------------------------------------------------------
double randomAngle()
{
   return math_constants::_2pi() * randomInterval();
}

//------------------------------------------------------------------------------
double randomInterval()
{
   return Ran2(&testSeed);
}

//------------------------------------------------------------------------------
int randomInteger(int N)
{
   return static_cast<int>(randomInterval()*N);
}

//------------------------------------------------------------------------------
size_t randomInteger(size_t N)
{
   return static_cast<size_t>(randomInterval()*static_cast<double>(N));
}

//------------------------------------------------------------------------------
double Ran2(long* idum)
{
   int   j;
   long  k;
   double temp;

   if (*idum <= 0) {
      if (-(*idum) < 1) *idum=1;
      else *idum = -(*idum);
      idum2=(*idum);
      for (j=NTAB+7; j>=0; j--) {
         k=(*idum)/IQ1;
         *idum=IA1*(*idum-k*IQ1)-k*IR1;
         if (*idum < 0) *idum += IM1;
         if (j < NTAB) iv[j] = *idum;
      }
      iy=iv[0];
   }
   k=(*idum)/IQ1;
   *idum=IA1*(*idum-k*IQ1)-k*IR1;
   if (*idum < 0) *idum += IM1;
   k=idum2/IQ2;
   idum2=IA2*(idum2-k*IQ2)-k*IR2;
   if (idum2 < 0) idum2 += IM2;
   j=static_cast<int>(iy/NDIV); // TODO: Is implicit loss of precision here OK?!?
   iy=iv[j]-idum2;
   iv[j] = *idum;
   if (iy < 1) iy += IMM1;
   if ((temp =AM*static_cast<double>(iy)) > RNMX)
      return RNMX;
   else
      return temp;
}

//------------------------------------------------------------------------------
void storeRandomState()
{
   // store the state into a RandomState struct
   storedState.kseed = testSeed;
   storedState.kidum2 = idum2;
   storedState.kiy = iy;
   for (int c = 0; c < NTAB; c++) {
      storedState.kiv[c] = iv[c];
   }
}

//------------------------------------------------------------------------------
RandomState getRandomState()
{
   RandomState state;
   // store the state into a RandomState struct
   state.kseed = testSeed;
   state.kidum2 = idum2;
   state.kiy = iy;
   for (int c = 0; c < NTAB; c++) {
      state.kiv[c] = iv[c];
   }
   return state;
}

//------------------------------------------------------------------------------
bool equalToStoredState(RandomState state)
{
   // check if anything has changed compared to the stored state
   if (testSeed != state.kseed)
      return false;
   if (idum2 != state.kidum2)
      return false;
   if (iy != state.kiy)
      return false;
   for (int c = 0; c < NTAB; c++) {
      if (iv[c] != state.kiv[c])
         return false;
   }
   return true;
}

//------------------------------------------------------------------------------
void initializeRandomState(RandomState state)
{
   testSeed = state.kseed;
   idum2 = state.kidum2;
   iy = state.kiy;
   for (int c = 0; c < NTAB; c++) {
      iv[c] = state.kiv[c];
   }
}

//------------------------------------------------------------------------------
void saveRandomState(std::string filename)
{
   // put the state into a string to store
   std::stringstream statestream("");

   statestream << storedState.kseed << "\n";
   statestream << storedState.kidum2 << "\n";
   statestream << storedState.kiy << "\n";
   for (int c = 0; c < NTAB; c++) {
      statestream << storedState.kiv[c] << "\n";
   }

   std::ofstream statefile(filename.c_str());

   // write the state out
   statefile << statestream.str();
   statefile.close();
   Log::info("Random") << "Saved state of RNG to " << filename << Log::end;
}

//------------------------------------------------------------------------------
void initializeFromStatefile(std::string filename)
{
   std::ifstream statefile(filename.c_str());
   if (statefile) {
      statefile >> storedState.kseed;
      statefile >> storedState.kidum2;
      statefile >> storedState.kiy;
      for (int c = 0; c < NTAB; c++) {
         statefile >> storedState.kiv[c];
      }
      // now initialize the state
      testSeed = storedState.kseed;
      idum2 = storedState.kidum2;
      iy = storedState.kiy;
      for (int c = 0; c < NTAB; c++) {
         iv[c] = storedState.kiv[c];
      }
      Log::verboseInfo("Random") << "Initialized RNG state from " << filename << Log::end;
   } else {
      Log::fatal("Random") << "Unable to read state from " << filename << Log::end;
   }
}

#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX

} // namespace Geneva
