//------------------------------------------------------------------------------
/// \file EventGeneratorAdaptive.cpp
//
// Author(s):
//    Christian Bauer, Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2016 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class EventGeneratorAdaptive
//------------------------------------------------------------------------------

#include "Geneva/Driver/EventGeneratorAdaptive.hpp"

#include "Geneva/Generators/IntegrationChannelManager.hpp"

#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/Utils/Random.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"
#include "Geneva/Core/Utils/FileUtils.hpp"

#include <fstream>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief Construct the Event generator
 *
 * \param gen The Generator used as the phase space map
 * \param calc The Calculation performed
 * \param optMap  The option map
 *
 * This sets up an EventGeneratorAdaptive. A separate instance of a SmartRandom is instantiated
 * for each IntegrationChannel required by the calculation, such that separate importance
 * sampling can be performed.
 */
//------------------------------------------------------------------------------
EventGeneratorAdaptive::EventGeneratorAdaptive(IGenerator* gen, ICalculation* calc, const OptionMap& optMap)
   : _gen(gen), _calc(calc), _optMap(optMap), _calcFailures(0), _lastChannel(0),
     _stem(SmartRandomHandler::getGridStem(optMap)), _seedTag(format_tag(optMap.getOption("seed")))
{
   _maxCalcFailures = size_t_from_int(optMap.getOption<int>("EventGenerator::maxCalcFailures"));
   _maxGenFailures = size_t_from_int(optMap.getOption<int>("EventGenerator::maxGenFailures"));

   _numBins = size_t_from_int(optMap.getOption<int>("Mint::numBins"));
   _numIterations = size_t_from_int(optMap.getOption<int>("Mint::iterationsGrid"));

   // Get the channels for the required signatures.
   _sigSet = _gen->getSignatures();
   if (_sigSet.empty())
      Log::error("EventGeneratorAdaptive::EventGeneratorAdaptive") << "The SignatureSet is empty." << Log::end;
   createChannels(_sigSet);
   // Set the weights for each channel to be equal.
   initializeChannelWeights();

   // Create SmartRandoms for all channels.
   for (ChannelIterator channel = _channels.begin(); channel != _channels.end(); ++channel) {
      size_t dimension = _gen->numberOfRandomsNeeded(*(channel->signatures().begin()));
      _smartRandomHandler.makeSmartRandom(SmartRandomHandler::kMint, *channel, dimension, _numBins);
   }

   setupOptimization();

   // Re-initialize the calculation after optimization has been set up.
   _calc->initialize();
   _calcFailures = 0;
}

//------------------------------------------------------------------------------
/**
 * \brief Create one event
 *
 * This function creates a single weighted event
 *
 * The steps are
 * 1) Ask SmartRandom for a set of random numbers
 * 2) Generate the configuration
 * 3) Calculate the cross-section
 */
const Event& EventGeneratorAdaptive::next()
{
   const std::pair<ChannelIterator, double> channelWeightPair = pickChannel();
   _lastChannel = &*channelWeightPair.first;
   next(channelWeightPair.first, channelWeightPair.second);
   return _event;
}

//------------------------------------------------------------------------------
void EventGeneratorAdaptive::finalize(const size_t numEvents)
{
   _gen->finalize();
   _calc->finalize(numEvents);
}

//------------------------------------------------------------------------------
// Protected and private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * Tries to generate a valid configuration. If an exception is thrown during
 * generation it is caught, and we try again. If more than _maxGenFailures
 * occur, a fatal error is raised.
 * If the phase space generator returns false (i.e. because of generation cuts),
 * returns a negative generation weight. In this case, an empty event should be
 * handed out.
 */
double EventGeneratorAdaptive::generateConfig(ChannelIterator channel, double channelWeight)
{
   size_t genFailures = 0;
   while (true) {
      try {
         double weight = channelWeight;
         // Pick a signature
         const std::pair<Signature, double> signatureWeightPair = channel->pickSignature();
         _signature = signatureWeightPair.first;
         weight *= signatureWeightPair.second;

         // Pick an appropriate number of random numbers
         std::vector<double> rand(_gen->numberOfRandomsNeeded(_signature), std::numeric_limits< double >::quiet_NaN());
         double prob = std::numeric_limits<double>::quiet_NaN();
         _smartRandomHandler.getRandom(*channel, rand, prob);
         weight *= prob;

         // Generate the configuration
         if (!_gen->generate(_signature, rand, _event.config()))
            return -1.;

         const double genWeight = _gen->getLastWeight();
         if (genWeight == 0.)
            Log::error("EventGeneratorAdaptive::generateConfig") << "The probability to generate an event should not be 0." << Log::end;

         weight *= genWeight;
         return weight;
      } catch (const NormalError&) {
         if (++genFailures <= _maxGenFailures)
            Log::warning("EventGeneratorAdaptive::generateConfig")
               << "Caught error from phase-space generator. Continuing with new phase-space point." << Log::end;
         else
            Log::fatal("EventGeneratorAdaptive::generateConfig")
               << "Phase-space generator failed more than " << _maxGenFailures << " times." << Log::end;

         // SmartRandoms need to be informed of failure during setup stage in order to provide a new random
         _smartRandomHandler.informOfFailure(*channel);
      }
   }
}

//------------------------------------------------------------------------------
std::pair<EventGeneratorAdaptive::ChannelIterator, double> EventGeneratorAdaptive::pickChannel()
{
   double jac;
   ChannelIterator channel = _channels.begin();
   if (_channels.size() == 1) {
      jac = 1.;
   } else {
      const double rand = randomInterval() * _cumulativeWeights.back();
      size_t count = 0;
      while (rand > _cumulativeWeights[count]) {
         ++channel;
         ++count;
         if (channel == _channels.end()) {
            Log::error("EventGeneratorAdaptive::pickChannel") << "Should not reach the end of the Channel vector" << Log::end;
         }
      }
      if (count == 0)
         jac = _cumulativeWeights[0] / _cumulativeWeights.back();
      else
         jac = (_cumulativeWeights[count] - _cumulativeWeights[count - 1]) / _cumulativeWeights.back();
   }
   if (jac < 0)
      Log::error("EventGeneratorAdaptive::pickChannel") << "The probability for Channel " << channel->name() << " is negative" << Log::end;
   return std::make_pair(channel, jac);
}

//------------------------------------------------------------------------------
void EventGeneratorAdaptive::next(ChannelIterator channel, double channelWeight)
{
   while (true) {
      try {
         const double genWeight = generateConfig(channel, channelWeight);
         if (genWeight > 0.) {
            // We have a valid configuration. Compute the cross section for it:
            // It could be zero if it fails the Calculation's Restriction, or it could
            // fail if an error occurs internally. In the second case the calculation
            // should throw an exception.
            _calc->informGenerationWeight(genWeight);
            const Sigma sigma = (*_calc)(_event.config());
            if (sigma.isDefined()) {
               _event.setSigma(sigma / genWeight);
               _totalRate.addWeight(_event.weightedSigma());
               return;
            }
         }
         // hand out empty event if generation or calculation failed nonexceptionally
         _event.clearSigma();
         _totalRate.addWeight();
         return;
      } catch (const NormalError&) {
         if (++_calcFailures <= _maxCalcFailures)
            Log::warning("EventGeneratorAdaptive::next")
               << "Caught error in calculation. Continuing with new phase-space point." << Log::end;
         else
            Log::fatal("EventGeneratorAdaptive::next")
               << "Calculation failed to evaluate more than " << _maxCalcFailures << " times." << Log::end;

         // SmartRandoms need to be informed of failure during setup stage in order to provide a new random
         _smartRandomHandler.informOfFailure(*channel);
      }
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Set up the optimization
 *
 * This function sets up the optimization used for importance sampling. It uses %SmartRandom.
 * There is a separate %SmartRandom instance for each IntegrationChannel.
 *
 * First, it tries to load the Grid information from existing text files
 *
 * Second, it uses a given number of iterations to improve the grids. The steps are:
 *
 * 1) Ask SmartRandom for a set of random numbers
 * 2) Use the PSGenerator to construct a Configuration for these random numbers
 * 3) Use the Calculation to return the differential cross section
 * 4) Pass the resulting differential cross section back to SmartRandom for importance sampling
 *
 * After each iteration, the grid is adjusted.
 *
 * At the end of all iterations, each SmartRandom is finalized.
 *
 */
void EventGeneratorAdaptive::setupOptimization()
{
   bool overWriteGrid = _optMap.getOption<bool>("Mint::overWriteGrid");
   bool refineGrid = _optMap.getOption<bool>("Mint::refineGrid");

   //Initialize the optimization, and try to load the information from the Grid files
   _smartRandomHandler.initializeOptimization();
   if (!overWriteGrid) {
      bool optimizationLoaded = _smartRandomHandler.loadOptimization(_stem);
      bool sigProbsLoaded = loadSignatureProbabilities();
      bool wasLoaded = optimizationLoaded && sigProbsLoaded;
      if (sigProbsLoaded)
         updateChannelWeights();
      if (!refineGrid) {
         if (!wasLoaded)
            Log::error("EventGeneratorAdaptive::setupOptimization") << "A valid grid is needed when refineGrid is false." << Log::end;
         _smartRandomHandler.finalizeOptimization();
         return;
      }
   }

   // we need to recreate or refine the grids
   size_t totPoints = size_t_from_int(_optMap.getOption<int>("Mint::maxPointsPerIterGrid"));
   size_t minNeeded = std::max(static_cast<size_t>(std::ceil(static_cast<double>(totPoints) * _optMap.getOption<double>("Mint::minFractionPerChannel"))), static_cast<size_t>(1));

   if (_numIterations == 0)
      Log::error("EventGeneratorAdaptive::setupOptimization") << "Cannot set iterations = 0 when creating the grids." << Log::end;

   for (size_t iteration = 1; iteration <= _numIterations; iteration++) {
      Log::info() << "-----------------------------------------------" << Log::end;
      Log::info() << "Iteration " << iteration << " of " << _numIterations << Log::end;
      Log::info() << "-----------------------------------------------" << Log::end;

      // loop over channels
      for (std::vector<IntegrationChannel>::iterator it = _channels.begin(); it != _channels.end(); ++it) {
         // If the total weight in a given channel is zero, go to the next channel.
         if (it->totalWeight() == 0.) {
            Log::warning("EventGeneratorAdaptive::setupOptimization") << "channel " << it->name() << " has zero total weight. Skipping it." << Log::end;
            continue;
         }

         // Determine the number of points according to the relative channel probability.
         // For the first iteration, this effectively uses the same number of points for each channel.
         size_t numUsed = static_cast<size_t>(std::ceil(static_cast<double>(totPoints) * it->totalWeight() / _cumulativeWeights.back()));
         if (numUsed < minNeeded) {
            Log::info() << "Requested number of " << numUsed << " points is less than minimum number of " << minNeeded << Log::end;
            numUsed = minNeeded;
         }
         Log::info() << "Using " << numUsed << " points for channel " << *it << Log::end;
         size_t writeout = numUsed / 10;
         if (writeout == 0)
            ++writeout;

         // generate numUsed points
         for (size_t num = 1; num <= numUsed; num++) {
            if (num % writeout == 0)
               Log::verboseInfo() << num << " points done" << Log::end;

            next(it, 1.);

            double result = 0.;
            if (_event.weightedSigma().isDefined())
               result = _event.weightedSigma().central();
            _smartRandomHandler.informOfCalculation(*it, result);
            it->addWeight(_signature, result);
         }
      }

      // at end of iteration, adjust and store the grid and update channel weights
      _smartRandomHandler.storeOptimization(_stem, _seedTag);
      _smartRandomHandler.optimize();
      for (std::vector<IntegrationChannel>::iterator it = _channels.begin(); it != _channels.end(); ++it) {
         writeCrossSection(it);
      }
      updateChannelWeights();
   }

   _smartRandomHandler.finalizeOptimization();
}

//------------------------------------------------------------------------------
//This divides the whole SignatureSet into channels, which each have their own SmartRandom.
void EventGeneratorAdaptive::createChannels(const SignatureSet& sigSet)
{
   assert(_channels.empty());
   _channels = IntegrationChannelManager().getChannels(sigSet, _optMap);
   assert(not _channels.empty());
}

//------------------------------------------------------------------------------
void EventGeneratorAdaptive::writeCrossSection(std::vector<IntegrationChannel>::iterator it)
{
   const std::string fileName = _stem + format_tag(it->name()) + _seedTag + ".dat";
   std::ofstream gridFile(fileName.c_str(), std::ofstream::app);
   gridFile << "###" << std::endl;
   SignatureSet sigs = it->signatures();

   for (SignatureSet::iterator sig = sigs.begin(); sig != sigs.end(); sig++) {
      gridFile << *sig << '\n' << it->crossSection(*sig) << std::endl;
   }
   gridFile.close();
   if (!gridFile)
      Log::fatal("EventGeneratorAdaptive::writeCrossSection")
         << "Could not write to grid file: " << fileName << Log::end;
}

//------------------------------------------------------------------------------
// This loads the information of the the cross sections that are stored inside the grid files into the IntegrationChannels
bool EventGeneratorAdaptive::loadSignatureProbabilities()
{
   for (std::vector<IntegrationChannel>::iterator it = _channels.begin(); it != _channels.end(); ++it) {
      const std::string fileName = _stem + format_tag(it->name()) + ".dat";
      std::ifstream gridFile(fileName.c_str());
      if (not gridFile.good())
         return false;
      std::string line;
      while (getline(gridFile, line)) {
         if (line == "###")
            break;
      }
      size_t sigCount = it->signatures().size();
      while (getline(gridFile, line)) {
         Signature sig;
         std::stringstream ss;
         ss << line;
         int idcode;
         ss >> idcode;
         sig.setInitialA(PdgID(idcode));
         ss >> idcode;
         sig.setInitialB(PdgID(idcode));
         while (ss >> idcode) {
            sig.insertFinal(PdgID(idcode));
         }
         if (getline(gridFile, line)) {
            std::stringstream ss2;
            ss2 << line;
            double result;
            ss2 >> result;
            it->setWeight(sig, result);
            it->setNumEvents(1);
         } else {
            Log::error("EventGeneratorAdaptive::loadSignatureProbabilities")
               << "Missing cross section information for Signature " << sig
               << "in grid file "<< fileName << Log::end;
         }
         sigCount--;
      }
      if (not gridFile.fail())
        return false;
      if (sigCount > 0)
         Log::error("EventGeneratorAdaptive::loadSignatureProbabilities") << "Not enough Signatures found in grid file " << fileName << Log::end;
      gridFile.close();
   }
   return true;
}

//------------------------------------------------------------------------------
void EventGeneratorAdaptive::initializeChannelWeights()
{
   _cumulativeWeights.clear();
   for (std::vector<IntegrationChannel>::iterator it = _channels.begin(); it != _channels.end(); ++it) {
      it->initialize();
      if (it == _channels.begin())
         _cumulativeWeights.push_back(1.);
      else
         _cumulativeWeights.push_back(_cumulativeWeights.back() + 1.);
   }
}

//------------------------------------------------------------------------------
void EventGeneratorAdaptive::updateChannelWeights()
{
   _cumulativeWeights.clear();
   for (std::vector<IntegrationChannel>::iterator it = _channels.begin(); it != _channels.end(); ++it) {
      it->reset();
      if (it == _channels.begin())
         _cumulativeWeights.push_back(it->totalWeight());
      else
         _cumulativeWeights.push_back(_cumulativeWeights.back() + it->totalWeight());
   }
}

} // namespace Geneva
