//------------------------------------------------------------------------------
/// \file EventGeneratorAdaptivePartialUnweighted.cpp
//
// Author(s):
//    Christian Bauer, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2016 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class EventGeneratorAdaptiveUnweighted
//------------------------------------------------------------------------------

#include "Geneva/Driver/EventGeneratorAdaptivePartialUnweighted.hpp"

#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/Utils/Random.hpp"

#include <vector>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief Constructs the event generator
 *
 * \param gen The Generator used as the phase space map
 * \param calc The Calculation performed
 * \param optMap The option Map
 */
EventGeneratorAdaptivePartialUnweighted::EventGeneratorAdaptivePartialUnweighted(IGenerator* gen, ICalculation* calc, const OptionMap& optMap)
   : EventGeneratorAdaptive(gen, calc, optMap)
{
   double targetMultiplier = optMap.getOption<double>("EventGenerator::AdaptivePartialUnweighted::targetMultiplier");
   double targetWeight = optMap.getOption<double>("EventGenerator::AdaptivePartialUnweighted::targetWeight");

   // Keeps the ability to set an absolute value of the target weight for backward compatibility.
   if (targetWeight > 0.) {
      // Check that we are not setting both targetWeight and targetMultiplier.
      if (targetMultiplier > 0.)
         Log::error("EventGeneratorAdaptivePartialUnweighted")
            << "Only one of targetWeight or targetMultiplier in EventGenerator::AdaptivePartialUnweighted should be greater than zero." << Log::end;
      _targetWeight = targetWeight;
   } else {
      const double precision = optMap.getOption<double>("EventGenerator::AdaptivePartialUnweighted::estimatePrecision");
      _targetWeight = targetMultiplier * getCrossSectionEstimate(precision);
   }
   if (_targetWeight == 0.)
      Log::warning("EventGeneratorAdaptivePartialUnweighted") << "Zero target weight. No unweighting will take place."<<Log::end;
   else
      Log::info("EventGeneratorAdaptivePartialUnweighted") << "Partially unweighting events to the target weight = " << _targetWeight << Log::end;
}

//------------------------------------------------------------------------------
/**
 * \brief Creates one partially unweighted event.
 *
 * The steps are
 * 1) Generate one event using nextCentral
 * 2) Reweight the event if we are exceeding the unweighting target
 * 3) If the event is kept evaluate the full calculation if needed
 */
void EventGeneratorAdaptivePartialUnweighted::next(ChannelIterator channel, double channelWeight)
{
   while (true) {
      try {
         // Generate an event with only the central weight.
         _event.clearSigma();
         double genWeight = generateConfig(channel, channelWeight);
         if (genWeight < 0.) {
            _totalRate.addWeight();
            return;
         }
         _calc->informGenerationWeight(genWeight);
         const RandomState randState = getRandomState();
         Sigma sigma = (*_calc)(_event.config(), false);
         if (!sigma.isDefined()) {
            _totalRate.addWeight();
            return;
         }
         const double weight = sigma.central() / genWeight;
         _totalRate.addWeight(weight);

         // Reweight, throw away event if weight < random * targetweight
         const double weightAbs = abs(weight);
         if (weightAbs < _targetWeight) {
            if (weightAbs <= randomInterval() * _targetWeight)
               return;
            // rescale the generation weight by the reweightFactor
            genWeight *= weightAbs / _targetWeight;
         }

         // We keep the event, evaluate the full calculation if it has variations
         if (_calc->getNumVariations() != 0) {
            initializeRandomState(randState);
            sigma = (*_calc)(_event.config(), true);
            // ensures that random sequence is independent of variations
            if (weightAbs < _targetWeight)
               randomInterval();
         }
         if (sigma.isDefined())
            _event.setSigma(sigma / genWeight);
         return;
      } catch (const NormalError&) {
         if (++_calcFailures <= _maxCalcFailures)
            Log::warning("EventGeneratorAdaptivePartialUnweighted::next")
               << "Caught error in calculation. Continuing with new phase-space point." << Log::end;
         else
            Log::fatal("EventGeneratorAdaptivePartialUnweighted::next")
               << "Calculation failed to evaluate more than " << _maxCalcFailures << " times." << Log::end;
      }
   }
}

//------------------------------------------------------------------------------
// Obtains an estimate of the total cross section.
double EventGeneratorAdaptivePartialUnweighted::getCrossSectionEstimate(double precision)
{
   RandomState randState = getRandomState();
   size_t failures = 0;
   Rate estimate;
   while (true) {
      const std::pair<ChannelIterator, double> channelAndWeight = pickChannel();
      try {
         const double genWeight = generateConfig(channelAndWeight.first, channelAndWeight.second);
         if (genWeight > 0.)
            estimate.addWeight(_calc->fastEstimate(_event.config()) / genWeight);
         else
            estimate.addWeight();
         if (estimate.totEntries() % 100 == 0) {
            if (estimate.error().central() < precision * abs(estimate.value().central()))
               break;
         }
      } catch (const NormalError&) {
         if (++failures > _maxCalcFailures)
            Log::fatal("EventGeneratorAdaptivePartialUnweighted::getCrossSectionEstimate")
               << "Calculation estimate failed to evaluate more than " << _maxCalcFailures << " times." << Log::end;
      }
   }
   // reset the random number seed
   initializeRandomState(randState);
   return estimate.value().central();
}

} // namespace Geneva
