//------------------------------------------------------------------------------
/// \file EventGeneratorWeighted.cpp
//
// Author(s):
//    Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class EventGeneratorWeighted
//------------------------------------------------------------------------------

#include "Geneva/Driver/EventGeneratorWeighted.hpp"

#include "Geneva/Generators/IGenerator.hpp"

#include "Geneva/Core/Physics/Configuration.hpp"
#include "Geneva/Core/UI/Log.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
const Event& EventGeneratorWeighted::next()
{
   Configuration& config = _event.config();
   Sigma sigma;
   while (true) {
      size_t genFailures = 0;
      // try to generate a valid configuration, if an exception is thrown
      // during generation increase a counter and start over. If too much
      // failures occur, kill the program.
      // if the generator is returning false (i.e. because of generation cuts),
      // just hand out an empty event
      while (true) {
         try {
            if (!_generator->generate(config)) {
               _event.clearSigma();
               _totalRate.addWeight();
               return _event;
            }
         } catch (const NormalError&) {
            if (++genFailures <= _maxGenFailures)
               Log::warning("EventGeneratorWeighted::next")
                  << "Caught error from phase-space generator. Continuing with new phase-space point." << Log::end;
            else
               Log::fatal("EventGeneratorWeighted::next")
                  << "Phase-space generator failed more than " << _maxGenFailures << " times." << Log::end;
            continue;
         }
         break;
      }
      // Now we have a valid configuration. Compute the cross section for it:
      // It could be zero if it fails the Calculation's Restriction, or it could
      // fail if an error occurs internally. In the second case the calculation
      // should throw an exception.
      try {
         _calculation->informGenerationWeight(_generator->getLastWeight());
         sigma = (*_calculation)(config);
      } catch (const NormalError&) {
         if (++_calcFailures <= _maxCalcFailures)
            Log::warning("EventGeneratorWeighted::next")
               << "Caught error in calculation. Continuing with new phase-space point." << Log::end;
         else
            Log::fatal("EventGeneratorWeighted::next")
               << "Calculation failed to evaluate more than " << _maxCalcFailures << " times." << Log::end;
         continue;
      }
      break;
   }
   if (sigma.isDefined()) {
      _event.setSigma(sigma / _generator->getLastWeight());
   } else {
      _event.clearSigma();
   }
   _totalRate.addWeight(_event.weightedSigma());
   return _event;
}

} // namespace Geneva
