//------------------------------------------------------------------------------
/// \file Geneva.cpp
//
// Author(s):
//    Chris Vermilion, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 LBNL, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class Geneva
//------------------------------------------------------------------------------

#include "Geneva/version.hpp"

#include "Geneva/Driver/Geneva.hpp"
#include "Geneva/Driver/YamlToXmlTranslation.hpp"
#include "Geneva/Driver/EventGeneratorWeighted.hpp"
#include "Geneva/Driver/EventGeneratorAdaptive.hpp"
#include "Geneva/Driver/EventGeneratorAdaptivePartialUnweighted.hpp"

#include "Geneva/Calculations/ICalculation.hpp"
#include "Geneva/Calculations/CalculationFactory.hpp"

#include "Geneva/Generators/IGenerator.hpp"
#include "Geneva/Generators/GeneratorFactory.hpp"

#include "Geneva/Core/UI/OptionParserYAML.hpp"
#include "Geneva/Core/UI/OptionParserXML.hpp"
#include "Geneva/Core/UI/OptionParserCL.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/Exception.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"
#include "Geneva/Core/Utils/Random.hpp"

#include <string>
#include <ctime>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * \brief Constructor using only default options.
 *
 * Only loads default options from DefaultOptions.xml.
 */
Geneva::Geneva()
   : _evtGen(0), _isInitialized(false)
{
   construct();
}

//------------------------------------------------------------------------------
/**
 * \brief Constructor using command line options.
 *
 * This loads options in the following order:
 * - default options from DefaultOptions.xml
 * - options from a file specified on command line via '--optionsFile filename'
 * - options specified on the command line
 */
Geneva::Geneva(int argc, const char* argv[])
   : _evtGen(0), _isInitialized(false)
{
   std::stringstream stream;
   for (int i = 1; i < argc; i++) {
      stream << argv[i] << ' ';
   }
   construct(stream.str());
}

//------------------------------------------------------------------------------
/**
 * \brief Constructor using a string of options.
 *
 * This loads options in the following order:
 * - default options from DefaultOptions.xml
 * - options from a file specified in the \a options string via '--optionsFile filename'
 * - options specified by \a options.
 */
Geneva::Geneva(const std::string& options)
   : _evtGen(0), _isInitialized(false)
{
   construct(options);
}

//------------------------------------------------------------------------------
Geneva::~Geneva()
{
   delete _evtGen;
   Log::info() << footer() << Log::end;
}

//------------------------------------------------------------------------------
/// Initialize the event generator.
void Geneva::initialize()
{
   // set up Calculation
   ICalculation* calc = CalculationFactory::instance().getCalculation(_options);
   if (!calc) {
      Log::error("Geneva::init") << "No calculation returned." << Log::end;
   }

   // set up Generator
   IGenerator* generator = GeneratorFactory::instance().getGenerator(_options);
   if (!generator) {
      Log::error("Geneva::init") << "No generator returned." << Log::end;
   }

   // use all Signatures requested by the Calculation
   SignatureSet sigSet = calc->neededSignatures();
   // unless explicitly asking for only a subset
   const size_t onlySize = size_t_from_int(_options.getOption<int>("EventGenerator::OnlySignatureSize"));
   if (onlySize != 0) {
      Log::warning("Geneva::init") << "Only selecting signatures of size " << onlySize << Log::end;
      size_t erased = sigSet.erase_if(std::not1(CheckSignatureSize(onlySize)));
      if ((erased == 0) && ((sigSet.copy_if(CheckSignatureSize(onlySize))).empty()))
         Log::error("Geneva::init") << "No signatures with given size " << onlySize << " found." << Log::end;
   }
   if (!generator->setSignatures(sigSet)) {
      Log::error("Geneva::init") << "Generator cannot handle signatures needed by the calculation." << Log::end;
   }

   // Option dump
   Log::verboseInfo("Geneva::init") << "Options set:\n"
   << "-----------------------------------------------------\n"
   << _options;

   // print overview of what we're going to do
   int numEvents = _options.getOption<int>("num");
   ComponentInfo components = calc->componentInfo();
   Log::info("Geneva::init") << "Generating " << numEvents << " events using\n"
   << "Generator: " << generator->name() << '\n'
   << "Calculation: " << calc->name() << '\n'
   << "with the following Components:\n";
   for (size_t i = 0; i < components.size(); ++i) {
      Log::info() << components[i].first << ": " << components[i].second << '\n';
   }

   if (_options.getOption("EventGenerator") == "Simple")
      _evtGen = new EventGeneratorWeighted(generator, calc, _options);
   else if (_options.getOption("EventGenerator") == "Adaptive")
      _evtGen = new EventGeneratorAdaptive(generator, calc, _options);
   else if (_options.getOption("EventGenerator") == "AdaptivePartialUnweighted")
      _evtGen = new EventGeneratorAdaptivePartialUnweighted(generator, calc, _options);
   else
      Log::error("Geneva::init") << "Unknown event generator (has to be one of 'Simple', 'Adaptive', 'AdaptivePartialUnweighted')" << Log::end;

   _isInitialized = true;
}

//------------------------------------------------------------------------------
const Event& Geneva::next()
{
   if (not _isInitialized)
      Log::fatal("Geneva::next") << "Geneva has not been initialized." << Log::end;
   return _evtGen->next();
}

//------------------------------------------------------------------------------
void Geneva::finalize(const size_t numEvents)
{
   if (not _isInitialized)
      Log::fatal("Geneva::finalize") << "Geneva has not been initialized." << Log::end;
   _evtGen->finalize(numEvents);
}

//------------------------------------------------------------------------------
std::string Geneva::header()
{
   std::stringstream head;
   head << "+------------------------------------------------------------------------------+\n";
   head << "|  Geneva " << std::setw(69) << std::left << config::version << "|\n";
   time_t rawtime = std::time(0);
   char formattedDate[20];
   char formattedTime[20];
   std::strftime(formattedDate, 20, "%Y-%m-%d", std::localtime(&rawtime));
   std::strftime(formattedTime, 20, "%H:%M:%S %Z", std::localtime(&rawtime));
   head << "|  run started on " << formattedDate << " at " << std::setw(47) << std::left << formattedTime << "|\n";
   head << "+------------------------------------------------------------------------------+\n";
   return head.str();
}

//------------------------------------------------------------------------------
std::string Geneva::footer()
{
   std::stringstream foot;
   foot << "+------------------------------------------------------------------------------+\n";
   time_t rawtime = std::time(0);
   char formattedDate[20];
   char formattedTime[20];
   std::strftime(formattedDate, 20, "%Y-%m-%d", std::localtime(&rawtime));
   std::strftime(formattedTime, 20, "%H:%M:%S %Z", std::localtime(&rawtime));
   foot << "|  run ended on " << formattedDate << " at " << std::setw(49) << std::left << formattedTime << "|\n";
   foot << "+------------------------------------------------------------------------------+\n";
   return foot.str();
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
void Geneva::setRandom()
{
   int randSeed = _options.getOption<int>("seed");
   if (randSeed < 1) {
      Log::fatal("Geneva::init") << "Cannot initialize with random seed '" << randSeed << "'. Seed must be a number greater than zero." << Log::end;
   }
   std::string randFile = _options.getOption("Random::initFile");
   if (randFile != "") {
      Log::warning("Geneva::init") << "You have an initialization file for the random number generator. Your choice of seed will be overwritten." << Log::end;
   }
   if (randFile != "")
      initializeFromStatefile(randFile);
   else
      setRandomSeed(randSeed);
}

//------------------------------------------------------------------------------
void Geneva::construct(const std::string& cmdline)
{
   Log::info() << header() << Log::end;

   // load default options
   OptionParserXML parserXML;
   _options.createDefaultOptions(parserXML);
   Log::verboseInfo("Geneva") << "Loaded default options from file: " << parserXML.fileName() << Log::end;

   // parse command line
   if (!cmdline.empty())
      parse(cmdline);

   if (_options.getOption<bool>("verbose"))
      Log::setVerbosity(Log::kVerboseInfo);

   setRandom();
}

//------------------------------------------------------------------------------
void Geneva::parse(const std::string& cmdline)
{
   OptionParserCL parserCL(cmdline);
   if (parserCL.hasOption("help")) {
      Log::info() << "TODO: Print some more useful help here ..." << Log::end;
      _options.printOptions();
      // exiting gracefully ...
      throw NoError();
   };

   // set options from file (overwriting defaults)
   OptionParserCL::const_iterator it;
   if (parserCL.find("optionsFile", it)) {
      // get the extension of the option file
      std::string extension;
      std::string::size_type idx;
      idx = it->second.rfind('.');
      if (idx != std::string::npos)
         extension = it->second.substr(idx+1);
      if (extension == "xml") {
         OptionParserXML parser(it->second);
         _options.setUserOptions(parser, false);
      } else if (extension == "yml") {
         // parse the yaml options
         OptionParserYAML parser(it->second);
         OptionMap yamlOptions;
         yamlOptions.createDefaultOptions(parser);

         // translate yaml options in new layout to old layout
         _options.setUserOptions(convertYAMLToXML(yamlOptions), true);
      } else
         Log::fatal("Geneva") << "Provided option file " << it->second << " is not a .xml or .yml file." << Log::end;
      Log::verboseInfo("Geneva") << "Loaded options from " << it->second << '.' << Log::end;
   }

   // set options from command line (overwriting defaults and options from file)
   _options.setUserOptions(parserCL, true);
   Log::verboseInfo("Geneva") << "Loaded command line options." << Log::end;
}

} // namespace Geneva
