//------------------------------------------------------------------------------
/// \file LHEFInterface.cpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2013 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class LHEFInterface
//------------------------------------------------------------------------------

#include "Geneva/version.hpp"

#include "Geneva/Driver/LHEFInterface.hpp"

#include "Geneva/PythiaInterface/ShowerScaleSCETppV012.hpp"

#include "Geneva/Components/ComponentFactory.hpp"
#include "Geneva/Components/ITreeMEColor.hpp"
#include "Geneva/Components/Collider.hpp"
#include "Geneva/Components/IPDF.hpp"

#include "Geneva/Core/Physics/LHEvent.hpp"
#include "Geneva/Core/Physics/Event.hpp"
#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/Utils/Random.hpp"

#include <iomanip>
#include <sstream>
#include <algorithm>

namespace Geneva
{

//------------------------------------------------------------------------------
LHEFInterface::LHEFInterface(const OptionMap& options, std::string fileName)
   : _fileName(fileName)
{
   Collider collider(options);
   _beam = std::make_pair(collider.beamA(), collider.beamB());
   _beamEnergy.first = 0.5 * options.getOption<double>("Ecm");
   _beamEnergy.second = 0.5 * options.getOption<double>("Ecm");

   if (collider.isHadronic()) {
      IPDF* pdf = ComponentFactory::instance().getPDF(options);
      _pdfID = std::make_pair(pdf->pdfIDA(), pdf->pdfIDB());
      delete pdf;
   } else {
      _pdfID = std::make_pair(-1, -1);
   }

   _calc = options.getOption<std::string>("calc");
   if (_calc == "CalcSCETppV012") {
      _tauNcut = options.getOption<double>("NJettiness::Tau0cut");
      _tauNp1cut = options.getOption<double>("NJettiness::Tau1cut");
   } else {
      Log::fatal("LHEFInterface") << "LHEF output for calculation " << _calc << " is not supported." << Log::end;
   }

   assert(nprup() == 1);
   _xsecup.resize(nprup(), 0.);
   _xerrup.resize(nprup(), 0.);
   _xmaxup.resize(nprup(), 0.);
   _lprup.push_back(1);

   // setup the output stream
   _out.open(_fileName.c_str());
   if (!_out.good())
      Log::fatal("LHEFInterface") << "Cannot open the LHE output file " << _fileName << Log::end;
   _out << std::setprecision(16) << std::scientific;
   _out << std::boolalpha;
   printHeader(_out, options);

   _colorCalculator = ComponentFactory::instance().getTreeMEColor(options);
   Log::verboseInfo("LHEFInterface") << "Large-Nc color flow provider: " << _colorCalculator->name() << Log::end;

   _normalizeWeights = options.getOption<bool>("LHEF::normalizeWeights");
   _runSeed = options.getOption<int>("seed");
}

//------------------------------------------------------------------------------
// Writes out the event in LHEF format.
void LHEFInterface::writeEvent(const Event& event)
{
   if (event.weightedSigma().isDefined()) {
      const double scalup = getShowerScale(event.config());
      if (!std::isfinite(scalup))
         Log::error("LHEFInterface") << "Obtained nonfinite value for shower scale (scalup): " << scalup << Log::end;

      LHEvent lhEvent(event, scalup);
      setColorFlow(lhEvent.config());
      printEvent(_out, lhEvent);
   }
   _eventRate.addWeight(event.weightedSigma());
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
void LHEFInterface::finalize()
{
   if (_totalRate.totEntries() == 0) {
      Log::warning("LHEFInterface::finalize") << "The total cross section was never set. Using the cross section calculated from the received events." << Log::end;
      _totalRate = _eventRate;
   }

   printFooter(_out);
   _out.close();

   if (_normalizeWeights)
      try {
         normalizeWeights();
      } catch (Geneva::NormalError&) {
         Log::warning("LHEFInterface::finalize")
            << "Failed to normalize weights. The original (unnormalized) file is in " << _fileName << Log::end;
      }

   delete _colorCalculator;
}

namespace
{
//------------------------------------------------------------------------------
bool string_begins_with_any(const std::string input, std::vector<std::string> matches)
{
   for (std::vector<std::string>::iterator it = matches.begin(); it != matches.end(); ++it) {
      if (input.compare(0, it->length(), *it) == 0)
         return true;
   }
   return false;
}
}

//------------------------------------------------------------------------------
std::ostream& LHEFInterface::printHeader(std::ostream& out, const OptionMap& options) const
{
   // the comment block
   out << "<LesHouchesEvents version=\"3.0\">\n";
   out << "<!--\n";
   out << "  ## File: " << _fileName << '\n';
   out << "  ## Created by Geneva version " << config::version << '\n';
   std::string tmp(config::gitChanges);
   if (!tmp.empty()) {
      size_t pos = 0.;
      while ((pos = tmp.find('\n', pos)) != std::string::npos)
         tmp.insert(++pos, "  ## ");
      out << "  ## " << tmp << '\n';
   }

   std::vector<std::string> excludeFromLhefHeader;
   excludeFromLhefHeader.push_back("Analyzer");
   excludeFromLhefHeader.push_back("UserHookRestriction");
   excludeFromLhefHeader.push_back("UHR");
   excludeFromLhefHeader.push_back("overwriteOutput");
   excludeFromLhefHeader.push_back("writeReport");
   excludeFromLhefHeader.push_back("writeEvents");
   excludeFromLhefHeader.push_back("showTimer");
   excludeFromLhefHeader.push_back("verbose");
   excludeFromLhefHeader.push_back("Mint::refineGrid");
   excludeFromLhefHeader.push_back("LHEF::normalizeWeights");
   out << "  ## Options:\n";
   for (OptionMap::const_iterator it = options.begin(); it != options.end(); ++it)
      if (!string_begins_with_any(it->first, excludeFromLhefHeader))
         out << "  # " << it->first << " : "  << it->second->getValueStr() << '\n';
   out << "-->\n";

   // the <init> block
   out << "<init>\n";
   out << ' ' << beamA() << ' ' << beamB()
      << ' ' << beamAEnergy() << ' ' << beamBEnergy()
      << " -1 -1"                                     // pdf group (PDFGUP) is outdated and unused
      << ' ' << beamApdfID() << ' ' << beamBpdfID()   //
      << ' ' << idwtup() << ' ' << nprup()
      << '\n';
   for (size_t i = 0; i < nprup(); ++i)
      out << ' ' << xsecup(i) << ' ' << xerrup(i) << ' ' << xmaxup(i) << ' ' << lprup(i) << '\n';

   out << "<generator name=\"Geneva\" version=\"" << config::version << "\"/>\n";
   printWeightInfo(out, options);
   out << "</init>\n";

   return out;
}

//------------------------------------------------------------------------------
/**
 * \warning Currently the info is hard-coded and only available for SCETppV012
 * calculation with scale=all.
 *
 * \todo Find a way to get the proper weight information for each Calculation.
 */
std::ostream& LHEFInterface::printWeightInfo(std::ostream& out, const OptionMap& options) const
{
   if (options.getOption<std::string>("calc") == "CalcSCETppV012"
       && options.getOption<std::string>("CalcSCETppV012::scale") == "all") {
      out << "<weightinfo name=\"central\"> central weight </weight>\n";
      out << "<weightgroup type=\"fixed-order scale variations\" combine=\"envelope\">\n";
      out << "<weightinfo name=\"muup\"> overall scale up by factor 2.0 </weightinfo>\n";
      out << "<weightinfo name=\"mudown\"> overall scale down by factor 0.5 </weightinfo>\n";
      out << "</weightgroup>\n";
      out << "<weightgroup type=\"resummation scale variations\" combine=\"symm_envelope\">\n";
      out << "<weightinfo name=\"muBup\"> beam scale up </weightinfo>\n";
      out << "<weightinfo name=\"muBdown\"> beam scale down </weightinfo>\n";
      out << "<weightinfo name=\"muSup\"> soft scale up </weightinfo>\n";
      out << "<weightinfo name=\"muSdown\"> soft scale down </weightinfo>\n";
      out << "<weightinfo name=\"muxup\"> profile transition up </weightinfo>\n";
      out << "<weightinfo name=\"muxdown\"> profile transition down </weightinfo>\n";
      out << "</weightgroup>\n";
      out << "<weightgroup type=\"inclusive scale variations\" combine=\"envelope\">\n";
      out << "<weightinfo name=\"muupincl\"> overall scale up by factor 2.0 for inclusive quantities </weightinfo>\n";
      out << "<weightinfo name=\"mudownincl\"> overall scale down by factor 0.5 for inclusive quantities </weightinfo>\n";
      out << "</weightgroup>\n";
   } else {
      Log::warning("LHEFInterface::printWeightInfo")
         << "Weight information is only available for SCETppV012 calculation with all scale variations." << Log::end;
   }
   return out;
}

//------------------------------------------------------------------------------
// Format the events
std::ostream& LHEFInterface::printEvent(std::ostream& out, const LHEvent& event) const
{
   out << "<event>\n";

   out << ' ' << event.nup()
      << ' ' << lprup()      // only support a single process
      << ' ' << event.xwgtup()
      << ' ' << event.scalup()
      << " -1."              // unused event.aQEDup()
      << " -1."              // unused event.aQCDup()
      << '\n';

   const Configuration& config = event.config();
   for (Configuration::iterator it = config.begin(); it != config.end(); ++it) {
      out << ' ' << std::setw(3) << it->id()
         << ' ' << std::setw(2) << event.istup(it)
         << ' ' << std::setw(2) << event.mothup1(it)
         << ' ' << std::setw(2) << event.mothup2(it)
         << ' ' << std::setw(4) << event.icolup1(it)
         << ' ' << std::setw(4) << event.icolup2(it)
         << ' ' << it->px() << ' ' << it->py() << ' ' << it->pz() << ' ' << it->energy() << ' ' << it->mass()
         << " 0."            // unused default lifetime event.vtimup()
         << " 9."            // unused default spin event.spinup()
         << '\n';
   }

   // put additional weights into <weights> section according to LHEF3 compressed format
   const Sigma& weights = event.weightedSigma();
   if (weights.variations().size() > 0) {
      out << "<weights> " << weights.central();
      for (Sigma::Variations::const_iterator it = weights.variations().begin(); it != weights.variations().end(); ++it)
         out << ' ' << *it;
      out << " </weights>\n";
   }

   out << "</event>\n";
   return out;
}

//------------------------------------------------------------------------------
// Writes out the footer.
std::ostream& LHEFInterface::printFooter(std::ostream& out) const
{
   out << "</LesHouchesEvents>\n";
   out << "<!--\n";
   out << "  >>> Additional Info (lowest block is most up-to-date)\n";
   out << "  ### Begin Run Information\n";
   out << "  #   random seed : " << _runSeed << "\n";
   out << "  #   number of events in file : " << _eventRate.entries() << "\n";
   out << "  #   weights normalized to total cross section : " << _normalizeWeights << "\n";
   if (_normalizeWeights)
      out << "  #   normalization factor : " << _totalRate.value().central() / _eventRate.sumWeights().central() * static_cast<double>(_eventRate.entries()) << "\n";
   out << "  ##  Total Cross Section\n";
   out << "  #   number of processed events : " << _totalRate.totEntries() << "\n";
   out << "  #   sum of processed weights : " << _totalRate.sumWeights().central() << "\n";
   out << "  #   sum of processed weights squared : " << _totalRate.sumWeightsSq().central() << "\n";
   out << "  #   total cross section value [pb] : " << _totalRate.value().central() << "\n";
   out << "  #   total cross section error [pb] : " << _totalRate.error().central() << "\n";
   out << "  ### End Run Information\n";
   out << "-->\n";
   return out;
}

//------------------------------------------------------------------------------
// Adjusts the events weights in the file to match the total cross section.
void LHEFInterface::normalizeWeights() const
{
   const double reweight = _totalRate.value().central() / _eventRate.sumWeights().central() * static_cast<double>(_eventRate.entries());

   _istream_type inFile(_fileName.c_str());
   if (!inFile.good())
      Log::error("LHEFInterface::normalizeWeights") << "Cannot read the original LHE file " << _fileName << Log::end;

   const std::string tmpFile = _fileName + ".tmp";
   _ostream_type outFile(tmpFile.c_str());
   if (!outFile.good())
      Log::error("LHEFInterface::normalizeWeights") << "Cannot write to temporary LHE file " << tmpFile << Log::end;
   outFile << std::setprecision(16) << std::scientific;

   std::string line;
   size_t numEvents = 0;
   while (outFile && std::getline(inFile, line)) {
      if (line.find("<weights>") != std::string::npos) {
         std::istringstream linestream(line);
         std::string pre;
         linestream >> pre;
         outFile << "<weights>";
         double weight;
         while (linestream >> weight) {
            outFile << " " << weight * reweight;
         }
         outFile << " </weights>\n";
      } else if (line.find("<event>") != std::string::npos) {
         outFile << line << '\n';
         // Parse the line after the <event> tag
         std::getline(inFile, line);
         std::istringstream linestream(line);
         int num, status;
         double weight;
         if (linestream >> num >> status >> weight)
            outFile << num << " " << status << " " << weight * reweight << linestream.rdbuf() << '\n';
         else
            break;
         numEvents++;
      } else {
         outFile << line << '\n';
      }
   }
   inFile.close();
   outFile.close();
   if (!inFile.eof())
      Log::error("LHEFInterface::normalizeWeights") << "Unexpected error while reading from " << _fileName << Log::end;
   if (!outFile)
      Log::error("LHEFInterface::normalizeWeights") << "Unexpected error while writing to " << tmpFile << Log::end;
   if (std::rename(tmpFile.c_str(), _fileName.c_str()))
      Log::error("LHEFInterface::normalizeWeights") << "Cannot rename reweighted LHE file " << tmpFile << " to final " << _fileName << Log::end;

   Log::verboseInfo("LHEFInterface::normalizeWeights") << "Normalized " << numEvents << " events to total cross section." << Log::end;
}

//------------------------------------------------------------------------------
// Determines the shower starting scale (scalup) for config.
double LHEFInterface::getShowerScale(const Configuration& config) const
{
   if (_calc == "CalcSCETppV012") {
      // NOTE: Currently this assumes maxFinal = 2
      return ShowerScaleSCETppV012(2, _tauNcut, _tauNp1cut, 1.0).getpTScale(config);
   }

   assert("Unsupported calculation" == 0);
   return 0.;
}

//------------------------------------------------------------------------------
// Sets the color flow for \a config.
void LHEFInterface::setColorFlow(Configuration& config) const
{
   _colorCalculator->calculateColorInfo(config);
   double total = 0.;
   for (size_t i = 0; i < _colorCalculator->numColorFlows(); ++i)
      total += _colorCalculator->valueLargeNc(i);

   const double thr = randomInterval() * total;
   double partial = _colorCalculator->valueLargeNc(0);
   size_t colorFlow = 0;
   while (partial < thr) {
      ++colorFlow;
      partial += _colorCalculator->valueLargeNc(colorFlow);
   }
   assert(colorFlow < _colorCalculator->numColorFlows());

   for (Configuration::iterator it = config.begin(); it != config.endFinal(); ++it) {
      // sanctioned const_cast to change configurations color info in-place
      std::pair<int, int> color = _colorCalculator->colorIndices(it, colorFlow);
      const_cast<Configuration::iterator::value_type&>(*it).color1() = color.first;
      const_cast<Configuration::iterator::value_type&>(*it).color2() = color.second;
   }
}

} // namespace Geneva
