//------------------------------------------------------------------------------
/// \file GeneratorFactory.cpp
//
// Author(s):
//    Chris Vermilion, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class GeneratorFactory
//------------------------------------------------------------------------------

#include "Geneva/Generators/GeneratorFactory.hpp"
#include "Geneva/Generators/GeneratorPP2BosonJets.hpp"

#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include <limits>

using std::string;
using std::vector;

namespace Geneva
{

//------------------------------------------------------------------------------
GeneratorFactory::~GeneratorFactory()
{
   for (size_t i = 0; i < _toDelete.size(); ++i)
      delete _toDelete[i];
}

//------------------------------------------------------------------------------
/// Given options in optionMap, create a Generator
IGenerator* GeneratorFactory::getGenerator(const OptionMap& optionMap, string genName)
{
   string name = (genName != "") ? genName : optionMap.getOption<string>("Generator");
   string calcName = optionMap.getOption<string>("calc");

   if (name == "PP2BosonJets") {
      double Ecm = optionMap.getOption<double>("Ecm");
      int nLight = optionMap.getOption<int>("NLightFlavors");
      std::string sampleVirtuality = optionMap.getOption<string>("PP2BosonJets::VirtualitySampling");
      GeneratorPP2BosonJets::SamplingMass vSample = GeneratorPP2BosonJets::kBreitWignerMass;
      if (sampleVirtuality == "BreitWigner") {
            vSample = GeneratorPP2BosonJets::kBreitWignerMass;
      } else if (sampleVirtuality =="Flat") {
            vSample = GeneratorPP2BosonJets::kFlatMass;
      } else {
         Log::error("GeneratorFactory::getGenerator") << "Unknown PP2BosonJets::VirtualitySampling mode " << sampleVirtuality << Log::end;
      }
      std::string sampleTau = optionMap.getOption<string>("PP2BosonJets::CMEnergySampling");
      GeneratorPP2BosonJets::SamplingTau tauSample = GeneratorPP2BosonJets::kPowerTau;
      if (sampleTau == "Power") {
         tauSample = GeneratorPP2BosonJets::kPowerTau;
      } else if (sampleTau == "Exp") {
         tauSample = GeneratorPP2BosonJets::kExponentialTau;
      } else if (sampleTau == "Sqrt") {
         tauSample = GeneratorPP2BosonJets::kSqrtTau;
      } else if (sampleTau == "Flat") {
         tauSample = GeneratorPP2BosonJets::kFlatTau;
      } else {
         Log::error("GeneratorFactory::getGenerator") << "Unknown PP2BosonJets::CMEnergySampling mode " << sampleTau << Log::end;
      }
      std::string sampleY = optionMap.getOption<string>("PP2BosonJets::CMRapiditySampling");
      GeneratorPP2BosonJets::SamplingY ySample = GeneratorPP2BosonJets::kPowerY;
      if (sampleY == "Power") {
         ySample = GeneratorPP2BosonJets::kPowerY;
      } else if (sampleY == "Flat") {
         ySample = GeneratorPP2BosonJets::kFlatY;
      } else {
         Log::error("GeneratorFactory::getGenerator") << "Unknown PP2BosonJets::CMRapiditySampling mode " << sampleY << Log::end;
      }

      std::string sampleThetaJet = optionMap.getOption<string>("PP2BosonJets::ThetaJetSampling");
      GeneratorPP2BosonJets::SamplingThetaJet thetaJetSample = GeneratorPP2BosonJets::kImportanceThetaJet;
      if (sampleThetaJet == "Importance") {
         thetaJetSample = GeneratorPP2BosonJets::kImportanceThetaJet;
      } else if (sampleThetaJet == "Flat") {
         thetaJetSample = GeneratorPP2BosonJets::kFlatThetaJet;
      } else {
         Log::error("GeneratorFactory::getGenerator") << "Unknown PP2BosonJets::ThetaJetSampling mode " << sampleThetaJet << Log::end;
      }

      PdgID idB;
      double massB = 0.;
      double widthB = 0;
      std::string boson = optionMap.getOption<string>("Process::DrellYan::BosonType");
      if (boson == "Z") {
         idB = PdgID::kZ;
         massB = optionMap.getOption<double>("ZMass");
         widthB = optionMap.getOption<double>("ZWidth");
      } else if (boson == "W+" || boson == "W-") {
         idB = (boson == "W+") ? PdgID::kWPlus : PdgID::kWMinus;
         massB = optionMap.getOption<double>("WMass");
         widthB = optionMap.getOption<double>("WWidth");
      } else {
         Log::error("GeneratorFactory::getGenerator") << "Unknown boson " << boson << Log::end;
      }
      double lowBmass = optionMap.getOption<double>("PP2BosonJets::BInvMassCutLow");
      double highBmass = optionMap.getOption<double>("PP2BosonJets::BInvMassCutHigh");
      bool cutBrapidity = optionMap.getOption<bool>("PP2BosonJets::BRapidityCut");
      double lowBrapidity = -std::numeric_limits< double >::infinity();
      double highBrapidity = std::numeric_limits< double >::infinity();
      if (cutBrapidity) {
         lowBrapidity = optionMap.getOption<double>("PP2BosonJets::BRapidityCut::Low");
         highBrapidity = optionMap.getOption<double>("PP2BosonJets::BRapidityCut::High");
      }
      double minKtJet = optionMap.getOption<double>("PP2BosonJets::KtJetCut");
      double minMassJetJet = optionMap.getOption<double>("PP2BosonJets::MJetJetCut");

      bool useBjjFKS = optionMap.getOption<bool>("PP2BosonJets::BjjFKS");

      std::string BjjTau1Restriction = optionMap.getOption<string>("PP2BosonJets::BjjTau1Restriction");
      double Tau1cut = optionMap.getOption<double>("PP2BosonJets::Tau1cut");

      bool useTau0Generator = optionMap.getOption<bool>("PP2BosonJets::useTau0Generator");

      _toDelete.push_back(new GeneratorPP2BosonJets(Ecm, nLight, massB, widthB, idB, lowBmass, highBmass, minKtJet, minMassJetJet, lowBrapidity, highBrapidity, GeneratorPP2BosonJets::kRandom, GeneratorPP2BosonJets::kRambo, vSample, tauSample, ySample, thetaJetSample, useBjjFKS, useTau0Generator, BjjTau1Restriction, Tau1cut));
      return _toDelete.back();
   } else {
      Log::error("GeneratorFactory::getGenerator") << "Unknown Generator " << name << Log::end;
      return 0;
   }
}

} // namespace Geneva
