//------------------------------------------------------------------------------
/// \file GeneratorPP2BosonJets.cpp
//
// Author(s):
//    Simone Alioli, Chris Vermilion
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class GeneratorPP2BosonJets
//------------------------------------------------------------------------------

#include "Geneva/config.hpp"

#include "Geneva/Generators/GeneratorPP2BosonJets.hpp"
#include "Geneva/Generators/GeneratorUtils.hpp"

#include "Geneva/Components/RegionInfoList.hpp"
#include "Geneva/Components/SignatureManager.hpp"
#include "Geneva/Components/FKSMapping.hpp"

#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/Physics/ConfigurationUtils.hpp"
#include "Geneva/Core/Physics/PhiRadFKS.hpp"
#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/UnitCubeRad.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/Random.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"
#include "Geneva/Core/Utils/NumericsUtils.hpp"

#include <gsl/gsl_roots.h>

#include <list>
#include <cassert>
#include <algorithm>
#include <iostream>
#include <iomanip>
#include <functional>
#include <iterator>

#define ACCURACY_THRESHOLD 1e-6

namespace Geneva
{

//------------------------------------------------------------------------------
struct GeneratorPP2BosonJets::_solver_type : gsl_root_fdfsolver {};

//------------------------------------------------------------------------------
GeneratorPP2BosonJets::GeneratorPP2BosonJets(double Ecm, int nLight, double massB, double widthB, PdgID idB,
                                             double lowBmass, double highBmass, double ktJetcut, double mJetJetCut, double yBcutLow, double yBcutHigh,
                                             SignatureAccessMode sigAccess, LargeMultiplicitiesGenerationMode genMode,
                                             SamplingMass importSampleMass, SamplingTau importSampleTau,
                                             SamplingY importSampleY, SamplingThetaJet importSampleThetaJet,
                                             bool useBjjFKS, bool useTau0Generator, std::string bjjTau1cut, double tau1cut)
   : _sigSet(), _cmEnergy(Ecm), _massB(massB), _widthB(widthB), _idB(idB), _nLight(nLight), _sigAccess(sigAccess), _genModeLargeN(genMode),
     _importanceSampleMass(importSampleMass), _importanceSampleTau(importSampleTau), _importanceSampleY(importSampleY), _importanceSampleThetaJet(importSampleThetaJet),
     _ktJetCut(ktJetcut),  _mJetJetCut(mJetJetCut), _yBcutLow(yBcutLow), _yBcutHigh(yBcutHigh),
     _signatureNumber(0), _numGenerated(0), _numFail(0), _useBjjFKS(useBjjFKS), _bjjTau1cut(bjjTau1cut),  _tau1cut(tau1cut), _useTau0Generator(useTau0Generator)
{
   bool zeroWidthApprox = (compareToZero(widthB) == 0);
   if (!zeroWidthApprox) {
      double lowBosonMassCut;
      if (_idB == PdgID::kZ && lowBmass < 1.0) {
         Log::warning(name()) << " Invariant mass cut at 1 GeV to avoid photon pole! " << Log::end;
         lowBosonMassCut = 1.0;
      } else {
         lowBosonMassCut = lowBmass;
      }
      double bosonMassWindowLow = (_massB - lowBosonMassCut) / widthB;
      double bosonMassWindowHigh = (highBmass - _massB) / widthB;
      _lowBosonMassSquared = pow(std::max(lowBosonMassCut, _massB - bosonMassWindowLow * _widthB), 2);
      _highBosonMassSquared = pow(std::min(_cmEnergy, _massB + bosonMassWindowHigh * _widthB), 2);
   } else {
      Log::warning(name()) << " Using zero-width approximation! " << Log::end;
      _lowBosonMassSquared = _massB * _massB;
      _highBosonMassSquared = _lowBosonMassSquared;
   }

   Log::verboseInfo(name()) << sqrt(_lowBosonMassSquared) << " < mB < " << sqrt(_highBosonMassSquared) << " , kT j > " << _ktJetCut << " , Mjj > " << _mJetJetCut << Log::end;

   if (_yBcutHigh <= _yBcutLow)
      Log::fatal(name()) << "Upper limit on boson rapidity smaller than lower limit. Halting execution!" << Log::end;

   if (_genModeLargeN == kRambo)
      _solver = static_cast<_solver_type*>(gsl_root_fdfsolver_alloc(gsl_root_fdfsolver_newton));
}

//------------------------------------------------------------------------------
// Destructor to delete the gsl root solver
GeneratorPP2BosonJets::~GeneratorPP2BosonJets()
{
   if (_genModeLargeN == kRambo)
      gsl_root_fdfsolver_free(_solver);
}

//------------------------------------------------------------------------------
void GeneratorPP2BosonJets::finalize()
{
   Log::verboseInfo(name()) << " Efficiency: " << double(_numGenerated) / double(_numGenerated + _numFail) << Log::end;
}

//------------------------------------------------------------------------------
// Returns the current set of Signatures used by the generator.
const SignatureSet& GeneratorPP2BosonJets::getSignatures() const
{
   return _sigSet;
}

//------------------------------------------------------------------------------
// Sets the Signatures for which to generate
bool GeneratorPP2BosonJets::setSignatures(const SignatureSet& sigSet)
{
   _sigSet.clear();
   for (SignatureSet::iterator it = sigSet.begin(); it != sigSet.end(); ++it)
      if (!addSignature(*it))
         return false;
   assert(_sigSet == sigSet);

   return true;
}

//------------------------------------------------------------------------------
// Adds a Signature to generate Configurations for.
bool GeneratorPP2BosonJets::addSignature(const Signature& sig)
{
   if (checkSignature(sig)) {
      _sigSet.insert(sig);
      return true;
   } else
      return false;
}

//------------------------------------------------------------------------------
// Checks if the generator can handle a Signature.
bool GeneratorPP2BosonJets::checkSignature(const Signature& sig) const
{
   // count leptons
   ptrdiff_t nLeptons = std::count_if(sig.beginFinal(), sig.endFinal(), std::mem_fun_ref(&PdgID::isLepton));

   // count weak and/or Higgs bosons, exclude gluons
   ptrdiff_t nBosons = std::count_if(sig.beginFinal(), sig.endFinal(), std::mem_fun_ref(&PdgID::isGaugeBoson)) - std::count_if(sig.beginFinal(), sig.endFinal(), std::mem_fun_ref(&PdgID::isGluon));

   if (includesDecay(sig)) {
      return ((nLeptons == 2)  && IsQCDPair()(*sig.initialA(), *sig.initialB()));
   } else {
      return ((nBosons == 1)  && IsQCDPair()(*sig.initialA(), *sig.initialB()));
   }
}

//------------------------------------------------------------------------------
// Generates a new Configuration.
bool GeneratorPP2BosonJets::generate(Configuration& configuration)
{
   if (_sigSet.size() == 0) {
      Log::fatal("GeneratorPP2BosonJets::generate") << "empty SignatureSet! " << Log::end;
   }
   _lastConfig.clear();

   if (_sigSet.size() > 1) {
      if (_sigAccess == kRandom) {
         _signatureNumber = randomInteger(_sigSet.size());
      } else if (_sigAccess == kSequential) {
         _signatureNumber++;
         if (_signatureNumber >= _sigSet.size())  _signatureNumber = 0;
      } else {
         Log::fatal("GeneratorPP2BosonJets::generate") << "Invalid signature access method" << Log::end;
      }
   }
   SignatureSet::iterator signature = _sigSet.begin();
   std::advance(signature, ptrdiff_t_from_size_t(_signatureNumber));

   if (!checkSignature(*signature)) {
      Log::fatal("GeneratorPP2BosonJets::generate") << "Signature " << *signature << " not allowed " << Log::end;
   }

   size_t numRand = numberOfRandomsNeeded(*signature);
   std::vector<double> rand;
   for (size_t i = 0; i < numRand; ++i)
      rand.push_back(randomInterval());

   bool success = generate(*signature, rand, configuration);
   _lastWeight /= static_cast<double>(_sigSet.size());
   return success;

}

//------------------------------------------------------------------------------
// Generates a new Configuration.
bool GeneratorPP2BosonJets::generate(const Signature& signature, std::vector<double> rand, Configuration& configuration)
{
   if (_sigSet.size() == 0) {
      Log::fatal("GeneratorPP2BosonJets::generate") << "empty SignatureSet! " << Log::end;
   }
   _lastConfig.clear();

   // count QCD particles in final state
   ptrdiff_t nPartons = std::count_if(signature.beginFinal(), signature.endFinal(), IsQCD());
   double jac = std::numeric_limits< double >::quiet_NaN();
   bool success = false;
   // Clear the previous configuration
   configuration.clear();
   // decide which generator to use according to parton multiplicity
   switch (nPartons) {
      case 0:
         success = generateB(signature, configuration, rand, jac);
         break;
      case 1:
         if (_useTau0Generator) {
            success = generateBjTau0(signature, configuration, rand, jac);
         } else {
            success = generateBj(signature, configuration, rand, jac);
         }
         break;
      case 2:
         if (_useBjjFKS) {
            success = generateBjjRecursiveFKS(signature, configuration, rand, jac);
         } else {
            success = generateBjj(signature, configuration, rand, jac);
         // success = generateBjetsRAMBO(*signature, configuration, jac);
         }
         break;
      default:
         Log::fatal("GeneratorPP2BosonJets::generate") << "LargeMultiplitiesGenerationMode not allowed " << Log::end;
   }
   if (!success) {
      _lastConfig.clear();
      _lastWeight = 0.0;
      return false;
   } else {
#ifdef REFINE_MOMENTA
      //TODO: Make RefineMomenta work for undecayed particles
      if (includesDecay(configuration)) {
         enforceMomentumConservation(configuration, "GeneratorPP2BosonJets::generate", REFINE_TARGET_PRECISION, REFINE_TARGET_ACCURACY);
      }
#endif
      _lastConfig = configuration;
      _lastWeight = 1. / abs(jac);  // Geneva wants 1/Jacobian as the event weight of the generator
      _numGenerated++;
      return true;
   }
}

//------------------------------------------------------------------------------
size_t GeneratorPP2BosonJets::numberOfRandomsNeeded(const Signature& sig)
{
   size_t nPartons = size_t_from_ptrdiff_t(std::count_if(sig.beginFinal(), sig.endFinal(), IsQCD()));
   return 2 + 3 * (nPartons + 2) - 4 - 1;
}

//------------------------------------------------------------------------------
// Returns the weight of the last generated configuration.
double GeneratorPP2BosonJets::getLastWeight()
{
   return _lastWeight;
}

//------------------------------------------------------------------------------
// Returns the name "GeneratorPP2BosonJets" of the Generator.
const std::string GeneratorPP2BosonJets::name() const
{
   return "GeneratorPP2BosonJets";
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

bool GeneratorPP2BosonJets::includesDecay(const Signature& sig) const
{
   return !(sig.includesFinal(PdgID::kZ)
            || sig.includesFinal(PdgID::kWPlus)
            || sig.includesFinal(PdgID::kWMinus));
}

//------------------------------------------------------------------------------
double GeneratorPP2BosonJets::generateInvMassSquared(double ran, bool doDecay, double& massSquared)
{
   double jac = 1.0;
   if (doDecay) {
      if ((_lowBosonMassSquared > _massB * _massB || _highBosonMassSquared < _massB * _massB) || _importanceSampleMass == kFlatMass) {
         assert(_lowBosonMassSquared < _highBosonMassSquared);
         // if the peak is not in the integration region or if explicitly requested  a flat sampling is used
         massSquared = _lowBosonMassSquared + (_highBosonMassSquared - _lowBosonMassSquared) * ran;
         jac *= (_highBosonMassSquared - _lowBosonMassSquared);
      } else {
         // otherwise importance sample around the peak
         assert (_importanceSampleMass == kBreitWignerMass);
//         double sampling = 3.;
         double sampling = 1.2;
         double ranMass   = (erf(sampling * (2.*ran - 1.)) / erf(sampling) + 1.) / 2.;
         jac *= 2.*sampling * exp(-pow(sampling * (2.*ran - 1.), 2)) / sqrt(math_constants::pi()) / erf(sampling);
         double zLow = atan((_lowBosonMassSquared - _massB * _massB) / (_massB * _widthB));
         double zHigh = atan((_highBosonMassSquared - _massB * _massB) / (_massB * _widthB));
         assert(zLow < zHigh);
         double z = zLow + (zHigh - zLow) * ranMass;
         jac *= (zHigh - zLow);
         massSquared = tan(z) * _massB * _widthB + _massB * _massB;
         jac *= _massB * _widthB / pow(cos(z), 2.);
      }
      jac *= math_constants::_1_4pi() / 4.;
   } else {
      // narrow width approximation
      massSquared = _massB * _massB;
      jac *= math_constants::_2pi();
   }
   return jac;
}

//-------------------------------------------------------------------------------
// Generates the tau value for a system in its rest frame decaying into two
// back-to back massive particles with a minimum transverse momentum
double GeneratorPP2BosonJets::generateTau(double ranTau, double invMass1Squared, double invMass2Squared, double kt, double& tau)
{
   double jac = 0.;
   double taumin =  pow((sqrt(invMass1Squared + pow(kt, 2)) + sqrt(invMass2Squared + pow(kt, 2))), 2) / (_cmEnergy * _cmEnergy);
   forceInUnitRange(taumin, "GeneratorPP2BosonJets::generateTau tauMin");
   //TODO: test whether a flat generation (or a different sampling) would be preferable for the final SCET calculation
   if (_importanceSampleTau == kFlatTau) {
      // flat generation in tau
      tau = taumin + ACCURACY_THRESHOLD + ranTau * (1. - taumin - 2.*ACCURACY_THRESHOLD);
      jac = 1. - taumin;
   } else if (_importanceSampleTau == kExponentialTau) {
      // generation importance sampled towards low tau values
      tau = exp(log(taumin) * (1. - ranTau * ranTau));
      jac = tau * abs(log(taumin)) * 2. * ranTau;
   } else if (_importanceSampleTau == kSqrtTau) {
      // generation importance sampled towards low tau values, trying to mimic pdf suppression
      const double oneOsqtm = 1. / sqrt(taumin);
      const double sqtran = sqrt(1. - ranTau);
      tau = 1. / (1. + sqtran * (oneOsqtm - 1.));
      tau *= tau;
      jac = abs((oneOsqtm - 1.) / (sqtran * pow((oneOsqtm - 1.) * sqtran + 1., 3)));
   } else if (_importanceSampleTau == kPowerTau) {
      // generation highly importance sampled towards low tau values, trying to mimic pdf suppression
      const double a = 4.;
      tau = pow(1. / (1. + (1. - ranTau) * (pow(taumin, -a) - 1.)), 1. / a);
      jac = abs((pow(taumin, -a) - 1.) / a * pow(tau, a + 1.));
   } else {
      Log::fatal("GeneratorPP2BosonJets::generateTau") << "importance sample mode" << _importanceSampleTau << " not allowed " << Log::end;
   }
   forceInRange(tau, taumin, 1.0, "GeneratorPP2BosonJets::generateTau tau");
   return jac;
}

//------------------------------------------------------------------------------
double GeneratorPP2BosonJets::generateY(double ranY, double tau, double yMin, double yMax, double& y)
{
   assert(yMin < yMax);
   double jac = 0.0;
   double halfLogTau = 0.5 * log(tau);
   double min = std::max(yMin, halfLogTau);
   double max = std::min(yMax, -halfLogTau);
   assert(min < max);
   double ave = 0.5 * (max + min);
   double diff = (max - min);
   if (_importanceSampleY == kFlatY) {
      y = ave - (1. - 2.*ranY) * 0.5 * diff;
      jac = diff;
   } else if (_importanceSampleY == kPowerY) {
      y = ave - pow((1. - 2.*ranY), 5) * 0.5 * diff;
      jac = diff * 5 * pow(1. - 2.*ranY, 4);
   } else {
      Log::fatal("GeneratorPP2BosonJets::generateY") << "importance sample mode" << _importanceSampleY << " not allowed " << Log::end;
   }
   return jac;
}

//------------------------------------------------------------------------------
void GeneratorPP2BosonJets::setxBarAxBarB(double tau, double y, double& xBarA, double& xBarB)
{
   xBarA = sqrt(tau) * exp(y);
   forceInUnitRange(xBarA, "GeneratorPP2BosonJets::setxBarAxBarB xBarA");
   xBarB = tau / xBarA;
   forceInUnitRange(xBarB, "GeneratorPP2BosonJets::setxBarAxBarB xBarB");
}

//------------------------------------------------------------------------------
// Generates a new Configuration with only B in the final state, given external randoms as input.
bool GeneratorPP2BosonJets::generateB(const Signature& signature, Configuration& configuration, std::vector<double> rand, double& jac)
{
   assert(rand.size() == 3 || rand.size() == 4);
   configuration.clear();
   jac = 1.0;

   double ranMass, ranY, ranCosTheta;
   // The variables above completely specify the kinematics
   // of a 2->2 event at a hadron collider.
   // The following is an overall rotation around the z-axis, whose coefficient
   // does not enter the jacobian and which is irrelavant for the integration.
   // No physical observable can depend on this, it is only customary not
   // to have e.g. the e- going always around the y-axis.

   //If we have specified 4 random numbers, use the last one for the azimuthal rotation. Otherwise generate a new random number
   double ranRotationAzimuth;
   if (rand.size() == 3)
      ranRotationAzimuth= randomAngle();
   else
      ranRotationAzimuth = math_constants::_2pi() * rand[3];

   ranMass = rand[0];
   ranY    = rand[1];
   ranCosTheta = rand[2];

   const bool doDecay = includesDecay(signature);
   // vector boson invariant mass squared
   double bosonInvMassSquared;
   jac *= generateInvMassSquared(ranMass, doDecay, bosonInvMassSquared);

   // angle of first vector boson decay products, if any
   double cosThetaDec = 0.;
   // azimuthal angle  of first vector boson decay products, if any
   double phiDec = 0.;
   if (doDecay) {
      // decay kinematics
      double ran = 1. - 2.*ranCosTheta;
      cosThetaDec = 1.5 * (ran - pow(ran, 3) / 3.);
      jac *= 3. * (1. - ran * ran);
   }

   // partonic rescaled CM energy
   double tau = 0.;
   double hadronicS = _cmEnergy * _cmEnergy;
   double partonicS =  bosonInvMassSquared;
   tau = bosonInvMassSquared / hadronicS;
   jac /= hadronicS;

   // rapidity of partonic CM frame
   double y = std::numeric_limits<double>::quiet_NaN();
   jac *= generateY(ranY, tau, _yBcutLow, _yBcutHigh, y);

   // momenta fractions for incoming partons
   double x1, x2;
   setxBarAxBarB(tau, y, x1, x2);

   // boost fro Lab to partonic CM frame
   BoostZHyp zBoost(sqrt(x1/x2));

   // Overall random rotation around the beam axis
   RotationZ zRotation(ranRotationAzimuth);

   // construct kinematics in partonic CM frame
   std::vector<FourVector> kinematics(4, FourVector());
   double eTmp = sqrt(partonicS) / 2.;

   // kPlus
   kinematics[0].setComponents(eTmp, 0., 0., eTmp);
   // kMinus
   kinematics[1].setComponents(eTmp, 0., 0., -eTmp);

   double xTmp = sqrt(1. - pow(cosThetaDec, 2)) * sin(phiDec) * eTmp;
   double yTmp = sqrt(1. - pow(cosThetaDec, 2)) * cos(phiDec) * eTmp;
   double zTmp = cosThetaDec * eTmp;

   kinematics[2].setComponents(eTmp, xTmp, yTmp, zTmp);
   kinematics[3].setComponents(eTmp, -xTmp, -yTmp, -zTmp);

   // get IDs for final particles; if boson is not decayed, it will be idEW1 and idEW2 will stay as kUnknown
   PdgID idEW1(PdgID::kUnknown), idEW2(PdgID::kUnknown);
   for (Signature::iterator it = signature.beginFinal(); it != signature.endFinal(); ++it) {
      if (idEW1 == PdgID::kUnknown)
         idEW1 = *it;
      else
         idEW2 = *it;
      assert(!IsQCD()(*it));
   }

   // Boost momenta from CM to Lab frame
   for (std::vector<FourVector>::iterator it = kinematics.begin(); it != kinematics.end(); ++it)
      it->apply(zBoost);

   // Overall rotation along z-axis
   if (doDecay) {
      kinematics[2].apply(zRotation);
      kinematics[3].apply(zRotation);
   }

   // construct the required configuration
   configuration.setInitialA(Particle(*signature.initialA(), kinematics[0]));
   configuration.setInitialB(Particle(*signature.initialB(), kinematics[1]));
   if (doDecay) {
      configuration.insertFinal(Particle(idEW1, kinematics[2]));
      configuration.insertFinal(Particle(idEW2, kinematics[3]));
   } else {
      FourVector kV = kinematics[2] + kinematics[3];
      configuration.insertFinal(Particle(idEW1, kV));
   }


   // check that B rapidity cuts are implemented correctly
   double yB = bosonRapidity(configuration);
   if ((yB < _yBcutLow) || (yB > _yBcutHigh))
      Log::error("GeneratorPP2BosonJets::generateB") << "B rapidity " << yB <<" outside ["<<_yBcutLow<<","<<_yBcutHigh<<"] range." <<Log::end;

   assert(std::isfinite(jac));
   return true;
}

//------------------------------------------------------------------------------
// Generates a new Configuration with B and 1 parton in the final state, given external randoms as input.
bool GeneratorPP2BosonJets::generateBj(const Signature& signature, Configuration& configuration, std::vector<double> rand, double& jac)
{
   assert(rand.size() == 6 || rand.size() == 7);
   configuration.clear();
   jac = 1.0;

   double ranMass, ranY, ranCosTheta;
   double ranTau, ranCosThetaDec, ranPhiDec;
   // The variables above completely specify the kinematics
   // of a 2->3 event at a hadron collider.
   // The following is an overall rotation around the z-axis, whose coefficient
   // does not enter the jacobian and which is irrelevant for the integration.
   // No physical observable can depend on this, it is only customary not
   // to have the outgoing parton going always around the y-axis.

   //If we have specified 7 random numbers, use the last one for the azimuthal rotation. Otherwise generate a new random number
   double ranRotationAzimuth;
   if (rand.size() == 6)
      ranRotationAzimuth= randomAngle();
   else
      ranRotationAzimuth = math_constants::_2pi() * rand[6];

   ranMass        = rand[0];
   ranY           = rand[1];
   ranCosTheta    = rand[2];
   ranTau         = rand[3];
   ranCosThetaDec = rand[4];
   ranPhiDec      = rand[5];

   const bool doDecay = includesDecay(signature);
   // vector boson invariant mass squared
   double bosonInvMassSquared;
   jac *= generateInvMassSquared(ranMass, doDecay, bosonInvMassSquared);

   // partonic rescaled CM energy
   double tau;
   double jetInvMassCut = 0.0;
   double jetKtCut = _ktJetCut + ACCURACY_THRESHOLD;

   jac *= generateTau(ranTau, bosonInvMassSquared, jetInvMassCut, jetKtCut, tau);
   double partonicS = tau *  _cmEnergy * _cmEnergy;
   // rapidity of partonic CM frame
   double y;
   double tmp = 0.5 * log(partonicS / bosonInvMassSquared);
   assert(std::isfinite(tmp));
   jac *= generateY(ranY, tau, _yBcutLow - tmp, _yBcutHigh + tmp, y);
   // jacobian factor coming from dPhi3 = ds12/(2pi) dPhi2(tau*S,s12,0) dPhi2(s12,0,0)
   // The following formula is divided by only pi and not 2pi beacuse the BW is normalized to pi
   jac *= (1. - bosonInvMassSquared / partonicS) / (math_constants::pi() * math_constants::_8pi() * math_constants::_8pi());
   // if in the NWA, divide by the full 2pi, since the delta function is normalized to 2pi
   if (!doDecay) jac /= 2.0;

   // the emission angle of the extra parton, importance sampled around -1,1
   double ran = 1. - 2.*ranCosTheta;
   double ranTmp;
   if (_importanceSampleThetaJet == kImportanceThetaJet) {
      ranTmp = 1.5 * (ran - pow(ran, 3) / 3.);
      jac *= 3. * (1. - ran * ran);
   } else {
      ranTmp = ran;
      jac *= 2;
   }
   // Restrict cosTheta due to the cuts on jet transverse momentum
   double jetEnergyCMFrame = (partonicS - bosonInvMassSquared) / (2.0 * sqrt(partonicS));
   double cosThetaHighJetKtCut = sqrt(abs(1.0 - pow(_ktJetCut / jetEnergyCMFrame, 2)));
   double cosThetaLowJetKtCut = -cosThetaHighJetKtCut;

   // Restrict cosTheta to the cuts on B rapidity
   double bosonEnergyCMFrame = sqrt(partonicS) - jetEnergyCMFrame;
   double cosThetaHighBrap = -bosonEnergyCMFrame / jetEnergyCMFrame * tanh(y-_yBcutHigh);
   double cosThetaLowBrap = -bosonEnergyCMFrame / jetEnergyCMFrame * tanh(y-_yBcutLow);

   double cosThetaHigh = std::min(1.0, std::min(cosThetaHighJetKtCut, cosThetaHighBrap));
   double cosThetaLow = std::max(-1.0, std::max(cosThetaLowJetKtCut, cosThetaLowBrap));
   assert(cosThetaHigh>cosThetaLow);

   double ave = 0.5 * (cosThetaLow + cosThetaHigh);
   double diff = cosThetaHigh - cosThetaLow;
   double cosTheta = ave + ranTmp * 0.5 * diff;
   jac *= 0.5 * diff;

   //  boson decay kinematics
   double cosThetaDec = 1. - 2.*ranCosThetaDec;
   jac *= 2.;
   double phiDec = math_constants::_2pi() * ranPhiDec;
   jac *= math_constants::_2pi();

   // momenta fractions for incoming partons
   double x1, x2;
   setxBarAxBarB(tau, y, x1, x2);

   // boost from Lab to partonic CM frame
   BoostZHyp zBoost(sqrt(x1/x2));

   // Overall random rotation around the beam axis
   RotationZ zRotation(ranRotationAzimuth);

   // construct kinematics in partonic CM frame
   std::vector<FourVector> kinematics(5, FourVector());
   double eTmp = sqrt(partonicS) / 2.;

   kinematics[0].setComponents(eTmp, 0., 0., eTmp);
   kinematics[1].setComponents(eTmp, 0., 0., -eTmp);

   // get IDs for final particles; if boson is not decayed, it will be EW1 and EW2 will stay as kUnknown
   PdgID idEW1(PdgID::kUnknown), idEW2(PdgID::kUnknown), idEmitted(PdgID::kUnknown);
   for (Signature::iterator it = signature.beginFinal(); it != signature.endFinal(); ++it) {
      if (IsQCD()(*it))
         idEmitted = *it;
      else {
         if (idEW1 == PdgID::kUnknown)
            idEW1 = *it;
         else
            idEW2 = *it;
      }
   }
   eTmp = 0.5 * sqrt(bosonInvMassSquared);
   double sinThetaDec = sqrt((1.-cosThetaDec)*(1.+cosThetaDec));
   double xTmp = eTmp * sinThetaDec * sin(phiDec);
   double yTmp = eTmp * sinThetaDec * cos(phiDec);
   double zTmp = eTmp * cosThetaDec;
   kinematics[2].setComponents(eTmp, xTmp, yTmp, zTmp);
   kinematics[3].setComponents(eTmp, -xTmp, -yTmp, -zTmp);

   double betaV = (partonicS - bosonInvMassSquared) / (partonicS + bosonInvMassSquared);
   Boost  boostV(ThreeVector(betaV * sqrt(1. - pow(cosTheta, 2)), 0., betaV * cosTheta));
   kinematics[2].apply(boostV);
   kinematics[3].apply(boostV);
   kinematics[4] = FourVector(sqrt(partonicS) - kinematics[2].p0() - kinematics[3].p0(), -(kinematics[2].pvec() + kinematics[3].pvec()));

   // sanity check
   assert((kinematics[4].perp() + ACCURACY_THRESHOLD) > _ktJetCut);

   // Boost momenta from CM to Lab frame
   for (std::vector<FourVector>::iterator it = kinematics.begin(); it != kinematics.end(); ++it)
      it->apply(zBoost);

   // Overall rotation along z-axis
   kinematics[2].apply(zRotation);
   kinematics[3].apply(zRotation);
   kinematics[4].apply(zRotation);

   // construct the required configuration
   configuration.clear();
   configuration.setInitialA(Particle(*signature.initialA(), kinematics[0]));
   configuration.setInitialB(Particle(*signature.initialB(), kinematics[1]));
   if (doDecay) {
      configuration.insertFinal(Particle(idEW1, kinematics[2]));
      configuration.insertFinal(Particle(idEW2, kinematics[3]));
   } else {
      FourVector kV = kinematics[2] + kinematics[3];
      configuration.insertFinal(Particle(idEW1, kV));
   }
   configuration.insertFinal(Particle(idEmitted, kinematics[4]));

   // check that B rapidity cuts are implemented correctly
   double yB = bosonRapidity(configuration);
   if ((yB < _yBcutLow) || (yB > _yBcutHigh))
      Log::error("GeneratorPP2BosonJets::generateBj") << "B rapidity " << yB <<" outside ["<<_yBcutLow<<","<<_yBcutHigh<<"] range." <<Log::end;

   assert(std::isfinite(jac));
   return true;
}

//------------------------------------------------------------------------------
// Generates a new Configuration with B and 1 parton in the final state, given external randoms as input.
bool GeneratorPP2BosonJets::generateBjTau0(const Signature& signature, Configuration& configuration, std::vector<double> rand, double& jac)
{
   assert(rand.size() == 6 || rand.size() == 7);
   configuration.clear();
   jac = 1.0;

   //TODO: Include generation cuts
   if (_ktJetCut > 0. || _yBcutHigh < 1000. || _yBcutLow > -1000.)
      Log::error("GeneratorPP2BosonJets::generateBjTau0") << "Cuts on kTJet and yB not implemented yet" << Log::end;

   double ranMass, ranY, ranTau0, ranZ;
   double ranCosThetaDec, ranPhiDec;
   // The variables above completely specify the kinematics
   // of a 2->3 event at a hadron collider.
   // The following is an overall rotation around the z-axis, whose coefficient
   // does not enter the jacobian and which is irrelevant for the integration.
   // No physical observable can depend on this, it is only customary not
   // to have the outgoing parton going always around the y-axis.

   //If we have specified 7 random numbers, use the last one for the azimuthal rotation. Otherwise generate a new random number
   double phi;
   if (rand.size() == 6)
      phi= randomAngle();
   else
      phi = math_constants::_2pi() * rand[6];

   jac = 2. * math_constants::pi();

   ranMass        = rand[0];
   ranY           = rand[1];
   ranTau0        = rand[2];
   ranZ           = rand[3];
   ranCosThetaDec = rand[4];
   ranPhiDec      = rand[5];

   const bool doDecay = includesDecay(signature);
   // vector boson invariant mass squared
   double bosonInvMassSquared;
   jac *= generateInvMassSquared(ranMass, doDecay, bosonInvMassSquared);
   double mV = sqrt(bosonInvMassSquared);
   //Convert dm^2 into 2m dm
   jac /= 2. * mV;
   //Take out the fact that I am including a 16pi in the jacobian returned by generateInvMassSquared
   jac *= 16. * math_constants::pi();

   // rapidity of partonic CM frame
   double y = std::numeric_limits<double>::quiet_NaN();
   jac *= generateY(ranY, bosonInvMassSquared / _cmEnergy / _cmEnergy, _yBcutLow, _yBcutHigh, y);

   double tau0 = std::numeric_limits<double>::quiet_NaN();
   double expY = exp(y);
   double expAbsY = exp(abs(y));
   jac *= generateTau0(ranTau0, mV, expAbsY, tau0);
   if (jac == 0)
      return 0.;

   //  boson decay kinematics
   double cosThetaDec = 1. - 2.*ranCosThetaDec;
   jac *= 2.;
   double phiDec = math_constants::_2pi() * ranPhiDec;
   jac *= math_constants::_2pi();


   std::string rad = "ISRA";
   if (randomInterval() < 0.5)
      rad = "ISRB";
   jac *= 2.;
   double z = std::numeric_limits<double>::quiet_NaN();
   jac *= generateZ(ranZ, mV, expY, tau0, rad, z);
   if (jac == 0)
      return 0.;


   // get IDs for final particles; if boson is not decayed, it will be EW1 and EW2 will stay as kUnknown
   PdgID idEW1(PdgID::kUnknown), idEW2(PdgID::kUnknown), idEmitted(PdgID::kUnknown);
   for (Signature::iterator it = signature.beginFinal(); it != signature.endFinal(); ++it) {
      if (IsQCD()(*it))
         idEmitted = *it;
      else {
         if (idEW1 == PdgID::kUnknown)
            idEW1 = *it;
         else
            idEW2 = *it;
      }
   }
   configuration.clear();
   double pPlus, pMinus, pPerpSq, qPlus, qMinus;
   if (rad == "ISRA") {
      //      if (z < phi0.initialA()->p().minus() / _cmEnergy)
      //         return false;
      pPlus = tau0 / expY;
      pMinus = mV * expY * (1. / z - 1);
      pPerpSq = pPlus * pMinus;
      qPlus = sqrt(pPerpSq + mV * mV) / expY;
      qMinus = sqrt(pPerpSq + mV * mV) * expY;
      if (!(pPlus + qPlus <= _cmEnergy) || !(pMinus + qMinus <= _cmEnergy) || !(pMinus / pPlus >= expY * expY))
         return 0.;
      configuration.setInitialA(Particle(*signature.initialA(), (pMinus + qMinus) / 2., 0., 0., (pMinus + qMinus) / 2.));
      configuration.setInitialB(Particle(*signature.initialB(), (pPlus + qPlus) / 2., 0., 0., -(pPlus + qPlus) / 2.));

   }
   else if (rad == "ISRB") {
      //      if (z < phi0.initialB()->p().plus() / _cmEnergy)
      //         return false;
      pMinus = tau0 * expY;
      pPlus = mV / expY * (1. / z - 1);
      pPerpSq = pPlus * pMinus;
      qPlus = sqrt(pPerpSq + mV * mV) / expY;
      qMinus = sqrt(pPerpSq + mV * mV) * expY;
      if (!(pPlus + qPlus <= _cmEnergy) || !(pMinus + qMinus <= _cmEnergy) || !(pMinus / pPlus <= expY * expY))
         return 0.;
      configuration.setInitialA(Particle(*signature.initialA(), (pMinus + qMinus) / 2., 0., 0., (pMinus + qMinus) / 2.));
      configuration.setInitialB(Particle(*signature.initialB(), (pPlus + qPlus) / 2., 0., 0., -(pPlus + qPlus) / 2.));
   }
   else {
      pMinus = std::numeric_limits<double>::quiet_NaN();
      pPlus = std::numeric_limits<double>::quiet_NaN();
      pPerpSq = std::numeric_limits<double>::quiet_NaN();
      qPlus = std::numeric_limits<double>::quiet_NaN();
      qMinus = std::numeric_limits<double>::quiet_NaN();
      Log::error("GeneratorPP2BosonJets::generateBjTau0") << "rad = " << rad << " needs to be ISRA or ISRB" << Log::end;
   }

   //Get the transverse components
   double px = sqrt(pPerpSq) * sin(phi);
   double py = sqrt(pPerpSq) * cos(phi);

   //Now add the new parton to the final state
   configuration.insertFinal(Particle(idEmitted, (pPlus + pMinus) / 2., px, py, (pMinus - pPlus) / 2.));

   //Now add the leptons to the final state
   double eTmp = sqrt(bosonInvMassSquared) / 2.;
   double xTmp = sqrt(1. - pow(cosThetaDec, 2)) * sin(phiDec) * eTmp;
   double yTmp = sqrt(1. - pow(cosThetaDec, 2)) * cos(phiDec) * eTmp;
   double zTmp = cosThetaDec * eTmp;

   std::vector<FourVector> kinematics(2, FourVector());
   kinematics[0].setComponents(eTmp, xTmp, yTmp, zTmp);
   kinematics[1].setComponents(eTmp, -xTmp, -yTmp, -zTmp);

   FourVector q((qPlus + qMinus) / 2., -px, -py, (qMinus - qPlus) / 2.);
   Boost toNewLab = Boost(q);
   kinematics[0].apply(toNewLab);
   kinematics[1].apply(toNewLab);
   configuration.insertFinal(Particle(idEW1, kinematics[0]));
   configuration.insertFinal(Particle(idEW2, kinematics[1]));

   double jac1 = (2. * mV * mV) / (z * z * _cmEnergy * _cmEnergy) / (256. * math_constants::pi4()) / 4. / math_constants::pi();
   //   double jac2 = 1. / 16. / math_constants::pi2() * mV / _cmEnergy / _cmEnergy;
   jac *= jac1;// / jac2;

   assert(std::isfinite(jac));
   return true;
}

//------------------------------------------------------------------------------
// Generates a new Configuration with  B and 2 partons in the final state, given external randoms as input.
bool GeneratorPP2BosonJets::generateBjj(const Signature& signature, Configuration& configuration, std::vector<double> rand, double& jac)
{
   assert(rand.size() == 9 || rand.size() == 10);
   configuration.clear();
   jac = 1.0;

   int samplingDijetMassSquared = 1;
   double samplingCosTheta = 1;
   double samplingCosThetaJets = 1.;
   double samplingCosThetaDec = 1.;

   double ranBosonMass, ranDijetMass, ranY, ranCosTheta;
   double ranTau, ranCosThetaDec, ranPhiDec, ranCosThetaJets, ranPhiJets;
   // The variables above completely specify the kinematics
   // of a 2->4 event at a hadron collider.

   // The following is an overall rotation around the z-axis, whose coefficient
   // does not enter the jacobian and which is irrelavant for the integration.
   // No physical observable can depend on this.

   //If we have specified 10 random numbers, use the last one for the azimuthal rotation. Otherwise generate a new random number
   double rand10;
   if (rand.size() == 9)
      rand10 = randomAngle();
   else
      rand10 = math_constants::_2pi() * rand[9];

   ranBosonMass   = rand[0];
   ranTau         = rand[1];
   ranY           = rand[2];
   ranDijetMass   = rand[3];
   ranCosTheta    = rand[4];
   ranCosThetaJets = rand[5];
   ranPhiJets     = rand[6];
   ranCosThetaDec = rand[7];
   ranPhiDec      = rand[8];

   std::vector<FourVector> kinematics(6, FourVector());

   const bool doDecay = includesDecay(signature);
   // vector boson invariant mass squared
   double bosonInvMassSquared;
   jac *= generateInvMassSquared(ranBosonMass, doDecay, bosonInvMassSquared);

   assert(jac > 0.0);
   // partonic rescaled CM energy
   double tau;
   const double dijetMassSquaredCut = _mJetJetCut * _mJetJetCut + ACCURACY_THRESHOLD;
   double ktVCut = ACCURACY_THRESHOLD;
   jac *= generateTau(ranTau, bosonInvMassSquared, dijetMassSquaredCut, ktVCut, tau);
   double partonicS = tau *  _cmEnergy * _cmEnergy;
   assert(jac > 0.0);
   // dijet invariant mass
   const double dijetMassSquaredMax = pow(sqrt(partonicS) - sqrt(bosonInvMassSquared), 2);

   //Importance sample the dijet invariant mass according to samplingDijetMassSquared
   const double dijetMassSquared = dijetMassSquaredCut + pow(ranDijetMass, samplingDijetMassSquared) * (dijetMassSquaredMax - dijetMassSquaredCut);
   //sanity check
   assert(sqrt(dijetMassSquared) >= _mJetJetCut);
   jac *= abs(samplingDijetMassSquared * pow(ranDijetMass, samplingDijetMassSquared - 1) * (dijetMassSquaredMax - dijetMassSquaredCut));
   assert(jac > 0.0);

   // rapidity of partonic CM frame
   double y;
   double tmpA = partonicS*partonicS + (dijetMassSquared - bosonInvMassSquared)*(dijetMassSquared - bosonInvMassSquared) - 2.*partonicS * (dijetMassSquared + bosonInvMassSquared);
   if (compareToZero(tmpA, ACCURACY_THRESHOLD) == -1) {
      Log::error("GeneratorPP2BosonJets::generateBjj") << "tmpA=" <<std::setprecision(16)<< tmpA << Log::end;
   }
   tmpA=sqrt(abs(tmpA));
   double tmpB = partonicS + bosonInvMassSquared - dijetMassSquared;
   double tmp = 0.5 * log((tmpB + tmpA)/(tmpB - tmpA));
   assert(std::isfinite(tmp));
   jac *= generateY(ranY, tau, _yBcutLow - tmp, _yBcutHigh + tmp , y);
   assert(jac > 0.0);

   const double bosonEnergyCMFrame = (partonicS + bosonInvMassSquared - dijetMassSquared) / (2.*sqrt(partonicS));
   const double momCMFrameSquared = pow(bosonEnergyCMFrame, 2) - bosonInvMassSquared;

   // Allow for small rounding errors that result in negative values
   if (compareToZero(momCMFrameSquared, ACCURACY_THRESHOLD) == -1) {
      Log::error("GeneratorPP2BosonJets::generateBjj") << "momCMFrameSquared=" << momCMFrameSquared << Log::end;
   }
   double momCMFrameMod = sqrt(abs(momCMFrameSquared));
   // taking the square root of too small a quantity results in  being rounded up to zero
   if (momCMFrameMod <= 0.0) momCMFrameMod += ACCURACY_THRESHOLD;
   // the boson emission angles in partonic CM frame
   // Sample it towards small values
   double ranTmp;
   jac *= sampleCosTheta(ranCosTheta, samplingCosTheta, ranTmp);
   double cosThetaHigh = std::min(1.0, -bosonEnergyCMFrame / momCMFrameMod * tanh(y-_yBcutHigh));
   double cosThetaLow = std::max(-1.0, -bosonEnergyCMFrame / momCMFrameMod * tanh(y-_yBcutLow));
   double ave = 0.5 * (cosThetaLow + cosThetaHigh);
   double diff = cosThetaHigh - cosThetaLow;
   double cosThetaBoson = ave + ranTmp * 0.5 * diff;
   jac *= 0.5 * diff;

   const double sinThetaBoson = sqrt(1.0 - cosThetaBoson * cosThetaBoson);
   const ThreeVector momCMFrame(momCMFrameMod * sinThetaBoson, 0.0, momCMFrameMod * cosThetaBoson);

   // jacobian factor coming from dPhi4 = ds12/(2pi) ds34/(2pi) dPhi2(tau*S,s12,s34) dPhi2(s12,0,0) dPhi2(s34,0,0)
   jac *=  2.0 * math_constants::_1_16pi2() * math_constants::_1_16pi2() * math_constants::_1_16pi2() * momCMFrameMod / sqrt(partonicS);

   // if in the NWA, divide by the full 2pi, since the delta function is normalized to 2pi
   if (!doDecay) jac /= 2.0;

   assert(jac > 0.0);
   // momenta fractions for incoming partons
   double x1, x2;
   setxBarAxBarB(tau, y, x1, x2);

   // set initial state momenta in Lab frame
   kinematics[0].setComponents(x1 * _cmEnergy / 2., 0., 0., x1 * _cmEnergy / 2.);
   kinematics[1].setComponents(x2 * _cmEnergy / 2., 0., 0., -x2 * _cmEnergy / 2.);

   //  decay kinematics in boson rest frame
   double cosThetaDec;
   jac *= sampleCosTheta(ranCosThetaDec, samplingCosThetaDec, cosThetaDec);
   const double phiDec = math_constants::_2pi() * ranPhiDec;
   jac *= math_constants::_2pi();
   assert(jac > 0.0);

   // get IDs for final particles; if boson is not decayed, it will be EW1 and EW2 will stay as kUnknown
   PdgID idEW1(PdgID::kUnknown), idEW2(PdgID::kUnknown);
   std::vector<PdgID> idEmitted;
   for (Signature::iterator it = signature.beginFinal(); it != signature.endFinal(); ++it) {
      if (IsQCD()(*it))
         idEmitted.push_back(*it);
      else {
         if (idEW1 == PdgID::kUnknown)
            idEW1 = *it;
         else
            idEW2 = *it;
      }
   }

   // construct boson decay products in its rest frame
   const double eLep = sqrt(bosonInvMassSquared) / 2.;
   const double sinThetaDec = sqrt(1. - pow(cosThetaDec, 2));
   const double xLep = sinThetaDec * sin(phiDec) * eLep;
   const double yLep = sinThetaDec * cos(phiDec) * eLep;
   const double zLep = cosThetaDec * eLep;

   kinematics[2].setComponents(eLep, xLep, yLep, zLep);

   // boost boson decay products to partonic CM frame
   Boost boostV(FourVector(bosonEnergyCMFrame, momCMFrame));
   kinematics[2].apply(boostV);

   // construct other lepton momentum by subtraction, to reduce numerical errors
   kinematics[3].setComponents(bosonEnergyCMFrame - kinematics[2].p0(), momCMFrame.p1() - kinematics[2].p1(), momCMFrame.p2() - kinematics[2].p2(), momCMFrame.p3() - kinematics[2].p3());

   // boost from partonic CM frame to Lab
   BoostZHyp zBoost(sqrt(x1/x2));
   // boost boson decay products  to Lab frame
   kinematics[2].apply(zBoost);
   kinematics[3].apply(zBoost);

   // Build dijet system in its rest frame
   const double eJet = sqrt(dijetMassSquared) / 2.;
   double cosThetaJets;
   jac *= sampleCosTheta(ranCosThetaJets, samplingCosThetaJets, cosThetaJets);
   const double phiJets = math_constants::_2pi() * ranPhiJets;
   jac *= math_constants::_2pi();
   double sinThetaJets = sqrt(1. - pow(cosThetaJets, 2));
   const double xJet = sinThetaJets * sin(phiJets) * eJet;
   const double yJet = sinThetaJets * cos(phiJets) * eJet;
   const double zJet = cosThetaJets * eJet;
   // build first jet in dijet rest frame
   kinematics[4].setComponents(eJet, xJet, yJet, zJet);

   ThreeVector vecJJLab = kinematics[0].pvec() + kinematics[1].pvec() - (kinematics[2].pvec() + kinematics[3].pvec());
   //Compute energy from invariant mass
   double energyJJLab = sqrt(dijetMassSquared + vecJJLab.square());
   Boost boostJJ(FourVector(energyJJLab, vecJJLab));
   // boost first jet from dijet-rest frame to CM frame
   kinematics[4].apply(boostJJ);
   // construct second jet momentum in CM frame by subtraction, to reduce numerical errors
   kinematics[5].setComponents(energyJJLab - kinematics[4].p0(), vecJJLab.p1() - kinematics[4].p1(), vecJJLab.p2() - kinematics[4].p2(), vecJJLab.p3() - kinematics[4].p3());

   // enforce the kt cuts on final state jets
   if (((kinematics[4].perp() + ACCURACY_THRESHOLD) < _ktJetCut) || ((kinematics[5].perp() + ACCURACY_THRESHOLD) < _ktJetCut)) {
//      std::cout << std::endl << "mJJ = " << sqrt(dijetMassSquared) << " > " << _mJetJetCut << std::endl;
//      std::cout << "pT1 = " << kinematics[4].perp() << " pT2 = " << kinematics[5].perp() << " < " << _ktJetCut << std::endl;
      return false;
   }

   // Overall random rotation around the beam axis
   RotationZ zRotation(rand10);

   kinematics[2].apply(zRotation);
   kinematics[3].apply(zRotation);
   kinematics[4].apply(zRotation);
   kinematics[5].apply(zRotation);

   configuration.clear();
   // construct the required configuration
   configuration.setInitialA(Particle(*signature.initialA(), kinematics[0]));
   configuration.setInitialB(Particle(*signature.initialB(), kinematics[1]));
   if (doDecay) {
      configuration.insertFinal(Particle(idEW1, kinematics[2]));
      configuration.insertFinal(Particle(idEW2, kinematics[3]));
   } else {
      FourVector kV = kinematics[2] + kinematics[3];
      configuration.insertFinal(Particle(idEW1, kV));
   }
   size_t ii = 4;
   for (std::vector<PdgID>::iterator it = idEmitted.begin(); it != idEmitted.end(); ++it, ii++)
      configuration.insertFinal(Particle(*it, kinematics[ii]));

   // check that B rapidity cuts are implemented correctly
   double yB = bosonRapidity(configuration);
   if ((yB < _yBcutLow) || (yB > _yBcutHigh))
      Log::error("GeneratorPP2BosonJets::generateBjj") << "B rapidity " << yB <<" outside ["<<_yBcutLow<<","<<_yBcutHigh<<"] range." <<Log::end;

   //Check for potential cuts on Tau1
   if (_bjjTau1cut != "none") {
      double tau1 = NJettinessppRecursive(configuration)(1);
      if (_bjjTau1cut == "above" && tau1 < _tau1cut)
         return false;
      else if (_bjjTau1cut == "below" && tau1 > _tau1cut)
         return false;
   }
   assert(std::isnormal(jac));
   return true;
}

// ------------------------------------------------------------------------------
// Generates a new Configuration with B and 2 or more partons in the final state using generateBjj and RAMBO.
// TODO: enforce the cuts on ktJet and mJetJet
// TODO: validate against analytical results V2, V3, V4 and V5 coded in GeneratorPPBosonJetsTest
bool GeneratorPP2BosonJets::generateBjetsRAMBO(const Signature& signature, Configuration& configuration, std::vector<double> rand, double& jac)
{
   //TODO: make sure this function is no longer used for large multiplicities and can be safely deprecated
   Log::fatal("generateBjetsRAMBO") <<"This function would need to be revalidated before usage."<<Log::end;

   assert(signature.size() >= 5);
   assert(rand.size()>= 12);
   configuration.clear();
   jac = 1.0;

   Signature sigBjj;
   sigBjj.setInitialA(*signature.initialA());
   sigBjj.setInitialB(*signature.initialB());
   for (Signature::iterator it = signature.beginFinal(); it != signature.endFinal(); ++it)
      if (!IsQCD() (*it)) sigBjj.insertFinal(*it);
   // the correct PdgID's will be included later
   sigBjj.insertFinal(PdgID::kUnknown);
   sigBjj.insertFinal(PdgID::kUnknown);

   Configuration configBjj;
   bool success = false;
   std::vector<double>::iterator itr = rand.begin()+9;
   double jacBjj = 0;
   success = generateBjj(sigBjj, configBjj, std::vector<double>(rand.begin(),itr), jacBjj);
   if (!success)
      return false;

   jac*=jacBjj;
   const BoostZHyp toCMframe(sqrt(configBjj.initialB()->energy()/ configBjj.initialA()->energy()));
   // boost to CM frame
   configBjj.apply(toCMframe);
#ifndef NDEBUG
   if (!isInCoMFrame(configBjj, 1e-8, 1e-10))
      Log::warning("GeneratorPP2BosonJets::generateBjetsRAMBO") << "configBjj is not in CM frame." << Log::end;
#endif

   FourVector diJetMomentum;

   // remove the fake particles, storing their total FourVector in CM frame
   for (Configuration::iterator it = configBjj.beginFinal(); it != configBjj.endFinal(); ++it) {
      if (it->id() == PdgID::kUnknown) {
         diJetMomentum += it->p();
      }
   }
   configBjj.erase(PdgID::kUnknown);

   Boost toDiJetRestFrame(diJetMomentum);

   // sanity checks
   assert(std::count_if(configBjj.beginFinal(), configBjj.endFinal(), IsQCD()) == 0);
   assert(configBjj.findFinal(PdgID::kUnknown) == configBjj.end());

   // generate the extra partons in dijet rest frame
   const bool doDecay = includesDecay(signature);
   size_t nFinalExtra = doDecay ?  signature.sizeFinal() - 2 : signature.sizeFinal() - 1;
   assert(rand.size() == 9+(nFinalExtra*4));
   std::vector<FourVector> q;
   for (size_t i = 0; i < nFinalExtra; i++) {
      double c = -1.0 + 2.0 * (* ++itr);
      double s = sqrt(1.0 - c * c);
      double f = math_constants::_2pi() * (* ++itr);
      double r12 = 0.0;
      while (r12 == 0.0) {
         double r1 = (* ++itr);
         double r2 = (* ++itr);
         r12 = r1 * r2;
      }
      double En = -log(r12);
      q.push_back(FourVector(En, En * s * cos(f), En * s * sin(f), En * c));
   }

   std::vector<FourVector> pNewRambo = GeneratorUtils::physicalFV(q, diJetMomentum.magnitude());
   std::vector<double> eNewRambo;
   std::vector<double> mRambo;
   std::vector<PdgID> idRambo;

   Signature sigFinalExtra = signature - sigBjj;
   size_t i = 0;
   for (Signature::iterator it = sigFinalExtra.beginFinal(); it != sigFinalExtra.endFinal(); ++it, ++i) {
      // only includes QCD partons not already generated by generateBjj
      //TODO: extend to massive case
      mRambo.push_back(0.0);
      eNewRambo.push_back(pNewRambo[i].p0());
      idRambo.push_back(*it);
   }

   //Find the parameter xi to include the masses
   double xi = GeneratorUtils::findxi(diJetMomentum.square(), eNewRambo, mRambo, _solver);
   // Correct the momenta in pnew for masses
   GeneratorUtils::massMomentaFactor(xi, mRambo, pNewRambo);

   // Randomize the momenta assignments, in order to guarantee more uniform distributions
   RandomInteger randInt;
   random_shuffle(pNewRambo.begin(), pNewRambo.end(), randInt);

   // calculate RAMBO weight
   const int n = int_from_size_t(nFinalExtra);
   double w0 = pow(math_constants::pi(), 2 * n - 3) * pow(2., 4 * n - 5) * pow(diJetMomentum.square(), 2 - n)
               * factorial(unsigned(n - 1)) * factorial(unsigned(n - 2)) / static_cast<double>(_sigSet.size());
   double wm = GeneratorUtils::massWeightFactor(pNewRambo, diJetMomentum.magnitude(), nFinalExtra);

   double weight = w0 * wm;
   jac *= weight;

   // boost back in CM frame
   for (std::vector<FourVector>::iterator it = pNewRambo.begin(); it != pNewRambo.end(); ++it)
      it->apply(toDiJetRestFrame.inverse());

   // handy container for event kinematics
   std::vector<FourVector> kinematics;
   // initial particle
   kinematics.push_back(configBjj.initialA()->p());
   kinematics.push_back(configBjj.initialB()->p());
   // decay products
   for (Configuration::iterator it = configBjj.beginFinal(); it != configBjj.endFinal(); ++it)
      kinematics.push_back(it->p());
   // everything else
   kinematics.insert(kinematics.end(), pNewRambo.begin(), pNewRambo.end());

   // construct the required configuration
   configuration.setInitialA(Particle(configBjj.initialA()->id(), kinematics[0]));
   configuration.setInitialB(Particle(configBjj.initialB()->id(), kinematics[1]));
   i = 0;
   for (Configuration::iterator it = configBjj.beginFinal(); it != configBjj.endFinal(); ++it, i++)
      configuration.insertFinal(Particle(it->id(), kinematics[2 + i]));
   size_t j = i;
   for (Signature::iterator it = sigFinalExtra.beginFinal(); it != sigFinalExtra.endFinal(); ++it, j++)
      configuration.insertFinal(Particle(*it, kinematics[j]));

#ifndef NDEBUG
   if (!isInCoMFrame(configuration, 1e-6, 1e-8))
      Log::error("GeneratorPP2BosonJets::generateBjetsRAMBO") << "configuration is not in CM frame." << Log::end;
#endif

   configuration.apply(toCMframe.inverse());
   assert(std::isnormal(jac));
   return true;
}

//------------------------------------------------------------------------------
// Generates a new Configuration with  V  and 2 or more parton in the final state using
// generateBj and applying a recursive FKS splitting procedure
bool GeneratorPP2BosonJets::generateBjjRecursiveFKS(const Signature& signature, Configuration& config, std::vector<double> rand, double& jac)
{
   assert(rand.size() == 9 || rand.size() == 10);
   config.clear();
   jac = 1.0;
   assert(_useBjjFKS);
   if (_idB != PdgID::kZ)
      Log::fatal("GeneratorPP2BosonJets::generateBjjRecursiveFKS") << " boson "<<_idB<< "not yet allowed"<<Log::end;
   // the code below assumes Z
   Signature::iterator firstLepton=std::find_if(signature.beginFinal(), signature.endFinal(), IsElectroWeak());
   int decay = abs(firstLepton->code());
   static RegionInfoList splittings1to2;
   if (!splittings1to2.isInitialized()) {
      static SignatureSet oneJet = SignatureManager().getSignaturespp2ZJets(PdgID(decay), 1, _nLight, true);
      static SignatureSet twoJets = SignatureManager().getSignaturespp2ZJets(PdgID(decay), 2, _nLight, true);
      splittings1to2.setRegionInfoList(oneJet, twoJets, _nLight);
   }
   assert(splittings1to2.isInitialized());

   //Pick a random mapping to go from 1-jet to 2-jets
   RegionInfoList::region_list_iterator randomMap;
   std::list<RegionInfo> mappings = splittings1to2.getSingularRegionsToSignature(signature);
   size_t mapSize = mappings.size();
   if (mapSize == 0) {
      // if there's no N-1 body signature that can bring to the current one,
      // we have to generate it separately - TODO: complete this for W+2 jets
      Log::fatal("GeneratorPP2BosonJets::generateBjjRecursiveFKS")<<"There is no signature that can bring to "<<signature<<Log::end;
   } else {
      // pick a random mapping
      randomMap = mappings.begin();
      std::advance(randomMap, ptrdiff_t_from_size_t(randomInteger(mapSize)));
   }

   //Create the PhiNcmFKS required for the mappings
   std::vector<double> rand1;
   for (size_t i = 0; i < 6; i++)
      rand1.push_back(rand[i]);
   if (rand.size() == 10)
      rand1.push_back(rand[9]);
   const Signature& sig1 = *randomMap->fN;
   Configuration config1;
   double jac1 = 1.0;
   if (_useTau0Generator) {
      if (!generateBjTau0(sig1, config1, rand1, jac1))
         return false;
   } else {
      if (!generateBj(sig1, config1, rand1, jac1))
         return false;
   }
   const BoostZHyp toCMframe(sqrt(config1.initialB()->energy() / config1.initialA()->energy()));
   Configuration config1CM(config1.transformed(toCMframe));
   PhiNcmFKS phi1(config1CM);

   //Create the UnitCubeRad
   UnitCubeRad xRad;
   //Set the flavor structures and splitting type
   xRad.idMother() = *randomMap->k;
   xRad.idEmitted() = *randomMap->i;
   xRad.idSister() = *randomMap->j;
   if (randomMap->hasFinalStateSingularity()) {
      xRad.radType() = UnitCubeRad::kFSR;
      assert(xRad.idEmitted().isParticle());
      assert(xRad.idEmitted().isGluon() || xRad.idSister().isAntiPartner(xRad.idEmitted()));
   } else if (randomMap->hasInitialStateSingularity()) {
      xRad.radType() = (randomMap->k->isInitialA()) ? UnitCubeRad::kISRA : UnitCubeRad::kISRB;
   } else {
      Log::fatal("GeneratorPP2BosonJets::generateBjjRecursiveFKS")<<"Radiation region unknown."<<Log::end;
   }

   //Create the xRad kinematics
   // TODO: read these values from the corresponding option rather than hardcoding them here
   double x1TinyCut = 1e-6;
   double x2TinyCut = 1e-6;
   xRad.weight() = 1.;
   xRad.x1() = x1TinyCut + rand[6] * (1. - x1TinyCut);
   xRad.weight() *= (1. - x1TinyCut);
   xRad.x2() = x2TinyCut + rand[7] * (1. - x2TinyCut);
   xRad.weight() *= (1. - x2TinyCut);
   xRad.x3() = rand[8];

   //Pick the emitter
   if (randomMap->k->isInitialA()) {
      phi1.emitter = phi1.configCM.initialA();
   } else if (randomMap->k->isInitialB()) {
      phi1.emitter = phi1.configCM.initialB();
   } else {
      // pick a random emitter between the equivalent ones
      std::pair<Configuration::iterator, Configuration::iterator> sameEmi = phi1.configCM.equalRangeFinal(*randomMap->k);
      phi1.emitter = sameEmi.first;
      std::advance(phi1.emitter, ptrdiff_t_from_size_t(randomInteger(size_t_from_ptrdiff_t(std::distance(sameEmi.first, sameEmi.second)))));
      // avoid an FSR emitter exactly along the beampipe, which will cause NaN in the jacobians
      if (phi1.emitter->p().perp2() <= 0.0)
         return false;
   }

   //Now do the splitting
   double xA = 0.;
   double xB = 0.;
   double jacSplit = 0.;
   Configuration config2CM;
   Configuration::iterator emitted, sister;
   double xBarA = 2. * config1.initialA()->energy() / _cmEnergy;
   double xBarB = 2. * config1.initialB()->energy() / _cmEnergy;

   if (!split(phi1.configCM, phi1.emitter, xBarA, xBarB, xRad, config2CM, emitted, sister, xA, xB, jacSplit))
      return false;

   const BoostZHyp toLabFrame2(sqrt(xA / xB));
   config = config2CM.transformed(toLabFrame2);

   //Check generation cuts
   double yV = bosonRapidity(config);
   if ((yV < _yBcutLow) || (yV > _yBcutHigh))
      return 0.;
   FourVector pJJ;
   for (Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it) {
      if (IsQCD()(*it)) {
         if (it->p().perp() < _ktJetCut)
            return 0.;
         else
            pJJ+=it->p();
      }
   }
   if (pJJ.magnitude() < _mJetJetCut)
      return 0.;

   //Check for potential cuts on Tau1
   if (_bjjTau1cut != "none") {
      double tau1 = NJettinessppRecursive(config2CM)(1);
      if (_bjjTau1cut == "above" && tau1 < _tau1cut)
         return 0.;
      else if (_bjjTau1cut == "below" && tau1 > _tau1cut)
         return 0.;
   }

   //Now deal with the multiple coverings of phase space
   //Since we have picked the mapping randomly. We now make sure that each singular
   //phase space region is covered predominantly from the appropriate splitting.
   //This effectively provides a 1 / N_map term, which should give the correct phase space volume.

   //The problem is that we can have accidentaly singularities, such that a phi2 phase space point with say
   //an ISR singularity is created from an FSR map. To deal with this, we should replace the 1 / N_map by a set of Sij functions

   //TODO: This currenty assumes a 2-body point

   double dMap = 0.;
   double dTotal = 0.;
   for (std::list<RegionInfo>::iterator map = mappings.begin(); map != mappings.end(); ++map) {
      double dThisMap = std::numeric_limits<double>::quiet_NaN();
      if (map->k->isFinal()) {
         //Sum over the 1 / dij of all final state particles
         dThisMap = 0.;
         for (Configuration::iterator it1 = config2CM.beginFinal(); it1 != config2CM.endFinal(); ++it1) {
            if (IsQCD()(it1->id())) {
               Configuration::iterator start2 = it1;
               std::advance(start2, 1);
               for (Configuration::iterator it2 = start2; it2 != config2CM.endFinal(); ++it2) {
                  if (IsQCD()(it2->id())) {
                     dThisMap += 1. / NJettinessMetricFSR()(it1->p(), it2->p());
                  }
               }
            }
         }
      }
      else if (map->k->isInitialA()) {
         //Sum over the 1 / diA of all final state particles
         dThisMap = 0.;
         for (Configuration::iterator it = config2CM.beginFinal(); it != config2CM.endFinal(); ++it) {
            if (IsQCD()(it->id())) {
               dThisMap += 1. / (it->p().p0() - it->p().p3());
            }
         }
      }
      else if (map->k->isInitialB()) {
         //Sum over the 1 / diB of all final state particles
         dThisMap = 0.;
         for (Configuration::iterator it = config2CM.beginFinal(); it != config2CM.endFinal(); ++it) {
            if (IsQCD()(it->id())) {
               dThisMap += 1. / (it->p().p0() + it->p().p3());
            }
         }
      }
      else {
         Log::error("GeneratorPP2BosonJets::generateBjjRecursiveFKS") << "The map has to be either FSR, ISRA or ISRB" << Log::end;
      }
      if (map == randomMap)
         dMap = dThisMap;
      dTotal += dThisMap;
   }

   xRad.weight() *= static_cast<double>(mapSize) * dMap / dTotal;
   jac *= jac1 * jacSplit * xRad.weight();
   assert(std::isnormal(jac));
   return true;
}

//------------------------------------------------------------------------------
double GeneratorPP2BosonJets::sampleCosTheta(double ran, double sampling, double& cosTheta)
{
   if (ran > 0.5) {
      cosTheta = - pow(2.*ran - 1., sampling);
      return sampling * 2.*pow(2.*ran - 1., sampling - 1.);
   } else {
      cosTheta = pow(1. - 2.*ran, sampling);
      return sampling * 2.*pow(1. - 2.*ran, sampling - 1.);
   }
}

//------------------------------------------------------------------------------
double GeneratorPP2BosonJets::bosonRapidity(const Configuration& config)
{
   Signature signature(config);
   if (!includesDecay(signature)) {
      return config.findFinal(_idB)->p().rapidity();
   } else {
      FourVector p;
      size_t countLep = 0;
      for (Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it) {
         if (it->id().isLepton()) {
            p += it->p();
            countLep++;
         }
      }
      if (countLep != 2)
         Log::error("GeneratorPP2BosonJets::bosonRapidity") << "Number of leptons in decayed Configuration different from 2:\n" << config << Log::end;
      return p.rapidity();
   }
}

//------------------------------------------------------------------------------
bool GeneratorPP2BosonJets::split(const Configuration& configNCM, const Configuration::iterator emitterCM, double xBarA, double xBarB, const UnitCubeRad& xRad, Configuration& configNplus1CM, Configuration::iterator& emittedCM, Configuration::iterator& sisterCM, double& xA, double& xB, double& jacobian)
{
   FKSMapping mapFKS;
   double y = -1. + 2.*xRad.x2();
   if (xRad.radType()==UnitCubeRad::kISRB) y *= -1.;
   double ximax = mapFKS.getXiMaxReal(PhiNcmFKS(configNCM, emitterCM), xBarA, xBarB, y);
   PhiRadFKS phiRad(xRad, ximax, jacobian);
   PhiNplus1FKS phiNplus1FKS;
   // test if the splitting is valid and get the split config
   if (mapFKS.generateSplit(PhiNcmFKS(configNCM, emitterCM), phiRad, xBarA, xBarB, phiNplus1FKS)) {
      configNplus1CM.moveFrom(phiNplus1FKS.config);
      emittedCM = phiNplus1FKS.FKSparton;
      sisterCM = phiNplus1FKS.sister;
      jacobian *= mapFKS.getJacobian();
      if (emitterCM->isFinal()) {
         xA = xBarA;
         xB = xBarB;
      } else {
         xA = mapFKS.getxA(xBarA, phiRad.xi(), phiRad.y());
         xB = mapFKS.getxB(xBarB, phiRad.xi(), phiRad.y());
      }
      return true;
   }
   // generateSplit should never fail for the FKS map
   Log::error("GeneratorPP2BosonJets::constructPhiNplus1") << "Failed FKS splitting" << Log::end;
   return false;
}

//------------------------------------------------------------------------------
double GeneratorPP2BosonJets::generateTau0(double ranTau0, double mV,  double expAbsY, double& tau0) {
   double tauMin = 0;
   double tauMax = (_cmEnergy * _cmEnergy / expAbsY - mV * mV * expAbsY) / 2. / _cmEnergy;
   //   tauMax = _cmEnergy;
   if (tauMax < tauMin)
      return 0.;
   tau0 = tauMin + ranTau0 * ranTau0 * ranTau0 * (tauMax - tauMin);
   return 3. * ranTau0 * ranTau0 * (tauMax - tauMin);
}

//------------------------------------------------------------------------------
double GeneratorPP2BosonJets::generateZ(double ranZ, double mV,  double expY, double tau0, std::string rad, double& z) {

   //Determine the lower limit of z
   double zMin = getZMin(mV, expY, tau0, rad);
   //Determine the upper limit of z
   double zMax = mV / (mV + tau0);
   if (zMax < zMin)
      return 0.;

//   zMin = 0.;
//   zMax = 1.;
   z = zMax + ranZ * ranZ * (zMin - zMax);
   return 2. * ranZ * (zMax - zMin);
}

//------------------------------------------------------------------------------
double GeneratorPP2BosonJets::getZMin(double mV,  double expY, double tau0, std::string rad) {
   //Determine the lower limit of z
   double zMin = 0.;
   //The formulas for ISRA and ISRB are the same, after the replacement YV->-YV
   if (rad == "ISRB") {
      expY = 1. / expY;
   }
   double zlim = ((mV * tau0) / (_cmEnergy * _cmEnergy * expY * expY - mV * mV - 2.*_cmEnergy * expY * tau0 + mV * tau0 + tau0 * tau0));
   if (zlim < 0) {
      return 0.;
   }
   if (zlim >= zMin)
      zMin = zlim;

   zlim = (2. * _cmEnergy * expY * mV + 2. * pow(expY, 2) * pow(mV, 2) + pow(expY, 2) * mV * tau0);
   zlim /= (2. * (pow(_cmEnergy, 2) + 2. * _cmEnergy * expY * mV + pow(expY, 2) * mV * tau0));
   zlim += sqrt((4. * pow(expY, 4) * pow(mV, 4) + 4. * _cmEnergy * pow(expY, 3) * pow(mV, 2) * tau0 + pow(expY, 4) * pow(mV, 2) * pow(tau0, 2)) /
                pow(pow(_cmEnergy, 2) + 2 * _cmEnergy * expY * mV + pow(expY, 2) * mV * tau0, 2)) / 2.;
   // NOTE: negative zlim means there is no solution in [0,1], so return a lower limit > 1. to indicate invalid phase space
   if (zlim < 0) {
      return 0.;
   }
   if (zlim >= zMin)
      zMin = zlim;

   return zMin;
}

} // namespace Geneva
