//------------------------------------------------------------------------------
/// \file GeneratorUtils.cpp
//
// Author(s):
//    Simon Berman, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2016 LBNL, DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class GeneratorUtils
//------------------------------------------------------------------------------

#include "Geneva/Generators/GeneratorUtils.hpp"

#include "Geneva/Core/Utils/Random.hpp"

#include <gsl/gsl_errno.h>

#include <limits>
#include <list>
#include <cassert>

namespace Geneva
{

//------------------------------------------------------------------------------
// tests if the given Signature is of the form e+e-.
bool GeneratorUtils::isEe(const Signature& sig)
{
   if (sig.initialA() == sig.end() || sig.initialB() == sig.end())
      return false;
   return ((*sig.initialA() == PdgID::kElectron && *sig.initialB() == PdgID::kPositron) ||
           (*sig.initialA() == PdgID::kPositron && *sig.initialB() == PdgID::kElectron));
}

namespace
{

//------------------------------------------------------------------------------
struct XiFunctParameters
{
   double Ecm;
   std::vector<double> E, m;
};

//------------------------------------------------------------------------------
double xifunct(double xi, void* params)
{
   XiFunctParameters* p = static_cast<XiFunctParameters*>(params);
   double result = 0.0;
   for (size_t i = 0; i < p->E.size(); i++) {
      result += sqrt(p->m[i]*p->m[i] + xi*xi * p->E[i]*p->E[i]);
   }
   result -= p->Ecm;
   return result;
}

//------------------------------------------------------------------------------
double xifunct_deriv(double xi, void* params)
{
   XiFunctParameters* p = static_cast<XiFunctParameters*>(params);
   double result = 0.0;
   for (size_t i = 0; i < p->E.size(); i++) {
      result += xi * p->E[i]*p->E[i] / sqrt(p->m[i]*p->m[i] + xi*xi * p->E[i]*p->E[i]);
   }
   return result;
}

//------------------------------------------------------------------------------
void xifunct_fdf(double xi, void* params, double* y, double* dy)
{
   XiFunctParameters* p = static_cast<XiFunctParameters*>(params);
   *y = 0.0;
   *dy = 0.0;
   for (size_t i = 0; i < p->E.size(); i++) {
      const double temp = sqrt(p->m[i]*p->m[i] + xi*xi * p->E[i]*p->E[i]);
      *y += temp;
      *dy += xi * p->E[i]*p->E[i] / temp;
   }
   *y -= p->Ecm;
}

} // anonymous namespace

//------------------------------------------------------------------------------
// Calculates the value xi required to included masses.
double GeneratorUtils::findxi(double Ecm, std::vector<double> E, std::vector<double> m, gsl_root_fdfsolver* s)
{
   int status;
   int iter = 0, max_iter = 100;
   double xi0, xi = 1.0;
   struct XiFunctParameters p = {Ecm,E,m};

   gsl_function_fdf FDF;
   FDF.f = &(xifunct);
   FDF.df = &(xifunct_deriv);
   FDF.fdf = &(xifunct_fdf);
   FDF.params = &p;
   gsl_root_fdfsolver_set(s, &FDF, xi);

   do {
      iter++;
      gsl_root_fdfsolver_iterate(s);
      xi0 = xi;
      xi = gsl_root_fdfsolver_root(s);
      status = gsl_root_test_delta(xi, xi0, 0, 1e-3);

      if (status == GSL_SUCCESS) {
      }
   } while (status == GSL_CONTINUE && iter < max_iter);

   return xi;
}

//------------------------------------------------------------------------------
// Uses xi to appropriately adjust FVs to include mass
std::vector<FourVector> GeneratorUtils::massMomentaFactor(double xi, std::vector<double> m, std::vector<FourVector>& zeroMass)
{
   for (size_t i = 0; i < zeroMass.size(); i++) {
      zeroMass[i].p0() = sqrt(m[i]*m[i]+xi*xi*zeroMass[i].p0()*zeroMass[i].p0());
      zeroMass[i].p1() *= xi;
      zeroMass[i].p2() *= xi;
      zeroMass[i].p3() *= xi;
   }
   return zeroMass;
}

//------------------------------------------------------------------------------
// Calculates factor for mass correction in differential weight
double GeneratorUtils::massWeightFactor(std::vector<FourVector> k, double Ecm, size_t numGenerated)
{
   double Wm;
   double bracket1 = 0.0;
   double bracket2 = 1.0;
   double bracket3 = 0.0;
   for (size_t i = 0; i < k.size(); i++) {
      const double tmp = k[i].pvec().magnitude();
      bracket1 += tmp/Ecm;
      bracket2 *= tmp/k[i].p0();
      bracket3 += tmp*tmp/k[i].p0();
   }
   Wm = Ecm * pow(bracket1, 2 * int_from_size_t(numGenerated) - 3) * bracket2 / bracket3;
   return Wm;
}

//------------------------------------------------------------------------------
// Boosts massless four momenta into a frame in which they satisfy energy conservation
/**
  * The implementation follows the RAMBO paper by Ellis, Kleiss, and Stirling
  */
std::vector<FourVector> GeneratorUtils::physicalFV(std::vector<FourVector> q, double Ecm)
{
   FourVector Q;
   for (size_t i = 0; i < q.size(); i++) {
      Q += q[i];
   }
   double M = Q.magnitude();
   double g = Q.p0()/M;
   double a = 1.0/(1.0+g);
   double x = Ecm / M;
   FourVector b(0.0,0.0,0.0,0.0);
   b.p1() = - Q.p1()/M;
   b.p2() = - Q.p2()/M;
   b.p3() = - Q.p3()/M;

   std::vector<FourVector> pnew;
   for (size_t i = 0; i < q.size(); i++) {
      const double bdotq = b.p1()*q[i].p1() + b.p2()*q[i].p2() + b.p3()*q[i].p3();
      FourVector newMomentum;
      newMomentum.p0() = x*(g*q[i].p0() + bdotq);
      newMomentum.p1() = x*(q[i].p1() + b.p1()*q[i].p0() + a*b.p1()*bdotq);
      newMomentum.p2() = x*(q[i].p2() + b.p2()*q[i].p0() + a*b.p2()*bdotq);
      newMomentum.p3() = x*(q[i].p3() + b.p3()*q[i].p0() + a*b.p3()*bdotq);
      pnew.push_back(newMomentum);
   }
   return pnew;
}

//------------------------------------------------------------------------------
// ???
int GeneratorUtils::getNewColorLabel(int& numColorLabel)
{
   int temp = (101+numColorLabel);
   numColorLabel++;
   return temp;
}

// -----------------------------------------------------------------------------
// Checks if id1 and id2 represent a valid QCD splitting
/**
 * Checks if the particle id's id1 and id2 can originate from a QCD splitting.
 */
bool GeneratorUtils::isValidSplit(PdgID idi, PdgID idj, int nTriplet)
{
   if (idi == PdgID::kGluon || idj == PdgID::kGluon)
      return true;
   if (idi.isQuark() && idi.isAntiPartner(idj) && (nTriplet > 1))
      return true;
   return false;
}

// -----------------------------------------------------------------------------
/**
 * \brief Generate the color information of a color history.
 *
 * Uses a recursive algorithm to run generateColor(coloredParticle) on each element in the linked list.
 */
void GeneratorUtils::generateColorFlowHistory(const Signature& sig, std::vector<int>& color, std::vector<int>& antiColor, int& numColorLabel)
{
   // create the color history as a linked list of splittings.
   _splitting_type* coloredMother = createColorHistory(sig, color, antiColor);

   // generate the color flow information from this linked list.
   generateColorFlow(coloredMother, color, antiColor, numColorLabel);
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

namespace
{
enum { kSinglet = 1, kTriplet = 3, kTripletBar = -3, kOctet = 8};
}

//------------------------------------------------------------------------------
struct GeneratorUtils::_splitting_type
{
   PdgID id;
   int colorCharge;  // 1, 3, -3, 8 for singlet, triplet, anti-triplet, and octet
   size_t sigLocation;
   int color;
   int antiColor;
   _splitting_type* left;
   _splitting_type* right;
};

//------------------------------------------------------------------------------
void GeneratorUtils::generateColorFlow(_splitting_type* coloredMother, std::vector<int>& color, std::vector<int>& antiColor, int& numColorLabel)
{
   assert(coloredMother); //Should never get here for coloredMother = 0

   if ((coloredMother->left) && (coloredMother->right)) {  //Check if mother has daughters
      generateColor(coloredMother, numColorLabel);    //If yes, generate color information
   } else {
      assert(coloredMother->sigLocation != std::numeric_limits<size_t>::max()); //If no, has to be part of the original signature
      color[coloredMother->sigLocation] = coloredMother->color; //Write out color
      antiColor[coloredMother->sigLocation] = coloredMother->antiColor; //Write out color
   }
   if ((coloredMother->left->left) && (coloredMother->left->right)) {  //Check if left daughter has daughters
      generateColorFlow(coloredMother->left, color, antiColor, numColorLabel);    //If yes, generate color information
   } else {
      assert(coloredMother->left->sigLocation != std::numeric_limits<size_t>::max()); //If no, has to be part of the original signature
      color[coloredMother->left->sigLocation] = coloredMother->left->color;//Write out color
      antiColor[coloredMother->left->sigLocation] = coloredMother->left->antiColor; //Write out color
      delete coloredMother->left; //Clean up after ourselves
   }
   if ((coloredMother->right->left) && (coloredMother->right->right)) {  //Check if right daughter has daughters
      generateColorFlow(coloredMother->right, color, antiColor, numColorLabel);    //If yes, generate color information
   } else {
      assert(coloredMother->right->sigLocation != std::numeric_limits<size_t>::max()); //If no, has to be part of the original signature
      color[coloredMother->right->sigLocation] = coloredMother->right->color;//Write out color
      antiColor[coloredMother->right->sigLocation] = coloredMother->right->antiColor;//Write out color
      delete coloredMother->right; //Clean up after ourselves
   }
   delete coloredMother; //Clean up after ourselves
}

//------------------------------------------------------------------------------
/**
 * \brief Create the color history and return the start of the linked list.
 *
 * An algorithm for randomly finding a  splitting history that is allowed by
 * the interactions of QCD. It uses a recombination algorithm to
 * construct a possible history and assign the color flow of this history in the
 * large N limit
 * TODO: is this algorithm efficient?
 */
GeneratorUtils::_splitting_type* GeneratorUtils::createColorHistory(const Signature& sig, std::vector<int>& color, std::vector<int>& antiColor)
{
   //Get the number of final state particles
   size_t numParticles = sig.sizeFinal();
   //Count the number of quarks and antiquarks
   int nTriplet = 0;

   std::list<_splitting_type*> coloredList;

   //First, go through all particles in the signature and put the information of each particle
   //into a struct that will hold all information required. Note that these are not splittings
   //yet, but the struct contains the pointers that hold the splitting information.
   color.resize(numParticles);
   antiColor.resize(numParticles);
   size_t i = 0;
   for (Signature::iterator it = sig.beginFinal(); it != sig.endFinal(); ++it, ++i) {
      color[i] = 0;
      antiColor[i] = 0;
      _splitting_type* temp = new _splitting_type;
      temp->id = *it;
      if (temp->id.isQuark()) {
         if (temp->id.isParticle())
            temp->colorCharge = kTriplet;
         else
            temp->colorCharge = kTripletBar;
      } else if (temp->id == PdgID::kGluon)
         temp->colorCharge = kOctet;
      else
         temp->colorCharge = 1;
      temp->sigLocation = i;
      temp->color = 0;
      temp->antiColor = 0;
      temp->left = 0;
      temp->right = 0;
      coloredList.push_back(temp);

      if (it->isQuark() && it->isParticle()) {
         nTriplet++; //Count the number of triplets, which is required for isValidSplit
      }
   }
   //Now define iterators and PdgID's that hold the information of two particles
   std::list<_splitting_type*>::iterator iteri,iterj;
   PdgID idi;
   PdgID idj;
   PdgID idm;
   while (numParticles > 2) {
      // Pick two particles i and j
      size_t ii, jj;
      while (true) {
         ii = randomInteger(numParticles);
         while (ii == (jj = randomInteger(numParticles))) //Make sure that j is different than i
         {};
         iteri = iterj = coloredList.begin(); //set the list iterator to the beginning of the list
         std::advance(iteri, ptrdiff_t_from_size_t(ii)); //set the list iterator to the location of particle i
         std::advance(iterj, ptrdiff_t_from_size_t(jj)); //set the list iterator to the location of particle j
         idi = (*iteri)->id; //get the id of particle i
         idj = (*iterj)->id; //get the id of particle j

         if (isValidSplit(idi, idj, nTriplet)) { //Make sure that a splitting into i and j is allowed by QCD
            break;
         }
      }

      // At this point we know that we have a valid QCD splitting

      // Get the id of the combined particle and remove a triplet if we have a g->qqbar splitting
      if (idi.isGluon())
         idm = idj;
      else if (idj.isGluon())
         idm = idi;
      else {
         idm = PdgID::kGluon;
         nTriplet--;
      };

      //create the link from mother to L and R daugthers
      _splitting_type* temp = new _splitting_type;
      temp->id = idm;
      if (idm.isQuark()) {
         if (idm.isParticle())
            temp->colorCharge = kTriplet;
         else
            temp->colorCharge = kTripletBar;
      } else
         temp->colorCharge = kOctet;
      temp->sigLocation = std::numeric_limits<size_t>::max(); //Not part of the usual signature
      temp->color = 0;
      temp->antiColor = 0;
      temp->left = *iteri;
      temp->right = *iterj;
      coloredList.push_back(temp); //Add the mother to the list
      coloredList.erase(iteri);    //Remove daughter i from list
      coloredList.erase(iterj);    //Remove daughter j from list
      numParticles--;

   }
   //Get here if only two particles left in the list
   iteri=coloredList.begin();
   iterj=iteri++;   //First and last element are easy to come by with only two elements

   //create the link from trunk to L and R daugthers
   _splitting_type* temp = new _splitting_type;
   temp->id = PdgID::kUnknown;
   temp->colorCharge = kSinglet;
   temp->sigLocation = std::numeric_limits<size_t>::max(); //Not part of the usual signature
   temp->color = 0;
   temp->antiColor = 0;
   temp->left = *iteri;
   temp->right = *iterj;
   return temp; //Return the trunk of the linked list
}

//------------------------------------------------------------------------------
/**
 * \brief Create the color history and return the start of the linked list directly using a Configuration
 *
 * An algorithm for randomly finding a  splitting history that is allowed by
 * the interactions of QCD. It uses a recombination algorithm to
 * construct a possible history and assign the color flow of this history in the
 * large N limit
 * TODO: is this algorithm efficient?
 */
GeneratorUtils::_splitting_type* GeneratorUtils::createColorHistory(Configuration config)
{
   //Get the number of final state particles
   size_t numParticles = config.sizeFinal();
   //Count the number of quarks and antiquarks
   int nTriplet = 0;

   std::list<_splitting_type*> coloredList;

   //First, go through all particles in the signature and put the information of each particle
   //into a struct that will hold all information required. Note that these are not splittings
   //yet, but the struct contains the pointers that hold the splitting information.
   size_t i = 0;
   for (Configuration::iterator it = config.beginFinal(); it != config.endFinal(); ++it, ++i) {
      _splitting_type* temp = new _splitting_type;
      temp->id = it->id();
      if (temp->id.isQuark()) {
         if (temp->id.isParticle())
            temp->colorCharge = kTriplet;
         else
            temp->colorCharge = kTripletBar;
      } else if (temp->id == PdgID::kGluon)
         temp->colorCharge = kOctet;
      else
         temp->colorCharge = 1;
      temp->sigLocation = i;
      temp->color = 0;
      temp->antiColor = 0;
      temp->left = 0;
      temp->right = 0;
      coloredList.push_back(temp);

      if (it->id().isQuark() && it->id().isParticle()) {
         nTriplet++; //Count the number of triplets, which is required for isValidSplit
      }
   }
   //Now define iterators and PdgID's that hold the information of two particles
   std::list<_splitting_type*>::iterator iteri,iterj;
   PdgID idi;
   PdgID idj;
   PdgID idm;
   while (numParticles > 2) {
      // Pick two particles i and j
      size_t j;
      while (true) {
         i = randomInteger(numParticles);
         while (i == (j = randomInteger(numParticles))) //Make sure that j is different than i
         {};
         iteri = iterj = coloredList.begin(); //set the list iterator to the beginning of the list
         std::advance(iteri, ptrdiff_t_from_size_t(i)); //set the list iterator to the location of particle i
         std::advance(iterj, ptrdiff_t_from_size_t(j)); //set the list iterator to the location of particle j
         idi = (*iteri)->id; //get the id of particle i
         idj = (*iterj)->id; //get the id of particle j

         if (isValidSplit(idi, idj, nTriplet)) { //Make sure that a splitting into i and j is allowed by QCD
            break;
         }
      }

      // At this point we know that we have a valid QCD splitting

      // Get the id of the combined particle and remove a triplet if we have a g->qqbar splitting
      if (idi.isGluon())
         idm = idj;
      else if (idj.isGluon())
         idm = idi;
      else {
         idm = PdgID::kGluon;
         nTriplet--;
      };

      //create the link from mother to L and R daugthers
      _splitting_type* temp = new _splitting_type;
      temp->id = idm;
      if (idm.isQuark()) {
         if (idm.isParticle())
            temp->colorCharge = kTriplet;
         else
            temp->colorCharge = kTripletBar;
      } else
         temp->colorCharge = kOctet;
      temp->sigLocation = std::numeric_limits<size_t>::max(); //Not part of the usual signature
      temp->color = 0;
      temp->antiColor = 0;
      temp->left = *iteri;
      temp->right = *iterj;
      coloredList.push_back(temp); //Add the mother to the list
      coloredList.erase(iteri);    //Remove daughter i from list
      coloredList.erase(iterj);    //Remove daughter j from list
      numParticles--;

   }
   //Get here if only two particles left in the list
   iteri=coloredList.begin();
   iterj=iteri++;   //First and last element are easy to come by with only two elements

   //create the link from trunk to L and R daugthers
   _splitting_type* temp = new _splitting_type;
   temp->id = PdgID::kUnknown;
   temp->colorCharge = kSinglet;
   temp->sigLocation = std::numeric_limits<size_t>::max(); //Not part of the usual signature
   temp->color = 0;
   temp->antiColor = 0;
   temp->left = *iteri;
   temp->right = *iterj;
   return temp; //Return the trunk of the linked list
}

// -----------------------------------------------------------------------------
/**
 * \brief Generate the color information of an element of a colorHist.
 *
 * Calculates the color flow of a single splitting based on the large N approximation.
 */
void GeneratorUtils::generateColor(_splitting_type* mother, int& numColorLabel)
{
   _splitting_type* left = mother->left;
   _splitting_type* right = mother->right;
   int temp = 0;
   if (mother->colorCharge == kSinglet) {
      switch (left->colorCharge) {
         case kSinglet:
            assert(right->colorCharge == false);
            // don't know how to deal with photons yet
            break;
         case kTriplet:
            assert(right->colorCharge == kTripletBar);
            temp = getNewColorLabel(numColorLabel);
            left->color = temp;
            right->antiColor = temp;
            break;
         case kTripletBar:
            assert(right->colorCharge == kTriplet);
            temp = getNewColorLabel(numColorLabel);
            left->antiColor = temp;
            right->color = temp;
            break;
         case kOctet:
            assert(false);
            break;
      }
   } else if (mother->colorCharge == kTriplet) {
      switch (left->colorCharge) {
         case kSinglet:
            assert(false);
            // don't know how to deal with photons yet
            break;
         case kTriplet:
            assert(right->colorCharge == kOctet);
            //large N approximation
            temp = getNewColorLabel(numColorLabel);
            left->color = temp;
            right->antiColor = temp;
            right->color = mother->color;
            break;
         case kTripletBar:
            assert(false);
            break;
         case kOctet:
            assert(right->colorCharge == kTriplet);
            // large N approximation
            temp = getNewColorLabel(numColorLabel);
            right->color = temp;
            left->antiColor = temp;
            left->color = mother->color;
            break;
      }
   } else if (mother->colorCharge == kTripletBar) {
      switch (left->colorCharge) {
         case kSinglet:
            assert(false);
            // don't know how to deal with photons yet
            break;
         case kTriplet:
            assert(false);
            break;
         case kTripletBar:
            assert(right->colorCharge == kOctet);
            //large N approximation
            temp = getNewColorLabel(numColorLabel);
            left->antiColor = temp;
            right->color = temp;
            right->antiColor = mother->antiColor;
            break;
         case kOctet:
            assert(right->colorCharge == kTripletBar);
            //large N approximation
            temp = getNewColorLabel(numColorLabel);
            right->antiColor = temp;
            left->color = temp;
            left->antiColor = mother->antiColor;
            break;
      }
   } else if (mother->colorCharge == kOctet) {
      switch (left->colorCharge) {
         case kSinglet:
            assert(false);
            break;
         case kTriplet:
            assert(right->colorCharge == kTripletBar);
            left->color = mother->color;
            right->antiColor = mother->antiColor;
            break;
         case kTripletBar:
            assert(right->colorCharge == kTriplet);
            right->color = mother->color;
            left->antiColor = mother->antiColor;
            break;
         case kOctet:
            assert(right->colorCharge == kOctet);
            // large N approximation
            temp = getNewColorLabel(numColorLabel);
            left->color = mother->color;
            left->antiColor = temp;
            right->color = temp;
            right->antiColor = mother->antiColor;
            break;
      }
   }
}

} // namespace Geneva
