//------------------------------------------------------------------------------
/// \file IntegrationChannel.cpp
//
// Author(s):
//    Simone Alioli, Christian Bauer
//
// Copyright:
//    Copyright (C) 2015 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class IntegrationChannel
//------------------------------------------------------------------------------

#include "Geneva/Generators/IntegrationChannel.hpp"

#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/Random.hpp"

#include <map>
#include <string>
#include <vector>
#include <numeric>
#include <iterator>

namespace Geneva
{

//------------------------------------------------------------------------------
std::pair< Signature, double > IntegrationChannel::pickSignature() const
{
   double jac;
   SignatureSet::iterator sig = _signatures.begin();
   if (_signatures.size() == 1) {
      jac = 1.;

   } else {
      double randMult = randomInterval() * _cumulativeWeights.back();
      size_t count = 0;
      while (randMult > _cumulativeWeights[count]) {
         sig++;
         count++;
         if (sig == _signatures.end()) {
            Log::error("IntegrationChannel::pickSignature") << "Should not reach the end of the SignatureSet" << Log::end;
         }
      }
      if (count == 0)
         jac = _cumulativeWeights[0] / _cumulativeWeights.back();
      else {
         jac = (_cumulativeWeights[count] - _cumulativeWeights[count - 1]) / _cumulativeWeights.back();
      }
   }
   if (jac < 0)
      Log::error("IntegrationChannel::pickSignature") << "The probability for Signature " << *sig << "is negative" << Log::end;

   return std::make_pair(*sig, jac);
}

//------------------------------------------------------------------------------
void IntegrationChannel::initialize()
{
   double n = 1.;
   for (SignatureSet::iterator it = _signatures.begin(); it != _signatures.end(); ++it) {
      if (it->size() != _signatureSize) {
         Log::error("IntegrationChannel") << "Including different-size Signatures in the same IntegrationChannel is not allowed!" << Log::end;
         break;
      }
      //Initialize all the Signatures to have the same probability
      _weight[*it] = 0.;
      _weightSq[*it] = 0.;
      _numEvents = 0;
      _cumulativeWeights.push_back(n / static_cast<double>(_signatures.size()));
      n++;
   }
}

//------------------------------------------------------------------------------
void IntegrationChannel::reset()
{
   _cumulativeWeights.clear();
   for (std::map<Signature, double>::iterator it = _weight.begin(); it != _weight.end(); ++it) {
      if (it == _weight.begin()) {
         if (_numEvents == 0)
            _cumulativeWeights.push_back(0.);
         else
            _cumulativeWeights.push_back(abs(it->second) / static_cast<double>(_numEvents));
      } else {
         if (_numEvents == 0)
            _cumulativeWeights.push_back(0.);
         else
            _cumulativeWeights.push_back(_cumulativeWeights.back() + abs(it->second) / static_cast<double>(_numEvents));
      }
      //Now that the weights have been transferred to _cumaltiveWeights, set them to zero
      it->second = 0.;
   }
   for (std::map<Signature, double>::iterator it = _weightSq.begin(); it != _weightSq.end(); ++it) {
      //Set the weightSq to zero
      it->second = 0.;
   }
   //Set the numEvents to zero
   _numEvents = 0;
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Streams the IntegrationChannel to output.
std::ostream& IntegrationChannel::print(std::ostream& out) const
{
   out << name() << " with Signatures ( ";
   std::copy(_signatures.begin(), --_signatures.end(), std::ostream_iterator<Signature>(out, ", "));
   out << *(--_signatures.end()) << ')';
   return out;
}

} // namespace Geneva
