//------------------------------------------------------------------------------
/// \file IntegrationChannelManager.cpp
//
// Author(s):
//    Simone Alioli
//
// Copyright:
//    Copyright (C) 2015 CERN
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class IntegrationChannelManager
//------------------------------------------------------------------------------

#include "Geneva/Generators/IntegrationChannelManager.hpp"
#include "Geneva/Generators/IntegrationChannel.hpp"
#include "Geneva/Generators/IGenerator.hpp"
#include "Geneva/Generators/GeneratorFactory.hpp"

#include "Geneva/Components/ITreeME.hpp"
#include "Geneva/Components/ComponentFactory.hpp"
#include "Geneva/Components/RegionInfoList.hpp"
#include "Geneva/Components/SignatureManager.hpp"

#include "Geneva/Core/Physics/SignatureSet.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/Utils/Random.hpp"

#include <iterator>
#include <string>

namespace Geneva
{

namespace
{

//------------------------------------------------------------------------------
/**
 * \brief An unary function object checking the flavors of the intial state particles A and B.
 */
class IsSpecificABFlavor : public std::unary_function<Signature, bool>
{
   private:
      PdgID _idA, _idB;

   public:
      IsSpecificABFlavor(PdgID idA, PdgID idB) : _idA(idA), _idB(idB) {}

      bool operator()(const Signature& sig) const
      {
         return (*(sig.initialA()) == _idA) && (*sig.initialB() == _idB);
      }
};

//------------------------------------------------------------------------------
/**
 * \brief An unary function object checking the flavor of the intial state particle A.
 */
class IsSpecificAFlavor : public std::unary_function<Signature, bool>
{
   private:
      PdgID _idA;

   public:
      IsSpecificAFlavor(PdgID idA) : _idA(idA) {}

      bool operator()(const Signature& sig) const
      {
         return (*(sig.initialA()) == _idA);
      }
};

//------------------------------------------------------------------------------
/**
 * \brief An unary function object checking the flavor of the intial state particle B.
 */
class IsSpecificBFlavor : public std::unary_function<Signature, bool>
{
   private:
      PdgID _idB;

   public:
      IsSpecificBFlavor(PdgID idB) : _idB(idB) {}

      bool operator()(const Signature& sig) const
      {
         return (*(sig.initialB()) == _idB);
      }
};

//------------------------------------------------------------------------------
/**
 * \brief An unary function object checking if the final state contains a gluon
 */
class ContainsFinalGluon : public std::unary_function<Signature, bool>
{
   public:
      bool operator()(const Signature& sig) const
      {
         return (sig.countFinal(PdgID::kGluon) > 0);
      }
};

//------------------------------------------------------------------------------
/**
 * \brief An unary function object checking if initialA is the antipartner of initialB
 */
class ANotAntiPartnerOfB : public std::unary_function<Signature, bool>
{
   public:
      bool operator()(const Signature& sig) const
      {
         return (!sig.initialA()->isAntiPartner(*(sig.initialB())));
      }
};

} // anonymous namespace

//------------------------------------------------------------------------------
std::vector<IntegrationChannel> IntegrationChannelManager::getChannels(const SignatureSet& sigSetIn, const OptionMap& optMap)
{
   std::vector<IntegrationChannel> channels;
   if (not optMap.getOption<bool>("IntegrationChannels"))
      return channels;
   const std::string opt = optMap.getOption("IntegrationChannels::separation");

   bool singleChannelSetup = optMap.getOption<bool>("IntegrationChannels::singleChannelSetup");
   size_t channelNum = std::numeric_limits<size_t>::quiet_NaN();
   if (singleChannelSetup)
      channelNum = size_t_from_int(optMap.getOption<int>("IntegrationChannels::channelNumberSetup"));

   if (opt == "Multiplicity") {
      size_t lowerMultiplicity = sigSetIn.begin()->size();
      size_t higherMultiplicity = sigSetIn.rbegin()->size();
      size_t channelAddNum = 0;
      for (size_t mult = lowerMultiplicity; mult <= higherMultiplicity; mult++) {
         SignatureSet sigSet = sigSetIn.copy_if(CheckSignatureSize(mult));
         assert(not sigSet.empty());
         std::stringstream ss;
         ss << mult;
         if (singleChannelSetup) {
            if (channelAddNum == channelNum) {
               channels.push_back(IntegrationChannel(sigSet, "mult_" + ss.str()));
               return channels;
            }
         }
         else
            channels.push_back(IntegrationChannel(sigSet, "mult_" + ss.str()));
         ++channelAddNum;
      }
   } else if (opt == "Signatures") {
      size_t channelAddNum = 0;
      for (SignatureSet::iterator it = sigSetIn.begin(); it != sigSetIn.end(); ++it) {
         std::stringstream ss;
         Signature::iterator last = it->end();
         --last;
         std::copy(it->begin(), last, std::ostream_iterator<PdgID>(ss, "_"));
         ss << *last;
         if (singleChannelSetup) {
            if (channelAddNum == channelNum) {
               channels.push_back(IntegrationChannel(*it, "sig_" + ss.str()));
               return channels;
            }
         }
         else
            channels.push_back(IntegrationChannel(*it, "sig_" + ss.str()));
         ++channelAddNum;
      }
   } else if (opt == "TreeME") {
      ITreeME* treeME = ComponentFactory::instance().getTreeME(optMap);
      IGenerator* gen = GeneratorFactory::instance().getGenerator(optMap);
      gen->setSignatures(sigSetIn);
      std::map<Signature, std::vector<double> > resultTree;
      RandomState ranState = getRandomState();
      for (SignatureSet::iterator it = sigSetIn.begin(); it != sigSetIn.end(); ++it) {
         std::vector<double> resTree;
         initializeRandomState(ranState);
         for (size_t count = 0; count < 10; count ++) {
            Configuration config;
            std::vector<double> randoms;
            for (size_t num = 0; num < gen->numberOfRandomsNeeded(*it); num++) {
               randoms.push_back(randomInterval());
            }
            if (not gen->generate(*it, randoms, config))
               Log::error("IntegrationChannelManager::getChannels") << "Generator cannot fail to create a Configuration at this stage" << Log::end;
            double tmpTree = (*treeME)(config);
            if (not std::isnormal(tmpTree))
               Log::error("IntegrationChannelManager::getChannels") << "TreeME cannot fail to evaluate a Configuration at this stage" << Log::end;
            resTree.push_back(tmpTree);
         }
         resultTree[*it] = resTree;
      }
      delete treeME;
      initializeRandomState(ranState);
      std::map<Signature, SignatureSet> propTreeME;
      // find Signatures that are proportional
      for (SignatureSet::iterator it_i = sigSetIn.begin(); it_i != sigSetIn.end(); ++it_i) {
         bool isProp = true;
         SignatureSet::iterator it_j = sigSetIn.begin();
         while (it_j != it_i) {
            isProp = true;
            double ratio = (*(resultTree[*it_i].begin())) / (*(resultTree[*it_j].begin()));
            for (size_t count = 1; count < resultTree[*it_i].size(); count++) {
               if (compareToZero(1.0 - resultTree[*it_i][count] / resultTree[*it_j][count] / ratio) != 0) {
                  isProp = false;
                  break;
               }
            }
            if (isProp) break;
            it_j++;
         }
         if (isProp) {
            propTreeME[*it_j].insert(*it_i);
         } else {
            propTreeME[*it_i].insert(*it_i);
         }
      }
      assert(propTreeME.size() > 0);

      size_t channelAddNum = 0;
      for (std::map<Signature, SignatureSet>::iterator it = propTreeME.begin(); it != propTreeME.end(); ++it) {
         std::stringstream ss;
         Signature::iterator last = it->first.end();
         --last;
         std::copy(it->first.begin(), last, std::ostream_iterator<PdgID>(ss, "_"));
         ss << *last;
         if (singleChannelSetup) {
            if (channelAddNum == channelNum) {
               channels.push_back(IntegrationChannel(it->second, "tree_" + ss.str()));
               return channels;
            }
         }
         else
            channels.push_back(IntegrationChannel(it->second, "tree_" + ss.str()));
         ++channelAddNum;
      }
   } else if (opt == "ppToZ2jets") {
      std::string calc = optMap.getOption("calc");
      if ( ((calc != "CalcSCETppV012") && (calc != "CalcNLOppV") && (calc != "CalcNLOppVj")) && (optMap.getOption("Process::DrellYan::BosonType") != "Z") )
         Log::error("IntegrationChannelManager::getChannels") << "Channel separation " << opt << " does not work for Calculation " << calc << Log::end;

      //Split up the SignatureSet into sets with a Z and 0, 1 and 2 final QCD particles
      SignatureSet sigSet0 = sigSetIn.copy_if(CheckSignatureSize(4));
      SignatureSet sigSet1 = sigSetIn.copy_if(CheckSignatureSize(5));
      SignatureSet sigSet2 = sigSetIn.copy_if(CheckSignatureSize(6));
      SignatureSet sigSet;
      size_t channelAddNum = 0;
      if (not sigSet0.empty()) {
         sigSet = sigSet0.move_if(IsSpecificABFlavor(PdgID::kUp, PdgID::kUpBar));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_uubar_Z"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_uubar_Z"));
         }
         ++channelAddNum;
         sigSet = sigSet0.move_if(IsSpecificABFlavor(PdgID::kDown, PdgID::kDownBar));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ddbar_Z"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ddbar_Z"));
         }
         ++channelAddNum;
         sigSet = sigSet0.move_if(IsSpecificABFlavor(PdgID::kUpBar, PdgID::kUp));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ubaru_Z"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ubaru_Z"));
         }
         ++channelAddNum;
         sigSet = sigSet0.move_if(IsSpecificABFlavor(PdgID::kDownBar, PdgID::kDown));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_dbard_Z"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_dbard_Z"));
         }
         ++channelAddNum;
         sigSet = sigSet0;
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ff_Z"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ff_Z"));
         }
         ++channelAddNum;
      }
      else
         channelAddNum += 5;
      if (not sigSet1.empty()) {
         sigSet = sigSet1.move_if(IsSpecificABFlavor(PdgID::kUp, PdgID::kUpBar));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_uubar_Zg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_uubar_Zg"));
         }
         ++channelAddNum;
         sigSet = sigSet1.move_if(IsSpecificABFlavor(PdgID::kDown, PdgID::kDownBar));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ddbar_Zg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ddbar_Zg"));
         }
         ++channelAddNum;
         sigSet = sigSet1.move_if(IsSpecificABFlavor(PdgID::kUpBar, PdgID::kUp));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ubaru_Zg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ubaru_Zg"));
         }
         ++channelAddNum;
         sigSet = sigSet1.move_if(IsSpecificABFlavor(PdgID::kDownBar, PdgID::kDown));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_dbard_Zg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_dbard_Zg"));
         }
         ++channelAddNum;
         sigSet = sigSet1.move_if(IsSpecificABFlavor(PdgID::kUp, PdgID::kGluon));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ug_Zu"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ug_Zu"));
         }
         ++channelAddNum;
         sigSet = sigSet1.move_if(IsSpecificABFlavor(PdgID::kDown, PdgID::kGluon));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_dg_Zd"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_dg_Zd"));
         }
         ++channelAddNum;
         sigSet = sigSet1.move_if(IsSpecificABFlavor(PdgID::kGluon, PdgID::kUp));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_gu_Zu"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_gu_Zu"));
         }
         ++channelAddNum;
         sigSet = sigSet1.move_if(IsSpecificABFlavor(PdgID::kGluon, PdgID::kDown));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_gd_Zd"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_gd_Zd"));
         }
         ++channelAddNum;
         sigSet = sigSet1.move_if(IsSpecificAFlavor(PdgID::kGluon));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_gf_Zf"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_gf_Zf"));
         }
         ++channelAddNum;
         sigSet = sigSet1.move_if(IsSpecificBFlavor(PdgID::kGluon));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_fg_Zf"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_fg_Zf"));
         }
         ++channelAddNum;
         sigSet = sigSet1;
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ff_Zg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ff_Zg"));
         }
         ++channelAddNum;
      }
      else
         channelAddNum += 11;
      if (not sigSet2.empty()) {
         SignatureSet sig2uubar = sigSet2.move_if(IsSpecificABFlavor(PdgID::kUp, PdgID::kUpBar));
         SignatureSet sig2ddbar = sigSet2.move_if(IsSpecificABFlavor(PdgID::kDown, PdgID::kDownBar));
         SignatureSet sig2ubaru = sigSet2.move_if(IsSpecificABFlavor(PdgID::kUpBar, PdgID::kUp));
         SignatureSet sig2dbard = sigSet2.move_if(IsSpecificABFlavor(PdgID::kDownBar, PdgID::kDown));
         sigSet = sig2uubar.move_if(ContainsFinalGluon());
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_uubar_Zgg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_uubar_Zgg"));
         }
         ++channelAddNum;
         sigSet = sig2uubar;
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_uubar_Zff"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_uubar_Zff"));
         }
         ++channelAddNum;
         sigSet = sig2ddbar.move_if(ContainsFinalGluon());
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ddbar_Zgg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ddbar_Zgg"));
         }
         ++channelAddNum;
         sigSet = sig2ddbar;
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ddbar_Zff"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ddbar_Zff"));
         }
         ++channelAddNum;
         sigSet = sig2ubaru.move_if(ContainsFinalGluon());
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ubaru_Zgg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ubaru_Zgg"));
         }
         ++channelAddNum;
         sigSet = sig2ubaru;
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ubaru_Zff"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ubaru_Zff"));
         }
         ++channelAddNum;
         sigSet = sig2dbard.move_if(ContainsFinalGluon());
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_dbard_Zgg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_dbard_Zgg"));
         }
         ++channelAddNum;
         sigSet = sig2dbard;
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_dbard_Zff"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_dbard_Zff"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificABFlavor(PdgID::kUp, PdgID::kGluon));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ug_Zug"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ug_Zug"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificABFlavor(PdgID::kGluon, PdgID::kUp));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_gu_Zug"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_gu_Zug"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificABFlavor(PdgID::kDown, PdgID::kGluon));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_dg_Zdg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_dg_Zdg"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificABFlavor(PdgID::kGluon, PdgID::kDown));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_gd_Zdg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_gd_Zdg"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificABFlavor(PdgID::kUp, PdgID::kUp));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_uu_Zuu"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_uu_Zuu"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificABFlavor(PdgID::kDown, PdgID::kDown));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_dd_Zdd"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_dd_Zdd"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificABFlavor(PdgID::kGluon, PdgID::kGluon));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_gg_Zff"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_gg_Zff"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificAFlavor(PdgID::kGluon));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_gf_Zfg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_gf_Zfg"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificBFlavor(PdgID::kGluon));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_fg_Zfg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_fg_Zfg"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(ContainsFinalGluon());
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ff_Zgg"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ff_Zgg"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificABFlavor(PdgID::kUp, PdgID::kDown));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ud_Zud"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ud_Zud"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificABFlavor(PdgID::kDown, PdgID::kUp));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_du_Zud"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_du_Zud"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificAFlavor(PdgID::kUp));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_uf_Zuf"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_uf_Zuf"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificAFlavor(PdgID::kDown));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_df_Zdf"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_df_Zdf"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificBFlavor(PdgID::kUp));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_fu_Zuf"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_fu_Zuf"));
         }
         ++channelAddNum;
         sigSet = sigSet2.move_if(IsSpecificBFlavor(PdgID::kDown));
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_fd_Zdf"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_fd_Zdf"));
         }
         ++channelAddNum;
         sigSet = sigSet2;
         if (not sigSet.empty()) {
            if (singleChannelSetup) {
               if (channelAddNum == channelNum) {
                  channels.push_back(IntegrationChannel(sigSet, "proc_ff_Zff"));
                  return channels;
               }
            }
            else
               channels.push_back(IntegrationChannel(sigSet, "proc_ff_Zff"));
         }
      }
   }
   else if  (opt == "ppToVJets") {
      std::string calc = optMap.getOption("calc");
      if ( (calc != "CalcSCETppV012") && (calc != "CalcNLOppV") && (calc != "CalcNLOppVj") )
         Log::error("IntegrationChannelManager::getChannels") << "Channel separation " << opt << " does not work for Calculation " << calc << Log::end;

      //A map holding a SignatureSet and its name
      std::map<std::string, SignatureSet> signatureChannels;

      //Get the signature sets of individual multiplicities
      SignatureSet sigSetMult[3];
      sigSetMult[0] = sigSetIn.copy_if(CheckSignatureSize(4));
      sigSetMult[1] = sigSetIn.copy_if(CheckSignatureSize(5));
      sigSetMult[2] = sigSetIn.copy_if(CheckSignatureSize(6));

      //In case some multiplicities are empty, generate them from the SignatureManager, so I can get proper regionInfo's
      int nLight = optMap.getOption<int>("NLightFlavors");
      std::string optString;
      PdgID lepton;
      optString = optMap.getOption<std::string>("Process::DrellYan::DecayChannel");
      if (optString == "electrons")
         lepton = PdgID::kElectron;
      else if (optString == "muons")
         lepton = PdgID::kMuon;
      else if (optString == "taus")
         lepton = PdgID::kTau;
      else
         Log::error("IntegrationChannelManager::getChannels") << "DecayChannel " << optString << " not allowed" << Log::end;

      optString = optMap.getOption<std::string>("Process::DrellYan::BosonType");
      SignatureManager sigManager;
      if (optString == "Z") {
         if (sigSetMult[0].empty())
            sigSetMult[0] = sigManager.getSignaturespp2ZJets(lepton, 0, nLight);
         if (sigSetMult[1].empty())
            sigSetMult[1] = sigManager.getSignaturespp2ZJets(lepton, 1, nLight);
         if (sigSetMult[2].empty())
            sigSetMult[2] = sigManager.getSignaturespp2ZJets(lepton, 2, nLight);
      }
      else if (optString == "W+") {
         if (sigSetMult[0].empty())
            sigSetMult[0] = sigManager.getSignaturespp2WPlusJets(lepton, 0, nLight);
         if (sigSetMult[1].empty())
            sigSetMult[1] = sigManager.getSignaturespp2WPlusJets(lepton, 1, nLight);
         if (sigSetMult[2].empty())
            sigSetMult[2] = sigManager.getSignaturespp2WPlusJets(lepton, 2, nLight);
      }
      else  if (optString == "W-") {
         if (sigSetMult[0].empty())
            sigSetMult[0] = sigManager.getSignaturespp2WMinusJets(lepton, 0, nLight);
         if (sigSetMult[1].empty())
            sigSetMult[1] = sigManager.getSignaturespp2WMinusJets(lepton, 1, nLight);
         if (sigSetMult[2].empty())
            sigSetMult[2] = sigManager.getSignaturespp2WMinusJets(lepton, 2, nLight);
      } else
         Log::error("IntegrationChannelManager::getChannels") << "BosonType " << optString << " not allowed" << Log::end;

      RegionInfoList regionInfoList[2];
      regionInfoList[0].setRegionInfoList(sigSetMult[0], sigSetMult[1], optMap.getOption<int>("NLightFlavors"));
      regionInfoList[1].setRegionInfoList(sigSetMult[1], sigSetMult[2], optMap.getOption<int>("NLightFlavors"));

      for (SignatureSet::iterator itSig = sigSetIn.begin(); itSig != sigSetIn.end(); ++itSig) {

         //Determine the number and flavors of the strongly interacting particles
         Signature::iterator itA = itSig->initialA();
         Signature::iterator itB = itSig->initialB();
         size_t numQCD = itSig->sizeFinal() - 2;

         //This stringstream will hold the name of the channel
         std::stringstream channelType;
         channelType << "sing_" << numQCD;

         //Determine if the initial states are valence, sea or glue
         if (itA->isGluon())
            channelType << "_G";
         else if ((*itA == PdgID::kUp) or (*itA == PdgID::kDown))
            channelType << "_V";
         else //Sea quark
            channelType << "_S";
         if (itB->isGluon())
            channelType << "_G";
         else if ((*itB == PdgID::kUp) or (*itB == PdgID::kDown))
            channelType << "_V";
         else //Sea quark
            channelType << "_S";

         //Determine the type of singularities the Signature has.
         //There are A, B, F, for ISRA, ISRB and FSR
         if (numQCD > 0) {
            std::list<RegionInfo> regions = regionInfoList[numQCD-1].getSingularRegionsToSignature(*itSig);
            bool hasISRA = false;
            bool hasISRB = false;
            bool hasFSR = false;
            for (std::list<RegionInfo>::iterator region = regions.begin(); region != regions.end(); ++region) {
               if (region->hasInitialStateSingularity()) {
                  if (region->k->isInitialA()) {
                     if (not hasISRA) {
                        channelType << "_A";
                        hasISRA = true;
                     }
                  }
                  if (region->k->isInitialB()) {
                     if (not hasISRB) {
                        channelType << "_B";
                        hasISRB = true;
                     }
                  }
               }
               if (region->hasFinalStateSingularity() and not hasFSR) {
                  channelType << "_F";
                  hasFSR = true;
               }
            }
         }
         signatureChannels[channelType.str()].insert(*itSig);
      }
      std::map<std::string, SignatureSet>::iterator sigChannelIt;
      for (sigChannelIt = signatureChannels.begin(); sigChannelIt != signatureChannels.end(); ++sigChannelIt) {
         Log::debug() << "ChannelName: " << sigChannelIt->first << "  " << sigChannelIt->second.size() << Log::end;
         Log::debug() << sigChannelIt->second << Log::end;
         if (not sigChannelIt->second.empty())
            channels.push_back(IntegrationChannel(sigChannelIt->second, sigChannelIt->first));
      }
      Log::debug() << channels.size() << Log::end;
   }
   else {
      Log::error("IntegrationChannelManager::getChannels") << "EventGeneratorAdaptive::channelSeparation option " << opt << " not allowed" << Log::end;
   }
   return channels;
}

//------------------------------------------------------------------------------
std::map<Signature, std::string> IntegrationChannelManager::getSignatureMap(const SignatureSet& sigSet, const OptionMap& optMap)
{
   std::map<Signature, std::string> signatureMap;
   std::vector<IntegrationChannel> channels = getChannels(sigSet, optMap);
   for (SignatureSet::iterator sig = sigSet.begin(); sig != sigSet.end(); ++sig) {
      for (std::vector<IntegrationChannel>::iterator channel = channels.begin(); channel != channels.end(); ++channel) {
      SignatureSet sigSetChannel = channel->signatures();
         if ((signatureMap.find(*sig) == signatureMap.end()) and (sigSetChannel.find(*sig) != sigSetChannel.end())) {
         signatureMap[*sig] =  channel->name();
         }
      }
   }
   return signatureMap;
}

} // namespace Geneva
