//------------------------------------------------------------------------------
/// \file SmartRandomHandler.cpp
//
// Author(s):
//    Christian Bauer, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017, 2018 Geneva Collaboration
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class SmartRandomHandler
//------------------------------------------------------------------------------

#include "Geneva/Generators/SmartRandomHandler.hpp"
#include "Geneva/Generators/SmartRandomMint.hpp"
#include "Geneva/Generators/IntegrationChannel.hpp"

#include "Geneva/Core/UI/Log.hpp"

#include <fstream>

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * Deletes all stored (and owned) SmartRandoms.
 */
SmartRandomHandler::~SmartRandomHandler()
{
   for (_iterator it = _smart_randoms.begin(); it != _smart_randoms.end(); ++it)
      delete it->second;
}

//------------------------------------------------------------------------------
/**
 * \brief Creates a new SmartRandom for \a channel.
 * \param type The type of SmartRandom to use (currently only \c kMint is supported)
 * \param channel The IntegrationChannel with which the SmartRandom will be associated
 * \param dimension
 * \param numBins
 * \return \c true if a new SmartRandom was created.
 *
 * Creates a new SmartRandomMint using the provided information, unless a
 * SmartRandom for \a channel already exists. The newly created SmartRandom
 * is considered not set up.
 */
bool SmartRandomHandler::makeSmartRandom(SmartRandomType type, const IntegrationChannel& channel,
                                         size_t dimension, size_t numBins)
{
   // Get the name of the SmartRandom.
   const std::string name = channel.name();

   // Add it if it is not already in the map.
   if (_smart_randoms.find(name) == _smart_randoms.end()) {
      ISmartRandom* smartRandom = 0;
      if (type == kMint) {
         smartRandom = new SmartRandomMint(dimension, numBins);
      }
      else {
         Log::fatal("SmartRandomHandler::makeSmartRandom") << "SmartRandomType must be kMint." << Log::end;
      }
      _smart_randoms[name] = smartRandom;
      _is_setup[smartRandom] = false;
      return true;
   }
   return false;
}

//------------------------------------------------------------------------------
/**
 * \brief Obtains random numbers from the SmartRandom for \a channel.
 *
 * Fills the vector of random numbers \a rand using the SmartRandom for the
 * IntegrationChannel \a channel.
 *
 * The generation probability is returned in \a prob.
 */
void SmartRandomHandler::getRandom(const IntegrationChannel& channel, std::vector<double>& rand, double& prob)
{
   ISmartRandom* smartRandom = _get_smart_random(channel);
   if (rand.size() != smartRandom->dimension())
      Log::fatal("SmartRandomHandler::getRandom") << "The dimensions of the random vector passed in does not agree with the dimension of the SmartRandom." << Log::end;

   smartRandom->nextRandom(rand, prob);
}

//------------------------------------------------------------------------------
/**
 * \brief Informs the SmartRandom for \a channel of the result of the calculation.
 *
 * Informs the SmartRandom associated with IntegrationChannel \a channel of the
 * result of the calculation given by \a calcResult.
 *
 * This does nothing if the SmartRandom is already set up,
 */
void SmartRandomHandler::informOfCalculation(const IntegrationChannel& channel, double calcResult)
{
   ISmartRandom* smartRandom = _get_smart_random(channel);
   if (!_is_setup[smartRandom])
      smartRandom->informOfCalculation(calcResult);
}

//------------------------------------------------------------------------------
/**
 * \brief Informs the SmartRandom for \a channel that the calculation failed.
 *
 * This does nothing if the SmartRandom is already set up,
 */
void SmartRandomHandler::informOfFailure(const IntegrationChannel& channel)
{
   ISmartRandom* smartRandom = _get_smart_random(channel);
   if (!_is_setup[smartRandom])
      smartRandom->informOfFailure();
}

//------------------------------------------------------------------------------
// Returns true if the SmartRandom for channel is set up.
bool SmartRandomHandler::isSetup(const IntegrationChannel& channel)
{
   return _is_setup[_get_smart_random(channel)];
}

//------------------------------------------------------------------------------
// Returns true if all SmartRandoms are set up.
bool SmartRandomHandler::isSetup()
{
   for (_setup_iterator it = _is_setup.begin(); it != _is_setup.end(); ++it) {
      if (!it->second)
         return false;
   }
   return true;
}

//------------------------------------------------------------------------------
/**
 * \brief Initializes the optimization for all SmartRandoms.
 *
 * Calls initializeOptimization() on all SmartRandoms that are not set up.
 *
 * \note This should be called before the start of the optimization.
 */
void SmartRandomHandler::initializeOptimization()
{
   Log::info("SmartRandomHandler") << "Setting up grids for adaptive integration." << Log::end;

   for (_iterator it = _smart_randoms.begin(); it != _smart_randoms.end(); ++it) {
      if (!_is_setup[it->second])
         it->second->initializeOptimization();
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Optimizes all SmartRandoms.
 *
 * Calls optimize() on all SmartRandoms that are not set up.
 *
 * \note This should be called after each iteration.
 */
void SmartRandomHandler::optimize()
{
   for (_iterator it = _smart_randoms.begin(); it != _smart_randoms.end(); ++it) {
      if (!_is_setup[it->second])
         it->second->optimize();
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Finalizes the optimization for all SmartRandoms.
 *
 * This calls finalizeOptimization() on all SmartRandoms, after which they are
 * considered as set up.
 *
 * \note This should be called after the last call to optimize.
 *
 * \fixme This should probably only be called on SmartRandoms that are not already set up?
 */
void SmartRandomHandler::finalizeOptimization()
{
   for (_iterator it = _smart_randoms.begin(); it != _smart_randoms.end(); ++it) {
      it->second->finalizeOptimization();
      _is_setup[it->second] = true;
   }

   Log::info("SmartRandomHandler") << "The grids have been initialized." << Log::end;
}

//------------------------------------------------------------------------------
/**
 * \brief Reads the optimization for all SmartRandoms from files.
 * \param fileStem The file stem for the input files.
 * \return \c true if the files for all channels have been found and loaded.
 *
 * The names of the files to read from are constructed as 'fileStem_name.dat'
 * where name is the name of the corresponding channel.
 *
 * \note If any one file is not found it returns \c false. A parsing error for
 * a found file will most likely result in an error being thrown.
 */
bool SmartRandomHandler::loadOptimization(const std::string& fileStem)
{
   bool success = true;
   for (_iterator it = _smart_randoms.begin(); it != _smart_randoms.end(); ++it) {
      if (!_is_setup[it->second]) {
         const std::string name = fileStem + format_tag(it->first) + ".dat";
         std::ifstream gridfile(name.c_str());
         if (gridfile && it->second->loadOptimization(gridfile))
            Log::verboseInfo() << "Loaded grid file: " << name << Log::end;
         else
            success = false;
      }
   }
   return success;
}

//------------------------------------------------------------------------------
/**
 * \brief Writes the optimization for all SmartRandoms to files.
 * \param fileStem The file stem for the output files.
 * \param seedTag The seed tag to add to each file.
 *
 * The file names are constructed as 'fileStem_name_seedTag.dat' where name is
 * the name of the corresponding channel.
 *
 * \note This needs to be called BEFORE calling optimize().
 */
void SmartRandomHandler::storeOptimization(const std::string& fileStem, const std::string& seedTag)
{
   for (_iterator it = _smart_randoms.begin(); it != _smart_randoms.end(); ++it) {
      if (!_is_setup[it->second]) {
         const std::string name = fileStem + format_tag(it->first) + seedTag + ".dat";
         std::ofstream gridfile(name.c_str());
         it->second->storeOptimization(gridfile);
         gridfile.close();
         if (!gridfile)
            Log::fatal("SmartRandomHandler::storeOptimization")
               << "Could not write to grid file: " << name << Log::end;
      }
   }
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Gets a pointer to the ISmartRandom for channel.
ISmartRandom* SmartRandomHandler::_get_smart_random(const IntegrationChannel& channel)
{
   _iterator it = _smart_randoms.find(channel.name());
   if (it == _smart_randoms.end())
      Log::fatal("SmartRandomHandler")
         << "The SmartRandom '" << channel.name() << "' does not exist." << Log::end;

   return it->second;
}

} // namespace Geneva
