//------------------------------------------------------------------------------
/// \file SmartRandomMint.cpp
//
// Author(s):
//   Christian Bauer
//
// Copyright:
//    Copyright (C) 2013 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of the SmartRandomMint Integrator and Generator class
//------------------------------------------------------------------------------

#include "Geneva/Generators/SmartRandomMint.hpp"

#include "Geneva/Core/Physics/Event.hpp"
#include "Geneva/Core/Physics/Signature.hpp"
#include "Geneva/Core/Physics/SignatureSet.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/Random.hpp"
#include "Geneva/Core/Utils/MathUtils.hpp"

#include <vector>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <sstream>

namespace Geneva
{
//------------------------------------------------------------------------------
SmartRandomMint::SmartRandomMint(size_t dimension, size_t numBins)
   : _dimension(dimension), _numbins(numBins)
{
   _weightPrevious = 0.;
   _weightNow = 0.;
   _weightSqPrevious = 0.;
   _weightSqNow = 0.;
   _numEventsNow = 0;
   _numEventsPrevious = 0;

   //initialize the vectors needed
   for (size_t dim = 0; dim < _dimension; dim++) {
      std::vector<double> gridfordim;
      std::vector<double> cumIntForDim;
      std::vector<double> cumBinProbForDim;

      for (size_t bin = 0; bin < _numbins + 1; bin++) {
         gridfordim.push_back((double)bin / ((double)_numbins));
         cumIntForDim.push_back((double)bin / ((double)_numbins));
      }
      for (size_t bin = 0; bin < _numbins; bin++) {
         cumBinProbForDim.push_back((double)(bin + 1) / ((double)_numbins));
      }
      _xGrid.push_back(gridfordim);
      _cumIntegral.push_back(cumIntForDim);
      _cumBinProb.push_back(cumBinProbForDim);

      _foldingInfo.push_back(1);
      _currentFolding.push_back(1);

      _currentBin.push_back(0);
      _currentFoldedBin.push_back(0);
      _posInBin.push_back(0.);
      _randomPoint.push_back(0.);
   }

   _wasInformed = true;
   _firstFolding = true;

   _gridExists = false;
   _inGridSetup = false;

   _hasChanged = false;
}

//------------------------------------------------------------------------------
/**
 * \brief Generate the next set of random number and return the generation probability
 * \param rand The random number to be filled
 * \param prob The probability with which this point was generated
 *
 * This function fills a new set of random numbers based on the
 * grid that is stored locally in SmartRandomMint in the vector passed in.
 * It saves the volume of the cell from which the random number was
 * generated from in volume. It returns true if the generation was
 * successful.
 *
 */
bool SmartRandomMint::nextRandom(std::vector<double>& rand, double& prob)
{
   if (!_wasInformed) {
      Log::error("SmartRandomMint::nextRandom") << "Asked for a new set of Random numbers from SmartRandomMint without passing the information of the calculation back" << Log::end;
      return false;
   }
   double volume = 1.;
   for (size_t dim = 0; dim < _dimension; dim++) {
      //Determine which bin to choose
      if (_numbins == 1) { //If there is only a single bin, no need to do anything
         _currentBin[dim] = 0;
      } else { //Determine the bin randomly using the cumulative bin probability
         double randCell = randomInterval();
         _currentBin[dim] = 0;
         while (randCell > _cumBinProb[dim][_currentBin[dim]])
            _currentBin[dim]++;
      }
      //Determine the position in the bin
      _posInBin[dim] = randomInterval();
      //Calculate the volume of the cell
      double dx = getBinSize(dim, _currentBin[dim]);
      volume *= dx * static_cast<double>(_numbins);
      //Get the value of the x's
      rand[dim] = _xGrid[dim][_currentBin[dim]] + _posInBin[dim] * dx;
   }
   prob = 1. / volume;
   if (_inGridSetup)
      _wasInformed = false; // This makes sure that we call informOfCalculation before the next set of random numbers is determined
   return true;
}

//------------------------------------------------------------------------------
/**
 * \brief Get the overall weight of the SmartRandom
 *
 * Returns the overall weight with which this channel should be sampled.
 * It is based on the average square of the weight, which optimizes the
 * total integral when summed over all channels.
 *
 */
double SmartRandomMint::overallWeight()
{
   return (_weightSqPrevious + _weightSqNow) / static_cast<double>(_numEventsNow + _numEventsPrevious);
}

//------------------------------------------------------------------------------
/**
 * \brief Initialize the optimization
 *
 * This sets all the things needed for the grid setup to the
 * starting values.
 *
 */
void SmartRandomMint::initializeOptimization()
{
   if (_gridExists) {
      Log::warning("SmartRandomMint::initializeOptimization") << "Trying to set up a Grid even though one already exists." << Log::end;
   }
   //First, initialize all things to zero
   _iteration = 0;
   _numEventsPrevious = 0;
   _numEventsNow = 0;
   _weightPrevious = 0.;
   _weightSqPrevious = 0.;
   _weightNow = 0.;
   _weightSqNow = 0.;

   //Set up the vectors that hold the cell integral and the cell count, and initialize them to zero
   _cellIntegral.clear();
   _cellCount.clear();

   //Set all the vectors to their starting values
   for (size_t dim = 0; dim < _dimension; dim++) {
      std::vector<double> cellIntegralForDim;
      std::vector<size_t> cellCountForDim;
      for (size_t bin = 0; bin < _numbins; bin++) {
         cellIntegralForDim.push_back(0.);
         cellCountForDim.push_back(0);
      }
      _cellIntegral.push_back(cellIntegralForDim);
      _cellCount.push_back(cellCountForDim);

   }
   //Set the flag stating that we are now in the setup phase of the grid
   _hasChanged = false;
   _inGridSetup = true;
}

//------------------------------------------------------------------------------
/**
 * \brief Informs the grid of the result of the calulation
 * \param result
 *
 * This function  takes the value of the function to be integrated and uses
 * its information to keep track of the information needed for the grid
 *
 * It updates the _cellIntegral and _cellCount, which provides the
 * information to adjust the grid later on. It also updates the counter _nuPoints
 * and the total square of the weights.
 *
 */
void SmartRandomMint::informOfCalculation(double result)
{
   //Will always work with the absolute value of the calculation to set up the grids
   result = abs(result);

   if (!_inGridSetup) {
      Log::warning("SmartRandomMint::informOfCalculation") << "No action taken if the grid is already adjusted" << Log::end;
      return;
   }

   //Keep track of the number of events seen and sum up the square of the result
   _numEventsNow++;
   _weightNow += result;
   _weightSqNow += result * result;

   //Keep track of the integral for each bin in each dimension
   for (size_t dim = 0; dim < _dimension; dim++) {
      _cellCount[dim][_currentBin[dim]]++;
      _cellIntegral[dim][_currentBin[dim]] += result;
   }
   _wasInformed = true;
   _hasChanged = true;
   return;
}

//------------------------------------------------------------------------------
/**
 * \brief Informs the grid an error has occurred
 * \param result
 *
 * This function informs that an error has occurred and reset the SmartRandom
 * to be able to generate new randoms
 *
 */
void SmartRandomMint::informOfFailure()
{
   if (!_inGridSetup) {
      Log::warning("SmartRandomMint::informOfFailure") << "No action taken if the grid is already adjusted" << Log::end;
      return;
   }
   _wasInformed = true;
   return;
}

//------------------------------------------------------------------------------
/**
 * \brief Adjusts the grid used in the importance sampling
 *
 * This function adjusts the grid in a way that the ratio of
 * Ii/Ni is constant for each bin and each dimension. In other
 * words, the function is adjusting the grids for the projection
 * of the function of a single variables (integrating over all
 * other variables). The number of bins per dimension is kept
 * fixed, and the grid lines are simply moved to accomplish this.
 * The heart of this function is using the cumulant of I/N, using
 * a linear interpolation betweeen gridpoints.
 *
 */
void SmartRandomMint::optimize()
{
   // only optimize if there was a change
   if (!_hasChanged)
      return;

   setCumIntegral(_cellIntegral, _cellCount);
   setNewxGrid(_xGrid, _cumIntegral);

   // reset parameters
   resetGridParameters();
   _hasChanged = false;
   _numEventsPrevious += _numEventsNow;
   _numEventsNow = 0;
   _weightPrevious += _weightNow;
   _weightNow = 0.;
   _weightSqPrevious += _weightSqNow;
   _weightSqNow = 0.;
   _iteration++;
}

//------------------------------------------------------------------------------
/**
 * \brief Finalize the optimization
 *
 * This resets a few variables
 *
 */
void SmartRandomMint::finalizeOptimization()
{
   _cellIntegral.clear();
   _cellCount.clear();

   _gridExists = true;
   _inGridSetup = false;
   _hasChanged = false;
}

//------------------------------------------------------------------------------
/**
 * \brief Sets the number of folding for a given dimension
 * \param dim
 * \param numFolding
 *
 * This sets the number of folding used for a given dimension to numFolding
 *
 */
void SmartRandomMint::setFolding(const size_t dim, const size_t numFolding)
{
   if (_numbins % numFolding != 0) {
      Log::error("SmartRandomMint::setFolding") << "The number of foldings for each dimension has to be a divisor of the number of bins" << Log::end;
   }
   _foldingInfo[dim] = numFolding;
}

//------------------------------------------------------------------------------
/**
 * \brief Generate the next set of random number and return the generation probability
 * \param rand The random number to be filled
 * \param prob The probability with which this point was generated
 *
 * This function fills a new set of random numbers based on the
 * grid that is stored locally in SmartRandomMint in the vector passed in.
 * It uses the folding, such that for subsequent calls the random numbers
 * are correlated.
 *
 * It saves the volume of the cell from which the random number was
 * generated from in volume. It returns true if the generatrion was
 * successful.
 *
 *
 */
bool SmartRandomMint::nextRandomFolded(std::vector<double>& rand, double& prob)
{
   //If I ask for the first folding, obtain the random number
   if (_firstFolding) {
      for (size_t dim = 0; dim < _dimension; dim++) {
         size_t numFolding = _numbins / _foldingInfo[dim];
         _currentFoldedBin[dim] = randomInteger(numFolding);
         _posInBin[dim] = randomInterval();
         _currentFolding[dim] = 1;
      }
      _firstFolding = false;
   }
   //Otherwise, find the next folding
   else {
      if (!findNextFolding()) {
         //If there are no more foldings, return false and reset the folding
         _firstFolding = true;
         return false;
      }
   }

   double volume = 1.;
   for (size_t dim = 0; dim < _dimension; dim++) {
      size_t numBinsFold = _numbins / _foldingInfo[dim];
      _currentBin[dim] = _currentFoldedBin[dim] + (_currentFolding[dim] - 1) * numBinsFold;
      //Calculate the volume of the cell
      double dx = getBinSize(dim, _currentBin[dim]);
      volume *= dx * static_cast<double>(numBinsFold);
      //Get the value of the x's
      rand[dim] = _xGrid[dim][_currentBin[dim]] + _posInBin[dim] * dx;
   }
   prob = 1. / volume;
   return true;
}

//------------------------------------------------------------------------------
/**
 * \brief Writes the grid information to \a out.
 *
 * It stores the square of the average weight for importance sampling amongst
 * different SmartRandomMints, as well as the grid itself.
 *
 * \todo Make this safer against corruption.
 */
bool SmartRandomMint::storeOptimization(std::ostream& out)
{
   out << std::fixed << std::setprecision(13);
   out << _numbins << "\n";
   out << _dimension << "\n";
   out << _weightPrevious << "\n";
   out << _weightNow << "\n";
   out << _weightSqPrevious << "\n";
   out << _weightSqNow << "\n";
   out << _numEventsPrevious << "\n";
   out << _numEventsNow << "\n";
   for (size_t dim = 0; dim < _dimension; dim++) {
      for (size_t bin = 0; bin < _numbins; bin++) {
         out << _cellCount[dim][bin] << " " << _cellIntegral[dim][bin] << " " << _xGrid[dim][bin] << "\n";
      }
   }
   return true;
}

//------------------------------------------------------------------------------
/**
 * \brief Reads the grid information from \a in.
 *
 * It reads the square of the average weight for importance sampling amongst
 * different SmartRandomMints, as well as the grid itself.
 *
 * \todo Make this safer against corrupted input. Should e.g. check if the read
 * _numbins and _dimension are consistent with what we want.
 */
bool SmartRandomMint::loadOptimization(std::istream& in)
{
   in >> _numbins;
   in >> _dimension;
   in >> _weightPrevious;
   in >> _weightNow;
   in >> _weightSqPrevious;
   in >> _weightSqNow;
   in >> _numEventsPrevious;
   in >> _numEventsNow;
   if (!in)
      Log::fatal("SmartRandomMint::loadOptimization")
         << "Could not parse the grid information." << Log::end;

   for (size_t dim = 0; dim < _dimension; dim++) {
      for (size_t bin = 0; bin < _numbins; bin++) {
         in >> _cellCount[dim][bin] >> _cellIntegral[dim][bin] >> _xGrid[dim][bin];
      }
      _xGrid[dim][_numbins] = 1.0;
   }
   if (!in)
      Log::fatal("SmartRandomMint::loadOptimization")
         << "Could not parse the grid information." << Log::end;

   _gridExists = true;

   setCumIntegral(_cellIntegral, _cellCount);
   setNewxGrid(_xGrid, _cumIntegral);

   // reset parameters
   resetGridParameters();
   _hasChanged = false;
   _numEventsPrevious += _numEventsNow;
   _numEventsNow = 0;
   _weightPrevious += _weightNow;
   _weightNow = 0.;
   _weightSqPrevious += _weightSqNow;
   _weightSqNow = 0.;

   return true;
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Reset the two arrays that keep track of the integration grid and count grid
 *
 * Sets the internal variables _cellIntegral and _cellCount to zero
 *
 */
void SmartRandomMint::resetGridParameters()
{
   for (size_t dim = 0; dim < _dimension; dim++) {
      for (size_t bin = 0; bin < _numbins; bin++) {
         _cellIntegral[dim][bin] = 0.;
         _cellCount[dim][bin] = 0.;
      }
   }

}

//------------------------------------------------------------------------------
/**
 * \brief Simple helper routine to return the size of a bin
 *
 * This simply calculates the difference beteen the two grid edges
 */
double SmartRandomMint::getBinSize(size_t dim, size_t bin)
{
   return _xGrid[dim][bin + 1] - _xGrid[dim][bin];
}

//------------------------------------------------------------------------------
/**
 * \brief Simple helper routine to obtain the next folding
 *
 * This cycles through all possible combinations of the folded dimensions
 */
bool SmartRandomMint::findNextFolding()
{
   size_t dim = _dimension;
   while (dim > 0) {
      --dim;
      if (_currentFolding[dim] < _foldingInfo[dim]) {
         _currentFolding[dim]++;
         return true;
      }
      _currentFolding[dim] = 1;
   }
   return false;
}

//------------------------------------------------------------------------------
/**
 * \brief Helper routine to calculate the cumulant integral
 *
 * This takes the information from _cellIntegral and _cellCount and
 * determins the _cumIntegral
 */
void SmartRandomMint::setCumIntegral(std::vector< std::vector <double> >cellIntegral, std::vector< std::vector <size_t> > cellCount)
{
   for (size_t dim = 0; dim < _dimension; dim++) {
      //Set the cumulative integral based on the integral filled in previous iteration
      _cumIntegral[dim][0] = 0.;
      for (size_t bin = 0; bin < _numbins; bin++) {
         if (_cellCount[dim][bin] != 0) {
            double RoverN = abs(cellIntegral[dim][bin]) / static_cast<double>(cellCount[dim][bin]);
            _cumIntegral[dim][bin + 1] = _cumIntegral[dim][bin] + RoverN;
         } else
            _cumIntegral[dim][bin + 1] = _cumIntegral[dim][bin];
      }
      // If there is no value, keep old grid and go to the next dimension
      if (_cumIntegral[dim][_numbins] == 0)
         continue;
      //Otherwise, normalize the cumulative integral
      for (size_t bin = 0; bin < _numbins + 1; bin++) {
         _cumIntegral[dim][bin] /= _cumIntegral[dim][_numbins];
      }
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Helper routine to calculate the new xGrid
 *
 * This takes the information from the old xGrid and the cumulative integral and
 * determins the _cumIntegral
 *
 * The algorithm works
 * by drawing a function with x on the x-axis and I/N on
 * the y axis. At each grid edge,
 * this function is equal to the recorded value, with
 * straight lines between the
 * different grid edges. It is then solving the
 * linear equation giving the points
 * on the x-axis (newGridForDim) based on a uniform
 * division of the y axis.
 */
void SmartRandomMint::setNewxGrid(std::vector< std::vector <double> >xGridOld, std::vector< std::vector <double> > cumIntegral)
{
   for (size_t dim = 0; dim < _dimension; dim++) {

      //If there was no integral in the given dimension, leave the xGrid unchanged
      if (cumIntegral[dim][_numbins] == 0)
         continue;

      //First, create a vector for the new grid
      std::vector<double> newGridForDim(_numbins + 1, 0.);
      for (size_t bin = 1; bin < _numbins; bin++) {
         double r = (double)bin / ((double)_numbins);
         size_t j = 1;
         while (r > cumIntegral[dim][j]) {
            j++;
         }
         assert(j > 0);
         //Now find the new grid locations
         if (cumIntegral[dim][j] == cumIntegral[dim][j - 1])
            Log::error("SmartRandomMint::setNewxGrid") << "Can not find a new grid position" << Log::end;
         newGridForDim[bin] = (r - cumIntegral[dim][j - 1]);
         newGridForDim[bin] /= cumIntegral[dim][j] - cumIntegral[dim][j - 1];
         newGridForDim[bin] *= xGridOld[dim][j] - xGridOld[dim][j - 1];
         newGridForDim[bin] += xGridOld[dim][j - 1];
      }
      newGridForDim[_numbins] = 1.;

      //Set the local grid to the new grid we just determined
      for (size_t bin = 0; bin <= _numbins; bin++) {
         _xGrid[dim][bin] = newGridForDim[bin];
      }
   }
}

} //namespace Geneva
