//------------------------------------------------------------------------------
/// \file PythiaDriver.cpp
//
// Author(s):
//    Christian Bauer, Simone Alioli
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class PythiaDriver
//------------------------------------------------------------------------------

#include "Geneva/config.hpp"

#include "Geneva/Driver/YamlToXmlTranslation.hpp"

#include "Geneva/PythiaInterface/PythiaDriver.hpp"
#include "Geneva/PythiaInterface/PythiaUtils.hpp"
#include "Geneva/PythiaInterface/UserHookRestriction.hpp"
#include "Geneva/PythiaInterface/UHRppB012.hpp"

#include "Geneva/Calculations/CalculationFactory.hpp"
#include "Geneva/Calculations/ICalculation.hpp"

#include "Geneva/Components/ComponentFactory.hpp"
#include "Geneva/Components/CouplingStrong.hpp"

#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"
#include "Geneva/Core/UI/OptionParserXML.hpp"
#include "Geneva/Core/UI/OptionParserYAML.hpp"
#include "Geneva/Core/UI/OptionParserLHEFHeader.hpp"

#ifdef geneva_USE_ZLIB
#include "Geneva/gzstream.hpp"
#endif

#include <iostream>

namespace Geneva
{

namespace
{

#ifdef geneva_USE_ZLIB
typedef gzstream::igzstream _istream_type;
#else
typedef std::ifstream _istream_type;
#endif

//------------------------------------------------------------------------------
bool string_begins_with_any(const std::string input, std::vector<std::string> matches)
{
   for (std::vector<std::string>::iterator it = matches.begin(); it != matches.end(); ++it) {
      if (input.compare(0, it->length(), *it) == 0)
         return true;
   }
   return false;
}

//------------------------------------------------------------------------------
OptionMap getOptMap(std::istream& lhefInputStream, const std::string& genevaOptionsFile, bool enforceLhefOptions)
{
   //Set all the default options
   OptionParserXML defaultOptions;
   OptionMap optMap;
   optMap.createDefaultOptions(defaultOptions);

   std::string extension;
   std::string::size_type idx;
   idx = genevaOptionsFile.rfind('.');
   if (idx != std::string::npos)
      extension = genevaOptionsFile.substr(idx+1);
   if (extension == "xml") {
      OptionParserXML parser(genevaOptionsFile);
      optMap.setUserOptions(parser, true);
   } else if (extension == "yml") {
      // parse the yaml options
      OptionParserYAML parser(genevaOptionsFile);
      OptionMap yamlOptions;
      yamlOptions.createDefaultOptions(parser);

      // translate yaml options in new layout to old layout
      optMap.setUserOptions(convertYAMLToXML(yamlOptions), true);
   } else
      Log::fatal("PythiaDriver::getOptMap") << "Provided option file " << genevaOptionsFile << " is not a .xml or .yml file." << Log::end;

   // Check consistency with the options in the LHEF header.
   // Explicitly overwrite the fileStem option as read from the LHEF header.
   // Exclude any options that start with the following strings:
   std::vector<std::string> excludeFromCheck;
   excludeFromCheck.push_back("Random");
   excludeFromCheck.push_back("EventGenerator");
   excludeFromCheck.push_back("IntegrationChannel");
   excludeFromCheck.push_back("Analyzer");
   excludeFromCheck.push_back("PP2BosonJets");
   excludeFromCheck.push_back("UserHookRestriction");
   excludeFromCheck.push_back("UHR");
   excludeFromCheck.push_back("TreeME");
   excludeFromCheck.push_back("OneLoopME");
   excludeFromCheck.push_back("overwriteOutput");
   excludeFromCheck.push_back("writeReport");
   excludeFromCheck.push_back("writeEvents");
   excludeFromCheck.push_back("showTimer");
   excludeFromCheck.push_back("verbose");
   excludeFromCheck.push_back("Mint");
   excludeFromCheck.push_back("LHEF");
   excludeFromCheck.push_back("optionsFile");
   excludeFromCheck.push_back("filePath");
   excludeFromCheck.push_back("logFile");
   excludeFromCheck.push_back("num");
   excludeFromCheck.push_back("maxTime");
   excludeFromCheck.push_back("Generator");
   excludeFromCheck.push_back("SplittingFunction");
   excludeFromCheck.push_back("CalcSCETppV012::ImprovedXSFix");
   bool hasDifference = false;
   OptionParserLHEFHeader parserHeader(lhefInputStream);
   while (parserHeader.hasNext()) {
      std::string name, value;
      parserHeader.getNext(name, value);
      //Replace the seed number in the optionMap with the one the lhe file was generated with
      if (name == "seed") {
         optMap.setOption(name, value);
      }
      if (string_begins_with_any(name, excludeFromCheck))
         continue;
      if (!optMap.optionExists(name)) {
         Log::warning("PythiaDriver::getOptMap") << "Unknown option \"" << name << "\" in LHEF header." << Log::end;
         hasDifference = true;
      } else if (name == "fileStem")
         optMap.setOption(name, value);
      else if (value != optMap.getOption(name)) {
         Log::warning("PythiaDriver::getOptMap")
            << "Option \"" << name << "\" changed from "
            << TextColor::kBoldRed << value
            << TextColor::kReset << " (LHEF header) to "
            << TextColor::kGreen << optMap.getOption(name)
            << TextColor::kReset << " (now)."  << Log::end;
         hasDifference = true;
      }
   }
   // throw error if we are enforcing no differences
   if (enforceLhefOptions && hasDifference)
      Log::fatal("PythiaDriver::getOptMap") << "Inconsistent options between LHEF header and user-provided options." << Log::end;

   return optMap;
}

//------------------------------------------------------------------------------
/// Reads the number of events from \a in.
size_t readNumEvents(_istream_type& in)
{
   size_t numEvents = 0;
#ifndef geneva_USE_ZLIB
   // First try to read it from the tag in the footer of the file.
   if (in.seekg(-4096, std::ios::end)) { //Random access not possible if file is gzipped
#endif
      std::string line;
      while (std::getline(in, line)) {
         if (line.find("#   number of events in file :") != std::string::npos) {
            std::stringstream ss;
            ss.str(line.substr(line.find(':') + 1));
            ss >> numEvents;
            break;
         }
      }
#ifndef geneva_USE_ZLIB
   }
#endif
   // If it was not possible to read the tag, try and count the events by explicitly
   // reading from the beginning
   // (This is not very efficient as it requires reading in the entire file, but
   // maintains backward compatibility with older files that don't have the tag.)
   if (numEvents == 0) {
#ifdef geneva_USE_ZLIB
      in.rewind();
#else
      in.clear();
      in.seekg(0);
#endif
      Log::warning("PythiaDriver::readNumEvents") << "Could not read number of events from the end of the LHE file. Trying to count events ..." << Log::end;
      std::string line;
      while (std::getline(in, line)) {
         if (line == "</event>")
            numEvents++;
      }
      if (numEvents == 0)
         Log::warning("PythiaDriver::readNumEvents") << "Could not determine number of events in the LHE file." << Log::end;
   }
   return numEvents;
}

} // anonymous namespace

//------------------------------------------------------------------------------
/**
 * \param lhefInputFile name of LHEF file to read from
 * \param genevaOptionsFile name of Geneva options file
 * \param enforceLhefOptions whether to enforce consistency with the options in
 *        LHEF header
 * \param pyOptionsFile name of Pythia8 options file
 * \param maxEvents Maximum number of events to shower
 * \param fractionPythiaFailAllowed Fraction of events allowed to fail
 *
 * If \a maxEvents is 0, all events from \a lhefInputFile will be used. If
 * \a fractionPythiaFailAllowed is <= 0. no events are allowed to fail the
 * showering, otherwise \a fractionPythiaFailAllowed times the total number of
 * events (rounded up) is allowed to fail, where the total number of events is
 * the smaller of maxEvents and the number of events in the file.
 */
PythiaDriver::PythiaDriver(const std::string& lhefInputFile, const std::string& genevaOptionsFile, bool enforceLhefOptions, const std::string& pyOptionsFile,
                           size_t maxEvents, double fractionPythiaFailAllowed)
   : _pythia(config::pythia8_data_dir, true), _nPythiaEvents(0), _nPythiaFailAllowed(0), _nPythiaFail(0), _nOverVeto(0)
{
   _istream_type lhefFile(lhefInputFile.c_str());
   if (!lhefFile.good())
      Log::fatal("PythiaDriver") << "Could not find the LHEF file " << lhefInputFile << Log::end;
   _optMap = getOptMap(lhefFile, genevaOptionsFile, enforceLhefOptions);
   Log::verboseInfo("PythiaDriver") << "Loaded options from " << genevaOptionsFile << '.' << Log::end;

   _numEvents = readNumEvents(lhefFile);
   lhefFile.close();
   if (_numEvents == 0) {
      // if numEvents could not be read from the file try to use up to maxEvents
      Log::info("PythiaDriver") << "Trying to process up to " << maxEvents << " events of file " << lhefInputFile << Log::end;
      _numEvents = maxEvents;
   } else if (maxEvents && maxEvents < _numEvents) {
      // only process first maxEvents
      Log::info("PythiaDriver") << "Processing " << maxEvents << " of " << _numEvents << " events of file " << lhefInputFile << Log::end;
      _numEvents = maxEvents;
   } else
      Log::info("PythiaDriver") << "Processing all " << _numEvents << " events of file " << lhefInputFile << Log::end;

   if (fractionPythiaFailAllowed > 0.)
      _nPythiaFailAllowed = static_cast<size_t>(fractionPythiaFailAllowed * static_cast<double>(_numEvents)) + 1;

   init(lhefInputFile, genevaOptionsFile, enforceLhefOptions, pyOptionsFile, _optMap.getOption<bool>("verbose"));
}

//------------------------------------------------------------------------------
PythiaDriver::~PythiaDriver()
{
   delete _userhook;
}

//------------------------------------------------------------------------------
UserHookRestriction* PythiaDriver::getUserHook(const std::string& lhefInputFile, const std::string& genevaOptionsFile, bool enforceLhefOptions)
{
   _istream_type lhefFile(lhefInputFile.c_str());
   if (!lhefFile.good())
      Log::fatal("PythiaDriver::getUserHook") << "Could not find the LHEF file " << lhefInputFile << Log::end;
   UserHookRestriction* userhook = getUserHook(lhefFile, genevaOptionsFile, enforceLhefOptions);
   lhefFile.close();
   return userhook;
}

//------------------------------------------------------------------------------
UserHookRestriction* PythiaDriver::getUserHook(std::istream& lhefInputStream, const std::string& genevaOptionsFile, bool enforceLhefOptions)
{
   OptionMap optMap = getOptMap(lhefInputStream, genevaOptionsFile, enforceLhefOptions);

   UserHookRestriction* userhook = 0;
   // setup the UHR based on options given
   std::string uhr = optMap.getOption<std::string>("UserHookRestriction");
   if (uhr == "None") {
      userhook = new UserHookRestrictionNone();
      Log::warning("PythiaDriver::getUserHook") << "Running Pythia with no UserHookRestriction" << Log::end;
   } else {
      const std::string calc = optMap.getOption<std::string>("calc");

      if (calc == "CalcSCETppV012") {
         if (uhr == "ppB012") {
            size_t maxJets = (optMap.getOption<std::string>("CalcSCETppV012::fixed1jet") == "LO") ? 1 : 2;
            //NOTE: the number of final state partons is not determined by the order of the 1-jet bin calculation in presence of Tau1 resummation
            if (maxJets == 1 and optMap.getOption<std::string>("CalcSCETppV012::run1") != "none")
               maxJets = 2;
            const double tau0cut = optMap.getOption<double>("NJettiness::Tau0cut");
            double tau1cut = optMap.getOption<double>("NJettiness::Tau1cut");
            if (optMap.getOption<std::string>("CalcSCETppV012::run1") != "none"
                && optMap.getOption<bool>("CalcSCETppV012::run1::split1Jet")) {
               CouplingStrong* alphas = ComponentFactory::instance().getCouplingStrong(optMap);
               tau1cut = std::max(optMap.getOption<double>("CalcSCETppV012::run1::split1Jet::Tau1IR"), 1.1 * alphas->lambdaQCD());
               delete alphas;
            }

            const bool removeMPI = optMap.getOption<bool>("UHRppB012::removeMPI");
            if (!removeMPI)
               Log::warning(" PythiaDriver::getUserHook") << "Running without UHRppB012::removeMPI == true is unphysical" << Log::end;
            const double pTmax = optMap.getOption<double>("UHRppB012::pTmax");
            const double lambda = optMap.getOption<double>("UHRppB012::lambda");
            size_t maxAttempts = size_t_from_int(optMap.getOption<int>("UHRppB012::maxRetry"));
            const int maxISRstep = optMap.getOption<int>("UHRppB012::maxISRstep");
            const int maxFSRstep = optMap.getOption<int>("UHRppB012::maxFSRstep");
            const bool useTauClust = optMap.getOption<bool>("UHRppB012::useTauClust");
            const bool sumTauClust = optMap.getOption<bool>("UHRppB012::useTauClust::sumTauClust");
            const bool useHardDirections = optMap.getOption<bool>("UHRppB012::useTauClust::useHardDirections");
            const double tau1MaxRatio = optMap.getOption<double>("CalcSCETppV012::run1::tau1MaxRatio");

            //Set some options that avoid having unneeded initialization files
            optMap.setOption("CalcSCETppV012::TotalXSFix", "off");
            optMap.setOption("CalcSCETppV012::ImprovedXSFix", "off");
            optMap.setOption("CalcSCETppV012::phi1", "mult");
            optMap.setOption("CalcSCETppV012::run1", "none");

            userhook = new UHRppB012(CalculationFactory::instance().getCalculation(optMap)->getRestriction(), maxAttempts, maxJets, tau0cut, tau1cut, lambda, removeMPI, pTmax, size_t_from_int(maxISRstep), size_t_from_int(maxFSRstep), useTauClust, sumTauClust, useHardDirections, tau1MaxRatio);
         } else {
            Log::error("PythiaDriver::getUserHook") << "Unknown Pythia UserHookRestriction '" << uhr << "' for calculation '" << calc << "'." << Log::end;
         }
      } else {
         Log::error("PythiaDriver::getUserHook") << "Cannot create UserHook for calculation '" << calc << "'." << Log::end;
      }
   }
   return userhook;
}

//------------------------------------------------------------------------------
std::vector<std::string> PythiaDriver::getRequiredSettings(const std::string& lhefInputFile, const std::string& genevaOptionsFile, bool enforceLhefOptions)
{
   _istream_type lhefFile(lhefInputFile.c_str());
   if (!lhefFile.good())
      Log::fatal("PythiaDriver::getRequiredSettings") << "Could not find the LHEF file " << lhefInputFile << Log::end;
   std::vector<std::string> requiredSettings = getRequiredSettings(lhefFile, genevaOptionsFile, enforceLhefOptions);
   lhefFile.close();
   return requiredSettings;
}

//------------------------------------------------------------------------------
/**
 * \brief Returns a list of required settings to be put on Pythia
 */
std::vector<std::string> PythiaDriver::getRequiredSettings(std::istream& lhefInputStream, const std::string& genevaOptionsFile, bool enforceLhefOptions)
{
   OptionMap optMap = getOptMap(lhefInputStream, genevaOptionsFile, enforceLhefOptions);

   std::vector<std::string> requiredSettings;
   //Some options required if we remove MPI
   const bool removeMPI = optMap.getOption<bool>("UHRppB012::removeMPI");
   if (removeMPI) {
      requiredSettings.push_back("MultipartonInteractions:allowRescatter = false");
      requiredSettings.push_back("SecondHard:generate = false");
   }

   // Add settings that are required for Pythia to be consistent with Geneva
   requiredSettings.push_back("SpaceShower:pTmaxMatch = 1");
   requiredSettings.push_back("TimeShower:pTmaxMatch = 1");
   requiredSettings.push_back("SpaceShower:pTmaxFudge = 1.00000e+00");
   requiredSettings.push_back("TimeShower:pTmaxFudge = 1.00000e+00");
   requiredSettings.push_back("SpaceShower:MEcorrections = false");
   requiredSettings.push_back("TimeShower:MEcorrections = false");
   requiredSettings.push_back("LesHouches:setQuarkMass = 0");

   return requiredSettings;
}
//------------------------------------------------------------------------------
/**
 * \brief Showers the next event.
 * \return \c true if Pythia succeeded, \c false otherwise (internal Pythia
 *         error or the end of the file has been reached)
 *
 * Calls pythia.next() to read the next event from the LHEF file and shower it.
 * Keeps track of the number of vetoes and pythia failures as well as the Sigma
 * of the last event.
 */
bool PythiaDriver::next()
{
   if (!_pythia.next()) {
      // unexpected end of file
      if (_pythia.info.atEndOfFile()) {
         if (_nPythiaEvents < _numEvents)
            Log::error("PythiaDriver::next") << "Unexpected end of file, could only read " << _nPythiaEvents << " of expected " << _numEvents << " events." << Log::end;
         return false;
      }

      // failure was due to something internal to pythia
      _nPythiaEvents++;
      _nPythiaFail++;
      Log::warning("PythiaDriver::next") << "Pythia failed to shower event " << _nPythiaEvents << ", discarding event." << Log::end;
      if (Log::debug().isOn()) {
         Log::debug("PythiaDriver::next") << "Discarded event:" << Log::end;
         std::streambuf* cout_rdbuf = std::cout.rdbuf();
         std::cout.rdbuf(Log::debug().outStream().rdbuf());
         _pythia.event.list();
         std::cout.rdbuf(cout_rdbuf);
      }
      if (_nPythiaFail > _nPythiaFailAllowed)
         Log::fatal("PythiaDriver::next") << "Pythia showering failed more than " << _nPythiaFailAllowed << " times." << Log::end;
      return false;
   }

   if (_userhook->isOverVetoed()) {
      Log::warning("PythiaDriver::next") << "Too many vetoes showering event " << _nPythiaEvents << ", keeping showered event." << Log::end;
      _nOverVeto++;

      const size_t nfinal = size_t_from_ptrdiff_t(std::count_if(_userhook->hardProcess().beginFinal(), _userhook->hardProcess().endFinal(), IsQCD()));
      _nOverVetoFromMultiplicity[nfinal]++;
   }

   _lastWeight.central() = _pythia.info.weight();
   // if multiple weights are present, add them as variations
   if (_pythia.info.weights_compressed && !_pythia.info.weights_compressed->empty()) {
      // cross check if the first of multiple weights is identical to the central
      assert(compare(_pythia.info.weight(), *_pythia.info.weights_compressed->begin()) == 0);
      _lastWeight.variations().clear();
      for (std::vector<double>::iterator it = ++(_pythia.info.weights_compressed->begin()); it != _pythia.info.weights_compressed->end(); ++it)
         _lastWeight.variations().push_back(*it);
   }
   _lastWeight.setDefined();

   _nPythiaEvents++;
   return true;
}

//------------------------------------------------------------------------------
Event PythiaDriver::getPartonicEvent() const
{
   Event event(Configuration(), lastWeight());
   processToConfig(_pythia.process, event.config());
   return event;
}

//------------------------------------------------------------------------------
Event PythiaDriver::getShoweredEvent() const
{
   Event event(Configuration(), lastWeight());
   partonLevelToConfig(_userhook->partonLevel(), event.config(), false);
   return event;
}

//------------------------------------------------------------------------------
Event PythiaDriver::getShoweredNoMPIEvent() const
{
   Event event(Configuration(), lastWeight());
   partonLevelToConfig(_userhook->partonLevel(), event.config(), true, 0.0);
   return event;
}

//------------------------------------------------------------------------------
Event PythiaDriver::getHadronizedEvent() const
{
   Event event(Configuration(), lastWeight());
   fullEventToConfig(_pythia.event, event.config());
   return event;
}

//------------------------------------------------------------------------------
void PythiaDriver::stat()
{
   _pythia.stat();
   Log::info("PythiaDriver") << "Number of read events: " << _nPythiaEvents << Log::end;
   Log::info("PythiaDriver") << "Number of successful events: " << _nPythiaEvents - _nPythiaFail - _nOverVeto << Log::end;
   Log::info("PythiaDriver") << "Number of over-vetoed (kept) events: " << _nOverVeto << Log::end;
   for (std::map<size_t, size_t>::iterator it = _nOverVetoFromMultiplicity.begin(); it != _nOverVetoFromMultiplicity.end(); ++it) {
      Log::info("PythiaDriver") << "(Over-vetoed " << it->first << "-parton events: " << it->second << ')' << Log::end;
   }
   Log::info("PythiaDriver") << "Number of failed (discarded) events: " << _nPythiaFail << Log::end;
   if (Log::verboseInfo().isOn())
      _userhook->printStatistics(Log::verboseInfo().outStream());
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/**
 * \brief Initialize the class based on internal _optMap
 *
 * - Set up the UserHookRestriction
 * - Set up the LHEF reader
 * - Initialize Pythia
 * - Set all pythia options required for consistency
 * - Set some counters needed
 */
void PythiaDriver::init(const std::string& lhefInputFile, const std::string& genevaOptionsFile, bool enforceLhefOptions, const std::string& pyOptionsFile, bool verbose)
{
   if (verbose)
      Log::setVerbosity(Log::kVerboseInfo);

   // Set the UserHookRestrition
   _userhook = getUserHook(lhefInputFile, genevaOptionsFile, enforceLhefOptions);
   _pythia.setUserHooksPtr(_userhook);

   // Set up the LHEF reader
   _pythia.readString("Beams:frameType = 4");
   _pythia.readString("Beams:LHEF = "+lhefInputFile);

   // Initialize Pythia
   // set Pythia options from given file
   if (!pyOptionsFile.empty()) {
      if (!_pythia.readFile(pyOptionsFile))
         Log::fatal("PythiaDriver::init") << "Pythia8 could not read the options file " << pyOptionsFile << Log::end;
   }

   //Set all required settings at the end, to make sure they are correct
   setRequiredOptions(getRequiredSettings(lhefInputFile, genevaOptionsFile, enforceLhefOptions));
   if (!_pythia.init())
      Log::fatal("PythiaDriver::init") << "Pythia failed to initialize." << Log::end;
}

//------------------------------------------------------------------------------
void PythiaDriver::setRequiredOptions(const std::vector<std::string>& requiredOptions)
{
   std::string delim = "=";
   for (std::vector<std::string>::const_iterator optString = requiredOptions.begin(); optString != requiredOptions.end(); ++optString) {
      size_t pos = optString->find(delim);
      if (pos == std::string::npos)
         Log::error("PythiaDriver::setRequiredOptions") << "The string " << *optString << " is not a valid Pythia option string" << Log::end;
      std::string key = optString->substr(0, pos);
      std::string value = optString->substr(pos+1, std::string::npos);
      //Remove any whitespaces
      key.erase(std::remove_if(key.begin(), key.end(), ::isspace), key.end());
      value.erase(std::remove_if(value.begin(), value.end(), ::isspace), value.end());
      std::string currentValue = _pythia.settings.output(key, false);
      if (currentValue != value)
      {
         Log::info("PythiaDriver::setRequiredOptions") << "Changing Pythia option " << key << " from current value " << currentValue << " to required value " << value << Log::end;
         _pythia.readString(*optString);
      }
   }
}

} // namespace Geneva
