//------------------------------------------------------------------------------
/// \file PythiaUtils.cpp
//
// Author(s):
//    Calvin Berggren, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 Frank Tackmann, LBL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Basic algorithms needed for the interface with Pythia.
//------------------------------------------------------------------------------

#include "Geneva/PythiaInterface/PythiaUtils.hpp"

#include "Geneva/Core/Physics/PdgID.hpp"
#include "Geneva/Core/Physics/Configuration.hpp"
#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/UI/Log.hpp"

#include "Pythia8/Event.h"

#include <set>

namespace Geneva
{

namespace
{

//------------------------------------------------------------------------------
void getAllInitialMothers(const Pythia8::Event& event, int index, std::set<int>& mothers)
{
   // trace back as far as is unambiguous
   while (event[index].mother1() > 2
          && (event[index].mother2() == 0 || event[index].mother2() == event[index].mother1()))
      index = event[index].mother1();

   // reached the beams
   if (event[index].mother1() <= 2) {
      mothers.insert(index);
      return;
   }

   // reached a point with two mothers, keep tracing both
   getAllInitialMothers(event, event[index].mother1(), mothers);
   getAllInitialMothers(event, event[index].mother2(), mothers);
}

//------------------------------------------------------------------------------
inline int getFirstInitialMother(const Pythia8::Event& event, int index)
{
   // trace back the first mother until it is one of the beams (1 or 2)
   while (event[index].mother1() > 2)
      index = event[index].mother1();
   return index;
}

}

//------------------------------------------------------------------------------
/**
 * \brief Converts a Pythia8 Event to a Configuration.
 * \param[in] event The Pythia8 Event to convert
 * \param[out] conf The Configuration to be filled with the converted Event
 *
 * This takes indiscriminantely all final-state particles and uses the beams as
 * initial particles.
 *
 * This is sensible for e+e- collisions (partonic or hadronized) and for
 * hadronic collisions after beam remnants (with or without further
 * hadronization). To convert a partonic event in hadronic collisions before
 * beam remnants use \c partonicEventToConfig().
 */
//------------------------------------------------------------------------------
void fullEventToConfig(const Pythia8::Event& event, Configuration& conf)
{
   conf.clear();
   for (int i = 0; i < event.size(); ++i) {
      if (event[i].isFinal())
         conf.insertFinal(Particle(PdgID(event[i].id()), event[i].e(), event[i].px(), event[i].py(), event[i].pz()));
   }

   // set initial-state particles from incoming beams which are assumed to be
   // in positions 1 and 2
   assert(event[1].statusAbs() == 12);
   assert(event[2].statusAbs() == 12);
   if (event[1].pz() > 0.) {
      conf.setInitialA(Particle(PdgID(event[1].id()), event[1].e(), event[1].px(), event[1].py(), event[1].pz()));
      assert(event[2].pz() <= 0.);
      conf.setInitialB(Particle(PdgID(event[2].id()), event[2].e(), event[2].px(), event[2].py(), event[2].pz()));
   } else {
      conf.setInitialB(Particle(PdgID(event[1].id()), event[1].e(), event[1].px(), event[1].py(), event[1].pz()));
      assert(event[2].pz() <= 0.);
      conf.setInitialA(Particle(PdgID(event[2].id()), event[2].e(), event[2].px(), event[2].py(), event[2].pz()));
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Converts a partonic Pythia8 Event to a Configuration.
 * \param[in] event The Pythia8 Event to convert
 * \param[in] removeMPI Whether to remove MPIs
 * \param[in] pTminMPI  Min pT of MPI to remove
 * \param[out] conf The Configuration to be filled with the converted Event
 *
 * This uses the hard incoming partons (at the end of backward ISR evolution) as
 * the initial particles and takes all final-state particles which are directly
 * connected to them. If \a removeMPI is \c false it considers both the primary
 * (status code 21) and subsequent (MPI, status code 31) incoming partons. When
 * \a removeMPI is \c true, only the MPI processes where the core collision has
 * pT < pTminMPI are included. Setting pTminMPI = 0.0 (the default) removes all
 * MPI contributions.
 *
 * \note The momentum of the incoming MPI partons are added to the initial
 * momenta in the Configuration, so the Configuration should satisfy momentum
 * conservation. The id of the incoming partons will be set to the first ones
 * in Pythia's event record (which should correspond to the primary hard
 * interaction).
 */
void partonLevelToConfig(const Pythia8::Event& event, Configuration& conf, bool removeMPI, double pTminMPI)
{
   conf.clear();
   std::set<int> hardInitial;
   for (int i = 0; i < event.size(); ++i) {
      if (event[i].statusAbs() == 21 || (event[i].statusAbs() == 31 && !removeMPI))
         hardInitial.insert(getFirstInitialMother(event, i));
      else if (removeMPI && event[i].statusAbs() == 33 && event[i].pT() < pTminMPI)
         // with removeMPI only include subprocesses below pTminMPI
         getAllInitialMothers(event, i, hardInitial);
   }

   // add all final-state particles not coming from mpi
   for (int i = 0; i < event.size(); i++) {
      // here it is sufficient to trace back the first mother
      if (event[i].isFinal() && hardInitial.count(getFirstInitialMother(event, i)))
         conf.insertFinal(Particle(PdgID(event[i].id()), event[i].e(), event[i].px(), event[i].py(), event[i].pz()));
   }

   // set initial-state particles
   // we use the first one as the nominal particle and add the momentum of
   // additional incoming ones to it. This is the best we can do within the limits of Configuration
   Particle pA, pB;
   for (std::set<int>::const_iterator it = hardInitial.begin(); it != hardInitial.end(); ++it) {
      if (event[*it].pz() > 0.) {
         if (pA.id() == PdgID::kUnknown)
            pA = Particle(PdgID(event[*it].id()), event[*it].e(), event[*it].px(), event[*it].py(), event[*it].pz());
         else
            pA.p() += FourVector(event[*it].e(), event[*it].px(), event[*it].py(), event[*it].pz());
      } else {
         if (pB.id() == PdgID::kUnknown)
            pB = Particle(PdgID(event[*it].id()), event[*it].e(), event[*it].px(), event[*it].py(), event[*it].pz());
         else
            pB.p() += FourVector(event[*it].e(), event[*it].px(), event[*it].py(), event[*it].pz());
      }
   }
   if (pA.id() != PdgID::kUnknown)
      conf.setInitialA(pA);
   if (pB.id() != PdgID::kUnknown)
      conf.setInitialB(pB);
}

//------------------------------------------------------------------------------
/**
 * \brief Converts a Pythia8 Event to a Configuration, only the hard process
 * \param[in] event The Pythia8 Event to convert
 * \param[out] conf The Configuration to be filled with the converted Event
 */
//------------------------------------------------------------------------------
void processToConfig(const Pythia8::Event& event, Configuration& conf)
{
   conf.clear();
   for (int i = 0; i < event.size(); ++i) {
      if (event[i].statusAbs() == 21) {
         if (event[i].pz() > 0.) {
            assert(conf.initialA() == conf.end());
            conf.setInitialA(Particle(PdgID(event[i].id()), event[i].e(), event[i].px(), event[i].py(), event[i].pz()));
         } else {
            assert(conf.initialB() == conf.end());
            conf.setInitialB(Particle(PdgID(event[i].id()), event[i].e(), event[i].px(), event[i].py(), event[i].pz()));
         }
      } else if (event[i].statusAbs() == 23) {
         conf.insertFinal(Particle(PdgID(event[i].id()), event[i].e(), event[i].px(), event[i].py(), event[i].pz()));
      } else if (event[i].statusAbs() == 22) {
         // make sure that this particle does not decay into other hard interaction particles; assumes that both daughters would be code 23
         if (event[i].isFinal() || (event[event[i].daughter1()].statusAbs() != 23 && event[event[i].daughter2()].statusAbs() != 23)) {
            conf.insertFinal(Particle(PdgID(event[i].id()), event[i].e(), event[i].px(), event[i].py(), event[i].pz()));
         }
      }
   }
}

//------------------------------------------------------------------------------
/**
 * \brief Calculates TauNPythia for e+e- collisions
 *
 * Clusters the event back according to Pythia's splitting history and computes
 * TauN from the resulting massive N-body phase-space point.
 * If \a sumContributions is \c true, it sums up the contributions to TauN from
 * all legs, otherwise it returns the largest contributions from a single leg.
 *
 * \note This function is only meant for parton-level events, it must not be
 * used after beam remnants or hadronization.
 */
double calculateTauNPythia(const Pythia8::Event& event, bool sumContributions)
{
   double TauNPY = 0.0;
   for (int i = 0; i < event.size(); ++i) {
      if (event[i].statusAbs() == 23 && (event[i].isGluon() || event[i].isQuark())) {
         FourVector clustered;
         for (int j = 0; j < event.size(); j++) {
            if (event[j].isFinal() && (event[j].isGluon() || event[j].isQuark())
                && event[j].isAncestor(i))
               clustered += FourVector(event[j].e(), event[j].px(), event[j].py(), event[j].pz());
         }
         const double tmp = NJettinessMetricFSR::measure(clustered);
         if (sumContributions)
            TauNPY += tmp;
         else if (tmp > TauNPY)
            TauNPY = tmp;
      }
   }
   return TauNPY;
}

//------------------------------------------------------------------------------
/**
 * \brief Calculates TauNPythia for hadronic collisions
 *
 * Clusters the event back according to Pythia's splitting history and computes
 * TauN from the resulting massive N-body phase-space point.
 * If \a sumContributions is \c true, it sums up the contributions to TauN from
 * all legs, otherwise it returns the largest contributions from a single leg.
 *
 * \note This function is only meant for parton-level events, it must not be
 * used after beam remnants or hadronization.
 */
double calculateTauNppPythia(const Pythia8::Event& event, double y,
                             bool sumContributions, bool useHardDirections)
{
   BoostZHyp boost(exp(-y));
   double TauNPY = 0.0;
   for (int i = 0; i < event.size(); ++i) {
      if (event[i].statusAbs() == 23 && (event[i].isGluon() || event[i].isQuark())) {
         // hard final-state particle
         FourVector clustered;
         for (int j = 0; j < event.size(); j++) {
            if (event[j].isFinal() && (event[j].isGluon() || event[j].isQuark())
                && event[j].isAncestor(i))
               clustered += FourVector(event[j].e(), event[j].px(), event[j].py(), event[j].pz());
         }
         double tmp;
         if (useHardDirections) {
            FourVector phard = boost(FourVector(event[i].e(), event[i].px(), event[i].py(), event[i].pz()));
            clustered.apply(boost);
            tmp = clustered.p0() - clustered.pvec() * phard.pvec() / phard.pvec().magnitude();
         } else
            tmp = NJettinessMetric::finalMeasure(boost(clustered));
         if (sumContributions)
            TauNPY += tmp;
         else if (tmp > TauNPY)
            TauNPY = tmp;
      } else if (event[i].statusAbs() == 21 && (event[i].isGluon() || event[i].isQuark())) {
         // hard initial-state particle
         FourVector clustered;
         for (int j = 0; j < event.size(); ++j) {
            if (event[j].isFinal() && (event[j].isGluon() || event[j].isQuark())
                && event[j].isAncestor(getFirstInitialMother(event, i)))
               clustered += FourVector(event[j].e(), event[j].px(), event[j].py(), event[j].pz());
         }
         double tmp;
         if (event[i].pz() > 0.)  // initialA
            tmp = NJettinessMetric::initialMeasureA(boost(clustered));
         else                     // initialB
            tmp = NJettinessMetric::initialMeasureB(boost(clustered));
         if (sumContributions)
            TauNPY += tmp;
         else if (tmp > TauNPY)
            TauNPY = tmp;
      }
   }

   return TauNPY;
}

} // namespace Geneva
