//------------------------------------------------------------------------------
/// \file UHRppB012.cpp
//
// Author(s):
//    Christian Bauer, Chris Vermilion, Calvin Berggren, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class UHRppB012
//------------------------------------------------------------------------------

#include "Geneva/PythiaInterface/UHRppB012.hpp"
#include "Geneva/PythiaInterface/ShowerScaleSCETppV012.hpp"
#include "Geneva/PythiaInterface/PythiaUtils.hpp"

#include "Geneva/Core/Physics/NJettiness.hpp"
#include "Geneva/Core/Physics/Transformation.hpp"
#include "Geneva/Core/Physics/IRestriction.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/Utils/Random.hpp"

#include <cassert>

namespace Geneva
{

//------------------------------------------------------------------------------
UHRppB012::UHRppB012(IRestriction* restriction, size_t maxAttempts, size_t maxFinal,
                     double tau0cut, double tau1cut, double lambda, bool removeMPI, double pTmaxMPI,
                     size_t maxISRstep, size_t maxFSRstep, bool useTauClust, bool sumTauClust,
                     bool useHardDirections, double tau1MaxRatio)
   : UserHookRestriction(maxAttempts), _restriction(restriction), _maxFinal(maxFinal),
     _tau0cut(tau0cut), _tau1cut(tau1cut), _lambda(lambda), _removeMPI(removeMPI), _pTmaxMPI(pTmaxMPI),
     _maxISRstep(maxISRstep), _maxFSRstep(maxFSRstep), _useTauClust(useTauClust), _sumTauClust(sumTauClust),
     _useHardDirections(useHardDirections), _tau1MaxRatio(tau1MaxRatio),
     _nveto(9, 0.)
{
   if (maxFinal > 2)
      Log::fatal("UHRppB012") << "Maximum number of final-state partons " << maxFinal << " is above allowed maximum of 2." << Log::end;
}

//------------------------------------------------------------------------------
std::ostream& UHRppB012::printStatistics(std::ostream& out)
{
   out << "Total number of shower vetoes: " << numVetoes()
       << "\ncalc vetoes (0-parton events): " << _nveto[0]
       << "\ncalc vetoes (1-parton events): " << _nveto[1]
       << "\ncalc vetoes (2-parton events): " << _nveto[2]
       << "\nfull vetoes (0-parton events): " << _nveto[3]
       << "\nfull vetoes (1-parton events): " << _nveto[4]
       << "\nfull vetoes (2-parton events): " << _nveto[5]
       << "\nstep vetoes (0-parton events): " << _nveto[6]
       << "\nstep vetoes (1-parton events): " << _nveto[7]
       << "\nstep vetoes (2-parton events): " << _nveto[8]
       << '\n';
   return out;
}

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/// Sets all the observables that only depend on the hard event
void UHRppB012::setHardObservables()
{
   ZeroJettinesspp Tau0(hardProcess());
   _numHardFinal = Tau0.numQCD();
   if (_numHardFinal > _maxFinal)
      Log::error("UHRppB012::setHardObservables") << "Number of hard final partons " << _numHardFinal
         << " is above allowed maximum of " << _maxFinal << Log::end;

   // save Tau0 and Y for later use
   _tau0 = Tau0();
   _yBoost = Tau0.y();

   // get the Tau shower starting scale
   ShowerScaleSCETppV012 scale(_maxFinal, _tau0cut, _tau1cut, _tau1MaxRatio, true);
   _tauShowerScale = scale.getTauScale(hardProcess());
}

//------------------------------------------------------------------------------
bool UHRppB012::vetoPartonLevelEarly(const Pythia8::Event& event)
{
   Configuration config;
   partonLevelToConfig(event, config, _removeMPI, _pTmaxMPI);
   NJettinessppRecursive TauFR(config);

   // fixed-order restriction for exclusive bins if Pythia event has <= _maxFinal partons
   if (_numHardFinal < _maxFinal && TauFR.numQCD() <= _maxFinal) {
      if (_restriction->countJets(config) > _numHardFinal) {
         _nveto[_numHardFinal]++;
         return true;
      } else {
         return false;
      }
   }

   // Restrict TauPY to be below the Tau shower scale
   double TauPY;
   if (_useTauClust && _numHardFinal == _maxFinal) {
      TauPY = calculateTauNppPythia(event, _yBoost, _sumTauClust, _useHardDirections);
   } else {
      if (_numHardFinal == 0 || (_numHardFinal == 1 && _tau0 < _tau0cut))
         TauPY = ZeroJettinesspp(config)();
      else
         TauPY = TauFR(_numHardFinal);
   }

   if ((_numHardFinal == _maxFinal && TauPY > _tauShowerScale)
       || (_numHardFinal < _maxFinal && fractionalRejection(TauPY, _tauShowerScale))) {
      _nveto[_numHardFinal + 3]++;
      return true;
   }
   return false;
}

//------------------------------------------------------------------------------
bool UHRppB012::vetoStep(int iPos, int nISR, int nFSR, const Pythia8::Event& event)
{
   // If above the maximum allowed steps, don't do anything
   if (size_t_from_int(nISR) > _maxISRstep || size_t_from_int(nFSR) > _maxFSRstep)
      return false;

   // Determine a local value for TauPY
   double TauStepPY;
   if (iPos == 3) {
      int i = event.size() - 2;
      FourVector pij(event[i].e(), event[i].px(), event[i].py(), event[i].pz());
      --i;
      pij += FourVector(event[i].e(), event[i].px(), event[i].py(), event[i].pz());
      TauStepPY = NJettinessMetricFSR::measure(BoostZHyp(exp(-_yBoost))(pij));
   } else if (iPos == 2) {
      const int i = event.size() - 1;
      const FourVector pi(event[i].e(), event[i].px(), event[i].py(), event[i].pz());
      TauStepPY = NJettinessMetricISR::measure(BoostZHyp(exp(-_yBoost))(pi));
   } else {
      return false;
   }

   // Veto if TauPY > TauHard
   if (TauStepPY > _tauShowerScale) {
      _nveto[_numHardFinal + 6]++;
      return true;
   }

   return false;
}

//------------------------------------------------------------------------------
bool UHRppB012::fractionalRejection(double TauPY, double TauHard)
{
   if (TauPY <= TauHard)
      return false;

   if (TauPY  >= (1. + _lambda) * TauHard)
      return true;

   return TauPY > (1. + randomInterval() * _lambda) * TauHard;
}

} // namespace Geneva
