//------------------------------------------------------------------------------
/// \file UserHookRestriction.cpp
//
// Author(s):
//    Christian Bauer, Chris Vermilion, Calvin Berggren
//
// Copyright:
//    Copyright (C) 2014 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class UserHookRestriction
//------------------------------------------------------------------------------

#include "Geneva/PythiaInterface/UserHookRestriction.hpp"
#include "Geneva/PythiaInterface/PythiaUtils.hpp"

namespace Geneva
{

//------------------------------------------------------------------------------
/**
 * This function sets the hardProcess variable that is used in all subsequent
 * userHook checks. By setting the hardprocess here, we only need to calculate
 * it once per event.
 *
 * \return Whether to veto the event
 */
bool UserHookRestriction::doVetoProcessLevel(Pythia8::Event& event)
{
   _numVetoesEvent = 0;
   processToConfig(event, _hardProcess);
   setHardObservables();
   return false;
}
//------------------------------------------------------------------------------
/**
 * This function is a wrapper around the real function defining the veto.
 * The purpose is to store the value of the event being passed in for later use
 * so that we will have access to the parton-level event even if hadronization
 * is turned on.
 *
 * \return Whether to veto the event
 */
bool UserHookRestriction::doVetoPartonLevelEarly(const Pythia8::Event& event)
{
   if (vetoPartonLevelEarly(event)) {
      _numVetoes++;
      _numVetoesEvent++;
      return _numVetoesEvent < _maxAttempts;
   }

   // store the parton-level event
   // the first time through, set this twice; possible bug in pythia where
   // particles are copied before the particle data ptr is ready
   if (!_init) {
      _partonLevel = event;
      _init = true;
   }
   _partonLevel = event;
   return false;
}

//------------------------------------------------------------------------------
/**
 * This function is a wrapper around the real function defining the veto.
 * \return Whether to veto the event
 */
bool UserHookRestriction::doVetoStep(int iPos, int nISR, int nFSR, const Pythia8::Event& event)
{
   if (vetoStep(iPos, nISR, nFSR, event)) {
      _numVetoes++;
      _numVetoesEvent++;
      return _numVetoesEvent < _maxAttempts;
   }
   return false;
}

} // namespace Geneva
