//------------------------------------------------------------------------------
/// \file GenevaWrapper.cpp
//
// Author(s):
//    Christian Bauer, Chris Vermilion, Simone Alioli, Calvin Berggren
//
// Copyright:
//    Copyright (C) 2012 LBNL
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Wrappers to Geneva main functions.
//------------------------------------------------------------------------------

#include "Geneva/Wrapper/GenevaWrapper.hpp"

#include "Geneva/Driver/Geneva.hpp"
#include "Geneva/Driver/LHEFInterface.hpp"

#include "Geneva/Analyzers/AnalyzerFactory.hpp"
#include "Geneva/Analyzers/IAnalyzer.hpp"
#include "Geneva/Analyzers/OutputStreamAnalyzer.hpp"

#include "Geneva/Core/Physics/Event.hpp"
#include "Geneva/Core/Physics/QCDUtils.hpp"
#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/OptionParserCL.hpp"
#include "Geneva/Core/UI/Exception.hpp"
#include "Geneva/Core/UI/GenerationReport.hpp"
#include "Geneva/Core/UI/Timing.hpp"
#include "Geneva/Core/Utils/CaptureOutput.hpp"
#include "Geneva/Core/Utils/Random.hpp"
#include "Geneva/Core/Utils/FPEUtils.hpp"
#include "Geneva/Core/Utils/CatchSignal.hpp"
#include "Geneva/Core/Utils/FileUtils.hpp"

#include "Geneva/Calculations/CalcSCETppV012.hpp"

#ifdef geneva_USE_PYTHIA8
#include "Geneva/PythiaInterface/PythiaDriver.hpp"
#ifdef geneva_USE_HEPMC
#include "HepMC/GenEvent.h"
#include "HepMC/IO_GenEvent.h"
#include "Pythia8Plugins/HepMC2.h"
#endif
#endif

#ifdef geneva_USE_RIVET
#include "Rivet/Rivet.hh"
#include "HepMC/GenEvent.h"
#include "HepMC/IO_GenEvent.h"
#endif

#ifdef geneva_USE_ZLIB
#include "Geneva/gzstream.hpp"
#endif

#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdlib>

namespace Geneva
{

TrapFPE trapFPE;

CatchSignal catcher;

//------------------------------------------------------------------------------
int GenevaWrapper::setupCrossSection(const std::string& cmdLine)
{
   try {

      Log::setVerbosity(Log::kInfo);

      CaptureOutput logger;

      //====================   Setup   ====================//
      ::Geneva::Geneva geneva(cmdLine);

      logger.direct_to(geneva.getOption<std::string>("logFile"), geneva.getOption<bool>("logFile::append"));

      geneva.initialize();

      const size_t numEvents = size_t_from_int(geneva.getOption<int>("num"));
      const bool show_timer = geneva.getOption<bool>("showTimer");
      const std::string filename = CalcSCETppV012::getSetupFilename(geneva.optionMap(), true);

      Log::warning().setMaxCounts(std::max(numEvents/10000, size_t(100000000)));

      // setup timer
      Timing timer;
      timer.setTargetValue(numEvents);
      timer.setCurrentValue(0);
      const double updateInterval = 1.0; // update every 1 second
      double timeElapsed = 1.0 - updateInterval; // first update at 1 second
      double timeElapsedSinceSave = 0.; // first save at 10 minutes

      //====================   Main Event Loop   ====================//
      timer.start();
      size_t num = 0;
      Rate sigmaGE1;
      try {
         for (; num < numEvents; ++num) {
            storeRandomState();

            // timing information to screen
            if (show_timer && num % 100 == 0 && timer.timeElapsedInSec() > timeElapsed + updateInterval) {
               timer.setCurrentValue(num);
               std::cout << timer.progressString() << std::flush;
               timeElapsed = timer.timeElapsedInSec();
            }

            const Event& evt = geneva.next();

            Sigma weights(evt.weightedSigma());
            if (std::count_if(evt.config().beginFinal(), evt.config().endFinal(), IsQCD()) > 0)
               sigmaGE1.addWeight(weights);
            else
               sigmaGE1.addWeight();

            if (num % 1000 == 0 && timer.timeElapsedInMin() > timeElapsedSinceSave + 1) {
               std::ofstream gridFile(filename.c_str());
               gridFile << sigmaGE1.totEntries() << '\n';
               gridFile << sigmaGE1.sumWeights().central();
               for (size_t i = 0; i < 5; ++i) {
                  gridFile << ' ';
                  if (i < sigmaGE1.sumWeights().variations().size())
                     gridFile << sigmaGE1.sumWeights().variations()[i];
                  else
                     gridFile << "0.0";
               }
               gridFile << '\n';
               gridFile << sigmaGE1.sumWeightsSq().central();
               for (size_t i = 0; i < 5; ++i) {
                  gridFile << ' ';
                  if (i < sigmaGE1.sumWeightsSq().variations().size())
                     gridFile << sigmaGE1.sumWeightsSq().variations()[i];
                  else
                     gridFile << "0.0";
               }
               gridFile.close();
               timeElapsedSinceSave = timer.timeElapsedInMin();
            }

         } // main event loop

      } catch (...) {
         Log::warning("setup-xsec") << "Exception caught during event " << num
            << " at runtime of " << timer.timeElapsedInSec() << " seconds." << Log::end;
         saveRandomState(filename + ".RandomStateDump");
         throw;
      }
      timer.stop();
      if (show_timer) {
         // one final progress string
         timer.setCurrentValue(num);
         std::cout << timer.progressString() << '\n'
                   << timer.timeElapsedString() << '\n'
                   << std::endl;
      }

      //====================   Write Results   ====================//
      geneva.finalize(num);

      std::ofstream gridFile(filename.c_str());
      gridFile << sigmaGE1.totEntries() << '\n';
      gridFile << sigmaGE1.sumWeights().central();
      for (size_t i = 0; i < 5; ++i) {
         gridFile << ' ';
         if (i < sigmaGE1.sumWeights().variations().size())
            gridFile << sigmaGE1.sumWeights().variations()[i];
         else
            gridFile << "0.0";
      }
      gridFile << '\n';
      gridFile << sigmaGE1.sumWeightsSq().central();
      for (size_t i = 0; i < 5; ++i) {
         gridFile << ' ';
         if (i < sigmaGE1.sumWeightsSq().variations().size())
            gridFile << sigmaGE1.sumWeightsSq().variations()[i];
         else
            gridFile << "0.0";
      }
      gridFile.close();
      Log::warning().writeCounts();
      Log::warning().resetCounts();
   } catch (...) {
      Log::warning().writeCounts();
      Log::warning().resetCounts();
      return mainExceptionHandler();
   }
   return EXIT_SUCCESS;
}

//------------------------------------------------------------------------------
int GenevaWrapper::generateEvents(const std::string& exeName, const std::string& cmdLine)
{
   try {

      Log::setVerbosity(Log::kInfo);

      CaptureOutput logger;

      //====================   Setup   ====================//
      ::Geneva::Geneva geneva(cmdLine);

      logger.direct_to(geneva.getOption<std::string>("logFile"), geneva.getOption<bool>("logFile::append"));

      geneva.initialize();

      // setup the report
      const std::string reportName = GenerationReport::getFilename(geneva.optionMap(), "lhef");
      bool needReport = geneva.getOption<bool>("writeReport");
      if (needReport) {
         if (std::ifstream(reportName.c_str())) {
            if (!geneva.getOption<bool>("overwriteOutput"))
               Log::fatal("main") << reportName << " already exists.\nPlease move the file or allow overwriting it by setting the 'overwrite_output' option." << Log::end;
            else
               Log::warning("main") << "Will overwrite existing file: " << reportName << Log::end;
         }
      }
      GenerationReport report(geneva.optionMap(), exeName);
      report.setTitle(GenerationReport::getDefaultTitle(geneva.optionMap(), "lhef"));

      // setup the .lhe output
      LHEFInterface* lhefInterface = 0;
      const bool writeEvents = geneva.getOption<bool>("writeEvents");
      if (writeEvents) {
         const std::string eventsFile = LHEFInterface::getFilename(geneva.optionMap());
         if (std::ifstream(eventsFile.c_str())) {
            if (!geneva.getOption<bool>("overwriteOutput"))
               Log::fatal("main") << eventsFile << " already exists.\nPlease move the file or allow overwriting it by setting the 'overwrite_output' option." << Log::end;
            else
               Log::warning("main") << "Will overwrite existing file " << eventsFile << Log::end;
         }
         lhefInterface = &LHEFInterface::instance(geneva.optionMap(), eventsFile);
      }

      // setup the analyzer
      IAnalyzer* analyzer = 0;
      if (geneva.getOption<bool>("Analyzer::Unshowered"))
         analyzer = AnalyzerFactory::instance().getAnalyzer(geneva.optionMap());
      const bool doAnalyze = (analyzer != 0);

      const bool show_timer = geneva.getOption<bool>("showTimer");
      const size_t numEvents = size_t_from_int(geneva.getOption<int>("num"));

      Log::warning().setMaxCounts(std::max(numEvents/10000, size_t(100)));

      // setup timer
      Timing timer;
      timer.setTargetValue(numEvents);
      timer.setCurrentValue(0);
      const double updateInterval = 1.0; // update every 1 second
      double timeElapsed = 1.0 - updateInterval; // first update at 1 second
      const double saveInterval = 10.0; // save analyzer and report every 10 minutes
      double timeElapsedSinceSave = 0.; // first save at 10 minutes
      const double max_time = geneva.getOption<double>("maxTime"); // maximum time to run in minutes

      //Allow for the possibility to catching spikes
      bool checkForSpikes = geneva.getOption<bool>("Debug::checkForSpikes");
      double spikeTrigger = std::numeric_limits<double>::quiet_NaN();
      if (checkForSpikes)
         spikeTrigger = geneva.getOption<double>("Debug::spikeTrigger");

      //====================   Main Event Loop   ====================//
      timer.start();
      size_t num = 0;
      try {
         for (; num < numEvents; ++num) {
            storeRandomState();
            // stop if we've been running too long
            if (timer.timeElapsedInMin() > max_time) {
               Log::warning("main: main event loop") << "Exceeded requested run time of " << max_time << " minutes.  Finished " << num << " events." << Log::end;
               break;
            }
            const Event& evt = geneva.next();
            if (doAnalyze)
               analyzer->analyzePartonic(evt);

            if (writeEvents)
               lhefInterface->writeEvent(evt);

            // timing information to screen
            if (show_timer && num % 100 == 0 && timer.timeElapsedInSec() > timeElapsed + updateInterval) {
               timer.setCurrentValue(num);
               std::cout << timer.progressString() << std::flush;
               timeElapsed = timer.timeElapsedInSec();
            }
            // save output periodically
            if (num % 1000 == 0 && timer.timeElapsedInMin() > timeElapsedSinceSave + saveInterval) {
               if (needReport)
                  writeReport(report, reportName, timer.timeElapsedInSec(), num, analyzer);
               timeElapsedSinceSave = timer.timeElapsedInMin();
            }

            if (evt.weightedSigma().isDefined()) {
               //Save the RandomState if we are triggering on Spikes and the weight is above the spikeTrigger
               if (checkForSpikes and (abs(evt.weightedSigma().central())  > spikeTrigger)) {
                  Log::info() << "Event #" << num << " triggers spike. Saving RandomState" << Log::end;
                  Log::info() << std::setprecision(16)<< evt.weightedSigma().central() << Log::end;
                  Log::info() << evt.config() << Log::end;
                  saveRandomState(reportName + ".RandomStateDump");
               }
            }
         } // main event loop

      } catch (...) {
         Log::warning("main") << "Exception caught during event " << num
            << " at runtime of " << timer.timeElapsedInSec() << " seconds. Attempting to write results ..." << Log::end;
            geneva.finalize(num);
         if (writeEvents)
            lhefInterface->setTotalRate(geneva.totalRate());
         if (needReport)
            writeReport(report, reportName, timer.timeElapsedInSec(), num, analyzer);
         saveRandomState(reportName + ".RandomStateDump");
         throw;
      }
      //====================   Write Results   ====================//
      timer.stop();
      if (show_timer) {
         // one final progress string
         timer.setCurrentValue(num);
         std::cout << timer.progressString() << '\n'
                   << timer.timeElapsedString() << '\n'
                   << std::endl;
      }

      geneva.finalize(num);

      if (writeEvents)
         lhefInterface->setTotalRate(geneva.totalRate());

      // write out basic rates
      if (doAnalyze)
         Log::info() << *analyzer << Log::end;

      if (needReport)
         writeReport(report, reportName, timer.timeElapsedInSec(), num, analyzer);

      // save last random state if explicitly asked to do so
      if (geneva.getOption<bool>("Random::saveLast")) {
         storeRandomState();
         saveRandomState(reportName + ".RandomStateDump");
      }
      Log::warning().writeCounts();
      Log::warning().resetCounts();
   } catch (...) {
      Log::warning().writeCounts();
      Log::warning().resetCounts();
      return mainExceptionHandler();
   }
   return EXIT_SUCCESS;
}

#ifdef geneva_USE_PYTHIA8

//------------------------------------------------------------------------------
void analyzeLhefUsage()
{
   Log::info() << "usage: geneva-lhef-analyze --infile <file>.lhe[.gz] --option-file <filename>\n";
   Log::info() << "                           [--help] [optional arguments...]\n";
   Log::info() << "\nAnalyzes Geneva LHEF events.\n";
   Log::info() << "\nrequired arguments:\n";
   Log::info() << "  --infile <file>.lhe[.gz]        (zipped) LHEF file to analyze\n";
   Log::info() << "  --option-file <filename>        Geneva option file\n";
   Log::info() << "\noptional arguments: \n";
   Log::info() << "  --help                          show this help message and exit\n";
   Log::info() << "  --outpath <path>                output path for the report files [analyze]\n";
   Log::info() << "  --max-time <minutes>            maximum runtime in minutes [0 = unused]\n";
   Log::info() << "  --max-events <num>              maximum number of events to analyze [0 = unused]\n";
   Log::info() << "  --enforce-lhef-options [on|off] enforce consistency with options in LHEF header [on]\n";
   Log::info() << "  --show-timer [on|off]           show timer progress bar [off]\n";
   Log::info() << "  --logfile <filename>            log file where to append the screen output [terminal]\n";
}

//------------------------------------------------------------------------------
int GenevaWrapper::analyzeLhef(const std::string& exeName, const std::string& options)
{
   try {
      if (options.empty()) {
         analyzeLhefUsage();
         return EXIT_SUCCESS;
      }

      Log::setVerbosity(Log::kInfo);

      OptionParserCL parserCL(options);
      if (parserCL.hasOption("help")) {
         analyzeLhefUsage();
         return EXIT_SUCCESS;
      };
      // parse the command-line options
      if (!parserCL.hasOption("infile"))
         Log::fatal("lhef-analyze")
            << "Please use --infile <filename> to specify an LHEF file to analyze. Use --help to see a complete list of arguments." << Log::end;
      if (!parserCL.hasOption("option-file"))
         Log::fatal("lhef-analyze")
            << "Please use --option-file <filename> to specify a Geneva option file. Use --help to see a complete list of arguments." << Log::end;

      std::string infile;
      std::string option_file;
      std::string outpath = "analyze";
      size_t max_time = 0;
      size_t max_events = 0;
      bool enforce_lhef_options = true;
      bool show_timer = false;
      std::string logfile;
      while (parserCL.hasNext()) {
         std::string name;
         std::string value;
         parserCL.getNext(name, value);
         if (name == "infile") {
            infile = value;
         } else if (name == "option-file") {
            option_file = value;
         } else if (name == "outpath") {
            outpath = value;
         } else if (name == "max-time") {
            max_time = std::strtoul(value.c_str(), 0, 10);
         } else if (name == "max-events") {
            max_events = std::strtoul(value.c_str(), 0, 10);
         } else if (name == "enforce-lhef-options") {
            enforce_lhef_options = Option<bool>(name, value).value();
         } else if (name == "show-timer") {
            show_timer = Option<bool>(name, value).value();
         } else if (name == "logfile") {
            logfile = value;
         } else {
            Log::fatal("lhef-analyze") << "Invalid argument \"" << name << "\". Use --help to see a complete list of arguments." << Log::end;
         }
      }

      CaptureOutput logger;
      logger.direct_to(logfile);

      // initialize the PythiaDriver
      PythiaDriver pythiaDriver(infile, option_file, enforce_lhef_options, "", max_events, 0.001);
      Pythia8::Pythia& pythia = pythiaDriver.getPythia();
      OptionMap optMap = pythiaDriver.getOptions();

      // override user-provided analyzer options
      if (!optMap.getOption<bool>("Analyzer::Unshowered")) {
         Log::warning("lhef-analyze") << "Discarded user-provided option Analyzer::Unshowered off" << Log::end;
         optMap.setOption("Analyzer::Unshowered", "on");
      }
      if (optMap.getOption<bool>("Analyzer::Showered")) {
         Log::warning("lhef-analyze") << "Discarded user-provided option Analyzer::Showered on" << Log::end;
         optMap.setOption("Analyzer::Showered", "off");
      }
      if (optMap.getOption<bool>("Analyzer::ShoweredNoMPI")) {
         Log::warning("lhef-analyze") << "Discarded user-provided option Analyzer::ShoweredNoMPI on" << Log::end;
         optMap.setOption("Analyzer::ShoweredNoMPI", "off");
      }
      if (optMap.getOption<bool>("Analyzer::Hadronized")) {
         Log::warning("lhef-analyze") << "Discarded user-provided option Analyzer::Hadronized on" << Log::end;
         optMap.setOption("Analyzer::Hadronized", "off");
      }
      IAnalyzer* analyzer = AnalyzerFactory::instance().getAnalyzer(optMap);
      if (!analyzer)
         Log::fatal("lhef-analyze") << "No analyzer specified." << Log::end;

      // setup the report
      std::string reportName = GenerationReport::getFilename(optMap, "lhef", outpath);
      if (std::ifstream(reportName.c_str())) {
         if (!optMap.getOption<bool>("overwriteOutput"))
            Log::fatal("lhef-analyze") << reportName << " already exists.\nPlease move the file or allow overwriting it by setting the 'overwrite_output' option." << Log::end;
         else
            Log::warning("lhef-analyze") << "Will overwrite existing file " << reportName << Log::end;
      }
      GenerationReport report(optMap, exeName + ' ' + options);
      report.setTitle(GenerationReport::getDefaultTitle(optMap, "lhef"));

      // setup timer
      size_t numEvents = pythiaDriver.numEvents();
      Timing timer;
      timer.setTargetValue(numEvents);
      timer.setCurrentValue(0);
      const double updateInterval = 1.0; // update every 1 second
      double timeElapsed = 1.0 - updateInterval; // first update at 1 second
      const double saveInterval = 10.0; // save analyzer and report every 10 minutes
      double timeElapsedSinceSave = 0.; // first save at 10 minutes

      Log::warning().setMaxCounts(static_cast<size_t>(ceil(static_cast<double>(numEvents) / 100.)));

      // reset Pythia options that allow to just load the LHEF event
      pythia.settings.flag("PartonLevel:all", false);
      pythia.settings.flag("HadronLevel:all", false);
      pythia.settings.flag("PartonLevel:Remnants", false);
      pythia.settings.flag("PartonLevel:FSR", false);
      pythia.settings.flag("PartonLevel:ISR", false);
      pythia.settings.flag("PartonLevel:MPI", false);
      pythia.settings.flag("Check:event", false);
      pythia.settings.flag("TimeShower:QEDshowerByQ", false);
      pythia.settings.flag("TimeShower:QEDshowerByL", false);
      pythia.settings.flag("TimeShower:QEDshowerByGamma", false);
      pythia.settings.flag("SpaceShower:QEDshowerByQ", false);
      pythia.settings.flag("SpaceShower:QEDshowerByL", false);
      pythia.settings.mode("LesHouches:setLeptonMass", false);
      pythia.settings.mode("LesHouches:setQuarkMass", false);

      pythia.init();

      //============   Main Event Loop   ============//
      size_t num = 0; // event counter
      timer.start();
      try {
         while (pythiaDriver.hasNext()) {
            // stop if we've been running too long
            if (max_time && timer.timeElapsedInMin() >= static_cast<double>(max_time)) {
               Log::warning("lhef-analyze") << "Exceeded requested run time of " << max_time << " minutes. Finished " << num << " events." << Log::end;
               break;
            }

            if (pythiaDriver.next()) {
               ++num;

               // analyze the result
               analyzer->analyzePartonic(pythiaDriver.getPartonicEvent());

               // timing information to screen
               if (show_timer && num % 100 == 0 && timer.timeElapsedInSec() > timeElapsed + updateInterval) {
                  timer.setCurrentValue(num);
                  std::cout << timer.progressString() << std::flush;
                  timeElapsed = timer.timeElapsedInSec();
               }
               // save output periodically
               if (num % 1000 == 0 && timer.timeElapsedInMin() > timeElapsedSinceSave + saveInterval) {
                  writeReport(report, reportName, timer.timeElapsedInSec(), num, analyzer);
                  timeElapsedSinceSave = timer.timeElapsedInMin();
               }
            }
         }
      } catch (...) {
         Log::warning("lhef-analyze") << "Exception caught during event " << num
            << " at runtime of " << timer.timeElapsedInSec() << " seconds. Attempting to write results ..." << Log::end;
         writeReport(report, reportName, timer.timeElapsedInSec(), num, analyzer);
         saveRandomState(reportName + ".RandomStateDump");
         throw;
      }

      if (num < numEvents)
         Log::warning("lhef-analyze") << "Only " << num << " of " << numEvents << " events in " << infile << " could be analyzed." << Log::end;

      //====================   Write Results   ====================//

      timer.stop();
      if (show_timer) {
         // one final progress string
         timer.setCurrentValue(num);
         std::cout << timer.progressString() << '\n'
                   << timer.timeElapsedString() << '\n'
                   << std::endl;
      }

      pythiaDriver.stat();

      // write out basic rates
      Log::info() << *analyzer << Log::end;

      writeReport(report, reportName, timer.timeElapsedInSec(), num, analyzer);

      Log::warning().writeCounts();
      Log::warning().resetCounts();
   } catch (...) {
      Log::warning().writeCounts();
      Log::warning().resetCounts();
      return mainExceptionHandler();
   }
   return EXIT_SUCCESS;
}

//------------------------------------------------------------------------------
// Wrapper to use Geneva random numbers in Pythia8
class GenevaRandom : public Pythia8::RndmEngine
{
   public:
      // Routine for generating a random number.
      double flat() { return randomInterval(); }
};

#ifdef geneva_USE_HEPMC
//------------------------------------------------------------------------------
class HepMCWriter
{
   private:
#ifdef geneva_USE_ZLIB
      typedef gzstream::ogzstream _ostream_type;
#else
      typedef std::ofstream _ostream_type;
#endif

      HepMC::Pythia8ToHepMC _toHepMC;  ///< Pythia8 to HepMC Converter

      std::string _filename;

      bool _multiStream;

      std::vector<_ostream_type*> _outfile;
      std::vector<HepMC::IO_GenEvent*> _output;

      HepMC::GenEvent _hepmcEvent;
      HepMC::GenCrossSection _fixedxsec;

   public:
      HepMCWriter(const std::string& filename, bool multiStream = false)
         : _filename(filename), _multiStream(multiStream)
      {
         if (_filename == "cout") {
            if (_multiStream)
               Log::fatal("HepMCWriter") << "Cannot use multiple streams with std::cout." << Log::end;
            // send the central scale only to std::cout
            _output.push_back(new HepMC::IO_GenEvent(std::cout));
         } else {
            _outfile.push_back(new _ostream_type(_filename.c_str()));
            if (!_outfile.back()->good()) {
               Log::fatal("HepMCWriter") << "Could not open file: " << _filename << Log::end;
            }
            _output.push_back(new HepMC::IO_GenEvent(*(_outfile.back())));
         }
         _toHepMC.set_store_pdf(false);
         _toHepMC.set_store_proc(false);
      };

      ~HepMCWriter()
      {
         for (size_t i = 0; i < _output.size(); ++i)
            delete _output[i];

         for (size_t i = 0; i < _outfile.size(); ++i) {
            _outfile[i]->close();
            delete _outfile[i];
         }
      }

      void setFixedCrossSection(double sigma, double error)
      {
         _fixedxsec.set_cross_section(sigma, error);
      }

      //Get a HepMC::GenEvent* from the last showered event
      void writeEvent(Pythia8::Pythia& pythia)
      {
         _hepmcEvent.clear();

         // fills the HepMC event from Pythia, including PDF info, accumulated xsec, etc.
         _toHepMC.fill_next_event(pythia, &_hepmcEvent);

         // overwrite the accumulated xsec from Pythia
         if (_fixedxsec.is_set())
            _hepmcEvent.set_cross_section(_fixedxsec);

         // Pythia8 converter only fills the central weight, so we have to overwrite the weights
         _hepmcEvent.weights().clear();
         if (pythia.info.weights_compressed && !pythia.info.weights_compressed->empty()) {
            std::vector<double>& weights = *pythia.info.weights_compressed;
            if (_multiStream) {
               initMultiOutput(weights.size());
               size_t i = 0;
               for (std::vector<double>::iterator it = weights.begin(); it != weights.end(); ++it, ++i) {
                  _hepmcEvent.weights()["000"] = *it;
                  HepMC::GenEvent* ptr = &_hepmcEvent;
                  *_output[i] << ptr;
               }
            } else {
               if (weights.size() > 1000)
                  Log::fatal("HepMCWriter::writeEvent") << "Number of weights (" << weights.size() << ") is larger than maximum allowed (1000)." << Log::end;
               size_t i = 0;
               for (std::vector<double>::iterator it = weights.begin(); it != weights.end(); ++it, ++i) {
                  std::stringstream ss;
                  ss << std::setw(3) << std::setfill('0') << i;
                  _hepmcEvent.weights()[ss.str()] = *it;
               }
               HepMC::GenEvent* ptr = &_hepmcEvent;
               *_output[0] << ptr;
            }
         } else {
            _hepmcEvent.weights()["000"] = pythia.info.weight();
            HepMC::GenEvent* ptr = &_hepmcEvent;
            *_output[0] << ptr;
         }
      };

      /**
       * \brief Provides the default filename for the HepMC file.
       * \param options The options to use.
       * \param path The filepath to use.
       * \param showerTag The shower tag to use.
       * \param addSeed Whether to append the seed as tag (defaults to `true`).
       * \return The full filename including the path and extension.
       *
       * The full filename is constructed as
       * 'path/fileStem_generateTag_showerTag_seed.hepmc[.gz]',
       * where the seed is only added if \a addSeed is `true`, and the .gz
       * extension applies when ZLIB is enabled.
       */
      static std::string getFilename(const OptionMap& options, const std::string& path,
                                     const std::string& showerTag, bool addSeed = true)
      {
         std::string name = format_path(path)
                            + options.getOption("fileStem")
                            + format_tag(options.getOption("generateTag"))
                            + format_tag(showerTag);
         if (addSeed)
            name += format_tag(options.getOption("seed"));
         name += ".hepmc";
#ifdef geneva_USE_ZLIB
         name += ".gz";
#endif
         return name;
      };

   private:
      void initMultiOutput(size_t numWeights)
      {
         if (_output.size() < numWeights) {
            std::string stem = _filename;
            std::string ext;
            // split of the trailing extension
            size_t pos = stem.rfind('.');
            if (pos != std::string::npos) {
               ext = stem.substr(pos);
               stem.erase(pos);
               // if the extension was .gz, split of the next one as well
               if (ext == ".gz") {
                  pos = stem.rfind('.');
                  if (pos != std::string::npos) {
                     ext = stem.substr(pos) + ".gz";
                     stem.erase(pos);
                  }
               }
            }
            size_t i = _output.size();
            _output.resize(numWeights, 0);
            _outfile.resize(numWeights, 0);
            for (; i < numWeights; ++i) {
               std::stringstream s;
               s << i;
               std::string filename = stem + format_tag("vary" + s.str()) + ext;
               _outfile[i] = new _ostream_type(filename.c_str());
               if (! _outfile[i]->good()) {
                  Log::fatal("HepMCWriter") << "Could not open file: " << filename << Log::end;
               }
               _output[i] = new HepMC::IO_GenEvent(*(_outfile[i]));
            }
         }
      }
};
#endif // geneva_USE_HEPMC

//------------------------------------------------------------------------------
void showerPythia8Usage()
{
   Log::info() << "usage: geneva-shower-pythia8 --infile <file>.lhe[.gz] --option-file <filename> --seed <seed>\n";
   Log::info() << "                             [--help] [optional arguments...]\n";
   Log::info() << "\nShowers Geneva events using Pythia8.\n";
   Log::info() << "\nrequired arguments:\n";
   Log::info() << "  --infile <file>.lhe[.gz]        (zipped) LHEF file to shower\n";
   Log::info() << "  --option-file <filename>        Geneva option file\n";
   Log::info() << "  --seed <seed>                   random seed\n";
   Log::info() << "\noptional arguments:\n";
   Log::info() << "  --help                          show this help message and exit\n";
   Log::info() << "  --pythia-options <filename>     Pythia8 option file\n";
   Log::info() << "  --outpath <path>                output path for the hepmc[.gz] and report files [shower]\n";
   Log::info() << "  --shower-tag <tag>              extra tag to append to output files [""]\n";
   Log::info() << "  --max-time <minutes>            maximum runtime in minutes [0 = unused]\n";
   Log::info() << "  --max-events <num>              maximum number of events to shower [0 = unused]\n";
   Log::info() << "  --enforce-lhef-options [on|off] enforce consistency with options in LHEF header [on]\n";
   Log::info() << "  --show_timer [on|off]            show timer progress bar [on]\n";
   Log::info() << "  --logfile <filename>            log file where to append the screen output [terminal]\n";
#ifdef geneva_USE_HEPMC
   Log::info() << "  --hepmc [on|off]                write out hepmc [on]\n";
   Log::info() << "  --hepmc-multi-stream [on|off]   write a separate hepmc file for each weight [off]\n";
   Log::info() << "  --sigma <xsec>                  overwrite accumulated cross section in hepmc output by fixed value <xsec> [unused]\n";
   Log::info() << "  --dsigma <error>                overwrite accumulated cross section error in hepmc file by fixed value <error> [unused]\n";
#endif
}

//------------------------------------------------------------------------------
int GenevaWrapper::showerPythia8(const std::string& exeName, const std::string& options)
{
   try {
      if (options.empty()) {
         showerPythia8Usage();
         return EXIT_SUCCESS;
      }

      Log::setVerbosity(Log::kInfo);

      OptionParserCL parserCL(options);
      if (parserCL.hasOption("help")) {
         showerPythia8Usage();
         return EXIT_SUCCESS;
      };
      // parse the command-line options
      if (!parserCL.hasOption("infile"))
         Log::fatal("shower-pythia8")
            << "Please use --infile <filename> to specify an LHEF file to shower. Use --help to see a complete list of arguments." << Log::end;
      if (!parserCL.hasOption("option-file"))
         Log::fatal("shower-pythia8")
            << "Please use --option-file <filename> to specify a Geneva option file. Use --help to see a complete list of arguments." << Log::end;
      if (!parserCL.hasOption("seed"))
         Log::fatal("shower-pythia8")
            << "Please use --seed <seed> to specify a random seed. Use --help to see a complete list of arguments." << Log::end;

      std::string infile;
      std::string option_file;
      int randomSeed = -1;
      std::string pythia_options;
      std::string outpath = "shower";
      std::string shower_tag;
      size_t max_time = 0;
      size_t max_events = 0;
      bool enforce_lhef_options = true;
      bool show_timer = false;
      std::string logfile;
#ifdef geneva_USE_HEPMC
      bool hepmc = true;
      bool hepmc_multi_stream = false;
      double sigma = std::numeric_limits<double>::quiet_NaN();
      double dsigma = std::numeric_limits<double>::quiet_NaN();
#else
      bool hepmc = false;
#endif
      while (parserCL.hasNext()) {
         std::string name;
         std::string value;
         parserCL.getNext(name, value);
         if (name == "infile") {
            infile = value;
         } else if (name == "option-file") {
            option_file = value;
         } else if (name == "seed") {
            randomSeed = std::atoi(value.c_str());
         } else if (name == "pythia-options") {
            pythia_options = value;
         } else if (name == "outpath") {
            outpath = value;
         } else if (name == "shower-tag") {
            shower_tag = value;
         } else if (name == "max-time") {
            max_time = std::strtoul(value.c_str(), 0, 10);
         } else if (name == "max-events") {
            max_events = std::strtoul(value.c_str(), 0, 10);
         } else if (name == "enforce-lhef-options") {
            enforce_lhef_options = Option<bool>(name, value).value();
         } else if (name == "show-timer") {
            show_timer = Option<bool>(name, value).value();
         } else if (name == "logfile") {
            logfile = value;
         } else if (name == "hepmc") {
            hepmc = Option<bool>(name, value).value();
#ifndef geneva_USE_HEPMC
            if (hepmc)
               Log::fatal("shower-pythia8") << "No HepMC support. Please compile with HepMC enabled to use HepMC output." << Log::end;
#else
         } else if (name == "hepmc-multi-stream") {
            hepmc_multi_stream = Option<bool>(name, value).value();
         } else if (name == "sigma") {
            sigma = std::atof(value.c_str());
         } else if (name == "dsigma") {
            dsigma = std::atof(value.c_str());
#endif
         } else {
            Log::fatal("shower-pythia8") << "Invalid argument \"" << name << "\". Use --help to see a complete list of arguments." << Log::end;
         }
      }

      CaptureOutput logger;
      logger.direct_to(logfile);

      // initialize the PythiaDriver
      PythiaDriver pythiaDriver(infile, option_file, enforce_lhef_options, pythia_options, max_events, 0.001);
      Pythia8::Pythia& pythia = pythiaDriver.getPythia();
      OptionMap optMap = pythiaDriver.getOptions();

      // setup the random sequence
      if (randomSeed > 0)
         setRandomSeed(randomSeed);
      else
         Log::fatal("shower-pythia8") << "Provide integer random seed" << Log::end;
      GenevaRandom genevaRandom;
      pythia.setRndmEnginePtr(&genevaRandom);

#ifdef geneva_USE_HEPMC
      // setup HepMC output
      HepMCWriter* hepmcWriter = 0;
      if (hepmc) {
         const std::string hepmcName = HepMCWriter::getFilename(optMap, outpath, shower_tag);
         if (std::ifstream(hepmcName.c_str())) {
            if (!optMap.getOption<bool>("overwriteOutput"))
               Log::fatal("shower-pythia8") << hepmcName << " already exists.\nPlease move the file or allow overwriting it by setting the 'overwrite_output' option." << Log::end;
            else
               Log::warning("shower-pythia8") << "Will overwrite existing file " << hepmcName << Log::end;
         }
         hepmcWriter = new HepMCWriter(hepmcName, hepmc_multi_stream);
         if (!isnan(sigma) && !isnan(dsigma))
            hepmcWriter->setFixedCrossSection(sigma, dsigma);
      }
#endif

      // setup Analyzer
      // This is done before the report, so the modified analyzer options get
      // stored in the report.
      const bool analyzePartonic = optMap.getOption<bool>("Analyzer::Unshowered");
      const bool analyzeShowered = optMap.getOption<bool>("Analyzer::Showered");
      const bool analyzeShowerNoMPI = pythia.settings.flag("PartonLevel:MPI") && optMap.getOption<bool>("Analyzer::ShoweredNoMPI");
      if (!analyzeShowerNoMPI)
         optMap.setOption("Analyzer::ShoweredNoMPI", "off");
      const bool analyzeHadronic = pythia.settings.flag("HadronLevel:all") && optMap.getOption<bool>("Analyzer::Hadronized");
      if (!analyzeHadronic)
         optMap.setOption("Analyzer::Hadronized", "off");
      IAnalyzer* analyzer = AnalyzerFactory::instance().getAnalyzer(optMap);
      const bool doAnalyze = (analyzer != 0);

      // setup the report
      // For the report and analyzer the added shower tag must be treated as an
      // addition to the generate tag, such that it is inserted in the
      // reportName and stored in the options in the report file.
      optMap.setOption("generateTag", optMap.getOption("generateTag") + format_tag(shower_tag));

      const std::string reportName = GenerationReport::getFilename(optMap, "", outpath);
      const bool needReport = optMap.getOption<bool>("writeReport");
      if (needReport) {
         if (std::ifstream(reportName.c_str())) {
            if (!optMap.getOption<bool>("overwriteOutput"))
               Log::fatal("shower-pythia8") << reportName << " already exists.\nPlease move the file or allow overwriting it by setting the 'overwrite_output' option." << Log::end;
            else
               Log::warning("shower-pythia8") << "Will overwrite existing file " << reportName << Log::end;
         }
      }
      GenerationReport report(optMap, exeName + ' ' + options, pythia_options);

      // setup timer
      const size_t numEvents = pythiaDriver.numEvents();
      Timing timer;
      timer.setTargetValue(numEvents);
      timer.setCurrentValue(0);
      const double updateInterval = 1.0; // update every 1 second
      double timeElapsed = 1.0 - updateInterval; // first update at 1 second
      const double saveInterval = 10.0; // save analyzer and report every 10 minutes
      double timeElapsedSinceSave = 0.; // first save at 10 minutes

      Log::warning().setMaxCounts(std::max(10ul, static_cast<size_t>(ceil(static_cast<double>(numEvents) / 100.))));

      //============   Main Event Loop   ============//
      size_t num = 0; // event counter
      timer.start();
      try {
         while (pythiaDriver.hasNext()) {
            // stop if we've been running too long
            if (max_time && timer.timeElapsedInMin() >= static_cast<double>(max_time)) {
               Log::info("shower-pythia8") << "Exceeded requested run time of " << max_time << " minutes. Finished " << num << " events." << Log::end;
               break;
            }

            if (pythiaDriver.next()) {
               ++num;

               // analyze the event
               if (doAnalyze) {
                  Event partonicEvent = pythiaDriver.getPartonicEvent();
                  if (analyzePartonic)
                     analyzer->analyzePartonic(partonicEvent);
                  if (analyzeShowered)
                     analyzer->analyzeShowered(pythiaDriver.getShoweredEvent(), partonicEvent.config());
                  if (analyzeShowerNoMPI)
                     analyzer->analyzeShoweredNoMPI(pythiaDriver.getShoweredNoMPIEvent(), partonicEvent.config());
                  if (analyzeHadronic)
                     analyzer->analyzeHadronized(pythiaDriver.getHadronizedEvent(), partonicEvent.config());
               }

#ifdef geneva_USE_HEPMC
               // write out hepmc
               if (hepmc)
                  hepmcWriter->writeEvent(pythia);
#endif

               // timing information to screen
               if (show_timer && num % 100 == 0 && timer.timeElapsedInSec() > timeElapsed + updateInterval) {
                  timer.setCurrentValue(num);
                  std::cout << timer.progressString() << std::flush;
                  timeElapsed = timer.timeElapsedInSec();
               }
               // save output periodically
               if (num % 1000 == 0 && timer.timeElapsedInMin() > timeElapsedSinceSave + saveInterval) {
                  if (needReport)
                     writeReport(report, reportName, timer.timeElapsedInSec(), num, analyzer);
                  timeElapsedSinceSave = timer.timeElapsedInMin();
               }
            }
         }
      } catch (...) {
         Log::warning("shower-pythia8") << "Exception caught during event " << num
            << " at runtime of " << timer.timeElapsedInSec() << " seconds. Attempting to write results ..." << Log::end;
#ifdef geneva_USE_HEPMC
         if (hepmc)
            delete hepmcWriter;
#endif
         if (needReport)
            writeReport(report, reportName, timer.timeElapsedInSec(), num, analyzer);
         saveRandomState(reportName + ".RandomStateDump");
         throw;
      }

      if (num < numEvents)
         Log::warning("shower-pythia8") << "Only " << num << " of " << numEvents << " events in " << infile << " could be showered." << Log::end;

      //====================   Write Results   ====================//

      timer.stop();
      if (show_timer) {
         // one final progress string
         timer.setCurrentValue(num);
         std::cout << timer.progressString() << '\n'
                   << timer.timeElapsedString() << '\n'
                   << std::endl;
      }

      pythiaDriver.stat();

#ifdef geneva_USE_HEPMC
      // close hepmc output
      if (hepmc)
         delete hepmcWriter;
#endif

      // write out basic rates
      if (doAnalyze)
         Log::info() << *analyzer << Log::end;

      if (needReport)
         writeReport(report, reportName, timer.timeElapsedInSec(), num, analyzer);

      Log::warning().writeCounts();
      Log::warning().resetCounts();
   } catch (...) {
      Log::warning().writeCounts();
      Log::warning().resetCounts();
      return mainExceptionHandler();
   }
   return EXIT_SUCCESS;
}

#endif // geneva_USE_PYTHIA8

#ifdef geneva_USE_RIVET
//------------------------------------------------------------------------------
class RivetInterface
{
   private:
#ifdef geneva_USE_ZLIB
      typedef gzstream::igzstream _istream_type;
#else
      typedef std::ifstream _istream_type;
#endif

      Rivet::AnalysisHandler _rivet;

      HepMC::GenEvent _hepmcEvent;

      _istream_type _infile;
      HepMC::IO_GenEvent _input;

      size_t _weight;

   public:
      RivetInterface(const std::string& infile, const std::string& runname, size_t weight = 1)
         : _rivet(runname), _infile(infile.c_str()), _input(_infile), _weight(weight)
      {
         if (!_infile) {
            Log::fatal("RivetInterface") << "Could not open file: " << infile << Log::end;
         }
         if (_input.rdstate() != 0) {
            Log::fatal("RivetInterface") << "Could not read file: " << infile << Log::end;
         }
         Log::info("RivetInterface") << "Rivet version: " << Rivet::version() << '\n'
            << "Using input file: " << infile << '\n'
            << "Using weight: " << _weight << Log::end;
      };

      void init(const std::set<std::string>& analyses, double xsec, const std::string& extra_analyses_path = "")
      {
         Rivet::addAnalysisLibPath(config::rivet_lib_dir);

         if (!extra_analyses_path.empty())
            Rivet::addAnalysisLibPath(extra_analyses_path);

         _rivet.addAnalyses(std::vector<std::string>(analyses.begin(), analyses.end()));

         // Read first event to define run conditions
         if (_input.rdstate() != 0 || !_input.fill_next_event(&_hepmcEvent))
            Log::fatal("RivetInterface") << "Could not read first event." << Log::end;
         if (_hepmcEvent.particles_size() == 0)
            Log::fatal("RivetInterface") << "Empty first event." << Log::end;

         _rivet.init(_hepmcEvent);

         if (!std::isnan(xsec)) {
            Log::info("RivetInterface") << "Setting user cross-section = " << xsec << " pb." << Log::end;
         }
         _rivet.setCrossSection(xsec);
      };

      void analyze()
      {
         // Set cross-section if found in event and not from command line
#ifdef HEPMC_HAS_CROSS_SECTION
         if (std::isnan(_rivet.crossSection()) && _hepmcEvent.cross_section()) {
            const double xs = _hepmcEvent.cross_section()->cross_section(); // in pb
            Log::debug("RivetInterface")
               << "Setting cross section = " << xs << " pb" << Log::end;
            _rivet.setCrossSection(xs);
         }
#endif
         // Complain about absence of cross-section if required!
         if (_rivet.needCrossSection() && !_rivet.hasCrossSection()) {
            Log::fatal("RivetInterface")
               << "Total cross section needed for at least one of the analyses. "
               << "Please set it via the --cross-section option." << Log::end;
         }

         // Replace the multiple weights with the requested one
         double weight = _hepmcEvent.weights()[_weight-1];
         _hepmcEvent.weights().clear();
         _hepmcEvent.weights().push_back(weight);
         _rivet.analyze(_hepmcEvent);
      };

      double crossSection()
      {
         return _rivet.crossSection();
      };

      bool readNext()
      {
         // Read the next event
         _hepmcEvent.clear();
         if ((_input.rdstate() != 0) || (!(_input.fill_next_event(&_hepmcEvent))) ) {
            Log::debug("RivetInterface") << "Could not read next event." << Log::end;
            return false;
         }
         return true;
      };

      void finalize(const std::string& outfile)
      {
         _rivet.finalize();
         Log::info("RivetInterface") << "Histograms written to " << outfile <<Log::end;
         _rivet.writeData(outfile);
      };

      /**
       * \brief Provides the default filename for the YODA file.
       * \param infilefull The input file, including its full path.
       * \param weight The weight to consider.
       * \param path The filepath to use.
       * \param rivetTag The rivet tag to use.
       * \return The full filename including the path and extension.
       *
       * The full filename is constructed as
       * 'path/fileStem_rivetTag_w<weight>_<seed>.yoda'
       */
      static std::string getFilename(const std::string& infilefull, size_t weight,
                                     const std::string& path, const std::string& rivetTag)
      {
         std::string infile;
         std::string::size_type pos = infilefull.rfind("/");
         if (pos == std::string::npos)
            infile = infilefull;
         else
            infile = infilefull.substr(pos);

         pos = infile.rfind(".");
         std::string stem;
         if (pos != std::string::npos) {
            std::string ext = infile.substr(pos);
            stem = infile.substr(0, pos);
            //If last extension was .gz, also remove the next-to-last one.
            if (ext == ".gz") {
               pos = stem.rfind(".");
               if (pos != std::string::npos)
                  stem = infile.substr(0, pos);
            }
         }
         pos = stem.rfind("_");
         if (pos == std::string::npos)
            pos = stem.size();
         std::string name = stem.substr(0,pos) + "_w"+ std::to_string(weight) + stem.substr(pos) + ".yoda";
         return path+name;
      };
};

//------------------------------------------------------------------------------
void analyzeRivetUsage()
{
   Log::info() << "usage: geneva-rivet-analyze --infile <file>.hepmc[.gz] --add-analysis <ana1> [--add-analysis <ana2> ...]\n";
   Log::info() << "                             [--help] [optional arguments...]\n";
   Log::info() << "\nAnalyzes an HepMC file with Rivet.\n";
   Log::info() << "\nrequired arguments:\n";
   Log::info() << "  --infile <file>.hepmc[.gz]      (zipped) input HepMC file with multiple weights\n";
   Log::info() << "  --add-analysis <ana>            add the analysis <ana> to the list of analyses to perform\n";
   Log::info() << "\noptional arguments:\n";
   Log::info() << "  --help                          show this help message and exit\n";
   Log::info() << "  --weight <weight>               weight number to use [1]\n";
   Log::info() << "  --outpath <path>                path for the Rivet output files [rivet]\n";
   Log::info() << "  --cross-section <xsec>          overwrite event cross section by fixed value <xsec> [unused]\n";
   Log::info() << "  --analyses-path <path>          extra path to append to analyses search path [unused]\n";
   Log::info() << "  --rivet-tag <tag>               extra tag to append to output files [""]\n";
   Log::info() << "  --max-time <minutes>            maximum runtime in minutes [0 = unused]\n";
   Log::info() << "  --max-events <num>              maximum number of events to analyze [0 = unused]\n";
   Log::info() << "  --logfile <filename>            log file where to append the screen output [terminal]\n";
}

//------------------------------------------------------------------------------
int GenevaWrapper::analyzeRivet(const std::string& exeName, const std::string& options)
{
   try {
      if (options.empty()) {
         analyzeRivetUsage();
         return EXIT_SUCCESS;
      }

      Log::setVerbosity(Log::kInfo);

      OptionParserCL parserCL(options);
      if (parserCL.hasOption("help")) {
         analyzeRivetUsage();
         return EXIT_SUCCESS;
      };
      // parse the command-line options
      if (!parserCL.hasOption("infile"))
         Log::fatal("rivet-analyze")
            << "Please use --infile <filename> to specify an HepMC file to analyse. Use --help to see a complete list of arguments." << Log::end;
      if (!parserCL.hasOption("add-analysis"))
         Log::fatal("rivet-analyze")
            << "Please use --add-analysis <ana> to specify at least one Rivet analysis. Use --help to see a complete list of arguments." << Log::end;

      std::string infile;
      std::set<std::string> analyses;
      size_t weight = 1;
      std::string outpath = "rivet";
      std::string rivet_tag;
      std::string logfile;
      std::string extra_analyses_path;
      size_t max_time = 0;
      size_t max_events = 0;
      double xsec = std::numeric_limits<double>::quiet_NaN();
      while (parserCL.hasNext()) {
         std::string name;
         std::string value;
         parserCL.getNext(name, value);
         if (name == "infile") {
            infile = value;
         } else if (name == "add-analysis") {
            analyses.insert(value);
         } else if (name == "weight") {
            weight = std::strtoul(value.c_str(), 0, 10);
         } else if (name == "cross-section") {
            xsec = std::atof(value.c_str());
         } else if (name == "outpath") {
            outpath = value;
         } else if (name == "rivet-tag") {
            rivet_tag = value;
         } else if (name == "max-time") {
            max_time = std::strtoul(value.c_str(), 0, 10);
         } else if (name == "max-events") {
            max_events = std::strtoul(value.c_str(), 0, 10);
         } else if (name == "logfile") {
            logfile = value;
         } else if (name == "analyses-path") {
            extra_analyses_path = value;
         } else {
            Log::fatal("rivet-analyze") << "Invalid argument \"" << name << "\". Use --help to see a complete list of arguments." << Log::end;
         }
      }

      CaptureOutput logger;
      logger.direct_to(logfile);

      // initialize Rivet
      RivetInterface rivet(infile, exeName, weight);
      rivet.init(analyses, xsec, extra_analyses_path);

      // setup timer
      Timing timer;
      timer.setCurrentValue(0);
      if (max_events != 0)
         timer.setTargetValue(max_events);
      else
         max_events = std::numeric_limits<int>::max();
      size_t num = 0;
      timer.start();
      try {
         while (true) {
            // stop if we've been running too long
            if (max_time && timer.timeElapsedInMin() >= static_cast<double>(max_time)) {
               Log::warning("rivet-analyze") << "Exceeded requested run time of " << max_time << " minutes. Finished " << num << " events." << Log::end;
               break;
            }
            // stop if we've been running too many events
            if (num >= max_events) {
               Log::warning("rivet-analyze") << "Exceeded requested number of " << max_events << " events." << Log::end;
               break;
            }
            // do analysis
            rivet.analyze();
            // print progress info
            ++num;
            if (num % 100 == 0)
               Log::info() << "Event " << num << " ("<< timer.timeElapsedInSec() << " seconds elapsed)" << Log::end;
            // stop on reading errror or when reaching end-of-file
            if (! rivet.readNext())
               break;
         }
      } catch (...) {
         Log::warning("rivet-analyze") << "Exception caught during event " << num
            << " at runtime of " << timer.timeElapsedInSec() << " seconds." << Log::end;
         throw;
      }
      timer.stop();
      Log::info("rivet-analyze") << "Finished event loop after " << timer.timeElapsedInSec() << " seconds." << Log::end;
      Log::info() << "Cross section = " << rivet.crossSection() << " pb" << Log::end;
      rivet.finalize(RivetInterface::getFilename(infile, weight, outpath, rivet_tag));
      Log::warning().writeCounts();
      Log::warning().resetCounts();
   } catch (...) {
      Log::warning().writeCounts();
      Log::warning().resetCounts();
      return mainExceptionHandler();
   }
   return EXIT_SUCCESS;
}

#endif // geneva_USE_RIVET

//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
void GenevaWrapper::writeReport(GenerationReport& report, const std::string& filename,
                                double runtime, size_t numEvents, IAnalyzer* analyzer)
{
   // write main report to file
   report.setRunTimeInSec(runtime);
   report.setNumEventsGenerated(numEvents);
   if (analyzer) {
      report.clearOutput();
      report.outputStream() << OutputStreamAnalyzer(analyzer, report.title());
   }

   std::ofstream file(filename.c_str());
   file << report;
   file.close();
   if (!file)
      Log::warning("writeReport") << "Could not write generation report to file: " << filename << Log::end;
   else
      Log::info("writeReport") << "Wrote generation report after " << numEvents
                               << " events to file: " << filename << Log::end;
}

} // namespace Geneva
