//------------------------------------------------------------------------------
/// \file compare-options.cpp
//
// Author(s):
//    Simone Alioli, Frank Tackmann
//
// Copyright:
//    Copyright (C) 2017 CERN, DESY
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Program to compare option files
//------------------------------------------------------------------------------

#include "Geneva/Driver/Geneva.hpp"

#include "Geneva/Core/UI/Log.hpp"
#include "Geneva/Core/UI/OptionMap.hpp"

#include <sstream>
#include <iomanip>

using namespace Geneva;

//------------------------------------------------------------------------------
int main(int argc, const char* argv[])
{
   try {
      if (argc != 3) {
         Log::info() << "usage: geneva-compare-options <file1> <file2>\n\n";
         Log::info() << "arguments:\n";
         Log::info() << "  <file1>   First (old) option file to compare\n";
         Log::info() << "  <file2>   Second (new) option file to compare\n\n";
         Log::info() << "Reads each option file and constructs a complete set of options\n";
         Log::info() << "as would be seen when calling 'geneva-main --optionsFile <file>'\n";
         Log::info() << "and prints out the differences between both sets of options.\n\n";
         Log::info() << "Unknown options, including deprecated no longer known ones,\n";
         Log::info() << "are warned about and are otherwise ignored in the comparison.\n";
         return EXIT_SUCCESS;
      }

      Log::setVerbosity(Log::kWarning);
      Log::warning().setMaxCounts(static_cast<size_t>(-1));

      ::Geneva::Geneva geneva1("--optionsFile " + std::string(argv[1]));
      ::Geneva::Geneva geneva2("--optionsFile " + std::string(argv[2]));

      const OptionMap& options1 = geneva1.optionMap();
      const OptionMap& options2 = geneva2.optionMap();

      std::stringstream diff;
      bool foundDiff = false;
      for (OptionMap::const_iterator it = options1.begin(); it != options1.end(); ++it) {
         try {
            const std::string val2 = options2.getOption(it->first);
            if (val2 != it->second->getValueStr()) {
               foundDiff = true;
               diff << std::left << std::setw(30) << it->first
                    << TextColor::kRed   << std::right << std::setw(25) << it->second->getValueStr()
                    << TextColor::kGreen << "  " << val2
                    << TextColor::kReset << '\n';
            }
         } catch (NormalError&) {
            foundDiff = true;
            diff << std::left << std::setw(30) << it->first
                 << TextColor::kRed   << std::right << std::setw(25) << it->second->getValueStr()
                 << TextColor::kBoldMagenta << "  <unknown>"
                 << TextColor::kReset << " (something is wrong)\n";
         }
      }
      Log::setVerbosity(Log::kInfo);
      if (foundDiff) {
         Log::info() << "\nDifferent options between " << TextColor::kRed << argv[1] << TextColor::kReset << " and " << TextColor::kGreen << argv[2]
                     << TextColor::kReset << "\n\n";
         Log::info() << diff.str() << Log::end;
      } else {
         Log::info() << "\nNo differences found between the two sets of options.\n\n";
      }
   } catch (...) {
      return mainExceptionHandler();
   }
   return EXIT_SUCCESS;
}
