//------------------------------------------------------------------------------
/// \file SignatureSet.cpp
//
// Author(s):
//    Frank Tackmann
//
// Copyright:
//    Copyright (C) 2012 MIT, Frank Tackmann
//
//    This file is part of the Geneva MC framework. Geneva is distributed under
//    the terms of the GNU General Public License version 3 (GPLv3), see the
//    COPYING file that comes with this distribution for details.
//    Please respect the academic usage guidelines in the GUIDELINES file.
//
// Description:
//    Implementation of class SignatureSet
//------------------------------------------------------------------------------

#include "Geneva/Core/Physics/SignatureSet.hpp"

#include <algorithm>
#include <iterator>
#include <ostream>

namespace Geneva
{

//------------------------------------------------------------------------------
// Accessing and Querying Signatures
//------------------------------------------------------------------------------
/**
 * \brief Tests if a %SignatureSet is contained in the %SignatureSet.
 * \param other The %SignatureSet to be searched for.
 * \return \c true if all Signatures in \a other are contained in the %SignatureSet,
 *         \c false otherwise.
 *
 * If \c true is returned, \a other is a subset of the %SignatureSet, i.e. their
 * intersection will be equal to \a other. Requires linear time.
 */
bool SignatureSet::includes(const SignatureSet& other) const
{
   return std::includes(begin(), end(), other.begin(), other.end());
}

//------------------------------------------------------------------------------
/**
 * \brief Tests if a %SignatureSet intersects with the %SignatureSet.
 * \param other The %SignatureSet to be tested.
 * \return \c true if at least one Signature in \a other is also contained in
 *         the %SignatureSet, \c false otherwise.
 *
 * If \c true is returned, the %SignatureSet has a non-empty intersection with
 * \a other. In this case this test is faster than computing the intersection
 * (because the algorithm can stop at the first common Signature). For \c false,
 * it takes exactly as long as computing the (empty) intersection (because all
 * Signatures have to be checked in this case). Hence, if the intersection is
 * later needed, it is overall faster to directly compute it and check for its
 * emptiness.
 */
bool SignatureSet::intersects(const SignatureSet& other) const
{
   iterator i1 = begin();
   iterator i2 = other.begin();
   const iterator end1 = end();
   const iterator end2 = other.end();

   // Since we only test for intersection we can stop as soon as we find one
   // common element.
   while (i1 != end1 && i2 != end2) {
      if (*i1 < *i2)
         ++i1;
      else if (*i2 < *i1)
         ++i2;
      else
         return true;
   }
   return false;
}

//------------------------------------------------------------------------------
// Algebra
//------------------------------------------------------------------------------
/**
 * \brief %SignatureSet subtraction (complement) assignment operator
 * \param rhs The %SignatureSet to take the complement with.
 * \return A reference to the %SignatureSet.
 *
 * Takes the complement of the %SignatureSet with \a rhs. As a result the
 * %SignatureSet will only contain Signatures that are not contained in \a rhs.
 * Signatures in \a rhs that are not part of the %SignatureSet have no effect
 * (apart from maybe slowing down the operation). Requires linear time.
 */
SignatureSet& SignatureSet::operator-=(const SignatureSet& rhs)
{
   iterator i1 = begin();
   iterator i2 = rhs.begin();
   const iterator end1 = end();
   const iterator end2 = rhs.end();

   // We exploit here that set iterators stay valid if elements are erased,
   // but it's important to use erase(i1++), which advances i1 and returns its
   // old value, which is then erased.
   while (i1 != end1 && i2 != end2) {
      if (*i1 < *i2)
         ++i1;
      else if (*i2 < *i1)
         ++i2;
      else {
         erase(i1++);
         ++i2;
      }
   }
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %SignatureSet AND (intersection) assignment operator
 * \param rhs The %SignatureSet to take the intersection with.
 * \return A reference to the %SignatureSet.
 *
 * Takes the intersection of the %SignatureSet with \a rhs. It only keeps
 * Signatures in the %SignatureSet that are also part of \a rhs, removing all
 * others. Signatures in \a rhs that are not part of the %SignatureSet have no
 * effect (apart from maybe slowing down the operation). Requires linear time.
 */
SignatureSet& SignatureSet::operator&=(const SignatureSet& rhs)
{
   iterator i1 = begin();
   iterator i2 = rhs.begin();
   const iterator end1 = end();
   const iterator end2 = rhs.end();

   // We exploit here that set iterators stay valid if elements are erased,
   // but it's important to use erase(i1++), which advances i1 and returns its
   // old value, which is then erased.
   while (i1 != end1 && i2 != end2) {
      if (*i1 < *i2)
         erase(i1++);
      else if (*i2 < *i1)
         ++i2;
      else {
         ++i1;
         ++i2;
      }
   }
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %SignatureSet XOR (symmetric difference) assignment operator
 * \param rhs The %SignatureSet to take the symmetric difference with.
 * \return A reference to the %SignatureSet.
 *
 * Takes the symmetric difference of the %SignatureSet with \a rhs, which
 * corresponds to taking the union and removing the intersection. As a result
 * the %SignatureSet will contain all Signatures that are either part of itself
 * or part of \a rhs, but not those common to both. Requires linear time.
 */
SignatureSet& SignatureSet::operator^=(const SignatureSet& rhs)
{
   iterator i1 = begin();
   iterator i2 = rhs.begin();
   const iterator end1 = end();
   const iterator end2 = rhs.end();

   // We exploit here that set iterators stay valid if elements are erased,
   // but it's important to use erase(i1++), which advances i1 and returns its
   // old value, which is then erased.
   while (i1 != end1 && i2 != end2) {
      if (*i1 < *i2)
         ++i1;
      else if (*i2 < *i1) {
         insert(i1, *i2);
         ++i2;
      } else {
         erase(i1++);
         ++i2;
      }
   }
   insert(i2, end2);
   return *this;
}

//------------------------------------------------------------------------------
/**
 * \brief %SignatureSet subtraction (complement) operator
 * \param lhs The %SignatureSet to take the complement of.
 * \param rhs The %SignatureSet to take the complement with.
 * \return The resulting %SignatureSet.
 *
 * Takes the complement of \a lhs with \a rhs and returns the result in a new
 * %SignatureSet. See operator-= for how the complement is taken. Requires
 * linear time.
 */
const SignatureSet operator-(const SignatureSet& lhs, const SignatureSet& rhs)
{
   SignatureSet tmp;
   std::set_difference(lhs.begin(), lhs.end(), rhs.begin(), rhs.end(), std::inserter(tmp, tmp.end()));
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Signature AND (intersection) operator
 * \param lhs The 1st %SignatureSet to take the intersection with.
 * \param rhs The 2nd %SignatureSet to take the intersection with.
 * \return The resulting %SignatureSet.
 *
 * Takes the intersection of \a lhs and \a rhs and returns the result in a new
 * %SignatureSet. See operator&= for how the intersection is taken. Requires
 * linear time.
 */
const SignatureSet operator&(const SignatureSet& lhs, const SignatureSet& rhs)
{
   SignatureSet tmp;
   std::set_intersection(lhs.begin(), lhs.end(), rhs.begin(), rhs.end(), std::inserter(tmp, tmp.end()));
   return tmp;
}

//------------------------------------------------------------------------------
/**
 * \brief %Signature XOR (symmetric difference) operator
 * \param lhs The 1st %SignatureSet to take the symmetric difference with.
 * \param rhs The 2nd %SignatureSet to take the symmetric difference with.
 * \return The resulting %SignatureSet.
 *
 * Takes the symmetric difference \a lhs and \a rhs and returns the result in a
 * new %SignatureSet. The symmetric difference is defined as
 * lhs ^ rhs == (lhs + rhs) - (lhs & rhs) == (lhs - rhs) + (rhs - lhs).
 */
const SignatureSet operator^(const SignatureSet& lhs, const SignatureSet& rhs)
{
   SignatureSet tmp;
   std::set_symmetric_difference(lhs.begin(), lhs.end(), rhs.begin(), rhs.end(), std::inserter(tmp, tmp.end()));
   return tmp;
}


//------------------------------------------------------------------------------
// Private member functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Streams the SignatureSet to output.
std::ostream& SignatureSet::print(std::ostream& out) const
{
   std::copy(begin(), end(), std::ostream_iterator<Signature>(out, "\n"));
   return out;
}

} // namespace Geneva
